Package: dm
Title: Relational Data Models
Version: 1.1.2
Date: 2026-05-17
Authors@R: 
    c(person(given = "Tobias",
             family = "Schieferdecker",
             role = "aut",
             email = "tobias@cynkra.com"),
      person(given = "Kirill",
             family = "Müller",
             role = c("aut", "cre"),
             email = "kirill@cynkra.com",
             comment = c(ORCID = "0000-0002-1416-3412")),
      person(given = "Antoine",
             family = "Fabri",
             role = "ctb"),
      person(given = "Darko",
             family = "Bergant",
             role = "aut",
             email = "darko.bergant@gmail.com"),
      person(given = "Katharina",
             family = "Brunner",
             role = "ctb"),
      person(given = "James",
             family = "Wondrasek",
             role = "ctb"),
      person(given = "Indrajeet",
             family = "Patil",
             comment = c(ORCID = "0000-0003-1995-6531"),
             role = "ctb"),
      person(given = "Maëlle",
             family = "Salmon",
             comment = c(ORCID = "0000-0002-2815-0399"),
             role = "ctb"),
      person(given = "energie360° AG",
             role = "fnd",
             email = "info@energie360.ch"),
      person(given = "cynkra GmbH",
             role = c("fnd", "cph"),
             email = "mail@cynkra.com",
             comment = c(ROR = "0335t7e62"))
    )
Description: Provides tools for working with multiple related
    tables, stored as data frames or in a relational database.  Multiple
    tables (data and metadata) are stored in a compound object, which can
    then be manipulated with a pipe-friendly syntax.
License: MIT + file LICENSE
URL: https://dm.cynkra.com/, https://github.com/cynkra/dm
BugReports: https://github.com/cynkra/dm/issues
Depends: R (>= 4.0)
Imports: backports, cli (>= 2.2.0), dplyr (>= 1.2.0), glue, lifecycle
        (>= 1.0.3), memoise, methods, purrr (>= 1.0.0), rlang (>=
        1.0.2), tibble (>= 3.0.0), tidyr (>= 1.0.0), tidyselect (>=
        1.2.0), vctrs (>= 0.3.2)
Suggests: brio, colourpicker, covr, DBI (>= 1.2.0), dbplyr (>= 2.3.4),
        DiagrammeR, DiagrammeRsvg, digest, duckdb (>= 0.4.0), duckplyr,
        fansi, forcats, htmltools, htmlwidgets, igraph (>= 2.2.0),
        jsonlite, keyring, knitr, labelled (>= 2.12.0), magrittr,
        nycflights13, odbc (>= 1.4.2), pillar, pixarfilms, pool,
        progress, reactable, RMariaDB (>= 1.3.3), rmarkdown, RPostgres,
        RSQLite (>= 2.2.8), rstudioapi, shiny, shinyAce,
        shinydashboard, testthat (>= 3.2.0), tidyverse, waldo, withr
Config/Needs/website: brio, bslib, cynkra/cynkratemplate, htmltools,
        pagedown, purrr, rmarkdown, whisker, xml2
Config/Needs/check: anthonynorth/roxyglobals
VignetteBuilder: knitr
Config/autostyle/scope: line_breaks
Config/autostyle/strict: true
Config/testthat/edition: 3
Config/testthat/parallel: false
Config/testthat/start-first: zzx-deprecated, flatten, dplyr, filter-dm,
        draw-dm, bind, rows-dm, learn
Encoding: UTF-8
Config/roxygen2/version: 8.0.0.9000
NeedsCompilation: no
Packaged: 2026-05-17 04:40:51 UTC; kirill
Author: Tobias Schieferdecker [aut],
  Kirill Müller [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1416-3412>),
  Antoine Fabri [ctb],
  Darko Bergant [aut],
  Katharina Brunner [ctb],
  James Wondrasek [ctb],
  Indrajeet Patil [ctb] (ORCID: <https://orcid.org/0000-0003-1995-6531>),
  Maëlle Salmon [ctb] (ORCID: <https://orcid.org/0000-0002-2815-0399>),
  energie360° AG [fnd],
  cynkra GmbH [fnd, cph] (ROR: <https://ror.org/0335t7e62>)
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2026-05-17 07:30:02 UTC
Built: R 4.4.3; ; 2026-05-17 08:11:21 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$",dm)
S3method("$",dm_zoomed)
S3method("$<-",dm)
S3method("[",dm)
S3method("[",dm_zoomed)
S3method("[<-",dm)
S3method("[[",dm)
S3method("[[",dm_zoomed)
S3method("[[<-",dm)
S3method("length<-",dm)
S3method("names<-",dm)
S3method(anti_join,dm)
S3method(anti_join,dm_keyed_tbl)
S3method(anti_join,dm_zoomed)
S3method(arrange,dm)
S3method(arrange,dm_keyed_tbl)
S3method(arrange,dm_zoomed)
S3method(as.list,dm)
S3method(as.list,dm_zoomed)
S3method(as_dm,DBIConnection)
S3method(as_dm,default)
S3method(as_dm,src)
S3method(collect,dm)
S3method(collect,dm_zoomed)
S3method(compute,dm)
S3method(compute,dm_zoomed)
S3method(copy_to,dm)
S3method(count,dm)
S3method(count,dm_keyed_tbl)
S3method(count,dm_zoomed)
S3method(cross_join,dm)
S3method(cross_join,dm_keyed_tbl)
S3method(cross_join,dm_zoomed)
S3method(db_schema_create,"Microsoft SQL Server")
S3method(db_schema_create,PqConnection)
S3method(db_schema_create,SQLiteConnection)
S3method(db_schema_create,src_dbi)
S3method(db_schema_drop,"Microsoft SQL Server")
S3method(db_schema_drop,PqConnection)
S3method(db_schema_drop,SQLiteConnection)
S3method(db_schema_drop,src_dbi)
S3method(db_schema_exists,"Microsoft SQL Server")
S3method(db_schema_exists,PqConnection)
S3method(db_schema_exists,SQLiteConnection)
S3method(db_schema_exists,src_dbi)
S3method(db_schema_list,"Microsoft SQL Server")
S3method(db_schema_list,Pool)
S3method(db_schema_list,PqConnection)
S3method(db_schema_list,SQLiteConnection)
S3method(db_schema_list,src_dbi)
S3method(dim,dm_zoomed)
S3method(dimnames,dm_zoomed)
S3method(distinct,dm)
S3method(distinct,dm_keyed_tbl)
S3method(distinct,dm_zoomed)
S3method(filter,dm)
S3method(filter,dm_keyed_tbl)
S3method(filter,dm_zoomed)
S3method(filter_out,dm)
S3method(filter_out,dm_keyed_tbl)
S3method(filter_out,dm_zoomed)
S3method(format,dm)
S3method(format,dm_cg_code_block)
S3method(format,dm_keys)
S3method(format,dm_zoomed)
S3method(format,dm_zoomed_df)
S3method(full_join,dm)
S3method(full_join,dm_keyed_tbl)
S3method(full_join,dm_zoomed)
S3method(glimpse,dm)
S3method(glimpse,dm_zoomed)
S3method(group_by,dm)
S3method(group_by,dm_keyed_tbl)
S3method(group_by,dm_zoomed)
S3method(group_data,dm)
S3method(group_data,dm_zoomed)
S3method(group_indices,dm)
S3method(group_indices,dm_zoomed)
S3method(group_keys,dm)
S3method(group_keys,dm_zoomed)
S3method(group_vars,dm)
S3method(group_vars,dm_zoomed)
S3method(groups,dm)
S3method(groups,dm_zoomed)
S3method(head,dm_zoomed)
S3method(inner_join,dm)
S3method(inner_join,dm_keyed_tbl)
S3method(inner_join,dm_zoomed)
S3method(json_nest,"tbl_Microsoft SQL Server")
S3method(json_nest,data.frame)
S3method(json_nest,tbl_lazy)
S3method(json_nest_aggregate,default)
S3method(json_nest_join,data.frame)
S3method(json_pack,"tbl_Microsoft SQL Server")
S3method(json_pack,data.frame)
S3method(json_pack,tbl_lazy)
S3method(json_pack_join,data.frame)
S3method(json_unnest,data.frame)
S3method(json_unpack,data.frame)
S3method(left_join,dm)
S3method(left_join,dm_keyed_tbl)
S3method(left_join,dm_zoomed)
S3method(length,dm)
S3method(length,dm_zoomed)
S3method(mutate,dm)
S3method(mutate,dm_keyed_tbl)
S3method(mutate,dm_zoomed)
S3method(names,dm)
S3method(names,dm_zoomed)
S3method(nest_join,dm)
S3method(nest_join,dm_zoomed)
S3method(pack_join,data.frame)
S3method(pack_join,dm)
S3method(pack_join,dm_zoomed)
S3method(print,dm)
S3method(print,dm_cg_code_block)
S3method(print,dm_examine_cardinalities)
S3method(print,dm_examine_constraints)
S3method(print,dm_zoomed)
S3method(pull,dm)
S3method(pull,dm_zoomed)
S3method(pull_tbl,dm)
S3method(pull_tbl,dm_zoomed)
S3method(reframe,dm)
S3method(reframe,dm_keyed_tbl)
S3method(reframe,dm_zoomed)
S3method(relocate,dm)
S3method(relocate,dm_keyed_tbl)
S3method(relocate,dm_zoomed)
S3method(rename,dm)
S3method(rename,dm_keyed_tbl)
S3method(rename,dm_zoomed)
S3method(right_join,dm)
S3method(right_join,dm_keyed_tbl)
S3method(right_join,dm_zoomed)
S3method(rows_truncate,data.frame)
S3method(rows_truncate,tbl_sql)
S3method(select,dm)
S3method(select,dm_keyed_tbl)
S3method(select,dm_zoomed)
S3method(semi_join,dm)
S3method(semi_join,dm_keyed_tbl)
S3method(semi_join,dm_zoomed)
S3method(separate,dm)
S3method(separate,dm_keyed_tbl)
S3method(separate,dm_zoomed)
S3method(slice,dm)
S3method(slice,dm_keyed_tbl)
S3method(slice,dm_zoomed)
S3method(sql_json_nest,PqConnection)
S3method(sql_json_pack,PqConnection)
S3method(sql_rows_truncate,tbl_SQLiteConnection)
S3method(sql_rows_truncate,tbl_duckdb_connection)
S3method(sql_rows_truncate,tbl_sql)
S3method(src_tbls,dm)
S3method(str,dm)
S3method(str,dm_zoomed)
S3method(summarise,dm)
S3method(summarise,dm_keyed_tbl)
S3method(summarise,dm_zoomed)
S3method(tail,dm_zoomed)
S3method(tally,dm)
S3method(tally,dm_keyed_tbl)
S3method(tally,dm_zoomed)
S3method(tbl,dm)
S3method(tbl_sum,dm_keyed_tbl)
S3method(tbl_sum,dm_zoomed_df)
S3method(tbl_vars,dm)
S3method(tbl_vars,dm_zoomed)
S3method(transmute,dm)
S3method(transmute,dm_keyed_tbl)
S3method(transmute,dm_zoomed)
S3method(ungroup,dm)
S3method(ungroup,dm_keyed_tbl)
S3method(ungroup,dm_zoomed)
S3method(unite,dm)
S3method(unite,dm_keyed_tbl)
S3method(unite,dm_zoomed)
S3method(vec_cast,dm_keys.dm_keys)
S3method(vec_proxy_compare,dm_keys)
S3method(vec_ptype2,dm_keys.dm_keys)
S3method(vec_ptype_abbr,dm_keys)
export("%>%")
export(anti_join)
export(arrange)
export(as_dm)
export(cdm_add_fk)
export(cdm_add_pk)
export(cdm_add_tbl)
export(cdm_apply_filters)
export(cdm_apply_filters_to_tbl)
export(cdm_check_constraints)
export(cdm_copy_to)
export(cdm_disambiguate_cols)
export(cdm_draw)
export(cdm_enum_fk_candidates)
export(cdm_enum_pk_candidates)
export(cdm_filter)
export(cdm_flatten_to_tbl)
export(cdm_get_all_fks)
export(cdm_get_all_pks)
export(cdm_get_available_colors)
export(cdm_get_colors)
export(cdm_get_con)
export(cdm_get_filter)
export(cdm_get_fk)
export(cdm_get_pk)
export(cdm_get_referencing_tables)
export(cdm_get_src)
export(cdm_get_tables)
export(cdm_has_fk)
export(cdm_has_pk)
export(cdm_insert_zoomed_tbl)
export(cdm_is_referenced)
export(cdm_join_to_tbl)
export(cdm_learn_from_db)
export(cdm_nrow)
export(cdm_nycflights13)
export(cdm_paste)
export(cdm_rename)
export(cdm_rename_tbl)
export(cdm_rm_fk)
export(cdm_rm_pk)
export(cdm_rm_tbl)
export(cdm_select)
export(cdm_select_tbl)
export(cdm_set_colors)
export(cdm_squash_to_tbl)
export(cdm_update_zoomed_tbl)
export(cdm_zoom_out)
export(cdm_zoom_to_tbl)
export(check_cardinality)
export(check_cardinality_0_1)
export(check_cardinality_0_n)
export(check_cardinality_1_1)
export(check_cardinality_1_n)
export(check_if_subset)
export(check_key)
export(check_set_equality)
export(check_subset)
export(collect)
export(compute)
export(copy_dm_to)
export(copy_to)
export(db_schema_create)
export(db_schema_drop)
export(db_schema_exists)
export(db_schema_list)
export(decompose_table)
export(dm)
export(dm_add_fk)
export(dm_add_pk)
export(dm_add_tbl)
export(dm_add_uk)
export(dm_apply_filters)
export(dm_apply_filters_to_tbl)
export(dm_bind)
export(dm_ddl_post)
export(dm_ddl_pre)
export(dm_deconstruct)
export(dm_disambiguate_cols)
export(dm_discard_zoomed)
export(dm_dml_load)
export(dm_draw)
export(dm_enum_fk_candidates)
export(dm_enum_pk_candidates)
export(dm_examine_cardinalities)
export(dm_examine_constraints)
export(dm_filter)
export(dm_financial)
export(dm_financial_sqlite)
export(dm_flatten)
export(dm_flatten_to_tbl)
export(dm_from_con)
export(dm_from_src)
export(dm_get_all_fks)
export(dm_get_all_pks)
export(dm_get_all_uks)
export(dm_get_available_colors)
export(dm_get_colors)
export(dm_get_con)
export(dm_get_filters)
export(dm_get_fk)
export(dm_get_pk)
export(dm_get_referencing_tables)
export(dm_get_src)
export(dm_get_table_description)
export(dm_get_tables)
export(dm_gui)
export(dm_has_fk)
export(dm_has_pk)
export(dm_insert_zoomed)
export(dm_is_referenced)
export(dm_join_to_tbl)
export(dm_mutate_tbl)
export(dm_nest_tbl)
export(dm_nrow)
export(dm_nycflights13)
export(dm_pack_tbl)
export(dm_paste)
export(dm_pixarfilms)
export(dm_ptype)
export(dm_rename)
export(dm_rename_tbl)
export(dm_reset_table_description)
export(dm_rm_fk)
export(dm_rm_pk)
export(dm_rm_tbl)
export(dm_rm_uk)
export(dm_rows_append)
export(dm_rows_delete)
export(dm_rows_insert)
export(dm_rows_patch)
export(dm_rows_truncate)
export(dm_rows_update)
export(dm_rows_upsert)
export(dm_select)
export(dm_select_tbl)
export(dm_set_colors)
export(dm_set_table_description)
export(dm_sql)
export(dm_squash_to_tbl)
export(dm_unnest_tbl)
export(dm_unpack_tbl)
export(dm_unwrap_tbl)
export(dm_update_zoomed)
export(dm_validate)
export(dm_wrap_tbl)
export(dm_zoom_to)
export(enum_fk_candidates)
export(enum_pk_candidates)
export(examine_cardinality)
export(filter)
export(full_join)
export(glimpse)
export(group_by)
export(inner_join)
export(is_dm)
export(json_nest)
export(json_nest_join)
export(json_pack)
export(json_pack_join)
export(json_unnest)
export(json_unpack)
export(left_join)
export(mutate)
export(new_dm)
export(pack_join)
export(pull_tbl)
export(rename)
export(reunite_parent_child)
export(reunite_parent_child_from_list)
export(right_join)
export(rows_append)
export(rows_delete)
export(rows_insert)
export(rows_patch)
export(rows_truncate)
export(rows_update)
export(rows_upsert)
export(select)
export(semi_join)
export(separate)
export(sql_rows_truncate)
export(sql_schema_create)
export(sql_schema_drop)
export(sql_schema_exists)
export(sql_schema_list)
export(src_tbls)
export(summarise)
export(summarize)
export(tbl)
export(tibble)
export(transmute)
export(ungroup)
export(unite)
export(validate_dm)
import(cli)
import(dplyr, except = c(data_frame))
import(glue)
import(lifecycle)
import(purrr, except = c(list_along, rep_along, invoke, modify, as_function, flatten_dbl, flatten_lgl, flatten_int, flatten_raw, flatten_chr, splice, flatten, prepend, `%@%`, `%||%`))
import(rlang)
import(tibble, except = c(data_frame))
import(tidyr, except = c(extract))
import(vctrs)
importFrom(dplyr,"%>%")
importFrom(dplyr,anti_join)
importFrom(dplyr,arrange)
importFrom(dplyr,collect)
importFrom(dplyr,compute)
importFrom(dplyr,copy_to)
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,inner_join)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,rename)
importFrom(dplyr,right_join)
importFrom(dplyr,rows_append)
importFrom(dplyr,rows_delete)
importFrom(dplyr,rows_insert)
importFrom(dplyr,rows_patch)
importFrom(dplyr,rows_update)
importFrom(dplyr,rows_upsert)
importFrom(dplyr,select)
importFrom(dplyr,semi_join)
importFrom(dplyr,src_tbls)
importFrom(dplyr,summarise)
importFrom(dplyr,summarize)
importFrom(dplyr,tbl)
importFrom(dplyr,transmute)
importFrom(dplyr,ungroup)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colors)
importFrom(grDevices,rgb)
importFrom(methods,is)
importFrom(tibble,glimpse)
importFrom(tibble,tibble)
importFrom(tidyr,separate)
importFrom(tidyr,unite)
importFrom(utils,head)
importFrom(utils,packageVersion)
importFrom(utils,str)
importFrom(utils,tail)

How to install Accelar/450 Management Software
----------------------------------------------

This file contains instructions to install the Device 
Management(DM) and VLAN Management(VM) software for AIX, 
HPUX and Solaris UNIX platforms.

The standard installation procedure involves copying the
compressed tar file containing the DM/VM software from 
the cdrom onto the user's hard-disk.  The file is then
uncompressed and untared.  Lastly, the user must manually 
add the "DMPATH" environment variable to their shell file.
(See the section "Standard Installation Procedure" below).

Optionally, if the user is running HP OpenView(without
Optivity) and would like to be able to launch the DM/VM 
software directly from the HPOV windows, the user can 
execute one additional script to install the DM/VM software
into HPOV. (See the section "Installing DM/VM into HPOV" 
below).



Standard Installation Procedure 
-------------------------------

Prerequisites

- AIX workstation with AIX 4.1 or higher
- HP workstation with HPUX 10.20 or higher
- SPARC workstaion with Solaris 2.4 or higher 
- 32 MB DRAM
- 3MB space on hard disk


Step #1 :

Copy the compressed tar file containing the DM/VM software 
to the host's hard-disk.  The instructions that follow assumes 
the user has mounted the cdrom drive onto the "/cdrom" directory.
If you do not know how to mount the cdrom drive, refer to your 
system's operation system user guide. 

Execute the following commands : 

a. Create a directory on the host's hard-disk where you
   want the DM/VM software to reside.  The location
   can be anywhere the user chooses.  Typically, 3rd party
   software is installed under the "/opt" directory.
   Depending on how your system is configured, you may need
   to be the root user to have write permission to the "/opt"
   directory.  If you do not have root access, you can install
   the software under "/usr/local/Bay".

   % mkdir /opt/Bay

b. Goto the directory on the cdrom where the DM/VM software is 
   located:

   % cd /cdrom/dm-unix

c. Execute the "install_dmvm" script by providing the version 
   of the software you want to install and target directory where 
   you have decided to install the DM/VM software: 

   % install_dmvm dm_v200 /opt/Bay


   If you encounter problems executing this script, refer
   to the "Trouble Shooting" section at the end of this file.


Step #2 :

Add the "DMPATH" environment variable to your shell file.  
If you are using c-shell, add a line like this to your .cshrc 
file:

setenv DMPATH /opt/Bay/dm_v[version_number]

where [version_number] is the version of the DM/VM software you 
just installed.  For instance, if you are installing version 
1.3.0 of the DM/VM software, the line you need to add would look 
like this:

setenv DMPATH /opt/Bay/dm_v200



Installing DM/VM into HPOV (Optional)
-------------------------------------

Prerequisites

- HP OpenView must be installed.
- Optivity should not be installed.  If you're running Optivity
  you should follow the Optivity installation procedures to
  upgrade your Optivity software with the newer Management
  Software. (Using the instructions that follow will cause Optivity
  to have uninstall problems).
- Device Management software must be installed.
- You must be logged on as root.


To integrate Device Management software:

a. Go to the directory where the Device Management software is 
   installed.

   % cd /opt/Bay/dm_v200

b. As the root user start Device Manager and point it to an 
   Accelar or 450 device (e.g., "dm a.b.c.d", where a.b.c.d is the 
   IP address of the Accelar/450 device or the name of the device 
   if your network is running some name service).

   % dm 132.32.1.5

c. From Device Manager, click on Help from menu bar and select 
   the "Run HP OpenView Install on Console" option.

   This will copy the Accelar/450 specific icons and MIB files to the 
   standard HP Openview software directories.  When this is completed,
   a list of commands will be shown on the console.  Manually execute 
   the commands as root to complete the installation process.

   Here's a brief description of each command:

   ------------------------------------------------------------
   Command 		| 	What it does ...	
   ------------------------------------------------------------
   1. ovw -fields	|	Update the fields database.
   ------------------------------------------------------------
   2. ovstop netmon	|	Stop network monitor to allow the
			|	database to update.
   ------------------------------------------------------------
   3. ovtopofix		| 	Update topology database.	
   ------------------------------------------------------------
   4. ovstart netmon	|	Restart network monitor.
   ------------------------------------------------------------
   5. ovw		|	Start HP OpenView.
   ------------------------------------------------------------

d. After HP Overview is started, load the Accelar/450 specific mibs by
   selecting the "Options->Load MIBs" option.  Load the "rapid_city.mib"
   file located in the "/var/opt/OV/share/snmp_mibs/Vendor" directory.



Trouble Shooting
----------------

o Why am I getting the "permission denied" message when executing the
  "install_dmvm" script?

  Make sure the "install_dmvm" file has executible permission.
  You make need to execute this command:

    % chmod a+x install_dmvm
  

o Why is the "install_dmvm" script failing?

  If you encounter problems executing the script, you can work-around
  the problem by doing the install manually.  Execute the following
  commands:  

     % mkdir /opt/Bay
     % cd /cdrom/dm-unix
     % cp dm_v200.tar.Z  /opt/Bay
     % cd /opt/Bay
     % uncompress dm_v200.tar.Z
     % tar -xv dm_v200.tar
     % cd dm_v200
     % standalone_setup

  Finish-up the installation procedure by performing Step#2 described
  in section "Standard Installation Procedure".


o When I start-up the Device Manager(DM) program, I'm getting this 
  message: "DMPATH variable not set". 
 
  Set the DMPATH environment variable in your shell file by performing
  Step #2 described in section "Standard Installation Procedure".



Revision : 09/02/98 [ET] - initial version
Revision : 11/24/98 [ET] - updated for rel1.3.3
Revision : 12/15/98 [ET] - updated for rel2.0.0

<!-- README.md is generated from README.Rmd. Please edit that file -->

# [dm](https://dm.cynkra.com/)

<!-- badges: start -->

[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R build
status](https://github.com/cynkra/dm/workflows/rcc/badge.svg)](https://github.com/cynkra/dm/actions)
[![Codecov test
coverage](https://codecov.io/gh/cynkra/dm/branch/main/graph/badge.svg)](https://app.codecov.io/gh/cynkra/dm?branch=main)
[![CRAN
status](https://www.r-pkg.org/badges/version/dm)](https://CRAN.R-project.org/package=dm)
[![Launch
rstudio.cloud](https://img.shields.io/badge/rstudio-cloud-blue.svg)](https://rstudio.cloud/project/523482)
<!-- badges: end -->

> Are you using multiple data frames or database tables in R? Organize
> them with dm.
>
> - Use it for data analysis today.
> - Build data models tomorrow.
> - Deploy the data models to your organization’s Relational Database
>   Management System (RDBMS) the day after.

## Overview

dm bridges the gap in the data pipeline between individual data frames
and relational databases. It’s a grammar of joined tables that provides
a consistent set of verbs for consuming, creating, and deploying
relational data models. For individual researchers, it broadens the
scope of datasets they can work with and how they work with them. For
organizations, it enables teams to quickly and efficiently create and
share large, complex datasets.

dm objects encapsulate relational data models constructed from local
data frames or lazy tables connected to an RDBMS. dm objects support the
full suite of dplyr data manipulation verbs along with additional
methods for constructing and verifying relational data models, including
key selection, key creation, and rigorous constraint checking. Once a
data model is complete, dm provides methods for deploying it to an
RDBMS. This allows it to scale from datasets that fit in memory to
databases with billions of rows.

## Features

dm makes it easy to bring an existing relational data model into your R
session. As the dm object behaves like a named list of tables it
requires little change to incorporate it within existing workflows. The
dm interface and behavior is modeled after dplyr, so you may already be
familiar with many of its verbs. dm also offers:

- visualization to help you understand relationships between entities
  represented by the tables
- simpler joins that “know” how tables are related, including a
  “flatten” operation that automatically follows keys and performs
  column name disambiguation
- consistency and constraint checks to help you understand (and fix) the
  limitations of your data

That’s just the tip of the iceberg. See [Getting
started](https://dm.cynkra.com/articles/dm.html) to hit the ground
running and explore all the features.

## Installation

The latest stable version of the {dm} package can be obtained from
[CRAN](https://CRAN.R-project.org/package=dm) with the command

``` r
install.packages("dm")
```

The latest development version of {dm} can be installed from R-universe:

``` r
# Enable repository from cynkra
options(
  repos = c(
    cynkra = "https://cynkra.r-universe.dev",
    CRAN = "https://cloud.r-project.org"
  )
)
# Download and install dm in R
install.packages('dm')
```

or from GitHub:

``` r
# install.packages("devtools")
devtools::install_github("cynkra/dm")
```

## Usage

Create a dm object (see [Getting
started](https://dm.cynkra.com/articles/dm.html) for details).

``` r
library(dm)
dm <- dm_nycflights13(table_description = TRUE)
dm
#> ── Metadata ────────────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 4
#> Foreign keys: 4
```

dm is a named list of tables:

``` r
names(dm)
#> [1] "airlines" "airports" "flights"  "planes"   "weather"
nrow(dm$airports)
#> [1] 86
dm$flights %>%
  count(origin)
#> # A tibble: 3 × 2
#>   origin     n
#>   <chr>  <int>
#> 1 EWR      641
#> 2 JFK      602
#> 3 LGA      518
```

Visualize relationships at any time:

``` r
dm %>%
  dm_draw()
```

<img src="man/figures/README-draw.svg" />

Simple joins:

``` r
dm %>%
  dm_flatten_to_tbl(flights)
#> Renaming ambiguous columns: %>%
#>   dm_rename(flights, year.flights = year) %>%
#>   dm_rename(flights, month.flights = month) %>%
#>   dm_rename(flights, day.flights = day) %>%
#>   dm_rename(flights, hour.flights = hour) %>%
#>   dm_rename(airlines, name.airlines = name) %>%
#>   dm_rename(airports, name.airports = name) %>%
#>   dm_rename(planes, year.planes = year) %>%
#>   dm_rename(weather, year.weather = year) %>%
#>   dm_rename(weather, month.weather = month) %>%
#>   dm_rename(weather, day.weather = day) %>%
#>   dm_rename(weather, hour.weather = hour)
#> # A tibble: 1,761 × 48
#>    year.flights month.…¹ day.f…² dep_t…³ sched…⁴ dep_d…⁵ arr_t…⁶ sched…⁷ arr_d…⁸
#>           <int>    <int>   <int>   <int>   <int>   <dbl>   <int>   <int>   <dbl>
#>  1         2013        1      10       3    2359       4     426     437     -11
#>  2         2013        1      10      16    2359      17     447     444       3
#>  3         2013        1      10     450     500     -10     634     648     -14
#>  4         2013        1      10     520     525      -5     813     820      -7
#>  5         2013        1      10     530     530       0     824     829      -5
#>  6         2013        1      10     531     540      -9     832     850     -18
#>  7         2013        1      10     535     540      -5    1015    1017      -2
#>  8         2013        1      10     546     600     -14     645     709     -24
#>  9         2013        1      10     549     600     -11     652     724     -32
#> 10         2013        1      10     550     600     -10     649     703     -14
#> # ℹ 1,751 more rows
#> # ℹ abbreviated names: ¹​month.flights, ²​day.flights, ³​dep_time,
#> #   ⁴​sched_dep_time, ⁵​dep_delay, ⁶​arr_time, ⁷​sched_arr_time, ⁸​arr_delay
#> # ℹ 39 more variables: carrier <chr>, flight <int>, tailnum <chr>,
#> #   origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>,
#> #   hour.flights <dbl>, minute <dbl>, time_hour <dttm>, name.airlines <chr>,
#> #   name.airports <chr>, lat <dbl>, lon <dbl>, alt <dbl>, tz <dbl>, dst <chr>,
#> #   tzone <chr>, year.planes <int>, type <chr>, manufacturer <chr>,
#> #   model <chr>, engines <int>, seats <int>, speed <int>, engine <chr>,
#> #   year.weather <int>, month.weather <int>, day.weather <int>,
#> #   hour.weather <int>, temp <dbl>, dewp <dbl>, humid <dbl>, wind_dir <dbl>,
#> #   wind_speed <dbl>, wind_gust <dbl>, precip <dbl>, pressure <dbl>, …
```

Check consistency:

``` r
dm %>%
  dm_examine_constraints()
#> ! Unsatisfied constraints:
#> • Table `flights`: foreign key `tailnum` into table `planes`: values of `flights$tailnum` not in `planes$tailnum`: N725MQ (6), N537MQ (5), N722MQ (5), N730MQ (5), N736MQ (5), …
```

Learn more in the [Getting
started](https://dm.cynkra.com/articles/dm.html) article.

## Getting help

If you encounter a clear bug, please file an issue with a minimal
reproducible example on [GitHub](https://github.com/cynkra/dm/issues).
For questions and other discussion, please use
[community.rstudio.com](https://forum.posit.co/).

------------------------------------------------------------------------

License: MIT © cynkra GmbH.

Funded by:

[![energie360°](man/figures/energie-72.png)](https://www.energie360.ch/de/)
<span style="padding-right:50px"> </span>
[![cynkra](man/figures/cynkra-72.png)](https://www.cynkra.com/)

------------------------------------------------------------------------

Please note that the ‘dm’ project is released with a [Contributor Code
of Conduct](https://dm.cynkra.com/CODE_OF_CONDUCT.html). By contributing
to this project, you agree to abide by its terms.

# Hammer.js 2.0 [![Build Status](https://travis-ci.org/hammerjs/hammer.js.svg)](https://travis-ci.org/hammerjs/hammer.js/)

Visit [hammerjs.github.io](http://hammerjs.github.io) for documentation.

You can get the pre-build versions from the Hammer.js website, or do this by yourself running 
`npm install -g grunt-cli && npm install && grunt build`

# jQuery

> jQuery is a fast, small, and feature-rich JavaScript library.

For information on how to get started and how to use jQuery, please see [jQuery's documentation](https://api.jquery.com/).
For source files and issues, please visit the [jQuery repo](https://github.com/jquery/jquery).

If upgrading, please see the [blog post for 3.6.0](https://blog.jquery.com/2021/03/02/jquery-3-6-0-released/). This includes notable differences from the previous version and a more readable changelog.

## Including jQuery

Below are some of the most common ways to include jQuery.

### Browser

#### Script tag

```html
<script src="https://code.jquery.com/jquery-3.6.0.min.js"></script>
```

#### Babel

[Babel](https://babeljs.io/) is a next generation JavaScript compiler. One of the features is the ability to use ES6/ES2015 modules now, even though browsers do not yet support this feature natively.

```js
import $ from "jquery";
```

#### Browserify/Webpack

There are several ways to use [Browserify](http://browserify.org/) and [Webpack](https://webpack.github.io/). For more information on using these tools, please refer to the corresponding project's documentation. In the script, including jQuery will usually look like this...

```js
var $ = require( "jquery" );
```

#### AMD (Asynchronous Module Definition)

AMD is a module format built for the browser. For more information, we recommend [require.js' documentation](https://requirejs.org/docs/whyamd.html).

```js
define( [ "jquery" ], function( $ ) {

} );
```

### Node

To include jQuery in [Node](https://nodejs.org/), first install with npm.

```sh
npm install jquery
```

For jQuery to work in Node, a window with a document is required. Since no such window exists natively in Node, one can be mocked by tools such as [jsdom](https://github.com/jsdom/jsdom). This can be useful for testing purposes.

```js
const { JSDOM } = require( "jsdom" );
const { window } = new JSDOM( "" );
const $ = require( "jquery" )( window );
```

# Demo for presentation at the satRday Neuch^atel,
# March 2020

# {dm}: Relational data models in R

options(tibble.print_min = 6)
options(tibble.print_max = 6)
options(rlang_backtrace_on_error = "none")
library(magrittr)

##
##
##
## RELATIONAL DATA MODELS WITH R
## -------------------------------------------------------
##
##
##
##

# Take home messages

## The dm package helps
## 1. **use** data models
## 2. **build** data models
## 3. **deploy** data models

### (on database or locally)

options(tibble.print_min = 6)
options(tibble.print_max = 6)
set.seed(20200314)


## ----connect, cache = FALSE------------------------------------
library(DBI) # <<

mydb <- dbConnect(
  RMariaDB::MariaDB(),
  user = "guest",
  password = "ctu-relational",
  dbname = "Financial_ijs",
  host = "relational.fel.cvut.cz"
)


## ----list-tables-----------------------------------------------
dbListTables(mydb)


## ----load-table, highlight.output = 1--------------------------
loans <- dbReadTable(mydb, "loans")

library(tidyverse)
as_tibble(loans)


## ----lazy-table, highlight.output = 2:3------------------------
library(dbplyr) # <<
loans <- tbl(mydb, "loans")
loans


## ----eval = FALSE----------------------------------------------
## accounts <- tbl(mydb, "accounts")
## cards <- tbl(mydb, "cards")
## clients <- tbl(mydb, "clients")
## disps <- tbl(mydb, "disps")
## districts <- tbl(mydb, "districts")
## loans <- tbl(mydb, "loans")
## orders <- tbl(mydb, "orders")
## tkeys <- tbl(mydb, "tkeys")
## trans <- tbl(mydb, "trans")
## ...
## ...
## ...
## ...
## ...
## ...
## ...
## ...
## ...
## ...
## ...
## ...
## ...

## ----dm-object, highlight.output = 2, cache = FALSE------------
library(dm)
dm <- dm_from_src(mydb)
dm


## ----dm-container----------------------------------------------
names(dm)
dm$accounts


## ----dm-download, highlight.output = 1-------------------------
dm_local <-
  dm %>%
  collect()
dm_local$accounts


## ----dm-download-size------------------------------------------
object.size(dm_local)


## ----dm-pk, cache = FALSE--------------------------------------
dm_pk <-
  dm_local %>%
  dm_add_pk(districts, id) %>%
  dm_add_pk(accounts, id) %>%
  dm_add_pk(clients, id) %>%
  dm_add_pk(loans, id) %>%
  dm_add_pk(orders, id) %>%
  dm_add_pk(trans, id) %>%
  dm_add_pk(disps, id) %>%
  dm_add_pk(cards, id)


## ----dm-pk-out, highlight.output = 6---------------------------
dm_pk


## ----dm-fk, cache = FALSE--------------------------------------
dm_fk <-
  dm_pk %>% # <<
  dm_add_fk(
    loans,
    account_id,
    accounts
  ) %>%
  dm_add_fk(
    orders,
    account_id,
    accounts
  ) %>%
  dm_add_fk(
    trans,
    account_id,
    accounts
  ) %>%
  dm_add_fk(
    disps,
    account_id,
    accounts
  ) %>%
  dm_add_fk(
    disps,
    client_id,
    clients
  ) %>%
  dm_add_fk(
    accounts,
    district_id,
    districts
  ) %>%
  dm_add_fk(
    clients,
    district_id,
    districts
  ) %>%
  dm_add_fk(
    cards,
    disp_id,
    disps
  )


## ----dm-fk-out, highlight.output = 7---------------------------
dm_fk


## ----dm-visualize----------------------------------------------
dm_fk %>%
  dm_draw()


## ----dm-tweak--------------------------------------------------
dm_fk %>%
  dm_rm_tbl(tkeys) %>%
  dm_set_colors(green = loans, red = trans, orange = districts) %>%
  dm_draw()


## ----dm-immutable, highlight.output = 4:5----------------------
dm_fk


## ----dm-tweak-assign, highlight.output = 4:5-------------------
dm_tweaked <- # <<
  dm_fk %>%
  dm_rm_tbl(tkeys) %>%
  dm_set_colors(green = loans, red = trans, orange = districts)
dm_tweaked # <<


## ----dm-constraints--------------------------------------------
dm_fk %>%
  dm_examine_constraints()
dm_nycflights13() %>%
  dm_examine_constraints()


## ----dm-flatten------------------------------------------------
dm_fk %>%
  dm_flatten_to_tbl(loans)


## ----starwars-show---------------------------------------------
starwars


## ----starwars-pk-----------------------------------------------
starwars %>%
  select_if(~ !is.list(.)) %>%
  enum_pk_candidates()


## ----starwars-nested-------------------------------------------
starwars %>%
  select(name, films)


## ----starwars-nested-expand------------------------------------
characters_films <-
  starwars %>%
  select(name, films) %>%
  unnest(films) %>%
  rename(film = films)
characters_films


## ----starwars-films--------------------------------------------
films <-
  characters_films %>%
  distinct(film) %>%
  mutate(year = round(runif(7, max = 40) + 1980))

films


## ----starwars-dm-create----------------------------------------
characters <-
  starwars %>%
  select(-films)

dm_starwars <-
  dm(
    # <<
    characters,
    films,
    characters_films
  ) %>%
  dm_add_pk(characters, name) %>%
  dm_add_pk(films, film) %>%
  dm_add_fk(
    characters_films,
    name,
    characters
  ) %>%
  dm_add_fk(characters_films, film, films)

dm_starwars %>%
  dm_examine_constraints()


## ----starwars-dm-draw------------------------------------------
dm_starwars %>%
  dm_draw()


## ----starwars-dm-clean-----------------------------------------
dm_starwars_clean <-
  dm_starwars %>%
  dm_zoom_to(characters) %>%
  select_if(~ !is.list(.)) %>% # <<
  dm_update_zoomed()

dm_starwars_clean$characters


## ----sqlite, cache = FALSE-------------------------------------
sqlite <- dbConnect(RSQLite::SQLite(), dbname = ":memory:")

dm_starwars_production <- copy_dm_to(sqlite, dm_starwars_clean)
dm_starwars_production


## ----sqlite-examine, highlight.output = 2----------------------
dm_starwars_production %>%
  dm_examine_constraints()

## ----setup, include = FALSE----------------------------------------------
source("setup/setup.R")

## ----zoom----------------------------------------------------------------
# library(dm)
# library(dplyr)
# 
# flights_dm <- dm_nycflights13(cycle = TRUE)
# flights_dm
# flights_keyed <-
#   flights_dm %>%
#   dm_get_tables(keyed = TRUE)
# 
# # The print output for a `dm_keyed_tbl` looks very much like that from a normal
# # `tibble`, with additional details about keys.
# flights_keyed$flights
# flights_tbl_mutate <-
#   flights_keyed$flights %>%
#   mutate(am_pm_dep = if_else(dep_time < 1200, "am", "pm"), .after = dep_time)
# 
# flights_tbl_mutate

## ----zoom2---------------------------------------------------------------
# updated_flights_dm <- dm(
#   flights = flights_tbl_mutate,
#   !!!flights_keyed[c("airlines", "airports", "planes", "weather")]
# )
# 
# # The only difference in the `dm` print output is the increased number of
# # columns
# updated_flights_dm
# # The schematic view of the data model remains unchanged
# dm_draw(updated_flights_dm)

## ------------------------------------------------------------------------
# library(tidyr)
# 
# flights_keyed$weather
# 
# # Maybe there is some hidden candidate for a primary key that we overlooked?
# enum_pk_candidates(flights_keyed$weather)
# # Seems we have to construct a column with unique values
# # This can be done by combining column `origin` with `time_hour`, if the latter
# # is converted to a single time zone first; all within the `dm`:
# weather_tbl_mutate <-
#   flights_keyed$weather %>%
#   # first convert all times to the same time zone:
#   mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
#   # paste together as character the airport code and the time
#   unite("origin_slot_id", origin, time_hour_fmt) %>%
#   select(origin_slot_id, everything())
# 
# # check if we the result is as expected:
# weather_tbl_mutate %>%
#   enum_pk_candidates() %>%
#   filter(candidate)
# # We apply the same transformation to create
# # the foreign key in the flights table:
# flights_tbl_mutate <-
#   flights_keyed$flights %>%
#   mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
#   unite("origin_slot_id", origin, time_hour_fmt) %>%
#   select(origin_slot_id, everything())
# 
# surrogate_flights_dm <-
#   dm(
#     weather = weather_tbl_mutate,
#     flights = flights_tbl_mutate,
#     !!!flights_keyed[c("airlines", "airports", "planes")]
#   ) %>%
#   dm_add_pk(weather, origin_slot_id) %>%
#   dm_add_fk(flights, origin_slot_id, weather)
# 
# surrogate_flights_dm %>%
#   dm_draw()

## ------------------------------------------------------------------------
# disentangled_flights_dm <-
#   dm(
#     destination = flights_keyed$airports,
#     origin = flights_keyed$airports,
#     !!!flights_keyed[c("flights", "airlines", "planes", "weather")]
#   ) %>%
#   # Key relations are also duplicated, so the wrong ones need to be removed
#   dm_rm_fk(flights, dest, origin) %>%
#   dm_rm_fk(flights, origin, destination)
# 
# disentangled_flights_dm %>%
#   dm_draw()

## ------------------------------------------------------------------------
# flights_derived <-
#   flights_dm %>%
#   pull_tbl(flights, keyed = TRUE) %>%
#   dplyr::count(origin, carrier)
# 
# derived_flights_dm <- dm(flights_derived, !!!flights_keyed)
# 
# derived_flights_dm %>%
#   dm_draw()

## ------------------------------------------------------------------------
# planes_for_join <-
#   flights_keyed$planes %>%
#   select(tailnum, plane_type = type)
# 
# joined_flights_tbl <-
#   flights_keyed$flights %>%
#   # let's first reduce the number of columns of flights
#   select(-dep_delay:-arr_delay, -air_time:-minute, -starts_with("sched_")) %>%
#   # in the {dm}-method for the joins you can specify which columns you want to
#   # add to the subsetted table
#   left_join(planes_for_join)
# 
# joined_flights_dm <- dm(
#   flights_plane_type = joined_flights_tbl,
#   !!!flights_keyed[c("airlines", "airports", "weather")]
# )
# 
# # this is how the table looks now
# joined_flights_dm$flights_plane_type
# # also here, the FK-relations are transferred to the new table
# joined_flights_dm %>%
#   dm_draw()

## ------------------------------------------------------------------------
# dm <- dm_nycflights13()
# dm_deconstruct(dm)

## ----setup, include = FALSE----------------------------------------------
source("setup/setup.R")

## ----zoom----------------------------------------------------------------
# library(dm)
# library(dplyr)
# flights_dm <- dm_nycflights13()
# flights_dm
# flights_zoomed <-
#   flights_dm %>%
#   dm_zoom_to(flights)
# # The print output for a `dm_zoomed` looks very much like that from a normal `tibble`.
# flights_zoomed
# 
# flights_zoomed_mutate <-
#   flights_zoomed %>%
#   mutate(am_pm_dep = if_else(dep_time < 1200, "am", "pm")) %>%
#   # in order to see our changes in the output we use `select()` for reordering the columns
#   select(year:dep_time, am_pm_dep, everything())
# 
# flights_zoomed_mutate
# 
# # To update the original `dm` with a new `flights` table we use `dm_update_zoomed()`:
# updated_flights_dm <-
#   flights_zoomed_mutate %>%
#   dm_update_zoomed()
# # The only difference in the `dm` print output is the increased number of columns
# updated_flights_dm
# # The schematic view of the data model remains unchanged
# dm_draw(updated_flights_dm)

## ------------------------------------------------------------------------
# library(tidyr)
# 
# weather_zoomed <-
#   flights_dm %>%
#   dm_zoom_to(weather)
# weather_zoomed
# # Maybe there is some hidden candidate for a primary key that we overlooked
# enum_pk_candidates(weather_zoomed)
# # Seems we have to construct a column with unique values
# # This can be done by combining column `origin` with `time_hour`, if the latter
# # is converted to a single time zone first; all within the `dm`:
# weather_zoomed_mutate <-
#   weather_zoomed %>%
#   # first convert all times to the same time zone:
#   mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
#   # paste together as character the airport code and the time
#   unite("origin_slot_id", origin, time_hour_fmt) %>%
#   select(origin_slot_id, everything())
# # check if we the result is as expected:
# enum_pk_candidates(weather_zoomed_mutate) %>% filter(candidate)
# flights_upd_weather_dm <-
#   weather_zoomed_mutate %>%
#   dm_update_zoomed() %>%
#   dm_add_pk(weather, origin_slot_id)
# flights_upd_weather_dm
# # creating the coveted FK relation between `flights` and `weather`
# extended_flights_dm <-
#   flights_upd_weather_dm %>%
#   dm_zoom_to(flights) %>%
#   mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
#   # need to keep `origin` as FK to airports, so `remove = FALSE`
#   unite("origin_slot_id", origin, time_hour_fmt, remove = FALSE) %>%
#   dm_update_zoomed() %>%
#   dm_add_fk(flights, origin_slot_id, weather)
# extended_flights_dm %>% dm_draw()

## ------------------------------------------------------------------------
# dm_draw(dm_nycflights13(cycle = TRUE))

## ------------------------------------------------------------------------
# disentangled_flights_dm <-
#   dm_nycflights13(cycle = TRUE) %>%
#   # zooming and immediately inserting essentially creates a copy of the original table
#   dm_zoom_to(airports) %>%
#   # reinserting the `airports` table under the name `destination`
#   dm_insert_zoomed("destination") %>%
#   # renaming the originally zoomed table
#   dm_rename_tbl(origin = airports) %>%
#   # Key relations are also duplicated, so the wrong ones need to be removed
#   dm_rm_fk(flights, dest, origin) %>%
#   dm_rm_fk(flights, origin, destination)
# dm_draw(disentangled_flights_dm)

## ------------------------------------------------------------------------
# dm_with_summary <-
#   flights_dm %>%
#   dm_zoom_to(flights) %>%
#   dplyr::count(origin, carrier) %>%
#   dm_insert_zoomed("dep_carrier_count")
# dm_draw(dm_with_summary)

## ------------------------------------------------------------------------
# joined_flights_dm <-
#   flights_dm %>%
#   dm_zoom_to(flights) %>%
#   # let's first reduce the number of columns of flights
#   select(-dep_delay:-arr_delay, -air_time:-time_hour) %>%
#   # in the {dm}-method for the joins you can specify which columns you want to add to the zoomed table
#   left_join(planes, select = c(tailnum, plane_type = type)) %>%
#   dm_insert_zoomed("flights_plane_type")
# # this is how the table looks now
# joined_flights_dm$flights_plane_type
# # also here, the FK-relations are transferred to the new table
# dm_draw(joined_flights_dm)

## ------------------------------------------------------------------------
# flights_dm %>%
#   dm_zoom_to(flights) %>%
#   select(-dep_delay:-arr_delay, -air_time:-time_hour) %>%
#   left_join(planes, select = c(tailnum, plane_type = type)) %>%
#   pull_tbl()

# Demo for presentation at the
# New York Open Statistical Programming Meetup,
# February 2020

# {dm}: Relational data models in R

options(tibble.print_min = 6)
options(tibble.print_max = 6)
options(rlang_backtrace_on_error = "none")
library(magrittr)

##
##
##
## Greeting
## -------------------------------------------------------
##
##
##

# Poll: Who is familiar with the {dplyr} package
#       (grammar of data manipulation)?

# Poll: Who is familiar with the {dbplyr} package
#       (using {dplyr} with databases)?

# Poll: Who has worked with databases, perhaps using SQL?

# Poll: Who has worked with a software where you work with
#       "THE DATASET"?

# Poll: Who uses more than one table/data frame
#       at the same time?

##
##
##
##
##
## Keys
## -------------------------------------------------------
##
##
##

# Data
library(nycflights13)

dm::dm_nycflights13(cycle = TRUE) %>%
  dm::dm_draw()

# `carrier` is a "primary key" in `airlines`
any(duplicated(airlines$carrier))

# `carrier` is a "foreign key" in `flights` into `airlines`
all(flights$carrier %in% airlines$carrier)


##
##
##
## Goal: work with one table
## -------------------------------------------------------
##
##
##

library(tidyverse)
library(nycflights13)

# Join one table
flights %>%
  left_join(airlines, by = "carrier") %>%
  select(name, carrier, year:dep_time)

# Use attributes from two tables after joining
flights %>%
  left_join(airlines, by = "carrier") %>%
  count(name, month) %>%
  ggplot() +
  aes(x = factor(month), y = name, fill = n) +
  geom_tile()

# Join many tables
flights_details <-
  flights %>%
  left_join(airlines, by = "carrier") %>%
  left_join(planes, by = "tailnum") %>%
  left_join(weather, by = c("origin", "time_hour")) %>%
  left_join(airports, by = c("origin" = "faa")) %>%
  left_join(airports, by = c("dest" = "faa"))

flights_details

##
##
##
## ...eventually
## -------------------------------------------------------
##
##
##

flights_details %>%
  count(is.na(lat.y), is.na(lon.y))

all(flights$dest %in% airports$faa)

##
##
##
## PROBLEMS (see appendix)
## -------------------------------------------------------
##
## - wrong keys
## - data mismatches
## - relationship unclear
## - combinatorial explosion
##
##
##

##
##
##
##
## Single source of truth
## -------------------------------------------------------
##
##
##

# Keep information in one place
flights %>%
  left_join(airlines) %>%
  select(name, carrier, tailnum)

# Update in one single location
airlines[airlines$carrier == "UA", "name"] <-
  "United broke my guitar"

airlines %>%
  filter(carrier == "UA")

# ...propagates to all related records
flights %>%
  left_join(airlines) %>%
  select(name, carrier, tailnum)

rm(airlines)

##
##
##
## Data model object
## -------------------------------------------------------
##
##
##

library(dm)

# Compound object: tables, relationships, data
dm_flights <- dm_nycflights13(cycle = TRUE)
dm_flights

dm_flights$airlines

# Table names
dm_flights %>%
  names()

# NB: [ and [[ also work

# Visualize
dm_flights %>%
  dm_draw()


##
##
##
##
## THREE USE CASES:
## -------------------------------------------------------
##
## 1. Work with a prepared dm object or connect to a
##    database
## 2. Build a data model for your own data
## 3. Publish a dm to a relational database (and load
##    from it)
##
##
##
##

##
##
##
##
##
##
##
##
##
##
##
##
## USE CASE 1: Work with a prepared dm object or connect
##             to a database
## -------------------------------------------------------
##
##
##

##
##
##
## Setup
## -------------------------------------------------------
##
##
##

# Get a dm from an existing function:
dm_flights <- dm_nycflights13()

# Selecting tables
dm_flights %>%
  dm_select_tbl(-weather)

# Selecting columns in tables
dm_flights %>%
  dm_select_tbl(-weather) %>%
  dm_select(
    flights,
    origin,
    dest,
    carrier,
    tailnum,
    year,
    month,
    day,
    dep_time
  )

# !!! Connect to a database !!!
try({
  con_pq <- DBI::dbConnect(RPostgres::Postgres())

  dm_flights <-
    dm_from_src(con_pq, schema = "nycflights13") %>%
    dm_rm_fk(flights, dest, airports)

  dm_flights
})

# Selecting more columns in tables
dm_flights %>%
  dm_select_tbl(-weather) %>%
  dm_select(
    flights,
    origin,
    dest,
    carrier,
    tailnum,
    year,
    month,
    day,
    dep_time
  ) %>%
  dm_select(planes, tailnum, year) %>%
  dm_select(airports, faa, lat, lon)

# Filtering rows in tables
dm_flights %>%
  dm_select_tbl(-weather) %>%
  dm_select(
    flights,
    origin,
    dest,
    carrier,
    tailnum,
    year,
    month,
    day,
    dep_time
  ) %>%
  dm_select(planes, tailnum, year) %>%
  dm_select(airports, faa, lat, lon) %>%
  dm_filter(flights, month == 2, day == 5)

# Filtering rows in other tables
dm_flights %>%
  dm_select_tbl(-weather) %>%
  dm_select(
    flights,
    origin,
    dest,
    carrier,
    tailnum,
    year,
    month,
    day,
    dep_time
  ) %>%
  dm_select(planes, tailnum, year) %>%
  dm_select(airports, faa, name, lat, lon) %>%
  dm_filter(flights, month == 2, day == 5) %>%
  dm_filter(airports, name == "John F Kennedy Intl") %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.")

# Immutable object: need to assign a name to persist
dm_flights_jfk_today <-
  dm_flights %>%
  dm_select_tbl(-weather) %>%
  dm_select(
    flights,
    origin,
    dest,
    carrier,
    tailnum,
    year,
    month,
    day,
    dep_time
  ) %>%
  dm_select(planes, tailnum, year) %>%
  dm_select(airports, faa, name, lat, lon) %>%
  dm_filter(flights, month == 2, day == 4) %>%
  dm_filter(airports, name == "John F Kennedy Intl") %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_apply_filters()

dm_flights_jfk_today

dm_flights_jfk_today %>%
  dm_draw()

# Lazy tables
dm_flights_jfk_today %>%
  pull_tbl(flights)

try(
  dm_flights_jfk_today %>%
    pull_tbl(flights) %>%
    dbplyr::sql_render()
)

# Load the entire data model into memory
dm_flights_jfk_today_df <-
  dm_flights_jfk_today %>%
  collect()

dm_flights_jfk_today_df


##
##
##
## Joining two tables
## -------------------------------------------------------
##
##
##

dm_flights <- dm_nycflights13()
dm_flights %>%
  dm_draw()

dm_flights %>%
  dm_join_to_tbl(airlines, flights)

airlines_flights <-
  dm_flights %>%
  dm_join_to_tbl(airlines, flights)

airlines_flights

try(
  dm_flights %>%
    dm_join_to_tbl(airports, airlines)
)


##
##
##
## Joining many tables
## -------------------------------------------------------
##
##
##

# Join "everything" into a flat table:
dm_flights %>%
  dm_flatten_to_tbl(flights)

# Manual disambiguation of column names:
dm_flights %>%
  dm_select(planes, -year) %>%
  dm_rename(airlines, airline_name = name) %>%
  dm_flatten_to_tbl(flights)

# Separate access to automatic disambiguation:
dm_flights %>%
  dm_disambiguate_cols() %>%
  dm_draw()


##
##
##
## NEW NEW NEW: Data manipulation in a dm
## -------------------------------------------------------
##
##
##

# A single table of a `dm` can be activated (or zoomed to),
# and subsequently be manipulated by many {dplyr}-verbs.
# Eventually, either the original table can be updated
# or the manipulated table can be inserted as a new table.

# The print output for a `dm_zoomed` looks very much
# like that from a normal `tibble`.
dm_flights %>%
  dm_zoom_to(flights)

# Many {dplyr} verbs work on zoomed tables:
dm_flights %>%
  dm_zoom_to(flights) %>%
  mutate(
    am_pm_dep = if_else(dep_time < 1200, "am", "pm")
  ) %>%
  select(year:dep_time, am_pm_dep, everything())

# Put back into the dm:
dm_flights %>%
  dm_zoom_to(flights) %>%
  mutate(
    am_pm_dep = if_else(dep_time < 1200, "am", "pm")
  ) %>%
  select(year:dep_time, am_pm_dep, everything()) %>%
  dm_update_zoomed()

# Immutable objects, like in {dplyr}
dm_flights

# Creation of a summary table:
dm_with_summary <-
  dm_flights %>%
  dm_zoom_to(flights) %>%
  count(origin) %>%
  dm_insert_zoomed("origin_count")

dm_with_summary$origin_count

dm_with_summary %>%
  dm_draw()

# All relationships still available in the summary
# are retained:
dm_flights %>%
  dm_zoom_to(flights) %>%
  count(carrier, origin) %>%
  dm_insert_zoomed("origin_carrier_count") %>%
  dm_draw()


##
##
##
## USE CASE 2: Build a data model for your own data
## -------------------------------------------------------
##
##
##

##
##
##
##
##
##
##
## Build up data model from scratch
## -------------------------------------------------------
##
##
##

# Use `dm()` with a syntax similar to `tibble()`:
nycflights13_tbl <-
  dm(airlines, airports, flights, planes, weather)
nycflights13_tbl

nycflights13_tbl %>%
  dm_draw()

# Alternatively, start from an empty `dm`
# and add tables via `dm_add_tbl()`:
dm() %>%
  dm_add_tbl(airlines, airports, flights, planes, weather)

# Tables are not connected yet:
nycflights13_tbl %>%
  dm_draw()

# Adding primary keys:
nycflights13_pk <-
  nycflights13_tbl %>%
  dm_add_pk(planes, tailnum) %>%
  dm_add_pk(airports, faa) %>%
  dm_add_pk(airlines, carrier)

nycflights13_pk %>%
  dm_draw()

# FIXME: Model weak constraints, show differently
# in diagram (#4)

# Adding foreign keys
nycflights13_fk <-
  nycflights13_pk %>%
  dm_add_fk(flights, tailnum, planes) %>%
  dm_add_fk(flights, origin, airports) %>%
  dm_add_fk(flights, dest, airports) %>%
  dm_add_fk(flights, carrier, airlines, check = TRUE)

nycflights13_fk %>%
  dm_draw()

# Color it!
dm_get_available_colors()

nycflights13_base <-
  nycflights13_fk %>%
  dm_set_colors(
    blue = c(airlines, planes, airports)
  )

nycflights13_base %>%
  dm_draw()

nycflights13_base %>%
  dm_paste()

# !!! NEW: Examine all constraints of a dm !!!
nycflights13_base %>%
  dm_examine_constraints()


##
##
##
## USE CASE 3: Publish a dm to a relational database
##             (and load from it)
## -------------------------------------------------------
##
##
##

##
##
##
##
##
## Copy to database
## -------------------------------------------------------
##
##
##

# All operations are designed to work locally
# and on the database
dm_flights_sqlite <-
  dm_flights %>%
  copy_dm_to(
    dbplyr::src_memdb(),
    .,
    unique_table_names = TRUE,
    set_key_constraints = FALSE
  )

dm_flights_sqlite

dm_flights_sqlite %>%
  dm_draw()

# Operations work on the database:
dm_flights_sqlite %>%
  dm_flatten_to_tbl(flights) %>%
  dbplyr::sql_render()


##
##
##
## Copy to a database, including key constraints
## -------------------------------------------------------
##
##
##

try({
  dm_flights <- dm_nycflights13(cycle = TRUE)

  con_pq <- DBI::dbConnect(RPostgres::Postgres())

  # Off by default, to ensure that no tables are
  # accidentally deleted
  if (FALSE) {
    DBI::dbExecute(
      con_pq,
      "DROP SCHEMA IF EXISTS nycflights13 CASCADE"
    )
    DBI::dbExecute(
      con_pq,
      "CREATE SCHEMA nycflights13"
    )
  }

  # Ensure referential integrity (FIXME: Better way):
  dm_flights_ref <-
    dm_flights %>%
    #
    # FIXME: Can't use cycles for now.
    dm_rm_fk(flights, origin, airports) %>%
    #
    # Fill bad links with NA/NULL values:
    dm_zoom_to(flights) %>%
    left_join(planes, select = c(tailnum, type)) %>%
    mutate(
      tailnum = ifelse(is.na(type), NA_character_, tailnum)
    ) %>%
    dm_update_zoomed() %>%
    dm_add_fk(flights, tailnum, planes) %>%
    #
    # Insert synthetic rows:
    dm_zoom_to(flights) %>%
    count(dest) %>%
    dm_insert_zoomed("dest") %>%
    dm_zoom_to(airports) %>%
    full_join(dest, select = dest) %>%
    dm_update_zoomed() %>%
    dm_select_tbl(-dest) %>%
    dm_add_fk(flights, origin, airports)

  qualified_names <-
    rlang::set_names(names(dm_flights_ref))
  qualified_names[] <- paste0(
    "nycflights13.",
    qualified_names
  )

  dm_flights_pq <-
    dm_flights_ref %>%
    copy_dm_to(
      con_pq,
      .,
      temporary = FALSE,
      table_names = qualified_names
    )

  dm_flights_from_pq <-
    dm_from_src(con_pq, schema = "nycflights13")

  dm_flights_from_pq %>%
    dm_draw()
})


##
##
##
##
## THREE USE CASES:
## -------------------------------------------------------
##
## 1. Work with a prepared dm object or connect to a
##    database
## 2. Build a data model for your own data
## 3. Publish a dm to a relational database (and load
##    from it)
## 4. Data documentation (?)
##
##
##
##

##
##
##
##
##
##
##
##
##
##
##
##
## Appendix
## ======================================================
##
##
##

##
##
##
## Analogy?
## -------------------------------------------------------
##
##
##

# Analogy: parallel vectors
# Multiple parallel vectors can be combined into a data
# frame.
# Multiple related tables can be combined into a dm.
x <- 1:5
x
y <- x + 1
y
z <- diff(y)
z

try(
  tibble(x = 1:5) %>%
    mutate(y = x + 1) %>%
    mutate(z = diff(y))
)

##
##
##
## Pitfall: wrong keys
## -------------------------------------------------------
##
##
##

# Same for airplanes?
planes

flights_base <-
  flights %>%
  select(
    carrier,
    tailnum,
    origin,
    dest,
    year,
    month,
    day,
    time_hour
  )

flights_base %>%
  left_join(planes)

flights_base %>%
  left_join(planes) %>%
  count(is.na(type))

# Take a closer look at the join
flights_base %>%
  left_join(planes, by = "tailnum")

flights_base %>%
  left_join(planes, by = "tailnum") %>%
  count(is.na(type))

##
##
##
## Pitfall: data mismatches
## -------------------------------------------------------
##
##
##

flights_base %>%
  left_join(planes, by = "tailnum") %>%
  summarize(.by = carrier, mismatch_rate = mean(is.na(type))) %>%
  filter(mismatch_rate > 0) %>%
  ggplot(aes(x = carrier, y = mismatch_rate)) +
  geom_col()

##
##
##
## Pitfall: relationship unclear
## -------------------------------------------------------
##
##
##

# Same for airports?
airports

try(
  flights_base %>%
    left_join(airports)
)

# Need to specify join variables!
flights_base %>%
  left_join(airports, by = c("origin" = "faa"))

# cleanup
rm(airlines)

##
##
##
## Pitfall: combinatorial explosion
## -------------------------------------------------------
##
##
##

# Row identifiers
t1 <- tibble(a = 1, b = letters[1:3])
t1
t2 <- tibble(a = 1, c = 1:2)
t2

# What happens here?
left_join(t1, t2)

# When joining, the column(s) must be unique in at least
# one participating table!

# Ensure uniqueness:
airlines %>%
  count(carrier)

airlines %>%
  count(carrier) %>%
  count(n)

planes %>%
  count(tailnum) %>%
  count(n)

# dm shortcut:
planes %>%
  check_key(tailnum)

try(
  planes %>%
    check_key(engines)
)

airports %>%
  check_key(faa)

# Why is name not a key candidate for airports?
try(
  airports %>%
    check_key(name)
)

# Friendly description
airports %>%
  enum_pk_candidates()

# Cleanup
rm(t1, t2)


##
##
##
## Advanced zooming: Use data manipulation to understand
## and establish relationships
## -------------------------------------------------------
##
##
##

# Determine key candidates
zoomed_weather <- dm_zoom_to(nycflights13_base, weather)
zoomed_weather

# `enum_pk_candidates()` works for both `tibbles` and
# `dm_zoomed`
enum_pk_candidates(zoomed_weather)

enum_pk_candidates(zoomed_weather) %>%
  count(candidate)

# It's tricky:
zoomed_weather %>%
  unite(
    "slot_id",
    origin,
    year,
    month,
    day,
    hour,
    remove = FALSE
  ) %>%
  count(slot_id) %>%
  filter(n > 1)

zoomed_weather %>%
  count(origin, time_hour) %>%
  filter(n > 1)

zoomed_weather %>%
  count(origin, format(time_hour)) %>%
  filter(n > 1)

# This looks like a good candidate:
zoomed_weather %>%
  count(origin, format(time_hour, tz = "UTC")) %>%
  filter(n > 1)

# FIXME: Support compound keys (#3)

# Currently, we need to create surrogate keys:
nycflights13_weather_link <-
  zoomed_weather %>%
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  unite("origin_slot_id", origin, time_hour_fmt) %>%
  # Update the original `weather` table
  dm_update_zoomed() %>%
  # Add a PK for the "enhanced" weather table
  dm_add_pk(weather, origin_slot_id)

nycflights13_weather_link$weather

nycflights13_weather_link %>%
  dm_draw()

# FIXME: zoom to multiple tables

nycflights13_weather_flights_link <-
  dm_zoom_to(nycflights13_weather_link, flights) %>%
  # same procedure with `flights` table
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  # for flights we need to keep the column `origin`,
  # since it is a FK pointing to `airports`
  unite("origin_slot_id", origin, time_hour_fmt, remove = FALSE) %>%
  select(origin_slot_id, everything(), -time_hour_fmt) %>%
  dm_update_zoomed()

# `dm_enum_fk_candidates()` of a `dm` gives info
# about potential FK columns from one table to another
dm_enum_fk_candidates(
  nycflights13_weather_flights_link,
  flights,
  weather
)

# well, it's almost perfect, let's add the FK anyway...

nycflights13_perfect <-
  nycflights13_weather_flights_link %>%
  dm_add_fk(flights, origin_slot_id, weather)

nycflights13_perfect %>%
  dm_draw()

# What are the missings?
nycflights13_perfect %>%
  dm_zoom_to(flights) %>%
  anti_join(weather) %>%
  count(origin_slot_id)

# Demo for presentation at Energie 360, November 2019

# {dm} facilitates working with multiple tables

options(tibble.print_min = 6)
options(tibble.print_max = 6)
options(rlang_backtrace_on_error = "none")

##
##
##
## Greeting
## --------------------------------------------------------------------
##
##
##

# Teaser
dm::dm_nycflights13(cycle = TRUE) %>%
  dm::dm_draw()

# Poll: Who has worked with a software that has
#       a concept of "THE DATASET"?

# Poll: Who has worked with databases?

# Poll: Who uses more than one table/data frame
#       at the same time?

##
##
##
## Why?
## --------------------------------------------------------------------
##
##
##

library(nycflights13)

# Example dataset: tables linked with each other
?flights

flights

##
##
##
## Data model
## --------------------------------------------------------------------
##
##
##

library(tidyverse)
flights_base <-
  flights %>%
  select(year, month, day, carrier, tailnum, origin, dest, time_hour)
flights_base

# `carrier` column also present in `airlines`, this table contains
# additional information
airlines

flights_base %>%
  left_join(airlines)

##
##
##
## Keys
## --------------------------------------------------------------------
##
##
##

# `carrier` is a "primary key" in `airlines`
any(duplicated(airlines$carrier))

# `carrier` is a "foreign key" in `flights` into `airlines`
all(flights$carrier %in% airlines$carrier)

##
##
##
## Single source of truth
## --------------------------------------------------------------------
##
##
##

# Update in one single location
airlines[airlines$carrier == "UA", "name"] <-
  "United broke my guitar"

airlines %>%
  filter(carrier == "UA")

# ...propagates to all related records
flights_base %>%
  left_join(airlines)

##
##
##
## Goal: work with one table
## --------------------------------------------------------------------
##
##
##

flights_base %>%
  left_join(airlines, by = "carrier") %>%
  left_join(planes, by = "tailnum") %>%
  left_join(weather, by = c("origin", "time_hour")) %>%
  left_join(airports, by = c("origin" = "faa")) %>%
  left_join(airports, by = c("dest" = "faa"))

##
##
##
## PROBLEMS (see appendix)
## --------------------------------------------------------------------
##
## - wrong keys
## - data mismatches
## - relationship unclear
## - combinatorial explosion
##
##
##

##
##
##
## Data model object
## --------------------------------------------------------------------
##
##
##

library(dm)

# Compound object: tables, relationships, data
dm_flights <- dm_nycflights13(cycle = TRUE)
dm_flights

dm_flights %>%
  dm_draw()

# Selection of tables
dm_flights %>%
  dm_select_tbl(flights, airlines) %>%
  dm_draw()

dm_flights %>%
  dm_select_tbl(airports, airlines) %>%
  dm_draw()

try(
  dm_flights %>%
    dm_select_tbl(bogus)
)

# Accessing tables
dm_flights %>%
  tbl("airlines")

dm_flights$airlines

# Table names
dm_flights %>%
  src_tbls()

# NB: [, $, [[ and names() also work

##
##
##
## Joining two tables
## --------------------------------------------------------------------
##
##
##

dm_flights <- dm_nycflights13()
dm_flights %>%
  dm_draw()

dm_flights %>%
  dm_join_to_tbl(airlines, flights)

try(
  dm_flights %>%
    dm_join_to_tbl(airports, airlines)
)

##
##
##
## Joining many tables
## --------------------------------------------------------------------
##
##
##

dm_flights %>%
  dm_flatten_to_tbl(flights)

##
##
##
## Copy to database
## --------------------------------------------------------------------
##
##
##

# All operations are designed to work locally and on the database
dm_flights_sqlite <-
  dm_flights %>%
  copy_dm_to(
    dbplyr::src_memdb(),
    .,
    unique_table_names = TRUE,
    set_key_constraints = FALSE
  )

dm_flights_sqlite

dm_flights_sqlite %>%
  dm_draw()

dm_flights_sqlite %>%
  dm_get_tables() %>%
  map(dbplyr::sql_render)

dm_flights_sqlite %>%
  dm_join_to_tbl(airlines, flights) %>%
  dbplyr::sql_render()

dm_flights_sqlite %>%
  dm_flatten_to_tbl(flights) %>%
  dbplyr::sql_render()

# Filtering on the database
dm_flights_sqlite %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(flights, day == 1) %>%
  tbl("flights")

# ... and the corresponding SQL statement and query plan
dm_flights_sqlite %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(flights, day == 1) %>%
  tbl("flights") %>%
  dbplyr::sql_render()

##
##
##
## Filtering for data models
## --------------------------------------------------------------------
##
##
##

# Filtering on a table returns a dm object
# with
# 1. the filter applied to the table in question
# 2. the filter condition(s) stored
dm_flights %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.")

# the resulting `dm` can then be filtered on another table
dm_flights %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl")

# ... and stored in another dm variable
delta_non_jfk_january <-
  dm_flights %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(planes, year < 2000) %>%
  dm_filter(flights, month == 1)
delta_non_jfk_january

# Querying a table applies the filters via semi-joins
# along the FK constraints to the requested table
delta_non_jfk_january %>%
  tbl("planes")

# FIXME: Can this work without applying all filters?

delta_non_jfk_january %>%
  dm_apply_filters() %>%
  dm_join_to_tbl(flights, airlines)

delta_non_jfk_january %>%
  dm_flatten_to_tbl(flights)


##
##
##
## Build up data model from scratch
## --------------------------------------------------------------------
##
##
##

# Use `dm()` with a syntax similar to `tibble()`:
nycflights13_tbl <- dm(airlines, airports, flights, planes, weather)
nycflights13_tbl

# Alternatively, start from an empty `dm`
# and add tables via `dm_add_tbl()`:
dm() %>%
  dm_add_tbl(airlines, airports, flights, planes, weather)

# Tables are not connected yet:
nycflights13_tbl %>%
  dm_draw()

# Adding primary keys:
nycflights13_pk <-
  nycflights13_tbl %>%
  dm_add_pk(planes, tailnum) %>%
  dm_add_pk(airports, faa) %>%
  dm_add_pk(airlines, carrier)

nycflights13_pk %>%
  dm_draw()

# FIXME: Model weak constraints, show differently in diagram (#4)

# Adding foreign keys
nycflights13_fk <-
  nycflights13_pk %>%
  dm_add_fk(flights, tailnum, planes) %>%
  dm_add_fk(flights, origin, airports) %>%
  dm_add_fk(flights, dest, airports) %>%
  dm_add_fk(flights, carrier, airlines, check = TRUE)

nycflights13_fk %>%
  dm_draw()

# Color it!
dm_get_available_colors()

nycflights13_base <-
  nycflights13_fk %>%
  dm_set_colors(
    airlines = ,
    planes = ,
    weather = ,
    airports = "blue"
  )

nycflights13_base %>%
  dm_draw()


##
##
##
## NEW NEW NEW: Data manipulation in a dm
## --------------------------------------------------------------------
##
##
##

# A single table of a `dm` can be activated (or zoomed to),
# and subsequently be manipulated by many {dplyr}-verbs.
# Eventually, either the original table can be updated
# or the manipulated table can be inserted as a new table.

# Determine key candidates
zoomed_weather <- dm_zoom_to(nycflights13_base, weather)
zoomed_weather

# `enum_pk_candidates()` works for both `tibbles` and `dm_zoomed`
enum_pk_candidates(zoomed_weather)

enum_pk_candidates(zoomed_weather) %>%
  count(candidate)

# It's tricky:
zoomed_weather %>%
  unite("slot_id", origin, year, month, day, hour, remove = FALSE) %>%
  count(slot_id) %>%
  filter(n > 1)

zoomed_weather %>%
  count(origin, time_hour) %>%
  filter(n > 1)

zoomed_weather %>%
  count(origin, format(time_hour)) %>%
  filter(n > 1)

# This looks like a good candidate:
zoomed_weather %>%
  count(origin, format(time_hour, tz = "UTC")) %>%
  filter(n > 1)

# FIXME: Support compound keys (#3)

# Currently, we need to create surrogate keys:
nycflights13_weather_link <-
  zoomed_weather %>%
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  unite("origin_slot_id", origin, time_hour_fmt) %>%
  # here the original 'weather' table is updated with the manipulated one
  dm_update_zoomed() %>%
  # here we are adding a PK for the "enhanced" weather table
  dm_add_pk(weather, origin_slot_id)

nycflights13_weather_link$weather

nycflights13_weather_link %>%
  dm_draw()

# FIXME: zoom to multiple tables

nycflights13_weather_flights_link <-
  dm_zoom_to(nycflights13_weather_link, flights) %>%
  # same procedure with `flights` table
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  # for flights we need to keep the column `origin`,
  # since it is a FK pointing to `airports`
  unite("origin_slot_id", origin, time_hour_fmt, remove = FALSE) %>%
  select(origin_slot_id, everything(), -time_hour_fmt) %>%
  dm_update_zoomed()

# `dm_enum_fk_candidates()` of a `dm` gives info
# about potential FK columns from one table to another
dm_enum_fk_candidates(nycflights13_weather_flights_link, flights, weather)

# well, it's almost perfect, let's add the FK anyway...

nycflights13_perfect <-
  nycflights13_weather_flights_link %>%
  dm_add_fk(flights, origin_slot_id, weather)

nycflights13_perfect %>%
  dm_draw()

# What are the missings?
nycflights13_perfect %>%
  dm_zoom_to(flights) %>%
  anti_join(weather) %>%
  count(origin_slot_id)

# Create table of aggregates, and insert it into the dm
nycflights13_perfect %>%
  dm_zoom_to(flights) %>%
  count(origin) %>%
  dm_insert_zoomed("flights_agg") %>%
  dm_draw()

##
##
##
## Import a dm from a database, including key constraints
## --------------------------------------------------------------------
##
##
##

try({
  con_pq <- DBI::dbConnect(RPostgres::Postgres())

  # FIXME: Schema support

  # Off by default, to ensure that no tables are accidentally deleted
  if (FALSE) {
    walk(
      names(dm_flights),
      ~ DBI::dbExecute(
        con_pq,
        paste0("DROP TABLE IF EXISTS ", ., " CASCADE")
      )
    )
  }

  # Import
  dm_flights_pq <-
    dm_flights %>%
    dm_filter(planes, TRUE) %>%
    dm_filter(flights, month == 1, day == 1) %>%
    copy_dm_to(con_pq, ., temporary = FALSE)

  dm_flights_from_pq <-
    dm_learn_from_db(con_pq)

  dm_flights_from_pq %>%
    dm_draw()
})

##
##
##
## Appendix
## ====================================================================
##
##
##

##
##
##
## Analogy?
## --------------------------------------------------------------------
##
##
##

# Analogy: parallel vectors
# Multiple parallel vectors can be combined into a data frame.
# Multiple related tables can be combined into a dm.
x <- 1:5
x
y <- x + 1
y
z <- diff(y)
z

try(
  tibble(x = 1:5) %>%
    mutate(y = x + 1) %>%
    mutate(z = diff(y))
)

##
##
##
## Pitfall: wrong keys
## --------------------------------------------------------------------
##
##
##

# Same for airplanes?
planes

flights_base %>%
  left_join(planes)

flights_base %>%
  left_join(planes) %>%
  count(is.na(type))

# Take a closer look at the join
flights_base %>%
  left_join(planes, by = "tailnum")

flights_base %>%
  left_join(planes, by = "tailnum") %>%
  count(is.na(type))

##
##
##
## Pitfall: data mismatches
## --------------------------------------------------------------------
##
##
##

flights_base %>%
  left_join(planes, by = "tailnum") %>%
  summarize(.by = carrier, mismatch_rate = mean(is.na(type))) %>%
  filter(mismatch_rate > 0) %>%
  ggplot(aes(x = carrier, y = mismatch_rate)) +
  geom_col()

##
##
##
## Pitfall: relationship unclear
## --------------------------------------------------------------------
##
##
##

# Same for airports?
airports

try(
  flights_base %>%
    left_join(airports)
)

# Need to specify join variables!
flights_base %>%
  left_join(airports, by = c("origin" = "faa"))

# cleanup
rm(airlines)

##
##
##
## Pitfall: combinatorial explosion
## --------------------------------------------------------------------
##
##
##

# Row identifiers
t1 <- tibble(a = 1, b = letters[1:3])
t1
t2 <- tibble(a = 1, c = 1:2)
t2

# What happens here?
left_join(t1, t2)

# When joining, the column(s) must be unique in at least one
# participating table!

# Ensure uniqueness:
airlines %>%
  count(carrier)

airlines %>%
  count(carrier) %>%
  count(n)

planes %>%
  count(tailnum) %>%
  count(n)

# dm shortcut:
planes %>%
  check_key(tailnum)

try(
  planes %>%
    check_key(engines)
)

airports %>%
  check_key(faa)

# Why is name not a key candidate for airports?
try(
  airports %>%
    check_key(name)
)

# Friendly description
airports %>%
  enum_pk_candidates()

# Cleanup
rm(t1, t2)

temp_folder <- withr::local_tempdir()
input <- knitr::current_input(dir = TRUE)
input_dir <- dirname(input)
final_dir <- file.path(input_dir, "cheatsheet")
withr::with_dir(
  temp_folder,
  {
    html_path <- withr::local_tempfile(fileext = ".html")
    rmd_path <- withr::local_tempfile(fileext = ".rmd")

    original <- brio::read_lines(input)
    # remove the part reading this very script to avoid an infinite loop :-s
    brio::write_lines(original[-c((length(original) - 4):length(original))], rmd_path)

    withr::with_envvar(
      new = list("CHEATSHEET" = "blabla"),
      {
        rmarkdown::render(
          rmd_path,
          output_format = rmarkdown::html_document(
            self_contained = FALSE,
            section_divs = FALSE
          ),
          output_file = html_path
        )
      }
    )

    html <- xml2::read_html(html_path)
    title <- xml2::xml_find_first(html, ".//h1") |> xml2::xml_text()
    subtitle <- xml2::xml_find_first(html, ".//blockquote/p") |> xml2::xml_text()

    # figures ----------

    img <- xml2::xml_find_all(html, ".//img")
    fix_src <- function(img) {
      if (grepl("shiny.png", xml2::xml_attr(img, "src"))) {
        xml2::xml_attr(img, "src") <- "cheatsheet-figures/shiny.png"
      } else {
        xml2::xml_attr(img, "src") <- sub(
          ".*man/figures/cheatsheet",
          "cheatsheet-figures",
          xml2::xml_attr(img, "src")
        )
      }
    }
    purrr::walk(img, fix_src)

    # content ---------
    find_col <- function(col) {
      xml2::xml_find_first(
        html,
        sprintf(
          "//div[@class='%s']",
          col
        )
      ) |>
        xml2::xml_contents() |>
        as.character() |>
        paste0(collapse = "\n")
    }
    col11 <- find_col("page1col1")
    col12 <- find_col("page1col2")
    col13 <- find_col("page1col3")
    col21 <- find_col("page2col1")
    col22 <- find_col("page2col2")
    col23 <- find_col("page2col3")

    # deps ------------
    bs_theme <- bslib::bs_theme(
      `font-size-base` = ".6rem",
      `h2-font-size` = ".9rem",
      `h3-font-size` = ".7rem",
      `h4-font-size` = ".7rem",
      `headings-color` = "#6366f1",
      `primary` = "#6366f1"
    )

    deps <- bslib::bs_theme_dependencies(bs_theme)
    fs::dir_create(file.path("cheatsheet", "deps"))
    deps <- lapply(deps, htmltools::copyDependencyToDir, file.path("cheatsheet", "deps"))
    deps <- lapply(deps, htmltools::makeDependencyRelative, getwd())
    deps <- htmltools::renderDependencies(deps, srcType = "file")

    # css -------
    css <- system.file("cheatsheet.css", package = "dm")
    file.copy(css, "cheatsheet.css")

    # logo -------
    fs::dir_copy(file.path(input_dir, "cheatsheet-figures"), "cheatsheet-figures")
    logo <- file.path(input_dir, "cheatsheet-figures", "logo.svg")
    dir.create("cheatsheet")
    file.copy(logo, file.path("cheatsheet", "logo.svg"))
    file.copy(logo, "logo.svg")
    file.copy(
      file.path(input_dir, "logo.png"),
      "logo.png"
    )

    # render --------------
    template <- paste0(
      brio::read_lines(system.file("cheatsheet-template.html", package = "dm")),
      collapse = ""
    )
    rendered <- whisker::whisker.render(template)
    brio::write_lines(rendered, "cheatsheet-printable.html")
    fs::dir_copy(getwd(), final_dir)
  }
)

# Demo for presentation at BaselR, December 2019

# {dm} facilitates working with multiple tables

options(tibble.print_min = 6)
options(tibble.print_max = 6)
options(rlang_backtrace_on_error = "none")

##
##
##
## Greeting
## --------------------------------------------------------------------
##
##
##

# Teaser
dm::dm_nycflights13(cycle = TRUE) %>%
  dm::dm_draw()

# Poll: Who is familiar with the {dplyr} package
#       (grammar of data manipulation)?

# Poll: Who is familiar with the {dbplyr} package
#       (using {dplyr} with databases)?

# Poll: Who has worked with databases?

# Poll: Who has worked with a software that has
#       a concept of "THE DATASET"?

# Poll: Who uses more than one table/data frame
#       at the same time?

##
##
##
## Why?
## --------------------------------------------------------------------
##
##
##

library(nycflights13)

# Example dataset: tables linked with each other
?flights

flights

##
##
##
## Data model
## --------------------------------------------------------------------
##
##
##

library(tidyverse)
flights_base <-
  flights %>%
  select(year, month, day, carrier, tailnum, origin, dest, time_hour)
flights_base

# `carrier` column also present in `airlines`, this table contains
# additional information
airlines

flights_base %>%
  left_join(airlines)

##
##
##
## Keys
## --------------------------------------------------------------------
##
##
##

# `carrier` is a "primary key" in `airlines`
any(duplicated(airlines$carrier))

# `carrier` is a "foreign key" in `flights` into `airlines`
all(flights$carrier %in% airlines$carrier)

##
##
##
## Single source of truth
## --------------------------------------------------------------------
##
##
##

# Update in one single location
airlines[airlines$carrier == "UA", "name"] <-
  "United broke my guitar"

airlines %>%
  filter(carrier == "UA")

# ...propagates to all related records
flights_base %>%
  left_join(airlines)

##
##
##
## Goal: work with one table
## --------------------------------------------------------------------
##
##
##

flights_base %>%
  left_join(airlines, by = "carrier") %>%
  left_join(planes, by = "tailnum") %>%
  left_join(weather, by = c("origin", "time_hour")) %>%
  left_join(airports, by = c("origin" = "faa")) %>%
  left_join(airports, by = c("dest" = "faa"))

##
##
##
## PROBLEMS (see appendix)
## --------------------------------------------------------------------
##
## - wrong keys
## - data mismatches
## - relationship unclear
## - combinatorial explosion
##
##
##

##
##
##
## Data model object
## --------------------------------------------------------------------
##
##
##

library(dm)

# Compound object: tables, relationships, data
dm_flights <- dm_nycflights13(cycle = TRUE)
dm_flights

dm_flights %>%
  dm_draw()

# Selection of tables
dm_flights %>%
  dm_select_tbl(flights, airlines) %>%
  dm_draw()

dm_flights %>%
  dm_select_tbl(airports, airlines) %>%
  dm_draw()

try(
  dm_flights %>%
    dm_select_tbl(bogus)
)

# Accessing tables
dm_flights %>%
  pull_tbl(airlines)

dm_flights$airlines

# Table names
dm_flights %>%
  src_tbls()

# NB: [, $, [[ and names() also work

##
##
##
##
##
##
## THREE USE CASES:
## --------------------------------------------------------------------
##
## 1. Work with a prepared dm object
## 2. Build a data model for your own data
## 3. Publish a dm to a relational database (and load from it)
##
##
##
##

##
##
##
##
##
##
##
##
##
##
##
##
## USE CASE 1: Work with a prepared dm object
## --------------------------------------------------------------------
##
##
##

##
##
##
##
##
## Joining two tables
## --------------------------------------------------------------------
##
##
##

dm_flights <- dm_nycflights13()
dm_flights %>%
  dm_draw()

dm_flights %>%
  dm_join_to_tbl(airlines, flights)

try(
  dm_flights %>%
    dm_join_to_tbl(airports, airlines)
)


##
##
##
## Joining many tables
## --------------------------------------------------------------------
##
##
##

dm_flights %>%
  dm_flatten_to_tbl(flights)

##
##
##
## Filtering for data models
## --------------------------------------------------------------------
##
##
##

# Filtering on a table returns a dm object
# with
# 1. the filter applied to the table in question
# 2. the filter condition(s) stored
dm_flights %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.")

# the resulting `dm` can then be filtered on another table
dm_flights %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl")

# ... and stored in another dm variable
delta_non_jfk_january <-
  dm_flights %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(planes, year < 2000) %>%
  dm_filter(flights, month == 1)
delta_non_jfk_january

# Querying a table applies the filters via semi-joins
# along the FK constraints to the requested table
delta_non_jfk_january %>%
  tbl("planes")

# FIXME: Can this work without applying all filters?

delta_non_jfk_january %>%
  dm_apply_filters() %>%
  dm_join_to_tbl(flights, airlines)

delta_non_jfk_january %>%
  dm_flatten_to_tbl(flights)

##
##
##
## NEW NEW NEW: Data manipulation in a dm
## --------------------------------------------------------------------
##
##
##

# A single table of a `dm` can be activated (or zoomed to),
# and subsequently be manipulated by many {dplyr}-verbs.
# Eventually, either the original table can be updated
# or the manipulated table can be inserted as a new table.

# The print output for a `dm_zoomed` looks very much like that from a normal `tibble`.
dm_flights %>%
  dm_zoom_to(flights)

# Many {dplyr} verbs work on zoomed tables:
dm_flights %>%
  dm_zoom_to(flights) %>%
  mutate(am_pm_dep = if_else(dep_time < 1200, "am", "pm")) %>%
  select(year:dep_time, am_pm_dep, everything())

# Put back into the dm:
dm_flights %>%
  dm_zoom_to(flights) %>%
  mutate(am_pm_dep = if_else(dep_time < 1200, "am", "pm")) %>%
  select(year:dep_time, am_pm_dep, everything()) %>%
  dm_update_zoomed()

# Immutable objects, like in {dplyr}
dm_flights

# Creation of a summary table:
dm_with_summary <-
  dm_flights %>%
  dm_zoom_to(flights) %>%
  count(origin) %>%
  dm_insert_zoomed("origin_count")

dm_with_summary$origin_count

dm_with_summary %>%
  dm_draw()

# All relationships still available in the summary are retained:
dm_flights %>%
  dm_zoom_to(flights) %>%
  count(carrier, origin) %>%
  dm_insert_zoomed("origin_carrier_count") %>%
  dm_draw()


##
##
##
## USE CASE 2: Build a data model for your own data
## --------------------------------------------------------------------
##
##
##

##
##
##
##
##
##
##
## Build up data model from scratch
## --------------------------------------------------------------------
##
##
##

# Use `dm()` with a syntax similar to `tibble()`:
nycflights13_tbl <- dm(airlines, airports, flights, planes, weather)
nycflights13_tbl

nycflights13_tbl %>%
  dm_draw()

# Alternatively, start from an empty `dm`
# and add tables via `dm_add_tbl()`:
dm() %>%
  dm_add_tbl(airlines, airports, flights, planes, weather)

# Tables are not connected yet:
nycflights13_tbl %>%
  dm_draw()

# Adding primary keys:
nycflights13_pk <-
  nycflights13_tbl %>%
  dm_add_pk(planes, tailnum) %>%
  dm_add_pk(airports, faa) %>%
  dm_add_pk(airlines, carrier)

nycflights13_pk %>%
  dm_draw()

# FIXME: Model weak constraints, show differently in diagram (#4)

# Adding foreign keys
nycflights13_fk <-
  nycflights13_pk %>%
  dm_add_fk(flights, tailnum, planes) %>%
  dm_add_fk(flights, origin, airports) %>%
  dm_add_fk(flights, dest, airports) %>%
  dm_add_fk(flights, carrier, airlines, check = TRUE)

nycflights13_fk %>%
  dm_draw()

# Color it!
dm_get_available_colors()

nycflights13_base <-
  nycflights13_fk %>%
  cdm_set_colors(
    airlines = ,
    planes = ,
    weather = ,
    airports = "blue"
  )

nycflights13_base %>%
  dm_draw()


##
##
##
## Use data manipulation to understand and establish relationships
## --------------------------------------------------------------------
##
##
##

# Determine key candidates
zoomed_weather <- dm_zoom_to(nycflights13_base, weather)
zoomed_weather

# `enum_pk_candidates()` works for both `tibbles` and `dm_zoomed`
enum_pk_candidates(zoomed_weather)

enum_pk_candidates(zoomed_weather) %>%
  count(candidate)

# It's tricky:
zoomed_weather %>%
  unite("slot_id", origin, year, month, day, hour, remove = FALSE) %>%
  count(slot_id) %>%
  filter(n > 1)

zoomed_weather %>%
  count(origin, time_hour) %>%
  filter(n > 1)

zoomed_weather %>%
  count(origin, format(time_hour)) %>%
  filter(n > 1)

# This looks like a good candidate:
zoomed_weather %>%
  count(origin, format(time_hour, tz = "UTC")) %>%
  filter(n > 1)

# FIXME: Support compound keys (#3)

# Currently, we need to create surrogate keys:
nycflights13_weather_link <-
  zoomed_weather %>%
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  unite("origin_slot_id", origin, time_hour_fmt) %>%
  # here the original 'weather' table is updated with the manipulated one
  dm_update_zoomed() %>%
  # here we are adding a PK for the "enhanced" weather table
  dm_add_pk(weather, origin_slot_id)

nycflights13_weather_link$weather

nycflights13_weather_link %>%
  dm_draw()

# FIXME: zoom to multiple tables

nycflights13_weather_flights_link <-
  dm_zoom_to(nycflights13_weather_link, flights) %>%
  # same procedure with `flights` table
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  # for flights we need to keep the column `origin`,
  # since it is a FK pointing to `airports`
  unite("origin_slot_id", origin, time_hour_fmt, remove = FALSE) %>%
  select(origin_slot_id, everything(), -time_hour_fmt) %>%
  dm_update_zoomed()

# `dm_enum_fk_candidates()` of a `dm` gives info
# about potential FK columns from one table to another
dm_enum_fk_candidates(nycflights13_weather_flights_link, flights, weather)

# well, it's almost perfect, let's add the FK anyway...

nycflights13_perfect <-
  nycflights13_weather_flights_link %>%
  dm_add_fk(flights, origin_slot_id, weather)

nycflights13_perfect %>%
  dm_draw()

# What are the missings?
nycflights13_perfect %>%
  dm_zoom_to(flights) %>%
  anti_join(weather) %>%
  count(origin_slot_id)

# Create table of aggregates, and insert it into the dm
nycflights13_perfect %>%
  dm_zoom_to(flights) %>%
  count(origin) %>%
  dm_insert_zoomed("flights_agg") %>%
  dm_draw()


##
##
##
## USE CASE 3: Publish a dm to a relational database (and load from it)
## --------------------------------------------------------------------
##
##
##

##
##
##
##
##
## Copy to database
## --------------------------------------------------------------------
##
##
##

# All operations are designed to work locally and on the database
dm_flights_sqlite <-
  dm_flights %>%
  copy_dm_to(
    dbplyr::src_memdb(),
    .,
    unique_table_names = TRUE,
    set_key_constraints = FALSE
  )

dm_flights_sqlite

dm_flights_sqlite %>%
  dm_draw()

dm_flights_sqlite %>%
  dm_get_tables() %>%
  map(dbplyr::sql_render)

dm_flights_sqlite %>%
  dm_join_to_tbl(airlines, flights) %>%
  dbplyr::sql_render()

dm_flights_sqlite %>%
  dm_flatten_to_tbl(flights) %>%
  dbplyr::sql_render()

# Filtering on the database
dm_flights_sqlite %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(flights, day == 1) %>%
  dm_apply_filters_to_tbl(flights)

# ... and the corresponding SQL statement and query plan
dm_flights_sqlite %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(flights, day == 1) %>%
  dm_apply_filters_to_tbl(flights) %>%
  dbplyr::sql_render()


##
##
##
## Import a dm from a database, including key constraints
## --------------------------------------------------------------------
##
##
##

try({
  con_pq <- DBI::dbConnect(RPostgres::Postgres())

  # FIXME: Schema support

  # Off by default, to ensure that no tables are accidentally deleted
  if (FALSE) {
    walk(
      names(dm_flights),
      ~ DBI::dbExecute(
        con_pq,
        paste0("DROP TABLE IF EXISTS ", ., " CASCADE")
      )
    )
  }

  # Import
  dm_flights_pq <-
    dm_flights %>%
    dm_filter(planes, TRUE) %>%
    dm_filter(flights, month == 1, day == 1) %>%
    copy_dm_to(con_pq, ., temporary = FALSE)

  dm_flights_from_pq <-
    dm_learn_from_db(con_pq)

  dm_flights_from_pq %>%
    dm_draw()
})


##
##
##
##
## THREE USE CASES:
## --------------------------------------------------------------------
##
## 1. Work with a prepared dm object
## 2. Build a data model for your own data
## 3. Publish a dm to a relational database (and load from it)
##
##
##
##

##
##
##
##
##
##
##
##
##
##
##
##
## Appendix
## ====================================================================
##
##
##

##
##
##
## Analogy?
## --------------------------------------------------------------------
##
##
##

# Analogy: parallel vectors
# Multiple parallel vectors can be combined into a data frame.
# Multiple related tables can be combined into a dm.
x <- 1:5
x
y <- x + 1
y
z <- diff(y)
z

try(
  tibble(x = 1:5) %>%
    mutate(y = x + 1) %>%
    mutate(z = diff(y))
)

##
##
##
## Pitfall: wrong keys
## --------------------------------------------------------------------
##
##
##

# Same for airplanes?
planes

flights_base %>%
  left_join(planes)

flights_base %>%
  left_join(planes) %>%
  count(is.na(type))

# Take a closer look at the join
flights_base %>%
  left_join(planes, by = "tailnum")

flights_base %>%
  left_join(planes, by = "tailnum") %>%
  count(is.na(type))

##
##
##
## Pitfall: data mismatches
## --------------------------------------------------------------------
##
##
##

flights_base %>%
  left_join(planes, by = "tailnum") %>%
  summarize(.by = carrier, mismatch_rate = mean(is.na(type))) %>%
  filter(mismatch_rate > 0) %>%
  ggplot(aes(x = carrier, y = mismatch_rate)) +
  geom_col()

##
##
##
## Pitfall: relationship unclear
## --------------------------------------------------------------------
##
##
##

# Same for airports?
airports

try(
  flights_base %>%
    left_join(airports)
)

# Need to specify join variables!
flights_base %>%
  left_join(airports, by = c("origin" = "faa"))

# cleanup
rm(airlines)

##
##
##
## Pitfall: combinatorial explosion
## --------------------------------------------------------------------
##
##
##

# Row identifiers
t1 <- tibble(a = 1, b = letters[1:3])
t1
t2 <- tibble(a = 1, c = 1:2)
t2

# What happens here?
left_join(t1, t2)

# When joining, the column(s) must be unique in at least one
# participating table!

# Ensure uniqueness:
airlines %>%
  count(carrier)

airlines %>%
  count(carrier) %>%
  count(n)

planes %>%
  count(tailnum) %>%
  count(n)

# dm shortcut:
planes %>%
  check_key(tailnum)

try(
  planes %>%
    check_key(engines)
)

airports %>%
  check_key(faa)

# Why is name not a key candidate for airports?
try(
  airports %>%
    check_key(name)
)

# Friendly description
airports %>%
  enum_pk_candidates()

# Cleanup
rm(t1, t2)

# Demo for presentation at e-Rum in Milano,
# June 2020

# {dm}: Relational data models in R

# Slides at https://krlmlr.github.io/dm-slides/

##
##
##
## RELATIONAL DATA MODELS WITH R
## -------------------------------------------------------
##
##
##

# The dm package helps

## 1. **consume** data models
## 2. **build** data models
## 3. **deploy** data models

### (on database or locally)

##
## ----setup, include = FALSE----
library(tidyverse)
library(dbplyr)
library(dm)
set.seed(20200314)


## ----connect, cache = FALSE----
library(DBI) # <<

mydb <- dbConnect(
  RMariaDB::MariaDB(),
  username = "guest",
  password = "ctu-relational",
  dbname = "Financial_ijs",
  host = "relational.fel.cvut.cz"
)


## ----dm, cache = FALSE, echo = FALSE----
mydm <- dm_from_src(mydb, learn_keys = FALSE)


## ----echo = FALSE------------
financial_dm <- function(mydb) {
  mydm %>%
    dm_add_pk(districts, id) %>%
    dm_add_pk(accounts, id) %>%
    dm_add_pk(clients, id) %>%
    dm_add_pk(loans, id) %>%
    dm_add_pk(orders, id) %>%
    dm_add_pk(trans, id) %>%
    dm_add_pk(disps, id) %>%
    dm_add_pk(cards, id) %>%
    dm_add_fk(
      loans,
      account_id,
      accounts
    ) %>%
    dm_add_fk(
      orders,
      account_id,
      accounts
    ) %>%
    dm_add_fk(
      trans,
      account_id,
      accounts
    ) %>%
    dm_add_fk(
      disps,
      account_id,
      accounts
    ) %>%
    dm_add_fk(
      disps,
      client_id,
      clients
    ) %>%
    dm_add_fk(
      accounts,
      district_id,
      districts
    ) %>%
    dm_add_fk(
      clients,
      district_id,
      districts
    ) %>%
    dm_add_fk(
      cards,
      disp_id,
      disps
    ) %>%
    dm_rm_tbl(tkeys) %>%
    dm_set_colors(orange = accounts)
}


## ----dm-load, cache = FALSE----
my_dm <- financial_dm()


## ----dm-visualize------------
dm_draw(my_dm)


## ----dm-squash---------------
dm_squash_to_tbl(my_dm, loans)


## ----dm-squash-vis, echo = FALSE----
my_dm %>%
  dm_set_colors(darkblue = loans, darkgreen = c(accounts, districts)) %>%
  dm_draw()


## ----flights-----------------
library(nycflights13)

flights_dm <-
  dm(flights, planes) %>%
  dm_add_pk(planes, tailnum) %>%
  dm_add_fk(flights, tailnum, planes)


## ----flights-examine---------
dm_draw(flights_dm)


## ----sqlite, cache = FALSE, highlight.output = 2----
sqlite <- dbConnect(RSQLite::SQLite(), dbname = ":memory:")

flights_dm_production <- copy_dm_to(sqlite, flights_dm)
flights_dm_production

# Demo for presentation at the 16th Berlin R meetup

# {dm} facilitates working with multiple tables

options(tibble.print_min = 6)
options(tibble.print_max = 6)

##
##
##
## Why?
## --------------------------------------------------------------------
##
##
##

library(nycflights13)

# Example dataset: tables linked with each other
?flights
?airports
?airlines
?planes
?weather

library(tidyverse)
flights_base <-
  flights %>%
  select(year, month, day, carrier, tailnum, origin, dest)
flights_base

# carrier column also present in `airlines`, this table contains
# additional information
airlines

flights_base %>%
  left_join(airlines)

# single source of truth: updating in one single location
airlines[airlines$carrier == "UA", "name"] <- "United broke my guitar"

# ...propagates to related records
flights_base %>%
  left_join(airlines)

# Same for airplanes?
planes

flights_base %>%
  left_join(planes)

flights_base %>%
  left_join(planes) %>%
  count(is.na(type))

# Take a closer look at the join
flights_base %>%
  left_join(planes, by = "tailnum")

# Same for airports?
airports

try(
  flights_base %>%
    left_join(airports)
)

# Need to specify join variables!
flights_base %>%
  left_join(airports, by = c("origin" = "faa"))

# cleanup
rm(airlines)

##
##
##
## Keys
## --------------------------------------------------------------------
##
##
##

# Row identifiers
t1 <- tibble(a = 1, b = letters[1:3])
t1
t2 <- tibble(a = 1, c = 1:2)
t2

# What happens here?
left_join(t1, t2)

# When joining, the column(s) must be unique in at least one
# participating table!

# Ensure uniqueness:
airlines %>%
  count(carrier)

airlines %>%
  count(carrier) %>%
  count(n)

planes %>%
  count(tailnum) %>%
  count(n)

# dm shortcut:
planes %>%
  dm::check_key(tailnum)

try(
  planes %>%
    dm::check_key(engines)
)

airports %>%
  dm::check_key(faa)

# FIXME: add dm function that explains why not key candidate

# Why is name not a key candidate for airports?
try(
  airports %>%
    dm::check_key(name)
)

airports %>%
  add_count(name) %>%
  filter(n > 1) %>%
  arrange(name)

# Cleanup
rm(t1, t2)

##
##
##
## Data model
## --------------------------------------------------------------------
##
##
##

library(dm)

# Compound object: tables, relationships, data
dm_nycflights13(cycle = TRUE)

dm_nycflights13(cycle = TRUE) %>%
  dm_draw()

# Selection of tables
dm_nycflights13(cycle = TRUE) %>%
  dm_select_tbl(flights, airlines) %>%
  dm_draw()

dm_nycflights13(cycle = TRUE) %>%
  dm_select_tbl(airports, airlines) %>%
  dm_draw()

try(
  dm_nycflights13() %>%
    dm_select_tbl(bogus)
)

# Accessing tables
dm_nycflights13() %>%
  tbl("airlines")

# Table names
dm_nycflights13() %>%
  src_tbls()

# NB: [, $, [[ and names() also work

# Analogy: parallel vectors in the global environment
x <- 1:5
y <- x + 1
z <- x * y
w <- map(z, ~ runif(.))
tibble(x, y, z, w)

tibble(x = 1:5) %>%
  mutate(y = x + 1) %>%
  mutate(z = x * y) %>%
  mutate(w = map(z, ~ runif(.)))

##
##
##
## Filtering for data models
## --------------------------------------------------------------------
##
##
##

dm_nycflights13()

# Filtering on a table returns a dm object with the filter condition(s) stored
(dm_nyc_filtered <-
  dm_nycflights13() %>%
  dm_filter(airlines, carrier == "AA"))

# Apply all filters and retrieve an "updated" `dm`
dm_nyc_filtered %>%
  dm_apply_filters()

# If a filter condition is phrased wrongly it will only fail, once the filter is being applied
(dm_nyc_fail <-
  dm_nycflights13() %>%
  dm_filter(airports, origin == "EWR"))
try(
  tbl(dm_nyc_fail, "flights")
)
# Mind: when accessing table from a `dm` (using one of: `tbl()`, `[[.dm()`, `$.dm()`),
# only the necessary filter conditions are applied:
tbl(dm_nyc_fail, "weather")

dm_nycflights13() %>%
  dm_filter(flights, origin == "EWR") %>%
  dm_apply_filters()

# ... which then can be filtered on another table
dm_nycflights13() %>%
  dm_filter(airlines, name == "American Airlines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_apply_filters()

aa_non_jfk_january <-
  dm_nycflights13() %>%
  dm_filter(airlines, name == "American Airlines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(flights, month == 1) %>%
  dm_apply_filters()
aa_non_jfk_january

# ... and processed further
aa_non_jfk_january %>%
  tbl("planes")

##
##
##
## Copy to database
## --------------------------------------------------------------------
##
##
##

# FIXME: SQLite with many rows

# FIXME: Make work with planes table

# FIXME: remove all_connected = TRUE for now, redo example above

# All operations are designed to work locally and on the database
nycflights13_sqlite <-
  dm_nycflights13() %>%
  dm_select_tbl(-planes) %>%
  dm_filter(flights, month == 1) %>%
  dm_apply_filters() %>%
  copy_dm_to(dbplyr::src_memdb(), ., unique_table_names = TRUE)

nycflights13_sqlite

nycflights13_sqlite %>%
  dm_draw()

nycflights13_sqlite %>%
  dm_get_tables() %>%
  map(dbplyr::sql_render)

# Filtering on the database
nycflights13_sqlite %>%
  dm_filter(airlines, name == "American Airlines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(flights, day == 1) %>%
  tbl("flights")

# ... and the corresponding SQL statement
nycflights13_sqlite %>%
  dm_filter(airlines, name == "American Airlines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(flights, day == 1) %>%
  tbl("flights") %>%
  dbplyr::sql_render()

##
##
##
## Joining tables
## --------------------------------------------------------------------
##
##
##

dm_nycflights13() %>%
  dm_join_to_tbl(airlines, flights, join = left_join)

dm_nycflights13() %>%
  dm_join_to_tbl(flights, airlines, join = left_join)

nycflights13_sqlite %>%
  dm_join_to_tbl(airlines, flights, join = left_join)

aa_non_jfk_january %>%
  dm_join_to_tbl(flights, airlines, join = left_join)

# FIXME: Multi-joins

try(
  dm_nycflights13() %>%
    dm_join_to_tbl(airports, airlines, join = left_join)
)

try(
  dm_nycflights13() %>%
    dm_join_to_tbl(flights, airports, airlines, join = left_join)
)

##
##
##
## Build up data model from scratch
## --------------------------------------------------------------------
##
##
##

# Linking the weather table

# Determine key candidates
weather %>%
  enum_pk_candidates()

weather %>%
  enum_pk_candidates() %>%
  count(candidate)

# It's tricky:
weather %>%
  unite("slot_id", origin, year, month, day, hour, remove = FALSE) %>%
  count(slot_id) %>%
  filter(n > 1)

weather %>%
  count(origin, time_hour) %>%
  filter(n > 1)

weather %>%
  count(origin, format(time_hour)) %>%
  filter(n > 1)

# This looks like a good candidate:
weather %>%
  count(origin, format(time_hour, tz = "UTC")) %>%
  filter(n > 1)

# FIXME: Support compound keys (#3)

# Currently, we need to create surrogate keys:
weather_link <-
  weather %>%
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  unite("origin_slot_id", origin, time_hour_fmt, remove = FALSE)

flights_link <-
  flights %>%
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  unite("origin_slot_id", origin, time_hour_fmt, remove = FALSE)

# one option to create a `dm` is to use `as_dm()`:
nycflights13_dm <- as_dm(list(
  airlines = airlines,
  airports = airports,
  flights = flights_link,
  planes = planes,
  weather = weather_link
))

# Copy to this environment
airlines_global <- airlines
airports_global <- airports
planes_global <- planes

global <-
  dm_from_src(src_df(env = .GlobalEnv))
global

global %>%
  dm_rename_tbl(
    airlines = airlines_global,
    airports = airports_global,
    planes = planes_global,
    flights = flights_link,
    weather = weather_link
  ) %>%
  dm_select_tbl(airlines, airports, planes, flights, weather)

# or better:
nycflights13_tbl <-
  global %>%
  dm_select_tbl(
    airlines = airlines_global,
    airports = airports_global,
    planes = planes_global,
    flights = flights_link,
    weather = weather_link
  )


nycflights13_tbl

nycflights13_tbl %>%
  dm_draw()

# Adding primary keys
nycflights13_pk <-
  nycflights13_tbl %>%
  dm_add_pk(weather, origin_slot_id) %>%
  dm_add_pk(planes, tailnum) %>%
  dm_add_pk(airports, faa) %>%
  dm_add_pk(airlines, carrier)

nycflights13_pk %>%
  dm_draw()

# FIXME: Model weak constraints, show differently in diagram (#4)

# Adding foreign keys
nycflights13_fk <-
  nycflights13_pk %>%
  dm_add_fk(flights, origin_slot_id, weather, check = FALSE) %>%
  dm_add_fk(flights, tailnum, planes, check = FALSE) %>%
  dm_add_fk(flights, origin, airports) %>%
  dm_add_fk(flights, dest, airports, check = FALSE) %>%
  dm_add_fk(flights, carrier, airlines)

nycflights13_fk %>%
  dm_draw()

# Color it!
dm_get_available_colors()

nycflights13_fk %>%
  dm_set_colors(airlines = , planes = , weather = , airports = "blue") %>%
  dm_draw()

##
##
##
## Import a dm from a database, including key constraints
## --------------------------------------------------------------------
##
##
##

try({
  # Import
  dm_pq <-
    dm_nycflights13() %>%
    dm_select_tbl(-planes) %>%
    dm_filter(flights, month == 1) %>%
    copy_dm_to(src_postgres(), ., temporary = FALSE)

  dm_from_pq <-
    dm_learn_from_db(src_postgres())
})

# Demo for presentation at the Zurich R meetup, September 2019

# {dm} facilitates working with multiple tables

options(tibble.print_min = 6)
options(tibble.print_max = 6)
options(rlang_backtrace_on_error = "none")

##
##
##
## Greeting
## --------------------------------------------------------------------
##
##
##

# Teaser
dm::dm_nycflights13(cycle = TRUE) %>%
  dm::dm_draw()

# Poll: Who has worked with a software that has
#       a concept of "THE DATASET"?

# Poll: Who has worked with databases?

# Poll: Who uses more than one table/data frame
#       at the same time?

##
##
##
## Why?
## --------------------------------------------------------------------
##
##
##

library(nycflights13)

# Example dataset: tables linked with each other
?flights

flights

##
##
##
## Data model
## --------------------------------------------------------------------
##
##
##

library(tidyverse)
flights_base <-
  flights %>%
  select(year, month, day, carrier, tailnum, origin, dest, time_hour)
flights_base

# `carrier` column also present in `airlines`, this table contains
# additional information
airlines

flights_base %>%
  left_join(airlines)

##
##
##
## Keys
## --------------------------------------------------------------------
##
##
##

# `carrier` is a "primary key" in `airlines`
any(duplicated(airlines$carrier))

# `carrier` is a "foreign key" in `flights` into `airlines`
all(flights$carrier %in% airlines$carrier)

##
##
##
## Single source of truth
## --------------------------------------------------------------------
##
##
##

# Update in one single location
airlines[airlines$carrier == "UA", "name"] <-
  "United broke my guitar"

airlines %>%
  filter(carrier == "UA")

# ...propagates to all related records
flights_base %>%
  left_join(airlines)

##
##
##
## Goal: work with one table
## --------------------------------------------------------------------
##
##
##

flights_base %>%
  left_join(airlines, by = "carrier") %>%
  left_join(planes, by = "tailnum") %>%
  left_join(weather, by = c("origin", "time_hour")) %>%
  left_join(airports, by = c("origin" = "faa")) %>%
  left_join(airports, by = c("dest" = "faa"))

##
##
##
## PROBLEMS (see appendix)
## --------------------------------------------------------------------
##
## - wrong keys
## - data mismatches
## - relationship unclear
## - combinatorial explosion
##
##
##

##
##
##
## Data model object
## --------------------------------------------------------------------
##
##
##

library(dm)

# Compound object: tables, relationships, data
dm_flights <- dm_nycflights13(cycle = TRUE)
dm_flights

dm_flights %>%
  dm_draw()

# Selection of tables
dm_flights %>%
  dm_select_tbl(flights, airlines) %>%
  dm_draw()

dm_flights %>%
  dm_select_tbl(airports, airlines) %>%
  dm_draw()

try(
  dm_flights %>%
    dm_select_tbl(bogus)
)

# Accessing tables
dm_flights %>%
  tbl("airlines")

# Table names
dm_flights %>%
  src_tbls()

# NB: [, $, [[ and names() also work

##
##
##
## Joining two tables
## --------------------------------------------------------------------
##
##
##

dm_flights <- dm_nycflights13()
dm_flights %>%
  dm_draw()

dm_flights %>%
  dm_join_to_tbl(airlines, flights)

try(
  dm_flights %>%
    dm_join_to_tbl(airports, airlines)
)

##
##
##
## NEW NEW NEW: Joining many tables
## --------------------------------------------------------------------
##
##
##

dm_flights %>%
  dm_flatten_to_tbl(flights)

##
##
##
## Copy to database
## --------------------------------------------------------------------
##
##
##

# All operations are designed to work locally and on the database
dm_flights_sqlite <-
  dm_flights %>%
  copy_dm_to(
    dbplyr::src_memdb(),
    .,
    unique_table_names = TRUE,
    set_key_constraints = FALSE
  )

dm_flights_sqlite

dm_flights_sqlite %>%
  dm_draw()

dm_flights_sqlite %>%
  dm_get_tables() %>%
  map(dbplyr::sql_render)

dm_flights_sqlite %>%
  dm_join_to_tbl(airlines, flights) %>%
  dbplyr::sql_render()

dm_flights_sqlite %>%
  dm_flatten_to_tbl(flights) %>%
  dbplyr::sql_render()

# Filtering on the database
dm_flights_sqlite %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(flights, day == 1) %>%
  tbl("flights")

# ... and the corresponding SQL statement and query plan
dm_flights_sqlite %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(flights, day == 1) %>%
  tbl("flights") %>%
  dbplyr::sql_render()

##
##
##
## Filtering for data models
## --------------------------------------------------------------------
##
##
##

# Filtering on a table returns a dm object
# with the filter condition(s) stored
dm_flights %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.")

# ... which then can be filtered on another table
dm_flights %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl")

# ... and stored in another dm variable
delta_non_jfk_january <-
  dm_flights %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(planes, year < 2000) %>%
  dm_filter(flights, month == 1)
delta_non_jfk_january

# Querying a table applies the filters
delta_non_jfk_january %>%
  tbl("planes")

# FIXME: Can this work without applying all filters?

delta_non_jfk_january %>%
  dm_apply_filters() %>%
  dm_join_to_tbl(flights, airlines)

delta_non_jfk_january %>%
  dm_flatten_to_tbl(flights)


##
##
##
## Build up data model from scratch
## --------------------------------------------------------------------
##
##
##

# Linking the weather table

# Determine key candidates
weather %>%
  enum_pk_candidates()

weather %>%
  enum_pk_candidates() %>%
  count(candidate)

# It's tricky:
weather %>%
  unite("slot_id", origin, year, month, day, hour, remove = FALSE) %>%
  count(slot_id) %>%
  filter(n > 1)

weather %>%
  count(origin, time_hour) %>%
  filter(n > 1)

weather %>%
  count(origin, format(time_hour)) %>%
  filter(n > 1)

# This looks like a good candidate:
weather %>%
  count(origin, format(time_hour, tz = "UTC")) %>%
  filter(n > 1)

# FIXME: Support compound keys (#3)

# Currently, we need to create surrogate keys:
weather_link <-
  weather %>%
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  unite("origin_slot_id", origin, time_hour_fmt, remove = FALSE)

weather_link

flights_link <-
  flights %>%
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  unite("origin_slot_id", origin, time_hour_fmt, remove = FALSE)

flights_link

# one option to create a `dm` is to use `as_dm()`:
nycflights13_tbl <- as_dm(list(
  airlines = airlines,
  airports = airports,
  flights = flights_link,
  planes = planes,
  weather = weather_link
))

nycflights13_tbl

nycflights13_tbl %>%
  dm_draw()

# Adding primary keys
nycflights13_pk <-
  nycflights13_tbl %>%
  dm_add_pk(weather, origin_slot_id) %>%
  dm_add_pk(planes, tailnum) %>%
  dm_add_pk(airports, faa) %>%
  dm_add_pk(airlines, carrier)

nycflights13_pk %>%
  dm_draw()

# FIXME: Model weak constraints, show differently in diagram (#4)

# Adding foreign keys
nycflights13_fk <-
  nycflights13_pk %>%
  dm_add_fk(flights, origin_slot_id, weather) %>%
  dm_add_fk(flights, tailnum, planes) %>%
  dm_add_fk(flights, origin, airports) %>%
  dm_add_fk(flights, dest, airports) %>%
  dm_add_fk(flights, carrier, airlines)

nycflights13_fk %>%
  dm_draw()

# Color it!
dm_get_available_colors()

nycflights13_fk %>%
  dm_set_colors(
    airlines = ,
    planes = ,
    weather = ,
    airports = "blue"
  ) %>%
  dm_draw()

##
##
##
## Import a dm from a database, including key constraints
## --------------------------------------------------------------------
##
##
##

try({
  con_pq <- DBI::dbConnect(RPostgres::Postgres())

  # FIXME: Schema support

  # Off by default, to ensure that no tables are accidentally deleted
  if (FALSE) {
    walk(
      names(dm_flights),
      ~ DBI::dbExecute(
        con_pq,
        paste0("DROP TABLE IF EXISTS ", ., " CASCADE")
      )
    )
  }

  # Import
  dm_flights_pq <-
    dm_flights %>%
    dm_filter(planes, TRUE) %>%
    dm_filter(flights, month == 1, day == 1) %>%
    copy_dm_to(con_pq, ., temporary = FALSE)

  dm_flights_from_pq <-
    dm_learn_from_db(con_pq)

  dm_flights_from_pq %>%
    dm_draw()
})

##
##
##
## Appendix
## ====================================================================
##
##
##

##
##
##
## Analogy?
## --------------------------------------------------------------------
##
##
##

# Analogy: parallel vectors
# Multiple parallel vectors can be combined into a data frame.
# Multiple related tables can be combined into a dm.
x <- 1:5
x
y <- x + 1
y
z <- diff(y)
z

try(
  tibble(x = 1:5) %>%
    mutate(y = x + 1) %>%
    mutate(z = diff(y))
)

##
##
##
## Pitfall: wrong keys
## --------------------------------------------------------------------
##
##
##

# Same for airplanes?
planes

flights_base %>%
  left_join(planes)

flights_base %>%
  left_join(planes) %>%
  count(is.na(type))

# Take a closer look at the join
flights_base %>%
  left_join(planes, by = "tailnum")

flights_base %>%
  left_join(planes, by = "tailnum") %>%
  count(is.na(type))

##
##
##
## Pitfall: data mismatches
## --------------------------------------------------------------------
##
##
##

flights_base %>%
  left_join(planes, by = "tailnum") %>%
  summarize(.by = carrier, mismatch_rate = mean(is.na(type))) %>%
  filter(mismatch_rate > 0) %>%
  ggplot(aes(x = carrier, y = mismatch_rate)) +
  geom_col()

##
##
##
## Pitfall: relationship unclear
## --------------------------------------------------------------------
##
##
##

# Same for airports?
airports

try(
  flights_base %>%
    left_join(airports)
)

# Need to specify join variables!
flights_base %>%
  left_join(airports, by = c("origin" = "faa"))

# cleanup
rm(airlines)

##
##
##
## Pitfall: combinatorial explosion
## --------------------------------------------------------------------
##
##
##

# Row identifiers
t1 <- tibble(a = 1, b = letters[1:3])
t1
t2 <- tibble(a = 1, c = 1:2)
t2

# What happens here?
left_join(t1, t2)

# When joining, the column(s) must be unique in at least one
# participating table!

# Ensure uniqueness:
airlines %>%
  count(carrier)

airlines %>%
  count(carrier) %>%
  count(n)

planes %>%
  count(tailnum) %>%
  count(n)

# dm shortcut:
planes %>%
  check_key(tailnum)

try(
  planes %>%
    check_key(engines)
)

airports %>%
  check_key(faa)

# Why is name not a key candidate for airports?
try(
  airports %>%
    check_key(name)
)

# NEW NEW NEW: Friendly description
airports %>%
  enum_pk_candidates()

# Cleanup
rm(t1, t2)

library(tidyverse)
pkgload::load_all()

#src <- dm::dm_get_src(dm_financial())
#src <- test_src_maria()
#src <- test_src_postgres()
src <- test_src_mssql()

# DuckDB doesn't have references
#src <- test_src_duckdb()

con <- src$con

if (FALSE) {
  try(DBI::dbRemoveTable(con, "airlines"))
  try(DBI::dbRemoveTable(con, "airports"))
  try(DBI::dbRemoveTable(con, "planes"))
  try(DBI::dbRemoveTable(con, "flights"))
  try(DBI::dbRemoveTable(con, "weather"))
  dm_nycflights13() %>%
    dm_zoom_to(flights) %>%
    semi_join(planes) %>%
    dm_update_zoomed() %>%
    copy_dm_to(con, ., temporary = FALSE)
}

src_pg <- test_src_postgres()
con_pg <- src_pg$con

dm_meta(con)$key_column_usage
dm_meta(con_pg)$key_column_usage
dm_meta(con)$constraint_column_usage
dm_meta(con_pg)$constraint_column_usage

DBI::dbListTables(con, schema_name = "information_schema")

obj <- DBI::dbListObjects(con)
obj %>% filter(is_prefix)

DBI::dbListObjects(con, DBI::Id(schema = "INFORMATION_SCHEMA"))

info <-
  dm_meta(con)

info %>%
  dm_draw()

info %>%
  dm_get_tables()

info %>%
  dm_examine_constraints()

info_local <-
  info %>%
  collect()

quote_fq_schema <- function(con, catalog, schema) {
  if (is_postgres(con) || is_mssql(con)) {
    catalog <- dbQuoteIdentifier(con, catalog)
    schema <- dbQuoteIdentifier(con, schema)
    paste0(catalog, ".", schema)
  } else {
    bla
  }
}

quote_fq_table <- function(con, fq_schema, table) {
  table <- dbQuoteIdentifier(con, table)
  paste0(fq_schema, ".", table)
}

# quote_fq_column <- function(con, fq_table, column) {
#   table <- dbQuoteIdentifier(con, column)
#   paste0(fq_table, ".", column)
# }

fq_r_table_if_needed <- function(catalog, schema, table) {
  fq <- tibble(catalog, schema, table)

  fq %>%
    group_by(table) %>%
    mutate(n = n()) %>%
    ungroup() %>%
    mutate(fq_table = if_else(n > 1, fq_r_table(catalog, schema, table), table)) %>%
    pull()
}

fq_r_table <- function(catalog, schema, table) {
  if (length(unique(catalog)) > 1) {
    catalog <- paste0(catalog, ".")
  } else {
    catalog <- ""
  }

  if (length(unique(schema)) > 1) {
    schema <- paste0(schema, ".")
  } else {
    schema <- ""
  }

  paste0(catalog, schema, table)
}

info_local_named <-
  info_local %>%

  # FIXME: Simplify with rekey, https://github.com/cynkra/dm/issues/519
  dm_zoom_to(schemata) %>%
  mutate(fq_schema_name = quote_fq_schema(!!con, catalog_name, schema_name), .before = catalog_name) %>%
  dm_update_zoomed() %>%

  dm_zoom_to(tables) %>%
  left_join(schemata, select = fq_schema_name) %>%
  mutate(fq_table_name = quote_fq_table(!!con, fq_schema_name, table_name), .before = table_catalog) %>%
  mutate(r_table_name = fq_r_table_if_needed(table_catalog, table_schema, table_name)) %>%
  dm_update_zoomed() %>%

  dm_zoom_to(table_constraints) %>%
  left_join(tables, select = fq_table_name) %>%
  mutate(fq_constraint_name = quote_fq_table(!!con, quote_fq_schema(!!con, constraint_catalog, constraint_schema), constraint_name), .before = constraint_catalog) %>%
  select(fq_constraint_name, fq_table_name, everything()) %>%
  dm_update_zoomed() %>%

  dm_zoom_to(columns) %>%
  left_join(tables, select = fq_table_name) %>%
  #mutate(fq_column_name = quote_fq_column(!!con, fq_table_name, column_name), .before = column_name) %>%
  select(fq_table_name, everything()) %>%
  #select(fq_table_name, fq_column_name, everything()) %>%
  dm_update_zoomed() %>%

  dm_zoom_to(key_column_usage) %>%
  left_join(columns, select = fq_table_name) %>%
  left_join(table_constraints, select = fq_constraint_name) %>%
  select(fq_constraint_name, fq_table_name, everything()) %>%
  dm_update_zoomed() %>%

  dm_zoom_to(constraint_column_usage) %>%
  left_join(columns, select = fq_table_name) %>%
  left_join(table_constraints, select = fq_constraint_name) %>%
  select(fq_constraint_name, fq_table_name, everything()) %>%
  dm_update_zoomed()

info_simple <-
  dm(!!!dm_get_tables(info_local_named)) %>%
  dm_add_pk(schemata, fq_schema_name) %>%
  dm_add_pk(tables, fq_table_name) %>%
  dm_add_fk(tables, fq_schema_name, schemata) %>%
  dm_add_pk(columns, c(fq_table_name, column_name)) %>%
  dm_add_fk(columns, fq_table_name, tables) %>%
  #dm_add_fk(table_constraints, table_schema, schemata) %>%
  dm_add_pk(table_constraints, fq_constraint_name) %>%
  dm_add_fk(table_constraints, fq_table_name, tables) %>%
  # constraint_schema vs. table_schema?

  # not on mssql:
  #dm_add_fk(referential_constraints, c(constraint_schema, table_name), tables) %>%
  #dm_add_fk(referential_constraints, c(constraint_schema, referenced_table_name), tables) %>%

  dm_add_pk(key_column_usage, c(fq_constraint_name, ordinal_position)) %>%
  dm_add_fk(key_column_usage, c(fq_table_name, column_name), columns) %>%
  dm_add_fk(key_column_usage, fq_constraint_name, table_constraints) %>%

  # not on mariadb;
  dm_add_pk(constraint_column_usage, c(fq_constraint_name, ordinal_position)) %>%
  dm_add_fk(constraint_column_usage, c(fq_table_name, column_name), columns) %>%
  dm_add_fk(constraint_column_usage, fq_constraint_name, table_constraints) %>%
  dm_add_fk(constraint_column_usage, c(fq_constraint_name, ordinal_position), key_column_usage) %>%

  dm_select(columns, -c(table_catalog, table_schema, table_name)) %>%
  dm_select(table_constraints, -c(table_catalog, table_schema, table_name)) %>%
  dm_select(key_column_usage, -c(table_catalog, table_schema, table_name)) %>%
  dm_select(key_column_usage, -c(constraint_catalog, constraint_schema, constraint_name)) %>%
  dm_select(constraint_column_usage, -c(table_catalog, table_schema, table_name)) %>%
  dm_select(constraint_column_usage, -c(constraint_catalog, constraint_schema, constraint_name)) %>%
  dm_set_colors(brown = c(tables, columns), blue = schemata, green4 = ends_with("_constraints"), orange = ends_with("_usage"))

info_simple %>%
  dm_draw()

info_simple %>%
  dm_get_tables()

key_dm <-
  info_simple %>%

  dm_zoom_to(table_constraints) %>%
  filter(constraint_type == "PRIMARY KEY") %>%
  dm_insert_zoomed("pk_constraints") %>%
  dm_zoom_to(key_column_usage) %>%
  semi_join(pk_constraints) %>%
  dm_insert_zoomed("pk") %>%

  dm_zoom_to(table_constraints) %>%
  filter(constraint_type == "FOREIGN KEY") %>%
  dm_update_zoomed() %>%
  dm_zoom_to(constraint_column_usage) %>%
  semi_join(table_constraints) %>%
  rename(fk_fq_table_name = fq_table_name, fk_column_name = column_name) %>%
  left_join(key_column_usage) %>%
  rename(pk_fq_table_name = fq_table_name, pk_column_name = column_name) %>%

  # Postgres: Can return int64 here
  mutate(ordinal_position = as.integer(ordinal_position)) %>%
  dm_insert_zoomed("fk") %>%
  dm_add_fk(fk, c(pk_fq_table_name, pk_column_name), columns) %>%

  dm_zoom_to(columns) %>%
  left_join(tables, select = r_table_name) %>%
  dm_update_zoomed() %>%

  dm_select_tbl(columns, pk, fk)

key_dm %>%
  dm_draw()

key_dm %>%
  dm_get_tables()

key_dm$fk %>%
  select(-fq_constraint_name)

# FIXME: Why is this missing?

#' Nest a table inside its dm
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `dm_nest_tbl()` converts a child table to a nested column in its parent
#' table.
#' The child table should not have children itself (i.e. it needs to be a
#' *terminal child table*).
#'
#' @param dm A dm.
#' @param child_table A terminal table with one parent table.
#' @param into The table to nest `child_tables` into, optional as it can be guessed
#'   from the foreign keys unambiguously but useful to be explicit.
#'
#' @seealso [dm::dm_wrap_tbl()], [dm::dm_unwrap_tbl()], [dm_pack_tbl()]
#' @export
#' @examples
#' nested_dm <-
#'   dm_nycflights13() %>%
#'   dm_select_tbl(airlines, flights) %>%
#'   dm_nest_tbl(flights)
#'
#' nested_dm
#'
#' nested_dm$airlines
dm_nest_tbl <- function(dm, child_table, into = NULL) {
  dm_local_error_call()

  # process args
  into <- enquo(into)
  # FIXME: Rename table_name to child_tables_name
  table_name <- dm_tbl_name(dm, {{ child_table }})

  # retrieve fk and parent_name
  fks <- dm_get_all_fks(dm)

  # retrieve fk and parent_name
  # FIXME: fix redundancies and DRY when we decide what we export
  fks <- dm_get_all_fks(dm)
  children <-
    fks %>%
    filter(parent_table == !!table_name) %>%
    pull(child_table)
  fk <- filter(fks, child_table == !!table_name)
  parent_fk <- unlist(fk$parent_key_cols)
  child_fk <- unlist(fk$child_fk_cols)
  child_pk <-
    dm_get_all_pks(dm) %>%
    filter(table == !!table_name) %>%
    pull(pk_col) %>%
    unlist()
  parent_name <- pull(fk, parent_table)

  # make sure we have a terminal child
  if (length(children) || !length(parent_name) || length(parent_name) > 1) {
    cli::cli_abort(
      c(
        "{.val {table_name}} can't be nested because it is not a terminal child table.",
        if (length(parent_name)) "parents: {.val {parent_name}}",
        if (length(children)) "children: {.val {children}}"
      ),
      call = dm_error_call()
    )
  }

  # check consistency of `into` if relevant
  if (!quo_is_null(into)) {
    into <- dm_tbl_name(dm, !!into)
    if (into != parent_name) {
      cli::cli_abort(
        "{.val {table_name}} can only be packed into {.val {child_name}}.",
        call = dm_error_call()
      )
    }
  }

  # fetch def and join tables
  def <- dm_get_def(dm, quiet = TRUE)
  table_data <- def$data[def$table == table_name][[1]]
  parent_data <- def$data[def$table == parent_name][[1]]
  nested_data <- nest_join(
    parent_data,
    table_data,
    by = set_names(child_fk, parent_fk),
    name = table_name
  )
  class(nested_data[[table_name]]) <- c("nested", class(nested_data[[table_name]]))

  # update def and rebuild dm
  def$data[def$table == parent_name] <- list(nested_data)
  old_parent_table_fk <- def[def$table == parent_name, ][["fks"]][[1]]
  new_parent_table_fk <- filter(old_parent_table_fk, table != table_name)
  def[def$table == parent_name, ][["fks"]][[1]] <- new_parent_table_fk
  def <- def[def$table != table_name, ]

  dm_from_def(def)
}

#' dm_pack_tbl()
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `dm_pack_tbl()` converts a parent table to a packed column in its child
#' table.
#' The parent table should not have parent tables itself (i.e. it needs to be a
#' *terminal parent table*).
#'
#' @param dm A dm.
#' @param parent_table A terminal table with one child table.
#' @param into The table to pack `parent_tables` into, optional as it can be guessed
#'   from the foreign keys unambiguously but useful to be explicit.
#'
#' @seealso [dm::dm_wrap_tbl()], [dm::dm_unwrap_tbl()], [dm_nest_tbl()].
#' @export
#' @examples
#' dm_packed <-
#'   dm_nycflights13() %>%
#'   dm_pack_tbl(planes)
#'
#' dm_packed
#'
#' dm_packed$flights
#'
#' dm_packed$flights$planes
dm_pack_tbl <- function(dm, parent_table, into = NULL) {
  dm_local_error_call()

  # process args
  into <- enquo(into)
  table_name <- dm_tbl_name(dm, {{ parent_table }})

  fks <- dm_get_all_fks(dm)
  fk <- filter(fks, parent_table == !!table_name)
  children_names <- pull(fk, child_table)

  check_table_can_be_packed(table_name, children_names, fks)
  child_name <- children_names # we checked we had only one

  pks <- dm_get_all_pks(dm)
  pk <- filter(pks, table == !!table_name)
  child_fk <- unlist(fk$child_fk_cols)
  parent_fk <- unlist(fk$parent_key_cols)
  parent_pk <- unlist(pk$pk_col)

  # check consistency of `into` if relevant
  if (!quo_is_null(into)) {
    into <- dm_tbl_name(dm, !!into)
    if (into != child_name) {
      cli::cli_abort(
        "{.val {table_name}} can only be packed into {.val {child_name}}.",
        call = dm_error_call()
      )
    }
  }

  # fetch def and join tables
  def <- dm_get_def(dm, quiet = TRUE)
  table_data <- def$data[def$table == table_name][[1]]
  child_data <- def$data[def$table == child_name][[1]]
  packed_data <- pack_join(
    child_data,
    table_data,
    by = set_names(parent_fk, child_fk),
    name = table_name
  )
  class(packed_data[[table_name]]) <- c("packed", class(packed_data[[table_name]]))

  # update def and rebuild dm
  def$data[def$table == child_name] <- list(packed_data)
  def <- def[def$table != table_name, ]

  dm_from_def(def)
}

check_table_can_be_packed <- function(table_name, children_names, fks) {
  # make sure we have a terminal parent
  parents <-
    fks %>%
    filter(child_table == !!table_name) %>%
    pull(parent_table)

  table_has_parents <- length(parents) > 0
  table_has_one_child <- length(children_names) == 1
  table_is_terminal_parent <- table_has_one_child && !table_has_parents
  if (!table_is_terminal_parent) {
    cli::cli_abort(
      c(
        "{.val {table_name}} can't be packed because it is not a terminal parent table.",
        if (length(parents) > 0) "parents : {.val {parents}}",
        if (length(children_names) > 0) "children: {.val {children_names}}"
      ),
      call = dm_error_call()
    )
  }
  invisible(NULL)
}

# FIXME: can we be more efficient ?
node_type_from_graph <- function(graph, drop = NULL) {
  vertices <- graph_vertices(graph)
  n_children <- map_dbl(vertices, ~ length(graph_neighbors(graph, .x, mode = "in")))
  n_parents <- map_dbl(vertices, ~ length(graph_neighbors(graph, .x, mode = "out")))
  node_types <- set_names(rep_along(vertices, "intermediate"), names(vertices))
  node_types[n_parents == 0 & n_children == 1] <- "terminal parent"
  node_types[n_children == 0 & n_parents == 1] <- "terminal child"
  node_types[n_children == 0 & n_parents == 0] <- "isolated"
  node_types[!names(node_types) %in% drop]
}

test_that("`dm_nest_tbl()` failure modes", {
  skip_if_remote_src()

  # is terminal parent
  expect_snapshot(error = TRUE, {
    dm_nest_tbl(dm_for_filter(), tf_1)
  })
  # has several parents
  expect_snapshot(error = TRUE, {
    dm_nest_tbl(dm_for_filter(), tf_2)
  })
  # has several children
  expect_snapshot(error = TRUE, {
    dm_nest_tbl(dm_for_filter(), tf_3)
  })
  # has both parent and child
  expect_snapshot(error = TRUE, {
    dm_nest_tbl(dm_for_filter(), tf_4)
  })
})

test_that("`dm_pack_tbl()`, `dm_unpack_tbl()`, `dm_nest_tbl()`, `dm_unnest_tbl()` work", {
  skip_if_remote_src()

  expect_snapshot({
    dm_packed <- dm_pack_tbl(dm_for_filter(), tf_1)
    dm_packed

    dm_packed_nested <- dm_nest_tbl(dm_packed, tf_2)
    dm_packed_nested

    dm_packed_nested_unnested <- dm_unnest_tbl(
      dm_packed_nested,
      tf_3,
      tf_2,
      ptype = dm_for_filter()
    )
    dm_packed_nested_unnested

    dm_packed_nested_unnested_unpacked <- dm_unpack_tbl(
      dm_packed_nested_unnested,
      tf_2,
      tf_1,
      ptype = dm_for_filter()
    )
    dm_packed_nested_unnested_unpacked
  })
})

test_that("`dm_pack_tbl()` failure modes", {
  # has several parents
  expect_snapshot(error = TRUE, {
    dm_pack_tbl(dm_for_filter(), tf_2)
  })
  # has several children
  expect_snapshot(error = TRUE, {
    dm_pack_tbl(dm_for_filter(), tf_3)
  })
  # has both parent and child
  expect_snapshot(error = TRUE, {
    dm_pack_tbl(dm_for_filter(), tf_4)
  })
})

ops_stack <- function(x) {
  new_ops_stack(list(x), 1)
}

new_ops_stack <- function(list, pos) {
  stopifnot(is.list(list))
  stopifnot(pos >= 1)
  stopifnot(pos <= length(list))
  list(
    list = list,
    pos = pos
  )
}

ops_stack_current <- function(stack) {
  stack$list[[stack$pos]]
}

ops_stack_append <- function(stack, x) {
  pos <- stack$pos
  new_ops_stack(
    c(stack$list[seq_len(pos)], list(x)),
    pos + 1
  )
}

ops_stack_undo <- function(stack) {
  pos <- stack$pos
  new_ops_stack(
    stack$list,
    max(pos - 1, 1)
  )
}

ops_stack_redo <- function(stack) {
  pos <- stack$pos
  new_ops_stack(
    stack$list,
    min(pos + 1, length(stack$list))
  )
}

repair_names_vec <- function(names, repair, quiet) {
  withCallingHandlers(
    vec_as_names(names, repair = repair, quiet = quiet),
    vctrs_error_names_must_be_unique = function(e) {
      abort_need_unique_names(names[duplicated(names)])
    }
  )
}

repair_table_names <- function(old_names, new_names, repair = "check_unique", quiet = FALSE) {
  all_names <- repair_names_vec(c(old_names, new_names), repair, quiet)
  all_names_ordered <- all_names[seq_along(old_names)]
  new_old_names <- set_names(old_names, all_names_ordered)
  old_new_names <- set_names(all_names_ordered, old_names)

  new_names <-
    all_names[seq2(length(old_names) + 1, length(all_names))]
  list(new_old_names = new_old_names, new_names = new_names, old_new_names = old_new_names)
}

dm_add_tbl_impl <- function(
  dm,
  tbls,
  table_name,
  filters = list_of(new_filter()),
  pks = list_of(new_pk()),
  fks = list_of(new_fk()),
  uks = list_of(new_uk())
) {
  def <- dm_get_def(dm)

  def_0 <- def[rep_along(table_name, NA_integer_), ]
  def_0$table <- table_name
  def_0$data <- unname(tbls)
  def_0$pks <- pks
  def_0$uks <- uks
  def_0$fks <- fks
  def_0$filters <- filters

  dm_from_def(vec_rbind(def, def_0))
}

check_new_tbls <- function(dm, tbls) {
  orig_tbls <- dm_get_tables_impl(dm)

  # are all new tables on the same source as the original ones?
  if (has_length(orig_tbls) && !all_same_source(c(orig_tbls[1], tbls))) {
    abort_not_same_src()
  }
}

#' Update tables in a [`dm`]
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Updates one or more existing tables in a [`dm`].
#' For now, the column names must be identical.
#' This restriction may be levied optionally in the future.
#'
#' @seealso [dm()], [dm_select_tbl()]
#'
#' @param dm A [`dm`] object.
#' @param ... One or more tables to update in the `dm`.
#'   Must be named.
#'
#' @export
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_mutate_tbl(flights = nycflights13::flights[1:3, ])
dm_mutate_tbl <- function(dm, ...) {
  check_not_zoomed(dm)

  old_names <- src_tbls_impl(dm)

  new_tables <- list2(...)
  stopifnot(is_named(new_tables))

  new_names <- names(new_tables)
  stopifnot(new_names %in% old_names)

  old_tables <- dm_get_tables_impl(dm)

  stopifnot(identical(map(new_tables, colnames), map(old_tables[new_names], colnames)))

  old_tables[new_names] <- new_tables

  def <- dm_get_def(dm)
  def$data <- unname(old_tables)
  dm_from_def(def)
}

dm_to_svg <- function(dm) {
  svg <-
    dm %>%
    dm::dm_draw(graph_attrs = 'bgcolor="transparent"', node_attrs = 'fontname="Helvetica"') %>%
    DiagrammeRsvg::export_svg()

  gsub('( class="(edge|node)")', '\\1 onclick="\\2Style(this.id)"', svg)
}

# graph code directly from {datamodelr} -----------------------------------------

bdm_create_graph <- function(
  data_model,
  rankdir = "BT",
  graph_name = "Data Model",
  graph_attrs = "",
  node_attrs = "",
  edge_attrs = "",
  view_type = "all",
  focus = NULL,
  col_attr = "column",
  columnArrows = FALSE,
  table_description = NULL,
  font_size
) {
  g_list <-
    bdm_create_graph_list(
      data_model = data_model,
      view_type = view_type,
      focus = focus,
      col_attr = col_attr,
      columnArrows = columnArrows,
      table_description = table_description,
      font_size = font_size
    )
  if (length(g_list$nodes$nodes) == 0) {
    warning("The number of tables to render is 0.")
  }
  graph <-
    list(
      graph_attrs = sprintf('rankdir=%s tooltip="%s" %s', rankdir, graph_name, graph_attrs),
      node_attrs = sprintf('margin=0 fontcolor = "#444444" %s', node_attrs),
      nodes_df = g_list$nodes,
      edges_df = g_list$edges,
      edge_attrs = c('color = "#555555"', "arrowsize = 1", edge_attrs)
    )
  class(graph) <- c("datamodelr_graph", class(graph))

  # re-create dot code for data model
  # (DiagrammeR does not support yet the HTML labels and clusters (v.0.8))
  graph$dot_code <- dot_graph(graph, columnArrows)

  graph
}

bdm_render_graph <- function(graph, width = NULL, height = NULL, top_level_fun = NULL) {
  check_suggested("DiagrammeR", top_level_fun)

  if (is.null(graph$dot_code)) {
    graph$dot_code <- dot_graph(graph)
  }
  DiagrammeR::grViz(graph$dot_code, allow_subst = FALSE, width, height)
}

bdm_create_graph_list <- function(
  data_model,
  view_type = "all",
  focus = NULL,
  col_attr = "column",
  columnArrows = FALSE,
  table_description = list(),
  font_size
) {
  # hidden tables

  if (!is.null(focus) && is.list(focus)) {
    if (!is.null(focus[["tables"]])) {
      data_model$tables <- data_model$tables[data_model$tables$table %in% focus$tables, ]
      data_model$columns <- data_model$columns[data_model$columns$table %in% focus$tables, ]
      if (is.null(focus[["external_ref"]]) || !focus[["external_ref"]]) {
        data_model$references <- data_model$references[
          data_model$references$table %in%
            focus$tables &
            data_model$references$ref %in% focus$tables,
        ]
      }
    }
  } else {
    # hide tables with display == "hide" attribute
    if (is.null(data_model$tables$display)) {
      data_model$tables$display <- NA
    }
    data_model$tables$display[is.na(data_model$tables$display)] <- "show"
    hidden_tables <- data_model$tables[data_model$tables$display == "hide", "table"]
    if (!is.null(hidden_tables)) {
      data_model$tables <- data_model$tables[!data_model$tables$table %in% hidden_tables, ]
      data_model$columns <- data_model$columns[!data_model$columns$table %in% hidden_tables, ]
      data_model$references <- data_model$references[
        !data_model$references$table %in% hidden_tables &
          !data_model$references$ref %in% hidden_tables,
      ]
    }
  }

  # remove hidden columns
  # data_model$columns <-
  #  data_model$columns[is.na(data_model$columns$display) | data_model$columns$display != "hide", ]

  tables <- split(data_model$columns, data_model$columns$table)

  switch(
    view_type,
    all = {},
    #
    keys_only = {
      tables <- lapply(tables, function(tab) {
        tab[tab[["key"]] > 0 | !is.na(tab[, "ref"]), ]
      })
    },
    #
    title_only = {
      tables <- lapply(tables, function(tab) {
        tab[0L, ]
      })
    }
  )
  g_labels <-
    sapply(names(tables), function(x) {
      dot_html_label(
        tables[[x]],
        title = x,
        palette_id = data_model$tables[data_model$tables$table == x, "display"],
        col_attr = col_attr,
        columnArrows = columnArrows,
        table_description = table_description[[x]],
        font_size = font_size
      )
    })

  nodes <-
    data.frame(
      nodes = names(tables),
      label = g_labels,
      shape = "plaintext",
      type = "upper",
      segment = data_model$tables[order(data_model$tables$table), "segment"],
      stringsAsFactors = FALSE
    )

  if (!is.null(data_model$references)) {
    edges <-
      with(
        data_model$references[data_model$references$ref_col_num == 1, ],
        data.frame(
          from = table,
          to = ref,
          fromCol = column,
          toCol = ref_col,
          keyId = keyId,
          uk_col = uk_col,
          stringsAsFactors = FALSE
        )
      )
  } else {
    edges <- NULL
  }

  ret <-
    list(nodes = nodes, edges = edges)

  ret
}

dot_graph <- function(graph, columnArrows = FALSE) {
  graph_type <- "digraph"

  dot_attr <- paste0(
    sprintf("graph [%s]\n\n", paste(graph$graph_attrs, collapse = ", ")),
    sprintf("node [%s]\n\n", paste(graph$node_attrs, collapse = ", ")),
    sprintf("edge [%s]\n\n", paste(graph$edge_attrs, collapse = ", ")),
    "pack=true\n",
    'packmode= "node"\n'
  )
  segments <- unique(graph$nodes_df$segment)
  segments <- segments[!is.na(segments)]
  segments <- stats::setNames(1:(length(segments)), segments)

  dot_nodes <- sapply(seq_len(nrow(graph$nodes_df)), function(n) {
    node <- graph$nodes_df[n, ]
    dot_node <- sprintf(
      '  "%s" [id = "%s", label = %s, shape = "%s"] \n',
      node$nodes,
      node$nodes,
      node$label,
      node$shape
    )
    if (!is.na(node[["segment"]])) {
      dot_node <- sprintf(
        "subgraph cluster_%s {\nlabel='%s'\ncolor=\"#DDDDDD\"\n%s\n}\n",
        segments[node[["segment"]]],
        node[["segment"]],
        dot_node
      )
    }
    dot_node
  })

  dot_seg_nodes <- paste(dot_nodes, collapse = "\n")
  dot_edges <- ""
  if (columnArrows) {
    dot_edges <- paste(
      sprintf(
        '"%s":"%s"->"%s":"%s" [id="%s"%s]',
        graph$edges_df$from,
        graph$edges_df$fromCol,
        graph$edges_df$to,
        graph$edges_df$toCol,
        graph$edges_df$keyId,
        graph$edges_df$uk_col
      ),
      collapse = "\n"
    )
  } else {
    dot_edges <- paste(
      sprintf(
        '"%s"->"%s" [id="%s"]',
        graph$edges_df$from,
        graph$edges_df$to,
        graph$edges_df$keyId
      ),
      collapse = "\n"
    )
  }
  ret <- sprintf(
    "#data_model\n%s {\n%s\n%s\n%s\n}",
    graph_type,
    dot_attr,
    dot_seg_nodes,
    dot_edges
  )
  ret
}

html_table <- function(x, ...) html_tag(x, tag = "TABLE", ident = 1, ...)
html_tr <- function(x, ...) html_tag(x, tag = "TR", ident = 2, ...)
html_td <- function(x, ...) html_tag(x, tag = "TD", ident = 3, nl = FALSE, ...)
html_font <- function(x, ...) html_tag(x, tag = "FONT", ident = 0, nl = FALSE, ...)

html_tag <- function(x, tag, ident = 0, nl = TRUE, atrs = NULL, collapse = "") {
  if (length(x) > 1 && !is.null(collapse)) {
    x <- paste(x, collapse = collapse)
  }
  space <- paste(rep("  ", ident), collapse = "")
  atrs <- paste(sprintf('%s="%s"', names(atrs), atrs), collapse = " ")
  if (nchar(atrs) > 0) {
    atrs <- paste0(" ", atrs)
  }

  htext <-
    if (nl) {
      sprintf("%s<%s%s>\n%s%s</%s>\n", space, tag, atrs, x, space, tag)
    } else {
      sprintf("%s<%s%s>%s</%s>\n", space, tag, atrs, x, tag)
    }
  paste(htext, collapse = "")
}

to_html_table <- function(
  x,
  title = "Table",
  attr_table,
  attr_header,
  attr_font,
  attr_td = NULL,
  trans = NULL,
  cols = names(x),
  table_description = NULL,
  font_size,
  attr_desc
) {
  html_table(
    atrs = attr_table,
    c(
      # header
      html_tr(
        html_td(
          html_font(title, atrs = c(attr_font, "POINT-SIZE" = font_size[["header"]])),
          atrs = attr_header,
          collapse = NULL
        )
      ),
      if (!is.null(table_description)) {
        table_description <- strsplit(table_description, "\n")[[1]]
        map_chr(table_description, function(desc) {
          html_tr(
            html_td(
              html_font(
                repair_html(desc),
                atrs = c(attr_desc, "POINT-SIZE" = font_size[["table_description"]] %||% 8L)
              ),
              atrs = attr_desc,
              collapse = NULL
            )
          )
        })
      },
      # rows
      unique(sapply(seq_len(nrow(x)), function(r) {
        html_tr(c(
          # cells
          sapply(cols, function(col_name) {
            value <- x[r, col_name]
            if (!is_null(trans)) {
              value <- trans(col_name, x[r, ], value, font_size[["column"]])
            }
            html_td(value, if (is.null(attr_td)) NULL else attr_td(col_name, x[r, ], value))
          })
        ))
      }))
    )
  )
}

dot_html_label <- function(
  x,
  title,
  palette_id = "default",
  col_attr = c("column"),
  columnArrows = FALSE,
  table_description = NULL,
  font_size
) {
  cols <- c("ref", col_attr)
  if (is.null(palette_id) || palette_id == "show") {
    palette_id <- "default"
  }
  # currently we always have a border around the tables
  border <- 1
  # test if palette_id is valid: either datamodelr or hexcode (converted in `dm_set_colors()` from colorname)
  if (palette_id == "default") {
    col <- list(
      line_color = "#555555",
      header_bgcolor = "#EFEBDD",
      header_font = "#000000",
      bgcolor = "#FFFFFF",
      desccol = "#F9F8F3"
    )
  } else {
    header_bgcol_rgb <- col2rgb(palette_id, alpha = TRUE)
    bodycol_rgb <- calc_bodycol_rgb(header_bgcol_rgb)
    bodycol <- hex_from_rgb(bodycol_rgb)
    desccol_rgb <- calc_bodycol_rgb(header_bgcol_rgb, ratio = 0.65)
    desccol <- hex_from_rgb(desccol_rgb)
    # if header background too dark, use white font color
    header_font <- if (is_dark_color(header_bgcol_rgb)) "#FFFFFF" else "#000000"
    line_color_rgb <- header_bgcol_rgb / 1.5
    line_color <- hex_from_rgb(line_color_rgb)

    col <- list(
      line_color = line_color,
      header_bgcolor = palette_id,
      header_font = header_font,
      bgcolor = bodycol,
      desccol = desccol
    )
  }

  attr_table <- list(
    ALIGN = "LEFT",
    BORDER = border,
    CELLBORDER = 0,
    CELLSPACING = 0
  )
  # border color
  if (border) {
    attr_table[["COLOR"]] <- col[["line_color"]]
  }
  attr_header <- list(
    COLSPAN = length(cols) - columnArrows,
    BGCOLOR = col[["header_bgcolor"]],
    BORDER = 0
  )
  attr_font <- list(COLOR = col[["header_font"]])

  attr_desc <- attr_header
  attr_desc[["COLOR"]] <- "#000000"
  attr_desc[["BGCOLOR"]] <- col[["desccol"]]
  attr_desc[["BORDER"]] <- 0

  attr_td <- function(col_name, row_values, value) {
    ret <- list(ALIGN = "LEFT", BGCOLOR = col[["bgcolor"]])
    if (col_name == "column" && columnArrows) {
      key <- row_values[["key"]]
      reference <- row_values[["ref"]]
      if (!is.na(reference) || key) {
        ret$PORT <- row_values[["column"]]
      }
    }
    ret
  }

  # value presentation transformation
  trans <- function(col_name, row_values, value, font_size) {
    if (col_name == "ref") {
      value <- ifelse(is.na(value), "", "~")
      if (columnArrows) {
        value <- NULL
      }
    }
    if (col_name == "column" && row_values[["key"]] == 1 && row_values[["kind"]] == "PK") {
      value <- sprintf("<U>%s</U>", value)
    } else if (col_name == "column" && row_values[["key"]] == 1 && row_values[["kind"]] != "PK") {
      value <- sprintf("<I>%s</I>", value)
    }
    if (!is.null(value) && is.na(value)) {
      value <- ""
    } else if (!is.null(font_size)) {
      value <- sprintf('<FONT POINT-SIZE=\"%i\">%s</FONT>', font_size, value)
    }
    return(value)
  }
  ret <- to_html_table(
    x,
    title = title,
    attr_table = attr_table,
    attr_header = attr_header,
    attr_font = attr_font,
    attr_td = attr_td,
    cols = cols,
    trans = trans,
    table_description = table_description,
    font_size = font_size,
    attr_desc
  )
  ret <- sprintf("<%s>", trimws(ret))
  ret
}

repair_html <- function(x) {
  x <- gsub("&", "&amp;", x)
  x <- gsub("'", "&#39;", x)
  x <- gsub("<", "&lt;", x)
  x <- gsub(">", "&gt;", x)
  x <- gsub("\"", "&#34;", x)
  x
}

test_that("datamodel-code for drawing", {
  local_options(max.print = 10000)

  expect_snapshot({
    dm_get_data_model(dm_for_filter())
    dm_get_data_model(dm_for_filter(), column_types = TRUE)
  })
})

test_that("snapshot test for datamodelr code", {
  dm <- dm_nycflights13(cycle = TRUE)
  data_model <- dm_get_data_model(dm)

  path <- tempfile(fileext = ".dot")
  writeLines(bdm_create_graph(data_model, font_size = list())$dot_code, path)

  expect_snapshot_file(path, "nycflights13.dot")
})

test_that("snapshot test 2 for datamodelr code", {
  dm <- dm_nycflights13(cycle = TRUE)
  data_model <- dm_get_data_model(dm)

  path <- tempfile(fileext = ".dot")
  writeLines(
    bdm_create_graph(
      data_model,
      table_description = list(
        "flights" = "Flüge",
        "planes" = "Flugzeuge\nl'avion\nel & <\"avión flying in the sky\">"
      ),
      font_size = list()
    )$dot_code,
    path
  )

  expect_snapshot_file(path, "nycflights13_table_desc_1.dot")
})

test_that("snapshot test 3 for datamodelr code", {
  dm <- dm_nycflights13(cycle = TRUE)
  data_model <- dm_get_data_model(dm)

  path <- tempfile(fileext = ".dot")
  writeLines(
    bdm_create_graph(
      data_model,
      table_description = list(
        "flights" = "Flüge",
        "planes" = "Flugzeuge\nl'avion\nel & <\"avión flying in the sky\">"
      ),
      font_size = list(table_description = 10L)
    )$dot_code,
    path
  )

  expect_snapshot_file(path, "nycflights13_table_desc_2.dot")
})

test_that("snapshot test 4 for datamodelr code", {
  dm <- dm_nycflights13(cycle = TRUE) %>%
    dm_add_uk(weather, time_hour)
  data_model <- dm_get_data_model(dm)

  path <- tempfile(fileext = ".dot")
  writeLines(
    bdm_create_graph(
      data_model,
      table_description = list(
        "flights" = "Flüge",
        "planes" = "Flugzeuge\nl'avion\nel & <\"avión flying in the sky\">"
      ),
      font_size = list(table_description = 10L)
    )$dot_code,
    path
  )

  expect_snapshot_file(path, "nycflights13_draw_uk_1.dot")
})

test_that("snapshot test 5 for datamodelr code", {
  dm <- dm_nycflights13(cycle = TRUE) %>%
    dm_add_fk(flights, time_hour, weather, time_hour)
  data_model <- dm_get_data_model(dm)

  path <- tempfile(fileext = ".dot")
  writeLines(
    bdm_create_graph(
      data_model,
      table_description = list(
        "flights" = "Flüge",
        "planes" = "Flugzeuge\nl'avion\nel & <\"avión flying in the sky\">"
      ),
      font_size = list(table_description = 10L)
    )$dot_code,
    path
  )

  expect_snapshot_file(path, "nycflights13_draw_uk_2.dot")
})

test_that("snapshot test 6 for datamodelr code", {
  dm <- dm_nycflights13(cycle = TRUE) %>%
    dm_add_fk(flights, time_hour, weather, time_hour)
  data_model <- dm_get_data_model(dm)

  path <- tempfile(fileext = ".dot")
  writeLines(
    bdm_create_graph(
      data_model,
      table_description = list(
        "flights" = "Flüge",
        "planes" = "Flugzeuge\nl'avion\nel & <\"avión flying in the sky\">"
      ),
      font_size = list(table_description = 10L, column = 12L, header = 19L)
    )$dot_code,
    path
  )

  expect_snapshot_file(path, "nycflights13_draw_uk_3.dot")
})


test_that("snapshot test for weird data models", {
  dm <- dm("a b" = tibble(`c d` = 1)) %>%
    dm_add_pk(`a b`, `c d`)

  data_model <- dm_get_data_model(dm)

  path <- tempfile(fileext = ".dot")
  writeLines(bdm_create_graph(data_model, font_size = list())$dot_code, path)

  expect_snapshot_file(path, "weird.dot")
})

#' @keywords internal
#' @aliases NULL dm-package
"_PACKAGE"

#' @import rlang
#' @import cli
#' @import glue
#' @import lifecycle
#' @import vctrs
#' @rawNamespace import(tibble, except = c(data_frame))
#' @rawNamespace import(dplyr, except = c(data_frame))
#' @rawNamespace import(purrr, except = c(list_along, rep_along, invoke, modify, as_function, flatten_dbl, flatten_lgl, flatten_int, flatten_raw, flatten_chr, splice, flatten, prepend, `%@%`, `%||%`))
#' @rawNamespace import(tidyr, except = c(extract))
## usethis namespace: start
#' @importFrom methods is
#' @importFrom utils head tail str packageVersion
#' @importFrom grDevices col2rgb colors rgb
## usethis namespace: end
NULL

#' Connection to SQL Financial Database
#' @description Connects to relational.fel.cvut.cz unless the service is
#' unavailable, in which case databases.pacha.dev is used as a fallback
#' @return A `MariaDBConnection` object
#' @noRd
financial_db_con <- function() {
  if (Sys.getenv("DM_OFFLINE") != "") {
    cli::cli_abort("Offline")
  }

  err_relational <- tryCatch(return(relational_con()), error = identity)
  err_dbedu <- tryCatch(return(dbedu_con()), error = identity)

  cli::cli_abort(paste0(
    "Can't connect to relational.fel.cvut.cz or databases.pacha.dev:\n",
    conditionMessage(err_relational),
    "\n",
    conditionMessage(err_dbedu)
  ))
}

relational_con <- function() {
  DBI::dbConnect(
    RMariaDB::MariaDB(),
    username = "guest",
    password = "ctu-relational",
    dbname = "Financial_ijs",
    host = "relational.fel.cvut.cz"
  )
}

dbedu_con <- function() {
  DBI::dbConnect(
    RMariaDB::MariaDB(),
    username = "student",
    password = "uQCy30sNP5arqMBGHVLZ",
    dbname = "financial",
    host = "databases.pacha.dev"
  )
}

unique_db_table_name <- local({
  i <- 0

  function(table_name) {
    i <<- i + 1
    glue("{table_name}_", as.character(i), "_", systime_convenient(), "_", get_pid())
  }
})

systime_convenient <- function() {
  # FIXME: Race condition here, but fast enough
  local_options(digits.secs = 6)

  if (Sys.getenv("IN_PKGDOWN") != "") {
    "20200828_071303"
  } else {
    time <- as.character(Sys.time())
    gsub("[-:.]", "", gsub(" ", "_", time))
  }
}

get_pid <- function() {
  if (Sys.getenv("IN_PKGDOWN") != "") {
    "12345"
  } else {
    as.character(Sys.getpid())
  }
}

is_db <- function(x) {
  inherits(x, "src_sql")
}

is_src_db <- function(dm) {
  is_db(dm_get_src_impl(dm))
}

is_duckdb <- function(dest) {
  inherits(dest, c("duckdb_connection", "src_duckdb_connection"))
}

is_sqlite <- function(dest) {
  inherits(dest, "SQLiteConnection")
}

is_mssql <- function(dest) {
  inherits(
    dest,
    c(
      "Microsoft SQL Server",
      "src_Microsoft SQL Server",
      "dblogConnection-Microsoft SQL Server",
      "src_dblogConnection-Microsoft SQL Server"
    )
  )
}

is_postgres <- function(dest) {
  inherits_any(
    dest,
    c(
      "src_PostgreSQLConnection",
      "src_PqConnection",
      "PostgreSQLConnection",
      "PqConnection",
      "src_PostgreSQL"
    )
  )
}

is_redshift <- function(dest) {
  inherits_any(
    dest,
    c(
      "src_RedshiftConnection",
      "RedshiftConnection"
    )
  )
}

is_mariadb <- function(dest) {
  inherits_any(
    dest,
    c(
      "MariaDBConnection",
      "src_MariaDBConnection",
      "MySQLConnection",
      "src_MySQLConnection",
      "src_DoltConnection",
      "src_DoltLocalConnection"
    )
  )
}

schema_supported_dbs <- function() {
  tibble::tribble(
    ~db_name     , ~id_function  , ~test_shortcut ,
    "SQL Server" , "is_mssql"    , "mssql"        ,
    "Postgres"   , "is_postgres" , "postgres"     ,
    "MariaDB"    , "is_mariadb"  , "maria"        ,
  )
}

is_schema_supported <- function(con) {
  funs <- schema_supported_dbs()[["id_function"]]

  any(purrr::map_lgl(funs, ~ do.call(., args = list(dest = con))))
}

src_from_src_or_con <- function(dest) {
  if (is.src(dest)) dest else dbplyr::src_dbi(dest)
}

con_from_src_or_con <- function(dest) {
  if (is.src(dest)) dest$con else dest
}

repair_table_names_for_db <- function(table_names, temporary, con, schema = NULL) {
  if (temporary) {
    if (!is.null(schema)) {
      abort_temporary_not_in_schema()
    }
    # FIXME: Better logic for temporary table names
    if (is_mssql(con)) {
      names <- paste0("#", table_names)
    } else {
      names <- table_names
    }
    names <- unique_db_table_name(names)
  } else {
    # permanent tables
    if (!is.null(schema) && !is_schema_supported(con)) {
      abort_no_schemas_supported(con = con)
    }
    names <- table_names
  }
  names <- set_names(names, table_names)
  quote_ids(names, con_from_src_or_con(con), schema)
}

find_name_clashes <- function(old, new) {
  # Any entries in `new` with more than one corresponding entry in `old`
  purrr::keep(split(old, new), ~ length(unique(.x)) > 1)
}

#' @autoglobal
get_src_tbl_names <- function(src, schema = NULL, dbname = NULL, names_pattern = "{.table}") {
  schema <- check_schema(src, schema)

  if (!is_mssql(src) && !is_postgres(src) && !is_redshift(src) && !is_mariadb(src)) {
    warn_if_arg_not(dbname, only_on = "MSSQL")
    tables <- src_tbls(src)
    out <- purrr::map(tables, ~ DBI::Id(table = .x))

    return(set_names(out, tables))
  }

  con <- con_from_src_or_con(src)

  if (is_mssql(src)) {
    # MSSQL
    dbname_sql <- dbname_mssql(con, dbname)
    names_table <- get_names_table_mssql(con, dbname_sql)
    dbname <- names(dbname_sql)
  } else if (is_postgres(src)) {
    # Postgres
    dbname <- warn_if_arg_not(dbname, only_on = "MSSQL")
    names_table <- get_names_table_postgres(con)
  } else if (is_redshift(src)) {
    # Redshift
    dbname <- warn_if_arg_not(dbname, only_on = "MSSQL")
    names_table <- get_names_table_redshift(con)
  } else if (is_mariadb(src)) {
    # MariaDB
    dbname <- warn_if_arg_not(dbname, only_on = "MSSQL")
    names_table <- get_names_table_mariadb(con)
  }

  names_table <- names_table %>%
    filter(
      schema_name %in% !!(if (inherits(schema, "sql")) glue_sql_collapse(schema) else schema)
    ) %>%
    collect() %>%
    # create remote names for the tables in the given schema (name is table_name; cannot be duplicated within a single schema)
    mutate(
      local_name = glue(names_pattern, .table = table_name, .schema = schema_name),
      remote_name = schema_if(schema_name, table_name, con, dbname)
    )

  # SQL table names are only guaranteed to be unique in a single schema, so if
  # we have multiple schemas, we might end up with the same local_name pointing
  # to more than one remote_name
  # In such a case, raise a warning, and keep only the first relevant schema
  if (length(schema) > 1) {
    # Order according to ordering of `schema`, so that in a moment we can keep "first" table in event of a clash
    names_table <- names_table %>%
      mutate(schema_name = factor(schema_name, levels = schema)) %>%
      arrange(schema_name)

    clashes <- with(names_table, find_name_clashes(remote_name, local_name))

    if (length(clashes) > 0) {
      cli::cli_warn(c(
        "Some table names aren't unique:",
        purrr::imap_chr(
          clashes,
          ~ cli::format_inline(
            "Local name {.field {.y}} will refer to {.cls {DBI::dbQuoteIdentifier(con, .x[[1]])}}, ",
            "rather than to {.or {.cls {map(.x[-1], DBI::dbQuoteIdentifier, conn = con)}}}"
          )
        ) %>%
          purrr::set_names(rep("*", length(clashes)))
      ))

      # Keep only first schema for each local_name
      names_table <- slice_head(names_table, by = local_name)
    }
  }

  names_table %>%
    select(local_name, remote_name) %>%
    deframe()
}

check_schema <- function(src, schema) {
  if (!is_mssql(src) && !is_postgres(src) && !is_redshift(src) && !is_mariadb(src)) {
    warn_if_arg_not(schema, only_on = c("MSSQL", "Postgres", "Redshift", "MariaDB"))
    return(schema)
  }

  con <- con_from_src_or_con(src)

  if (!is.null(schema)) {
    check_param_class(schema, "character")
  }

  if (is_mssql(src)) {
    schema_mssql(con, schema)
  } else if (is_postgres(src)) {
    schema_postgres(con, schema)
  } else if (is_redshift(src)) {
    schema_redshift(con, schema)
  } else if (is_mariadb(src)) {
    schema_mariadb(con, schema)
  }
}

# `schema_*()` : default schema if NULL, otherwise unchanged
schema_mssql <- function(con, schema) {
  if (is_null(schema)) {
    schema <- "dbo"
  }
  schema
}

schema_postgres <- function(con, schema) {
  if (is_null(schema)) {
    schema <- "public"
  }
  schema
}

schema_redshift <- schema_postgres

schema_mariadb <- function(con, schema) {
  if (is_null(schema)) {
    # Extra parenthesis because it is used in a dbplyr-generated IN clause later
    schema <- sql("(database())")
  }
  schema
}

dbname_mssql <- function(con, dbname) {
  if (is_null(dbname)) {
    dbname <- ""
    dbname_sql <- ""
  } else {
    check_param_class(dbname, "character")
    dbname_sql <- paste0(DBI::dbQuoteIdentifier(con, dbname), ".")
  }
  set_names(dbname_sql, dbname)
}


get_names_table_mssql <- function(con, dbname_sql) {
  tbl(
    con,
    sql(glue::glue(
      "
      SELECT tabs.name AS table_name, schemas.name AS schema_name
      FROM {dbname_sql}sys.tables tabs
      INNER JOIN {dbname_sql}sys.schemas schemas ON
      tabs.schema_id = schemas.schema_id
    "
    ))
  )
}

get_names_table_postgres <- function(con) {
  tbl(
    con,
    sql(
      "SELECT table_schema as schema_name, table_name as table_name from information_schema.tables"
    )
  )
}

get_names_table_redshift <- get_names_table_postgres

get_names_table_mariadb <- get_names_table_postgres

test_that("table identifiers are quoted", {
  con_db <- my_db_test_con()

  test_dm <- copy_dm_to(
    con_db,
    dm(
      test_table_123 = tibble(a = 1),
      test_table_321 = tibble(b = 2)
    ),
    temporary = FALSE,
    table_names = ~ unique_db_table_name(.x)
  )

  remote_tbl_names_copied <- map_chr(dm_get_tables(test_dm), dbplyr::remote_name)

  on.exit({
    walk(
      remote_tbl_names_copied,
      ~ try(DBI::dbExecute(con_db, paste0("DROP TABLE ", .x)))
    )
  })

  dm <-
    suppress_mssql_warning(dm_from_con(con_db, learn_keys = FALSE)) %>%
    dm_select_tbl(
      !!!map(
        DBI::dbUnquoteIdentifier(con_db, DBI::SQL(remote_tbl_names_copied)),
        ~ .x@name[[length(.x@name)]]
      )
    )

  remote_tbl_names_learned <-
    dm %>%
    dm_get_tables() %>%
    map_chr(remote_name_qual)

  # `gsub()`, cause schema names are part of the remote_names (also standard schemas "dbo" for MSSQL and "public" for Postgres).
  expect_setequal(
    gsub("^.*\\.", "", unname(remote_tbl_names_learned)),
    unclass(DBI::dbQuoteIdentifier(con_db, names(dm)))
  )
})

test_that("table identifiers are quoted with learn_keys = FALSE", {
  con_db <- my_db_test_con()

  test_dm <- copy_dm_to(
    con_db,
    dm(
      test_table_123 = tibble(a = 1),
      test_table_321 = tibble(b = 2)
    ),
    temporary = FALSE,
    table_names = ~ unique_db_table_name(.x)
  )
  remote_tbl_names_copied <- map_chr(
    src_tbls_impl(test_dm),
    ~ remote_name_qual(test_dm[[.x]])
  )

  on.exit({
    walk(
      remote_tbl_names_copied,
      ~ try(DBI::dbExecute(con_db, paste0("DROP TABLE ", .x)))
    )
  })

  dm <- suppress_mssql_warning(dm_from_con(con_from_src_or_con(con_db), learn_keys = FALSE))
  remote_names <-
    dm %>%
    dm_get_tables() %>%
    map_chr(remote_name_qual)

  con <- dm_get_con(dm)
  expect_equal(
    gsub("^.*\\.", "", DBI::SQL(unname(remote_names))),
    DBI::dbQuoteIdentifier(con, names(dm))
  )
})


test_that("dm_from_src() deprecated", {
  con_db <- my_db_test_con()

  expect_deprecated(dm_from_src(src_from_src_or_con(con_db), learn_keys = FALSE))
})

test_that("DB helpers work for MSSQL", {
  skip_if_src_not("mssql")
  con_mssql <- my_test_src()$con
  expect_identical(schema_mssql(con_mssql, "schema"), "schema")
  expect_identical(schema_mssql(con_mssql, NULL), "dbo")
  expect_identical(
    dbname_mssql(con_mssql, "database_2"),
    set_names("\"database_2\".", "database_2")
  )
  expect_identical(dbname_mssql(con_mssql, NULL), set_names("", ""))

  withr::defer({
    try(DBI::dbExecute(con_mssql, "DROP TABLE test_db_helpers"))
    try(DBI::dbExecute(con_mssql, "DROP TABLE test_db_helpers_2"))
    try(DBI::dbExecute(con_mssql, "DROP TABLE schema_db_helpers.test_db_helpers_2"))
    try(DBI::dbExecute(con_mssql, "DROP SCHEMA schema_db_helpers"))
    try(DBI::dbExecute(con_mssql, "DROP TABLE [db_helpers_db].[dbo].[test_db_helpers_3]"))
    try(DBI::dbExecute(
      con_mssql,
      "DROP TABLE [db_helpers_db].[schema_db_helpers_2].[test_db_helpers_4]"
    ))
    # dropping schema is unnecessary
    try(DBI::dbExecute(con_mssql, "DROP DATABASE db_helpers_db"))
  })

  # create tables in 'dbo'
  DBI::dbWriteTable(
    con_mssql,
    DBI::Id(schema = "dbo", table = "test_db_helpers"),
    value = tibble(a = 1)
  )
  DBI::dbWriteTable(
    con_mssql,
    DBI::Id(schema = "dbo", table = "test_db_helpers_2"),
    value = tibble(a = 1)
  )
  # create table in a schema
  DBI::dbExecute(con_mssql, "CREATE SCHEMA schema_db_helpers")
  DBI::dbWriteTable(
    con_mssql,
    DBI::Id(schema = "schema_db_helpers", table = "test_db_helpers_2"),
    value = tibble(a = 1)
  )
  # create table on 'dbo' on another DB
  DBI::dbExecute(con_mssql, "CREATE DATABASE db_helpers_db")
  DBI::dbWriteTable(
    con_mssql,
    DBI::Id(catalog = "db_helpers_db", schema = "dbo", table = "test_db_helpers_3"),
    value = tibble(a = 1)
  )
  # create table in a schema on another DB
  original_dbname <- attributes(con_mssql)$info$dbname
  DBI::dbExecute(con_mssql, "USE db_helpers_db")
  DBI::dbExecute(con_mssql, "CREATE SCHEMA schema_db_helpers_2")
  DBI::dbExecute(con_mssql, paste0("USE ", original_dbname))
  DBI::dbWriteTable(
    con_mssql,
    DBI::Id(catalog = "db_helpers_db", schema = "schema_db_helpers_2", table = "test_db_helpers_4"),
    value = tibble(a = 1)
  )

  expect_identical(
    get_src_tbl_names(my_test_src())[["test_db_helpers"]],
    DBI::Id(schema = "dbo", table = "test_db_helpers")
  )
  expect_identical(
    get_src_tbl_names(my_test_src(), schema = "schema_db_helpers")[["test_db_helpers_2"]],
    DBI::Id(schema = "schema_db_helpers", table = "test_db_helpers_2")
  )
  expect_identical(
    get_src_tbl_names(my_test_src(), dbname = "db_helpers_db")[["test_db_helpers_3"]],
    DBI::Id(catalog = "db_helpers_db", schema = "dbo", table = "test_db_helpers_3")
  )
  expect_identical(
    get_src_tbl_names(my_test_src(), dbname = "db_helpers_db", schema = "schema_db_helpers_2")[[
      "test_db_helpers_4"
    ]],
    DBI::Id(catalog = "db_helpers_db", schema = "schema_db_helpers_2", table = "test_db_helpers_4")
  )
  expect_identical(
    get_src_tbl_names(
      my_test_src(),
      schema = c("dbo", "schema_db_helpers"),
      names_pattern = "{.schema}.{.table}"
    )[["dbo.test_db_helpers_2"]],
    DBI::Id(schema = "dbo", table = "test_db_helpers_2")
  )
  expect_identical(
    get_src_tbl_names(
      my_test_src(),
      schema = c("dbo", "schema_db_helpers"),
      names_pattern = "{.schema}.{.table}"
    )[["schema_db_helpers.test_db_helpers_2"]],
    DBI::Id(schema = "schema_db_helpers", table = "test_db_helpers_2")
  )
  expect_warning(
    out <- get_src_tbl_names(
      my_test_src(),
      schema = c("dbo", "schema_db_helpers"),
      names_pattern = "{.table}"
    )["test_db_helpers_2"],
    'Local name test_db_helpers_2 will refer to <"dbo"."test_db_helpers_2">, rather than to <"schema_db_helpers"."test_db_helpers_2">',
    fixed = TRUE
  )
  expect_identical(
    out,
    list(
      test_db_helpers_2 = DBI::Id(
        schema = "dbo",
        table = "test_db_helpers_2"
      )
    )
  )
  expect_warning(
    out <- get_src_tbl_names(
      my_test_src(),
      schema = c("schema_db_helpers", "dbo"),
      names_pattern = "{.table}"
    )["test_db_helpers_2"],
    'Local name test_db_helpers_2 will refer to <"schema_db_helpers"."test_db_helpers_2">, rather than to <"dbo"."test_db_helpers_2">',
    fixed = TRUE
  )
  expect_identical(
    out,
    list(
      test_db_helpers_2 = DBI::Id(
        schema = "schema_db_helpers",
        table = "test_db_helpers_2"
      )
    )
  )
})


test_that("DB helpers work for Postgres", {
  skip_if_src_not("postgres")
  con_postgres <- my_test_src()$con
  expect_identical(schema_postgres(con_postgres, "schema"), "schema")
  expect_identical(schema_postgres(con_postgres, NULL), "public")

  withr::defer({
    try(DBI::dbExecute(con_postgres, "DROP TABLE test_db_helpers"))
    try(DBI::dbExecute(con_postgres, "DROP TABLE test_db_helpers_2"))
    try(DBI::dbExecute(con_postgres, "DROP TABLE schema_db_helpers.test_db_helpers_2"))
    try(DBI::dbExecute(con_postgres, "DROP SCHEMA schema_db_helpers"))
  })

  # create tables in 'public'
  DBI::dbWriteTable(
    con_postgres,
    DBI::Id(schema = "public", table = "test_db_helpers"),
    value = tibble(a = 1)
  )
  DBI::dbWriteTable(
    con_postgres,
    DBI::Id(schema = "public", table = "test_db_helpers_2"),
    value = tibble(a = 1)
  )
  # create table in a schema
  DBI::dbExecute(con_postgres, "CREATE SCHEMA schema_db_helpers")
  DBI::dbWriteTable(
    con_postgres,
    DBI::Id(schema = "schema_db_helpers", table = "test_db_helpers_2"),
    value = tibble(a = 1)
  )

  expect_identical(
    get_src_tbl_names(my_test_src())["test_db_helpers"][[1]],
    DBI::Id(schema = "public", table = "test_db_helpers")
  )
  expect_identical(
    get_src_tbl_names(my_test_src(), schema = "schema_db_helpers")["test_db_helpers_2"][[1]],
    DBI::Id(schema = "schema_db_helpers", table = "test_db_helpers_2")
  )
  expect_identical(
    get_src_tbl_names(
      my_test_src(),
      schema = c("public", "schema_db_helpers"),
      names_pattern = "{.schema}.{.table}"
    )["public.test_db_helpers_2"][[1]],
    DBI::Id(schema = "public", table = "test_db_helpers_2")
  )
  expect_identical(
    get_src_tbl_names(
      my_test_src(),
      schema = c("public", "schema_db_helpers"),
      names_pattern = "{.schema}.{.table}"
    )["schema_db_helpers.test_db_helpers_2"][[1]],
    DBI::Id(schema = "schema_db_helpers", table = "test_db_helpers_2")
  )
  expect_warning(
    out <- get_src_tbl_names(
      my_test_src(),
      schema = c("public", "schema_db_helpers"),
      names_pattern = "{.table}"
    )["test_db_helpers_2"],
    'Local name test_db_helpers_2 will refer to <"public"."test_db_helpers_2">, rather than to <"schema_db_helpers"."test_db_helpers_2">',
    fixed = TRUE
  )
  expect_identical(
    out,
    list(
      test_db_helpers_2 = DBI::Id(
        schema = "public",
        table = "test_db_helpers_2"
      )
    )
  )
  expect_warning(
    out <- get_src_tbl_names(
      my_test_src(),
      schema = c("schema_db_helpers", "public"),
      names_pattern = "{.table}"
    )["test_db_helpers_2"],
    'Local name test_db_helpers_2 will refer to <"schema_db_helpers"."test_db_helpers_2">, rather than to <"public"."test_db_helpers_2">',
    fixed = TRUE
  )
  expect_identical(
    out,
    list(
      test_db_helpers_2 = DBI::Id(
        schema = "schema_db_helpers",
        table = "test_db_helpers_2"
      )
    )
  )
})

test_that("DB helpers work for other DBMS than MSSQL or Postgres", {
  # FIXME: Why does it fail for those databases?
  skip_if_src("mssql", "postgres")

  # for other DBMS than "MSSQL" or "Postgrs", get_src_tbl_names() translates to `src_tbls_impl()`
  con_db <- my_db_test_src()$con
  DBI::dbWriteTable(
    con_db,
    DBI::Id(table = "test_db_helpers"),
    value = tibble(a = 1)
  )
  withr::defer({
    try(DBI::dbExecute(con_db, "DROP TABLE test_db_helpers"))
  })

  skip_if_src("maria")

  # test for 2 warnings and if the output contains the new table
  expect_dm_warning(
    expect_dm_warning(
      expect_true(
        "test_db_helpers" %in%
          names(get_src_tbl_names(my_db_test_src(), schema = "schema", dbname = "dbname"))
      ),
      class = "arg_not"
    ),
    class = "arg_not"
  )

  skip_if_src("mssql", "postgres")

  # test for warning and if the output contains the new table
  expect_dm_warning(
    expect_true(
      "test_db_helpers" %in% names(get_src_tbl_names(my_db_test_src(), dbname = "dbname"))
    ),
    class = "arg_not"
  )
})

test_that("find name clashes", {
  # If all old names change to different new names...
  res <- find_name_clashes(
    c("one", "two", "three"),
    c("uno", "dos", "tres")
  )
  # ... we shouldn't get anything
  expect_length(res, 0)

  # If multiple old names change to the same new name...
  res <- find_name_clashes(
    c("one", "two", "three"),
    c("uno", "uno", "tres")
  )
  # We should get a list, with one element per "clashing" new name
  expect_named(res, "uno")
  expect_equal(res[["uno"]], c("one", "two"))
})

test_that("data source found", {
  expect_false(is_null(my_test_src_fun()))
  expect_silent(my_test_src_fun()())
})

test_that("copy_dm_to() copies data frames to databases", {
  skip_if_ide()

  expect_equivalent_dm(
    copy_dm_to(my_db_test_src(), collect(dm_for_filter())),
    dm_for_filter()
  )

  expect_equivalent_dm(
    copy_dm_to(my_db_test_src(), dm_for_filter()),
    dm_for_filter()
  )

  # FIXME: How to test writing permanent tables without and be sure they are removed at the end independent what 'my_test_src()' is?
})

test_that("copy_dm_to() copies data frames from any source", {
  expect_snapshot(error = TRUE, {
    copy_dm_to(default_local_src(), dm_for_filter())
  })
})

# FIXME: Add test that set_key_constraints = FALSE doesn't set key constraints,
# in combination with dm_learn_from_db

test_that("copy_dm_to() rejects overwrite and types arguments", {
  expect_snapshot(error = TRUE, {
    copy_dm_to(my_test_src(), dm_for_filter(), overwrite = TRUE)
  })

  expect_snapshot(error = TRUE, {
    copy_dm_to(my_test_src(), dm_for_filter(), types = character())
  })
})

test_that("copy_dm_to() fails with duplicate table names", {
  bad_names <- set_names(names(dm_for_filter()))
  bad_names[[2]] <- bad_names[[1]]

  expect_dm_error(
    copy_dm_to(my_db_test_src(), dm_for_filter(), table_names = bad_names),
    class = "copy_dm_to_table_names_duplicated"
  )
})

test_that("default table repair works", {
  con <- con_from_src_or_con(my_db_test_src())

  table_names <- c("t1", "t2", "t3")

  calls <- 0

  my_unique_db_table_name <- function(table_name) {
    calls <<- calls + 1
    glue::glue("{table_name}_2020_05_15_10_45_29_0")
  }

  local_mocked_bindings(
    unique_db_table_name = function(table_name) my_unique_db_table_name(table_name)
  )
  repair_table_names_for_db(table_names, temporary = FALSE, con)
  expect_equal(calls, 0)
  repair_table_names_for_db(table_names, temporary = TRUE, con)
  expect_gt(calls, 0)
})

test_that("copy_dm_to() fails legibly if target schema missing for MSSQL & Postgres", {
  skip_if_src_not(c("mssql", "postgres"))

  src_db <- my_test_src()
  local_dm <- dm_for_filter() %>% collect()

  expect_deprecated(expect_false(db_schema_exists(src_db, "copy_dm_to_schema")))

  expect_error(
    copy_dm_to(src_db, local_dm, schema = "copy_dm_to_schema", temporary = FALSE)
  )
})

test_that("copy_dm_to() fails legibly with schema argument for MSSQL & Postgres", {
  skip_if_src_not(c("mssql", "postgres"))

  src_db <- my_test_src()
  local_dm <- dm_for_filter() %>% collect()

  expect_false(db_schema_exists(src_db$con, "copy_dm_to_schema"))

  db_schema_create(src_db$con, "copy_dm_to_schema")

  withr::defer({
    try(DBI::dbExecute(src_db$con, "DROP SCHEMA copy_dm_to_schema"))
  })

  expect_dm_error(
    copy_dm_to(src_db, local_dm, schema = "copy_dm_to_schema"),
    "temporary_not_in_schema"
  )

  expect_dm_error(
    copy_dm_to(
      src_db,
      local_dm,
      schema = "copy_dm_to_schema",
      temporary = FALSE,
      table_names = set_names(letters[1:6], src_tbls_impl(local_dm))
    ),
    "one_of_schema_table_names"
  )
})

test_that("copy_dm_to() works with schema argument for MSSQL & Postgres", {
  skip_if_src_not(c("mssql", "postgres"))

  src_db <- my_test_src()
  local_dm <- dm_for_filter() %>% collect()

  expect_false(db_schema_exists(src_db$con, "copy_dm_to_schema"))

  db_schema_create(src_db$con, "copy_dm_to_schema")

  withr::defer({
    order_of_deletion <- c("tf_2", "tf_1", "tf_5", "tf_6", "tf_4", "tf_3")
    walk(
      dm_get_tables_impl(remote_dm)[order_of_deletion],
      ~ try(DBI::dbExecute(src_db$con, paste0("DROP TABLE ", remote_name_qual(.x))))
    )
    try(DBI::dbExecute(src_db$con, "DROP SCHEMA copy_dm_to_schema"))
  })

  expect_silent(
    remote_dm <- copy_dm_to(
      src_db,
      local_dm,
      schema = "copy_dm_to_schema",
      temporary = FALSE
    )
  )

  if (is_postgres(src_db)) {
    table_tibble <- sql_schema_table_list_postgres(src_db, "copy_dm_to_schema")
  } else if (is_mssql(src_db)) {
    table_tibble <- sql_schema_table_list_mssql(src_db, "copy_dm_to_schema")
  }

  tbl_names <- names(remote_dm)
  # compare names and remote names
  expect_identical(
    table_tibble %>% arrange(table_name) %>% deframe(),
    purrr::map(set_names(tbl_names), ~ DBI::Id(schema = "copy_dm_to_schema", table = .x))
  )
})

test_that("copy_dm_to() fails with schema argument for databases where schema is unsupported", {
  skip_if_schema_supported()

  local_dm <- dm_for_filter() %>% collect()

  expect_dm_error(
    copy_dm_to(
      my_test_src(),
      local_dm,
      temporary = FALSE,
      schema = "test"
    ),
    "no_schemas_supported"
  )
})

test_that("copy_dm_to() works with autoincrement PKs and FKS on selected DBs", {
  skip_if_src_not(c("postgres", "sqlite", "mssql", "maria"))

  con_db <- my_test_con()
  local_dm_ptype <-
    dm_for_autoinc_1() %>%
    dm_ptype() %>%
    dm_add_pk(t1, a, autoincrement = TRUE) %>%
    dm_add_pk(t2, c, autoincrement = TRUE) %>%
    dm_add_fk(t2, d, t1) %>%
    dm_add_fk(t3, e, t1) %>%
    dm_add_fk(t4, h, t2) %>%
    # Avoid clash with other test that uses the same table names
    dm_rename_tbl(xt1 = t1, xt2 = t2, xt3 = t3, xt4 = t4)

  withr::defer({
    order_of_deletion <- c("xt4", "xt2", "xt3", "xt1")
    walk(
      dm_get_tables_impl(remote_dm)[order_of_deletion],
      ~ try(DBI::dbExecute(con_db, paste0("DROP TABLE ", dbplyr::remote_name(.x))))
    )
  })

  # FIXME: how to check if autoincrement is actually set on DB?
  expect_silent(
    remote_dm <- copy_dm_to(
      con_db,
      local_dm_ptype,
      temporary = FALSE
    )
  )

  collected_dm <- remote_dm %>%
    collect()

  expect_equivalent_dm(
    local_dm_ptype,
    collected_dm
  )
})

test_that("copy_dm_to(table_names = ) (#250)", {
  dm <- dm(x = tibble(a = 1))
  dm_copy <- copy_dm_to(
    my_db_test_con(),
    dm,
    table_names = c(x = "y"),
    temporary = FALSE
  )
  withr::defer({
    try(DBI::dbExecute(my_db_test_con(), "DROP TABLE y", immediate = TRUE))
  })
  expect_equal(dbplyr::remote_name(dm_copy$x), "y")

  dm <- dm(x = tibble(a = 1), y = tibble(b = 1))
  dm_copy <- copy_dm_to(
    my_db_test_con(),
    dm,
    table_names = ~ paste0(.x, "1"),
    temporary = FALSE
  )
  withr::defer({
    try(DBI::dbExecute(my_db_test_con(), "DROP TABLE x1", immediate = TRUE))
    try(DBI::dbExecute(my_db_test_con(), "DROP TABLE y1", immediate = TRUE))
  })
  expect_equal(dbplyr::remote_name(dm_copy$x), "x1")
  expect_equal(dbplyr::remote_name(dm_copy$y), "y1")
})

test_that("build_copy_queries snapshot test for pixarfilms", {
  src_db <- my_db_test_src()

  # build regular dm from `dm_pixarfilms()`
  pixar_dm <-
    # fetch sample dm
    dm_pixarfilms() %>%
    # make it regular
    dm_filter(pixar_films = (!is.na(film))) %>%
    dm_select_tbl(-pixar_people)

  skip_if_not_installed("testthat", "3.1.1")

  expect_snapshot(
    variant = my_test_src_name,
    {
      pixar_dm %>%
        build_copy_queries(
          src_db,
          .
        ) %>%
        as.list() # to print full queries
    }
  )
})


test_that("build_copy_queries snapshot test for dm_for_filter()", {
  src_db <- my_db_test_src()

  skip_if_not_installed("testthat", "3.1.1")

  expect_snapshot(
    variant = my_test_src_name,
    {
      dm_for_filter() %>%
        collect() %>%
        build_copy_queries(
          src_db,
          .
        ) %>%
        as.list() # to print full queries
    }
  )
})


test_that("build_copy_queries avoids duplicate indexes", {
  src_db <- my_db_test_src()

  # build a dm whose index might be duplicated if naively build (child__a__key)
  ambiguous_dm <- dm(
    parent1 = tibble(key = 1),
    parent2 = tibble(a__key = 1),
    child = tibble(a__key = 1),
    child__a = tibble(key = 1)
  ) %>%
    dm_add_pk(parent1, key) %>%
    dm_add_pk(parent2, a__key) %>%
    dm_add_fk(child, a__key, parent2) %>%
    dm_add_fk(child__a, key, parent2)

  queries <-
    build_copy_queries(
      src_db,
      ambiguous_dm,
      table_names = names(ambiguous_dm) %>%
        repair_table_names_for_db(temporary = FALSE, con = src_db, schema = NULL)
    )

  expect_equal(anyDuplicated(unlist(queries$index_name)), 0)

  skip_if_not_installed("testthat", "3.1.1")

  expect_snapshot(
    variant = my_test_src_name,
    {
      as.list(queries)
    }
  )
})

#' \pkg{dplyr} table manipulation methods for zoomed dm objects
#'
#' @description
#' Use these methods without the '.dm_zoomed' suffix (see examples).
#' @param .data object of class `dm_zoomed`
#' @param ... see corresponding function in package \pkg{dplyr} or \pkg{tidyr}
#' @inheritParams dplyr::filter
#' @inheritParams dplyr::arrange
#' @inheritParams dplyr::group_by
#' @inheritParams dplyr::summarise
#' @inheritParams dplyr::mutate
#' @name dplyr_table_manipulation
#' @examplesIf rlang::is_installed("nycflights13")
#' zoomed <- dm_nycflights13() %>%
#'   dm_zoom_to(flights) %>%
#'   arrange(desc(day)) %>%
#'   summarize(.by = month, avg_air_time = mean(air_time, na.rm = TRUE))
#' zoomed
#' dm_insert_zoomed(zoomed, new_tbl_name = "avg_air_time_per_month")
NULL

#' @export
filter.dm <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @export
filter.dm_zoomed <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  tbl <- tbl_zoomed(.data)
  filtered_tbl <- filter(tbl, ..., .by = {{ .by }}, .preserve = .preserve)
  replace_zoomed_tbl(.data, filtered_tbl)
}

#' @rdname dplyr_table_manipulation
#' @export
filter.dm_keyed_tbl <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  filtered_tbl <- filter(tbl, ..., .by = {{ .by }}, .preserve = .preserve)
  new_keyed_tbl_from_keys_info(filtered_tbl, keys_info)
}

#' @export
filter_out.dm <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @export
filter_out.dm_zoomed <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  tbl <- tbl_zoomed(.data)
  filtered_tbl <- filter_out(tbl, ..., .by = {{ .by }}, .preserve = .preserve)
  replace_zoomed_tbl(.data, filtered_tbl)
}

#' @rdname dplyr_table_manipulation
#' @export
filter_out.dm_keyed_tbl <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  filtered_tbl <- filter_out(tbl, ..., .by = {{ .by }}, .preserve = .preserve)
  new_keyed_tbl_from_keys_info(filtered_tbl, keys_info)
}

#' @export
mutate.dm <- function(
  .data,
  ...,
  .by = NULL,
  .keep = c("all", "used", "unused", "none"),
  .before = NULL,
  .after = NULL
) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @export
mutate.dm_zoomed <- function(
  .data,
  ...,
  .by = NULL,
  .keep = c("all", "used", "unused", "none"),
  .before = NULL,
  .after = NULL
) {
  tbl <- tbl_zoomed(.data)
  mutated_tbl <- mutate(
    tbl,
    ...,
    .by = {{ .by }},
    .keep = .keep,
    .before = {{ .before }},
    .after = {{ .after }}
  )
  # #663: user responsibility: those columns are tracked whose names remain
  selected <- set_names(intersect(colnames(tbl), colnames(mutated_tbl)))
  new_tracked_cols_zoom <- new_tracked_cols(.data, selected)
  replace_zoomed_tbl(.data, mutated_tbl, new_tracked_cols_zoom)
}

#' @rdname dplyr_table_manipulation
#' @export
mutate.dm_keyed_tbl <- function(
  .data,
  ...,
  .by = NULL,
  .keep = c("all", "used", "unused", "none"),
  .before = NULL,
  .after = NULL
) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  out <- mutate(
    tbl,
    ...,
    .by = {{ .by }},
    .keep = .keep,
    .before = {{ .before }},
    .after = {{ .after }}
  )
  new_keyed_tbl_from_keys_info(out, keys_info)
}

#' @export
transmute.dm <- function(.data, ...) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @export
transmute.dm_zoomed <- function(.data, ...) {
  tbl <- tbl_zoomed(.data)
  # groups are "selected"; key tracking will continue for them
  groups <- set_names(map_chr(groups(tbl), as_string))
  transmuted_tbl <- transmute(tbl, ...)

  # #663: user responsibility: those columns are tracked whose names remain
  selected <- set_names(intersect(colnames(tbl), colnames(transmuted_tbl)))
  new_tracked_cols_zoom <- new_tracked_cols(.data, selected)

  replace_zoomed_tbl(.data, transmuted_tbl, new_tracked_cols_zoom)
}

#' @rdname dplyr_table_manipulation
#' @export
transmute.dm_keyed_tbl <- function(.data, ...) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  out <- transmute(tbl, ...)
  new_keyed_tbl_from_keys_info(out, keys_info)
}

#' @export
select.dm <- function(.data, ...) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @export
select.dm_zoomed <- function(.data, ...) {
  tbl <- tbl_zoomed(.data)

  selected <- eval_select_both(quo(c(...)), colnames(tbl))
  selected_tbl <- select(tbl, !!!selected$indices)
  new_tracked_cols_zoom <- new_tracked_cols(.data, selected$names)

  replace_zoomed_tbl(.data, selected_tbl, new_tracked_cols_zoom)
}

#' @rdname dplyr_table_manipulation
#' @export
select.dm_keyed_tbl <- function(.data, ...) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  out <- select(tbl, ...)
  new_keyed_tbl_from_keys_info(out, keys_info)
}

#' @export
relocate.dm <- function(.data, ..., .before = NULL, .after = NULL) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @inheritParams dplyr::relocate
#' @export
relocate.dm_zoomed <- function(.data, ..., .before = NULL, .after = NULL) {
  tbl <- tbl_zoomed(.data)

  relocated_tbl <- relocate(tbl, ..., .before = {{ .before }}, .after = {{ .after }})
  replace_zoomed_tbl(.data, relocated_tbl)
}

#' @rdname dplyr_table_manipulation
#' @export
relocate.dm_keyed_tbl <- function(.data, ..., .before = NULL, .after = NULL) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  relocated_tbl <- relocate(tbl, ..., .before = {{ .before }}, .after = {{ .after }})
  new_keyed_tbl_from_keys_info(relocated_tbl, keys_info)
}

#' @export
rename.dm <- function(.data, ...) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @export
rename.dm_zoomed <- function(.data, ...) {
  tbl <- tbl_zoomed(.data)

  renamed <- eval_rename_both(quo(c(...)), colnames(tbl))
  renamed_tbl <- rename(tbl, !!!renamed$indices)
  new_tracked_cols_zoom <- new_tracked_cols(.data, renamed$all_names)

  replace_zoomed_tbl(.data, renamed_tbl, new_tracked_cols_zoom)
}

#' @rdname dplyr_table_manipulation
#' @export
rename.dm_keyed_tbl <- function(.data, ...) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  out <- rename(tbl, ...)
  new_keyed_tbl_from_keys_info(out, keys_info)
}

#' @export
distinct.dm <- function(.data, ..., .keep_all = FALSE) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @param .keep_all For `distinct.dm_zoomed()`: see [dplyr::distinct()]
#' @export
distinct.dm_zoomed <- function(.data, ..., .keep_all = FALSE) {
  tbl <- tbl_zoomed(.data)
  distinct_tbl <- distinct(tbl, ..., .keep_all = .keep_all)

  # #663: user responsibility: those columns are tracked whose names remain
  selected <- set_names(intersect(colnames(tbl), colnames(distinct_tbl)))
  new_tracked_cols_zoom <- new_tracked_cols(.data, selected)

  replace_zoomed_tbl(.data, distinct_tbl, new_tracked_cols_zoom)
}

#' @rdname dplyr_table_manipulation
#' @export
distinct.dm_keyed_tbl <- function(.data, ..., .keep_all = FALSE) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  distinct_tbl <- distinct(tbl, ..., .keep_all = .keep_all)
  new_keyed_tbl_from_keys_info(distinct_tbl, keys_info)
}

#' @export
arrange.dm <- function(.data, ..., .by_group = FALSE, .locale = NULL) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @export
arrange.dm_zoomed <- function(.data, ..., .by_group = FALSE, .locale = NULL) {
  tbl <- tbl_zoomed(.data)
  arranged_tbl <- arrange(tbl, ..., .by_group = .by_group, .locale = .locale)
  replace_zoomed_tbl(.data, arranged_tbl)
}

#' @rdname dplyr_table_manipulation
#' @export
arrange.dm_keyed_tbl <- function(.data, ..., .by_group = FALSE, .locale = NULL) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  arranged_tbl <- arrange(tbl, ..., .by_group = .by_group, .locale = .locale)
  new_keyed_tbl_from_keys_info(arranged_tbl, keys_info)
}

#' @export
slice.dm <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @param .keep_pk For `slice.dm_zoomed`: Logical, if `TRUE`, the primary key will be retained during this transformation. If `FALSE`, it will be dropped.
#' By default, the value is `NULL`, which causes the function to issue a message in case a primary key is available for the zoomed table.
#' This argument is specific for the `slice.dm_zoomed()` method.
#' @export
slice.dm_zoomed <- function(.data, ..., .by = NULL, .preserve = FALSE, .keep_pk = NULL) {
  sliced_tbl <- slice(tbl_zoomed(.data), ..., .by = {{ .by }}, .preserve = .preserve)
  orig_pk <- dm_get_pk_impl(.data, orig_name_zoomed(.data))
  tracked_cols <- col_tracker_zoomed(.data)
  if (is_null(.keep_pk)) {
    if (has_length(orig_pk) && any(unlist(orig_pk) %in% tracked_cols)) {
      message(
        paste(
          "Keeping PK column, but `slice.dm_zoomed()` can potentially damage the uniqueness of PK columns (duplicated indices).",
          "Set argument `.keep_pk` to `TRUE` or `FALSE` to ensure the behavior you intended."
        )
      )
    }
  } else if (!.keep_pk) {
    tracked_cols <- discard(tracked_cols, tracked_cols == orig_pk)
  }
  replace_zoomed_tbl(.data, sliced_tbl, tracked_cols)
}

#' @rdname dplyr_table_manipulation
#' @export
slice.dm_keyed_tbl <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  sliced_tbl <- slice(tbl, ..., .by = {{ .by }}, .preserve = .preserve)
  new_keyed_tbl_from_keys_info(sliced_tbl, keys_info)
}

#' @export
group_by.dm <- function(.data, ..., .add = FALSE, .drop = group_by_drop_default(.data)) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @export
group_by.dm_zoomed <- function(.data, ..., .add = FALSE, .drop = group_by_drop_default(.data)) {
  tbl <- tbl_zoomed(.data)
  grouped_tbl <- group_by(tbl, ..., .add = .add, .drop = .drop)

  replace_zoomed_tbl(.data, grouped_tbl)
}

#' @rdname dplyr_table_manipulation
#' @export
group_by.dm_keyed_tbl <- function(.data, ..., .add = FALSE, .drop = group_by_drop_default(.data)) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)
  grouped_tbl <- group_by(tbl, ..., .add = .add, .drop = .drop)
  new_keyed_tbl_from_keys_info(grouped_tbl, keys_info)
}

#' @export
group_data.dm <- function(.data) {
  check_zoomed(.data)
}

#' @export
group_data.dm_zoomed <- function(.data) {
  tbl <- tbl_zoomed(.data)
  group_data(tbl)
}

#' @export
group_keys.dm <- function(.tbl, ...) {
  check_zoomed(.tbl)
}

#' @export
group_keys.dm_zoomed <- function(.tbl, ...) {
  .data <- .tbl
  tbl <- tbl_zoomed(.data)
  group_keys(tbl, ...)
}

#' @export
group_indices.dm <- function(.data, ...) {
  check_zoomed(.data)
}

#' @export
group_indices.dm_zoomed <- function(.data, ...) {
  tbl <- tbl_zoomed(.data)
  group_indices(tbl, ...)
}

#' @export
group_vars.dm <- function(x) {
  check_zoomed(x)
}

#' @export
group_vars.dm_zoomed <- function(x) {
  .data <- x
  tbl <- tbl_zoomed(.data)
  group_vars(tbl)
}

#' @export
groups.dm <- function(x) {
  check_zoomed(x)
}

#' @export
groups.dm_zoomed <- function(x) {
  .data <- x
  tbl <- tbl_zoomed(.data)
  groups(tbl)
}

#' @export
ungroup.dm <- function(x, ...) {
  check_zoomed(x)
}

#' @rdname dplyr_table_manipulation
#' @param x For `ungroup.dm_zoomed`: object of class `dm_zoomed`
#' @export
ungroup.dm_zoomed <- function(x, ...) {
  tbl <- tbl_zoomed(x)
  ungrouped_tbl <- ungroup(tbl, ...)

  replace_zoomed_tbl(x, ungrouped_tbl)
}

#' @rdname dplyr_table_manipulation
#' @export
ungroup.dm_keyed_tbl <- function(x, ...) {
  keys_info <- keyed_get_info(x)
  tbl <- unclass_keyed_tbl(x)
  ungrouped_tbl <- ungroup(tbl, ...)
  new_keyed_tbl_from_keys_info(ungrouped_tbl, keys_info)
}

#' @export
summarise.dm <- function(.data, ..., .by = NULL, .groups = NULL) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @export
summarise.dm_zoomed <- function(.data, ..., .by = NULL, .groups = NULL) {
  tbl <- tbl_zoomed(.data)
  by <- enquo(.by)
  if (!quo_is_null(by)) {
    by_loc <- tidyselect::eval_select(by, data = tbl)
    groups <- set_names(names(by_loc))
  } else {
    # groups are "selected"; key tracking will continue for them
    # #663: user responsibility: if group columns are manipulated, they are still tracked
    groups <- set_names(map_chr(groups(tbl), as_string))
  }
  summarized_tbl <- summarize(tbl, ..., .by = !!by, .groups = .groups)
  new_tracked_cols_zoom <- new_tracked_cols(.data, groups)
  replace_zoomed_tbl(.data, summarized_tbl, new_tracked_cols_zoom)
}


#' @rdname dplyr_table_manipulation
#' @export
summarise.dm_keyed_tbl <- function(.data, ..., .by = NULL, .groups = NULL) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)

  by <- enquo(.by)
  if (!quo_is_null(by)) {
    by_loc <- tidyselect::eval_select(by, data = tbl)
    new_pk <- names(by_loc)
  } else {
    new_pk <- group_vars(.data)
  }
  if (length(new_pk) == 0) {
    new_pk <- NULL
  }

  summarised_tbl <- summarise(tbl, ..., .by = !!by, .groups = .groups)

  # FIXME: Add original FKs for the remaining columns
  # (subsets of the grouped columns), use new UUID
  new_keyed_tbl(
    summarised_tbl,
    pk = new_pk,
    uuid = keys_info$uuid,
    zoom2 = keys_info$zoom2
  )
}

#' @export
reframe.dm <- function(.data, ..., .by = NULL) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @export
reframe.dm_zoomed <- function(.data, ..., .by = NULL) {
  tbl <- tbl_zoomed(.data)
  reframed_tbl <- reframe(tbl, ..., .by = {{ .by }})
  # reframe() always returns ungrouped data, no group tracking needed
  # #663: user responsibility: those columns are tracked whose names remain
  selected <- set_names(intersect(colnames(tbl_zoomed(.data)), colnames(reframed_tbl)))
  new_tracked_cols_zoom <- new_tracked_cols(.data, selected)
  replace_zoomed_tbl(.data, reframed_tbl, new_tracked_cols_zoom)
}

#' @rdname dplyr_table_manipulation
#' @export
reframe.dm_keyed_tbl <- function(.data, ..., .by = NULL) {
  keys_info <- keyed_get_info(.data)
  tbl <- unclass_keyed_tbl(.data)

  reframed_tbl <- reframe(tbl, ..., .by = {{ .by }})

  # reframe() can return any number of rows per group,

  # so the primary key is not preserved
  new_keyed_tbl(
    reframed_tbl,
    uuid = keys_info$uuid,
    zoom2 = keys_info$zoom2
  )
}

#' @export
count.dm <- function(
  x,
  ...,
  wt = NULL,
  sort = FALSE,
  name = NULL,
  .drop = group_by_drop_default(x)
) {
  check_zoomed(x)
}

#' @rdname dplyr_table_manipulation
#' @inheritParams dplyr::count
#' @export
count.dm_zoomed <- function(
  x,
  ...,
  wt = NULL,
  sort = FALSE,
  name = NULL,
  .drop = group_by_drop_default(x)
) {
  tbl <- tbl_zoomed(x)

  if (!missing(...)) {
    out <- group_by(tbl, ..., .add = TRUE, .drop = .drop)
  } else {
    out <- tbl
  }

  groups <- set_names(map_chr(groups(out), as_string))

  out <- tally(out, wt = !!enquo(wt), sort = sort, name = name)

  # Ensure grouping is transient
  if (is.data.frame(tbl)) {
    out <- dplyr_reconstruct(out, tbl)
  }

  new_tracked_cols_zoom <- new_tracked_cols(x, groups)
  replace_zoomed_tbl(x, out, new_tracked_cols_zoom)
}

#' @rdname dplyr_table_manipulation
#' @export
count.dm_keyed_tbl <- function(
  x,
  ...,
  wt = NULL,
  sort = FALSE,
  name = NULL,
  .drop = group_by_drop_default(x)
) {
  keys_info <- keyed_get_info(x)
  tbl <- unclass_keyed_tbl(x)
  counted_tbl <- count(tbl, ..., wt = {{ wt }}, sort = sort, name = name, .drop = .drop)
  new_keyed_tbl(
    counted_tbl,
    uuid = keys_info$uuid,
    zoom2 = keys_info$zoom2
  )
}

#' @export
tally.dm <- function(x, wt = NULL, sort = FALSE, name = NULL) {
  check_zoomed(x)
}

#' @rdname dplyr_table_manipulation
#' @export
tally.dm_zoomed <- function(x, wt = NULL, sort = FALSE, name = NULL) {
  tbl <- tbl_zoomed(x)
  groups <- set_names(map_chr(groups(tbl), as_string))

  out <- tally(tbl, wt = {{ wt }}, sort = sort, name = name)

  # Ensure grouping is transient
  if (is.data.frame(tbl)) {
    out <- dplyr_reconstruct(out, tbl)
  }

  new_tracked_cols_zoom <- new_tracked_cols(x, groups)
  replace_zoomed_tbl(x, out, new_tracked_cols_zoom)
}

#' @rdname dplyr_table_manipulation
#' @export
tally.dm_keyed_tbl <- function(x, wt = NULL, sort = FALSE, name = NULL) {
  keys_info <- keyed_get_info(x)
  tbl <- unclass_keyed_tbl(x)
  tallied_tbl <- tally(tbl, wt = {{ wt }}, sort = sort, name = name)
  new_keyed_tbl(
    tallied_tbl,
    uuid = keys_info$uuid,
    zoom2 = keys_info$zoom2
  )
}

#' @export
pull.dm <- function(.data, var = -1, name = NULL, ...) {
  check_zoomed(.data)
}

#' @rdname dplyr_table_manipulation
#' @inheritParams dplyr::pull
#' @export
pull.dm_zoomed <- function(.data, var = -1, name = NULL, ...) {
  tbl <- tbl_zoomed(.data)
  pull(tbl, var = {{ var }}, name = {{ name }}, ...)
}

#' @rdname dplyr_table_manipulation
#' @export
compute.dm_zoomed <- function(x, ...) {
  dm_zoomed_df <-
    tbl_zoomed(x) %>%
    compute(...)
  replace_zoomed_tbl(x, dm_zoomed_df)
}

#' \pkg{dplyr} join methods for zoomed dm objects
#'
#' Use these methods without the '.dm_zoomed' suffix (see examples).
#' @name dplyr_join
#' @param x,y tbls to join. `x` is the `dm_zoomed` and `y` is another table in the `dm`.
#' @param by If left `NULL` (default), the join will be performed by via the foreign key relation that exists between the originally zoomed table (now `x`)
#' and the other table (`y`).
#' If you provide a value (for the syntax see [`dplyr::join`]), you can also join tables that are not connected in the `dm`.
#' @param copy Disabled, since all tables in a `dm` are by definition on the same `src`.
#' @param suffix Disabled, since columns are disambiguated automatically if necessary, changing the column names to `table_name.column_name`.
#' @param select Select a subset of the \strong{RHS-table}'s columns, the syntax being `select = c(col_1, col_2, col_3)` (unquoted or quoted).
#' This argument is specific for the `join`-methods for `dm_zoomed`.
#' The table's `by` column(s) are automatically added if missing in the selection.
#' @param ... see [`dplyr::join`]
#' @inheritParams dplyr::left_join
#' @examplesIf rlang::is_installed("nycflights13")
#' flights_dm <- dm_nycflights13()
#' dm_zoom_to(flights_dm, flights) %>%
#'   left_join(airports, select = c(faa, name))
#'
#' # this should illustrate that tables don't necessarily need to be connected
#' dm_zoom_to(flights_dm, airports) %>%
#'   semi_join(airlines, by = "name")
NULL

#' @export
left_join.dm <- function(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL
) {
  check_zoomed(x)
}

#' @rdname dplyr_join
#' @export
left_join.dm_zoomed <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL,
  select = NULL
) {
  y_name <- as_string(enexpr(y))
  join_data <- prepare_join(x, {{ y }}, by, {{ select }}, suffix, copy)
  joined_tbl <- left_join(
    join_data$x_tbl,
    join_data$y_tbl,
    join_data$by,
    copy = FALSE,
    ...,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    unmatched = unmatched,
    relationship = relationship
  )
  replace_zoomed_tbl(x, joined_tbl, join_data$new_col_names)
}

#' @rdname dplyr_join
#' @export
left_join.dm_keyed_tbl <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = FALSE,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL
) {
  if (!is_dm_keyed_tbl(y)) {
    return(NextMethod())
  }

  join_spec <- keyed_build_join_spec(x, y, by, suffix)
  joined_tbl <- left_join(
    join_spec$x_tbl,
    join_spec$y_tbl,
    deframe(join_spec$by),
    copy = copy,
    suffix = join_spec$suffix,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    unmatched = unmatched,
    relationship = relationship,
    ...
  )

  new_keyed_tbl(
    joined_tbl,
    pk = join_spec$new_pk,
    fks_in = join_spec$new_fks_in,
    fks_out = join_spec$new_fks_out,
    uuid = join_spec$new_uuid,
    zoom2 = keyed_get_info(x)$zoom2
  )
}

#' @export
inner_join.dm <- function(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL
) {
  check_zoomed(x)
}

#' @rdname dplyr_join
#' @export
inner_join.dm_zoomed <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL,
  select = NULL
) {
  y_name <- as_string(enexpr(y))
  join_data <- prepare_join(x, {{ y }}, by, {{ select }}, suffix, copy)
  joined_tbl <- inner_join(
    join_data$x_tbl,
    join_data$y_tbl,
    join_data$by,
    copy = FALSE,
    ...,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    unmatched = unmatched,
    relationship = relationship
  )
  replace_zoomed_tbl(x, joined_tbl, join_data$new_col_names)
}

#' @rdname dplyr_join
#' @export
inner_join.dm_keyed_tbl <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = FALSE,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL
) {
  if (!is_dm_keyed_tbl(y)) {
    return(NextMethod())
  }

  join_spec <- keyed_build_join_spec(x, y, by, suffix)
  joined_tbl <- inner_join(
    join_spec$x_tbl,
    join_spec$y_tbl,
    deframe(join_spec$by),
    copy = copy,
    suffix = join_spec$suffix,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    unmatched = unmatched,
    relationship = relationship,
    ...
  )

  new_keyed_tbl(
    joined_tbl,
    pk = join_spec$new_pk,
    fks_in = join_spec$new_fks_in,
    fks_out = join_spec$new_fks_out,
    uuid = join_spec$new_uuid,
    zoom2 = keyed_get_info(x)$zoom2
  )
}

#' @export
full_join.dm <- function(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  relationship = NULL
) {
  check_zoomed(x)
}

#' @rdname dplyr_join
#' @export
full_join.dm_zoomed <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  relationship = NULL,
  select = NULL
) {
  y_name <- as_string(enexpr(y))
  join_data <- prepare_join(x, {{ y }}, by, {{ select }}, suffix, copy)
  joined_tbl <- full_join(
    join_data$x_tbl,
    join_data$y_tbl,
    join_data$by,
    copy = FALSE,
    ...,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    relationship = relationship
  )
  replace_zoomed_tbl(x, joined_tbl, join_data$new_col_names)
}

#' @rdname dplyr_join
#' @export
full_join.dm_keyed_tbl <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = FALSE,
  na_matches = c("na", "never"),
  multiple = "all",
  relationship = NULL
) {
  if (!is_dm_keyed_tbl(y)) {
    return(NextMethod())
  }

  join_spec <- keyed_build_join_spec(x, y, by, suffix)
  joined_tbl <- full_join(
    join_spec$x_tbl,
    join_spec$y_tbl,
    deframe(join_spec$by),
    copy = copy,
    suffix = join_spec$suffix,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    relationship = relationship,
    ...
  )

  new_keyed_tbl(
    joined_tbl,
    pk = join_spec$new_pk,
    fks_in = join_spec$new_fks_in,
    fks_out = join_spec$new_fks_out,
    uuid = join_spec$new_uuid,
    zoom2 = keyed_get_info(x)$zoom2
  )
}

#' @export
right_join.dm <- function(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL
) {
  check_zoomed(x)
}

#' @rdname dplyr_join
#' @export
right_join.dm_zoomed <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL,
  select = NULL
) {
  y_name <- as_string(enexpr(y))
  join_data <- prepare_join(x, {{ y }}, by, {{ select }}, suffix, copy)
  joined_tbl <- right_join(
    join_data$x_tbl,
    join_data$y_tbl,
    join_data$by,
    copy = FALSE,
    ...,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    unmatched = unmatched,
    relationship = relationship
  )
  replace_zoomed_tbl(x, joined_tbl, join_data$new_col_names)
}

#' @rdname dplyr_join
#' @export
right_join.dm_keyed_tbl <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = FALSE,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL
) {
  if (!is_dm_keyed_tbl(y)) {
    return(NextMethod())
  }

  join_spec <- keyed_build_join_spec(x, y, by, suffix)
  joined_tbl <- right_join(
    join_spec$x_tbl,
    join_spec$y_tbl,
    deframe(join_spec$by),
    copy = copy,
    suffix = join_spec$suffix,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    unmatched = unmatched,
    relationship = relationship,
    ...
  )

  new_keyed_tbl(
    joined_tbl,
    pk = join_spec$new_pk,
    fks_in = join_spec$new_fks_in,
    fks_out = join_spec$new_fks_out,
    uuid = join_spec$new_uuid,
    zoom2 = keyed_get_info(x)$zoom2
  )
}

#' @export
semi_join.dm <- function(x, y, by = NULL, copy = FALSE, ..., na_matches = c("na", "never")) {
  check_zoomed(x)
}

#' @rdname dplyr_join
#' @export
semi_join.dm_zoomed <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  ...,
  na_matches = c("na", "never"),
  suffix = NULL,
  select = NULL
) {
  y_name <- as_string(enexpr(y))
  join_data <- prepare_join(x, {{ y }}, by, {{ select }}, suffix, copy, disambiguate = FALSE)
  joined_tbl <- semi_join(
    join_data$x_tbl,
    join_data$y_tbl,
    join_data$by,
    copy = FALSE,
    ...,
    na_matches = na_matches
  )
  replace_zoomed_tbl(x, joined_tbl, join_data$new_col_names)
}

#' @rdname dplyr_join
#' @export
semi_join.dm_keyed_tbl <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  ...,
  na_matches = c("na", "never")
) {
  if (!is_dm_keyed_tbl(y)) {
    return(NextMethod())
  }

  if (is.null(by)) {
    by <- keyed_by(x, y)
  }

  joined_tbl <- semi_join(
    unclass_keyed_tbl(x),
    unclass_keyed_tbl(y),
    by,
    copy = copy,
    ...
  )

  new_keyed_tbl_from_keys_info(joined_tbl, keyed_get_info(x))
}

#' @export
anti_join.dm <- function(x, y, by = NULL, copy = FALSE, ..., na_matches = c("na", "never")) {
  check_zoomed(x)
}

#' @rdname dplyr_join
#' @export
anti_join.dm_zoomed <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  ...,
  na_matches = c("na", "never"),
  suffix = NULL,
  select = NULL
) {
  y_name <- as_string(enexpr(y))
  join_data <- prepare_join(x, {{ y }}, by, {{ select }}, suffix, copy, disambiguate = FALSE)
  joined_tbl <- anti_join(
    join_data$x_tbl,
    join_data$y_tbl,
    join_data$by,
    copy = FALSE,
    ...,
    na_matches = na_matches
  )
  replace_zoomed_tbl(x, joined_tbl, join_data$new_col_names)
}

#' @rdname dplyr_join
#' @export
anti_join.dm_keyed_tbl <- function(
  x,
  y,
  by = NULL,
  copy = NULL,
  ...,
  na_matches = c("na", "never")
) {
  if (!is_dm_keyed_tbl(y)) {
    return(NextMethod())
  }

  if (is.null(by)) {
    by <- keyed_by(x, y)
  }

  joined_tbl <- anti_join(
    unclass_keyed_tbl(x),
    unclass_keyed_tbl(y),
    by,
    copy = copy,
    ...
  )

  new_keyed_tbl_from_keys_info(joined_tbl, keyed_get_info(x))
}

#' @export
nest_join.dm <- function(
  x,
  y,
  by = NULL,
  copy = FALSE,
  keep = NULL,
  name = NULL,
  ...,
  na_matches = c("na", "never"),
  unmatched = "drop"
) {
  check_zoomed(x)
}

#' @rdname dplyr_join
#' @inheritParams dplyr::nest_join
#' @export
nest_join.dm_zoomed <- function(
  x,
  y,
  by = NULL,
  copy = FALSE,
  keep = NULL,
  name = NULL,
  ...,
  na_matches = c("na", "never"),
  unmatched = "drop"
) {
  y_name <- as_string(enexpr(y))
  name <- name %||% y_name
  join_data <- prepare_join(x, {{ y }}, by, NULL, NULL, NULL, disambiguate = FALSE)
  joined_tbl <- nest_join(join_data$x_tbl, join_data$y_tbl, join_data$by, copy, keep, name, ...)
  replace_zoomed_tbl(x, joined_tbl, join_data$new_col_names)
}

#' @export
cross_join.dm <- function(x, y, ..., copy = FALSE, suffix = c(".x", ".y")) {
  check_zoomed(x)
}

#' @rdname dplyr_join
#' @export
cross_join.dm_zoomed <- function(x, y, ..., copy = NULL, suffix = c(".x", ".y")) {
  if (!is_null(copy)) {
    message("Tables in a `dm` are necessarily on the same `src`, setting `copy = FALSE`.")
  }
  y_name <- dm_tbl_name(x, {{ y }})
  zoomed <- dm_get_zoom(x, c("table", "zoom", "col_tracker_zoom"))
  x_tbl <- zoomed$zoom[[1]]
  y_tbl <- dm_get_tables_impl(x)[[y_name]]
  new_col_names <- zoomed$col_tracker_zoom[[1]]
  joined_tbl <- cross_join(x_tbl, y_tbl, ..., copy = FALSE, suffix = suffix)
  replace_zoomed_tbl(x, joined_tbl, new_col_names)
}

#' @rdname dplyr_join
#' @export
cross_join.dm_keyed_tbl <- function(x, y, ..., copy = NULL, suffix = c(".x", ".y")) {
  if (!is_dm_keyed_tbl(y)) {
    return(NextMethod())
  }

  joined_tbl <- cross_join(
    unclass_keyed_tbl(x),
    unclass_keyed_tbl(y),
    ...,
    copy = copy,
    suffix = suffix
  )

  new_keyed_tbl(
    joined_tbl,
    uuid = keyed_get_info(x)$uuid,
    zoom2 = keyed_get_info(x)$zoom2
  )
}

#' @autoglobal
prepare_join <- function(
  x,
  y,
  by,
  selected,
  suffix,
  copy,
  disambiguate = TRUE,
  error_call = caller_env()
) {
  y_name <- dm_tbl_name(x, {{ y }})
  select_quo <- enquo(selected)

  if (!is_null(suffix)) {
    message("Column names are disambiguated if necessary, `suffix` ignored.")
  }
  if (!is_null(copy)) {
    message("Tables in a `dm` are necessarily on the same `src`, setting `copy = FALSE`.")
  }

  zoomed <- dm_get_zoom(x, c("table", "zoom", "col_tracker_zoom"))

  x_tbl <- zoomed$zoom[[1]]
  x_orig_name <- zoomed$table
  y_tbl <- dm_get_tables_impl(x)[[y_name]]
  all_cols_y <- colnames(y_tbl)

  if (quo_is_null(select_quo)) {
    select_quo <- quo(everything())
  }

  selected <- eval_select_both(select_quo, colnames(y_tbl), error_call = error_call)$names

  new_col_names <- zoomed$col_tracker_zoom[[1]]

  if (is_null(by)) {
    by <- get_by(x, x_orig_name, y_name)

    # If the original FK-relation between original `x` and `y` got lost, `by` needs to be provided explicitly
    if (!all(names(by) %in% new_col_names)) abort_fk_not_tracked(x_orig_name, y_name)
  }

  by <- normalize_join_by(by)

  # selection without RHS `by`; only this is needed for disambiguation and by-columns are added later on for all join-types
  selected_wo_by <- selected[selected %in% setdiff(selected, by)]

  if (disambiguate) {
    x_disambig_name <- x_orig_name
    y_disambig_name <- y_name
    if (x_disambig_name == y_disambig_name) {
      x_disambig_name <- paste0(x_disambig_name, ".x")
      y_disambig_name <- paste0(y_disambig_name, ".y")
    }

    table_colnames <-
      vec_rbind(
        tibble(table = x_disambig_name, column = colnames(x_tbl)),
        tibble(table = y_disambig_name, column = names(selected_wo_by))
      )

    recipe <- compute_disambiguate_cols_recipe(table_colnames, sep = ".")
    explain_col_rename(recipe)

    x_renames <-
      recipe %>%
      filter(table == x_disambig_name) %>%
      pull(renames)
    y_renames <-
      recipe %>%
      filter(table == y_disambig_name) %>%
      pull(renames)

    if (has_length(x_renames)) {
      x_tbl <- x_tbl %>% rename(!!!x_renames[[1]])
      names(by) <- recode_compat(names2(by), prep_recode(x_renames[[1]]))
      names(new_col_names) <- recode_compat(names(new_col_names), prep_recode(x_renames[[1]]))
    }

    if (has_length(y_renames)) {
      names(selected_wo_by) <- recode_compat(names(selected_wo_by), prep_recode(y_renames[[1]]))
    }
  }

  # inform user in case RHS `by` column(s) are added
  if (!all(by %in% selected)) {
    new_cols <- glue_collapse(tick_if_needed(setdiff(by, selected)), ", ")
  }

  # rename RHS `by` columns in the tibble to avoid after-the-fact disambiguation collision
  prefix <- unique_prefix(names(selected_wo_by))
  by_rhs_rename <- by
  names(by_rhs_rename) <- paste0(prefix, seq_along(by_rhs_rename))
  stopifnot(!any(names(selected_wo_by) %in% names(by_rhs_rename)))

  # complete vector of selection consisting of selection without by and the newly named by-columns
  selected_repaired <- c(selected_wo_by, by_rhs_rename)

  y_tbl <- select(y_tbl, !!!selected_repaired)

  # the `by` argument needs to be updated: LHS stays, RHS needs to be replaced with new names
  repaired_by <- set_names(recode_compat(by, prep_recode(by_rhs_rename)), names(by))

  # in case key columns of x_tbl have the same name as selected columns of y_tbl
  # the column names of x will be adapted (not for `semi_join()` and `anti_join()`)
  # We can track the new column names
  list(x_tbl = x_tbl, y_tbl = y_tbl, by = repaired_by, new_col_names = new_col_names)
}

unique_prefix <- function(x) {
  if (is_empty(x)) {
    return("...")
  }

  dots <- max(max(nchar(x, "bytes")), 3)
  paste(rep(".", dots), collapse = "")
}

# Workaround for dev dplyr + dbplyr
safe_count <- function(
  x,
  ...,
  wt = NULL,
  sort = FALSE,
  name = NULL,
  .drop = group_by_drop_default(x)
) {
  quos <- enquos(...)

  if (has_length(quos)) {
    named <- names2(quos) != ""
    if (any(named)) {
      quos <- as.list(quos)
      named_quos <- quos[named]
      x <- mutate(x, !!!named_quos)
      quos[named] <- syms(names2(quos)[named])
      names(quos) <- NULL
    }
    out <- group_by(x, !!!quos, .add = FALSE, .drop = .drop)
  } else {
    out <- ungroup(x)
  }

  # Compatibility for dplyr < 1.0.0
  if (is.null(name)) {
    out <- tally(out, wt = !!enquo(wt), sort = sort)
  } else {
    out <- tally(out, wt = !!enquo(wt), sort = sort, name = name)
  }
  ungroup(out)
}

new_tracked_cols <- function(dm, selected) {
  tracked_cols <- col_tracker_zoomed(dm)
  old_tracked_names <- names(tracked_cols)
  # the new tracked keys need to be the remaining original column names
  # and their name needs to be the newest one (tidyselect-syntax)
  # `intersect(selected, old_tracked_names)` is empty, return `NULL`

  selected_match <- selected[selected %in% old_tracked_names]
  set_names(
    tracked_cols[selected_match],
    names(selected_match)
  )
}

rlang::on_load({
  if (rlang::is_installed("shiny")) {
    shiny::registerInputHandler("dm_nodes_edges", force = TRUE, shiny_input_dm_nodes_edges)
  }
})

utils::globalVariables(c(
  ".",
  "child_table",
  "id",
  "name",
  "on_delete",
  "op",
  "parent_table",
  "pk_col",
  NULL
))

"%:::%" <- function(p, f) {
  get(f, envir = asNamespace(p))
}

new_uuid <- function() {
  out <- as.raw(sample.int(16, 16) - 1L)
  paste0(
    paste0(out[1:4], collapse = ""),
    "-",
    paste0(out[5:6], collapse = ""),
    "-",
    paste0(out[7:8], collapse = ""),
    "-",
    paste0(out[9:10], collapse = ""),
    "-",
    paste0(out[11:16], collapse = "")
  )
}

vec_new_uuid_along <- function(x) {
  map_chr(x, function(.x) new_uuid())
}

new_fks <- function(
  ...,
  child_uuid = NULL,
  child_fk_cols = NULL,
  parent_uuid = NULL,
  parent_key_cols = NULL
) {
  check_dots_empty0(...)

  child_fk_cols <- new_keys(child_fk_cols)
  parent_key_cols <- new_keys(parent_key_cols)
  tibble(child_uuid, child_fk_cols, parent_uuid, parent_key_cols)
}

new_fks_in <- function(child_uuid = NULL, child_fk_cols = NULL, parent_key_cols = NULL) {
  new_fks(
    child_uuid = vec_cast(child_uuid, character()) %||% character(),
    child_fk_cols = child_fk_cols,
    parent_key_cols = parent_key_cols
  )
}

new_fks_out <- function(child_fk_cols = NULL, parent_uuid = NULL, parent_key_cols = NULL) {
  new_fks(
    child_fk_cols = child_fk_cols,
    parent_uuid = vec_cast(parent_uuid, character()) %||% character(),
    parent_key_cols = parent_key_cols
  )
}

new_keyed_tbl <- function(
  x,
  ...,
  pk = NULL,
  uks = NULL,
  fks_in = NULL,
  fks_out = NULL,
  uuid = NULL,
  zoom2 = NULL
) {
  check_dots_empty()

  pk <- vec_cast(pk, character())
  uks <- vec_cast(uks, new_uk()) %||% new_uk()
  fks_in <- vec_cast(fks_in, new_fks_in()) %||% new_fks_in()
  fks_out <- vec_cast(fks_out, new_fks_out()) %||% new_fks_out()

  if (is.null(uuid)) {
    uuid <- new_uuid()
  }

  class(x) <- unique(c("dm_keyed_tbl", class(x)))
  dm_key_info <- list(
    pk = pk,
    uks = uks,
    fks_in = fks_in,
    fks_out = fks_out,
    uuid = uuid
  )

  if (!is.null(zoom2)) {
    dm_key_info$zoom2 <- zoom2
  }

  attr(x, "dm_key_info") <- dm_key_info

  x
}

new_keyed_tbl_from_keys_info <- function(tbl, keys_info) {
  new_keyed_tbl(
    tbl,
    pk = keys_info$pk,
    fks_in = keys_info$fks_in,
    fks_out = keys_info$fks_out,
    uuid = keys_info$uuid,
    zoom2 = keys_info$zoom2
  )
}

keyed_get_info <- function(x) {
  stopifnot(is_dm_keyed_tbl(x))
  attr(x, "dm_key_info")
}

is_dm_keyed_tbl <- function(x) {
  inherits(x, "dm_keyed_tbl")
}

#' @export
tbl_sum.dm_keyed_tbl <- function(x, ...) {
  info <- keyed_get_info(x)

  if (is.null(info$pk)) {
    pk_info <- cli::symbol$em_dash
  } else {
    pk_info <- commas(tick(info$pk))
  }

  fks_in_info <- nrow(info$fks_in)
  fks_out_info <- nrow(info$fks_out)

  c(
    NextMethod(),
    Keys = paste0(pk_info, " | ", fks_in_info, " | ", fks_out_info)
  )
}

#' @title Remove `"dm_keyed_tbl"` class
#'
#' @return If entered table has `"dm_keyed_tbl"` class, it will be removed. All
#' other classes will be preserved.
#'
#' @examples
#' dm <- dm_nycflights13()
#' class(dm$airlines)
#' class(dm:::unclass_keyed_tbl(dm$airlines))
#' @keywords internal
#' @noRd
unclass_keyed_tbl <- function(tbl) {
  if (inherits(tbl, "dm_keyed_tbl")) {
    new_classes <- class(tbl)[class(tbl) != "dm_keyed_tbl"]
    class(tbl) <- new_classes
  }

  attr(tbl, "dm_key_info") <- NULL

  tbl
}

pks_df_from_keys_info <- function(tables) {
  pks <- map(unname(tables), new_pks_from_keys_info)
  tibble(table = names2(tables), pks)
}

new_pks_from_keys_info <- function(tbl) {
  df_keys <- keyed_get_info(tbl)
  if (is.null(df_keys$pk)) {
    NULL
  } else if (length(setdiff(df_keys$pk, colnames(tbl))) > 0) {
    NULL
  } else {
    new_pk(list(df_keys$pk))
  }
}

uks_df_from_keys_info <- function(tables) {
  uks <- map(unname(tables), new_uks_from_keys_info)
  tibble(table = names2(tables), uks)
}

new_uks_from_keys_info <- function(tbl) {
  df_keys <- keyed_get_info(tbl)

  if (nrow(df_keys$uks) == 0) {
    return(new_uk())
  }

  missing_uks <- map(df_keys$uks$column, setdiff, colnames(tbl))
  valid_uks <- (lengths(missing_uks) == 0)

  new_uk(df_keys$uks$column[valid_uks])
}

#' @autoglobal
fks_df_from_keys_info <- function(tables) {
  info <- map(tables, keyed_get_info)

  fks_in <-
    map_dfr(info, ~ tibble(.x$fks_in, parent_uuid = .x$uuid))

  fks_out <-
    map_dfr(info, ~ tibble(child_uuid = .x$uuid, .x$fks_out))

  fks <-
    vec_rbind(
      fks_out,
      fks_in,
      .ptype = new_fks(child_uuid = character(), parent_uuid = character())
    ) %>%
    distinct()

  uuid_lookup <- tibble(
    table = names2(tables),
    data = map(unname(tables), unclass_keyed_tbl),
    uuid = map_chr(info, "uuid")
  )

  child_uuid_lookup <- set_names(uuid_lookup, paste0("child_", names(uuid_lookup)))
  parent_uuid_lookup <- set_names(uuid_lookup, paste0("parent_", names(uuid_lookup)))

  fks %>%
    # Multiple tables with the same uuid can occur due to aliasing,
    # hence `multiple = "all"` and `relationship = "many-to-many"`
    left_join(
      child_uuid_lookup,
      by = "child_uuid",
      multiple = "all",
      relationship = "many-to-many"
    ) %>%
    left_join(
      parent_uuid_lookup,
      by = "parent_uuid",
      multiple = "all",
      relationship = "many-to-many"
    ) %>%
    select(-child_uuid, -parent_uuid) %>%
    filter(map2_lgl(child_fk_cols, child_data, ~ all(.x %in% colnames(.y)))) %>%
    filter(map2_lgl(parent_key_cols, parent_data, ~ all(.x %in% colnames(.y)))) %>%
    group_by(parent_table) %>%
    # FIXME: Capture on_delete
    summarize(
      fks = list(new_fk(as.list(parent_key_cols), child_table, as.list(child_fk_cols), "no_action"))
    ) %>%
    ungroup() %>%
    rename(table = parent_table)
}

new_fks_from_keys_info <- function(tbl) {
  if (is_dm_keyed_tbl(tbl)) {
    df_keys <- keyed_get_info(tbl)$fks_in
    df_fks <- tibble(
      ref_column = map(df_keys$parent_key_cols, as.character),
      table = df_keys$child_table,
      column = map(df_keys$child_fk_cols, as.character),
      on_delete = "no_action"
    )

    new_fk(df_fks$ref_column, df_fks$table, df_fks$column, df_fks$on_delete)
  }
}

# FIXME: Pass keep when ready
keyed_build_join_spec <- function(x, y, by = NULL, suffix = NULL) {
  info_x <- keyed_get_info(x)
  info_y <- keyed_get_info(y)

  if (is.null(by)) {
    by <- keyed_by(x, y)
  } else {
    by <- normalize_join_by(by)
  }

  if (is.null(suffix)) {
    suffix <- c(".x", ".y")
  }

  rename <- join_rename_rules(colnames(x), colnames(y), by, suffix)

  # Is one of the "by" column sets a primary key? Keep the primary key of the *other* table!
  if (keyed_is_pk(x, names(by))) {
    new_pk <- join_rename(info_y$pk, rename$y)
  } else if (keyed_is_pk(y, by)) {
    new_pk <- join_rename(info_x$pk, rename$x)
  } else {
    new_pk <- NULL
  }

  # Keep all foreign keys
  new_fks_in <-
    vec_rbind(
      join_rename_key(info_x$fks_in, "parent_key_cols", rename$x),
      join_rename_key(info_y$fks_in, "parent_key_cols", rename$y),
    )

  new_fks_out <-
    vec_rbind(
      join_rename_key(info_x$fks_out, "child_fk_cols", rename$x),
      join_rename_key(info_y$fks_out, "child_fk_cols", rename$y),
    )

  # need to remove the `"dm_keyed_tbl"` class to avoid infinite recursion
  # while joining
  x_tbl <- unclass_keyed_tbl(x)
  y_tbl <- unclass_keyed_tbl(y)

  list(
    x_tbl = x_tbl,
    y_tbl = y_tbl,
    by = enframe(by, "x", "y"),
    suffix = suffix,
    new_pk = new_pk,
    new_fks_in = new_fks_in,
    new_fks_out = new_fks_out,
    new_uuid = new_uuid()
  )
}

keyed_by <- function(x, y) {
  fks_df <- fks_df_from_keys_info(list(x = x, y = y))

  if (nrow(fks_df) == 0) {
    cli::cli_abort("Can't infer {.arg by}: foreign key information lost?", call = dm_error_call())
  }

  stopifnot(map_int(fks_df$fks, NROW) > 0)

  if (nrow(fks_df) > 1) {
    cli::cli_abort(
      "Can't infer {.arg by}: foreign key available in both directions.",
      call = dm_error_call()
    )
  }

  if (nrow(fks_df$fks[[1]]) > 1) {
    cli::cli_abort(
      "Can't infer {.arg by}: multiple foreign keys available.",
      call = dm_error_call()
    )
  }

  fk <- fks_df$fks[[1]][1, ]

  if (fks_df$table == "x") {
    set_names(fk$column[[1]], fk$ref_column[[1]])
  } else {
    set_names(fk$ref_column[[1]], fk$column[[1]])
  }
}

keyed_is_pk <- function(x, cols) {
  info <- keyed_get_info(x)

  # cols must include at least all pk columns
  !is.null(info$pk) && all(info$pk %in% cols)
}

join_rename_rules <- function(x, y, by, suffix) {
  by_idx <- match(by, y)
  stopifnot(!anyNA(by_idx))
  stopifnot(length(by_idx) > 0)
  conflicts <- intersect(x, y[-by_idx])

  x_new <- x
  x_new[x_new %in% conflicts()] <- paste0(x_new, suffix[[1]])

  y_new <- y
  y_new[by_idx] <- names(by)
  y_new[y_new %in% conflicts()] <- paste0(y_new, suffix[[2]])

  list(
    x = set_names(x_new, x),
    y = set_names(y_new, y)
  )
}

join_rename <- function(x, rename) {
  if (is.null(x)) {
    return(NULL)
  }
  stopifnot(all(x %in% names(rename)))
  unname(rename[x])
}

join_rename_key <- function(x, colname, rename) {
  stopifnot(colname %in% names(x))
  x[[colname]] <- new_keys(map(x[[colname]], join_rename, rename))
  x
}

# TODO: can probably be deleted once this feature branch is ready for a merge

# helpers ----------------------------------

test_that("`new_fks_in()` generates expected tibble", {
  expect_snapshot({
    new_fks_in(
      child_uuid = "flights-uuid",
      child_fk_cols = new_keys(list(list("origin", "dest"))),
      parent_key_cols = new_keys(list(list("faa")))
    )
  })
})

test_that("`new_fks_out()` generates expected tibble", {
  expect_snapshot({
    new_fks_out(
      child_fk_cols = new_keys(list(list("origin", "dest"))),
      parent_uuid = "airports-uuid",
      parent_key_cols = new_keys(list(list("faa")))
    )
  })
})

test_that("`new_keyed_tbl()` generates expected output", {
  expect_snapshot({
    dm <- dm_nycflights13(cycle = TRUE)

    # should look similar to `dm_get_all_fks_impl(dm, "airports")`
    keyed_tbl <- new_keyed_tbl(
      x = dm$airports,
      pk = "faa",
      fks_in = new_fks_in(
        child_uuid = "flights-uuid",
        child_fk_cols = new_keys(list("origin", "dest")),
        parent_key_cols = new_keys(list("faa"))
      ),
      fks_out = new_fks_out(
        child_fk_cols = new_keys(list("origin", "dest")),
        parent_uuid = "airports-uuid",
        parent_key_cols = new_keys(list("faa"))
      ),
      uuid = "0a0c060f-0d01-0b03-0402-05090800070e"
    )

    keyed_get_info(keyed_tbl)
  })

  expect_equal(dm$airports, keyed_tbl, ignore_attr = TRUE)
})

test_that("dm_get_keyed_tables_impl()", {
  withr::local_seed(20220715)

  expect_snapshot({
    dm_nycflights13(cycle = TRUE) %>%
      dm_get_keyed_tables_impl() %>%
      map(keyed_get_info)
  })
})


test_that("`new_keyed_tbl()` formatting", {
  local_options(
    pillar.min_title_chars = NULL,
    pillar.max_title_chars = NULL,
    pillar.max_footer_lines = NULL,
    pillar.bold = NULL,
  )

  expect_snapshot({
    keyed_tbl_impl(dm_nycflights13(cycle = TRUE), "flights")
    keyed_tbl_impl(dm_nycflights13(cycle = TRUE), "airports")
    keyed_tbl_impl(dm_nycflights13(cycle = TRUE), "airlines")
  })
})

# subsetting ----------------------------------

test_that("both subsetting operators for `dm` produce the same object", {
  dm <- dm_nycflights13(cycle = TRUE)

  expect_equal(dm$airlines, dm[["airlines"]])
  expect_equal(dm[[1]], dm[["airlines"]])
})

test_that("subsetting `dm` produces `dm_keyed_tbl` objects", {
  dm <- dm_nycflights13(cycle = TRUE)

  skip("keyed = TRUE")

  expect_s3_class(dm$airlines, "dm_keyed_tbl")
  expect_s3_class(dm[[1]], "dm_keyed_tbl")
  expect_s3_class(dm[["airlines"]], "dm_keyed_tbl")
})

# constructors ----------------------------------

test_that("`dm()` and `new_dm()` can handle a list of `dm_keyed_tbl` objects", {
  dm <- dm_nycflights13(cycle = TRUE)

  y1 <- keyed_tbl_impl(dm, "weather") %>%
    mutate() %>%
    select(everything())
  y2 <- keyed_tbl_impl(dm, "airports") %>%
    mutate() %>%
    select(everything())

  expect_s3_class(y1, "dm_keyed_tbl")
  expect_s3_class(y2, "dm_keyed_tbl")

  dm_output <- dm(d1 = y1, d2 = y2)
  expect_s3_class(dm_output, "dm")

  new_dm_output <- new_dm(list(d1 = y1, d2 = y2))
  expect_s3_class(new_dm_output, "dm")

  # there shouldn't be any keys
  expect_snapshot(tbl_sum(keyed_tbl_impl(dm_output, "d1")))
  expect_snapshot(tbl_sum(keyed_tbl_impl(dm_output, "d2")))
  expect_snapshot(tbl_sum(keyed_tbl_impl(new_dm_output, "d1")))
  expect_snapshot(tbl_sum(keyed_tbl_impl(new_dm_output, "d2")))
})

test_that("`dm()` and `new_dm()` can handle a mix of tables and `dm_keyed_tbl` objects", {
  dm <- dm_nycflights13(cycle = TRUE)

  y1 <- keyed_tbl_impl(dm, "weather") %>%
    mutate() %>%
    select(everything())
  y2 <- dm$airports

  expect_s3_class(y1, "dm_keyed_tbl")
  expect_s3_class(y2, "tbl_df")

  dm_output <- dm(d1 = y1, d2 = y2)
  expect_s3_class(dm_output, "dm")

  new_dm_output <- new_dm(list(d1 = y1, d2 = y2))
  expect_s3_class(new_dm_output, "dm")

  # there shouldn't be any keys
  expect_snapshot(tbl_sum(keyed_tbl_impl(dm_output, "d1")))
  expect_snapshot(tbl_sum(keyed_tbl_impl(dm_output, "d2")))
  expect_snapshot(tbl_sum(keyed_tbl_impl(new_dm_output, "d1")))
  expect_snapshot(tbl_sum(keyed_tbl_impl(new_dm_output, "d2")))
})

test_that("`dm()` handles missing key column names gracefully", {
  dm <-
    dm(x = tibble(a = 1, b = 1), y = tibble(a = 1, b = 1)) %>%
    dm_add_pk(y, c(a, b)) %>%
    dm_add_fk(x, c(a, b), y)

  keyed <-
    dm %>%
    dm_get_tables(keyed = TRUE)

  expect_snapshot({
    dm(x = keyed$x["b"], y = keyed$y) %>%
      dm_paste()
    dm(x = keyed$x, y = keyed$y["b"]) %>%
      dm_paste()
  })
})

# joins ----------------------------------

test_that("keyed_by()", {
  withr::local_seed(20220715)

  dm <-
    dm(x = tibble(a = 1), y = tibble(b = 1)) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")

  expect_snapshot({
    keyed_by(x, y)
    keyed_by(y, x)
  })
})

test_that("keyed_by() failure modes", {
  withr::local_seed(20220715)

  # No foreign key information
  dm_no_fk <-
    dm(x = tibble(a = 1), y = tibble(b = 1))

  x_no_fk <- keyed_tbl_impl(dm_no_fk, "x")
  y_no_fk <- keyed_tbl_impl(dm_no_fk, "y")

  expect_snapshot(error = TRUE, {
    keyed_by(x_no_fk, y_no_fk)
  })

  # Foreign keys in both directions
  dm_both_fk <-
    dm(x = tibble(a = 1), y = tibble(b = 1)) %>%
    dm_add_pk(x, a) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y) %>%
    dm_add_fk(y, b, x)

  x_both_fk <- keyed_tbl_impl(dm_both_fk, "x")
  y_both_fk <- keyed_tbl_impl(dm_both_fk, "y")

  expect_snapshot(error = TRUE, {
    keyed_by(x_both_fk, y_both_fk)
  })

  # Multiple foreign keys in same direction
  dm_multi_fk <-
    dm(x = tibble(a = 1, a2 = 1), y = tibble(b = 1)) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y) %>%
    dm_add_fk(x, a2, y)

  x_multi_fk <- keyed_tbl_impl(dm_multi_fk, "x")
  y_multi_fk <- keyed_tbl_impl(dm_multi_fk, "y")

  expect_snapshot(error = TRUE, {
    keyed_by(x_multi_fk, y_multi_fk)
  })
})

test_that("joins without child PK", {
  withr::local_seed(20220715)

  dm <-
    dm(x = tibble(a = 1), y = tibble(b = 1)) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")

  expect_snapshot({
    keyed_build_join_spec(x, y) %>%
      to_snapshot_json()
    dm(x, y, r = left_join(x, y)) %>%
      dm_paste(options = c("select", "keys"))
    keyed_build_join_spec(y, x) %>%
      to_snapshot_json()
    dm(x, y, r = left_join(y, x)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("joins with other child PK", {
  withr::local_seed(20220715)

  dm <-
    dm(x = tibble(a = 1, c = 1), y = tibble(b = 1)) %>%
    dm_add_pk(x, c) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")

  expect_snapshot({
    keyed_build_join_spec(x, y) %>%
      to_snapshot_json()
    dm(x, y, r = left_join(x, y)) %>%
      dm_paste(options = c("select", "keys"))
    keyed_build_join_spec(y, x) %>%
      to_snapshot_json()
    dm(x, y, r = left_join(y, x)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("joins with other child PK and name conflict", {
  withr::local_seed(20220715)

  dm <-
    dm(x = tibble(a = 1, b = 1), y = tibble(b = 1)) %>%
    dm_add_pk(x, b) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")

  expect_snapshot({
    keyed_build_join_spec(x, y) %>%
      to_snapshot_json()
    dm(x, y, r = left_join(x, y)) %>%
      dm_paste(options = c("select", "keys"))
    keyed_build_join_spec(y, x) %>%
      to_snapshot_json()
    dm(x, y, r = left_join(y, x)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("joins with same child PK", {
  withr::local_seed(20220715)

  dm <-
    dm(x = tibble(a = 1), y = tibble(b = 1)) %>%
    dm_add_pk(x, a) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")

  expect_snapshot({
    keyed_build_join_spec(x, y) %>%
      to_snapshot_json()
    dm(x, y, r = left_join(x, y)) %>%
      dm_paste(options = c("select", "keys"))
    keyed_build_join_spec(y, x) %>%
      to_snapshot_json()
    dm(x, y, r = left_join(y, x)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("joins with same child PK and same name", {
  withr::local_seed(20220715)

  dm <-
    dm(x = tibble(b = 1), y = tibble(b = 1)) %>%
    dm_add_pk(x, b) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, b, y)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")

  expect_snapshot({
    keyed_build_join_spec(x, y) %>%
      to_snapshot_json()
    dm(x, y, r = left_join(x, y)) %>%
      dm_paste(options = c("select", "keys"))
    keyed_build_join_spec(y, x) %>%
      to_snapshot_json()
    dm(x, y, r = left_join(y, x)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("joins with other FK from parent", {
  withr::local_seed(20220715)

  dm <-
    dm(x = tibble(a = 1), y = tibble(b = 1, c = 1), z = tibble(c = 1)) %>%
    dm_add_pk(x, a) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y) %>%
    dm_add_fk(y, c, z, c)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")
  z <- keyed_tbl_impl(dm, "z")

  expect_snapshot({
    keyed_build_join_spec(x, y) %>%
      to_snapshot_json()
    dm(x, y, z, r = left_join(x, y)) %>%
      dm_paste(options = c("select", "keys"))
    keyed_build_join_spec(y, x) %>%
      to_snapshot_json()
    dm(x, y, z, r = left_join(y, x)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("joins with other FK from parent and name conflict", {
  withr::local_seed(20220715)

  dm <-
    dm(x = tibble(a = 1), y = tibble(b = 1, a = 1), z = tibble(a = 1)) %>%
    dm_add_pk(x, a) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y) %>%
    dm_add_fk(y, a, z, a)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")
  z <- keyed_tbl_impl(dm, "z")

  expect_snapshot({
    keyed_build_join_spec(x, y) %>%
      to_snapshot_json()
    dm(x, y, z, r = left_join(x, y)) %>%
      dm_paste(options = c("select", "keys"))
    keyed_build_join_spec(y, x) %>%
      to_snapshot_json()
    dm(x, y, z, r = left_join(y, x)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("joins with other FK from child", {
  withr::local_seed(20220715)

  dm <-
    dm(x = tibble(a = 1, c = 1), y = tibble(b = 1), z = tibble(c = 1)) %>%
    dm_add_pk(x, a) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y) %>%
    dm_add_fk(x, c, z, c)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")
  z <- keyed_tbl_impl(dm, "z")

  expect_snapshot({
    keyed_build_join_spec(x, y) %>%
      to_snapshot_json()
    dm(x, y, z, r = left_join(x, y)) %>%
      dm_paste(options = c("select", "keys"))
    keyed_build_join_spec(y, x) %>%
      to_snapshot_json()
    dm(x, y, z, r = left_join(y, x)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("joins with other FK from child and name conflict", {
  withr::local_seed(20220715)

  dm <-
    dm(x = tibble(a = 1, b = 1), y = tibble(b = 1), z = tibble(b = 1)) %>%
    dm_add_pk(x, a) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y) %>%
    dm_add_fk(x, b, z, b)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")
  z <- keyed_tbl_impl(dm, "z")

  expect_snapshot({
    keyed_build_join_spec(x, y) %>%
      to_snapshot_json()
    dm(x, y, z, r = left_join(x, y)) %>%
      dm_paste(options = c("select", "keys"))
    keyed_build_join_spec(y, x) %>%
      to_snapshot_json()
    dm(x, y, z, r = left_join(y, x)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("left join works as expected with keyed tables", {
  withr::local_seed(20220717)

  local_options(
    pillar.min_title_chars = NULL,
    pillar.max_title_chars = NULL,
    pillar.max_footer_lines = NULL,
    pillar.bold = NULL,
  )

  expect_snapshot({
    dm <- dm_nycflights13()
    keyed_tbl_impl(dm, "weather") %>% left_join(keyed_tbl_impl(dm, "flights"), multiple = "all")
  })

  # results should be similar to zooming
  zd1 <-
    dm %>%
    dm_rename(weather, year.weather = year) %>%
    dm_rename(weather, month.weather = month) %>%
    dm_rename(weather, day.weather = day) %>%
    dm_rename(weather, hour.weather = hour) %>%
    dm_rename(flights, year.flights = year) %>%
    dm_rename(flights, month.flights = month) %>%
    dm_rename(flights, day.flights = day) %>%
    dm_rename(flights, hour.flights = hour) %>%
    dm_zoom_to(weather) %>%
    left_join(flights, multiple = "all")

  zd2 <-
    dm %>%
    dm_rename(flights, year.flights = year) %>%
    dm_rename(flights, month.flights = month) %>%
    dm_rename(flights, day.flights = day) %>%
    dm_rename(flights, hour.flights = hour) %>%
    dm_rename(weather, year.weather = year) %>%
    dm_rename(weather, month.weather = month) %>%
    dm_rename(weather, day.weather = day) %>%
    dm_rename(weather, hour.weather = hour) %>%
    dm_zoom_to(flights) %>%
    left_join(weather)

  jd1 <- keyed_tbl_impl(dm, "weather") %>%
    left_join(keyed_tbl_impl(dm, "flights"), multiple = "all")
  jd2 <- keyed_tbl_impl(dm, "flights") %>% left_join(keyed_tbl_impl(dm, "weather"))

  expect_equal(ncol(jd1), ncol(jd2))
  expect_equal(dim(zd2), dim(jd2))
})

# semi_join ----------------------------------

test_that("semi_join()", {
  withr::local_seed(20220720)

  local_options(
    pillar.min_title_chars = NULL,
    pillar.max_title_chars = NULL,
    pillar.max_footer_lines = NULL,
    pillar.bold = NULL,
  )

  dm <-
    dm(x = tibble(a = 1), y = tibble(b = 1)) %>%
    dm_add_pk(y, b) %>%
    dm_add_fk(x, a, y)

  x <- keyed_tbl_impl(dm, "x")
  y <- keyed_tbl_impl(dm, "y")

  expect_snapshot({
    dm(x, y, r = semi_join(x, y)) %>%
      dm_paste(options = c("select", "keys"))
    dm(x, y, r = semi_join(y, x)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

# arrange ----------------------------------

test_that("arrange for keyed tables produces expected output", {
  dm <- dm_nycflights13(cycle = TRUE)

  expect_snapshot({
    keyed_tbl_impl(dm, "airlines") %>% arrange(desc(name))
  })
})

# group_by ----------------------------------

test_that("group_by for keyed tables produces expected output", {
  local_options(
    pillar.min_title_chars = NULL,
    pillar.max_title_chars = NULL,
    pillar.max_footer_lines = NULL,
    pillar.bold = NULL,
  )

  expect_snapshot({
    dm <- dm_nycflights13(cycle = TRUE)

    keyed_tbl_impl(dm, "flights") %>% group_by(month)

    keyed_tbl_impl(dm, "airports") %>% group_by(tzone)

    # grouping by the primary key works as well
    keyed_tbl_impl(dm, "airports") %>% group_by(faa)
  })
})

# summarize ----------------------------------

test_that("summarize for keyed tables produces expected output", {
  # FIXME: Brittle tests?
  local_options(dplyr.summarise.inform = FALSE)

  expect_snapshot({
    dm <- dm_nycflights13(cycle = TRUE)

    keyed_tbl_impl(dm, "airports") %>%
      summarise(mean_alt = mean(alt))

    keyed_tbl_impl(dm, "airports") %>%
      group_by(tzone, dst) %>%
      summarise(mean_alt = mean(alt))
  })
})


test_that("summarize for keyed tables produces same output as zooming", {
  local_options(
    pillar.min_title_chars = NULL,
    pillar.max_title_chars = NULL,
    pillar.max_footer_lines = NULL,
    pillar.bold = NULL,
  )

  dm <- dm_nycflights13(cycle = TRUE)

  z_summary <- dm %>%
    dm_zoom_to(flights) %>%
    group_by(month) %>%
    arrange(desc(day)) %>%
    summarize(avg_air_time = mean(air_time, na.rm = TRUE))

  k_summary <- keyed_tbl_impl(dm, "flights") %>%
    group_by(month) %>%
    arrange(desc(day)) %>%
    summarize(avg_air_time = mean(air_time, na.rm = TRUE))

  # zoomed and keyed approaches should provide same summaries
  expect_equal(dim(z_summary), dim(k_summary))
  expect_equal(z_summary$month, k_summary$month)
  expect_equal(z_summary$avg_air_time, k_summary$avg_air_time)
})

# reconstruction ----------------------------------

test_that("pks_df_from_keys_info()", {
  withr::local_seed(20220715)

  dm <- dm_nycflights13(cycle = TRUE)

  expect_snapshot({
    dm %>%
      dm_get_keyed_tables_impl() %>%
      pks_df_from_keys_info() %>%
      to_snapshot_json()
  })
})

test_that("uks_df_from_keys_info()", {
  withr::local_seed(20220715)

  dm <- dm_for_filter() %>%
    dm_add_uk(tf_5, l)

  expect_snapshot({
    dm %>%
      dm_get_keyed_tables_impl() %>%
      uks_df_from_keys_info() %>%
      to_snapshot_json()
  })
})

test_that("fks_df_from_keys_info()", {
  withr::local_seed(20220715)

  dm <- dm_nycflights13(cycle = TRUE)

  expect_snapshot({
    dm %>%
      dm_get_keyed_tables_impl() %>%
      fks_df_from_keys_info() %>%
      to_snapshot_json()
  })
})

test_that("primary and foreign keys survive the round trip", {
  dm <- dm_nycflights13(cycle = TRUE)
  tbl <- keyed_tbl_impl(dm, "weather")
  tbl_mutate <- tbl %>% select(everything())

  dm2 <- dm(
    weather = tbl_mutate,
    airlines = keyed_tbl_impl(dm, "airlines"),
    airports = keyed_tbl_impl(dm, "airports"),
    planes = keyed_tbl_impl(dm, "planes"),
    flights = keyed_tbl_impl(dm, "flights"),
  )

  original_def <- dm_get_def(dm) %>% arrange(table)
  new_def <- dm_get_def(dm2) %>% arrange(table)

  expect_equal(original_def$pks, new_def$pks)
  expect_equal(original_def$fks, new_def$fks)
})

#' Validate your data model
#'
#' @description
#' This function returns a tibble with information about
#' which key constraints are met (`is_key = TRUE`) or violated (`FALSE`).
#' The printing for this object is special, use [as_tibble()]
#' to print as a regular tibble.
#'
#' @param .dm A `dm` object.
#' @inheritParams rlang::args_dots_empty
#' @param .progress Whether to display a progress bar, if `NA` (the default)
#'   hide in non-interactive mode, show in interactive mode. Requires the
#'   'progress' package.
#' @param .max_value Maximum number of distinct problematic values to report
#'   in the `problem` column, defaults to `6`.
#'   Set to `Inf` to report all values.
#' @param dm,progress `r lifecycle::badge("deprecated")`
#'
#' @return A tibble with the following columns:
#'   \describe{
#'     \item{`table`}{the table in the `dm`,}
#'     \item{`kind`}{"PK" or "FK",}
#'     \item{`columns`}{the table columns that define the key,}
#'     \item{`ref_table`}{for foreign keys, the referenced table,}
#'     \item{`is_key`}{logical,}
#'     \item{`problem`}{if `is_key = FALSE`, the reason for that.}
#'   }
#'
#' @details For the primary key constraints, it is tested if the values in the respective columns are all unique.
#' For the foreign key constraints, the tests check if for each foreign key constraint, the values of the foreign key column
#' form a subset of the values of the referenced column.
#'
#' @export
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_examine_constraints()
#' @autoglobal
dm_examine_constraints <- function(
  .dm,
  ...,
  .progress = NA,
  .max_value = 6L,
  dm = deprecated(),
  progress = deprecated()
) {
  check_dots_empty()
  dm_local_error_call()

  if (!is_missing(dm)) {
    deprecate_warn("1.0.0", "dm_examine_constraints(dm = )", "dm_examine_constraints(.dm = )")
  }

  if (is_missing(.dm)) {
    return(dm_examine_constraints(
      dm,
      .progress = .progress,
      .max_value = .max_value,
      progress = progress
    ))
  }

  if (!is_missing(progress)) {
    if (is.na(progress)) {
      progress <- .progress
    }
    deprecate_warn(
      "1.0.0",
      "dm_examine_constraints(progress = )",
      "dm_examine_constraints(.progress = )"
    )
  }

  check_not_zoomed(.dm)
  .dm %>%
    dm_examine_constraints_impl(
      progress = .progress,
      top_level_fun = "dm_examine_constraints",
      max_value = .max_value
    ) %>%
    rename(columns = column) %>%
    mutate(columns = new_keys(columns)) %>%
    new_dm_examine_constraints()
}

dm_examine_constraints_impl <- function(
  dm,
  progress = NA,
  top_level_fun = NULL,
  max_value = MAX_COMMAS
) {
  pk_results <- check_pk_constraints(
    dm,
    progress,
    top_level_fun = top_level_fun,
    max_value = max_value
  )
  fk_results <- check_fk_constraints(
    dm,
    progress,
    top_level_fun = top_level_fun,
    max_value = max_value
  )
  bind_rows(
    pk_results,
    fk_results
  ) %>%
    arrange(is_key, desc(kind), table)
}

new_dm_examine_constraints <- function(x) {
  class(x) <- c("dm_examine_constraints", class(x))
  x
}

#' @export
#' @autoglobal
print.dm_examine_constraints <- function(x, ...) {
  key_df <-
    x %>%
    as_tibble()
  problem_df <-
    key_df %>%
    filter(problem != "")

  if (nrow(key_df) == 0) {
    cli::cli_alert_info("No constraints defined.")
  } else if (nrow(problem_df) == 0) {
    cli::cli_alert_info("All constraints satisfied.")
  } else {
    cli::cli_alert_warning("Unsatisfied constraints:")

    problem_df %>%
      mutate(
        into = if_else(kind == "FK", paste0(" into table ", tick(ref_table)), "")
      ) %>%
      # FIXME: Use cli styles
      mutate(
        text = paste0(
          "Table ",
          tick(table),
          ": ",
          kind_to_long(kind),
          " ",
          format(map(problem_df$columns, tick), justify = "none"),
          into,
          ": ",
          problem
        )
      ) %>%
      pull(text) %>%
      cli::cat_bullet(bullet_col = "red")
  }

  invisible(x)
}

kind_to_long <- function(kind) {
  case_when(
    kind == "PK" ~ "primary key",
    kind == "UK" ~ "unique key",
    # FIXME: `case_when()` gets the `.default` arg in v1.1.0, then:
    # .default = "foreign key"
    TRUE ~ "foreign key"
  )
}

#' @autoglobal
check_pk_constraints <- function(dm, progress = NA, top_level_fun = NULL, max_value = MAX_COMMAS) {
  pks <- bind_rows(
    list(
      PK = dm_get_all_pks_impl(dm),
      UK = dm_get_all_uks_impl(dm) %>% rename(pk_col = uk_col) %>% select(-kind)
    ),
    .id = "kind"
  ) %>%
    distinct(table, pk_col, .keep_all = TRUE)
  if (nrow(pks) == 0) {
    return(tibble(
      table = character(),
      kind = character(),
      column = new_keys(),
      ref_table = character(),
      is_key = logical(),
      problem = character()
    ))
  }
  table_names <- pks$table
  columns <- pks$pk_col

  ticker <- new_ticker(
    "checking pk constraints",
    n = length(table_names),
    progress = progress,
    top_level_fun = top_level_fun
  )

  candidates <- map2(
    set_names(table_names),
    columns,
    ticker(
      ~ {
        tbl <- tbl_impl(dm, .x)
        enum_pk_candidates_impl(tbl, list(.y), max_value = max_value)
      }
    )
  )

  tbl_is_pk <-
    tibble(table = table_names, candidate = candidates) %>%
    unnest_df(
      "candidate",
      tibble(column = new_keys(), candidate = logical(), why = character())
    ) %>%
    rename(is_key = candidate, problem = why)

  tibble(
    table = table_names,
    kind = pks$kind,
    column = pks$pk_col,
    ref_table = NA_character_
  ) %>%
    left_join(tbl_is_pk, by = c("table", "column"))
}

#' @autoglobal
check_fk_constraints <- function(dm, progress = NA, top_level_fun = NULL, max_value = MAX_COMMAS) {
  fks <- dm_get_all_fks_impl(dm)
  pts <- map(fks$parent_table, tbl_impl, dm = dm)
  cts <- map(fks$child_table, tbl_impl, dm = dm)
  fks_tibble <-
    mutate(fks, t1 = cts, t2 = pts) %>%
    select(
      t1,
      t1_name = child_table,
      colname = child_fk_cols,
      t2,
      t2_name = parent_table,
      pk = parent_key_cols
    )

  ticker <- new_ticker(
    "checking fk constraints",
    n = nrow(fks_tibble),
    progress = progress,
    top_level_fun = top_level_fun
  )

  fks_tibble %>%
    mutate(
      problem = pmap_chr(fks_tibble, ticker(check_fk), max_value = max_value),
      is_key = (problem == ""),
      kind = "FK"
    ) %>%
    select(table = t1_name, kind, column = colname, ref_table = t2_name, is_key, problem)
}

# FIXME: Adapt argument list as provided/required by the Shiny app
mw_cg_make_dm_set_colors <- function(dm, ..., table_names = NULL, color_name = NULL) {
  check_dots_empty()
  # Checks
  stopifnot(is.character(table_names))
  stopifnot(table_names %in% names(dm))

  if (length(table_names) == 1) {
    table_sym <- sym(table_names)
  } else {
    table_sym <- expr(c(!!!syms(table_names)))
  }

  table_syms <- exprs(!!sym(color_name) := !!table_sym)

  list(
    call = expr(dm_set_colors(
      .,
      !!!table_syms
    ))
  )
}

#' Modifying rows for multiple tables
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' These functions provide a framework for updating data in existing tables.
#' Unlike [compute()], [copy_to()] or [copy_dm_to()], no new tables are created
#' on the database.
#' All operations expect that both existing and new data are presented
#' in two compatible [dm] objects on the same data source.
#'
#' The functions make sure that the tables in the target dm
#' are processed in topological order so that parent (dimension)
#' tables receive insertions before child (fact) tables.
#'
#' These operations, in contrast to all other operations,
#' may lead to irreversible changes to the underlying database.
#' Therefore, in-place operation must be requested explicitly with `in_place = TRUE`.
#' By default, an informative message is given.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams dplyr::rows_insert
#' @inheritParams dm_examine_constraints
#' @param x Target `dm` object.
#' @param y `dm` object with new data.
#'
#' @return A dm object of the same [dm_ptype()] as `x`.
#'   If `in_place = TRUE`, the underlying data is updated as a side effect,
#'   and `x` is returned, invisibly.
#'
#' @name rows-dm
#' @examplesIf rlang::is_installed(c("RSQLite", "nycflights13", "dbplyr"))
#' # Establish database connection:
#' sqlite <- DBI::dbConnect(RSQLite::SQLite())
#'
#' # Entire dataset with all dimension tables populated
#' # with flights and weather data truncated:
#' flights_init <-
#'   dm_nycflights13() %>%
#'   dm_zoom_to(flights) %>%
#'   filter(FALSE) %>%
#'   dm_update_zoomed() %>%
#'   dm_zoom_to(weather) %>%
#'   filter(FALSE) %>%
#'   dm_update_zoomed()
#'
#' # Target database:
#' flights_sqlite <- copy_dm_to(sqlite, flights_init, temporary = FALSE)
#' print(dm_nrow(flights_sqlite))
#'
#' # First update:
#' flights_jan <-
#'   dm_nycflights13() %>%
#'   dm_select_tbl(flights, weather) %>%
#'   dm_zoom_to(flights) %>%
#'   filter(month == 1) %>%
#'   dm_update_zoomed() %>%
#'   dm_zoom_to(weather) %>%
#'   filter(month == 1) %>%
#'   dm_update_zoomed()
#' print(dm_nrow(flights_jan))
#'
#' # Copy to temporary tables on the target database:
#' flights_jan_sqlite <- copy_dm_to(sqlite, flights_jan)
#'
#' # Dry run by default:
#' dm_rows_append(flights_sqlite, flights_jan_sqlite)
#' print(dm_nrow(flights_sqlite))
#'
#' # Explicitly request persistence:
#' dm_rows_append(flights_sqlite, flights_jan_sqlite, in_place = TRUE)
#' print(dm_nrow(flights_sqlite))
#'
#' # Second update:
#' flights_feb <-
#'   dm_nycflights13() %>%
#'   dm_select_tbl(flights, weather) %>%
#'   dm_zoom_to(flights) %>%
#'   filter(month == 2) %>%
#'   dm_update_zoomed() %>%
#'   dm_zoom_to(weather) %>%
#'   filter(month == 2) %>%
#'   dm_update_zoomed()
#'
#' # Copy to temporary tables on the target database:
#' flights_feb_sqlite <- copy_dm_to(sqlite, flights_feb)
#'
#' # Explicit dry run:
#' flights_new <- dm_rows_append(
#'   flights_sqlite,
#'   flights_feb_sqlite,
#'   in_place = FALSE
#' )
#' print(dm_nrow(flights_new))
#' print(dm_nrow(flights_sqlite))
#'
#' # Check for consistency before applying:
#' flights_new %>%
#'   dm_examine_constraints()
#'
#' # Apply:
#' dm_rows_append(flights_sqlite, flights_feb_sqlite, in_place = TRUE)
#' print(dm_nrow(flights_sqlite))
#'
#' DBI::dbDisconnect(sqlite)
NULL


#' dm_rows_insert
#'
#' `dm_rows_insert()` adds new records via [rows_insert()] with `conflict = "ignore"`.
#' Duplicate records will be silently discarded.
#' This operation requires primary keys on all tables, use `dm_rows_append()`
#' to insert unconditionally.
#' @rdname rows-dm
#' @aliases dm_rows_...
#' @export
dm_rows_insert <- function(x, y, ..., in_place = NULL, progress = NA) {
  dm_local_error_call()
  check_dots_empty()

  dm_rows(x, y, "insert", top_down = TRUE, in_place, require_keys = TRUE, progress = progress)
}

#' dm_rows_append
#'
#' `dm_rows_append()` adds new records via [rows_append()].
#' The primary keys must differ from existing records.
#' This must be ensured by the caller and might be checked by the underlying database.
#' Use `in_place = FALSE` and apply [dm_examine_constraints()] to check beforehand.
#' @rdname rows-dm
#' @export
dm_rows_append <- function(x, y, ..., in_place = NULL, progress = NA) {
  dm_local_error_call()
  check_dots_empty()

  dm_rows(x, y, "append", top_down = TRUE, in_place, require_keys = FALSE, progress = progress)
}

#' dm_rows_update
#'
#' `dm_rows_update()` updates existing records via [rows_update()].
#' Primary keys must match for all records to be updated.
#'
#' @rdname rows-dm
#' @export
dm_rows_update <- function(x, y, ..., in_place = NULL, progress = NA) {
  dm_local_error_call()
  check_dots_empty()

  dm_rows(x, y, "update", top_down = TRUE, in_place, require_keys = TRUE, progress = progress)
}

#' dm_rows_patch
#'
#' `dm_rows_patch()` updates missing values in existing records
#' via [rows_patch()].
#' Primary keys must match for all records to be patched.
#'
#' @rdname rows-dm
#' @export
dm_rows_patch <- function(x, y, ..., in_place = NULL, progress = NA) {
  dm_local_error_call()
  check_dots_empty()

  dm_rows(x, y, "patch", top_down = TRUE, in_place, require_keys = TRUE, progress = progress)
}

#' dm_rows_upsert
#'
#' `dm_rows_upsert()` updates existing records and adds new records,
#' based on the primary key, via [rows_upsert()].
#'
#' @rdname rows-dm
#' @export
dm_rows_upsert <- function(x, y, ..., in_place = NULL, progress = NA) {
  dm_local_error_call()
  check_dots_empty()

  dm_rows(x, y, "upsert", top_down = TRUE, in_place, require_keys = TRUE, progress = progress)
}

#' dm_rows_delete
#'
#' `dm_rows_delete()` removes matching records via [rows_delete()],
#' based on the primary key.
#' The order in which the tables are processed is reversed.
#'
#' @rdname rows-dm
#' @export
dm_rows_delete <- function(x, y, ..., in_place = NULL, progress = NA) {
  dm_local_error_call()
  check_dots_empty()

  dm_rows(x, y, "delete", top_down = FALSE, in_place, require_keys = TRUE, progress = progress)
}

dm_rows <- function(x, y, operation_name, top_down, in_place, require_keys, progress = NA) {
  dm_rows_check(x, y)

  if (is_null(in_place)) {
    cli::cli_inform(
      "Result is returned as a dm object with lazy tables. Use {.code in_place = FALSE} to mute this message, or {.code in_place = TRUE} to write to the underlying tables."
    )
    in_place <- FALSE
  }

  dm_rows_run(x, y, operation_name, top_down, in_place, require_keys, progress = progress)
}

dm_rows_check <- function(x, y) {
  check_not_zoomed(x)
  check_not_zoomed(y)

  check_same_src(x, y)
  check_tables_superset(x, y)
  tables <- dm_get_tables_impl(y)
  walk2(dm_get_tables_impl(x)[names(tables)], tables, check_columns_superset)
  check_keys_compatible(x, y)
}

check_same_src <- function(x, y) {
  tables <- c(dm_get_tables_impl(x), dm_get_tables_impl(y))
  if (!all_same_source(tables)) {
    abort_not_same_src()
  }
}

check_tables_superset <- function(x, y) {
  tables_missing <- setdiff(src_tbls_impl(y), src_tbls_impl(x))
  if (has_length(tables_missing)) {
    abort_tables_missing(tables_missing)
  }
}

check_columns_superset <- function(target_tbl, tbl) {
  columns_missing <- setdiff(colnames(tbl), colnames(target_tbl))
  if (has_length(columns_missing)) {
    abort_columns_missing(columns_missing)
  }
}

check_keys_compatible <- function(x, y) {
  # FIXME
}

get_dm_rows_op <- function(operation_name) {
  switch(
    operation_name,
    "insert" = list(fun = do_rows_insert, pb_label = "inserting rows"),
    "append" = list(fun = do_rows_append, pb_label = "appending rows"),
    "update" = list(fun = do_rows_update, pb_label = "updating rows"),
    "patch" = list(fun = do_rows_patch, pb_label = "patching rows"),
    "upsert" = list(fun = do_rows_upsert, pb_label = "upserting rows"),
    "delete" = list(fun = do_rows_delete, pb_label = "deleting rows"),
    "truncate" = list(fun = rows_truncate_, pb_label = "truncating rows")
  )
}

do_rows_insert <- function(x, y, by = NULL, ..., autoinc_col = NULL) {
  stopifnot(is.null(autoinc_col))
  rows_insert(x, y, by = by, ..., conflict = "ignore")
}

do_rows_append <- function(x, y, by = NULL, ..., in_place = FALSE, autoinc_col = NULL) {
  if (is.null(autoinc_col)) {
    return(rows_append(x, y, ..., in_place = in_place))
  } else if (inherits(x, "data.frame")) {
    return(rows_append_ai_local(x, y, ..., autoinc_col))
  }

  # Assuming in-place append operation on dbplyr data source
  stopifnot(inherits(x, "tbl_dbi"))
  stopifnot(in_place)

  # FIXME: optimize if unique key exists
  returning <- sym(autoinc_col)
  key_values <-
    y %>%
    select(!!returning) %>%
    collect() %>%
    pull()

  if (anyDuplicated(key_values)) {
    cli::cli_abort("Duplicate values for autoincrement primary key {.field {autoinc_col}}.")
  }

  autoinc_col_new <- paste0(autoinc_col, "_new")
  if (length(key_values) == 0) {
    return(tibble(
      !!sym(autoinc_col) := integer(),
      !!sym(autoinc_col_new) := integer(),
    ))
  }

  source_rows <- map(key_values, ~ select(filter(y, !!returning == !!.x), -!!returning))

  con <- dbplyr::remote_con(x)
  # FIXME can be removed after depending on dbplyr >= 2.4.0
  append_args <- rlang::fn_fmls_names(dbplyr::sql_query_append)
  old_interface <- identical(append_args, c("con", "x_name", "y", "...", "returning_cols"))
  if (old_interface) {
    target_name <- remote_name_qual(x)
    insert_queries <- map(
      source_rows,
      ~ dbplyr::sql_query_append(
        con,
        target_name,
        .x,
        returning_cols = autoinc_col
      )
    )
  } else {
    # FIXME: dbplyr::remote_table() private in dbplyr 2.3.3, public in dbplyr 2.4.0
    dbplyr_ns <- asNamespace("dbplyr")
    remote_table <- mget("remote_table", dbplyr_ns, mode = "function", ifnotfound = list(NULL))[[1]]

    insert_queries <- map(
      source_rows,
      ~ dbplyr::sql_query_append(
        con,
        remote_table(x),
        from = dbplyr::sql_render(.x, con),
        insert_cols = colnames(.x),
        returning_cols = autoinc_col
      )
    )
  }

  autoinc_col_orig <- paste0(autoinc_col, "_orig")

  insert_sql <- sql(unlist(insert_queries))

  # Run INSERT INTO queries, side effect!
  # Must run queries individually, e.g. on Postgres:
  # > WITH clause containing a data-modifying statement must be at the top level
  insert_res <- map(insert_queries, ~ DBI::dbGetQuery(con, .x))

  out <- tibble(
    bind_rows(!!!insert_res),
    !!sym(autoinc_col_orig) := !!key_values
  )

  set_names(out[2:1], c(autoinc_col, autoinc_col_new))
}

do_rows_update <- function(x, y, by = NULL, ..., autoinc_col = NULL) {
  stopifnot(is.null(autoinc_col))
  rows_update(x, y, by = by, ..., unmatched = "ignore")
}

do_rows_patch <- function(x, y, by = NULL, ..., autoinc_col = NULL) {
  stopifnot(is.null(autoinc_col))
  rows_patch(x, y, by = by, ..., unmatched = "ignore")
}

do_rows_upsert <- function(x, y, by = NULL, ..., autoinc_col = NULL) {
  stopifnot(is.null(autoinc_col))
  rows_upsert(x, y, by = by, ...)
}

do_rows_delete <- function(x, y, by = NULL, ..., autoinc_col = NULL) {
  stopifnot(is.null(autoinc_col))
  rows_delete(x, y, by = by, ..., unmatched = "ignore")
}

dm_rows_run <- function(
  x,
  y,
  rows_op_name,
  top_down,
  in_place,
  require_keys,
  progress = NA
) {
  # topologically sort tables
  graph <- create_graph_from_dm(x, directed = TRUE)
  topo <- graph_topo_sort(graph, mode = if (top_down) "in" else "out")
  tables <- intersect(names(topo), src_tbls_impl(y))

  # Use tables and keys
  target_tbls <- dm_get_tables_impl(x)[tables]
  tbls <- dm_get_tables_impl(y)[tables]

  if (require_keys) {
    all_pks <- dm_get_all_pks(x)
    if (!(all(tables %in% all_pks$table))) {
      cli::cli_abort(
        "{.fun dm_rows_{rows_op_name}} requires the {.cls dm} object to have primary keys for all target tables.",
        call = dm_error_call()
      )
    }
    keys <- all_pks$pk_col[match(tables, all_pks$table)]
  } else {
    keys <- rep_along(tables, list(NULL))
  }

  # FIXME: Use keyholder?

  rows_op <- get_dm_rows_op(rows_op_name)
  ticker <- new_ticker(rows_op$pb_label, length(tables), progress)
  op_ticker <- ticker(rows_op$fun)

  if (!in_place || (!is_src_db(x) && rows_op_name == "append")) {
    op_results <- target_tbls
  }

  # run operation(target_tbl, tbl, in_place = in_place) for each table
  for (i in seq_along(tables)) {
    table <- tables[[i]]
    tbl <- tbls[[i]]
    # FIXME: implement for in_place = FALSE
    if (is_src_db(x) && in_place && (rows_op_name == "append")) {
      autoinc_col <- get_autoinc_col(x, table, colnames(tbl))
    } else if (!is_src_db(x) && (rows_op_name == "append")) {
      # `colnames(tbl)` could not contain the PK col, if it's left out in `y`
      # in order to apply the AI logic for local dm's we need `colnames(x[[table]])`
      autoinc_col <- get_autoinc_col(x, table, colnames(x[[table]]))
    } else {
      autoinc_col <- NULL
    }

    # Returns
    # 1. tibble if autoinc_col is set and remote
    # 2. list of tibbles (res_tibble and ai-lookup) autoinc_col is set and local
    # 3. `target_tbl` otherwise
    res <- op_ticker(
      target_tbls[[i]],
      tbl,
      by = keys[[i]],
      in_place = in_place,
      autoinc_col = autoinc_col
    )

    if (!is.null(autoinc_col) && is_src_db(x)) {
      tbls <- align_autoinc_fks(tbls, x, table, res)
    } else if (!is.null(autoinc_col) && !is_src_db(x)) {
      op_results[[i]] <- res$x_new
      # `if` is necessary, since in case y doesn't contain the AI column, the target
      # table should still have the AI values, but no update of potentially existing
      # FKs should take place
      if (!is.null(get_autoinc_col(x, table, colnames(tbl)))) {
        tbls <- align_autoinc_fks(tbls, x, table, res$ai_lookup)
      }
    } else if (!in_place) {
      op_results[[i]] <- res
    }
  }

  if (in_place && is_src_db(x)) {
    invisible(x)
  } else {
    x %>%
      dm_patch_tbl(!!!op_results)
  }
}

dm_patch_tbl <- function(dm, ...) {
  check_not_zoomed(dm)

  new_tables <- list2(...)

  # FIXME: Better error message for unknown tables

  def <- dm_get_def(dm)
  idx <- match(names(new_tables), def$table)
  def[idx, "data"] <- list(unname(new_tables))
  dm_from_def(def)
}

get_autoinc_col <- function(x, table, cols) {
  my_pk <- dm_get_all_pks_impl(x, table)
  stopifnot(nrow(my_pk) %in% 0:1)

  if (!isTRUE(my_pk$autoincrement)) {
    return(NULL)
  }

  pk_col <- get_key_cols(my_pk$pk_col[[1]])
  if (!(pk_col %in% cols)) {
    return(NULL)
  }

  pk_col
}

align_autoinc_fks <- function(tbls, target_dm, table, returning_rows) {
  fks <- dm_get_all_fks_impl(target_dm, table)
  fks_target <- fks[fks$child_table %in% names(tbls), ]

  all_target_tables <- dm_get_tables(target_dm)[fks_target$child_table]
  all_target_cols <- unique(unlist(map(all_target_tables, colnames)))

  pk_col <- names(returning_rows)[[1]]
  new_pk_col <- derive_temp_column_name(all_target_cols, pk_col)
  names(returning_rows)[[2]] <- new_pk_col

  # Structure: <pk_col>, <new_pk_col>
  if (is_src_db(target_dm)) {
    align_tbl <- dbplyr::copy_inline(dm_get_con(target_dm), returning_rows)
  } else {
    align_tbl <- returning_rows
  }

  for (i in seq_along(fks_target$child_table)) {
    child_table <- fks_target$child_table[[i]]
    child_fk_col <- get_key_cols(fks_target$child_fk_cols[[i]])
    tbl <- tbls[[child_table]]

    if (child_fk_col %in% colnames(tbl)) {
      tbls[[child_table]] <-
        tbl %>%
        left_join(align_tbl, by = vec_c(!!child_fk_col := pk_col)) %>%
        select(-!!sym(child_fk_col), !!sym(child_fk_col) := sym(new_pk_col)) %>%
        select(!!!colnames(tbl))
    }
  }

  tbls
}

derive_temp_column_name <- function(tbl_names, base, suffix = "_new") {
  new_name <- paste0(base, suffix)
  tbl_names <- colnames(tbl)

  repeat {
    if (!(new_name %in% tbl_names)) {
      return(new_name)
    }
    new_name <- paste0(new_name, "_")
  }
}

rows_append_ai_local <- function(x, y, autoinc_col) {
  num_new_rows <- nrow(y)
  if (is_empty(x[[autoinc_col]])) {
    init_ai_val <- 1L
  } else {
    init_ai_val <- max(x[[autoinc_col]]) + 1L
  }
  new_col_name <- derive_temp_column_name(colnames(x), autoinc_col)
  ai_lu <- tibble(
    !!autoinc_col := y[[autoinc_col]],
    !!new_col_name := seq.int(init_ai_val, length.out = num_new_rows)
  )
  y_new <-
    y %>%
    mutate(!!new_col_name := !!ai_lu[[new_col_name]]) %>%
    select(
      -!!intersect(colnames(y), autoinc_col),
      !!autoinc_col := !!new_col_name,
      !!setdiff(colnames(x), autoinc_col)
    )
  list(x_new = rows_append(x, y_new), ai_lookup = ai_lu)
}

# Errors ------------------------------------------------------------------

abort_columns_missing <- function(...) {
  # FIXME
  cli::cli_abort("Column mismatch between target and source tables.", call = dm_error_call())
}

error_txt_columns_missing <- function(...) {
  # FIXME
}

abort_tables_missing <- function(...) {
  # FIXME
  cli::cli_abort("Source tables are missing from target dm.", call = dm_error_call())
}

error_txt_tables_missing <- function(...) {
  # FIXME
}

test_that("dumma", {
  expect_snapshot({
    "dummy"
  })
})

test_that("dm_rows_insert()", {
  skip_if_not_installed("RSQLite")
  local_options(lifecycle_verbosity = "quiet")

  # Entire dataset with all dimension tables populated
  # with flights and weather data truncated:
  flights_init <-
    dm_nycflights13() %>%
    dm_zoom_to(flights) %>%
    filter(FALSE) %>%
    dm_update_zoomed() %>%
    dm_zoom_to(weather) %>%
    filter(FALSE) %>%
    dm_update_zoomed()

  # Must use SQLite because other databases have strict foreign key constraints
  sqlite <- DBI::dbConnect(RSQLite::SQLite())

  # Target database:
  flights_sqlite <- copy_dm_to(sqlite, flights_init, temporary = FALSE)

  expect_snapshot({
    print(dm_nrow(flights_sqlite))

    # First update:
    flights_hour10 <-
      dm_nycflights13() %>%
      dm_select_tbl(flights, weather) %>%
      dm_zoom_to(flights) %>%
      filter(month == 1, day == 10, hour == 10) %>%
      dm_update_zoomed() %>%
      dm_zoom_to(weather) %>%
      filter(month == 1, day == 10, hour == 10) %>%
      dm_update_zoomed()
    print(dm_nrow(flights_hour10))

    # Copy to temporary tables on the target database:
    flights_hour10_sqlite <- copy_dm_to(sqlite, flights_hour10)

    # Dry run by default:
    out <- dm_rows_append(flights_sqlite, flights_hour10_sqlite)
    print(dm_nrow(flights_sqlite))

    # Explicitly request persistence:
    dm_rows_append(flights_sqlite, flights_hour10_sqlite, in_place = TRUE)
    print(dm_nrow(flights_sqlite))

    # Second update:
    flights_hour11 <-
      dm_nycflights13() %>%
      dm_select_tbl(flights, weather) %>%
      dm_zoom_to(flights) %>%
      filter(month == 1, day == 10, hour == 11) %>%
      dm_update_zoomed() %>%
      dm_zoom_to(weather) %>%
      filter(month == 1, day == 10, hour == 11) %>%
      dm_update_zoomed()

    # Copy to temporary tables on the target database:
    flights_hour11_sqlite <- copy_dm_to(sqlite, flights_hour11)

    # Explicit dry run:
    flights_new <- dm_rows_append(
      flights_sqlite,
      flights_hour11_sqlite,
      in_place = FALSE
    )
    print(dm_nrow(flights_new))
    print(dm_nrow(flights_sqlite))

    # Check for consistency before applying:
    flights_new %>%
      dm_examine_constraints()

    # Apply:
    dm_rows_append(flights_sqlite, flights_hour11_sqlite, in_place = TRUE)
    print(dm_nrow(flights_sqlite))

    # Disconnect
    DBI::dbDisconnect(sqlite)
  })
})

test_that("dm_rows_update()", {
  # Test bad column order
  dm_filter_rearranged <-
    dm_for_filter() %>%
    dm_select(tf_2, d, everything()) %>%
    dm_select(tf_4, i, everything()) %>%
    dm_select(tf_5, l, m, everything())

  dm_copy <- suppressMessages(copy_dm_to(my_db_test_src(), dm_filter_rearranged))

  dm_update_local <- dm(
    tf_1 = tibble(
      a = 2L,
      b = "q"
    ),
    tf_4 = tibble(
      h = "e",
      i = "sieben",
    ),
    tf_5 = tibble(
      k = 3L,
      ww = 3,
    ),
  )

  dm_update_copy <- suppressMessages(copy_dm_to(my_db_test_src(), dm_update_local))

  expect_snapshot({
    dm_copy %>%
      pull_tbl(tf_2) %>%
      arrange_all()

    dm_copy %>%
      dm_rows_update(dm_update_copy) %>%
      pull_tbl(tf_2) %>%
      arrange_all()

    dm_copy %>%
      pull_tbl(tf_2) %>%
      arrange_all()

    dm_copy %>%
      dm_rows_update(dm_update_copy, in_place = FALSE) %>%
      pull_tbl(tf_2) %>%
      arrange_all()

    dm_copy %>%
      dm_get_tables() %>%
      map(arrange_all)

    dm_copy %>%
      dm_rows_update(dm_update_copy, in_place = TRUE)

    dm_copy %>%
      dm_get_tables() %>%
      map(arrange_all)
  })
})

test_that("dm_rows_truncate()", {
  local_options(lifecycle_verbosity = "warning")

  suppressMessages(dm_copy <- copy_dm_to(my_db_test_src(), dm_for_filter()))

  dm_truncate_local <- dm(
    tf_2 = tibble(
      c = c("worm"),
      d = 10L,
    ),
    tf_5 = tibble(
      k = 3L,
      m = "tree",
    ),
  )

  dm_truncate_copy <- suppressMessages(copy_dm_to(my_db_test_src(), dm_truncate_local))

  expect_snapshot({
    dm_copy %>%
      pull_tbl(tf_2) %>%
      arrange_all()

    dm_copy %>%
      dm_rows_truncate(dm_truncate_copy) %>%
      pull_tbl(tf_2) %>%
      arrange_all()

    dm_copy %>%
      pull_tbl(tf_2) %>%
      arrange_all()

    dm_copy %>%
      dm_rows_truncate(dm_truncate_copy, in_place = FALSE) %>%
      pull_tbl(tf_2) %>%
      arrange_all()

    dm_copy %>%
      dm_get_tables() %>%
      map(arrange_all)

    dm_copy %>%
      dm_rows_truncate(dm_truncate_copy, in_place = TRUE)

    dm_copy %>%
      dm_get_tables() %>%
      map(arrange_all)
  })
})


test_that("dm_rows_update() requires primary keys", {
  skip_if_remote_src()
  local_options(lifecycle_verbosity = "quiet")

  x <- dm(a = tibble::tibble(id = 1L, val = "x"))
  y <- dm(a = tibble::tibble(id = 1L, val = "y"))

  expect_snapshot(error = TRUE, {
    dm_rows_update(x, y, in_place = FALSE)
  })
})


# tests for compound keys -------------------------------------------------

test_that("output for compound keys", {
  skip("COMPOUND")
  local_options(lifecycle_verbosity = "warning")

  expect_snapshot({
    target_dm <- dm_filter(nyc_comp(), weather, pressure > 1010) %>% dm_apply_filters()
    insert_dm <-
      dm_filter(nyc_comp(), weather, pressure <= 1010) %>%
      dm_apply_filters() %>%
      dm_select_tbl(flights, weather)
    dm_rows_insert(target_dm, insert_dm, in_place = FALSE)
    dm_rows_truncate(nyc_comp(), insert_dm, in_place = FALSE)
  })
})


# tests for autoincrement PKs ---------------------------------------------

test_that("dm_rows_append() works with autoincrement PKs and FKS for selected DBs", {
  skip_if_src_not(c("postgres", "mssql", "sqlite"))

  con_db <- my_test_con()

  # Setup
  dm_ai_w_keys <-
    dm_for_autoinc_1() %>%
    dm_add_pk(t1, a, autoincrement = TRUE) %>%
    dm_add_pk(t2, c, autoincrement = TRUE) %>%
    dm_add_pk(t4, g, autoincrement = TRUE) %>%
    dm_add_fk(t2, d, t1) %>%
    dm_add_fk(t3, e, t1) %>%
    dm_add_fk(t4, h, t2)

  local_dm <-
    dm_ai_w_keys %>%
    collect()

  withr::defer({
    order_of_deletion <- c("t4", "t2", "t3", "t1")
    walk(
      order_of_deletion,
      ~ try(DBI::dbExecute(con_db, paste0("DROP TABLE ", .x)))
    )
  })

  dm_ai_empty_remote <-
    local_dm %>%
    dm_ptype() %>%
    copy_dm_to(con_db, ., temporary = FALSE)

  # Tests
  dm_ai_insert <-
    dm_for_autoinc_1() %>%
    # Remove one PK column, only provided by database
    dm_select(t4, -g) %>%
    dm_zoom_to(t3) %>%
    filter(0L == 1L) %>%
    dm_update_zoomed()

  expect_silent(
    filled_dm <- dm_rows_append(
      dm_ai_empty_remote,
      dm_ai_insert,
      in_place = FALSE,
      progress = FALSE
    ) %>%
      collect()
  )

  expect_silent(
    filled_dm_in_place <- dm_rows_append(
      dm_ai_empty_remote,
      dm_ai_insert,
      in_place = TRUE,
      progress = FALSE
    ) %>%
      collect()
  )

  expect_silent(
    filled_dm_in_place_twice <- dm_rows_append(
      dm_ai_empty_remote,
      dm_ai_insert,
      in_place = TRUE,
      progress = FALSE
    ) %>%
      collect()
  )

  expect_snapshot(
    variant = my_test_src_name,
    {
      local_dm$t1
      local_dm$t2
      local_dm$t3
      local_dm$t4
      filled_dm$t1
      filled_dm$t2
      filled_dm$t3
      filled_dm$t4
      filled_dm_in_place$t1
      filled_dm_in_place$t2
      filled_dm_in_place$t3
      filled_dm_in_place$t4
      filled_dm_in_place_twice$t1
      filled_dm_in_place_twice$t2
      filled_dm_in_place_twice$t3
      filled_dm_in_place_twice$t4
    }
  )
})


test_that("dm_rows_append() works with autoincrement PKs and FKS locally", {
  skip_if_remote_src()

  # Setup
  local_dm <-
    dm_for_autoinc_1() %>%
    dm_add_pk(t1, a, autoincrement = TRUE) %>%
    dm_add_pk(t2, c, autoincrement = TRUE) %>%
    dm_add_pk(t4, g, autoincrement = TRUE) %>%
    dm_add_fk(t2, d, t1) %>%
    dm_add_fk(t3, e, t1) %>%
    dm_add_fk(t4, h, t2)

  dm_ai_empty <-
    local_dm %>%
    dm_ptype()

  # Corner case: empty + empty = empty
  expect_identical(
    expect_silent(dm_rows_append(
      dm_ai_empty,
      dm_ai_empty,
      in_place = FALSE,
      progress = FALSE
    )),
    dm_ai_empty
  )

  # Tests
  dm_ai_insert <-
    dm_for_autoinc_1() %>%
    # Remove one PK column, only provided by local logic
    dm_select(t4, -g) %>%
    dm_zoom_to(t3) %>%
    filter(0L == 1L) %>%
    dm_update_zoomed()

  expect_silent(
    filled_dm <- dm_rows_append(
      dm_ai_empty,
      dm_ai_insert,
      in_place = FALSE,
      progress = FALSE
    )
  )

  # Corner case: data + empty = data
  expect_identical(
    expect_silent(dm_rows_append(
      filled_dm,
      dm_ai_empty,
      in_place = FALSE,
      progress = FALSE
    )),
    filled_dm
  )

  expect_error(
    dm_rows_append(
      dm_ai_empty,
      dm_ai_insert,
      in_place = TRUE,
      progress = FALSE
    )
  )

  expect_silent(
    filled_twice_dm <- dm_rows_append(
      filled_dm,
      dm_ai_insert,
      in_place = FALSE,
      progress = FALSE
    )
  )

  expect_snapshot(
    variant = my_test_src_name,
    {
      local_dm$t1
      local_dm$t2
      local_dm$t3
      local_dm$t4
      filled_dm$t1
      filled_dm$t2
      filled_dm$t3
      filled_dm$t4
      filled_twice_dm$t1
      filled_twice_dm$t2
      filled_twice_dm$t3
      filled_twice_dm$t4
    }
  )
})

#' Unpack a JSON column
#'
#' A wrapper around [tidyr::unpack()] that extracts its data from a JSON column.
#' The inverse of [json_pack()].
#'
#' @param data A data frame, a data frame extension (e.g. a tibble), or  a lazy
#'   data frame (e.g. from dbplyr or dtplyr).
#' @inheritParams rlang::args_dots_used
#' @inheritParams tidyr::unpack
#'
#' @return An object of the same type as `data`
#' @export
#' @examples
#' tibble(a = 1, b = '{ "c": 2, "d": 3 }') %>%
#'   json_unpack(b)
json_unpack <- function(data, cols, ..., names_sep = NULL, names_repair = "check_unique") {
  check_dots_used()

  UseMethod("json_unpack")
}

#' @export
json_unpack.data.frame <- function(data, cols, ..., names_sep = NULL, names_repair = "check_unique") {
  if (missing(cols)) {
    abort("The `cols` argument must be provided")
  }
  data %>%
    mutate(across({{ cols }}, ~ map_dfr(., ~ jsonlite::fromJSON(.) %>% as_tibble()))) %>%
    unpack(
      {{ cols }},
      names_sep = names_sep,
      names_repair = names_repair
    )
}

# @export
json_unpack.tbl_lazy <- function(data, cols, names_sep = NULL, names_repair = "check_unique") {
  # select string_agg(quotename(k) + case t
  #                   when 0 then ' nchar(1)'       -- javascript null
  #                   when 1 then ' nvarchar(max)'  -- javascript string
  #                   when 2 then ' float'          -- javascript number
  #                   when 3 then ' bit'            -- javascript boolean
  #                   else ' nvarchar(max) as json' -- javascript array or object
  #                   end, ', ') within group (order by k)
  # from (
  #   select j2.[key], max(j2.[type])
  #   from test
  #   cross apply openjson(case when json_col like '{%}' then '[' + json_col + ']' else json_col end) as j1
  #   cross apply openjson(j1.value) as j2
  #   group by j2.[key]
  # ) as kt(k, t)
}

#' @autoglobal
dm_upgrade <- function(dm, quiet) {
  # Versioned dm objects introduced in dm 0.2.1, unversioned dm same as version 0
  version <- attr(dm, "version") %||% 0L

  if (version < 1L) {
    # FIXME: Can't give unconditional message, because IDE can tickle object in environment,
    # and this also produces a message. Don't emit message when calling str()?
    if (!quiet) {
      message("Upgrading dm object created with dm <= 0.2.1.")
    }
    def <- unclass(dm)$def
    def$fks <- list_of(
      !!!map2(
        def$fks,
        def$pks,
        ~ {
          .x[["ref_column"]] <- .y[["column"]]
          .x <- .x[c("ref_column", "table", "column")]
          .x
        }
      )
    )
    dm <- dm_from_def(def, zoomed = is_zoomed(dm), validate = FALSE)
  }

  if (version < 2L) {
    # FIXME: Can't give unconditional message, because IDE can tickle object in environment,
    # and this also produces a message. Don't emit message when calling str()?
    if (!quiet) {
      message("Upgrading dm object created with dm <= 0.2.4.")
    }
    def <- unclass(dm)$def
    def$fks <- list_of(!!!map(def$fks, mutate, on_delete = "no_action"))
    dm <- dm_from_def(def, zoomed = is_zoomed(dm), validate = FALSE)
  }

  if (version < 3L) {
    # FIXME: Can't give unconditional message, because IDE can tickle object in environment,
    # and this also produces a message. Don't emit message when calling str()?
    if (!quiet) {
      message("Upgrading dm object created with dm <= 0.3.0.")
    }
    def <- unclass(dm)$def
    def$uuid <- vec_new_uuid_along(def$table)
    dm <- dm_from_def(def, zoomed = is_zoomed(dm), validate = FALSE)
  }

  if (version < 4L) {
    if (!quiet) {
      message("Upgrading dm object created with dm <= 1.0.3.")
    }
    def <- unclass(dm)$def
    def$pks <- map(def$pks, mutate, autoincrement = FALSE) %>%
      vctrs::as_list_of(new_pk())
    def <- mutate(def, uks = vctrs::list_of(new_uk()), .after = pks)
    dm <- dm_from_def(def, zoomed = is_zoomed(dm))
  }

  dm
}

.onLoad <- function(libname, pkgname) {
  backports::import(pkgname, c("...length"))
  backports::import(pkgname, c("...names"))

  if (getRversion() >= "3.4") {
    dm_financial <<- memoise::memoise(dm_financial, cache = cache_attach())
    dm_financial_sqlite <<- memoise::memoise(dm_financial_sqlite, cache = cache_attach())
  }

  nycflights_subset <<- memoise::memoise(nycflights_subset, cache = cache_attach())

  dm_has_financial <<- memoise::memoise(dm_has_financial, cache = cache_attach())

  check_suggested <<- memoise::memoise(check_suggested)

  register_pkgdown_methods()

  check_version_on_load
  vctrs::s3_register("waldo::compare_proxy", "dm")

  check_version_on_load(
    "RSQLite",
    "2.2.8",
    "to use the {.code returning} argument in {.code rows_*()}."
  )
  check_version_on_load("dbplyr", "2.2.0", "to use the {.code rows_*()} functions.")

  # rigg(enum_pk_candidates_impl)
  # rigg(build_copy_data)
  # rigg(dm_insert_zoomed_outgoing_fks)
  # rigg(dm_upgrade)
  # rigg(check_df_structure)
  # rigg(dm_insert_zoomed)
  # rigg(dm_select_tbl_impl)
  # rigg(dm_insert_zoomed_outgoing_fks)
  # rigg(dm_rm_pk_impl)
  # rigg(dm_rm_fk_impl)
  # rigg(cdm_rm_fk)

  rlang::run_on_load()
}

.onAttach <- function(libname, pkgname) {
  if (!isTRUE(getOption("dm.suppress_dplyr_startup_message"))) {
    attached <- search()
    missing_pkgs <- setdiff(
      "dplyr",
      sub("^package:", "", grep("^package:", attached, value = TRUE))
    )
    if (length(missing_pkgs) > 0) {
      packageStartupMessage(cli::format_message(c(
        "!" = paste0(
          "In a coming version, {.pkg dm} will no longer reexport {.pkg dplyr} functions. ",
          'For best results, run {.code library("dplyr")} before {.code library("dm")}.'
        ),
        "i" = "To suppress this message unconditionally, set {.code options(dm.suppress_dplyr_startup_message = TRUE)}."
      )))
    }
  }
}

setup_graph_functions <- function(ns) {
  use_igraph <- getOption("dm.use_igraph")

  if (is.null(use_igraph) && !rlang::is_installed("igraph")) {
    packageStartupMessage(cli::format_message(c(
      "!" = "Package {.pkg igraph} is not installed. {.pkg dm} will use a pure R fallback.",
      "i" = "Install {.pkg igraph} for better performance: {.run install.packages(\"igraph\")}",
      "i" = "Set {.code options(dm.use_igraph = FALSE)} to suppress this message."
    )))
  }

  if (isTRUE(use_igraph)) {
    rlang::check_installed("igraph", reason = "because `options(dm.use_igraph = TRUE)` is set.")
  } else if (isFALSE(use_igraph) || !rlang::is_installed("igraph")) {
    graph_from_data_frame <<- graph_from_data_frame_fallback
    graph_vertices <<- graph_vertices_fallback
    graph_edges <<- graph_edges_fallback
    graph_dfs <<- graph_dfs_fallback
    graph_topo_sort <<- graph_topo_sort_fallback
    graph_distances <<- graph_distances_fallback
    graph_induced_subgraph <<- graph_induced_subgraph_fallback
    graph_shortest_paths <<- graph_shortest_paths_fallback
    graph_delete_vertices <<- graph_delete_vertices_fallback
    graph_neighbors <<- graph_neighbors_fallback
    graph_vcount <<- graph_vcount_fallback
    graph_girth <<- graph_girth_fallback
  }
}

rigg <- function(fun) {
  name <- deparse(substitute(fun))

  rig <- get("rig", asNamespace("boomer"), mode = "function")

  assign(name, rig(fun, ignore = c("~", "{", "(")), getNamespace("dm"))
}

check_version_on_load <- function(package, version, reason) {
  check_version <- function(...) {
    if (utils::packageVersion(package) < version) {
      message <- c(
        paste0("You need {.pkg {package} >= {version}} ", reason),
        `i` = 'Install with {.code install.packages("{package}")}'
      )

      cli::cli_inform(message)
    }
  }

  # Always register hook in case package is later unloaded & reloaded
  setHook(packageEvent(package, "onLoad"), check_version)

  # Avoid registration failures during loading (pkgload or regular)
  if (isNamespaceLoaded(package)) {
    check_version()
  }

  invisible()
}

#' Wrap dm into a single tibble dm
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `dm_wrap_tbl()` creates a single tibble dm containing the `root` table
#' enhanced with all the data related to it
#' through the relationships stored in the dm.
#' It runs a sequence of [dm_nest_tbl()] and [dm_pack_tbl()] operations
#' on the dm.
#'
#' @param strict Whether to fail for cyclic dms that cannot be wrapped into a
#'   single table, if `FALSE` a partially wrapped dm will be returned.
#' @param dm A cycle free dm object.
#' @param root Table to wrap the dm into (unquoted).
#' @param progress Whether to display a progress bar, if `NA` (the default)
#'   hide in non-interactive mode, show in interactive mode. Requires the
#'   'progress' package.
#'
#' @details
#' `dm_wrap_tbl()` is an inverse to `dm_unwrap_tbl()`,
#' i.e., wrapping after unwrapping returns the same information
#' (disregarding row and column order).
#' The opposite is not generally true:
#' since `dm_wrap_tbl()` keeps only rows related directly or indirectly to
#' rows in the `root` table.
#' Even if all referential constraints are satisfied,
#' unwrapping after wrapping loses rows in parent tables
#' that don't have a corresponding row in the child table.
#'
#' This function differs from `dm_flatten_to_tbl()` and `dm_squash_to_tbl()` ,
#' which always return a single table, and not a `dm` object.
#'
#' @return A `dm` object.
#' @export
#' @seealso [dm_unwrap_tbl()], [dm_nest_tbl()],
#'   [dm_examine_constraints()],
#'   [dm_examine_cardinalities()].
#' @examples
#' dm_nycflights13() %>%
#'   dm_wrap_tbl(root = airlines)
dm_wrap_tbl <- function(dm, root, strict = TRUE, progress = NA) {
  dm_local_error_call()
  check_dm(dm)

  wrap_plan <- dm_wrap_tbl_plan(dm, {{ root }})

  ticker <- new_ticker(
    "Wrapping dm object",
    n = nrow(wrap_plan),
    progress = progress,
    top_level_fun = "dm_wrap_tbl"
  )

  wrapped_dm <- reduce2(
    wrap_plan$action,
    wrap_plan$table,
    ticker(function(dm, f, table) exec(f, dm, table)),
    .init = dm
  )

  # inform or fail if we have a cycle
  if (length(wrapped_dm) > 1) {
    if (strict) {
      # FIXME: Detect earlier
      cli::cli_abort(
        "The {.cls dm} is not cycle free and can't be wrapped into a single tibble.",
        call = dm_error_call()
      )
    }
  }

  wrapped_dm
}

dm_wrap_tbl_plan <- function(dm, root) {
  # process args
  root_name <- dm_tbl_name(dm, {{ root }})

  # initiate graph and positions
  graph <- create_graph_from_dm(dm, directed = TRUE)
  positions <- node_type_from_graph(graph, drop = root_name)

  # build plan of actions to wrap terminal nodes as long as they're not the root
  wrap_plan <- tibble(action = character(0), table = character(0))
  repeat {
    child_name <- names(positions)[positions == "terminal child"][1]
    has_terminal_child <- !is.na(child_name)
    if (has_terminal_child) {
      wrap_plan <- add_row(wrap_plan, action = "dm_nest_tbl", table = child_name)
      graph <- graph_delete_vertices(graph, child_name)
      positions <- node_type_from_graph(graph, drop = root_name)
    }
    parent_name <- names(positions)[positions == "terminal parent"][1]
    has_terminal_parent <- !is.na(parent_name)
    if (has_terminal_parent) {
      wrap_plan <- add_row(wrap_plan, action = "dm_pack_tbl", table = parent_name)
      graph <- graph_delete_vertices(graph, parent_name)
      positions <- node_type_from_graph(graph, drop = root_name)
    }
    if (!has_terminal_child && !has_terminal_parent) break
  }
  wrap_plan
}


#' Unwrap a single table dm
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `dm_unwrap_tbl()` unwraps all tables in a dm object so that the resulting dm
#' matches a given ptype dm.
#' It runs a sequence of [dm_unnest_tbl()] and [dm_unpack_tbl()] operations
#' on the dm.
#'
#' @param dm A dm.
#' @param ptype A dm, only used to query names of primary and foreign keys.
#' @inheritParams dm_wrap_tbl
#' @return A dm.
#' @seealso [dm_wrap_tbl()], [dm_unnest_tbl()],
#'   [dm_examine_constraints()],
#'   [dm_examine_cardinalities()],
#'   [dm_ptype()].
#' @export
#' @examples
#'
#' roundtrip <-
#'   dm_nycflights13() %>%
#'   dm_wrap_tbl(root = flights) %>%
#'   dm_unwrap_tbl(ptype = dm_ptype(dm_nycflights13()))
#' roundtrip
#'
#' # The roundtrip has the same structure but fewer rows:
#' dm_nrow(dm_nycflights13())
#' dm_nrow(roundtrip)
dm_unwrap_tbl <- function(dm, ptype, progress = NA) {
  check_dm(ptype)

  unwrap_plan <- dm_get_tables(dm) %>%
    imap(dm_unwrap_tbl_plan) %>%
    unlist(recursive = FALSE) %>%
    purrr::discard(~ nrow(.) == 0)

  ticker <- new_ticker(
    "Unwrapping dm object",
    n = length(unwrap_plan),
    progress = progress,
    top_level_fun = "dm_unwrap_tbl"
  )

  unwrapped_dm <- reduce(
    unwrap_plan,
    function(dm, row) {
      exec(row$action, dm, row$table, row$col, ptype)
    },
    .init = dm
  )
  unwrapped_dm
}

dm_unwrap_tbl_plan <- function(table, table_name) {
  nms <- names(table)

  children <- nms[map_lgl(table, inherits, "nested")]
  parents <- nms[map_lgl(table, inherits, "packed")]

  unnest_plan <-
    tibble(
      action = "dm_unnest_tbl",
      table = table_name,
      col = children
    ) %>%
    split.data.frame(seq_along(children))

  unpack_plan <-
    tibble(
      action = "dm_unpack_tbl",
      table = table_name,
      col = parents
    ) %>%
    split.data.frame(seq_along(parents))

  unwrap_plan_from_children <-
    # note: we cannot use bind_rows() because of https://github.com/tidyverse/dplyr/issues/6447,
    #   or even vec_rbind() because of https://github.com/r-lib/vctrs/issues/1640
    map(children, ~ dm_unwrap_tbl_plan(vec_c(!!!table[[.x]]), .x)) %>%
    unlist(recursive = FALSE)

  unwrap_plan_from_parents <-
    map(parents, ~ dm_unwrap_tbl_plan(table[[.x]], .x)) %>%
    unlist(recursive = FALSE)

  c(
    unnest_plan,
    unpack_plan,
    unwrap_plan_from_children,
    unwrap_plan_from_parents
  )
}

test_that("`dm_wrap_tbl()` and `dm_unwrap_tbl()` work", {
  skip_if_remote_src()

  dm1 <- dm_for_filter()
  dm_wrapped <- dm_wrap_tbl(dm1, tf_4)
  expect_length(dm_wrapped, 1)
  expect_equal(names(dm_wrapped), "tf_4")
  dm_unwrapped <- dm_unwrap_tbl(dm_wrapped, dm1)
  expect_equivalent_dm(dm_unwrapped, dm_unwrap_tbl(dm_wrapped, dm1))

  # to tibble
  expect_snapshot({
    wrapped <- dm_wrap_tbl(dm_for_filter(), tf_4)
    wrapped
    wrapped$tf_4
    wrapped$tf_4$tf_3$tf_2[[3]]
    wrapped$tf_4$tf_5[[2]]
  })

  # back to dm
  expect_snapshot({
    unwrapped <- dm_unwrap_tbl(dm_wrap_tbl(dm_for_filter(), tf_4), dm_for_filter())
    unwrapped
    unwrapped$tf_4
    unwrapped$tf_1
    unwrapped$tf_6
  })
})

test_that("`dm_wrap_tbl()` and `dm_unwrap_tbl()` round trip", {
  skip_if_remote_src()

  dm <- dm_for_filter()
  reduced_dm <-
    dm %>%
    dm_filter(tf_1 = TRUE)
  roundtrip_dm <- dm_unwrap_tbl(dm_wrap_tbl(dm, tf_1), dm)
  expect_equivalent_dm(
    roundtrip_dm,
    reduced_dm,
    sort = TRUE,
    ignore_on_delete = TRUE,
    ignore_autoincrement = TRUE
  )

  dm <- dm_for_filter()
  reduced_dm <-
    dm %>%
    dm_filter(tf_2 = TRUE)
  roundtrip_dm <- dm_unwrap_tbl(dm_wrap_tbl(dm, tf_2), dm)
  expect_equivalent_dm(
    roundtrip_dm,
    reduced_dm,
    sort = TRUE,
    ignore_on_delete = TRUE,
    ignore_autoincrement = TRUE
  )

  dm <- dm_for_filter()
  reduced_dm <-
    dm %>%
    dm_filter(tf_3 = TRUE)
  roundtrip_dm <- dm_unwrap_tbl(dm_wrap_tbl(dm, tf_3), dm)
  expect_equivalent_dm(
    roundtrip_dm,
    reduced_dm,
    sort = TRUE,
    ignore_on_delete = TRUE,
    ignore_autoincrement = TRUE
  )

  dm <- dm_for_filter()
  reduced_dm <-
    dm %>%
    dm_filter(tf_4 = TRUE)
  roundtrip_dm <- dm_unwrap_tbl(dm_wrap_tbl(dm, tf_4), dm)
  expect_equivalent_dm(
    roundtrip_dm,
    reduced_dm,
    sort = TRUE,
    ignore_on_delete = TRUE,
    ignore_autoincrement = TRUE
  )

  dm <- dm_for_filter()
  reduced_dm <-
    dm %>%
    dm_filter(tf_5 = TRUE)
  roundtrip_dm <- dm_unwrap_tbl(dm_wrap_tbl(dm, tf_5), dm)
  expect_equivalent_dm(
    roundtrip_dm,
    reduced_dm,
    sort = TRUE,
    ignore_on_delete = TRUE,
    ignore_autoincrement = TRUE
  )

  dm <- dm_for_filter()
  reduced_dm <-
    dm %>%
    dm_filter(tf_6 = TRUE)
  roundtrip_dm <- dm_unwrap_tbl(dm_wrap_tbl(dm, tf_6), dm)
  expect_equivalent_dm(
    roundtrip_dm,
    reduced_dm,
    sort = TRUE,
    ignore_on_delete = TRUE,
    ignore_autoincrement = TRUE
  )
})

test_that("`node_type_from_graph()` works", {
  dm1 <- dm_for_filter()
  graph <- create_graph_from_dm(dm1, directed = TRUE)
  expect_snapshot({
    node_type_from_graph(graph)
  })
  expect_snapshot({
    node_type_from_graph(graph, drop = "tf_4")
  })
})

test_that("`dm_wrap_tbl()` fails for cyclic dm", {
  skip_if_remote_src()

  # Create a cyclic dm: a -> b -> a
  a <- tibble::tibble(id = 1L, b_id = 1L)
  b <- tibble::tibble(id = 1L, a_id = 1L)
  cyclic_dm <- dm(a, b) %>%
    dm_add_pk(a, id) %>%
    dm_add_pk(b, id) %>%
    dm_add_fk(a, b_id, b) %>%
    dm_add_fk(b, a_id, a)

  expect_snapshot(error = TRUE, {
    dm_wrap_tbl(cyclic_dm, a)
  })
})

#' JSON nest join
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' A wrapper around [dplyr::nest_join()] which stores the joined data into a JSON column.
#' `json_nest_join()` returns all rows and columns in `x` with a new JSON columns that contains all nested matches from `y`.
#'
#' @inheritParams dplyr::nest_join
#' @param x,y A pair of data frames or data frame extensions (e.g. a tibble).
#' @seealso [dplyr::nest_join()], [json_pack_join()]
#' @export
#' @examples
#' df1 <- tibble::tibble(x = 1:3)
#' df2 <- tibble::tibble(x = c(1, 1, 2), y = c("first", "second", "third"))
#' df3 <- json_nest_join(df1, df2)
#' df3
#' df3$df2
json_nest_join <- function(x, y, by = NULL, ..., copy = FALSE, keep = FALSE, name = NULL) {
  UseMethod("json_nest_join")
}

#' @export
json_nest_join.data.frame <- function(x, y, by = NULL, ..., copy = FALSE, keep = FALSE, name = NULL) {
  check_suggested("jsonlite", use = TRUE, top_level_fun = "json_nest_join")
  check_dots_empty()

  name_var <- name %||% as_label(enexpr(y))
  dplyr::nest_join(x, y, by, copy = copy, keep = keep, name = name_var) %>%
    mutate(!!name_var := map(!!sym(name_var), jsonlite::toJSON, digits = NA))
}

test_that("json joins work", {
  df1 <- tibble(col1 = 1:2, key = letters[1:2])
  df2 <- tibble(col2 = c(3:4, pi), col3 = c("X", "Y", "Z"), key = letters[c(1, 1:2)])

  expect_snapshot({
    packed <- json_pack_join(df1, df2, by = "key")
    packed
    packed$df2
  })

  expect_snapshot({
    nested <- json_nest_join(df1, df2, by = "key")
    nested
    nested$df2
  })

  # Make sure precision is not lost in JSON conversion
  packed <- json_pack_join(df1, df2, by = "key")
  nested <- json_nest_join(df1, df2, by = "key")
  expect_equal(jsonlite::fromJSON(packed$df2[[3]])[["col2"]], pi)
  expect_equal(jsonlite::fromJSON(nested$df2[[2]])[["col2"]], pi)
})

# Graph abstraction layer.
# When igraph is installed, graph_* functions use igraph for best performance.
# When igraph is not installed, all graph_* functions are overwritten at load
# time with pure R fallbacks from R/igraph-fallback.R.
# All functions accept and return "dm_graph" objects (or the "dm_igraph" subclass).

rlang::on_load({
  setup_graph_functions(getNamespace("dm"))
})

# dm_igraph: wraps an igraph object; used when igraph is installed.
new_dm_igraph <- function(ig) {
  structure(
    list(igraph = ig),
    class = c("dm_igraph", "dm_graph")
  )
}

print.dm_graph <- function(x, ...) {
  n_v <- length(x$vnames)
  n_e <- length(x$from)
  directed_str <- if (isTRUE(x$directed)) "directed" else "undirected"
  cat(
    sprintf(
      "<dm_graph> %s, %d %s, %d %s\n",
      directed_str,
      n_v,
      if (n_v == 1L) "vertex" else "vertices",
      n_e,
      if (n_e == 1L) "edge" else "edges"
    )
  )
  if (n_v > 0L) {
    cat("Vertices:", paste(x$vnames, collapse = ", "), "\n")
  }
  if (n_e > 0L) {
    edge_strs <- paste(x$vnames[x$from], x$vnames[x$to], sep = "|")
    cat("Edges:", paste(edge_strs, collapse = ", "), "\n")
  }
  invisible(x)
}

# Forward to print.dm_graph by creating a plain dm_graph from igraph data.
print.dm_igraph <- function(x, ...) {
  ig <- x$igraph
  fallback <- new_dm_graph(
    directed = igraph::is_directed(ig),
    vnames = names(igraph::V(ig)),
    from = as.integer(igraph::tail_of(ig, igraph::E(ig))),
    to = as.integer(igraph::head_of(ig, igraph::E(ig)))
  )
  print.dm_graph(fallback, ...)
}

# graph_from_data_frame -------------------------------------------------------

graph_from_data_frame <- function(d, directed, vertices = NULL) {
  ig <- igraph::graph_from_data_frame(d, directed = directed, vertices = vertices)
  new_dm_igraph(ig)
}

# V: vertex accessor -----------------------------------------------------------
# Returns a named integer vector: values are 1-based indices, names are vertex names.

graph_vertices <- function(g) {
  igraph::V(g$igraph)
}

# E: edge accessor -------------------------------------------------------------
# Returns an integer vector with a "vnames" attribute (e.g. "from|to").

graph_edges <- function(g) {
  igraph::E(g$igraph)
}

# dfs -------------------------------------------------------------------------
# Returns a list with $order (named integer), $dist (named numeric), $parent (named integer).

graph_dfs <- function(g, root, unreachable = TRUE, parent = FALSE, dist = FALSE) {
  igraph::dfs(g$igraph, root, unreachable = unreachable, parent = parent, dist = dist)
}

# topo_sort -------------------------------------------------------------------
# Returns a named integer vector (vertex indices with vertex names as names).

graph_topo_sort <- function(g, mode = "out") {
  igraph::topo_sort(g$igraph, mode = mode)
}

# distances -------------------------------------------------------------------
# Returns a matrix: rows = sources, columns = all vertices.

graph_distances <- function(g, v = NULL) {
  igraph::distances(g$igraph, v)
}

# induced_subgraph ------------------------------------------------------------
# Returns a new dm_igraph containing only the specified vertices.

graph_induced_subgraph <- function(g, vids) {
  new_dm_igraph(igraph::induced_subgraph(g$igraph, vids))
}

# shortest_paths --------------------------------------------------------------
# Returns a list with $predecessors.

graph_shortest_paths <- function(g, from, to, predecessors = FALSE) {
  igraph::shortest_paths(g$igraph, from, to, predecessors = predecessors)
}

# delete_vertices -------------------------------------------------------------
# Returns a new dm_igraph with the specified vertices removed.

graph_delete_vertices <- function(g, v) {
  new_dm_igraph(igraph::delete_vertices(g$igraph, v))
}

# neighbors -------------------------------------------------------------------
# Returns the neighbors of vertex v.

graph_neighbors <- function(g, v, mode = "all") {
  igraph::neighbors(g$igraph, v, mode = mode)
}

# vcount ----------------------------------------------------------------------
# Returns the number of vertices in the graph.

graph_vcount <- function(g) {
  igraph::vcount(g$igraph)
}

# girth -----------------------------------------------------------------------
# Returns a list with $circle: vertices forming the shortest cycle.

graph_girth <- function(g) {
  igraph::girth(g$igraph)
}

test_that("functions working with graphs do the right thing?", {
  join_list_tbl_1 <- tibble(
    lhs = c("tf_2", "tf_3", "tf_4", "tf_5", "tf_6"),
    rhs = c("tf_1", "tf_2", "tf_3", "tf_4", "tf_5"),
    rank = as.numeric(2:6),
    has_father = rep(TRUE, 5)
  )

  join_list_tbl_3 <- tibble::tribble(
    ~lhs   , ~rhs   , ~rank , ~has_father ,
    "tf_2" , "tf_3" ,     2 , TRUE        ,
    "tf_4" , "tf_3" ,     3 , TRUE        ,
    "tf_1" , "tf_2" ,     4 , TRUE        ,
    "tf_5" , "tf_4" ,     5 , TRUE        ,
    "tf_6" , "tf_5" ,     6 , TRUE
  )

  expect_identical_graph(
    graph_from_data_frame(
      tibble(
        tables = c("tf_1", "tf_2", "tf_2", "tf_3", "tf_4", "tf_5", "tf_6"),
        ref_tables = c("tf_2", "tf_7", "tf_3", "tf_4", "tf_5", "tf_6", "tf_7")
      ),
      directed = FALSE
    ),
    create_graph_from_dm(dm_for_filter_w_cycle())
  )

  expect_snapshot({
    attr(graph_edges(create_graph_from_dm(nyc_comp())), "vnames")
  })
})

test_that("empty graph", {
  g0 <- create_graph_from_dm(empty_dm())
  g1 <- create_graph_from_dm(dm(x = tibble(a = 1)))
  expect_snapshot({
    print(g0)
    print(g1)
  })
})

# Tests for the pure R graph implementations in R/igraph.R.
# Tests use graph_from_data_frame() to obtain the correct graph type
# (igraph when installed, dm_graph otherwise), so they run correctly
# in both cases.

# Helper: simple directed graph a->b->c, b->d
make_graph_abc <- function(directed = TRUE) {
  graph_from_data_frame(
    tibble::tibble(
      from = c("a", "b", "b"),
      to = c("b", "c", "d")
    ),
    directed = directed,
    vertices = c("a", "b", "c", "d")
  )
}

# Helper: directed cycle a->b->c->a
make_graph_cycle <- function() {
  graph_from_data_frame(
    tibble::tibble(
      from = c("a", "b", "c"),
      to = c("b", "c", "a")
    ),
    directed = TRUE,
    vertices = c("a", "b", "c")
  )
}

# graph_from_data_frame ----------------------------------------------------

test_that("graph_from_data_frame: print output is consistent", {
  d <- tibble::tibble(from = c("a", "b"), to = c("b", "c"))
  g_directed <- graph_from_data_frame(d, directed = TRUE, vertices = c("a", "b", "c"))
  g_undirected <- graph_from_data_frame(d, directed = FALSE, vertices = c("a", "b", "c"))
  expect_snapshot({
    print(g_directed)
    print(g_undirected)
  })
})

test_that("graph_from_data_frame: non-empty data frame builds correct graph", {
  d <- tibble::tibble(from = c("a", "b"), to = c("b", "c"))
  g <- graph_from_data_frame(d, directed = TRUE, vertices = c("a", "b", "c"))
  expect_setequal(names(graph_vertices(g)), c("a", "b", "c"))
  expect_equal(length(graph_edges(g)), 2L)
  expect_setequal(attr(graph_edges(g), "vnames"), c("a|b", "b|c"))
})

test_that("graph_from_data_frame: zero-row data frame builds graph with no edges", {
  d <- tibble::tibble(from = character(0), to = character(0))
  g <- graph_from_data_frame(d, directed = FALSE, vertices = c("x", "y"))
  expect_setequal(names(graph_vertices(g)), c("x", "y"))
  expect_equal(length(graph_edges(g)), 0L)
  expect_snapshot(print(g))
})

test_that("graph_from_data_frame: NULL vertices derives vertex names from edges", {
  d <- tibble::tibble(from = c("a", "b"), to = c("b", "c"))
  g <- graph_from_data_frame(d, directed = FALSE)
  expect_setequal(names(graph_vertices(g)), c("a", "b", "c"))
  expect_equal(length(graph_edges(g)), 2L)
})

# graph_vertices ------------------------------------------------------------------------

test_that("graph_vertices returns vertices with correct names", {
  g <- make_graph_abc()
  expect_setequal(names(graph_vertices(g)), c("a", "b", "c", "d"))
})

test_that("graph_vertices on graph with no vertices returns empty result", {
  g <- graph_from_data_frame(
    tibble::tibble(from = character(0), to = character(0)),
    directed = FALSE,
    vertices = character(0)
  )
  expect_equal(length(graph_vertices(g)), 0L)
})

# graph_edges ------------------------------------------------------------------------

test_that("graph_edges returns edges with vnames attribute", {
  g <- make_graph_abc()
  e <- graph_edges(g)
  expect_equal(length(e), 3L)
  expect_setequal(attr(e, "vnames"), c("a|b", "b|c", "b|d"))
})

test_that("graph_edges on graph with no edges returns empty result", {
  g <- graph_from_data_frame(
    tibble::tibble(from = character(0), to = character(0)),
    directed = FALSE,
    vertices = c("x")
  )
  expect_equal(length(graph_edges(g)), 0L)
})

# graph_vcount -------------------------------------------------------------------

test_that("graph_vcount returns correct vertex count", {
  expect_equal(graph_vcount(make_graph_abc()), 4L)
})

test_that("graph_vcount returns 0 for empty graph", {
  g <- graph_from_data_frame(
    tibble::tibble(from = character(0), to = character(0)),
    directed = FALSE,
    vertices = character(0)
  )
  expect_equal(graph_vcount(g), 0L)
})

# graph_dfs ----------------------------------------------------------------------

test_that("graph_dfs visits all reachable vertices from root", {
  g <- make_graph_abc()
  dfs <- graph_dfs(g, "a", unreachable = FALSE, dist = TRUE)
  visited <- names(dfs$order)[!is.na(names(dfs$order))]
  expect_setequal(visited, c("a", "b", "c", "d"))
})

test_that("graph_dfs dist=TRUE gives 0 distance for root", {
  g <- make_graph_abc()
  dfs <- graph_dfs(g, "a", unreachable = FALSE, dist = TRUE)
  expect_equal(dfs$dist[["a"]], 0)
  expect_equal(dfs$dist[["b"]], 1)
})

test_that("graph_dfs parent=TRUE gives NA parent for root", {
  g <- make_graph_abc()
  dfs <- graph_dfs(g, "a", unreachable = FALSE, parent = TRUE)
  # parent structure has one entry per vertex; unclass converts to integer for both igraph and dm_graph
  parent_indices <- unclass(dfs$parent)
  a_idx <- which(names(graph_vertices(g)) == "a")
  expect_true(is.na(parent_indices[[a_idx]]))
})

# graph_topo_sort ----------------------------------------------------------------

test_that("graph_topo_sort mode='in' places parent-tables before children", {
  # FK graph: child -> parent
  g <- graph_from_data_frame(
    tibble::tibble(from = c("child"), to = c("parent")),
    directed = TRUE,
    vertices = c("parent", "child")
  )
  topo <- graph_topo_sort(g, mode = "in")
  topo_names <- names(topo)
  expect_lt(which(topo_names == "parent"), which(topo_names == "child"))
})

test_that("graph_topo_sort returns all vertices", {
  g <- make_graph_abc()
  topo <- graph_topo_sort(g, mode = "in")
  expect_setequal(names(topo), c("a", "b", "c", "d"))
})

test_that("graph_topo_sort mode='in' and mode='out' are reversed for linear chain", {
  g <- graph_from_data_frame(
    tibble::tibble(from = c("a", "b"), to = c("b", "c")),
    directed = TRUE,
    vertices = c("a", "b", "c")
  )
  topo_in <- names(graph_topo_sort(g, mode = "in"))
  topo_out <- names(graph_topo_sort(g, mode = "out"))
  expect_equal(topo_in, rev(topo_out))
})

# graph_distances ----------------------------------------------------------------

test_that("graph_distances returns 0 for self-distance", {
  g <- make_graph_abc(directed = FALSE)
  d <- graph_distances(g, "a")
  expect_equal(d[1, "a"], 0)
})

test_that("graph_distances returns 1 for adjacent vertex", {
  g <- make_graph_abc(directed = FALSE)
  d <- graph_distances(g, "a")
  expect_equal(d[1, "b"], 1)
})

test_that("graph_distances returns Inf for disconnected vertex", {
  # a-b connected, c disconnected
  g <- graph_from_data_frame(
    tibble::tibble(from = c("a"), to = c("b")),
    directed = FALSE,
    vertices = c("a", "b", "c")
  )
  d <- graph_distances(g, "a")
  expect_true(is.infinite(d[1, "c"]))
})

# graph_induced_subgraph ---------------------------------------------------------

test_that("graph_induced_subgraph keeps only specified vertices", {
  g <- make_graph_abc()
  sub <- graph_induced_subgraph(g, c("a", "b", "c"))
  expect_setequal(names(graph_vertices(sub)), c("a", "b", "c"))
})

test_that("graph_induced_subgraph removes edges to excluded vertices", {
  g <- make_graph_abc()
  # Keeping a, b, c (not d): edges a->b and b->c remain, b->d is removed
  sub <- graph_induced_subgraph(g, c("a", "b", "c"))
  expect_equal(length(graph_edges(sub)), 2L)
})

# graph_delete_vertices ----------------------------------------------------------

test_that("graph_delete_vertices removes specified vertex", {
  g <- make_graph_abc()
  g2 <- graph_delete_vertices(g, "b")
  expect_false("b" %in% names(graph_vertices(g2)))
})

test_that("graph_delete_vertices removes incident edges", {
  g <- make_graph_abc()
  g2 <- graph_delete_vertices(g, "b")
  # all edges involving b are removed (a->b, b->c, b->d)
  expect_equal(length(graph_edges(g2)), 0L)
})

# graph_neighbors ----------------------------------------------------------------

test_that("graph_neighbors mode='out' gives outgoing neighbors", {
  g <- make_graph_abc()
  nbrs <- graph_neighbors(g, "b", mode = "out")
  expect_setequal(names(nbrs), c("c", "d"))
})

test_that("graph_neighbors mode='in' gives incoming neighbors", {
  g <- make_graph_abc()
  nbrs <- graph_neighbors(g, "b", mode = "in")
  expect_setequal(names(nbrs), c("a"))
})

test_that("graph_neighbors mode='all' gives all neighbors", {
  g <- make_graph_abc()
  nbrs <- graph_neighbors(g, "b", mode = "all")
  expect_setequal(names(nbrs), c("a", "c", "d"))
})

# graph_girth --------------------------------------------------------------------

test_that("graph_girth returns Inf for acyclic graph", {
  g <- make_graph_abc()
  gi <- graph_girth(g)
  expect_true(is.infinite(gi$girth))
  expect_equal(length(gi$circle), 0L)
})

test_that("graph_girth detects directed cycle", {
  g <- make_graph_cycle()
  gi <- graph_girth(g)
  expect_true(is.finite(gi$girth))
  expect_gt(length(gi$circle), 0L)
  expect_true(all(names(gi$circle) %in% c("a", "b", "c")))
})

# graph_shortest_paths -----------------------------------------------------------

test_that("graph_shortest_paths predecessors has one entry per vertex", {
  g <- make_graph_abc(directed = FALSE)
  sp <- graph_shortest_paths(g, "a", names(graph_vertices(g)), predecessors = TRUE)
  # predecessors should have as many entries as the number of vertices
  expect_equal(length(sp$predecessors), graph_vcount(g))
})

test_that("graph_shortest_paths source vertex predecessor is NA", {
  g <- make_graph_abc(directed = FALSE)
  sp <- graph_shortest_paths(g, "a", names(graph_vertices(g)), predecessors = TRUE)
  # source vertex 'a' is the first vertex; its predecessor name should be NA
  a_idx <- which(names(graph_vertices(g)) == "a")
  expect_true(is.na(names(sp$predecessors)[[a_idx]]))
})

test_that("graph_shortest_paths predecessor name of adjacent vertex is source", {
  g <- make_graph_abc(directed = FALSE)
  sp <- graph_shortest_paths(g, "a", names(graph_vertices(g)), predecessors = TRUE)
  # vertex 'b' is adjacent to 'a', so predecessor of 'b' should be 'a'
  b_idx <- which(names(graph_vertices(g)) == "b")
  expect_equal(names(sp$predecessors)[[b_idx]], "a")
})

# Integration tests using dm objects ------------------------------------------

test_that("create_graph_from_dm produces graph with correct vertex count", {
  skip_if_not_installed("nycflights13")
  dm <- dm_nycflights13()
  g <- create_graph_from_dm(dm)
  expect_equal(length(graph_vertices(g)), length(dm))
})

test_that("graph_topo_sort via create_graph_from_dm: parent tables before children", {
  skip_if_not_installed("nycflights13")
  dm <- dm_nycflights13()
  g <- create_graph_from_dm(dm, directed = TRUE)
  topo <- graph_topo_sort(g, mode = "in")
  expect_lt(which(names(topo) == "airlines"), which(names(topo) == "flights"))
})

test_that("dm_filter works correctly with graph wrappers", {
  skip_if_not_installed("nycflights13")
  dm_filtered <- dm_nycflights13() %>%
    dm_filter(airlines = (carrier == "UA"))
  expect_lt(nrow(dm_filtered$airlines), nrow(dm_nycflights13()$airlines))
})

test_that("dm_flatten_to_tbl works correctly with graph wrappers", {
  skip_if_not_installed("nycflights13")
  flat <- dm_nycflights13() %>%
    dm_select_tbl(flights, airlines) %>%
    dm_flatten_to_tbl(.start = flights)
  expect_true(is.data.frame(flat))
  expect_true("name" %in% names(flat))
})

test_that("dm_wrap_tbl works correctly with graph wrappers", {
  skip_if_not_installed("nycflights13")
  wrapped <- dm_nycflights13() %>%
    dm_wrap_tbl(root = airlines)
  expect_equal(length(wrapped), 1L)
})

# Fallback vs igraph parity tests -----------------------------------------------
# Require igraph to compare results. Build a dm_graph directly to feed fallbacks.

# Helper: simple directed graph a->b->c, b->d
make_graph_abc <- function(directed = TRUE) {
  graph_from_data_frame(
    tibble::tibble(
      from = c("a", "b", "b"),
      to = c("b", "c", "d")
    ),
    directed = directed,
    vertices = c("a", "b", "c", "d")
  )
}

# Helper: directed cycle a->b->c->a
make_graph_cycle <- function() {
  graph_from_data_frame(
    tibble::tibble(
      from = c("a", "b", "c"),
      to = c("b", "c", "a")
    ),
    directed = TRUE,
    vertices = c("a", "b", "c")
  )
}

make_dm_graph_abc <- function(directed = TRUE) {
  new_dm_graph(
    directed = directed,
    vnames = c("a", "b", "c", "d"),
    from = c(1L, 2L, 2L),
    to = c(2L, 3L, 4L)
  )
}

make_dm_graph_cycle <- function() {
  new_dm_graph(
    directed = TRUE,
    vnames = c("a", "b", "c"),
    from = c(1L, 2L, 3L),
    to = c(2L, 3L, 1L)
  )
}

test_that("fallback graph_vertices matches igraph graph_vertices", {
  skip_if_not_installed("igraph")
  fb <- graph_vertices_fallback(make_dm_graph_abc())
  ig <- graph_vertices(make_graph_abc())
  expect_setequal(names(fb), names(ig))
})

test_that("fallback graph_edges matches igraph graph_edges", {
  skip_if_not_installed("igraph")
  fb <- graph_edges_fallback(make_dm_graph_abc())
  ig <- graph_edges(make_graph_abc())
  expect_equal(length(fb), length(ig))
  expect_setequal(attr(fb, "vnames"), attr(ig, "vnames"))
})

test_that("fallback graph_vcount matches igraph graph_vcount", {
  skip_if_not_installed("igraph")
  expect_equal(
    graph_vcount_fallback(make_dm_graph_abc()),
    graph_vcount(make_graph_abc())
  )
})

test_that("fallback graph_dfs matches igraph graph_dfs: visited vertices", {
  skip_if_not_installed("igraph")
  fb <- graph_dfs_fallback(make_dm_graph_abc(), "a", unreachable = FALSE, dist = TRUE)
  ig <- graph_dfs(make_graph_abc(), "a", unreachable = FALSE, dist = TRUE)
  fb_visited <- names(fb$order)[!is.na(names(fb$order))]
  ig_visited <- names(ig$order)[!is.na(names(ig$order))]
  expect_setequal(fb_visited, ig_visited)
  expect_equal(fb$dist[["a"]], ig$dist[["a"]])
  expect_equal(fb$dist[["b"]], ig$dist[["b"]])
})

test_that("fallback graph_topo_sort matches igraph graph_topo_sort: ordering constraint", {
  skip_if_not_installed("igraph")
  g_fb <- new_dm_graph(
    directed = TRUE,
    vnames = c("parent", "child"),
    from = 2L,
    to = 1L
  )
  g_ig <- graph_from_data_frame(
    tibble::tibble(from = "child", to = "parent"),
    directed = TRUE,
    vertices = c("parent", "child")
  )
  topo_fb <- names(graph_topo_sort_fallback(g_fb, mode = "in"))
  topo_ig <- names(graph_topo_sort(g_ig, mode = "in"))
  expect_lt(which(topo_fb == "parent"), which(topo_fb == "child"))
  expect_lt(which(topo_ig == "parent"), which(topo_ig == "child"))
})

test_that("fallback graph_distances matches igraph graph_distances", {
  skip_if_not_installed("igraph")
  g_fb <- make_dm_graph_abc(directed = FALSE)
  g_ig <- make_graph_abc(directed = FALSE)
  fb <- graph_distances_fallback(g_fb, "a")
  ig <- graph_distances(g_ig, "a")
  expect_equal(fb[1, "a"], ig[1, "a"])
  expect_equal(fb[1, "b"], ig[1, "b"])
})

test_that("fallback graph_induced_subgraph matches igraph: correct vertices and edges", {
  skip_if_not_installed("igraph")
  g_fb <- make_dm_graph_abc()
  g_ig <- make_graph_abc()
  sub_fb <- graph_induced_subgraph_fallback(g_fb, c("a", "b", "c"))
  sub_ig <- graph_induced_subgraph(g_ig, c("a", "b", "c"))
  expect_setequal(names(graph_vertices_fallback(sub_fb)), names(graph_vertices(sub_ig)))
  expect_equal(length(graph_edges_fallback(sub_fb)), length(graph_edges(sub_ig)))
})

test_that("fallback graph_delete_vertices matches igraph: correct vertices remain", {
  skip_if_not_installed("igraph")
  g_fb <- make_dm_graph_abc()
  g_ig <- make_graph_abc()
  r_fb <- graph_delete_vertices_fallback(g_fb, "b")
  r_ig <- graph_delete_vertices(g_ig, "b")
  expect_setequal(names(graph_vertices_fallback(r_fb)), names(graph_vertices(r_ig)))
  expect_equal(length(graph_edges_fallback(r_fb)), length(graph_edges(r_ig)))
})

test_that("fallback graph_neighbors matches igraph: mode='out'", {
  skip_if_not_installed("igraph")
  g_fb <- make_dm_graph_abc()
  g_ig <- make_graph_abc()
  fb <- graph_neighbors_fallback(g_fb, "b", mode = "out")
  ig <- graph_neighbors(g_ig, "b", mode = "out")
  expect_setequal(names(fb), names(ig))
})

test_that("fallback graph_neighbors matches igraph: mode='in'", {
  skip_if_not_installed("igraph")
  g_fb <- make_dm_graph_abc()
  g_ig <- make_graph_abc()
  fb <- graph_neighbors_fallback(g_fb, "b", mode = "in")
  ig <- graph_neighbors(g_ig, "b", mode = "in")
  expect_setequal(names(fb), names(ig))
})

test_that("fallback graph_girth: acyclic graph returns Inf", {
  skip_if_not_installed("igraph")
  fb <- graph_girth_fallback(make_dm_graph_abc())
  ig <- graph_girth(make_graph_abc())
  expect_true(is.infinite(fb$girth))
  expect_true(is.infinite(ig$girth))
  expect_equal(length(fb$circle), length(ig$circle))
})

test_that("fallback graph_girth: cycle detected matches igraph cycle length", {
  skip_if_not_installed("igraph")
  fb <- graph_girth_fallback(make_dm_graph_cycle())
  ig <- graph_girth(make_graph_cycle())
  expect_equal(fb$girth, ig$girth)
  expect_equal(length(fb$circle), length(ig$circle))
})

test_that("fallback graph_shortest_paths: predecessors count matches igraph", {
  skip_if_not_installed("igraph")
  g_fb <- make_dm_graph_abc(directed = FALSE)
  g_ig <- make_graph_abc(directed = FALSE)
  all_v <- c("a", "b", "c", "d")
  fb <- graph_shortest_paths_fallback(g_fb, "a", all_v, predecessors = TRUE)
  ig <- graph_shortest_paths(g_ig, "a", all_v, predecessors = TRUE)
  expect_equal(length(fb$predecessors), length(ig$predecessors))
  # source vertex predecessor is NA in both
  a_idx_fb <- which(names(graph_vertices_fallback(g_fb)) == "a")
  a_idx_ig <- which(names(graph_vertices(g_ig)) == "a")
  expect_true(is.na(names(fb$predecessors)[[a_idx_fb]]))
  expect_true(is.na(names(ig$predecessors)[[a_idx_ig]]))
})

#' Creates a dm object for the \pkg{nycflights13} data
#'
#' @description
#' Creates an example [`dm`] object from the tables in \pkg{nycflights13},
#' along with the references.
#' See [`nycflights13::flights`] for a description of the data.
#' As described in [`nycflights13::planes`], the relationship
#' between the `flights` table and the `planes` tables is "weak", it does not satisfy
#' data integrity constraints.
#'
#' @inheritParams rlang::args_dots_empty
#' @param cycle Boolean.
#'   If `FALSE` (default), only one foreign key relation
#'   (from `flights$origin` to `airports$faa`) between the `flights` table and the `airports` table is
#'   established.
#'   If `TRUE`, a `dm` object with a double reference
#'   between those tables will be produced.
#' @param color Boolean, if `TRUE` (default), the resulting `dm` object will have
#'   colors assigned to different tables for visualization with `dm_draw()`.
#' @param subset Boolean, if `TRUE` (default), the `flights` table is reduced to flights with column `day` equal to 10.
#' @param compound Boolean, if `FALSE`, no link will be established between tables `flights` and `weather`,
#'   because this requires compound keys.
#' @param table_description Boolean, if `TRUE`, a description will be added for each table that will be displayed
#'   when drawing the table with [dm_draw()].
#'
#' @return A `dm` object consisting of \pkg{nycflights13} tables, complete with primary and foreign keys and optionally colored.
#'
#' @export
#' @seealso `vignette("howto-dm-df")`
#' @examplesIf rlang::is_installed("DiagrammeR")
#' dm_nycflights13() %>%
#'   dm_draw()
#' @autoglobal
dm_nycflights13 <- function(
  ...,
  cycle = FALSE,
  color = TRUE,
  subset = TRUE,
  compound = TRUE,
  table_description = FALSE
) {
  check_dots_empty()

  if (subset) {
    data <- nycflights_subset()
    flights <- data$flights
    weather <- data$weather
    airlines <- data$airlines
    airports <- data$airports
    planes <- data$planes
  } else {
    check_suggested("nycflights13", "dm_nycflights13")

    flights <- nycflights13::flights
    weather <- nycflights13::weather
    airlines <- nycflights13::airlines
    airports <- nycflights13::airports
    planes <- nycflights13::planes
  }

  dm <-
    dm(airlines, airports, flights, planes, weather) %>%
    dm_add_pk(planes, tailnum) %>%
    dm_add_pk(airlines, carrier) %>%
    dm_add_pk(airports, faa) %>%
    dm_add_fk(flights, tailnum, planes) %>%
    dm_add_fk(flights, carrier, airlines) %>%
    dm_add_fk(flights, origin, airports)

  if (compound) {
    dm <-
      dm %>%
      dm_add_pk(weather, c(origin, time_hour)) %>%
      dm_add_fk(flights, c(origin, time_hour), weather)
  }

  if (color) {
    dm <-
      dm %>%
      dm_set_colors(
        "#5B9BD5" = flights,
        "#ED7D31" = c(starts_with("air"), planes),
        "#70AD47" = weather
      )
  }

  if (cycle) {
    dm <-
      dm %>%
      dm_add_fk(flights, dest, airports, check = FALSE)
  }

  if (table_description) {
    dm <-
      dm %>%
      dm_set_table_description(
        rlang::set_names(
          c("flights", "airports", "planes", "weather", "airlines"),
          c(
            paste(
              "On-time data for all flights that",
              "departed NYC (i.e. JFK, LGA or EWR) in 2013",
              sep = "\n"
            ),
            "Airports of origin or destination of the flights",
            "Planes used for the flights",
            "Hourly meteorological data for LGA, JFK and EWR in 2013",
            "Airlines that operated the flights"
          )
        )
      )
  }

  dm
}

nycflights_subset <- function() {
  readRDS(system.file("extdata/nycflights13-small.rds", package = "dm"))
}

withr::local_seed(20220706)

test_that("waldo", {
  skip_if_not_installed("waldo", "0.3.0")

  dm <- dm_nycflights13()

  expect_snapshot({
    dm %>%
      waldo::compare(dm, max_diffs = 10)
  })
  expect_snapshot({
    dm %>%
      dm_select_tbl(-airlines) %>%
      waldo::compare(dm, max_diffs = 10)
  })
  expect_snapshot({
    dm %>%
      dm_select(airlines, -name) %>%
      waldo::compare(dm, max_diffs = 10)
  })
  expect_snapshot({
    dm %>%
      dm_rm_fk() %>%
      waldo::compare(dm, max_diffs = 10)
  })
  expect_snapshot({
    dm %>%
      dm_rm_pk() %>%
      waldo::compare(dm, max_diffs = 10)
  })
  expect_snapshot({
    dm %>%
      dm_set_colors("yellow" = flights) %>%
      waldo::compare(dm, max_diffs = 10)
  })
  expect_snapshot({
    dm %>%
      dm_zoom_to(flights) %>%
      waldo::compare(dm, max_diffs = 10)
  })
})

test_that("snapshot test", {
  expect_snapshot({
    dm <- dm_nycflights13()
    dm_deconstruct(dm)
  })
})

test_that("non-syntactic names", {
  expect_snapshot({
    dm <- dm(`if` = tibble(a = 1), `a b` = tibble(b = 1))
    dm_deconstruct(dm)
  })
})

# ---
# repo: cynkra/dm
# file: standalone-check_suggested.R
# last-updated: 2023-02-23
# license: https://unlicense.org
# imports: rlang
# ---
#
# This file provides a wrapper around `rlang::check_installed()` that skips tests
# and supports optional usage.
#
# Needs functions from rlang, and purrr or standalone-purrr.R.
#
# ## Changelog
#
# 2023-10-19:
# * Initial

# nocov start

#' Check if suggested package is installed
#'
#' @param packages Vector of package names to check. Can supply a version
#'   between parenthesis. (See examples).
#' @param top_level_fun the name of the top level function called.
#' @param use whether to trigger the check, `NA` means `TRUE` if `is_interactive()`
#'   and `FALSE` otherwise
#' @return whether check was triggered and all packages are installed
#' @noRd
#' @examples
#' check_suggested(c("testthat (>= 3.2.0)", "xxx"), "foo")
check_suggested <- function(packages, top_level_fun, use = TRUE) {
  # If NA, inform that package isn't installed, but only in interactive mode
  only_msg <- is.na(use)
  if (only_msg) {
    use <- is_interactive()
  }

  if (!use) {
    return(FALSE)
  }

  # Check installation status if `use` was not `FALSE`
  installed <- map_lgl(packages, is_installed)

  if (all(installed)) {
    return(TRUE)
  }

  if (only_msg) {
    pkgs_not_installed <- packages[!installed]
    message <- "{.fn {top_level_fun}} is improved by the {.pkg {.val {pkgs_not_installed}}} package{?s}. Consider installing {?it/them}."
    cli::cli_inform(message)

    return(FALSE)
  }

  # Skip if some packages are not installed when testing
  # And say which package was not installed.
  if (identical(Sys.getenv("TESTTHAT"), "true")) {
    pkgs_not_installed <- packages[!installed]
    message <- cli::format_inline(
      "{.fn {top_level_fun}} needs the {.pkg {.val {pkgs_not_installed}}} package{?s}."
    )
    testthat::skip(message)
  }

  # If in interactive session, a prompt will ask user if they want
  # to install the package.
  # check_installed() uses pak for installation
  # if it's installed on the user system.

  # Which message to display in the prompt
  check_installed(packages, reason = glue("to use `{top_level_fun}()`."))

  # If check_installed() returns, all packages are installed
  TRUE
}

# nocov end

test_that("`check_suggested()` works", {
  expect_snapshot({
    check_suggested("rlang", TRUE, top_level_fun = "foo")

    check_suggested("dm", NA, top_level_fun = "foo")
    check_suggested("iurtnkjvmomweicopbt", NA, top_level_fun = "foo")
    check_suggested("iurtnkjvmomweicopbt", NA, message = "not installed!")

    check_suggested("dm", FALSE, top_level_fun = "foo")
    check_suggested("iurtnkjvmomweicopbt", FALSE, top_level_fun = "foo")
    check_suggested("iurtnkjvmomweicopbt", FALSE, message = "not installed!")
  })
})

test_that("`check_suggested()` works for error messages", {
  expect_snapshot(error = TRUE, {
    check_suggested("iurtnkjvmomweicopbt", TRUE, top_level_fun = "foo")
    check_suggested("iurtnkjvmomweicopbt", TRUE, message = "not installed!")
  })
})

test_that("`check_suggested()` works", {
  withr::local_envvar("TESTTHAT" = "")
  expect_snapshot({
    check_suggested("rlang", "foo")
    check_suggested("dm", "foo", use = NA)
    check_suggested("iurtnkjvmomweicopbt", "foo", use = NA)
    check_suggested("iurtnkjvmomweicopbt", use = NA)

    check_suggested("dm", "foo", use = FALSE)
    check_suggested("iurtnkjvmomweicopbt", "foo", use = FALSE)
    check_suggested("iurtnkjvmomweicopbt", use = FALSE)
  })
})

test_that("`check_suggested()` works for error messages", {
  withr::local_envvar("TESTTHAT" = "")
  expect_snapshot(error = TRUE, {
    check_suggested("iurtnkjvmomweicopbt", "foo")
  })
  rlang::local_interactive(TRUE)
  expect_snapshot({
    # A message
    check_suggested(c("iurtnkjvmomweicopbt (>= 0.5)", "xxx", "cli"), "foo", use = NA)
  })
})


test_that("`check_suggested()` will skip.", {
  # this should be skipped.
  check_suggested("iurtnkjvmomweicopbt", "foo")
  # Catching an error if it is not skipped.
  expect_equal("skipped", "not skipped")
})

#' Add a primary key
#'
#' @description
#' `dm_add_pk()` marks the specified columns as the primary key of the specified table.
#' If `check == TRUE`, then it will first check if
#' the given combination of columns is a unique key of the table.
#' If `force == TRUE`, the function will replace an already
#' set key, without altering foreign keys previously pointing to that primary key.
#'
#' @details There can be only one primary key per table in a [`dm`].
#' It's possible though to set an unlimited number of unique keys using [dm_add_uk()]
#' or adding foreign keys pointing to columns other than the primary key columns with [dm_add_fk()].
#'
#' @inheritParams rlang::args_dots_empty
#' @param dm A `dm` object.
#' @param table A table in the `dm`.
#' @param columns Table columns, unquoted.
#'   To define a compound key, use `c(col1, col2)`.
#' @param check Boolean, if `TRUE`, a check is made if the combination of columns is a unique key of the table.
#' @param force Boolean, if `FALSE` (default), an error will be thrown if there is already a primary key
#'   set for this table.
#'   If `TRUE`, a potential old `pk` is deleted before setting a new one.
#' @param autoincrement
#'   `r lifecycle::badge("experimental")`
#'   If `TRUE`, the  column specified in `columns` will be populated
#'   automatically with a sequence of integers.
#'
#' @family primary key functions
#'
#' @return An updated `dm` with an additional primary key.
#'
#' @examplesIf rlang::is_installed(c("nycflights13", "DiagrammeR"))
#' nycflights_dm <- dm(
#'   planes = nycflights13::planes,
#'   airports = nycflights13::airports,
#'   weather = nycflights13::weather
#' )
#'
#' nycflights_dm %>%
#'   dm_draw()
#'
#' # Create primary keys:
#' nycflights_dm %>%
#'   dm_add_pk(planes, tailnum) %>%
#'   dm_add_pk(airports, faa, check = TRUE) %>%
#'   dm_add_pk(weather, c(origin, time_hour)) %>%
#'   dm_draw()
#'
#' # Keys can be checked during creation:
#' try(
#'   nycflights_dm %>%
#'     dm_add_pk(planes, manufacturer, check = TRUE)
#' )
#' @export
dm_add_pk <- function(
  dm,
  table,
  columns,
  ...,
  autoincrement = FALSE,
  check = FALSE,
  force = FALSE
) {
  check_dots_empty()
  dm_local_error_call()

  check_not_zoomed(dm)

  table_name <- dm_tbl_name(dm, {{ table }})
  table <- dm_get_tables_impl(dm)[[table_name]]

  check_required(columns)
  col_expr <- enexpr(columns)
  col_name <- names(eval_select_indices(col_expr, colnames(table)))

  if (autoincrement && length(col_name) > 1L) {
    cli::cli_abort(c(
      "Composite primary keys cannot be autoincremented.",
      "i" = "Provide only a single column name to {.arg columns}."
    ))
  }

  if (check) {
    table_from_dm <- dm_get_filtered_table(dm, table_name)
    eval_tidy(expr(check_key(!!sym(table_name), !!col_expr)), list2(!!table_name := table_from_dm))
  }

  dm_add_pk_impl(dm, table_name, col_name, autoincrement, force)
}

# both "table" and "column" must be characters
# in {datamodelr}, a primary key may consist of more than one columns
# a key will be added, regardless of whether it is a unique key or not; not to be exported
dm_add_pk_impl <- function(dm, table, column, autoincrement, force) {
  def <- dm_get_def(dm)
  i <- which(def$table == table)

  if (!force && NROW(def$pks[[i]]) > 0) {
    if (
      !dm_is_strict_keys(dm) &&
        identical(def$pks[[i]]$column[[1]], column)
    ) {
      return(dm)
    }

    abort_key_set_force_false(table)
  }

  def$pks[[i]] <- new_pk(column = list(column), autoincrement = autoincrement)

  dm_from_def(def)
}

#' Check for primary key
#'
#' @description
#' `dm_has_pk()` checks if a given table has columns marked as its primary key.
#'
#' @inheritParams dm_add_pk
#'
#' @family primary key functions
#'
#' @return A logical value: `TRUE` if the given table has a primary key, `FALSE` otherwise.
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_has_pk(flights)
#' dm_nycflights13() %>%
#'   dm_has_pk(planes)
#' @export
dm_has_pk <- function(dm, table, ...) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})
  dm_has_pk_impl(dm, table_name)
}

dm_has_pk_impl <- function(dm, table) {
  has_length(dm_get_pk_impl(dm, table))
}

#' Primary key column names
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' This function is deprecated because of its limited use
#' and its unintuitive return value.
#' Use [dm_get_all_pks()] instead.
#'
#' @export
#' @keywords internal
dm_get_pk <- function(dm, table, ...) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)

  deprecate_warn("0.2.1", "dm::dm_get_pk()", "dm::dm_get_all_pks()")

  table_name <- dm_tbl_name(dm, {{ table }})
  new_keys(dm_get_pk_impl(dm, table_name))
}

dm_get_pk_impl <- function(dm, table_name) {
  # Optimized
  def <- dm_get_def(dm)
  pks <- def$pks[[which(def$table == table_name)]]
  pks$column
}

#' Get all primary keys of a [`dm`] object
#'
#' @description
#' `dm_get_all_pks()` checks the `dm` object for primary keys and
#' returns the tables and the respective primary key columns.
#'
#' @family primary key functions
#' @param table One or more table names, unquoted,
#'   to return primary key information for.
#'   If given, primary keys are returned in that order.
#'   The default `NULL` returns information for all tables.
#'
#' @inheritParams dm_add_pk
#'
#' @return A tibble with the following columns:
#'   \describe{
#'     \item{`table`}{table name,}
#'     \item{`pk_col`}{column name(s) of primary key, as list of character vectors.}
#'   }
#'
#' @export
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_get_all_pks()
dm_get_all_pks <- function(dm, table = NULL, ...) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)
  table_expr <- enexpr(table) %||% src_tbls_impl(dm, quiet = TRUE)
  table_names <- eval_select_table(table_expr, set_names(src_tbls_impl(dm, quiet = TRUE)))
  dm_get_all_pks_impl(dm, table_names)
}

dm_get_all_pks_impl <- function(dm, table = NULL) {
  dm %>%
    dm_get_def() %>%
    dm_get_all_pks_def_impl(table)
}

dm_get_all_pks_def_impl <- function(def, table = NULL) {
  # Optimized for speed

  def_sub <- def[c("table", "pks")]

  if (!is.null(table)) {
    idx <- match(table, def_sub$table)
    def_sub <- def_sub[match(table, def_sub$table), ]
  }

  out <-
    def_sub %>%
    unnest_df("pks", tibble(column = list(), autoincrement = logical())) %>%
    set_names(c("table", "pk_col", "autoincrement"))

  out$pk_col <- new_keys(out$pk_col)
  out
}


#' Remove a primary key
#'
#' @description
#' If a table name is provided, `dm_rm_pk()` removes the primary key from this table and leaves the [`dm`] object otherwise unaltered.
#' If no table is given, the `dm` is stripped of all primary keys at once.
#' An error is thrown if no primary key matches the selection criteria.
#' If the selection criteria are ambiguous, a message with unambiguous replacement code is shown.
#' Foreign keys are never removed.
#'
#' @inheritParams dm_add_pk
#' @param table A table in the `dm`.
#'   Pass `NULL` to remove all matching keys.
#' @param columns Table columns, unquoted.
#'   To refer to a compound key, use `c(col1, col2)`.
#'   Pass `NULL` (the default) to remove all matching keys.
#' @param fail_fk `r lifecycle::badge("deprecated")`
#'
#' @family primary key functions
#'
#' @return An updated `dm` without the indicated primary key(s).
#'
#' @export
#' @examplesIf rlang::is_installed(c("nycflights13", "DiagrammeR"))
#' dm_nycflights13() %>%
#'   dm_rm_pk(airports) %>%
#'   dm_draw()
dm_rm_pk <- function(dm, table = NULL, columns = NULL, ..., fail_fk = NULL) {
  dm_local_error_call()
  if (!is.null(fail_fk)) {
    lifecycle::deprecate_warn(
      "1.0.4",
      "dm_rm_pk(fail_fk =)",
      details = "When removing a primary key, potential associated foreign keys will be pointing at an implicit unique key."
    )
  }
  dm_rm_pk_(dm, {{ table }}, {{ columns }}, ...)
}

dm_rm_pk_ <- function(dm, table, columns, ..., rm_referencing_fks = NULL) {
  check_dots_empty()
  check_not_zoomed(dm)

  if (!is.null(rm_referencing_fks)) {
    deprecate_warn(
      "0.2.1",
      "dm::dm_rm_pk(rm_referencing_fks = )",
      details = "When removing a primary key, potential associated foreign keys will be pointing at an implicit unique key."
    )
  }

  table_name <- dm_tbl_name_null(dm, {{ table }})
  columns <- enexpr(columns)

  dm_rm_pk_impl(dm, table_name, columns)
}

dm_rm_pk_impl <- function(dm, table_name, columns, error_call = caller_env()) {
  def <- dm_get_def(dm)

  if (is.null(table_name)) {
    i <- which(map_int(def$pks, vec_size) > 0)
  } else {
    i <- which(def$table == table_name)
    if (nrow(def$pks[[i]]) == 0) {
      i <- integer()
    }
  }

  if (!quo_is_null(columns)) {
    ii <- map2_lgl(
      def$data[i],
      def$pks[i],
      ~ tryCatch(
        {
          vars <- eval_select_indices(columns, colnames(.x), error_call = error_call)
          identical(names(vars), .y$column[[1]])
        },
        error = function(e) {
          FALSE
        }
      )
    )

    i <- i[ii]
  }

  if (length(i) == 0 && dm_is_strict_keys(dm)) {
    abort_pk_not_defined()
  }

  # Talk about it
  if (is.null(table_name)) {
    message("Removing primary keys: %>%")
    message("  ", glue_collapse(glue("dm_rm_pk({tick_if_needed(def$table[i])})"), " %>%\n  "))
  }

  # Execute
  def$pks[i] <- list_of(new_pk())

  dm_from_def(def)
}


#' Primary key candidate
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `enum_pk_candidates()` checks for each column of a
#' table if the column contains only unique values, and is thus
#' a suitable candidate for a primary key of the table.
#'
#' @return A tibble with the following columns:
#'   \describe{
#'     \item{`columns`}{columns of `table`,}
#'     \item{`candidate`}{boolean: are these columns a candidate for a primary key,}
#'     \item{`why`}{if not a candidate for a primary key column, explanation for this.}
#'   }
#'
#' @section Life cycle:
#' These functions are marked "experimental" because we are not yet sure about
#' the interface, in particular if we need both `dm_enum...()` and `enum...()`
#' variants.
#' Changing the interface later seems harmless because these functions are
#' most likely used interactively.
#'
#' @rdname dm_enum_pk_candidates
#' @export
#' @examplesIf rlang::is_installed("nycflights13")
#' nycflights13::flights %>%
#'   enum_pk_candidates()
#' @autoglobal
enum_pk_candidates <- function(table, ...) {
  check_dots_empty()
  # a list of ayes and noes:
  if (is_dm(table) && is_zoomed(table)) {
    table <- tbl_zoomed(table)
  }

  table %>%
    enum_pk_candidates_impl() %>%
    rename(columns = column) %>%
    mutate(columns = new_keys(columns))
}

#' @description `dm_enum_pk_candidates()` performs these checks
#' for a table in a [dm] object.
#'
#' @family primary key functions
#'
#' @inheritParams dm_add_pk
#'
#' @export
#' @examplesIf rlang::is_installed("nycflights13")
#'
#' dm_nycflights13() %>%
#'   dm_enum_pk_candidates(airports)
dm_enum_pk_candidates <- function(dm, table, ...) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)
  # FIXME: with "direct" filter maybe no check necessary: but do we want to check
  # for tables retrieved with `tbl()` or with `dm_get_tables()[[table_name]]`
  check_no_filter(dm)

  table_name <- dm_tbl_name(dm, {{ table }})

  table <- dm_get_tables_impl(dm)[[table_name]]
  table %>%
    enum_pk_candidates_impl() %>%
    rename(columns = column) %>%
    mutate(columns = new_keys(columns))
}

#' @autoglobal
enum_pk_candidates_impl <- function(
  table,
  columns = new_keys(colnames(table)),
  max_value = MAX_COMMAS
) {
  tibble(column = new_keys(columns)) %>%
    mutate(why = map_chr(column, ~ check_pk(table, .x, max_value = max_value))) %>%
    mutate(candidate = (why == "")) %>%
    select(column, candidate, why) %>%
    arrange(desc(candidate), column)
}

check_pk <- function(table, columns, max_value = MAX_COMMAS) {
  duplicate_values <- is_unique_key_se(table, columns, max_value = max_value)
  if (duplicate_values$unique) {
    return("")
  }

  dup_data <- duplicate_values$data[[1]]

  fun <- ~ format(.x, trim = TRUE, justify = "none")

  values <- dup_data$value
  n <- dup_data$n
  values_na <- is.na(values)

  if (any(values_na)) {
    missing <- paste0(sum(n[values_na]), " missing values")
    values <- values[!values_na]
    n <- n[!values_na]
  } else {
    missing <- NULL
  }

  if (length(values) > 0) {
    values_count <- paste0(values, " (", n[!values_na], ")")
    values_text <- commas(values_count, max_commas = max_value, capped = TRUE, fun = fun)
    duplicate <- paste0("duplicate values: ", values_text)
  } else {
    duplicate <- NULL
  }

  problem <- glue_collapse(c(missing, duplicate), sep = "", last = ", and ")
  paste0("has ", problem)
}


# Error -------------------------------------------------------------------

abort_pk_not_defined <- function() {
  cli::cli_abort(
    "No primary keys to remove.",
    class = dm_error_full("pk_not_defined"),
    call = dm_error_call()
  )
}

abort_key_set_force_false <- function(table) {
  cli::cli_abort(
    "Table {.field {table}} already has a primary key. Use {.code force = TRUE} to change the existing primary key.",
    class = dm_error_full("key_set_force_false"),
    call = dm_error_call()
  )
}

abort_first_rm_fks <- function(table, fk_tables) {
  cli::cli_abort(
    "{cli::qty(length(fk_tables))}There {?is/are} {?a foreign key/foreign keys} pointing from table{?s} {.field {fk_tables}} to table {.field {table}}. First remove those, or set {.code fail_fk = FALSE}.",
    class = dm_error_full("first_rm_fks"),
    call = dm_error_call()
  )
}

test_that("'nest_join_dm_zoomed()'-method for `dm_zoomed` works", {
  skip("FIXME")
  skip_if_remote_src()
  expect_equivalent_tbl(
    dm_zoomed_2() %>%
      nest_join_dm_zoomed() %>%
      pull_tbl(),
    tf_3() %>%
      nest_join(tf_2(), name = "tf_2", by = c("f" = "e", "f1" = "e1")) %>%
      mutate(tf_2 = as_list_of(tf_2)) %>%
      nest_join(tf_4(), name = "tf_4", by = c("f" = "j")) %>%
      mutate(tf_4 = as_list_of(tf_4))
  )

  expect_equivalent_tbl(
    dm_zoomed_2() %>%
      nest_join_dm_zoomed(tf_4, tf_2) %>%
      pull_tbl(),
    tf_3() %>%
      nest_join(tf_4(), name = "tf_4", by = c("f" = "j")) %>%
      mutate(tf_4 = as_list_of(tf_4)) %>%
      nest_join(tf_2(), name = "tf_2", by = c("f" = "e", "f1" = "e1")) %>%
      mutate(tf_2 = as_list_of(tf_2))
  )

  expect_message(
    expect_equivalent_dm(
      dm_rm_pk(dm_for_filter(), tf_3, TRUE) %>% dm_zoom_to(tf_3) %>% nest_join_dm_zoomed(),
      dm_rm_pk(dm_for_filter(), tf_3, TRUE) %>% dm_zoom_to(tf_3)
    ),
    "didn't have a primary key"
  )

  expect_dm_error(
    dm_zoomed_2() %>% select(g) %>% nest_join_dm_zoomed(),
    "pk_not_tracked"
  )
})

test_that("'nest_join_dm_zoomed()' fails for DB-'dm'", {
  expect_dm_error(
    dm_zoom_to(dm_for_filter_duckdb(), tf_3) %>% nest_join_dm_zoomed(),
    "only_for_local_src"
  )
})

# tests for compound keys -------------------------------------------------

# FIXME: COMPOUND:: `nest_join_dm_zoomed()` cannot deal with composite keys yet
# verify_output(
#   "out/compound-nest.txt", {
#     dm_zoom_to(nyc_comp(), weather) %>%
#       nest_join_dm_zoomed()
#   }
# )

x <- tibble::tibble(
  x_id = integer(0L),
  x_data = character(0L),
)
y <- tibble::tibble(
  y_id = integer(0L),
  x_id = integer(0L),
  y_data = character(0L),
)

dm <- dm(x, y) %>%
  dm_add_pk(x, x_id, autoincrement = TRUE) %>%
  dm_add_pk(y, y_id, autoincrement = FALSE) %>%
  dm_add_fk(y, x_id, x)

test_that("autoincrement produces valid R code", {
  expect_snapshot(dm)
})

test_that("autoincrement produces valid SQL queries and R code - RSQLite", {
  skip_if_not_installed("RSQLite")

  con <- DBI::dbConnect(RSQLite::SQLite())
  df <- dm:::build_copy_queries(con, dm)

  expect_snapshot(df$sql_table)
  expect_snapshot(dm_paste(dm))
})

# NB. The following changes were made to the snapshotted files:
#
# 1. Calls to `rlang::quo()` have been suppressed, since these calls capture
#    an ephemeral environment that cannot be reproduced across separate test
#    runs.
#
# 2. `cg_block` invocations are excluded, since they belong to a separate and
#    independent testing unit.

test_that("snapshot of code_generation_middleware.R is unchanged", {
  expect_snapshot({
    parent <- tibble(a = c(1L, 1:3), b = -1)
    child <- tibble(a = 1:4, c = 3)

    dm <- dm(parent, child)

    # This is done by the Shiny app
    ops <- enum_ops(dm)
    ops

    enum_ops(dm, op_name = "dm_add_pk")
    enum_ops(dm, op_name = "dm_add_pk", table_names = "parent")

    enum_ops(dm, op_name = "dm_add_pk", table_names = "parent", column_names = "a")$call
    enum_ops(dm, op_name = "dm_add_pk", table_names = "parent", column_names = c("a", "b"))$call
  })
})

test_that("snapshot of code_generation_middleware_2.R is unchanged", {
  expect_snapshot({
    parent <- tibble(a = c(1L, 1:3), b = -1)
    child <- tibble(a = 1:4, c = 3)

    dm <- dm(parent, child)

    # This is done by the Shiny app
    ops <- enum_ops(dm)
    ops

    ops$multiple

    # Choice options (the user can choose arbitrarily many):
    user_choice <- list(
      table_names = "parent"
    )

    # Called by the Shiny app, always with the same pattern:
    ops_2 <- exec(
      # Always enum_ops
      enum_ops,
      # Returned by the previous call to enum_ops()
      !!!ops$input,
      # Input selected by the user
      !!!user_choice
    )
    # Same as:
    # ops_2 <- enum_ops(dm = ops$input$dm, table_names = table_names)

    ops_2

    # Choice options (the user can choose only one of those):
    ops_2$single

    # User's choice:
    user_choice_2 <- list(op_name = "dm_add_pk")

    # Called by the Shiny app, always with the same pattern:
    ops_3 <- exec(
      # Always enum_ops
      enum_ops,
      # Returned by the previous call to enum_ops()
      !!!ops_2$input,
      # Input selected by the user
      !!!user_choice_2
    )

    ops_3

    # Choice options (the user can choose arbitrarily many):
    ops_3$multiple

    # User's choice:
    user_choice_3 <- list(column_names = c("a"))

    # Called by the Shiny app, always with the same pattern:
    ops_4 <- exec(
      # Always enum_ops
      enum_ops,
      # Returned by the previous call to enum_ops()
      !!!ops_3$input,
      # Input selected by the user
      !!!user_choice_3
    )

    ops_4

    # Final result:
    ops_4$call
  })
})

test_that("snapshot of code_generation_middleware_3.R is unchanged", {
  expect_snapshot({
    parent <- tibble(a = c(1L, 1:3), b = -1)
    child <- tibble(a = 1:4, c = 3)

    dm <- dm(parent, child)

    # This is done by the Shiny app
    ops <- enum_ops(dm)
    ops

    ops$multiple

    # Choice options (the user can choose arbitrarily many):
    user_choice <- list(
      table_names = "parent"
    )

    # Called by the Shiny app, always with the same pattern:
    ops_2 <- exec(
      # Always enum_ops
      enum_ops,
      # Returned by the previous call to enum_ops()
      !!!ops$input,
      # Input selected by the user
      !!!user_choice
    )
    # Same as:
    # ops_2 <- enum_ops(dm = ops$input$dm, table_names = table_names)

    ops_2

    # Choice options (the user can choose arbitrarily many):
    ops_2$multiple

    # User's choice:
    user_choice_2 <- list(column_names = c("a"))

    # Called by the Shiny app, always with the same pattern:
    ops_3 <- exec(
      # Always enum_ops
      enum_ops,
      # Returned by the previous call to enum_ops()
      !!!ops_2$input,
      # Input selected by the user
      !!!user_choice_2
    )

    ops_3

    # Choice options (the user can choose only one of those):
    ops_3$single

    # User's choice:
    user_choice_3 <- list(op_name = "dm_add_pk")

    # Called by the Shiny app, always with the same pattern:
    ops_4 <- exec(
      # Always enum_ops
      enum_ops,
      # Returned by the previous call to enum_ops()
      !!!ops_3$input,
      # Input selected by the user
      !!!user_choice_3
    )

    ops_4

    # Final result:
    ops_4$call
  })
})

test_that("snapshot of code_generation_middleware_4.R is unchanged", {
  expect_snapshot({
    parent <- tibble(a = c(1L, 1:3), b = -1)
    child <- tibble(a = 1:4, c = 3)

    dm <-
      dm(parent, child) %>%
      dm_add_pk(parent, b)

    # This is done by the Shiny app
    ops <- enum_ops(dm)
    ops

    ops$multiple

    # Choice options (the user can choose arbitrarily many):
    user_choice <- list(
      table_names = "parent"
    )

    # Called by the Shiny app, always with the same pattern:
    ops_2 <- exec(
      # Always enum_ops
      enum_ops,
      # Returned by the previous call to enum_ops()
      !!!ops$input,
      # Input selected by the user
      !!!user_choice
    )
    # Same as:
    # ops_2 <- enum_ops(dm = ops$input$dm, table_names = table_names)

    ops_2

    # Choice options (the user can choose arbitrarily many):
    ops_2$multiple

    # User's choice:
    user_choice_2 <- list(column_names = c("a"))

    # Called by the Shiny app, always with the same pattern:
    ops_3 <- exec(
      # Always enum_ops
      enum_ops,
      # Returned by the previous call to enum_ops()
      !!!ops_2$input,
      # Input selected by the user
      !!!user_choice_2
    )

    ops_3

    # Choice options (the user can choose only one of those):
    ops_3$single

    # User's choice:
    user_choice_3 <- list(op_name = "dm_add_pk")

    # Called by the Shiny app, always with the same pattern:
    ops_4 <- exec(
      # Always enum_ops
      enum_ops,
      # Returned by the previous call to enum_ops()
      !!!ops_3$input,
      # Input selected by the user
      !!!user_choice_3
    )

    ops_4

    # Final result:
    ops_4$call
  })
})

# End-to-end snapshot tests for error ownership
# Each test triggers an error through exported functions to track that the
# error is reported with the correct origin (the exported function name).

test_that("dm_add_pk() - abort_key_set_force_false", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_add_pk(a, x)
  expect_snapshot(error = TRUE, {
    dm_add_pk(d, a, x)
  })
})

test_that("dm_add_pk() - abort_not_unique_key (via check)", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = c(1, 1)))
  expect_snapshot(error = TRUE, {
    dm_add_pk(d, a, x, check = TRUE)
  })
})

test_that("dm_rm_pk() - abort_pk_not_defined", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_rm_pk(d, a)
  })
})

test_that("dm_add_fk() - abort_ref_tbl_has_no_pk", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_add_fk(d, a, x, ref_table = b)
  })
})

test_that("dm_add_fk() - abort_fk_exists", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>%
    dm_add_pk(b, x) %>%
    dm_add_fk(a, x, ref_table = b)
  expect_snapshot(error = TRUE, {
    dm_add_fk(d, a, x, ref_table = b)
  })
})

test_that("dm_add_fk() - abort_not_subset_of (via check)", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 2), b = tibble(x = 1)) %>% dm_add_pk(b, x)
  expect_snapshot(error = TRUE, {
    dm_add_fk(d, a, x, ref_table = b, check = TRUE)
  })
})

test_that("dm_rm_fk() - abort_is_not_fkc", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_rm_fk(d, a, x, ref_table = b)
  })
})

test_that("dm_add_uk() - abort_no_uk_if_pk (PK exists)", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_add_pk(a, x)
  expect_snapshot(error = TRUE, {
    dm_add_uk(d, a, x)
  })
})

test_that("dm_add_uk() - abort_no_uk_if_pk (UK exists)", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_add_uk(a, x)
  expect_snapshot(error = TRUE, {
    dm_add_uk(d, a, x)
  })
})

test_that("dm_rm_uk() - abort_uk_not_defined", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_rm_uk(d, a)
  })
})

test_that("check_key() - abort_not_unique_key", {
  local_options(lifecycle_verbosity = "quiet")

  t <- tibble(x = c(1, 1))
  expect_snapshot(error = TRUE, {
    check_key(t, x)
  })
})

test_that("check_cardinality_1_1() - abort_not_bijective", {
  local_options(lifecycle_verbosity = "quiet")

  parent <- tibble(x = 1:3)
  child <- tibble(x = c(1, 1, 2, 3))
  expect_snapshot(error = TRUE, {
    check_cardinality_1_1(parent, child, by_position = TRUE)
  })
})

test_that("check_cardinality_0_1() - abort_not_injective", {
  local_options(lifecycle_verbosity = "quiet")

  parent <- tibble(x = 1:4)
  child <- tibble(x = c(1, 1, 2))
  expect_snapshot(error = TRUE, {
    check_cardinality_0_1(parent, child, by_position = TRUE)
  })
})

test_that("dm_set_colors() - abort_is_not_dm", {
  local_options(lifecycle_verbosity = "quiet")

  expect_snapshot(error = TRUE, {
    dm_set_colors("not_a_dm", blue = a)
  })
})

test_that("dm_get_colors() - abort_is_not_dm", {
  local_options(lifecycle_verbosity = "quiet")

  expect_snapshot(error = TRUE, {
    dm_get_colors("not_a_dm")
  })
})

test_that("dm_discard_zoomed() - abort_is_not_dm", {
  local_options(lifecycle_verbosity = "quiet")

  expect_snapshot(error = TRUE, {
    dm_discard_zoomed("not_a_dm")
  })
})

test_that("dm_wrap_tbl() - abort_is_not_dm", {
  local_options(lifecycle_verbosity = "quiet")

  expect_snapshot(error = TRUE, {
    dm_wrap_tbl("not_a_dm", root = a)
  })
})

test_that("dm_set_colors() - abort_only_named_args", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_set_colors(d, a)
  })
})

test_that("dm_set_colors() - abort_wrong_syntax_set_cols", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_set_colors(d, a = "blue")
  })
})

test_that("dm_set_colors() - abort_cols_not_avail", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_set_colors(d, nonexistent_color_xyz = a)
  })
})

test_that("dm_get_con() - abort_con_only_for_dbi", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_get_con(d)
  })
})

test_that("$<-.dm - abort_update_not_supported", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    d$a <- tibble(x = 2)
  })
})

test_that("[<-.dm - abort_update_not_supported", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    d[["a"]] <- tibble(x = 2)
  })
})

test_that("dm_validate() - abort_is_not_dm", {
  local_options(lifecycle_verbosity = "quiet")

  expect_snapshot(error = TRUE, {
    dm_validate("not_a_dm")
  })
})

test_that("dm_validate() - abort_dm_invalid", {
  local_options(lifecycle_verbosity = "quiet")

  bad <- structure(list(bad = "dm"), class = "dm")
  expect_snapshot(error = TRUE, {
    dm_validate(bad)
  })
})

test_that("dm_zoom_to() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_zoom_to(d, b)
  })
})

test_that("dm_update_zoomed() on unzoomed dm - abort_only_possible_w_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_update_zoomed(d)
  })
})

test_that("dm_flatten_to_tbl() with recursive and unsupported join - abort_squash_limited", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>%
    dm_add_pk(b, x) %>%
    dm_add_fk(a, x, ref_table = b)
  expect_snapshot(error = TRUE, {
    dm_flatten_to_tbl(d, a, .recursive = TRUE, .join = right_join)
  })
})

test_that("dm_flatten_to_tbl() with nest_join - abort_no_flatten_with_nest_join", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>%
    dm_add_pk(b, x) %>%
    dm_add_fk(a, x, ref_table = b)
  expect_snapshot(error = TRUE, {
    dm_flatten_to_tbl(d, a, .join = nest_join)
  })
})

test_that("dm_rename_tbl() - abort_need_unique_names", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_rename_tbl(d, a = b)
  })
})

test_that("decompose_table() - abort_dupl_new_id_col_name", {
  local_options(lifecycle_verbosity = "quiet")

  t <- tibble(x = 1, y = 2)
  expect_snapshot(error = TRUE, {
    decompose_table(t, x, y)
  })
})

test_that("dm_paste() - abort_unknown_option", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_paste(d, options = "nonexistent_option")
  })
})

test_that("check_set_equality() - abort_sets_not_equal", {
  local_options(lifecycle_verbosity = "quiet")

  parent <- tibble(x = 1:3)
  child <- tibble(x = c(1, 2))
  expect_snapshot(error = TRUE, {
    check_set_equality(parent, child, by_position = TRUE)
  })
})

test_that("dm_draw() - unsupported backend_opts", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_draw(d, backend_opts = list(unsupported_option = TRUE))
  })
})

test_that("copy_to.dm() - abort_only_data_frames_supported", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    copy_to(d, 42, name = "b")
  })
})

test_that("copy_to.dm() - abort_no_overwrite", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    copy_to(d, tibble(y = 1), name = "b", overwrite = TRUE)
  })
})

test_that("copy_to.dm() - abort_one_name_for_copy_to", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    copy_to(d, tibble(y = 1), name = c("b", "c"))
  })
})

test_that("pull_tbl() on dm with table not in dm - abort_table_not_in_dm", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    pull_tbl(d, nonexistent)
  })
})

test_that("pull_tbl() on zoomed dm with wrong table - abort_table_not_zoomed", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    pull_tbl(d, b)
  })
})

test_that("dm_flatten_to_tbl() with unrelated tables - abort_tables_not_reachable_from_start", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1), c = tibble(x = 1)) %>%
    dm_add_pk(b, x) %>%
    dm_add_fk(a, x, ref_table = b)
  expect_snapshot(error = TRUE, {
    dm_flatten_to_tbl(d, a, c)
  })
})

test_that("dm_flatten_to_tbl() with grandparent - abort_only_parents", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1), c = tibble(x = 1)) %>%
    dm_add_pk(b, x) %>%
    dm_add_pk(c, x) %>%
    dm_add_fk(a, x, ref_table = b) %>%
    dm_add_fk(b, x, ref_table = c)
  expect_snapshot(error = TRUE, {
    dm_flatten_to_tbl(d, a, b, c)
  })
})

test_that("dm_flatten_to_tbl() with cycle - abort_no_cycles", {
  local_options(lifecycle_verbosity = "quiet")

  expect_snapshot(error = TRUE, {
    dm_flatten_to_tbl(dm_for_filter_w_cycle(), tf_5, .recursive = TRUE)
  })
})

test_that("dm_select_tbl() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_select_tbl(d, a)
  })
})

test_that("dm_add_pk() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_add_pk(d, b, x)
  })
})

test_that("dm_add_fk() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_add_fk(d, a, x, ref_table = b)
  })
})

test_that("dm_get_con() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_get_con(d)
  })
})

test_that("dm_insert_zoomed() on unzoomed dm - abort_only_possible_w_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    dm_insert_zoomed(d)
  })
})

test_that("dm_draw() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_draw(d)
  })
})

test_that("dm_paste() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_paste(d)
  })
})

# --- Additional error conditions ---

test_that("dm_flatten() - abort_tables_not_neighbors", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1), c = tibble(x = 1)) %>%
    dm_add_pk(b, x) %>%
    dm_add_fk(a, x, ref_table = b)
  expect_snapshot(error = TRUE, {
    dm_flatten(d, a, parent_tables = c)
  })
})

test_that("pull_tbl() - abort_no_table_provided", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_snapshot(error = TRUE, {
    pull_tbl(d, )
  })
})

test_that("dm_flatten_to_tbl() - abort_no_flatten_with_nest_join (via dm_join_to_tbl)", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>%
    dm_add_pk(b, x) %>%
    dm_add_fk(a, x, ref_table = b)
  expect_snapshot(error = TRUE, {
    dm_join_to_tbl(d, a, b, join = nest_join)
  })
})

test_that("dm_examine_constraints() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_examine_constraints(d)
  })
})

test_that("dm_disambiguate_cols() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_disambiguate_cols(d)
  })
})

test_that("dm_rm_pk() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>%
    dm_add_pk(a, x) %>%
    dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_rm_pk(d, a)
  })
})

test_that("dm_rm_fk() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>%
    dm_add_pk(b, x) %>%
    dm_add_fk(a, x, ref_table = b) %>%
    dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_rm_fk(d, a, x, ref_table = b)
  })
})

test_that("dm_add_uk() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_add_uk(d, a, x)
  })
})

test_that("dm_rm_uk() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>%
    dm_add_uk(a, x) %>%
    dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_rm_uk(d, a)
  })
})

test_that("dm_has_pk() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_has_pk(d, a)
  })
})

test_that("dm_get_pk() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_get_pk(d, a)
  })
})

test_that("dm_get_all_pks() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_get_all_pks(d)
  })
})

test_that("dm_get_all_fks() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_get_all_fks(d)
  })
})

test_that("dm_get_all_uks() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_get_all_uks(d)
  })
})

test_that("dm_enum_pk_candidates() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_enum_pk_candidates(d, a)
  })
})

test_that("dm_enum_fk_candidates() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1), b = tibble(x = 1)) %>%
    dm_add_pk(b, x) %>%
    dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_enum_fk_candidates(d, a, b)
  })
})

test_that("dm_rename() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_rename(d, a, y = x)
  })
})

test_that("dm_select() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_select(d, a, x)
  })
})

test_that("dm_discard_zoomed() on unzoomed dm returns silently", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1))
  expect_identical(dm_discard_zoomed(d), d)
})

test_that("dm_set_table_description() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_set_table_description(d, a = "test")
  })
})

test_that("dm_get_tables() on zoomed dm - abort_only_possible_wo_zoom", {
  local_options(lifecycle_verbosity = "quiet")

  d <- dm(a = tibble(x = 1)) %>% dm_zoom_to(a)
  expect_snapshot(error = TRUE, {
    dm_get_tables(d)
  })
})

test_that("check_cardinality_0_n() - abort_not_subset_of", {
  local_options(lifecycle_verbosity = "quiet")

  parent <- tibble(x = 1:2)
  child <- tibble(x = c(1, 3))
  expect_snapshot(error = TRUE, {
    check_cardinality_0_n(parent, child, by_position = TRUE)
  })
})

test_that("check_cardinality_1_n() - abort_not_unique_key and not_subset_of", {
  local_options(lifecycle_verbosity = "quiet")

  parent <- tibble(x = c(1, 1))
  child <- tibble(x = c(1, 2))
  expect_snapshot(error = TRUE, {
    check_cardinality_1_n(parent, child, by_position = TRUE)
  })
})

test_that("check_subset() - abort_not_subset_of", {
  local_options(lifecycle_verbosity = "quiet")

  parent <- tibble(x = 1:2)
  child <- tibble(x = c(1, 3))
  expect_snapshot(error = TRUE, {
    check_subset(child, parent, by_position = TRUE)
  })
})

test_that("dm_add_pk() works as intended?", {
  expect_silent(
    dm_test_obj() %>%
      dm_add_pk(dm_table_1, a)
  )

  expect_silent(
    dm_test_obj() %>%
      dm_add_pk(dm_table_2, c)
  )

  expect_silent(expect_equivalent_dm(
    dm_test_obj() %>%
      dm_add_pk(dm_table_1, a) %>%
      dm_add_pk(dm_table_1, b, force = TRUE),
    dm_test_obj() %>%
      dm_add_pk(dm_table_1, b)
  ))

  expect_error(
    dm_add_pk(dm_test_obj(), dm_table_1, qq),
    class = "vctrs_error_subscript_oob"
  )

  expect_snapshot(error = TRUE, {
    dm(x = tibble(x = integer())) %>%
      dm_add_pk(x)
  })

  expect_dm_error(
    dm_test_obj() %>%
      dm_add_pk(dm_table_1, a) %>%
      dm_add_pk(dm_table_1, b),
    class = "key_set_force_false"
  )
  expect_dm_error(
    dm_test_obj() %>%
      dm_add_pk(dm_table_1, a) %>%
      dm_add_pk(dm_table_1, a),
    class = "key_set_force_false"
  )
  expect_dm_error(
    dm_add_pk(dm_test_obj(), dm_table_2, c, check = TRUE),
    class = "not_unique_key"
  )
})

test_that("dm_rm_pk() works as intended?", {
  expect_silent(
    dm_test_obj() %>%
      dm_add_pk(dm_table_1, a) %>%
      dm_rm_pk(dm_table_1)
  )
  expect_dm_error(
    dm_test_obj() %>%
      dm_rm_pk(dm_table_1),
    class = "pk_not_defined"
  )
  expect_dm_error(
    dm_test_obj() %>%
      dm_add_pk(dm_table_1, a) %>%
      dm_rm_pk(dm_table_bogus),
    class = "table_not_in_dm"
  )

  expect_deprecated(
    expect_equivalent_dm(
      dm_for_filter() %>%
        dm_rm_pk(tf_4, fail_fk = TRUE),
      dm_for_filter() %>%
        dm_get_def() %>%
        mutate(pks = if_else(table == "tf_4", list_of(new_pk()), pks)) %>%
        dm_from_def()
    ),
    "fail_fk"
  )

  # test if error is thrown if col not found
  expect_dm_error(
    dm_for_filter() %>%
      dm_rm_pk(tf_5, x),
    "pk_not_defined"
  )

  # test if error is thrown if any col not found
  expect_dm_error(
    dm_for_filter() %>%
      dm_rm_pk(columns = x),
    "pk_not_defined"
  )
})

test_that("dm_rm_pk() supports partial filters", {
  expect_snapshot({
    # test logic if argument `fail_fk = FALSE`
    dm_for_filter() %>%
      dm_rm_pk(tf_4) %>%
      get_all_keys()

    dm_for_filter() %>%
      dm_rm_pk(tf_3) %>%
      get_all_keys()

    # no failure if pk not used in relationship
    dm_for_filter() %>%
      dm_rm_pk(tf_6) %>%
      get_all_keys()

    # dprecated argument name
    dm_for_filter() %>%
      dm_rm_pk(tf_4, rm_referencing_fks = TRUE) %>%
      get_all_keys()

    # partial match for columns
    dm_for_filter() %>%
      dm_rm_pk(columns = c) %>%
      get_all_keys()

    # partial match for columns, with compound key
    dm_for_filter() %>%
      dm_rm_pk(columns = c(f, f1)) %>%
      get_all_keys()

    # partial match for all tables
    dm_for_filter() %>%
      dm_rm_pk(fail_fk = FALSE) %>%
      get_all_keys()
  })
})

test_that("dm_has_pk() works as intended?", {
  expect_false(
    dm_has_pk(dm_test_obj(), dm_table_2)
  )
  expect_true(
    dm_test_obj() %>%
      dm_add_pk(dm_table_1, a) %>%
      dm_has_pk(dm_table_1)
  )
})

test_that("dm_get_pk() works as intended?", {
  local_options(lifecycle_verbosity = "quiet")

  expect_identical(
    dm_get_pk(dm_test_obj(), dm_table_1),
    new_keys(character(0))
  )
  expect_identical(
    dm_test_obj() %>%
      dm_add_pk(dm_table_1, a) %>%
      dm_get_pk(dm_table_1),
    new_keys("a")
  )
})

test_that("dm_enum_pk_candidates() works properly?", {
  skip_if_ide()

  expect_snapshot({
    dm_enum_pk_candidates(dm_test_obj(), dm_table_1)
    dm_enum_pk_candidates(dm_test_obj(), dm_table_2)
    dm_enum_pk_candidates(dm_test_obj(), dm_table_5)
    dm_enum_pk_candidates(dm_test_obj(), dm_table_6)
  })
})

test_that("enum_pk_candidates() works properly", {
  expect_silent(
    expect_identical(
      enum_pk_candidates(dm_zoomed()),
      enum_pk_candidates(tf_2())
    )
  )
})

test_that("output", {
  expect_snapshot(error = TRUE, {
    dm(x = tibble(a = c(1, 1))) %>%
      dm_add_pk(x, a, check = TRUE)
  })
})


# all primary keys --------------------------------------------------------------------

test_that("dm_get_all_pks() and order", {
  dm <- nyc_comp()
  pks_all <- dm_get_all_pks(dm)
  pks_12 <- dm_get_all_pks(dm, !!names(dm)[1:2])
  pks_21 <- dm_get_all_pks(dm, !!names(dm)[2:1])

  expect_equal(pks_all[1:2, ], pks_12)
  expect_equal(pks_all[2:1, ], pks_21)
})

test_that("dm_get_all_pks() with table arg", {
  expect_snapshot({
    nyc_comp() %>%
      dm_get_all_pks("weather")

    nyc_comp() %>%
      dm_get_all_pks(c(airlines, weather))

    # test tidyselect functions for parent_table arg
    nyc_comp() %>%
      dm_get_all_pks(starts_with("weat"))

    nyc_comp() %>%
      dm_get_all_pks(matches("^a.*s$"))
  })
})

test_that("dm_get_all_pks() with table arg fails nicely", {
  skip_if(packageVersion("tidyselect") > "1.2.0")

  expect_snapshot_error({
    nyc_comp() %>%
      dm_get_all_pks(c(airlines, weather, timetable, tabletime))
  })
})

# tests for compound keys -------------------------------------------------

test_that("dm_get_all_pks() with compound keys", {
  # FIXME: COMPOUND: Need proper test
  skip_if_remote_src()

  expect_snapshot({
    nyc_comp()

    nyc_comp() %>%
      dm_get_all_pks()
  })
})

# autoincrement -------------------------------------------------

x <- tibble(
  x_id = integer(0),
  z = integer(0),
  x_data = character(0),
)

test_that("autoincrement fails with compound keys", {
  expect_snapshot(error = TRUE, {
    dm(x) %>%
      dm_add_pk(x, columns = c(x_id, z), autoincrement = TRUE)
  })
})

test_that("set autoincrement PK", {
  expect_snapshot({
    dm(x, y = x) %>%
      dm_add_pk(x, columns = c(x_id), autoincrement = TRUE) %>%
      dm_add_pk(y, columns = c(x_id, z)) %>%
      dm_get_all_pks()
  })
})

new_ticker <- function(label, n, progress = NA, top_level_fun = NULL) {
  stopifnot(is_bare_string(label))
  stopifnot(is_bare_integerish(n, 1))

  suggested <- check_suggested("progress", top_level_fun, use = progress)
  if (!suggested) {
    return(identity)
  }

  # pb to be updated by reference by output function
  pb <- progress::progress_bar$new(
    format = sprintf(
      "  %s [:bar] :percent in :elapsed",
      label
    ),
    total = n,
    clear = FALSE
  )
  # output a function that curries f to tick, updating pb
  function(f) {
    f <- purrr::as_mapper(f)
    function(...) {
      pb$tick()
      f(...)
    }
  }
}

#' Copy data model to data source
#'
#' @description
#' `copy_dm_to()` takes a [`dbplyr::src_dbi`] object or a [`DBI::DBIConnection-class`] object as its first argument
#' and a [`dm`] object as its second argument.
#' The latter is copied to the former.
#' The default is to create temporary tables, set `temporary = FALSE` to create permanent tables.
#' Unless `set_key_constraints` is `FALSE`, primary key, foreign key, and unique constraints
#' are set, and indexes for foreign keys are created, on all databases.
#'
#' @inheritParams dm_examine_constraints
#'
#' @param dest An object of class `"src"` or `"DBIConnection"`.
#' @param dm A `dm` object.
#' @inheritParams rlang::args_dots_empty
#' @param set_key_constraints If `TRUE` will mirror
#'   the primary, foreign, and unique key constraints
#'   and create indexes for foreign key constraints
#'   for the primary and foreign keys in the `dm` object.
#'   Set to `FALSE` if your data model currently does not satisfy
#'   primary or foreign key constraints.
#' @param temporary If `TRUE`, only temporary tables will be created.
#'   These tables will vanish when disconnecting from the database.
#' @param schema Name of schema to copy the `dm` to.
#'   If `schema` is provided, an error will be thrown if `temporary = FALSE` or
#'   `table_names` is not `NULL`.
#'
#'   Not all DBMS are supported.
#' @param table_names Desired names for the tables on `dest`; the names within the `dm` remain unchanged.
#'   Can be `NULL`, a named character vector, or a vector of [`DBI::Id`] objects.
#'
#'   If left `NULL` (default), the names will be determined automatically depending on the `temporary` argument:
#'
#'   1. `temporary = TRUE` (default): unique table names based on the names of the tables in the `dm` are created.
#'   1. `temporary = FALSE`: the table names in the `dm` are used as names for the tables on `dest`.
#'
#'   If a function or one-sided formula, `table_names` is converted to a function
#'   using [rlang::as_function()].
#'   This function is called with the unquoted table names of the `dm` object
#'   as the only argument.
#'   The output of this function is processed by [DBI::dbQuoteIdentifier()],
#'   that result should be a vector of identifiers of the same length
#'   as the original table names.
#'
#'   Use a variant of
#'   `table_names = ~ DBI::SQL(paste0("schema_name", ".", .x))`
#'   to specify the same schema for all tables.
#'   Use `table_names = identity` with `temporary = TRUE`
#'   to avoid giving temporary tables unique names.
#'
#'   If a named character vector,
#'   the names of this vector need to correspond to the table names in the `dm`,
#'   and its values are the desired names on `dest`.
#'   The value is processed by [DBI::dbQuoteIdentifier()],
#'   that result should be a vector of identifiers of the same length
#'   as the original table names.
#'
#'   Use qualified names corresponding to your database's syntax
#'   to specify e.g. database and schema for your tables.
#' @param unique_table_names,copy_to Must be `NULL`.
#'
#' @family DB interaction functions
#'
#' @return A `dm` object on the given `src` with the same table names
#'   as the input `dm`.
#'
#' @examplesIf rlang::is_installed(c("RSQLite", "nycflights13", "dbplyr"))
#' con <- DBI::dbConnect(RSQLite::SQLite())
#'
#' # Copy to temporary tables, unique table names by default:
#' temp_dm <- copy_dm_to(
#'   con,
#'   dm_nycflights13(),
#'   set_key_constraints = FALSE
#' )
#'
#' # Persist, explicitly specify table names:
#' persistent_dm <- copy_dm_to(
#'   con,
#'   dm_nycflights13(),
#'   temporary = FALSE,
#'   table_names = ~ paste0("flights_", .x)
#' )
#' dbplyr::remote_name(persistent_dm$planes)
#'
#' DBI::dbDisconnect(con)
#' @export
copy_dm_to <- function(
  dest,
  dm,
  ...,
  set_key_constraints = TRUE,
  table_names = NULL,
  temporary = TRUE,
  schema = NULL,
  progress = NA,
  unique_table_names = NULL,
  copy_to = NULL
) {
  if (!is.null(unique_table_names)) {
    deprecate_stop(
      "0.1.4",
      "dm::copy_dm_to(unique_table_names = )",
      details = "Use `table_names = set_names(names(dm))` to use unchanged names for temporary tables."
    )
  }

  if (!is.null(copy_to)) {
    deprecate_stop(
      "1.0.0",
      "dm::copy_dm_to(copy_to = )",
      details = "Use `dm_sql()` for more control over the schema creation process."
    )
  }

  check_dots_empty()
  dm_local_error_call()

  check_not_zoomed(dm)

  check_suggested("dbplyr", "copy_dm_to")

  dest <- src_from_src_or_con(dest)

  if (!is_db(dest)) {
    deprecate_stop(
      "0.1.6",
      "dm::copy_dm_to(dest = 'must refer to a DBI connection')",
      "dm::collect.dm()"
    )
  }

  src_names <- src_tbls_impl(dm)
  dest_con <- con_from_src_or_con(dest)

  # in case `table_names` was chosen by the user, check if the input makes sense:
  # 1. is there one name per dm-table?
  # 2. are there any duplicated table names?
  # 3. is it a named character or ident_q vector with the correct names?
  if (is.null(table_names)) {
    table_names_out <- repair_table_names_for_db(src_names, temporary, dest_con, schema)
    # https://github.com/tidyverse/dbplyr/issues/487
    if (is_mssql(dest)) {
      temporary <- FALSE
    }
  } else {
    if (!is.null(schema)) {
      abort_one_of_schema_table_names()
    }
    if (is_function(table_names) || is_bare_formula(table_names)) {
      table_name_fun <- as_function(table_names)
      table_names_out <- set_names(table_name_fun(src_names), src_names)
    } else {
      table_names_out <- table_names
    }
    check_naming(names(table_names_out), src_names)

    if (anyDuplicated(table_names_out)) {
      problem <- table_names_out[duplicated(table_names_out)][[1]]
      abort_copy_dm_to_table_names_duplicated(problem)
    }

    names(table_names_out) <- src_names
  }

  # Must be done here because table types may depend on string length, #2066
  dm <- collect(dm, progress = progress)

  if (isTRUE(set_key_constraints)) {
    dm_for_sql <- dm
  } else {
    def_no_keys <- dm_get_def(dm)
    # FIXME: Add test
    def_no_keys$pks[] <- list(new_pk())
    def_no_keys$uks[] <- list(new_uk())
    def_no_keys$fks[] <- list(new_fk())
    # Must keep primary keys
    dm_for_sql <- dm_from_def(def_no_keys)
  }

  sql <- dm_sql(dm_for_sql, dest_con, table_names_out, temporary)

  # FIXME: Extract function
  # FIXME: Make descriptions part of the dm_sql() output

  pre <- unlist(sql$pre)
  load <- unlist(sql$load)
  post <- unlist(sql$post)

  ticker_pre <- new_ticker(
    "creating tables",
    n = length(pre),
    progress = progress,
    top_level_fun = "copy_dm_to"
  )

  # create tables
  walk(
    pre,
    ticker_pre(
      ~ {
        DBI::dbExecute(dest_con, .x, immediate = TRUE)
      }
    )
  )

  ticker_load <- new_ticker(
    "populating tables",
    n = length(load),
    progress = progress,
    top_level_fun = "copy_dm_to"
  )

  # populate tables
  walk(
    load,
    ticker_load(
      ~ {
        DBI::dbExecute(dest_con, .x, immediate = TRUE)
      }
    )
  )

  ticker_post <- new_ticker(
    "creating indexes",
    n = length(post),
    progress = progress,
    top_level_fun = "copy_dm_to"
  )

  # create indexes
  walk(
    post,
    ticker_post(
      ~ {
        DBI::dbExecute(dest_con, .x, immediate = TRUE)
      }
    )
  )

  # remote dm is same as source dm with replaced data
  # FIXME: Extract function
  def <- dm_get_def(dm)

  remote_tables <- map2(
    table_names_out,
    map(def$data, colnames),
    ~ tbl(dest_con, .x, vars = .y)
  )

  def$data <- unname(remote_tables)
  remote_dm <- dm_from_def(def)

  invisible(debug_dm_validate(remote_dm))
}

check_naming <- function(table_names, dm_table_names) {
  if (!identical(sort(table_names), sort(dm_table_names))) {
    abort_copy_dm_to_table_names()
  }
}

db_append_table <- function(
  con,
  remote_table,
  table,
  progress,
  top_level_fun = "copy_dm_to",
  autoinc = logical(0)
) {
  stopifnot(is.data.frame(table))
  if (nrow(table) == 0 || ncol(table) == 0) {
    return(invisible())
  }

  remote_table_name <- DBI::dbQuoteIdentifier(con, remote_table)

  if (is_mssql(con)) {
    # FIXME: Make adaptive
    chunk_size <- 1000L
    n_chunks <- ceiling(nrow(table) / chunk_size)

    remote_table_id <- DBI::dbQuoteIdentifier(con, remote_table)

    ticker <- new_ticker(
      paste0("inserting into ", remote_table_id),
      n = n_chunks,
      progress = progress,
      top_level_fun = top_level_fun
    )

    walk(
      seq_len(n_chunks),
      ticker(
        ~ {
          end <- .x * chunk_size
          idx <- seq2(end - (chunk_size - 1), min(end, nrow(table)))
          values <- map(table[idx, , drop = FALSE], mssql_escape, con = con)
          # Can't use dbAppendTable(): https://github.com/r-dbi/odbc/issues/480
          sql <- DBI::sqlAppendTable(con, remote_table_id, values, row.names = FALSE)
          if (length(autoinc) > 1L) {
            cli::cli_abort("More than one autoincrement key in one table.")
          }
          if (!is_empty(autoinc) && autoinc) {
            sql <- DBI::SQL(paste0(
              "SET IDENTITY_INSERT ",
              remote_table_name,
              " ON\n",
              sql,
              "\n",
              "SET IDENTITY_INSERT ",
              remote_table_name,
              " OFF"
            ))
          }
          DBI::dbExecute(con, sql, immediate = TRUE)
        }
      )
    )
  } else if (is_postgres(con) || is_redshift(con)) {
    # https://github.com/r-dbi/RPostgres/issues/384
    table <- as.data.frame(table)
    # https://github.com/r-dbi/RPostgres/issues/382
    DBI::dbAppendTable(con, remote_table, table, copy = FALSE)
  } else {
    DBI::dbAppendTable(con, remote_table, table)
  }

  invisible()
}


# Errors ------------------------------------------------------------------

abort_copy_dm_to_table_names <- function() {
  cli::cli_abort(
    "{.arg table_names} must have names that are the same as the table names in {.arg dm}.",
    class = dm_error_full("copy_dm_to_table_names"),
    call = dm_error_call()
  )
}

abort_copy_dm_to_table_names_duplicated <- function(problem) {
  cli::cli_abort(
    c(
      "{.arg table_names} must be unique.",
      i = "Duplicate: {.code {problem}}"
    ),
    class = dm_error_full("copy_dm_to_table_names_duplicated"),
    call = dm_error_call()
  )
}

test_that("dummy", {
  expect_snapshot({
    "dummy"
  })
})

test_that("dm_sql()", {
  # Need skip in every test block, unfortunately
  skip_if_src_not("sqlite")

  # https://github.com/tidyverse/dbplyr/pull/1190
  skip_if(is(my_test_con(), "MySQLConnection"))

  expect_snapshot({
    dm_for_filter_df() %>%
      dm_sql(my_test_con())
  })

  expect_snapshot({
    dm(x = data.frame(a = strrep("x", 300))) %>%
      dm_sql(my_test_con())
  })

  expect_snapshot({
    dm(x = data.frame(a = strrep("x", 10000))) %>%
      dm_sql(my_test_con())
  })
})

test_that("long text columns with copy_dm_to()", {
  # Need skip in every test block, unfortunately
  skip_if_src_not("sqlite")

  dm <- dm(x = data.frame(a = strrep("x", 300)))
  dm_out <- copy_dm_to(my_test_con(), dm)
  expect_equal(collect(dm_out$x)$a, dm$x$a)
  dm_out_out <- copy_dm_to(my_test_con(), dm_out)
  expect_equal(collect(dm_out_out$x)$a, dm$x$a)

  dm <- dm(x = data.frame(a = strrep("x", 10000)))
  dm_out <- copy_dm_to(my_test_con(), dm)
  expect_equal(collect(dm_out$x)$a, dm$x$a)
  dm_out_out <- copy_dm_to(my_test_con(), dm_out)
  expect_equal(collect(dm_out_out$x)$a, dm$x$a)
})

test_that("dummy", {
  expect_snapshot({
    "dummy"
  })
})

test_that("dm_sql()", {
  # Need skip in every test block, unfortunately
  skip_if_src_not("postgres")

  # https://github.com/tidyverse/dbplyr/pull/1190
  skip_if(is(my_test_con(), "MySQLConnection"))

  expect_snapshot({
    dm_for_filter_df() %>%
      dm_sql(my_test_con())
  })

  expect_snapshot({
    dm(x = data.frame(a = strrep("x", 300))) %>%
      dm_sql(my_test_con())
  })

  expect_snapshot({
    dm(x = data.frame(a = strrep("x", 10000))) %>%
      dm_sql(my_test_con())
  })
})

test_that("long text columns with copy_dm_to()", {
  # Need skip in every test block, unfortunately
  skip_if_src_not("postgres")

  dm <- dm(x = data.frame(a = strrep("x", 300)))
  dm_out <- copy_dm_to(my_test_con(), dm)
  expect_equal(collect(dm_out$x)$a, dm$x$a)
  dm_out_out <- copy_dm_to(my_test_con(), dm_out)
  expect_equal(collect(dm_out_out$x)$a, dm$x$a)

  dm <- dm(x = data.frame(a = strrep("x", 10000)))
  dm_out <- copy_dm_to(my_test_con(), dm)
  expect_equal(collect(dm_out$x)$a, dm$x$a)
  dm_out_out <- copy_dm_to(my_test_con(), dm_out)
  expect_equal(collect(dm_out_out$x)$a, dm$x$a)
})

#' Draw a diagram of the data model
#'
#' @description
#' `dm_draw()` draws a diagram, a visual representation of the data model.
#'
#' @details
#' Currently, \pkg{dm} uses \pkg{DiagrammeR} to draw diagrams.
#' Use [DiagrammeRsvg::export_svg()] to convert the diagram to an SVG file.
#'
#' The backend for drawing the diagrams might change in the future.
#' If you rely on DiagrammeR, pass an explicit value for the `backend` argument.
#'
#' @param dm A [`dm`] object.
#' @param rankdir Graph attribute for direction (e.g., 'BT' = bottom --> top).
#' @param col_attr Deprecated, use `colummn_types` instead.
#' @param view_type Can be "keys_only" (default), "all" or "title_only".
#'   It defines the level of details for rendering tables
#'   (only primary and foreign keys, all columns, or no columns).
#' @inheritParams rlang::args_dots_empty
#' @param column_types Set to `TRUE` to show column types.
#' @param backend Currently, only the default `"DiagrammeR"` is accepted.
#'   Pass this value explicitly if your code relies on the type of the return value.
#' @param backend_opts A named list of backend-specific options.
#'   For the `"DiagrammeR"` backend, supported options are:
#'   - `graph_attrs`: Additional graph attributes (default `""`).
#'   - `node_attrs`: Additional node attributes (default `""`).
#'   - `edge_attrs`: Additional edge attributes (default `""`).
#'   - `focus`: A list of parameters for rendering (table filter).
#'   - `graph_name`: The name of the graph (default `"Data Model"`).
#'   - `column_arrow`: Edges from columns to columns (default `TRUE`).
#'   - `font_size`: `r lifecycle::badge("experimental")`
#'     Font size for `header` (default `16`), `column` (default `16`),
#'     and `table_description` (default `8`).
#'     Can be set as a named integer vector, e.g.
#'     `c(table_headers = 18L, table_description = 6L)`.
#' @param columnArrows `r lifecycle::badge("deprecated")` Use
#'   `backend_opts = list(column_arrow = ...)` instead.
#' @param graph_attrs `r lifecycle::badge("deprecated")` Use
#'   `backend_opts = list(graph_attrs = ...)` instead.
#' @param node_attrs `r lifecycle::badge("deprecated")` Use
#'   `backend_opts = list(node_attrs = ...)` instead.
#' @param edge_attrs `r lifecycle::badge("deprecated")` Use
#'   `backend_opts = list(edge_attrs = ...)` instead.
#' @param focus `r lifecycle::badge("deprecated")` Use
#'   `backend_opts = list(focus = ...)` instead.
#' @param graph_name `r lifecycle::badge("deprecated")` Use
#'   `backend_opts = list(graph_name = ...)` instead.
#' @param font_size `r lifecycle::badge("deprecated")` Use
#'   `backend_opts = list(font_size = ...)` instead.
#'
#' @seealso [dm_set_colors()] for defining the table colors.
#' @seealso [dm_set_table_description()] for adding details to one or more tables in the diagram
#'
#' @export
#'
#' @return An object with a [print()] method, which,
#' when printed, produces the output seen in the viewer as a side effect.
#' Currently, this is an object of class `grViz` (see also
#' [DiagrammeR::grViz()]), but this is subject to change.
#'
#' @examplesIf rlang::is_installed(c("nycflights13", "DiagrammeR"))
#' dm_nycflights13() %>%
#'   dm_draw()
#'
#' dm_nycflights13(cycle = TRUE) %>%
#'   dm_draw(view_type = "title_only")
#'
#' head(dm_get_available_colors())
#' length(dm_get_available_colors())
#'
#' dm_nycflights13() %>%
#'   dm_get_colors()
dm_draw <- function(
  dm,
  rankdir = "LR",
  ...,
  col_attr = NULL,
  view_type = c("keys_only", "all", "title_only"),
  column_types = NULL,
  backend = c("DiagrammeR"),
  backend_opts = list(),
  columnArrows = lifecycle::deprecated(),
  graph_attrs = lifecycle::deprecated(),
  node_attrs = lifecycle::deprecated(),
  edge_attrs = lifecycle::deprecated(),
  focus = lifecycle::deprecated(),
  graph_name = lifecycle::deprecated(),
  font_size = lifecycle::deprecated()
) {
  dm_local_error_call()
  check_not_zoomed(dm)
  check_dots_empty()

  # Handle deprecated DiagrammeR-specific arguments
  if (lifecycle::is_present(columnArrows)) {
    lifecycle::deprecate_soft(
      "1.1.0",
      "dm_draw(columnArrows = )",
      details = "Use `backend_opts = list(column_arrow = ...)` instead."
    )
    backend_opts[["column_arrow"]] <- backend_opts[["column_arrow"]] %||% columnArrows
  }
  if (lifecycle::is_present(graph_attrs)) {
    lifecycle::deprecate_soft(
      "1.1.0",
      "dm_draw(graph_attrs = )",
      details = "Use `backend_opts = list(graph_attrs = ...)` instead."
    )
    backend_opts[["graph_attrs"]] <- backend_opts[["graph_attrs"]] %||% graph_attrs
  }
  if (lifecycle::is_present(node_attrs)) {
    lifecycle::deprecate_soft(
      "1.1.0",
      "dm_draw(node_attrs = )",
      details = "Use `backend_opts = list(node_attrs = ...)` instead."
    )
    backend_opts[["node_attrs"]] <- backend_opts[["node_attrs"]] %||% node_attrs
  }
  if (lifecycle::is_present(edge_attrs)) {
    lifecycle::deprecate_soft(
      "1.1.0",
      "dm_draw(edge_attrs = )",
      details = "Use `backend_opts = list(edge_attrs = ...)` instead."
    )
    backend_opts[["edge_attrs"]] <- backend_opts[["edge_attrs"]] %||% edge_attrs
  }
  if (lifecycle::is_present(focus)) {
    lifecycle::deprecate_soft(
      "1.1.0",
      "dm_draw(focus = )",
      details = "Use `backend_opts = list(focus = ...)` instead."
    )
    backend_opts[["focus"]] <- backend_opts[["focus"]] %||% focus
  }
  if (lifecycle::is_present(graph_name)) {
    lifecycle::deprecate_soft(
      "1.1.0",
      "dm_draw(graph_name = )",
      details = "Use `backend_opts = list(graph_name = ...)` instead."
    )
    backend_opts[["graph_name"]] <- backend_opts[["graph_name"]] %||% graph_name
  }
  if (lifecycle::is_present(font_size)) {
    lifecycle::deprecate_soft(
      "1.1.0",
      "dm_draw(font_size = )",
      details = "Use `backend_opts = list(font_size = ...)` instead."
    )
    backend_opts[["font_size"]] <- backend_opts[["font_size"]] %||% font_size
  }

  tbl_names <- src_tbls_impl(dm, quiet = TRUE)
  table_description <- dm_get_table_description_impl(
    dm,
    set_names(seq_along(tbl_names), tbl_names)
  ) %>%
    prep_recode()

  view_type <- arg_match(view_type)

  if (!is.null(col_attr)) {
    deprecate_warn("0.1.13", "dm::dm_draw(col_attr = )", "dm::dm_draw(column_types = )")
    if (is.null(column_types) && "type" %in% col_attr) {
      column_types <- TRUE
    }
  }

  backend <- arg_match(backend)

  supported_backend_opts <- c(
    "graph_attrs",
    "node_attrs",
    "edge_attrs",
    "focus",
    "graph_name",
    "column_arrow",
    "font_size"
  )
  unsupported <- setdiff(names(backend_opts), supported_backend_opts)
  if (length(unsupported) > 0) {
    cli::cli_abort(
      c(
        "Unsupported {.arg backend_opts} for backend {.val {backend}}: {.field {unsupported}}.",
        i = "Supported options are: {.field {supported_backend_opts}}."
      ),
      class = dm_error_full("unsupported_backend_opts"),
      call = dm_error_call()
    )
  }

  if (is_empty(dm)) {
    message("The dm cannot be drawn because it is empty.")
    return(invisible(NULL))
  }

  column_types <- isTRUE(column_types)

  data_model <- dm_get_data_model(dm, column_types)

  # DiagrammeR backend (default)
  graph <- bdm_create_graph(
    data_model,
    rankdir = rankdir,
    col_attr = c("column", if (column_types) "type"),
    view_type = view_type,
    columnArrows = backend_opts[["column_arrow"]] %||% TRUE,
    graph_attrs = backend_opts[["graph_attrs"]] %||% "",
    node_attrs = backend_opts[["node_attrs"]] %||% "",
    edge_attrs = backend_opts[["edge_attrs"]] %||% "",
    focus = backend_opts[["focus"]],
    graph_name = backend_opts[["graph_name"]] %||% "Data Model",
    table_description = as.list(table_description),
    font_size = as.list(backend_opts[["font_size"]])
  )
  bdm_render_graph(graph, top_level_fun = "dm_draw")
}

#' Get data_model
#'
#' `dm_get_data_model()` converts a `dm` to a \pkg{datamodelr}
#' data model object for drawing.
#'
#' @noRd
#' @autoglobal
dm_get_data_model <- function(x, column_types = FALSE) {
  def <- dm_get_def(x)

  tables <- data.frame(
    table = def$table,
    segment = def$segment,
    display = def$display,
    stringsAsFactors = FALSE
  )

  all_uks <- dm_get_all_uks_impl(x)
  references_for_columns <- dm_get_all_fks_impl(x, id = TRUE) %>%
    left_join(all_uks, by = c("parent_table" = "table", "parent_key_cols" = "uk_col")) %>%
    rename(uk_col = kind) %>%
    transmute(
      table = child_table,
      column = format(child_fk_cols),
      ref = parent_table,
      ref_col = format(parent_key_cols),
      keyId = id,
      uk_col = if_else(uk_col != "PK", ", style=\"dashed\"", "")
    )

  references <-
    references_for_columns %>%
    mutate(ref_id = row_number(), ref_col_num = 1L)

  keys_pk <-
    all_uks %>%
    mutate(column = format(uk_col)) %>%
    select(table, column, kind) %>%
    mutate(key = 1L)

  keys_fk <-
    dm_get_all_fks_impl(x) %>%
    mutate(column = format(parent_key_cols)) %>%
    select(table = parent_table, column) %>%
    mutate(key_fk = 2L) %>%
    # `parent_table` and `column` can be referenced by multiple child tables
    distinct()

  if (column_types) {
    types <- dm_get_all_column_types(x)
  } else {
    types <- dm_get_all_columns(x)
  }

  columns <-
    types %>%
    full_join(keys_pk, by = c("table", "column")) %>%
    full_join(keys_fk, by = c("table", "column")) %>%
    # there is a legitimate interest to have duplicates in `table` and `column`
    # in table `references_for_columns`.
    # When using a dplyr version >= 1.1.0, we get a warning in that case, thus
    # we need `multiple = "all"`.
    # FIXME: is there another way? like this we need a min dplyr version 1.1.0.
    full_join(references_for_columns, by = c("table", "column"), multiple = "all") %>%
    # Order matters: key == 2 if foreign key points to non-default primary key
    mutate(key = coalesce(key, key_fk, 0L)) %>%
    select(-key_fk) %>%
    # I don't understand why this is necessary
    distinct() %>%
    # for compatibility with print method from {datamodelr}
    as.data.frame()

  new_data_model(
    tables,
    columns,
    references
  )
}

dm_get_all_columns <- function(x) {
  x %>%
    dm_get_tables_impl() %>%
    map(colnames) %>%
    map(~ enframe(., "id", "column")) %>%
    enframe("table") %>%
    unnest_df("value", tibble(id = integer(), column = character())) %>%
    select(table, column, id)
}

#' @autoglobal
dm_get_all_column_types <- function(x) {
  x %>%
    dm_get_tables_impl() %>%
    map(
      ~ mutate(
        enframe(as.list(collect(head(.x, 0))), "column"),
        id = row_number()
      )
    ) %>%
    enframe("table") %>%
    unnest_df("value", tibble(column = character(), value = list(), id = integer())) %>%
    mutate(type = map_chr(value, vec_ptype_abbr), .keep = "unused")
}

#' Color in database diagrams
#'
#' @description
#' `dm_set_colors()` allows to define the colors that will be used to display the tables of the data model with [dm_draw()].
#' The colors can either be specified with hex color codes or using the names of the built-in R colors.
#' An overview of the colors corresponding to the standard color names can be found at
#' the bottom of
#' [https://rpubs.com/krlmlr/colors](https://rpubs.com/krlmlr/colors).
#'
#' @inheritParams dm_draw
#' @param ... Colors to set in the form `color = table`.
#' Allowed colors are all hex coded colors (quoted) and the color names from `dm_get_available_colors()`.
#' `tidyselect` is supported, see [dplyr::select()] for details on the semantics.
#' @return For `dm_set_colors()`: the updated data model.
#'
#' @export
#' @examplesIf rlang::is_installed(c("nycflights13", "DiagrammeR"))
#' dm_nycflights13(color = FALSE) %>%
#'   dm_set_colors(
#'     darkblue = starts_with("air"),
#'     "#5986C4" = flights
#'   ) %>%
#'   dm_draw()
#'
#' # Splicing is supported:
#' nyc_cols <-
#'   dm_nycflights13() %>%
#'   dm_get_colors()
#' nyc_cols
#'
#' dm_nycflights13(color = FALSE) %>%
#'   dm_set_colors(!!!nyc_cols) %>%
#'   dm_draw()
#' @autoglobal
dm_set_colors <- function(dm, ...) {
  dm_local_error_call()
  check_dm(dm)
  quos <- enquos(...)
  if (any(names(quos) == "")) {
    abort_only_named_args("dm_set_colors", "the colors")
  }
  cols <- names(quos)
  if (
    !all(cols[!is_hex_color(cols)] %in% dm_get_available_colors()) &&
      all(cols %in% src_tbls_impl(dm))
  ) {
    abort_wrong_syntax_set_cols()
  }

  # get table names for each color (name_spec argument is not needed)
  selected_tables <- eval_select_table(quo(c(...)), src_tbls_impl(dm), unique = FALSE)

  # convert color names to hex color codes (if already hex code this is a no-op)
  # avoid error from mutate()
  names(selected_tables) <- col_to_hex(names(selected_tables))

  display_df <-
    selected_tables %>%
    enframe(name = "new_display", value = "table") %>%
    # needs to be done like this, `distinct()` would keep the first one
    filter(!duplicated(table, fromLast = TRUE))

  def <-
    dm_get_def(dm) %>%
    left_join(display_df, by = "table") %>%
    mutate(display = coalesce(new_display, display)) %>%
    select(-new_display)

  dm_from_def(def)
}

color_quos_to_display <- function(...) {
  quos <- enquos(..., .named = TRUE, .ignore_empty = "none", .homonyms = "error")
  missing <- map_lgl(quos, quo_is_missing)
  if (has_length(missing) && missing[[length(missing)]]) {
    abort_last_col_missing()
  }
  avail <- !missing
  idx <- rev(cumsum(rev(avail)))
  values <- map_chr(quos[avail], eval_tidy)

  set_names(names(quos), rev(values)[idx])
}

#' dm_get_colors()
#'
#' `dm_get_colors()` returns the colors defined for a data model.
#'
#' @return For `dm_get_colors()`, a named character vector of table names
#'   with the colors in the names.
#'   This allows calling `dm_set_colors(!!!dm_get_colors(...))`.
#'   Use [tibble::enframe()] to convert this to a tibble.
#'
#' @rdname dm_set_colors
#' @export
dm_get_colors <- function(dm) {
  dm_local_error_call()
  check_dm(dm)
  dm %>%
    dm_get_def() %>%
    select(table, display) %>%
    select(display, table) %>%
    mutate(display = coalesce(display, "default")) %>%
    deframe()
}

#' dm_get_available_colors()
#'
#' `dm_get_available_colors()` returns an overview of the names of the available colors
#' These are the standard colors also returned by [grDevices::colors()] plus a default
#' table color with the name "default".
#'
#' @return For `dm_get_available_colors()`, a vector with the available colors.
#'
#' @rdname dm_set_colors
#' @export
dm_get_available_colors <- function() {
  c("default", colors())
}

test_that("API", {
  expect_identical(
    color_quos_to_display(
      flights = "blue",
      airlines = ,
      airports = "orange",
      planes = "green_nb"
    ),
    set_names(
      c("flights", "airlines", "airports", "planes"),
      c("blue", "orange", "orange", "green_nb")
    )
  )
})

test_that("`dm_set_colors()` works", {
  expect_snapshot({
    dm_nycflights_small() %>%
      dm_set_colors(
        blue = starts_with("air"),
        green = contains("h")
      ) %>%
      dm_get_colors()
  })

  skip_if_not_installed("testthat", "3.1.1")

  colset <- c(blue = "flights", green = "airports")

  # test splicing
  expect_snapshot(
    variant = if (packageVersion("testthat") > "3.1.0") "testthat-new" else "testthat-legacy",
    {
      dm_nycflights_small() %>%
        dm_set_colors(!!!colset) %>%
        dm_get_colors()
    }
  )
})

test_that("`dm_set_colors()` errors if old syntax used", {
  expect_dm_error(
    dm_set_colors(
      dm_nycflights_small(),
      airports = ,
      airlines = "blue",
      flights = ,
      weather = "green"
    ),
    class = "wrong_syntax_set_cols"
  )
})

test_that("`dm_set_colors()` errors with unnamed args", {
  expect_dm_error(
    dm_set_colors(
      dm_nycflights_small(),
      airports
    ),
    class = "only_named_args"
  )
})

test_that("last", {
  expect_dm_error(
    color_quos_to_display(
      flights = "blue",
      airlines =
    ),
    class = "last_col_missing"
  )
})

test_that("bad color", {
  skip_if_not(getRversion() >= "3.5")

  expect_dm_error(
    dm_set_colors(
      dm_nycflights_small(),
      "zzz-bogus" = flights
    ),
    class = "cols_not_avail"
  )
})

test_that("getter", {
  expect_equal(
    dm_get_colors(dm_nycflights13()),
    c(
      "#ED7D31FF" = "airlines",
      "#ED7D31FF" = "airports",
      "#5B9BD5FF" = "flights",
      "#ED7D31FF" = "planes",
      "#70AD47FF" = "weather"
    )
  )
})

test_that("get available colors", {
  expect_length(
    dm_get_available_colors(),
    length(colors()) + 1
  )
})

test_that("helpers", {
  expect_snapshot({
    dm_get_all_columns(dm_for_filter())
  })

  expect_snapshot({
    dm_get_all_column_types(dm_for_filter())
  })
})

test_that("output", {
  skip_if_not_installed("testthat", "3.1.1")

  # 444: types
  expect_snapshot_diagram(
    dm_nycflights13() %>%
      dm_draw(column_types = TRUE),
    "nycflight-dm-types.svg"
  )

  expect_snapshot_diagram(
    dm_nycflights13(cycle = TRUE) %>%
      dm_zoom_to(planes) %>%
      # Multi-fk (#37)
      dm_insert_zoomed("planes_copy") %>%
      # Loose table
      dm(loose = tibble(a = 1)) %>%
      # Non-default fk (#402)
      dm(agency = tibble(airline_name = character())) %>%
      dm_add_fk(agency, airline_name, airlines, name) %>%
      dm_draw(),
    "nycflight-dm.svg"
  )

  # empty table corner cases
  expect_snapshot_diagram(
    dm(a = tibble()) %>%
      dm_draw(),
    "single-empty-table-dm.svg"
  )

  expect_snapshot_diagram(
    dm(x = tibble(a = 1), y = tibble(b = 1), a = tibble()) %>%
      dm_draw(view_type = "all"),
    "empty-table-in-dm.svg"
  )
})

test_that("table_description works", {
  expect_error(dm_set_table_description(dm_nycflights_small(), "flight" = "Flüge"))
  expect_snapshot_diagram(
    dm_nycflights_small() %>%
      dm_set_table_description("high in the sky\nflying from NY" = flights) %>%
      dm_draw(),
    "table-desc-1-dm.svg"
  )

  expect_snapshot_diagram(
    dm_nycflights_small() %>%
      dm_set_table_description("high in the sky\nflying from NY" = flights) %>%
      dm_draw(backend_opts = list(font_size = list(table_description = 6L))),
    "table-desc-2-dm.svg"
  )

  expect_snapshot_diagram(
    dm_nycflights_small() %>%
      dm_set_table_description("high in the sky\nflying from NY" = flights) %>%
      dm_draw(
        backend_opts = list(font_size = c(table_description = 6L, header = 19L, column = 14L))
      ),
    "table-desc-3-dm.svg"
  )

  expect_snapshot_diagram(
    dm_nycflights13(table_description = TRUE) %>%
      dm_draw(
        backend_opts = list(font_size = c(table_description = 6L, header = 19L, column = 14L))
      ),
    "table-desc-4-dm.svg"
  )
})

test_that("UK support works", {
  expect_snapshot_diagram(
    dm_nycflights_small() %>%
      dm_add_uk(weather, time_hour) %>%
      dm_set_table_description("high in the sky\nflying from NY" = flights) %>%
      dm_draw(view_type = "all"),
    "table-uk-1-dm.svg"
  )

  expect_snapshot_diagram(
    dm_nycflights_small() %>%
      dm_add_fk(flights, time_hour, weather, time_hour) %>%
      dm_set_table_description("Wetter\npogoda" = weather) %>%
      dm_draw(),
    "table-uk-2-dm.svg"
  )
})

test_that("DiagrammeR-specific options are soft-deprecated", {
  skip_if_not_installed("DiagrammeR")

  expect_snapshot({
    dm_nycflights13() |>
      dm_draw(graph_attrs = "rankdir=LR") |>
      invisible()
  })

  expect_snapshot({
    dm_nycflights13() |>
      dm_draw(font_size = 14L) |>
      invisible()
  })
})

test_that("unsupported backend_opts fail with clear error", {
  expect_snapshot(
    dm_nycflights13() |> dm_draw(backend_opts = list(columnArrows = TRUE)),
    error = TRUE
  )

  expect_snapshot(
    dm_nycflights13() |> dm_draw(backend_opts = list(foo = 1, bar = 2)),
    error = TRUE
  )
})

test_that("column_arrow backend option works", {
  skip_if_not_installed("DiagrammeR")

  expect_snapshot({
    dm_nycflights13() |>
      dm_draw(columnArrows = FALSE) |>
      invisible()
  })

  # Directly using column_arrow in backend_opts should work without warning
  expect_no_warning(
    dm_nycflights13() |>
      dm_draw(backend_opts = list(column_arrow = FALSE)) |>
      invisible()
  )
})

new_cg_block <- function(cg_input_object = list(), cg_f_list = list()) {
  structure(
    list(cg_input_object = cg_input_object, cg_f_list = cg_f_list),
    class = "dm_cg_code_block"
  )
}

cg_add_call <- function(cg_block, fn_call) {
  fn <- function(.) NULL
  body(fn) <- enexpr(fn_call)
  new_cg_block(cg_block$cg_input_object, append(cg_block$cg_f_list, fn))
}

#' @export
format.dm_cg_code_block <- function(x, ...) {
  c(
    cg_format_input_object(x, is_empty(x$cg_f_list)),
    cg_format_f_list(x$cg_f_list)
  )
}

#' @export
print.dm_cg_code_block <- function(x, ...) {
  writeLines(format(x, ...))
  invisible(x)
}

cg_format_input_object <- function(x, has_no_f_call) {
  if (has_no_f_call) {
    if (!is_empty(x$cg_input_object)) {
      cg_clean_in_obj_text(x)
    }
  } else {
    paste0(cg_clean_in_obj_text(x), " %>%")
  }
}

cg_format_f_list <- function(x) {
  if (is_empty(x)) {
    character()
  } else {
    purrr::map_chr(x, ~ body(.) %>% call_to_char()) %>%
      paste0("  ", ., collapse = " %>%\n")
  }
}

call_to_char <- function(body) {
  stopifnot(identical(body[[2]], sym(".")))
  body <- as.call(as.list(body[-2]))
  trimws(expr_deparse(body)) %>% paste0(collapse = " ")
}

cg_eval_block <- function(cg_block) {
  if (is_empty(cg_block$cg_f_list)) {
    eval_tidy(cg_block$cg_input_object)
  } else {
    magrittr::freduce(eval_tidy(cg_block$cg_input_object), cg_block$cg_f_list)
  }
}

cg_clean_in_obj_text <- function(x) {
  gsub("\033[39m", "", trimws(expr_deparse(x$cg_input_object)), fixed = TRUE) %>%
    gsub("\\^", "", .)
}

test_that("code generation works", {
  expect_identical(
    new_cg_block(),
    structure(list(cg_input_object = list(), cg_f_list = list()), class = "dm_cg_code_block")
  )

  expect_snapshot({
    call_to_char(body(function(.) dm_add_tbl(., weather)))
    call_to_char(expr(dm_add_tbl(
      .,
      weather,
      airports,
      flights,
      airlines,
      planes,
      mtcars,
      penguins
    )))
  })

  expect_snapshot({
    new_cg_block()
    new_cg_block(quo(dm_nycflights13()), list(function(.) dm_add_pk(., flights, flight_id)))
  })

  expect_snapshot({
    table <- "flights"
    columns <- "carrier"
    cg_block <- new_cg_block(quo(dm_nycflights13())) %>%
      cg_add_call(dm_rm_fk(
        .,
        table = !!sym(table),
        columns = !!sym(columns),
        ref_table = airlines
      )) %>%
      cg_add_call(dm_rm_fk(
        .,
        table = flights,
        columns = c(origin, time_hour),
        ref_table = weather
      )) %>%
      cg_add_call(dm_add_fk(
        .,
        table = !!sym(table),
        columns = !!sym(columns),
        ref_table = airlines
      ))
    cg_block
    cg_eval_block(cg_block)
  })

  expect_snapshot({
    cg_block_2 <- new_cg_block(
      cg_block$cg_input_object,
      list(
        function(.) dm(., mtcars),
        function(.) dm_select_tbl(., -planes)
      )
    )
    cg_block_2
    cg_eval_block(cg_block_2)
  })

  expect_snapshot({
    format(new_cg_block(quo(dm_nycflights13()), list(function(.) dm_add_pk(., flights, flight_id))))
  })
})

d1 <- tibble::tibble(a = 1:5, b = letters[1:5])
d2 <- tibble::tibble(a = c(1,3:6), b = letters[1:5])
d3 <- tibble::tibble(c = 1:5)
d4 <- tibble::tibble(c = c(1:5,5))
d5 <- tibble::tibble(a = 1:5)
d6 <- tibble::tibble(c = 1:4)



check_cardinality_0_n(d1, a, d2, a)
check_cardinality_0_n(d1, a, d3, c)
check_cardinality_1_n(d1, a, d3, c)
check_cardinality_1_1(d1, a, d3, c)
check_cardinality_1_1(d1, a, d4, c)
check_cardinality_1_1(d4, c, d1, a)
check_set_equality(d1, a, d3, c)
check_cardinality_0_1(d1, a, d2, a)
check_cardinality_0_1(d1, a, d4, c)
check_cardinality_0_1(d4, c, d1, a)
check_cardinality_0_n(d4, c, d5, a)
check_cardinality_0_n(d5, a, d4, c)
check_cardinality_1_1(d5, a, d4, c)
check_cardinality_0_1(d5, a, d6, c)

#' Resolve column name ambiguities
#'
#' @description
#' This function ensures that all columns in a `dm` have unique names.
#'
#' @details
#' The function first checks if there are any column names that are not unique.
#' If there are, those columns will be assigned new, unique, names by prefixing their existing name
#' with the name of their table and a separator.
#' Columns that act as primary or foreign keys will not be renamed
#' because only the foreign key column will remain when two tables are joined,
#' making that column name "unique" as well.
#'
#' @inheritParams dm_add_pk
#' @inheritParams rlang::args_dots_empty
#' @param .sep The character variable that separates the names of the table and the names of the ambiguous columns.
#' @param .quiet Boolean.
#'   By default, this function lists the renamed columns in a message, pass `TRUE` to suppress this message.
#' @param .position
#'   `r lifecycle::badge("experimental")`
#'   By default, table names are appended to the column names to resolve conflicts.
#'   Prepending table names was the default for versions before 1.0.0,
#'   use `"prefix"` to achieve this behavior.
#'
#' @return A `dm` whose column names are unambiguous.
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_disambiguate_cols()
#' @export
dm_disambiguate_cols <- function(
  dm,
  .sep = ".",
  ...,
  .quiet = FALSE,
  .position = c("suffix", "prefix")
) {
  dm_local_error_call()
  check_not_zoomed(dm)
  check_dots_empty()
  .position <- arg_match(.position)
  dm_disambiguate_cols_impl(
    dm,
    tables = NULL,
    sep = .sep,
    quiet = .quiet,
    position = .position
  )
}

dm_disambiguate_cols_impl <- function(dm, tables, sep = ".", quiet = FALSE, position = "suffix") {
  table_colnames <- get_table_colnames(dm, tables, exclude_pk = FALSE)
  recipe <- compute_disambiguate_cols_recipe(table_colnames, sep = sep, position = position)
  if (!quiet) {
    explain_col_rename(recipe)
  }
  col_rename(dm, recipe)
}

get_table_colnames <- function(dm, tables = NULL, exclude_pk = TRUE) {
  def <- dm_get_def(dm)

  if (!is.null(tables)) {
    def <- def[def$table %in% tables, ]
  }

  table_colnames <-
    tibble(table = def$table, column = map(def$data, colnames)) %>%
    unnest_col("column", character())

  if (exclude_pk) {
    pks <- dm_get_all_pks_def_impl(def)

    keep_colnames <-
      pks[c("table", "pk_col", "autoincrement")] %>%
      set_names(c("table", "column", "autoincrement")) %>%
      unnest_col("column", character())

    table_colnames <-
      table_colnames %>%
      # in case of flattening, the primary key columns will never be responsible for the name
      # of the resulting column in the end, so they do not need to be disambiguated
      anti_join(keep_colnames, by = c("table", "column"))
  }

  table_colnames
}

#' create a disambiguation recipe tibble
#'
#' It will contain :
#'   * table: the table name
#'   * renames: a list of named symbols to be substituted in
#'     `db_rename(dm, tbl, new = old)`
#'   * name and a list of tibbles containing character cols `new_name` and `column`
#'     that will be used to print`db_rename` instructions through explain_col_rename
#' @param table_colnames a table containing table name and col names of dm
#' @param sep separator used to create new names for dupe cols
#' @noRd
#' @autoglobal
compute_disambiguate_cols_recipe <- function(table_colnames, sep, position = "suffix") {
  dupes <- vec_duplicate_detect(table_colnames$column)
  dup_colnames <- table_colnames[dupes, ]

  if (position == "prefix") {
    dup_colnames$new_name <- paste0(dup_colnames$table, sep, dup_colnames$column)
  } else {
    dup_colnames$new_name <- paste0(dup_colnames$column, sep, dup_colnames$table)
  }

  dup_data <- dup_colnames[c("new_name", "column")]
  dup_data$column_sym <- syms(dup_data$column)

  dup_nested <-
    vec_split(dup_data, dup_colnames$table) %>%
    set_names("table", "renames")

  dup_nested$names <- map(dup_nested$renames, select, new_name, column)
  dup_nested$renames <- map(dup_nested$renames, ~ deframe(select(., -column)))
  as_tibble(dup_nested)
}


#' Describe renaming of cols by printing code
#'
#' @param recipe created by `compute_disambiguate_cols_recipe`
#' @noRd
#' @autoglobal
explain_col_rename <- function(recipe) {
  if (nrow(recipe) == 0) {
    return()
  }

  disambiguation <-
    recipe %>%
    unnest(names) %>%
    mutate(
      text = glue(
        "dm_rename({tick_if_needed(table)}, {tick_if_needed(new_name)} = {tick_if_needed(column)})"
      )
    ) %>%
    pull(text)

  message("Renaming ambiguous columns: %>%\n  ", glue_collapse(disambiguation, " %>%\n  "))
}

col_rename <- function(dm, recipe) {
  reduce2(recipe$table, recipe$renames, ~ dm_rename(..1, !!..2, !!!..3), .init = dm)
}

#' Mark table for manipulation (v2)
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `dm_zoom2_to()` zooms to the given table, returning a keyed table
#' that has the dm object as an attribute.
#' Key column tracking (primary and foreign) is the responsibility
#' of the `dm_keyed_tbl` object.
#'
#' `dm_update_zoom2ed()` overwrites the originally zoomed table
#' with the manipulated table.
#'
#' `dm_insert_zoom2ed()` adds the manipulated table as a new table to the dm.
#'
#' `dm_discard_zoom2ed()` discards the zoomed table and returns the
#' original `dm` as it was before zooming.
#'
#' @inheritParams dm_zoom_to
#'
#' @return For `dm_zoom2_to()`: A `dm_keyed_tbl` object with zoom2 info
#'   stored in `dm_key_info`.
#'
#' @noRd
#' @examplesIf rlang::is_installed(c("nycflights13", "DiagrammeR"))
#' flights_keyed <- dm_zoom2_to(dm_nycflights13(), flights)
#'
#' flights_keyed
#'
#' flights_keyed_transformed <-
#'   flights_keyed %>%
#'   mutate(am_pm_dep = ifelse(dep_time < 1200, "am", "pm"))
#'
#' # replace table `flights` with the zoomed table
#' flights_keyed_transformed %>%
#'   dm_update_zoom2ed()
#'
#' # insert the zoomed table as a new table
#' flights_keyed_transformed %>%
#'   dm_insert_zoom2ed("extended_flights")
#'
#' # discard changes and return original dm
#' flights_keyed_transformed %>%
#'   dm_discard_zoom2ed()
dm_zoom2_to <- function(dm, table) {
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})

  keyed_tables <- dm_get_keyed_tables_impl(dm)
  keyed_tbl <- keyed_tables[[table_name]]

  # Store zoom2 info inside dm_key_info so it survives all dplyr/tidyr verbs
  keys_info <- keyed_get_info(keyed_tbl)
  keys_info$zoom2 <- list(dm = dm, table_name = table_name)
  attr(keyed_tbl, "dm_key_info") <- keys_info

  keyed_tbl
}

#' @rdname dm_zoom2_to
#' @param zoomed_tbl A `dm_keyed_tbl` object returned by `dm_zoom2_to()`
#'   or modified via dplyr operations.
#'
#' @return For `dm_update_zoom2ed()`, `dm_insert_zoom2ed()` and
#'   `dm_discard_zoom2ed()`: A `dm` object.
#'
#' @noRd
dm_update_zoom2ed <- function(zoomed_tbl) {
  zoom2_info <- zoom2_get_info(zoomed_tbl)
  dm <- zoom2_info$dm
  table_name <- zoom2_info$table_name

  keyed_tables <- dm_get_keyed_tables_impl(dm)
  keyed_tables[[table_name]] <- zoom2_clean_keys_info(zoomed_tbl)

  # Preserve table order by using names from original keyed_tables
  new_dm(keyed_tables)
}

#' @rdname dm_zoom2_to
#' @param new_tbl_name Name of the new table.
#' @inheritParams vctrs::vec_as_names
#'
#' @noRd
dm_insert_zoom2ed <- function(zoomed_tbl, new_tbl_name = NULL, repair = "unique", quiet = FALSE) {
  zoom2_info <- zoom2_get_info(zoomed_tbl)
  dm <- zoom2_info$dm
  table_name <- zoom2_info$table_name

  if (is.null(new_tbl_name)) {
    new_tbl_name <- table_name
  }

  keyed_tables <- dm_get_keyed_tables_impl(dm)

  # Repair names if needed
  names_list <- repair_table_names(
    old_names = names(keyed_tables),
    new_names = new_tbl_name,
    repair,
    quiet
  )

  # Rename existing tables if name repair occurred
  names(keyed_tables) <- names_list$old_new_names[names(keyed_tables)]
  new_tbl_name <- names_list$new_names

  # Add the new table
  keyed_tables[[new_tbl_name]] <- zoom2_clean_keys_info(zoomed_tbl)

  out <- new_dm(keyed_tables)

  # Transfer color from original table to new table
  orig_def <- dm_get_def(dm)
  orig_display <- orig_def$display[orig_def$table == table_name]
  if (!is.na(orig_display)) {
    out <- dm_set_colors(out, !!!set_names(new_tbl_name, orig_display))
  }

  out
}

#' @rdname dm_zoom2_to
#' @noRd
dm_discard_zoom2ed <- function(zoomed_tbl) {
  zoom2_info <- zoom2_get_info(zoomed_tbl)
  zoom2_info$dm
}


# Internal helpers --------------------------------------------------------

zoom2_get_info <- function(zoomed_tbl) {
  if (!is_dm_keyed_tbl(zoomed_tbl)) {
    cli::cli_abort("This object was not created by {.fn dm_zoom2_to}.")
  }

  keys_info <- keyed_get_info(zoomed_tbl)
  zoom2 <- keys_info$zoom2

  if (is.null(zoom2)) {
    cli::cli_abort("This object was not created by {.fn dm_zoom2_to}.")
  }

  zoom2
}

zoom2_clean_keys_info <- function(x) {
  if (is_dm_keyed_tbl(x)) {
    keys_info <- keyed_get_info(x)
    keys_info$zoom2 <- NULL
    attr(x, "dm_key_info") <- keys_info
  }
  x
}

test_that("dm_zoom2_to() works", {
  # returns a keyed table
  result <- dm_zoom2_to(dm_for_filter(), tf_1)

  expect_s3_class(result, "dm_keyed_tbl")

  # has zoom2 info in dm_key_info
  keys_info <- keyed_get_info(result)
  expect_false(is.null(keys_info$zoom2))
  expect_equal(keys_info$zoom2$table_name, "tf_1")
})


test_that("dm_zoom2_to() preserves table content", {
  expect_equivalent_tbl(
    dm_zoom2_to(dm_for_filter(), tf_2) %>%
      unclass_keyed_tbl(),
    tf_2()
  )

  expect_equivalent_tbl(
    dm_zoom2_to(dm_for_filter(), tf_3) %>%
      unclass_keyed_tbl(),
    tf_3()
  )
})


test_that("dm_insert_zoom2ed() works", {
  # test that a new tbl is inserted, based on the requested one
  expect_equivalent_dm(
    dm_zoom2_to(dm_for_filter(), tf_4) %>%
      dm_insert_zoom2ed("tf_4_new"),
    dm_for_filter() %>%
      dm(tf_4_new = tf_4()) %>%
      dm_add_pk(tf_4_new, h) %>%
      dm_add_fk(tf_4_new, c(j, j1), tf_3) %>%
      dm_add_fk(tf_5, l, tf_4_new),
    ignore_on_delete = TRUE,
    ignore_autoincrement = TRUE
  )

  # test that an error is thrown if 'repair = check_unique' and duplicate table names
  expect_dm_error(
    dm_zoom2_to(dm_for_filter(), tf_4) %>% dm_insert_zoom2ed("tf_4", repair = "check_unique"),
    "need_unique_names"
  )

  # test that in case of 'repair = unique' and duplicate table names -> renames of old and new
  expect_equivalent_dm(
    dm_for_filter() %>%
      dm_zoom2_to(tf_4) %>%
      dm_insert_zoom2ed("tf_4", repair = "unique", quiet = TRUE),
    dm_for_filter() %>%
      dm_rename_tbl(tf_4...4 = tf_4) %>%
      dm(tf_4...7 = tf_4()) %>%
      dm_add_pk(tf_4...7, h) %>%
      dm_add_fk(tf_4...7, c(j, j1), tf_3) %>%
      dm_add_fk(tf_5, l, tf_4...7),
    ignore_on_delete = TRUE,
    ignore_autoincrement = TRUE
  )
})


test_that("dm_update_zoom2ed() works", {
  # zooming to a table and updating should yield the same dm
  expect_equivalent_dm(
    dm_for_filter() %>%
      dm_zoom2_to(tf_2) %>%
      dm_update_zoom2ed(),
    dm_for_filter(),
    ignore_on_delete = TRUE,
    ignore_autoincrement = TRUE
  )
})


test_that("dm_update_zoom2ed() preserves mutated data", {
  original_dm <- dm_for_filter()
  result_dm <-
    original_dm %>%
    dm_zoom2_to(tf_2) %>%
    mutate(new_col = 1) %>%
    dm_update_zoom2ed()

  # The updated table should have the new column
  expect_true("new_col" %in% colnames(result_dm$tf_2))
})


test_that("dm_discard_zoom2ed() works", {
  original_dm <- dm_for_filter()
  result_dm <-
    original_dm %>%
    dm_zoom2_to(tf_2) %>%
    mutate(new_col = 1) %>%
    dm_discard_zoom2ed()

  # Should return the original dm unchanged
  expect_equivalent_dm(
    result_dm,
    original_dm
  )
  expect_false("new_col" %in% colnames(result_dm$tf_2))
})


test_that("dm_discard_zoom2ed() works after summarise", {
  original_dm <- dm_for_filter()
  result_dm <-
    original_dm %>%
    dm_zoom2_to(tf_2) %>%
    summarise(n = n()) %>%
    dm_discard_zoom2ed()

  expect_equivalent_dm(
    result_dm,
    original_dm
  )
})


# tests for compound keys -------------------------------------------------

test_that("zoom2 output for compound keys", {
  # FIXME: COMPOUND: Need proper test
  skip_if_remote_src()

  expect_snapshot({
    nyc_comp() %>% dm_zoom2_to(weather)
    nyc_comp() %>%
      dm_zoom2_to(weather) %>%
      dm_update_zoom2ed()
    nyc_comp_2 <-
      nyc_comp() %>%
      dm_zoom2_to(weather) %>%
      dm_insert_zoom2ed("weather_2")
    nyc_comp_2 %>%
      get_all_keys()

    nyc_comp_3 <-
      nyc_comp() %>%
      dm_zoom2_to(flights) %>%
      dm_insert_zoom2ed("flights_2")
    nyc_comp_3 %>%
      get_all_keys()
  })
})


# test that inserting a zoomed table retains the color --------------------

test_that("dm_insert_zoom2ed() retains color", {
  expect_identical(
    dm_for_filter() %>%
      dm_set_colors("cyan" = tf_2) %>%
      dm_zoom2_to(tf_2) %>%
      dm_insert_zoom2ed("tf_2_new") %>%
      dm_get_def() %>%
      filter(table == "tf_2_new") %>%
      pull(display),
    "#00FFFFFF"
  )
})


# dplyr verb snapshot tests -----------------------------------------------
# Each test constructs a dm inline, applies a dplyr verb via dm_zoom2_to(),
# and uses dm_paste() to show key structure after dm_update_zoom2ed().

# --- Verbs that change key columns ---

test_that("zoom2 select() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      select(id) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 select() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      select(child_id, val) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 rename() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      rename(parent_id = id) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 rename() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      rename(cid = child_id, pid = parent_id) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 summarise() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      summarise(n = n()) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 summarise() on child table with group_by", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      group_by(parent_id) %>%
      summarise(n = n()) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 summarise() insert", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      group_by(parent_id) %>%
      summarise(n = n()) %>%
      dm_insert_zoom2ed("child_summary") %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 summarise() on child table with .by", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      summarise(n = n(), .by = parent_id) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 summarise() insert with .by", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      summarise(n = n(), .by = parent_id) %>%
      dm_insert_zoom2ed("child_summary") %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 reframe() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      reframe(n = n()) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 reframe() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      reframe(n = n(), .by = parent_id) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 tally() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      tally() %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 tally() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      tally() %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 left_join()", {
  skip_if_remote_src()
  expect_snapshot({
    d <- dm(
      grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")),
      parent = tibble(id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(grandparent, gp_id) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(parent, gp_id, grandparent) %>%
      dm_add_fk(child, parent_id, parent)

    d %>%
      dm_zoom2_to(child) %>%
      left_join(dm_zoom2_to(d, parent)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 inner_join()", {
  skip_if_remote_src()
  expect_snapshot({
    d <- dm(
      grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")),
      parent = tibble(id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(grandparent, gp_id) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(parent, gp_id, grandparent) %>%
      dm_add_fk(child, parent_id, parent)

    d %>%
      dm_zoom2_to(child) %>%
      inner_join(dm_zoom2_to(d, parent)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 semi_join()", {
  skip_if_remote_src()
  expect_snapshot({
    d <- dm(
      grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")),
      parent = tibble(id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(grandparent, gp_id) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(parent, gp_id, grandparent) %>%
      dm_add_fk(child, parent_id, parent)

    d %>%
      dm_zoom2_to(child) %>%
      semi_join(dm_zoom2_to(d, parent)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 anti_join()", {
  skip_if_remote_src()
  expect_snapshot({
    d <- dm(
      grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")),
      parent = tibble(id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(grandparent, gp_id) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(parent, gp_id, grandparent) %>%
      dm_add_fk(child, parent_id, parent)

    d %>%
      dm_zoom2_to(child) %>%
      anti_join(dm_zoom2_to(d, parent)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 right_join()", {
  skip_if_remote_src()
  expect_snapshot({
    d <- dm(
      grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")),
      parent = tibble(id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(grandparent, gp_id) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(parent, gp_id, grandparent) %>%
      dm_add_fk(child, parent_id, parent)

    d %>%
      dm_zoom2_to(child) %>%
      right_join(dm_zoom2_to(d, parent)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 full_join()", {
  skip_if_remote_src()
  expect_snapshot({
    d <- dm(
      grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")),
      parent = tibble(id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(grandparent, gp_id) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(parent, gp_id, grandparent) %>%
      dm_add_fk(child, parent_id, parent)

    d %>%
      dm_zoom2_to(child) %>%
      full_join(dm_zoom2_to(d, parent)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 cross_join()", {
  skip_if_remote_src()
  expect_snapshot({
    d <- dm(
      grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")),
      parent = tibble(id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(grandparent, gp_id) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(parent, gp_id, grandparent) %>%
      dm_add_fk(child, parent_id, parent)

    d %>%
      dm_zoom2_to(child) %>%
      cross_join(dm_zoom2_to(d, parent)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 left_join() insert", {
  skip_if_remote_src()
  expect_snapshot({
    d <- dm(
      grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")),
      parent = tibble(id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(grandparent, gp_id) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(parent, gp_id, grandparent) %>%
      dm_add_fk(child, parent_id, parent)

    d %>%
      dm_zoom2_to(child) %>%
      left_join(dm_zoom2_to(d, parent)) %>%
      dm_insert_zoom2ed("child_parent") %>%
      dm_paste(options = "all")
  })
})

# --- Verbs that don't change key columns (parent + child) ---

test_that("zoom2 filter() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      filter(id > 1) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 filter() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      filter(val != "a") %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 mutate() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      mutate(name2 = toupper(name)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 mutate() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      mutate(val2 = toupper(val)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 arrange() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      arrange(desc(id)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 arrange() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      arrange(desc(child_id)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 distinct() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      distinct() %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 distinct() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      distinct() %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 slice() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      slice(1:2) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 slice() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      slice(1:3) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 transmute() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      transmute(id, name_upper = toupper(name)) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 transmute() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      transmute(child_id, parent_id) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 relocate() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      relocate(name, .before = id) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 relocate() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      relocate(val, .before = child_id) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 group_by() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      group_by(name) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 group_by() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      group_by(parent_id) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 ungroup() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      group_by(name) %>%
      ungroup() %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 ungroup() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      group_by(parent_id) %>%
      ungroup() %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 count() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      count(name) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 count() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      count(parent_id) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

# --- tidyr verbs ---

test_that("zoom2 unite() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, first = c("a", "b", "c"), last = c("x", "y", "z")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      tidyr::unite(full_name, first, last) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 unite() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(
        child_id = 1:4,
        parent_id = c(1L, 1L, 2L, 3L),
        first = c("a", "b", "c", "d"),
        last = c("w", "x", "y", "z")
      )
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      tidyr::unite(full_name, first, last) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 separate() on parent table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, full_name = c("a_x", "b_y", "c_z")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(parent) %>%
      tidyr::separate(full_name, into = c("first", "last")) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

test_that("zoom2 separate() on child table", {
  skip_if_remote_src()
  expect_snapshot({
    dm(
      parent = tibble(id = 1:3, name = c("a", "b", "c")),
      child = tibble(
        child_id = 1:4,
        parent_id = c(1L, 1L, 2L, 3L),
        full_val = c("a_1", "b_2", "c_3", "d_4")
      )
    ) %>%
      dm_add_pk(parent, id) %>%
      dm_add_pk(child, child_id) %>%
      dm_add_fk(child, parent_id, parent) %>%
      dm_zoom2_to(child) %>%
      tidyr::separate(full_val, into = c("letter", "number")) %>%
      dm_update_zoom2ed() %>%
      dm_paste(options = "all")
  })
})

old <- fs::dir_ls("R", glob = "R/gui-*.R")
fs::file_delete(old)

input <- setdiff(fs::dir_ls("../dmSVG/R"), "../dmSVG/R/dmSVG-package.R")

new_names <- paste0("gui-", fs::path_file(input))
new_paths <- fs::path("R", new_names)
fs::file_copy(input, new_paths)

if (fs::dir_exists("inst/htmlwidgets")) {
  fs::dir_delete("inst/htmlwidgets")
}

fs::dir_copy("../dmSVG/inst/htmlwidgets", "inst")

## SQL script backend for copy_dm_to()
## taken to new file to avoid merge conflicts to ongoing changes in db-interface.R #1739

#' @name dm_sql
#'
#' @title Create \emph{DDL} and \emph{DML} scripts for a `dm` a and database connection
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Generate SQL scripts to create tables, load data and set constraints, keys and indices.
#' This function powers [copy_dm_to()] and is useful if you need more control
#' over the process of copying a `dm` to a database.
#'
#' @param dm A `dm` object.
#' @param dest Connection to database.
#' @param table_names A named character vector or named vector of [DBI::Id],
#'   [DBI::SQL] or dbplyr objects created with [dbplyr::ident()], [dbplyr::in_schema()]
#'   or [dbplyr::in_catalog()],
#'   with one unique element for each table in `dm`.
#'   The default, `NULL`, means to use the original table names.
#' @param temporary Should the tables be marked as \emph{temporary}? Defaults to `TRUE`.
#'
#' @details
#' \itemize{
#'   \item{ `dm_ddl_pre()` generates `CREATE TABLE` statements (including `PRIMARY KEY` definition). }
#'   \item{ `dm_dml_load()` generates `INSERT INTO` statements. }
#'   \item{ `dm_ddl_post()` generates scripts for `FOREIGN KEY`, `UNIQUE KEY` and `INDEX`. }
#'   \item{ `dm_sql()` calls all three above and returns a complete set of scripts. }
#' }
#'
#' @return Nested list of SQL statements.
#'
#' @export
#' @examplesIf rlang::is_installed(c("RSQLite", "dbplyr"))
#' con <- DBI::dbConnect(RSQLite::SQLite())
#' dm <- dm_nycflights13()
#' s <- dm_sql(dm, con)
#' s
#' DBI::dbDisconnect(con)
dm_sql <- function(
  dm,
  dest,
  table_names = NULL,
  temporary = TRUE
) {
  #
  check_suggested("dbplyr", "dm_sql")

  stopifnot(is(dest, "DBIConnection"))

  table_names <- ddl_quote_table_names(table_names, dm, dest)

  dm <- ddl_reorder_dm(dm, dest)

  list(
    ## CREATE TABLE and PRIMARY KEY, FOREIGN KEYS and UNIQUE KEYS on DuckDB and SQLite
    pre = dm_ddl_pre(dm, dest, table_names, temporary),
    ## INSERT INTO, handle autoincrement
    load = dm_dml_load(dm, dest, table_names, temporary),
    ## FOREIGN KEYS and UNIQUE KEYS (except DuckDB and SQLite), INDEXES
    post = dm_ddl_post(dm, dest, table_names, temporary)
  )
}

ddl_quote_table_names <- function(table_names, dm, dest) {
  if (is.null(table_names)) {
    table_names <- set_names(names(dm))
  } else {
    stopifnot(names(dm) %in% names(table_names))
  }

  map(
    table_names,
    ~ {
      if (inherits(.x, "sql")) {
        # no op, need this because sql is also a character
      } else if (is.character(.x) || is(.x, "Id")) {
        .x <- DBI::dbQuoteIdentifier(dest, .x)
      }

      dbplyr::escape(.x, con = dest)
    }
  )
}

ddl_reorder_dm <- function(dm, con) {
  if (!ddl_need_early_constraints(con)) {
    return(dm)
  }

  ## For databases that do not support adding constraints after table creation,
  ## reorder queries according to topological sort so uks are created before associated fks
  graph <- create_graph_from_dm(dm, directed = TRUE)
  topo <- names(graph_topo_sort(graph, mode = "in"))

  if (length(topo) == length(dm)) {
    dm <- dm[topo]
  } else {
    cli_warn(c(
      # FIXME: show cycle
      "Cycles in the relationship graph prevent creation of foreign key constraints.",
      i = if (is_duckdb(dest)) "DuckDB does not support adding constraints after table creation.",
      i = if (is_sqlite(dest)) "SQLite does not support adding constraints after table creation.",
      NULL
    ))
    dm <- dm_rm_fk(dm)
  }

  dm
}

#' @rdname dm_sql
#' @export
#' @autoglobal
dm_ddl_pre <- function(
  dm,
  dest,
  table_names = NULL,
  temporary = TRUE
) {
  #
  check_suggested("dbplyr", "dm_ddl_pre")

  table_names <- ddl_quote_table_names(table_names, dm, dest)

  dm <- ddl_reorder_dm(dm, dest)

  ## use 0-rows object if from database
  if (is_src_db(dm)) {
    ptype_dm <- collect(dm_ptype(dm))
  } else {
    ptype_dm <- dm
  }

  con <- con_from_src_or_con(dest)

  ## fetch types, keys and uniques
  pks <- dm_get_all_pks_impl(ptype_dm)
  uks <- dm_get_all_uks_impl(ptype_dm)
  fks <- dm_get_all_fks_impl(ptype_dm)

  ## build sql definitions to use in `CREATE TABLE ...`

  tbl_defs <- tibble(name = names(ptype_dm))

  # column definitions
  col_defs <-
    ptype_dm %>%
    dm_get_tables() %>%
    ddl_get_col_defs(con, table_names, pks) %>%
    group_by(name) %>%
    summarize(
      col_defs = paste(col_def, collapse = ",\n  ")
    ) %>%
    ungroup()

  # primary key definitions
  pk_defs <-
    pks %>%
    ddl_get_pk_def(con, table_names)

  # Only add constraints if not adding them later
  if (!ddl_need_early_constraints(dest)) {
    uks <- vec_slice(uks, 0)
    fks <- vec_slice(fks, 0)
  }

  uk_defs <-
    uks %>%
    ddl_get_uk_defs(con, table_names) %>%
    group_by(name) %>%
    summarize(uk_defs = paste(uk_def, collapse = ",\n  ")) %>%
    ungroup()

  fk_defs <-
    fks %>%
    ddl_get_fk_defs(con, table_names) %>%
    group_by(name) %>%
    summarize(fk_defs = paste(fk_def, collapse = ",\n  ")) %>%
    ungroup()

  ## compile `CREATE TABLE ...` queries
  create_table_queries <-
    tbl_defs %>%
    left_join(col_defs, by = "name") %>%
    left_join(pk_defs, by = "name") %>%
    left_join(uk_defs, by = "name") %>%
    left_join(fk_defs, by = "name") %>%
    rowwise() %>%
    mutate(
      remote_name = table_names[name],
      all_defs = paste(
        Filter(
          Negate(is.na),
          c(col_defs, pk_def, uk_defs, fk_defs)
        ),
        collapse = ",\n  "
      )
    ) %>%
    ungroup() %>%
    transmute(
      name,
      remote_name,
      sql_table = DBI::SQL(glue(
        "CREATE {if (temporary) 'TEMPORARY ' else ''}TABLE {remote_name} (\n  {all_defs}\n)"
      ))
    )

  set_names(map(create_table_queries$sql_table, DBI::SQL), create_table_queries$name)
}

#' @rdname dm_sql
#' @export
dm_dml_load <- function(
  dm,
  dest,
  table_names = NULL,
  temporary = TRUE
) {
  #
  check_suggested("dbplyr", "dm_dml_load")

  table_names <- ddl_quote_table_names(table_names, dm, dest)

  dm <- ddl_reorder_dm(dm, dest)

  if (is_mssql(dest)) {
    pks <- dm_get_all_pks_impl(dm)
  } else {
    pks <- NULL
  }

  dm %>%
    names() %>%
    set_names() %>%
    map(dml_tbl_load, dm, table_names, pks, dest) %>%
    compact()
}

#' @rdname dm_sql
#' @export
dm_ddl_post <- function(
  dm,
  dest,
  table_names = NULL,
  temporary = TRUE
) {
  #
  check_suggested("dbplyr", "dm_ddl_post")

  table_names <- ddl_quote_table_names(table_names, dm, dest)

  ## use 0-rows object
  ptype_dm <- collect(dm_ptype(dm))

  con <- con_from_src_or_con(dest)

  ## fetch types, keys and uniques
  uks <- dm_get_all_uks_impl(ptype_dm)
  fks <- dm_get_all_fks_impl(ptype_dm)

  ## build sql definitions to use in `CREATE TABLE ...`

  tbl_defs <- tibble(name = names(ptype_dm))

  # Get this before we perhaps erase fks
  index_defs <- ddl_get_index_defs(fks, con, table_names)

  # unique constraint definitions
  if (nrow(uks) == 0) {
    # No action
  } else if (ddl_need_early_constraints(dest)) {
    uks <- vec_slice(uks, 0)
  }

  uk_defs <-
    ddl_get_uk_defs(uks, con, table_names) %>%
    group_by(name) %>%
    summarize(
      uk_defs = if (length(remote_name) == 0) {
        list(NULL)
      } else {
        list(DBI::SQL(glue(
          # FIXME: Designate temporary table if possible
          "ALTER TABLE {remote_name[[1]]} ADD {uk_def}"
        )))
      }
    ) %>%
    ungroup()

  # foreign key definitions and indexing queries
  # https://github.com/r-lib/rlang/issues/1422
  if (nrow(fks) == 0) {
    # No action
  } else if (is_mariadb(con) && temporary) {
    if (!is_testing()) {
      warn(
        "MySQL and MariaDB don't support foreign keys for temporary tables, these won't be set in the remote database but are preserved in the `dm`"
      )
    }
    fks <- vec_slice(fks, 0)
  } else if (ddl_need_early_constraints(dest)) {
    fks <- vec_slice(fks, 0)
  }

  fk_defs <-
    ddl_get_fk_defs(fks, con, table_names) %>%
    group_by(name) %>%
    summarize(
      fk_defs = if (length(remote_name) == 0) {
        list(NULL)
      } else {
        list(DBI::SQL(glue(
          # FIXME: Designate temporary table if possible
          "ALTER TABLE {remote_name[[1]]} ADD {fk_def}"
        )))
      }
    ) %>%
    ungroup()

  queries <-
    tbl_defs %>%
    left_join(uk_defs, by = "name") %>%
    left_join(fk_defs, by = "name") %>%
    left_join(index_defs, by = "name")

  list(
    uk = compact(set_names(queries$uk_defs, queries$name)),
    fk = compact(set_names(queries$fk_defs, queries$name)),
    indexes = compact(set_names(queries$index_defs, queries$name))
  )
}

ddl_need_early_constraints <- function(con) {
  is_duckdb(con) || is_sqlite(con)
}

ddl_quote_enum_col <- function(x, con) {
  map_chr(x, ~ paste(DBI::dbQuoteIdentifier(.x, conn = con), collapse = ", "))
}

ddl_get_col_defs <- function(tables, con, table_names, pks) {
  duckdb_auto_warned <- FALSE

  get_sql_col_types <- function(tbl, name) {
    # autoincrementing is not possible for composite keys, so `pk_col` is guaranteed
    # to be a scalar
    pk_col <- vec_slice(pks, pks$table == name)

    types <- DBI::dbDataType(con, tbl)

    # database-specific type conversions
    if (is_mariadb(con)) {
      # FIXME: This is wrong in general, only needed for index columns
      if (nrow(tbl) == 0) {
        types[types == "TEXT"] <- "VARCHAR(255)"
      } else {
        for (i in which(types == "TEXT")) {
          chars <- max(nchar(tbl[[i]], type = "bytes", keepNA = FALSE))
          if (chars <= 255) {
            types[[i]] <- paste0("VARCHAR(255)")
          }
        }
      }
    }
    if (is_sqlite(con)) {
      types[types == "INT"] <- "INTEGER"
    }

    # database-specific autoincrementing column types
    if (isTRUE(pk_col$autoincrement)) {
      # extract column name representing primary key
      pk_col_name <- pk_col$pk_col[[1]]

      # Postgres/Redshift:
      if (is_postgres(con) || is_redshift(con)) {
        types[pk_col_name] <- "SERIAL"
      }

      # SQL Server:
      if (is_mssql(con)) {
        types[pk_col_name] <- paste0(types[pk_col_name], " IDENTITY")
      }

      # MariaDB:
      # Doesn't have a special data type. Uses `AUTO_INCREMENT` attribute instead.
      # Ref: https://mariadb.com/kb/en/auto_increment/
      if (is_mariadb(con)) {
        types[pk_col_name] <- paste0(types[pk_col_name], " AUTO_INCREMENT")
      }

      # DuckDB:
      # Doesn't have a special data type. Uses `CREATE SEQUENCE` instead.
      # Ref: https://duckdb.org/docs/sql/statements/create_sequence
      # https://stackoverflow.com/a/72883259/946850
      if (is_duckdb(con)) {
        if (!duckdb_auto_warned) {
          cli_warn(
            "Autoincrementing columns not yet supported for DuckDB, these won't be set in the remote database but are preserved in the `dm`"
          )
          duckdb_auto_warned <<- TRUE
        }
      }

      # SQLite:
      # For a primary key, autoincrementing works by default, and it is almost never
      # necessary to use the `AUTOINCREMENT` keyword. So nothing we need to do here.
      # Ref: https://www.sqlite.org/autoinc.html
    }

    enframe(types, "col", "type")
  }

  tables %>%
    imap(get_sql_col_types) %>%
    list_rbind(names_to = "name") %>%
    mutate(col_def = glue("{DBI::dbQuoteIdentifier(con, col)} {type}"))
}

ddl_get_pk_def <- function(pks, con, table_names) {
  pk_defs <- tibble(
    name = character(),
    remote_name = unname(vec_ptype(table_names)),
    pk_def = character()
  )
  if (nrow(pks) == 0) {
    return(pk_defs)
  }

  out <-
    pks %>%
    rename(name = table) %>%
    transmute(
      name,
      remote_name = unname(table_names[name]),
      pk_def = paste0("PRIMARY KEY (", ddl_quote_enum_col(pk_col, con), ")")
    )

  vec_assert(out, pk_defs)
  out
}

ddl_get_uk_defs <- function(uks, con, table_names) {
  uk_defs <- tibble(
    name = character(),
    remote_name = unname(vec_ptype(table_names)),
    uk_def = character()
  )
  if (nrow(uks) == 0) {
    return(uk_defs)
  }

  out <-
    uks %>%
    rename(name = table) %>%
    transmute(
      name,
      remote_name = unname(table_names[name]),
      uk_def = paste0("UNIQUE (", ddl_quote_enum_col(uk_col, con), ")")
    )

  vec_assert(out, uk_defs)
  out
}

ddl_get_fk_defs <- function(fks, con, table_names) {
  fk_defs <- tibble(
    name = character(),
    remote_name = unname(vec_ptype(table_names)),
    fk_def = character()
  )
  if (nrow(fks) == 0) {
    return(fk_defs)
  }

  ## helper to set on delete statement for fks if required
  set_on_delete_col <- function(x) {
    if (is_duckdb(con) && any(x == "cascade")) {
      inform(glue('`on_delete = "cascade"` not supported for duckdb'))
      ""
    } else {
      map_chr(
        x,
        ~ {
          switch(
            .x,
            "no_action" = "",
            "cascade" = " ON DELETE CASCADE",
            cli::cli_abort('{.code on_delete = {.val {.x}}} is not supported.')
          )
        }
      )
    }
  }

  out <-
    fks %>%
    transmute(
      name = child_table,
      remote_name = unname(table_names[name]),
      fk_def = paste0(
        "FOREIGN KEY (",
        ddl_quote_enum_col(child_fk_cols, con),
        ") REFERENCES ",
        table_names[fks$parent_table],
        " (",
        ddl_quote_enum_col(parent_key_cols, con),
        ")",
        set_on_delete_col(on_delete)
      )
    )

  vec_assert(out, fk_defs)
  out
}

dml_tbl_load <- function(tbl_name, dm, table_names, pks, dest) {
  tbl <- collect(dm[[tbl_name]])
  if (nrow(tbl) == 0) {
    return(NULL)
  }

  # pks is NULL if not mssql
  mssql_autoinc <- !is.null(pks) && isTRUE(pks$autoincrement[tbl_name == pks$table])

  remote_name <- table_names[[tbl_name]]
  selectvals <- dbplyr::sql_render(dbplyr::copy_inline(dest, tbl))

  if (is_mariadb(dest)) {
    # Work around https://github.com/tidyverse/dbplyr/pull/1195
    selectvals <- strsplit(selectvals, "\n", fixed = TRUE)[[1]]
    from <- grep("^FROM", selectvals)[[1]]
    idx <- seq_len(from - 1L)
    # https://github.com/tidyverse/dbplyr/pull/1195
    selectvals[idx] <- gsub(" AS NUMERIC", " AS DOUBLE", selectvals[idx])
    # https://github.com/tidyverse/dbplyr/pull/1375
    selectvals[idx] <- gsub("CAST[(]`[^`]*(?:``[^`]*)*` AS INTEGER[)] AS ", "", selectvals[idx])
    selectvals <- paste(selectvals, collapse = "\n")
  }

  ## for some DBes we could skip columns specification, but only when no autoincrement, easier to just specify that always
  ## duckdb autoincrement tests are skipped, could be added by inserting from sequence
  out <- paste0(
    "INSERT INTO ",
    remote_name,
    " (",
    paste(DBI::dbQuoteIdentifier(dest, names(tbl)), collapse = ", "),
    ")\n",
    selectvals
  )
  DBI::SQL(paste0(
    if (mssql_autoinc) paste0("SET IDENTITY_INSERT ", remote_name, " ON\n"),
    out,
    if (mssql_autoinc) paste0("\nSET IDENTITY_INSERT ", remote_name, " OFF")
  ))
}

ddl_get_index_defs <- function(fks, con, table_names) {
  index_defs <- tibble(
    name = character(0),
    index_defs = list(),
    index_name = list()
  )
  if (nrow(fks) == 0) {
    return(index_defs)
  }

  out <-
    fks %>%
    mutate(
      name = child_table,
      index_name = map_chr(child_fk_cols, paste, collapse = "_"),
      remote_name = table_names[name],
      remote_name_unquoted = map_chr(
        DBI::dbUnquoteIdentifier(con, DBI::SQL(remote_name)),
        ~ .x@name[[length(.x@name)]]
      ),
      index_name = make.unique(paste0(remote_name_unquoted, "__", index_name), sep = "__")
    ) %>%
    group_by(name) %>%
    summarize(
      index_defs = list(DBI::SQL(paste0(
        "CREATE INDEX ",
        index_name,
        " ON ",
        remote_name,
        " (",
        ddl_quote_enum_col(child_fk_cols, con),
        ")"
      ))),
      index_name = list(index_name)
    )

  vec_assert(out, index_defs)
  out
}

test_that("dummy", {
  expect_snapshot({
    "dummy"
  })
})

test_that("dm_sql()", {
  # Need skip in every test block, unfortunately
  skip_if_src_not("mssql")

  # https://github.com/tidyverse/dbplyr/pull/1190
  skip_if(is(my_test_con(), "MySQLConnection") && packageVersion("dbplyr") < "2.4.0")

  expect_snapshot({
    dm_for_filter_df() %>%
      dm_sql(my_test_con())
  })
})

# From \d+ information_schema.constraint_column_usage, adapted to include colnum
postgres_column_constraints <- "SELECT current_database()::information_schema.sql_identifier AS table_catalog,
    x.tblschema::information_schema.sql_identifier AS table_schema,
    x.tblname::information_schema.sql_identifier AS table_name,
    x.colname::information_schema.sql_identifier AS column_name,
    current_database()::information_schema.sql_identifier AS constraint_catalog,
    x.cstrschema::information_schema.sql_identifier AS constraint_schema,
    x.cstrname::information_schema.sql_identifier AS constraint_name,
    x.colnum::information_schema.cardinal_number AS ordinal_position

   FROM ( SELECT DISTINCT nr.nspname,
            r.relname,
            r.relowner,
            a.attname,
            NULL::smallint AS attnum,
            nc.nspname,
            c.conname
           FROM pg_namespace nr,
            pg_class r,
            pg_attribute a,
            pg_depend d,
            pg_namespace nc,
            pg_constraint c
          WHERE nr.oid = r.relnamespace AND r.oid = a.attrelid AND d.refclassid = 'pg_class'::regclass::oid AND d.refobjid = r.oid AND d.refobjsubid = a.attnum AND d.classid = 'pg_constraint'::regclass::oid AND d.objid = c.oid AND c.connamespace = nc.oid AND c.contype = 'c'::\"char\" AND (r.relkind = ANY (ARRAY['r'::\"char\", 'p'::\"char\"])) AND NOT a.attisdropped
        UNION ALL
         SELECT nr.nspname,
            r.relname,
            r.relowner,
            a.attname,
            array_position(
              CASE c.contype
                    WHEN 'f'::\"char\" THEN c.confkey
                    ELSE c.conkey
              END,
              a.attnum
            ) AS attnum,
            nc.nspname,
            c.conname
           FROM pg_namespace nr,
            pg_class r,
            pg_attribute a,
            pg_namespace nc,
            pg_constraint c
          WHERE nr.oid = r.relnamespace AND r.oid = a.attrelid AND nc.oid = c.connamespace AND r.oid =
                CASE c.contype
                    WHEN 'f'::\"char\" THEN c.confrelid
                    ELSE c.conrelid
                END AND (a.attnum = ANY (
                CASE c.contype
                    WHEN 'f'::\"char\" THEN c.confkey
                    ELSE c.conkey
                END)) AND NOT a.attisdropped AND (c.contype = ANY (ARRAY['f'::\"char\"])) AND (r.relkind = ANY (ARRAY['r'::\"char\", 'p'::\"char\"]))) x(tblschema, tblname, tblowner, colname, colnum, cstrschema, cstrname)
  WHERE colnum IS NOT NULL
"

# FIXME: Slightly redundant `array_position()` and `a.attnum = ANY`

#' Check foreign key reference
#'
#' `dm_is_referenced()` is soft-deprecated, use the information returned from
#' [dm_get_all_fks()] instead.
#'
#' @inheritParams dm_add_pk
#'
#' @rdname deprecated
#' @export
dm_is_referenced <- function(dm, table) {
  deprecate_warn("0.3.0", "dm::dm_is_referenced()", "dm::dm_get_all_fks()")

  check_not_zoomed(dm)
  has_length(dm_get_referencing_tables(dm, !!ensym(table)))
}

#' Get the names of referencing tables
#'
#' `dm_get_referencing_tables()` is soft-deprecated, use the information
#' returned from [dm_get_all_fks()] instead.
#'
#' @inheritParams dm_is_referenced
#' @rdname deprecated
#'
#' @export
dm_get_referencing_tables <- function(dm, table) {
  deprecate_warn("0.3.0", "dm::dm_get_referencing_tables()", "dm::dm_get_all_fks()")

  check_not_zoomed(dm)
  table <- dm_tbl_name(dm, {{ table }})

  def <- dm_get_def(dm)
  i <- which(def$table == table)
  def$fks[[i]]$table
}

#' @autoglobal
create_graph_from_dm <- function(dm, directed = FALSE) {
  def <- dm_get_def(dm)
  def %>%
    select(ref_table = table, fks) %>%
    unnest_list_of_df("fks") %>%
    select(table, ref_table) %>%
    graph_from_data_frame(directed = directed, vertices = def$table)
}

get_names_of_connected <- function(g, start, squash) {
  dfs <- graph_dfs(g, start, unreachable = FALSE, dist = TRUE)
  # `purrr::discard()` in case `list_of_pts` is `NA`
  if (squash) {
    setdiff(names(dfs[["order"]]), start) %>% discard(is.na)
  } else {
    # FIXME: Enumerate outgoing edges
    setdiff(names(dfs[["order"]]), c(start, names(dfs$dist[dfs$dist > 1]))) %>% discard(is.na)
  }
}

data <- tibble(a = 1:3, b = letters[c(1:2, NA)], c = 0.5 + 0:2)
data

rows_insert(data, tibble(a = 4, b = "z"))
rows_update(data, tibble(a = 2:3, b = "z"))
rows_update(data, tibble(b = "z", a = 2:3), by = "a")
rows_patch(data, tibble(a = 2:3, b = "z"))
rows_upsert(data, tibble(a = 2:4, b = "z"))
rows_delete(data, tibble(a = 2:4))
rows_delete(data, tibble(a = 2:4, b = "b"))
rows_delete(data, tibble(a = 2:4, b = "b"), by = c("a", "b"))
rows_truncate(data)

# nocov start
replace_if_dplyr_has <- function(fun) {
  dplyr_ns <- asNamespace("dplyr")

  fun <- as_string(ensym(fun))
  value <- mget(fun, dplyr_ns, mode = "function", ifnotfound = list(NULL))[[1]]
  if (!is.null(value)) {
    assign(fun, value, inherits = TRUE)
    "dplyr"
  } else {
    "dm"
  }
}

register_if_dplyr_hasnt <- function(...) {
  dplyr_ns <- asNamespace("dplyr")

  # Register our method implementations only if dplyr doesn't provide them
  methods <- enquos(..., .named = TRUE)
  dplyr_methods <- mget(names(methods), dplyr_ns, mode = "function", ifnotfound = list(NULL))
  methods <- methods[map_lgl(dplyr_methods, is.null)]

  if (is_empty(methods)) {
    return()
  }

  methods <- map(methods, eval_tidy)
  classes <- sub("^[^.]*.", "", names(methods))
  fun <- sub("[.].*$", "", names(methods)[[1]])

  map2(classes, methods, s3_register, generic = paste0("dm::", fun))
  invisible()
}
# nocov end

# FIXME: Adapt argument list as provided/required by the Shiny app
mw_cg_make_dm_rm_fk <- function(dm, ..., edge_table) {
  check_dots_empty()

  # Checks
  stopifnot(is.character(edge_table$child_table))
  stopifnot(is.character(edge_table$parent_table))
  stopifnot(edge_table$child_table %in% names(dm))
  stopifnot(edge_table$parent_table %in% names(dm))

  # FIXME: check also cols
  purrr::pmap(
    edge_table,
    function(child_table, child_fk_cols, parent_table, parent_key_cols) {
      if (length(child_fk_cols) > 1) {
        call <- expr(dm_rm_fk(
          .,
          table = !!sym(child_table),
          columns = c(!!!syms(child_fk_cols)),
          ref_table = !!sym(parent_table),
          ref_columns = c(!!!syms(parent_key_cols))
        ))
      } else {
        call <- expr(dm_rm_fk(
          .,
          table = !!sym(child_table),
          columns = !!sym(child_fk_cols),
          ref_table = !!sym(parent_table),
          ref_columns = !!sym(parent_key_cols)
        ))
      }
    }
  )
}

test_src_df <- function() {
  NULL
}

test_src_sqlite <- function() {
  testthat::skip_if_not_installed("DBI")
  testthat::skip_if_not_installed("dbplyr")
  testthat::skip_if_not_installed("RSQLite")
  dbplyr::src_dbi(DBI::dbConnect(RSQLite::SQLite(), ":memory:"), auto_disconnect = TRUE)
}

test_src_duckdb <- function() {
  testthat::skip_if_not_installed("DBI")
  testthat::skip_if_not_installed("dbplyr")
  testthat::skip_if_not_installed("duckdb")
  dbplyr::src_dbi(DBI::dbConnect(duckdb::duckdb()), auto_disconnect = TRUE)
}

test_src_postgres <- function() {
  # Check for PostgreSQL-specific host first, then fall back to general docker host
  postgres_host <- Sys.getenv("DM_TEST_POSTGRES_HOST")
  docker_host <- Sys.getenv("DM_TEST_DOCKER_HOST")

  if (postgres_host != "" || docker_host != "") {
    host <- if (postgres_host != "") postgres_host else docker_host
    con <- DBI::dbConnect(
      RPostgres::Postgres(),
      host = host,
      user = "compose",
      password = "YourStrong!Passw0rd"
    )
  } else {
    con <- DBI::dbConnect(RPostgres::Postgres())
  }
  dbplyr::src_dbi(con, auto_disconnect = TRUE)
}

test_src_maria <- function(root = FALSE) {
  # Check for MariaDB-specific host first, then fall back to general docker host
  maria_host <- Sys.getenv("DM_TEST_MARIA_HOST")
  docker_host <- Sys.getenv("DM_TEST_DOCKER_HOST")

  if (maria_host != "" || docker_host != "") {
    host <- if (maria_host != "") maria_host else docker_host
    con <- DBI::dbConnect(
      RMariaDB::MariaDB(),
      host = host,
      username = if (root) "root" else "compose",
      password = "YourStrong!Passw0rd",
      dbname = "test"
    )
  } else {
    con <- DBI::dbConnect(RMariaDB::MariaDB(), dbname = "test")
  }
  dbplyr::src_dbi(con, auto_disconnect = TRUE)
}

test_src_mssql <- function(database = TRUE) {
  # Check for SQL Server-specific host first, then fall back to general docker host
  mssql_host <- Sys.getenv("DM_TEST_MSSQL_HOST")
  docker_host <- Sys.getenv("DM_TEST_DOCKER_HOST")

  if (mssql_host != "" || docker_host != "") {
    host <- if (mssql_host != "") mssql_host else docker_host
    con <- DBI::dbConnect(
      odbc::odbc(),
      driver = "ODBC Driver 18 for SQL Server",
      server = host,
      database = if (database) "test",
      uid = "SA",
      pwd = "YourStrong!Passw0rd",
      port = 1433,
      TrustServerCertificate = "yes"
    )
  } else if (Sys.getenv("CI") != "") {
    con <- DBI::dbConnect(
      odbc::odbc(),
      "mssql-test",
      uid = "SA",
      pwd = "YourStrong!Passw0rd",
      port = 1433
    )
  } else {
    con <- DBI::dbConnect(
      odbc::odbc(),
      "dm-test",
      uid = keyring::key_get("mssql", "dm-test-user"),
      pwd = keyring::key_get("mssql", "dm-test-password")
    )
  }

  DBI::dbExecute(con, "SET IMPLICIT_TRANSACTIONS OFF", immediate = TRUE)

  dbplyr::src_dbi(con, auto_disconnect = TRUE)
}

#' Learn about your data model
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' This function returns a tibble with information about
#' the cardinality of the FK constraints.
#' The printing for this object is special, use [as_tibble()]
#' to print as a regular tibble.
#'
# Can't @inheritParams dm_examine_constraints for some reason
#' @param .dm A `dm` object.
#' @inheritParams rlang::args_dots_empty
#' @param .progress Whether to display a progress bar, if `NA` (the default)
#'   hide in non-interactive mode, show in interactive mode. Requires the
#'   'progress' package.
#' @param dm,progress `r lifecycle::badge("deprecated")`
#'
#' @return A tibble with the following columns:
#'   \describe{
#'     \item{`child_table`}{child table,}
#'     \item{`child_fk_cols`}{foreign key column(s) in child table as list of character vectors,}
#'     \item{`parent_table`}{parent table,}
#'     \item{`parent_key_cols`}{key column(s) in parent table as list of character vectors,}
#'     \item{`cardinality`}{the nature of cardinality along the foreign key.}
#'   }
#'
#' @details Uses [examine_cardinality()] on each foreign key that is defined in the [`dm`].
#'
#' @family cardinality functions
#'
#' @export
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_examine_cardinalities()
dm_examine_cardinalities <- function(
  .dm,
  ...,
  .progress = NA,
  dm = deprecated(),
  progress = deprecated()
) {
  check_dots_empty()
  dm_local_error_call()

  if (!is_missing(dm)) {
    deprecate_warn("1.0.0", "dm_examine_cardinalities(dm = )", "dm_examine_cardinalities(.dm = )")
  }

  if (is_missing(.dm)) {
    return(dm_examine_cardinalities(dm, .progress = .progress, progress = progress))
  }

  if (!is_missing(progress)) {
    if (is.na(progress)) {
      progress <- .progress
    }
    deprecate_warn(
      "1.0.0",
      "dm_examine_cardinalities(progress = )",
      "dm_examine_cardinalities(.progress = )"
    )
  }

  check_not_zoomed(.dm)
  .dm %>%
    dm_examine_cardinalities_impl(
      progress = .progress,
      top_level_fun = "dm_examine_cardinalities"
    ) %>%
    new_dm_examine_cardinalities()
}

#' @autoglobal
dm_examine_cardinalities_impl <- function(dm, progress = NA, top_level_fun = NULL) {
  fks <-
    dm_get_all_fks_impl(dm) %>%
    select(-on_delete)

  dm_def <- as.list(dm)
  fks_data <-
    fks %>%
    transmute(
      x_label = parent_table,
      y_label = child_table,
      x = map2(dm_def[x_label], parent_key_cols, ~ select(.x, all_of(.y))),
      y = map2(dm_def[y_label], child_fk_cols, ~ select(.x, all_of(.y))),
    )
  ticker <- new_ticker(
    "checking fk cardinalities",
    n = nrow(fks),
    progress = progress,
    top_level_fun = top_level_fun
  )

  fks %>%
    mutate(cardinality = pmap_chr(fks_data, ticker(examine_cardinality_impl0)))
}

new_dm_examine_cardinalities <- function(x) {
  class(x) <- c("dm_examine_cardinalities", class(x))
  x
}

#' @export
print.dm_examine_cardinalities <- function(x, ...) {
  if (nrow(x) == 0) {
    cli::cli_alert_warning("No FKs available in `dm`.")
    return(invisible(x))
  }
  x %>%
    mutate(
      cardinalities = pmap_chr(
        x,
        function(parent_table, parent_key_cols, child_table, child_fk_cols, cardinality) {
          paste0(
            "FK: ",
            child_table,
            "$(",
            commas(tick(child_fk_cols)),
            ") -> ",
            parent_table,
            "$(",
            commas(tick(parent_key_cols)),
            "): ",
            cardinality
          )
        }
      )
    ) %>%
    bullets_cardinalities()
}

#' @autoglobal
bullets_cardinalities <- function(x) {
  x <- mutate(
    x,
    col = if_else(grepl("mapping", cardinality), "black", "red")
  ) %>%
    arrange(col)
  walk2(x$cardinalities, x$col, ~ cli::cat_bullet(.x, bullet_col = .y))
  if (sum(x$col == "red") > 0) {
    cli::cli_alert_warning(
      "Not all FK constraints satisfied, call `dm_examine_constraints()` for details."
    )
  }
  invisible(x)
}

test_that("dm_examine_cardinalities() works", {
  expect_snapshot({
    dm_examine_cardinalities(dm_for_card())
    dm_examine_cardinalities(dm_for_card()) %>% as_tibble()
    dm_for_card() %>%
      dm_rm_fk(dc_6, c, dc_1, a) %>%
      dm_rm_fk(dc_4, c(b, a), dc_3, c(b, a)) %>%
      dm_examine_cardinalities()
    dm_examine_cardinalities(dm())
  })
})

test_that("`dm_examine_cardinalities()` API", {
  local_options(lifecycle_verbosity = "warning")

  expect_snapshot({
    dm_examine_cardinalities(dm_test_obj(), progress = FALSE)
    dm_examine_cardinalities(dm = dm_test_obj())
  })
})

test_that("`dm_examine_cardinalities()` API (2)", {
  expect_snapshot(error = TRUE, {
    dm_examine_cardinalities(dm_test_obj(), foo = "bar")
  })
})

# internal helper functions:

# validates, that `table` is character and is part of the `dm` object
dm_tbl_name <- function(dm, table) {
  table_name <- as_name(ensym(table))

  # Missing argument?
  if (table_name == "") {
    arg_name <- deparse(uncurly(substitute(table)))
    abort_table_missing(arg_name)
  }

  if (!(table_name %in% src_tbls_impl(dm))) {
    abort_table_not_in_dm(table_name, src_tbls_impl(dm))
  }

  table_name
}

dm_tbl_name_null <- function(dm, table) {
  table_expr <- enexpr(table)

  # Missing argument?
  if (quo_is_null(table_expr)) {
    return(NULL)
  }

  table_name <- as_name(table_expr)

  if (!(table_name %in% src_tbls_impl(dm))) {
    abort_table_not_in_dm(table_name, src_tbls_impl(dm))
  }

  table_name
}

uncurly <- function(call) {
  # Transforms {{ x }} to x
  # Doesn't work for other expression patterns
  call[[2]][[2]]
}

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

is_this_a_test <- function() {
  # Only run if the top level call is devtools::test() or testthat::test_check()

  calls <-
    sys.calls() %>%
    as.list() %>%
    map(as.list) %>%
    map(1) %>%
    map_chr(as_label)

  is_test_call <- any(
    calls %in%
      c("devtools::test", "testthat::test_check", "testthat::test_file", "testthis:::test_this")
  )

  is_testing <- rlang::is_installed("testthat") && testthat::is_testing()

  is_test_call || is_testing
}


# more general 'check'-type functions -------------------------------------

check_param_class <- function(
  param_value,
  correct_class,
  param_name = deparse(substitute(param_value))
) {
  if (!inherits(param_value, correct_class)) {
    abort_parameter_not_correct_class(
      parameter = param_name,
      correct_class = correct_class,
      class = class(param_value)
    )
  }
}

check_param_length <- function(
  param_value,
  correct_length = 1,
  param_name = deparse(substitute(param_value))
) {
  if (length(param_value) != correct_length) {
    abort_parameter_not_correct_length(
      parameter = param_name,
      correct_length = correct_length,
      param_value
    )
  }
}

# general error: table not part of `dm` -----------------------------------

abort_table_missing <- function(arg_name) {
  cli::cli_abort(
    "Must pass {.arg {arg_name}} argument.",
    class = dm_error_full("table_missing"),
    call = dm_error_call()
  )
}

abort_table_not_in_dm <- function(table_name, dm_tables) {
  cli::cli_abort(
    "Table {.field {table_name}} not in {.cls dm} object. Available table names: {.field {dm_tables}}.",
    class = dm_error_full("table_not_in_dm"),
    call = dm_error_call()
  )
}

test_that("general check-functions work", {
  expect_silent(check_param_class(1, "numeric", "one"))
  expect_dm_error(check_param_class("1", "numeric", "one"), "parameter_not_correct_class")
  expect_silent(check_param_length(1:3, 3, "count_to_three"))
  expect_dm_error(check_param_length(1:4, 3, "count_to_three"), "parameter_not_correct_length")
})

#' dm as data source
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' These methods are deprecated because of their limited use,
#' and because the notion of a "source" seems to be getting phased out from dplyr.
#' Use other ways to access the tables in a `dm`.
#'
#' @details
#' Use [dm_get_con()] instead of `dm_get_src()` to get the DBI connetion for a
#' `dm` object
#'
#' @name dplyr_src
#' @export
#' @keywords internal
dm_get_src <- function(x) {
  deprecate_warn(
    "0.2.0",
    "dm::dm_get_src()",
    details = "Use `dm_get_con(dm)` for databases, or `class(dm[[1]])` to get the class of a table."
  )

  check_not_zoomed(x)
  dm_get_src_impl(x)
}

#' @details
#' Use [`[[`][base::Extract] instead of `tbl()` to access individual tables in a `dm` object.
#' @param src A `dm` object.
#' @param from A length one character variable containing the name of the requested table
#' @param ... See original function documentation
#' @export
#' @rdname dplyr_src
#' @keywords internal
tbl.dm <- function(src, from, ...) {
  deprecate_warn(
    "0.2.0",
    "dm::tbl.dm()",
    details = "Use `dm[[table_name]]` instead to access a specific table."
  )

  check_not_zoomed(src)

  # The src argument here is a dm object
  tbl_impl(src, from)
}

#' @details
#' Get the names from [dm_get_tables()] instead of calling `dm_get_src()`
#' to list the table names in a `dm` object.
#' @rdname dplyr_src
#' @keywords internal
#' @export
src_tbls.dm <- function(x, ...) {
  deprecate_warn("0.2.0", "dm::src_tbls.dm()", details = "Use `names(dm_get_tables(dm))` instead.")

  check_not_zoomed(x)

  src_tbls_impl(x)
}

#' @details
#' Use [copy_to()] on a table and then [dm()] instead of `copy_to()`
#' on a `dm` object.
#' @param dest For `copy_to.dm()`: The `dm` object to which a table should be copied.
#' @param df For `copy_to.dm()`: A table (can be on a different `src`)
#' @param name For `copy_to.dm()`: See [dplyr::copy_to()]
#' @param overwrite For `copy_to.dm()`: See [dplyr::copy_to()]; `TRUE` leads to an error
#' @param temporary For `copy_to.dm()`: If the `dm` is on a DB, the copied version of `df` will only be written temporarily to the DB.
#' After the connection is reset it will no longer be available.
#' @param repair,quiet Name repair options; cf. [vctrs::vec_as_names()]
#' @export
#' @rdname dplyr_src
#' @keywords internal
copy_to.dm <- function(
  dest,
  df,
  name = deparse(substitute(df)),
  overwrite = FALSE,
  temporary = TRUE,
  repair = "unique",
  quiet = FALSE,
  ...
) {
  dm_local_error_call()
  deprecate_warn(
    "0.2.0",
    "dm::copy_to.dm()",
    details = "Use `copy_to(dm_get_con(dm), ...)` and `dm()`."
  )

  check_not_zoomed(dest)

  if (!(inherits(df, "data.frame") || inherits(df, "tbl_dbi"))) {
    abort_only_data_frames_supported()
  }
  if (overwrite) {
    abort_no_overwrite()
  }
  if (length(name) != 1) {
    abort_one_name_for_copy_to(name)
  }
  # src: if `df` on a different src:
  # if `df_list` is on DB and `dest` is local, collect `df_list`
  # if `df_list` is local and `dest` is on DB, copy `df_list` to respective DB
  dest_src <- dm_get_src_impl(dest)
  if (is.null(dest_src)) {
    df <- as_tibble(collect(df))
  } else {
    # FIXME: should we allow `overwrite` argument?
    df <- copy_to(dest_src, df, unique_db_table_name(name), temporary = temporary, ...)
  }
  names_list <- repair_table_names(src_tbls_impl(dest), name, repair, quiet)
  # rename old tables with potentially new names
  dest <- dm_rename_tbl(dest, !!!names_list$new_old_names)
  # `repair` argument is `unique` by default
  dm_add_tbl_impl(dest, list(df), names_list$new_names)
}

MAX_COMMAS <- 6L

commas <- function(x, max_commas = MAX_COMMAS, capped = FALSE, fun = identity) {
  if (is_null(max_commas)) {
    max_commas <- MAX_COMMAS
  }
  fun <- as_function(fun)

  if (is_empty(x)) {
    x <- ""
  } else if (length(x) > max_commas) {
    cap <- if (!capped) paste0(" (", length(x), " total)")

    x <- c(
      fun(x[seq_len(max_commas - 1)]),
      paste0(cli::symbol$ellipsis, cap)
    )
  } else {
    x <- fun(x)
  }

  glue_collapse(x, sep = ", ")
}

tick <- function(x) {
  if (is_empty(x)) {
    return(character())
  }
  paste0("`", x, "`")
}

deparse_keys <- function(x) {
  ticked <- map(x, tick_if_needed)
  out <- map_chr(ticked, paste, collapse = ", ")
  add_c <- lengths(ticked) != 1
  out[add_c] <- paste0("c(", out[add_c], ")")
  out
}

# next 2 are borrowed from {tibble}:
tick_if_needed <- function(x) {
  needs_ticks <- !is_syntactic(x)
  x[needs_ticks] <- tick(x[needs_ticks])
  x
}

is_syntactic <- function(x) {
  x == make.names(x)
}

if_pkg_version <- function(pkg, min_version, if_true, if_false = NULL) {
  if (packageVersion(pkg) >= min_version) if_true else if_false
}

format_classes <- function(class) {
  commas(tick(class))
}

trim_width <- function(x, width) {
  if (nchar(x) > width) {
    paste0(strtrim(x, width), cli::symbol$ellipsis)
  } else {
    x
  }
}

test_that("output", {
  expect_snapshot({
    commas(character())
    commas(letters[1])
    commas(letters[1:2])
    commas(letters[1:3])
    commas(letters[seq_len(MAX_COMMAS - 1)])
    commas(letters[seq_len(MAX_COMMAS)])
    commas(letters[seq_len(MAX_COMMAS + 1)])
    commas(letters[1:4], max_commas = 3)
    commas(letters, capped = TRUE)
    commas(letters, fun = tick)
  })
})

#' @rdname deprecated
#' @keywords internal
#' @export
check_if_subset <- function(t1, c1, t2, c2) {
  deprecate_warn("0.1.0", "dm::check_if_subset()", "dm::check_subset()")
  t1q <- enquo(t1)
  t2q <- enquo(t2)
  c1q <- ensym(c1)
  c2q <- ensym(c2)
  if (is_subset(eval_tidy(t1q), !!c1q, eval_tidy(t2q), !!c2q)) {
    return(invisible(eval_tidy(t1q)))
  }
  v1 <- pull(eval_tidy(t1q), !!ensym(c1q))
  v2 <- pull(eval_tidy(t2q), !!ensym(c2q))
  setdiff_v1_v2 <- setdiff(v1, v2)
  print(filter(eval_tidy(t1q), !!c1q %in% setdiff_v1_v2))
  abort_not_subset_of(
    as_name(t1q),
    as_name(c1q),
    as_name(t2q),
    as_name(c2q)
  )
}

#' @rdname deprecated
#' @keywords internal
#' @export
check_cardinality <- function(parent_table, pk_column, child_table, fk_column) {
  deprecate_warn("0.1.0", "dm::check_cardinality()", "dm::examine_cardinality()")
  pt <- enquo(parent_table)
  pkc <- enexpr(pk_column)
  ct <- enquo(child_table)
  fkc <- enexpr(fk_column)
  if (!is_unique_key(eval_tidy(pt), !!pkc)$unique) {
    return(glue(
      "Column(s) {tick(commas(as_string(pkc)))} not ",
      "a unique key of {tick('parent_table')}."
    ))
  }
  if (!is_subset(!!ct, !!fkc, !!pt, !!pkc)) {
    return(glue(
      "Column(s) {tick(commas(as_string(fkc)))} of {tick('child_table')} not ",
      "a subset of column(s) {tick(commas(as_string(pkc)))} of {tick('parent_table')}."
    ))
  }
  min_1 <- is_subset(!!pt, !!pkc, !!ct, !!fkc)
  max_1 <- pull(is_unique_key(eval_tidy(ct), !!fkc), unique)
  if (min_1 && max_1) {
    return("bijective mapping (child: 1 -> parent: 1)")
  } else if (min_1) {
    return("surjective mapping (child: 1 to n -> parent: 1)")
  } else if (max_1) {
    return("injective mapping (child: 0 or 1 -> parent: 1)")
  }
  "generic mapping (child: 0 to n -> parent: 1)"
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_src <- function(x) {
  deprecate_warn("0.1.0", "dm::cdm_get_src()", "dm::dm_get_con()")
  check_not_zoomed(x)
  out <- dm_get_src_impl(x)
  if (is.null(out)) {
    out <- default_local_src()
  }
  out
}

default_local_src <- function() {
  structure(
    list(tbl_f = as_tibble, name = "<environment: R_GlobalEnv>", env = .GlobalEnv),
    class = c("src_local", "src")
  )
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_con <- function(x) {
  deprecate_warn("0.1.0", "dm::cdm_get_con()", "dm::dm_get_con()")
  dm_get_con(dm = x)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_tables <- function(x) {
  deprecate_warn("0.1.0", "dm::cdm_get_tables()", "dm::dm_get_tables()")
  dm_get_tables(x = x)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_filter <- function(x) {
  deprecate_warn("0.1.0", "dm::cdm_get_filter()", "dm::dm_get_filters()")
  dm_get_filters(dm = x)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_add_tbl <- function(dm, ..., repair = "unique", quiet = FALSE) {
  deprecate_warn("0.1.0", "dm::cdm_add_tbl()", "dm::dm_add_tbl()")
  dm_add_tbl(dm = dm, ... = ..., repair = repair, quiet = quiet)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_rm_tbl <- function(dm, ...) {
  deprecate_warn("0.1.0", "dm::cdm_rm_tbl()", "dm::dm_select_tbl()")
  dm_rm_tbl_impl(dm = dm, ... = ...)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_copy_to <- function(
  dest,
  dm,
  ...,
  types = NULL,
  overwrite = NULL,
  indexes = NULL,
  unique_indexes = NULL,
  set_key_constraints = TRUE,
  unique_table_names = FALSE,
  table_names = NULL,
  temporary = TRUE
) {
  deprecate_warn("0.1.0", "dm::cdm_copy_to()", "dm::copy_dm_to()")

  if (!is_null(unique_table_names)) {
    if (is.null(table_names) && temporary && !unique_table_names) {
      table_names <- identity
    }
  }

  inject(copy_dm_to(
    dest = dest,
    dm = dm,
    ... = ...,
    !!!compact(list(
      types = types,
      overwrite = overwrite,
      indexes = indexes,
      unique_indexes = unique_indexes
    )),
    set_key_constraints = set_key_constraints,
    table_names = table_names,
    temporary = temporary
  ))
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_disambiguate_cols <- function(dm, sep = ".", quiet = FALSE) {
  deprecate_warn("0.1.0", "dm::cdm_disambiguate_cols()", "dm::dm_disambiguate_cols()")
  dm_disambiguate_cols(dm = dm, .sep = sep, .quiet = quiet)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_draw <- function(
  dm,
  rankdir = "LR",
  col_attr = "column",
  view_type = "keys_only",
  columnArrows = TRUE,
  graph_attrs = "",
  node_attrs = "",
  edge_attrs = "",
  focus = NULL,
  graph_name = "Data Model"
) {
  deprecate_warn("0.1.0", "dm::cdm_draw()", "dm::dm_draw()")
  dm_draw(
    dm = dm,
    rankdir = rankdir,
    col_attr = col_attr,
    view_type = view_type,
    columnArrows = columnArrows,
    graph_attrs = graph_attrs,
    node_attrs = node_attrs,
    edge_attrs = edge_attrs,
    focus = focus,
    graph_name = graph_name
  )
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_set_colors <- function(dm, ...) {
  deprecate_warn("0.1.0", "dm::cdm_set_colors()", "dm::dm_set_colors()")
  display <- color_quos_to_display(...)
  dm_set_colors(dm, !!!display)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_colors <- function(dm) {
  deprecate_warn("0.1.0", "dm::cdm_get_colors()", "dm::dm_get_colors()")
  prep_recode(dm_get_colors(dm))
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_available_colors <- function() {
  deprecate_warn(
    "0.1.0",
    "dm::cdm_get_available_colors()",
    "dm::dm_get_available_colors()"
  )
  dm_get_available_colors()
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_filter <- function(dm, table, ...) {
  deprecate_warn("0.1.0", "dm::cdm_filter()", "dm::dm_filter()")
  dm %>%
    dm_zoom_to({{ table }}) %>%
    dm_filter_impl(..., set_filter = TRUE) %>%
    dm_update_zoomed()
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_nrow <- function(dm) {
  deprecate_warn("0.1.0", "dm::cdm_nrow()", "dm::dm_nrow()")
  dm_nrow(dm = dm)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_flatten_to_tbl <- function(dm, start, ..., join = left_join) {
  deprecate_warn("0.1.0", "dm::cdm_flatten_to_tbl()", "dm::dm_flatten_to_tbl()")
  join_name <- deparse(substitute(join))
  start <- dm_tbl_name(dm, {{ start }})

  vars <- setdiff(src_tbls_impl(dm), start)
  list_of_pts <- eval_select_table(quo(c(...)), vars)

  dm_flatten_to_tbl_impl(
    dm,
    start,
    list_of_pts,
    join = join,
    join_name = join_name,
    squash = FALSE,
    .position = "prefix"
  )
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_squash_to_tbl <- function(dm, start, ..., join = left_join) {
  deprecate_warn("0.1.0", "dm::cdm_squash_to_tbl()", "dm::dm_squash_to_tbl()")
  join_name <- deparse(substitute(join))
  if (!(join_name %in% c("left_join", "full_join", "inner_join"))) {
    abort_squash_limited()
  }
  start <- dm_tbl_name(dm, {{ start }})

  vars <- setdiff(src_tbls_impl(dm), start)
  list_of_pts <- eval_select_table(quo(c(...)), vars)

  dm_flatten_to_tbl_impl(
    dm,
    start,
    list_of_pts,
    join = join,
    join_name = join_name,
    squash = TRUE,
    .position = "prefix"
  )
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_join_to_tbl <- function(dm, table_1, table_2, join = left_join) {
  force(join)
  deprecate_warn("0.1.0", "dm::cdm_join_to_tbl()", "dm::dm_join_to_tbl()")
  stopifnot(is_function(join))
  join_name <- deparse(substitute(join))

  t1_name <- dm_tbl_name(dm, {{ table_1 }})
  t2_name <- dm_tbl_name(dm, {{ table_2 }})

  rel <- parent_child_table(dm, {{ table_1 }}, {{ table_2 }})
  start <- rel$child_table
  other <- rel$parent_table

  dm_flatten_to_tbl_impl(
    dm,
    start,
    other,
    join = join,
    join_name = join_name,
    squash = FALSE,
    .position = "prefix"
  )
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_apply_filters <- function(dm) {
  deprecate_warn("0.1.0", "dm::cdm_apply_filters()", "dm::dm_apply_filters()")
  dm_apply_filters(dm = dm)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_apply_filters_to_tbl <- function(dm, table) {
  deprecate_warn(
    "0.1.0",
    "dm::cdm_apply_filters_to_tbl()",
    "dm::dm_apply_filters_to_tbl()"
  )
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})
  dm_get_filtered_table(dm, table_name)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_add_pk <- function(dm, table, column, check = FALSE, force = FALSE) {
  deprecate_warn("0.1.0", "dm::cdm_add_pk()", "dm::dm_add_pk()")
  dm_add_pk(dm, {{ table }}, {{ column }}, check = check, force = force)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_add_fk <- function(dm, table, column, ref_table, check = FALSE) {
  deprecate_warn("0.1.0", "dm::cdm_add_fk()", "dm::dm_add_fk()")
  dm_add_fk(dm, {{ table }}, {{ column }}, {{ ref_table }}, check = check)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_has_fk <- function(dm, table, ref_table) {
  deprecate_warn("0.1.0", "dm::cdm_has_fk()", "dm::dm_has_fk()")
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})
  ref_table_name <- dm_tbl_name(dm, {{ ref_table }})
  dm_has_fk_impl(dm, table_name, ref_table_name)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_fk <- function(dm, table, ref_table) {
  deprecate_warn("0.1.0", "dm::cdm_get_fk()", "dm::dm_get_fk()")
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})
  ref_table_name <- dm_tbl_name(dm, {{ ref_table }})
  new_keys(dm_get_fk_impl(dm, table_name, ref_table_name))
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_all_fks <- function(dm) {
  deprecate_warn("0.1.0", "dm::cdm_get_all_fks()", "dm::dm_get_all_fks()")
  dm %>%
    dm_get_all_fks_impl() %>%
    mutate(child_fk_cols = as.character(unclass(child_fk_cols))) %>%
    mutate(parent_key_cols = as.character(unclass(parent_key_cols)))
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_rm_fk <- function(dm, table, columns, ref_table) {
  deprecate_warn("0.1.0", "dm::cdm_rm_fk()", "dm::dm_rm_fk()")
  check_not_zoomed(dm)
  column_quo <- enquo(columns)
  if (quo_is_missing(column_quo)) {
    abort_rm_fk_col_missing()
  }
  table_name <- dm_tbl_name(dm, {{ table }})
  ref_table_name <- dm_tbl_name(dm, {{ ref_table }})
  fk_cols <- dm_get_fk_impl(dm, table_name, ref_table_name)
  if (is_empty(fk_cols)) {
    return(dm)
  }
  if (quo_is_null(column_quo)) {
    cols <- get_key_cols(fk_cols)
  } else {
    cols <- as_name(ensym(columns))
    if (!all(cols %in% fk_cols)) {
      abort_is_not_fkc()
    }
  }
  dm_rm_fk_impl(dm, table_name, cols, ref_table_name, NULL)
}

#' @rdname deprecated
#' @keywords internal
#' @export
#' @autoglobal
cdm_enum_fk_candidates <- function(dm, table, ref_table) {
  deprecate_warn("0.1.0", "dm::cdm_enum_fk_candidates()", "dm::dm_enum_fk_candidates()")
  check_not_zoomed(dm)
  check_no_filter(dm)
  table_name <- dm_tbl_name(dm, {{ table }})
  ref_table_name <- dm_tbl_name(dm, {{ ref_table }})

  ref_tbl_pk <- dm_get_pk_impl(dm, ref_table_name)
  ref_tbl <- tbl_impl(dm, ref_table_name)
  tbl <- tbl_impl(dm, table_name)
  enum_fk_candidates_impl(
    table_name,
    tbl,
    ref_table_name,
    ref_tbl,
    ref_tbl_pk
  ) %>%
    rename(columns = column) %>%
    mutate(columns = new_keys(columns))
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_is_referenced <- function(dm, table) {
  deprecate_warn("0.1.0", "dm::cdm_is_referenced()", "dm::dm_is_referenced()")
  check_not_zoomed(dm)
  has_length(dm_get_referencing_tables(dm, !!ensym(table)))
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_referencing_tables <- function(dm, table) {
  deprecate_warn(
    "0.1.0",
    "dm::cdm_get_referencing_tables()",
    "dm::dm_get_referencing_tables()"
  )
  check_not_zoomed(dm)
  table <- dm_tbl_name(dm, {{ table }})
  def <- dm_get_def(dm)
  i <- which(def$table == table)
  def$fks[[i]]$table
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_learn_from_db <- function(dest) {
  deprecate_warn("0.1.0", "dm::cdm_learn_from_db()", "dm::dm_from_src()")
  dm_from_con(con_from_src_or_con(dest))
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_check_constraints <- function(dm) {
  deprecate_warn("0.1.0", "dm::cdm_check_constraints()", "dm::dm_examine_constraints()")
  dm_examine_constraints_impl(dm = dm, progress = FALSE, top_level_fun = "cdm_check_constraints")
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_nycflights13 <- function(cycle = FALSE, color = TRUE, subset = TRUE) {
  deprecate_warn("0.1.0", "dm::cdm_nycflights13()", "dm::dm_nycflights13()")
  dm_nycflights13(cycle = cycle, color = color, subset = subset)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_paste <- function(dm, select = FALSE, tab_width = 2) {
  deprecate_warn("0.1.0", "dm::cdm_paste()", "dm::dm_paste()")

  if (isTRUE(select)) {
    options <- c("keys", "select")
  } else {
    options <- c("keys")
  }

  code <- dm_paste_impl(dm = dm, options, tab_width)

  # without "\n" in the end it looks weird when a warning is issued
  cat(code, "\n")

  invisible(dm)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_has_pk <- function(dm, table) {
  deprecate_warn("0.1.0", "dm::cdm_has_pk()", "dm::dm_has_pk()")
  check_not_zoomed(dm)
  has_length(dm_get_pk(dm, {{ table }}))
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_pk <- function(dm, table) {
  deprecate_warn("0.1.0", "dm::cdm_get_pk()", "dm::dm_get_pk()")
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})
  new_keys(dm_get_pk_impl(dm, table_name))
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_get_all_pks <- function(dm) {
  deprecate_warn("0.1.0", "dm::cdm_get_all_pks()", "dm::dm_get_all_pks()")
  dm %>%
    dm_get_all_pks_impl() %>%
    mutate(pk_col = as.character(unclass(pk_col)))
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_rm_pk <- function(dm, table, rm_referencing_fks = FALSE) {
  deprecate_warn("0.1.0", "dm::cdm_rm_pk()", "dm::dm_rm_pk()")
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})
  def <- dm_get_def(dm)
  if (!rm_referencing_fks && dm_is_referenced(dm, !!table_name)) {
    affected <- dm_get_referencing_tables(dm, !!table_name)
    abort_first_rm_fks(table_name, affected)
  }
  def$pks[def$table == table_name] <- list(new_pk())
  def$fks[def$table == table_name] <- list(new_fk())
  dm_from_def(def)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_enum_pk_candidates <- function(dm, table) {
  deprecate_warn("0.1.0", "dm::cdm_enum_pk_candidates()", "dm::dm_enum_pk_candidates()")
  check_no_filter(dm)

  table_name <- dm_tbl_name(dm, {{ table }})

  table <- dm_get_tables_impl(dm)[[table_name]]
  enum_pk_candidates_impl(table)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_select_tbl <- function(dm, ...) {
  deprecate_warn("0.1.0", "dm::cdm_select_tbl()", "dm::dm_select_tbl()")
  dm_select_tbl(dm = dm, ... = ...)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_rename_tbl <- function(dm, ...) {
  deprecate_warn("0.1.0", "dm::cdm_rename_tbl()", "dm::dm_rename_tbl()")
  dm_rename_tbl(dm = dm, ... = ...)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_select <- function(dm, table, ...) {
  deprecate_warn("0.1.0", "dm::cdm_select()", "dm::dm_select()")
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})

  dm %>%
    dm_zoom_to(!!table_name) %>%
    select(...) %>%
    dm_update_zoomed()
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_rename <- function(dm, table, ...) {
  deprecate_warn("0.1.0", "dm::cdm_rename()", "dm::dm_rename()")
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})

  dm %>%
    dm_zoom_to(!!table_name) %>%
    rename(...) %>%
    dm_update_zoomed()
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_zoom_to_tbl <- function(dm, table) {
  deprecate_warn("0.1.0", "dm::cdm_zoom_to_tbl()", "dm::dm_zoom_to()")
  check_not_zoomed(dm)
  zoom <- dm_tbl_name(dm, {{ table }})

  cols <- list(get_all_cols(dm, zoom))
  dm %>%
    dm_get_def() %>%
    mutate(
      zoom = if_else(table == !!zoom, data, list(NULL)),
      col_tracker_zoom = if_else(table == !!zoom, cols, list(NULL))
    ) %>%
    dm_from_def(zoomed = TRUE)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_insert_zoomed_tbl <- function(dm, new_tbl_name = NULL, repair = "unique", quiet = FALSE) {
  deprecate_warn("0.1.0", "dm::cdm_insert_zoomed_tbl()", "dm::dm_insert_zoomed()")

  check_zoomed(dm)
  if (is_null(enexpr(new_tbl_name))) {
    new_tbl_name_chr <- orig_name_zoomed(dm)
  } else {
    if (is_symbol(enexpr(new_tbl_name))) {
      warning("The argument `new_tbl_name` in `dm_insert_zoomed()` should be of class `character`.")
    }
    new_tbl_name_chr <- as_string(enexpr(new_tbl_name))
  }
  names_list <- repair_table_names(
    old_names = src_tbls_impl(dm),
    new_names = new_tbl_name_chr,
    repair,
    quiet
  )
  dm <- dm_select_tbl_zoomed_impl(dm, names_list$new_old_names)
  new_tbl_name_chr <- names_list$new_names
  old_tbl_name <- orig_name_zoomed(dm)
  new_tbl <- list(tbl_zoomed(dm))
  all_filters <- filters_zoomed(dm)
  old_filters <- all_filters %>% filter(!zoomed)
  new_filters <-
    all_filters %>%
    filter(zoomed) %>%
    mutate(zoomed = FALSE)
  upd_pk <- list_of(update_zoomed_pk(dm))
  upd_inc_fks <- list_of(update_zoomed_incoming_fks(dm))
  dm_wo_outgoing_fks <-
    dm %>%
    update_filter(old_tbl_name, list_of(old_filters)) %>%
    dm_add_tbl_zoomed_impl(new_tbl, new_tbl_name_chr, list_of(new_filters)) %>%
    dm_get_def() %>%
    mutate(
      pks = if_else(table == new_tbl_name_chr, !!upd_pk, pks),
      fks = if_else(table == new_tbl_name_chr, !!upd_inc_fks, fks)
    ) %>%
    dm_from_def(zoomed = TRUE, validate = FALSE)

  dm_wo_outgoing_fks %>%
    dm_insert_zoomed_outgoing_fks(
      new_tbl_name_chr,
      names_list$old_new_names[old_tbl_name],
      col_tracker_zoomed(dm)
    ) %>%
    dm_clean_zoomed()
}

dm_select_tbl_zoomed_impl <- function(dm, selected) {
  if (anyDuplicated(names(selected))) {
    abort_need_unique_names(names(selected[duplicated(names(selected))]))
  }

  def <-
    dm_get_def(dm) %>%
    filter_recode_table_def(selected) %>%
    filter_recode_table_fks(selected)

  dm_from_def(def, zoomed = TRUE)
}

dm_add_tbl_zoomed_impl <- function(
  dm,
  tbls,
  table_name,
  filters = list_of(new_filter()),
  pks = list_of(new_pk()),
  fks = list_of(new_fk())
) {
  def <- dm_get_def(dm)

  def_0 <- def[rep_along(table_name, NA_integer_), ]
  def_0$table <- table_name
  def_0$data <- unname(tbls)
  def_0$pks <- pks
  def_0$fks <- fks
  def_0$filters <- filters

  dm_from_def(vec_rbind(def, def_0), zoomed = TRUE)
}


#' @rdname deprecated
#' @keywords internal
#' @export
cdm_update_zoomed_tbl <- function(dm) {
  deprecate_warn("0.1.0", "dm::cdm_update_zoomed_tbl()", "dm::dm_update_zoomed()")
  dm_update_zoomed(dm = dm)
}

#' @rdname deprecated
#' @keywords internal
#' @export
cdm_zoom_out <- function(dm) {
  deprecate_warn("0.1.0", "dm::cdm_zoom_out()", "dm::dm_discard_zoomed()")
  dm_discard_zoomed(dm = dm)
}

#' @rdname deprecated
#' @keywords internal
#' @export
dm_rm_tbl <- function(dm, ...) {
  deprecate_warn("1.0.0", "dm::dm_rm_tbl()", "dm::dm_select_tbl()")
  # since deprecated cdm_rm_tbl() was calling dm_rm_tbl() we need dm_rm_tbl_impl
  dm_rm_tbl_impl(dm, ...)
}

dm_rm_tbl_impl <- function(dm, ...) {
  check_not_zoomed(dm)
  deselected_ind <- eval_select_table_indices(quo(c(...)), src_tbls_impl(dm))
  selected_ind <- setdiff(seq_along(dm), deselected_ind)

  dm_select_tbl(dm, !!!selected_ind)
}

abort_rm_fk_col_missing <- function() {
  cli::cli_abort(
    "Parameter {.arg columns} has to be set. Pass {.val {I('NULL')}} for removing all references.",
    class = dm_error_full("rm_fk_col_missing")
  )
}

#' @description
#' `dm_add_tbl` is deprecated as of dm 1.0.0, because the same functionality
#' is offered by [dm()] with `.name_repair = "unique"`.
#' @rdname deprecated
#' @keywords internal
#' @export
dm_add_tbl <- function(dm, ..., repair = "unique", quiet = FALSE) {
  deprecate_warn(
    "1.0.0",
    "dm_add_tbl()",
    "dm()",
    details = 'Use `.name_repair = "unique"` if necessary.'
  )

  check_not_zoomed(dm)

  new_names <- names(exprs(..., .named = TRUE))
  new_tables <- list2(...)

  check_new_tbls(dm, new_tables)

  old_names <- src_tbls_impl(dm)
  names_list <- repair_table_names(old_names, new_names, repair, quiet)
  # rename old tables in case name repair changed their names

  dm <- dm_select_tbl_impl(dm, names_list$new_old_names)
  dm_add_tbl_impl(dm, new_tables, names_list$new_names)
}

#' @description
#' `dm_bind()`  is deprecated as of dm 1.0.0, because the same functionality
#' is offered by [dm()].
#'
#' @rdname deprecated
#' @keywords internal
#' @export
dm_bind <- function(..., repair = "check_unique", quiet = FALSE) {
  deprecate_warn("1.0.0", "dm_bind()", "dm()")

  dms <- list2(...)

  new_def <- dm_bind_impl(dms, repair, quiet)
  dm_from_def(new_def)
}

#' @description
#' `dm_squash_to_tbl()`  is deprecated as of dm 1.0.0, because the same functionality
#' is offered by [dm_flatten_to_tbl()] with `.recursive = TRUE`.
#'
#' @rdname deprecated
#' @keywords internal
#'
#' @export
dm_squash_to_tbl <- function(dm, start, ..., join = left_join) {
  deprecate_warn(
    "1.0.0",
    "dm_squash_to_tbl()",
    details = "Please use `.recursive = TRUE` in `dm_flatten_to_tbl()` instead."
  )

  check_not_zoomed(dm)
  join_name <- as_label(enexpr(join))
  if (!(join_name %in% c("left_join", "full_join", "inner_join"))) {
    abort_squash_limited()
  }
  start <- dm_tbl_name(dm, {{ start }})

  vars <- setdiff(src_tbls_impl(dm), start)
  list_of_pts <- eval_select_table(quo(c(...)), vars)

  dm_flatten_to_tbl_impl(
    dm,
    start,
    list_of_pts,
    join = join,
    join_name = join_name,
    squash = TRUE,
    .position = "prefix"
  )
}

#' @description
#' `rows_truncate()` is deprecated as of dm 1.0.0, because it's a DDL operation
#' and requires different permissions than the
#' [`dplyr::rows_*()`][dplyr::rows_insert()] functions.
#'
#' @rdname deprecated
#' @keywords internal
#' @export
rows_truncate <- function(x, ..., in_place = FALSE) {
  deprecate_warn("1.0.0", "rows_truncate()")

  check_dots_used(action = warn)
  UseMethod("rows_truncate", x)
}

rows_truncate_ <- function(x, y, ..., in_place = FALSE) {
  UseMethod("rows_truncate", x)
}

#' @export
rows_truncate.data.frame <- function(x, ..., in_place = NULL) {
  stopifnot(is.null(in_place) || !in_place)
  x[0, ]
}

#' @export
rows_truncate.tbl_sql <- function(x, ..., in_place = NULL) {
  name <- target_table_name(x, in_place)

  if (!is_null(name)) {
    con <- dbplyr::remote_con(x)
    sql <- sql_rows_truncate(x)
    DBI::dbExecute(con, sql, immediate = TRUE)
    invisible(x)
  } else {
    x %>%
      filter(0L == 1L)
  }
}

#' @rdname deprecated
#' @keywords internal
#'
#' @export
sql_rows_truncate <- function(x, ...) {
  deprecate_warn("1.0.0", "sql_rows_truncate()")

  check_dots_used()
  UseMethod("sql_rows_truncate")
}

#' @export
sql_rows_truncate.tbl_sql <- function(x, ...) {
  name <- dbplyr::remote_name(x)
  paste0("TRUNCATE TABLE ", name)
}

#' @export
sql_rows_truncate.tbl_SQLiteConnection <- function(x, ...) {
  name <- dbplyr::remote_name(x)
  paste0("DELETE FROM ", name)
}

#' @export
sql_rows_truncate.tbl_duckdb_connection <- sql_rows_truncate.tbl_SQLiteConnection

target_table_name <- function(x, in_place) {
  name <- dbplyr::remote_name(x)

  # Only write if requested
  if (!is_null(name) && is_true(in_place)) {
    return(name)
  }

  # Abort if requested but can't write
  if (is_null(name) && is_true(in_place)) {
    cli::cli_abort(
      "Can't determine name for target table. Set {.code in_place = FALSE} to return a lazy table."
    )
  }

  # Verbose by default
  if (is_null(in_place)) {
    if (is_null(name)) {
      inform(
        "Result is returned as lazy table, because `x` does not correspond to a table that can be updated. Use `in_place = FALSE` to mute this message."
      )
    } else {
      inform(
        "Result is returned as lazy table. Use `in_place = FALSE` to mute this message, or `in_place = TRUE` to write to the underlying table."
      )
    }
  }

  # Never write unless handled above
  NULL
}

#' @rdname deprecated
#' @keywords internal
#'
#' @export
dm_rows_truncate <- function(x, y, ..., in_place = NULL, progress = NA) {
  deprecate_warn("1.0.0", "dm_rows_truncate()")
  check_dots_empty()

  dm_rows(x, y, "truncate", top_down = FALSE, in_place, require_keys = FALSE, progress = progress)
}

test_that("cdm_add_tbl() works", {
  skip_on_cran()
  skip_if_remote_src()
  local_options(lifecycle_verbosity = "quiet")

  mtcars_tbl <- test_src_frame(!!!mtcars)

  expect_equivalent_dm(
    cdm_add_tbl(dm_for_filter_simple(), cars_table = mtcars_tbl),
    dm(dm_for_filter_simple(), cars_table = mtcars_tbl)
  )
})

test_that("cdm_rm_tbl() works", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_dm(
    cdm_rm_tbl(dm_for_flatten(), starts_with("dim")),
    dm_rm_tbl(dm_for_flatten(), starts_with("dim"))
  )
})

test_that("cdm_copy_to() behaves correctly", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_dm(
    cdm_copy_to(duckdb_test_src(), dm_for_filter_simple(), unique_table_names = TRUE),
    dm_for_filter_simple()
  )
})

test_that("cdm_disambiguate_cols() works as intended", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_dm(
    expect_message_obj(cdm_disambiguate_cols(dm_for_disambiguate())),
    expect_message_obj(dm_disambiguate_cols(dm_for_disambiguate()))
  )
})

test_that("cdm_get_colors() behaves as intended", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equal(
    cdm_get_colors(cdm_nycflights13()),
    set_names(
      c("#ED7D31FF", "#ED7D31FF", "#5B9BD5FF", "#ED7D31FF", "#70AD47FF"),
      c("airlines", "airports", "flights", "planes", "weather")
    )
  )
})

test_that("cdm_filter() behaves correctly", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_tbl(
    cdm_filter(dm_for_filter_simple(), tf_1, a > 4) %>% dm_apply_filters_to_tbl(tf_2),
    tf_2_simple() %>% filter(d > 4)
  )

  expect_equivalent_tbl(
    dm_filter(dm_for_filter_simple(), tf_1, a > 4) %>% cdm_apply_filters_to_tbl(tf_2),
    tf_2_simple() %>% filter(d > 4)
  )

  skip_if_remote_src()
  expect_snapshot({
    dm_for_filter_simple() %>%
      dm_filter(tf_1, a > 3, a < 8) %>%
      cdm_apply_filters() %>%
      dm_get_tables()
  })
})

test_that("cdm_nrow() works?", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equal(
    sum(cdm_nrow(dm_test_obj())),
    rows_dm_obj
  )
})

test_that("`cdm_flatten_to_tbl()`, `cdm_join_to_tbl()` and `dm_squash_to_tbl()` work", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")
  # FIXME: Debug GHA fail
  # expect_equivalent_tbl(
  #   expect_message_obj(cdm_flatten_to_tbl(dm_for_flatten(), fact)),
  #   result_from_flatten()
  # )

  expect_equivalent_tbl(
    expect_message_obj(cdm_join_to_tbl(dm_for_flatten(), fact, dim_3)),
    select(result_from_flatten(), fact:fact.something, dim_3.something)
  )
})

test_that("cdm_get_src() works", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_dm_error(
    cdm_get_src(1),
    class = "is_not_dm"
  )

  expect_identical(
    class(cdm_get_src(dm_for_filter_simple_db())),
    class(my_db_test_src())
  )
})

test_that("cdm_get_con() works", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_dm_error(
    cdm_get_con(1),
    class = "is_not_dm"
  )

  if (is.null(my_test_src())) {
    expect_dm_error(
      cdm_get_con(dm_for_filter_simple()),
      class = "con_only_for_dbi"
    )
  } else {
    expect_silent(cdm_get_con(dm_for_filter_simple()))
  }
})


test_that("cdm_get_tables() works", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_tbl_lists(
    cdm_get_tables(dm_for_filter_simple()),
    dm_get_tables(dm_for_filter_simple())
  )
})

test_that("cdm_get_filter() works", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_identical(
    cdm_get_filter(dm_for_filter_simple()),
    dm_get_filters(dm_for_filter_simple())
  )

  expect_identical(
    cdm_get_filter(dm_filter(dm_for_filter_simple(), tf_1, a > 3, a < 8)),
    dm_get_filters(dm_filter(dm_for_filter_simple(), tf_1, a > 3, a < 8))
  )
})

test_that("cdm_add_pk() and cdm_add_fk() work", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_dm(
    cdm_add_pk(dm_test_obj(), dm_table_4, c),
    dm_add_pk(dm_test_obj(), dm_table_4, c)
  )

  expect_equivalent_dm(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      cdm_add_fk(dm_table_1, a, dm_table_4),
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4)
  )
})

test_that("other FK functions work", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_true(cdm_has_fk(dm_for_filter_simple(), tf_2, tf_1))

  expect_false(cdm_has_fk(dm_for_filter_simple(), tf_1, tf_2))

  expect_identical(
    cdm_get_fk(dm_for_filter_simple(), tf_2, tf_1),
    dm_get_fk(dm_for_filter_simple(), tf_2, tf_1)
  )

  expect_identical(
    dm_for_filter_simple() %>%
      cdm_get_all_fks() %>%
      mutate(child_fk_cols = new_keys(child_fk_cols), parent_key_cols = new_keys(parent_key_cols)),
    dm_get_all_fks(dm_for_filter_simple())
  )

  expect_equivalent_dm(
    cdm_rm_fk(dm_for_filter_simple(), tf_2, d, tf_1),
    dm_rm_fk(dm_for_filter_simple(), tf_2, d, tf_1)
  )

  skip_if_remote_src()
  expect_identical(
    dm_for_filter_simple() %>%
      cdm_enum_fk_candidates(tf_2, tf_1) %>%
      mutate(why = if_else(why != "", "<reason>", "")),
    dm_for_filter_simple() %>%
      dm_enum_fk_candidates(tf_2, tf_1) %>%
      mutate(why = if_else(why != "", "<reason>", ""))
  )
})

test_that("graph-functions work", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_identical(
    cdm_is_referenced(dm_for_filter_simple(), tf_3),
    dm_is_referenced(dm_for_filter_simple(), tf_3)
  )

  expect_identical(
    cdm_get_referencing_tables(dm_for_filter_simple(), tf_3),
    dm_get_referencing_tables(dm_for_filter_simple(), tf_3)
  )
})

test_that("cdm_learn_from_db() works from PG", {
  skip("not testing deprecated learning from DB: test too slow")
  local_options(lifecycle_verbosity = "quiet")

  src_postgres <- skip_if_error(src_test("postgres"))
  con_postgres <- src_postgres$con

  # create an object on the Postgres-DB that can be learned
  if (is_postgres_empty()) {
    copy_dm_to(con_postgres, dm_for_filter_simple(), unique_table_names = TRUE, temporary = FALSE)
  }

  expect_equivalent_dm(
    cdm_learn_from_db(con_postgres),
    dm_learn_from_db(con_postgres)
  )
  clear_postgres()
})

test_that("cdm_examine_constraints() works", {
  skip_on_cran()
  skip_if_remote_src()
  local_options(lifecycle_verbosity = "quiet")

  expect_identical(
    cdm_check_constraints(bad_dm()),
    dm_examine_constraints_impl(bad_dm(), progress = FALSE)
  )
})

test_that("cdm_nycflights13() works", {
  skip("not testing deprecated cdm_nycflights13(): test too slow")

  expect_equivalent_dm(
    cdm_nycflights13(),
    dm_nycflights13()
  )
})

test_that("cdm_paste() works", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_output(
    cdm_paste(dm_for_filter_simple(), FALSE, 4),
    paste0(
      "dm::dm(\n    tf_1,\n    tf_2,\n    tf_3,\n    tf_4,\n    tf_5,\n    tf_6,\n) %>%\n    dm::dm_add_pk(tf_1, a) %>%\n    dm::dm_add_pk(tf_2, c) %>%",
      "\n    dm::dm_add_pk(tf_3, f) %>%\n    dm::dm_add_pk(tf_4, h) %>%\n    dm::dm_add_pk(tf_5, k) %>%\n    ",
      "dm::dm_add_pk(tf_6, n) %>%\n    dm::dm_add_fk(tf_2, d, tf_1) %>%\n    dm::dm_add_fk(tf_2, e, tf_3) %>%\n    ",
      "dm::dm_add_fk(tf_4, j, tf_3) %>%\n    dm::dm_add_fk(tf_5, l, tf_4) %>%\n    dm::dm_add_fk(tf_5, m, tf_6)"
    ),
    fixed = TRUE
  )
})

test_that("other PK functions work", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_identical(
    cdm_has_pk(dm_for_filter_simple(), tf_1),
    dm_has_pk(dm_for_filter_simple(), tf_1)
  )

  expect_identical(
    cdm_get_pk(dm_for_filter_simple(), tf_1),
    dm_get_pk(dm_for_filter_simple(), tf_1)
  )

  expect_identical(
    dm_for_filter_simple() %>%
      cdm_get_all_pks() %>%
      mutate(pk_col = new_keys(pk_col)),
    dm_get_all_pks(dm_for_filter_simple())
  )

  expect_equivalent_dm(
    cdm_rm_pk(dm_for_filter_simple(), tf_2),
    dm_rm_pk(dm_for_filter_simple(), tf_2)
  )

  expect_identical(
    dm_for_disambiguate() %>%
      cdm_enum_pk_candidates(iris_1) %>%
      rename(columns = column) %>%
      mutate(columns = new_keys(columns)),
    dm_enum_pk_candidates(dm_for_disambiguate(), iris_1)
  )
})

test_that("dm_select_tbl() and dm_rename_tbl() work", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_dm(
    cdm_select_tbl(dm_for_filter_simple(), tf_1_new = tf_1, tf_2, new_tf_6 = tf_6),
    dm_select_tbl(dm_for_filter_simple(), tf_1_new = tf_1, tf_2, new_tf_6 = tf_6)
  )

  expect_equivalent_dm(
    cdm_rename_tbl(dm_for_filter_simple(), tf_1_new = tf_1, new_tf_6 = tf_6),
    dm_rename_tbl(dm_for_filter_simple(), tf_1_new = tf_1, new_tf_6 = tf_6)
  )
})

test_that("dm_select() and dm_rename() work", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_tbl(
    cdm_select(dm_for_filter_simple(), tf_1, a_new = a) %>% tbl_impl("tf_1"),
    dm_select(dm_for_filter_simple(), tf_1, a_new = a) %>% tbl_impl("tf_1")
  )

  expect_equivalent_tbl(
    cdm_rename(dm_for_filter_simple(), tf_1, a_new = a) %>% tbl_impl("tf_1"),
    dm_rename(dm_for_filter_simple(), tf_1, a_new = a) %>% tbl_impl("tf_1")
  )
})

test_that("dm_zoom_to() and related functions work", {
  skip_on_cran()
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_dm(
    cdm_zoom_to_tbl(dm_for_filter_simple(), tf_1),
    dm_zoom_to(dm_for_filter_simple(), tf_1)
  )

  expect_equivalent_dm(
    dm_zoom_to(dm_for_filter_simple(), tf_1) %>% cdm_insert_zoomed_tbl("another_name"),
    dm_zoom_to(dm_for_filter_simple(), tf_1) %>% dm_insert_zoomed("another_name")
  )

  expect_equivalent_dm(
    dm_zoom_to(dm_for_filter_simple(), tf_1) %>% cdm_update_zoomed_tbl(),
    dm_zoom_to(dm_for_filter_simple(), tf_1) %>% dm_update_zoomed()
  )

  expect_equivalent_dm(
    dm_zoom_to(dm_for_filter_simple(), tf_1) %>% cdm_zoom_out(),
    dm_zoom_to(dm_for_filter_simple(), tf_1) %>% dm_discard_zoomed()
  )
})

test_that("default_local_src() works", {
  expect_s3_class(default_local_src(), "src")
})

test_that("dm_squash_to_tbl() deprecation warning is correct", {
  # Test that the deprecation warning shows the correct parameter name .recursive
  expect_snapshot({
    dm_squash_to_tbl(dm_for_flatten(), fact)
  })
})

#' Unnest columns from a wrapped table
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `dm_unnest_tbl()` target a specific column to unnest
#' from the given table in a given dm.
#' A ptype or a set of keys should be given, not both.
#'
#' @details
#' [dm_nest_tbl()] is an inverse operation to `dm_unnest_tbl()`
#' if differences in row and column order are ignored.
#' The opposite is true if referential constraints between both tables
#' are satisfied.
#'
#' @inheritParams dm_unwrap_tbl
#' @param parent_table A table in the dm with nested columns.
#' @param col The column to unnest (unquoted).
#'
#' @return A dm.
#' @seealso [dm_unwrap_tbl()], [dm_unpack_tbl()],
#'   [dm_nest_tbl()], [dm_pack_tbl()], [dm_wrap_tbl()],
#'   [dm_examine_constraints()], [dm_examine_cardinalities()],
#'   [dm_ptype()].
#' @export
#'
#' @examples
#' airlines_wrapped <-
#'   dm_nycflights13() %>%
#'   dm_wrap_tbl(airlines)
#'
#' # The ptype is required for reconstruction.
#' # It can be an empty dm, only primary and foreign keys are considered.
#' ptype <- dm_ptype(dm_nycflights13())
#'
#' airlines_wrapped %>%
#'   dm_unnest_tbl(airlines, flights, ptype)
dm_unnest_tbl <- function(dm, parent_table, col, ptype) {
  # process args and build names
  parent_table_name <- dm_tbl_name(dm, {{ parent_table }})
  table <- dm_get_tables_impl(dm)[[parent_table_name]]
  col_expr <- enexpr(col)
  new_child_table_name <- names(eval_select_indices(col_expr, colnames(table)))

  child_pk_names <-
    dm_get_all_pks(ptype) %>%
    filter(table == new_child_table_name) %>%
    pull(pk_col) %>%
    unlist()
  fk <-
    dm_get_all_fks(ptype) %>%
    filter(child_table == new_child_table_name, parent_table == parent_table_name)
  parent_fk_names <- unlist(fk$parent_key_cols)
  child_fk_names <- unlist(fk$child_fk_cols)

  # extract nested table
  new_table <-
    table %>%
    select(!!!set_names(parent_fk_names, child_fk_names), !!new_child_table_name) %>%
    unnest(!!new_child_table_name) %>%
    distinct()

  # update the dm by adding new table, removing nested col and setting keys
  dm <- dm(dm, !!new_child_table_name := new_table)
  dm <- dm_select(dm, !!parent_table_name, -all_of(new_child_table_name))
  if (length(parent_fk_names)) {
    dm <- dm_add_fk(
      dm,
      !!new_child_table_name,
      !!child_fk_names,
      !!parent_table_name,
      !!parent_fk_names
    )
  }
  if (length(child_pk_names)) {
    dm <- dm_add_pk(dm, !!new_child_table_name, !!child_pk_names)
  }

  dm
}

#' Unpack columns from a wrapped table
#'
#' #' @description
#' `r lifecycle::badge("experimental")`
#'
#' `dm_unpack_tbl()` targets a specific column to unpack
#' from the given table in a given dm.
#' A ptype or a set of keys should be given,
#' not both.
#'
#' [dm_pack_tbl()] is an inverse operation to `dm_unpack_tbl()`
#' if differences in row and column order are ignored.
#' The opposite is true if referential constraints between both tables
#' are satisfied
#' and if all rows in the parent table have at least one child row,
#' i.e. if the relationship is of cardinality 1:n or 1:1.
#'
#' @inheritParams dm_unwrap_tbl
#' @param child_table A table in the dm with packed columns.
#' @param col The column to unpack (unquoted).
#'
#' @seealso [dm_unwrap_tbl()], [dm_unnest_tbl()],
#'   [dm_nest_tbl()], [dm_pack_tbl()], [dm_wrap_tbl()],
#'   [dm_examine_constraints()], [dm_examine_cardinalities()],
#'   [dm_ptype()].
#' @export
#' @examples
#' flights_wrapped <-
#'   dm_nycflights13() %>%
#'   dm_wrap_tbl(flights)
#'
#' # The ptype is required for reconstruction.
#' # It can be an empty dm, only primary and foreign keys are considered.
#' ptype <- dm_ptype(dm_nycflights13())
#'
#' flights_wrapped %>%
#'   dm_unpack_tbl(flights, airlines, ptype)
dm_unpack_tbl <- function(dm, child_table, col, ptype) {
  # process args and build names
  child_table_name <- dm_tbl_name(dm, {{ child_table }})
  table <- dm_get_tables_impl(dm)[[child_table_name]]
  col_expr <- enexpr(col)
  new_parent_table_name <- names(eval_select_indices(col_expr, colnames(table)))

  parent_pk_names <- dm_get_all_pks(ptype) %>%
    filter(table == new_parent_table_name) %>%
    pull(pk_col) %>%
    unlist()
  fk <- dm_get_all_fks(ptype) %>%
    filter(child_table == child_table_name, parent_table == new_parent_table_name)
  child_fk_names <- unlist(fk$child_fk_cols)
  parent_fk_names <- unlist(fk$parent_key_cols)

  # extract packed table
  new_table <-
    table %>%
    select(!!!set_names(child_fk_names, parent_fk_names), !!new_parent_table_name) %>%
    unpack(!!new_parent_table_name) %>%
    distinct()

  # update the dm by adding new table, removing packed col and setting keys
  dm <- dm(dm, !!new_parent_table_name := new_table)
  dm <- dm_select(dm, !!child_table_name, -all_of(new_parent_table_name))
  if (length(child_fk_names)) {
    dm <- dm_add_fk(
      dm,
      !!child_table_name,
      !!child_fk_names,
      !!new_parent_table_name,
      !!parent_fk_names
    )
  }
  if (length(parent_pk_names)) {
    dm <- dm_add_pk(dm, !!new_parent_table_name, !!parent_pk_names)
  }

  dm
}

test_that("`dm_unnest_tbl()` and `dm_unpack_tbl()` failure modes", {
  skip_if_remote_src()

  # trying to unpack or unnest a column that doesn't exist gives informative error
  expect_error(
    dm_unnest_tbl(dm_for_filter(), tf_2, tf_1, dm_for_filter()),
    "Column `tf_1` doesn't exist"
  )
})

test_that("`dm_unpack_tbl()` failure modes", {
  skip_if_remote_src()

  # trying to unpack a column that doesn't exist gives informative error
  expect_error(
    dm_unpack_tbl(dm_for_filter(), tf_2, tf_1, dm_for_filter()),
    "Column `tf_1` doesn't exist"
  )
})

#' Add a unique key
#'
#' @description
#' `dm_add_uk()` marks the specified columns as a unique key of the specified table.
#' If `check == TRUE`, then it will first check if
#' the given combination of columns is a unique key of the table.
#'
#' @inheritParams dm_add_pk
#'
#' @details The difference between a primary key (PK) and a unique key (UK) consists in the following:
#' - When a local `dm` is copied to a database (DB) with `copy_dm_to()`, a PK will be set on the DB by default, whereas a UK is being ignored.
#' - A PK can be set as an `autoincrement` key (also implemented on certain DBMS when the `dm` is transferred to the DB)
#' - There can be only one PK for each table, whereas there can be unlimited UKs
#' - A UK will be used, if the same table has an autoincrement PK in addition, to ensure that during delta load processes
#'   on the DB (cf. [dm_rows_append()]) the foreign keys are updated accordingly.
#'   If no UK is available, the insertion is done row-wise, which also ensures a correct matching, but can be much slower.
#' - A UK can generally enhance the data model by adding additional information
#' - There can also be implicit UKs, when the columns addressed by a foreign key are neither a PK nor a UK.
#'   These implicit UKs are also listed by [`dm_get_all_uks()`]
#'
#' @family primary key functions
#'
#' @return An updated `dm` with an additional unqiue key.
#'
#' @examplesIf rlang::is_installed(c("nycflights13", "DiagrammeR"))
#'
#' nycflights_dm <- dm(
#'   planes = nycflights13::planes,
#'   airports = nycflights13::airports,
#'   weather = nycflights13::weather
#' )
#'
#' # Create unique keys:
#' nycflights_dm %>%
#'   dm_add_uk(planes, tailnum) %>%
#'   dm_add_uk(airports, faa, check = TRUE) %>%
#'   dm_add_uk(weather, c(origin, time_hour)) %>%
#'   dm_get_all_uks()
#'
#' # Keys can be checked during creation:
#' try(
#'   nycflights_dm %>%
#'     dm_add_uk(planes, manufacturer, check = TRUE)
#' )
#' @export
dm_add_uk <- function(dm, table, columns, ..., check = FALSE) {
  check_dots_empty()
  dm_local_error_call()

  check_not_zoomed(dm)

  table_name <- dm_tbl_name(dm, {{ table }})
  table <- dm_get_tables_impl(dm)[[table_name]]

  check_required(columns)
  col_expr <- enexpr(columns)
  col_name <- names(eval_select_indices(col_expr, colnames(table)))

  if (check) {
    table_from_dm <- dm_get_filtered_table(dm, table_name)
    eval_tidy(expr(check_key(!!sym(table_name), !!col_expr)), list2(!!table_name := table_from_dm))
  }

  dm_add_uk_impl(dm, table_name, col_name)
}

# both "table" and "column" must be characters
# in {datamodelr}, a primary key may consist of more than one columns
# a key will be added, regardless of whether it is a unique key or not; not to be exported
dm_add_uk_impl <- function(dm, table, column) {
  def <- dm_get_def(dm)
  i <- which(def$table == table)

  if (!is_empty(def$pks[[i]]$column) && identical(def$pks[[i]]$column[[1]], column)) {
    abort_no_uk_if_pk(table, column)
  }
  if (any(map_lgl(def$uks[[i]]$column, identical, column))) {
    abort_no_uk_if_pk(table, column, type = "UK")
  }

  def$uks[[i]] <- vctrs::vec_rbind(
    def$uks[[i]],
    new_uk(list(column))
  )

  dm_from_def(def)
}

#' Get all unique keys of a [`dm`] object
#'
#' @description
#' `dm_get_all_uks()` checks the `dm` object for unique keys
#' (primary keys, explicit and implicit unique keys) and returns the tables and
#' the respective unique key columns.
#'
#' @family primary key functions
#' @param table One or more table names, unquoted,
#'   to return unique key information for.
#'   The default `NULL` returns information for all tables.
#'
#' @details There are 3 kinds of unique keys:
#'    - `PK`: Primary key, set by [`dm_add_pk()`]
#'    - `explicit UK`: Unique key, set by [`dm_add_uk()`]
#'    - `implicit UK`: Unique key, not explicitly set, but referenced by a foreign key.
#'
#' @inheritParams dm_add_uk
#'
#' @return A tibble with the following columns:
#'   \describe{
#'     \item{`table`}{table name,}
#'     \item{`uk_col`}{column name(s) of primary key, as list of character vectors,}
#'     \item{`kind`}{kind of unique key, see details.}
#'   }
#'
#' @export
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_get_all_uks()
dm_get_all_uks <- function(dm, table = NULL, ...) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)
  table_expr <- enexpr(table) %||% src_tbls_impl(dm, quiet = TRUE)
  table_names <- eval_select_table(table_expr, set_names(src_tbls_impl(dm, quiet = TRUE)))
  dm_get_all_uks_impl(dm, table_names)
}

dm_get_all_uks_impl <- function(dm, table = NULL) {
  def <- dm %>%
    dm_get_def()

  if (!is.null(table)) {
    idx <- match(table, def$table)
    def <- def[match(table, def$table), ]
  }

  all_pks <- dm_get_all_pks_def_impl(def, table) %>%
    select(-autoincrement) %>%
    rename(uk_col = pk_col)

  all_explicit_uks <- dm_get_all_uks_def_impl(def)

  all_explicit <- bind_rows(
    mutate(all_pks, kind = "PK"),
    mutate(all_explicit_uks, kind = "explicit UK")
  )

  all_implicit <- dm_get_all_implicit_uks_def_impl(def, table, all_explicit) %>%
    mutate(kind = "implicit UK")

  bind_rows(
    all_explicit,
    all_implicit
  )
}

dm_get_all_uks_def_impl <- function(def, table = NULL) {
  # Optimized for speed

  def_sub <- def[c("table", "uks")]
  out <-
    def_sub %>%
    unnest_df("uks", tibble(column = list())) %>%
    set_names(c("table", "uk_col")) %>%
    mutate(kind = "explicit UK")

  out$uk_col <- new_keys(out$uk_col)
  out
}

dm_get_all_implicit_uks_def_impl <- function(def, table = NULL, all_explicit) {
  dm_get_all_fks_def_impl(def) %>%
    select(table = parent_table, uk_col = parent_key_cols) %>%
    distinct() %>%
    # rm those that are PKs or explicit UKs
    anti_join(all_explicit, c("table", "uk_col"))
}


#' Remove a unique key
#'
#' @description
#' `dm_rm_uk()` removes one or more unique keys from a table and leaves the [`dm`] object otherwise unaltered.
#' An error is thrown if no unique key matches the selection criteria.
#' If the selection criteria are ambiguous, a message with unambiguous replacement code is shown.
#' Foreign keys are never removed.
#'
#' @inheritParams dm_rm_pk
#'
#' @family primary key functions
#'
#' @return An updated `dm` without the indicated unique key(s).
#'
#' @export
dm_rm_uk <- function(dm, table = NULL, columns = NULL, ...) {
  dm_local_error_call()
  dm_rm_uk_(dm, {{ table }}, {{ columns }}, ...)
}

dm_rm_uk_ <- function(dm, table, columns, ...) {
  check_dots_empty()
  check_not_zoomed(dm)

  table_name <- dm_tbl_name_null(dm, {{ table }})
  columns <- enexpr(columns)

  dm_rm_uk_impl(dm, table_name, columns)
}

dm_rm_uk_impl <- function(dm, table_name, columns, error_call = caller_env()) {
  def <- dm_get_def(dm)

  if (is.null(table_name)) {
    i <- which(map_int(def$uks, vec_size) > 0)
  } else {
    i <- which(def$table == table_name)
    if (nrow(def$uks[[i]]) == 0) {
      i <- integer()
    }
  }
  if (length(i) == 0 && dm_is_strict_keys(dm)) {
    abort_uk_not_defined()
  }

  ii <- if (!quo_is_null(columns)) {
    ii_col <- map2(
      def$data[i],
      def$uks[i],
      ~ tryCatch(
        {
          vars <- eval_select_indices(columns, colnames(.x), error_call = error_call)
          map_lgl(.y$column, ~ identical(names(vars), .x))
        },
        error = function(e) {
          FALSE
        }
      )
    )

    # if `columns` is not NULL, it refers to only one UK, therefore we can choose
    # the first element of the list created by `map2()`

    # FIXME: error message should be more informative: which UKs are available
    # for the given table? What was the user input for `columns`?
    if (!any(ii_col[[1]])) {
      abort_uk_not_defined()
    }
    ii_col[[1]]
  } else {
    # if no `column` is provided by user, use all columns for matching
    TRUE
  }

  # Talk about it
  if (is.null(table_name) || quo_is_null(columns)) {
    n_uk_per_table <- map_int(i, ~ nrow(def$uks[[.x]]))
    message("Removing unique keys: %>%")
    message(
      "  ",
      glue_collapse(
        glue(
          "dm_rm_uk({tick_if_needed(rep(def$table[i], n_uk_per_table))}, {list_c(map(i, ~ deparse_keys(def$uks[[.x]]$column)))})"
        ),
        " %>%\n  "
      )
    )
  }
  # Execute
  # in case `length(i) > 1`: all tables have all their UKs removed, respectively
  def$uks[i] <- if (length(i) > 1) {
    list_of(new_uk())
  } else {
    list_of(filter(def$uks[[i]], !ii))
  }

  dm_from_def(def)
}

# Error -------------------------------------------------------------------

abort_uk_not_defined <- function() {
  cli::cli_abort(
    "No unique keys to remove.",
    class = dm_error_full("uk_not_defined"),
    call = dm_error_call()
  )
}

abort_no_uk_if_pk <- function(table, column, type = "PK") {
  cli::cli_abort(
    "A {type} ({commas(tick(column))}) for table {.field {table}} already exists, not adding UK.",
    class = dm_error_full("no_uk_if_pk"),
    call = dm_error_call()
  )
}

test_that("unique keys", {
  expect_snapshot({
    nyc_1_uk <- dm_add_uk(
      dm_nycflights_small(),
      flights,
      everything()
    )

    # add 1 UK
    nyc_1_uk %>%
      dm_get_all_uks()
    # add 2 UK to same table and 1 to another table
    nyc_1_uk %>%
      dm_add_uk(
        flights,
        c(origin, dest, time_hour)
      ) %>%
      dm_add_uk(
        planes,
        c(year, manufacturer, model)
      ) %>%
      dm_get_all_uks()

    # add 2 UK to same table and 1 to another table and remove 1 of the 2 for the first table
    nyc_1_uk %>%
      dm_add_uk(
        flights,
        c(origin, dest, time_hour)
      ) %>%
      dm_add_uk(
        planes,
        c(year, manufacturer, model)
      ) %>%
      dm_rm_uk(flights, c(origin, dest, time_hour)) %>%
      dm_get_all_uks()

    # add 2 UK to same table and 1 to another table and remove 1 of the 2 for
    # the first table using the tidyselect function used to create the UK
    nyc_1_uk %>%
      dm_add_uk(
        flights,
        c(origin, dest, time_hour)
      ) %>%
      dm_add_uk(
        planes,
        c(year, manufacturer, model)
      ) %>%
      dm_rm_uk(flights, everything()) %>%
      dm_get_all_uks()

    # dm_rm_uk()
    nyc_1_uk %>%
      dm_rm_uk() %>%
      dm_get_all_uks()
    nyc_1_uk %>%
      dm_rm_uk(flights) %>%
      dm_get_all_uks()
    nyc_1_uk %>%
      dm_rm_uk(flights, everything()) %>%
      dm_get_all_uks()

    # test for code generation:
    # add 2 UK to same table and 1 of length > 1 to another table and remove all UKs
    nyc_1_uk %>%
      dm_add_uk(
        flights,
        c(origin, dest, time_hour)
      ) %>%
      dm_add_uk(
        planes,
        c(year, manufacturer, model)
      ) %>%
      dm_rm_uk() %>%
      dm_get_all_uks()

    # test for code generation:
    # add 2 UK to same table and 1 of length 1  to another table and remove all UKs
    nyc_1_uk %>%
      dm_add_uk(
        flights,
        c(origin, dest, time_hour)
      ) %>%
      dm_add_uk(
        planes,
        manufacturer
      ) %>%
      dm_rm_uk() %>%
      dm_get_all_uks()

    # dm_examine_constraints()
    dm_examine_constraints(
      dm_nycflights_small() %>%
        dm_add_uk(
          planes,
          c(year, manufacturer, model)
        )
    )

    dm_add_fk(
      dm_nycflights_small(),
      flights,
      time_hour,
      weather,
      time_hour
    ) %>%
      dm_get_all_uks()

    dm_add_fk(
      dm_nycflights_small(),
      flights,
      time_hour,
      weather,
      time_hour
    ) %>%
      dm_add_uk(weather, time_hour) %>%
      dm_get_all_uks()

    # key tracking needs to work also for UKs
    dm_rename(dm_for_filter(), tf_6, p = n) %>% dm_get_all_uks()
    dm_rename(dm_for_filter(), tf_6, p = n) %>% dm_get_all_fks()
    dm_select(dm_for_filter(), tf_6, -n) %>% dm_get_all_uks()
    dm_select(dm_for_filter(), tf_6, -n) %>% dm_get_all_fks()

    # test table arg for dm_get_all_uks()
    nyc_1_uk %>%
      dm_get_all_uks(flights)

    nyc_1_uk %>%
      dm_get_all_uks("airports")

    nyc_1_uk %>%
      dm_get_all_uks(c(airports, weather, flights, airlines))

    # test tidyselect functions
    nyc_1_uk %>%
      dm_get_all_uks(starts_with("a"))

    nyc_1_uk %>%
      dm_get_all_uks(everything())
  })

  expect_snapshot_error(
    # failing check upon addition of UK
    dm_add_uk(
      dm_nycflights_small(),
      planes,
      c(year, manufacturer, model),
      check = TRUE
    ),
    class = dm_error("not_unique_key")
  )

  expect_snapshot_error(
    # trying to add a UK for which a PK already exists
    dm_add_uk(
      dm_nycflights_small(),
      airlines,
      carrier,
    ),
    class = dm_error("no_uk_if_pk")
  )

  expect_snapshot_error(
    # trying to add a UK for which a PK already exists
    dm_add_uk(
      nyc_1_uk,
      flights,
      everything()
    ),
    class = dm_error("no_uk_if_pk")
  )
})

test_that("dm_get_all_uks() with table arg fails nicely", {
  expect_snapshot_error(
    # trying to request a table not part of the dm
    nyc_1_uk %>%
      dm_get_all_uks(timetable)
  )

  expect_snapshot_error(
    # trying to request 2 tables that are not part of the dm and a few others
    nyc_1_uk %>%
      dm_get_all_uks(c(timetable, weather, flights, tabletime))
  )
})

eval_select_table <- function(quo, table_names, unique = TRUE, error_call = caller_env()) {
  indexes <- eval_select_table_indices(quo, table_names, unique = unique, error_call = error_call)
  set_names(table_names[indexes], names(indexes))
}

eval_rename_table_all <- function(quo, table_names) {
  indexes <- eval_rename_table_indices(quo, table_names)
  names <- table_names
  names[indexes] <- names(indexes)
  set_names(table_names, names)
}

eval_select_table_indices <- function(quo, table_names, unique = TRUE, error_call = caller_env()) {
  withCallingHandlers(
    eval_select_indices(quo, table_names, unique = unique, error_call = error_call),
    vctrs_error_subscript = function(cnd) {
      cnd$subscript_elt <- "table"
      cnd_signal(cnd)
    }
  )
}

eval_rename_table_indices <- function(quo, table_names) {
  withCallingHandlers(
    eval_rename_indices(quo, table_names),
    vctrs_error_subscript = function(cnd) {
      cnd$subscript_elt <- "table"
      cnd_signal(cnd)
    }
  )
}

eval_select_both <- function(quo, names, error_call = caller_env()) {
  indices <- eval_select_indices(quo, names, error_call = error_call)
  names <- set_names(names[indices], names(indices))
  list(indices = indices, names = names)
}

eval_select_indices <- function(quo, names, unique = TRUE, error_call = caller_env()) {
  pos <- tidyselect::eval_select(quo, set_names(names), error_call = error_call)

  if (unique) {
    # Called for side effects.
    # Normally done by tidyselect if the `data` argument
    # to eval_select() is a data frame.
    vec_as_names(names(pos), repair = "check_unique", call = error_call)
  }

  pos
}

eval_rename_both <- function(quo, names) {
  indices <- eval_rename_indices(quo, names)
  all_names <- set_names(names)
  names(all_names)[indices] <- names(indices)
  list(indices = indices, all_names = all_names)
}

eval_rename_indices <- function(quo, names) {
  pos <- tidyselect::eval_rename(quo, set_names(names))

  # Called for side effects.
  # Normally done by tidyselect if the `data` argument
  # to eval_rename() is a data frame.
  vec_as_names(names(pos), repair = "check_unique")

  pos
}

test_that("tidyselecting tables works", {
  table_names <- c("table_1", "table_2", "table_3")
  quo <- quo(c(table_3_new = table_3, table_1_new = table_1))

  expect_identical(
    eval_select_table_indices(quo, table_names),
    c(table_3_new = 3L, table_1_new = 1L)
  )

  expect_identical(
    eval_select_table(quo, table_names),
    set_names(c("table_3", "table_1"), c("table_3_new", "table_1_new"))
  )

  expect_identical(
    eval_rename_table_all(quo, table_names),
    set_names(table_names, c("table_1_new", "table_2", "table_3_new"))
  )
})

test_that("output", {
  skip_if_not_installed("vctrs", "0.4.1")
  skip_if(packageVersion("tidyselect") > "1.2.0")

  expect_snapshot(error = TRUE, {
    dm_for_filter() %>%
      dm_select_tbl(tf_7)

    dm_for_filter() %>%
      dm_rename_tbl(tf_0 = tf_7)
  })
})

# nocov start
replace_if_dbplyr_has <- function(fun) {
  if (!requireNamespace("dbplyr", quietly = TRUE)) {
    return()
  }

  dbplyr_ns <- asNamespace("dbplyr")

  fun <- as_string(ensym(fun))
  value <- mget(fun, dbplyr_ns, mode = "function", ifnotfound = list(NULL))[[1]]
  if (!is.null(value)) {
    assign(fun, value, inherits = TRUE)
    "dbplyr"
  } else {
    "dm"
  }
}

register_if_dbplyr_hasnt <- function(...) {
  methods <- enquos(..., .named = TRUE)
  if (requireNamespace("dbplyr", quietly = TRUE)) {
    dbplyr_ns <- asNamespace("dbplyr")

    # Register our method implementations only if dbplyr doesn't provide them
    dbplyr_methods <- mget(names(methods), dbplyr_ns, mode = "function", ifnotfound = list(NULL))

    methods <- methods[map_lgl(dbplyr_methods, is.null)]

    if (is_empty(methods)) {
      return()
    }
  }

  methods <- map(methods, eval_tidy)
  classes <- sub("^[^.]*.", "", names(methods))
  fun <- sub("[.].*$", "", names(methods)[[1]])

  map2(classes, methods, s3_register, generic = paste0("dm::", fun))
  invisible()
}
# nocov end

remote_name_qual <- function(x) {
  dbplyr_ns <- asNamespace("dbplyr")

  value <- mget("remote_table", dbplyr_ns, mode = "function", ifnotfound = list(NULL))[[1]]
  if (!is.null(value)) {
    table <- value(x)
    dbplyr::escape(table, con = dbplyr::remote_con(x))
  } else {
    dbplyr::remote_name(x)
  }
}

# db_schema_list() -------------------------------------------------------

#' List schemas on a database
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `db_schema_list()` lists the available schemas on the database.
#'
#' @param con An object of class `"src"` or `"DBIConnection"`.
#' @param include_default Boolean, if `TRUE` (default), also the default schema
#' on the database is included in the result
#' @param ... Passed on to the individual methods.
#'
#' @details Methods are not available for all DBMS.
#'
#' Additional arguments are:
#'
#'   - `dbname`: supported for MSSQL. List schemas on a different database on the connected MSSQL-server;
#'   default: database addressed by `con`.
#'
#' @return A tibble with the following columns:
#'   \describe{
#'     \item{`schema_name`}{the names of the schemas,}
#'     \item{`schema_owner`}{the schema owner names.}
#'   }
#'
#' @export
#' @family schema handling functions
#'
#' @name db_schema_list
db_schema_list <- function(con, include_default = TRUE, ...) {
  check_param_class(include_default, "logical")

  # If we check in the method, we need to specify the user_env argument
  if (inherits(con, "src_dbi")) {
    deprecate_warn(
      "0.2.5",
      'dm::db_schema_list(con = "must be a DBI connection, not a dbplyr source,")',
    )
  }

  UseMethod("db_schema_list")
}

#' @export
db_schema_list.src_dbi <- function(con, include_default = TRUE, ...) {
  db_schema_list(con$con, include_default = include_default, ...)
}

#' @export
`db_schema_list.Microsoft SQL Server` <- function(con, include_default = TRUE, dbname = NULL, ...) {
  dbname_sql <- if (is_null(dbname)) {
    ""
  } else {
    check_param_class(dbname, "character")
    check_param_length(dbname)
    paste0(DBI::dbQuoteIdentifier(con, dbname), ".")
  }
  default_if_true <- if_else(include_default, "", " AND NOT s.name = 'dbo'")
  # ignore built-in schemas for backward compatibility:
  # https://docs.microsoft.com/en-us/sql/relational-databases/security/authentication-access/ownership-and-user-schema-separation?view=sql-server-ver15
  DBI::dbGetQuery(
    con,
    glue::glue(
      "SELECT s.name as schema_name
    FROM {dbname_sql}sys.schemas s
    WHERE s.name NOT IN ('sys', 'guest', 'INFORMATION_SCHEMA', 'db_accessadmin',
          'db_backupoperator', 'db_datareader', 'db_datawriter', 'db_ddladmin',
          'db_denydatareader', 'db_denydatawriter', 'db_owner',
          'db_securityadmin'){default_if_true}"
    )
  ) %>%
    as_tibble()
}

#' @export
db_schema_list.PqConnection <- function(con, include_default = TRUE, ...) {
  default_if_true <- if_else(include_default, "", ", 'public'")
  DBI::dbGetQuery(
    con,
    glue::glue(
      "SELECT schema_name, schema_owner FROM information_schema.schemata WHERE
    schema_name NOT IN ('information_schema', 'pg_catalog'{default_if_true})
    AND schema_name NOT LIKE 'pg_toast%'
    AND schema_name NOT LIKE 'pg_temp_%'
    ORDER BY schema_name"
    )
  ) %>%
    as_tibble()
}

#' @export
db_schema_list.SQLiteConnection <- function(con, include_default = TRUE, ...) {
  abort_no_schemas_supported("SQLite")
}

#' @export
db_schema_list.Pool <- function(con, include_default = TRUE, ...) {
  pool_con <- pool::poolCheckout(con)
  on.exit(pool::poolReturn(pool_con))
  db_schema_list(pool_con, include_default, ...)
}


# db_schema_exists() -----------------------------------------------------

#' Check for existence of a schema on a database
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `db_schema_exists()` checks, if a schema exists on the database.
#'
#' @inheritParams db_schema_list
#' @param schema Class `character` or `SQL`, name of the schema
#'
#' @details Methods are not available for all DBMS.
#'
#' Additional arguments are:
#'
#'   - `dbname`: supported for MSSQL. Check if a schema exists on a different
#'   database on the connected MSSQL-server; default: database addressed by `con`.
#' @return A boolean: `TRUE` if schema exists, `FALSE` otherwise.
#'
#' @family schema handling functions
#' @export
db_schema_exists <- function(con, schema, ...) {
  check_param_class(schema, "character")
  check_param_length(schema)

  # If we check in the method, we need to specify the user_env argument
  if (inherits(con, "src_dbi")) {
    deprecate_warn(
      "0.2.5",
      'dm::db_schema_exists(con = "must be a DBI connection, not a dbplyr source,")',
    )
  }

  UseMethod("db_schema_exists")
}

#' @export
db_schema_exists.src_dbi <- function(con, schema, ...) {
  db_schema_exists(con$con, schema, ...)
}

#' @export
`db_schema_exists.Microsoft SQL Server` <- function(con, schema, dbname = NULL, ...) {
  sql_to_character(con, schema) %in% db_schema_list(con, dbname = dbname)$schema_name
}


#' @export
db_schema_exists.PqConnection <- function(con, schema, ...) {
  sql_to_character(con, schema) %in% db_schema_list(con)$schema_name
}

#' @export
db_schema_exists.SQLiteConnection <- function(con, schema, ...) {
  abort_no_schemas_supported("SQLite")
}


# db_schema_create() -----------------------------------------------------

#' Create a schema on a database
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `db_schema_create()` creates a schema on the database.
#'
#' @inheritParams db_schema_list
#' @param schema Class `character` or `SQL` (cf. Details), name of the schema
#'
#' @details Methods are not available for all DBMS.
#'
#' An error is thrown if a schema of that name already exists.
#'
#' The argument `schema` (and `dbname` for MSSQL) can be provided as `SQL` objects.
#' Keep in mind, that in this case it is assumed that they are already correctly quoted as identifiers
#' using [DBI::dbQuoteIdentifier()].
#'
#' Additional arguments are:
#'
#'   - `dbname`: supported for MSSQL. Create a schema in a different
#'   database on the connected MSSQL-server; default: database addressed by `con`.
#'
#' @return `NULL` invisibly.
#'
#' @family schema handling functions
#' @export
db_schema_create <- function(con, schema, ...) {
  check_param_class(schema, "character")
  check_param_length(schema)

  # If we check in the method, we need to specify the user_env argument
  if (inherits(con, "src_dbi")) {
    deprecate_warn(
      "0.2.5",
      'dm::db_schema_create(con = "must be a DBI connection, not a dbplyr source,")',
    )
  }

  UseMethod("db_schema_create")
}

#' @export
db_schema_create.src_dbi <- function(con, schema, ...) {
  db_schema_create(con$con, schema, ...)
}

#' @export
db_schema_create.PqConnection <- function(con, schema, ...) {
  DBI::dbExecute(con, glue("CREATE SCHEMA {DBI::dbQuoteIdentifier(con, schema)}"))
  message(glue::glue("Schema {tick(sql_to_character(con, schema))} created."))
  invisible(NULL)
}

#' @export
`db_schema_create.Microsoft SQL Server` <- function(con, schema, dbname = NULL, ...) {
  if (!is_null(dbname)) {
    original_dbname <- attributes(con)$info$dbname
    DBI::dbExecute(con, glue::glue("USE {DBI::dbQuoteIdentifier(con, dbname)}"))
    withr::defer(DBI::dbExecute(
      con,
      glue::glue("USE {DBI::dbQuoteIdentifier(con, original_dbname)}")
    ))
  }
  msg_suffix <- fix_msg(sql_to_character(con, dbname))
  DBI::dbExecute(con, glue("CREATE SCHEMA {DBI::dbQuoteIdentifier(con, schema)}"))
  message(glue::glue("Schema {tick(sql_to_character(con, schema))} created{msg_suffix}."))
  invisible(NULL)
}

#' @export
db_schema_create.SQLiteConnection <- function(con, schema, ...) {
  abort_no_schemas_supported("SQLite")
}

# sql_schema_table_list() -------------------------------------------------

# List the tables in a schema on a database
#
# @description `sql_schema_table_list()` list the tables in a schema on the database.
#
# @inheritParams db_schema_exists
#
# @details Methods are not available for all DBMS.
#
# An error is thrown if no schema of that name exists.
#
# Additional arguments are:
#
#   - `dbname`: supported for MSSQL. Look for tables on a different
#   database on the connected MSSQL-server; default: database addressed by `con`.
#
# @return A tibble with the following columns:
#   \describe{
#     \item{`table_name`}{name of the table,}
#     \item{`remote_name`}{identifier of the table on the DBMS.
#     Can be used to access the listed tables with the syntax
#     `tbl(con, remote_name).`}
#   }
#
# @family schema handling functions
# @export
# sql_schema_table_list <- function(con, schema = NULL, ...) {
#   if (!is_null(schema)) {
#     check_param_class(schema, "character")
#     check_param_length(schema)
#   }
#   if (!is_null(schema) && !db_schema_exists(con, schema, ...)) {
#     abort_no_schema_exists(sql_to_character(con_from_src_or_con(con), schema), ...)
#   }
#   UseMethod("sql_schema_table_list")
# }

# FIXME: this should be done using a dplyr function
sql_schema_table_list_mssql <- function(con, schema = NULL, dbname = NULL) {
  src <- src_from_src_or_con(con)
  if (!is_null(schema)) {
    check_param_class(schema, "character")
    check_param_length(schema)
  }
  if (!is_null(dbname)) {
    check_param_class(dbname, "character")
    check_param_length(dbname)
  }
  enframe(
    get_src_tbl_names(src, schema = sql_to_character(src$con, schema), dbname = dbname),
    name = "table_name",
    value = "remote_name"
  )
}

# FIXME: this should be done using a dplyr function
sql_schema_table_list_postgres <- function(con, schema = NULL) {
  src <- src_from_src_or_con(con)
  if (!is_null(schema)) {
    check_param_class(schema, "character")
    check_param_length(schema)
  }
  enframe(
    get_src_tbl_names(src, schema = sql_to_character(src$con, schema)),
    name = "table_name",
    value = "remote_name"
  )
}

# db_schema_drop() -------------------------------------------------------

#' Remove a schema from a database
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `db_schema_drop()` deletes a schema from the database.
#' For certain DBMS it is possible to force the removal of a non-empty schema, see below.
#'
#' @inheritParams db_schema_create
#' @param force Boolean, default `FALSE`. Set to `TRUE` to drop a schema and
#' all objects it contains at once. Currently only supported for Postgres/Redshift.
#'
#' @details Methods are not available for all DBMS.
#'
#' An error is thrown if no schema of that name exists.
#'
#' The argument `schema` (and `dbname` for MSSQL) can be provided as `SQL` objects.
#' Keep in mind, that in this case it is assumed that they are already correctly quoted as identifiers.
#'
#' Additional arguments are:
#'
#'   - `dbname`: supported for MSSQL. Remove a schema from a different
#'   database on the connected MSSQL-server; default: database addressed by `con`.
#'
#' @return `NULL` invisibly.
#'
#' @family schema handling functions
#' @export
db_schema_drop <- function(con, schema, force = FALSE, ...) {
  check_param_class(schema, "character")
  check_param_length(schema)
  check_param_class(force, "logical")
  check_param_length(force)

  # If we check in the method, we need to specify the user_env argument
  if (inherits(con, "src_dbi")) {
    deprecate_warn(
      "0.2.5",
      'dm::db_schema_drop(con = "must be a DBI connection, not a dbplyr source,")',
    )
  }

  UseMethod("db_schema_drop")
}

#' @export
db_schema_drop.src_dbi <- function(con, schema, force = FALSE, ...) {
  db_schema_drop(con$con, schema, force, ...)
}

#' @export
db_schema_drop.PqConnection <- function(con, schema, force = FALSE, ...) {
  if (force) {
    force_infix <- " and all objects it contained"
    force_suffix <- " CASCADE"
  } else {
    force_infix <- ""
    force_suffix <- ""
  }
  DBI::dbExecute(
    con,
    glue("DROP SCHEMA {DBI::dbQuoteIdentifier(con, schema)}{force_suffix}")
  )
  message(glue::glue("Dropped schema {tick(sql_to_character(con, schema))}{force_infix}."))
  invisible(NULL)
}

#' @export
`db_schema_drop.Microsoft SQL Server` <- function(con, schema, force = FALSE, dbname = NULL, ...) {
  warn_if_arg_not(
    force,
    only_on = c("Postgres", "Redshift"),
    correct = FALSE,
    additional_msg = "Please remove potential objects from the schema manually."
  )
  if (!is_null(dbname)) {
    check_param_class(dbname, "character")
    check_param_length(dbname)
    original_dbname <- attributes(con)$info$dbname
    DBI::dbExecute(con, glue("USE {dbname}"))
    withr::defer(DBI::dbExecute(con, glue("USE {original_dbname}")))
  }
  msg_infix <- fix_msg(sql_to_character(con, dbname))
  DBI::dbExecute(con, glue("DROP SCHEMA {DBI::dbQuoteIdentifier(con, schema)}"))
  message(glue::glue("Dropped schema {tick(sql_to_character(con, schema))}{msg_infix}."))
  invisible(NULL)
}

#' @export
db_schema_drop.SQLiteConnection <- function(con, schema, force = FALSE, ...) {
  abort_no_schemas_supported("SQLite")
}

fix_msg <- function(dbname) {
  if (!is_null(dbname)) {
    msg_suffix <- paste0(" on database ", tick(dbname))
  } else {
    msg_suffix <- ""
  }
}

sql_to_character <- function(con, sql_or_char) {
  if (inherits(sql_or_char, "SQL")) {
    sql_or_char <- map_chr(
      sql_or_char,
      ~ DBI::dbUnquoteIdentifier(con, .x)[[1]] %>% pluck("name")
    )
  }
  sql_or_char
}

test_that("schema handling on MSSQL and Postgres works", {
  skip_if_src_not(c("mssql", "postgres"))

  src_db <- my_test_src()
  con_db <- src_db$con
  sql_schema_table_list <- if (is_postgres(src_db)) {
    sql_schema_table_list_postgres
  } else if (is_mssql(src_db)) {
    sql_schema_table_list_mssql
  }

  expect_dm_error(db_schema_exists(con_db, 1), "parameter_not_correct_class")
  expect_dm_error(db_schema_exists(con_db, letters[1:2]), "parameter_not_correct_length")
  expect_dm_error(db_schema_exists(src_db, 1), "parameter_not_correct_class")
  expect_dm_error(db_schema_exists(src_db, letters[1:2]), "parameter_not_correct_length")

  withr::defer({
    try(DBI::dbExecute(con_db, "DROP TABLE test_schema_1"))
    try(DBI::dbExecute(con_db, DBI::SQL('DROP TABLE "1-dm_schema_TEST"."test_schema_2"')))
    try(DBI::dbExecute(con_db, DBI::SQL('DROP SCHEMA "1-dm_schema_TEST"')))
  })

  expect_false(db_schema_exists(con_db, "1-dm_schema_TEST"))
  expect_deprecated(expect_false(db_schema_exists(src_db, "1-dm_schema_TEST")))

  # create a table in the default schema
  expect_message(db_schema_create(con_db, "1-dm_schema_TEST"), "created")
  expect_error(db_schema_create(con_db, "1-dm_schema_TEST"))
  expect_identical(
    con_db %>%
      db_schema_list(include_default = FALSE) %>%
      filter(schema_name == "1-dm_schema_TEST") %>%
      pull(schema_name),
    "1-dm_schema_TEST"
  )
  expect_true(db_schema_exists(con_db, "1-dm_schema_TEST"))
  expect_message(db_schema_drop(con_db, "1-dm_schema_TEST"), "Dropped schema")
  expect_error(db_schema_drop(con_db, "1-dm_schema_TEST"))
  expect_false(db_schema_exists(con_db, "1-dm_schema_TEST"))

  expect_deprecated(expect_message(db_schema_create(src_db, "1-dm_schema_TEST"), "created"))
  expect_deprecated(expect_true(db_schema_exists(src_db, "1-dm_schema_TEST")))
  expect_message(db_schema_drop(con_db, "1-dm_schema_TEST"), "Dropped schema")
  expect_deprecated(expect_false(db_schema_exists(src_db, "1-dm_schema_TEST")))

  expect_false("test_schema_1" %in% sql_schema_table_list(con_db)$table_name)
  expect_false("test_schema_1" %in% sql_schema_table_list(src_db)$table_name)

  DBI::dbWriteTable(
    con_db,
    DBI::Id(table = "test_schema_1"),
    value = tibble(a = 1:5)
  )

  expect_true("test_schema_1" %in% sql_schema_table_list(con_db)$table_name)
  expect_true("test_schema_1" %in% sql_schema_table_list(src_db)$table_name)

  remote_table_1 <-
    src_db %>%
    sql_schema_table_list() %>%
    filter(table_name == "test_schema_1") %>%
    pull(remote_name)
  expect_identical(
    tbl(src_db, remote_table_1[[1]]) %>% collect(),
    tibble(a = 1:5)
  )

  expect_message(expect_deprecated(db_schema_create(src_db, "1-dm_schema_TEST")), "created")

  DBI::dbWriteTable(
    con_db,
    DBI::Id(schema = "1-dm_schema_TEST", table = "test_schema_2"),
    value = tibble(b = letters[1:5])
  )

  schema_list <- sql_schema_table_list(con_db, schema = "1-dm_schema_TEST")

  expect_identical(
    schema_list,
    tibble(
      table_name = "test_schema_2",
      remote_name = list(DBI::Id(schema = "1-dm_schema_TEST", table = "test_schema_2"))
    )
  )

  remote_table_2 <-
    schema_list %>%
    filter(table_name == "test_schema_2") %>%
    pull(remote_name)

  expect_identical(
    tbl(src_db, remote_table_2[[1]]) %>% collect(),
    tibble(b = letters[1:5])
  )
})

test_that("schema handling on Postgres works", {
  skip_if_src_not("postgres")

  src_db <- my_test_src()
  con_db <- src_db$con

  db_schema_create(con_db, "2-dm_schema_TEST")
  DBI::dbWriteTable(
    con_db,
    DBI::Id(schema = "2-dm_schema_TEST", table = "test_schema_2"),
    value = tibble(b = letters[1:5])
  )

  expect_identical(
    sort(
      filter(
        db_schema_list(con_db, include_default = TRUE),
        schema_name == "public" | schema_name == "2-dm_schema_TEST"
      ) %>%
        pull(schema_name)
    ),
    c("2-dm_schema_TEST", "public")
  )

  expect_message(
    db_schema_drop(con_db, "2-dm_schema_TEST", force = TRUE),
    "all objects"
  )
})

test_that("schema handling on MSSQL works for different DBs", {
  skip_if_src_not("mssql")

  src_db <- my_test_src()
  con_db <- src_db$con

  withr::defer({
    try(DBI::dbExecute(con_db, "DROP DATABASE test_db_for_schema_dm"))
  })

  original_dbname <- attributes(con_db)$info$dbname
  DBI::dbExecute(con_db, "CREATE DATABASE test_db_for_schema_dm")
  expect_false(db_schema_exists(con_db, schema = "test_schema", dbname = "test_db_for_schema_dm"))

  expect_message(
    db_schema_create(con_db, schema = "test_schema", dbname = "test_db_for_schema_dm"),
    "on database `test_db_for_schema_dm`"
  )

  expect_error(
    db_schema_create(con_db, schema = "test_schema", dbname = "test_db_for_schema_dm")
  )

  expect_identical(
    sort(
      db_schema_list(con_db, include_default = TRUE, dbname = "test_db_for_schema_dm")$schema_name
    ),
    c("dbo", "test_schema")
  )

  expect_message(
    db_schema_drop(con_db, schema = "test_schema", dbname = "test_db_for_schema_dm"),
    "on database `test_db_for_schema_dm`"
  )

  expect_message(
    db_schema_create(con_db, schema = "test_schema", dbname = "test_db_for_schema_dm"),
    "on database `test_db_for_schema_dm`"
  )

  expect_true(db_schema_exists(con_db, schema = "test_schema", dbname = "test_db_for_schema_dm"))

  expect_identical(
    sql_schema_table_list_mssql(con_db, schema = "test_schema", dbname = "test_db_for_schema_dm"),
    tibble(
      table_name = character(),
      remote_name = list()
    )
  )

  DBI::dbWriteTable(
    con_db,
    DBI::Id(catalog = "test_db_for_schema_dm", schema = "test_schema", table = "test_1"),
    value = tibble(c = c(5L))
  )

  expect_error(
    expect_warning(
      db_schema_drop(con_db, "test_schema", dbname = "test_db_for_schema_dm", force = TRUE),
      "Argument `force` ignored:"
    )
  )

  expect_identical(
    sql_schema_table_list_mssql(con_db, schema = "test_schema", dbname = "test_db_for_schema_dm"),
    tibble(
      table_name = "test_1",
      remote_name = list(DBI::Id(
        catalog = "test_db_for_schema_dm",
        schema = "test_schema",
        table = "test_1"
      ))
    )
  )
})

test_that("schema handling on SQLite all throw errors", {
  skip_if_src_not("sqlite")

  src_db <- my_test_src()
  con_db <- src_db$con

  expect_dm_error(
    expect_deprecated(db_schema_exists(src_db, "test"), "DBI connection"),
    "no_schemas_supported"
  )
  expect_dm_error(
    expect_deprecated(db_schema_list(src_db), "DBI connection"),
    "no_schemas_supported"
  )
  expect_dm_error(
    expect_deprecated(db_schema_drop(src_db, "test"), "DBI connection"),
    "no_schemas_supported"
  )
  expect_dm_error(
    expect_deprecated(db_schema_exists(src_db, "test"), "DBI connection"),
    "no_schemas_supported"
  )

  expect_dm_error(
    db_schema_exists(con_db, "test"),
    "no_schemas_supported"
  )
  expect_dm_error(
    db_schema_list(con_db),
    "no_schemas_supported"
  )
  expect_dm_error(
    db_schema_drop(con_db, "test"),
    "no_schemas_supported"
  )
  expect_dm_error(
    db_schema_exists(con_db, "test"),
    "no_schemas_supported"
  )
})

#' Create R code for a dm object
#'
#' @description
#' `dm_paste()` takes an existing `dm` and emits the code necessary for its creation.
#'
#' @inheritParams dm_add_pk
#' @param select
#'   Deprecated, see `"select"` in the `options` argument.
#' @param ... Must be empty.
#' @param tab_width Indentation width for code from the second line onwards
#' @param options Formatting options. A character vector containing some of:
#'   - `"tables"`: [tibble()] calls for empty table definitions
#'     derived from [dm_ptype()], overrides `"select"`.
#'   - `"select"`: [dm_select()] statements for columns that are part
#'     of the dm.
#'   - `"keys"`: [dm_add_pk()], [dm_add_fk()] and [dm_add_uk()] statements for adding keys.
#'   - `"color"`: [dm_set_colors()] statements to set color.
#'   - `"all"`: All options above except `"select"`
#'
#'   Default `NULL` is equivalent to `c("keys", "color")`
#' @param path Output file, if `NULL` the code is printed to the console.
#'
#' @details
#' The code emitted by the function reproduces the structure of the `dm` object.
#' The `options` argument controls the level of detail: keys, colors,
#' table definitions.
#' Data in the tables is never included, see [dm_ptype()] for the underlying logic.
#'
#' @return Code for producing the prototype of the given `dm`.
#'
#' @export
#' @examples
#' dm() %>%
#'   dm_paste()
#' @examplesIf rlang::is_installed("nycflights13")
#'
#' dm_nycflights13() %>%
#'   dm_paste()
#'
#' dm_nycflights13() %>%
#'   dm_paste(options = "select")
dm_paste <- function(dm, select = NULL, ..., tab_width = 2, options = NULL, path = NULL) {
  dm_local_error_call()
  check_dots_empty(action = warn)

  options <- check_paste_options(options, select, caller_env())

  if (!is.null(path)) {
    check_suggested("brio", "dm_paste")
  }

  code <- dm_paste_impl(dm = dm, options, tab_width = tab_width)

  if (is.null(path)) {
    cli::cli_code(code)
  } else {
    brio::write_lines(code, path)
  }
  invisible(dm)
}

check_paste_options <- function(options, select, env) {
  allowed_options <- c("all", "tables", "keys", "select", "color")

  if (is.null(options)) {
    options <- c("keys", "color")
  } else {
    if (!all(options %in% allowed_options)) {
      abort_unknown_option(options, allowed_options)
    }
  }

  if (!is.null(select)) {
    deprecate_warn(
      "0.1.2",
      "dm::dm_paste(select = )",
      "dm::dm_paste(options = 'select')",
      env = env
    )
    if (isTRUE(select)) {
      options <- c(options, "select")
    }
  }

  if ("all" %in% options) {
    options <- allowed_options
  }

  if ("tables" %in% options) {
    options <- setdiff(options, "select")
  }

  options
}

dm_paste_impl <- function(dm, options, tab_width, chunk_size = 100) {
  check_not_zoomed(dm)
  check_no_filter(dm)

  tab <- paste0(rep(" ", tab_width), collapse = "")

  # code for including table definitions
  code_tables <- if ("tables" %in% options) dm_paste_tables(dm, tab)

  # code for including the tables
  code_construct <- dm_paste_construct(dm, tab)

  # adding code for selection of columns
  code_select <- if ("select" %in% options) dm_paste_select(dm)

  # adding code for establishing PKs
  code_pks <- if ("keys" %in% options) dm_paste_pks(dm)

  # adding code for establishing UKs
  code_uks <- if ("keys" %in% options) dm_paste_uks(dm)

  # adding code for establishing FKs
  code_fks <- if ("keys" %in% options) dm_paste_fks(dm)

  # adding code for color
  code_color <- if ("color" %in% options) dm_paste_color(dm)

  # collect operations as named parts, preserving group boundaries for chunking
  parts <- Filter(
    function(x) length(x) > 0,
    list(code_select, code_pks, code_uks, code_fks, code_color)
  )

  # combine dm construction and operations, chunking at group boundaries if needed
  code_dm <- dm_paste_chunk_operations(
    code_construct,
    parts,
    tab,
    chunk_size = chunk_size
  )

  paste0(code_tables, code_dm)
}

dm_paste_tables <- function(dm, tab) {
  ptype <- dm_ptype(dm)

  tables <-
    ptype %>%
    dm_get_tables() %>%
    map_chr(df_paste, tab)

  glue_collapse1(
    glue("{tick_if_needed(names(tables))} <- {tables}\n\n", .trim = FALSE)
  )
}

dm_paste_construct <- function(dm, tab) {
  if (length(dm) == 0) {
    return("dm::dm(\n)")
  }

  paste0(
    "dm::dm(\n",
    paste0(tab, tick_if_needed(src_tbls_impl(dm)), ",\n", collapse = ""),
    ")"
  )
}

#' @autoglobal
dm_paste_select <- function(dm) {
  tbl_select <-
    dm %>%
    dm_get_def() %>%
    mutate(cols = map(data, colnames)) %>%
    mutate(cols = map_chr(cols, ~ glue_collapse1(glue(", {tick_if_needed(.x)}")))) %>%
    mutate(code = glue("dm::dm_select({tick_if_needed(table)}{cols})")) %>%
    pull()
}

dm_paste_pks <- function(dm) {
  dm %>%
    dm_get_all_pks_impl() %>%
    mutate(
      code = if_else(
        !is.na(autoincrement) & autoincrement,
        glue(
          "dm::dm_add_pk({tick_if_needed(table)}, {deparse_keys(pk_col)}, autoincrement = TRUE)"
        ),
        glue("dm::dm_add_pk({tick_if_needed(table)}, {deparse_keys(pk_col)})")
      )
    ) %>%
    pull()
}

dm_paste_uks <- function(dm) {
  dm %>%
    dm_get_def() %>%
    dm_get_all_uks_def_impl() %>%
    mutate(code = glue("dm::dm_add_uk({tick_if_needed(table)}, {deparse_keys(uk_col)})")) %>%
    pull()
}

dm_paste_fks <- function(dm) {
  pks <-
    dm %>%
    dm_get_all_pks_impl() %>%
    set_names(c("parent_table", "parent_default_pk_cols", "autoincrement"))

  fks <-
    dm %>%
    dm_get_all_fks_impl()

  fpks <-
    left_join(fks, pks, by = "parent_table")

  need_non_default <- !map2_lgl(fpks$parent_key_cols, fpks$parent_default_pk_cols, identical)
  fpks$non_default_parent_key_cols <- ""
  fpks$non_default_parent_key_cols[need_non_default] <- paste0(
    ", ",
    deparse_keys(fpks$parent_key_cols[need_non_default])
  )

  on_delete <- if_else(
    fpks$on_delete != "no_action",
    glue(", on_delete = \"{fpks$on_delete}\""),
    ""
  )

  glue(
    "dm::dm_add_fk({tick_if_needed(fpks$child_table)}, {deparse_keys(fpks$child_fk_cols)}, {tick_if_needed(fpks$parent_table)}{fpks$non_default_parent_key_cols}{on_delete})"
  )
}

dm_paste_color <- function(dm) {
  colors <- dm_get_colors(dm)
  colors <- colors[names(colors) != "default"]
  glue("dm::dm_set_colors({tick_if_needed(names(colors))} = {tick_if_needed(colors)})")
}

df_paste <- function(x, tab) {
  cols <- map_chr(x, deparse_line)
  if (is_empty(x)) {
    cols <- character()
  } else {
    cols <- paste0(tab, tick_if_needed(names(cols)), " = ", cols, ",\n", collapse = "")
  }

  paste0("tibble::tibble(\n", cols, ")")
}

dm_paste_chunk_operations <- function(code_construct, parts, tab, chunk_size = 100) {
  # Build chunks by accumulating complete groups (parts), never splitting a group
  # across chunks unless the group itself exceeds chunk_size.
  chunks_ops <- list()
  current_ops <- character()

  for (part in parts) {
    if (length(current_ops) + length(part) <= chunk_size) {
      # Entire part fits in the current chunk
      current_ops <- c(current_ops, part)
    } else {
      # Part doesn't fit — flush the current chunk and handle the overflow
      chunks_ops <- c(chunks_ops, list(current_ops))
      current_ops <- character()
      # Split the part into sub-chunks when it exceeds chunk_size on its own
      sub_chunks <- split(part, ceiling(seq_along(part) / chunk_size))
      for (j in seq_along(sub_chunks)) {
        if (j < length(sub_chunks)) {
          chunks_ops <- c(chunks_ops, list(sub_chunks[[j]]))
        } else {
          current_ops <- sub_chunks[[j]]
        }
      }
    }
  }
  chunks_ops <- c(chunks_ops, list(current_ops))

  # Remove a trailing empty chunk (can arise if the last part was exactly split
  # into sub-chunks), but always keep at least one chunk for the construct.
  if (length(chunks_ops) > 1 && length(chunks_ops[[length(chunks_ops)]]) == 0) {
    chunks_ops <- chunks_ops[-length(chunks_ops)]
  }

  n_chunks <- length(chunks_ops)

  # Single chunk: use a simple pipe chain with no intermediate variables
  if (n_chunks == 1) {
    return(glue_collapse(
      c(code_construct, chunks_ops[[1]]),
      sep = glue(" %>%\n{tab}", .trim = FALSE)
    ))
  }

  # Multiple chunks: generate intermediate variable assignments
  result_lines <- character()
  for (i in seq_along(chunks_ops)) {
    is_last <- i == n_chunks
    var_name <- if (!is_last) paste0("dm_step_", i) else NULL
    ops <- chunks_ops[[i]]

    if (i == 1) {
      chunk_code <- glue_collapse(
        c(code_construct, ops),
        sep = glue(" %>%\n{tab}", .trim = FALSE)
      )
    } else {
      prev_var <- paste0("dm_step_", i - 1)
      chunk_code <- paste0(
        prev_var,
        " %>%\n",
        tab,
        glue_collapse(ops, sep = glue(" %>%\n{tab}", .trim = FALSE))
      )
    }

    result_lines <- c(
      result_lines,
      if (!is.null(var_name)) paste0(var_name, " <- ", chunk_code) else chunk_code
    )
  }

  paste(result_lines, collapse = "\n\n")
}

deparse_line <- function(x) {
  attrs <- attributes(x)
  # Workaround necessary for R < 3.5:
  if (length(attrs) > 0) {
    attributes(x) <- attrs[sort(names(attrs))]
  }
  x <- deparse(x, width.cutoff = 500, backtick = TRUE)
  # need paste0() because of https://github.com/cynkra/dm/issues/1510
  paste0(gsub(" *\n *", " ", x), collapse = "")
}

glue_collapse1 <- function(x, ...) {
  if (is_empty(x)) {
    ""
  } else {
    glue_collapse(x, ...)
  }
}

dquote <- function(x) {
  if (is_empty(x)) {
    return(character())
  }
  paste0('"', x, '"')
}

# Errors ------------------------------------------------------------------

abort_unknown_option <- function(options, all_options) {
  bad_options <- setdiff(options, all_options)
  cli::cli_abort(
    "{qty(length(bad_options))}Option{?s} unknown: {.val {bad_options}}. Must be one of {.val {all_options}}.",
    class = dm_error_full("unknown_option"),
    call = dm_error_call()
  )
}

test_that("path argument", {
  path <- tempfile()
  dm() %>% dm_paste(path = path)
  expect_identical(readLines(path), c("dm::dm(", ")"))
})

test_that("output", {
  local_options(lifecycle_verbosity = "warning")

  expect_snapshot({
    "empty"
    empty_dm() %>% dm_paste()

    "empty table"
    dm(a = tibble()) %>% dm_paste(options = "tables")

    "baseline"
    dm_for_filter() %>% dm_paste()

    "changing the tab width"
    dm_for_filter() %>% dm_paste(tab_width = 4)

    "we don't care if the tables really exist"
    dm_for_filter() %>%
      dm_rename_tbl(tf_1_new = tf_1) %>%
      dm_paste()

    "produce `dm_select()` statements in addition to the rest"
    dm_for_filter() %>%
      dm_select(tf_5, k = k, m) %>%
      dm_select(tf_1, a) %>%
      dm(x = copy_to_my_test_src(tibble(q = 1L), qq)) %>%
      dm_paste(options = "select")

    "produce code with colors"
    dm_for_filter() %>%
      dm_set_colors("orange" = tf_1:tf_3, "darkgreen" = tf_5:tf_6) %>%
      dm_paste()

    "tick if needed"
    a <- tibble(x = 1)
    names(a) <- "a b"
    dm(a) %>%
      dm_zoom_to(a) %>%
      dm_insert_zoomed("a b") %>%
      dm_add_pk(a, "a b") %>%
      dm_add_fk("a b", "a b", a) %>%
      dm_set_colors(green = "a b") %>%
      dm_paste(options = "all")

    "FK referencing non default PK"
    b <- tibble(x = 1, y = "A", z = "A")
    c <- tibble(x = "A", y = "A")

    dm(b, c) %>%
      dm_add_pk(c, x) %>%
      dm_add_fk(b, y, c) %>%
      dm_add_fk(b, z, c, y) %>%
      dm_paste()

    # UKs
    dm_for_filter() %>%
      dm_add_uk(tf_5, l) %>%
      dm_add_uk(tf_6, n) %>%
      dm_paste()

    "on_delete if needed"
    dm(b, c) %>%
      dm_add_pk(c, x) %>%
      dm_add_fk(b, y, c, on_delete = "cascade") %>%
      dm_add_fk(b, z, c, y, on_delete = "no_action") %>%
      dm_paste()

    "all of nycflights13"
    dm_nycflights13() %>%
      dm_paste(options = "all")

    "deprecation warning for select argument"
    dm() %>%
      dm_paste(select = TRUE)

    "error for bad option"
    writeLines(conditionMessage(
      expect_error(dm_paste(dm(), options = c("bogus", "all", "mad")))
    ))
  })
})

test_that("output 2", {
  skip_if(getRversion() < "4.2")
  local_options(lifecycle_verbosity = "warning")

  expect_snapshot({
    "no error for factor column that leads to code with width > 500"
    dm(tibble(
      a = factor(
        levels = expand.grid(
          letters,
          as.character(1:5)
        ) %>%
          transmute(x = paste0(Var1, Var2)) %>%
          pull()
      )
    )) %>%
      dm_paste(options = "tables")
  })
})

test_that("chunking behavior for large dm", {
  create_large_dm <- function(n_tables) {
    main_table <- tibble(id = integer(0))
    tables <- list(main = main_table)
    for (i in 1:n_tables) {
      table_name <- paste0("table_", i)
      tables[[table_name]] <- tibble(
        id = integer(0),
        main_id = integer(0)
      )
    }
    dm_obj <- do.call(dm, tables)
    dm_obj <- dm_obj %>% dm_add_pk(main, id)
    for (i in 1:n_tables) {
      table_name <- paste0("table_", i)
      dm_obj <- dm_obj %>%
        dm_add_pk(!!table_name, id) %>%
        dm_add_fk(!!table_name, main_id, main)
    }
    dm_obj
  }

  # Use chunk_size = 3 with a 3-table dm (7 operations) to verify chunking output
  expect_snapshot(
    writeLines(dm:::dm_paste_impl(create_large_dm(3), c("keys"), 2, chunk_size = 3))
  )
})

cache_attach <- function(
  algo = "sha512",
  base_attach = attach,
  name = paste0(utils::packageName(), "_cache")
) {
  force(algo)

  if (!rlang::is_attached(name)) {
    env <- rlang::new_environment(list(...cache = rlang::new_environment()))
    base_attach(env, pos = length(search()) - 1, name = name)
  }
  cache <- rlang::search_env(name)$...cache

  cache_reset <- function() {
    rm(list = ls(cache), envir = cache)
  }

  cache_set <- function(key, value) {
    assign(key, value, envir = cache)
  }

  cache_get <- function(key) {
    get(key, envir = cache, inherits = FALSE)
  }

  cache_has_key <- function(key) {
    exists(key, envir = cache, inherits = FALSE)
  }

  list(
    digest = function(...) digest::digest(..., algo = algo),
    reset = cache_reset,
    set = cache_set,
    get = cache_get,
    has_key = cache_has_key,
    keys = function() ls(cache)
  )
}

replay_html_dm_draw <- function(x, ...) {
  unclass(x)
}

# Use custom name to avoid clash if pkgload is active
pkgdown_print_grViz <- function(x, ...) {
  structure(DiagrammeRsvg::export_svg(x), class = "dm_draw")
}

register_pkgdown_methods <- function() {
  if (Sys.getenv("IN_PKGDOWN") == "") {
    return()
  }

  # FIXME Should we use Config/Needs/website in DESCRIPTION instead?
  check_suggested(c("DiagrammeR", "DiagrammeRsvg"), "register_pkgdown_methods")

  # For dev pkgdown
  s3_register("downlit::replay_html", "dm_draw", replay_html_dm_draw)
  s3_register("pkgdown::pkgdown_print", "grViz", pkgdown_print_grViz)
}

# FIXME: make more robust as we implement all json functions
unjson_nested <- function(data) {
  safe_fromJSON <- function(x) {
    tryCatch(map(x, jsonlite::fromJSON), error = function(e) x)
  }
  mutate(data, across(where(is.character), safe_fromJSON))
}

#' @autoglobal
dm_disentangle <- function(dm, start, quiet = FALSE) {
  check_not_zoomed(dm)
  start <- dm_tbl_name(dm, {{ start }})
  recipes <- enumerate_all_paths(dm, start)
  changed <- arrange(recipes$table_mapping, table, new_table)

  if (!quiet) {
    msgs <-
      changed %>%
      group_by(table) %>%
      summarize(
        msg = glue::glue("Replaced table {tick(unique(table))} with {commas(tick(new_table))}.")
      ) %>%
      ungroup()

    walk(msgs$msg, message)
  }
  fk_table <- fk_table_to_def_fks(
    recipes$new_fks,
    child_table = "new_child_table",
    child_fk_cols = "child_cols",
    parent_table = "new_parent_table",
    parent_key_cols = "parent_cols"
  ) %>%
    rename(new_fks = fks)

  dm_get_def(dm) %>%
    left_join(changed, by = "table", multiple = "all") %>%
    mutate(table = coalesce(new_table, table)) %>%
    select(-new_table) %>%
    left_join(fk_table, by = c("table" = "new_parent_table")) %>%
    select(-fks) %>%
    relocate(fks = new_fks, .after = uks) %>%
    mutate(fks = vctrs::as_list_of(map(fks, ~ .x %||% new_fk()))) %>%
    dm_from_def()
}

test_that("dm_disentangle() works", {
  expect_snapshot({
    dm_disentangle(dm_for_filter_w_cycle(), tf_1) %>%
      dm_get_all_fks()
    dm_disentangle(dm_for_filter_w_cycle(), tf_5) %>%
      dm_get_all_fks()
    dm_disentangle(entangled_dm(), a, quiet = TRUE) %>%
      dm_get_all_fks()
    dm_disentangle(entangled_dm(), c) %>%
      dm_get_all_fks()
    dm_disentangle(entangled_dm_2(), a) %>%
      dm_get_all_fks()
    dm_disentangle(entangled_dm_2(), d, quiet = TRUE) %>%
      dm_get_all_fks()
  })

  expect_snapshot_error(class = dm_error("only_possible_wo_zoom"), {
    dm_for_filter_w_cycle() %>%
      dm_zoom_to(tf_1) %>%
      dm_disentangle(tf_1)
  })

  expect_snapshot_error({
    dm_for_filter_w_cycle() %>%
      dm_disentangle()
  })
})

is_dark_color <- function(rgb) {
  # according to https://stackoverflow.com/questions/3942878/how-to-decide-font-color-in-white-or-black-depending-on-background-color
  # and https://www.w3.org/WAI/WCAG21/Techniques/general/G18.html
  # you should calculate luminance first and then compare its contrast to white and black
  #
  # But after evaluating results from this approach and taking the alpha channel into account,
  # it seems more natural to use a formula like this:
  sum(rgb[1:3] / 255. * c(0.3, 0.5, 0.2)) + (255. - rgb[4]) / 255. * 0.2 < 0.6
  #  + (255 - rgb[4]) / 255. * 0.15
  # the weights are inspired by the sources mentioned above (only the differences between them are much smaller here)
}

is_hex_color <- function(x) {
  grepl("^#[A-Fa-f0-9]{6}([A-Fa-f0-9]{2})?$", x)
}

col_to_hex <- function(x) {
  # if not all colors that are not hex coded colors are available, abort
  if (!all(x[!is_hex_color(x)] %in% dm_get_available_colors())) {
    abort_cols_not_avail(setdiff(
      x[!is_hex_color(x)],
      dm_get_available_colors()
    ))
  }

  # from hex or name to rgb; "default" should remain "default"
  is_not_default <- which(x != "default")
  x[is_not_default] <- hex_from_rgb(col2rgb(x[is_not_default], alpha = TRUE))
  x
}

hex_from_rgb <- function(col_rgb) {
  rgb(col_rgb[1, ], col_rgb[2, ], col_rgb[3, ], col_rgb[4, ], maxColorValue = 255)
}

calc_bodycol_rgb <- function(header_bgcol_rgb, ratio = 0.8) {
  # alpha channel remains the same for the body color
  header_bgcol_rgb[1:3, ] <- header_bgcol_rgb[1:3, ] + (255 - header_bgcol_rgb[1:3, ]) * ratio
  header_bgcol_rgb
}

test_that("color related functions work", {
  # is_dark_color()
  expect_true(is_dark_color(rgb = c(100, 100, 100, 255)))
  expect_false(is_dark_color(rgb = c(142, 164, 143, 255)))

  # is_hex_color()
  expect_true(is_hex_color("#92A87F"))
  expect_true(is_hex_color("#92A87F4C"))
  expect_false(is_hex_color("#92A87F4"))
  expect_false(is_hex_color("92A87F"))

  # col_to_hex()
  expect_identical(
    col_to_hex(c("brown", "#92A87F", "darkgreen", "#92A87F4C")),
    c("#A52A2AFF", "#92A87FFF", "#006400FF", "#92A87F4C")
  )

  expect_dm_error(
    col_to_hex("darklightpink"),
    "cols_not_avail"
  )

  # hex_from_rgb()
  expect_identical(
    hex_from_rgb(matrix(100, 4, 1)),
    "#64646464"
  )

  expect_identical(
    hex_from_rgb(matrix(100:107, 4, 2)),
    c("#64656667", "#68696A6B")
  )

  expect_identical(
    calc_bodycol_rgb(matrix(c(100, 20, 200, 155), 4, 1)),
    matrix(c(224, 208, 244, 155), 4, 1)
  )
})

#' \pkg{utils} table manipulation methods for `dm_zoomed` objects
#'
#' @description
#' Extract the first or last rows from a table.
#' Use these methods without the '.dm_zoomed' suffix (see examples).
#' The methods for regular `dm` objects extract the first or last tables.
#'
#' @param x object of class `dm_zoomed`
#' @inheritParams utils::head
#' @rdname utils_table_manipulation
#'
#' @return A `dm_zoomed` object.
#'
#' @details see manual for the corresponding functions in \pkg{utils}.
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' zoomed <- dm_nycflights13() %>%
#'   dm_zoom_to(flights) %>%
#'   head(4)
#' zoomed
#' dm_insert_zoomed(zoomed, new_tbl_name = "head_flights")
#' @export
head.dm_zoomed <- function(x, n = 6L, ...) {
  # dm method provided by utils
  replace_zoomed_tbl(x, head(tbl_zoomed(x), n, ...))
}

#' @rdname utils_table_manipulation
#' @export
tail.dm_zoomed <- function(x, n = 6L, ...) {
  # dm method provided by utils
  replace_zoomed_tbl(x, tail(tbl_zoomed(x), n, ...))
}

fast_tibble <- function(...) {
  out <- vctrs::df_list(...)
  tibble::new_tibble(out)
}

recode_compat <- function(x, y) {
  x <- unname(x)
  idx <- vctrs::vec_match(names(y), x)
  vctrs::vec_assign(x, idx, y)
}

library(tidyverse)
library(dbplyr)
pkgload::load_all()

dm <- dm_nycflights13(cycle = FALSE)

src <- src_sqlite(":memory:", create = TRUE)

dm_sqlite <- dm_copy_to(src, dm)

dm1 <-
  dm_sqlite %>%
  dm_filter(airports, faa == "EWR")

dm1 %>% tbl_impl("airports")
dm1 %>% tbl_impl("flights")
dm1 %>% tbl_impl("weather")
dm1 %>% tbl_impl("airlines")
dm1 %>% tbl_impl("planes")

dm1 %>% dm_get_tables()

dm1 %>% dm_get_tables() %>% map(nrow)

dm1 %>% tbl_impl("airports") %>% sql_render()
dm1 %>% tbl_impl("flights") %>% sql_render()
dm1 %>% tbl_impl("weather") %>% sql_render()
dm1 %>% tbl_impl("airlines") %>% sql_render()
dm1 %>% tbl_impl("planes") %>% sql_render()

dm1 %>% tbl_impl("airports") %>% count() %>% explain()
dm1 %>% tbl_impl("flights") %>% count() %>% explain()
dm1 %>% tbl_impl("weather") %>% count() %>% explain()
dm1 %>% tbl_impl("airlines") %>% count() %>% explain()
dm1 %>% tbl_impl("planes") %>% count() %>% explain()

dm1 %>% tbl_impl("airports") %>% count()
dm1 %>% tbl_impl("flights") %>% count()
dm1 %>% tbl_impl("weather") %>% count()
dm1 %>% tbl_impl("airlines") %>% count()
dm1 %>% tbl_impl("planes") %>% count()

new_data_model <- function(tables, columns, references) {
  empty_tables <- setdiff(tables$table, columns$table)
  if (length(empty_tables)) {
    empty_table_columns <- tibble(table = empty_tables, column = "", key = 0)
    columns <- bind_rows(columns, empty_table_columns)
  }
  stopifnot(nrow(tables) > 0)
  stopifnot(nrow(columns) > 0)

  structure(
    list(
      tables = tables,
      columns = columns,
      references = references
    ),
    class = "data_model"
  )
}

#' Flatten a table in a `dm` by joining its parent tables
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `dm_flatten()` updates a table in-place by joining its parent tables into it,
#' and removes the now integrated parent tables from the dm.
#'
#' @inheritParams dm_flatten_to_tbl
#' @param table The table to flatten by joining its parent tables.
#'   An interesting choice could be
#'   for example a fact table in a star schema.
#' @param ... These dots are for future extensions and must be empty.
#' @param parent_tables
#'   `r lifecycle::badge("experimental")`
#'
#'   Unquoted names of the parent tables to be joined into `table`.
#'   The order of the tables here determines the order of the joins.
#'   If `NULL` (the default), all direct parent tables
#'   are joined in non-recursive mode,
#'   or all reachable ancestor tables in recursive mode.
#'   `tidyselect` is supported, see [dplyr::select()] for details on the semantics.
#' @param recursive Logical, defaults to `FALSE`.
#'   If `TRUE`, recursively flatten parent tables before joining them
#'   into `table`.
#'   Uses simple recursion: recursively flattening the parents
#'   and then doing a join in order.
#'   If `FALSE`, fails if a parent table has further parents
#'   (unless `allow_deep` is `TRUE`).
#'   Cannot be `TRUE` when `allow_deep` is `TRUE`.
#' @param allow_deep Logical, defaults to `FALSE`.
#'   Only relevant if `recursive = FALSE`.
#'   If `TRUE`, parent tables with further parents are allowed
#'   and will remain in the result with a
#'   foreign-key relationship to the flattened table.
#'   Cannot be `TRUE` when `recursive` is `TRUE`.
#' @param join The type of join to use when combining parent tables,
#'   see [dplyr::join()].
#'   Defaults to [dplyr::left_join()].
#'   `nest_join` is not supported.
#'   When `recursive = TRUE`, only [dplyr::left_join()],
#'   [dplyr::inner_join()], and [dplyr::full_join()] are supported.
#'
#' @return A [`dm`] object with the flattened table and removed parent tables.
#'
#' @family flattening functions
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_select_tbl(-weather) %>%
#'   dm_flatten(flights, recursive = TRUE)
#'
#' @export
dm_flatten <- function(
  dm,
  table,
  ...,
  parent_tables = NULL,
  recursive = FALSE,
  allow_deep = FALSE,
  join = left_join
) {
  check_not_zoomed(dm)
  check_no_filter(dm)
  check_dots_empty()
  dm_local_error_call()

  join_name <- as_label(enexpr(join))

  if (join_name == "nest_join") {
    abort_no_flatten_with_nest_join("dm_flatten")
  }

  if (recursive && !(join_name %in% c("left_join", "full_join", "inner_join"))) {
    abort_squash_limited("dm_flatten", "recursive")
  }

  if (recursive && allow_deep) {
    cli::cli_abort(
      "{.arg allow_deep} can't be {.code TRUE} when {.arg recursive} is {.code TRUE}.",
      class = dm_error_full("recursive_and_allow_deep"),
      call = dm_error_call()
    )
  }

  start <- dm_tbl_name(dm, {{ table }})

  parent_tables_quo <- enquo(parent_tables)

  all_fks <- dm_get_all_fks_impl(dm, ignore_on_delete = TRUE)

  # Find direct parents of start
  direct_parents <- all_fks %>%
    filter(child_table == start) %>%
    pull(parent_table) %>%
    unique()

  # Auto-detect or evaluate parent tables
  if (quo_is_null(parent_tables_quo)) {
    if (recursive) {
      # In recursive mode, default to all reachable ancestor tables
      g <- create_graph_from_dm(dm, directed = TRUE)
      list_of_pts <- get_names_of_connected(g, start, squash = TRUE)
    } else {
      list_of_pts <- direct_parents
    }
  } else {
    vars <- setdiff(src_tbls_impl(dm), start)
    list_of_pts <- eval_select_table(quo(c(!!parent_tables_quo)), vars)
  }

  # Early return if nothing to flatten
  if (is_empty(list_of_pts)) {
    return(dm)
  }

  # Validate: all listed tables must be direct parents (or reachable in recursive mode)
  non_parents <- setdiff(list_of_pts, direct_parents)
  if (length(non_parents) > 0) {
    if (recursive) {
      g <- create_graph_from_dm(dm, directed = TRUE)
      reachable <- get_names_of_connected(g, start, squash = TRUE)
      non_reachable <- setdiff(non_parents, reachable)
      if (length(non_reachable) > 0) {
        abort_tables_not_reachable_from_start("table")
      }
    } else {
      g <- create_graph_from_dm(dm, directed = TRUE)
      reachable <- get_names_of_connected(g, start, squash = TRUE)
      if (all(non_parents %in% reachable)) {
        abort_only_parents("dm_flatten", "table", "recursive")
      } else {
        abort_tables_not_reachable_from_start("table")
      }
    }
  }

  # Non-recursive: check for deeper hierarchy
  if (!recursive) {
    has_parents <- all_fks %>%
      filter(child_table %in% list_of_pts) %>%
      nrow()
    if (has_parents > 0 && !allow_deep) {
      abort_only_parents("dm_flatten", "table", "recursive")
    }
  }

  # Run DFS once for cycle detection and ordering (used by recursive path)
  if (recursive) {
    g <- create_graph_from_dm(dm, directed = TRUE)
    g_sub <- graph_induced_subgraph(g, c(start, list_of_pts))
    if (length(graph_vertices(g_sub)) - 1 != length(graph_edges(g_sub))) {
      abort_no_cycles(g_sub)
    }

    dfs <- graph_dfs(g_sub, start, unreachable = FALSE, dist = TRUE)
    dfs_order <- names(dfs$order) %>% discard(is.na)
  } else {
    dfs_order <- NULL
  }

  out <- dm_flatten_impl(dm, start, list_of_pts, dfs_order, join)

  # Handle allow_deep: transfer FKs from absorbed parents to start
  if (allow_deep) {
    parents <- intersect(direct_parents, list_of_pts)
    out$dm <- dm_flatten_transfer_fks(out$dm, start, parents, out$col_renames, all_fks)
  }

  dm_flatten_explain_renames(out$all_renames)

  out$dm
}

#' Flatten a start table by joining its direct parents
#'
#' Workhorse that performs joining of parent tables into a start table.
#' In recursive mode (`dfs_order` not `NULL`), first recursively flattens each
#' direct parent's own ancestors via a loop, then unconditionally joins
#' the resulting direct parents into the start table.
#'
#' @param dm A `dm` object.
#' @param start Name of the table to flatten into.
#' @param list_of_pts Character vector of parent table names to absorb.
#' @param dfs_order Character vector giving DFS traversal order from the main
#'   function, or `NULL` for non-recursive mode.
#' @param join The join function to use (e.g. [dplyr::left_join()]).
#'
#' @return A list with components:
#'   - `dm`: the updated `dm` with parents joined into `start` and removed.
#'   - `all_renames`: list of rename entries from column disambiguation.
#'   - `col_renames`: named list mapping parent names to their column renames.
#'
#' @noRd
#' @autoglobal
dm_flatten_impl <- function(dm, start, list_of_pts, dfs_order, join) {
  all_fks <- dm_get_all_fks_impl(dm, ignore_on_delete = TRUE)

  # Find direct parents of start
  direct_parents <- all_fks %>%
    filter(child_table == start) %>%
    pull(parent_table) %>%
    unique()

  all_renames <- list()

  # --- Recursive part: pre-flatten each direct parent's ancestors ---
  if (!is.null(dfs_order)) {
    parents_to_process <- intersect(direct_parents, list_of_pts)
    parents_to_process <- intersect(dfs_order, parents_to_process)

    for (pt in parents_to_process) {
      current_fks <- dm_get_all_fks_impl(dm, ignore_on_delete = TRUE)
      pt_direct_parents <- current_fks %>%
        filter(child_table == pt) %>%
        pull(parent_table) %>%
        unique()
      pt_parents_in_list <- intersect(pt_direct_parents, list_of_pts)
      pt_parents_in_list <- intersect(pt_parents_in_list, src_tbls_impl(dm))

      if (length(pt_parents_in_list) > 0) {
        # Recurse: flatten pt's ancestors into pt
        out <- dm_flatten_impl(dm, pt, pt_parents_in_list, dfs_order, join)
        dm <- out$dm
        all_renames <- c(all_renames, out$all_renames)
      }
    }
  }

  # --- Non-recursive part: join direct parents into start ---
  current_fks <- dm_get_all_fks_impl(dm, ignore_on_delete = TRUE)
  parents_to_join <- current_fks %>%
    filter(child_table == start) %>%
    pull(parent_table) %>%
    unique()
  parents_to_join <- intersect(parents_to_join, intersect(direct_parents, list_of_pts))
  parents_to_join <- intersect(parents_to_join, src_tbls_impl(dm))

  out <- dm_flatten_join(dm, start, parents_to_join, join)
  out$all_renames <- c(all_renames, out$all_renames)
  out
}

#' Join parent tables into a start table and remove them
#'
#' Performs sequential `left_join()` operations to merge each parent table into
#' the start table.
#' Conflicting column names are disambiguated by appending
#' `.parent_name` as a suffix.
#' The parent tables are removed from the dm after joining.
#'
#' @param dm A `dm` object.
#' @param start Name of the table to join parents into.
#' @param parents Character vector of parent table names to join.
#' @param join The join function to use (e.g. [dplyr::left_join()]).
#'
#' @return A list with components:
#'   - `dm`: the updated `dm` with parents joined and removed.
#'   - `all_renames`: list of rename entries recording disambiguated columns.
#'   - `col_renames`: named list mapping each parent to its column renames
#'     (used by `dm_flatten_transfer_fks()`).
#'
#' @noRd
#' @autoglobal
dm_flatten_join <- function(dm, start, parents, join) {
  if (is_empty(parents)) {
    return(list(dm = dm, all_renames = list(), col_renames = list()))
  }

  all_fks <- dm_get_all_fks_impl(dm, ignore_on_delete = TRUE)

  start_tbl <- tbl_impl(dm, start)
  current_cols <- colnames(start_tbl)

  # Track renames per parent for allow_deep FK transfer
  col_renames <- list()
  all_renames <- list()

  for (pt in parents) {
    # Get FK between start and parent
    fk_info <- all_fks %>%
      filter(child_table == start, parent_table == pt)
    if (nrow(fk_info) == 0) {
      abort_tables_not_neighbors(start, pt)
    }
    if (nrow(fk_info) > 1) {
      abort_no_cycles(create_graph_from_dm(dm))
    }

    child_fk_cols <- fk_info$child_fk_cols[[1]]
    parent_key_cols <- fk_info$parent_key_cols[[1]]
    by <- set_names(parent_key_cols, child_fk_cols)

    parent_tbl <- tbl_impl(dm, pt)
    parent_cols <- colnames(parent_tbl)

    # Columns from parent that will appear in the result
    # (all parent columns EXCEPT the key columns used in the join)
    parent_incoming_cols <- setdiff(parent_cols, parent_key_cols)

    # Find conflicts with current columns
    conflicting <- intersect(current_cols, parent_incoming_cols)

    # Rename conflicting columns in parent table
    renames <- character(0)
    if (length(conflicting) > 0) {
      new_names <- paste0(conflicting, ".", pt)
      rename_vec <- set_names(conflicting, new_names)
      parent_tbl <- rename(parent_tbl, !!!rename_vec)
      renames <- set_names(new_names, conflicting)
      all_renames <- c(all_renames, list(list(table = pt, renames = renames)))
    }
    col_renames[[pt]] <- renames

    # Perform the join
    start_tbl <- join(start_tbl, parent_tbl, by = by)

    # Update current columns for next iteration
    current_cols <- colnames(start_tbl)
  }

  # Build result dm
  def <- dm_get_def(dm)
  start_idx <- which(def$table == start)
  def$data[[start_idx]] <- start_tbl

  dm_result <- dm_from_def(def)

  # Remove parent tables
  remaining <- setdiff(def$table, parents)
  remaining <- set_names(remaining)

  list(
    dm = dm_select_tbl_impl(dm_result, remaining),
    all_renames = all_renames,
    col_renames = col_renames
  )
}

#' Report column renames to the user
#'
#' Formats and prints a message describing renamed columns,
#' using the same style as `dm_flatten_to_tbl()`.
#'
#' @param all_renames List of rename entries, each with `table` and `renames`
#'   components.
#'   `renames` is a named character vector mapping old names to new names.
#'
#' @return Called for its side effect (message).
#'   Returns `invisible()`.
#'
#' @noRd
dm_flatten_explain_renames <- function(all_renames) {
  if (is_empty(all_renames)) {
    return(invisible())
  }

  lines <- map_chr(all_renames, function(entry) {
    rename_strs <- paste0(
      tick_if_needed(entry$renames),
      " = ",
      tick_if_needed(names(entry$renames))
    )
    paste0(
      "dm_rename(",
      tick_if_needed(entry$table),
      ", ",
      paste0(rename_strs, collapse = ", "),
      ")"
    )
  })

  message(
    "Renaming ambiguous columns: %>%\n  ",
    paste0(lines, collapse = " %>%\n  ")
  )
}

#' Transfer foreign keys from absorbed parents to the flattened table
#'
#' When `allow_deep = TRUE`, parent tables that reference grandparent tables
#' are absorbed into the start table.
#' This function re-points those FK
#' relationships so that the start table now references the grandparent
#' directly, accounting for column renames and dropped join keys.
#'
#' @param dm A `dm` object (after parents have been joined and removed).
#' @param start Name of the flattened table.
#' @param parents Character vector of parent table names that were absorbed.
#' @param col_renames Named list mapping each parent to its column renames
#'   (from `dm_flatten_join()`).
#' @param all_fks Data frame of all FK relationships from the original dm
#'   (before absorption).
#'
#' @return An updated `dm` with transferred FK relationships.
#'
#' @noRd
#' @autoglobal
dm_flatten_transfer_fks <- function(dm, start, parents, col_renames, all_fks) {
  def <- dm_get_def(dm)

  for (pt in parents) {
    # Find FKs where pt is the child (pt references other tables as parent)
    pt_child_fks <- all_fks %>%
      filter(child_table == pt)

    for (i in seq_len(nrow(pt_child_fks))) {
      gp <- pt_child_fks$parent_table[i]
      # Skip if grandparent is also being absorbed or is the start table
      if (gp %in% parents || gp == start) {
        next
      }

      child_cols <- pt_child_fks$child_fk_cols[[i]]
      parent_cols <- pt_child_fks$parent_key_cols[[i]]

      # Apply renames: if any FK columns were renamed during disambiguation
      renames <- col_renames[[pt]]
      if (length(renames) > 0) {
        for (k in seq_along(child_cols)) {
          if (child_cols[k] %in% names(renames)) {
            child_cols[k] <- renames[[child_cols[k]]]
          }
        }
      }

      # Handle the case where a FK column is the same as the join key
      # (it was dropped during the join, so use start's FK column instead)
      fk_to_parent <- all_fks %>%
        filter(child_table == start, parent_table == pt)
      if (nrow(fk_to_parent) > 0) {
        parent_key <- fk_to_parent$parent_key_cols[[1]]
        start_fk <- fk_to_parent$child_fk_cols[[1]]
        for (k in seq_along(child_cols)) {
          key_idx <- match(child_cols[k], parent_key)
          if (!is.na(key_idx)) {
            child_cols[k] <- start_fk[key_idx]
          }
        }
      }

      # Add FK from start to grandparent
      gp_idx <- which(def$table == gp)
      def$fks[[gp_idx]] <- vec_rbind(
        def$fks[[gp_idx]],
        new_fk(list(parent_cols), start, list(child_cols), "no_action")
      )
    }
  }

  dm_from_def(def)
}

dm_for_flatten_deep <- function() {
  dm(
    A = tibble::tibble(id = 1:3, b_id = 1:3, val_a = letters[1:3]),
    B = tibble::tibble(id = 1:3, c_id = 1:3, val_b = LETTERS[1:3]),
    C = tibble::tibble(id = 1:3, val_c = c("x", "y", "z"))
  ) %>%
    dm_add_pk(B, id) %>%
    dm_add_pk(C, id) %>%
    dm_add_fk(A, b_id, B) %>%
    dm_add_fk(B, c_id, C)
}

test_that("`dm_flatten()` works for basic star schema", {
  # Basic flatten: join all parents into fact
  result <- expect_message_obj(dm_flatten(dm_for_flatten(), fact))

  # Result should be a dm
  expect_s3_class(result, "dm")

  # fact should still exist
  expect_true("fact" %in% names(dm_get_tables(result)))

  # Parents should be removed
  expect_false("dim_1" %in% names(dm_get_tables(result)))
  expect_false("dim_2" %in% names(dm_get_tables(result)))
  expect_false("dim_3" %in% names(dm_get_tables(result)))
  expect_false("dim_4" %in% names(dm_get_tables(result)))

  # fact should have columns from parents (renamed for disambiguation)
  fact_result <- pull_tbl(result, fact)

  # The "something" column exists in all tables, so parents' should be renamed
  expect_true("something" %in% colnames(fact_result))
  expect_true("something.dim_1" %in% colnames(fact_result))
  expect_true("something.dim_2" %in% colnames(fact_result))
  expect_true("something.dim_3" %in% colnames(fact_result))
  expect_true("something.dim_4" %in% colnames(fact_result))

  # Result should have correct number of rows
  expect_equal(nrow(fact_result), nrow(fact()))
})

test_that("`dm_flatten()` snapshot with `dm_paste()`", {
  expect_snapshot({
    dm_flatten(dm_for_flatten(), fact) %>%
      dm_paste(options = c("select", "keys"))
  })

  expect_snapshot({
    dm_flatten(dm_for_flatten(), fact, parent_tables = c(dim_1, dim_2)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("`dm_flatten()` recursive snapshot with `dm_paste()`", {
  expect_snapshot({
    dm_flatten(dm_more_complex(), tf_5, parent_tables = c(tf_4, tf_3), recursive = TRUE) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("`dm_flatten()` allow_deep snapshot with `dm_paste()`", {
  dm_deep <- dm_for_flatten_deep()

  expect_snapshot({
    dm_flatten(dm_deep, A, allow_deep = TRUE) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("`dm_flatten()` with no parents returns unchanged dm", {
  # A one-table dm
  one_tbl_dm <- dm_for_flatten() %>% dm_select_tbl(fact)
  result <- dm_flatten(one_tbl_dm, fact)
  expect_identical(names(dm_get_tables(result)), "fact")
  expect_equivalent_tbl(pull_tbl(result, fact), fact())
})

test_that("`dm_flatten()` with explicit parent selection", {
  # Flatten only dim_1 and dim_2
  result <- expect_message_obj(dm_flatten(dm_for_flatten(), fact, parent_tables = c(dim_1, dim_2)))

  expect_true("fact" %in% names(dm_get_tables(result)))
  expect_false("dim_1" %in% names(dm_get_tables(result)))
  expect_false("dim_2" %in% names(dm_get_tables(result)))
  # dim_3 and dim_4 should remain
  expect_true("dim_3" %in% names(dm_get_tables(result)))
  expect_true("dim_4" %in% names(dm_get_tables(result)))

  # fact should have columns from dim_1 and dim_2
  fact_result <- pull_tbl(result, fact)
  expect_true("something.dim_1" %in% colnames(fact_result))
  expect_true("something.dim_2" %in% colnames(fact_result))

  # FKs to dim_3 and dim_4 should still exist
  fks <- dm_get_all_fks(result)
  expect_true(any(fks$parent_table == "dim_3"))
  expect_true(any(fks$parent_table == "dim_4"))
})

test_that("`dm_flatten()` errors with deeper hierarchy by default", {
  expect_dm_error(
    dm_flatten(dm_more_complex(), tf_5, parent_tables = c(tf_4, tf_3)),
    class = "only_parents"
  )
})

test_that("`dm_flatten()` with recursive = TRUE works", {
  # dm_more_complex has tf_5 -> tf_4 -> tf_3
  result <- dm_flatten(dm_more_complex(), tf_5, parent_tables = c(tf_4, tf_3), recursive = TRUE)

  expect_s3_class(result, "dm")
  expect_true("tf_5" %in% names(dm_get_tables(result)))
  expect_false("tf_4" %in% names(dm_get_tables(result)))
  expect_false("tf_3" %in% names(dm_get_tables(result)))

  # tf_5 should have columns from tf_4 and tf_3
  tf5_result <- pull_tbl(result, tf_5)
  expected <-
    tf_5() %>%
    left_join(tf_4(), by = c("l" = "h")) %>%
    left_join(tf_3(), by = c("j" = "f", "j1" = "f1"))
  expect_equivalent_tbl(tf5_result, expected)
})

test_that("`dm_flatten()` with recursive = TRUE auto-detect", {
  result <- dm_flatten(dm_more_complex(), tf_5, recursive = TRUE)

  expect_s3_class(result, "dm")
  expect_true("tf_5" %in% names(dm_get_tables(result)))

  # All reachable ancestor tables should have been absorbed
  tf5_cols <- colnames(pull_tbl(result, tf_5))
  # tf_5 absorbs columns from tf_4, tf_3, tf_4_2, tf_6
  expect_true(length(tf5_cols) >= 10)
})

test_that("`dm_flatten()` with allow_deep = TRUE keeps grandparents", {
  dm_deep <- dm_for_flatten_deep()

  result <- dm_flatten(dm_deep, A, allow_deep = TRUE)

  # A should exist, B should be removed, C should remain
  expect_true("A" %in% names(dm_get_tables(result)))
  expect_false("B" %in% names(dm_get_tables(result)))
  expect_true("C" %in% names(dm_get_tables(result)))

  # A should have B's columns
  a_result <- pull_tbl(result, A)
  expect_true("val_b" %in% colnames(a_result))

  # FK from A to C should exist
  fks <- dm_get_all_fks(result)
  expect_true(any(fks$child_table == "A" & fks$parent_table == "C"))
})

test_that("`dm_flatten()` errors when recursive and allow_deep are both TRUE", {
  expect_dm_error(
    dm_flatten(dm_for_flatten(), fact, recursive = TRUE, allow_deep = TRUE),
    class = "recursive_and_allow_deep"
  )
})

test_that("`dm_flatten()` preserves primary keys of start table", {
  dm_test <- dm(
    child = tibble::tibble(id = 1:3, parent_id = 1:3, val = letters[1:3]),
    parent = tibble::tibble(id = 1:3, info = LETTERS[1:3])
  ) %>%
    dm_add_pk(child, id) %>%
    dm_add_pk(parent, id) %>%
    dm_add_fk(child, parent_id, parent)

  result <- dm_flatten(dm_test, child)

  # child's PK should be preserved
  pk <- dm_get_all_pks(result)
  expect_true(any(pk$table == "child"))
})

test_that("`dm_flatten()` with unreachable table errors", {
  expect_dm_error(
    dm_flatten(dm_for_filter(), tf_2, parent_tables = c(tf_3, tf_4)),
    class = "tables_not_reachable_from_start"
  )
})

test_that("`dm_flatten()` errors with cycles", {
  expect_dm_error(
    dm_flatten(
      dm_nycflights_small() %>%
        dm_add_fk(flights, origin, airports),
      flights
    ),
    "no_cycles"
  )
})

test_that("`dm_flatten()` with tidyselect via parent_tables", {
  # Deselection
  result1 <- expect_message_obj(
    dm_flatten(dm_for_flatten(), fact, parent_tables = c(-dim_2, -dim_4))
  )
  result2 <- expect_message_obj(
    dm_flatten(dm_for_flatten(), fact, parent_tables = c(dim_1, dim_3))
  )

  expect_equal(names(dm_get_tables(result1)), names(dm_get_tables(result2)))
  expect_equivalent_tbl(
    pull_tbl(result1, fact),
    pull_tbl(result2, fact)
  )

  # Select helpers
  expect_snapshot({
    dm_flatten(dm_for_flatten(), fact, parent_tables = c(ends_with("3"), ends_with("1"))) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("`dm_flatten()` handles column disambiguation correctly", {
  expect_snapshot({
    dm_flatten(dm_for_flatten(), fact, parent_tables = c(dim_1, dim_2)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("`dm_flatten()` reports renames", {
  expect_snapshot({
    dm_flatten(dm_for_flatten(), fact, parent_tables = c(dim_1, dim_2)) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("`dm_flatten()` errors on zoomed dm", {
  expect_dm_error(
    dm_for_flatten() %>% dm_zoom_to(fact) %>% dm_flatten(fact),
    class = "only_possible_wo_zoom"
  )
})

# --- Helper function tests ---

test_that("`dm_flatten_impl()` joins single parent without recursion", {
  dm_test <- dm(
    child = tibble::tibble(id = 1:3, pid = 1:3, val = letters[1:3]),
    parent = tibble::tibble(id = 1:3, info = LETTERS[1:3])
  ) %>%
    dm_add_pk(parent, id) %>%
    dm_add_fk(child, pid, parent)

  out <- dm_flatten_impl(dm_test, "child", "parent", dfs_order = NULL, left_join)

  expect_s3_class(out$dm, "dm")
  expect_equal(names(dm_get_tables(out$dm)), "child")
  expect_true("info" %in% colnames(pull_tbl(out$dm, child)))
  expect_equal(out$all_renames, list())
})

test_that("`dm_flatten_impl()` recurses with dfs_order", {
  dm_deep <- dm_for_flatten_deep()

  g <- create_graph_from_dm(dm_deep, directed = TRUE)
  g_sub <- graph_induced_subgraph(g, c("A", "B", "C"))
  dfs <- graph_dfs(g_sub, "A", unreachable = FALSE, dist = TRUE)
  dfs_order <- names(dfs$order) %>% purrr::discard(is.na)

  out <- dm_flatten_impl(dm_deep, "A", c("B", "C"), dfs_order, left_join)

  expect_equal(names(dm_get_tables(out$dm)), "A")
  a_tbl <- pull_tbl(out$dm, A)
  expect_true("val_b" %in% colnames(a_tbl))
  expect_true("val_c" %in% colnames(a_tbl))
})

test_that("`dm_flatten_impl()` returns empty renames when no conflicts", {
  dm_test <- dm(
    child = tibble::tibble(id = 1:2, pid = 1:2),
    parent = tibble::tibble(id = 1:2, extra = c("a", "b"))
  ) %>%
    dm_add_pk(parent, id) %>%
    dm_add_fk(child, pid, parent)

  out <- dm_flatten_impl(dm_test, "child", "parent", dfs_order = NULL, left_join)
  expect_equal(out$all_renames, list())
  expect_equal(out$col_renames, list(parent = character(0)))
})

test_that("`dm_flatten_join()` disambiguates conflicting columns", {
  dm_test <- dm(
    child = tibble::tibble(id = 1:2, pid = 1:2, val = c("a", "b")),
    parent = tibble::tibble(id = 1:2, val = c("X", "Y"))
  ) %>%
    dm_add_pk(parent, id) %>%
    dm_add_fk(child, pid, parent)

  out <- dm_flatten_join(dm_test, "child", "parent", left_join)

  child_tbl <- pull_tbl(out$dm, child)
  expect_true("val" %in% colnames(child_tbl))
  expect_true("val.parent" %in% colnames(child_tbl))

  expect_length(out$all_renames, 1)
  expect_equal(out$all_renames[[1]]$table, "parent")
  expect_equal(out$all_renames[[1]]$renames, c(val = "val.parent"))
})

test_that("`dm_flatten_join()` returns col_renames for each parent", {
  dm_test <- dm(
    child = tibble::tibble(id = 1:2, pid = 1:2, val = c("a", "b")),
    parent = tibble::tibble(id = 1:2, val = c("X", "Y"))
  ) %>%
    dm_add_pk(parent, id) %>%
    dm_add_fk(child, pid, parent)

  out <- dm_flatten_join(dm_test, "child", "parent", left_join)
  expect_true("parent" %in% names(out$col_renames))
  expect_equal(out$col_renames[["parent"]], c(val = "val.parent"))
})

test_that("`dm_flatten_join()` removes parent tables", {
  dm_test <- dm(
    child = tibble::tibble(id = 1:2, pid = 1:2),
    parent = tibble::tibble(id = 1:2, extra = c("a", "b")),
    other = tibble::tibble(id = 1:2)
  ) %>%
    dm_add_pk(parent, id) %>%
    dm_add_fk(child, pid, parent)

  out <- dm_flatten_join(dm_test, "child", "parent", left_join)
  expect_equal(sort(names(dm_get_tables(out$dm))), c("child", "other"))
})

test_that("`dm_flatten_join()` with empty parents returns unchanged dm", {
  dm_test <- dm(child = tibble::tibble(id = 1:2))
  out <- dm_flatten_join(dm_test, "child", character(0), left_join)
  expect_equal(names(dm_get_tables(out$dm)), "child")
  expect_equal(out$all_renames, list())
  expect_equal(out$col_renames, list())
})

test_that("`dm_flatten_explain_renames()` produces message for renames", {
  renames <- list(
    list(table = "dim_1", renames = c(val = "val.dim_1"))
  )
  expect_message(dm_flatten_explain_renames(renames), "Renaming")
})

test_that("`dm_flatten_explain_renames()` is silent when no renames", {
  expect_silent(dm_flatten_explain_renames(list()))
})

test_that("`dm_flatten_transfer_fks()` re-points FKs to start table", {
  dm_deep <- dm_for_flatten_deep()
  all_fks <- dm_get_all_fks_impl(dm_deep, ignore_on_delete = TRUE)

  # Simulate: join B into A, remove B
  out <- dm_flatten_join(dm_deep, "A", "B", left_join)

  # Transfer B's FK to C so it now points from A to C
  result <- dm_flatten_transfer_fks(out$dm, "A", "B", out$col_renames, all_fks)

  fks <- dm_get_all_fks(result)
  expect_true(any(fks$child_table == "A" & fks$parent_table == "C"))
})

test_that("`dm_flatten_transfer_fks()` handles renamed FK columns", {
  # Create dm where parent's FK column conflicts with child's column
  dm_test <- dm(
    child = tibble::tibble(id = 1:2, pid = 1:2, ref = c("x", "y")),
    parent = tibble::tibble(id = 1:2, ref = 1:2),
    grandparent = tibble::tibble(id = 1:2, info = c("a", "b"))
  ) %>%
    dm_add_pk(parent, id) %>%
    dm_add_pk(grandparent, id) %>%
    dm_add_fk(child, pid, parent) %>%
    dm_add_fk(parent, ref, grandparent)

  all_fks <- dm_get_all_fks_impl(dm_test, ignore_on_delete = TRUE)
  out <- dm_flatten_join(dm_test, "child", "parent", left_join)

  # ref was renamed to ref.parent
  expect_equal(out$col_renames[["parent"]], c(ref = "ref.parent"))

  result <- dm_flatten_transfer_fks(out$dm, "child", "parent", out$col_renames, all_fks)
  fks <- dm_get_all_fks(result)

  # FK should now be from child.ref.parent to grandparent.id
  fk_row <- fks[fks$child_table == "child" & fks$parent_table == "grandparent", ]
  expect_equal(nrow(fk_row), 1)
  expect_equal(fk_row$child_fk_cols[[1]], "ref.parent")
})

# --- Join argument tests ---

test_that("`dm_flatten()` with `join = inner_join`", {
  dm_test <- dm(
    child = tibble::tibble(id = 1:4, pid = c(1L, 2L, 3L, NA), val = letters[1:4]),
    parent = tibble::tibble(id = 1:3, info = LETTERS[1:3])
  ) %>%
    dm_add_pk(parent, id) %>%
    dm_add_fk(child, pid, parent)

  result <- dm_flatten(dm_test, child, join = inner_join)

  child_tbl <- pull_tbl(result, child)
  # inner_join drops unmatched rows (pid = NA)
  expect_equal(nrow(child_tbl), 3)
  expect_true("info" %in% colnames(child_tbl))
})

test_that("`dm_flatten()` with `join = full_join`", {
  dm_test <- dm(
    child = tibble::tibble(id = 1:2, pid = c(1L, 3L)),
    parent = tibble::tibble(id = 1:3, info = LETTERS[1:3])
  ) %>%
    dm_add_pk(parent, id) %>%
    dm_add_fk(child, pid, parent)

  result <- dm_flatten(dm_test, child, join = full_join)

  child_tbl <- pull_tbl(result, child)
  # full_join keeps all rows from both sides
  expect_true(nrow(child_tbl) >= 2)
  expect_true("info" %in% colnames(child_tbl))
})

test_that("`dm_flatten()` with `join = semi_join`", {
  dm_test <- dm(
    child = tibble::tibble(id = 1:4, pid = c(1L, 2L, 3L, NA), val = letters[1:4]),
    parent = tibble::tibble(id = 1:3, info = LETTERS[1:3])
  ) %>%
    dm_add_pk(parent, id) %>%
    dm_add_fk(child, pid, parent)

  result <- dm_flatten(dm_test, child, join = semi_join)

  child_tbl <- pull_tbl(result, child)
  # semi_join keeps only matching child rows, no parent columns added
  expect_equal(nrow(child_tbl), 3)
  expect_false("info" %in% colnames(child_tbl))
})

test_that("`dm_flatten()` with `join = anti_join`", {
  dm_test <- dm(
    child = tibble::tibble(id = 1:4, pid = c(1L, 2L, 3L, NA), val = letters[1:4]),
    parent = tibble::tibble(id = 1:3, info = LETTERS[1:3])
  ) %>%
    dm_add_pk(parent, id) %>%
    dm_add_fk(child, pid, parent)

  result <- dm_flatten(dm_test, child, join = anti_join)

  child_tbl <- pull_tbl(result, child)
  # anti_join keeps only non-matching child rows
  expect_equal(nrow(child_tbl), 1)
  expect_false("info" %in% colnames(child_tbl))
})

test_that("`dm_flatten()` errors with `join = nest_join`", {
  expect_dm_error(
    dm_flatten(dm_for_flatten(), fact, join = nest_join),
    class = "no_flatten_with_nest_join"
  )
})

test_that("`dm_flatten()` recursive with non-left join errors", {
  expect_dm_error(
    dm_flatten(dm_more_complex(), tf_5, recursive = TRUE, join = semi_join),
    class = "squash_limited"
  )
})

test_that("`dm_flatten()` recursive with `join = inner_join`", {
  result <- dm_flatten(
    dm_more_complex(),
    tf_5,
    parent_tables = c(tf_4, tf_3),
    recursive = TRUE,
    join = inner_join
  )

  expect_s3_class(result, "dm")
  expect_true("tf_5" %in% names(dm_get_tables(result)))
  expect_false("tf_4" %in% names(dm_get_tables(result)))
  expect_false("tf_3" %in% names(dm_get_tables(result)))
})

test_that("`dm_flatten()` join argument snapshot with `dm_paste()`", {
  expect_snapshot({
    dm_flatten(dm_for_flatten(), fact, parent_tables = c(dim_1, dim_2), join = inner_join) %>%
      dm_paste(options = c("select", "keys"))
  })
})

mssql_sys_db <- function(con, dbname, sys_table, vars = NULL) {
  if (is.na(dbname)) {
    id <- DBI::Id(schema = "sys", table = sys_table)
    sql_name <- sql("DB_NAME()")
  } else {
    id <- DBI::Id(catalog = dbname, schema = "sys", table = sys_table)
    sql_name <- dbname
  }
  tbl(con, id, vars = vars) %>%
    mutate(catalog = !!sql_name) %>%
    select(catalog, everything())
}

#' @autoglobal
mssql_constraint_column_usage <- function(con, table_constraints, dbname) {
  info_fkc <-
    table_constraints %>%
    select(constraint_catalog, constraint_schema, constraint_name, constraint_type) %>%
    filter(constraint_type == "FOREIGN KEY")

  fkc <- mssql_sys_db(
    con,
    dbname,
    "foreign_key_columns",
    vars = c(
      "constraint_object_id",
      "constraint_column_id",
      "referenced_object_id",
      "referenced_column_id"
    )
  )

  columns <-
    mssql_sys_db(
      con,
      dbname,
      "columns",
      vars = c(
        "name",
        "object_id",
        "column_id"
      )
    ) %>%
    rename(column_name = name)

  tables <-
    mssql_sys_db(
      con,
      dbname,
      "tables",
      vars = c(
        "schema_id",
        "name",
        "object_id"
      )
    ) %>%
    rename(table_name = name)

  schemas <-
    mssql_sys_db(
      con,
      dbname,
      "schemas",
      vars = c(
        "schema_id",
        "name"
      )
    ) %>%
    rename(table_schema = name)

  objects <-
    mssql_sys_db(
      con,
      dbname,
      "objects",
      vars = c(
        "name",
        "object_id"
      )
    ) %>%
    select(constraint_name = name, object_id)

  sys_fkc_column_usage <-
    fkc %>%
    left_join(
      columns,
      by = c("catalog", "referenced_object_id" = "object_id", "referenced_column_id" = "column_id")
    ) %>%
    left_join(tables, by = c("catalog", "referenced_object_id" = "object_id")) %>%
    collapse() %>%
    left_join(schemas, by = c("catalog", "schema_id")) %>%
    collapse() %>%
    left_join(objects, by = c("constraint_object_id" = "object_id")) %>%
    # table_schema is used twice
    transmute(
      constraint_catalog = catalog,
      constraint_schema = table_schema,
      constraint_name,
      table_schema,
      table_name,
      column_name,
      ordinal_position = constraint_column_id
    )

  tbl_lc(
    con,
    "information_schema.constraint_column_usage",
    vars = c(
      "table_catalog",
      "table_schema",
      "table_name",
      "column_name",
      "constraint_catalog",
      "constraint_schema",
      "constraint_name"
    )
  ) %>%
    semi_join(info_fkc, by = c("constraint_catalog", "constraint_schema", "constraint_name")) %>%
    select(-table_schema, -table_name, -column_name) %>%
    distinct() %>%
    left_join(
      sys_fkc_column_usage,
      by = c("constraint_catalog", "constraint_schema", "constraint_name")
    )
}

mssql_escape <- function(x, con) {
  # https://github.com/tidyverse/dbplyr/issues/934
  if (is.logical(x)) {
    dbplyr::sql(if_else(x, "1", "0", "NULL"))
  } else {
    dbplyr::escape(x, parens = FALSE, collapse = NULL, con = con)
  }
}

#' Number of rows
#'
#' @description
#' Returns a named vector with the number of rows for each table.
#'
#' @param dm A [`dm`] object.
#'
#' @return A named vector with the number of rows for each table.
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_filter(airports = (faa %in% c("EWR", "LGA"))) %>%
#'   dm_nrow()
#' @export
dm_nrow <- function(dm) {
  check_not_zoomed(dm)
  # FIXME: with "direct" filter maybe no check necessary: but do we want to issue
  # a message in case the filters haven't been applied yet?
  check_no_filter(dm)
  map_dbl(dm_get_tables_impl(dm), ~ as.numeric(pull(collect(safe_count(.)))))
}

get_by <- function(dm, lhs_name, rhs_name) {
  if (dm_has_fk_impl(dm, lhs_name, rhs_name)) {
    cols <- dm_get_fk2_impl(dm, lhs_name, rhs_name)
  } else if (dm_has_fk_impl(dm, rhs_name, lhs_name)) {
    cols <- dm_get_fk2_impl(dm, rhs_name, lhs_name)[2:1]
  } else {
    abort_tables_not_neighbors(lhs_name, rhs_name)
  }

  if (nrow(cols) > 1) {
    abort_no_cycles(create_graph_from_dm(dm))
  }

  # Construct a `by` argument of the form `c("lhs_col[1]" = "rhs_col[1]", ...)`
  # as required by `*_join()`
  by <- get_key_cols(cols[[2]])
  names(by) <- get_key_cols(cols[[1]])
  by
}

# Normalize a `by` argument: convert `dplyr_join_by` objects to named character
# vectors, and ensure unnamed elements get names equal to their values.
normalize_join_by <- function(by) {
  if (inherits(by, "dplyr_join_by")) {
    return(set_names(by$y, by$x))
  }
  if (!is_named2(by)) {
    by <- set_names(by, by)
  }
  bad <- which(names2(by) == "")
  names(by)[bad] <- by[bad]
  by
}

update_filter <- function(dm, table_name, filters) {
  def <- dm_get_def(dm)
  def$filters[def$table == table_name] <- filters
  dm_from_def(def, zoomed = TRUE)
}

#' Create a new state for the code-generation middleware
#'
#' @param dm A dm object, e.g. `dm_nycflights13()`.
#' @noRd
new_mw_cg <- function(dm) {
  structure(
    list(
      dm = dm,
      cg = new_cg_block(quo(dm))
    ),
    class = "dm_mw_cg"
  )
}

#' Run a command for the code-generation middleware
#'
#' @param mw_cg A state object as created by [new_mw_cg()] or a previous
#'   `mw_cg_run()` call.
#' @param op_name An operation name as string, see `mw_cg_make()` for the
#'   supported ones.
#' @param ... Named arguments as required by the individual `mw_cg_make_...()`
#'   functions.
#' @param confirmation_function A function or function-like with one argument
#'   that contains a confirmation message to show.
#'   This function is called whenever the middleware requests confirmation
#'   from the user.
#'   This function must return a scalar logical that indicates whether to
#'   proceed.
#'   Processed with [rlang::as_function()].
#' @param abort_function A function or function-like with one argument
#'   that contains an error message to show.
#'   This function is called whenever the middleware encounters a failure and
#'   must abort the operation.
#'   Processed with [rlang::as_function()].
#' @noRd
mw_cg_run <- function(
  mw_cg,
  op_name,
  ...,
  confirmation_function = abort_function,
  abort_function = abort
) {
  confirmation_function <- as_function(confirmation_function)
  abort_function <- as_function(abort_function)

  recipe <- tryCatch(
    mw_cg_make(mw_cg$dm, op_name = op_name, ...),
    error = function(e) {
      abort_function(paste0("Error from `mw_cg_make()`: ", conditionMessage(e)))
      NULL
    }
  )
  if (is.null(recipe)) {
    return(mw_cg)
  }

  if (!is.null(recipe$confirmation_message)) {
    out <- confirmation_function(recipe$confirmation_message)
    if (!is_scalar_logical(out)) {
      abort_function(paste0(
        "Internal error in `mw_cg_run()`: `confirmation_function()` doesn't return a scalar logical."
      ))
    }
    if (!isTRUE(out)) {
      return(mw_cg)
    }
  }

  mw_cg$cg <- purrr::reduce(recipe, ~ cg_add_call(..1, ..2), .init = mw_cg$cg)
  tryCatch(
    mw_cg$dm <- cg_eval_block(mw_cg$cg),
    error = function(e) {
      abort_function(paste0("Unexpected error in `mw_cg_run()`: ", conditionMessage(e)))
    }
  )
  mw_cg
}

mw_cg_make <- function(dm, op_name, ...) {
  # FIXME: Add more operations, call e.g. use_cg_make("dm_select_tbl")
  # to create templates
  switch(
    op_name,
    "dm_select_tbl" = mw_cg_make_dm_select_tbl(dm, ...),
    "dm_add_pk" = mw_cg_make_dm_add_pk(dm, ...),
    "dm_add_fk" = mw_cg_make_dm_add_fk(dm, ...),
    "dm_rename" = mw_cg_make_dm_rename(dm, ...),
    "dm_rm_fk" = mw_cg_make_dm_rm_fk(dm, ...),
    "dm_set_colors" = mw_cg_make_dm_set_colors(dm, ...),
    "dm_disentangle" = mw_cg_make_dm_disentangle(dm, ...),
    "dm_select" = mw_cg_make_dm_select(dm, ...),
    cli::cli_abort("Unknown op in {.fn mw_cg_make}: {.val {op}}.")
  )
}

check_is_character_vec <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (!is_character(x)) {
    cli::cli_abort("{.arg {arg}} must be a character vector.", call = call)
  }
}

check_tbl_in_dm <- function(
  x,
  y,
  arg_x = caller_arg(x),
  arg_y = caller_arg(y),
  call = caller_env()
) {
  not_in_dm <- setdiff(y, names(x))
  if (length(not_in_dm) != 0) {
    cli::cli_abort("Table {.arg_y {not_in_dm}} is not in {.arg_x {names(x)}}.", call = call)
  }
}

check_at_least_one_col <- function(cols, call = current_call()) {
  if (length(cols) < 1) {
    cli::cli_abort("At least one column must be selected.", call = call)
  }
}

#' Unnest a JSON column
#'
#' A wrapper around [tidyr::unnest()] that extracts its data from a JSON column.
#' The inverse of [json_nest()].
#'
#' @param data A data frame, a data frame extension (e.g. a tibble), or  a lazy
#'   data frame (e.g. from dbplyr or dtplyr).
#' @inheritParams rlang::args_dots_used
#' @inheritParams tidyr::unnest
#'
#' @return An object of the same type as `data`
#' @export
#' @examples
#' tibble(a = 1, b = '[{ "c": 2 }, { "c": 3 }]') %>%
#'   json_unnest(b)
json_unnest <- function(data, cols, ..., names_sep = NULL, names_repair = "check_unique") {
  check_dots_used()

  UseMethod("json_unnest")
}

#' @export
json_unnest.data.frame <- function(
  data,
  cols,
  ...,
  keep_empty = FALSE,
  ptype = NULL,
  names_sep = NULL,
  names_repair = "check_unique"
) {
  if (missing(cols)) {
    cli::cli_abort("The {.arg cols} argument must be provided.")
  }
  data %>%
    mutate(across({{ cols }}, ~ map(., ~ jsonlite::fromJSON(.) %>% as_tibble()))) %>%
    unnest(
      {{ cols }},
      keep_empty = keep_empty,
      ptype = ptype,
      names_sep = names_sep,
      names_repair = names_repair
    )
}

# @export
# json_unnest.tbl_lazy <- function(data, cols, ..., keep_empty = FALSE, ptype = NULL,
#                                  names_sep = NULL, names_repair = "check_unique") {
# select string_agg(quotename(k) + case t
#                   when 0 then ' nchar(1)'       -- javascript null
#                   when 1 then ' nvarchar(max)'  -- javascript string
#                   when 2 then ' float'          -- javascript number
#                   when 3 then ' bit'            -- javascript boolean
#                   else ' nvarchar(max) as json' -- javascript array or object
#                   end, ', ') within group (order by k)
# from (
#   select j2.[key], max(j2.[type])
#   from test
#   cross apply openjson(case when json_col like '{%}' then '[' + json_col + ']' else json_col end) as j1
#   cross apply openjson(j1.value) as j2
#   group by j2.[key]
# ) as kt(k, t)
# }

#' Rename columns
#'
#' @description
#' Rename the columns of your [`dm`] using syntax that is similar to [dplyr::rename()].
#'
#' @inheritParams dm_zoom_to
#' @param ... One or more unquoted expressions separated by commas.
#'   You can treat
#'   variable names as if they were positions, and use expressions like `x:y`
#'   to select the ranges of variables.
#'
#'   Use named arguments, e.g. `new_name = old_name`, to rename the selected variables.
#'
#'   The arguments in ... are automatically quoted and evaluated in a context where
#'   column names represent column positions.
#'   They also support unquoting and splicing.
#'   See `vignette("programming", package = "dplyr")` for an introduction to those concepts.
#'
#'   See select helpers for more details, and the examples about [tidyselect
#'   helpers][tidyselect::select_helpers], such as `starts_with()`, `everything()`, etc.
#'
#' @details If key columns are renamed, then the meta-information of the `dm` is updated accordingly.
#'
#' @return An updated `dm` with the columns of `table` renamed.
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_rename(airports, code = faa, altitude = alt)
#' @export
dm_rename <- function(dm, table, ...) {
  dm_local_error_call()
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})

  dm %>%
    dm_zoom_to(!!table_name) %>%
    rename(...) %>%
    dm_update_zoomed()
}

#' Select columns
#'
#' @description
#' Select columns of your [`dm`] using syntax that is similar to [dplyr::select()].
#'
#' @inheritParams dm_rename
#' @details If key columns are renamed, then the meta-information of the `dm` is updated accordingly.
#' If key columns are removed, then all related relations are dropped as well.
#'
#' @return An updated `dm` with the columns of `table` reduced and/or renamed.
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_select(airports, code = faa, altitude = alt)
#' @export
dm_select <- function(dm, table, ...) {
  dm_local_error_call()
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})

  dm %>%
    dm_zoom_to(!!table_name) %>%
    select(...) %>%
    dm_update_zoomed()
}

test_that("dm_rename() works for replacing pk", {
  expect_snapshot({
    dm_for_filter() %>%
      dm_rename(tf_3, new_f = f) %>%
      dm_get_all_pks_impl()
  })
})

test_that("dm_rename() works for replacing fks", {
  expect_snapshot({
    dm_for_filter() %>%
      dm_rename(tf_2, new_d = d, new_e = e) %>%
      dm_get_all_fks_impl()
  })
})

test_that("dm_select() works for replacing pk", {
  expect_snapshot({
    dm_for_filter() %>%
      dm_select(tf_3, new_f = f) %>%
      dm_get_all_pks_impl()
  })
})

test_that("dm_select() keeps pks up to date", {
  expect_identical(
    dm_for_filter() %>%
      dm_select(tf_3, f) %>%
      dm_get_all_pks_impl(),
    dm_for_filter() %>%
      dm_get_all_pks_impl() %>%
      filter(table != "tf_3")
  )

  expect_identical(
    dm_for_filter() %>%
      dm_select(tf_3, new_f = f, f1) %>%
      dm_get_all_pks_impl(),
    dm_for_filter() %>%
      dm_get_all_pks_impl() %>%
      # https://github.com/r-lib/vctrs/issues/1371
      mutate(pk_col = new_keys(if_else(table == "tf_3", list(c("new_f", "f1")), unclass(pk_col))))
  )
})

test_that("dm_select() works for replacing fks, and removes missing ones", {
  expect_snapshot({
    dm_for_filter() %>%
      dm_select(tf_2, new_d = d) %>%
      dm_get_all_fks_impl()
  })
})

test_that("dm_select() removes fks if not in selection", {
  expect_equal(
    dm_for_filter() %>%
      dm_select(tf_2, c, e, e1) %>%
      dm_get_all_fks_impl(),
    dm_for_filter() %>%
      dm_get_all_fks_impl() %>%
      filter(child_table != "tf_2" | parent_table != "tf_1")
  )

  expect_equal(
    dm_for_filter() %>%
      dm_select(tf_2, c, e, ex = e1) %>%
      dm_get_all_fks_impl(),
    dm_for_filter() %>%
      dm_get_all_fks_impl() %>%
      filter(child_table != "tf_2" | parent_table != "tf_1") %>%
      mutate(child_fk_cols = new_keys(map(child_fk_cols, ~ gsub("e1", "ex", .x))))
  )

  expect_equal(
    dm_for_filter() %>%
      dm_select(tf_2, c, e) %>%
      dm_get_all_fks_impl(),
    dm_for_filter() %>%
      dm_get_all_fks_impl() %>%
      filter(child_table != "tf_2")
  )
})


# tests for compound keys -------------------------------------------------

test_that("output for compound keys", {
  expect_snapshot({
    dm_select(dm_for_flatten(), fact, dim_1_key_1, dim_1_key_2) %>%
      dm_paste(options = c("select", "keys"))
    dm_select(dm_for_flatten(), dim_1, dim_1_pk_1, dim_1_pk_2) %>%
      dm_paste(options = c("select", "keys"))
    dm_select(dm_for_flatten(), fact, -dim_1_key_1) %>% dm_paste(options = c("select", "keys"))
    dm_select(dm_for_flatten(), dim_1, -dim_1_pk_1) %>% dm_paste(options = c("select", "keys"))
  })
})

dm_bind_impl <- function(dms, repair, quiet, repair_arg = "", caller = caller_env()) {
  dm_local_error_call()
  if (length(dms) == 0) {
    return(new_dm_def())
  }

  walk(dms, check_dm)
  walk(dms, check_not_zoomed)
  if (!all_same_source(map(dms, dm_get_tables_impl) %>% list_c())) {
    abort_not_same_src(dm_bind = TRUE)
  }

  # repair table names
  table_names <- map(dms, src_tbls_impl) %>% list_c()
  new_table_names <- vec_as_names(
    table_names,
    repair = repair,
    quiet = quiet,
    repair_arg = repair_arg,
    call = caller
  )

  # need to individually rename tables for each `dm`
  ntables_dms <- map(dms, length)
  dms_indices <- map(ntables_dms, seq_len)
  renaming_recipe <- map2(
    dms_indices,
    lag(cumsum(ntables_dms), default = 0),
    ~ set_names(.x, new_table_names[.x + .y])
  )

  dms_renamed <- map2(dms, renaming_recipe, bind_rename_tbl)
  new_defs <- map(dms_renamed, dm_get_def)
  vec_rbind(!!!new_defs)
}

bind_rename_tbl <- function(dm, renamed) {
  if (length(dm) == 0) {
    return(dm)
  }

  def <- dm_get_def(dm)
  renamed_names <- set_names(def$table[renamed], names(renamed))

  def %>%
    bind_filter_recode_table_def(renamed) %>%
    filter_recode_table_fks(renamed_names) %>%
    dm_from_def()
}

bind_filter_recode_table_def <- function(def, renamed) {
  def$table[renamed] <- names(renamed)
  def
}

dm_meta <- function(con, catalog = NA, schema = NULL, simple = FALSE, error_call = caller_env()) {
  if (is_sqlite(con)) {
    return(dm_meta_sqlite(
      con,
      catalog = catalog,
      schema = schema,
      simple = simple,
      error_call = error_call
    ))
  }

  need_collect <- FALSE

  if (is_mssql(con)) {
    if (is.null(catalog)) {
      # FIXME: Classed error message?
      cli::cli_abort("SQL server only supports learning from one database.")
    }

    if (!is.na(catalog)) {
      message("Temporarily switching to database ", tick(catalog), ".")
      old_dbname <- DBI::dbGetQuery(con, "SELECT DB_NAME()")[[1]]
      sql <- paste0("USE ", DBI::dbQuoteIdentifier(con, catalog))
      old_sql <- paste0("USE ", DBI::dbQuoteIdentifier(con, old_dbname))
      DBI::dbExecute(con, sql, immediate = TRUE)
      withr::defer({
        DBI::dbExecute(con, old_sql, immediate = TRUE)
      })
      need_collect <- TRUE
    }
  }

  if (simple) {
    out <-
      dm_meta_simple_raw(con) %>%
      filter_dm_meta_simple(catalog, schema)
  } else {
    out <-
      dm_meta_raw(con, catalog) %>%
      select_dm_meta() %>%
      filter_dm_meta(catalog, schema)
  }

  if (need_collect) {
    out <-
      out %>%
      collect()
  }

  out
}

#' @autoglobal
dm_meta_raw <- function(con, catalog) {
  src <- src_from_src_or_con(con)

  local_options(digits.secs = 6)

  schemata <- tbl_lc(
    src,
    "information_schema.schemata",
    vars = vec_c(
      "catalog_name",
      "schema_name",
      "default_character_set_name",
      # Optional, not MySQL:
      # "schema_owner", "default_character_set_catalog", "default_character_set_schema",
    )
  )
  tables <- tbl_lc(
    src,
    "information_schema.tables",
    vars = vec_c(
      "table_catalog",
      "table_schema",
      "table_name",
      "table_type",
    )
  )
  columns <- tbl_lc(
    src,
    "information_schema.columns",
    vars = vec_c(
      "table_catalog",
      "table_schema",
      "table_name",
      "column_name",
      "ordinal_position",
      "column_default",
      "is_nullable",
      "data_type",
      "character_maximum_length",
      "character_octet_length",
      "numeric_precision",
      "numeric_scale",
      "datetime_precision",
      "character_set_name",
      "collation_name",
      if (is_mariadb(src)) "extra" else NULL,

      # Optional, not RMySQL:
      # "numeric_precision_radix",
      # "character_set_catalog", "character_set_schema",
      # "collation_catalog", "collation_schema", "domain_catalog",
      # "domain_schema", "domain_name"
    )
  )

  # add is_autoincrement column to columns table
  if (is_mssql(src)) {
    columns <- columns %>%
      mutate(
        is_autoincrement = sql(
          "CAST(COLUMNPROPERTY(object_id(TABLE_SCHEMA+'.'+TABLE_NAME), COLUMN_NAME, 'IsIdentity') AS BIT)"
        )
      )
  } else if (is_postgres(src) || is_redshift(src)) {
    columns <- columns %>%
      mutate(
        is_autoincrement = sql(
          "CASE WHEN column_default IS NULL THEN FALSE ELSE column_default SIMILAR TO '%nextval%' END"
        )
      )
  } else if (is_mariadb(src)) {
    columns <- columns %>%
      mutate(is_autoincrement = sql("extra REGEXP 'auto_increment'")) %>%
      select(-extra)
  } else {
    cli::cli_alert_warning("unable to fetch autoincrement metadata for src '{class(src)[1]}'")
    columns <- columns %>%
      mutate(is_autoincrement = NA)
  }

  if (is_mariadb(src)) {
    table_constraints <- tbl_lc(
      src,
      "information_schema.table_constraints",
      vars = vec_c(
        "constraint_catalog",
        "constraint_schema",
        "constraint_name",
        "table_name",
        "constraint_type"
      )
    ) %>%
      mutate(
        table_catalog = constraint_catalog,
        table_schema = constraint_schema,
        .before = table_name
      ) %>%
      mutate(
        constraint_name = if_else(
          constraint_type == "PRIMARY KEY",
          paste0("pk_", table_name),
          constraint_name
        )
      ) %>%
      left_join(
        tbl_lc(
          src,
          "information_schema.referential_constraints",
          vars = vec_c(
            "constraint_catalog",
            "constraint_schema",
            "constraint_name",
            # "unique_constraint_catalog", "unique_constraint_schema",
            # "unique_constraint_name", "match_option", "update_rule",
            # "table_name", "referenced_table_name"
            "delete_rule"
          )
        ),
        by = c("constraint_catalog", "constraint_schema", "constraint_name")
      )
  } else {
    table_constraints <- tbl_lc(
      src,
      "information_schema.table_constraints",
      vars = vec_c(
        "constraint_catalog",
        "constraint_schema",
        "constraint_name",
        "table_catalog",
        "table_schema",
        "table_name",
        "constraint_type",
        "is_deferrable",
        "initially_deferred",
      )
    ) %>%
      left_join(
        tbl_lc(
          src,
          "information_schema.referential_constraints",
          vars = vec_c(
            "constraint_catalog",
            "constraint_schema",
            "constraint_name",
            # "unique_constraint_catalog", "unique_constraint_schema",
            # "unique_constraint_name", "match_option", "update_rule",
            "delete_rule"
          )
        ),
        by = c("constraint_catalog", "constraint_schema", "constraint_name")
      )
  }

  key_column_usage <- tbl_lc(
    src,
    "information_schema.key_column_usage",
    vars = vec_c(
      "constraint_catalog",
      "constraint_schema",
      "constraint_name",
      "table_catalog",
      "table_schema",
      "table_name",
      "column_name",
      "ordinal_position",
    )
  )

  if (is_postgres(src)) {
    # Need hand-crafted query for now
    constraint_column_usage <-
      tbl(
        src,
        sql(postgres_column_constraints),
        vars = c(
          "table_catalog",
          "table_schema",
          "table_name",
          "column_name",
          "constraint_catalog",
          "constraint_schema",
          "constraint_name",
          "ordinal_position"
        )
      )
  } else if (is_redshift(src)) {
    constraint_column_usage <-
      tbl_lc(
        src,
        "information_schema.key_column_usage",
        vars = c(
          "table_catalog",
          "table_schema",
          "table_name",
          "column_name",
          "constraint_catalog",
          "constraint_schema",
          "constraint_name",
          "ordinal_position"
        )
      ) %>%
      filter(!is.na(table_name))
  } else if (is_mssql(src)) {
    constraint_column_usage <- mssql_constraint_column_usage(src, table_constraints, catalog)
  } else {
    # Alternate constraint names for uniqueness
    key_column_usage <-
      key_column_usage %>%
      left_join(
        tbl_lc(
          src,
          "information_schema.table_constraints",
          vars = vec_c(
            "constraint_catalog",
            "constraint_schema",
            "constraint_name",
            "table_name",
            "constraint_type"
          )
        ),
        by = vec_c(
          "constraint_catalog",
          "constraint_schema",
          "constraint_name",
          "table_name",
        )
      ) %>%
      mutate(
        constraint_name = if_else(
          constraint_type == "PRIMARY KEY",
          paste0("pk_", table_name),
          constraint_name
        )
      ) %>%
      select(-constraint_type)

    constraint_column_usage <-
      tbl_lc(
        src,
        "information_schema.key_column_usage",
        vars = c(
          "table_catalog",
          "referenced_table_schema",
          "referenced_table_name",
          "referenced_column_name",
          "constraint_catalog",
          "constraint_schema",
          "constraint_name",
          "ordinal_position"
        )
      ) %>%
      filter(!is.na(referenced_table_name)) %>%
      rename(
        table_schema = referenced_table_schema,
        table_name = referenced_table_name,
        column_name = referenced_column_name,
      )
  }

  dm(schemata, tables, columns, table_constraints, key_column_usage, constraint_column_usage) %>%
    dm_meta_add_keys()
}

#' @autoglobal
dm_meta_add_keys <- function(dm_meta) {
  dm_meta %>%
    dm_meta_simple_add_keys() %>%
    # dm_add_fk(table_constraints, table_schema, schemata) %>%
    dm_add_pk(table_constraints, c(constraint_catalog, constraint_schema, constraint_name)) %>%
    dm_add_fk(table_constraints, c(table_catalog, table_schema, table_name), tables) %>%
    # constraint_schema vs. table_schema?

    # not on mssql:
    # dm_add_fk(referential_constraints, c(constraint_schema, table_name), tables) %>%
    # dm_add_fk(referential_constraints, c(constraint_schema, referenced_table_name), tables) %>%

    dm_add_pk(
      key_column_usage,
      c(constraint_catalog, constraint_schema, constraint_name, ordinal_position)
    ) %>%
    dm_add_fk(
      key_column_usage,
      c(table_catalog, table_schema, table_name, column_name),
      columns
    ) %>%
    dm_add_fk(
      key_column_usage,
      c(constraint_catalog, constraint_schema, constraint_name),
      table_constraints
    ) %>%
    #
    # not on mariadb;
    dm_add_pk(
      constraint_column_usage,
      c(constraint_catalog, constraint_schema, constraint_name, ordinal_position)
    ) %>%
    dm_add_fk(
      constraint_column_usage,
      c(table_catalog, table_schema, table_name, column_name),
      columns
    ) %>%
    dm_add_fk(
      constraint_column_usage,
      c(constraint_catalog, constraint_schema, constraint_name),
      table_constraints
    ) %>%
    dm_add_fk(
      constraint_column_usage,
      c(constraint_catalog, constraint_schema, constraint_name, ordinal_position),
      key_column_usage
    ) %>%
    #
    dm_set_colors(green4 = ends_with("_constraints"), orange = ends_with("_usage"))
}

#' @autoglobal
dm_meta_sqlite <- function(
  con,
  catalog = NA,
  schema = NULL,
  simple = FALSE,
  error_call = caller_env()
) {
  if (!is.na(catalog)) {
    cli::cli_abort("{.arg catalog} must be {.code NA} for SQLite connections.", call = error_call)
  }

  catalog <- NA_character_
  if (is.null(schema)) {
    schema <- "main"
  }

  # FIXME: Use dbListObjects() when it works, https://github.com/r-dbi/RSQLite/issues/689
  schema_q <- DBI::dbQuoteIdentifier(con, schema)
  table_names_df <- DBI::dbGetQuery(
    con,
    paste0(
      "SELECT name FROM ",
      schema_q,
      ".sqlite_master",
      " WHERE type = 'table' AND name NOT LIKE 'sqlite_%'"
    )
  )
  table_names <- table_names_df$name

  schemata <- tibble(catalog_name = catalog, schema_name = schema)

  tables <- tibble(
    table_catalog = catalog,
    table_schema = schema,
    table_name = table_names,
    table_type = "BASE TABLE",
  )

  pragma_table_info <- map(set_names(table_names), function(t) {
    t_quoted <- DBI::dbQuoteIdentifier(con, t)
    DBI::dbGetQuery(con, paste0("PRAGMA ", schema_q, ".table_info(", t_quoted, ")"))
  })

  pragma_fk_list <- map(set_names(table_names), function(t) {
    t_quoted <- DBI::dbQuoteIdentifier(con, t)
    DBI::dbGetQuery(con, paste0("PRAGMA ", schema_q, ".foreign_key_list(", t_quoted, ")"))
  })

  columns <- imap_dfr(pragma_table_info, function(info, tbl) {
    if (nrow(info) == 0) {
      return(NULL)
    }
    tibble(
      table_catalog = catalog,
      table_schema = schema,
      table_name = tbl,
      column_name = info$name,
      ordinal_position = as.integer(info$cid + 1L),
      column_default = as.character(info$dflt_value),
      is_nullable = ifelse(info$notnull == 0L, "YES", "NO"),
      data_type = tolower(info$type),
      is_autoincrement = FALSE,
    )
  })

  if (simple) {
    return(
      dm(schemata, tables, columns) %>%
        dm_meta_simple_add_keys()
    )
  }

  # PK constraints: one row per table that has PK columns

  pk_constraints <- imap_dfr(pragma_table_info, function(info, tbl) {
    pk_cols <- info[info$pk > 0, , drop = FALSE]
    if (nrow(pk_cols) == 0) {
      return(NULL)
    }
    tibble(
      constraint_catalog = catalog,
      constraint_schema = schema,
      constraint_name = paste0("pk_", tbl),
      table_catalog = catalog,
      table_schema = schema,
      table_name = tbl,
      constraint_type = "PRIMARY KEY",
      delete_rule = NA_character_,
    )
  })

  # FK constraints: one row per FK relationship
  fk_constraints <- imap_dfr(pragma_fk_list, function(fks, tbl) {
    if (nrow(fks) == 0) {
      return(NULL)
    }
    fks %>%
      distinct(.data$id, .keep_all = TRUE) %>%
      transmute(
        constraint_catalog = catalog,
        constraint_schema = schema,
        constraint_name = paste0("fk_", tbl, "_", .data$id),
        table_catalog = catalog,
        table_schema = schema,
        table_name = tbl,
        constraint_type = "FOREIGN KEY",
        delete_rule = .data$on_delete,
      )
  })

  empty_constraints <- tibble(
    constraint_catalog = character(),
    constraint_schema = character(),
    constraint_name = character(),
    table_catalog = character(),
    table_schema = character(),
    table_name = character(),
    constraint_type = character(),
    delete_rule = character(),
  )
  table_constraints <- vec_rbind(empty_constraints, pk_constraints, fk_constraints)

  # key_column_usage: PK columns
  pk_key_usage <- imap_dfr(pragma_table_info, function(info, tbl) {
    pk_cols <- info[info$pk > 0, , drop = FALSE]
    if (nrow(pk_cols) == 0) {
      return(NULL)
    }
    tibble(
      constraint_catalog = catalog,
      constraint_schema = schema,
      constraint_name = paste0("pk_", tbl),
      table_catalog = catalog,
      table_schema = schema,
      table_name = tbl,
      column_name = pk_cols$name,
      ordinal_position = as.integer(pk_cols$pk)
    )
  })

  # key_column_usage: FK child columns
  fk_key_usage <- imap_dfr(pragma_fk_list, function(fks, tbl) {
    if (nrow(fks) == 0) {
      return(NULL)
    }
    tibble(
      constraint_catalog = catalog,
      constraint_schema = schema,
      constraint_name = paste0("fk_", tbl, "_", fks$id),
      table_catalog = catalog,
      table_schema = schema,
      table_name = tbl,
      column_name = fks$from,
      ordinal_position = as.integer(fks$seq + 1L),
    )
  })

  empty_key_usage <- tibble(
    constraint_catalog = character(),
    constraint_schema = character(),
    constraint_name = character(),
    table_catalog = character(),
    table_schema = character(),
    table_name = character(),
    column_name = character(),
    ordinal_position = integer(),
  )
  key_column_usage <- vec_rbind(empty_key_usage, pk_key_usage, fk_key_usage)

  # constraint_column_usage: FK parent (referenced) columns
  empty_ccu <- tibble(
    table_catalog = character(),
    table_schema = character(),
    table_name = character(),
    column_name = character(),
    constraint_catalog = character(),
    constraint_schema = character(),
    constraint_name = character(),
    ordinal_position = integer(),
  )
  constraint_column_usage <- imap_dfr(pragma_fk_list, function(fks, tbl) {
    if (nrow(fks) == 0) {
      return(NULL)
    }
    tibble(
      table_catalog = catalog,
      table_schema = schema,
      table_name = fks$table,
      column_name = fks$to,
      constraint_catalog = catalog,
      constraint_schema = schema,
      constraint_name = paste0("fk_", tbl, "_", fks$id),
      ordinal_position = as.integer(fks$seq + 1L),
    )
  })
  constraint_column_usage <- vec_rbind(empty_ccu, constraint_column_usage)

  dm(schemata, tables, columns, table_constraints, key_column_usage, constraint_column_usage) %>%
    dm_meta_add_keys()
}

dm_meta_simple_raw <- function(con) {
  src <- src_from_src_or_con(con)

  local_options(digits.secs = 6)

  schemata <- tbl_lc(
    src,
    "information_schema.schemata",
    vars = c(
      "catalog_name",
      "schema_name"
    )
  )
  tables <- tbl_lc(
    src,
    "information_schema.tables",
    vars = c(
      "table_catalog",
      "table_schema",
      "table_name",
      "table_type"
    )
  )
  columns <- tbl_lc(
    src,
    "information_schema.columns",
    vars = c(
      "table_catalog",
      "table_schema",
      "table_name",
      "column_name",
      "ordinal_position",
      "column_default",
      "is_nullable",
      "data_type"
    )
  )

  dm(schemata, tables, columns) %>%
    dm_meta_simple_add_keys()
}

#' @autoglobal
dm_meta_simple_add_keys <- function(dm_meta) {
  dm_meta %>%
    dm_add_pk(schemata, c(catalog_name, schema_name)) %>%
    dm_add_pk(tables, c(table_catalog, table_schema, table_name)) %>%
    dm_add_fk(tables, c(table_catalog, table_schema), schemata) %>%
    dm_add_pk(columns, c(table_catalog, table_schema, table_name, column_name)) %>%
    dm_add_fk(columns, c(table_catalog, table_schema, table_name), tables) %>%
    #
    dm_set_colors(brown = c(tables, columns), blue = schemata)
}

tbl_lc <- function(con, name, vars) {
  # For discovery only!
  if (is.null(vars)) {
    from <- name
  } else {
    quoted_vars <- DBI::dbQuoteIdentifier(con_from_src_or_con(con), vars)
    from <- sql(paste0(
      "SELECT ",
      # Be especially persuasive for MySQL
      paste0(quoted_vars, " AS ", quoted_vars, collapse = ", "),
      "\nFROM ",
      name
    ))
  }

  out <- tbl(con, from, vars = vars)
  if (is.null(vars)) {
    out <-
      out %>%
      rename(!!!set_names(colnames(out), tolower(colnames(out))))
  }
  out
}

#' @autoglobal
select_dm_meta <- function(dm_meta) {
  dm_meta %>%
    dm_select(schemata, catalog_name, schema_name) %>%
    dm_select(tables, table_catalog, table_schema, table_name, table_type) %>%
    dm_select(
      columns,
      table_catalog,
      table_schema,
      table_name,
      column_name,
      ordinal_position,
      column_default,
      is_nullable,
      is_autoincrement
    ) %>%
    dm_select(
      table_constraints,
      constraint_catalog,
      constraint_schema,
      constraint_name,
      table_catalog,
      table_schema,
      table_name,
      constraint_type,
      delete_rule
    ) %>%
    dm_select(
      key_column_usage,
      constraint_catalog,
      constraint_schema,
      constraint_name,
      table_catalog,
      table_schema,
      table_name,
      column_name,
      ordinal_position
    ) %>%
    dm_select(
      constraint_column_usage,
      table_catalog,
      table_schema,
      table_name,
      column_name,
      constraint_catalog,
      constraint_schema,
      constraint_name,
      ordinal_position
    )
}

#' @autoglobal
#' @global DATABASE
filter_dm_meta <- function(dm_meta, catalog = NULL, schema = NULL) {
  force(catalog)
  force(schema)

  if (length(schema) > 1 && anyNA(schema)) {
    cli::cli_abort("{.arg schema} must not contain NA if it has more than one element.")
  }

  schemata <- dm_meta$schemata
  tables <- dm_meta$tables
  columns <- dm_meta$columns
  table_constraints <- dm_meta$table_constraints
  key_column_usage <- dm_meta$key_column_usage
  constraint_column_usage <- dm_meta$constraint_column_usage

  if (!is.null(catalog) && !is.na(catalog)) {
    schemata <- schemata %>% filter(catalog_name %in% !!catalog)
    tables <- tables %>% filter(table_catalog %in% !!catalog)
    columns <- columns %>% filter(table_catalog %in% !!catalog)
    table_constraints <- table_constraints %>% filter(table_catalog %in% !!catalog)
    key_column_usage <- key_column_usage %>% filter(table_catalog %in% !!catalog)
    constraint_column_usage <- constraint_column_usage %>% filter(table_catalog %in% !!catalog)
  }

  if (!is.null(schema) && !anyNA(schema)) {
    schemata <- schemata %>% filter(schema_name %in% !!schema)
    tables <- tables %>% filter(table_schema %in% !!schema)
    columns <- columns %>% filter(table_schema %in% !!schema)
    table_constraints <- table_constraints %>% filter(table_schema %in% !!schema)
    key_column_usage <- key_column_usage %>% filter(table_schema %in% !!schema)
    constraint_column_usage <- constraint_column_usage %>% filter(table_schema %in% !!schema)
  } else if (!isTRUE(is.na(schema)) && is_mariadb(dm_get_con(dm_meta))) {
    schemata <- schemata %>% filter(schema_name == DATABASE() | is.na(DATABASE()))
    tables <- tables %>% filter(table_schema == DATABASE() | is.na(DATABASE()))
    columns <- columns %>% filter(table_schema == DATABASE() | is.na(DATABASE()))
    table_constraints <- table_constraints %>%
      filter(table_schema == DATABASE() | is.na(DATABASE()))
    key_column_usage <- key_column_usage %>% filter(table_schema == DATABASE() | is.na(DATABASE()))
    constraint_column_usage <- constraint_column_usage %>%
      filter(table_schema == DATABASE() | is.na(DATABASE()))
  }

  dm(
    schemata,
    tables,
    columns,
    table_constraints,
    key_column_usage,
    constraint_column_usage
  ) %>%
    dm_meta_add_keys()
}

#' @autoglobal
filter_dm_meta_simple <- function(dm_meta, catalog = NULL, schema = NULL) {
  force(catalog)
  force(schema)

  schemata <- dm_meta$schemata
  tables <- dm_meta$tables
  columns <- dm_meta$columns

  if (!is.null(catalog) && !is.na(catalog)) {
    schemata <- schemata %>% filter(catalog_name %in% !!catalog)
    tables <- tables %>% filter(table_catalog %in% !!catalog)
    columns <- columns %>% filter(table_catalog %in% !!catalog)
  }

  if (!is.null(schema)) {
    schemata <- schemata %>% filter(schema_name %in% !!schema)
    tables <- tables %>% filter(table_schema %in% !!schema)
    columns <- columns %>% filter(table_schema %in% !!schema)
  } else if (!is.null(schema) && !is.na(schema) && is_mariadb(dm_get_con(dm_meta))) {
    schemata <- schemata %>% filter(schema_name == DATABASE() | is.na(DATABASE()))
    tables <- tables %>% filter(table_schema == DATABASE() | is.na(DATABASE()))
    columns <- columns %>% filter(table_schema == DATABASE() | is.na(DATABASE()))
  }

  dm(schemata, tables, columns) %>%
    dm_meta_simple_add_keys()
}

#' \pkg{tidyr} table manipulation methods for zoomed dm objects
#'
#' @description
#' Use these methods without the '.dm_zoomed' suffix (see examples).
#' @param data object of class `dm_zoomed`
#' @param ... For `unite.dm_zoomed`: see [tidyr::unite()]
#'
#' For `separate.dm_zoomed`: see [tidyr::separate()]
#' @inheritParams tidyr::unite
#' @inheritParams tidyr::separate
#' @name tidyr_table_manipulation
#' @examplesIf rlang::is_installed("nycflights13")
#' zoom_united <- dm_nycflights13() %>%
#'   dm_zoom_to(flights) %>%
#'   select(year, month, day) %>%
#'   unite("month_day", month, day)
#' zoom_united
#' zoom_united %>%
#'   separate(month_day, c("month", "day"))
NULL
#' @export
unite.dm <- function(data, col, ..., sep = "_", remove = TRUE, na.rm = FALSE) {
  check_zoomed(data)
}

#' @rdname tidyr_table_manipulation
#' @export
unite.dm_zoomed <- function(data, col, ..., sep = "_", remove = TRUE, na.rm = FALSE) {
  tbl <- tbl_zoomed(data)
  united_tbl <- unite(tbl, col = !!col, ..., sep = sep, remove = remove, na.rm = na.rm)

  # all columns that are not not removed count as "selected"; names of "selected" are identical to "selected"
  if (remove) {
    deselected <- eval_select_both(quo(c(...)), colnames(tbl))
  } else {
    deselected <- eval_select_both(quo(c()), colnames(tbl))
  }
  selected <- set_names(setdiff(names(col_tracker_zoomed(data)), deselected$names))
  new_tracked_cols_zoom <- new_tracked_cols(data, selected)

  replace_zoomed_tbl(data, united_tbl, new_tracked_cols_zoom)
}

#' @rdname tidyr_table_manipulation
#' @export
unite.dm_keyed_tbl <- function(data, col, ..., sep = "_", remove = TRUE, na.rm = FALSE) {
  keys_info <- keyed_get_info(data)
  tbl <- unclass_keyed_tbl(data)
  out <- unite(tbl, col = {{ col }}, ..., sep = sep, remove = remove, na.rm = na.rm)
  new_keyed_tbl_from_keys_info(out, keys_info)
}

#' @export
separate.dm <- function(
  data,
  col,
  into,
  sep = "[^[:alnum:]]+",
  remove = TRUE,
  convert = FALSE,
  extra = "warn",
  fill = "warn",
  ...
) {
  check_zoomed(data)
}

#' @rdname tidyr_table_manipulation
#' @export
separate.dm_zoomed <- function(
  data,
  col,
  into,
  sep = "[^[:alnum:]]+",
  remove = TRUE,
  convert = FALSE,
  extra = "warn",
  fill = "warn",
  ...
) {
  tbl <- tbl_zoomed(data)
  col <- tidyselect::vars_pull(names(tbl), !!enquo(col))
  separated_tbl <- separate(
    tbl,
    col = !!col,
    into = into,
    sep = sep,
    remove = remove,
    convert = convert,
    extra = extra,
    fill = fill,
    ...
  )
  # all columns that are not removed count as "selected"; names of "selected" are identical to "selected"
  deselected <- if (remove) col else character()
  selected <- set_names(setdiff(names(col_tracker_zoomed(data)), deselected))
  new_tracked_cols_zoom <- new_tracked_cols(data, selected)
  replace_zoomed_tbl(data, separated_tbl, new_tracked_cols_zoom)
}

#' @rdname tidyr_table_manipulation
#' @export
separate.dm_keyed_tbl <- function(
  data,
  col,
  into,
  sep = "[^[:alnum:]]+",
  remove = TRUE,
  convert = FALSE,
  extra = "warn",
  fill = "warn",
  ...
) {
  keys_info <- keyed_get_info(data)
  tbl <- unclass_keyed_tbl(data)
  out <- separate(
    tbl,
    col = {{ col }},
    into = into,
    sep = sep,
    remove = remove,
    convert = convert,
    extra = extra,
    fill = fill,
    ...
  )
  new_keyed_tbl_from_keys_info(out, keys_info)
}

#' Prototype for a dm object
#'
#' @description
#' The prototype contains all tables, all primary and foreign keys,
#' but no data.
#' All tables are truncated and converted to zero-row tibbles,
#' also for remote data models.
#' Columns retain their type.
#' This is useful for performing creation and population of a database
#' in separate steps.
#'
#' @inheritParams dm_has_fk
#' @export
#' @examplesIf dm:::dm_has_financial()
#' dm_financial() %>%
#'   dm_ptype()
#'
#' dm_financial() %>%
#'   dm_ptype() %>%
#'   dm_nrow()
dm_ptype <- function(dm) {
  check_not_zoomed(dm)

  # collect() doesn't support n argument for data frames
  # collect() requires n > 0: https://github.com/tidyverse/dbplyr/issues/415
  dm %>%
    dm_get_def() %>%
    mutate(data = map(data, ~ head(.x, 0))) %>%
    dm_from_def() %>%
    collect()
}

# helper to check if key tracking works
get_all_keys <- function(dm, table_name = NULL) {
  fks <-
    dm_get_all_fks_impl(dm, table_name)
  pks <-
    dm_get_all_pks_impl(dm, table_name)

  list(
    pks = pks,
    fks = fks
  )
}

c_list_of <- function(x) {
  if (length(x) == 0) {
    # vctrs internals
    return(attr(x, "ptype"))
  }

  vec_c(!!!x, .name_spec = zap())
}

unnest_col <- function(x, col, ptype) {
  col_data <- x[[col]]
  out <- x[rep(seq_len(nrow(x)), lengths(col_data)), ]

  if (length(col_data) > 0) {
    col_data <- unlist(col_data, recursive = FALSE, use.names = FALSE)
    if (!identical(col_data[0], ptype)) {
      cli::cli_abort(
        "Internal: {.fun unnest_col} ptype mismatch, must be {.cls {class(ptype)[[1]]}}, not {.cls {class(col_data[0])[[1]]}}."
      )
    }
  } else {
    col_data <- ptype
  }

  out[[col]] <- col_data
  out
}

unnest_list_of_df <- function(x, col) {
  col_data <- x[[col]]
  stopifnot(is_list_of(col_data))

  out <- x[rep(seq_len(nrow(x)), map_int(col_data, vec_size)), setdiff(names(x), col)]
  out <- vec_cbind(out, c_list_of(col_data))
  out
}

unnest_df <- function(x, col, ptype) {
  col_data <- x[[col]]

  out <- x[rep(seq_len(nrow(x)), map_int(col_data, vec_size)), setdiff(names(x), col)]

  if (length(col_data) > 0) {
    col_data <- vec_rbind(!!!col_data)
    if (!identical(col_data[0, ], ptype)) {
      cli::cli_abort("Internal: {.fun unnest_df} ptype mismatch.")
    }
  } else {
    col_data <- ptype
  }

  out <- vec_cbind(out, col_data)
  out
}

#' Flatten a part of a `dm` into a wide table
#'
#' `dm_flatten_to_tbl()` gathers all information of interest in one place in a wide table.
#' It performs a disambiguation of column names and a cascade of joins.
#'
#' @inheritParams dm_join_to_tbl
#' @param .start The table from which all outgoing foreign key relations are considered
#'   when establishing a processing order for the joins.
#'   An interesting choice could be
#'   for example a fact table in a star schema.
#' @param ...
#'   `r lifecycle::badge("experimental")`
#'
#'   Unquoted names of the tables to be included in addition to the `.start` table.
#'   The order of the tables here determines the order of the joins.
#'   If the argument is empty, all tables that can be reached will be included.
#'   `tidyselect` is supported, see [dplyr::select()] for details on the semantics.
#' @param .recursive Logical, defaults to `FALSE`. Should not only parent tables be joined to `.start`, but also their ancestors?
#' @param .join The type of join to be performed, see [dplyr::join()].
#' @family flattening functions
#'
#' @details
#' With `...` left empty, this function will join together all the tables of your [`dm`]
#' object that can be reached from the `.start` table, in the direction of the foreign key relations
#' (pointing from the child tables to the parent tables), using the foreign key relations to
#' determine the argument `by` for the necessary joins.
#' The result is one table with unique column names.
#' Use the `...` argument if you would like to control which tables should be joined to the `.start` table.
#'
#' Mind that calling `dm_flatten_to_tbl()` with `.join = right_join` and no table order determined in the `...` argument
#' will not lead to a well-defined result if two or more foreign tables are to be joined to `.start`.
#' The resulting
#' table would depend on the order the tables that are listed in the `dm`.
#' Therefore, trying this will result in a warning.
#'
#' Since `.join = nest_join` does not make sense in this direction (LHS = child table, RHS = parent table: for valid key constraints
#' each nested column entry would be a tibble of one row), an error will be thrown if this method is chosen.
#'
#' The difference between `.recursive = FALSE` and `.recursive = TRUE` is
#' the following (see the examples):
#'
#' - `.recursive = FALSE` allows only one level of hierarchy
#'   (i.e., direct neighbors to table `.start`), while
#'
#' - `.recursive = TRUE` will go through all levels of hierarchy while joining.
#'
#' Additionally, these functions differ from `dm_wrap_tbl()`, which always
#' returns a `dm` object.
#'
#' @return A single table that results from consecutively joining all affected tables to the `.start` table.
#'
#' @examplesIf dm:::dm_has_financial()
#'
#' dm_financial() %>%
#'   dm_select_tbl(-loans) %>%
#'   dm_flatten_to_tbl(.start = cards)
#'
#' dm_financial() %>%
#'   dm_select_tbl(-loans) %>%
#'   dm_flatten_to_tbl(.start = cards, .recursive = TRUE)
#'
#' @export
dm_flatten_to_tbl <- function(dm, .start, ..., .recursive = FALSE, .join = left_join) {
  dm_local_error_call()
  check_not_zoomed(dm)
  join_name <- as_label(enexpr(.join))
  if (.recursive && !(join_name %in% c("left_join", "full_join", "inner_join"))) {
    abort_squash_limited()
  }

  start <- dm_tbl_name(dm, {{ .start }})

  vars <- setdiff(src_tbls_impl(dm), start)
  list_of_pts <- eval_select_table(quo(c(...)), vars)

  dm_flatten_to_tbl_impl(
    dm,
    start,
    list_of_pts,
    join = .join,
    join_name = join_name,
    squash = .recursive
  )
}

dm_flatten_to_tbl_impl <- function(
  dm,
  start,
  list_of_pts,
  join,
  join_name,
  squash,
  .position = "suffix"
) {
  if (join_name == "nest_join") {
    abort_no_flatten_with_nest_join()
  }

  force(join)
  stopifnot(is_function(join))

  # in case of `semi_join()` and `anti_join()` no renaming necessary
  gotta_rename <- !(join_name %in% c("semi_join", "anti_join"))

  # early returns for some of the possible joins would be possible for "perfect" key relations,
  # but since it is generally possible to have imperfect FK relations, `semi_join` and `anti_join` might
  # produce results, that are of interest, e.g.
  # dm_flatten_to_tbl(dm_nycflights13(cycle = TRUE) %>% dm_rm_fk(flights, origin, airports), flights, airports, join = anti_join)

  # need to work with directed graph here, since we only want to go in the direction
  # the foreign key is pointing to
  g <- create_graph_from_dm(dm, directed = TRUE)

  # If no tables are given, we use all reachable tables
  auto_detect <- is_empty(list_of_pts)
  if (auto_detect) {
    list_of_pts <- get_names_of_connected(g, start, squash)
  }
  # We use the induced subgraph right away
  g <- graph_induced_subgraph(g, c(start, list_of_pts))

  # each next table needs to be accessible from the former table (note: directed relations)
  # we achieve this with a depth-first-search (DFS) with param `unreachable = FALSE`
  dfs <- graph_dfs(g, start, unreachable = FALSE, parent = TRUE, dist = TRUE)

  # compute all table names
  order_df <-
    tibble(
      name = names(dfs[["order"]]),
      pred = names(graph_vertices(g))[unclass(dfs[["parent"]])[name]]
    )

  # function to detect any reason for abort()
  check_flatten_to_tbl(
    join_name,
    (nrow(dm_get_filters_impl(dm)) > 0) && !is_empty(list_of_pts),
    anyNA(order_df$name) || nrow(order_df) < graph_vcount(g),
    g,
    auto_detect,
    nrow(order_df) > 2,
    any(dfs$dist > 1),
    squash
  )

  # rename dm and replace table `.start` by its filtered, renamed version
  prep_dm <- prepare_dm_for_flatten(dm, order_df$name, gotta_rename, position = .position)

  # Drop the first table in the list of join partners. (We have at least one table, `.start`.)
  # (Working with `reduce2()` here and the `.init`-argument is the first table)
  # in the case of only one table in the `dm` (table "start"), all code below is a no-op
  order_df <- order_df[-1, ]
  # the order given in the ellipsis determines the join-list; if empty ellipsis, this is a no-op.
  # `unname()` to avoid warning (tibble version ‘2.99.99.9012’ retains names in column vectors)
  order_df <- left_join(tibble(name = unname(list_of_pts)), order_df, by = "name")

  # list of join partners
  ordered_table_list <- dm_get_tables(prep_dm)[order_df$name]
  by <- map2(order_df$pred, order_df$name, ~ get_by(prep_dm, .x, .y))

  # perform the joins according to the list, starting with table `initial_LHS`
  reduce2(ordered_table_list, by, ~ join(..1, ..2, by = ..3), .init = tbl_impl(prep_dm, start))
}

#' Join two tables
#'
#' `dm_join_to_tbl()` is deprecated in favor of [dm_flatten_to_tbl()].
#'
#' @param dm A [`dm`] object.
#' @param table_1 One of the tables involved in the join.
#' @param table_2 The second table of the join.
#' @param join The type of join to be performed, see [dplyr::join()].
#'
#' @rdname deprecated
#' @export
dm_join_to_tbl <- function(dm, table_1, table_2, join = left_join) {
  dm_local_error_call()
  deprecate_warn("0.3.0", "dm::dm_join_to_tbl()", "dm::dm_flatten_to_tbl()")

  check_not_zoomed(dm)
  force(join)
  stopifnot(is_function(join))
  join_name <- deparse(substitute(join))

  t1_name <- dm_tbl_name(dm, {{ table_1 }})
  t2_name <- dm_tbl_name(dm, {{ table_2 }})

  rel <- parent_child_table(dm, {{ table_1 }}, {{ table_2 }})
  start <- rel$child_table
  other <- rel$parent_table

  dm_flatten_to_tbl_impl(
    dm,
    start,
    other,
    join = join,
    join_name = join_name,
    squash = FALSE,
    .position = "prefix"
  )
}

parent_child_table <- function(dm, table_1, table_2) {
  t1_name <- dm_tbl_name(dm, {{ table_1 }})
  t2_name <- dm_tbl_name(dm, {{ table_2 }})

  rel <-
    dm_get_all_fks(dm) %>%
    filter(
      (child_table == t1_name & parent_table == t2_name) |
        (child_table == t2_name & parent_table == t1_name)
    )

  if (nrow(rel) == 0) {
    abort_tables_not_neighbors(t1_name, t2_name)
  }

  if (nrow(rel) > 1) {
    abort_no_cycles(create_graph_from_dm(dm))
  }

  rel
}

check_flatten_to_tbl <- function(
  join_name,
  part_cond_abort_filters,
  any_not_reachable,
  g,
  auto_detect,
  more_than_1_pt,
  has_grandparent,
  squash
) {
  # argument checking, or filter and recompute induced subgraph
  # for subsequent check
  if (any_not_reachable) {
    abort_tables_not_reachable_from_start()
  }

  # Cycles not yet supported
  if (length(graph_vertices(g)) - 1 != length(graph_edges(g))) {
    abort_no_cycles(g)
  }
  if (join_name == "nest_join") {
    abort_no_flatten_with_nest_join()
  }
  if (part_cond_abort_filters && join_name %in% c("full_join", "right_join")) {
    abort_apply_filters_first(join_name)
  }
  # the result for `right_join()` depends on the order of the dim-tables in the `dm`
  # if 2 or more of them are joined to the fact table and ellipsis is empty.

  # If called by `dm_join_to_tbl()` or `dm_flatten_to_tbl()` with `.recursive = FALSE`.
  # Then only one level of hierarchy is allowed (direct neighbors to table `.start`).
  if (!squash && has_grandparent) {
    abort_only_parents("dm_join_to_tbl", ".start", ".recursive")
  }

  if (join_name == "right_join" && auto_detect && more_than_1_pt) {
    warning(
      paste0(
        "Result for `dm_flatten_to_tbl()` with `right_join()` dependend on order of tables in `dm`, when ",
        "more than 2 tables involved and no explicit order given in `...`."
      )
    )
  }
}

prepare_dm_for_flatten <- function(dm, tables, gotta_rename, position = "suffix") {
  start <- tables[1]
  # filters need to be empty, for the disambiguation to work
  # renaming will be minimized if we reduce the `dm` to the necessary tables here
  red_dm <-
    dm_reset_all_filters(dm) %>%
    dm_select_tbl(!!!tables)
  # Only need to compute `dm[[start]]`, `dm_apply_filters()` not necessary
  # Need to use `dm` and not `clean_dm` here, because of possible filter conditions.
  start_tbl <- dm_get_filtered_table(dm, start)

  if (gotta_rename) {
    table_colnames <- get_table_colnames(red_dm)
    recipe <- compute_disambiguate_cols_recipe(table_colnames, sep = ".", position = position)
    explain_col_rename(recipe)
    # prepare `dm` by disambiguating columns (on a reduced dm)
    clean_dm <-
      col_rename(red_dm, recipe)
    # the column names of start_tbl need to be updated, since taken from `dm` and not `clean_dm`,
    # therefore we need a named variable containing the new and old names
    renames <-
      pluck(recipe$renames[recipe$table == start], 1)
    start_tbl <- start_tbl %>% rename(!!!renames)
  } else {
    # for `anti_join()` and `semi_join()` no renaming necessary
    clean_dm <- red_dm
    renames <- character(0)
  }

  def <- dm_get_def(clean_dm)
  def$data[[which(def$table == start)]] <- start_tbl
  dm_from_def(def)
}

test_that("`dm_flatten_to_tbl()` does the right things for 'left_join()'", {
  skip_if_src_not(c("df", "duckdb"))

  local_options(
    pillar.min_title_chars = NULL,
    pillar.max_title_chars = NULL,
    pillar.max_footer_lines = NULL,
    pillar.bold = NULL,
  )

  # FIXME: Debug GHA fail
  # for left join test the basic flattening also on all DBs
  # expect_equivalent_tbl(
  #   expect_message_obj(dm_flatten_to_tbl(dm_for_flatten(), fact)),
  #   result_from_flatten_new()
  # )

  expect_snapshot(
    {
      prepare_dm_for_flatten(
        dm_for_flatten(),
        tables = c("fact", "dim_1", "dim_2", "dim_3", "dim_4"),
        gotta_rename = TRUE
      ) %>%
        dm_get_tables()
      dm_flatten_to_tbl(dm_for_flatten(), fact)
      result_from_flatten_new()
    },
    variant = my_test_src_name
  )

  # a one-table-dm
  expect_equivalent_tbl(
    dm_for_flatten() %>%
      dm_select_tbl(fact) %>%
      dm_flatten_to_tbl(fact),
    fact()
  )

  # explicitly choose parent tables
  out <- expect_message_obj(dm_flatten_to_tbl(
    dm_for_flatten(),
    fact,
    dim_1,
    dim_2
  ))
  expect_equivalent_tbl(
    out,
    left_join(
      fact_clean_new(),
      dim_1_clean_new(),
      by = c("dim_1_key_1" = "dim_1_pk_1", "dim_1_key_2" = "dim_1_pk_2")
    ) %>%
      left_join(dim_2_clean_new(), by = c("dim_2_key" = "dim_2_pk"))
  )

  # change order of parent tables
  out <- expect_message_obj(dm_flatten_to_tbl(
    dm_for_flatten(),
    fact,
    dim_2,
    dim_1
  ))
  expect_equivalent_tbl(
    out,
    left_join(
      fact_clean_new(),
      dim_2_clean_new(),
      by = c("dim_2_key" = "dim_2_pk")
    ) %>%
      left_join(
        dim_1_clean_new(),
        by = c("dim_1_key_1" = "dim_1_pk_1", "dim_1_key_2" = "dim_1_pk_2")
      )
  )

  # with grandparent table
  expect_dm_error(
    dm_flatten_to_tbl(dm_more_complex(), tf_5, tf_4, tf_3),
    class = "only_parents"
  )

  # table unreachable
  expect_dm_error(
    dm_flatten_to_tbl(dm_for_filter(), tf_2, tf_3, tf_4),
    class = "tables_not_reachable_from_start"
  )

  # deeper hierarchy available and `auto_detect = TRUE`
  # for flatten: columns from tf_5 + tf_4 + tf_4_2 + tf_6 are combined in one table, 8 cols in total
  expect_identical(
    ncol(dm_flatten_to_tbl(dm_more_complex(), tf_5)),
    11L
  )
})

test_that("`dm_flatten_to_tbl()` does the right things for 'inner_join()'", {
  local_options(
    pillar.min_title_chars = NULL,
    pillar.max_title_chars = NULL,
    pillar.max_footer_lines = NULL,
    pillar.bold = NULL,
  )

  out <- expect_message_obj(
    arrange(
      dm_flatten_to_tbl(dm_for_flatten(), fact, .join = inner_join),
      pick(everything())
    )
  )
  # FIXME: Debug GHA fail
  # expect_equivalent_tbl(out, result_from_flatten_new())
  expect_snapshot(
    {
      out
    },
    variant = my_test_src_name
  )
})

test_that("`dm_flatten_to_tbl()` does the right things for 'full_join()'", {
  skip_if_src("sqlite")
  skip_if_src("maria")
  out <- expect_message_obj(dm_flatten_to_tbl(
    dm_for_flatten(),
    fact,
    .join = full_join
  ))
  expect_equivalent_tbl(
    out,
    fact_clean_new() %>%
      full_join(
        dim_1_clean_new(),
        by = c("dim_1_key_1" = "dim_1_pk_1", "dim_1_key_2" = "dim_1_pk_2")
      ) %>%
      full_join(dim_2_clean_new(), by = c("dim_2_key" = "dim_2_pk")) %>%
      full_join(dim_3_clean_new(), by = c("dim_3_key" = "dim_3_pk")) %>%
      full_join(dim_4_clean_new(), by = c("dim_4_key" = "dim_4_pk"))
  )
})

test_that("`dm_flatten_to_tbl()` does the right things for 'semi_join()'", {
  expect_equivalent_tbl(
    dm_flatten_to_tbl(dm_for_flatten(), fact, .join = semi_join),
    fact()
  )
})

test_that("`dm_flatten_to_tbl()` does the right things for 'anti_join()'", {
  expect_equivalent_tbl(
    dm_flatten_to_tbl(dm_for_flatten(), fact, .join = anti_join),
    fact() %>% filter(1 == 0)
  )
})

test_that("`dm_flatten_to_tbl()` does the right things for 'nest_join()'", {
  expect_dm_error(
    dm_flatten_to_tbl(dm_for_flatten(), fact, .join = nest_join),
    class = "no_flatten_with_nest_join"
  )
})


test_that("`dm_flatten_to_tbl()` does the right things for 'right_join()'", {
  skip_if_src("sqlite")
  expect_equivalent_tbl(
    expect_message_obj(expect_warning_obj(
      dm_flatten_to_tbl(dm_for_flatten(), fact, .join = right_join),
      "right_join"
    )),
    fact_clean_new() %>%
      right_join(
        dim_1_clean_new(),
        by = c("dim_1_key_1" = "dim_1_pk_1", "dim_1_key_2" = "dim_1_pk_2")
      ) %>%
      right_join(dim_2_clean_new(), by = c("dim_2_key" = "dim_2_pk")) %>%
      right_join(dim_3_clean_new(), by = c("dim_3_key" = "dim_3_pk")) %>%
      right_join(dim_4_clean_new(), by = c("dim_4_key" = "dim_4_pk"))
  )

  # change order of parent tables
  out <- expect_message_obj(dm_flatten_to_tbl(
    dm_for_flatten(),
    fact,
    dim_2,
    dim_1,
    .join = right_join
  ))
  expect_equivalent_tbl(
    out,
    right_join(
      fact_clean_new(),
      dim_2_clean_new(),
      by = c("dim_2_key" = "dim_2_pk")
    ) %>%
      right_join(
        dim_1_clean_new(),
        by = c("dim_1_key_1" = "dim_1_pk_1", "dim_1_key_2" = "dim_1_pk_2")
      )
  )
})

test_that("`dm_squash_to_tbl()` is deprecated but still works", {
  # with grandparent table
  # left_join:
  expect_deprecated(
    expect_equivalent_tbl(
      dm_squash_to_tbl(dm_more_complex(), tf_5, tf_4, tf_3),
      tf_5() %>%
        left_join(tf_4(), by = c("l" = "h")) %>%
        left_join(tf_3(), by = c("j" = "f", "j1" = "f1"))
    )
  )
})

test_that("`dm_flatten_to_tbl(.recursive = TRUE)` does the right things", {
  # with grandparent table
  # left_join:
  expect_equivalent_tbl(
    dm_flatten_to_tbl(dm_more_complex(), tf_5, tf_4, tf_3, .recursive = TRUE),
    tf_5() %>%
      left_join(tf_4(), by = c("l" = "h")) %>%
      left_join(tf_3(), by = c("j" = "f", "j1" = "f1"))
  )

  # deeper hierarchy available and `auto_detect = TRUE`
  # for flatten: columns from tf_5 + tf_4 + tf_3 + tf_4_2 + tf_6 are combined in one table, 10 cols in total
  expect_identical(
    ncol(dm_flatten_to_tbl(dm_more_complex(), tf_5, .recursive = TRUE)),
    12L
  )

  # semi_join:
  expect_dm_error(
    dm_flatten_to_tbl(dm_more_complex(), tf_5, tf_4, tf_3, .join = semi_join, .recursive = TRUE),
    class = "squash_limited"
  )

  # anti_join:
  expect_dm_error(
    dm_flatten_to_tbl(dm_more_complex(), tf_5, tf_4, tf_3, .join = anti_join, .recursive = TRUE),
    class = "squash_limited"
  )

  # fails when there is a cycle:
  expect_dm_error(
    dm_flatten_to_tbl(dm_for_filter_w_cycle(), tf_5, .recursive = TRUE),
    "no_cycles"
  )

  skip_if_src("sqlite")
  skip_if_src("maria")

  # full_join:
  expect_equivalent_tbl(
    dm_flatten_to_tbl(dm_more_complex(), tf_5, tf_4, tf_3, .join = full_join, .recursive = TRUE),
    tf_5() %>%
      full_join(tf_4(), by = c("l" = "h")) %>%
      full_join(tf_3(), by = c("j" = "f", "j1" = "f1"))
  )

  # skipping inner_join, not gaining new info

  # right_join:
  expect_dm_error(
    dm_flatten_to_tbl(dm_more_complex(), tf_5, tf_4, tf_3, .join = right_join, .recursive = TRUE),
    class = "squash_limited"
  )
})

test_that("prepare_dm_for_flatten() works", {
  # with rename
  out <- expect_message_obj(prepare_dm_for_flatten(
    dm_for_flatten(),
    c("fact", "dim_1", "dim_3"),
    gotta_rename = TRUE
  ))
  expect_equivalent_dm(
    out,
    dm_select_tbl(dm_for_flatten(), fact, dim_1, dim_3) %>% dm_disambiguate_cols(.quiet = TRUE)
  )

  # without rename
  expect_equivalent_dm(
    prepare_dm_for_flatten(dm_for_flatten(), c("fact", "dim_1", "dim_3"), gotta_rename = FALSE),
    dm_select_tbl(dm_for_flatten(), fact, dim_1, dim_3)
  )
})

test_that("tidyselect works for flatten", {
  # test if deselecting works
  expect_equivalent_tbl(
    expect_message_obj(dm_flatten_to_tbl(dm_for_flatten(), fact, -dim_2, dim_3, -dim_4, dim_1)),
    expect_message_obj(dm_flatten_to_tbl(dm_for_flatten(), fact, dim_1, dim_3))
  )

  # test if select helpers work
  expect_equivalent_tbl(
    expect_message_obj(dm_flatten_to_tbl(dm_for_flatten(), fact, ends_with("3"), ends_with("1"))),
    expect_message_obj(dm_flatten_to_tbl(dm_for_flatten(), fact, dim_3, dim_1))
  )

  expect_equivalent_tbl(
    expect_message_obj(dm_flatten_to_tbl(dm_for_flatten(), fact, everything())),
    expect_message_obj(dm_flatten_to_tbl(dm_for_flatten(), fact))
  )

  # if only deselecting one potential candidate for flattening, the tables that are not
  # candidates will generally be part of the choice
  expect_dm_error(
    dm_flatten_to_tbl(dm_for_filter(), tf_2, -tf_1),
    class = "tables_not_reachable_from_start"
  )

  # trying to deselect table that doesn't exist:
  expect_error(
    dm_flatten_to_tbl(dm_for_filter(), tf_2, -tf_101),
    class = "vctrs_error_subscript"
  )
})

test_that("`dm_join_to_tbl()` works", {
  expect_deprecated(
    expect_equivalent_tbl(
      expect_message_obj(dm_join_to_tbl(dm_for_flatten(), fact, dim_3), "Renaming"),
      left_join(
        fact_clean(),
        dim_3_clean(),
        by = c("dim_3_key" = "dim_3_pk")
      )
    )
  )

  expect_dm_error(
    expect_deprecated(dm_join_to_tbl(dm_for_filter(), tf_7, tf_8)),
    "table_not_in_dm"
  )
})

# tests that do not work on DB when keys are set ('bad_dm' and 'nycflights'; currently PG and MSSQL)
test_that("tests with 'bad_dm' work", {
  # can't create bad_dm() on Postgres due to strict constraint checks
  skip_if_src("postgres")

  # duckdb doesn't work before R 4.0
  skip_if(getRversion() < "4.0")

  # flatten bad_dm() (no referential integrity)
  if (is_db(my_test_src()) || utils::packageVersion("dplyr") >= "1.1.0.9000") {
    expect_equivalent_tbl(
      dm_flatten_to_tbl(bad_dm(), tbl_1, tbl_2, tbl_3),
      tbl_1() %>%
        left_join(tbl_2(), by = c("a" = "id", "x")) %>%
        left_join(tbl_3(), by = c("b" = "id"))
    )
  }

  skip_if_src("maria")

  # filtered `dm`
  bad_filtered_dm <- dm_filter(bad_dm(), tbl_1 = (a != 4))

  expect_equivalent_tbl(
    dm_flatten_to_tbl(bad_filtered_dm, tbl_1),
    bad_filtered_dm %>% dm_flatten_to_tbl(tbl_1)
  )

  # filtered `dm`
  expect_equivalent_tbl(
    dm_flatten_to_tbl(bad_filtered_dm, tbl_1, .join = semi_join),
    bad_filtered_dm %>% dm_flatten_to_tbl(tbl_1, .join = semi_join)
  )

  # fails when there is a cycle
  expect_dm_error(
    dm_nycflights_small() %>%
      dm_add_fk(flights, origin, airports) %>%
      dm_flatten_to_tbl(flights),
    "no_cycles"
  )
})

test_that("tests with 'bad_dm' work (2)", {
  # can't create bad_dm() on Postgres due to strict constraint checks
  skip_if_src("postgres")

  # full & right join not available on SQLite and MariaDB
  skip_if_src("sqlite", "maria")

  # duckdb doesn't work before R 4.0
  skip_if(getRversion() < "4.0")

  bad_filtered_dm <- dm_filter(bad_dm(), tbl_1 = (a != 4))

  # flatten bad_dm() (no referential integrity)
  if (is_db(my_test_src()) || utils::packageVersion("dplyr") >= "1.1.0.9000") {
    expect_equivalent_tbl(
      dm_flatten_to_tbl(bad_dm(), tbl_1, tbl_2, tbl_3, .join = full_join),
      tbl_1() %>%
        full_join(tbl_2(), by = c("a" = "id", "x")) %>%
        full_join(tbl_3(), by = c("b" = "id"))
    )
  }
})

test_that("tests with 'bad_dm' work (3)", {
  # can't create bad_dm() on Postgres due to strict constraint checks
  skip_if_src("postgres")

  # full & right join not available on SQLite
  skip_if_src("sqlite")

  # duckdb doesn't work before R 4.0
  skip_if(getRversion() < "4.0")

  bad_filtered_dm <- dm_filter(bad_dm(), tbl_1 = (a != 4))

  # flatten bad_dm() (no referential integrity)
  if (is_db(my_test_src()) || utils::packageVersion("dplyr") >= "1.1.0.9000") {
    expect_equivalent_tbl(
      dm_flatten_to_tbl(bad_dm(), tbl_1, tbl_2, tbl_3, .join = right_join),
      tbl_1() %>%
        right_join(tbl_2(), by = c("a" = "id", "x")) %>%
        right_join(tbl_3(), by = c("b" = "id"))
    )
  }

  # flatten bad_dm() (no referential integrity); different order
  if (is_db(my_test_src()) || utils::packageVersion("dplyr") >= "1.1.0.9000") {
    expect_equivalent_tbl(
      dm_flatten_to_tbl(bad_dm(), tbl_1, tbl_3, tbl_2, .join = right_join),
      tbl_1() %>%
        right_join(tbl_3(), by = c("b" = "id")) %>%
        right_join(tbl_2(), by = c("a" = "id", "x"))
    )
  }
})

#' Deprecated functions
#'
#' These functions are deprecated in favor of better alternatives.
#' Most functions with the `cdm_` prefix have an identical alternative
#' with a `dm_` prefix.
#'
#' @name deprecated
NULL

#' sql_schema_create()
#'
#' `sql_schema_*()` functions have been replaced with the corresponding
#' `db_schema_*()` functions.
#'
#' @keywords internal
#' @rdname deprecated
#' @export
sql_schema_create <- function(dest, schema, ...) {
  # FIXME: Use sql_*() methods to construct the SQL code
  # Challenge: How to run multi-statement code with cleanup?
  deprecate_warn("0.2.5", "dm::sql_schema_create()", "dm::db_schema_create()")
  check_dots_empty()
  db_schema_create(dest, schema)
}

#' sql_schema_drop()
#'
#' @keywords internal
#' @rdname deprecated
#' @export
sql_schema_drop <- function(dest, schema, force = FALSE, ...) {
  # FIXME: Use sql_*() methods to construct the SQL code
  # Challenge: How to run multi-statement code with cleanup?
  deprecate_warn("0.2.5", "dm::sql_schema_drop()", "dm::db_schema_drop()")
  check_dots_empty()
  db_schema_drop(dest, schema, force)
}

#' sql_schema_exists()
#'
#' @keywords internal
#' @rdname deprecated
#' @export
sql_schema_exists <- function(dest, schema, ...) {
  # FIXME: Use sql_*() methods to construct the SQL code
  # Challenge: How to run multi-statement code with cleanup?
  deprecate_warn("0.2.5", "dm::sql_schema_exists()", "dm::db_schema_exists()")
  check_dots_empty()
  db_schema_exists(dest, schema)
}

#' sql_schema_list()
#'
#' @keywords internal
#' @rdname deprecated
#' @export
sql_schema_list <- function(dest, include_default = TRUE, ...) {
  # FIXME: Use sql_*() methods to construct the SQL code
  # Challenge: How to run multi-statement code with cleanup?
  deprecate_warn("0.2.5", "dm::sql_schema_list()", "dm::db_schema_list()")
  check_dots_empty()
  db_schema_list(dest, include_default)
}

#' Add foreign keys
#'
#' @description
#' `dm_add_fk()` marks the specified `columns` as the foreign key of table `table` with
#' respect to a key of table `ref_table`.
#' Usually the referenced columns are a primary key in `ref_table`.
#' However, it is also possible to specify other columns via the `ref_columns` argument.
#' If `check == TRUE`, then it will first check if the values in `columns` are a subset
#' of the values of the key in table `ref_table`.
#'
#' @inheritParams dm_add_pk
#' @param columns The columns of `table` which are to become the foreign key columns that
#'   reference `ref_table`.
#'   To define a compound key, use `c(col1, col2)`.
#' @param ref_table The table which `table` will be referencing.
#' @param ref_columns The column(s) of `table` which are to become the referenced column(s) in `ref_table`.
#'   By default, the primary key is used.
#'   To define a compound key, use `c(col1, col2)`.
#' @param check Boolean, if `TRUE`, a check will be performed to determine if the values of
#'   `columns` are a subset of the values of the key column(s) of `ref_table`.
#' @param on_delete
#'   `r lifecycle::badge("experimental")`
#'
#'   Defines behavior if a row in the parent table is deleted.
#'     - `"no_action"`, the default, means that no action is taken
#'        and the operation is aborted if child rows exist
#'     - `"cascade"` means that the child row is also deleted
#'   This setting is picked up by [copy_dm_to()] with `set_key_constraints = TRUE`,
#'   and by [dm_sql()],
#'   and might be considered by [dm_rows_delete()] in a future version.
#'
#' @family foreign key functions
#'
#' @details
#' It is possible that a foreign key (FK) is pointing to columns that are neither primary (PK) nor
#' explicit unique keys (UK).
#' This can happen
#'   1. when a FK is added without a corresponding PK or UK being present in the parent table
#'   1. when the PK or UK is removed ([`dm_rm_pk()`]/[`dm_rm_uk()`]) without first removing the associated FKs.
#'
#' These columns are then a so-called "implicit unique key" of the referenced table and can be listed via [`dm_get_all_uks()`].
#'
#' @rdname dm_add_fk
#'
#' @return An updated `dm` with an additional foreign key relation.
#'
#' @export
#' @examplesIf rlang::is_installed(c("nycflights13", "DiagrammeR"))
#' nycflights_dm <- dm(
#'   planes = nycflights13::planes,
#'   flights = nycflights13::flights,
#'   weather = nycflights13::weather
#' )
#'
#' nycflights_dm %>%
#'   dm_draw()
#'
#' # Create foreign keys:
#' nycflights_dm %>%
#'   dm_add_pk(planes, tailnum) %>%
#'   dm_add_fk(flights, tailnum, planes) %>%
#'   dm_add_pk(weather, c(origin, time_hour)) %>%
#'   dm_add_fk(flights, c(origin, time_hour), weather) %>%
#'   dm_draw()
#'
#' # Keys can be checked during creation:
#' try(
#'   nycflights_dm %>%
#'     dm_add_pk(planes, tailnum) %>%
#'     dm_add_fk(flights, tailnum, planes, check = TRUE)
#' )
dm_add_fk <- function(
  dm,
  table,
  columns,
  ref_table,
  ref_columns = NULL,
  ...,
  check = FALSE,
  on_delete = c("no_action", "cascade")
) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)
  table_name <- dm_tbl_name(dm, {{ table }})
  ref_table_name <- dm_tbl_name(dm, {{ ref_table }})
  on_delete <- arg_match(on_delete)

  table_obj <- tbl_impl(dm, table_name)
  col_expr <- enexpr(columns)
  col_name <- names(eval_select_indices(col_expr, colnames(table_obj)))

  ref_table_obj <- tbl_impl(dm, ref_table_name)
  ref_col_expr <- enexpr(ref_columns)
  if (is.null(ref_col_expr)) {
    ref_key <- dm_get_pk_impl(dm, ref_table_name)

    if (is_empty(ref_key)) {
      abort_ref_tbl_has_no_pk(ref_table_name)
    }

    ref_col_name <- get_key_cols(ref_key)
  } else {
    ref_col_name <- names(eval_select_indices(ref_col_expr, colnames(ref_table_obj)))
  }

  # FIXME: COMPOUND:: Clean check with proper error message
  stopifnot(length(ref_col_name) == length(col_name))

  if (check) {
    if (!is_subset(table_obj, !!col_name, ref_table_obj, !!ref_col_name)) {
      abort_not_subset_of(table_name, col_name, ref_table_name, ref_col_name)
    }
  }

  dm_add_fk_impl(dm, table_name, list(col_name), ref_table_name, list(ref_col_name), on_delete)
}

dm_add_fk_impl <- function(dm, table, column, ref_table, ref_column, on_delete) {
  column <- unclass(column)
  ref_column <- unclass(ref_column)

  on_delete <- vec_recycle(on_delete, length(ref_table))

  loc <- which(!duplicated(ref_table))
  n_loc <- length(loc)
  if (n_loc > 1) {
    my_ref_table <- ref_table[[loc[[n_loc]]]]

    my <- ref_table == my_ref_table
    where_other <- which(!my)
    dm <- dm_add_fk_impl(
      dm,
      table[where_other],
      column[where_other],
      ref_table[where_other],
      ref_column[where_other],
      on_delete[where_other]
    )

    table <- table[my]
    column <- column[my]
    ref_column <- ref_column[my]
    on_delete <- on_delete[my]
    # ref_table must be scalar, unlike the others
    ref_table <- my_ref_table
  } else if (n_loc == 0) {
    return(dm)
  } else {
    my_ref_table <- ref_table[[1]]
  }

  def <- dm_get_def(dm)

  i <- which(def$table == ref_table)

  fks <- def$fks[[i]]

  existing <- fks$table == table & !is.na(vec_match(fks$column, column))
  if (any(existing)) {
    if (dm_is_strict_keys(dm)) {
      first_existing <- which(existing)[[1]]
      abort_fk_exists(table[[first_existing]], column[[first_existing]], ref_table)
    }

    stopifnot(all(existing))

    return(dm)
  }

  def$fks[[i]] <- vec_rbind(
    fks,
    new_fk(ref_column, table, column, on_delete)
  )

  dm_from_def(def)
}

#' Check if foreign keys exists
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' These functions are deprecated because of their limited use
#' since the introduction of foreign keys to arbitrary columns in dm 0.2.1.
#' Use [dm_get_all_fks()] with table manipulation functions instead.
#'
#' @inheritParams dm_add_fk
#' @export
#' @keywords internal
dm_has_fk <- function(dm, table, ref_table, ...) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)

  deprecate_warn("0.2.1", "dm::dm_has_fk()", "dm::dm_get_all_fks()")

  table_name <- dm_tbl_name(dm, {{ table }})
  ref_table_name <- dm_tbl_name(dm, {{ ref_table }})
  dm_has_fk_impl(dm, table_name, ref_table_name)
}

dm_has_fk_impl <- function(dm, table_name, ref_table_name) {
  has_length(dm_get_fk_impl(dm, table_name, ref_table_name))
}

#' @rdname dm_has_fk
#' @export
dm_get_fk <- function(dm, table, ref_table, ...) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)

  deprecate_warn("0.2.1", "dm::dm_get_fk()", "dm::dm_get_all_fks()")

  table_name <- dm_tbl_name(dm, {{ table }})
  ref_table_name <- dm_tbl_name(dm, {{ ref_table }})

  new_keys(dm_get_fk_impl(dm, table_name, ref_table_name))
}

dm_get_fk_impl <- function(dm, table_name, ref_table_name) {
  def <- dm_get_def(dm)
  i <- which(def$table == ref_table_name)

  fks <- def$fks[[i]]
  fks$column[fks$table == table_name]
}

dm_get_fk2_impl <- function(dm, table_name, ref_table_name) {
  # FIXME: Revisit instances of dm_get_fk_impl()
  def <- dm_get_def(dm)
  i <- which(def$table == ref_table_name)

  fks <- def$fks[[i]]
  fks[fks$table == table_name, c("column", "ref_column")]
}

#' Get foreign key constraints
#'
#' @description
#' Get a summary of all foreign key relations in a [`dm`].
#'
#' @return A tibble with the following columns:
#'   \describe{
#'     \item{`child_table`}{child table,}
#'     \item{`child_fk_cols`}{foreign key column(s) in child table as list of character vectors,}
#'     \item{`parent_table`}{parent table,}
#'     \item{`parent_key_cols`}{key column(s) in parent table as list of character vectors.}
#'     \item{`on_delete`}{behavior on deletion of rows in the parent table.}
#'   }
#'
#' @inheritParams dm_has_fk
#' @param parent_table One or more table names, unquoted,
#'   to return foreign key information for.
#'   If given, foreign keys are returned in that order.
#'   The default `NULL` returns information for all tables.
#'
#' @family foreign key functions
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_get_all_fks()
#' @export
dm_get_all_fks <- function(dm, parent_table = NULL, ...) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)
  table_expr <- enexpr(parent_table) %||% src_tbls_impl(dm, quiet = TRUE)
  table_names <- eval_select_table(table_expr, set_names(src_tbls_impl(dm, quiet = TRUE)))
  dm_get_all_fks_impl(dm, table_names)
}

dm_get_all_fks_impl <- function(dm, parent_table = NULL, ignore_on_delete = FALSE, id = FALSE) {
  def <- dm_get_def(dm)

  dm_get_all_fks_def_impl(
    def = def,
    parent_table = parent_table,
    ignore_on_delete = ignore_on_delete,
    id = id
  )
}

dm_get_all_fks_def_impl <- function(
  def,
  parent_table = NULL,
  ignore_on_delete = FALSE,
  id = FALSE
) {
  def_sub <- def[c("table", "fks")]
  names(def_sub)[[1]] <- "parent_table"

  if (!is.null(parent_table)) {
    idx <- match(parent_table, def_sub$parent_table)
    if (anyNA(idx)) {
      abort_table_not_in_dm(parent_table[which(is.na(idx))], def$table)
    }
    def_sub <- def_sub[idx, ]
  }

  flat <- unnest_list_of_df(def_sub, "fks")

  names(flat) <- c("parent_table", "parent_key_cols", "child_table", "child_fk_cols", "on_delete")
  flat[[2]] <- new_keys(flat[[2]])
  flat[[4]] <- new_keys(flat[[4]])
  out <- flat[c(3:4, 1:2, if (!ignore_on_delete) 5L)]
  if (id) {
    out <-
      out %>%
      group_by(child_table) %>%
      mutate(id = paste0(child_table, "_", row_number())) %>%
      ungroup()
  }
  out
}

#' Remove foreign keys
#'
#' @description
#' `dm_rm_fk()` can remove either one reference between two tables, or multiple references at once (with a message).
#' An error is thrown if no matching foreign key is found.
#'
#' @family foreign key functions
#'
#' @inheritParams dm_rm_pk
#' @param ref_table The table referenced by the `table` argument.
#'   Pass `NULL` to remove all matching keys.
#' @param ref_columns The columns of `table` that should no longer be referencing the primary key of `ref_table`.
#'   To refer to a compound key, use `c(col1, col2)`.
#'
#' @return An updated `dm` without the matching foreign key relation(s).
#'
#' @export
#' @examplesIf rlang::is_installed(c("nycflights13", "DiagrammeR"))
#' dm_nycflights13(cycle = TRUE) %>%
#'   dm_rm_fk(flights, dest, airports) %>%
#'   dm_draw()
dm_rm_fk <- function(dm, table = NULL, columns = NULL, ref_table = NULL, ref_columns = NULL, ...) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)

  table_name <- dm_tbl_name_null(dm, {{ table }})
  column_expr <- enexpr(columns)
  ref_table_name <- dm_tbl_name_null(dm, {{ ref_table }})
  ref_column_expr <- enexpr(ref_columns)

  dm_rm_fk_impl(dm, table_name, column_expr, ref_table_name, ref_column_expr)
}

#' @autoglobal
dm_rm_fk_impl <- function(
  dm,
  table_name,
  cols,
  ref_table_name,
  ref_cols,
  error_call = caller_env()
) {
  def <- dm_get_def(dm)

  # Filter by each argument if given:

  # ref_table_name: keyed by def$table, simplest
  if (is.null(ref_table_name)) {
    idx <- seq_along(def$table)
  } else {
    idx <- which(def$table == ref_table_name)
  }

  # other args: inside def$fks, maintaining list of indexes
  idx_fk <- map(def$fks[idx], ~ seq_len(nrow(.x)))

  # table_name: keep FK entries pointing to the other table
  if (!is.null(table_name)) {
    idx_fk <- map2(
      def$fks[idx],
      idx_fk,
      ~ {
        ii <- (.x$table[.y] == table_name)
        .y[ii]
      }
    )

    # Prune after each step (this also ensures that negative selection works further below)
    keep <- (lengths(idx_fk) > 0)
    idx <- idx[keep]
    idx_fk <- idx_fk[keep]
  }

  # ref_cols: find column names once for each ref_table
  if (!is.null(ref_cols)) {
    idx_fk <- pmap(
      list(def$fks[idx], idx_fk, def$data[idx]),
      ~ {
        ii <- tryCatch(
          {
            names_vars <- names(eval_select_indices(
              ref_cols,
              colnames(..3),
              error_call = error_call
            ))
            map_lgl(.x$ref_column[.y], identical, names_vars)
          },
          error = function(e) {
            0
          }
        )
        .y[ii]
      }
    )

    # Prune after each step (this also ensures that negative selection works further below)
    keep <- (lengths(idx_fk) > 0)
    idx <- idx[keep]
    idx_fk <- idx_fk[keep]
  }

  # cols: find column inside each fks entry
  if (!is.null(cols)) {
    all_tables <- set_names(def$data, def$table)

    idx_fk <- map2(
      def$fks[idx],
      idx_fk,
      ~ {
        ii <- map2_lgl(
          .x$table[.y],
          .x$column[.y],
          ~ {
            tryCatch(
              {
                names_vars <- names(eval_select_indices(
                  cols,
                  colnames(all_tables[[.x]]),
                  error_call = error_call
                ))
                identical(.y, names_vars)
              },
              error = function(e) {
                FALSE
              }
            )
          }
        )
        .y[ii]
      }
    )

    # Prune after each step (this also ensures that negative selection works further below)
    keep <- (lengths(idx_fk) > 0)
    idx <- idx[keep]
    idx_fk <- idx_fk[keep]
  }

  # Check if empty
  if (length(idx) == 0) {
    abort_is_not_fkc()
  }

  # Talk about it
  if (is.null(table_name) || is.null(cols) || is.null(ref_table_name)) {
    show_disambiguation <- TRUE
  } else if (!is.null(ref_cols)) {
    show_disambiguation <- FALSE
  } else {
    # Check if all FKs (being removed) point to the primary key
    show_disambiguation <- !all(pmap_lgl(
      list(def$fks[idx], idx_fk, def$pks[idx]),
      ~ {
        all(map_lgl(..1$ref_column[..2], identical, ..3$column[[1]]))
      }
    ))
  }

  if (show_disambiguation) {
    def_rm <- def[idx, c("table", "pks", "fks")]
    def_rm$fks <- map2(def_rm$fks, idx_fk, vec_slice)
    def_rm$fks <- map2(
      def_rm$fks,
      def_rm$pks,
      ~ {
        .x$need_ref <- !map_lgl(.x$ref_column, identical, .y$column[[1]])
        .x
      }
    )

    disambiguation <-
      def_rm %>%
      select(ref_table = table, fks) %>%
      unnest(-ref_table) %>%
      mutate(ref_col_text = if_else(need_ref, glue(", {deparse_keys(ref_column)})"), "")) %>%
      mutate(
        text = glue(
          "dm_rm_fk({tick_if_needed(table)}, {deparse_keys(column)}, {tick_if_needed(ref_table)}{ref_col_text})"
        )
      ) %>%
      pull()

    message("Removing foreign keys: %>%\n  ", glue_collapse(disambiguation, " %>%\n  "))
  }

  # Execute
  def$fks[idx] <- map2(def$fks[idx], idx_fk, ~ .x[-.y, ])

  dm_from_def(def)
}

#' Foreign key candidates
#'
#' @description `r lifecycle::badge("experimental")`
#'
#' Determine which columns would be good candidates to be used as foreign keys of a table,
#' to reference the primary key column of another table of the [`dm`] object.
#'
#' @inheritParams dm_add_fk
#' @param table The table whose columns should be tested for suitability as foreign keys.
#' @param ref_table A table with a primary key.
#'
#' @details `dm_enum_fk_candidates()` first checks if `ref_table` has a primary key set,
#' if not, an error is thrown.
#'
#' If `ref_table` does have a primary key, then a join operation will be tried using
#' that key as the `by` argument of join() to match it to each column of `table`.
#' Attempting to join incompatible columns triggers an error.
#'
#' The outcome of the join operation determines the value of the `why` column in the result:
#'
#' - an empty value for a column of `table` that is a suitable foreign key candidate
#' - the count and percentage of missing matches for a column that is not suitable
#' - the error message triggered for unsuitable candidates that may include the types of mismatched columns
#'
#' @section Life cycle:
#' These functions are marked "experimental" because we are not yet sure about
#' the interface, in particular if we need both `dm_enum...()` and `enum...()`
#' variants.
#' Changing the interface later seems harmless because these functions are
#' most likely used interactively.
#'
#' @return A tibble with the following columns:
#'   \describe{
#'     \item{`columns`}{columns of `table`,}
#'     \item{`candidate`}{boolean: are these columns a candidate for a foreign key,}
#'     \item{`why`}{if not a candidate for a foreign key, explanation for for this.}
#'   }
#'
#' @family foreign key functions
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_enum_fk_candidates(flights, airports)
#'
#' dm_nycflights13() %>%
#'   dm_zoom_to(flights) %>%
#'   enum_fk_candidates(airports)
#' @export
dm_enum_fk_candidates <- function(dm, table, ref_table, ...) {
  check_dots_empty()
  dm_local_error_call()
  check_not_zoomed(dm)
  # FIXME: with "direct" filter maybe no check necessary: but do we want to check
  # for tables retrieved with `tbl()` or with `dm_get_tables()[[table_name]]`
  check_no_filter(dm)
  table_name <- dm_tbl_name(dm, {{ table }})
  ref_table_name <- dm_tbl_name(dm, {{ ref_table }})

  ref_tbl_pk <- dm_get_pk_impl(dm, ref_table_name)

  ref_tbl <- tbl_impl(dm, ref_table_name)
  tbl <- tbl_impl(dm, table_name)

  table_name %>%
    enum_fk_candidates_impl(tbl, ref_table_name, ref_tbl, ref_tbl_pk) %>%
    rename(columns = column) %>%
    mutate(columns = new_keys(columns))
}

#' @details `enum_fk_candidates()` works like `dm_enum_fk_candidates()` with the zoomed table as `table`.
#'
#' @rdname dm_enum_fk_candidates
#' @param dm_zoomed A `dm` with a zoomed table.
#' @export
enum_fk_candidates <- function(dm_zoomed, ref_table, ...) {
  check_dots_empty()
  check_zoomed(dm_zoomed)
  check_no_filter(dm_zoomed)

  table_name <- orig_name_zoomed(dm_zoomed)
  ref_table_name <- dm_tbl_name(dm_zoomed, {{ ref_table }})

  ref_tbl_pk <- dm_get_pk_impl(dm_zoomed, ref_table_name)

  ref_tbl <- dm_get_tables_impl(dm_zoomed)[[ref_table_name]]
  enum_fk_candidates_impl(
    table_name,
    tbl_zoomed(dm_zoomed),
    ref_table_name,
    ref_tbl,
    ref_tbl_pk
  ) %>%
    rename(columns = column) %>%
    mutate(columns = new_keys(columns))
}

#' @autoglobal
enum_fk_candidates_impl <- function(table_name, tbl, ref_table_name, ref_tbl, ref_tbl_pk) {
  if (is_empty(ref_tbl_pk)) {
    abort_ref_tbl_has_no_pk(ref_table_name)
  }
  ref_tbl_cols <- get_key_cols(ref_tbl_pk)

  tbl_colnames <- colnames(tbl)
  tibble(
    column = tbl_colnames,
    why = map_chr(column, ~ check_fk(tbl, table_name, .x, ref_tbl, ref_table_name, ref_tbl_cols))
  ) %>%
    mutate(candidate = ifelse(why == "", TRUE, FALSE)) %>%
    select(column, candidate, why) %>%
    arrange(desc(candidate))
}

check_fk <- function(t1, t1_name, colname, t2, t2_name, pk, max_value = MAX_COMMAS) {
  stopifnot(length(colname) == length(pk))

  val_names <- paste0("value", seq_along(colname))
  t1_vals <- syms(colname)
  names(t1_vals) <- val_names
  t2_vals <- syms(pk)
  names(t2_vals) <- val_names

  t1_join <-
    t1 %>%
    count(!!!t1_vals) %>%
    ungroup()
  t2_join <-
    t2 %>%
    count(!!!t2_vals) %>%
    ungroup()

  val_names_na_expr <- map(syms(val_names), ~ call("is.na", .x))
  any_value_na_expr <- reduce(val_names_na_expr, ~ call("|", .x, .y))

  # Work around weird bug in R 3.6 and before
  if (getRversion() < "4.0" && inherits(t1_join, "tbl_lazy")) {
    dbplyr::sql_render(t1_join)
  }

  res_tbl <- tryCatch(
    t1_join %>%
      # if value* is NULL, this also counts as a match -- consistent with fk semantics
      filter(!(!!any_value_na_expr)) %>%
      anti_join(t2_join, by = val_names) %>%
      arrange(desc(n), !!!syms(val_names)) %>%
      head(max_value + 1L) %>%
      collect(),
    error = identity
  )

  # return error message if error occurred (possibly types didn't match etc.)
  if (is_condition(res_tbl)) {
    return(conditionMessage(res_tbl))
  }

  # return empty character if candidate
  if (nrow(res_tbl) == 0) {
    return("")
  }

  res_tbl[val_names] <- map(res_tbl[val_names], format, trim = TRUE, justify = "none")
  res_tbl[val_names[-1]] <- map(res_tbl[val_names[-1]], ~ paste0(", ", .x))
  res_tbl$value <- exec(paste0, !!!res_tbl[val_names])

  vals_formatted <- commas(
    glue("{res_tbl$value} ({res_tbl$n})"),
    max_commas = max_value,
    capped = TRUE
  )
  glue(
    "values of ",
    "{commas(tick(glue('{t1_name}${colname}')), Inf)} not in {commas(tick(glue('{t2_name}${pk}')), Inf)}: {vals_formatted}"
  )
}

fk_table_to_def_fks <- function(
  table,
  child_table = "child_table",
  child_fk_cols = "child_fk_cols",
  parent_table = "parent_table",
  parent_key_cols = "parent_key_cols"
) {
  table %>%
    group_by(!!ensym(parent_table)) %>%
    summarize(
      fks = list_of(new_fk(
        ref_column = as.list(!!ensym(parent_key_cols)),
        table = !!ensym(child_table),
        column = as.list(!!ensym(child_fk_cols)),
        on_delete = on_delete
      ))
    )
}

# Errors ------------------------------------------------------------------

abort_fk_exists <- function(child_table_name, colnames, parent_table_name) {
  cli::cli_abort(
    "({commas(tick(colnames))}) is already a foreign key of table {.field {child_table_name}} into table {.field {parent_table_name}}.",
    class = dm_error_full("fk_exists"),
    call = dm_error_call()
  )
}

abort_is_not_fkc <- function() {
  cli::cli_abort(
    "No foreign keys to remove.",
    class = dm_error_full("is_not_fkc"),
    call = dm_error_call()
  )
}

test_that("dm_add_fk() works as intended?", {
  expect_dm_error(
    dm_test_obj() %>%
      dm_add_fk(dm_table_1, a, dm_table_4),
    class = "ref_tbl_has_no_pk"
  )

  expect_dm_error(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      dm_add_fk(dm_table_1, a, dm_table_4),
    class = "fk_exists"
  )

  expect_snapshot({
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      get_all_keys()
  })
})

test_that("dm_has_fk() and dm_get_fk() work as intended?", {
  local_options(lifecycle_verbosity = "quiet")

  expect_identical(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      dm_add_fk(dm_table_2, c, dm_table_4) %>%
      dm_get_fk(dm_table_1, dm_table_4),
    new_keys("a")
  )

  expect_true(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      dm_add_fk(dm_table_2, c, dm_table_4) %>%
      dm_has_fk(dm_table_2, dm_table_4)
  )

  expect_identical(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      dm_add_fk(dm_table_2, c, dm_table_4) %>%
      dm_get_fk(dm_table_2, dm_table_4),
    new_keys("c")
  )

  expect_false(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      dm_add_fk(dm_table_2, c, dm_table_4) %>%
      dm_has_fk(dm_table_3, dm_table_4)
  )

  expect_identical(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      dm_add_fk(dm_table_2, c, dm_table_4) %>%
      dm_get_fk(dm_table_3, dm_table_4),
    new_keys(character(0))
  )
})

test_that("dm_rm_fk() works as intended?", {
  expect_silent(expect_true(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      dm_add_fk(dm_table_2, c, dm_table_4) %>%
      dm_rm_fk(dm_table_2, c, dm_table_4) %>%
      dm_has_fk_impl("dm_table_1", "dm_table_4")
  ))

  expect_silent(expect_false(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      dm_add_fk(dm_table_2, c, dm_table_4) %>%
      dm_rm_fk(dm_table_2, c, dm_table_4) %>%
      dm_has_fk_impl("dm_table_2", "dm_table_4")
  ))

  expect_message(expect_false(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      dm_add_fk(dm_table_2, c, dm_table_4) %>%
      dm_rm_fk(dm_table_2, NULL, dm_table_4) %>%
      dm_has_fk_impl("dm_table_2", "dm_table_4")
  ))

  expect_dm_error(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_1, a, dm_table_4) %>%
      dm_add_fk(dm_table_2, c, dm_table_4) %>%
      dm_rm_fk(dm_table_2, z, dm_table_4),
    class = "is_not_fkc"
  )

  expect_dm_error(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_add_fk(dm_table_2, c, dm_table_4) %>%
      dm_rm_fk(dm_table_2, c, dm_table_4) %>%
      dm_rm_fk(dm_table_2, c, dm_table_4),
    class = "is_not_fkc"
  )

  # Bad input
  expect_dm_error(
    dm_for_filter() %>%
      dm_rm_fk(tf_x),
    class = "table_not_in_dm"
  )

  expect_dm_error(
    dm_for_filter() %>%
      dm_rm_fk(columns = x),
    class = "is_not_fkc"
  )

  expect_dm_error(
    dm_for_filter() %>%
      dm_rm_fk(ref_table = tf_x),
    class = "table_not_in_dm"
  )

  expect_dm_error(
    dm_for_filter() %>%
      dm_rm_fk(ref_columns = x),
    class = "is_not_fkc"
  )
})


test_that("dm_rm_fk() works with partial matching", {
  expect_snapshot({
    # Only table
    dm_for_filter() %>%
      dm_rm_fk(tf_5) %>%
      get_all_keys()

    # Only columns
    dm_for_filter() %>%
      dm_rm_fk(columns = l) %>%
      get_all_keys()

    # Only columns, compound
    dm_for_filter() %>%
      dm_rm_fk(columns = c(e, e1)) %>%
      get_all_keys()

    # Only ref_table
    dm_for_filter() %>%
      dm_rm_fk(ref_table = tf_3) %>%
      get_all_keys()

    # Only ref_columns, compound
    dm_for_filter() %>%
      dm_rm_fk(ref_columns = c(f, f1)) %>%
      get_all_keys()

    # All foreign keys
    dm_for_filter() %>%
      dm_rm_fk() %>%
      get_all_keys()
  })
})


test_that("dm_rm_fk() produces no message when removing FK to PK in presence of FK to non-PK", {
  p <- tibble(p_id = 1, p2_id = 1)
  c1 <- tibble(p_id = 1)
  c2 <- tibble(p2_id = 1)

  my_dm <-
    dm(p, c1, c2) %>%
    dm_add_pk(p, p_id) %>%
    dm_add_fk(c1, p_id, p) %>%
    dm_add_fk(c2, p2_id, p, p2_id)

  expect_silent(my_dm %>% dm_rm_fk(c1, p_id, p))
})


test_that("dm_enum_fk_candidates() works as intended?", {
  skip_if_ide()

  # `anti_join()` doesn't distinguish between `dbl` and `int`
  tbl_fk_candidates_tf_1_tf_4 <- tribble(
    ~column , ~candidate , ~why       ,
    "a"     , TRUE       , ""         ,
    "b"     , FALSE      , "<reason>"
  ) %>%
    rename(columns = column) %>%
    mutate(columns = new_keys(columns))

  expect_identical(
    dm_test_obj() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_enum_fk_candidates(dm_table_1, dm_table_4) %>%
      mutate(why = if_else(why != "", "<reason>", "")) %>%
      collect(),
    tbl_fk_candidates_tf_1_tf_4
  )

  tbl_tf_3_tf_4 <- tibble::tribble(
    ~column , ~candidate , ~why       ,
    "c"     , FALSE      , "<reason>"
  ) %>%
    rename(columns = column) %>%
    mutate(columns = new_keys(columns))

  expect_identical(
    dm_test_obj_2() %>%
      dm_add_pk(dm_table_4, c) %>%
      dm_enum_fk_candidates(dm_table_3, dm_table_4) %>%
      mutate(why = if_else(why != "", "<reason>", "")),
    tbl_tf_3_tf_4
  )

  tbl_tf_4_tf_3 <- tibble::tribble(
    ~column , ~candidate , ~why ,
    "c"     , TRUE       , ""
  ) %>%
    rename(columns = column) %>%
    mutate(columns = new_keys(columns))

  expect_identical(
    dm_test_obj() %>%
      dm_add_pk(dm_table_3, c) %>%
      dm_enum_fk_candidates(dm_table_4, dm_table_3),
    tbl_tf_4_tf_3
  )

  expect_dm_error(
    dm_enum_fk_candidates(dm_test_obj(), dm_table_1, dm_table_4),
    class = "ref_tbl_has_no_pk"
  )

  expect_snapshot({
    dm_nycflights13() %>%
      dm_enum_fk_candidates(flights, airports) %>%
      mutate(why = if_else(why != "", "<reason>", ""))
  })
})

test_that("enum_fk_candidates() works properly", {
  # FIXME: COMPOUND: Test for tf_2 -> tf_3 and other combinations too
  expect_silent(
    expect_equivalent_why(
      enum_fk_candidates(dm_zoomed(), tf_1),
      dm_enum_fk_candidates(dm_for_filter(), tf_2, tf_1)
    )
  )
})

test_that("can add foreign key with cascade", {
  expect_snapshot({
    dm(a = tibble(x = 1), b = tibble(x = 1)) %>%
      dm_add_fk(a, x, b, x, on_delete = "cascade") %>%
      dm_get_all_fks()
  })
})

test_that("bogus arguments are rejected", {
  expect_snapshot(error = TRUE, {
    dm(a = tibble(x = 1), b = tibble(x = 1)) %>%
      dm_add_fk(a, x, b, x, on_delete = "bogus")
    dm(a = tibble(x = 1), b = tibble(x = 1)) %>%
      dm_add_fk(a, x, b, x, on_delete = letters)
  })
})


# all foreign keys --------------------------------------------------------------------

test_that("dm_get_all_fks() and order", {
  dm <- dm_for_filter()
  fks_all <- dm_get_all_fks(dm)
  fks_1 <- dm_get_all_fks(dm, "tf_1")
  fks_3 <- dm_get_all_fks(dm, tf_3)
  fks_4 <- dm_get_all_fks(dm, "tf_4")
  fks_6 <- dm_get_all_fks(dm, "tf_6")
  fks_34 <- dm_get_all_fks(dm, c(tf_3, tf_4))
  fks_43 <- dm_get_all_fks(dm, c(tf_4, tf_3))

  expect_equal(fks_all, bind_rows(fks_1, fks_3, fks_4, fks_6))
  expect_equal(fks_34, bind_rows(fks_3, fks_4))
  expect_equal(fks_43, bind_rows(fks_4, fks_3))
})

test_that("dm_get_all_fks() with parent_table arg", {
  expect_snapshot({
    nyc_comp() %>%
      dm_get_all_fks(weather)

    nyc_comp() %>%
      dm_get_all_fks(c("airlines", "weather"))

    # test tidyselect functions for parent_table arg
    nyc_comp() %>%
      dm_get_all_fks(ends_with("ports"))

    nyc_comp() %>%
      dm_get_all_fks(everything())
  })
})

test_that("dm_get_all_fks() with parent_table arg fails nicely", {
  skip_if(packageVersion("tidyselect") > "1.2.0")

  expect_snapshot_error({
    nyc_comp() %>%
      dm_get_all_fks(c(airlines, weather, timetable, tabletime))
  })
})

flights <-
  nycflights13::flights %>%
  filter(day == 10, month %in% c(1, 2))

weather <-
  nycflights13::weather %>%
  filter(day == 10, month %in% c(1, 2))

airlines <-
  nycflights13::airlines %>%
  semi_join(flights, by = "carrier")

airports <-
  nycflights13::airports %>%
  left_join(flights %>% select(origin, origin_day = day), by = c("faa" = "origin")) %>%
  left_join(flights %>% select(dest, dest_day = day), by = c("faa" = "dest")) %>%
  filter(!is.na(origin_day) | !is.na(dest_day)) %>%
  select(-origin_day, -dest_day) %>%
  distinct()

planes <-
  nycflights13::planes %>%
  semi_join(flights, by = "tailnum")

data <- tibble::lst(flights, weather, airlines, airports, planes)

dir.create("inst/extdata", showWarnings = FALSE)
saveRDS(data, "inst/extdata/nycflights13-small.rds", compress = "gzip", version = 2)

#' Decompose a table into two linked tables
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Perform table surgery by extracting a 'parent table' from a table, linking the original table and the new table by a key, and returning both tables.
#'
#' `decompose_table()` accepts a data frame, a name for the 'ID column' that will be newly created, and the names
#' of the columns that will be extracted into the new data frame.
#'
#' It creates a 'parent table', which consists of the columns specified in the ellipsis, and a new 'ID column'.
#' Then it removes those
#' columns from the original table, which is now called the 'child table, and adds the 'ID column'.
#'
#' @param .data Data frame from which columns `...` are to be extracted.
#' @param new_id_column Name of the identifier column (primary key column) for the parent table.
#'   A column of this name is also added in 'child table'.
#' @param ... The columns to be extracted from the `.data`.
#'
#'   One or more unquoted expressions separated by commas.
#'   You can treat variable names as if they were positions, so you
#'   can use expressions like x:y to select ranges of variables.
#'
#'   The arguments in ... are automatically quoted and evaluated in a context where column names represent column positions.
#'   They also support
#'   unquoting and splicing.
#'   See vignette("programming") for an introduction to those concepts.
#'
#'   See select helpers for more details, and the examples about tidyselect helpers, such as starts_with(), everything(), ...
#'
#' @family table surgery functions
#'
#' @return A named list of length two:
#'   - entry "child_table": the child table with column `new_id_column` referring to the same column in `parent_table`,
#'   - entry "parent_table": the "lookup table" for `child_table`.
#'
#' @section Life cycle:
#' This function is marked "experimental" because it seems more useful
#' when applied to a table in a dm object.
#' Changing the interface later seems harmless because these functions are
#' most likely used interactively.
#'
#' @examples
#' decomposed_table <- decompose_table(mtcars, new_id, am, gear, carb)
#' decomposed_table$child_table
#' decomposed_table$parent_table
#' @export
decompose_table <- function(.data, new_id_column, ...) {
  dm_local_error_call()
  table_name <- deparse(substitute(.data))
  avail_cols <- colnames(.data)
  id_col_q <- ensym(new_id_column)

  if (as_string(id_col_q) %in% avail_cols) {
    abort_dupl_new_id_col_name(table_name)
  }

  sel_vars <- eval_select_both(quo(c(...)), avail_cols)

  parent_table <-
    select(.data, !!!sel_vars$indices) %>%
    distinct() %>%
    # Without as.integer(), RPostgres creates integer64 column (#15)
    mutate(
      !!id_col_q := as.integer(coalesce(row_number(!!sym(names(sel_vars$indices)[[1]])), 0L))
    ) %>%
    select(!!id_col_q, everything())

  non_key_indices <-
    setdiff(seq_along(avail_cols), sel_vars$indices)

  child_table <-
    .data %>%
    left_join(
      parent_table,
      by = prep_recode(sel_vars$names)
    ) %>%
    select(!!!non_key_indices, !!id_col_q)
  # FIXME: Think about a good place for the target column,
  # perhaps if this operation is run in a data model?

  list("child_table" = child_table, "parent_table" = parent_table)
}

#' Merge two tables that are linked by a foreign key relation
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Perform table fusion by combining two tables by a common (key) column, and then removing this column.
#'
#' `reunite_parent_child()`: After joining the two tables by the column `id_column`, this column will be removed.
#' The transformation is roughly the
#' inverse of what `decompose_table()` does.
#'
#' @param child_table Table (possibly created by `decompose_table()`) that references `parent_table`
#' @param parent_table Table (possibly created by `decompose_table()`).
#' @param id_column Identical name of referencing / referenced column in `child_table`/`parent_table`.
#'
#' @family table surgery functions
#'
#' @return A wide table produced by joining the two given tables.
#'
#' @section Life cycle:
#' These functions are marked "experimental" because they seem more useful
#' when applied to a table in a dm object.
#' Changing the interface later seems harmless because these functions are
#' most likely used interactively.
#'
#' @name reunite_parent_child
#'
#' @export
#' @examples
#' decomposed_table <- decompose_table(mtcars, new_id, am, gear, carb)
#' ct <- decomposed_table$child_table
#' pt <- decomposed_table$parent_table
#'
#' reunite_parent_child(ct, pt, new_id)
#' reunite_parent_child_from_list(decomposed_table, new_id)
reunite_parent_child <- function(child_table, parent_table, id_column) {
  id_col_q <- ensym(id_column)

  id_col_chr <-
    as_name(id_col_q)

  child_table %>%
    left_join(parent_table, by = id_col_chr) %>%
    select(-!!id_col_q)
}

#' Merge two tables that are linked by a foreign key relation
#'
#' @description `reunite_parent_child_from_list()`: After joining the two tables
#' by the column `id_column`, `id_column` is removed.
#'
#' This function is almost exactly the inverse of `decompose_table()` (the order
#' of the columns is not retained, and the original row names are lost).
#'
#' @param list_of_parent_child_tables Cf arguments `child_table` and `parent_table` from
#'   `reunite_parent_child()`, but both in a named list (as created by `decompose_table()`).
#'
#' @rdname reunite_parent_child
#' @export
reunite_parent_child_from_list <- function(list_of_parent_child_tables, id_column) {
  id_col_q <- ensym(id_column)

  id_col_chr <-
    as_name(id_col_q)

  child_table <- list_of_parent_child_tables[["child_table"]]

  parent_table <- list_of_parent_child_tables[["parent_table"]]

  child_table %>%
    left_join(parent_table, by = id_col_chr) %>%
    select(-!!id_col_q)
}

test_that("decompose_table() decomposes tables nicely on chosen source", {
  skip_if_src("maria")

  out <- decompose_table(data_ts(), aef_id, a, e, f)

  expect_equivalent_tbl(
    out$parent_table,
    list_of_data_ts_parent_and_child()$parent_table,
    # https://github.com/tidyverse/dbplyr/pull/496/files#r523986061
    across(where(is.integer), as.numeric)
  )
  expect_equivalent_tbl(
    out$child_table,
    list_of_data_ts_parent_and_child()$child_table,
    # https://github.com/tidyverse/dbplyr/pull/496/files#r523986061
    across(where(is.integer), as.numeric)
  )
})

test_that("decompose_table() decomposes everything() to the original", {
  out <- decompose_table(data_ts(), abcdef_id, everything())$parent_table
  expect_equivalent_tbl(
    out %>% select(-abcdef_id),
    data_ts()
  )
})

test_that("decomposition works with {tidyselect}", {
  pt_iris <-
    iris %>%
    select(starts_with("Sepal")) %>%
    distinct() %>%
    mutate(Sepal_id = row_number(Sepal.Length)) %>%
    select(Sepal_id, everything())

  ct_iris <-
    left_join(iris, pt_iris, by = c("Sepal.Length", "Sepal.Width")) %>%
    select(-Sepal.Length, -Sepal.Width)

  reference_flower_object <- list(
    child_table = ct_iris,
    parent_table = pt_iris
  )

  out <- decompose_table(iris, Sepal_id, starts_with("Sepal"))
  expect_equivalent_tbl(
    out$parent_table,
    reference_flower_object$parent_table
  )
  expect_equivalent_tbl(
    out$child_table,
    reference_flower_object$child_table
  )
})

test_that("reunite_parent_child() reunites parent and child nicely on chosen source", {
  out <- reunite_parent_child(
    list_of_data_ts_parent_and_child()$child_table,
    list_of_data_ts_parent_and_child()$parent_table,
    aef_id
  )
  ref <-
    left_join(
      list_of_data_ts_parent_and_child()$child_table,
      list_of_data_ts_parent_and_child()$parent_table,
      by = "aef_id"
    ) %>%
    select(-aef_id)

  expect_equivalent_tbl(out, ref)
})

test_that("reunite_parent_child_from_list() reunites parent and child nicely on chosen source", {
  out <- reunite_parent_child_from_list(
    list_of_data_ts_parent_and_child(),
    aef_id
  )
  ref <-
    left_join(
      list_of_data_ts_parent_and_child()$child_table,
      list_of_data_ts_parent_and_child()$parent_table,
      by = "aef_id"
    ) %>%
    select(-aef_id)

  expect_equivalent_tbl(out, ref)
})


test_that("table surgery functions fail in the expected ways?", {
  expect_error(
    decompose_table(data_ts(), aex_id, a, e, x),
    class = if_pkg_version("vctrs", "0.2.99.9004", "vctrs_error_subscript_oob")
  )

  expect_dm_error(
    decompose_table(data_ts(), a, a, e, x),
    class = "dupl_new_id_col_name"
  )
})

test_that("decompose_table() doesn't create NAs in key column", {
  data <- tibble(a = c(1, 2, NA, 3), b = letters[1:4])
  decomposed <- decompose_table(data, id, a)
  expect_true(!anyNA(decomposed$parent_table$id))
})

# do not add 'btn-default' class, so we can customize
actionButton2 <- function(inputId, label, icon = NULL, width = NULL, ...) {
  value <- shiny::restoreInput(id = inputId, default = NULL)
  shiny::tags$button(
    id = inputId,
    style = htmltools::css(width = htmltools::validateCssUnit(width)),
    type = "button",
    class = "btn action-button",
    `data-val` = value,
    list(("shiny" %:::% "validateIcon")(icon), label),
    ...
  )
}


# FIXME no need for custom HTML (I guess)
html_logo <- function() {
  shiny::tags$span(
    class = "logo",
    shiny::span(style = "font-size: 100%; font-weight: bold; letter-spacing: .1rem;", "DM"),
    shiny::span(style = "font-size: 70%; letter-spacing: .1rem;", "R DATA MODELS")
  )
}

html_header <- function() {
  shiny::tags$header(
    class = "main-header",
    html_logo(),
    shiny::tags$nav(
      class = "navbar navbar-static-top",
      role = "navigation",
      shiny::tags$span(
        style = "display:none;",
        shiny::tags$i(class = "fa fa-bars")
      ),
      shiny::tags$a(
        href = "#",
        class = "sidebar-toggle",
        `data-toggle` = "offcanvas",
        role = "button",
        shiny::tags$span(class = "sr-only", "Toggle navigation")
      ),
      shiny::tags$div(
        class = "navbar-custom-menu",
        shiny::tags$ul(
          class = "nav navbar-nav",
          shiny::tags$div(
            class = "navbar-custom-menu",
            shiny::tags$ul(
              class = "nav navbar-nav",
              shiny::tags$li(actionButton2(
                "i_undo",
                label = NULL,
                icon = shiny::icon(verify_fa = FALSE, "undo"),
                class = "btn btn-default",
                style = "margin-top: 9px; margin-right: 15px;"
              )),
              shiny::tags$li(actionButton2(
                "i_redo",
                label = NULL,
                icon = shiny::icon(verify_fa = FALSE, "redo"),
                class = "btn btn-default",
                style = "margin-top: 9px; margin-right: 15px;"
              )),
              shiny::tags$li(shiny::tags$button(
                id = "i_return",
                style = "margin-top: 9px; margin-right: 15px;",
                href = "#",
                type = "button",
                class = "btn btn-primary btn action-button btn-navbar",
                shiny::icon(verify_fa = FALSE, "fas fa-sign-out-alt", ),
                "To Console"
              ))
            )
          )
        )
      )
    )
  )
}


gui_ui <- function(ns, select_tables = TRUE) {
  shinydashboard::dashboardPage(
    skin = "black",
    html_header(),
    shinydashboard::dashboardSidebar(
      disable = TRUE
    ),
    shinydashboard::dashboardBody(
      shiny::tags$head(
        shiny::tags$link(href = "docs.css", rel = "stylesheet")
      ),
      shiny::fluidRow(
        shiny::column(
          6,
          style = "padding: 0px;",
          shinydashboard::box(
            status = "primary",
            width = 12,
            # FIXME: Incorporate selectize into dmSVG(), manage state internally
            if (select_tables) {
              shiny::tagList(
                shiny::selectizeInput(
                  "i_selected_tables",
                  label = NULL,
                  multiple = TRUE,
                  choices = NULL,
                  options = list(
                    placeholder = "Select or search one or several tables",
                    plugins = list("remove_button")
                  ),
                  width = "100%"
                ),
                dmSVGOutput("o_svg", height = "570px")
              )
            } else {
              dmSVGOutput("o_svg", height = "626px")
            },
            footer = shiny::tagList(
              actionButton2(
                "i_zoom_to_selected",
                "Zoom",
                icon = shiny::icon(verify_fa = FALSE, "binoculars"),
                class = "btn btn-info btn-app"
              ),
              actionButton2(
                "i_rm_tbl",
                "Delete table",
                icon = shiny::icon(verify_fa = FALSE, "trash-alt"),
                class = "btn btn-info btn-app"
              ),
              actionButton2(
                "i_select_tbl",
                "Select table",
                icon = shiny::icon(verify_fa = FALSE, "table"),
                class = "btn btn-info btn-app"
              ),
              actionButton2(
                "i_rm_fk",
                "Remove foreign key",
                icon = shiny::icon(verify_fa = FALSE, "unlink"),
                class = "btn btn-info btn-app"
              ),
              actionButton2(
                "i_disentangle",
                "Remove cycles",
                icon = shiny::icon(verify_fa = FALSE, "circle-notch"),
                class = "btn btn-info btn-app"
              ),
              shiny::tags$style(htmltools::HTML(
                "
                #colorpicker-user>.shiny-input-container {
                  display: inline-block !important;
                  float: right !important;
                  width: 140px !important;
                  margin-right: 10px !important;
                }
                .colourpicker-panel {
                  left: -50px;
                  top: -130px;
                }
              "
              )),
              htmltools::span(
                id = "colorpicker-user",
                colourpicker::colourInput(
                  "i_choose_color",
                  label = "Choose color",
                  value = "white",
                  showColour = "text",
                  returnName = TRUE,
                  palette = "limited",
                  closeOnClick = TRUE
                )
              )
            )
          )
        ),
        shinydashboard::box(
          title = "DM Call",
          # status = "default",
          width = 6,
          shinyAce::aceEditor(
            outputId = "i_call",
            fontSize = 18,
            mode = "r",
            readOnly = TRUE,
            theme = "tomorrow",
            height = "224px"
          )
        ),
        shiny::conditionalPanel(
          condition = "output.is_one_table_selected",
          shinydashboard::box(
            title = "Edit Table",
            status = "success",
            width = 6,
            reactable::reactableOutput("o_column"),
            footer = shiny::conditionalPanel(
              condition = "output.is_one_table_selected",
              shiny::tagList(
                actionButton2(
                  "i_add_pk",
                  "Add primary key",
                  icon = shiny::icon(verify_fa = FALSE, "key"),
                  class = "btn btn-info btn-app"
                ),
                # actionButton("i_rename_col", "Rename columns"),
                actionButton2(
                  "i_rm_col",
                  "Delete column",
                  icon = shiny::icon(verify_fa = FALSE, "trash"),
                  class = "btn btn-app"
                )
              )
            )
          )
        ),
        shiny::conditionalPanel(
          condition = "output.are_two_tables_selected",
          shinydashboard::box(
            title = "Manage relationships between two tables",
            status = "warning",
            width = 6,
            shiny::fluidRow(
              shiny::column(6, reactable::reactableOutput("o_column_1")),
              shiny::column(6, reactable::reactableOutput("o_column_2")),
            ),
            footer = shiny::tagList(
              # p("Select two tables to manage the relationship between them."),
              actionButton2(
                "i_add_fk",
                "Add foreign key",
                icon = shiny::icon(verify_fa = FALSE, "link"),
                class = "btn btn-info btn-app"
              ),
              actionButton2(
                "i_switch_tbls",
                "Switch tables",
                icon = shiny::icon(verify_fa = FALSE, "exchange-alt"),
                class = "btn btn-info btn-app"
              )
            )
          )
        )
      )
    )
  )
}

gui_server <- function(input, output, session) {
  g_table_names <- character()

  # debug <- isTRUE(getOption("shiny.fullstacktrace"))
  # if (!debug) {
  #   print <- function(x, ...) invisible(x)
  # }

  # data and filters -------------------------------------------------------------

  r_ops_stack <- shiny::reactiveVal(
    ops_stack(new_mw_cg(getOption("dm.gui_init_dm")))
  )

  r_ops <- shiny::reactive({
    ops_stack_current(r_ops_stack())
  })

  r_svg_elements <- shiny::reactiveVal()
  shiny::observe({
    r_svg_elements(input$svg_elements)
  })

  r_dm <- shiny::reactive({
    r_ops()$dm
  })

  # Not every update in dm needs to reset the svg values
  # trigger by: r_reset_svg(r_reset_svg() + 1)
  r_reset_svg <- shiny::reactiveVal(0)
  shiny::observeEvent(r_reset_svg(), {
    # reset input reactive value
    # FIXME: Move to htmlwidget
    session$sendCustomMessage(type = "reset_value", message = "svg_elements")
    r_svg_elements(NULL)
  })

  r_call <- shiny::reactive({
    paste(utils::capture.output(print.dm_cg_code_block(r_ops()$cg)), collapse = "\n")
    # format(r_ops()$cg)
  })

  shiny::observe({
    shinyAce::updateAceEditor(session, "i_call", r_call())
  })

  # edge selector -----------------------------------------------------------

  r_edge_names <- shiny::reactive({
    dm <- shiny::req(r_dm())
    svg_elements <- r_svg_elements()
    as.character(svg_elements$edge)
  })

  r_edge_table <- shiny::reactive({
    dm <- shiny::req(r_dm())
    all_fks <- dm_get_all_fks_impl(r_dm(), id = TRUE) %>%
      filter(id %in% r_edge_names()) %>%
      select(-on_delete, -id)
  })

  # table selector -------------------------------------------------------------

  # a) from SVG
  r_table_names_svg <- shiny::reactive({
    dm <- shiny::req(r_dm())
    svg_elements <- r_svg_elements()
    as.character(svg_elements$nodes)
  })

  # Update by:
  # - svg
  # - switch
  # - select works through svg
  r_table_names_ordered <- shiny::reactiveVal()

  shiny::observe({
    table_names <- r_table_names_svg()
    table_names <- c(intersect(g_table_names, table_names), setdiff(table_names, g_table_names))
    g_table_names <<- table_names
    r_table_names_ordered(table_names)
  })
  shiny::observeEvent(input$i_switch_tbls, {
    table_names <- rev(r_table_names_ordered())
    g_table_names <<- table_names
    r_table_names_ordered(table_names)
  })

  # Is there a better solution for this? Seems like a standard problem:
  # Two input widgets that update each other.

  g_table_names_select <- character()
  g_update_time <- Sys.time()

  r_table_names_select <- shiny::reactive({
    selected_tables <- input$i_selected_tables
    if (is.null(selected_tables)) {
      selected_tables <- character(0)
      session$sendCustomMessage(type = "reset_value", message = "svg_elements")
    }
    g_table_names_select <<- selected_tables
    g_update_time <<- Sys.time()

    selected_tables
  })

  # On start, after new dm, remove choices
  shiny::observe({
    choices <- names(r_dm())
    shiny::updateSelectizeInput(
      session,
      "i_selected_tables",
      choices = choices,
      selected = shiny::isolate(r_table_names_ordered())
    )
  })

  shiny::observeEvent(
    input$i_selected_tables,
    {
      selected_tables <- input$i_selected_tables
      r_table_names_ordered(selected_tables)
    },
    ignoreInit = TRUE
  )

  # Update select only if changed by svg or switch
  shiny::observe({
    choices <- names(shiny::isolate(r_dm()))
    table_names <- r_table_names_ordered()

    # do not self-update: changes in svg or should only trigger an update of
    # select if it was NOT induced by select itself.
    if (identical(table_names, g_table_names_select)) {
      return(NULL)
    }

    # To avoid infite loops
    # Not perfect: One could update and then click on SVG
    if (Sys.time() - g_update_time < 0.3) {
      return(NULL)
    }

    shiny::updateSelectizeInput(
      session,
      "i_selected_tables",
      choices = choices,
      selected = table_names
    )
  })

  # view modes -----------------------------------------------------------------

  # so we can use shiny::conditionalPanel() and avoid renderUI for now
  output$are_tables_selected <- shiny::reactive({
    length(r_table_names_ordered()) > 0
  })
  shiny::outputOptions(output, "are_tables_selected", suspendWhenHidden = FALSE)

  output$are_two_tables_selected <- shiny::reactive({
    length(r_table_names_ordered()) == 2
  })
  shiny::outputOptions(output, "are_two_tables_selected", suspendWhenHidden = FALSE)

  output$is_one_table_selected <- shiny::reactive({
    length(r_table_names_ordered()) == 1
  })
  shiny::outputOptions(output, "is_one_table_selected", suspendWhenHidden = FALSE)

  output$is_one_or_two_table_selected <- shiny::reactive({
    length(r_table_names_ordered()) %in% 1:2
  })
  shiny::outputOptions(output, "is_one_or_two_table_selected", suspendWhenHidden = FALSE)

  # table mode (1 table selected) ----------------------------------------------

  r_data_column <- shiny::reactiveVal()

  shiny::observe({
    table_name <- r_table_names_ordered()
    if (!(length(table_name) %in% 1:2)) {
      r_data_column(NULL)
    } else {
      dm <- r_dm()
      data <- data_column(dm, table_name[[1]])
      r_data_column(data)
    }
  })

  output$o_column <- reactable::renderReactable({
    table_name <- shiny::req(r_table_names_ordered())
    data <- shiny::req(r_data_column())
    reactable_column(data, table_name[1])
  })

  # FIXME: Clean up
  r_column_names <- shiny::reactive({
    table_name <- r_table_names_ordered()
    if (length(table_name) == 1) {
      n <- reactable::getReactableState("o_column", "selected")
      shiny::req(r_data_column())[n, ] %>% dplyr::pull(name)
    } else if (length(table_name) == 2) {
      n <- reactable::getReactableState("o_column_1", "selected")
      shiny::req(r_data_column())[n, ] %>% dplyr::pull(name)
    }
  })

  # so we can use shiny::conditionalPanel() and avoid renderUI for now
  output$are_columns_selected <- shiny::reactive({
    length(r_column_names()) > 0
  })
  shiny::outputOptions(output, "are_columns_selected", suspendWhenHidden = FALSE)

  # 2 table mode (2 tables selected) -------------------------------------------

  r_data_column_2 <- shiny::reactiveVal()

  shiny::observe({
    table_name <- r_table_names_ordered()
    if (length(table_name) != 2) {
      r_data_column_2(NULL)
    } else {
      dm <- r_dm()
      data <- data_column(dm, table_name[[2]])
      r_data_column_2(data)
    }
  })

  output$o_column_2 <- reactable::renderReactable({
    data <- shiny::req(r_data_column_2())
    table_name <- shiny::req(r_table_names_ordered())
    reactable_column(data, table_name = table_name[2])
  })

  # FIXME: single view is most likely different from dual view
  output$o_column_1 <- reactable::renderReactable({
    table_name <- shiny::req(r_table_names_ordered())
    data <- shiny::req(r_data_column())
    reactable_column(data, table_name[1])
  })

  r_column_2_names <- shiny::reactive({
    n <- reactable::getReactableState("o_column_2", "selected")
    shiny::req(r_data_column_2())[n, ] %>% dplyr::pull(name)
  })

  # so we can use shiny::conditionalPanel() and avoid renderUI for now
  output$are_columns_2_selected <- shiny::reactive({
    length(r_column_2_names()) > 0
  })
  shiny::outputOptions(output, "are_columns_2_selected", suspendWhenHidden = FALSE)

  # calls to middleware --------------------------------------------------------

  # update a reactive r_obs in each step

  shiny::observeEvent(input$i_rm_tbl, {
    ans <- mw_cg_run(
      r_ops(),
      "dm_select_tbl",
      table_names = r_table_names_ordered(),
      rm = TRUE,
      abort_function = ~ showNotification(.x, type = "error")
    )
    r_reset_svg(r_reset_svg() + 1)
    r_ops_stack(ops_stack_append(r_ops_stack(), ans))
  })

  shiny::observeEvent(input$i_select_tbl, {
    ans <- mw_cg_run(
      r_ops(),
      "dm_select_tbl",
      table_names = r_table_names_ordered(),
      abort_function = ~ showNotification(.x, type = "error")
    )
    r_reset_svg(r_reset_svg() + 1)
    r_ops_stack(ops_stack_append(r_ops_stack(), ans))
  })

  shiny::observeEvent(input$i_add_pk, {
    ans <- mw_cg_run(
      r_ops(),
      "dm_add_pk",
      table_names = r_table_names_ordered(),
      column_names = r_column_names(),
      abort_function = ~ showNotification(.x, type = "error")
    )
    r_ops_stack(ops_stack_append(r_ops_stack(), ans))
  })

  shiny::observeEvent(input$i_add_fk, {
    ans <- mw_cg_run(
      r_ops(),
      "dm_add_fk",
      table_names = r_table_names_ordered(),
      column_names = r_column_names(),
      other_column_names = r_column_2_names(),
      abort_function = ~ showNotification(.x, type = "error")
    )
    r_ops_stack(ops_stack_append(r_ops_stack(), ans))
  })

  shiny::observeEvent(input$i_rm_fk, {
    ans <- mw_cg_run(
      r_ops(),
      "dm_rm_fk",
      edge_table = r_edge_table(),
      abort_function = ~ showNotification(.x, type = "error")
    )

    r_ops_stack(ops_stack_append(r_ops_stack(), ans))
  })

  shiny::observeEvent(input$i_disentangle, {
    ans <- mw_cg_run(
      r_ops(),
      "dm_disentangle",
      table_names = r_table_names_ordered(),
      abort_function = ~ showNotification(.x, type = "error")
    )

    r_ops_stack(ops_stack_append(r_ops_stack(), ans))
  })

  shiny::observeEvent(
    input$i_choose_color,
    {
      ans <- mw_cg_run(
        r_ops(),
        "dm_set_colors",
        table_names = r_table_names_ordered(),
        color_name = input$i_choose_color,
        abort_function = ~ showNotification(.x, type = "error")
      )

      r_ops_stack(ops_stack_append(r_ops_stack(), ans))
    },
    ignoreInit = TRUE
  )

  shiny::observeEvent(input$i_rm_col, {
    ans <- mw_cg_run(
      r_ops(),
      "dm_select",
      table_names = r_table_names_ordered(),
      column_names = r_column_names(),
      rm = TRUE,
      abort_function = ~ showNotification(.x, type = "error")
    )

    r_ops_stack(ops_stack_append(r_ops_stack(), ans))
  })

  shiny::observeEvent(input$i_return, {
    shiny::stopApp(r_call())
  })

  shiny::observeEvent(input$i_undo, {
    # ops_stack_can_undo(r_ops_stack())
    r_ops_stack(ops_stack_undo(r_ops_stack()))
  })

  shiny::observeEvent(input$i_redo, {
    # ops_stack_can_redo(r_ops_stack())
    r_ops_stack(ops_stack_redo(r_ops_stack()))
  })

  # output ---------------------------------------------------------------------

  r_node_to_zoom <- shiny::reactiveVal(character())
  shiny::observeEvent(input$i_zoom_to_selected, {
    r_node_to_zoom(r_table_names_ordered())
  })

  output$o_svg <- renderDmSVG({
    dm <- r_dm()
    dmSVG(
      dm,
      viewBox = FALSE,
      node_to_zoom = r_node_to_zoom(),
      nodes_to_select = I(r_table_names_select())
    )
  })
}

# Cf. https://github.com/cynkra/dm/issues/144 (Review error messages)

# error call infrastructure -----------------------------------------------

# Call in exported functions to mark them as the error origin.
# Similar to dplyr:::dplyr_local_error_call().
dm_local_error_call <- function(call = frame, frame = caller_env()) {
  frame$.__dm_error_call__. <- call
  invisible(NULL)
}

# Call in internal functions to retrieve the error call.
# Similar to dplyr:::dplyr_error_call().
dm_error_call <- function(call) {
  if (is_missing(call)) {
    call <- caller_env()
  }

  # Walk up the call stack to find the exported function that set
  # .__dm_error_call__. via dm_local_error_call().
  frame <- call
  n <- sys.nframe()
  for (i in seq_len(n)) {
    caller <- eval_bare(quote(base::parent.frame()), frame)
    if (
      identical(caller, frame) ||
        identical(caller, global_env()) ||
        identical(caller, base_env()) ||
        identical(caller, empty_env())
    ) {
      break
    }
    caller_call <- caller[[".__dm_error_call__."]]
    if (!is_null(caller_call)) {
      call <- caller_call
      break
    }
    frame <- caller
  }

  call
}

# error class generator ---------------------------------------------------

dm_error <- function(x) {
  paste0("dm_error_", x)
}

dm_error_full <- function(x) {
  c(dm_error(x), "dm_error")
}

dm_abort <- function(bullets, class) {
  cli::cli_abort(bullets, class = dm_error_full(class), call = dm_error_call())
}

dm_warning <- function(x) {
  paste0("dm_warning_", x)
}

dm_warning_full <- function(x) {
  c(dm_warning(x), "dm_warning")
}

dm_warn <- function(bullets, class) {
  cli::cli_warn(bullets, class = dm_warning_full(class))
}

# abort and text for key-helper functions ---------------------------------

abort_not_unique_key <- function(table_name, column_names) {
  cli::cli_abort(
    "({commas(tick(column_names))}) not a unique key of {.code {table_name}}.",
    class = dm_error_full("not_unique_key"),
    call = dm_error_call()
  )
}

# error: is not subset of -------------------------------------------------

abort_not_subset_of <- function(table_name_1, colname_1, table_name_2, colname_2) {
  plural <- s_if_plural(colname_1)
  cli::cli_abort(
    "Column{plural['n']} ({commas(tick(colname_1))}) of table {.field {table_name_1}} contain{plural['v']} values (see examples above) that are not present in column{plural['n']} ({commas(tick(colname_2))}) of table {.field {table_name_2}}.",
    class = dm_error_full("not_subset_of"),
    call = dm_error_call()
  )
}

# error sets not equal ----------------------------------------------------

abort_sets_not_equal <- function(error_msgs) {
  cli::cli_abort(error_msgs, class = dm_error_full("sets_not_equal"), call = dm_error_call())
}

# cardinality check errors ------------------------------------------------

abort_not_bijective <- function(child_table_name, fk_col_name) {
  plural <- s_if_plural(fk_col_name)
  cli::cli_abort(
    "1..1 cardinality (bijectivity) is not given: Column{plural['n']} ({commas(tick(fk_col_name))}) in table {.field {child_table_name}} contain{plural['v']} duplicate values.",
    class = dm_error_full("not_bijective"),
    call = dm_error_call()
  )
}

abort_not_injective <- function(child_table_name, fk_col_name) {
  plural <- s_if_plural(fk_col_name)
  cli::cli_abort(
    "0..1 cardinality (injectivity from child table to parent table) is not given: Column{plural['n']} ({commas(tick(fk_col_name))}) in table {.field {child_table_name}} contain{plural['v']} duplicate values.",
    class = dm_error_full("not_injective"),
    call = dm_error_call()
  )
}

# errors in fk handling --------------------------------------------------

abort_ref_tbl_has_no_pk <- function(ref_table_name) {
  cli::cli_abort(
    "ref_table {.field {ref_table_name}} needs a primary key first. Use {.fn dm_enum_pk_candidates} to find appropriate columns and {.fn dm_add_pk} to define a primary key.",
    class = dm_error_full("ref_tbl_has_no_pk"),
    call = dm_error_call()
  )
}

# error helpers for draw_dm -----------------------------------------------

abort_last_col_missing <- function() {
  cli::cli_abort(
    "The last color can't be missing.",
    class = dm_error_full("last_col_missing"),
    call = dm_error_call()
  )
}

# errors in graph-functions -----------------------------------------------

abort_no_cycles <- function(g) {
  shortest_cycle <-
    graph_girth(g) %>%
    pluck("circle") %>%
    names()
  # add the first element after the last element, so it's more clear that it's a cycle
  shortest_cycle <- paste(c(shortest_cycle, shortest_cycle[1]), collapse = " -> ")
  # FIXME: extract, also identify parallel edges as circles
  cli::cli_abort(
    c(
      "Cycles in the relationship graph not yet supported.",
      i = "Shortest cycle: {shortest_cycle}"
    ),
    class = dm_error_full("no_cycles"),
    call = dm_error_call()
  )
}

# error in dm_flatten_to_tbl() ----------------------------------------------

abort_tables_not_reachable_from_start <- function(table_arg = ".start") {
  cli::cli_abort(
    "All selected tables must be reachable from {.arg {table_arg}}.",
    class = dm_error_full("tables_not_reachable_from_start"),
    call = dm_error_call()
  )
}

# errors in table surgery -------------------------------------------------

abort_dupl_new_id_col_name <- function(table_name) {
  cli::cli_abort(
    "{.arg new_id_column} can't have an identical name as one of the columns of {.field {table_name}}.",
    class = dm_error_full("dupl_new_id_col_name"),
    call = dm_error_call()
  )
}

abort_no_overwrite <- function() {
  cli::cli_abort(
    "The {.arg overwrite} argument is not supported.",
    class = dm_error_full("no_overwrite"),
    call = dm_error_call()
  )
}

abort_update_not_supported <- function() {
  cli::cli_abort(
    "Updating {.cls dm} objects not supported.",
    class = dm_error_full("update_not_supported"),
    call = dm_error_call()
  )
}

# errors when filters are set but they shouldn't be ------------------------------

abort_only_possible_wo_filters <- function(fun_name) {
  cli::cli_abort(
    "Not supported on a {.cls dm} with filter conditions. Consider using {.fn dm_apply_filters} first.",
    class = dm_error_full("only_possible_wo_filters"),
    call = dm_error_call()
  )
}

# no foreign key relation -------------------------------------------------

abort_tables_not_neighbors <- function(t1_name, t2_name) {
  cli::cli_abort(
    "Tables {.field {t1_name}} and {.field {t2_name}} are not directly linked by a foreign key relation.",
    class = dm_error_full("tables_not_neighbors"),
    call = dm_error_call()
  )
}

# `dm_flatten_to_tbl()` and `dm_join_to_tbl()` only supported for parents

abort_only_parents <- function(func, table_arg, recursive_arg) {
  cli::cli_abort(
    "All join partners of table {.arg {table_arg}} must be its direct neighbors. Use {.code {recursive_arg} = TRUE} for recursive flattening.",
    class = dm_error_full("only_parents"),
    call = dm_error_call()
  )
}

# not all tables have the same src ----------------------------------------

abort_not_same_src <- function(dm_bind = FALSE) {
  if (!dm_bind) {
    cli::cli_abort(
      "Not all tables in the object share the same {.arg src}.",
      class = dm_error_full("not_same_src"),
      call = dm_error_call()
    )
  } else {
    cli::cli_abort(
      "All {.cls dm} objects need to share the same {.arg src}.",
      class = dm_error_full("not_same_src"),
      call = dm_error_call()
    )
  }
}

# Something other than tables are put in a `dm` ------------------

abort_what_a_weird_object <- function(class) {
  cli::cli_abort(
    "Don't know how to determine table source for object of class {.cls {class}}.",
    class = dm_error_full("what_a_weird_object"),
    call = dm_error_call()
  )
}

abort_squash_limited <- function(func = "dm_flatten_to_tbl", recursive_arg = ".recursive") {
  cli::cli_abort(
    "Recursive flattening only supports {.fn left_join}, {.fn inner_join}, or {.fn full_join}.",
    class = dm_error_full("squash_limited"),
    call = dm_error_call()
  )
}

abort_apply_filters_first <- function(join_name) {
  cli::cli_abort(
    "{.fn dm_..._to_tbl} with join using {.fn {join_name}} generally wouldn't produce the correct result when filters are set. Please consider calling {.fn dm_apply_filters} first.",
    class = dm_error_txt_apply_filters_first(join_name),
    call = dm_error_call()
  )
}

dm_error_txt_apply_filters_first <- function(join_name) {
  dm_error(c(paste0("apply_filters_first_", join_name), "apply_filters_first"))
}

abort_no_flatten_with_nest_join <- function(func = "dm_..._to_tbl") {
  cli::cli_abort(
    "{.code join = nest_join} is not supported. Consider {.code join = left_join}.",
    class = dm_error_full("no_flatten_with_nest_join"),
    call = dm_error_call()
  )
}


# object is not a `dm` (but should be one) --------------------------------
abort_is_not_dm <- function(obj_class) {
  cli::cli_abort(
    "Required class {.cls dm} but instead is {.cls {obj_class}}.",
    class = dm_error_full("is_not_dm"),
    call = dm_error_call()
  )
}


# local `dm` has no con ---------------------------------------------------
abort_con_only_for_dbi <- function() {
  cli::cli_abort(
    "A local {.cls dm} doesn't have a DB connection.",
    class = dm_error_full("con_only_for_dbi"),
    call = dm_error_call()
  )
}

# when zoomed and it shouldn't be ------------------------------

abort_only_possible_wo_zoom <- function(fun_name) {
  cli::cli_abort(
    "Not supported on a {.cls dm_zoomed}. Consider using one of {.fn dm_update_zoomed}, {.fn dm_insert_zoomed} or {.fn dm_discard_zoomed} first.",
    class = dm_error_full("only_possible_wo_zoom"),
    call = dm_error_call()
  )
}

# when not zoomed and it should be ------------------------------

abort_only_possible_w_zoom <- function(fun_name) {
  cli::cli_abort(
    "Not supported on an unzoomed {.cls dm}. Consider using {.fn dm_zoom_to} first.",
    class = dm_error_full("only_possible_w_zoom"),
    call = dm_error_call()
  )
}

# errors for `copy_to.dm()` ----------------------------------------------

abort_only_data_frames_supported <- function() {
  cli::cli_abort(
    "Only class {.cls data.frame} is supported for argument {.arg df}.",
    class = dm_error_full("only_data_frames_supported"),
    call = dm_error_call()
  )
}

abort_one_name_for_copy_to <- function(name) {
  cli::cli_abort(
    "Argument {.arg name} must have length 1, not length {length(name)}.",
    class = dm_error_full("one_name_for_copy_to"),
    call = dm_error_call()
  )
}

# new table name needs to be unique ---------------------------------------

abort_need_unique_names <- function(duplicate_names) {
  dupl <- unique(duplicate_names)
  cli::cli_abort(
    "{cli::qty(length(dupl))}Each new table needs to have a unique name. Duplicate new name{?s}: {.field {dupl}}.",
    class = dm_error_full("need_unique_names"),
    call = dm_error_call()
  )
}

# lost track of by-column (FK-relation) -----------------------------------

abort_fk_not_tracked <- function(x_orig_name, y_name) {
  cli::cli_abort(
    "The foreign key that existed between the originally zoomed table {.field {x_orig_name}} and {.field {y_name}} got lost in transformations. Please explicitly provide the {.arg by} argument.",
    class = dm_error_full("fk_not_tracked"),
    call = dm_error_call()
  )
}

# lost track of PK-column(s) -----------------------------------

abort_pk_not_tracked <- function(orig_table, orig_pk) {
  cli::cli_abort(
    "The primary key column(s) {commas(tick(orig_pk))} of the originally zoomed table {.field {orig_table}} got lost in transformations.",
    class = dm_error_full("pk_not_tracked"),
    call = dm_error_call()
  )
}


# only for local src ------------------------------------------------------

abort_only_for_local_src <- function(src_dm) {
  cli::cli_abort(
    "Only supported for a local {.arg src}, not on a database with {.arg src}-class: {.cls {class(src_dm)}}.",
    class = dm_error_full("only_for_local_src"),
    call = dm_error_call()
  )
}

# Errors for `pull_tbl.dm()` -----------------------------

abort_no_table_provided <- function() {
  cli::cli_abort(
    "Argument {.arg table} is missing.",
    class = dm_error_full("no_table_provided"),
    call = dm_error_call()
  )
}

abort_table_not_zoomed <- function(table_name, zoomed_tables) {
  cli::cli_abort(
    "Table {.code {table_name}} not zoomed, zoomed tables: {.code {zoomed_tables}}.",
    class = dm_error_full("table_not_zoomed"),
    call = dm_error_call()
  )
}

abort_not_pulling_multiple_zoomed <- function() {
  cli::cli_abort(
    "If more than one zoomed table is available, you need to specify argument {.arg table}.",
    class = dm_error_full("not_pulling_multiple_zoomed"),
    call = dm_error_call()
  )
}

abort_cols_not_avail <- function(wrong_col) {
  cli::cli_abort(
    "{cli::qty(length(wrong_col))}The color{?s} {.val {wrong_col}} {?is/are} not available. Call {.fn dm_get_available_colors} for possible color names or use hex color codes.",
    class = dm_error_full("cols_not_avail"),
    call = dm_error_call()
  )
}

abort_only_named_args <- function(fun_name, name_meaning) {
  cli::cli_abort(
    "All {.arg ...} arguments must be named. The names represent {name_meaning}.",
    class = dm_error_full("only_named_args"),
    call = dm_error_call()
  )
}

abort_wrong_syntax_set_cols <- function() {
  cli::cli_abort(
    "You seem to be using outdated syntax for setting colors, type {.code ?dm_set_colors()} for examples.",
    class = dm_error_full("wrong_syntax_set_cols"),
    call = dm_error_call()
  )
}

abort_parameter_not_correct_class <- function(parameter, correct_class, class) {
  cli::cli_abort(
    "Parameter {.arg {parameter}} needs to be of class {.cls {correct_class}} but is of class {.cls {class}}.",
    class = dm_error_full("parameter_not_correct_class"),
    call = dm_error_call()
  )
}

abort_parameter_not_correct_length <- function(parameter, correct_length, parameter_value) {
  cli::cli_abort(
    "Argument {.arg {parameter}} needs to be of length {.val {correct_length}} but is of length {.val {length(parameter_value)}} ({.val {parameter_value}}).",
    class = dm_error_full("parameter_not_correct_length"),
    call = dm_error_call()
  )
}

warn_if_arg_not <- function(
  arg,
  only_on,
  arg_name = deparse(substitute(arg)),
  correct = NULL,
  additional_msg = ""
) {
  if (!identical(arg, correct)) {
    only_on_string <- glue::glue_collapse(only_on, sep = ", ", last = " and ")
    msg <- cli::format_inline(
      "Argument {.arg {arg_name}} ignored: currently only supported for {only_on_string}."
    )
    dm_warn(
      c(msg, if (!is.null(additional_msg) && nzchar(additional_msg)) c(i = additional_msg)),
      class = "arg_not"
    )
  }
  NULL
}

# Errors for schema handling functions ------------------------------------

abort_no_schemas_supported <- function(dbms = NULL, con = NULL) {
  if (!is.null(dbms)) {
    cli::cli_abort(
      "The concept of schemas is not supported for DBMS {.code {dbms}}.",
      class = dm_error_full("no_schemas_supported"),
      call = dm_error_call()
    )
  } else if (!is.null(con)) {
    cli::cli_abort(
      "Currently schemas are not supported for a connection of class {.cls {class(con)}}.",
      class = dm_error_full("no_schemas_supported"),
      call = dm_error_call()
    )
  } else {
    # if local src, `con = NULL`
    cli::cli_abort(
      "Schemas are not available locally.",
      class = dm_error_full("no_schemas_supported"),
      call = dm_error_call()
    )
  }
}

abort_temporary_not_in_schema <- function() {
  cli::cli_abort(
    "If argument {.arg temporary} is {.val {TRUE}}, argument {.arg schema} has to be {.val {I('NULL')}}.",
    class = dm_error_full("temporary_not_in_schema"),
    call = dm_error_call()
  )
}

abort_one_of_schema_table_names <- function() {
  cli::cli_abort(
    "Only one of the arguments {.arg schema} and {.arg table_names} can be different from {.val {I('NULL')}}.",
    class = dm_error_full("one_of_schema_table_names"),
    call = dm_error_call()
  )
}

s_if_plural <- function(vec) {
  if (length(vec) > 1) {
    # n = noun, v = verb
    c(n = "s", v = "")
  } else {
    c(n = "", v = "s")
  }
}

expect_identical_graph <- function(g1, g2) {
  if (inherits(g1, "dm_igraph") && inherits(g2, "dm_igraph")) {
    expect_true(igraph::identical_graphs(g1$igraph, g2$igraph))
  } else {
    expect_setequal(g1$vnames, g2$vnames)
    edges1 <- sort(paste(g1$vnames[g1$from], g1$vnames[g1$to]))
    edges2 <- sort(paste(g2$vnames[g2$from], g2$vnames[g2$to]))
    expect_equal(edges1, edges2)
  }
}

expect_equivalent_dm <- function(
  object,
  expected,
  sort = FALSE,
  ...,
  sort_tables = sort,
  sort_columns = sort,
  sort_keys = sort,
  ignore_on_delete = FALSE,
  ignore_autoincrement = FALSE
) {
  tables1 <- dm_get_tables_impl(object) %>% map(collect)
  tables2 <- dm_get_tables_impl(expected) %>% map(collect)

  expect_equivalent_tbl_lists(
    tables1,
    tables2,
    sort_tables = sort_tables,
    sort_columns = sort_columns
  )

  if (sort_keys) {
    if (ignore_autoincrement) {
      expect_equivalent_tbl(
        dm_get_all_pks_impl(object) %>% select(-autoincrement),
        dm_get_all_pks_impl(expected) %>% select(-autoincrement)
      )
    } else {
      expect_equivalent_tbl(dm_get_all_pks_impl(object), dm_get_all_pks_impl(expected))
    }
    expect_equivalent_tbl(
      dm_get_all_fks_impl(object, ignore_on_delete = ignore_on_delete),
      dm_get_all_fks_impl(expected, ignore_on_delete = ignore_on_delete)
    )
  } else {
    if (ignore_autoincrement) {
      expect_equivalent_tbl(
        dm_get_all_pks_impl(object) %>% select(-autoincrement),
        dm_get_all_pks_impl(expected) %>% select(-autoincrement)
      )
    } else {
      expect_equivalent_tbl(dm_get_all_pks_impl(object), dm_get_all_pks_impl(expected))
    }
    expect_equal(
      dm_get_all_fks_impl(object, ignore_on_delete = ignore_on_delete),
      dm_get_all_fks_impl(expected, ignore_on_delete = ignore_on_delete)
    )
  }
}

expect_equivalent_why <- function(ex1, ex2) {
  if (inherits(my_test_src(), "src_dbi")) {
    ex1 <-
      ex1 %>%
      mutate(why = (why != ""))
    ex2 <-
      ex2 %>%
      mutate(why = (why != ""))
  }

  expect_identical(ex1, ex2)
}

expect_dm_error <- function(expr, class) {
  expect_error(expr, class = dm_error(class))
}

expect_dm_warning <- function(expr, class) {
  expect_warning(out <- expr, class = dm_warning(class))
  out
}

expect_name_repair_message <- function(expr) {
  expect_message(out <- expr)
  out
}

arrange_if_no_list <- function(tbl) {
  if (inherits(tbl, "tbl_dbi")) {
    arrange_all(tbl)
  } else {
    arrange(tbl, across(where(~ !is.list(.))))
  }
}

harmonize_tbl <- function(tbl, ...) {
  tbl %>%
    collect() %>%
    mutate(...) %>%
    arrange_if_no_list()
}

# are two tables identical minus the `src`
expect_equivalent_tbl <- function(
  tbl_1,
  tbl_2,
  ...,
  .label = NULL,
  .expected_label = NULL,
  .sort_columns = FALSE
) {
  if (.sort_columns) {
    tbl_1 <- select(tbl_1, !!!sort(names(tbl_1)))
    tbl_2 <- select(tbl_2, !!!sort(names(tbl_2)))
  }
  tbl_1_lcl <- harmonize_tbl(tbl_1, ...)
  tbl_2_lcl <- harmonize_tbl(tbl_2, ...)
  expect_identical(tbl_1_lcl, tbl_2_lcl, label = .label, expected.label = .expected_label)
}

# are two lists of tables identical minus the `src`
expect_equivalent_tbl_lists <- function(
  object,
  expected,
  sort_tables = FALSE,
  sort_columns = FALSE
) {
  expect_equal(length(object), length(expected))
  if (length(object) == length(expected)) {
    if (sort_tables) {
      object <- object[sort(names(object))]
      expected <- expected[sort(names(expected))]
    }

    expect_identical(names(object), names(expected))

    recipe <- tibble(
      tbl_1 = object,
      tbl_2 = expected,
      .label = paste0("object$", names(object)),
      .expected_label = paste0("expected$", names(expected)),
    )
    pwalk(recipe, expect_equivalent_tbl, .sort_columns = sort_columns)
  }
}

expect_snapshot_diagram <- function(diagram, name) {
  skip_if_not_installed("DiagrammeR")
  skip_if_not_installed("DiagrammeRsvg")

  dir <- withr::local_tempdir()
  path <- file.path(dir, name)

  diagram %>%
    DiagrammeRsvg::export_svg() %>%
    writeLines(path)

  expect_snapshot_file(path, compare = compare_file_text)
}

expect_same <- function(object, ...) {
  others <- enquos(...)
  walk(others, function(.x) expect_identical(!!.x, {{ object }}))
}

test_that("output", {
  expect_snapshot(error = TRUE, {
    abort_key_set_force_false("pk_table")
    abort_not_unique_key("Christmas", "Rudolph")
    abort_not_unique_key("Christmas", c("elves", "Rudolph", "mulled_wine"))
    abort_table_not_in_dm("laziness", "hard_work")
    abort_not_subset_of("playing", "game", "hunting", "game")
    abort_sets_not_equal(c(
      "A problem occurred.",
      "And another, even worse problem, occurred shortly after."
    ))
    # The mentioned reasons for not being bijective and injective are the same, but this is intended:
    # If surjectivity is not given during the test for bijectivity it will fail earlier
    abort_not_bijective("child_table_name", "fk_col_name")
    abort_not_injective("child_table_name", "fk_col_name")
    abort_ref_tbl_has_no_pk("parent_table")
    abort_is_not_fkc()
    abort_rm_fk_col_missing()
    abort_last_col_missing()
    abort_no_cycles(create_graph_from_dm(dm_for_filter_w_cycle()))
    abort_tables_not_reachable_from_start()
    abort_dupl_new_id_col_name("tibbletable")
    abort_no_overwrite()
    abort_pk_not_defined()
    abort_fk_exists("child", c("child_1", "child_2"), "parent")
    abort_first_rm_fks("parent", c("child_1", "child_2"))
    abort_update_not_supported()
    abort_only_possible_wo_filters("find_wisdom")
    abort_tables_not_neighbors("subjects", "king")
    abort_only_parents("dm_flatten", "table", "recursive")
    abort_not_same_src()
    abort_what_a_weird_object("monster")
    abort_not_same_src()
    abort_squash_limited()
    abort_apply_filters_first("join_tightly")
    abort_no_flatten_with_nest_join()
    abort_is_not_dm("blob")
    abort_con_only_for_dbi()
    abort_only_possible_wo_zoom("dm_zoom_to")
    abort_only_possible_w_zoom("dm_update_zoomed")
    abort_learn_keys(error_cnd(message = "some reason"))
    abort_tbl_access("accessdenied")
    abort_need_unique_names("clone")
    abort_fk_not_tracked("hook", "eye")
    abort_dm_invalid("it's ugly.")
    abort_no_table_provided()
    abort_table_not_zoomed("blur", c("focus_1", "focus_2"))
    abort_not_pulling_multiple_zoomed()
    abort_cols_not_avail(c("pink5", "elephant"))
    abort_only_named_args("give_names", "frobnicability")
    abort_wrong_syntax_set_cols()
    abort_pk_not_tracked("house", "house_number")
    abort_only_for_local_src(mtcars)
    abort_parameter_not_correct_class("number", correct_class = "numeric", class = "logical")
    abort_parameter_not_correct_length("length_1_parameter", 1, letters[1:26])
    warn_if_arg_not("NULL", c("MSSQL", "Postgres"), arg_name = "dbms_dependent_arg")
    abort_no_schemas_supported("FantasticDatabaseManagementSystem", "hyperconnection")
    abort_no_schemas_supported(con = 1)
    abort_no_schemas_supported()
    abort_temporary_not_in_schema()
    abort_one_of_schema_table_names()
  })
})

test_that("singular and plural", {
  expect_snapshot(error = TRUE, {
    # abort_not_subset_of: compound keys (plural)
    abort_not_subset_of("t1", c("a", "b"), "t2", c("c", "d"))
    # abort_not_bijective: compound keys (plural)
    abort_not_bijective("child_tbl", c("col_a", "col_b"))
    # abort_not_injective: compound keys (plural)
    abort_not_injective("child_tbl", c("col_a", "col_b"))
    # abort_need_unique_names: plural
    abort_need_unique_names(c("clone_a", "clone_b"))
    # abort_first_rm_fks: singular
    abort_first_rm_fks("parent", "child_1")
    # abort_tbl_access: plural
    abort_tbl_access(c("table_a", "table_b"))
    # abort_cols_not_avail: singular
    abort_cols_not_avail("pink5")
  })
  expect_snapshot({
    # warn_tbl_access: singular and plural
    warn_tbl_access("table_a")
    warn_tbl_access(c("table_a", "table_b"))
  })
})

library(dm)

dm <- dmSVG:::get_dm("AdventureWorks2014")

con <- DBI::dbConnect(RPostgres::Postgres())

DBI::dbExecute(con, "CREATE SCHEMA av2014")

copy_dm_to(con, dm, schema = "av2014", temporary = FALSE)

enumerate_all_paths <- function(dm, start) {
  all_fks <-
    dm %>%
    dm_get_all_fks() %>%
    rename(child_cols = child_fk_cols, parent_cols = parent_key_cols) %>%
    mutate(edge_id = row_number())

  helper_env <- new_environment()
  helper_env$tbl_node <- list()
  helper_env$all_paths <- tibble(
    child_table = character(),
    child_cols = new_keys(),
    parent_table = character(),
    parent_cols = new_keys(),
    new_child_table = character(),
    new_parent_table = character()
  )

  enumerate_all_paths_impl(start, all_fks = all_fks, helper_env = helper_env)

  all_paths <- helper_env$all_paths
  # need to take into account FKs from unconnected components (graph representation)
  fks_from_unconnected <- anti_join(
    all_fks,
    all_paths,
    by = c("child_table", "parent_table")
  ) %>%
    mutate(new_child_table = child_table, new_parent_table = parent_table)
  all_paths %>%
    rename_unique() %>%
    bind_rows(fks_from_unconnected) %>%
    split_to_list()
}

enumerate_all_paths_impl <- function(
  node,
  edge_id = NULL,
  path = set_names(node),
  all_fks,
  helper_env
) {
  if (!is.null(edge_id)) {
    # increase tbl_node[[node]] by 1, return this index in a suffix
    usage_idx <- inc_tbl_node(node, helper_env)
    new_node <- paste0(node, "-", usage_idx)
    # new nodes appended to the front
    path <- c(set_names(node, new_node), path)
    # the first two elements serve as (reverse) lookup for new table names
    node_lookup <- prep_recode(path[1:2])

    add_path_to_all_paths(all_fks, edge_id, node_lookup, helper_env)
  }

  in_edges <-
    all_fks %>%
    filter(parent_table == !!node) %>%
    filter(!(child_table %in% !!path)) %>%
    select(node = child_table, edge_id)

  out_edges <-
    all_fks %>%
    filter(child_table == !!node) %>%
    filter(!(parent_table %in% !!path)) %>%
    select(node = parent_table, edge_id)

  bind_rows(in_edges, out_edges) %>%
    pwalk(enumerate_all_paths_impl, path, all_fks, helper_env)
}

#' @autoglobal
rename_unique <- function(all_paths) {
  node_lookup <-
    bind_rows(
      select(all_paths, new_table = new_child_table, table = child_table),
      select(all_paths, new_table = new_parent_table, table = parent_table)
    ) %>%
    distinct() %>%
    arrange(table, new_table) %>%
    add_count(table) %>%
    mutate(table = if_else(n == 1L, table, new_table)) %>%
    select(new_table, table) %>%
    deframe()

  all_paths %>%
    mutate(
      new_child_table = (!!node_lookup)[new_child_table],
      new_parent_table = (!!node_lookup)[new_parent_table]
    )
}

inc_tbl_node <- function(node, helper_env) {
  tbl_node <- helper_env$tbl_node
  out <- (tbl_node[[node]] %||% 0) + 1
  tbl_node[[node]] <- out
  helper_env$tbl_node <- tbl_node
  out
}

add_path_to_all_paths <- function(all_fks, edge_id, node_lookup, helper_env) {
  all_paths <- helper_env$all_paths
  path_element <-
    all_fks %>%
    filter(edge_id == !!edge_id)

  helper_env$all_paths <- bind_rows(
    all_paths,
    path_element %>%
      mutate(
        new_child_table = (!!node_lookup)[child_table],
        new_parent_table = (!!node_lookup)[parent_table]
      )
  )
}

#' @autoglobal
split_to_list <- function(all_paths) {
  table_mapping <- bind_rows(
    select(all_paths, new_table = new_child_table, table = child_table),
    select(all_paths, new_table = new_parent_table, table = parent_table)
  ) %>%
    filter(new_table != table) %>%
    distinct() %>%
    mutate(new_table = unname(new_table)) %>%
    arrange()

  new_fks <- select(
    all_paths,
    new_child_table,
    child_cols,
    new_parent_table,
    parent_cols,
    on_delete
  )
  list(table_mapping = table_mapping, new_fks = new_fks)
}

test_that("enumerate_all_paths() works", {
  expect_snapshot({
    enumerate_all_paths(dm_for_filter_w_cycle(), "tf_1")
    enumerate_all_paths(dm_for_filter_w_cycle(), "tf_5")
    enumerate_all_paths(entangled_dm(), "a")
    enumerate_all_paths(entangled_dm(), "c")
    enumerate_all_paths(entangled_dm_2(), "a")
    enumerate_all_paths(entangled_dm_2(), "d")
  })
})

#' Mark table for manipulation
#'
#' @description
#' Zooming to a table of a [`dm`] allows for the use of many `dplyr`-verbs directly on this table, while retaining the
#' context of the `dm` object.
#'
#' `dm_zoom_to()` zooms to the given table.
#'
#' `dm_update_zoomed()` overwrites the originally zoomed table with the manipulated table.
#' The filter conditions for the zoomed table are added to the original filter conditions.
#'
#' `dm_insert_zoomed()` adds a new table to the `dm`.
#'
#' `dm_discard_zoomed()` discards the zoomed table and returns the `dm` as it was before zooming.
#'
#' Please refer to `vignette("tech-db-zoom", package = "dm")`
#' for a more detailed introduction.
#'
#' @inheritParams dm_add_pk
#' @inheritParams vctrs::vec_as_names
#'
#' @details
#' Whenever possible, the key relations of the original table are transferred to the resulting table
#' when using `dm_insert_zoomed()` or `dm_update_zoomed()`.
#'
#' Functions from `dplyr` that are supported for a `dm_zoomed`: [group_by()], [summarise()], [mutate()],
#' [transmute()], [filter()], [select()], [rename()] and [ungroup()].
#' You can use these functions just like you would
#' with a normal table.
#'
#' Calling [filter()] on a zoomed `dm` is different from calling [dm_filter()]:
#' only with the latter, the filter expression is added to the list of table filters stored in the dm.
#'
#' Furthermore, different `join()`-variants from \pkg{dplyr} are also supported,
#' e.g. [left_join()] and [semi_join()].
#' (Support for [nest_join()] is planned.)
#' The join-methods for `dm_zoomed` infer the columns to join by from the primary and foreign keys,
#' and have an extra argument `select` that allows choosing the columns of the RHS table.
#'
#' And -- last but not least -- also the \pkg{tidyr}-functions [unite()] and [separate()] are supported for `dm_zoomed`.
#'
#' @rdname dm_zoom_to
#' @aliases zoomed_df
#' @aliases dm_zoomed_df
#'
#' @return For `dm_zoom_to()`: A `dm_zoomed` object.
#'
#' @export
#' @examplesIf rlang::is_installed(c("nycflights13", "DiagrammeR"))
#' flights_zoomed <- dm_zoom_to(dm_nycflights13(), flights)
#'
#' flights_zoomed
#'
#' flights_zoomed_transformed <-
#'   flights_zoomed %>%
#'   mutate(am_pm_dep = ifelse(dep_time < 1200, "am", "pm")) %>%
#'   # `by`-argument of `left_join()` can be explicitly given
#'   # otherwise the key-relation is used
#'   left_join(airports) %>%
#'   select(year:dep_time, am_pm_dep, everything())
#'
#' flights_zoomed_transformed
#'
#' # replace table `flights` with the zoomed table
#' flights_zoomed_transformed %>%
#'   dm_update_zoomed()
#'
#' # insert the zoomed table as a new table
#' flights_zoomed_transformed %>%
#'   dm_insert_zoomed("extended_flights") %>%
#'   dm_draw()
#'
#' # discard the zoomed table
#' flights_zoomed_transformed %>%
#'   dm_discard_zoomed()
dm_zoom_to <- function(dm, table) {
  dm_local_error_call()
  check_not_zoomed(dm)
  # for now only one table can be zoomed on
  zoom <- dm_tbl_name(dm, {{ table }})

  def <- dm_get_def(dm)
  where <- which(def$table == zoom)
  stopifnot(length(where) == 1)

  zoomed_tbl <- def$data[[where]]
  def$zoom[[where]] <- zoomed_tbl
  def$col_tracker_zoom[[where]] <- set_names(colnames(zoomed_tbl))

  dm_from_def(def, zoomed = TRUE)
}

is_zoomed <- function(dm) {
  inherits(dm, c("dm_zoomed", "zoomed_dm"))
}

#' @rdname dm_zoom_to
#' @param new_tbl_name Name of the new table.
#' @inheritParams vctrs::vec_as_names
#'
#' @return For `dm_insert_zoomed()`, `dm_update_zoomed()` and `dm_discard_zoomed()`: A `dm` object.
#'
#' @export
dm_insert_zoomed <- function(dm, new_tbl_name = NULL, repair = "unique", quiet = FALSE) {
  dm_local_error_call()
  check_zoomed(dm)
  if (is_null(enexpr(new_tbl_name))) {
    new_tbl_name_chr <- orig_name_zoomed(dm)
  } else {
    if (is_symbol(enexpr(new_tbl_name))) {
      warning(
        "The argument `new_tbl_name` in `dm_insert_zoomed()` should be of class `character`."
      )
    }
    new_tbl_name_chr <- as_string(enexpr(new_tbl_name))
  }

  zoomed <- dm_get_zoom(dm, c("table", "display", "zoom", "filters", "col_tracker_zoom"))
  old_tbl_name <- zoomed$table
  new_tbl <- zoomed$zoom
  # filters need to be split: old_filters belong to the old table, new filters to the inserted table
  all_filters <- zoomed$filters[[1]]
  old_filters <- all_filters %>% filter(!zoomed)
  new_filters <-
    all_filters %>%
    filter(zoomed) %>%
    mutate(zoomed = FALSE)

  # rename dm in case of name repair
  names_list <- repair_table_names(
    old_names = src_tbls_impl(dm),
    new_names = new_tbl_name_chr,
    repair,
    quiet
  )

  dm_unzoomed <-
    dm %>%
    update_filter(old_tbl_name, list_of(old_filters)) %>%
    dm_clean_zoomed() %>%
    dm_select_tbl_impl(names_list$new_old_names)

  new_tbl_name_chr <- names_list$new_names
  old_tbl_name <- names_list$old_new_names[[old_tbl_name]]

  dm_wo_outgoing_fks <-
    dm_unzoomed %>%
    dm_add_tbl_impl(
      new_tbl,
      new_tbl_name_chr,
      filters = list_of(new_filters),

      # PK: either the same primary key as in the old table, renamed in the new table, or no primary key if none available
      pks = list_of(update_zoomed_pk(dm)),

      # UK: either the same unique key as in the old table, renamed in the new table, or no unique key if none available
      uks = list_of(update_zoomed_uk(dm)),

      # incoming FKs: in the new row, based on the old table;
      # if PK available, foreign key relations can be copied from the old table
      # if PK vanished, the entry will be empty
      fks = list_of(update_zoomed_incoming_fks(dm))
    )

  # outgoing FKs: potentially in several rows, based on the old table;
  # renamed(?) FK columns if they still exist
  new_dm <- dm_wo_outgoing_fks %>%
    dm_insert_zoomed_outgoing_fks(new_tbl_name_chr, old_tbl_name, zoomed$col_tracker_zoom[[1]])
  if (!is.na(zoomed$display)) {
    new_dm %>%
      dm_set_colors(!!!set_names(new_tbl_name_chr, zoomed$display))
  } else {
    new_dm
  }
}

#' @rdname dm_zoom_to
#' @export
dm_update_zoomed <- function(dm) {
  dm_local_error_call()
  check_zoomed(dm)

  def <- dm_get_def(dm)

  where <- which(lengths(def$zoom) != 0)
  table_name <- def$table[[where]]

  orig_colnames <- colnames(def$data[[where]])
  tracked_cols <- def$col_tracker_zoom[[where]]

  # Test if keys need to be updated (TRUE, if at least one column was renamed or lost)
  upd_keys <- !all(orig_colnames %in% tracked_cols) || !all(names(tracked_cols) == tracked_cols)

  upd_filter <- def$filters[[where]]
  upd_filter$zoomed <- FALSE

  new_def <- def
  new_def$data[[where]] <- new_def$zoom[[where]]
  new_def$filters[[where]] <- upd_filter

  if (upd_keys) {
    new_def$pks[[where]] <- update_zoomed_pk(dm)
    new_def$uks[[where]] <- update_zoomed_uk(dm)
    new_def$fks[[where]] <- update_zoomed_incoming_fks(dm)

    tracked_cols <- new_def$col_tracker_zoom[[where]]
    new_def$fks <- as_list_of(map(new_def$fks, update_zoomed_outgoing, table_name, tracked_cols))
  }

  new_def %>%
    clean_zoom() %>%
    dm_from_def()
}

#' @rdname dm_zoom_to
#' @export
#' @autoglobal
dm_discard_zoomed <- function(dm) {
  dm_local_error_call()
  check_dm(dm)
  if (!is_zoomed(dm)) {
    return(dm)
  }

  def <- dm_get_def(dm)

  where <- which(lengths(def$zoom) != 0)
  old_tbl_name <- def$table[[where]]
  upd_filter <-
    def$filters[[where]] %>%
    filter(zoomed == FALSE)

  def$filters[[where]] <- upd_filter

  def %>%
    clean_zoom() %>%
    dm_from_def()
}

dm_clean_zoomed <- function(dm) {
  dm %>%
    dm_get_def() %>%
    clean_zoom() %>%
    dm_from_def()
}

clean_zoom <- function(def) {
  empty <- rep(list(NULL), length(def$data))
  def$zoom <- empty
  def$col_tracker_zoom <- empty
  def
}

update_zoomed_pk <- function(dm) {
  old_tbl_name <- orig_name_zoomed(dm)
  tracked_cols <- col_tracker_zoomed(dm)
  orig_pk <- dm_get_pk_impl(dm, old_tbl_name)

  if (has_length(orig_pk) && all(get_key_cols(orig_pk) %in% tracked_cols)) {
    upd_pk <- new_pk(list(recode2(get_key_cols(orig_pk), tracked_cols)))
  } else {
    upd_pk <- new_pk()
  }

  upd_pk
}

update_zoomed_uk <- function(dm) {
  old_tbl_name <- orig_name_zoomed(dm)
  tracked_cols <- col_tracker_zoomed(dm)
  orig_uk <- dm %>%
    dm_get_def() %>%
    dm_get_all_uks_def_impl(old_tbl_name)
  if (has_length(orig_uk$uk_col) && all(get_key_cols(orig_uk$uk_col) %in% tracked_cols)) {
    upd_uk <- new_uk(map(orig_uk$uk_col, ~ recode2(get_key_cols(.x), tracked_cols)))
  } else {
    upd_uk <- new_uk()
  }

  upd_uk
}

update_zoomed_incoming_fks <- function(dm) {
  old_tbl_name <- orig_name_zoomed(dm)
  tracked_cols <- col_tracker_zoomed(dm)
  def <- dm_get_def(dm)

  orig_idx <- which(def$table == old_tbl_name)
  orig_fk <- def$fks[[orig_idx]]

  orig_fk$ref_column <- map(
    orig_fk$ref_column,
    ~ {
      if (all(.x %in% tracked_cols)) {
        recode2(.x, tracked_cols)
      } else {
        NULL
      }
    }
  )

  orig_fk[lengths(orig_fk$ref_column) > 0, ]
}

update_zoomed_outgoing <- function(fks, tbl_name, tracked_cols) {
  idx <- which(fks$table == tbl_name)
  if (is_empty(idx)) {
    return(fks)
  }

  remove <- which(!map_lgl(fks$column[idx], ~ all(.x %in% tracked_cols)))
  if (has_length(remove)) {
    fks <- fks[-idx[remove], ]
    # Nontrivial shifts in idx may occur, need to recompute
    idx <- which(fks$table == tbl_name)
  }

  fks$column[idx] <- map(fks$column[idx], ~ recode2(.x, tracked_cols))
  fks
}

update_zoomed_fks <- function(dm, old_tbl_name, tracked_cols) {
  dm_get_all_fks_impl(dm) %>%
    filter(child_table == !!old_tbl_name) %>%
    filter(map_lgl(child_fk_cols, ~ all(.x %in% !!tracked_cols))) %>%
    distinct() %>%
    mutate(
      child_fk_cols = new_keys(map(
        child_fk_cols,
        ~ (!!names(tracked_cols))[match(.x, !!tracked_cols, nomatch = 0L)]
      ))
    )
}

dm_insert_zoomed_outgoing_fks <- function(dm, new_tbl_name, old_tbl_name, tracked_cols) {
  new_out_keys <- update_zoomed_fks(dm, old_tbl_name, tracked_cols)

  dm %>%
    dm_add_fk_impl(
      rep_len(new_tbl_name, length(new_out_keys$child_fk_cols)),
      new_out_keys$child_fk_cols,
      new_out_keys$parent_table,
      new_out_keys$parent_key_cols,
      new_out_keys$on_delete
    )
}

col_tracker_zoomed <- function(dm) {
  dm_get_zoom(dm, "col_tracker_zoom")[[1]][[1]]
}

tbl_zoomed <- function(dm, quiet = FALSE) {
  dm_get_zoom(dm, "zoom", quiet = quiet)[[1]][[1]]
}

orig_name_zoomed <- function(dm) {
  dm_get_zoom(dm, "table")[[1]]
}

filters_zoomed <- function(dm) {
  dm_get_zoom(dm, "filters")[[1]][[1]]
}

replace_zoomed_tbl <- function(dm, new_zoomed_tbl, tracked_cols = NULL) {
  table <- orig_name_zoomed(dm)
  def <- dm_get_def(dm)
  where <- which(def$table == table)
  def$zoom[[where]] <- new_zoomed_tbl
  # the tracked columns are only replaced if they changed, otherwise this function is called with default `NULL`
  if (!is_null(tracked_cols)) {
    def$col_tracker_zoom[[where]] <- tracked_cols
  }
  dm_from_def(def, zoomed = TRUE)
}

check_zoomed <- function(dm) {
  check_dm(dm)
  if (is_zoomed(dm)) {
    return()
  }

  fun_name <- as_string(sys.call(-1)[[1]])
  # if a method for `dm_zoomed()` is used for a `dm`, we don't want `fun_name = method.dm` but rather `fun_name = method`
  fun_name <- sub("\\.dm", "", fun_name)
  abort_only_possible_w_zoom(fun_name)
}

check_not_zoomed <- function(dm) {
  check_dm(dm)
  if (!is_zoomed(dm)) {
    return()
  }

  fun_name <- gsub(".dm_zoomed", "", as_string(sys.call(-1)[[1]]))
  # if a method for `dm()` is used for a `dm_zoomed`, we don't want `fun_name = method.dm_zoomed` but rather `fun_name = method`
  fun_name <- sub("\\.dm_zoomed", "", fun_name)
  abort_only_possible_wo_zoom(fun_name)
}

# For `nest.dm_zoomed()`, we need the incoming foreign keys of the originally zoomed table
get_orig_in_fks <- function(dm_zoomed, orig_table) {
  # FIXME: maybe there is a more efficient implementation possible?
  dm_zoomed %>%
    dm_get_all_fks_impl() %>%
    filter(parent_table == orig_table) %>%
    select(-parent_table)
}

get_all_cols <- function(dm, table_name) {
  set_names(colnames(tbl_impl(dm, table_name)))
}

# basic tests -------------------------------------------------------------

test_that("basic test: 'group_by()'-methods work", {
  expect_equivalent_tbl(
    group_by(dm_zoomed(), e) %>% tbl_zoomed(),
    group_by(tf_2(), e)
  )

  expect_dm_error(
    group_by(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'select()'-methods work", {
  expect_equivalent_tbl(
    select(dm_zoomed(), e, a = c) %>% tbl_zoomed(),
    select(tf_2(), e, a = c)
  )

  expect_dm_error(
    select(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'relocate()'-methods work", {
  expect_equivalent_tbl(
    relocate(dm_zoomed(), e) %>% tbl_zoomed(),
    relocate(tf_2(), e)
  )

  expect_equivalent_tbl(
    relocate(dm_zoomed(), e, .after = e1) %>% tbl_zoomed(),
    relocate(tf_2(), e, .after = e1)
  )

  expect_dm_error(
    relocate(dm_for_filter()),
    "only_possible_w_zoom"
  )
})


test_that("basic test: 'rename()'-methods work", {
  expect_equivalent_tbl(
    rename(dm_zoomed(), a = c) %>% tbl_zoomed(),
    rename(tf_2(), a = c)
  )

  expect_dm_error(
    rename(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'mutate()'-methods work", {
  expect_equivalent_tbl(
    mutate(dm_zoomed(), d_2 = d * 2) %>% tbl_zoomed(),
    mutate(tf_2(), d_2 = d * 2)
  )

  expect_dm_error(
    mutate(dm_for_filter()),
    "only_possible_w_zoom"
  )
})


test_that("basic test: 'transmute()'-methods work", {
  expect_equivalent_tbl(
    transmute(dm_zoomed(), d_2 = d * 2) %>% tbl_zoomed(),
    transmute(tf_2(), d_2 = d * 2)
  )

  expect_dm_error(
    transmute(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'ungroup()'-methods work", {
  expect_equivalent_tbl(
    group_by(dm_zoomed(), e) %>% ungroup() %>% tbl_zoomed(),
    group_by(tf_2(), e) %>% ungroup()
  )

  expect_dm_error(
    ungroup(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'summarise()'-methods work", {
  expect_equivalent_tbl(
    summarise(dm_zoomed(), d_2 = mean(d, na.rm = TRUE)) %>% tbl_zoomed(),
    summarise(tf_2(), d_2 = mean(d, na.rm = TRUE))
  )

  expect_dm_error(
    summarise(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'count()'-method works", {
  expect_equivalent_tbl(
    count(dm_zoomed()) %>% tbl_zoomed(),
    count(tf_2())
  )

  expect_equivalent_tbl(
    count(dm_zoomed(), c) %>% tbl_zoomed(),
    count(tf_2(), c)
  )

  expect_equivalent_tbl(
    count(dm_zoomed(), wt = d) %>% tbl_zoomed(),
    count(tf_2(), wt = d)
  )

  expect_equivalent_tbl(
    count(dm_zoomed(), sort = TRUE) %>% tbl_zoomed(),
    count(tf_2(), sort = TRUE)
  )

  expect_equivalent_tbl(
    count(dm_zoomed(), name = "COUNT") %>% tbl_zoomed(),
    count(tf_2(), name = "COUNT")
  )

  expect_dm_error(
    count(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'tally()'-method works", {
  expect_equivalent_tbl(
    tally(dm_zoomed()) %>% tbl_zoomed(),
    tally(tf_2())
  )

  expect_dm_error(
    tally(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'filter()'-methods work", {
  skip_if_src("maria")

  expect_equivalent_tbl(
    dm_zoomed() %>%
      filter(d > mean(d, na.rm = TRUE)) %>%
      dm_update_zoomed() %>%
      tbl_impl("tf_2"),
    tf_2() %>%
      filter(d > mean(d, na.rm = TRUE))
  )
})

test_that("basic test: 'filter()'-methods work (2)", {
  expect_dm_error(
    filter(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'distinct()'-methods work", {
  expect_equivalent_tbl(
    distinct(dm_zoomed(), d_new = d) %>% dm_update_zoomed() %>% tbl_impl("tf_2"),
    distinct(tf_2(), d_new = d)
  )

  expect_dm_error(
    distinct(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'arrange()'-methods work", {
  # standard arrange
  expect_equivalent_tbl(
    arrange(dm_zoomed(), e) %>% tbl_zoomed(),
    arrange(tf_2(), e)
  )

  # arrange within groups
  expect_equivalent_tbl(
    group_by(dm_zoomed(), e) %>% arrange(desc(d), .by_group = TRUE) %>% tbl_zoomed(),
    arrange(group_by(tf_2(), e), desc(d), .by_group = TRUE)
  )

  expect_dm_error(
    arrange(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'slice()'-methods work", {
  skip_if_remote_src()
  expect_message(
    expect_equivalent_tbl(slice(dm_zoomed(), 3:6) %>% tbl_zoomed(), slice(tf_2(), 3:6)),
    "`slice.dm_zoomed\\(\\)` can potentially"
  )

  # silent when no PK available
  expect_silent(
    expect_equivalent_tbl(
      dm_for_disambiguate() %>%
        dm_zoom_to(iris_3) %>%
        slice(1:3) %>%
        tbl_zoomed(),
      iris_3() %>%
        slice(1:3)
    )
  )

  # changed for #663: mutate() tracks now all cols that remain
  expect_message(
    mutate(dm_zoomed(), c = 1) %>% slice(1:3),
    "Keeping PK column"
  )

  expect_silent(
    expect_equivalent_tbl(
      slice(dm_zoomed(), if_else(d < 5, 1:6, 7:2), .keep_pk = FALSE) %>% tbl_zoomed(),
      slice(tf_2(), if_else(d < 5, 1:6, 7:2))
    )
  )

  expect_dm_error(
    slice(dm_for_filter(), 2),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'join()'-methods for `zoomed.dm` work", {
  expect_equivalent_tbl(
    left_join(dm_zoomed(), tf_1) %>% dm_update_zoomed() %>% tbl_impl("tf_2"),
    left_join(tf_2(), tf_1(), by = c("d" = "a"))
  )

  expect_equivalent_tbl(
    inner_join(dm_zoomed(), tf_1) %>% dm_update_zoomed() %>% tbl_impl("tf_2"),
    inner_join(tf_2(), tf_1(), by = c("d" = "a"))
  )

  expect_equivalent_tbl(
    semi_join(dm_zoomed(), tf_1) %>% dm_update_zoomed() %>% tbl_impl("tf_2"),
    semi_join(tf_2(), tf_1(), by = c("d" = "a"))
  )

  expect_equivalent_tbl(
    anti_join(dm_zoomed(), tf_1) %>% dm_update_zoomed() %>% tbl_impl("tf_2"),
    anti_join(tf_2(), tf_1(), by = c("d" = "a"))
  )

  # SQLite doesn't implement right join
  skip_if_src("sqlite")
  skip_if_src("maria")
  expect_equivalent_tbl(
    full_join(dm_zoomed(), tf_1) %>% dm_update_zoomed() %>% tbl_impl("tf_2"),
    full_join(tf_2(), tf_1(), by = c("d" = "a"))
  )

  expect_equivalent_tbl(
    right_join(dm_zoomed(), tf_1) %>% dm_update_zoomed() %>% tbl_impl("tf_2"),
    right_join(tf_2(), tf_1(), by = c("d" = "a"))
  )

  # these databases don't implement nest join
  skip_if_src("mssql", "postgres", "sqlite", "maria")
  # https://github.com/duckdb/duckdb/pull/3829
  skip_if_src("duckdb")
  expect_equivalent_tbl(
    nest_join(dm_zoomed(), tf_1) %>% dm_update_zoomed() %>% tbl_impl("tf_2"),
    nest_join(tf_2(), tf_1(), by = c("d" = "a"), name = "tf_1")
  )
})

test_that("basic test: 'join()'-methods for `zoomed.dm` work (2)", {
  # fails if RHS not linked to zoomed table and no by is given
  expect_dm_error(
    left_join(dm_zoomed(), tf_4),
    "tables_not_neighbors"
  )

  # works, if by is given
  if (is_db(my_test_src())) {
    expect_equivalent_tbl(
      left_join(dm_zoomed(), tf_4, by = c("e" = "j")) %>% dm_update_zoomed() %>% tbl_impl("tf_2"),
      left_join(tf_2(), tf_4(), by = c("e" = "j"))
    )

    expect_equivalent_tbl(
      left_join(dm_zoomed(), tf_4, by = c("e" = "j", "e1" = "j1")) %>%
        dm_update_zoomed() %>%
        tbl_impl("tf_2"),
      left_join(tf_2(), tf_4(), by = c("e" = "j", "e1" = "j1"))
    )

    # explicitly select columns from RHS using argument `select`
    expect_equivalent_tbl(
      left_join(dm_zoomed_2(), tf_2, select = c(starts_with("c"), e, e1)) %>%
        dm_update_zoomed() %>%
        tbl_impl("tf_3"),
      left_join(tf_3(), select(tf_2(), c, e, e1), by = c("f" = "e", "f1" = "e1"))
    )

    # explicitly select and rename columns from RHS using argument `select`
    expect_equivalent_tbl(
      left_join(dm_zoomed_2(), tf_2, select = c(starts_with("c"), d_new = d, e, e1)) %>%
        dm_update_zoomed() %>%
        tbl_impl("tf_3"),
      left_join(tf_3(), select(tf_2(), c, d_new = d, e, e1), by = c("f" = "e", "f1" = "e1"))
    )
  } else {
    if (utils::packageVersion("dplyr") >= "1.1.0.9000") {
      expect_equivalent_tbl(
        left_join(dm_zoomed(), tf_4, by = c("e" = "j"), relationship = "many-to-many") %>%
          dm_update_zoomed() %>%
          tbl_impl("tf_2"),
        left_join(tf_2(), tf_4(), by = c("e" = "j"), relationship = "many-to-many")
      )

      expect_equivalent_tbl(
        left_join(
          dm_zoomed(),
          tf_4,
          by = c("e" = "j", "e1" = "j1"),
          relationship = "many-to-many"
        ) %>%
          dm_update_zoomed() %>%
          tbl_impl("tf_2"),
        left_join(tf_2(), tf_4(), by = c("e" = "j", "e1" = "j1"), relationship = "many-to-many")
      )
    }

    # explicitly select columns from RHS using argument `select`
    expect_equivalent_tbl(
      left_join(dm_zoomed_2(), tf_2, select = c(starts_with("c"), e, e1), multiple = "all") %>%
        dm_update_zoomed() %>%
        tbl_impl("tf_3"),
      left_join(tf_3(), select(tf_2(), c, e, e1), by = c("f" = "e", "f1" = "e1"), multiple = "all")
    )

    # explicitly select and rename columns from RHS using argument `select`
    expect_equivalent_tbl(
      left_join(
        dm_zoomed_2(),
        tf_2,
        select = c(starts_with("c"), d_new = d, e, e1),
        multiple = "all"
      ) %>%
        dm_update_zoomed() %>%
        tbl_impl("tf_3"),
      left_join(
        tf_3(),
        select(tf_2(), c, d_new = d, e, e1),
        by = c("f" = "e", "f1" = "e1"),
        multiple = "all"
      )
    )
  }

  # a former FK-relation could not be tracked
  expect_dm_error(
    dm_zoomed() %>% select(-e) %>% left_join(tf_3),
    "fk_not_tracked"
  )

  expect_snapshot({
    "keys are correctly tracked if selected columns from 'y' have same name as key columns from 'x'"
    dm_zoomed() %>%
      left_join(tf_3, select = c(d = g, f, f1)) %>%
      dm_update_zoomed() %>%
      get_all_keys()

    "keys are correctly tracked if selected columns from 'y' have same name as key columns from 'x'"
    dm_zoomed() %>%
      semi_join(tf_3, select = c(d = g, f, f1)) %>%
      dm_update_zoomed() %>%
      get_all_keys()
  })
})

test_that("basic test: 'join()'-methods for `zoomed.dm` work (3)", {
  skip_if_src("maria")
  # multi-column "by" argument
  out <- expect_message_obj(
    dm_for_disambiguate() %>%
      dm_zoom_to(iris_2) %>%
      left_join(iris_2, by = c("key", "Sepal.Width", "other_col")) %>%
      tbl_zoomed()
  )
  expect_equivalent_tbl(
    out,
    left_join(
      iris_2() %>%
        rename_at(vars(matches("^[PS]")), ~ paste0(., ".iris_2.x")) %>%
        rename(Sepal.Width = Sepal.Width.iris_2.x),
      iris_2() %>% rename_at(vars(matches("^[PS]")), ~ paste0(., ".iris_2.y")),
      by = c("key", "Sepal.Width" = "Sepal.Width.iris_2.y", "other_col")
    )
  )
})

test_that("basic test: 'join()'-methods for `zoomed.dm` work (3)", {
  skip_if_src("sqlite")
  # test RHS-by name collision
  if (is_db(my_test_src())) {
    expect_equivalent_dm(
      dm_for_filter() %>%
        dm_rename(tf_2, "...1" = d) %>%
        dm_zoom_to(tf_3) %>%
        right_join(tf_2) %>%
        dm_update_zoomed(),
      dm_for_filter() %>%
        dm_zoom_to(tf_3) %>%
        right_join(tf_2) %>%
        dm_update_zoomed() %>%
        dm_rename(tf_3, "...1" = d) %>%
        dm_rename(tf_2, "...1" = d)
    )
  } else {
    expect_equivalent_dm(
      dm_for_filter() %>%
        dm_rename(tf_2, "...1" = d) %>%
        dm_zoom_to(tf_3) %>%
        right_join(tf_2, multiple = "all") %>%
        dm_update_zoomed(),
      dm_for_filter() %>%
        dm_zoom_to(tf_3) %>%
        right_join(tf_2, multiple = "all") %>%
        dm_update_zoomed() %>%
        dm_rename(tf_3, "...1" = d) %>%
        dm_rename(tf_2, "...1" = d)
    )
  }
})

test_that("basic test: 'join()'-methods for `dm` throws error", {
  expect_dm_error(
    left_join(dm_for_filter()),
    "only_possible_w_zoom"
  )

  expect_dm_error(
    inner_join(dm_for_filter()),
    "only_possible_w_zoom"
  )

  expect_dm_error(
    semi_join(dm_for_filter()),
    "only_possible_w_zoom"
  )

  expect_dm_error(
    anti_join(dm_for_filter()),
    "only_possible_w_zoom"
  )

  expect_dm_error(
    full_join(dm_for_filter()),
    "only_possible_w_zoom"
  )

  expect_dm_error(
    right_join(dm_for_filter()),
    "only_possible_w_zoom"
  )

  expect_dm_error(
    inner_join(dm_zoom_to(dm_for_filter(), tf_1), tf_7),
    "table_not_in_dm"
  )

  expect_dm_error(
    nest_join(dm_for_filter()),
    "only_possible_w_zoom"
  )

  expect_dm_error(
    pack_join(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'across' works properly", {
  expect_equivalent_tbl(
    dm_for_filter() %>%
      dm_zoom_to(tf_2) %>%
      mutate(across(c(1, 3), ~"C")) %>%
      pull_tbl(),
    dm_for_filter() %>%
      pull_tbl(tf_2) %>%
      mutate(across(c(1, 3), ~"C"))
  )

  expect_equivalent_tbl(
    dm_for_filter() %>%
      dm_zoom_to(tf_2) %>%
      summarize(across(c(c, e), ~"C")) %>%
      pull_tbl(),
    dm_for_filter() %>%
      pull_tbl(tf_2) %>%
      summarize(across(c(c, e), ~"C"))
  )

  expect_equivalent_tbl(
    dm_for_filter() %>%
      dm_zoom_to(tf_2) %>%
      group_by(d) %>%
      summarize(across(c(1, 3), ~"C")) %>%
      pull_tbl(),
    dm_for_filter() %>%
      pull_tbl(tf_2) %>%
      group_by(d) %>%
      summarize(across(c(1, 3), ~"C"))
  )
})

# test key tracking for all methods ---------------------------------------

# dm_for_filter(), zoomed to tf_2; PK: c; 2 outgoing FKs: d, e; no incoming FKS
zoomed_grouped_out_dm <- dm_zoom_to(dm_for_filter(), tf_2) %>% group_by(c, e, e1)

# dm_for_filter(), zoomed to tf_3; PK: f; 2 incoming FKs: tf_4$j, tf_2$e; no outgoing FKS:
zoomed_grouped_in_dm <- dm_zoom_to(dm_for_filter(), tf_3) %>% group_by(g)

test_that("key tracking works", {
  expect_snapshot({
    "rename()"

    zoomed_grouped_out_dm %>%
      rename(c_new = c) %>%
      dm_update_zoomed() %>%
      get_all_keys()

    zoomed_grouped_out_dm %>%
      rename(e_new = e) %>%
      dm_update_zoomed() %>%
      get_all_keys()

    # FKs should not be dropped when renaming the PK they are pointing to; tibble from `dm_get_all_fks()` shouldn't change
    zoomed_grouped_in_dm %>%
      rename(f_new = f) %>%
      dm_update_zoomed() %>%
      get_all_keys()

    "summarize()"

    # grouped by two key cols: "c" and "e" -> these two remain
    zoomed_grouped_out_dm %>%
      summarize(d_mean = mean(d)) %>%
      dm_insert_zoomed("new_tbl") %>%
      get_all_keys()

    # grouped_by non-key col means, that no keys remain
    zoomed_grouped_in_dm %>%
      summarize(g_list = list(g)) %>%
      dm_insert_zoomed("new_tbl") %>%
      get_all_keys()
  })
})

test_that("key tracking works (2)", {
  # https://github.com/tidyverse/dbplyr/issues/670
  skip_if_remote_src()

  expect_snapshot({
    "transmute()"

    # grouped by three key cols: "c", "e", "e1" -> these three remain
    zoomed_grouped_out_dm %>%
      transmute(d_mean = mean(d)) %>%
      dm_insert_zoomed("new_tbl") %>%
      get_all_keys()
  })
})

test_that("key tracking works (3)", {
  # FKs that point to a PK that vanished, should also vanish
  expect_snapshot({
    # grouped_by non-key col means, that no keys remain
    zoomed_grouped_in_dm %>%
      transmute(g_list = list(g)) %>%
      dm_insert_zoomed("new_tbl") %>%
      get_all_keys()
  })
})

test_that("key tracking works (4)", {
  # FKs that point to a PK that vanished, should also vanish
  expect_snapshot({
    "mutate()"

    # grouped by three key cols: "c", "e" and "e1 -> these three remain
    zoomed_grouped_out_dm %>%
      mutate(d_mean = mean(d)) %>%
      select(-d) %>%
      dm_insert_zoomed("new_tbl") %>%
      get_all_keys()

    # grouped_by non-key col means, that only key-columns that remain in the
    # result tibble are tracked for mutate()
    zoomed_grouped_in_dm %>%
      mutate(f = paste0(g, g)) %>%
      dm_insert_zoomed("new_tbl") %>%
      get_all_keys()

    # grouped_by non-key col means, that only key-columns that remain in the
    # result tibble are tracked for transmute()
    zoomed_grouped_in_dm %>%
      mutate(g_new = list(g)) %>%
      dm_insert_zoomed("new_tbl") %>%
      get_all_keys()
  })
})

test_that("key tracking works (5)", {
  skip_if_remote_src()

  expect_snapshot({
    "chain of renames & other transformations"

    zoomed_grouped_out_dm %>%
      summarize(d_mean = mean(d)) %>%
      ungroup() %>%
      rename(e_new = e) %>%
      group_by(e_new, e1) %>%
      transmute(c = paste0(c, "_animal")) %>%
      dm_insert_zoomed("new_tbl") %>%
      get_all_keys()
  })
})

test_that("key tracking works (6)", {
  # FKs that point to a PK that vanished, should also vanish
  expect_snapshot({
    zoomed_grouped_in_dm %>%
      select(g_new = g) %>%
      get_all_keys("tf_3")
  })
})

test_that("key tracking works for distinct() and arrange()", {
  expect_identical(
    dm_zoomed() %>%
      distinct(d_new = d) %>%
      dm_update_zoomed() %>%
      dm_get_all_fks_impl(),
    dm_for_filter() %>%
      dm_get_all_fks_impl() %>%
      filter(child_table != "tf_2")
  )

  expect_identical(
    dm_zoomed() %>%
      arrange(e) %>%
      dm_update_zoomed() %>%
      dm_get_all_fks_impl(),
    dm_for_filter() %>%
      dm_get_all_fks_impl()
  )

  expect_identical(
    dm_for_flatten() %>%
      dm_zoom_to(fact) %>%
      select(dim_1_key_1, dim_1_key_2, dim_3_key, dim_2_key) %>%
      dm_update_zoomed() %>%
      dm_get_all_fks_impl(),
    dm_for_flatten() %>%
      dm_get_all_fks_impl() %>%
      filter(child_fk_cols != new_keys("dim_4_key"))
  )

  # it should be possible to combine 'filter' on a dm_zoomed with all other dplyr-methods; example: 'rename'
  expect_equivalent_dm(
    dm_for_filter() %>%
      dm_zoom_to(tf_2) %>%
      filter(d < 6) %>%
      rename(c_new = c, d_new = d) %>%
      dm_update_zoomed() %>%
      dm_select_tbl(tf_2) %>%
      dm_rm_pk(tf_2),
    dm_for_filter()$tf_2 %>%
      filter(d < 6) %>%
      rename(c_new = c, d_new = d) %>%
      dm(tf_2 = .)
  )

  # dm_nycflights13() (with FK constraints) doesn't work on DB
  # here, FK constraints are not implemented on the DB

  expect_equivalent_tbl(
    dm_nycflights_small() %>%
      dm_zoom_to(weather) %>%
      summarize(avg_wind_speed = mean(wind_speed, na.rm = TRUE)) %>%
      tbl_zoomed(),
    tbl_impl(dm_nycflights_small(), "weather") %>%
      summarize(avg_wind_speed = mean(wind_speed, na.rm = TRUE))
  )

  expect_equivalent_tbl(
    dm_nycflights_small() %>%
      dm_zoom_to(weather) %>%
      transmute(celsius_temp = (temp - 32) * 5 / 9) %>%
      tbl_zoomed(),
    tbl_impl(dm_nycflights_small(), "weather") %>% transmute(celsius_temp = (temp - 32) * 5 / 9)
  )

  expect_equivalent_tbl(
    dm_nycflights_small() %>%
      dm_zoom_to(weather) %>%
      summarize(avg_wind_speed = mean(wind_speed, na.rm = TRUE)) %>%
      tbl_zoomed(),
    tbl_impl(dm_nycflights_small(), "weather") %>%
      summarize(avg_wind_speed = mean(wind_speed, na.rm = TRUE))
  )

  expect_equivalent_tbl(
    dm_nycflights_small() %>%
      dm_zoom_to(weather) %>%
      transmute(celsius_temp = (temp - 32) * 5 / 9) %>%
      tbl_zoomed(),
    tbl_impl(dm_nycflights_small(), "weather") %>% transmute(celsius_temp = (temp - 32) * 5 / 9)
  )

  # slice() doesn't work on DB and reformatting a datetime on a DB is
  # currently not possible with a mere `format()` call -> skipping; cf. #358
  skip_if_remote_src()
  # keys tracking when there are no keys to track
  expect_equivalent_tbl(
    dm_nycflights_small() %>%
      dm_zoom_to(weather) %>%
      mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
      tbl_zoomed(),
    tbl_impl(dm_nycflights_small(), "weather") %>%
      mutate(time_hour_fmt = format(time_hour, tz = "UTC"))
  )
})


test_that("key tracking works for slice()", {
  skip_if_remote_src()
  expect_identical(
    slice(dm_zoomed(), if_else(d < 5, 1:6, 7:2), .keep_pk = FALSE) %>% col_tracker_zoomed(),
    set_names(c("d", "e", "e1"))
  )
  expect_message(
    expect_identical(
      slice(dm_zoomed(), if_else(d < 5, 1:6, 7:2)) %>% col_tracker_zoomed(),
      set_names(c("c", "d", "e", "e1"))
    ),
    "Keeping PK"
  )
  expect_identical(
    slice(dm_zoomed(), if_else(d < 5, 1:6, 7:2), .keep_pk = TRUE) %>% col_tracker_zoomed(),
    set_names(c("c", "d", "e", "e1"))
  )
})


test_that("can use column as primary and foreign key", {
  f <- tibble(data_card_1 = 1:3)
  data_card_1 <- tibble(data_card_1 = 1:3)

  dm <-
    dm(f, data_card_1) %>%
    dm_add_pk(f, data_card_1) %>%
    dm_add_pk(data_card_1, data_card_1) %>%
    dm_add_fk(f, data_card_1, data_card_1)

  expect_equivalent_dm(
    dm %>%
      dm_zoom_to(f) %>%
      dm_update_zoomed(),
    dm
  )
})

test_that("'summarize_at()' etc. work", {
  expect_equivalent_tbl(
    dm_nycflights_small() %>%
      dm_zoom_to(airports) %>%
      summarize_at(vars(lat, lon), list(mean = mean, min = min, max = max), na.rm = TRUE) %>%
      tbl_zoomed(),
    dm_nycflights_small() %>%
      pull_tbl(airports) %>%
      summarize_at(vars(lat, lon), list(mean = mean, min = min, max = max), na.rm = TRUE)
  )

  expect_equivalent_tbl(
    dm_nycflights_small() %>%
      dm_zoom_to(airports) %>%
      select(3:6) %>%
      summarize_all(list(mean = mean, sum = sum), na.rm = TRUE) %>%
      tbl_zoomed(),
    dm_nycflights_small() %>%
      pull_tbl(airports) %>%
      select(3:6) %>%
      summarize_all(list(mean = mean, sum = sum), na.rm = TRUE)
  )

  skip_if_remote_src()
  expect_equivalent_tbl(
    dm_nycflights_small() %>%
      dm_zoom_to(airports) %>%
      summarize_if(is_double, list(mean = mean, sum = sum), na.rm = TRUE) %>%
      tbl_zoomed(),
    dm_nycflights_small() %>%
      pull_tbl(airports) %>%
      summarize_if(is_double, list(mean = mean, sum = sum), na.rm = TRUE)
  )
})

test_that("unique_prefix()", {
  expect_equal(unique_prefix(character()), "...")
  expect_equal(unique_prefix(c("a", "bc", "ef")), "...")
  expect_equal(unique_prefix(c("a", "bcd", "ef")), "...")
  expect_equal(unique_prefix(c("a", "....", "ef")), "....")
})


# compound tests ----------------------------------------------------------

test_that("output for compound keys", {
  # FIXME: COMPOUND: Need proper test
  skip_if_remote_src()

  zoomed_comp_dm <-
    nyc_comp() %>%
    dm_zoom_to(weather)
  # grouped by one key col and one other col
  grouped_zoomed_comp_dm_1 <-
    zoomed_comp_dm %>%
    group_by(time_hour, wind_dir)
  # grouped by the two key cols
  grouped_zoomed_comp_dm_2 <-
    zoomed_comp_dm %>%
    group_by(time_hour, origin)

  expect_snapshot({
    # TRANSFORMATION VERBS

    # mutate()
    grouped_zoomed_comp_dm_1 %>%
      mutate(count = n()) %>%
      col_tracker_zoomed()
    grouped_zoomed_comp_dm_2 %>%
      mutate(count = n()) %>%
      col_tracker_zoomed()
    # transmute()
    grouped_zoomed_comp_dm_1 %>%
      transmute(count = n()) %>%
      dm_update_zoomed()
    grouped_zoomed_comp_dm_2 %>%
      transmute(count = n()) %>%
      dm_update_zoomed()
    # summarize()
    grouped_zoomed_comp_dm_1 %>%
      summarize(count = n()) %>%
      dm_update_zoomed()
    grouped_zoomed_comp_dm_2 %>%
      summarize(count = n()) %>%
      dm_update_zoomed()
    # select()
    zoomed_comp_dm %>%
      select(time_hour, wind_dir) %>%
      dm_update_zoomed()
    zoomed_comp_dm %>%
      select(time_hour, origin, wind_dir) %>%
      dm_update_zoomed()
    # rename()
    zoomed_comp_dm %>%
      rename(th = time_hour, wd = wind_dir) %>%
      dm_update_zoomed()
    # distinct()
    zoomed_comp_dm %>%
      distinct(origin, wind_dir) %>%
      dm_update_zoomed()
    zoomed_comp_dm %>%
      distinct(origin, wind_dir, time_hour) %>%
      dm_update_zoomed()
    # filter() (cf. #437)
    zoomed_comp_dm %>%
      filter(pressure < 1020) %>%
      dm_update_zoomed()
    # pull()
    zoomed_comp_dm %>%
      pull(origin) %>%
      unique()
    # slice()
    zoomed_comp_dm %>%
      slice(c(1:3, 5:3))
    zoomed_comp_dm %>%
      slice(c(1:3, 5:3), .keep_pk = TRUE) %>%
      col_tracker_zoomed()
    # FIXME: COMPOUND:: .keep_pk = FALSE cannot deal with compound keys ATM
    # zoomed_comp_dm %>%
    #   slice(c(1:3, 5:3), .keep_pk = FALSE) %>%
    #   get_tracked_cols()

    # JOINS

    # left_join()
    zoomed_comp_dm %>%
      left_join(flights, multiple = "all") %>%
      nrow()
    # right_join()
    zoomed_comp_dm %>%
      right_join(flights, multiple = "all") %>%
      nrow()
    # inner_join()
    zoomed_comp_dm %>%
      inner_join(flights, multiple = "all") %>%
      nrow()
    # full_join()
    zoomed_comp_dm %>%
      full_join(flights, multiple = "all") %>%
      nrow()
    # semi_join()
    zoomed_comp_dm %>%
      semi_join(flights) %>%
      nrow()
    # anti_join()
    zoomed_comp_dm %>%
      anti_join(flights) %>%
      nrow()
    # nest_join()
    zoomed_comp_dm %>%
      nest_join(flights) %>%
      nrow()
  })
})


# dplyr 1.2.0 tests -------------------------------------------------------

test_that("basic test: 'filter_out()'-methods work", {
  skip_if_remote_src()

  expect_equivalent_tbl(
    dm_zoomed() %>%
      filter_out(d < mean(d, na.rm = TRUE)) %>%
      tbl_zoomed(),
    tf_2() %>%
      filter_out(d < mean(d, na.rm = TRUE))
  )

  expect_dm_error(
    filter_out(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'reframe()'-methods work", {
  expect_equivalent_tbl(
    dm_zoomed() %>%
      group_by(e) %>%
      reframe(d_mean = mean(d, na.rm = TRUE)) %>%
      tbl_zoomed(),
    tf_2() %>%
      group_by(e) %>%
      reframe(d_mean = mean(d, na.rm = TRUE))
  )

  expect_dm_error(
    reframe(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'cross_join()'-methods for `zoomed.dm` work", {
  skip_if_remote_src()

  expect_equivalent_tbl(
    cross_join(dm_zoomed(), tf_3) %>% tbl_zoomed(),
    cross_join(tf_2(), tf_3())
  )

  expect_dm_error(
    cross_join(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

# dm_keyed_tbl tests -------------------------------------------------------

test_that("dm_keyed_tbl methods preserve keyed class", {
  skip_if_remote_src()

  dm <- dm_for_filter()
  tbl <- keyed_tbl_impl(dm, "tf_2")

  expect_s3_class(filter(tbl, d > 5), "dm_keyed_tbl")
  expect_s3_class(filter_out(tbl, d > 5), "dm_keyed_tbl")
  expect_s3_class(mutate(tbl, d2 = d * 2), "dm_keyed_tbl")
  expect_s3_class(transmute(tbl, d2 = d * 2), "dm_keyed_tbl")
  expect_s3_class(select(tbl, c, d), "dm_keyed_tbl")
  expect_s3_class(relocate(tbl, e, .before = c), "dm_keyed_tbl")
  expect_s3_class(rename(tbl, c2 = c), "dm_keyed_tbl")
  expect_s3_class(distinct(tbl, e), "dm_keyed_tbl")
  expect_s3_class(arrange(tbl, desc(d)), "dm_keyed_tbl")
  expect_s3_class(slice(tbl, 1:2), "dm_keyed_tbl")
  expect_s3_class(ungroup(group_by(tbl, e)), "dm_keyed_tbl")
  expect_s3_class(count(tbl, e), "dm_keyed_tbl")
  expect_s3_class(tally(tbl), "dm_keyed_tbl")
  expect_s3_class(reframe(group_by(tbl, e), d_mean = mean(d, na.rm = TRUE)), "dm_keyed_tbl")
})

# Signature alignment tests ------------------------------------------------

test_that("dm method signatures match dplyr data.frame method signatures", {
  skip_on_cran()

  dplyr_ns <- asNamespace("dplyr")
  dm_ns <- asNamespace("dm")

  # All dplyr verbs for which we provide methods, mapped to their classes
  verbs <- c(
    "filter",
    "filter_out",
    "mutate",
    "transmute",
    "select",
    "relocate",
    "rename",
    "distinct",
    "arrange",
    "slice",
    "group_by",
    "ungroup",
    "summarise",
    "reframe",
    "count",
    "tally",
    "pull",
    "left_join",
    "right_join",
    "inner_join",
    "full_join",
    "semi_join",
    "anti_join",
    "nest_join",
    "cross_join"
  )

  for (verb in verbs) {
    df_method <- tryCatch(
      get(paste0(verb, ".data.frame"), envir = dplyr_ns),
      error = function(e) NULL
    )
    if (is.null(df_method)) {
      next
    }

    df_args <- names(formals(df_method))

    for (cls in c("dm", "dm_zoomed", "dm_keyed_tbl")) {
      method_name <- paste0(verb, ".", cls)
      dm_method <- tryCatch(
        get(method_name, envir = dm_ns),
        error = function(e) NULL
      )
      if (is.null(dm_method)) {
        next
      }

      dm_args <- names(formals(dm_method))
      missing_args <- setdiff(df_args, dm_args)
      expect_true(
        length(missing_args) == 0,
        label = paste0(
          method_name,
          " is missing args from ",
          verb,
          ".data.frame: ",
          paste(missing_args, collapse = ", ")
        )
      )
    }
  }
})

# join_by() tests ----------------------------------------------------------

test_that("zoomed joins work with join_by()", {
  skip_if_remote_src()

  # left_join with join_by() using FK column mapping (tf_2.e,e1 -> tf_3.f,f1)
  expect_equivalent_tbl(
    dm_zoomed() %>%
      left_join(tf_3, by = join_by(e == f, e1 == f1)) %>%
      tbl_zoomed(),
    left_join(tf_2(), tf_3(), by = join_by(e == f, e1 == f1))
  )

  # semi_join with join_by()
  expect_equivalent_tbl(
    dm_zoomed() %>%
      semi_join(tf_3, by = join_by(e == f, e1 == f1)) %>%
      tbl_zoomed(),
    semi_join(tf_2(), tf_3(), by = join_by(e == f, e1 == f1))
  )

  # anti_join with join_by()
  expect_equivalent_tbl(
    dm_zoomed() %>%
      anti_join(tf_3, by = join_by(e == f, e1 == f1)) %>%
      tbl_zoomed(),
    anti_join(tf_2(), tf_3(), by = join_by(e == f, e1 == f1))
  )
})

test_that("keyed joins work with join_by()", {
  skip_if_remote_src()

  dm <- dm_for_filter()
  tbl_2 <- keyed_tbl_impl(dm, "tf_2")
  tbl_3 <- keyed_tbl_impl(dm, "tf_3")

  result <- left_join(tbl_2, tbl_3, by = join_by(e == f, e1 == f1))
  expect_s3_class(result, "dm_keyed_tbl")
  expect_true(nrow(result) > 0)

  result <- inner_join(tbl_2, tbl_3, by = join_by(e == f, e1 == f1))
  expect_s3_class(result, "dm_keyed_tbl")
  expect_true(nrow(result) > 0)
})

# dplyr 1.2.0 compatibility tests -----------------------------------------

test_that(".by works with zoomed filter()", {
  skip_if_remote_src()

  expect_equivalent_tbl(
    dm_zoomed() %>%
      filter(d == max(d), .by = e) %>%
      tbl_zoomed(),
    tf_2() %>%
      filter(d == max(d), .by = e)
  )
})

test_that(".by works with keyed filter()", {
  skip_if_remote_src()

  dm <- dm_for_filter()
  tbl <- keyed_tbl_impl(dm, "tf_2")

  result <- filter(tbl, d == max(d), .by = e)
  expect_s3_class(result, "dm_keyed_tbl")

  expected <- filter(tibble::as_tibble(tbl), d == max(d), .by = e)
  expect_equal(nrow(result), nrow(expected))
})

test_that(".by works with zoomed mutate()", {
  skip_if_remote_src()

  expect_equivalent_tbl(
    dm_zoomed() %>%
      mutate(d_mean = mean(d, na.rm = TRUE), .by = e) %>%
      tbl_zoomed(),
    tf_2() %>%
      mutate(d_mean = mean(d, na.rm = TRUE), .by = e)
  )
})

test_that(".by works with keyed mutate()", {
  skip_if_remote_src()

  dm <- dm_for_filter()
  tbl <- keyed_tbl_impl(dm, "tf_2")

  result <- mutate(tbl, d_mean = mean(d, na.rm = TRUE), .by = e)
  expect_s3_class(result, "dm_keyed_tbl")
  expect_true("d_mean" %in% colnames(result))
})

test_that(".by works with zoomed summarise()", {
  expect_equivalent_tbl(
    dm_zoomed() %>%
      summarise(d_mean = mean(d, na.rm = TRUE), .by = e) %>%
      tbl_zoomed(),
    tf_2() %>%
      summarise(d_mean = mean(d, na.rm = TRUE), .by = e)
  )
})

test_that(".by key tracking works with zoomed summarise()", {
  expect_snapshot({
    # .by should track keys like group_by does
    dm_zoom_to(dm_for_filter(), tf_2) %>%
      summarize(d_mean = mean(d), .by = c(c, e, e1)) %>%
      dm_insert_zoomed("new_tbl") %>%
      get_all_keys()

    # .by with non-key col means no keys remain
    dm_zoom_to(dm_for_filter(), tf_3) %>%
      summarize(g_list = list(g), .by = g) %>%
      dm_insert_zoomed("new_tbl") %>%
      get_all_keys()
  })
})

test_that(".by works with keyed summarise()", {
  skip_if_remote_src()

  dm <- dm_for_filter()
  tbl <- keyed_tbl_impl(dm, "tf_2")

  result <- summarise(tbl, d_mean = mean(d, na.rm = TRUE), .by = e)
  expect_s3_class(result, "dm_keyed_tbl")
  expect_true("d_mean" %in% colnames(result))
})

test_that(".by works with zoomed reframe()", {
  expect_equivalent_tbl(
    dm_zoomed() %>%
      reframe(d_mean = mean(d, na.rm = TRUE), .by = e) %>%
      tbl_zoomed(),
    tf_2() %>%
      reframe(d_mean = mean(d, na.rm = TRUE), .by = e)
  )
})

test_that(".by works with keyed slice()", {
  skip_if_remote_src()

  dm <- dm_for_filter()
  tbl <- keyed_tbl_impl(dm, "tf_2")

  result <- slice(tbl, 1, .by = e)
  expect_s3_class(result, "dm_keyed_tbl")
  expect_true(nrow(result) > 0)
})

test_that("mutate .keep and .before/.after work with zoomed dm", {
  skip_if_remote_src()

  # .keep = "used"
  expect_equivalent_tbl(
    dm_zoomed() %>%
      mutate(d2 = d * 2, .keep = "used") %>%
      tbl_zoomed(),
    tf_2() %>%
      mutate(d2 = d * 2, .keep = "used")
  )

  # .after
  result <- dm_zoomed() %>%
    mutate(d2 = d * 2, .after = d) %>%
    tbl_zoomed()
  expected <- tf_2() %>%
    mutate(d2 = d * 2, .after = d)
  expect_equivalent_tbl(result, expected)
  expect_equal(colnames(result), colnames(expected))
})

test_that("mutate .before/.after work with keyed tbl", {
  skip_if_remote_src()

  dm <- dm_for_filter()
  tbl <- keyed_tbl_impl(dm, "tf_2")

  result <- mutate(tbl, d2 = d * 2, .after = d)
  expect_s3_class(result, "dm_keyed_tbl")
  d_pos <- which(colnames(result) == "d")
  d2_pos <- which(colnames(result) == "d2")
  expect_equal(d2_pos, d_pos + 1)
})

test_that("arrange .locale works with zoomed dm", {
  skip_if_remote_src()

  result <- dm_zoomed() %>%
    arrange(e1, .locale = "en") %>%
    tbl_zoomed()
  expected <- tf_2() %>%
    arrange(e1, .locale = "en")
  expect_equivalent_tbl(result, expected)
})

test_that("cross_join works with keyed tables", {
  skip_if_remote_src()

  dm <- dm_for_filter()
  tbl_2 <- keyed_tbl_impl(dm, "tf_2")
  tbl_3 <- keyed_tbl_impl(dm, "tf_3")

  result <- cross_join(tbl_2, tbl_3)
  expect_s3_class(result, "dm_keyed_tbl")
  expect_equal(nrow(result), nrow(tbl_2) * nrow(tbl_3))
})

test_that("filter_out works correctly with zoomed dm", {
  skip_if_remote_src()

  # filter_out should drop matching rows, keeping NAs
  expect_equivalent_tbl(
    dm_zoomed() %>%
      filter_out(d > 5) %>%
      tbl_zoomed(),
    tf_2() %>%
      filter_out(d > 5)
  )
})

test_that("filter_out works correctly with keyed tbl", {
  skip_if_remote_src()

  dm <- dm_for_filter()
  tbl <- keyed_tbl_impl(dm, "tf_2")

  result <- filter_out(tbl, d > 5)
  expect_s3_class(result, "dm_keyed_tbl")

  expected <- filter_out(tibble::as_tibble(tbl), d > 5)
  expect_equal(nrow(result), nrow(expected))
})

test_that("reframe returns any number of rows per group", {
  skip_if_remote_src()

  # reframe can return multiple rows per group
  expect_equivalent_tbl(
    dm_zoomed() %>%
      group_by(e) %>%
      reframe(d_vals = range(d, na.rm = TRUE)) %>%
      tbl_zoomed(),
    tf_2() %>%
      group_by(e) %>%
      reframe(d_vals = range(d, na.rm = TRUE))
  )
})

test_that("count .drop works with zoomed dm", {
  skip_if_remote_src()

  result <- dm_zoomed() %>%
    count(e, .drop = FALSE) %>%
    tbl_zoomed()
  expected <- tf_2() %>%
    count(e, .drop = FALSE)
  expect_equivalent_tbl(result, expected)
})

test_that("utils-methods work", {
  expect_equivalent_tbl(
    dm_zoomed() %>% arrange(d) %>% head(3) %>% tbl_zoomed(),
    head(tf_2() %>% arrange(d), 3)
  )

  # Not implemented for remote sources:
  skip_if_remote_src()
  expect_equivalent_tbl(
    dm_zoomed() %>% tail(2) %>% tbl_zoomed(),
    tail(tf_2(), 2)
  )
})

test_that("dm_zoom_to() works", {
  # no zoom in unzoomed `dm`
  expect_false(
    is_zoomed(dm_for_filter())
  )

  # zoom in zoomed `dm`
  expect_true(
    is_zoomed(dm_for_filter() %>% dm_zoom_to(tf_1))
  )

  expect_s3_class(
    dm_for_filter() %>% dm_zoom_to(tf_3),
    c("dm_zoomed", "dm")
  )
})


test_that("dm_discard_zoomed() works", {
  # no zoom in zoomed out from zoomed `dm`
  expect_false(is_zoomed(dm_for_filter() %>% dm_zoom_to(tf_1) %>% dm_discard_zoomed()))

  expect_s3_class(
    dm_for_filter() %>% dm_zoom_to(tf_3) %>% dm_discard_zoomed(),
    c("dm")
  )
})

test_that("print() and format() methods for subclass `dm_zoomed` work", {
  expect_snapshot(
    dm_for_filter() %>% dm_zoom_to(tf_5) %>% as_dm_zoomed_df() %>% tbl_sum()
  )

  expect_snapshot(
    dm_for_filter() %>% dm_zoom_to(tf_2) %>% as_dm_zoomed_df() %>% tbl_sum()
  )
})


test_that("dm_get_zoom() and tbl_zoomed() works", {
  expect_identical(
    dm_for_filter() %>%
      dm_zoom_to(tf_2) %>%
      dm_get_zoom() %>%
      pluck("table"),
    "tf_2"
  )
  expect_equivalent_tbl(
    dm_for_filter() %>%
      dm_zoom_to(tf_2) %>%
      dm_get_zoom() %>%
      pluck("zoom") %>%
      pluck(1),
    tf_2()
  )

  # function for getting only the tibble itself works
  expect_equivalent_tbl(
    dm_for_filter() %>% dm_zoom_to(tf_3) %>% tbl_zoomed(),
    tf_3()
  )
})

test_that("dm_insert_zoomed() works", {
  # test that a new tbl is inserted, based on the requested one
  expect_equivalent_dm(
    dm_zoom_to(dm_for_filter(), tf_4) %>%
      dm_insert_zoomed("tf_4_new"),
    dm_for_filter() %>%
      dm(tf_4_new = tf_4()) %>%
      dm_add_pk(tf_4_new, h) %>%
      dm_add_fk(tf_4_new, c(j, j1), tf_3) %>%
      dm_add_fk(tf_5, l, tf_4_new, on_delete = "cascade")
  )

  # test that an error is thrown if 'repair = check_unique' and duplicate table names
  expect_dm_error(
    dm_zoom_to(dm_for_filter(), tf_4) %>% dm_insert_zoomed("tf_4", repair = "check_unique"),
    "need_unique_names"
  )

  # test that in case of 'repair = unique' and duplicate table names -> renames of old and new
  expect_equivalent_dm(
    # FIXME: This produced occasional warnings on GitHub Actions, why?
    dm_for_filter() %>%
      dm_zoom_to(tf_4) %>%
      dm_insert_zoomed("tf_4", repair = "unique", quiet = TRUE),
    dm_for_filter() %>%
      dm_rename_tbl(tf_4...4 = tf_4) %>%
      dm(tf_4...7 = tf_4()) %>%
      dm_add_pk(tf_4...7, h) %>%
      dm_add_fk(tf_4...7, c(j, j1), tf_3) %>%
      dm_add_fk(tf_5, l, tf_4...7, on_delete = "cascade")
  )
})

test_that("dm_update_tbl() works", {
  # setting table tf_7 as zoomed table for tf_6 and removing its primary key and foreign keys pointing to it
  new_dm_for_filter <-
    dm_get_def(dm_for_filter()) %>%
    mutate(
      zoom = if_else(table == "tf_6", list(tf_7()), list(NULL)),
      col_tracker_zoom = if_else(table == "tf_6", list(character()), list(NULL)),
    ) %>%
    dm_from_def(zoomed = TRUE)

  # test that the old table is updated correctly
  expect_equivalent_dm(
    dm_update_zoomed(new_dm_for_filter),
    dm_for_filter() %>%
      dm_select_tbl(-tf_6) %>%
      dm(tf_6 = tf_7())
  )
})

# after #271:
test_that("all cols are tracked in zoomed table", {
  expect_identical(
    dm_nycflights_small() %>%
      dm_zoom_to(flights) %>%
      col_tracker_zoomed(),
    set_names(colnames(dm_nycflights_small()$flights))
  )
})


# tests for compound keys -------------------------------------------------

test_that("zoom output for compound keys", {
  # FIXME: COMPOUND: Need proper test
  skip_if_remote_src()

  expect_snapshot({
    nyc_comp() %>% dm_zoom_to(weather)
    nyc_comp() %>%
      dm_zoom_to(weather) %>%
      dm_update_zoomed()
    nyc_comp_2 <-
      nyc_comp() %>%
      dm_zoom_to(weather) %>%
      dm_insert_zoomed("weather_2")
    nyc_comp_2 %>%
      get_all_keys()
    attr(graph_edges(create_graph_from_dm(nyc_comp_2)), "vnames")

    nyc_comp_3 <-
      nyc_comp() %>%
      dm_zoom_to(flights) %>%
      dm_insert_zoomed("flights_2")
    nyc_comp_3 %>%
      get_all_keys()
    attr(graph_edges(create_graph_from_dm(nyc_comp_3)), "vnames")
  })
})

test_that("dm_get_zoom() works to zoom on empty tables", {
  zdm <- dm(x = tibble()) %>% dm_zoom_to(x)
  expect_identical(
    dm_get_zoom(zdm),
    tibble(table = "x", zoom = list(tibble()))
  )
})


# test that inserting a zoomed table retains the color --------------------

test_that("dm_insert_zoomed() retains color", {
  expect_identical(
    dm_for_filter() %>%
      dm_set_colors("cyan" = tf_2) %>%
      dm_zoom_to(tf_2) %>%
      dm_insert_zoomed("tf_2_new") %>%
      dm_get_def() %>%
      filter(table == "tf_2_new") %>%
      pull(display),
    "#00FFFFFF"
  )
})

test_that("error message for non-dm object", {
  # FIXME: Add a similar test to all callers of dm_get_def()
  expect_snapshot(error = TRUE, {
    dm_get_def(structure(list(table = "a"), class = "bogus"))
  })
})

test_that("can upgrade from v1", {
  if (FALSE) {
    saveRDS(dm_for_filter(), "tests/testthat/dm/v1.rds", version = 2)
    saveRDS(dm_for_filter() %>% dm_zoom_to(tf_2), "tests/testthat/dm/v1_zoomed.rds", version = 2)
  }

  dm_v1 <- readRDS(test_path("dm/v1.rds"))
  expect_snapshot({
    def <- dm_get_def(dm_v1, quiet = TRUE)
    def <- dm_get_def(dm_v1)
    dm <- dm_from_def(def)
    dm_validate(dm)
    is_zoomed(dm)
  })
})

test_that("can upgrade zoomed from v1", {
  dm_v1_zoomed <- readRDS(test_path("dm/v1_zoomed.rds"))
  expect_snapshot({
    def <- dm_get_def(dm_v1_zoomed, quiet = TRUE)
    def <- dm_get_def(dm_v1_zoomed)
    dm <- dm_from_def(def, zoomed = TRUE)
    dm_validate(dm)
    is_zoomed(dm)
  })
})

test_that("can upgrade from v2", {
  if (FALSE) {
    saveRDS(dm_for_filter(), "tests/testthat/dm/v2.rds", version = 2)
    saveRDS(dm_for_filter() %>% dm_zoom_to(tf_2), "tests/testthat/dm/v2_zoomed.rds", version = 2)
  }

  dm_v2 <- readRDS(test_path("dm/v2.rds"))
  expect_snapshot({
    def <- dm_get_def(dm_v2, quiet = TRUE)
    def <- dm_get_def(dm_v2)
    dm <- dm_from_def(def)
    dm_validate(dm)
    is_zoomed(dm)
  })
})

test_that("can upgrade zoomed from v2", {
  dm_v2_zoomed <- readRDS(test_path("dm/v2_zoomed.rds"))
  expect_snapshot({
    def <- dm_get_def(dm_v2_zoomed, quiet = TRUE)
    def <- dm_get_def(dm_v2_zoomed)
    dm <- dm_from_def(def, zoomed = TRUE)
    dm_validate(dm)
    is_zoomed(dm)
  })
})

test_that("can upgrade from v3", {
  if (FALSE) {
    # TODO:
    # - Run this code with a version of dm before the format change
    # - Search for attr(x, "version") and change
    saveRDS(dm_for_filter(), "tests/testthat/dm/v3.rds", version = 2)
    saveRDS(dm_for_filter() %>% dm_zoom_to(tf_2), "tests/testthat/dm/v3_zoomed.rds", version = 2)
  }

  dm_v3 <- readRDS(test_path("dm/v3.rds"))
  expect_snapshot({
    def <- dm_get_def(dm_v3, quiet = TRUE)
    def <- dm_get_def(dm_v3)
    dm <- dm_from_def(def)
    dm_validate(dm)
    is_zoomed(dm)
    dm_get_all_uks(dm_v3)
    dm_get_all_pks(dm_v3)
    dm_get_all_fks(dm_v3)
  })
})

test_that("can upgrade zoomed from v3", {
  dm_v3_zoomed <- readRDS(test_path("dm/v3_zoomed.rds"))
  expect_snapshot({
    def <- dm_get_def(dm_v3_zoomed, quiet = TRUE)
    def <- dm_get_def(dm_v3_zoomed)
    dm <- dm_from_def(def, zoomed = TRUE)
    dm_validate(dm)
    is_zoomed(dm)
  })
})

test_that("can upgrade to v4", {
  if (FALSE) {
    saveRDS(dm_for_filter(), "tests/testthat/dm/v4.rds", version = 2)
    saveRDS(dm_for_filter() %>% dm_zoom_to(tf_2), "tests/testthat/dm/v4_zoomed.rds", version = 2)
  }

  dm_v4 <- readRDS(test_path("dm/v4.rds"))
  expect_snapshot({
    def <- dm_get_def(dm_v4, quiet = TRUE)
    def <- dm_get_def(dm_v4)
    dm <- dm_from_def(def)
    dm_validate(dm)
    is_zoomed(dm)
    dm_get_all_uks(dm_v4)
    dm_get_all_pks(dm_v4)
    dm_get_all_fks(dm_v4)
  })
})

test_that("can upgrade zoomed to v4", {
  dm_v4_zoomed <- readRDS(test_path("dm/v4_zoomed.rds"))
  expect_snapshot({
    def <- dm_get_def(dm_v4_zoomed, quiet = TRUE)
    def <- dm_get_def(dm_v4_zoomed)
    dm <- dm_from_def(def, zoomed = TRUE)
    dm_validate(dm)
    is_zoomed(dm)
  })
})

test_that("Check assumptions on duckplyr class", {
  skip_if_not_installed("duckplyr", "1.1.99")
  expect_true(inherits(duckplyr::duckdb_tibble(a = 1), "duckplyr_df"))
})

test_that("Simple duckplyr test", {
  skip_if_not_installed("duckplyr", "1.1.99")

  expect_snapshot({
    dm(a = duckplyr::duckdb_tibble(x = 1:3, .prudence = "stingy")) |>
      dm_zoom_to(a) |>
      summarize(mean_x = mean(x))
  })
})

test_that("`pack_join()` works", {
  df1 <- tibble(col1 = 1:2, key = letters[1:2])
  df2 <- tibble(col2 = 3:4, key = letters[1:2])
  df3 <- tibble(col3 = 3:4, key3 = letters[1:2])
  expect_snapshot(pack_join(df1, df2))
  expect_snapshot(pack_join(df1, df2, name = "packed_col"))
  expect_snapshot(pack_join(df1, df3, by = c(key = "key3")))
  expect_snapshot(pack_join(df1, df3, by = c(key = "key3"), keep = TRUE))
})

test_that("`pack_join()` works with remote table", {
  df1 <- tibble(col1 = 1:2, key = letters[1:2])

  # fails with remote table
  dm_remote <- dm_for_filter_duckdb()
  expect_snapshot_error(pack_join(df1, dm_remote$tf_1, by = c(col1 = "a")))
  # unless copy = TRUE
  expect_snapshot(pack_join(df1, dm_remote$tf_1, by = c(col1 = "a"), copy = TRUE))

  # when we have conflicting columns, the column in x is overwritten silently
  # consistent with dplyr::nest_join
  df4 <- tibble(df5 = integer())
  df5 <- tibble(col = integer())
  expect_snapshot(pack_join(df4, df5, by = c(df5 = "col")))

  # No conflict occurs when packing `y` before the join
  df6 <- tibble(df6 = integer())
  expect_snapshot(pack_join(df5, df6, by = c(col = "df6")))
})

test_that("`pack_join()` works with dm_zoomed", {
  dm_nyc <- dm_nycflights13()

  name_packed_df <- "packed_flights"
  by_column <- "carrier"

  dm_nyc_new <- dm_nyc %>%
    dm_zoom_to(airlines) %>%
    pack_join(flights, by = by_column, name = name_packed_df) %>%
    dm_update_zoomed()

  # key relations should stay the same after packing
  expect_equal(get_all_keys(dm_nyc), get_all_keys(dm_nyc_new))

  # the new table should have only one additional (packed) column
  expect_equal(colnames(dm_nyc_new$airlines), c(colnames(dm_nyc$airlines), name_packed_df))

  # the packed table should have the same number of rows as the unpacked one
  expect_equal(nrow(dm_nyc_new$airlines$packed_flights), nrow(dm_nyc$flights))

  # but it should have fewer columns
  expect_equal(
    setdiff(colnames(dm_nyc$flights), colnames(dm_nyc_new$airlines$packed_flights)),
    by_column
  )
})

test_that("basic test: 'unite()'-methods work", {
  # see issue #361
  skip_if_remote_src()
  expect_equivalent_tbl(
    dm_zoomed() %>%
      unite("new_col", c, e) %>%
      tbl_zoomed(),
    unite(tf_2(), "new_col", c, e)
  )

  expect_equivalent_tbl(
    dm_for_filter() %>%
      pull_tbl(tf_2, keyed = TRUE) %>%
      unite("new_col", c, e) %>%
      unclass_keyed_tbl(),
    unite(tf_2(), "new_col", c, e)
  )

  expect_dm_error(
    unite(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("basic test: 'separate()'-methods work", {
  skip_if_remote_src()
  expect_equivalent_tbl(
    dm_zoomed() %>%
      unite("new_col", c, e) %>%
      separate("new_col", c("c", "e")) %>%
      select(c, d, e, e1) %>%
      tbl_zoomed(),
    tf_2()
  )

  expect_equivalent_tbl(
    dm_for_filter() %>%
      pull_tbl(tf_2, keyed = TRUE) %>%
      unite("new_col", c, e) %>%
      separate("new_col", c("c", "e")) %>%
      select(c, d, e, e1),
    dm_for_filter() %>%
      pull_tbl(tf_2, keyed = TRUE)
  )

  expect_dm_error(
    separate(dm_for_filter()),
    "only_possible_w_zoom"
  )
})

test_that("key tracking works", {
  skip_if_remote_src()
  expect_snapshot({
    dm_zoomed() %>%
      unite("new_col", c, e) %>%
      dm_update_zoomed() %>%
      get_all_keys()

    dm_zoomed() %>%
      unite("new_col", c, e, remove = FALSE) %>%
      dm_update_zoomed() %>%
      get_all_keys()

    dm_zoomed() %>%
      unite("new_col", c, e, remove = FALSE) %>%
      dm_update_zoomed() %>%
      dm_add_fk(tf_2, new_col, tf_6) %>%
      dm_zoom_to(tf_2) %>%
      separate(new_col, c("c", "e"), remove = TRUE) %>%
      dm_update_zoomed() %>%
      get_all_keys()

    dm_zoomed() %>%
      unite("new_col", c, e, remove = FALSE) %>%
      dm_update_zoomed() %>%
      dm_add_fk(tf_2, new_col, tf_6) %>%
      dm_zoom_to(tf_2) %>%
      separate(new_col, c("c", "e"), remove = FALSE) %>%
      dm_update_zoomed() %>%
      get_all_keys()
  })
})


# tests for compound keys -------------------------------------------------

test_that("output for compound keys", {
  # FIXME: COMPOUND: Need proper test
  skip_if_remote_src()

  expect_snapshot({
    unite_weather_dm <-
      nyc_comp() %>%
      dm_zoom_to(weather) %>%
      mutate(chr_col = "airport") %>%
      unite("new_col", origin, chr_col) %>%
      dm_update_zoomed()
    unite_weather_dm %>% get_all_keys()
    unite_weather_dm %>% get_all_keys()
    unite_flights_dm <-
      nyc_comp() %>%
      dm_zoom_to(flights) %>%
      mutate(chr_col = "airport") %>%
      unite("new_col", origin, chr_col) %>%
      dm_update_zoomed()
    unite_flights_dm %>% get_all_keys()
    unite_flights_dm %>% get_all_keys()
    nyc_comp() %>%
      dm_zoom_to(weather) %>%
      separate(origin, c("o1", "o2"), sep = "^..", remove = TRUE) %>%
      dm_update_zoomed()
    nyc_comp() %>%
      dm_zoom_to(weather) %>%
      separate(origin, c("o1", "o2"), sep = "^..", remove = FALSE) %>%
      dm_update_zoomed()
  })
})

# Signature alignment tests ------------------------------------------------

test_that("dm tidyr method signatures match tidyr data.frame method signatures", {
  skip_on_cran()

  tidyr_ns <- asNamespace("tidyr")
  dm_ns <- asNamespace("dm")

  verbs <- c("unite", "separate")

  for (verb in verbs) {
    df_method <- tryCatch(
      get(paste0(verb, ".data.frame"), envir = tidyr_ns),
      error = function(e) NULL
    )
    if (is.null(df_method)) {
      next
    }

    df_args <- names(formals(df_method))

    for (cls in c("dm", "dm_zoomed", "dm_keyed_tbl")) {
      method_name <- paste0(verb, ".", cls)
      dm_method <- tryCatch(
        get(method_name, envir = dm_ns),
        error = function(e) NULL
      )
      if (is.null(dm_method)) {
        next
      }

      dm_args <- names(formals(dm_method))
      missing_args <- setdiff(df_args, dm_args)
      expect_true(
        length(missing_args) == 0,
        label = paste0(
          method_name,
          " is missing args from ",
          verb,
          ".data.frame: ",
          paste(missing_args, collapse = ", ")
        )
      )
    }
  }
})

#' Pack Join
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `pack_join()` returns all rows and columns in `x` with a new packed column
#' that contains all matches from `y`.
#' @inheritParams dplyr::nest_join
#' @param x,y A pair of data frames or data frame extensions (e.g. a tibble).
#'
#' @export
#' @seealso [dplyr::nest_join()], [tidyr::pack()]
#' @examples
#' df1 <- tibble::tibble(x = 1:3)
#' df2 <- tibble::tibble(x = c(1, 1, 2), y = c("first", "second", "third"))
#' pack_join(df1, df2)
pack_join <- function(x, y, by = NULL, ..., copy = FALSE, keep = FALSE, name = NULL) {
  UseMethod("pack_join")
}

#' @export
pack_join.dm <- function(x, ...) {
  check_zoomed(x)
}

#' @rdname pack_join
#' @export
pack_join.dm_zoomed <- function(x, y, by = NULL, ..., copy = FALSE, keep = FALSE, name = NULL) {
  check_dots_empty()

  y_name <- dm_tbl_name(x, {{ y }})
  zoomed <- dm_get_zoom(x, c("table", "zoom", "col_tracker_zoom"))
  x_tbl <- zoomed$zoom[[1]]
  y_tbl <- dm_get_tables_impl(x)[[y_name]]

  joined_tbl <- pack_join(x_tbl, y_tbl, by, ..., copy = copy, keep = keep, name = name)
  replace_zoomed_tbl(x, joined_tbl)
}

#' @export
pack_join.data.frame <- function(x, y, by = NULL, ..., copy = FALSE, keep = FALSE, name = NULL) {
  check_dots_empty()
  name_var <- name %||% as_label(enexpr(y))
  if (!copy && inherits(y, "tbl_lazy")) {
    cli::cli_abort(
      "{.arg x} and {.arg y} must share the same src, set {.code copy = TRUE} (may be slow)."
    )
  }
  y_local <- collect(y)
  x_nms <- colnames(x)
  name_var_unique <- last(make.unique(c(names(y_local), x_nms, name_var)))
  y_keys <- unname(by) %||% intersect(names(x), names(y_local))
  y_packed <- tidyr::pack(y_local, !!name_var_unique := -all_of(y_keys))
  if (keep) {
    y_packed[[name_var_unique]][y_keys] <- y_packed[y_keys]
    # sort packed cols in original order
    y_packed[[name_var_unique]] <- y_packed[[name_var_unique]][(names(y_local))]
  }
  joined <- left_join(x, y_packed, by = by, copy = copy, keep = FALSE, multiple = "all")
  # overwrite existing column silently in x if collision, not very safe but consistent with dplyr::nest_join
  if (name_var %in% x_nms) {
    joined[[name_var]] <- NULL
  }
  rename(joined, !!name_var := !!name_var_unique)
}

# FIXME: Dynamic lookup by parsing https://svn.r-project.org/R/tags/
get_deps <- function() {
  # Determine package dependencies
  if (!requireNamespace("desc", quietly = TRUE)) {
    install.packages("desc")
  }

  deps_df <- desc::desc_get_deps()
  deps_df_optional <- deps_df$package[deps_df$type %in% c("Suggests", "Enhances")]
  deps_df_hard <- deps_df$package[deps_df$type %in% c("Depends", "Imports", "LinkingTo")]
  deps_df_base <- unlist(tools::standard_package_names(), use.names = FALSE)

  packages <- sort(deps_df_optional)
  packages <- intersect(packages, rownames(available.packages()))

  # Too big to fail, or can't be avoided:
  off_limits <- c("testthat", "rmarkdown", "rcmdcheck", deps_df_hard, deps_df_base)
  off_limits_dep <- unlist(tools::package_dependencies(off_limits, recursive = TRUE, which = "strong"))
  setdiff(packages, c(off_limits, off_limits_dep))
}

if (Sys.getenv("GITHUB_BASE_REF") != "") {
  print(Sys.getenv("GITHUB_BASE_REF"))
  system("git fetch origin ${GITHUB_BASE_REF}")
  # Use .. to avoid having to fetch the entire history
  # https://github.com/krlmlr/actions-sync/issues/45
  has_diff <- (system("git diff origin/${GITHUB_BASE_REF}.. | egrep '^[+][^+]' | grep -q ::") == 0)
  if (has_diff) {
    system("git diff origin/${GITHUB_BASE_REF}.. | egrep '^[+][^+]' | grep -q ::")
    packages <- get_deps()
  } else {
    writeLines("No changes using :: found, not checking without suggested packages")
    packages <- character()
  }
} else {
  packages <- get_deps()
}

if (length(packages) > 0) {
  json <- paste0(
    '{"package":[',
    paste0('"', packages, '"', collapse = ","),
    ']}'
  )
  writeLines(paste0("matrix=", json), Sys.getenv("GITHUB_OUTPUT"))
  writeLines(json)
} else {
  writeLines("No suggested packages found.")
}

# Only run if the top level call is devtools::test() or testthat::test_check()
if (is_this_a_test()) {
  # trim both the print output of `tbl_df` and `tbl_sql` so they match
  # FIXME: what if a table has more than 10 rows?
  if (rlang::is_installed("dbplyr")) {
    tbl_sum.tbl_sql <- function(x, ...) c()
    s3_register("tibble::tbl_sum", "tbl_sql", tbl_sum.tbl_sql)
    s3_register("pillar::tbl_sum", "tbl_sql", tbl_sum.tbl_sql)
    tbl_format_header.tbl_sql <- function(x, ...) invisible()
    s3_register("pillar::tbl_format_header", "tbl_sql", tbl_format_header.tbl_sql)
  }
  # Not sure why this works without s3_register(), and is also required
  tbl_sum.tbl_df <- function(x, ...) c()
}

#' JSON pack
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' A wrapper around [tidyr::pack()] which stores the packed data into JSON columns.
#'
#' @param .data A data frame, a data frame extension (e.g. a tibble), or  a lazy data frame (e.g. from dbplyr or dtplyr).
#' @param .names_sep If `NULL`, the default, the names will be left as is.
#' @param ... <[`tidy-select`][tidyr_tidy_select]> Columns to pack, specified
#'   using name-variable pairs of the form `new_col = c(col1, col2, col3)`.
#'   The right hand side can be any valid tidy select expression.
#' @seealso [tidyr::pack()], [json_pack_join()]
#' @export
#' @examples
#' df <- tibble::tibble(x1 = 1:3, x2 = 4:6, x3 = 7:9, y = 1:3)
#' packed <- json_pack(df, x = c(x1, x2, x3), y = y)
#' packed
json_pack <- function(.data, ..., .names_sep = NULL) {
  UseMethod("json_pack")
}

#' @export
json_pack.data.frame <- function(.data, ..., .names_sep = NULL) {
  check_suggested("jsonlite", "json_pack")
  dot_nms <- ...names()
  tidyr::pack(.data, ..., .names_sep = .names_sep) %>%
    mutate(across(all_of(dot_nms), to_packed_json))
}

to_packed_json <- function(x) {
  con <- textConnection(NULL, open = "w")
  on.exit(close(con))
  jsonlite::stream_out(x, con, digits = NA, verbose = FALSE)
  textConnectionValue(con)
}

#' @export
json_pack.tbl_lazy <- function(.data, ..., .names_sep = NULL) {
  dots <- quos(...)
  if ("" %in% names2(dots)) {
    cli::cli_abort("All elements of {.arg ...} must be named.")
  }

  col_nms <- colnames(.data)
  pack_cols <- purrr::map(dots, ~ tidyselect::vars_select(col_nms, !!.x))
  id_cols <- setdiff(col_nms, unlist(unique(pack_cols)))

  sql_exprs <- purrr::imap(
    pack_cols,
    ~ sql_json_pack(
      dbplyr::remote_con(.data),
      cols = names(.x),
      names_sep = .names_sep,
      packed_col = .y,
      data = .data
    )
  )

  .data %>%
    transmute(!!!syms(id_cols), !!!sql_exprs)
}

sql_json_pack <- function(con, cols, names_sep, packed_col, data, ...) {
  UseMethod("sql_json_pack")
}

#' @export
#' @autoglobal
#' @global JSON_BUILD_OBJECT
sql_json_pack.PqConnection <- function(con, cols, names_sep, packed_col, data, ...) {
  check_dots_empty()

  inside_cols <- remove_prefix_and_sep(cols, prefix = packed_col, sep = names_sep)
  inside_cols_idented <- dbplyr::ident(inside_cols)
  exprs <- vctrs::vec_interleave(as.list(inside_cols_idented), syms(cols))
  dbplyr::translate_sql(JSON_BUILD_OBJECT(!!!exprs), con = con)
}

remove_prefix_and_sep <- function(x, prefix, sep) {
  if (is.null(sep)) {
    return(x)
  }
  prefix_and_sep <- paste0(prefix, sep)
  prefixed_lgl <- startsWith(x, prefix_and_sep)
  # `substr()` rather than `sub()` to avoid escaping special regex chars
  replacements <- substr(x[prefixed_lgl], nchar(prefix_and_sep) + 1, nchar(x[prefixed_lgl]))
  replace(x, prefixed_lgl, replacements)
}


#' @export
`json_pack.tbl_Microsoft SQL Server` <- function(.data, ..., .names_sep = NULL) {
  dots <- quos(...)
  if ("" %in% names2(dots)) {
    cli::cli_abort("All elements of {.arg ...} must be named.")
  }
  con <- dbplyr::remote_con(.data)
  in_query <- dbplyr::sql_render(.data)

  col_nms <- colnames(.data)

  # packing_plan contains information and subqueries, one row for each packing column
  packing_plan <- purrr::imap_dfr(dots, function(quo_selection, packing_name) {
    packed_names <- tidyselect::vars_select(col_nms, !!quo_selection)
    if (!is.null(.names_sep)) {
      new_packed_names <-
        remove_prefix_and_sep(packed_names, prefix = packing_name, sep = .names_sep)
      selected <-
        glue_sql("{`packed_names`} {`new_packed_names`}", .con = con) %>%
        glue_collapse(", ")
    } else {
      new_packed_names <- packed_names
      selected <- glue_sql("{`packed_names`*}", .con = con)
    }

    for_json_path_query <- glue_sql(
      "(SELECT value FROM OPENJSON((SELECT ",
      selected,
      " FOR JSON PATH))) AS {`packing_name`}",
      .con = con
    )

    tibble::tibble_row(
      packing_name = packing_name,
      packed_names = list(packed_names),
      new_packed_names = list(new_packed_names),
      selected = selected,
      for_json_path_query = for_json_path_query,
    )
  })

  id_cols <- setdiff(col_nms, unlist(packing_plan$packed_names))
  temp_alias <- "*tmp*"

  # build final query
  out_query <- glue_sql(
    "SELECT {`id_cols`*}, ",
    glue_collapse(packing_plan$for_json_path_query, ", "),
    " FROM (",
    in_query,
    ") {`temp_alias`}",
    .con = con
  )

  tbl(con, sql(out_query), vars = c(id_cols, packing_plan$packing_name))
}

#' Data model class
#'
#' @description
#' The `dm` class holds a list of tables and their relationships.
#' It is inspired by [datamodelr](https://github.com/bergant/datamodelr),
#' and extends the idea by offering operations to access the data in the tables.
#'
#' `dm()` creates a `dm` object from [tbl] objects
#' (tibbles or lazy data objects).
#'
#' @param ... Tables or existing `dm` objects to add to the `dm` object.
#'   Unnamed tables are auto-named, `dm` objects must not be named.
#' @param .name_repair,.quiet Options for name repair.
#'   Forwarded as `repair` and `quiet` to [vctrs::vec_as_names()].
#'
#' @return For `dm()`, `new_dm()`, `as_dm()`: A `dm` object.
#'
#' @seealso
#'
#' - [dm_from_con()] for connecting to all tables in a database
#'   and importing the primary and foreign keys
#' - [dm_get_tables()] for returning a list of tables
#' - [dm_add_pk()] and [dm_add_fk()] for adding primary and foreign keys
#' - [copy_dm_to()] for DB interaction
#' - [dm_draw()] for visualization
#' - [dm_flatten_to_tbl()] for flattening
#' - [dm_filter()] for filtering
#' - [dm_select_tbl()] for creating a `dm` with only a subset of the tables
#' - [dm_nycflights13()]  for creating an example `dm` object
#' - [decompose_table()] for table surgery
#' - [check_key()] and [check_subset()] for checking for key properties
#' - [examine_cardinality()] for checking the cardinality of the relation between two tables
#'
#' @export
#' @examples
#' dm(trees, mtcars)
#'
#' new_dm(list(trees = trees, mtcars = mtcars))
#'
#' as_dm(list(trees = trees, mtcars = mtcars))
#' @examplesIf rlang::is_installed(c("nycflights13", "dbplyr"))
#'
#' is_dm(dm_nycflights13())
#'
#' dm_nycflights13()$airports
#'
#' dm_nycflights13()["airports"]
#'
#' dm_nycflights13()[["airports"]]
#'
#' dm_nycflights13() %>% names()
dm <- function(
  ...,
  .name_repair = c("check_unique", "unique", "universal", "minimal"),
  .quiet = FALSE
) {
  quos <- enquos(...)
  names <- names2(quos)

  dots <- map(quos, eval_tidy)

  is_dm <- map_lgl(dots, is_dm)

  for (i in which(is_dm)) {
    if (names[[i]] != "") {
      cli::cli_abort(c(
        "All dm objects passed to {.fun dm} must be unnamed.",
        i = paste0("Argument ", i, " has name ", tick(names[[i]]), ".")
      ))
    }

    if (is_zoomed(dots[[i]])) {
      cli::cli_abort(c(
        "All dm objects passed to {.fun dm} must be unzoomed.",
        i = paste0("Argument ", i, " is a zoomed dm.")
      ))
    }
  }

  # FIXME: check not zoomed, prettier
  stopifnot(names2(quos)[is_dm] == "")

  dm_tbl <- dm_impl(dots[!is_dm], names(quos_auto_name(quos[!is_dm])))
  def <- dm_bind_impl(c(dots[is_dm], list(dm_tbl)), .name_repair, .quiet, repair_arg = "")

  # Validation occurs in CI/CD
  dm_from_def(def)
}

dm_impl <- function(tbls, names) {
  names(tbls) <- names
  new_dm(tbls)
}

#' A low-level constructor
#'
#' @description
#' `new_dm()` is a low-level constructor that creates a new `dm` object.
#'
#' - If called without arguments, it will create an empty `dm`.
#'
#' - If called with arguments, no validation checks will be made to ascertain that
#'   the inputs are of the expected class and internally consistent;
#'   use [dm_validate()] to double-check the returned object.
#'
#' @param tables A named list of the tables (tibble-objects, not names),
#'   to be included in the `dm` object.
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' library(dm)
#' library(nycflights13)
#'
#' # using `data.frame` objects
#' new_dm(tibble::lst(weather, airports))
#'
#' # using `dm_keyed_tbl` objects
#' dm <- dm_nycflights13()
#' y1 <- dm$planes %>%
#'   mutate() %>%
#'   select(everything())
#' y2 <- dm$flights %>%
#'   left_join(dm$airlines, by = "carrier")
#'
#' new_dm(list("tbl1" = y1, "tbl2" = y2))
#'
#' @rdname dm
#' @export
new_dm <- function(tables = list()) {
  def <- new_keyed_dm_def(tables)
  dm_from_def(def)
}

new_keyed_dm_def <- function(tables = list()) {
  is_keyed <- map_lgl(unname(tables), is_dm_keyed_tbl)
  stopifnot(!anyDuplicated(names(tables)[is_keyed]))
  if (!any(is_keyed)) {
    return(new_dm_def(tables))
  }

  # data should be saved as a tibble
  unclassed_tables <- map(tables, unclass_keyed_tbl)

  pks_df <- pks_df_from_keys_info(tables[is_keyed])
  uks_df <- uks_df_from_keys_info(tables[is_keyed])
  fks_df <- fks_df_from_keys_info(tables[is_keyed])

  new_dm_def(unclassed_tables, pks_df, uks_df, fks_df)
}


new_dm_def <- function(tables = list(), pks_df = NULL, uks_df = NULL, fks_df = NULL) {
  # Legacy
  data <- unname(tables)
  table <- names2(tables)

  stopifnot(all(pks_df$table %in% table))
  stopifnot(all(uks_df$table %in% table))
  stopifnot(all(fks_df$table %in% table))

  def <- fast_tibble(
    table = table,
    data = data,
    segment = NA_character_,
    display = NA_character_,
    pks = list_of(new_pk()),
    uks = list_of(new_uk()),
    fks = list_of(new_fk()),
    filters = list_of(new_filter()),
    zoom = list(NULL),
    col_tracker_zoom = list(NULL),
    uuid = vec_new_uuid_along(table),
  )

  if (!is.null(pks_df)) {
    def$pks[match(pks_df$table, def$table)] <- map(pks_df$pks, `%||%`, new_pk())
  }
  if (!is.null(uks_df)) {
    def$uks[match(uks_df$table, def$table)] <- map(uks_df$uks, `%||%`, new_pk())
  }
  if (!is.null(fks_df)) {
    def$fks[match(fks_df$table, def$table)] <- map(fks_df$fks, `%||%`, new_pk())
  }

  def
}

dm_from_def <- function(def, zoomed = FALSE, validate = TRUE) {
  if (is.null(def[["uuid"]])) {
    def$uuid <- vec_new_uuid_along(def$table)
  } else {
    missing <- which(is.na(def$uuid))
    if (length(missing) > 0) {
      def$uuid[missing] <- vec_new_uuid_along(missing)
    }
  }

  class <- c(
    if (zoomed) "dm_zoomed",
    "dm"
  )
  out <- structure(list(def = def), class = class, version = 4L)

  # Enable for strict tests (search for INSTRUMENT in .github/workflows):
  # if (validate) { dm_validate(out) } # INSTRUMENT: validate

  out
}

dm_get_def <- function(x, quiet = FALSE) {
  # FIXME: Move that check to callers, for speed
  # Most callers already call it, but not all
  check_dm(x)

  if (!identical(attr(x, "version"), 4L)) {
    x <- dm_upgrade(x, quiet)
  }

  unclass(x)$def
}

new_pk <- function(column = list(), autoincrement = logical(length(column))) {
  stopifnot(is.list(column), is.logical(autoincrement))
  fast_tibble(column = column, autoincrement = autoincrement)
}

new_uk <- function(column = list()) {
  stopifnot(is.list(column))
  fast_tibble(column = column)
}

new_fk <- function(
  ref_column = list(),
  table = character(),
  column = list(),
  on_delete = character()
) {
  stopifnot(
    is.list(column),
    is.list(ref_column),
    length(table) == length(column),
    length(table) == length(ref_column),
    length(on_delete) %in% c(1L, length(table))
  )

  fast_tibble(
    ref_column = ref_column,
    table = table,
    column = column,
    on_delete = on_delete
  )
}

new_filter <- function(quos = list(), zoomed = logical()) {
  fast_tibble(filter_expr = unclass(quos), zoomed = zoomed)
}

unnest_pks <- function(def) {
  # Optimized
  pk_df <- tibble(
    table = rep(def$table, map_int(def$pks, nrow))
  )

  pk_df <- vec_cbind(pk_df, vec_rbind(!!!def$pks))

  # FIXME: Should work better with dplyr 0.9.0
  if (!("column" %in% names(pk_df))) {
    pk_df$column <- character()
  }

  pk_df
}

dm_get_zoom <- function(x, cols = c("table", "zoom"), quiet = FALSE) {
  # Performance
  def <- dm_get_def(x, quiet)
  zoom <- def$zoom
  where <- which(!map_lgl(zoom, is.null))
  if (length(where) != 1) {
    # FIXME: Better error message?
    abort_not_pulling_multiple_zoomed()
  }
  def[where, cols]
}

#' Check class
#'
#' `is_dm()` returns `TRUE` if the input is of class `dm`.
#'
#' @param x An object.
#'
#' @rdname dm
#'
#' @return For `is_dm()`: A scalar logical, `TRUE` if is this object is a `dm`.
#'
#' @export
is_dm <- function(x) {
  inherits(x, "dm")
}


#' Coerce
#'
#' `as_dm()` coerces objects to the `dm` class
#'
#' @rdname dm
#' @export
as_dm <- function(x, ...) {
  UseMethod("as_dm")
}

#' @export
as_dm.default <- function(x, ...) {
  check_dots_empty()

  if (!is.list(x) || is.object(x)) {
    cli::cli_abort("Can't coerce {.cls {class(x)[[1]]}} to {.cls dm}.")
  }

  # Automatic name repair
  names(x) <- vec_as_names(names2(x), repair = "unique")
  new_dm(x)
}

tbl_src <- function(x) {
  if (is_empty(x) || is.data.frame(x)) {
    NULL
  } else if (inherits(x, "tbl_sql")) {
    dbplyr::remote_src(x)
  } else {
    abort_what_a_weird_object(class(x)[[1]])
  }
}

#' @export
as_dm.src <- function(x, ...) {
  check_dots_empty()

  dm_from_con(con = con_from_src_or_con(x), table_names = NULL)
}

#' @export
as_dm.DBIConnection <- function(x, ...) {
  check_dots_empty()

  dm_from_con(con = x, table_names = NULL)
}

#' @export
print.dm <- function(x, ...) {
  # for both dm and dm_zoomed
  show_dm(x)
  invisible(x)
}

#' @export
print.dm_zoomed <- function(x, ..., n = NULL, width = NULL, n_extra = NULL) {
  format(x, ..., n = NULL, width = NULL, n_extra = NULL)
}

show_dm <- function(x) {
  def <- dm_get_def(x)
  if (nrow(def) == 0) {
    cat_line("dm()")
    return()
  }

  src <- dm_get_src_impl(x)
  if (!is.null(src)) {
    cat_rule("Table source", col = "green")
    db_info <- NULL

    tbl_str <- pillar::tbl_sum(def$data[[1]])
    if ("Database" %in% names(tbl_str)) {
      db_info <- paste0("src:  ", tbl_str[["Database"]])
    }
    if (is.null(db_info)) {
      db_info <- strsplit(format(src), "\n")[[1]][[1]]
    }

    cat_line(db_info)
  }

  cat_rule("Metadata", col = "violet")

  cat_line("Tables: ", commas(tick(def$table)))
  cat_line("Columns: ", def_get_n_columns(def))
  cat_line("Primary keys: ", def_get_n_pks(def))
  cat_line("Foreign keys: ", def_get_n_fks(def))

  filters <- dm_get_filters_impl(x)
  if (nrow(filters) > 0) {
    cat_rule("Filters", col = "orange")
    walk2(filters$table, filters$filter, ~ cat_line(paste0(.x, ": ", as_label(.y))))
  }
}

#' @export
format.dm <- function(x, ...) {
  # for both dm and dm_zoomed
  def <- dm_get_def(x)
  glue(
    "dm: {def_get_n_tables(def)} tables, {def_get_n_columns(def)} columns, {def_get_n_pks(def)} primary keys, {def_get_n_fks(def)} foreign keys"
  )
}

#' @export
format.dm_zoomed <- function(x, ..., n = NULL, width = NULL, n_extra = NULL) {
  # so far only 1 table can be zoomed on
  dm_zoomed_df <- as_dm_zoomed_df(x)
  cat_line(format(dm_zoomed_df, ..., n = n, width = width, n_extra = n_extra))
  invisible(x)
}

def_get_n_tables <- function(def) {
  nrow(def)
}

def_get_n_columns <- function(def) {
  sum(map_int(map(def$data, colnames), length))
}

def_get_n_pks <- function(def) {
  sum(map_int(def$pks, vec_size))
}

def_get_n_fks <- function(def) {
  sum(map_int(def$fks, vec_size))
}

as_dm_zoomed_df <- function(x) {
  zoomed <- dm_get_zoom(x)

  # for tests
  new_dm_zoomed_df(
    zoomed$zoom[[1]],
    name_df = zoomed$table
  )
}

new_dm_zoomed_df <- function(x, ...) {
  if (is.data.frame(x) && !inherits(x, "duckplyr_df")) {
    # need this in order to avoid star (from rownames, automatic from `structure(...)`)
    # in print method for local tibbles
    return(new_tibble(
      x,
      class = c("dm_zoomed_df", class(x), c("tbl_df", "tbl", "data.frame")),
      nrow = nrow(x),
      ...
    ))
  }

  # Careful to not touch duckplyr row names
  class(x) <- c("dm_zoomed_df", class(x))
  extra_attrs <- list(...)
  for (name in names(extra_attrs)) {
    attr(x, name) <- extra_attrs[[name]]
  }
  x
}

# this is called from `tibble:::trunc_mat()`, which is called from `tibble::format.tbl()`
# therefore, we need to have our own subclass but the main class needs to be `tbl`
#' @export
tbl_sum.dm_zoomed_df <- function(x, ...) {
  c(
    structure(attr(x, "name_df"), names = "Zoomed table"),
    NextMethod()
  )
}

#' @export
format.dm_zoomed_df <- function(x, ..., n = NULL, width = NULL, n_extra = NULL) {
  NextMethod()
}

#' @export
`$.dm` <- function(x, name) {
  # for both dm and dm_zoomed
  table <- dm_tbl_name(x, {{ name }})
  tbl_impl(x, table)
}

#' @export
`$.dm_zoomed` <- function(x, name) {
  name <- ensym(name)
  eval_tidy(quo(`$`(tbl_zoomed(x), !!name)))
}

#' @export
`$<-.dm` <- function(x, name, value) {
  dm_local_error_call()
  abort_update_not_supported()
}

#' @export
`[[.dm` <- function(x, id, ...) {
  check_dots_empty()

  # for both dm and dm_zoomed
  if (is.numeric(id)) {
    id <- src_tbls_impl(x)[id]
  } else {
    id <- as_string(id)
  }
  tbl_impl(x, id, quiet = TRUE)
}

#' @export
`[[.dm_zoomed` <- function(x, id) {
  `[[`(tbl_zoomed(x, quiet = TRUE), id)
}

#' @export
`[[<-.dm` <- function(x, name, value) {
  dm_local_error_call()
  abort_update_not_supported()
}

#' @export
`[.dm` <- function(x, id) {
  if (is.numeric(id)) {
    id <- src_tbls_impl(x)[id]
  }
  id <- as.character(id)
  dm_select_tbl(x, !!!id)
}

#' @export
`[.dm_zoomed` <- function(x, id) {
  # for both dm and dm_zoomed
  `[`(tbl_zoomed(x), id)
}


#' @export
`[<-.dm` <- function(x, name, value) {
  dm_local_error_call()
  abort_update_not_supported()
}

#' @export
names.dm <- function(x) {
  # for both dm and dm_zoomed
  src_tbls_impl(x, quiet = TRUE)
}

#' @export
names.dm_zoomed <- function(x) {
  names(tbl_zoomed(x, quiet = TRUE))
}


#' @export
`names<-.dm` <- function(x, value) {
  dm_local_error_call()
  abort_update_not_supported()
}

#' @export
length.dm <- function(x) {
  # for both dm and dm_zoomed
  length(src_tbls_impl(x, quiet = TRUE))
}

#' @export
length.dm_zoomed <- function(x) {
  length(tbl_zoomed(x, quiet = TRUE))
}

#' @export
`length<-.dm` <- function(x, value) {
  dm_local_error_call()
  abort_update_not_supported()
}

#' @export
#' @autoglobal
str.dm <- function(object, ...) {
  # for both dm and dm_zoomed
  object <-
    dm_get_def(object, quiet = TRUE) %>%
    select(table, pks, fks, filters)
  str(object)
}

#' @export
#' @autoglobal
str.dm_zoomed <- function(object, ...) {
  object <-
    dm_get_def(object, quiet = TRUE) %>%
    mutate(zoom = if_else(map_lgl(zoom, is_null), NA_character_, table)) %>%
    select(zoom, table, pks, fks, filters)
  str(object)
}

keyed_tbl_impl <- function(dm, from) {
  tbl_impl(dm, from, keyed = TRUE)
}

tbl_impl <- function(dm, from, quiet = FALSE, keyed = FALSE) {
  def <- dm_get_def(dm, quiet = quiet)
  idx <- match(from, def$table)
  if (is.na(idx)) {
    abort_table_not_in_dm(from, src_tbls_impl(dm))
  }

  tbl_def_impl(def, idx, keyed)
}

#' @autoglobal
tbl_def_impl <- function(def, idx, keyed) {
  data <- def$data[[idx]]

  if (!keyed) {
    return(data)
  }

  uuid_lookup <- def[c("table", "uuid")]

  pk_def <- def$pks[[idx]]
  if (nrow(pk_def) > 0) {
    pk <- pk_def$column[[1]]
  } else {
    pk <- NULL
  }

  uks <- def$uks[[idx]]

  fks_in_def <-
    def$fks[[idx]] %>%
    left_join(uuid_lookup, by = "table")

  fks_in <- new_fks_in(
    fks_in_def$uuid,
    fks_in_def$column,
    fks_in_def$ref_column
  )

  fks_out_def <-
    map2_dfr(def$uuid, def$fks, ~ tibble(ref_uuid = .x, .y)) %>%
    filter(table == !!def$table[[idx]]) %>%
    select(ref_uuid, ref_column, column)

  fks_out <- new_fks_out(
    fks_out_def$column,
    fks_out_def$ref_uuid,
    fks_out_def$ref_column
  )

  new_keyed_tbl(
    data,
    pk = pk,
    uks = uks,
    fks_in = fks_in,
    fks_out = fks_out,
    uuid = def$uuid[[idx]]
  )
}

dm_get_keyed_tables_impl <- function(dm) {
  def <- dm_get_def(dm)
  tables <- map(seq_along(def$table), ~ tbl_def_impl(def, .x, keyed = TRUE))
  set_names(tables, def$table)
}

src_tbls_impl <- function(dm, quiet = FALSE) {
  dm_get_def(dm, quiet)$table
}

#' Materialize
#'
#' @description
#' `compute()` materializes all tables in a `dm` to new temporary
#' tables on the database.
#'
#' @details
#' Called on a `dm` object, these methods create a copy of all tables in the `dm`.
#' Depending on the size of your data this may take a long time.
#'
#' To create permament tables, first create the database schema using [copy_dm_to()]
#' or [dm_sql()], and then use [dm_rows_append()].
#'
#' @inheritParams dm_get_tables
#' @param ... Passed on to [compute()].
#' @param temporary Must remain `TRUE`.
#' @return A `dm` object of the same structure as the input.
#' @name materialize
#' @export
#' @examplesIf dm:::dm_has_financial() && rlang::is_installed("RSQLite")
#' financial <- dm_financial_sqlite()
#'
#' financial %>%
#'   pull_tbl(districts) %>%
#'   dbplyr::remote_name()
#'
#' # compute() copies the data to new tables:
#' financial %>%
#'   compute() %>%
#'   pull_tbl(districts) %>%
#'   dbplyr::remote_name()
#'
#' # collect() returns a local dm:
#' financial %>%
#'   collect() %>%
#'   pull_tbl(districts) %>%
#'   class()
compute.dm <- function(x, ..., temporary = TRUE) {
  if (!isTRUE(temporary)) {
    cli::cli_abort("{.fun compute.dm} does not support {.code temporary = FALSE}.")
  }

  # for both dm and dm_zoomed
  x %>%
    dm_apply_filters_impl() %>%
    dm_get_def() %>%
    mutate(data = map(data, compute, ...)) %>%
    dm_from_def()
}

#' Materialize
#'
#' `collect()` downloads the tables in a `dm` object as local [tibble]s.
#'
#' @inheritParams dm_examine_constraints
#'
#' @rdname materialize
#' @export
collect.dm <- function(x, ..., progress = NA) {
  # for both dm and dm_zoomed
  x <- dm_apply_filters_impl(x)

  def <- dm_get_def(x)

  ticker <- new_ticker("downloading data", nrow(def), progress = progress)
  def$data <- map(def$data, ticker(collect), ...)
  dm_from_def(def, zoomed = is_zoomed(x))
}

#' @export
collect.dm_zoomed <- function(x, ...) {
  check_dots_empty()

  inform(c(
    "Detaching table from dm.",
    i = "Use `. %>% pull_tbl() %>% collect()` instead to silence this message."
  ))

  collect(pull_tbl(x))
}


# FIXME: what about 'dim.dm()'?
#' @export
dim.dm_zoomed <- function(x) {
  # dm method provided by base
  dim(tbl_zoomed(x, quiet = TRUE))
}

#' @export
dimnames.dm_zoomed <- function(x) {
  # dm method provided by base
  dimnames(tbl_zoomed(x))
}

#' @export
tbl_vars.dm <- function(x) {
  check_zoomed(x)
}

#' @export
tbl_vars.dm_zoomed <- function(x) {
  tbl_vars(tbl_zoomed(x))
}

dm_reset_all_filters <- function(dm) {
  def <- dm_get_def(dm)
  def$filters <- list_of(new_filter())
  dm_from_def(def)
}

all_same_source <- function(tables) {
  # Use `NULL` if `tables` is empty
  first_table <- tables[1][[1]]
  is.null(detect(tables[-1], ~ !same_src(., first_table)))
}

# creates an empty `dm`-object, `src` is defined by implementation of `dm_get_src_impl()`.
empty_dm <- function() {
  dm_from_def(
    tibble(
      table = character(),
      data = list(),
      segment = character(),
      display = character(),
      pks = list_of(new_pk()),
      uks = list_of(new_uk()),
      fks = list_of(new_fk()),
      filters = list_of(new_filter()),
      zoom = list(),
      col_tracker_zoom = list(),
      uuid = character(),
    )
  )
}

#' Retrieve a table
#'
#' @description
#' This generic has methods for both `dm` classes:
#' 1. With `pull_tbl.dm()` you can chose which table of the `dm` you want to retrieve.
#' 1. With `pull_tbl.dm_zoomed()` you will retrieve the zoomed table in the current state.
#'
#' @inheritParams dm_add_pk
#' @param table One unquoted table name for `pull_tbl.dm()`, ignored for `pull_tbl.dm_zoomed()`.
#' @inheritParams dm_get_tables
#'
#' @seealso [dm_deconstruct()] to generate code of the form
#'   `pull_tbl(..., keyed = TRUE)` from an existing `dm` object.
#'
#' @return The requested table.
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' # For an unzoomed dm you need to specify the table to pull:
#' dm_nycflights13() %>%
#'   pull_tbl(airports)
#'
#' # If zoomed, pulling detaches the zoomed table from the dm:
#' dm_nycflights13() %>%
#'   dm_zoom_to(airports) %>%
#'   pull_tbl()
#' @export
pull_tbl <- function(dm, table, ..., keyed = FALSE) {
  UseMethod("pull_tbl")
}

#' @export
pull_tbl.dm <- function(dm, table, ..., keyed = FALSE) {
  check_dots_empty()
  dm_local_error_call()

  # for both dm and dm_zoomed
  # FIXME: shall we issue a special error in case someone tries sth. like: `pull_tbl(dm_for_filter, c(t4, t3))`?
  table_name <- as_string(enexpr(table))
  if (table_name == "") {
    abort_no_table_provided()
  }
  tbl_impl(dm, table_name, keyed = keyed)
}

#' @export
pull_tbl.dm_zoomed <- function(dm, table, ..., keyed = FALSE) {
  dm_local_error_call()

  if (isTRUE(keyed)) {
    cli::cli_abort(
      "{.code keyed = TRUE} not supported for zoomed dm objects.",
      call = dm_error_call()
    )
  }

  check_dots_empty()

  table_name <- as_string(enexpr(table))
  zoomed <- dm_get_zoom(dm)
  if (table_name == "") {
    if (nrow(zoomed) == 1) {
      zoomed$zoom[[1]]
    } else {
      abort_not_pulling_multiple_zoomed()
    }
  } else if (!(table_name %in% zoomed$table)) {
    abort_table_not_zoomed(table_name, zoomed$table)
  } else {
    zoomed$zoom[[1]]
  }
}

#' @export
as.list.dm <- function(x, ...) {
  check_dots_empty()

  # for both dm and dm_zoomed
  dm_get_tables_impl(x)
}

#' @export
as.list.dm_zoomed <- function(x, ...) {
  check_dots_empty()

  as.list(tbl_zoomed(x))
}

#' Get a glimpse of your `dm` object
#'
#' @inheritParams dm_get_tables
#' @param width Controls the maximum number of columns on a line used in
#'   printing. If `NULL`, `getOption("width")` will be consulted.
#' @param ... Passed to [pillar::glimpse()].
#'
#' @description
#' `glimpse()` provides an overview (dimensions, column data types, primary
#' keys, etc.) of all tables included in the `dm` object. It will additionally
#' print details about outgoing foreign keys for the child table.
#'
#' `glimpse()` is provided by the pillar package, and re-exported by \pkg{dm}.
#'  See [pillar::glimpse()] for more details.
#'
#' @examples
#'
#' dm_nycflights13() %>% glimpse()
#'
#' dm_nycflights13() %>%
#'   dm_zoom_to(flights) %>%
#'   glimpse()
#'
#' @export
glimpse.dm <- function(x, width = NULL, ...) {
  glimpse_width <- width %||% getOption("width")
  table_names_list <- names(dm_get_tables_impl(x))

  print_glimpse_table_meta(x, glimpse_width)
  if (!is_empty(table_names_list)) {
    print_rule_between_tables()
  }
  walk(table_names_list, ~ print_glimpse_table(x, .x, glimpse_width, ...))

  invisible(x)
}

#' @rdname glimpse.dm
#' @export
glimpse.dm_zoomed <- function(x, width = NULL, ...) {
  glimpse_width <- width %||% getOption("width")

  table_name <- dm_get_zoom(x)$table[[1]]

  print_glimpse_table_meta(x, glimpse_width)
  print_glimpse_table(x, table_name, glimpse_width, ...)

  invisible(x)
}

#' Print details about all tables included in the `dm` object (zoomed or not)
#' @keywords internal
#' @noRd
print_glimpse_table_meta <- function(x, width) {
  table_list <- dm_get_tables_impl(x)

  if (length(table_list) == 0) {
    cat_line(trim_width("dm of 0 tables", width))
    return(invisible(x))
  }

  cat_line(
    trim_width(
      paste0("dm of ", length(table_list), " tables: ", toString(tick(names(table_list)))),
      width
    )
  )
}

#' @keywords internal
#' @noRd
print_rule_between_tables <- function() {
  cat("\n")
  cat_rule()
}


#' Print glimpse for a single table included in the `dm` object (zoomed or not)
#' @keywords internal
#' @noRd
print_glimpse_table <- function(x, table_name, width, ...) {
  if (is_zoomed(x)) {
    table <- dm_get_zoom(x)$zoom[[1]]
  } else {
    table <- x[[table_name]]
  }

  # `print_glimpse_table_meta()` is not part of this because it needs to be
  # printed only once for the entire object
  print_glimpse_table_name(x, table_name, width)
  print_glimpse_table_pk(x, table_name, width)
  print_glimpse_table_fk(x, table_name, width)
  # emtpy line to demarcate clearly information about keys and the glimpse info
  cat("\n")
  glimpse(table, width = width, ...)
  # in case the object is not zoomed, the following will help visually
  # distinguish between glimpse outputs for individual tables
  if (!is_zoomed(x)) print_rule_between_tables()
}

#' Print table name for a given table in the `dm` object (zoomed or not)
#' @keywords internal
#' @noRd
print_glimpse_table_name <- function(x, table_name, width) {
  if (is_zoomed(x)) {
    cat_line("\n", trim_width(paste0("Zoomed table: ", tick(table_name)), width))
  } else {
    cat_line("\n", trim_width(paste0("Table: ", tick(table_name)), width))
  }
}

#' Print details about primary key for a given table in the `dm` object (zoomed or not)
#' @keywords internal
#' @noRd
print_glimpse_table_pk <- function(x, table_name, width) {
  pk <- dm_get_pk_impl(x, table_name)

  # anticipate that some key columns could have been removed by the user
  if (is_zoomed(x)) {
    pk <-
      update_zoomed_pk(x) %>%
      pull(column)
  }

  pk <- pk %>%
    map_chr(~ collapse_key_names(.x))

  if (!is_empty(pk)) {
    # FIXME: needs to change if #622 is solved
    cat_line(trim_width(paste0("Primary key: ", pk), width))
  }
}

collapse_key_names <- function(keys, tab = FALSE) {
  tab <- ifelse(tab, "  ", "")
  if (length(keys) > 1L) {
    paste0(tab, "(", paste0(tick(keys), collapse = ", "), ")")
  } else {
    paste0(tab, tick(keys), collapse = ", ")
  }
}

#' Print details about foreign keys for a given table in the `dm` object (zoomed or not)
#' @keywords internal
#' @noRd
print_glimpse_table_fk <- function(x, table_name, width) {
  all_fks <- if (is_zoomed(x)) {
    # anticipate that some key columns could have been removed/renamed by the user
    update_zoomed_fks(x, table_name, col_tracker_zoomed(x))
  } else {
    dm_get_all_fks_impl(x)
  }

  fk <- all_fks %>%
    filter(child_table == !!table_name) %>%
    select(-child_table) %>%
    pmap_chr(
      function(child_fk_cols, parent_table, parent_key_cols, on_delete) {
        trim_width(
          paste0(
            collapse_key_names(child_fk_cols, tab = TRUE),
            " -> ",
            collapse_key_names(paste0(parent_table, "$", parent_key_cols)),
            " ",
            on_delete
          ),
          width
        )
      }
    )

  if (!is_empty(fk)) {
    cat_line(length(fk), " outgoing foreign key(s):")
    cat_line(fk)
  }
}

test_that("dm_bind() works?", {
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_dm(
    dm_bind(dm_for_filter()),
    dm_for_filter()
  )

  expect_equivalent_dm(
    dm_bind(dm_for_filter(), dm_for_flatten(), dm_for_disambiguate()),
    bind_rows(
      dm_get_def(dm_for_filter()),
      dm_get_def(dm_for_flatten()),
      dm_get_def(dm_for_disambiguate())
    ) %>%
      dm_from_def()
  )
})

test_that("are empty_dm() and empty ellipsis handled correctly?", {
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_dm(
    dm_bind(empty_dm()),
    empty_dm()
  )

  expect_equivalent_dm(
    dm_bind(empty_dm(), empty_dm(), empty_dm()),
    empty_dm()
  )

  expect_equivalent_dm(
    dm_bind(),
    empty_dm()
  )
})

test_that("errors: duplicate table names, src mismatches", {
  local_options(lifecycle_verbosity = "warning")

  expect_snapshot(error = TRUE, {
    dm_bind(dm_for_filter(), dm_for_flatten(), dm_for_filter())
  })
})

test_that("errors: src mismatches", {
  local_options(lifecycle_verbosity = "quiet")

  skip_if_not(getRversion() >= "4.0")
  expect_dm_error(dm_bind(dm_for_flatten(), dm_for_filter_duckdb()), "not_same_src")
})

test_that("auto-renaming works", {
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_dm(
    expect_name_repair_message(
      dm_bind(dm_for_filter(), dm_for_flatten(), dm_for_filter(), repair = "unique")
    ),
    bind_rows(
      dm_get_def(
        dm_rename_tbl(
          dm_for_filter(),
          tf_1...1 = tf_1,
          tf_2...2 = tf_2,
          tf_3...3 = tf_3,
          tf_4...4 = tf_4,
          tf_5...5 = tf_5,
          tf_6...6 = tf_6
        )
      ),
      dm_get_def(dm_for_flatten()),
      dm_get_def(dm_rename_tbl(
        dm_for_filter(),
        tf_1...12 = tf_1,
        tf_2...13 = tf_2,
        tf_3...14 = tf_3,
        tf_4...15 = tf_4,
        tf_5...16 = tf_5,
        tf_6...17 = tf_6
      ))
    ) %>%
      dm_from_def()
  )

  expect_silent(
    dm_bind(dm_for_filter(), dm_for_flatten(), dm_for_filter(), repair = "unique", quiet = TRUE)
  )
})

test_that("test error output for src mismatches", {
  local_options(lifecycle_verbosity = "warning")

  skip_if_not(getRversion() >= "4.0")

  expect_snapshot({
    writeLines(conditionMessage(expect_error(
      dm_bind(dm_for_flatten(), dm_for_filter_duckdb())
    )))
  })
})

test_that("output", {
  local_options(lifecycle_verbosity = "warning")

  expect_snapshot({
    dm_bind()
    dm_bind(empty_dm())
    dm_bind(dm_for_filter()) %>% collect()
    dm_bind(dm_for_filter(), dm_for_flatten(), dm_for_filter(), repair = "unique", quiet = TRUE) %>%
      collect()
    writeLines(conditionMessage(expect_error(
      dm_bind(dm_for_filter(), dm_for_flatten(), dm_for_filter())
    )))
  })

  expect_snapshot({
    dm_bind(dm_for_filter(), dm_for_flatten(), dm_for_filter(), repair = "unique") %>% collect()
  })
})

test_that("output for compound keys", {
  local_options(lifecycle_verbosity = "warning")

  expect_snapshot({
    dm_bind(dm_for_filter(), dm_for_flatten()) %>% dm_paste(options = c("select", "keys"))
    dm_bind(dm_for_flatten(), dm_for_filter()) %>% dm_paste(options = c("select", "keys"))
  })

  expect_snapshot({
    dm_bind(dm_for_flatten(), dm_for_flatten(), repair = "unique") %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("validator is silent", {
  expect_identical(
    expect_silent(dm_validate(new_dm())),
    empty_dm()
  )

  expect_identical(
    expect_silent(dm_validate(dm_for_filter_w_cycle())),
    dm_for_filter_w_cycle()
  )

  # Corner case
  expect_silent(
    dm(a = tibble(x = 1)) %>%
      dm_add_pk(a, x) %>%
      dm_validate()
  )
})

test_that("validator speaks up when something's wrong", {
  # col tracker of non-zoomed dm contains entries
  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(col_tracker_zoom = list(1)) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )

  # zoom column of `dm_zoomed` is empty
  expect_dm_error(
    dm_for_filter() %>%
      dm_zoom_to(tf_1) %>%
      dm_get_def() %>%
      mutate(zoom = list(NULL)) %>%
      dm_from_def(zoomed = TRUE) %>%
      dm_validate(),
    "dm_invalid"
  )

  # col tracker of zoomed dm is empty
  expect_dm_error(
    dm_for_filter() %>%
      dm_zoom_to(tf_1) %>%
      dm_get_def() %>%
      mutate(col_tracker_zoom = list(NULL)) %>%
      dm_from_def(zoomed = TRUE) %>%
      dm_validate(),
    "dm_invalid"
  )

  # table name is missing
  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(table = "") %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )

  # zoom column of un-zoomed dm contains a (nonsensical) entry
  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(zoom = list(1)) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )

  # zoom column of a zoomed dm contains a nonsensical entry
  expect_dm_error(
    dm_for_filter() %>%
      dm_zoom_to(tf_1) %>%
      dm_get_def() %>%
      mutate(zoom = if_else(table == "tf_1", list(1), list(NULL))) %>%
      dm_from_def(zoomed = TRUE) %>%
      dm_validate(),
    "dm_invalid"
  )

  # zoom column of a zoomed dm contains more than one entry
  expect_dm_error(
    dm_for_filter() %>%
      dm_zoom_to(tf_1) %>%
      dm_get_def() %>%
      mutate(zoom = list(tf_1)) %>%
      dm_from_def(zoomed = TRUE) %>%
      dm_validate(),
    "dm_invalid"
  )

  # data column of un-zoomed dm contains non-tibble entries
  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(data = list(1, 2, 3, 4, 5, 6)) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )

  # PK metadata wrong (colname doesn't exist)
  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(pks = if_else(table == "tf_1", list_of(new_pk(list("z"))), pks)) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )

  # FK metadata wrong (table doesn't exist)
  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(
        fks = if_else(
          table == "tf_3",
          list_of(new_fk(
            ref_column = list("y"),
            table = "tf_8",
            column = list("z"),
            on_delete = "no_action"
          )),
          fks
        )
      ) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )

  # FK metadata wrong (`column` doesn't exist)
  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(
        fks = if_else(
          table == "tf_1",
          list_of(new_fk(list("a"), "tf_2", list("z"), "no_action")),
          fks
        )
      ) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )

  # FK metadata wrong (`ref_column` doesn't exist)
  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(
        fks = if_else(
          table == "tf_1",
          list_of(new_fk(list("z"), "tf_2", list("d"), "no_action")),
          fks
        )
      ) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )

  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(
        fks = vctrs::as_list_of(map2(
          fks,
          table,
          ~ if (.y == "tf_2") {
            NULL
          } else {
            .x
          }
        ))
      ) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )

  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(
        pks = vctrs::as_list_of(map2(
          pks,
          table,
          ~ if (.y == "tf_2") {
            NULL
          } else {
            .x
          }
        ))
      ) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )

  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(
        filters = vctrs::as_list_of(map2(
          filters,
          table,
          ~ if (.y == "tf_2") {
            NULL
          } else {
            .x
          }
        ))
      ) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )
})

test_that("validator speaks up (sqlite())", {
  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(data = if_else(table == "tf_1", list(dm_for_filter_duckdb()$tf_1), data)) %>%
      dm_from_def() %>%
      dm_validate(),
    "dm_invalid"
  )
})

test_that("validator accepts tbl_sql which is not tbl_dbi (from copy_inline())", {
  skip_if_src("df")

  con <- my_test_con()
  test_table <- dbplyr::copy_inline(con, tibble(a = 1:3))
  # as of dbplyr v2.2.1, `copy_inline()` creates an object that's `tbl_sql`, but not `tbl_dbi`
  testthat::expect_false(inherits(test_table, "tbl_dbi"))
  expect_silent(dm_validate(dm(test_table)))
})

test_that("Standard learning from MSSQL (schema 'dbo') or Postgres (schema 'public') and get_src_tbl_names() works?", {
  skip_if_schema_not_supported()
  skip_if(identical(Sys.getenv("R_COVR"), "true"))

  # dm_learn_from_mssql() --------------------------------------------------
  con_db <- my_test_con()

  # create an object on the MSSQL-DB that can be learned
  withr::defer(
    try(walk(
      remote_tbl_names,
      ~ try(DBI::dbExecute(con_db, paste0("DROP TABLE ", .x)))
    ))
  )

  dm_for_filter_copied <- copy_dm_to(
    con_db,
    dm_for_filter(),
    temporary = FALSE,
    table_names = ~ unique_db_table_name(.x)
  )
  order_of_deletion <- c("tf_2", "tf_1", "tf_5", "tf_6", "tf_4", "tf_3")

  remote_tbl_names <-
    map_chr(
      dm_get_tables(dm_for_filter_copied)[order_of_deletion],
      dbplyr::remote_name
    ) %>%
    DBI::SQL() %>%
    DBI::dbUnquoteIdentifier(conn = con_db) %>%
    map_chr(~ .x@name[[length(.x@name)]])

  remote_tbl_map <- set_names(remote_tbl_names, gsub("^(tf_.).*$", "\\1", remote_tbl_names))

  expect_silent(dm_db_learned_all <- dm_from_con(con_db, learn_keys = TRUE))

  # Select and fix table names
  dm_db_learned <-
    dm_db_learned_all %>%
    dm_select_tbl(!!!remote_tbl_map)

  expect_equivalent_dm(
    dm_db_learned,
    dm_for_filter()[order_of_deletion],
    ignore_on_delete = FALSE,
    ignore_autoincrement = FALSE
  )

  # learning without keys:
  expect_silent(dm_db_learned_no_keys <- dm_from_con(con_db, learn_keys = FALSE))

  # for learning from DB without learning the key relations
  dm_for_filter_no_keys <-
    dm_for_filter()[order_of_deletion] %>%
    dm_get_def() %>%
    mutate(
      pks = list_of(new_pk()),
      fks = list_of(new_fk())
    ) %>%
    dm_from_def()

  # Select and fix table names
  dm_db_learned_no_keys <-
    dm_db_learned_no_keys %>%
    dm_select_tbl(!!!remote_tbl_map)

  expect_equivalent_dm(
    dm_db_learned_no_keys,
    dm_for_filter_no_keys
  )
})


test_that("Learning from specific schema works?", {
  skip_if_schema_not_supported()

  # produces a randomized schema name with a length of 4-10 characters
  # consisting of the symbols in `reservoir`
  random_schema <- function() {
    reservoir <- c(letters)
    how_long <- 10
    paste0(reservoir[sample(seq_len(length(reservoir)), how_long, replace = TRUE)], collapse = "")
  }

  schema_name <- random_schema()

  con_db <- my_test_con()

  schema_name_q <- DBI::dbQuoteIdentifier(con_db, schema_name)

  DBI::dbExecute(con_db, paste0("CREATE SCHEMA ", schema_name_q))

  dm_for_disambiguate_copied <- copy_dm_to(
    con_db,
    dm_for_disambiguate(),
    temporary = FALSE,
    schema = schema_name
  )
  order_of_deletion <- c("iris_3", "iris_2", "iris_1")
  remote_tbl_names <- set_names(
    paste0(schema_name_q, ".", DBI::dbQuoteIdentifier(con_db, order_of_deletion)),
    order_of_deletion
  )

  withr::defer({
    walk(
      remote_tbl_names,
      ~ try(DBI::dbExecute(con_db, paste0("DROP TABLE ", .x)))
    )
    try(DBI::dbExecute(con_db, paste0("DROP SCHEMA ", schema_name_q)))
  })

  normalize_table_name <- function(x) {
    tolower(gsub('["`]', "", x))
  }

  # test 'get_src_tbl_names()'
  src_tbl_names <- purrr::map_chr(
    get_src_tbl_names(con_db, schema = schema_name),
    ~ DBI::dbQuoteIdentifier(con_db, .x)
  )
  expect_identical(
    sort(normalize_table_name(DBI::SQL(src_tbl_names))),
    DBI::SQL(sort(normalize_table_name(remote_tbl_names)))
  )

  # learning with keys:
  dm_db_learned <-
    dm_from_con(con_db, schema = schema_name, learn_keys = TRUE) %>%
    dm_select_tbl(!!!order_of_deletion)

  expect_equivalent_dm(
    dm_db_learned,
    dm_for_disambiguate()[order_of_deletion],
    ignore_autoincrement = FALSE,
    ignore_on_delete = FALSE
  )

  # learning without keys:
  dm_db_learned_no_keys <-
    expect_silent(dm_from_con(con_db, schema = schema_name, learn_keys = FALSE)) %>%
    dm_select_tbl(!!!order_of_deletion)

  dm_for_disambiguate_no_keys <-
    dm_for_disambiguate()[order_of_deletion] %>%
    dm_get_def() %>%
    mutate(
      pks = list_of(new_pk()),
      fks = list_of(new_fk())
    ) %>%
    dm_from_def()

  expect_equivalent_dm(
    dm_db_learned_no_keys,
    dm_for_disambiguate_no_keys
  )
})

test_that("Learning from SQLite works (#288)?", {
  con_sqlite <- skip_if_error(DBI::dbConnect(RSQLite::SQLite(), ":memory:"))
  withr::defer(DBI::dbDisconnect(con_sqlite))

  DBI::dbWriteTable(con_sqlite, "test", tibble(a = 1:3))

  expect_equivalent_dm(
    dm_from_con(con_sqlite) %>%
      dm_select_tbl(test) %>%
      collect(),
    dm(test = tibble(a = 1:3))
  )
})

test_that("Learning keys from SQLite works", {
  con_sqlite <- skip_if_error(DBI::dbConnect(RSQLite::SQLite(), ":memory:"))
  withr::defer(DBI::dbDisconnect(con_sqlite))

  DBI::dbExecute(con_sqlite, "CREATE TABLE first (id INTEGER PRIMARY KEY)")
  DBI::dbExecute(
    con_sqlite,
    paste(
      "CREATE TABLE second (",
      "id INTEGER PRIMARY KEY, first_id INTEGER,",
      "FOREIGN KEY (first_id) REFERENCES first (id))"
    )
  )
  DBI::dbExecute(
    con_sqlite,
    paste(
      "CREATE TABLE third (",
      "id INTEGER PRIMARY KEY, first_id INTEGER, second_id INTEGER,",
      "FOREIGN KEY (first_id) REFERENCES first (id),",
      "FOREIGN KEY (second_id) REFERENCES second (id))"
    )
  )

  learned_dm <- dm_from_con(con_sqlite, learn_keys = TRUE) %>%
    collect()

  expect_snapshot(dm_paste(learned_dm, options = "keys"))
})

test_that("Learning keys from an attached SQLite database works", {
  con_sqlite <- skip_if_error(DBI::dbConnect(RSQLite::SQLite(), ":memory:"))
  withr::defer(DBI::dbDisconnect(con_sqlite))

  # Create a second (attached) database in a temp file
  db2_path <- withr::local_tempfile(fileext = ".sqlite")
  local({
    con2 <- DBI::dbConnect(RSQLite::SQLite(), db2_path)
    on.exit(DBI::dbDisconnect(con2))
    DBI::dbExecute(con2, "CREATE TABLE parent (id INTEGER PRIMARY KEY, val TEXT NOT NULL)")
    DBI::dbExecute(
      con2,
      paste(
        "CREATE TABLE child (",
        "id INTEGER PRIMARY KEY, parent_id INTEGER,",
        "FOREIGN KEY (parent_id) REFERENCES parent (id))"
      )
    )
  })

  # Attach the second database under a schema alias
  DBI::dbExecute(
    con_sqlite,
    paste0("ATTACH DATABASE '", db2_path, "' AS other")
  )

  learned_dm <- dm_from_con(
    con_sqlite,
    schema = "other",
    learn_keys = TRUE,
    .names = "{.schema}.{.table}_tbl"
  ) %>%
    collect()

  expect_snapshot(dm_paste(learned_dm, options = "keys"))
})


test_that("'schema_if()' works", {
  con_db <- my_db_test_src()$con

  # all 3 naming parameters set ('table' is required)
  expect_equal(
    schema_if(
      schema = "schema",
      table = "table",
      con = con_db,
      dbname = "database"
    )[[1]],
    DBI::Id(catalog = "database", schema = "schema", table = "table")
  )

  # schema and table set
  expect_equal(
    schema_if(
      schema = "schema",
      table = "table",
      con = con_db
    )[[1]],
    DBI::Id(schema = "schema", table = "table")
  )

  # dbname and table set
  expect_error(schema_if(schema = NA, con = con_db, table = "table", dbname = "db"))

  # only table set
  expect_equal(
    schema_if(schema = NA, table = "table", con = con_db)[[1]],
    DBI::Id(table = "table")
  )
})


# Learning from other DB on MSSQL -----------------------------------------
test_that("Learning from MSSQL (schema 'dbo') on other DB works?", {
  skip_if_src_not("mssql")
  # dm_learn_from_mssql() --------------------------------------------------
  con_db <- my_test_con()

  # delete database after test
  withr::defer({
    try(DBI::dbExecute(con_db, "DROP TABLE [test_database_dm].[dbo].[test_2]"))
    try(DBI::dbExecute(con_db, "DROP TABLE [test_database_dm].[dbo].[test_1]"))
    try(DBI::dbExecute(con_db, "DROP DATABASE test_database_dm"))
  })

  # create another DB and 2 connected tables
  DBI::dbExecute(con_db, "CREATE DATABASE test_database_dm")
  DBI::dbWriteTable(
    con_db,
    DBI::Id(catalog = "test_database_dm", schema = "dbo", table = "test_1"),
    value = tibble(a = c(5L, 5L, 4L, 2L, 1L), b = 1:5)
  )
  DBI::dbWriteTable(
    con_db,
    DBI::Id(catalog = "test_database_dm", schema = "dbo", table = "test_2"),
    value = tibble(c = c(1L, 1L, 1L, 5L, 4L), d = c(10L, 11L, 10L, 10L, 11L))
  )
  # set PK
  DBI::dbExecute(
    con_db,
    "ALTER TABLE [test_database_dm].[dbo].[test_1] ALTER COLUMN [b] INTEGER NOT NULL"
  )
  DBI::dbExecute(con_db, "ALTER TABLE [test_database_dm].[dbo].[test_1] ADD PRIMARY KEY ([b])")
  # set FK relation
  DBI::dbExecute(
    con_db,
    "ALTER TABLE [test_database_dm].[dbo].[test_2] ADD FOREIGN KEY ([c]) REFERENCES [test_database_dm].[dbo].[test_1] ([b]) ON DELETE NO ACTION ON UPDATE NO ACTION"
  )

  # test 'get_src_tbl_names()'
  expect_identical(
    get_src_tbl_names(con_db, dbname = "test_database_dm"),
    list(
      test_1 = DBI::Id(catalog = "test_database_dm", schema = "dbo", table = "test_1"),
      test_2 = DBI::Id(catalog = "test_database_dm", schema = "dbo", table = "test_2")
    )
  )

  dm_local_no_keys <- dm(
    test_1 = tibble(a = c(5L, 5L, 4L, 2L, 1L), b = 1:5),
    test_2 = tibble(c = c(1L, 1L, 1L, 5L, 4L), d = c(10L, 11L, 10L, 10L, 11L))
  )

  expect_message(
    dm_db_learned <- dm_from_con(con_db, dbname = "test_database_dm"),
    "queried successfully",
    fixed = TRUE
  )
  dm_learned <- dm_db_learned %>% collect()
  expect_equivalent_dm(
    dm_learned,
    dm_local_no_keys %>%
      dm_add_pk(test_1, b) %>%
      dm_add_fk(test_2, c, test_1),
    ignore_autoincrement = TRUE
  )

  # learning without keys:
  expect_silent(
    dm_learned_no_keys <- dm_from_con(
      con_db,
      dbname = "test_database_dm",
      learn_keys = FALSE
    ) %>%
      collect()
  )
  expect_equivalent_dm(
    dm_learned_no_keys[c("test_1", "test_2")],
    dm_local_no_keys
  )
})


# Learning from a specific schema in another DB on MSSQL -----------------------------------------
test_that("Learning from a specific schema in another DB for MSSQL works?", {
  skip_if_src_not("mssql")
  # dm_learn_from_mssql() --------------------------------------------------
  con_db <- my_test_con()

  original_dbname <- attributes(con_db)$info$dbname

  # delete database after test
  withr::defer({
    # dropping tables and schema is unnecessary
    try(DBI::dbExecute(con_db, "DROP DATABASE test_database_dm"))
  })

  # create another DB, a schema and 2 connected tables
  DBI::dbExecute(con_db, "CREATE DATABASE test_database_dm")
  DBI::dbExecute(con_db, "USE test_database_dm")
  DBI::dbExecute(con_db, "CREATE SCHEMA dm_test")
  DBI::dbExecute(con_db, paste0("USE ", original_dbname))

  DBI::dbWriteTable(
    con_db,
    DBI::Id(catalog = "test_database_dm", schema = "dm_test", table = "test_1"),
    value = tibble(a = c(5L, 5L, 4L, 2L, 1L), b = 1:5)
  )
  DBI::dbWriteTable(
    con_db,
    DBI::Id(catalog = "test_database_dm", schema = "dm_test", table = "test_2"),
    value = tibble(c = c(1L, 1L, 1L, 5L, 4L), d = c(10L, 11L, 10L, 10L, 11L))
  )
  # set PK
  DBI::dbExecute(
    con_db,
    "ALTER TABLE [test_database_dm].[dm_test].[test_1] ALTER COLUMN [b] INTEGER NOT NULL"
  )
  DBI::dbExecute(con_db, "ALTER TABLE [test_database_dm].[dm_test].[test_1] ADD PRIMARY KEY ([b])")
  # set FK relation
  DBI::dbExecute(
    con_db,
    "ALTER TABLE [test_database_dm].[dm_test].[test_2] ADD FOREIGN KEY ([c]) REFERENCES [test_database_dm].[dm_test].[test_1] ([b]) ON DELETE NO ACTION ON UPDATE NO ACTION"
  )

  # test 'get_src_tbl_names()'
  expect_identical(
    get_src_tbl_names(con_db, schema = "dm_test", dbname = "test_database_dm"),
    list(
      test_1 = DBI::Id(catalog = "test_database_dm", schema = "dm_test", table = "test_1"),
      test_2 = DBI::Id(catalog = "test_database_dm", schema = "dm_test", table = "test_2")
    )
  )

  dm_local_no_keys <- dm(
    test_1 = tibble(a = c(5L, 5L, 4L, 2L, 1L), b = 1:5),
    test_2 = tibble(c = c(1L, 1L, 1L, 5L, 4L), d = c(10L, 11L, 10L, 10L, 11L))
  )

  dm_db_learned <- dm_from_con(
    con_db,
    schema = "dm_test",
    dbname = "test_database_dm",
    learn_keys = TRUE
  )
  dm_learned <- dm_db_learned %>% collect()
  expect_equivalent_dm(
    dm_learned[c("test_1", "test_2")],
    dm_local_no_keys %>%
      dm_add_pk(test_1, b) %>%
      dm_add_fk(test_2, c, test_1),
    ignore_autoincrement = FALSE,
    ignore_on_delete = FALSE
  )

  # learning without keys:
  dm_learned_no_keys <- expect_silent(
    dm_from_con(
      con_db,
      schema = "dm_test",
      dbname = "test_database_dm",
      learn_keys = FALSE
    ) %>%
      collect()
  )
  expect_equivalent_dm(
    dm_learned_no_keys[c("test_1", "test_2")],
    dm_local_no_keys
  )
})

# Must be in the same file as the other learning tests
# to avoid race conditions.
test_that("dm_meta() contents", {
  skip_if_schema_not_supported()

  # produces a randomized schema name with a length of 4-10 characters
  # consisting of the symbols in `reservoir`
  random_schema <- function() {
    reservoir <- c(letters)
    how_long <- 10
    paste0(reservoir[sample(seq_len(length(reservoir)), how_long, replace = TRUE)], collapse = "")
  }

  schema_name <- random_schema()

  # dm_learn_from_mssql() --------------------------------------------------
  con_db <- my_test_con()

  schema_name_q <- DBI::dbQuoteIdentifier(con_db, schema_name)

  DBI::dbExecute(con_db, paste0("CREATE SCHEMA ", schema_name_q))

  # create an object on the MSSQL-DB that can be learned
  withr::defer({
    try(walk(
      order_of_deletion,
      ~ try(DBI::dbExecute(con_db, paste0("DROP TABLE ", schema_name_q, ".", .x, " CASCADE")))
    ))
    try(DBI::dbExecute(con_db, paste0("DROP SCHEMA ", schema_name_q)))
  })

  dm_for_filter_copied <- copy_dm_to(
    con_db,
    dm_for_filter(),
    temporary = FALSE,
    schema = schema_name
  )
  order_of_deletion <- c("tf_2", "tf_1", "tf_5", "tf_6", "tf_4", "tf_3")

  meta <- dm_meta(con_db, schema = schema_name)

  constraints <- dm_examine_constraints(meta, .progress = FALSE)
  expect_true(all(constraints$is_key))

  arrange_all_but_constraint_name <- function(.x) {
    names <- setdiff(colnames(.x), "constraint_name")
    .x %>%
      arrange(!!!syms(names))
  }

  expect_snapshot({
    meta %>%
      dm_select_tbl(-schemata) %>%
      dm_zoom_to(table_constraints) %>%
      filter(constraint_type %in% c("PRIMARY KEY", "FOREIGN KEY")) %>%
      dm_update_zoomed() %>%
      dm_get_tables() %>%
      map(select, -any_of("column_default"), -contains("catalog"), -contains("schema")) %>%
      map(collect) %>%
      map(arrange_all_but_constraint_name) %>%
      map(
        ~ if ("constraint_name" %in% colnames(.x)) {
          .x %>% mutate(constraint_name = as.integer(forcats::fct_inorder(constraint_name)))
        } else {
          .x
        }
      ) %>%
      imap(
        ~ if (is_mariadb(con_db) && .y == "columns") {
          # mariadb output on autoincrement column is integer
          # transform this to boolean
          mutate(.x, is_autoincrement = as.logical(is_autoincrement))
        } else {
          .x
        }
      ) %>%
      imap(
        ~ if (is_mariadb(con_db) && .y == "table_constraints") {
          # mariadb default action for delete_rule is RESTRICT (synonym for NO ACTION)
          # https://mariadb.com/kb/en/foreign-keys/#constraints
          mutate(.x, delete_rule = if_else(delete_rule == "RESTRICT", "NO ACTION", delete_rule))
        } else {
          .x
        }
      ) %>%
      map(arrange_all) %>%
      jsonlite::toJSON(pretty = TRUE) %>%
      gsub(schema_name, "schema_name", .) %>%
      gsub('(_catalog": ")[^"]*(")', "\\1catalog\\2", .) %>%
      writeLines()
  })
})

test_that("dm_from_con() with mariaDB", {
  skip_if_offline()
  skip_if_not(dm_has_financial())

  my_db <- RMariaDB::dbConnect(
    RMariaDB::MariaDB(),
    username = "guest",
    password = "ctu-relational",
    dbname = "Financial_ijs",
    host = "relational.fel.cvut.cz"
  )
  my_dm <- dm_from_con(my_db, learn_keys = TRUE)
  expect_snapshot(dm::dm_get_all_fks(my_dm))
  expect_snapshot(dm::dm_get_all_pks(my_dm))

  # multiple schemata work
  my_dm <- dm_from_con(
    my_db,
    schema = c("Accidents", "Ad", "Financial_std"),
    learn_keys = TRUE
  )
  expect_snapshot(dm::dm_get_all_fks(my_dm))
  expect_snapshot(dm::dm_get_all_pks(my_dm))
})

test_that("dm() API", {
  expect_snapshot({
    dm(a = tibble(), a = tibble(), .name_repair = "unique")
    dm(a = tibble(), a = tibble(), .name_repair = "unique", .quiet = TRUE)
  })
  expect_snapshot(error = TRUE, {
    dm(a = tibble(), a = tibble())
  })
  expect_snapshot(error = TRUE, {
    dm(a = dm())
  })
  expect_snapshot(error = TRUE, {
    dm(a = tibble(), dm_zoom_to(dm_for_filter(), tf_1))
  })
})

test_that("dm() works for adding tables", {
  # is a table added?
  expect_identical(
    length(dm_get_tables(dm(dm_for_filter(), data_card_1()))),
    7L
  )

  # can I retrieve the tibble under its old name?
  expect_equivalent_tbl(
    dm(dm_for_filter(), data_card_1())[["data_card_1()"]],
    data_card_1()
  )

  # can I retrieve the tibble under a new name?
  expect_equivalent_tbl(
    dm(dm_for_filter(), test = data_card_1())[["test"]],
    data_card_1()
  )

  # use special names with :=
  expect_identical(
    names(dm(dm_for_filter(), dm := data_card_1(), repair := data_card_2())),
    c(names(dm_for_filter()), "dm", "repair")
  )

  # we accept even weird table names, as long as they are unique
  expect_equivalent_tbl(
    dm(dm_for_filter(), . = data_card_1())[["."]],
    data_card_1()
  )

  # do I avoid the warning when piping the table but setting the name?
  expect_silent(
    expect_equivalent_tbl(
      dm_for_filter() %>% dm(new_name = data_card_1()) %>% pull_tbl(new_name),
      data_card_1()
    )
  )

  # adding more than 1 table:
  # 1. Is the resulting number of tables correct?
  expect_identical(
    length(dm_get_tables(dm(dm_for_filter(), data_card_1(), data_card_2()))),
    8L
  )

  # 2. Is the resulting order of the tables correct?
  expect_identical(
    src_tbls_impl(dm(dm_for_filter(), data_card_1(), data_card_2())),
    c(src_tbls_impl(dm_for_filter()), "data_card_1()", "data_card_2()")
  )

  # Is an error thrown in case I try to give the new table an old table's name if `repair = "check_unique"`?
  expect_snapshot(error = TRUE, {
    dm(dm_for_filter(), tf_1 = data_card_1(), .name_repair = "check_unique")
  })

  # are in the default case (`repair = 'unique'`) the tables renamed (old table AND new table) according to "unique" default setting
  expect_identical(
    dm(dm_for_filter(), tf_1 = data_card_1(), .name_repair = "unique", .quiet = TRUE) %>%
      src_tbls_impl(),
    c("tf_1...1", "tf_2", "tf_3", "tf_4", "tf_5", "tf_6", "tf_1...7")
  )

  expect_name_repair_message(
    expect_equivalent_dm(
      dm(dm_for_filter(), tf_1 = data_card_1(), .name_repair = "unique"),
      dm_for_filter() %>%
        dm_rename_tbl(tf_1...1 = tf_1) %>%
        dm(tf_1...7 = data_card_1())
    )
  )

  # can I use dm_select_tbl(), selecting among others the new table?
  expect_silent(
    dm(dm_for_filter(), tf_7_new = tf_7()) %>% dm_select_tbl(tf_1, tf_7_new, everything())
  )

  # error in case table srcs don't match
  expect_dm_error(
    dm(dm_for_filter(), data_card_1_duckdb()),
    "not_same_src"
  )

  # adding tables to an empty `dm` works for all sources
  expect_equivalent_tbl(
    dm(dm(), test = data_card_1_duckdb())$test,
    data_card_1()
  )
})

test_that("dm() for adding tables with compound keys", {
  expect_snapshot({
    dm(dm_for_flatten(), res_flat = result_from_flatten()) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("dm() works for dm objects", {
  expect_equivalent_dm(
    dm(dm_for_filter()),
    dm_for_filter()
  )

  expect_equivalent_dm(
    dm(dm_for_filter(), dm_for_flatten(), dm_for_disambiguate()),
    bind_rows(
      dm_get_def(dm_for_filter()),
      dm_get_def(dm_for_flatten()),
      dm_get_def(dm_for_disambiguate())
    ) %>%
      dm_from_def()
  )
})

test_that("are empty_dm() and empty ellipsis handled correctly?", {
  expect_equivalent_dm(
    dm(empty_dm()),
    empty_dm()
  )

  expect_equivalent_dm(
    dm(empty_dm(), empty_dm(), empty_dm()),
    empty_dm()
  )

  expect_equivalent_dm(
    dm(),
    empty_dm()
  )
})

test_that("errors: duplicate table names, src mismatches", {
  expect_snapshot(error = TRUE, {
    dm(dm_for_filter(), dm_for_flatten(), dm_for_filter())
  })

  expect_dm_error(dm(dm_for_flatten(), dm_for_filter_duckdb()), "not_same_src")
})

test_that("auto-renaming works", {
  expect_equivalent_dm(
    expect_name_repair_message(
      dm(dm_for_filter(), dm_for_flatten(), dm_for_filter(), .name_repair = "unique")
    ),
    bind_rows(
      dm_get_def(
        dm_rename_tbl(
          dm_for_filter(),
          tf_1...1 = tf_1,
          tf_2...2 = tf_2,
          tf_3...3 = tf_3,
          tf_4...4 = tf_4,
          tf_5...5 = tf_5,
          tf_6...6 = tf_6
        )
      ),
      dm_get_def(dm_for_flatten()),
      dm_get_def(dm_rename_tbl(
        dm_for_filter(),
        tf_1...12 = tf_1,
        tf_2...13 = tf_2,
        tf_3...14 = tf_3,
        tf_4...15 = tf_4,
        tf_5...16 = tf_5,
        tf_6...17 = tf_6
      ))
    ) %>%
      dm_from_def()
  )

  expect_silent(
    dm(dm_for_filter(), dm_for_flatten(), dm_for_filter(), .name_repair = "unique", .quiet = TRUE)
  )
})

test_that("test error output for src mismatches", {
  skip_if_not(getRversion() >= "4.0")

  expect_snapshot({
    writeLines(conditionMessage(expect_error(
      dm(dm_for_flatten(), dm_for_filter_duckdb())
    )))
  })
})

test_that("output for dm() with dm", {
  expect_snapshot({
    dm()
    dm(empty_dm())
    dm(dm_for_filter()) %>% collect()
    dm(
      dm_for_filter(),
      dm_for_flatten(),
      dm_for_filter(),
      .name_repair = "unique",
      .quiet = TRUE
    ) %>%
      collect()
  })
})

test_that("output for dm() with dm (2)", {
  expect_snapshot(error = TRUE, {
    dm(dm_for_filter(), dm_for_flatten(), dm_for_filter())
  })
})

test_that("output for dm() with dm (3)", {
  expect_snapshot({
    dm(dm_for_filter(), dm_for_flatten(), dm_for_filter(), .name_repair = "unique") %>% collect()
  })
})

test_that("output dm() for dm for compound keys", {
  expect_snapshot({
    dm(dm_for_filter(), dm_for_flatten()) %>% dm_paste(options = c("select", "keys"))
    dm(dm_for_flatten(), dm_for_filter()) %>% dm_paste(options = c("select", "keys"))
  })

  expect_snapshot({
    dm(dm_for_flatten(), dm_for_flatten(), .name_repair = "unique") %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("can create dm with as_dm()", {
  expect_equivalent_dm(as_dm(dm_get_tables(dm_test_obj())), dm_test_obj())
})

test_that("as_dm() rejects non-list objects", {
  expect_snapshot(error = TRUE, {
    as_dm(1)
  })
})

test_that("creation of empty `dm` works", {
  expect_true(
    is_empty(dm())
  )

  expect_true(
    is_empty(new_dm())
  )
})

test_that("'collect.dm()' collects tables on DB", {
  def <-
    dm_for_filter() %>%
    dm_filter(tf_1 = a > 3) %>%
    collect() %>%
    dm_get_def()

  is_df <- map_lgl(def$data, is.data.frame)
  expect_true(all(is_df))
})

test_that("'collect.dm_zoomed()' collects tables, with message", {
  dm_zoomed_for_collect <-
    dm_for_filter() %>%
    dm_zoom_to(tf_1) %>%
    mutate(c = a + 1)

  expect_message(
    out <- dm_zoomed_for_collect %>% collect(),
    "pull_tbl"
  )

  expect_s3_class(out, "data.frame")
})

test_that("'compute.dm()' computes tables on DB", {
  def <-
    dm_for_filter_duckdb() %>%
    dm_filter(tf_1 = a > 3) %>%
    {
      suppress_mssql_message(compute(.))
    } %>%
    dm_get_def()

  remote_names <- map(def$data, dbplyr::remote_name)
  expect_equal(lengths(remote_names), rep_along(remote_names, 1))
})

test_that("'compute.dm_zoomed()' computes tables on DB", {
  dm_zoomed_for_compute <-
    dm_for_filter_duckdb() %>%
    dm_zoom_to(tf_1) %>%
    mutate(c = a + 1)

  # without computing
  def <-
    dm_zoomed_for_compute %>%
    dm_update_zoomed() %>%
    dm_get_def()

  remote_names <- map(def$data, dbplyr::remote_name)
  expect_true(any(map_lgl(remote_names, is_null)))

  # with computing
  def <-
    suppress_mssql_message(compute(dm_zoomed_for_compute)) %>%
    dm_update_zoomed() %>%
    dm_get_def()

  remote_names <- map(def$data, dbplyr::remote_name)
  expect_equal(lengths(remote_names), rep_along(remote_names, 1))
})

test_that("'compute.dm()' fails with `temporary = FALSE` (#2059)", {
  expect_snapshot(error = TRUE, {
    dm_for_filter_duckdb() %>%
      compute(temporary = FALSE)
  })
})

test_that("some methods/functions for `dm_zoomed` work", {
  expect_identical(
    colnames(dm_zoom_to(dm_for_filter(), tf_1)),
    c("a", "b")
  )

  expect_identical(
    ncol(dm_zoom_to(dm_for_filter(), tf_1)),
    2L
  )

  expect_equivalent_tbl_lists(
    as.list(dm_for_filter()),
    dm_get_tables(dm_for_filter())
  )

  skip_if_remote_src()
  expect_identical(
    dim(dm_zoom_to(dm_for_filter(), tf_1)),
    c(10L, 2L)
  )
  expect_identical(
    names(dm_zoom_to(dm_for_filter(), tf_2)),
    colnames(tf_2())
  )
})

test_that("length and names for dm work", {
  expect_length(dm_for_filter(), 6L)
  expect_identical(names(dm_for_filter()), src_tbls_impl(dm_for_filter()))
})

test_that("`pull_tbl()`-methods work", {
  expect_equivalent_tbl(
    pull_tbl(dm_for_filter(), tf_5),
    tf_5()
  )

  expect_equal(
    pull_tbl(dm_for_filter(), tf_5, keyed = TRUE),
    dm_get_tables(dm_for_filter(), keyed = TRUE)[["tf_5"]]
  )
})

test_that("`pull_tbl()` keyed=TRUE fails for zoomed dm", {
  expect_snapshot(error = TRUE, {
    dm_for_filter() %>%
      dm_zoom_to(tf_1) %>%
      pull_tbl(keyed = TRUE)
  })
})

test_that("`pull_tbl()`-methods work for (0)", {
  tbl <-
    dm_nycflights_small() %>%
    dm_set_table_description("Flugzeuge" = planes) %>%
    pull_tbl(planes, keyed = TRUE)

  skip_if_not_installed("labelled")

  expect_identical(labelled::label_attribute(tbl), "Flugzeuge")
})

test_that("`pull_tbl()`-methods work for (1)", {
  skip_if_src("maria")
  expect_equivalent_tbl(
    dm_for_filter() %>%
      dm_zoom_to(tf_3) %>%
      mutate(new_col = row_number(f) * 3) %>%
      pull_tbl(),
    mutate(tf_3(), new_col = row_number(f) * 3)
  )
})

test_that("`pull_tbl()`-methods work (2)", {
  expect_equivalent_tbl(
    dm_zoom_to(dm_for_filter(), tf_1) %>% pull_tbl(tf_1),
    tf_1()
  )

  expect_dm_error(
    dm_zoom_to(dm_for_filter(), tf_1) %>% pull_tbl(tf_2),
    "table_not_zoomed"
  )

  expect_dm_error(
    pull_tbl(dm_for_filter()),
    "no_table_provided"
  )

  expect_dm_error(
    dm_for_filter() %>%
      dm_get_def() %>%
      mutate(zoom = list(tf_1)) %>%
      dm_from_def(zoomed = TRUE, validate = FALSE) %>%
      pull_tbl(),
    "not_pulling_multiple_zoomed"
  )
})

test_that("numeric subsetting works", {
  # check specifically for the right output in one case
  expect_equivalent_tbl(dm_for_filter()[[4]], tf_4())

  # compare numeric subsetting and subsetting by name on chosen src
  expect_equivalent_tbl(
    dm_for_filter()[["tf_2"]],
    dm_for_filter()[[2]]
  )

  # check if reducing `dm` size (and reordering) works on chosen src
  expect_equivalent_dm(
    dm_for_filter()[c(1, 5, 3)],
    dm_select_tbl(dm_for_filter(), 1, 5, 3)
  )
})

test_that("subsetting `dm` works", {
  expect_equivalent_tbl(dm_for_filter()$tf_5, tf_5())
  expect_equivalent_tbl(dm_for_filter()[["tf_3"]], tf_3())
})

test_that("subsetting `dm_zoomed` works", {
  skip_if_remote_src()
  expect_identical(
    dm_zoom_to(dm_for_filter(), tf_2)$c,
    pull(tf_2(), c)
  )

  expect_identical(
    dm_zoom_to(dm_for_filter(), tf_3)[["g"]],
    pull(tf_3(), g)
  )

  expect_identical(
    dm_zoom_to(dm_for_filter(), tf_3)[c("g", "f", "g")],
    tf_3()[c("g", "f", "g")]
  )
})

test_that("as.list()-method works for local `dm_zoomed`", {
  skip_if_remote_src()
  expect_identical(
    as.list(dm_for_filter() %>% dm_zoom_to(tf_4)),
    as.list(tf_4())
  )
})

# test getters: -----------------------------------------------------------

test_that("dm_get_src() works", {
  local_options(lifecycle_verbosity = "quiet")

  expect_dm_error(
    dm_get_src(1),
    class = "is_not_dm"
  )

  expect_identical(
    class(dm_get_src(dm_for_filter())),
    class(my_test_src())
  )
})

test_that("dm_get_con() errors", {
  expect_dm_error(
    dm_get_con(1),
    class = "is_not_dm"
  )

  skip_if_remote_src()
  expect_dm_error(
    dm_get_con(dm_for_filter()),
    class = "con_only_for_dbi"
  )
})

test_that("dm_get_con() works", {
  expect_identical(
    dm_get_con(dm_for_filter_db()),
    con_from_src_or_con(my_db_test_src())
  )
})

test_that("str()", {
  # https://github.com/cynkra/dm/pull/542/checks?check_run_id=2506393322#step:11:88
  skip("FIXME: Unstable on GHA?")

  expect_snapshot({
    dm_for_filter() %>%
      str()

    dm_for_filter() %>%
      dm_zoom_to(tf_2) %>%
      str()
  })
})

test_that("output", {
  expect_snapshot({
    print(dm())

    nyc_flights_dm <- dm_nycflights_small()
    collect(nyc_flights_dm)

    nyc_flights_dm %>%
      format()

    nyc_flights_dm %>%
      dm_filter(flights = (origin == "EWR")) %>%
      collect()
  })
})


# Compound tests ----------------------------------------------------------

test_that("output for compound keys", {
  # FIXME: COMPOUND: Need proper test
  skip_if_remote_src()

  # Can't be inside the snapshot
  car_table <- test_src_frame(!!!mtcars)

  expect_snapshot({
    copy_to(nyc_comp(), mtcars, "car_table")
    dm(nyc_comp(), car_table)
    nyc_comp() %>%
      collect()
    nyc_comp() %>%
      dm_filter(flights = (day == 10)) %>%
      collect() %>%
      dm_get_def() %>%
      select(-uuid)
    nyc_comp() %>%
      dm_zoom_to(weather) %>%
      mutate(origin_new = paste0(origin, " airport")) %>%
      compute() %>%
      dm_update_zoomed() %>%
      collect() %>%
      dm_get_def() %>%
      select(-uuid)
    nyc_comp() %>%
      dm_zoom_to(weather) %>%
      collect()
    pull_tbl(nyc_comp(), weather)
    nyc_comp() %>%
      dm_zoom_to(weather) %>%
      pull_tbl()
  })
})

test_that("glimpse.dm() works", {
  skip_if_remote_src()
  expect_snapshot({
    glimpse(empty_dm())

    # glimpse 'standard' dm object
    glimpse(dm_for_disambiguate())

    # glimpse 'standard' dm object with different width
    glimpse(dm_for_disambiguate(), width = 40)

    # option "width" inside test_that-environment should always be 80
    getOption("width")

    # # glimpse dm with long names for tables and/or columns
    glimpse(
      dm_for_disambiguate() %>%
        dm_rename(
          iris_1,
          gdsjgiodsjgdisogjdsiogjdsigjsdiogjisdjgiodsjgiosdjgiojsdiogjgrjihjrehoierjhiorejhrieojhreiojhieorhjioerjhierjhioerjhioerjhioerjiohjeriosdiogjsdjigjsd = key
        ) %>%
        dm_rename_tbl(
          gdsjgiodsjgdisogjdsiogjdsigjsdiogjisdjgiodsjgiosdjgiojsdiogjgrjihjrehoierjhiorejhrieojhreiojhieorhjioerjhierjhioerjhioerjhioerjiohjeriosdiogjsdjigjsd = iris_1
        )
    )

    # in case no primary keys are present, nothing about primary keys should be printed
    dm_nycflights13() %>%
      dm_select_tbl(weather) %>%
      dm_select(weather, -origin) %>%
      glimpse()
  })
})

test_that("glimpse.dm_zoomed() works", {
  skip_if_remote_src()
  expect_snapshot({
    # doesn't have foreign keys to print
    dm_nycflights13() %>%
      dm_zoom_to(airports) %>%
      glimpse()

    # has foreign keys to print
    dm_nycflights13() %>%
      dm_zoom_to(flights) %>%
      glimpse(width = 100)

    # if any primary key has been removed, no primary key is displayed
    dm_nycflights13() %>%
      dm_zoom_to(weather) %>%
      select(-origin) %>%
      glimpse()

    # anticipate primary keys being renamed by users
    dm_nycflights13() %>%
      dm_zoom_to(weather) %>%
      rename(origin_location = origin) %>%
      glimpse()

    # if any foreign key has been removed, corresponding composite key is not displayed
    dm_nycflights13() %>%
      dm_zoom_to(flights) %>%
      select(-carrier) %>%
      glimpse()
    dm_nycflights13() %>%
      dm_zoom_to(flights) %>%
      select(-origin) %>%
      glimpse()

    # anticipate foreign keys being renamed by users
    dm_nycflights13() %>%
      dm_zoom_to(flights) %>%
      rename(origin_location = origin) %>%
      glimpse()
  })
})

test_that("dm_select_tbl() selects a part of a larger `dm` as a reduced `dm`?", {
  def <-
    dm_for_filter() %>%
    dm_rm_fk(tf_5, m, tf_6, n) %>%
    dm_rm_fk(tf_2, d, tf_1) %>%
    dm_get_def()

  dm_for_filter_smaller <- dm_from_def(def[def$table %in% c("tf_1", "tf_6"), ])

  expect_equivalent_dm(
    dm_select_tbl(dm_for_filter(), -tf_2, -tf_3, -tf_4, -tf_5),
    dm_for_filter_smaller
  )
})

test_that("dm_select_tbl() can reorder the tables in a `dm`", {
  reordered_dm_for_filter <-
    dm_for_filter() %>%
    dm_get_def() %>%
    arrange(c(3:1, 6:4)) %>%
    dm_from_def()

  expect_equivalent_dm(
    dm_select_tbl(dm_for_filter(), tf_3:tf_1, tf_6:tf_4),
    reordered_dm_for_filter
  )
})

test_that("dm_select_tbl() remembers all FKs", {
  expect_snapshot({
    dm_nycflights_small() %>%
      dm_add_fk(flights, origin, airports) %>%
      dm_select_tbl(airports, flights) %>%
      dm_paste()
  })
})


test_that("dm_rename_tbl() renames a `dm`", {
  # FIXME: PR #313: do these test-tibbles here also need to be copied to `my_test_src()`?
  dm_rename <-
    as_dm(list(a = tibble(x = 1), b = tibble(y = 1))) %>%
    dm_add_pk(b, y) %>%
    dm_add_fk(a, x, b)

  dm_rename_a <-
    as_dm(list(c = tibble(x = 1), b = tibble(y = 1))) %>%
    dm_add_pk(b, y) %>%
    dm_add_fk(c, x, b)

  dm_rename_b <-
    as_dm(list(a = tibble(x = 1), e = tibble(y = 1))) %>%
    dm_add_pk(e, y) %>%
    dm_add_fk(a, x, e)

  dm_rename_bd <-
    as_dm(list(a = tibble(x = 1), d = tibble(y = 1))) %>%
    dm_add_pk(d, y) %>%
    dm_add_fk(a, x, d)

  expect_equivalent_dm(
    dm_rename_tbl(dm_rename, c = a),
    dm_rename_a
  )

  expect_equivalent_dm(
    dm_rename_tbl(dm_rename, e = b),
    dm_rename_b
  )

  skip("dm argument")
  expect_equivalent_dm(
    dm_rename_tbl(dm_rename, d = b),
    dm_rename_bd
  )
})

test_that("errors for selecting and renaming tables work", {
  expect_error(
    dm_select_tbl(dm_for_filter(), t_new = c(tf_1, tf_2)),
    class = "vctrs_error_names_must_be_unique"
  )

  expect_error(
    dm_rename_tbl(dm_for_filter(), t_new = c(tf_1, tf_2)),
    class = "vctrs_error_names_must_be_unique"
  )
})

test_that("dm_add_tbl() works", {
  local_options(lifecycle_verbosity = "quiet")

  # is a table added?
  expect_identical(
    length(dm_get_tables(dm_add_tbl(dm_for_filter(), data_card_1()))),
    7L
  )

  # can I retrieve the tibble under its old name?
  expect_equivalent_tbl(
    dm_add_tbl(dm_for_filter(), data_card_1())[["data_card_1()"]],
    data_card_1()
  )

  # can I retrieve the tibble under a new name?
  expect_equivalent_tbl(
    dm_add_tbl(dm_for_filter(), test = data_card_1())[["test"]],
    data_card_1()
  )

  # use special names with :=
  expect_identical(
    names(dm_add_tbl(dm_for_filter(), dm := data_card_1(), repair := data_card_2())),
    c(names(dm_for_filter()), "dm", "repair")
  )

  # we accept even weird table names, as long as they are unique
  expect_equivalent_tbl(
    dm_add_tbl(dm_for_filter(), . = data_card_1())[["."]],
    data_card_1()
  )

  # do I avoid the warning when piping the table but setting the name?
  expect_silent(
    expect_equivalent_tbl(
      dm_for_filter() %>% dm_add_tbl(new_name = data_card_1()) %>% pull_tbl(new_name),
      data_card_1()
    )
  )

  # adding more than 1 table:
  # 1. Is the resulting number of tables correct?
  expect_identical(
    length(dm_get_tables(dm_add_tbl(dm_for_filter(), data_card_1(), data_card_2()))),
    8L
  )

  # 2. Is the resulting order of the tables correct?
  expect_identical(
    src_tbls_impl(dm_add_tbl(dm_for_filter(), data_card_1(), data_card_2())),
    c(src_tbls_impl(dm_for_filter()), "data_card_1()", "data_card_2()")
  )

  # are in the default case (`repair = 'unique'`) the tables renamed (old table AND new table) according to "unique" default setting
  expect_identical(
    dm_add_tbl(dm_for_filter(), tf_1 = data_card_1(), quiet = TRUE) %>% src_tbls_impl(),
    c("tf_1...1", "tf_2", "tf_3", "tf_4", "tf_5", "tf_6", "tf_1...7")
  )

  expect_name_repair_message(
    expect_equivalent_dm(
      dm_add_tbl(dm_for_filter(), tf_1 = data_card_1(), repair = "unique"),
      dm_for_filter() %>%
        dm_rename_tbl(tf_1...1 = tf_1) %>%
        dm_add_tbl(tf_1...7 = data_card_1())
    )
  )

  # can I use dm_select_tbl(), selecting among others the new table?
  expect_silent(
    dm_add_tbl(dm_for_filter(), tf_7_new = tf_7()) %>% dm_select_tbl(tf_1, tf_7_new, everything())
  )

  # error in case table srcs don't match
  expect_dm_error(
    dm_add_tbl(dm_for_filter(), data_card_1_duckdb()),
    "not_same_src"
  )

  # adding tables to an empty `dm` works for all sources
  expect_equivalent_tbl(
    dm_add_tbl(dm(), test = data_card_1_duckdb())$test,
    data_card_1()
  )
})

test_that("dm_add_tbl() snapshots", {
  local_options(lifecycle_verbosity = "warning")

  # Is an error thrown in case I try to give the new table an old table's name if `repair = "check_unique"`?
  expect_snapshot(error = TRUE, {
    dm_add_tbl(dm_for_filter(), tf_1 = data_card_1(), repair = "check_unique")
  })

  expect_snapshot({
    dm_add_tbl(dm_for_flatten(), res_flat = result_from_flatten()) %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("dm_rm_tbl() works", {
  local_options(lifecycle_verbosity = "quiet")

  # removes a table
  expect_equivalent_dm(
    dm_rm_tbl(dm_for_filter_w_cycle(), tf_7),
    dm_for_filter()
  )

  # removes more than one table
  expect_equivalent_dm(
    dm_rm_tbl(dm_for_filter_w_cycle(), tf_7, tf_5, tf_3),
    dm_select_tbl(dm_for_filter(), tf_1, tf_2, tf_4, tf_6)
  )

  # fails when table name is wrong
  expect_error(
    dm_rm_tbl(dm_for_filter(), tf_9),
    class = "vctrs_error_subscript"
  )

  # select-helpers work for 'dm_rm_tbl()'
  expect_identical(
    dm_rm_tbl(dm_for_disambiguate(), everything()),
    empty_dm()
  )

  # corner case: not removing any table
  expect_identical(
    dm_rm_tbl(dm_for_disambiguate()),
    dm_for_disambiguate()
  )
})

test_that("dm_rm_tbl() snapshot", {
  local_options(lifecycle_verbosity = "warning")
  expect_snapshot({
    dm_rm_tbl(dm_for_flatten(), dim_1) %>% dm_paste(options = c("select", "keys"))
    dm_rm_tbl(dm_for_flatten(), fact) %>% dm_paste(options = c("select", "keys"))
  })
})

test_that("dm_mutate_tbl() works", {
  expect_equivalent_dm(
    dm_for_filter_w_cycle() %>%
      dm_mutate_tbl(
        tf_1 = dm_for_filter_w_cycle()$tf_1,
        tf_7 = dm_for_filter_w_cycle()$tf_7 %>%
          filter(0L == 1L)
      ),
    dm_for_filter_w_cycle() %>%
      dm_zoom_to(tf_7) %>%
      filter(0L == 1L) %>%
      dm_update_zoomed()
  )

  # Table doesn't exist yet
  expect_error(
    dm_for_filter_w_cycle() %>%
      dm_mutate_tbl(
        tf_99 = dm_for_filter_w_cycle()$tf_7
      )
  )

  # Wrong column structure
  expect_error(
    dm_for_filter_w_cycle() %>%
      dm_mutate_tbl(
        tf_1 = dm_for_filter_w_cycle()$tf_7
      )
  )
})

#' JSON nest
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' A wrapper around [tidyr::nest()] which stores the nested data into JSON columns.
#'
#' @param .data A data frame, a data frame extension (e.g. a tibble), or  a lazy data frame (e.g. from dbplyr or dtplyr).
#' @param .names_sep If `NULL`, the default, the names will be left as is.
#' @param ... <[`tidy-select`][tidyr_tidy_select]> Columns to pack, specified
#'   using name-variable pairs of the form `new_col = c(col1, col2, col3)`.
#'   The right hand side can be any valid tidy select expression.
#' @seealso [tidyr::nest()], [json_nest_join()]
#' @export
#' @examples
#' df <- tibble::tibble(x = c(1, 1, 1, 2, 2, 3), y = 1:6, z = 6:1)
#' nested <- json_nest(df, data = c(y, z))
#' nested
json_nest <- function(.data, ..., .names_sep = NULL) {
  UseMethod("json_nest")
}

#' @export
json_nest.data.frame <- function(.data, ..., .names_sep = NULL) {
  check_suggested("jsonlite", "json_nest")
  dot_nms <- ...names()
  # `{tidyr}` only warns but since we don't need backward compatibility we're
  #   better off failing
  if (is_null(dot_nms) || "" %in% dot_nms) {
    cli::cli_abort("All elements of {.arg ...} must be named.")
  }
  tidyr::nest(.data, ..., .names_sep = .names_sep) %>%
    mutate(across(all_of(dot_nms), ~ map_chr(., jsonlite::toJSON, digits = NA)))
}

#' @export
json_nest.tbl_lazy <- function(.data, ..., .names_sep = NULL) {
  dots <- quos(...)
  if ("" %in% names2(dots)) {
    cli::cli_abort("All elements of {.arg ...} must be named.")
  }

  col_nms <- colnames(.data)
  nest_cols <- purrr::map(dots, ~ tidyselect::vars_select(col_nms, !!.x))
  id_cols <- setdiff(col_nms, unlist(unique(nest_cols)))

  sql_exprs <- purrr::imap(
    nest_cols,
    ~ sql_json_nest(
      dbplyr::remote_con(.data),
      cols = names(.x),
      names_sep = .names_sep,
      packed_col = .y,
      id_cols = id_cols,
      data = .data
    )
  )

  json_nest_aggregate(dbplyr::remote_con(.data), .data, id_cols, sql_exprs)
}

json_nest_aggregate <- function(con, data, id_cols, sql_exprs, ...) {
  UseMethod("json_nest_aggregate")
}

#' @export
json_nest_aggregate.default <- function(con, data, id_cols, sql_exprs, ...) {
  check_dots_empty()

  data %>%
    group_by(across(!!!syms(id_cols))) %>%
    summarize(!!!sql_exprs) %>%
    ungroup()
}

sql_json_nest <- function(con, cols, names_sep, packed_col, id_cols, data, ...) {
  UseMethod("sql_json_nest")
}

#' @export
#' @autoglobal
#' @global JSON_AGG JSON_BUILD_OBJECT
sql_json_nest.PqConnection <- function(con, cols, names_sep, packed_col, id_cols, data, ...) {
  check_dots_empty()

  inside_cols <- remove_prefix_and_sep(cols, prefix = packed_col, sep = names_sep)
  inside_cols_idented <- dbplyr::ident(inside_cols)
  exprs <- vctrs::vec_interleave(as.list(inside_cols_idented), syms(cols))
  dbplyr::translate_sql(JSON_AGG(JSON_BUILD_OBJECT(!!!exprs)), con = con)
}

#' @export
`json_nest.tbl_Microsoft SQL Server` <- function(.data, ..., .names_sep = NULL) {
  dots <- quos(...)
  if ("" %in% names2(dots)) {
    cli::cli_abort("All elements of {.arg ...} must be named.")
  }
  con <- dbplyr::remote_con(.data)
  in_query <- dbplyr::sql_render(.data)

  col_nms <- colnames(.data)

  # nesting_plan contains information and subqueries, one row for each nesting column
  nesting_plan <- purrr::imap_dfr(dots, function(quo_selection, nesting_name) {
    nested_names <- tidyselect::vars_select(col_nms, !!quo_selection)
    if (!is.null(.names_sep)) {
      new_nested_names <-
        remove_prefix_and_sep(nested_names, prefix = nesting_name, sep = .names_sep)
      selected <-
        glue_sql("{`nested_names`} {`new_nested_names`}", .con = con) %>%
        glue_collapse(", ")
    } else {
      new_nested_names <- nested_names
      selected <- glue_sql("{`nested_names`*}", .con = con)
    }
    inner_query_alias <- glue("*tmp_{nesting_name}*")
    inner_query <- glue::glue_sql(
      "SELECT ",
      selected,
      " FROM (",
      in_query,
      ") {`inner_query_alias`}",
      .con = con
    )

    tibble::tibble_row(
      nesting_name = nesting_name,
      nested_names = list(nested_names),
      new_nested_names = list(new_nested_names),
      selected = selected,
      inner_query = inner_query,
    )
  })

  id_cols <- setdiff(col_nms, unlist(nesting_plan$nested_names))
  temp_alias <- "*tmp*"
  join_constraint <-
    glue_sql("{`id_cols`} = {`temp_alias`}.{`id_cols`}", .con = con) %>%
    glue_collapse(" AND ")

  nesting_plan$from_json_path_query <- glue::glue_sql(
    "(",
    nesting_plan$inner_query,
    " WHERE (",
    join_constraint,
    ") FOR JSON PATH) AS {`nesting_plan$nesting_name`}",
    .con = con
  )

  # build final query
  out_query <- glue_sql(
    "SELECT {`id_cols`*}, ",
    glue_collapse(nesting_plan$from_json_path_query, ", "),
    " FROM (",
    in_query,
    ") {`temp_alias`} GROUP BY {`id_cols`*}",
    .con = con
  )

  tbl(con, sql(out_query), vars = c(id_cols, nesting_plan$nesting_name))
}

#' dmSVG
#'
#' Widget
#'
#' @noRd
dmSVG <- function(
  dm,
  viewBox = TRUE,
  node_to_zoom = NULL,
  nodes_to_select = NULL,
  ...,
  width = NULL,
  height = NULL,
  elementId = NULL
) {
  svg_text <- dm_to_svg(dm)

  # forward options using x
  x <- list(
    svg = htmltools::HTML(svg_text),
    config = list(...),
    options = list(
      viewBox = viewBox
    ),
    panZoomOpts = list(
      node_id = node_to_zoom,
      nodes_to_select = nodes_to_select
    )
  )

  # create widget
  htmlwidgets::createWidget(
    name = "dmSVG",
    x,
    width = width,
    height = height,
    package = utils::packageName(),
    elementId = elementId
  )
}

#' Shiny bindings for dmSVG
#'
#' Output and render functions for using dmSVG within Shiny
#' applications and interactive Rmd documents.
#'
#' @param outputId output variable to read from
#' @param width,height Must be a valid CSS unit (like \code{'100\%'},
#'   \code{'400px'}, \code{'auto'}) or a number, which will be coerced to a
#'   string and have \code{'px'} appended.
#' @param expr An expression that generates a dmSVG
#' @param env The environment in which to evaluate \code{expr}.
#' @param quoted Is \code{expr} a quoted expression (with \code{quote()})? This
#'   is useful if you want to save an expression in a variable.
#'
#' @name dmSVG-shiny
#'
#' @noRd
dmSVGOutput <- function(outputId, width = "100%", height = "400px") {
  htmlwidgets::shinyWidgetOutput(outputId, "dmSVG", width, height, package = "dm")
}

#' @rdname dmSVG-shiny
#' @noRd
renderDmSVG <- function(expr, env = parent.frame(), quoted = FALSE) {
  if (!quoted) {
    expr <- substitute(expr)
  } # force quoted
  htmlwidgets::shinyRenderWidget(expr, dmSVGOutput, env, quoted = TRUE)
}

library(tidyverse)
library(dm)

library(rlang)

dm <- dm_nycflights13()
dm %>% dm_draw()

create_exist <- function(dm, child, parent) {
  child <- ensym(child)
  parent <- ensym(parent)
  parent.exist <- sym(paste0(as.character(parent), ".exist"))
  pk_col <- sym(dm_get_pk(dm, !!parent))

  dm %>%
    dm_zoom_to(!!parent) %>%
    mutate(!!parent.exist := 1L) %>%
    dm_update_zoomed() %>%
    dm_zoom_to(!!child) %>%
    left_join(!!parent, select = c(!!pk_col, !!parent.exist)) %>%
    mutate(!!parent.exist := coalesce(!!parent.exist, 0L)) %>%
    dm_update_zoomed() %>%
    dm_zoom_to(!!parent) %>%
    select(-!!parent.exist) %>%
    dm_update_zoomed()
}

dm %>%
  create_exist(flights, airlines) %>%
  create_exist(flights, planes) %>%
  create_exist(flights, airports)

#' Shiny app for defining dm objects
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' This function starts a Shiny application that allows to define `dm` objects
#' from a database or from local data frames.
#' The application generates R code that can be inserted or copy-pasted
#' into an R script or function.
#'
#' @details
#' In a future release, the app will also allow composing `dm` objects directly
#' from database connections or data frames.
#'
#' The signature of this function is subject to change without notice.
#' This should not pose too many problems, because it will usually be run
#' interactively.
#'
#' @inheritParams rlang::args_dots_empty
#' @param dm An initial dm object, currently required.
#' @param select_tables Show selectize input to select tables?
#' @param debug Set to `TRUE` to simplify debugging of the app.
#'
#' @export
#' @examples
#' \dontrun{
#' dm <- dm_nycflights13(cycle = TRUE)
#' dm_gui(dm = dm)
#' }
dm_gui <- function(..., dm = NULL, select_tables = TRUE, debug = FALSE) {
  check_dots_empty()

  check_suggested(
    c(
      "colourpicker",
      "htmltools",
      "htmlwidgets",
      "reactable",
      "rstudioapi",
      "shiny",
      "shinyAce",
      "shinydashboard"
    ),
    "dm_gui"
  )

  dm <- dm_ptype(dm)

  gui_run(dm, select_tables, debug)
}

#' @autoglobal
build_copy_queries <- function(
  dest,
  dm,
  set_key_constraints = TRUE,
  temporary = TRUE,
  table_names = set_names(names(dm))
) {
  stopifnot(!is_src_db(dm))

  ## Reorder queries according to topological sort so pks are created before associated fks
  graph <- create_graph_from_dm(dm, directed = TRUE)
  topo <- names(graph_topo_sort(graph, mode = "in"))

  if (length(topo) == length(dm)) {
    dm <- dm[topo]
  }

  con <- con_from_src_or_con(dest)

  ## helper to quote all elements of a column and enumerate (concat) element wise
  quote_enum_col <- function(x) {
    map_chr(x, ~ toString(map_chr(.x, DBI::dbQuoteIdentifier, conn = con)))
  }

  ## helper to set on delete statement for fks if required
  set_on_delete_col <- function(x) {
    if (is_duckdb(dest) && any(x == "cascade")) {
      inform(glue('`on_delete = "cascade"` not supported for duckdb'))
      ""
    } else {
      map_chr(
        x,
        ~ {
          switch(
            .x,
            "no_action" = "",
            "cascade" = " ON DELETE CASCADE",
            cli::cli_abort('{.code on_delete = {.val {.x}}} is not supported.')
          )
        }
      )
    }
  }

  ## fetch types, keys and uniques
  pks <- dm_get_all_pks_impl(dm) %>% rename(name = table)
  fks <- dm_get_all_fks_impl(dm)

  # if a that doesn't match a pk, this non-pk col should be unique
  uniques <-
    fks %>%
    select(name = parent_table, pk_col = parent_key_cols) %>%
    anti_join(pks, by = c("name", "pk_col")) %>%
    distinct()

  ## build sql definitions to use in `CREATE TABLE ...`

  # column definitions
  get_sql_col_types <- function(x) {
    # autoincrementing is not possible for composite keys, so `pk_col` is guaranteed
    # to be a scalar
    pk_col <-
      dm %>%
      dm_get_all_pks(!!x) %>%
      filter(autoincrement) %>%
      pull(pk_col)

    tbl <- tbl_impl(dm, x)
    types <- DBI::dbDataType(con, tbl)
    autoincrement_attribute <- ""

    # database-specific type conversions
    if (is_mariadb(dest)) {
      if (nrow(tbl) == 0) {
        types[types == "TEXT"] <- "VARCHAR(255)"
      } else {
        for (i in which(types == "TEXT")) {
          chars <- max(nchar(tbl[[i]], type = "bytes", keepNA = FALSE))
          if (chars <= 255) {
            types[[i]] <- paste0("VARCHAR(255)")
          }
        }
      }
    }
    if (is_sqlite(dest)) {
      types[types == "INT"] <- "INTEGER"
    }

    # database-specific autoincrementing column types
    if (length(pk_col) > 0L) {
      # extract column name representing primary key
      pk_col <- pk_col[[1]]

      # Postgres:
      if (is_postgres(dest) || is_redshift(dest)) {
        types[pk_col] <- "SERIAL"
      }

      # SQL Server:
      if (is_mssql(dest)) {
        types[pk_col] <- paste0(types[pk_col], " IDENTITY")
      }

      # MariaDB:
      # Doesn't have a special data type. Uses `AUTO_INCREMENT` attribute instead.
      # Ref: https://mariadb.com/kb/en/auto_increment/
      if (is_mariadb(dest)) {
        autoincrement_attribute <- " AUTO_INCREMENT"
      }

      # DuckDB:
      # Doesn't have a special data type. Uses `CREATE SEQUENCE` instead.
      # Ref: https://duckdb.org/docs/sql/statements/create_sequence

      # SQLite:
      # For a primary key, autoincrementing works by default, and it is almost never
      # necessary to use the `AUTOINCREMENT` keyword. So nothing we need to do here.
      # Ref: https://www.sqlite.org/autoinc.html
    }
    df_col_types <-
      enframe(types, "col", "type") %>%
      mutate(autoincrement_attribute = "")

    if (length(pk_col) > 0L) {
      df_col_types <-
        df_col_types %>%
        mutate(
          autoincrement_attribute = if_else(
            col == pk_col,
            !!autoincrement_attribute,
            autoincrement_attribute
          )
        )
    }

    df_col_types
  }

  tbl_defs <- tibble(name = names(dm))

  col_defs <-
    dm %>%
    src_tbls_impl() %>%
    set_names() %>%
    map_dfr(get_sql_col_types, .id = "name") %>%
    mutate(col_def = glue("{DBI::dbQuoteIdentifier(con, col)} {type}{autoincrement_attribute}")) %>%
    group_by(name) %>%
    summarize(
      col_defs = paste(col_def, collapse = ",\n  "),
      columns = list(col)
    )

  # default values
  pk_defs <- tibble(name = character(0), pk_defs = character(0))
  fk_defs <- tibble(name = character(0), fk_defs = character(0))
  unique_defs <- tibble(name = character(0), unique_defs = character(0))
  index_queries <- tibble(
    name = character(0),
    sql_index = list(),
    index_name = list()
  )

  if (set_key_constraints) {
    # primary key definitions
    pk_defs <-
      pks %>%
      transmute(
        name,
        pk_defs = paste0("PRIMARY KEY (", quote_enum_col(pk_col), ")")
      )

    # unique constraint definitions
    unique_defs <-
      uniques %>%
      transmute(
        name,
        unique_def = paste0(
          "UNIQUE (",
          quote_enum_col(pk_col),
          ")"
        )
      ) %>%
      group_by(name) %>%
      summarize(unique_defs = paste(unique_def, collapse = ",\n  "))

    # foreign key definitions and indexing queries
    # https://github.com/r-lib/rlang/issues/1422
    if (is_mariadb(con) && temporary) {
      if (nrow(fks) > 0 && !is_testing()) {
        warn(
          "MySQL and MariaDB don't support foreign keys for temporary tables, these won't be set in the remote database but are preserved in the `dm`"
        )
      }
    } else {
      fk_defs <-
        fks %>%
        transmute(
          name = child_table,
          fk_def = paste0(
            "FOREIGN KEY (",
            quote_enum_col(child_fk_cols),
            ") REFERENCES ",
            purrr::map_chr(table_names[fks$parent_table], ~ DBI::dbQuoteIdentifier(con, .x)),
            " (",
            quote_enum_col(parent_key_cols),
            ")",
            set_on_delete_col(on_delete)
          )
        ) %>%
        group_by(name) %>%
        summarize(fk_defs = paste(fk_def, collapse = ",\n  "))

      index_queries <-
        fks %>%
        mutate(
          name = child_table,
          index_name = map_chr(child_fk_cols, paste, collapse = "_"),
          remote_name = purrr::map_chr(table_names[name], ~ DBI::dbQuoteIdentifier(con, .x)),
          remote_name_unquoted = map_chr(
            DBI::dbUnquoteIdentifier(con, DBI::SQL(remote_name)),
            ~ .x@name[[length(.x@name)]]
          ),
          index_name = make.unique(paste0(remote_name_unquoted, "__", index_name), sep = "__")
        ) %>%
        group_by(name) %>%
        summarize(
          sql_index = list(DBI::SQL(paste0(
            "CREATE INDEX ",
            index_name,
            " ON ",
            remote_name,
            " (",
            quote_enum_col(child_fk_cols),
            ")"
          ))),
          index_name = list(index_name)
        )
    }
  }

  ## compile `CREATE TABLE ...` queries
  create_table_queries <-
    tbl_defs %>%
    left_join(col_defs, by = "name") %>%
    left_join(pk_defs, by = "name") %>%
    left_join(unique_defs, by = "name") %>%
    left_join(fk_defs, by = "name") %>%
    group_by(name, columns) %>%
    mutate(
      remote_name = table_names[name],
      all_defs = paste(
        Filter(
          Negate(is.na),
          c(col_defs, pk_defs, unique_defs, fk_defs)
        ),
        collapse = ",\n  "
      )
    ) %>%
    ungroup() %>%
    transmute(
      name,
      remote_name,
      columns,
      sql_table = DBI::SQL(glue(
        "CREATE {if (temporary) 'TEMPORARY ' else ''}TABLE {purrr::map_chr(remote_name, ~ DBI::dbQuoteIdentifier(con, .x))} (\n  {all_defs}\n)"
      ))
    )

  queries <-
    tbl_defs %>%
    left_join(create_table_queries, by = "name") %>%
    left_join(index_queries, by = "name")

  queries
}

# Pure R fallback implementations of all graph_* wrapper functions.
# These are assigned to the graph_* names in .onLoad() if igraph is not installed.
# Each function takes a dm_graph object (not dm_igraph) and returns a dm_graph.

# dm_graph: minimal pure R graph representation.
# Fields: directed (logical), vnames (character), from (integer), to (integer).
new_dm_graph <- function(directed, vnames, from, to) {
  structure(
    list(directed = directed, vnames = vnames, from = from, to = to),
    class = "dm_graph"
  )
}

graph_from_data_frame_fallback <- function(d, directed, vertices = NULL) {
  if (is.null(vertices)) {
    vnames <- unique(c(as.character(d[[1]]), as.character(d[[2]])))
  } else {
    vnames <- as.character(vertices)
  }
  vidx <- set_names(seq_along(vnames), vnames)
  from <- unname(vidx[as.character(d[[1]])])
  to <- unname(vidx[as.character(d[[2]])])
  # For undirected graphs, normalize so the lower-indexed vertex is always
  # "from" and the higher-indexed is "to", matching igraph's behavior.
  if (!directed) {
    swap <- from > to
    tmp <- from[swap]
    from[swap] <- to[swap]
    to[swap] <- tmp
  }
  new_dm_graph(directed = directed, vnames = vnames, from = from, to = to)
}

# Returns a named integer vector: values are 1-based indices, names are vertex names.
graph_vertices_fallback <- function(g) {
  set_names(seq_along(g$vnames), g$vnames)
}

# Returns an integer vector with a "vnames" attribute (e.g. "from|to").
graph_edges_fallback <- function(g) {
  out <- seq_along(g$from)
  attr(out, "vnames") <- paste(g$vnames[g$from], g$vnames[g$to], sep = "|")
  out
}

# Returns a list with $order (named integer), $dist (named numeric),
# $parent (named integer).
graph_dfs_fallback <- function(g, root, unreachable = TRUE, parent = FALSE, dist = FALSE) {
  n <- length(g$vnames)
  vidx <- set_names(seq_along(g$vnames), g$vnames)
  root_idx <- vidx[[root]]

  visited <- logical(n)
  visit_order <- integer(0)
  parent_vec <- set_names(rep(NA_integer_, n), g$vnames)
  dist_vec <- set_names(rep(Inf, n), g$vnames)

  dfs_visit <- function(v, d) {
    visited[v] <<- TRUE
    visit_order <<- c(visit_order, v)
    dist_vec[v] <<- d
    nbrs <- if (g$directed) g$to[g$from == v] else unique(c(g$to[g$from == v], g$from[g$to == v]))
    for (nbr in nbrs) {
      if (!visited[nbr]) {
        parent_vec[nbr] <<- v
        dfs_visit(nbr, d + 1)
      }
    }
  }

  dfs_visit(root_idx, 0)

  n_visited <- length(visit_order)
  n_unvisited <- n - n_visited
  order_result <- set_names(
    c(visit_order, rep(NA_integer_, n_unvisited)),
    c(g$vnames[visit_order], rep(NA_character_, n_unvisited))
  )

  result <- list(order = order_result)
  if (dist) {
    result$dist <- dist_vec
  }
  if (parent) {
    result$parent <- parent_vec
  }
  result
}

# Returns a named integer vector (vertex indices, vertex names as names).
# mode = "out": for edge u→v, u comes before v.
# mode = "in":  for edge u→v, v comes before u.
graph_topo_sort_fallback <- function(g, mode = "out") {
  n <- length(g$vnames)
  if (mode == "in") {
    from <- g$to
    to <- g$from
  } else {
    from <- g$from
    to <- g$to
  }

  # Kahn's algorithm for topological sort
  in_degree <- tabulate(to, nbins = n)
  queue <- which(in_degree == 0L)
  result <- integer(0)

  while (length(queue) > 0L) {
    v <- queue[[1L]]
    queue <- queue[-1L]
    result <- c(result, v)
    for (u in to[from == v]) {
      in_degree[[u]] <- in_degree[[u]] - 1L
      if (in_degree[[u]] == 0L) {
        queue <- c(queue, u)
      }
    }
  }

  set_names(result, g$vnames[result])
}

# Returns a matrix: rows = sources, columns = all vertices.
graph_distances_fallback <- function(g, v = NULL) {
  n <- length(g$vnames)

  # Build adjacency list (undirected: use both directions)
  adj <- vector("list", n)
  for (i in seq_along(g$from)) {
    adj[[g$from[[i]]]] <- c(adj[[g$from[[i]]]], g$to[[i]])
    adj[[g$to[[i]]]] <- c(adj[[g$to[[i]]]], g$from[[i]])
  }

  bfs_dist <- function(start_idx) {
    d <- rep(Inf, n)
    d[[start_idx]] <- 0
    queue <- start_idx
    while (length(queue) > 0L) {
      curr <- queue[[1L]]
      queue <- queue[-1L]
      for (nbr in adj[[curr]]) {
        if (is.infinite(d[[nbr]])) {
          d[[nbr]] <- d[[curr]] + 1
          queue <- c(queue, nbr)
        }
      }
    }
    d
  }

  vidx <- set_names(seq_along(g$vnames), g$vnames)
  start_indices <- vidx[v]
  dist_mat <- t(vapply(start_indices, bfs_dist, numeric(n)))
  rownames(dist_mat) <- v
  colnames(dist_mat) <- g$vnames
  dist_mat
}

# Returns a new dm_graph containing only the specified vertices and their edges.
graph_induced_subgraph_fallback <- function(g, vids) {
  keep_mask <- g$vnames %in% vids
  new_vnames <- g$vnames[keep_mask]
  old_indices <- which(keep_mask)
  # Map old integer indices to new integer indices
  idx_map <- set_names(seq_along(old_indices), as.character(old_indices))
  keep_edges <- g$from %in% old_indices & g$to %in% old_indices
  new_from <- unname(idx_map[as.character(g$from[keep_edges])])
  new_to <- unname(idx_map[as.character(g$to[keep_edges])])
  new_dm_graph(directed = g$directed, vnames = new_vnames, from = new_from, to = new_to)
}

# Returns a list with $predecessors: named integer vector of predecessor vertex indices.
# names(result$predecessors) are the predecessor vertex names (NA for source vertex).
graph_shortest_paths_fallback <- function(g, from, to, predecessors = FALSE) {
  n <- length(g$vnames)
  vidx <- set_names(seq_along(g$vnames), g$vnames)
  start_idx <- vidx[[from]]

  # Build adjacency list (undirected)
  adj <- vector("list", n)
  for (i in seq_along(g$from)) {
    adj[[g$from[[i]]]] <- c(adj[[g$from[[i]]]], g$to[[i]])
    adj[[g$to[[i]]]] <- c(adj[[g$to[[i]]]], g$from[[i]])
  }

  pred_idx <- set_names(rep(NA_integer_, n), g$vnames)
  visited <- logical(n)
  visited[[start_idx]] <- TRUE
  queue <- start_idx

  while (length(queue) > 0L) {
    curr <- queue[[1L]]
    queue <- queue[-1L]
    for (nbr in adj[[curr]]) {
      if (!visited[[nbr]]) {
        visited[[nbr]] <- TRUE
        pred_idx[[nbr]] <- curr
        queue <- c(queue, nbr)
      }
    }
  }

  result <- list()
  if (predecessors) {
    pred_names <- ifelse(is.na(pred_idx), NA_character_, g$vnames[pred_idx])
    result$predecessors <- set_names(pred_idx, pred_names)
  }
  result
}

# Returns a new dm_graph with the specified vertices (and incident edges) removed.
graph_delete_vertices_fallback <- function(g, v) {
  graph_induced_subgraph_fallback(g, setdiff(g$vnames, v))
}

# Returns a named integer vector of neighbor vertex indices.
# mode = "in": vertices with edges pointing TO v
# mode = "out": vertices with edges FROM v
# mode = "all": both
graph_neighbors_fallback <- function(g, v, mode = "all") {
  v_idx <- if (is.character(v)) {
    set_names(seq_along(g$vnames), g$vnames)[[v]]
  } else {
    as.integer(v)
  }
  incoming <- if (mode %in% c("in", "all")) g$from[g$to == v_idx] else integer(0)
  outgoing <- if (mode %in% c("out", "all")) g$to[g$from == v_idx] else integer(0)
  nbrs <- unique(c(incoming, outgoing))
  set_names(nbrs, g$vnames[nbrs])
}

# Returns the number of vertices in the graph.
graph_vcount_fallback <- function(g) {
  length(g$vnames)
}

# Returns a list with $girth (numeric) and $circle (named integer vector).
graph_girth_fallback <- function(g) {
  n <- length(g$vnames)
  if (n == 0L) {
    return(list(girth = Inf, circle = set_names(integer(0), character(0))))
  }

  # Build adjacency list (always undirected, matching igraph::girth which
  # treats directed graphs as undirected)
  adj <- vector("list", n)
  for (i in seq_along(g$from)) {
    adj[[g$from[[i]]]] <- c(adj[[g$from[[i]]]], g$to[[i]])
    adj[[g$to[[i]]]] <- c(adj[[g$to[[i]]]], g$from[[i]])
  }

  # BFS-based girth algorithm (matches igraph's igraph_girth implementation)
  mincirc <- .Machine$integer.max
  minvertex <- 1L
  t1 <- 0L
  t2 <- 0L
  stoplevel <- n + 1L
  triangle <- FALSE

  for (node in seq_len(n)) {
    if (triangle) {
      break
    }

    level <- integer(n)
    level[[node]] <- 1L
    queue <- node

    while (length(queue) > 0L) {
      actnode <- queue[[1L]]
      queue <- queue[-1L]
      actlevel <- level[[actnode]]

      if (actlevel >= stoplevel) {
        break
      }

      neis <- adj[[actnode]]
      if (is.null(neis)) {
        next
      }

      for (nei in neis) {
        neilevel <- level[[nei]]
        if (neilevel != 0L) {
          if (neilevel == actlevel - 1L) {
            # Parent edge, skip
          } else {
            # Found cycle
            stoplevel <- neilevel
            if (actlevel < mincirc) {
              # Cycle length: levels are 1-based, so path lengths are (actlevel-1) + (neilevel-1) + 1
              mincirc <- actlevel + neilevel - 1L
              minvertex <- node
              t1 <- actnode
              t2 <- nei
              if (neilevel == 2L) {
                triangle <- TRUE
              }
            }
            if (neilevel == actlevel) break
          }
        } else {
          queue <- c(queue, nei)
          level[[nei]] <- actlevel + 1L
        }
      }
    }
  }

  if (mincirc >= .Machine$integer.max) {
    return(list(girth = Inf, circle = set_names(integer(0), character(0))))
  }

  # Reconstruct cycle via BFS from minvertex
  circle <- integer(mincirc)
  parent <- integer(n)
  parent[[minvertex]] <- minvertex
  queue <- minvertex

  while (parent[[t1]] == 0L || parent[[t2]] == 0L) {
    actnode <- queue[[1L]]
    queue <- queue[-1L]
    neis <- adj[[actnode]]
    if (!is.null(neis)) {
      for (nei in neis) {
        if (parent[[nei]] == 0L) {
          parent[[nei]] <- actnode
          queue <- c(queue, nei)
        }
      }
    }
  }

  # Trace path from t1 to minvertex
  idx <- 1L
  curr <- t1
  while (curr != minvertex) {
    circle[[idx]] <- curr
    idx <- idx + 1L
    curr <- parent[[curr]]
  }
  circle[[idx]] <- minvertex

  # Trace path from t2 to minvertex (fill from the end)
  idx <- mincirc
  curr <- t2
  while (curr != minvertex) {
    circle[[idx]] <- curr
    idx <- idx - 1L
    curr <- parent[[curr]]
  }

  list(girth = mincirc, circle = set_names(circle, g$vnames[circle]))
}

# FIXME: Adapt argument list as provided/required by the Shiny app
mw_cg_make_dm_select <- function(dm, ..., table_names = NULL, column_names = NULL, rm = FALSE) {
  check_dots_empty()

  # Checks
  stopifnot(is.character(table_names))
  stopifnot(table_names %in% names(dm))
  stopifnot(length(table_names) == 1)

  col_names_sel_rm <- if (rm) {
    sapply(syms(column_names), function(x) {
      call("-", x)
    })
  } else {
    syms(column_names)
  }

  list(
    call = expr(dm_select(
      .,
      !!sym(table_names),
      !!!col_names_sel_rm
    ))
  )
}

#' Check if suggested package is installed
#'
#' @param packages vector of package names to check
#' @param use whether to trigger the check, if `NA` the check is triggered only
#'   if `is_interactive()` is `TRUE`
#' @param top_level_fun the name of the top level function called, not used if
#'   `message` is provided
#' @param message optional custom message, by default the message follows a template
#' @return whether check was triggered and all packages are installed
#' @noRd
check_suggested <- function(packages,
                            use,
                            top_level_fun = NULL,
                            message = NULL,
                            version = NULL) {
  # If NA, inform that package isn't installed, but only in interactive mode
  if (is.na(use)) {
    use <- is_interactive()
    if (!use) {
      return(FALSE)
    }

    installed <- map_lgl(packages, ~ {
      installed <- is_installed(., version = version)
      if (!installed) {
        if (is.null(message)) {
          # FIXME: Mention version
          message <- glue("`{top_level_fun}()` is improved by the '{.}' package. Consider `install.packages(\"{.}\")`.")
        }
        inform(message)
      }
      installed
    })
    return(all(installed))
  }

  # If FALSE, hide
  if (!use) {
    return(FALSE)
  }

  # If TRUE, fail if package is not installed
  for (pkg in packages) {
    if (!is_installed(pkg)) {
      if (is.null(message)) {
        message <- glue("`{top_level_fun}()` needs the '{pkg}' package. Do you need `install.packages(\"{pkg}\")` ?")
      }
      if (is_testing()) {
        testthat::skip(message)
      } else {
        abort(message)
      }
    }
  }
  TRUE
}

if (!rlang::is_attached("dm_cache")) {
  ((attach))(new_environment(), pos = length(search()) - 1, name = "dm_cache")
}
cache <- search_env("dm_cache")

`%<-%` <- function(lhs, rhs, env = caller_env()) {
  defer_assign({{ lhs }}, copy_to_my_test_src(rhs, {{ lhs }}), env)
}

`%<--%` <- function(lhs, rhs, env = caller_env()) {
  defer_assign({{ lhs }}, rhs, env)
}

defer_assign <- function(lhs, rhs, env) {
  lhs <- as_name(ensym(lhs))

  value <- get0(lhs, cache)
  if (is.null(value)) {
    message("Deferring ", lhs)

    # Enable this for eager assignment:
    # force(rhs)

    value <- function() {
      # message("Querying ", lhs)
      out <- rhs
      out
    }
    assign(lhs, value, cache)
  } else {
    message("Using cached ", lhs)
  }
  assign(lhs, value, env)
  invisible(value)
}

copy_to_my_test_src <- function(rhs, lhs) {
  name <- as_name(ensym(lhs))
  # message("Evaluating ", name)

  src <- my_test_src()
  if (is.null(src)) {
    rhs
  } else if (is_dm(rhs)) {
    # We want all dm operations to work with key constraints on the database
    # (except for bad_dm)
    # message(name)
    suppressMessages(copy_dm_to(src, rhs))
  } else if (inherits(rhs, "list")) {
    suppressMessages(
      map(rhs, ~ copy_to(src, .x, name = unique_db_table_name(name), temporary = TRUE))
    )
  } else {
    suppressMessages(copy_to(src, rhs, name = name, temporary = TRUE))
  }
}

my_test_src_name <- {
  src <- Sys.getenv("DM_TEST_SRC")
  # Allow set but empty DM_TEST_SRC environment variable
  if (src == "") {
    src <- "df"
  }
  name <- gsub("^.*-", "", src)
  cli::cli_inform(cli::col_green("Testing on {name}"))
  name
}

is_db_test_src <- function() {
  my_test_src_name != "df"
}

my_test_src_fun %<--%
  {
    fun <- paste0("test_src_", my_test_src_name)
    get0(fun, inherits = TRUE)
  }

my_test_src_cache %<--%
  {
    my_test_src_fun()()
  }

my_test_src <- function() {
  testthat::skip_if_not_installed("dbplyr")

  fun <- my_test_src_fun()
  if (is.null(fun)) {
    abort(paste0("Data source not known: ", my_test_src_name))
  }
  tryCatch(
    my_test_src_cache(),
    error = function(e) {
      abort(paste0("Data source ", my_test_src_name, " not accessible: ", conditionMessage(e)))
    }
  )
}

my_test_con <- function() {
  # FIXME: Remove my_test_src()
  con_from_src_or_con(my_test_src())
}

duckdb_test_src %<--%
  {
    if (getRversion() < "4.0") {
      testthat::skip("duckdb failing for R < 4.0")
    }
    testthat::skip_if_not_installed("duckdb")
    dbplyr::src_dbi(DBI::dbConnect(duckdb::duckdb()), auto_disconnect = TRUE)
  }

my_db_test_src <- function() {
  testthat::skip_if_not_installed("dbplyr")

  if (is_db_test_src()) {
    my_test_src()
  } else {
    duckdb_test_src()
  }
}

my_db_test_con <- function() {
  con_from_src_or_con(my_db_test_src())
}

test_src_frame <- function(..., .temporary = TRUE, .env = parent.frame(), .unique_indexes = NULL) {
  src <- my_test_src()

  df <- tibble(...)
  if (is.null(src)) {
    return(df)
  }

  if (!.temporary) {
    name <- unique_db_table_name("test_frame")
    temporary <- FALSE
  } else if (is_mssql(src)) {
    name <- paste0("#", unique_db_table_name("test_frame"))
    temporary <- FALSE
  } else {
    name <- unique_db_table_name("test_frame")
    temporary <- TRUE
  }

  out <- copy_to(src, df, name = name, temporary = temporary, unique_indexes = .unique_indexes)
  out
}

test_db_src_frame <- function(
  ...,
  .temporary = TRUE,
  .env = parent.frame(),
  .unique_indexes = NULL
) {
  if (is_db_test_src()) {
    return(test_src_frame(
      ...,
      .temporary = .temporary,
      .env = .env,
      .unique_indexes = .unique_indexes
    ))
  }

  src <- my_db_test_src()

  df <- tibble(...)

  name <- unique_db_table_name("test_frame")

  out <- copy_to(src, df, name = name, temporary = .temporary, unique_indexes = .unique_indexes)

  if (!.temporary) {
    withr::defer(DBI::dbRemoveTable(con_from_src_or_con(src), name), envir = .env)
  }

  out
}


# for examine_cardinality...() ----------------------------------------------

data_card_1 %<-% tibble::tibble(a = 1:5, b = letters[1:5])
data_card_1_duckdb %<--% copy_to(duckdb_test_src(), data_card_1())
data_card_2 %<-% tibble::tibble(a = c(1, 3:6), b = letters[1:5])
data_card_3 %<-% tibble::tibble(c = 1:5)
data_card_4 %<-% tibble::tibble(c = c(1:5, 5L))
data_card_5 %<-% tibble::tibble(a = 1:5)
data_card_6 %<-% tibble::tibble(c = 1:4)
data_card_7 %<-% tibble::tibble(c = c(1:5, 5L, 6L))
data_card_8 %<-% tibble::tibble(c = c(1:6))
data_card_9 %<-% tibble::tibble(c = c(1:5, NA))
data_card_10 %<-% tibble::tibble(c = c(1:3, 4:3, NA))
data_card_11 %<-% tibble::tibble(a = 1:4, b = letters[1:4])
data_card_12 %<-% tibble::tibble(a = c(1:5, 5L), b = letters[c(1:5, 5L)])
data_card_13 %<-% tibble::tibble(a = 1:6, b = letters[1:6])

dm_for_card %<--%
  {
    dm(
      dc_1 = data_card_1(),
      dc_2 = data_card_11(),
      dc_3 = data_card_12(),
      dc_4 = data_card_13(),
      dc_5 = suppress_mssql_message(compute(data_card_1())),
      dc_6 = data_card_7()
    ) %>%
      dm_add_fk(dc_2, c(a, b), dc_1, c(a, b)) %>%
      dm_add_fk(dc_3, c(a, b), dc_1, c(a, b)) %>%
      dm_add_fk(dc_3, c(b, a), dc_4, c(b, a)) %>%
      dm_add_fk(dc_4, c(b, a), dc_3, c(b, a)) %>%
      dm_add_fk(dc_5, c(b, a), dc_1, c(b, a)) %>%
      dm_add_fk(dc_6, c, dc_1, a)
  }

# for check_key() ---------------------------------------------------------

data_mcard %<-%
  tribble(
    ~c1 , ~c2 , ~c3 ,
      1 ,   2 ,   3 ,
      4 ,   5 ,   3 ,
      1 ,   2 ,   4
  )

data_mcard_1 %<-% tibble(a = c(1, 2, 1), b = c(1, 4, 1), c = c(5, 6, 7))
data_mcard_2 %<-% tibble(a = c(1, 2, 3), b = c(4, 5, 6), c = c(7, 8, 9))
data_mcard_3 %<-% tibble(a = c(2, 1, 2), b = c(4, 5, 6), c = c(7, 8, 9))

# for table-surgery functions ---------------------------------------------

data_ts %<-%
  tibble(
    a = as.integer(c(1, 2, 1)),
    b = c(1.1, 4.2, 1.1),
    c = as.integer(c(5, 6, 7)),
    d = c("a", "b", "c"),
    e = c("c", "b", "c"),
    f = c(TRUE, FALSE, TRUE)
  )

data_ts_child %<-%
  tibble(
    b = c(1.1, 4.2, 1.1),
    c = as.integer(c(5, 6, 7)),
    d = c("a", "b", "c"),
    aef_id = as.integer(c(1, 2, 1)),
  )

data_ts_parent %<-%
  tibble(
    aef_id = as.integer(c(1, 2)),
    a = as.integer(c(1, 2)),
    e = c("c", "b"),
    f = c(TRUE, FALSE)
  )

list_of_data_ts_parent_and_child %<--%
  list(
    child_table = data_ts_child(),
    parent_table = data_ts_parent()
  )

# for testing filter and semi_join ---------------------------------------------

# the following is for testing the filtering functionality:
tf_1 %<-%
  tibble(
    a = 1:10,
    b = LETTERS[1:10]
  )

tf_2_simple %<-%
  tibble(
    c = c("elephant", "lion", "seal", "worm", "dog", "cat"),
    d = 2:7,
    e = c(LETTERS[4:7], LETTERS[5:6])
  )

tf_2 %<-%
  tibble(
    c = c("elephant", "lion", "seal", "worm", "dog", "cat"),
    d = 2:7,
    e = c(LETTERS[4:7], LETTERS[5:6]),
    e1 = c(4:7, 5:6),
  )

tf_3_simple %<-%
  tibble(
    f = LETTERS[2:11],
    g = c("one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten")
  )

tf_3 %<-%
  tibble(
    f = LETTERS[c(3, 3:11)],
    f1 = c(2:7, 7L, 7L, 10:11),
    g = c("one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten")
  )

tf_4 %<-%
  tibble(
    h = letters[1:5],
    i = c("three", "four", "five", "six", "seven"),
    j = c(LETTERS[3:6], LETTERS[6]),
    j1 = c(3:6, 6L),
  )

tf_5 %<-%
  tibble(
    ww = 2L,
    k = 1:4,
    l = letters[2:5],
    m = c("house", "tree", "streetlamp", "streetlamp")
  )

tf_6 %<-%
  tibble(
    zz = 1L,
    n = c("house", "tree", "hill", "streetlamp", "garden"),
    o = letters[5:9]
  )

tf_7 %<-%
  tibble(
    p = letters[4:9],
    q = c("elephant", "lion", "seal", "worm", "dog", "cat")
  )

dm_for_filter_w_cycle %<-%
  {
    dm(
      tf_1 = tf_1(),
      tf_2 = tf_2(),
      tf_3 = tf_3(),
      tf_4 = tf_4(),
      tf_5 = tf_5(),
      tf_6 = tf_6(),
      tf_7 = tf_7()
    ) %>%
      dm_add_pk(tf_1, a, autoincrement = TRUE) %>%
      #
      dm_add_pk(tf_3, c(f, f1)) %>%
      dm_add_uk(tf_3, g) %>%
      #
      dm_add_pk(tf_2, c) %>%
      dm_add_fk(tf_2, d, tf_1) %>%
      dm_add_fk(tf_2, c(e, e1), tf_3) %>%
      #
      dm_add_pk(tf_4, h) %>%
      dm_add_fk(tf_4, c(j, j1), tf_3) %>%
      #
      dm_add_pk(tf_7, p) %>%
      dm_add_fk(tf_7, q, tf_2) %>%
      #
      dm_add_pk(tf_6, o) %>%
      dm_add_fk(tf_6, o, tf_7) %>%
      #
      dm_add_pk(tf_5, k) %>%
      dm_add_fk(tf_5, l, tf_4, on_delete = "cascade") %>%
      dm_add_fk(tf_5, m, tf_6, n)
  }

dm_for_filter %<-%
  {
    dm_for_filter_w_cycle() %>%
      dm_select_tbl(-tf_7)
  }

dm_for_filter_db %<--%
  {
    copy_dm_to(my_db_test_src(), dm_for_filter())
  }

dm_for_filter_df %<--%
  {
    # FIXME: Do it the other way round, data frame first, then copy to db
    dm_for_filter() %>%
      collect() %>%
      dm_zoom_to(tf_1) %>%
      arrange(pick(everything())) %>%
      dm_update_zoomed() %>%
      dm_zoom_to(tf_2) %>%
      arrange(pick(everything())) %>%
      dm_update_zoomed() %>%
      dm_zoom_to(tf_3) %>%
      arrange(pick(everything())) %>%
      dm_update_zoomed() %>%
      dm_zoom_to(tf_4) %>%
      arrange(pick(everything())) %>%
      dm_update_zoomed() %>%
      dm_zoom_to(tf_5) %>%
      arrange(pick(everything())) %>%
      dm_update_zoomed() %>%
      dm_zoom_to(tf_6) %>%
      arrange(pick(everything())) %>%
      dm_update_zoomed()
  }

dm_for_filter_duckdb %<--% copy_dm_to(duckdb_test_src(), dm_for_filter())

dm_for_filter_rev %<-%
  {
    def_dm_for_filter <- dm_get_def(dm_for_filter())
    dm_from_def(def_dm_for_filter[rev(seq_len(nrow(def_dm_for_filter))), ])
  }

# Deprecated tests
dm_for_filter_simple %<-%
  {
    dm(
      tf_1 = tf_1(),
      tf_2 = tf_2_simple(),
      tf_3 = tf_3_simple(),
      tf_4 = tf_4(),
      tf_5 = tf_5(),
      tf_6 = tf_6()
    ) %>%
      dm_add_pk(tf_1, a) %>%
      dm_add_pk(tf_3, f) %>%
      #
      dm_add_pk(tf_2, c) %>%
      dm_add_fk(tf_2, d, tf_1) %>%
      dm_add_fk(tf_2, e, tf_3) %>%
      #
      dm_add_pk(tf_4, h) %>%
      dm_add_fk(tf_4, j, tf_3) %>%
      #
      dm_add_pk(tf_6, n) %>%
      #
      dm_add_pk(tf_5, k) %>%
      dm_add_fk(tf_5, l, tf_4) %>%
      dm_add_fk(tf_5, m, tf_6)
  }

dm_for_filter_simple_db %<--%
  {
    copy_dm_to(my_db_test_src(), dm_for_filter_simple())
  }

# for tests on `dm` objects: dm_add_pk(), dm_add_fk() ------------------------

dm_test_obj %<-%
  as_dm(list(
    dm_table_1 = data_card_2(),
    dm_table_2 = data_card_4(),
    dm_table_3 = data_card_7(),
    dm_table_4 = data_card_8(),
    dm_table_5 = data_card_9(),
    dm_table_6 = data_card_10()
  ))

dm_test_obj_2 %<-%
  as_dm(list(
    dm_table_1 = data_card_4(),
    dm_table_2 = data_card_7(),
    dm_table_3 = data_card_8(),
    dm_table_4 = data_card_6()
  ))

# for `dm_nrow()` ---------------------------------------------------------

rows_dm_obj <- 36L

# Complicated `dm` --------------------------------------------------------

dm_more_complex_part %<-%
  {
    dm(
      tf_6_2 = tibble(p = letters[1:6], f = LETTERS[6:11], f1 = c(6:7, 7L, 7L, 10:11)),
      tf_4_2 = tibble(
        r = letters[2:6],
        s = c("three", "five", "six", "seven", "eight"),
        t = c(LETTERS[4:7], LETTERS[5])
      ),
      a = tibble(a_1 = letters[10:18], a_2 = 5:13),
      b = tibble(b_1 = LETTERS[12:15], b_2 = letters[12:15], b_3 = 9:6),
      c = tibble(c_1 = 4:10),
      d = tibble(d_1 = 1:6, b_1 = LETTERS[c(12:14, 13:15)]),
      e = tibble(e_1 = 1:2, b_1 = LETTERS[c(12:13)])
    )
  }

dm_more_complex %<-%
  {
    dm(
      !!!dm_get_tables(dm_for_filter_w_cycle()),
      !!!dm_get_tables(dm_more_complex_part())
    ) %>%
      dm_add_pk(tf_1, a) %>%
      dm_add_pk(tf_2, c) %>%
      dm_add_pk(tf_3, c(f, f1)) %>%
      dm_add_pk(tf_4, h) %>%
      dm_add_pk(tf_4_2, r) %>%
      dm_add_pk(tf_5, k) %>%
      dm_add_pk(tf_6, n) %>%
      dm_add_pk(tf_6_2, p) %>%
      dm_add_pk(a, a_1) %>%
      dm_add_pk(b, b_1) %>%
      dm_add_pk(c, c_1) %>%
      dm_add_pk(d, d_1) %>%
      dm_add_pk(e, e_1) %>%
      dm_add_fk(tf_2, d, tf_1) %>%
      dm_add_fk(tf_2, c(e, e1), tf_3) %>%
      dm_add_fk(tf_4, c(j, j1), tf_3) %>%
      dm_add_fk(tf_5, l, tf_4) %>%
      dm_add_fk(tf_5, l, tf_4_2) %>%
      dm_add_fk(tf_5, m, tf_6) %>%
      dm_add_fk(tf_6_2, c(f, f1), tf_3) %>%
      dm_add_fk(b, b_2, a) %>%
      dm_add_fk(b, b_3, c) %>%
      dm_add_fk(d, b_1, b) %>%
      dm_add_fk(e, b_1, b)
  }

# for testing `dm_disambiguate_cols()` ----------------------------------------

iris_1 %<-%
  {
    datasets::iris %>%
      as_tibble() %>%
      mutate(Species = as.character(Species)) %>%
      mutate(key = row_number()) %>%
      select(key, everything())
  }
iris_2 %<-%
  {
    iris_1() %>%
      mutate(other_col = 1L)
  }
iris_3 %<-%
  {
    iris_2() %>%
      mutate(one_more_col = 1)
  }

iris_1_dis %<-%
  {
    iris_1() %>%
      rename_at(2:6, ~ sub("^", "iris_1.", .))
  }
iris_2_dis %<-%
  {
    iris_2() %>%
      rename_at(1:7, ~ sub("^", "iris_2.", .))
  }
iris_3_dis %<-%
  {
    iris_3() %>%
      rename_at(1:7, ~ sub("^", "iris_3.", .))
  }

dm_for_disambiguate %<-%
  {
    list(iris_1 = iris_1(), iris_2 = iris_2(), iris_3 = iris_3()) %>%
      as_dm() %>%
      dm_add_pk(iris_1, key) %>%
      dm_add_fk(iris_2, key, iris_1)
  }

# star schema data model for testing `dm_flatten_to_tbl()` ------

fact %<-%
  tibble(
    fact = c(
      "acorn",
      "blubber",
      "cinderella",
      "depth",
      "elysium",
      "fantasy",
      "gorgeous",
      "halo",
      "ill-advised",
      "jitter"
    ),
    dim_1_key_1 = 14:5,
    dim_1_key_2 = LETTERS[14:5],
    dim_2_key = letters[3:12],
    dim_3_key = LETTERS[24:15],
    dim_4_key = 7:16,
    something = 1:10
  )

fact_clean %<-%
  {
    fact() %>%
      rename(
        fact.something = something
      )
  }

fact_clean_new %<-%
  {
    fact() %>%
      rename(
        something.fact = something
      )
  }

dim_1 %<-%
  tibble(
    dim_1_pk_1 = 1:20,
    dim_1_pk_2 = LETTERS[1:20],
    something = letters[3:22]
  )
dim_1_clean %<-%
  {
    dim_1() %>%
      rename(dim_1.something = something)
  }
dim_1_clean_new %<-%
  {
    dim_1() %>%
      rename(something.dim_1 = something)
  }

dim_2 %<-%
  tibble(
    dim_2_pk = letters[1:20],
    something = LETTERS[5:24]
  )
dim_2_clean %<-%
  {
    dim_2() %>%
      rename(dim_2.something = something)
  }
dim_2_clean_new %<-%
  {
    dim_2() %>%
      rename(something.dim_2 = something)
  }

dim_3 %<-%
  tibble(
    dim_3_pk = LETTERS[5:24],
    something = 3:22
  )
dim_3_clean %<-%
  {
    dim_3() %>%
      rename(dim_3.something = something)
  }
dim_3_clean_new %<-%
  {
    dim_3() %>%
      rename(something.dim_3 = something)
  }

dim_4 %<-%
  tibble(
    dim_4_pk = 19:7,
    something = 19:31
  )
dim_4_clean %<-%
  {
    dim_4() %>%
      rename(dim_4.something = something)
  }
dim_4_clean_new %<-%
  {
    dim_4() %>%
      rename(something.dim_4 = something)
  }

# dm for testing dm_disentangle() -----------------------------------------

entangled_dm %<-%
  {
    dm(
      a = tf_5() %>% rename(a = k),
      b = tf_5() %>% rename(b = k),
      c = tf_5() %>% rename(c = k),
      d = tf_5() %>% rename(d = k),
      e = tf_5() %>% rename(e = k),
      f = tf_5() %>% rename(f = k),
      g = tf_5() %>% rename(g = k),
      h = tf_5() %>% rename(h = k)
    ) %>%
      dm_add_pk(b, b) %>%
      dm_add_pk(c, c) %>%
      dm_add_pk(d, d) %>%
      dm_add_pk(e, e) %>%
      dm_add_pk(f, f) %>%
      dm_add_pk(g, g) %>%
      dm_add_pk(h, h) %>%
      dm_add_fk(a, a, b) %>%
      dm_add_fk(a, a, c) %>%
      dm_add_fk(b, b, d) %>%
      dm_add_fk(c, c, d) %>%
      dm_add_fk(d, d, e) %>%
      dm_add_fk(d, d, f) %>%
      dm_add_fk(e, e, g) %>%
      dm_add_fk(f, f, g) %>%
      dm_add_fk(g, g, h)
  }

entangled_dm_2 %<-%
  {
    dm(
      a = tf_5() %>% rename(a = k),
      b = tf_5() %>% rename(b = k),
      c = tf_5() %>% rename(c = k),
      d = tf_5() %>% rename(d = k),
      e = tf_5() %>% rename(e = k),
      f = tf_5() %>% rename(f = k),
      g = tf_5() %>% rename(g = k)
    ) %>%
      dm_add_pk(b, b) %>%
      dm_add_pk(c, c) %>%
      dm_add_pk(d, d) %>%
      dm_add_pk(e, e) %>%
      dm_add_pk(f, f) %>%
      dm_add_pk(g, g) %>%
      dm_add_fk(a, a, d) %>%
      dm_add_fk(b, b, d) %>%
      dm_add_fk(c, c, d) %>%
      dm_add_fk(a, a, e) %>%
      dm_add_fk(d, d, e) %>%
      dm_add_fk(f, f, g)
  }

# dm_flatten() ------------------------------------------------------------

dm_for_flatten %<-%
  {
    as_dm(list(
      fact = fact(),
      dim_1 = dim_1(),
      dim_2 = dim_2(),
      dim_3 = dim_3(),
      dim_4 = dim_4()
    )) %>%
      dm_add_pk(dim_1, c(dim_1_pk_1, dim_1_pk_2)) %>%
      dm_add_pk(dim_2, dim_2_pk) %>%
      dm_add_pk(dim_3, dim_3_pk) %>%
      dm_add_pk(dim_4, dim_4_pk) %>%
      dm_add_fk(fact, c(dim_1_key_1, dim_1_key_2), dim_1) %>%
      dm_add_fk(fact, dim_2_key, dim_2) %>%
      dm_add_fk(fact, dim_3_key, dim_3) %>%
      dm_add_fk(fact, dim_4_key, dim_4)
  }

result_from_flatten %<-%
  {
    fact_clean() %>%
      left_join(
        dim_1_clean(),
        by = c("dim_1_key_1" = "dim_1_pk_1", "dim_1_key_2" = "dim_1_pk_2")
      ) %>%
      left_join(dim_2_clean(), by = c("dim_2_key" = "dim_2_pk")) %>%
      left_join(dim_3_clean(), by = c("dim_3_key" = "dim_3_pk")) %>%
      left_join(dim_4_clean(), by = c("dim_4_key" = "dim_4_pk"))
  }

result_from_flatten_new %<-%
  {
    fact_clean_new() %>%
      left_join(
        dim_1_clean_new(),
        by = c("dim_1_key_1" = "dim_1_pk_1", "dim_1_key_2" = "dim_1_pk_2")
      ) %>%
      left_join(dim_2_clean_new(), by = c("dim_2_key" = "dim_2_pk")) %>%
      left_join(dim_3_clean_new(), by = c("dim_3_key" = "dim_3_pk")) %>%
      left_join(dim_4_clean_new(), by = c("dim_4_key" = "dim_4_pk"))
  }

# 'bad' dm (no ref. integrity) for testing dm_flatten_to_tbl() --------

tbl_1 %<-% tibble(a = as.integer(c(1, 2, 4, 5, NA)), x = LETTERS[3:7], b = a)
tbl_2 %<-% tibble(id = c(1:3, 3), x = LETTERS[c(3:5, 5)], c = letters[1:4])
tbl_3 %<-% tibble(id = c(2:4, 4), d = letters[2:5])

bad_dm_base %<-%
  {
    as_dm(list(tbl_1 = tbl_1(), tbl_2 = tbl_2(), tbl_3 = tbl_3()))
  }

# avoid copying constraints for invalid dm
bad_dm %<--%
  {
    bad_dm_base() %>%
      dm_add_pk(tbl_2, c(id, x)) %>%
      dm_add_pk(tbl_3, id) %>%
      dm_add_fk(tbl_1, c(a, x), tbl_2) %>%
      dm_add_fk(tbl_1, b, tbl_3)
  }

dm_nycflights_small_base %<-%
  {
    tables <- dm_get_tables(dm_nycflights13())
    # https://github.com/tidyverse/dbplyr/pull/1195
    tables$flights <- mutate(tables$flights, time_hour = as.character(time_hour))
    tables$weather <- mutate(tables$weather, time_hour = as.character(time_hour))
    dm(!!!tables)
  }

# Do not add PK and FK constraints to the database
dm_nycflights_small %<--%
  {
    dm_nycflights_small_base() %>%
      dm_add_pk(planes, tailnum) %>%
      dm_add_pk(airlines, carrier) %>%
      dm_add_pk(airports, faa) %>%
      dm_add_fk(flights, tailnum, planes) %>%
      dm_add_fk(flights, carrier, airlines) %>%
      dm_add_fk(flights, dest, airports)
  }

dm_nycflights_small_cycle %<--%
  {
    dm_nycflights_small() %>%
      dm_add_fk(flights, origin, airports)
  }

nyc_comp %<--%
  {
    dm_nycflights_small() %>%
      dm_add_pk(weather, c(origin, time_hour)) %>%
      dm_add_fk(flights, c(origin, time_hour), weather)
  }

dm_zoomed <- function() dm_zoom_to(dm_for_filter(), tf_2)
dm_zoomed_2 <- function() dm_zoom_to(dm_for_filter(), tf_3)

dm_for_autoinc_1 %<-%
  {
    dm(
      t1 = tibble(a = 5:7, o = letters[1:3]),
      t2 = tibble(c = 10:8, d = 7:5, o = letters[3:1]),
      t3 = tibble(e = c(6L, 5L, 7L), o = letters[c(2, 1, 3)]),
      t4 = tibble(g = 1:3, h = 8:10, o = letters[1:3])
    )
  }

# FIXME: regarding PR #313: everything below this line needs to be at least reconsidered if not just dumped.

# for database tests -------------------------------------------------

# postgres needs to be cleaned of t?_2019_* tables for learn-test
get_test_tables_from_postgres <- function() {
  src_postgres <- my_test_src()
  con_postgres <- src_postgres$con

  con_postgres %>%
    DBI::dbGetQuery(
      "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public'"
    ) %>%
    as_tibble() %>%
    filter(grepl(
      "^tf_[0-9]{1}_[0-9]{4}_[0-9]{2}_[0-9]{2}_[0-9]{2}_[0-9]{2}_[0-9]{2}_[0-9]+",
      table_name
    ))
}

is_postgres_empty <- function() {
  nrow(get_test_tables_from_postgres()) == 0
}

clear_postgres <- function() {
  src_postgres <- my_test_src()
  con_postgres <- src_postgres$con

  walk(
    get_test_tables_from_postgres() %>%
      pull(),
    ~ DBI::dbExecute(con_postgres, glue("DROP TABLE {.x} CASCADE"))
  )
}

test_that("dummy", {
  # To avoid deletion of file
  expect_snapshot({
    TRUE
  })
})

test_that("dm_meta() data model", {
  skip_if_schema_not_supported()

  expect_snapshot({
    dm_meta(my_test_src()) %>%
      dm_paste(options = c("select", "keys", "color"))
  })
})

test_that("dm_meta(simple = TRUE) columns", {
  # Still stored as snapshot in columns.csv, never cleared
  skip("Dependent on database version, find better way to record this info")

  columns <- tryCatch(
    my_db_test_src() %>%
      dm_meta(simple = TRUE) %>%
      .$columns %>%
      filter(tolower(table_schema) == "information_schema") %>%
      arrange(table_name, ordinal_position) %>%
      select(-table_catalog) %>%
      collect(),
    error = function(e) {
      data.frame(error = conditionMessage(e))
    }
  )

  path <- withr::local_tempfile(fileext = ".csv")
  write.csv(columns, path, na = "")

  expect_snapshot_file(path, name = "columns.csv", variant = my_test_src_name)
})

test_that("`json_pack()` works", {
  expect_snapshot({
    df <- tibble::tibble(x1 = 1:3, x2 = 4:6, x3 = 7:9, y = 1:3)
    packed <- json_pack(df, x = c(x1, x2, x3), y = y)
    packed
  })

  df_roundtrip <- json_unpack(packed, c(x, y))
  expect_equal(df, df_roundtrip)
})

test_that("`json_pack()` works remotely", {
  skip_if_src_not("postgres", "mssql")
  con <- my_test_src()$con

  local <- tibble(grp = c(1, 1, 2, 2), a_i = letters[1:4], a_j = LETTERS[1:4])
  remote <- test_db_src_frame(!!!local)

  expect_snapshot(variant = my_test_src_name, {
    query <- json_pack(remote, a = starts_with("a")) %>% dbplyr::sql_render()
    # For stable POSTGRES tests
    gsub("test_frame_[_0-9]+", "test_frame_...", query)
    json_pack(remote, a = starts_with("a"))
    query <- json_pack(remote, a = starts_with("a"), .names_sep = "_") %>% dbplyr::sql_render()
    # For stable POSTGRES tests
    gsub("test_frame_[_0-9]+", "test_frame_...", query)
    json_pack(remote, a = starts_with("a"), .names_sep = "_")
  })

  expect_identical(
    local %>% json_pack(a = starts_with("a")) %>% unjson_nested(),
    remote %>% json_pack(a = starts_with("a")) %>% collect() %>% unjson_nested()
  )
})

test_that("`json_unpack()` fails without cols argument", {
  df <- tibble::tibble(x = '{"a":1}')
  expect_snapshot(error = TRUE, {
    json_unpack(df)
  })
})

test_that("`json_nest()` and `json_unnest()` work", {
  expect_snapshot({
    df <- tibble::tibble(x = c(1, 1, 1, 2, 2, 3), y = 1:6, z = 6:1)
    nested <- json_nest(df, data = c(y, z))
    nested
  })

  df_roundtrip <- json_unnest(nested, data)
  df_roundtrip
  expect_equal(df, df_roundtrip)
})

test_that("`json_nest()` fails with unnamed elements", {
  df <- tibble::tibble(x = 1, y = 2, z = 3)
  expect_snapshot(error = TRUE, {
    json_nest(df, c(y, z))
  })
})

test_that("`json_nest()` works remotely", {
  skip_if_src_not("postgres", "mssql")
  con <- my_test_src()$con

  local <- tibble(grp = c(1, 1, 2, 2), a_i = letters[1:4], a_j = LETTERS[1:4])
  remote <- test_db_src_frame(!!!local)

  expect_snapshot(variant = my_test_src_name, {
    query <- remote %>%
      json_nest(a = starts_with("a")) %>%
      arrange(grp) %>%
      dbplyr::sql_render()
    # For stable POSTGRES tests
    gsub("test_frame_[_0-9]+", "test_frame_...", query)
    remote %>%
      json_nest(a = starts_with("a")) %>%
      arrange(grp) %>%
      collect()
    query <- remote %>%
      json_nest(a = starts_with("a"), .names_sep = "_") %>%
      arrange(grp) %>%
      dbplyr::sql_render()
    # For stable POSTGRES tests
    gsub("test_frame_[_0-9]+", "test_frame_...", query)
    remote %>%
      json_nest(a = starts_with("a"), .names_sep = "_") %>%
      arrange(grp) %>%
      collect()
  })

  expect_equivalent_tbl(
    local %>% json_nest(A = starts_with("a")) %>% unjson_nested(),
    remote %>% json_nest(A = starts_with("a")) %>% collect() %>% unjson_nested()
  )
})

test_that("`json_unnest()` fails without cols argument", {
  df <- tibble::tibble(x = '{"a":1}')
  expect_snapshot(error = TRUE, {
    json_unnest(df)
  })
})

mw_cg_make_dm_rename <- function(
  dm,
  ...,
  table_names = NULL,
  column_names = NULL,
  new_column_names = NULL
) {
  check_dots_empty()

  # Checks
  stopifnot(!is.null(table_names))
  stopifnot(length(table_names) == 1)
  stopifnot(table_names %in% names(dm))
  stopifnot(inherits(column_names, "character"))
  stopifnot(length(column_names) == 1)
  stopifnot(inherits(new_column_names, "character"))
  stopifnot(length(new_column_names) == 1)

  rename_col <- set_names(column_names, new_column_names)

  # Return call object
  list(
    call = expr(dm_rename(
      .,
      !!sym(table_names),
      !!!syms(rename_col)
    ))
  )
}

library(tidyverse)
library(rlang)
devtools::load_all()

parent <- tibble(a = c(1L, 1:3), b = -1)
child <- tibble(a = 1:4, c = 3)

dm <- dm(parent, child)

# This is done by the Shiny app
ops <- enum_ops(quo(dm))
ops

ops$multiple

# Choice options (the user can choose arbitrarily many):
user_choice <- list(
  table_names = "parent"
)

# Called by the Shiny app, always with the same pattern:
ops_2 <- exec(
  # Always enum_ops
  enum_ops,
  # Returned by the previous call to enum_ops()
  !!!ops$input,
  # Input selected by the user
  !!!user_choice
)
# Same as:
# ops_2 <- enum_ops(dm = ops$input$dm, table_names = table_names)

ops_2

# Choice options (the user can choose only one of those):
ops_2$single

# User's choice:
user_choice_2 <- list(op_name = "dm_add_pk")

# Called by the Shiny app, always with the same pattern:
ops_3 <- exec(
  # Always enum_ops
  enum_ops,
  # Returned by the previous call to enum_ops()
  !!!ops_2$input,
  # Input selected by the user
  !!!user_choice_2
)

ops_3

# Choice options (the user can choose arbitrarily many):
ops_3$multiple

# User's choice:
user_choice_3 <- list(column_names = c("a"))

# Called by the Shiny app, always with the same pattern:
ops_4 <- exec(
  # Always enum_ops
  enum_ops,
  # Returned by the previous call to enum_ops()
  !!!ops_3$input,
  # Input selected by the user
  !!!user_choice_3
)

ops_4

# Final result:
ops_4$call

cg_block <- new_cg_block(ops_4$input$dm)

cg_block %>%
  cg_add_call(!!ops_4$call)

cg_block %>%
  cg_add_call(!!ops_4$call) %>%
  cg_eval_block()

# pkgload::load_all()
library(dm)
library(dplyr)

dm <- dm_nycflights13()

dm %>%
  dm_draw()

tbls <-
  dm %>%
  dm_get_tables(keyed = TRUE)

flights <- tbls$flights
airports <- tbls$airports

flights
airports

flights %>%
  left_join(airports)

by_origin <-
  flights %>%
  group_by(origin) %>%
  summarize(n = n()) %>%
  ungroup()

by_origin

result_dm <- dm(flights, airports, by_origin)
result_dm %>%
  dm_draw()

#' Get tables
#'
#' `dm_get_tables()` returns a named list of \pkg{dplyr} [tbl] objects
#' of a `dm` object.
#'
#' @param x A `dm` object.
#' @inheritParams rlang::args_dots_empty
#' @param keyed `r lifecycle::badge("experimental")`
#'   Set to `TRUE` to return objects of the internal class `"dm_keyed_tbl"`
#'   that will contain information on primary and foreign keys
#'   in the individual table objects.
#'   This allows using dplyr workflows on those tables and later reconstruct them
#'   into a `dm` object.
#'   See [dm_deconstruct()] for a function that generates corresponding code
#'   for an existing dm object, and `vignette("tech-dm-keyed")` for details.
#'
#' @return A named list with the tables (data frames or lazy tables)
#'   constituting the `dm`.
#'
#' @seealso [dm()] and [new_dm()] for constructing a `dm` object from tables.
#'
#' @export
#' @examples
#' dm_nycflights13() %>%
#'   dm_get_tables()
#'
#' dm_nycflights13() %>%
#'   dm_get_tables(keyed = TRUE)
#'
#' dm_nycflights13() %>%
#'   dm_get_tables(keyed = TRUE) %>%
#'   new_dm()
dm_get_tables <- function(x, ..., keyed = FALSE) {
  check_dots_empty()
  dm_local_error_call()

  check_not_zoomed(x)
  if (isTRUE(keyed)) {
    dm_get_keyed_tables_impl(x)
  } else {
    dm_get_tables_impl(x)
  }
}

dm_get_tables_impl <- function(x, quiet = FALSE) {
  def <- dm_get_def(x, quiet)
  set_names(def$data, def$table)
}

data.frame(
  os = "ubuntu-22.04",
  r = "release",
  test_src = c(
    "test-mssql",
    "test-postgres",
    "test-maria",
    "test-mysql-maria",
    "test-duckdb",
    "test-sqlite"
  ),
  covr = "true",
  desc = c(
    "SQL Server with covr",
    "Postgres with covr",
    "MariaDB with covr",
    "MySQL with covr",
    "DuckDB with covr",
    "SQLite with covr"
  )
)

#' Filtering
#'
#' @description
#' Filtering a table of a [`dm`] object may affect other tables that are connected to it
#' directly or indirectly via foreign key relations.
#'
#' `dm_filter()` can be used to define filter conditions for tables using syntax that is similar to [dplyr::filter()].
#' The filters work across related tables:
#' The resulting `dm` object only contains rows that are related
#' (directly or indirectly) to rows that remain after applying the filters
#' on all tables.
#'
#' @details
#' As of dm 1.0.0, these conditions are no longer stored in the `dm` object,
#' instead they are applied to all tables during the call to `dm_filter()`.
#' Calling `dm_apply_filters()` or `dm_apply_filters_to_tbl()` is no longer necessary.
#'
#' Use [dm_zoom_to()] and [dplyr::filter()] to filter rows without affecting related tables.
#'
#' @inheritParams dm_examine_constraints
#' @param ...
#'   Named logical predicates.
#'   The names correspond to tables in the `dm` object.
#'   The predicates are defined in terms of the variables in the corresponding table,
#'   they are passed on to [dplyr::filter()].
#'
#'   Multiple conditions are combined with `&`.
#'   Only the rows where the condition evaluates
#'   to `TRUE` are kept.
#'
#' @return An updated `dm` object with filters executed across all tables.
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nyc <- dm_nycflights13()
#' dm_nyc %>%
#'   dm_nrow()
#'
#' dm_nyc_filtered <-
#'   dm_nycflights13() %>%
#'   dm_filter(airports = (name == "John F Kennedy Intl"))
#'
#' dm_nyc_filtered %>%
#'   dm_nrow()
#'
#' # If you want to keep only those rows in the parent tables
#' # whose primary key values appear as foreign key values in
#' # `flights`, you can set a `TRUE` filter in `flights`:
#' dm_nyc %>%
#'   dm_filter(flights = (1 == 1)) %>%
#'   dm_nrow()
#' # note that in this example, the only affected table is
#' # `airports` because the departure airports in `flights` are
#' # only the three New York airports.
#' @export
dm_filter <- function(.dm, ...) {
  dm_local_error_call()
  dm_filter_api0({{ .dm }}, ..., target = dm_filter_impl0, apply_target = dm_apply_filters_impl)
}

dm_filter_api0 <- function(
  ...,
  dm = NULL,
  call = caller_env(),
  user_env = caller_env(2),
  target = make_dm_filter_api_call,
  apply_target = make_dm_apply_filters_call
) {
  if (!is.null(dm)) {
    deprecate_warn("1.0.0", "dm_filter(dm = )", "dm_filter(.dm = )", user_env = user_env)
    dm_filter_api1(
      dm,
      ...,
      call = call,
      user_env = user_env,
      target = target,
      apply_target = apply_target
    )
  } else {
    dm_filter_api1(
      ...,
      call = call,
      user_env = user_env,
      target = target,
      apply_target = apply_target
    )
  }
}

dm_filter_api1 <- function(.dm, ..., table = NULL, call, user_env, target, apply_target) {
  quos <- enquos(...)
  table <- enquo(table)

  if (is_named2(quos)) {
    # New-style API: apply immediately
    out <- reduce2(names(quos), quos, dm_filter_api, .init = .dm, target = target)
    apply_target(out)
  } else {
    deprecate_warn(
      "1.0.0",
      "dm_filter(table = )",
      user_env = user_env,
      details = "`dm_filter()` now takes named filter expressions, the names correspond to the tables to be filtered. You no longer need to call `dm_apply_filters()` to materialize the filters."
    )

    if (quo_is_null(table)) {
      table_idx <- match("", names2(quos))
      if (is.na(table_idx)) {
        abort_table_missing("table")
      }
      table <- quos[[table_idx]]
      quos <- quos[-table_idx]
    }

    stopifnot(quo_is_symbol(table))

    reduce(quos, dm_filter_api, table = as_name(table), .init = .dm, target = target)
  }
}

dm_filter_api <- function(dm, table, expr, target) {
  target(dm, {{ table }}, {{ expr }})
}

make_dm_filter_api_call <- function(dm, table, expr) {
  call2("%>%", dm, call2("dm_filter_api", table, expr))
}

make_dm_apply_filters_call <- function(dm) {
  call2("%>%", dm, call2("dm_apply_filters"))
}

dm_filter_impl0 <- function(dm, table, expr) {
  check_not_zoomed(dm)
  dm %>%
    dm_zoom_to(!!table) %>%
    dm_filter_impl({{ expr }}, set_filter = TRUE) %>%
    dm_update_zoomed()
}

dm_filter_impl <- function(dm_zoomed, ..., set_filter) {
  # valid table and empty ellipsis provided
  filter_quos <- enquos(...)
  if (is_empty(filter_quos)) {
    return(dm_zoomed)
  }

  tbl <- tbl_zoomed(dm_zoomed)
  filtered_tbl <- filter(tbl, ...)

  # attribute filter expression to zoomed table. Needs to be flagged with `zoomed = TRUE`, since
  # in case of `dm_insert_zoomed()` the filter exprs needs to be transferred
  if (set_filter) {
    dm_zoomed <-
      dm_zoomed %>%
      set_filter_for_table(orig_name_zoomed(dm_zoomed), map(filter_quos, quo_get_expr), TRUE)
  }

  replace_zoomed_tbl(dm_zoomed, filtered_tbl)
}

set_filter_for_table <- function(dm, table, filter_exprs, zoomed) {
  def <- dm_get_def(dm)

  i <- which(def$table == table)
  def$filters[[i]] <- vec_rbind(def$filters[[i]], new_filter(filter_exprs, zoomed))
  dm_from_def(def, zoomed = zoomed)
}


#' @rdname deprecated
#' @export
dm_apply_filters <- function(dm) {
  check_not_zoomed(dm)

  filters <- dm_get_filters_impl(dm)
  if (nrow(filters) == 0) {
    deprecate_warn(
      "1.0.0",
      "dm_apply_filters()",
      details = "Calling `dm_apply_filters()` after `dm_filter()` is no longer necessary."
    )
  }

  dm_apply_filters_impl(dm)
}

dm_apply_filters_impl <- function(dm) {
  def <- dm_get_def(dm)

  def$data <- map(def$table, ~ dm_get_filtered_table(dm, .))

  dm_reset_all_filters(dm_from_def(def))
}

#' @rdname deprecated
#' @export
dm_apply_filters_to_tbl <- function(dm, table) {
  check_not_zoomed(dm)

  filters <- dm_get_filters_impl(dm)
  if (nrow(filters) == 0) {
    deprecate_warn(
      "1.0.0",
      "dm_apply_filters_to_tbl()",
      details = "Access tables directly after `dm_filter()`."
    )
  }

  dm_apply_filters_to_tbl_impl(dm, {{ table }})
}

dm_apply_filters_to_tbl_impl <- function(dm, table) {
  table_name <- dm_tbl_name(dm, {{ table }})
  dm_get_filtered_table(dm, table_name)
}

# calculates the necessary semi-joins from all tables that were filtered to
# the requested table
#' @autoglobal
dm_get_filtered_table <- function(dm, from) {
  filters <- dm_get_filters_impl(dm)
  # Shortcut for speed, not really necessary
  if (nrow(filters) == 0) {
    return(dm_get_tables(dm)[[from]])
  }

  fc <- get_all_filtered_connected(dm, from)

  fc_children <-
    fc %>%
    filter(node != parent) %>%
    select(-distance) %>%
    nest(semi_join = -parent) %>%
    rename(table = parent)

  recipe <-
    fc %>%
    select(table = node) %>%
    left_join(fc_children, by = "table")

  list_of_tables <- dm_get_tables(dm)

  for (i in seq_len(nrow(recipe))) {
    table_name <- recipe$table[i]
    table <- list_of_tables[[table_name]]

    semi_joins <- recipe$semi_join[[i]]
    if (!is_null(semi_joins)) {
      semi_joins <- pull(semi_joins)
      semi_joins_tbls <- list_of_tables[semi_joins]
      table <-
        reduce2(
          semi_joins_tbls,
          semi_joins,
          ~ semi_join(..1, ..2, by = get_by(dm, table_name, ..3)),
          .init = table
        )
    }
    list_of_tables[[table_name]] <- table
  }
  list_of_tables[[from]]
}


#' Get filter expressions
#'
#' `dm_get_filters()` returns the filter expressions that have been applied to a `dm` object.
#' These filter expressions are not intended for evaluation, only for
#' information.
#'
#' @section Life cycle:
#' This function is marked "questioning" because it seems wrong
#' to tightly couple filtering with the data model.
#' On the one hand, an overview of active filters is useful
#' when specifying the base data set for an analysis in terms of column selections
#' and row filters.
#' However, these filter condition should be only of informative nature
#' and never affect the results of other operations.
#' We are working on formalizing the semantics of the underlying operations
#' in order to present them in a cleaner interface.
#'
#' Use [dm_zoom_to()] and [dplyr::filter()] to filter rows without registering
#' the filter.
#'
#' @seealso [dm_filter()], [dm_apply_filters()]
#'
#' @return A tibble with the following columns:
#'   \describe{
#'     \item{`table`}{table that was filtered,}
#'     \item{`filter`}{the filter expression,}
#'     \item{`zoomed`}{logical, does the filter condition relate to the zoomed table.}
#'   }
#'
#' @examplesIf rlang::is_installed(c("nycflights13", "dbplyr"))
#' dm_nycflights13() %>%
#'   dm_get_filters()
#' @noRd
NULL

#' @rdname deprecated
#' @export
dm_get_filters <- function(dm) {
  check_not_zoomed(dm)

  filters <- dm_get_filters_impl(dm)
  if (nrow(filters) == 0) {
    deprecate_warn(
      "1.0.0",
      "dm_get_filters()",
      details = "Filter conditions are no longer stored with the dm object."
    )
  }
  filters
}

#' @autoglobal
dm_get_filters_impl <- function(dm) {
  filter_df <-
    dm_get_def(dm) %>%
    select(table, filters) %>%
    unnest_list_of_df("filters")

  # FIXME: Should work better with dplyr 0.9.0
  # if (!("filter_expr" %in% names(filter_df))) {
  #   filter_df$filter_expr <- list()
  # }

  filter_df %>%
    rename(filter = filter_expr) %>%
    mutate(filter = unname(filter))
}

get_all_filtered_connected <- function(dm, table) {
  filtered_tables <- unique(dm_get_filters_impl(dm)$table)
  graph <- create_graph_from_dm(dm)

  # Computation of distances and shortest paths uses the same algorithm
  # internally, but s.p. doesn't return distances and distances don't return
  # the predecessor.
  distances <- graph_distances(graph, table)[1, ]
  finite_distances <- distances[is.finite(distances)]

  # Using only nodes with finite distances (=in the same connected component)
  # as target. This avoids a warning.
  target_tables <- names(finite_distances)

  if (is_empty(intersect(target_tables, filtered_tables))) {
    return(new_filtered_edges(table))
  }

  # use only subgraph to
  # 1. speed things up
  # 2. make it possible to easily test for a cycle (cycle if: N(E) >= N(V))
  graph <- graph_induced_subgraph(graph, target_tables)
  if (length(graph_edges(graph)) >= length(graph_vertices(graph))) {
    abort_no_cycles(graph)
  }
  paths <- graph_shortest_paths(graph, table, target_tables, predecessors = TRUE)

  # All edges with finite distance as tidy data frame
  all_edges <-
    new_filtered_edges(
      node = names(graph_vertices(graph)),
      parent = names(paths$predecessors),
      # all of `graph`, `paths` and `finite_distances` are based on the same subset of tables,
      # hence the resulting tibble is correct
      distance = finite_distances
    )

  # igraph >= 2.0.0:
  all_edges$parent[is.na(all_edges$parent)] <- all_edges$node[is.na(all_edges$parent)]
  stopifnot(!anyNA(all_edges$parent))

  # Edges of interest, will be grown until source node `table` is reachable
  # from all nodes
  edges <-
    all_edges %>%
    filter(node %in% !!c(filtered_tables, table))

  # Recursive join
  repeat {
    missing <- setdiff(edges$parent, edges$node)
    if (is_empty(missing)) {
      break
    }

    edges <- bind_rows(edges, filter(all_edges, node %in% !!missing))
  }

  # Keeping the sentinel row (node == parent) to simplify further processing
  # and testing
  edges %>%
    arrange(-distance)
}

new_filtered_edges <- function(node, parent = node, distance = 0) {
  tibble(node, parent, distance)
}

check_no_filter <- function(dm) {
  def <-
    dm_get_def(dm)

  if (detect_index(def$filters, ~ vec_size(.) > 0) == 0) {
    return()
  }

  fun_name <- as_string(sys.call(-1)[[1]])
  abort_only_possible_wo_filters(fun_name)
}

test_that("dm_filter() legacy API", {
  local_options(lifecycle_verbosity = "quiet")

  # Autogenerated from
  expect_same(
    dm_filter_api0(expr(dm), a, b > 1),
    dm_filter_api0(table = a, expr(dm), b > 1),
    dm_filter_api0(expr(dm), table = a, b > 1),
    dm_filter_api0(expr(dm), b > 1, table = a)
  )
  expect_same(
    dm_filter_api0(expr(dm), a, b > 1),
    dm_filter_api0(dm = expr(dm), a, b > 1),
    dm_filter_api0(a, dm = expr(dm), b > 1),
    dm_filter_api0(a, b > 1, dm = expr(dm))
  )
  expect_same(
    dm_filter_api0(expr(dm), a, b > 1),
    dm_filter_api0(dm = expr(dm), table = a, b > 1),
    dm_filter_api0(dm = expr(dm), b > 1, table = a),
    dm_filter_api0(table = a, dm = expr(dm), b > 1),
    dm_filter_api0(table = a, b > 1, dm = expr(dm)),
    dm_filter_api0(b > 1, dm = expr(dm), table = a),
    dm_filter_api0(b > 1, table = a, dm = expr(dm))
  )

  # Equivalence with new-style API
  expect_same(
    dm_filter_api0(expr(dm), a = b > 1, apply_target = identity),
    dm_filter_api0(expr(dm), a, b > 1),
    dm_filter_api0(.dm = expr(dm), a = b > 1, apply_target = identity),
    dm_filter_api0(a = b > 1, .dm = expr(dm), apply_target = identity),
    dm_filter_api0(.dm = expr(dm), a, b > 1),
    dm_filter_api0(a, .dm = expr(dm), b > 1),
    dm_filter_api0(a, b > 1, .dm = expr(dm)),
  )
  expect_same(
    dm_filter_api0(expr(dm), a = b > 1, c = d < 2, apply_target = identity),
    # Can't use magrittr pipe here
    dm_filter_api0(dm_filter_api0(expr(dm), a, b > 1), c, d < 2),
  )
})

test_that("dm_filter() deprecations", {
  local_options(lifecycle_verbosity = "warning")

  expect_snapshot({
    dm_filter(dm_for_filter(), tf_1, a > 4)
    dm_filter(dm = dm_for_filter(), tf_1, a > 4)
    dm_filter(dm_for_filter(), tf_1, a > 4) %>% dm_apply_filters()
    dm_filter(dm_for_filter(), tf_1 = a > 4) %>% dm_apply_filters()
    dm_filter(dm_for_filter(), tf_1, a > 4) %>% dm_apply_filters_to_tbl(tf_2)
    dm_filter(dm_for_filter(), tf_1 = a > 4) %>% dm_apply_filters_to_tbl(tf_2)
    dm_filter(dm_for_filter(), tf_1, a > 4) %>% dm_get_filters()
    dm_filter(dm_for_filter(), tf_1 = a > 4) %>% dm_get_filters()
  })
})

test_that("data structure", {
  expect_snapshot({
    dm_more_complex() %>%
      dm_paste(options = c("select", "keys"))
  })
})

test_that("get_all_filtered_connected() calculates the paths correctly", {
  # Legacy API
  local_options(lifecycle_verbosity = "quiet")

  # Only need to run for local sources
  skip_if_remote_src()

  fc <-
    dm_more_complex() %>%
    dm_filter(tf_2, TRUE) %>%
    dm_filter(tf_6, TRUE) %>%
    get_all_filtered_connected("tf_5")
  expect_pred_chain(fc, c("tf_2", "tf_3", "tf_4", "tf_5"))
  expect_pred_chain(fc, c("tf_6", "tf_5"))
  expect_not_pred(fc, c("tf_1", "tf_4_2"))

  # more complicated graph structure:
  fc <-
    dm_more_complex() %>%
    dm_filter(tf_6, TRUE) %>%
    dm_filter(tf_6_2, TRUE) %>%
    get_all_filtered_connected("tf_4")
  expect_pred_chain(fc, c("tf_6", "tf_5", "tf_4"))
  expect_pred_chain(fc, c("tf_6_2", "tf_3", "tf_4"))

  # filter in an unconnected component:
  fc <-
    dm_more_complex() %>%
    dm_filter(tf_6, TRUE) %>%
    get_all_filtered_connected("a")
  expect_identical(fc$node, "a")

  fc <-
    dm_more_complex() %>%
    dm_filter(tf_5, TRUE) %>%
    get_all_filtered_connected("tf_3")
  expect_pred_chain(fc, c("tf_5", "tf_4", "tf_3"))

  f <-
    dm_more_complex() %>%
    dm_filter(tf_4_2, TRUE) %>%
    dm_filter(tf_6, TRUE)

  fc_tf_4 <- get_all_filtered_connected(f, "tf_4")

  expect_pred_chain(fc_tf_4, c("tf_4_2", "tf_5", "tf_4"))
  expect_pred_chain(fc_tf_4, c("tf_6", "tf_5", "tf_4"))
  expect_not_pred(fc_tf_4, c("tf_6_2", "tf_3", "tf_2", "tf_1"))

  f <-
    dm_more_complex() %>%
    dm_filter(tf_4_2, TRUE) %>%
    dm_filter(tf_6, TRUE, FALSE) %>%
    dm_filter(tf_5, TRUE)

  fc_tf_4 <- get_all_filtered_connected(f, "tf_4")

  expect_pred_chain(fc_tf_4, c("tf_4_2", "tf_5", "tf_4"))
  expect_pred_chain(fc_tf_4, c("tf_6", "tf_5", "tf_4"))
  expect_not_pred(fc_tf_4, c("tf_6_2", "tf_3", "tf_2", "tf_1"))

  # fails when cycle is present
  expect_dm_error(
    dm_for_filter_w_cycle() %>% dm_filter(tf_1, a > 3) %>% dm_get_filtered_table("tf_3"),
    "no_cycles"
  )

  # Cycles in other components don't affect filtering
  expect_equivalent_dm(
    dm_for_filter_w_cycle() %>%
      dm(tf_8 = tibble(r = 1)) %>%
      dm_filter(tf_8, TRUE) %>%
      dm_apply_filters(),
    dm_for_filter_w_cycle() %>%
      dm(tf_8 = tibble(r = 1))
  )

  # FIXME: fails, when it could actually work (check diagram of `dm_for_filter_w_cycle()`)
  # expect_identical(
  #   dm_for_filter_w_cycle() %>% dm_filter(tf_1, a > 3) %>% dm_get_filtered_table("tf_2"),
  #   semi_join(tf_2, filter(tf_1, a > 3))
  # )
})

test_that("legacy: we get unfiltered tables without dm_apply_filters()", {
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_tbl(
    dm_filter(dm_for_filter(), tf_1, a > 4) %>% tbl_impl("tf_2"),
    tf_2()
  )

  expect_equivalent_tbl(
    dm_filter(dm_for_filter(), tf_1 = a > 4) %>% dm_apply_filters_to_tbl(tf_2),
    tf_2() %>% semi_join(filter(tf_1(), a > 4), by = c("d" = "a"))
  )
})

test_that("we get filtered/unfiltered tables with respective funs", {
  expect_equivalent_tbl(
    dm_filter(dm_for_filter(), tf_1 = a > 4) %>% tbl_impl("tf_1"),
    filter(tf_1(), a > 4)
  )

  expect_snapshot({
    dm_for_filter() %>%
      dm_filter(tf_1 = a > 3 & a < 8) %>%
      dm_get_tables() %>%
      map(harmonize_tbl)
  })
})

test_that("dm_filter() works as intended for reversed dm", {
  expect_snapshot({
    dm_for_filter_rev() %>%
      dm_filter(tf_1 = a < 8 & a > 3) %>%
      dm_get_tables() %>%
      map(harmonize_tbl)
  })
})

test_that("dm_filter() works as intended for inbetween table", {
  expect_snapshot({
    dm_for_filter() %>%
      dm_filter(tf_3 = g == "five") %>%
      dm_get_tables() %>%
      map(harmonize_tbl)
  })
})

test_that("dm_filter() works without primary keys", {
  expect_silent(
    dm_for_filter() %>%
      dm_rm_pk(tf_5) %>%
      dm_filter(tf_5 = (l == "c")) %>%
      collect()
  )
})

test_that("dm_filter() returns original `dm` object when ellipsis empty", {
  local_options(lifecycle_verbosity = "quiet")

  expect_equivalent_dm(
    dm_filter(dm_for_filter(), tf_3),
    dm_for_filter()
  )
})

test_that("dm_filter() is a no-op when no table name is provided", {
  expect_equivalent_dm(
    dm_filter(dm_for_filter()),
    dm_for_filter()
  )
})

test_that("dm_get_filters() works", {
  local_options(lifecycle_verbosity = "quiet")

  expect_identical(
    dm_get_filters(dm_for_filter()),
    tibble(table = character(), filter = list(), zoomed = logical())
  )

  expect_identical(
    dm_get_filters(dm_filter(dm_for_filter(), tf_1, a > 3, a < 8)),
    tibble(table = "tf_1", filter = unname(exprs(a > 3, a < 8)), zoomed = FALSE)
  )
})

# tests for compound keys -------------------------------------------------

test_that("dm_filter() output for compound keys", {
  expect_snapshot({
    nyc_comp() %>%
      dm_filter(flights = sched_dep_time <= 1200) %>%
      dm_nrow()
    nyc_comp() %>%
      dm_filter(weather = pressure < 1020) %>%
      dm_nrow()
  })
})

# Generated by roxyglobals: do not edit by hand

utils::globalVariables(c(
  "parent_table", # <build_copy_queries>
  "parent_key_cols", # <build_copy_queries>
  "autoincrement", # <build_copy_queries>
  "name", # <build_copy_queries>
  "col_def", # <build_copy_queries>
  "pk_col", # <build_copy_queries>
  "unique_def", # <build_copy_queries>
  "child_table", # <build_copy_queries>
  "child_fk_cols", # <build_copy_queries>
  "on_delete", # <build_copy_queries>
  "fk_def", # <build_copy_queries>
  "remote_name", # <build_copy_queries>
  "remote_name_unquoted", # <build_copy_queries>
  "index_name", # <build_copy_queries>
  "columns", # <build_copy_queries>
  "schema_name", # <get_src_tbl_names>
  "table_name", # <get_src_tbl_names>
  "remote_name", # <get_src_tbl_names>
  "local_name", # <get_src_tbl_names>
  "child_uuid", # <fks_df_from_keys_info>
  "parent_uuid", # <fks_df_from_keys_info>
  "child_fk_cols", # <fks_df_from_keys_info>
  "child_data", # <fks_df_from_keys_info>
  "parent_key_cols", # <fks_df_from_keys_info>
  "parent_data", # <fks_df_from_keys_info>
  "parent_table", # <fks_df_from_keys_info>
  "child_table", # <fks_df_from_keys_info>
  "new_name", # <compute_disambiguate_cols_recipe>
  "column", # <compute_disambiguate_cols_recipe>
  "text", # <explain_col_rename>
  "new_table", # <dm_disentangle>
  "fks", # <dm_disentangle>
  "uks", # <dm_disentangle>
  "child_table", # <dm_flatten_impl>
  "parent_table", # <dm_flatten_impl>
  "child_table", # <dm_flatten_join>
  "parent_table", # <dm_flatten_join>
  "child_table", # <dm_flatten_transfer_fks>
  "parent_table", # <dm_flatten_transfer_fks>
  "pks", # <str.dm>
  "fks", # <str.dm>
  "filters", # <str.dm>
  "zoom", # <str.dm_zoomed>
  "pks", # <str.dm_zoomed>
  "fks", # <str.dm_zoomed>
  "filters", # <str.dm_zoomed>
  "ref_uuid", # <tbl_def_impl>
  "ref_column", # <tbl_def_impl>
  "column", # <tbl_def_impl>
  "name", # <dm_ddl_pre>
  "col_def", # <dm_ddl_pre>
  "uk_def", # <dm_ddl_pre>
  "fk_def", # <dm_ddl_pre>
  "pk_def", # <dm_ddl_pre>
  "remote_name", # <dm_ddl_pre>
  "renames", # <prepare_join>
  "kind", # <dm_get_data_model>
  "child_table", # <dm_get_data_model>
  "child_fk_cols", # <dm_get_data_model>
  "parent_table", # <dm_get_data_model>
  "parent_key_cols", # <dm_get_data_model>
  "id", # <dm_get_data_model>
  "uk_col", # <dm_get_data_model>
  "column", # <dm_get_data_model>
  "key", # <dm_get_data_model>
  "key_fk", # <dm_get_data_model>
  "value", # <dm_get_all_column_types>
  "new_display", # <dm_set_colors>
  "display", # <dm_set_colors>
  "new_child_table", # <rename_unique>
  "child_table", # <rename_unique>
  "new_parent_table", # <rename_unique>
  "parent_table", # <rename_unique>
  "new_table", # <rename_unique>
  "new_child_table", # <split_to_list>
  "child_table", # <split_to_list>
  "new_parent_table", # <split_to_list>
  "parent_table", # <split_to_list>
  "new_table", # <split_to_list>
  "child_cols", # <split_to_list>
  "parent_cols", # <split_to_list>
  "on_delete", # <split_to_list>
  "on_delete", # <dm_examine_cardinalities_impl>
  "parent_table", # <dm_examine_cardinalities_impl>
  "child_table", # <dm_examine_cardinalities_impl>
  "x_label", # <dm_examine_cardinalities_impl>
  "parent_key_cols", # <dm_examine_cardinalities_impl>
  "y_label", # <dm_examine_cardinalities_impl>
  "child_fk_cols", # <dm_examine_cardinalities_impl>
  "cardinality", # <bullets_cardinalities>
  "column", # <dm_examine_constraints>
  "columns", # <dm_examine_constraints>
  "problem", # <print.dm_examine_constraints>
  "kind", # <print.dm_examine_constraints>
  "ref_table", # <print.dm_examine_constraints>
  "into", # <print.dm_examine_constraints>
  "text", # <print.dm_examine_constraints>
  "uk_col", # <check_pk_constraints>
  "kind", # <check_pk_constraints>
  "pk_col", # <check_pk_constraints>
  "candidate", # <check_pk_constraints>
  "why", # <check_pk_constraints>
  "t1", # <check_fk_constraints>
  "child_table", # <check_fk_constraints>
  "child_fk_cols", # <check_fk_constraints>
  "t2", # <check_fk_constraints>
  "parent_table", # <check_fk_constraints>
  "parent_key_cols", # <check_fk_constraints>
  "problem", # <check_fk_constraints>
  "t1_name", # <check_fk_constraints>
  "kind", # <check_fk_constraints>
  "colname", # <check_fk_constraints>
  "t2_name", # <check_fk_constraints>
  "is_key", # <check_fk_constraints>
  "node", # <dm_get_filtered_table>
  "parent", # <dm_get_filtered_table>
  "distance", # <dm_get_filtered_table>
  "filters", # <dm_get_filters_impl>
  "filter_expr", # <dm_get_filters_impl>
  "fks", # <dm_rm_fk_impl>
  "ref_table", # <dm_rm_fk_impl>
  "need_ref", # <dm_rm_fk_impl>
  "column", # <enum_fk_candidates_impl>
  "why", # <enum_fk_candidates_impl>
  "candidate", # <enum_fk_candidates_impl>
  "fks", # <create_graph_from_dm>
  "ref_table", # <create_graph_from_dm>
  "JSON_AGG", # <sql_json_nest.PqConnection>
  "JSON_BUILD_OBJECT", # <sql_json_nest.PqConnection>
  "JSON_BUILD_OBJECT", # <sql_json_pack.PqConnection>
  "key", # <is_unique_key_se>
  "any_na", # <is_unique_key_se>
  ".", # <is_unique_key_se>
  "data", # <is_unique_key_se>
  "schemata", # <dm_learn_from_db>
  "table_catalog", # <dm_learn_from_db>
  "table_schema", # <dm_learn_from_db>
  "table_name", # <dm_learn_from_db>
  "catalog", # <dm_learn_from_db>
  "columns", # <dm_learn_from_db>
  "ordinal_position", # <dm_learn_from_db>
  "constraint_column_usage", # <dm_learn_from_db>
  "key_column_usage", # <dm_learn_from_db>
  "table_constraints", # <dm_learn_from_db>
  "column_name", # <dm_learn_from_db>
  "constraint_type", # <dm_learn_from_db>
  "constraint_catalog", # <dm_learn_from_db>
  "constraint_schema", # <dm_learn_from_db>
  "constraint_name", # <dm_learn_from_db>
  "is_autoincrement", # <dm_learn_from_db>
  "pks", # <dm_learn_from_db>
  "delete_rule", # <dm_learn_from_db>
  "constraint_column_usage.dm_name", # <dm_learn_from_db>
  "constraint_column_usage.column_name", # <dm_learn_from_db>
  "key_column_usage.dm_name", # <dm_learn_from_db>
  "key_column_usage.column_name", # <dm_learn_from_db>
  "ref_table", # <dm_learn_from_db>
  "ref_column", # <dm_learn_from_db>
  "column", # <dm_learn_from_db>
  "fks", # <dm_learn_from_db>
  "extra", # <dm_meta_raw>
  "constraint_catalog", # <dm_meta_raw>
  "constraint_schema", # <dm_meta_raw>
  "table_name", # <dm_meta_raw>
  "constraint_type", # <dm_meta_raw>
  "constraint_name", # <dm_meta_raw>
  "referenced_table_name", # <dm_meta_raw>
  "referenced_table_schema", # <dm_meta_raw>
  "referenced_column_name", # <dm_meta_raw>
  "table_constraints", # <dm_meta_add_keys>
  "constraint_catalog", # <dm_meta_add_keys>
  "constraint_schema", # <dm_meta_add_keys>
  "constraint_name", # <dm_meta_add_keys>
  "table_catalog", # <dm_meta_add_keys>
  "table_schema", # <dm_meta_add_keys>
  "table_name", # <dm_meta_add_keys>
  "tables", # <dm_meta_add_keys>
  "key_column_usage", # <dm_meta_add_keys>
  "ordinal_position", # <dm_meta_add_keys>
  "column_name", # <dm_meta_add_keys>
  "columns", # <dm_meta_add_keys>
  "constraint_column_usage", # <dm_meta_add_keys>
  "schemata", # <dm_meta_simple_add_keys>
  "catalog_name", # <dm_meta_simple_add_keys>
  "schema_name", # <dm_meta_simple_add_keys>
  "tables", # <dm_meta_simple_add_keys>
  "table_catalog", # <dm_meta_simple_add_keys>
  "table_schema", # <dm_meta_simple_add_keys>
  "table_name", # <dm_meta_simple_add_keys>
  "columns", # <dm_meta_simple_add_keys>
  "column_name", # <dm_meta_simple_add_keys>
  "schemata", # <select_dm_meta>
  "catalog_name", # <select_dm_meta>
  "schema_name", # <select_dm_meta>
  "tables", # <select_dm_meta>
  "table_catalog", # <select_dm_meta>
  "table_schema", # <select_dm_meta>
  "table_name", # <select_dm_meta>
  "table_type", # <select_dm_meta>
  "columns", # <select_dm_meta>
  "column_name", # <select_dm_meta>
  "ordinal_position", # <select_dm_meta>
  "column_default", # <select_dm_meta>
  "is_nullable", # <select_dm_meta>
  "is_autoincrement", # <select_dm_meta>
  "table_constraints", # <select_dm_meta>
  "constraint_catalog", # <select_dm_meta>
  "constraint_schema", # <select_dm_meta>
  "constraint_name", # <select_dm_meta>
  "constraint_type", # <select_dm_meta>
  "delete_rule", # <select_dm_meta>
  "key_column_usage", # <select_dm_meta>
  "constraint_column_usage", # <select_dm_meta>
  "DATABASE", # <filter_dm_meta>
  "catalog_name", # <filter_dm_meta>
  "table_catalog", # <filter_dm_meta>
  "schema_name", # <filter_dm_meta>
  "table_schema", # <filter_dm_meta>
  "catalog_name", # <filter_dm_meta_simple>
  "table_catalog", # <filter_dm_meta_simple>
  "schema_name", # <filter_dm_meta_simple>
  "table_schema", # <filter_dm_meta_simple>
  "constraint_catalog", # <mssql_constraint_column_usage>
  "constraint_schema", # <mssql_constraint_column_usage>
  "constraint_name", # <mssql_constraint_column_usage>
  "constraint_type", # <mssql_constraint_column_usage>
  "name", # <mssql_constraint_column_usage>
  "object_id", # <mssql_constraint_column_usage>
  "catalog", # <mssql_constraint_column_usage>
  "table_schema", # <mssql_constraint_column_usage>
  "table_name", # <mssql_constraint_column_usage>
  "column_name", # <mssql_constraint_column_usage>
  "constraint_column_id", # <mssql_constraint_column_usage>
  "tailnum", # <dm_nycflights13>
  "carrier", # <dm_nycflights13>
  "faa", # <dm_nycflights13>
  "origin", # <dm_nycflights13>
  "time_hour", # <dm_nycflights13>
  "dest", # <dm_nycflights13>
  "data", # <dm_paste_select>
  "cols", # <dm_paste_select>
  "film", # <dm_pixarfilms>
  "award_type", # <dm_pixarfilms>
  "genre", # <dm_pixarfilms>
  "category", # <dm_pixarfilms>
  "value", # <dm_pixarfilms>
  "column", # <enum_pk_candidates>
  "columns", # <enum_pk_candidates>
  "column", # <enum_pk_candidates_impl>
  "why", # <enum_pk_candidates_impl>
  "candidate", # <enum_pk_candidates_impl>
  "pks", # <dm_upgrade>
  "ref_table", # <dm_validate>
  "ref_column", # <dm_validate>
  "zoomed", # <dm_discard_zoomed>
  "column", # <cdm_enum_fk_candidates>
  "columns", # <cdm_enum_fk_candidates>
  NULL
))

# Registered in .onLoad()
compare_proxy.dm <- function(x, path = NULL, ...) {
  list(
    object = x %>%
      dm_get_def() %>%
      mutate(., data = transpose(select(., -table))) %>%
      select(table, data) %>%
      deframe() %>%
      map(as.list) %>%
      map(
        ~ {
          .x$pks <- as.list(.x$pks)
          fk_data <-
            .x$fks %>%
            select(-table) %>%
            transpose()
          .x$fks <- set_names(fk_data, .x$fks$table)
          .x
        }
      ),
    path = path
  )
}

skip_if_error <- function(expr) {
  tryCatch(
    force(expr),
    error = function(e) {
      skip(e$message)
    }
  )
}

skip_if_remote_src <- function(src = my_test_src()) {
  if (is_db(src)) skip("works only locally")
}

skip_if_src <- function(...) {
  if (my_test_src_name %in% c(...)) {
    skip(glue::glue("does not work on {commas(tick(c(...)))}"))
  }
}

skip_if_src_not <- function(...) {
  if (!(my_test_src_name %in% c(...))) {
    skip(glue::glue("only works on {commas(tick(c(...)))}"))
  }
}

skip_if_schema_supported <- function() {
  not_ok_dbs <- schema_supported_dbs()[["test_shortcut"]]
  skip_if_src(not_ok_dbs)
}
skip_if_schema_not_supported <- function() {
  not_ok_dbs <- schema_supported_dbs()[["test_shortcut"]]
  skip_if_src_not(not_ok_dbs)
}

skip_if_ide <- function() {
  if (!isTRUE(as.logical(Sys.getenv("CI")))) {
    skip("Slow test. To run, set CI=true")
  }
}

skip_if_offline <- function() {
  if (Sys.getenv("DM_OFFLINE") != "") {
    skip("DM_OFFLINE environment variable set")
  }
  testthat::skip_if_offline()
}

suppress_mssql_message <- function(code) {
  if (my_test_src_name == "mssql") {
    suppressMessages(code)
  } else {
    code
  }
}

suppress_mssql_warning <- function(code) {
  if (my_test_src_name == "mssql") {
    suppressWarnings(code)
  } else {
    code
  }
}

test_that("`dm_examine_constraints()` works", {
  # case of no constraints:
  expect_identical(
    dm_examine_constraints(dm_test_obj()),
    tibble(
      table = character(),
      kind = character(),
      columns = new_keys(),
      ref_table = character(),
      is_key = logical(),
      problem = character()
    ) %>%
      new_dm_examine_constraints()
  )

  skip_if_src("maria")

  # case of some constraints, all met:
  expect_identical(
    dm_examine_constraints(dm_for_disambiguate()),
    tibble(
      table = c("iris_1", "iris_2"),
      kind = c("PK", "FK"),
      columns = new_keys("key"),
      ref_table = c(NA, "iris_1"),
      is_key = TRUE,
      problem = ""
    ) %>%
      new_dm_examine_constraints()
  )
})

test_that("output", {
  skip_if_ide()

  expect_snapshot({
    dm() %>% dm_examine_constraints()

    dm_nycflights_small() %>% dm_examine_constraints()
    dm_nycflights_small_cycle() %>% dm_examine_constraints()
    dm_nycflights_small_cycle() %>%
      dm_select_tbl(-flights) %>%
      dm_examine_constraints()

    "n column"
    dm_for_filter_w_cycle() %>%
      dm_examine_constraints()
  })
})

test_that("output as tibble", {
  skip_if_ide()

  expect_snapshot({
    dm_nycflights_small_cycle() %>%
      dm_examine_constraints() %>%
      as_tibble()
  })
})

test_that(".max_value parameter works", {
  skip_if_ide()

  expect_snapshot({
    dm_nycflights_small_cycle() %>%
      dm_examine_constraints(.max_value = Inf)
  })
})


# test compound keys ------------------------------------------------------

test_that("output for compound keys", {
  # FIXME: COMPOUND: Need proper test
  skip_if_remote_src()

  expect_snapshot({
    bad_dm() %>%
      dm_examine_constraints()
  })
})

# Test unique keys for weak FK (no explicit PK set) -----------------------

test_that("Non-explicit PKs should be tested too", {
  expect_snapshot(
    # dm_for_card() has no PKs set, only FKs
    dm_for_card() %>%
      dm_examine_constraints()
  )
})

test_that("`dm_examine_constraints()` API", {
  local_options(lifecycle_verbosity = "warning")

  expect_snapshot({
    dm_examine_constraints(dm_test_obj(), progress = FALSE)
    dm_examine_constraints(dm = dm_test_obj())
  })
})

test_that("`dm_examine_constraints()` API (2)", {
  expect_snapshot(error = TRUE, {
    dm_examine_constraints(dm_test_obj(), foo = "bar")
  })
})

test_that("can access tables", {
  local_options(lifecycle_verbosity = "quiet")

  expect_identical(tbl(dm_nycflights13(), "airlines"), nycflights_subset()$airlines)
  expect_dm_error(
    tbl_impl(dm_nycflights13(), "x"),
    class = "table_not_in_dm"
  )
})

test_that("'copy_to.dm()' works", {
  local_options(lifecycle_verbosity = "quiet")

  expect_dm_error(
    copy_to(dm_for_filter(), letters[1:5], name = "letters"),
    "only_data_frames_supported"
  )

  expect_dm_error(
    copy_to(dm_for_filter(), list(mtcars, iris)),
    "only_data_frames_supported"
  )

  expect_dm_error(
    copy_to(dm_for_filter(), mtcars, overwrite = TRUE),
    "no_overwrite"
  )

  skip_if_src_not("df", "mssql")

  # `tibble()` call necessary, #322
  car_table <- test_src_frame(!!!mtcars)

  expect_equivalent_dm(
    suppress_mssql_message(copy_to(dm_for_filter(), mtcars, "car_table")),
    dm(dm_for_filter(), car_table)
  )

  # FIXME: Why do we do name repair in copy_to()?
  expect_equivalent_dm(
    suppress_mssql_message(expect_name_repair_message(
      copy_to(dm_for_filter(), mtcars, "")
    )),
    dm(dm_for_filter(), ...7 = car_table)
  )
})

test_that("'copy_to.dm()' works (2)", {
  skip_if_ide()

  local_options(lifecycle_verbosity = "quiet")

  expect_dm_error(
    copy_to(dm(), mtcars, c("car_table", "another_table")),
    "one_name_for_copy_to"
  )

  # rename old and new tables if `repair = unique`
  expect_name_repair_message(
    expect_equivalent_dm(
      copy_to(dm(mtcars), mtcars),
      dm(mtcars...1 = mtcars, mtcars...2 = tibble(mtcars))
    )
  )

  expect_equivalent_dm(
    expect_silent(
      copy_to(dm(mtcars), mtcars, quiet = TRUE)
    ),
    dm(mtcars...1 = mtcars, mtcars...2 = tibble(mtcars))
  )

  # throw error if duplicate table names and `repair = check_unique`
  expect_dm_error(
    dm(mtcars) %>% copy_to(mtcars, repair = "check_unique"),
    "need_unique_names"
  )

  # copying `tibble` from chosen src to sqlite() `dm`
  expect_equivalent_dm(
    copy_to(dm_for_filter_duckdb(), data_card_1(), "test_table"),
    dm(dm_for_filter_duckdb(), test_table = data_card_1_duckdb())
  )

  # copying sqlite() `tibble` to `dm` on src of choice
  expect_equivalent_dm(
    suppress_mssql_message(copy_to(dm_for_filter(), data_card_1_duckdb(), "test_table_1")),
    dm(dm_for_filter(), test_table_1 = data_card_1())
  )
})

# Create a snapshot of the code and of the resulting dm object
# This requires a certain of the code.
expect_mw_cg_snapshot <- function(...) {
  testthat::expect_snapshot(...)
  quo <- enquos(...)[[1]]
  out <- eval_tidy(quo)

  expr <- quo_get_expr(quo)
  if (!identical(expr[[1]], quote(`{`)) || length(expr) != 2) {
    result <- glue::as_glue("expect_mw_cg_snapshot(): check structure of your code block")
    testthat::expect_snapshot(result)
    return()
  }

  dm_expr <- expr[[2]][[2]]

  dm <- eval_tidy(dm_expr, env = quo_get_env(quo))

  result <-
    new_cg_block(dm) %>%
    cg_add_call(!!out$call) %>%
    cg_eval_block()

  testthat::expect_snapshot(result)
}

get_dm <- function(name) {
  if (name == "nycflights13") {
    return(dm::dm_nycflights13())
  }

  basename <- paste0(name, ".R")
  path <- system.file("relational", basename, package = utils::packageName(), mustWork = TRUE)
  source(path, local = TRUE)$value
}

# utils::Rprof(paste0("~/", Sys.getpid(), ".Rprof"), line.profiling = TRUE)
#
# withr::defer(utils::Rprof(NULL), teardown_env())

local_options(
  pillar.min_title_chars = NULL,
  pillar.max_title_chars = NULL,
  pillar.max_footer_lines = NULL,
  pillar.bold = NULL,
  .frame = teardown_env()
)

if (rlang::is_installed("dbplyr")) {
  has_tbl_sum <- exists("tbl_sum.tbl_sql", envir = asNamespace("dbplyr"))
  bindings <- compact(list(
    tbl_sum.tbl_sql = if (has_tbl_sum) function(x, ...) c(),
    tbl_format_header.tbl_sql = function(x, ...) invisible()
  ))
  local_mocked_bindings(!!!bindings, .package = "dbplyr", .env = teardown_env())
}

#' Check if column(s) can be used as keys
#'
#' @description
#' `check_key()` accepts a data frame and, optionally, columns.
#' It throws an error
#' if the specified columns are NOT a unique key of the data frame.
#' If the columns given in the ellipsis ARE a key, the data frame itself is returned silently, so that it can be used for piping.
#'
#' @param x The data frame whose columns should be tested for key properties.
#' @param ... The names of the columns to be checked, processed with
#'   [dplyr::select()]. If omitted, all columns will be checked.
#' @param .data Deprecated.
#'
#' @return Returns `x`, invisibly, if the check is passed.
#'   Otherwise an error is thrown and the reason for it is explained.
#'
#' @export
#' @examples
#' data <- tibble::tibble(a = c(1, 2, 1), b = c(1, 4, 1), c = c(5, 6, 7))
#' # this is failing:
#' try(check_key(data, a, b))
#'
#' # this is passing:
#' check_key(data, a, c)
#' check_key(data)
check_key <- function(x, ..., .data = deprecated()) {
  dm_local_error_call()
  if (!is_missing(.data)) {
    deprecate_warn("1.0.0", "check_key(.data = )", "check_key(x = )")
    return(check_key_impl0({{ .data }}, {{ x }}, ...))
  }

  check_key_impl({{ x }}, ...)
}

check_key_impl <- function(.data, ...) {
  data_q <- enquo(.data)
  original_data <- eval_tidy(data_q)

  selected_data <- original_data
  if (dots_n(...) > 0) {
    selected_data <- selected_data %>% select(...)
  }

  check_key_impl0(selected_data, as_label(data_q))
  invisible(original_data)
}

check_key_impl0 <- function(x, x_label) {
  orig_names <- colnames(x)
  cols_chosen <- syms(set_names(orig_names, glue("...{seq_along(orig_names)}")))

  if (inherits(x, "data.frame")) {
    any_duplicate_rows <- vctrs::vec_duplicate_any(x)
  } else {
    duplicate_rows <-
      x %>%
      safe_count(!!!cols_chosen) %>%
      select(n) %>%
      filter(n > 1) %>%
      head(1) %>%
      collect()
    any_duplicate_rows <- nrow(duplicate_rows) != 0
  }

  if (any_duplicate_rows) {
    abort_not_unique_key(x_label, orig_names)
  }
}

# an internal function to check if a column is a unique key of a table
is_unique_key <- function(.data, column) {
  col_q <- enexpr(column)
  col_name <- names(eval_select_indices(col_q, colnames(.data)))

  is_unique_key_se(.data, col_name)
}

#' @autoglobal
is_unique_key_se <- function(.data, colname, max_value = MAX_COMMAS) {
  val_names <- paste0("value", seq_along(colname))
  col_syms <- syms(colname)
  names(col_syms) <- val_names

  any_value_na_expr <-
    syms(val_names) %>%
    map(call2, .fn = quote(is.na)) %>%
    reduce(call2, .fn = quote(`|`))

  if (inherits(.data, "data.frame")) {
    count_tbl <-
      .data %>%
      select(!!!col_syms) %>%
      vctrs::vec_count() %>%
      unpack(key) %>%
      rename(n = count)
  } else {
    count_tbl <-
      .data %>%
      safe_count(!!!col_syms)
  }
  res_tbl <-
    count_tbl %>%
    mutate(any_na = if_else(!!any_value_na_expr, 1L, 0L)) %>%
    filter(n != 1 | any_na != 0L) %>%
    arrange(desc(n), !!!syms(val_names)) %>%
    utils::head(max_value + 1) %>%
    collect()

  res_tbl[val_names] <- map(res_tbl[val_names], format, trim = TRUE, justify = "none")
  res_tbl[val_names[-1]] <- map(res_tbl[val_names[-1]], ~ paste0(", ", .x))
  # Work around bit64 incompatibility
  res_tbl$value <- NA_character_
  res_tbl$value[res_tbl$any_na == 0] <- exec(paste0, !!!res_tbl[res_tbl$any_na == 0, val_names])

  duplicate_rows <-
    res_tbl %>%
    {
      # https://github.com/tidyverse/tidyr/issues/734
      tibble(data = list(.))
    } %>%
    mutate(unique = map_lgl(data, ~ nrow(.) == 0))

  duplicate_rows
}

#' Check column values for set equality
#'
#' @description
#' `check_set_equality()` is a wrapper of [check_subset()].
#'
#' It tests if one table is a subset of another and vice versa, i.e., if both sets are the same.
#' If not, it throws an error.
#'
#' @param x,y A data frame or lazy table.
#' @inheritParams rlang::args_dots_empty
#' @param x_select,y_select Key columns to restrict the check, processed with
#'   [dplyr::select()].
#' @param by_position Set to `TRUE` to ignore column names and match
#'   by position instead.
#'   The default means matching by name, use `x_select` and/or `y_select`
#'   to align the names.
#'
#' @return Returns `x`, invisibly, if the check is passed.
#'   Otherwise an error is thrown and the reason for it is explained.
#'
#' @export
#' @examples
#' data_1 <- tibble::tibble(a = c(1, 2, 1), b = c(1, 4, 1), c = c(5, 6, 7))
#' data_2 <- tibble::tibble(a = c(1, 2, 3), b = c(4, 5, 6), c = c(7, 8, 9))
#' # this is failing:
#' try(check_set_equality(data_1, data_2, x_select = a, y_select = a))
#'
#' data_3 <- tibble::tibble(a = c(2, 1, 2), b = c(4, 5, 6), c = c(7, 8, 9))
#' # this is passing:
#' check_set_equality(data_1, data_3, x_select = a, y_select = a)
#' # this is still failing:
#' try(check_set_equality(data_2, data_3))
check_set_equality <- function(x, y, ..., x_select = NULL, y_select = NULL, by_position = NULL) {
  dm_local_error_call()
  check_api(
    {{ x }},
    {{ y }},
    ...,
    x_select = {{ x_select }},
    y_select = {{ y_select }},
    by_position = by_position,
    target = check_set_equality_impl0
  )
  invisible(x)
}

check_set_equality_impl0 <- function(x, y, x_label, y_label) {
  catcher_1 <- tryCatch(
    {
      check_subset_impl0(x, y, x_label, y_label)
      NULL
    },
    error = identity
  )

  catcher_2 <- tryCatch(
    {
      check_subset_impl0(y, x, y_label, x_label)
      NULL
    },
    error = identity
  )

  catchers <- compact(list(catcher_1, catcher_2))

  if (length(catchers) > 0) {
    abort_sets_not_equal(map_chr(catchers, conditionMessage))
  }
}

#' Check column values for subset
#'
#' @description
#' `check_subset()` tests if `x` is a subset of `y`.
#' For convenience, the `x_select` and `y_select` arguments allow restricting the check
#' to a set of key columns without affecting the return value.
#'
#' @inheritParams check_set_equality
#'
#' @return Returns `x`, invisibly, if the check is passed.
#'   Otherwise an error is thrown and the reason for it is explained.
#'
#' @export
#' @examples
#' data_1 <- tibble::tibble(a = c(1, 2, 1), b = c(1, 4, 1), c = c(5, 6, 7))
#' data_2 <- tibble::tibble(a = c(1, 2, 3), b = c(4, 5, 6), c = c(7, 8, 9))
#' # this is passing:
#' check_subset(data_1, data_2, x_select = a, y_select = a)
#'
#' # this is failing:
#' try(check_subset(data_2, data_1))
check_subset <- function(x, y, ..., x_select = NULL, y_select = NULL, by_position = NULL) {
  dm_local_error_call()
  check_api(
    {{ x }},
    {{ y }},
    ...,
    x_select = {{ x_select }},
    y_select = {{ y_select }},
    by_position = by_position,
    target = check_subset_impl0
  )
  invisible(x)
}

check_subset_impl0 <- function(x, y, x_label, y_label) {
  # not using `is_subset()`, since then we would do the same job of finding
  # missing values/combinations twice
  res <- anti_join(x, y, by = set_names(colnames(y), colnames(x)))
  if (pull(count(head(res, 1))) == 0) {
    return()
  }

  # collect() for robust test output
  print(collect(head(res, n = 10)))

  abort_not_subset_of(x_label, colnames(x), y_label, colnames(y))
}

# similar to `check_subset()`, but evaluates to a boolean
is_subset <- function(t1, c1, t2, c2) {
  t1q <- enquo(t1)
  t2q <- enquo(t2)

  t1s <- eval_tidy(t1q) %>% select({{ c1 }})
  t2s <- eval_tidy(t2q) %>% select({{ c2 }})

  is_subset_se(t1s, t2s)
}

is_subset_se <- function(x, y) {
  res <- anti_join(x, y, by = set_names(colnames(y), colnames(x)))
  pull(count(head(res, 1))) == 0
}

check_api <- function(
  x,
  y,
  ...,
  x_select = NULL,
  y_select = NULL,
  by_position = NULL,
  call = caller_env(),
  target = list
) {
  if (dots_n(...) >= 2) {
    name <- as.character(frame_call(call)[[1]] %||% "check_api")
    deprecate_warn(
      "1.0.0",
      paste0(name, "(c1 = )"),
      paste0(name, "(x_select = )"),
      details = c(
        "Use `y_select` instead of `c2`, and `x` and `y` instead of `t1` and `t2`.",
        "Using `by_position = TRUE` for compatibility."
      )
    )
    stopifnot(is.null(by_position))
    check_api_impl(
      {{ x }},
      {{ y }},
      ...,
      by_position = TRUE,
      target = target
    )
  } else {
    check_dots_empty(call = call)
    check_api_impl(
      {{ x }},
      {{ x_select }},
      {{ y }},
      {{ y_select }},
      by_position = by_position %||% FALSE,
      target = target
    )
  }
}

check_api_impl <- function(t1, c1, t2, c2, ..., by_position, target) {
  t1q <- enquo(t1)
  t2q <- enquo(t2)

  c1q <- enquo(c1)
  c2q <- enquo(c2)

  if (!quo_is_null(c1q)) {
    t1 <- t1 %>% select(!!c1q)
  }

  if (!quo_is_null(c2q)) {
    t2 <- t2 %>% select(!!c2q)
  }

  if (!isTRUE(by_position)) {
    y_idx <- match(colnames(t1), colnames(t2))
    if (anyNA(y_idx)) {
      cli::cli_abort(
        "{.code by_position = FALSE} or {.code by_position = NULL} require column names in {.arg x} to match those in {.arg y}.",
        call = dm_error_call()
      )
    }

    t2 <-
      t2 %>%
      select(!!y_idx)
  }

  target(x = t1, y = t2, x_label = as_label(t1q), y_label = as_label(t2q))
}

test_that("check_key() API", {
  local_options(lifecycle_verbosity = "warning")

  expect_snapshot({
    check_key(tibble(a = 1), a)
    check_key(.data = tibble(a = 1), a)
    check_key(a, .data = tibble(a = 1))
  })
})

test_that("check_key() returns data frame for valid keys", {
  # Test the specific issue from #2221: check_key should return the data frame, not NULL
  dat <- tibble(a = c(1, 2, 3), b = c(5, 5, 6), c = c(7, 8, 9))

  # Single column key
  result <- check_key(dat, a)
  expect_identical(result, dat)

  # Multiple column key
  result2 <- check_key(dat, a, c)
  expect_identical(result2, dat)

  # All columns key
  result3 <- check_key(dat)
  expect_identical(result3, dat)

  # Test that it can be used in piping
  piped_result <- dat %>%
    check_key(a) %>%
    select(a)
  expect_equal(piped_result, select(dat, a))
})

test_that("check_key() checks primary key properly?", {
  expect_dm_error(
    check_key(data_mcard(), c1, c2),
    class = "not_unique_key"
  )

  expect_silent(check_key(data_mcard(), c1, c3))

  expect_silent(check_key(data_mcard(), c2, c3))

  test_tbl <- tibble(nn = 1:5, n = 6:10)
  expect_silent(
    check_key(test_tbl, !!!c("n1" = sym("n"), "n2" = sym("nn")))
  )

  expect_silent(
    check_key(test_tbl, !!!c(sym("n"), sym("nn")))
  )

  expect_silent(
    check_key(test_tbl, everything())
  )

  # Since dm 1.0.0:
  expect_silent(check_key(test_tbl))

  # if {tidyselect} selects nothing
  # cf. issue #360
  skip_if_remote_src()
  expect_dm_error(
    check_key(data_mcard(), starts_with("d")),
    "not_unique_key"
  )

  skip("Need to think about it")
  expect_silent(
    dm_nycflights_small() %>%
      dm_zoom_to(airlines) %>%
      check_key(carrier)
  )
})

test_that("check_key() returns data frame when key is valid", {
  # Test from issue #2221: check_key() should return the data frame when columns are a valid key
  dat <- tibble(a = c(1, 2, 3), b = c(5, 5, 6), c = c(7, 8, 9))

  # Should return the original data frame, not NULL
  result <- check_key(dat, a)
  expect_identical(result, dat)

  # Also test with multiple columns as key
  result2 <- check_key(dat, a, c)
  expect_identical(result2, dat)

  # Also test with data_mcard examples
  result3 <- check_key(data_mcard(), c1, c3)
  expect_identical(result3, data_mcard())
})

test_that("check_api() new interface", {
  local_options(lifecycle_verbosity = "quiet")

  expect_same(
    check_api(data_mcard_1(), data_mcard_2(), x_select = a, y_select = c(a = b)),
    check_api(x = data_mcard_1(), data_mcard_2(), x_select = a, y_select = c(a = b)),
    check_api(data_mcard_1(), y = data_mcard_2(), x_select = a, y_select = c(a = b)),
    check_api(x = data_mcard_1(), y = data_mcard_2(), x_select = a, y_select = c(a = b)),
    check_api(y = data_mcard_2(), x = data_mcard_1(), x_select = a, y_select = c(a = b)),
    check_api(data_mcard_1(), a, data_mcard_2(), c(a = b))
  )

  expect_same(
    check_api(data_mcard_1(), data_mcard_2(), x_select = a, y_select = b, by_position = TRUE),
    check_api(x = data_mcard_1(), data_mcard_2(), x_select = a, y_select = b, by_position = TRUE),
    check_api(data_mcard_1(), y = data_mcard_2(), x_select = a, y_select = b, by_position = TRUE),
    check_api(
      x = data_mcard_1(),
      y = data_mcard_2(),
      x_select = a,
      y_select = b,
      by_position = TRUE
    ),
    check_api(
      y = data_mcard_2(),
      x = data_mcard_1(),
      x_select = a,
      y_select = b,
      by_position = TRUE
    ),
    check_api(data_mcard_1(), a, data_mcard_2(), b)
  )
})

test_that("check_api() compatibility", {
  local_options(lifecycle_verbosity = "quiet")

  expect_same(
    check_api(data_mcard_1(), a, data_mcard_2(), b),
    check_api(t1 = data_mcard_1(), c1 = a, t2 = data_mcard_2(), c2 = b)
  )
  expect_same(
    check_api(c2 = b, data_mcard_1(), a, data_mcard_2()),
    check_api(data_mcard_1(), c2 = b, a, data_mcard_2()),
    check_api(data_mcard_1(), a, c2 = b, data_mcard_2()),
    check_api(data_mcard_1(), a, data_mcard_2(), c2 = b)
  )
  expect_same(
    check_api(t2 = data_mcard_2(), data_mcard_1(), a, b),
    check_api(data_mcard_1(), t2 = data_mcard_2(), a, b),
    check_api(data_mcard_1(), a, t2 = data_mcard_2(), b),
    check_api(data_mcard_1(), a, b, t2 = data_mcard_2())
  )
  expect_same(
    check_api(t2 = data_mcard_2(), c2 = b, data_mcard_1(), a),
    check_api(t2 = data_mcard_2(), data_mcard_1(), c2 = b, a),
    check_api(t2 = data_mcard_2(), data_mcard_1(), a, c2 = b),
    check_api(c2 = b, t2 = data_mcard_2(), data_mcard_1(), a),
    check_api(c2 = b, data_mcard_1(), t2 = data_mcard_2(), a),
    check_api(c2 = b, data_mcard_1(), a, t2 = data_mcard_2()),
    check_api(data_mcard_1(), t2 = data_mcard_2(), c2 = b, a),
    check_api(data_mcard_1(), t2 = data_mcard_2(), a, c2 = b),
    check_api(data_mcard_1(), c2 = b, t2 = data_mcard_2(), a),
    check_api(data_mcard_1(), c2 = b, a, t2 = data_mcard_2()),
    check_api(data_mcard_1(), a, t2 = data_mcard_2(), c2 = b),
    check_api(data_mcard_1(), a, c2 = b, t2 = data_mcard_2())
  )
  expect_same(
    check_api(c1 = a, data_mcard_1(), data_mcard_2(), b),
    check_api(data_mcard_1(), c1 = a, data_mcard_2(), b),
    check_api(data_mcard_1(), data_mcard_2(), c1 = a, b),
    check_api(data_mcard_1(), data_mcard_2(), b, c1 = a)
  )
  expect_same(
    check_api(c1 = a, c2 = b, data_mcard_1(), data_mcard_2()),
    check_api(c1 = a, data_mcard_1(), c2 = b, data_mcard_2()),
    check_api(c1 = a, data_mcard_1(), data_mcard_2(), c2 = b),
    check_api(c2 = b, c1 = a, data_mcard_1(), data_mcard_2()),
    check_api(c2 = b, data_mcard_1(), c1 = a, data_mcard_2()),
    check_api(c2 = b, data_mcard_1(), data_mcard_2(), c1 = a),
    check_api(data_mcard_1(), c1 = a, c2 = b, data_mcard_2()),
    check_api(data_mcard_1(), c1 = a, data_mcard_2(), c2 = b),
    check_api(data_mcard_1(), c2 = b, c1 = a, data_mcard_2()),
    check_api(data_mcard_1(), c2 = b, data_mcard_2(), c1 = a),
    check_api(data_mcard_1(), data_mcard_2(), c1 = a, c2 = b),
    check_api(data_mcard_1(), data_mcard_2(), c2 = b, c1 = a)
  )
  expect_same(
    check_api(c1 = a, t2 = data_mcard_2(), data_mcard_1(), b),
    check_api(c1 = a, data_mcard_1(), t2 = data_mcard_2(), b),
    check_api(c1 = a, data_mcard_1(), b, t2 = data_mcard_2()),
    check_api(t2 = data_mcard_2(), c1 = a, data_mcard_1(), b),
    check_api(t2 = data_mcard_2(), data_mcard_1(), c1 = a, b),
    check_api(t2 = data_mcard_2(), data_mcard_1(), b, c1 = a),
    check_api(data_mcard_1(), c1 = a, t2 = data_mcard_2(), b),
    check_api(data_mcard_1(), c1 = a, b, t2 = data_mcard_2()),
    check_api(data_mcard_1(), t2 = data_mcard_2(), c1 = a, b),
    check_api(data_mcard_1(), t2 = data_mcard_2(), b, c1 = a),
    check_api(data_mcard_1(), b, c1 = a, t2 = data_mcard_2()),
    check_api(data_mcard_1(), b, t2 = data_mcard_2(), c1 = a)
  )
  expect_same(
    check_api(c1 = a, t2 = data_mcard_2(), c2 = b, data_mcard_1()),
    check_api(c1 = a, t2 = data_mcard_2(), data_mcard_1(), c2 = b),
    check_api(c1 = a, c2 = b, t2 = data_mcard_2(), data_mcard_1()),
    check_api(c1 = a, c2 = b, data_mcard_1(), t2 = data_mcard_2()),
    check_api(c1 = a, data_mcard_1(), t2 = data_mcard_2(), c2 = b),
    check_api(c1 = a, data_mcard_1(), c2 = b, t2 = data_mcard_2()),
    check_api(t2 = data_mcard_2(), c1 = a, c2 = b, data_mcard_1()),
    check_api(t2 = data_mcard_2(), c1 = a, data_mcard_1(), c2 = b),
    check_api(t2 = data_mcard_2(), c2 = b, c1 = a, data_mcard_1()),
    check_api(t2 = data_mcard_2(), c2 = b, data_mcard_1(), c1 = a),
    check_api(t2 = data_mcard_2(), data_mcard_1(), c1 = a, c2 = b),
    check_api(t2 = data_mcard_2(), data_mcard_1(), c2 = b, c1 = a),
    check_api(c2 = b, c1 = a, t2 = data_mcard_2(), data_mcard_1()),
    check_api(c2 = b, c1 = a, data_mcard_1(), t2 = data_mcard_2()),
    check_api(c2 = b, t2 = data_mcard_2(), c1 = a, data_mcard_1()),
    check_api(c2 = b, t2 = data_mcard_2(), data_mcard_1(), c1 = a),
    check_api(c2 = b, data_mcard_1(), c1 = a, t2 = data_mcard_2()),
    check_api(c2 = b, data_mcard_1(), t2 = data_mcard_2(), c1 = a),
    check_api(data_mcard_1(), c1 = a, t2 = data_mcard_2(), c2 = b),
    check_api(data_mcard_1(), c1 = a, c2 = b, t2 = data_mcard_2()),
    check_api(data_mcard_1(), t2 = data_mcard_2(), c1 = a, c2 = b),
    check_api(data_mcard_1(), t2 = data_mcard_2(), c2 = b, c1 = a),
    check_api(data_mcard_1(), c2 = b, c1 = a, t2 = data_mcard_2()),
    check_api(data_mcard_1(), c2 = b, t2 = data_mcard_2(), c1 = a)
  )
  expect_same(
    check_api(t1 = data_mcard_1(), a, data_mcard_2(), b),
    check_api(a, t1 = data_mcard_1(), data_mcard_2(), b),
    check_api(a, data_mcard_2(), t1 = data_mcard_1(), b),
    check_api(a, data_mcard_2(), b, t1 = data_mcard_1())
  )
  expect_same(
    check_api(t1 = data_mcard_1(), c2 = b, a, data_mcard_2()),
    check_api(t1 = data_mcard_1(), a, c2 = b, data_mcard_2()),
    check_api(t1 = data_mcard_1(), a, data_mcard_2(), c2 = b),
    check_api(c2 = b, t1 = data_mcard_1(), a, data_mcard_2()),
    check_api(c2 = b, a, t1 = data_mcard_1(), data_mcard_2()),
    check_api(c2 = b, a, data_mcard_2(), t1 = data_mcard_1()),
    check_api(a, t1 = data_mcard_1(), c2 = b, data_mcard_2()),
    check_api(a, t1 = data_mcard_1(), data_mcard_2(), c2 = b),
    check_api(a, c2 = b, t1 = data_mcard_1(), data_mcard_2()),
    check_api(a, c2 = b, data_mcard_2(), t1 = data_mcard_1()),
    check_api(a, data_mcard_2(), t1 = data_mcard_1(), c2 = b),
    check_api(a, data_mcard_2(), c2 = b, t1 = data_mcard_1())
  )
  expect_same(
    check_api(t1 = data_mcard_1(), t2 = data_mcard_2(), a, b),
    check_api(t1 = data_mcard_1(), a, t2 = data_mcard_2(), b),
    check_api(t1 = data_mcard_1(), a, b, t2 = data_mcard_2()),
    check_api(t2 = data_mcard_2(), t1 = data_mcard_1(), a, b),
    check_api(t2 = data_mcard_2(), a, t1 = data_mcard_1(), b),
    check_api(t2 = data_mcard_2(), a, b, t1 = data_mcard_1()),
    check_api(a, t1 = data_mcard_1(), t2 = data_mcard_2(), b),
    check_api(a, t1 = data_mcard_1(), b, t2 = data_mcard_2()),
    check_api(a, t2 = data_mcard_2(), t1 = data_mcard_1(), b),
    check_api(a, t2 = data_mcard_2(), b, t1 = data_mcard_1()),
    check_api(a, b, t1 = data_mcard_1(), t2 = data_mcard_2()),
    check_api(a, b, t2 = data_mcard_2(), t1 = data_mcard_1())
  )
  expect_same(
    check_api(t1 = data_mcard_1(), t2 = data_mcard_2(), c2 = b, a),
    check_api(t1 = data_mcard_1(), t2 = data_mcard_2(), a, c2 = b),
    check_api(t1 = data_mcard_1(), c2 = b, t2 = data_mcard_2(), a),
    check_api(t1 = data_mcard_1(), c2 = b, a, t2 = data_mcard_2()),
    check_api(t1 = data_mcard_1(), a, t2 = data_mcard_2(), c2 = b),
    check_api(t1 = data_mcard_1(), a, c2 = b, t2 = data_mcard_2()),
    check_api(t2 = data_mcard_2(), t1 = data_mcard_1(), c2 = b, a),
    check_api(t2 = data_mcard_2(), t1 = data_mcard_1(), a, c2 = b),
    check_api(t2 = data_mcard_2(), c2 = b, t1 = data_mcard_1(), a),
    check_api(t2 = data_mcard_2(), c2 = b, a, t1 = data_mcard_1()),
    check_api(t2 = data_mcard_2(), a, t1 = data_mcard_1(), c2 = b),
    check_api(t2 = data_mcard_2(), a, c2 = b, t1 = data_mcard_1()),
    check_api(c2 = b, t1 = data_mcard_1(), t2 = data_mcard_2(), a),
    check_api(c2 = b, t1 = data_mcard_1(), a, t2 = data_mcard_2()),
    check_api(c2 = b, t2 = data_mcard_2(), t1 = data_mcard_1(), a),
    check_api(c2 = b, t2 = data_mcard_2(), a, t1 = data_mcard_1()),
    check_api(c2 = b, a, t1 = data_mcard_1(), t2 = data_mcard_2()),
    check_api(c2 = b, a, t2 = data_mcard_2(), t1 = data_mcard_1()),
    check_api(a, t1 = data_mcard_1(), t2 = data_mcard_2(), c2 = b),
    check_api(a, t1 = data_mcard_1(), c2 = b, t2 = data_mcard_2()),
    check_api(a, t2 = data_mcard_2(), t1 = data_mcard_1(), c2 = b),
    check_api(a, t2 = data_mcard_2(), c2 = b, t1 = data_mcard_1()),
    check_api(a, c2 = b, t1 = data_mcard_1(), t2 = data_mcard_2()),
    check_api(a, c2 = b, t2 = data_mcard_2(), t1 = data_mcard_1())
  )
  expect_same(
    check_api(t1 = data_mcard_1(), c1 = a, data_mcard_2(), b),
    check_api(t1 = data_mcard_1(), data_mcard_2(), c1 = a, b),
    check_api(t1 = data_mcard_1(), data_mcard_2(), b, c1 = a),
    check_api(c1 = a, t1 = data_mcard_1(), data_mcard_2(), b),
    check_api(c1 = a, data_mcard_2(), t1 = data_mcard_1(), b),
    check_api(c1 = a, data_mcard_2(), b, t1 = data_mcard_1()),
    check_api(data_mcard_2(), t1 = data_mcard_1(), c1 = a, b),
    check_api(data_mcard_2(), t1 = data_mcard_1(), b, c1 = a),
    check_api(data_mcard_2(), c1 = a, t1 = data_mcard_1(), b),
    check_api(data_mcard_2(), c1 = a, b, t1 = data_mcard_1()),
    check_api(data_mcard_2(), b, t1 = data_mcard_1(), c1 = a),
    check_api(data_mcard_2(), b, c1 = a, t1 = data_mcard_1())
  )
  expect_same(
    check_api(t1 = data_mcard_1(), c1 = a, c2 = b, data_mcard_2()),
    check_api(t1 = data_mcard_1(), c1 = a, data_mcard_2(), c2 = b),
    check_api(t1 = data_mcard_1(), c2 = b, c1 = a, data_mcard_2()),
    check_api(t1 = data_mcard_1(), c2 = b, data_mcard_2(), c1 = a),
    check_api(t1 = data_mcard_1(), data_mcard_2(), c1 = a, c2 = b),
    check_api(t1 = data_mcard_1(), data_mcard_2(), c2 = b, c1 = a),
    check_api(c1 = a, t1 = data_mcard_1(), c2 = b, data_mcard_2()),
    check_api(c1 = a, t1 = data_mcard_1(), data_mcard_2(), c2 = b),
    check_api(c1 = a, c2 = b, t1 = data_mcard_1(), data_mcard_2()),
    check_api(c1 = a, c2 = b, data_mcard_2(), t1 = data_mcard_1()),
    check_api(c1 = a, data_mcard_2(), t1 = data_mcard_1(), c2 = b),
    check_api(c1 = a, data_mcard_2(), c2 = b, t1 = data_mcard_1()),
    check_api(c2 = b, t1 = data_mcard_1(), c1 = a, data_mcard_2()),
    check_api(c2 = b, t1 = data_mcard_1(), data_mcard_2(), c1 = a),
    check_api(c2 = b, c1 = a, t1 = data_mcard_1(), data_mcard_2()),
    check_api(c2 = b, c1 = a, data_mcard_2(), t1 = data_mcard_1()),
    check_api(c2 = b, data_mcard_2(), t1 = data_mcard_1(), c1 = a),
    check_api(c2 = b, data_mcard_2(), c1 = a, t1 = data_mcard_1()),
    check_api(data_mcard_2(), t1 = data_mcard_1(), c1 = a, c2 = b),
    check_api(data_mcard_2(), t1 = data_mcard_1(), c2 = b, c1 = a),
    check_api(data_mcard_2(), c1 = a, t1 = data_mcard_1(), c2 = b),
    check_api(data_mcard_2(), c1 = a, c2 = b, t1 = data_mcard_1()),
    check_api(data_mcard_2(), c2 = b, t1 = data_mcard_1(), c1 = a),
    check_api(data_mcard_2(), c2 = b, c1 = a, t1 = data_mcard_1())
  )
  expect_same(
    check_api(t1 = data_mcard_1(), c1 = a, t2 = data_mcard_2(), b),
    check_api(t1 = data_mcard_1(), c1 = a, b, t2 = data_mcard_2()),
    check_api(t1 = data_mcard_1(), t2 = data_mcard_2(), c1 = a, b),
    check_api(t1 = data_mcard_1(), t2 = data_mcard_2(), b, c1 = a),
    check_api(t1 = data_mcard_1(), b, c1 = a, t2 = data_mcard_2()),
    check_api(t1 = data_mcard_1(), b, t2 = data_mcard_2(), c1 = a),
    check_api(c1 = a, t1 = data_mcard_1(), t2 = data_mcard_2(), b),
    check_api(c1 = a, t1 = data_mcard_1(), b, t2 = data_mcard_2()),
    check_api(c1 = a, t2 = data_mcard_2(), t1 = data_mcard_1(), b),
    check_api(c1 = a, t2 = data_mcard_2(), b, t1 = data_mcard_1()),
    check_api(c1 = a, b, t1 = data_mcard_1(), t2 = data_mcard_2()),
    check_api(c1 = a, b, t2 = data_mcard_2(), t1 = data_mcard_1()),
    check_api(t2 = data_mcard_2(), t1 = data_mcard_1(), c1 = a, b),
    check_api(t2 = data_mcard_2(), t1 = data_mcard_1(), b, c1 = a),
    check_api(t2 = data_mcard_2(), c1 = a, t1 = data_mcard_1(), b),
    check_api(t2 = data_mcard_2(), c1 = a, b, t1 = data_mcard_1()),
    check_api(t2 = data_mcard_2(), b, t1 = data_mcard_1(), c1 = a),
    check_api(t2 = data_mcard_2(), b, c1 = a, t1 = data_mcard_1()),
    check_api(b, t1 = data_mcard_1(), c1 = a, t2 = data_mcard_2()),
    check_api(b, t1 = data_mcard_1(), t2 = data_mcard_2(), c1 = a),
    check_api(b, c1 = a, t1 = data_mcard_1(), t2 = data_mcard_2()),
    check_api(b, c1 = a, t2 = data_mcard_2(), t1 = data_mcard_1()),
    check_api(b, t2 = data_mcard_2(), t1 = data_mcard_1(), c1 = a),
    check_api(b, t2 = data_mcard_2(), c1 = a, t1 = data_mcard_1())
  )
  expect_same(
    check_api(t1 = data_mcard_1(), c1 = a, t2 = data_mcard_2(), c2 = b),
    check_api(t1 = data_mcard_1(), c1 = a, c2 = b, t2 = data_mcard_2()),
    check_api(t1 = data_mcard_1(), t2 = data_mcard_2(), c1 = a, c2 = b),
    check_api(t1 = data_mcard_1(), t2 = data_mcard_2(), c2 = b, c1 = a),
    check_api(t1 = data_mcard_1(), c2 = b, c1 = a, t2 = data_mcard_2()),
    check_api(t1 = data_mcard_1(), c2 = b, t2 = data_mcard_2(), c1 = a),
    check_api(c1 = a, t1 = data_mcard_1(), t2 = data_mcard_2(), c2 = b),
    check_api(c1 = a, t1 = data_mcard_1(), c2 = b, t2 = data_mcard_2()),
    check_api(c1 = a, t2 = data_mcard_2(), t1 = data_mcard_1(), c2 = b),
    check_api(c1 = a, t2 = data_mcard_2(), c2 = b, t1 = data_mcard_1()),
    check_api(c1 = a, c2 = b, t1 = data_mcard_1(), t2 = data_mcard_2()),
    check_api(c1 = a, c2 = b, t2 = data_mcard_2(), t1 = data_mcard_1()),
    check_api(t2 = data_mcard_2(), t1 = data_mcard_1(), c1 = a, c2 = b),
    check_api(t2 = data_mcard_2(), t1 = data_mcard_1(), c2 = b, c1 = a),
    check_api(t2 = data_mcard_2(), c1 = a, t1 = data_mcard_1(), c2 = b),
    check_api(t2 = data_mcard_2(), c1 = a, c2 = b, t1 = data_mcard_1()),
    check_api(t2 = data_mcard_2(), c2 = b, t1 = data_mcard_1(), c1 = a),
    check_api(t2 = data_mcard_2(), c2 = b, c1 = a, t1 = data_mcard_1()),
    check_api(c2 = b, t1 = data_mcard_1(), c1 = a, t2 = data_mcard_2()),
    check_api(c2 = b, t1 = data_mcard_1(), t2 = data_mcard_2(), c1 = a),
    check_api(c2 = b, c1 = a, t1 = data_mcard_1(), t2 = data_mcard_2()),
    check_api(c2 = b, c1 = a, t2 = data_mcard_2(), t1 = data_mcard_1()),
    check_api(c2 = b, t2 = data_mcard_2(), t1 = data_mcard_1(), c1 = a),
    check_api(c2 = b, t2 = data_mcard_2(), c1 = a, t1 = data_mcard_1())
  )
})

test_that("check_subset() checks if tf_1$c1 column values are subset of tf_2$c2 properly?", {
  expect_silent(check_subset(data_mcard_1(), data_mcard_2(), x_select = a, y_select = a))
})

test_that("output for legacy API", {
  expect_snapshot({
    check_subset(data_mcard_1(), a, data_mcard_2(), a)
  })
})

test_that("output", {
  expect_snapshot(error = TRUE, {
    check_subset(data_mcard_1(), data_mcard_2(), x_select = c(x = a))
  })
  expect_snapshot(error = TRUE, {
    check_subset(data_mcard_2(), data_mcard_1(), x_select = a)
  })
})

test_that("output for compound keys", {
  expect_snapshot(error = TRUE, {
    check_subset(data_mcard_2(), data_mcard_1(), x_select = c(a, b))
  })
})

test_that("check_set_equality() checks properly if 2 sets of values are equal?", {
  expect_silent(check_set_equality(data_mcard_1(), data_mcard_3(), x_select = a))

  expect_snapshot(error = TRUE, {
    check_set_equality(data_mcard_1(), data_mcard_2(), x_select = c(a, c))
  })
})

# FIXME: COMPOUND:: regarding compound keys: should `check_subset()` and `check_set_equality()`
# also work for multiple columns? (matching needs to be provided, implicitly by order?)

#' Check table relations
#'
#' @description
#' All `check_cardinality_...()` functions test the following conditions:
#' 1. Are all rows in `x` unique?
#' 1. Are the rows in `y` a subset of the rows in `x`?
#' 1. Does the relation between `x` and `y` meet the cardinality requirements?
#'     One row from `x` must correspond to the requested number of rows in `y`,
#'     e.g. `_0_1` means that there must be zero or one rows in `y` for each
#'     row in `x`.
#'
#' `examine_cardinality()` also checks the first two points and subsequently determines the type of cardinality.
#'
#' For convenience, the `x_select` and `y_select` arguments allow restricting the check
#' to a set of key columns without affecting the return value.
#'
#' @details
#' All cardinality functions accept a parent and a child table (`x` and `y`).
#' All rows in `x` must be unique, and all rows in `y` must be a subset of the
#' rows in `x`.
#' The `x_select` and `y_select` arguments allow restricting the check
#' to a set of key columns without affecting the return value.
#' If given, both arguments must refer to the same number of key columns.
#'
#' The cardinality specifications "0_n", "1_n", "0_1", "1_1" refer to the expected relation that the child table has with the parent table.
#' "0", "1" and "n" refer to the occurrences of value combinations
#' in `y` that correspond to each combination in the
#' columns of the parent table.
#' "n" means "more than one" in this context, with no upper limit.
#'
# FIXME: Should/do we check that there is at least one with 0, 1 or 2 matches?
#' **"0_n"**: no restrictions, each row in `x` has at least 0 and at most
#' n corresponding occurrences in `y`.
#'
#' **"1_n"**: each row in `x` has at least 1 and at most
#' n corresponding occurrences in `y`.
#' This means that there is a "surjective" mapping from the child table
#' to the parent table, i.e. each parent table row exists at least once in the
#' child table.
#'
#' **"0_1"**: each row in `x` has at least 0 and at most
#' 1 corresponding occurrence in `y`.
#' This means that there is a "injective" mapping from the child table
#' to the parent table, i.e. no combination of values in the
#' parent table columns is addressed multiple times.
#' But not all parent table rows have to be referred to.
#'
#' **"1_1"**: each row in `x` occurs exactly once in `y`.
#' This means that there is a "bijective" ("injective" AND "surjective") mapping
#' between the child table and the parent table, i.e. the
#' sets of rows are identical.
#'
#' Finally, `examine_cardinality()` tests for and returns the nature of the relationship
#' (injective, surjective, bijective, or none of these)
#' between the two given sets of columns.
#' If either `x` is not unique or there are rows in `y` that are missing from `x`,
#' the requirements for a cardinality test is not fulfilled.
#' No error will be thrown, but
#' the result will contain the information which prerequisite was violated.
#' @param x Parent table, data frame or lazy table.
#' @param y Child table, data frame or lazy table.
#' @inheritParams rlang::args_dots_empty
#' @inheritParams check_set_equality
#'
#' @family cardinality functions
#'
#' @name examine_cardinality
#'
#' @return `check_cardinality_...()` return `x`, invisibly,
#' if the check is passed, to support pipes.
#' Otherwise an error is thrown and the reason for it is explained.
#'
#' `examine_cardinality()` returns a character variable specifying the type of relationship between the two columns.
#'
#' @aliases check_cardinality_...
#' @export
#' @examples
#' d1 <- tibble::tibble(a = 1:5)
#' d2 <- tibble::tibble(a = c(1:4, 4L))
#' d3 <- tibble::tibble(c = c(1:5, 5L), d = 0)
#' # This does not pass, `a` is not unique key of d2:
#' try(check_cardinality_0_n(d2, d1))
#'
#' # Columns are matched by name by default:
#' try(check_cardinality_0_n(d1, d3))
#'
#' # This passes, multiple values in d3$c are allowed:
#' check_cardinality_0_n(d1, d2)
#'
#' # This does not pass, injectivity is violated:
#' try(check_cardinality_1_1(d1, d3, y_select = c(a = c)))
#' try(check_cardinality_0_1(d1, d3, x_select = c(c = a)))
#'
#' # What kind of cardinality is it?
#' examine_cardinality(d1, d3, x_select = c(c = a))
#' examine_cardinality(d1, d2)
check_cardinality_0_n <- function(x, y, ..., x_select = NULL, y_select = NULL, by_position = NULL) {
  dm_local_error_call()
  check_card_api(
    {{ x }},
    {{ y }},
    ...,
    x_select = {{ x_select }},
    y_select = {{ y_select }},
    by_position = by_position,
    target = check_cardinality_0_n_impl0
  )
}

check_cardinality_0_n_impl0 <- function(x, y, x_label, y_label) {
  check_key_impl0(x, x_label)

  check_subset_impl0(y, x, y_label, x_label)

  invisible(x)
}

#' @rdname examine_cardinality
#' @export
check_cardinality_1_n <- function(x, y, ..., x_select = NULL, y_select = NULL, by_position = NULL) {
  dm_local_error_call()
  check_card_api(
    {{ x }},
    {{ y }},
    ...,
    x_select = {{ x_select }},
    y_select = {{ y_select }},
    by_position = by_position,
    target = check_cardinality_1_n_impl0
  )
}

check_cardinality_1_n_impl0 <- function(x, y, x_label, y_label) {
  check_key_impl0(x, x_label)

  check_set_equality_impl0(y, x, y_label, x_label)

  invisible(x)
}

#' @rdname examine_cardinality
#' @export
check_cardinality_1_1 <- function(x, y, ..., x_select = NULL, y_select = NULL, by_position = NULL) {
  dm_local_error_call()
  check_card_api(
    {{ x }},
    {{ y }},
    ...,
    x_select = {{ x_select }},
    y_select = {{ y_select }},
    by_position = by_position,
    target = check_cardinality_1_1_impl0
  )
}

check_cardinality_1_1_impl0 <- function(x, y, x_label, y_label) {
  check_key_impl0(x, x_label)

  check_set_equality_impl0(y, x, y_label, x_label)

  tryCatch(
    {
      check_key_impl0(y, y_label)
      NULL
    },
    error = function(e) abort_not_bijective(y_label, colnames(y))
  )

  invisible(x)
}

#' @rdname examine_cardinality
#' @export
check_cardinality_0_1 <- function(x, y, ..., x_select = NULL, y_select = NULL, by_position = NULL) {
  dm_local_error_call()
  check_card_api(
    {{ x }},
    {{ y }},
    ...,
    x_select = {{ x_select }},
    y_select = {{ y_select }},
    by_position = by_position,
    target = check_cardinality_0_1_impl0
  )
}

check_cardinality_0_1_impl0 <- function(x, y, x_label, y_label) {
  check_key_impl0(x, x_label)

  check_subset_impl0(y, x, y_label, x_label)

  tryCatch(
    {
      check_key_impl0(y, y_label)
      NULL
    },
    error = function(e) abort_not_injective(y_label, colnames(y))
  )

  invisible(x)
}

#' @rdname examine_cardinality
#' @export
examine_cardinality <- function(x, y, ..., x_select = NULL, y_select = NULL, by_position = NULL) {
  check_card_api(
    {{ x }},
    {{ y }},
    ...,
    x_select = {{ x_select }},
    y_select = {{ y_select }},
    by_position = by_position,
    target = examine_cardinality_impl0
  )
}

examine_cardinality_impl0 <- function(x, y, x_label, y_label) {
  if (!is_unique_key_se(x, colnames(x))$unique) {
    plural <- s_if_plural(colnames(x))
    return(
      glue(
        "Column{plural['n']} ({commas(tick(colnames(x)))}) not ",
        "a unique key of {tick(x_label)}."
      )
    )
  }
  if (!is_subset_se(y, x)) {
    plural <- s_if_plural(colnames(x))
    return(
      glue(
        "Column{plural['n']} ({commas(tick(colnames(y)))}) of table {tick(y_label)} not ",
        "a subset of column{plural['n']} ({commas(tick(colnames(x)))}) of table {tick(x_label)}."
      )
    )
  }
  min_1 <- is_subset_se(x, y)
  max_1 <- pull(is_unique_key_se(y, colnames(y)), unique)

  if (min_1 && max_1) {
    return("bijective mapping (child: 1 -> parent: 1)")
  } else if (min_1) {
    return("surjective mapping (child: 1 to n -> parent: 1)")
  } else if (max_1) {
    return("injective mapping (child: 0 or 1 -> parent: 1)")
  }
  "generic mapping (child: 0 to n -> parent: 1)"
}


check_card_api <- function(
  x,
  y,
  ...,
  x_select = NULL,
  y_select = NULL,
  by_position = NULL,
  call = caller_env(),
  target = list
) {
  dm_local_error_call(call)
  if (dots_n(...) >= 2) {
    name <- as.character(frame_call(call)[[1]] %||% "check_card_api")
    deprecate_warn(
      "1.0.0",
      paste0(name, "(pk_column)"),
      paste0(name, "(x_select = )"),
      details = c(
        "*" = "Use `y_select` instead of `fk_column`, and `x` and `y` instead of `parent_table` and `child_table`.",
        "*" = "Using `by_position = TRUE` for compatibility."
      )
    )
    stopifnot(is.null(by_position))
    check_card_api_impl(
      {{ x }},
      {{ y }},
      ...,
      by_position = TRUE,
      target = target
    )
  } else {
    check_dots_empty(call = call)
    check_card_api_impl(
      {{ x }},
      {{ x_select }},
      {{ y }},
      {{ y_select }},
      by_position = by_position %||% FALSE,
      target = target
    )
  }
}

check_card_api_impl <- function(
  parent_table,
  pk_column,
  child_table,
  fk_column,
  ...,
  by_position,
  target
) {
  ptq <- enquo(parent_table)
  ctq <- enquo(child_table)

  pkcq <- enquo(pk_column)
  fkcq <- enquo(fk_column)

  if (!quo_is_null(pkcq)) {
    parent_table <- parent_table %>% select(!!pkcq)
  }

  if (!quo_is_null(fkcq)) {
    child_table <- child_table %>% select(!!fkcq)
  }

  if (!isTRUE(by_position)) {
    y_idx <- match(colnames(parent_table), colnames(child_table))
    if (anyNA(y_idx)) {
      cli::cli_abort(
        "{.code by_position = FALSE} or {.code by_position = NULL} require column names in {.arg x} to match those in {.arg y}.",
        call = dm_error_call()
      )
    }

    child_table <-
      child_table %>%
      select(!!y_idx)
  }

  target(parent_table, child_table, as_label(ptq), as_label(ctq))
}

test_that("check_card_api() new interface", {
  local_options(lifecycle_verbosity = "quiet")

  expect_same(
    check_card_api(data_mcard_1(), data_mcard_2(), x_select = a, y_select = c(a = b)),
    check_card_api(x = data_mcard_1(), data_mcard_2(), x_select = a, y_select = c(a = b)),
    check_card_api(data_mcard_1(), y = data_mcard_2(), x_select = a, y_select = c(a = b)),
    check_card_api(x = data_mcard_1(), y = data_mcard_2(), x_select = a, y_select = c(a = b)),
    check_card_api(y = data_mcard_2(), x = data_mcard_1(), x_select = a, y_select = c(a = b)),
    check_card_api(data_mcard_1(), a, data_mcard_2(), c(a = b))
  )

  expect_same(
    check_card_api(data_mcard_1(), data_mcard_2(), x_select = a, y_select = b, by_position = TRUE),
    check_card_api(
      x = data_mcard_1(),
      data_mcard_2(),
      x_select = a,
      y_select = b,
      by_position = TRUE
    ),
    check_card_api(
      data_mcard_1(),
      y = data_mcard_2(),
      x_select = a,
      y_select = b,
      by_position = TRUE
    ),
    check_card_api(
      x = data_mcard_1(),
      y = data_mcard_2(),
      x_select = a,
      y_select = b,
      by_position = TRUE
    ),
    check_card_api(
      y = data_mcard_2(),
      x = data_mcard_1(),
      x_select = a,
      y_select = b,
      by_position = TRUE
    ),
    check_card_api(data_mcard_1(), a, data_mcard_2(), b)
  )
})

test_that("check_cardinality_...() functions work without `x_select` and `y_select`", {
  expect_silent(check_cardinality_0_n(data_card_1(), data_card_11()))
  expect_silent(check_cardinality_1_n(data_card_1(), data_card_12()))
  # compute(): Can't open the same table twice on MySQL
  expect_silent(check_cardinality_1_1(
    data_card_1(),
    suppress_mssql_message(compute(data_card_1()))
  ))
  expect_silent(check_cardinality_0_1(data_card_1(), data_card_11()))

  expect_snapshot({
    examine_cardinality(data_card_1(), data_card_11())
    examine_cardinality(data_card_1(), data_card_12())
    # compute(): Can't open the same table twice on MySQL
    examine_cardinality(data_card_1(), suppress_mssql_message(compute(data_card_1())))
    examine_cardinality(data_card_1(), data_card_11())
  })
})

test_that("check_card_api() compatibility", {
  local_options(lifecycle_verbosity = "quiet")

  expect_same(
    check_card_api(data_mcard_1(), a, data_mcard_2(), b),
    check_card_api(
      parent_table = data_mcard_1(),
      pk_column = a,
      child_table = data_mcard_2(),
      fk_column = b
    )
  )
  expect_same(
    check_card_api(fk_column = b, data_mcard_1(), a, data_mcard_2()),
    check_card_api(data_mcard_1(), fk_column = b, a, data_mcard_2()),
    check_card_api(data_mcard_1(), a, fk_column = b, data_mcard_2()),
    check_card_api(data_mcard_1(), a, data_mcard_2(), fk_column = b)
  )
  expect_same(
    check_card_api(child_table = data_mcard_2(), data_mcard_1(), a, b),
    check_card_api(data_mcard_1(), child_table = data_mcard_2(), a, b),
    check_card_api(data_mcard_1(), a, child_table = data_mcard_2(), b),
    check_card_api(data_mcard_1(), a, b, child_table = data_mcard_2())
  )
  expect_same(
    check_card_api(child_table = data_mcard_2(), fk_column = b, data_mcard_1(), a),
    check_card_api(child_table = data_mcard_2(), data_mcard_1(), fk_column = b, a),
    check_card_api(child_table = data_mcard_2(), data_mcard_1(), a, fk_column = b),
    check_card_api(fk_column = b, child_table = data_mcard_2(), data_mcard_1(), a),
    check_card_api(fk_column = b, data_mcard_1(), child_table = data_mcard_2(), a),
    check_card_api(fk_column = b, data_mcard_1(), a, child_table = data_mcard_2()),
    check_card_api(data_mcard_1(), child_table = data_mcard_2(), fk_column = b, a),
    check_card_api(data_mcard_1(), child_table = data_mcard_2(), a, fk_column = b),
    check_card_api(data_mcard_1(), fk_column = b, child_table = data_mcard_2(), a),
    check_card_api(data_mcard_1(), fk_column = b, a, child_table = data_mcard_2()),
    check_card_api(data_mcard_1(), a, child_table = data_mcard_2(), fk_column = b),
    check_card_api(data_mcard_1(), a, fk_column = b, child_table = data_mcard_2())
  )
  expect_same(
    check_card_api(pk_column = a, data_mcard_1(), data_mcard_2(), b),
    check_card_api(data_mcard_1(), pk_column = a, data_mcard_2(), b),
    check_card_api(data_mcard_1(), data_mcard_2(), pk_column = a, b),
    check_card_api(data_mcard_1(), data_mcard_2(), b, pk_column = a)
  )
  expect_same(
    check_card_api(pk_column = a, fk_column = b, data_mcard_1(), data_mcard_2()),
    check_card_api(pk_column = a, data_mcard_1(), fk_column = b, data_mcard_2()),
    check_card_api(pk_column = a, data_mcard_1(), data_mcard_2(), fk_column = b),
    check_card_api(fk_column = b, pk_column = a, data_mcard_1(), data_mcard_2()),
    check_card_api(fk_column = b, data_mcard_1(), pk_column = a, data_mcard_2()),
    check_card_api(fk_column = b, data_mcard_1(), data_mcard_2(), pk_column = a),
    check_card_api(data_mcard_1(), pk_column = a, fk_column = b, data_mcard_2()),
    check_card_api(data_mcard_1(), pk_column = a, data_mcard_2(), fk_column = b),
    check_card_api(data_mcard_1(), fk_column = b, pk_column = a, data_mcard_2()),
    check_card_api(data_mcard_1(), fk_column = b, data_mcard_2(), pk_column = a),
    check_card_api(data_mcard_1(), data_mcard_2(), pk_column = a, fk_column = b),
    check_card_api(data_mcard_1(), data_mcard_2(), fk_column = b, pk_column = a)
  )
  expect_same(
    check_card_api(pk_column = a, child_table = data_mcard_2(), data_mcard_1(), b),
    check_card_api(pk_column = a, data_mcard_1(), child_table = data_mcard_2(), b),
    check_card_api(pk_column = a, data_mcard_1(), b, child_table = data_mcard_2()),
    check_card_api(child_table = data_mcard_2(), pk_column = a, data_mcard_1(), b),
    check_card_api(child_table = data_mcard_2(), data_mcard_1(), pk_column = a, b),
    check_card_api(child_table = data_mcard_2(), data_mcard_1(), b, pk_column = a),
    check_card_api(data_mcard_1(), pk_column = a, child_table = data_mcard_2(), b),
    check_card_api(data_mcard_1(), pk_column = a, b, child_table = data_mcard_2()),
    check_card_api(data_mcard_1(), child_table = data_mcard_2(), pk_column = a, b),
    check_card_api(data_mcard_1(), child_table = data_mcard_2(), b, pk_column = a),
    check_card_api(data_mcard_1(), b, pk_column = a, child_table = data_mcard_2()),
    check_card_api(data_mcard_1(), b, child_table = data_mcard_2(), pk_column = a)
  )
  expect_same(
    check_card_api(pk_column = a, child_table = data_mcard_2(), fk_column = b, data_mcard_1()),
    check_card_api(pk_column = a, child_table = data_mcard_2(), data_mcard_1(), fk_column = b),
    check_card_api(pk_column = a, fk_column = b, child_table = data_mcard_2(), data_mcard_1()),
    check_card_api(pk_column = a, fk_column = b, data_mcard_1(), child_table = data_mcard_2()),
    check_card_api(pk_column = a, data_mcard_1(), child_table = data_mcard_2(), fk_column = b),
    check_card_api(pk_column = a, data_mcard_1(), fk_column = b, child_table = data_mcard_2()),
    check_card_api(child_table = data_mcard_2(), pk_column = a, fk_column = b, data_mcard_1()),
    check_card_api(child_table = data_mcard_2(), pk_column = a, data_mcard_1(), fk_column = b),
    check_card_api(child_table = data_mcard_2(), fk_column = b, pk_column = a, data_mcard_1()),
    check_card_api(child_table = data_mcard_2(), fk_column = b, data_mcard_1(), pk_column = a),
    check_card_api(child_table = data_mcard_2(), data_mcard_1(), pk_column = a, fk_column = b),
    check_card_api(child_table = data_mcard_2(), data_mcard_1(), fk_column = b, pk_column = a),
    check_card_api(fk_column = b, pk_column = a, child_table = data_mcard_2(), data_mcard_1()),
    check_card_api(fk_column = b, pk_column = a, data_mcard_1(), child_table = data_mcard_2()),
    check_card_api(fk_column = b, child_table = data_mcard_2(), pk_column = a, data_mcard_1()),
    check_card_api(fk_column = b, child_table = data_mcard_2(), data_mcard_1(), pk_column = a),
    check_card_api(fk_column = b, data_mcard_1(), pk_column = a, child_table = data_mcard_2()),
    check_card_api(fk_column = b, data_mcard_1(), child_table = data_mcard_2(), pk_column = a),
    check_card_api(data_mcard_1(), pk_column = a, child_table = data_mcard_2(), fk_column = b),
    check_card_api(data_mcard_1(), pk_column = a, fk_column = b, child_table = data_mcard_2()),
    check_card_api(data_mcard_1(), child_table = data_mcard_2(), pk_column = a, fk_column = b),
    check_card_api(data_mcard_1(), child_table = data_mcard_2(), fk_column = b, pk_column = a),
    check_card_api(data_mcard_1(), fk_column = b, pk_column = a, child_table = data_mcard_2()),
    check_card_api(data_mcard_1(), fk_column = b, child_table = data_mcard_2(), pk_column = a)
  )
  expect_same(
    check_card_api(parent_table = data_mcard_1(), a, data_mcard_2(), b),
    check_card_api(a, parent_table = data_mcard_1(), data_mcard_2(), b),
    check_card_api(a, data_mcard_2(), parent_table = data_mcard_1(), b),
    check_card_api(a, data_mcard_2(), b, parent_table = data_mcard_1())
  )
  expect_same(
    check_card_api(parent_table = data_mcard_1(), fk_column = b, a, data_mcard_2()),
    check_card_api(parent_table = data_mcard_1(), a, fk_column = b, data_mcard_2()),
    check_card_api(parent_table = data_mcard_1(), a, data_mcard_2(), fk_column = b),
    check_card_api(fk_column = b, parent_table = data_mcard_1(), a, data_mcard_2()),
    check_card_api(fk_column = b, a, parent_table = data_mcard_1(), data_mcard_2()),
    check_card_api(fk_column = b, a, data_mcard_2(), parent_table = data_mcard_1()),
    check_card_api(a, parent_table = data_mcard_1(), fk_column = b, data_mcard_2()),
    check_card_api(a, parent_table = data_mcard_1(), data_mcard_2(), fk_column = b),
    check_card_api(a, fk_column = b, parent_table = data_mcard_1(), data_mcard_2()),
    check_card_api(a, fk_column = b, data_mcard_2(), parent_table = data_mcard_1()),
    check_card_api(a, data_mcard_2(), parent_table = data_mcard_1(), fk_column = b),
    check_card_api(a, data_mcard_2(), fk_column = b, parent_table = data_mcard_1())
  )
  expect_same(
    check_card_api(parent_table = data_mcard_1(), child_table = data_mcard_2(), a, b),
    check_card_api(parent_table = data_mcard_1(), a, child_table = data_mcard_2(), b),
    check_card_api(parent_table = data_mcard_1(), a, b, child_table = data_mcard_2()),
    check_card_api(child_table = data_mcard_2(), parent_table = data_mcard_1(), a, b),
    check_card_api(child_table = data_mcard_2(), a, parent_table = data_mcard_1(), b),
    check_card_api(child_table = data_mcard_2(), a, b, parent_table = data_mcard_1()),
    check_card_api(a, parent_table = data_mcard_1(), child_table = data_mcard_2(), b),
    check_card_api(a, parent_table = data_mcard_1(), b, child_table = data_mcard_2()),
    check_card_api(a, child_table = data_mcard_2(), parent_table = data_mcard_1(), b),
    check_card_api(a, child_table = data_mcard_2(), b, parent_table = data_mcard_1()),
    check_card_api(a, b, parent_table = data_mcard_1(), child_table = data_mcard_2()),
    check_card_api(a, b, child_table = data_mcard_2(), parent_table = data_mcard_1())
  )
  expect_same(
    check_card_api(parent_table = data_mcard_1(), child_table = data_mcard_2(), fk_column = b, a),
    check_card_api(parent_table = data_mcard_1(), child_table = data_mcard_2(), a, fk_column = b),
    check_card_api(parent_table = data_mcard_1(), fk_column = b, child_table = data_mcard_2(), a),
    check_card_api(parent_table = data_mcard_1(), fk_column = b, a, child_table = data_mcard_2()),
    check_card_api(parent_table = data_mcard_1(), a, child_table = data_mcard_2(), fk_column = b),
    check_card_api(parent_table = data_mcard_1(), a, fk_column = b, child_table = data_mcard_2()),
    check_card_api(child_table = data_mcard_2(), parent_table = data_mcard_1(), fk_column = b, a),
    check_card_api(child_table = data_mcard_2(), parent_table = data_mcard_1(), a, fk_column = b),
    check_card_api(child_table = data_mcard_2(), fk_column = b, parent_table = data_mcard_1(), a),
    check_card_api(child_table = data_mcard_2(), fk_column = b, a, parent_table = data_mcard_1()),
    check_card_api(child_table = data_mcard_2(), a, parent_table = data_mcard_1(), fk_column = b),
    check_card_api(child_table = data_mcard_2(), a, fk_column = b, parent_table = data_mcard_1()),
    check_card_api(fk_column = b, parent_table = data_mcard_1(), child_table = data_mcard_2(), a),
    check_card_api(fk_column = b, parent_table = data_mcard_1(), a, child_table = data_mcard_2()),
    check_card_api(fk_column = b, child_table = data_mcard_2(), parent_table = data_mcard_1(), a),
    check_card_api(fk_column = b, child_table = data_mcard_2(), a, parent_table = data_mcard_1()),
    check_card_api(fk_column = b, a, parent_table = data_mcard_1(), child_table = data_mcard_2()),
    check_card_api(fk_column = b, a, child_table = data_mcard_2(), parent_table = data_mcard_1()),
    check_card_api(a, parent_table = data_mcard_1(), child_table = data_mcard_2(), fk_column = b),
    check_card_api(a, parent_table = data_mcard_1(), fk_column = b, child_table = data_mcard_2()),
    check_card_api(a, child_table = data_mcard_2(), parent_table = data_mcard_1(), fk_column = b),
    check_card_api(a, child_table = data_mcard_2(), fk_column = b, parent_table = data_mcard_1()),
    check_card_api(a, fk_column = b, parent_table = data_mcard_1(), child_table = data_mcard_2()),
    check_card_api(a, fk_column = b, child_table = data_mcard_2(), parent_table = data_mcard_1())
  )
  expect_same(
    check_card_api(parent_table = data_mcard_1(), pk_column = a, data_mcard_2(), b),
    check_card_api(parent_table = data_mcard_1(), data_mcard_2(), pk_column = a, b),
    check_card_api(parent_table = data_mcard_1(), data_mcard_2(), b, pk_column = a),
    check_card_api(pk_column = a, parent_table = data_mcard_1(), data_mcard_2(), b),
    check_card_api(pk_column = a, data_mcard_2(), parent_table = data_mcard_1(), b),
    check_card_api(pk_column = a, data_mcard_2(), b, parent_table = data_mcard_1()),
    check_card_api(data_mcard_2(), parent_table = data_mcard_1(), pk_column = a, b),
    check_card_api(data_mcard_2(), parent_table = data_mcard_1(), b, pk_column = a),
    check_card_api(data_mcard_2(), pk_column = a, parent_table = data_mcard_1(), b),
    check_card_api(data_mcard_2(), pk_column = a, b, parent_table = data_mcard_1()),
    check_card_api(data_mcard_2(), b, parent_table = data_mcard_1(), pk_column = a),
    check_card_api(data_mcard_2(), b, pk_column = a, parent_table = data_mcard_1())
  )
  expect_same(
    check_card_api(parent_table = data_mcard_1(), pk_column = a, fk_column = b, data_mcard_2()),
    check_card_api(parent_table = data_mcard_1(), pk_column = a, data_mcard_2(), fk_column = b),
    check_card_api(parent_table = data_mcard_1(), fk_column = b, pk_column = a, data_mcard_2()),
    check_card_api(parent_table = data_mcard_1(), fk_column = b, data_mcard_2(), pk_column = a),
    check_card_api(parent_table = data_mcard_1(), data_mcard_2(), pk_column = a, fk_column = b),
    check_card_api(parent_table = data_mcard_1(), data_mcard_2(), fk_column = b, pk_column = a),
    check_card_api(pk_column = a, parent_table = data_mcard_1(), fk_column = b, data_mcard_2()),
    check_card_api(pk_column = a, parent_table = data_mcard_1(), data_mcard_2(), fk_column = b),
    check_card_api(pk_column = a, fk_column = b, parent_table = data_mcard_1(), data_mcard_2()),
    check_card_api(pk_column = a, fk_column = b, data_mcard_2(), parent_table = data_mcard_1()),
    check_card_api(pk_column = a, data_mcard_2(), parent_table = data_mcard_1(), fk_column = b),
    check_card_api(pk_column = a, data_mcard_2(), fk_column = b, parent_table = data_mcard_1()),
    check_card_api(fk_column = b, parent_table = data_mcard_1(), pk_column = a, data_mcard_2()),
    check_card_api(fk_column = b, parent_table = data_mcard_1(), data_mcard_2(), pk_column = a),
    check_card_api(fk_column = b, pk_column = a, parent_table = data_mcard_1(), data_mcard_2()),
    check_card_api(fk_column = b, pk_column = a, data_mcard_2(), parent_table = data_mcard_1()),
    check_card_api(fk_column = b, data_mcard_2(), parent_table = data_mcard_1(), pk_column = a),
    check_card_api(fk_column = b, data_mcard_2(), pk_column = a, parent_table = data_mcard_1()),
    check_card_api(data_mcard_2(), parent_table = data_mcard_1(), pk_column = a, fk_column = b),
    check_card_api(data_mcard_2(), parent_table = data_mcard_1(), fk_column = b, pk_column = a),
    check_card_api(data_mcard_2(), pk_column = a, parent_table = data_mcard_1(), fk_column = b),
    check_card_api(data_mcard_2(), pk_column = a, fk_column = b, parent_table = data_mcard_1()),
    check_card_api(data_mcard_2(), fk_column = b, parent_table = data_mcard_1(), pk_column = a),
    check_card_api(data_mcard_2(), fk_column = b, pk_column = a, parent_table = data_mcard_1())
  )
  expect_same(
    check_card_api(parent_table = data_mcard_1(), pk_column = a, child_table = data_mcard_2(), b),
    check_card_api(parent_table = data_mcard_1(), pk_column = a, b, child_table = data_mcard_2()),
    check_card_api(parent_table = data_mcard_1(), child_table = data_mcard_2(), pk_column = a, b),
    check_card_api(parent_table = data_mcard_1(), child_table = data_mcard_2(), b, pk_column = a),
    check_card_api(parent_table = data_mcard_1(), b, pk_column = a, child_table = data_mcard_2()),
    check_card_api(parent_table = data_mcard_1(), b, child_table = data_mcard_2(), pk_column = a),
    check_card_api(pk_column = a, parent_table = data_mcard_1(), child_table = data_mcard_2(), b),
    check_card_api(pk_column = a, parent_table = data_mcard_1(), b, child_table = data_mcard_2()),
    check_card_api(pk_column = a, child_table = data_mcard_2(), parent_table = data_mcard_1(), b),
    check_card_api(pk_column = a, child_table = data_mcard_2(), b, parent_table = data_mcard_1()),
    check_card_api(pk_column = a, b, parent_table = data_mcard_1(), child_table = data_mcard_2()),
    check_card_api(pk_column = a, b, child_table = data_mcard_2(), parent_table = data_mcard_1()),
    check_card_api(child_table = data_mcard_2(), parent_table = data_mcard_1(), pk_column = a, b),
    check_card_api(child_table = data_mcard_2(), parent_table = data_mcard_1(), b, pk_column = a),
    check_card_api(child_table = data_mcard_2(), pk_column = a, parent_table = data_mcard_1(), b),
    check_card_api(child_table = data_mcard_2(), pk_column = a, b, parent_table = data_mcard_1()),
    check_card_api(child_table = data_mcard_2(), b, parent_table = data_mcard_1(), pk_column = a),
    check_card_api(child_table = data_mcard_2(), b, pk_column = a, parent_table = data_mcard_1()),
    check_card_api(b, parent_table = data_mcard_1(), pk_column = a, child_table = data_mcard_2()),
    check_card_api(b, parent_table = data_mcard_1(), child_table = data_mcard_2(), pk_column = a),
    check_card_api(b, pk_column = a, parent_table = data_mcard_1(), child_table = data_mcard_2()),
    check_card_api(b, pk_column = a, child_table = data_mcard_2(), parent_table = data_mcard_1()),
    check_card_api(b, child_table = data_mcard_2(), parent_table = data_mcard_1(), pk_column = a),
    check_card_api(b, child_table = data_mcard_2(), pk_column = a, parent_table = data_mcard_1())
  )
  expect_same(
    check_card_api(
      parent_table = data_mcard_1(),
      pk_column = a,
      child_table = data_mcard_2(),
      fk_column = b
    ),
    check_card_api(
      parent_table = data_mcard_1(),
      pk_column = a,
      fk_column = b,
      child_table = data_mcard_2()
    ),
    check_card_api(
      parent_table = data_mcard_1(),
      child_table = data_mcard_2(),
      pk_column = a,
      fk_column = b
    ),
    check_card_api(
      parent_table = data_mcard_1(),
      child_table = data_mcard_2(),
      fk_column = b,
      pk_column = a
    ),
    check_card_api(
      parent_table = data_mcard_1(),
      fk_column = b,
      pk_column = a,
      child_table = data_mcard_2()
    ),
    check_card_api(
      parent_table = data_mcard_1(),
      fk_column = b,
      child_table = data_mcard_2(),
      pk_column = a
    ),
    check_card_api(
      pk_column = a,
      parent_table = data_mcard_1(),
      child_table = data_mcard_2(),
      fk_column = b
    ),
    check_card_api(
      pk_column = a,
      parent_table = data_mcard_1(),
      fk_column = b,
      child_table = data_mcard_2()
    ),
    check_card_api(
      pk_column = a,
      child_table = data_mcard_2(),
      parent_table = data_mcard_1(),
      fk_column = b
    ),
    check_card_api(
      pk_column = a,
      child_table = data_mcard_2(),
      fk_column = b,
      parent_table = data_mcard_1()
    ),
    check_card_api(
      pk_column = a,
      fk_column = b,
      parent_table = data_mcard_1(),
      child_table = data_mcard_2()
    ),
    check_card_api(
      pk_column = a,
      fk_column = b,
      child_table = data_mcard_2(),
      parent_table = data_mcard_1()
    ),
    check_card_api(
      child_table = data_mcard_2(),
      parent_table = data_mcard_1(),
      pk_column = a,
      fk_column = b
    ),
    check_card_api(
      child_table = data_mcard_2(),
      parent_table = data_mcard_1(),
      fk_column = b,
      pk_column = a
    ),
    check_card_api(
      child_table = data_mcard_2(),
      pk_column = a,
      parent_table = data_mcard_1(),
      fk_column = b
    ),
    check_card_api(
      child_table = data_mcard_2(),
      pk_column = a,
      fk_column = b,
      parent_table = data_mcard_1()
    ),
    check_card_api(
      child_table = data_mcard_2(),
      fk_column = b,
      parent_table = data_mcard_1(),
      pk_column = a
    ),
    check_card_api(
      child_table = data_mcard_2(),
      fk_column = b,
      pk_column = a,
      parent_table = data_mcard_1()
    ),
    check_card_api(
      fk_column = b,
      parent_table = data_mcard_1(),
      pk_column = a,
      child_table = data_mcard_2()
    ),
    check_card_api(
      fk_column = b,
      parent_table = data_mcard_1(),
      child_table = data_mcard_2(),
      pk_column = a
    ),
    check_card_api(
      fk_column = b,
      pk_column = a,
      parent_table = data_mcard_1(),
      child_table = data_mcard_2()
    ),
    check_card_api(
      fk_column = b,
      pk_column = a,
      child_table = data_mcard_2(),
      parent_table = data_mcard_1()
    ),
    check_card_api(
      fk_column = b,
      child_table = data_mcard_2(),
      parent_table = data_mcard_1(),
      pk_column = a
    ),
    check_card_api(
      fk_column = b,
      child_table = data_mcard_2(),
      pk_column = a,
      parent_table = data_mcard_1()
    )
  )
})

test_that("check_cardinality_...() API errors", {
  skip_if_src_not("df")

  expect_snapshot({
    check_card_api(data_mcard_1(), a, data_mcard_2(), b)
  })
  expect_snapshot(error = TRUE, {
    check_card_api(data_mcard_1(), data_mcard_2(), x_select = a, y_select = c)
  })
})

test_that("check_cardinality_...() functions are checking the cardinality correctly?", {
  #  expecting silent: ------------------------------------------------------

  expect_silent(check_cardinality_0_n(
    data_card_1(),
    data_card_3(),
    x_select = a,
    y_select = c(a = c)
  ))

  expect_silent(check_cardinality_1_n(
    data_card_1(),
    data_card_3(),
    x_select = a,
    y_select = c(a = c)
  ))

  expect_silent(check_cardinality_1_1(
    data_card_1(),
    data_card_3(),
    x_select = a,
    y_select = c(a = c)
  ))

  expect_silent(check_set_equality(data_card_1(), data_card_3(), x_select = a, y_select = c(a = c)))

  expect_silent(check_cardinality_0_n(
    data_card_5(),
    data_card_4(),
    x_select = a,
    y_select = c(a = c)
  ))

  expect_silent(check_cardinality_0_1(
    data_card_5(),
    data_card_6(),
    x_select = a,
    y_select = c(a = c)
  ))

  # scenarios for examine_cardinality() -------------------------------------

  expect_identical(
    examine_cardinality(data_card_8(), data_card_2(), x_select = c, y_select = c(c = a)),
    "injective mapping (child: 0 or 1 -> parent: 1)"
  )

  expect_identical(
    examine_cardinality(data_card_5(), data_card_4(), x_select = a, y_select = c(a = c)),
    "surjective mapping (child: 1 to n -> parent: 1)"
  )

  expect_identical(
    examine_cardinality(data_card_8(), data_card_4(), x_select = c, y_select = c),
    "generic mapping (child: 0 to n -> parent: 1)"
  )

  expect_identical(
    examine_cardinality(data_card_1(), data_card_3(), x_select = a, y_select = c(a = c)),
    "bijective mapping (child: 1 -> parent: 1)"
  )

  # expect specific errors and sometimes specific output due to errors ---------------

  expect_snapshot({
    expect_dm_error(
      check_cardinality_0_n(
        data_card_1(),
        data_card_2(),
        x_select = a,
        y_select = a
      ),
      class = "not_subset_of"
    )

    expect_dm_error(
      check_cardinality_1_1(data_card_5(), data_card_4(), x_select = a, y_select = c(a = c)),
      class = "not_bijective"
    )

    expect_dm_error(
      check_cardinality_1_1(data_card_4(), data_card_5(), x_select = c, y_select = c(c = a)),
      class = "not_unique_key"
    )

    expect_dm_error(
      check_cardinality_1_1(data_card_4(), data_card_1(), x_select = c, y_select = c(c = a)),
      class = "not_unique_key"
    )

    expect_dm_error(
      check_cardinality_0_1(data_card_1(), data_card_4(), x_select = a, y_select = c(a = c)),
      class = "not_injective"
    )

    expect_dm_error(
      check_cardinality_0_n(data_card_4(), data_card_1(), x_select = c, y_select = c(c = a)),
      class = "not_unique_key"
    )

    expect_dm_error(
      check_cardinality_1_1(data_card_4(), data_card_1(), x_select = c, y_select = c(c = a)),
      class = "not_unique_key"
    )

    expect_dm_error(
      check_cardinality_1_1(data_card_1(), data_card_4(), x_select = a, y_select = c(a = c)),
      class = "not_bijective"
    )
  })
})


# tests for compound keys -------------------------------------------------

test_that("check_cardinality_...() functions are supporting compound keys", {
  # successes
  expect_silent(check_cardinality_0_n(
    data_card_1(),
    data_card_11(),
    x_select = c(a, b),
    y_select = c(a, b)
  ))
  expect_silent(check_cardinality_1_n(
    data_card_1(),
    data_card_12(),
    x_select = c(a, b),
    y_select = c(a, b)
  ))
  # compute(): Can't open the same table twice on MySQL
  expect_silent(check_cardinality_1_1(
    data_card_1(),
    suppress_mssql_message(compute(data_card_1())),
    x_select = c(a, b),
    y_select = c(a, b)
  ))
  expect_silent(check_set_equality(
    data_card_12(),
    data_card_1(),
    x_select = c(a, b),
    y_select = c(a, b)
  ))
  expect_silent(check_cardinality_0_1(
    data_card_1(),
    data_card_11(),
    x_select = c(a, b),
    y_select = c(a, b)
  ))

  # scenarios for examine_cardinality() -------------------------------------

  expect_identical(
    examine_cardinality(data_card_1(), data_card_11(), x_select = c(a, b), y_select = c(a, b)),
    "injective mapping (child: 0 or 1 -> parent: 1)"
  )

  expect_identical(
    examine_cardinality(data_card_1(), data_card_12(), x_select = c(a, b), y_select = c(a, b)),
    "surjective mapping (child: 1 to n -> parent: 1)"
  )

  expect_identical(
    examine_cardinality(data_card_13(), data_card_12(), x_select = c(b, a), y_select = c(b, a)),
    "generic mapping (child: 0 to n -> parent: 1)"
  )

  expect_identical(
    # compute(): Can't open the same table twice on MySQL
    examine_cardinality(
      data_card_1(),
      suppress_mssql_message(compute(data_card_1())),
      x_select = c(b, a),
      y_select = c(b, a)
    ),
    "bijective mapping (child: 1 -> parent: 1)"
  )

  # expect specific errors and sometimes specific output due to errors ---------------

  expect_snapshot({
    expect_dm_error(
      check_cardinality_0_n(
        data_card_1(),
        data_card_2(),
        x_select = c(a, b),
        y_select = c(a, b)
      ),
      class = "not_subset_of"
    )

    expect_dm_error(
      check_cardinality_1_1(data_card_1(), data_card_12(), x_select = c(a, b), y_select = c(a, b)),
      class = "not_bijective"
    )

    expect_dm_error(
      check_cardinality_1_1(data_card_12(), data_card_1(), x_select = c(a, b), y_select = c(a, b)),
      class = "not_unique_key"
    )

    expect_dm_error(
      check_cardinality_0_1(data_card_1(), data_card_12(), x_select = c(b, a), y_select = c(b, a)),
      class = "not_injective"
    )
  })
})

new_keys <- function(x = list()) {
  # both c("a", "b") and list("a", "b") is accepted
  if (!is.list(x)) {
    x <- as.list(as.character(x))
  }
  new_list_of(x, character(), class = "dm_keys")
}

#' @export
vec_ptype2.dm_keys.dm_keys <- function(x, y, ...) new_keys()

#' @export
vec_cast.dm_keys.dm_keys <- function(x, to, ...) x

#' @export
vec_ptype_abbr.dm_keys <- function(x, ..., prefix_named, suffix_shape) {
  "keys"
}

#' @export
vec_proxy_compare.dm_keys <- function(x, ...) {
  # Not called: https://github.com/r-lib/vctrs/issues/1373
  x_raw <- vec_data(x)

  if (length(x_raw) > 0) {
    # First figure out the maximum length
    n <- max(lengths(x_raw))

    # Then expand all vectors to this length by filling in with zeros
    full <- map(x_raw, function(x) c(x, rep("", n - length(x))))
  } else {
    full <- list()
  }

  # Then turn into a data frame
  as.data.frame(do.call(rbind, full))
}

pillar_shaft.dm_keys <- function(x, ...) {
  x <- map_chr(x, commas, max_commas = 3)
  pillar::pillar_shaft(x)
}
on_load({
  s3_register("pillar::pillar_shaft", "dm_keys", pillar_shaft.dm_keys)
})

#' @export
format.dm_keys <- function(x, ...) {
  map_chr(x, commas, max_commas = Inf)
}

get_key_cols <- function(x) {
  stopifnot(length(x) == 1)
  x[[1]]
}

# Infrastructure for "strict mode" of a dm
# where re-adding existing keys gives an error
#
# Idea: "strict mode" is set during construction of a dm
#
# For now, strict mode only, easier to relax later
# than to strengthen after the fact

dm_is_strict_keys <- function(dm) {
  TRUE
}

#' Add info about a dm's tables
#'
#' @description
#' When creating a diagram from a `dm` using [dm_draw()] the table descriptions set with `dm_set_table_description()` will be displayed.
#'
#' @inheritParams dm_draw
#' @param ...
#' For `dm_set_table_description()`: Descriptions for tables to set in the form `description = table`.
#' `tidyselect` is supported, see [dplyr::select()] for details on the semantics.
#'
#' For `dm_get_table_description()` and `dm_reset_table_description()`: These dots are for future extensions and must be empty.
#'
#' @details
#' Multi-line descriptions can be achieved using the newline symbol `\n`.
#' Descriptions are set with `dm_set_table_description()`.
#' The currently set descriptions can be checked using `dm_get_table_description()`.
#' Descriptions can be removed using `dm_reset_table_description()`.
#'
#' @return For `dm_set_table_description()`: A `dm` object containing descriptions for specified tables.
#' @export
#'
#' @examplesIf rlang::is_installed(c("nycflights13", "labelled", "DiagrammeR"))
#' desc_flights <- rlang::set_names(
#'   "flights",
#'   paste(
#'     "On-time data for all flights",
#'     "that departed NYC (i.e. JFK, LGA or EWR) in 2013.",
#'     sep = "\n"
#'   )
#' )
#' nyc_desc <- dm_nycflights13() %>%
#'   dm_set_table_description(
#'     !!desc_flights,
#'     "Weather at the airport of\norigin at time of departure" = weather
#'   )
#' nyc_desc %>%
#'   dm_draw()
#'
#' dm_get_table_description(nyc_desc)
#' dm_reset_table_description(nyc_desc, flights) %>%
#'   dm_draw(font_size = c(header = 18L, table_description = 9L, column = 15L))
#'
#' pull_tbl(nyc_desc, flights) %>%
#'   labelled::label_attribute()
dm_set_table_description <- function(dm, ...) {
  dm_local_error_call()
  check_not_zoomed(dm)

  check_suggested("labelled (>= 2.12.0)", "dm_set_table_description")

  def <- dm_get_def(dm, quiet = TRUE)
  selected <- eval_select_indices(quo(c(...)), src_tbls_impl(dm))
  labels <- names(selected)

  out <- dm_set_table_description_impl(def, selected, labels)
  dm_from_def(out)
}

dm_set_table_description_impl <- function(def, selected, labels) {
  reduce2(
    selected,
    labels,
    function(def, table, desc) {
      labelled::label_attribute(def$data[[table]]) <- desc
      def
    },
    .init = def
  )
}


#' @rdname dm_set_table_description
#'
#' @return For `dm_get_table_description`: A named vector of tables, with the descriptions in the names.
#'
#' @export
dm_get_table_description <- function(dm, table = NULL, ...) {
  check_dots_empty()
  check_not_zoomed(dm)

  check_suggested("labelled (>= 2.12.0)", "dm_get_table_description")

  table_expr <- enexpr(table) %||% src_tbls_impl(dm, quiet = TRUE)
  tables <- eval_select_indices(table_expr, set_names(src_tbls_impl(dm, quiet = TRUE)))

  dm_get_table_description_impl(dm, tables)
}

dm_get_table_description_impl <- function(dm, tables) {
  # FIXME: Is this correct?
  if (!is_installed("labelled")) {
    return(set_names(character()))
  }

  def <- dm_get_def(dm, quiet = TRUE)
  map(
    tables,
    ~ labelled::label_attribute(def$data[[.x]])
  ) %>%
    purrr::discard(is.null) %>%
    prep_recode()
}

#' @inheritParams dm_get_all_pks
#' @rdname dm_set_table_description
#' @param table One or more table names, unquoted, for which to
#'
#' 1. get information about the current description(s) with [dm_get_table_description()].
#' 2. remove descriptions with [dm_reset_table_description()].
#'
#' In both cases the default applies to all tables in the `dm`.
#'
#' @return For `dm_reset_table_description()`: A `dm` object without descriptions for specified tables.
#' @export
dm_reset_table_description <- function(dm, table = NULL, ...) {
  check_dots_empty()
  check_not_zoomed(dm)

  check_suggested("labelled (>= 2.12.0)", "dm_reset_table_description")

  table_expr <- enexpr(table) %||% src_tbls_impl(dm, quiet = TRUE)
  def <- dm_get_def(dm, quiet = TRUE)
  tables <- eval_select_indices(table_expr, set_names(src_tbls_impl(dm, quiet = TRUE)))
  labels <- rep(list(NULL), length(tables))

  out <- dm_set_table_description_impl(def, tables, labels)
  dm_from_def(out)
}

test_that("table_description works", {
  expect_identical(
    dm_nycflights_small() %>%
      dm_set_table_description("high in the sky\nflying from NY" = flights) %>%
      dm_get_table_description(),
    set_names("flights", "high in the sky\nflying from NY")
  )
  expect_identical(
    dm_nycflights_small() %>%
      dm_set_table_description("high in the sky\nflying from NY" = flights) %>%
      dm_get_table_description(flights),
    set_names("flights", "high in the sky\nflying from NY")
  )
  expect_identical(
    dm_nycflights_small() %>%
      dm_set_table_description("high in the sky\nflying from NY" = flights) %>%
      dm_get_table_description(planes),
    set_names(character())
  )
  expect_identical(
    dm_nycflights_small() %>%
      dm_set_table_description(
        "high in the sky\nflying from NY" = flights,
        "Flugzeuge" = planes
      ) %>%
      dm_get_table_description(planes),
    set_names("planes", "Flugzeuge")
  )

  table_desc <- dm_nycflights_small() %>%
    dm_set_table_description(
      "high in the sky\nflying from NY" = flights,
      "Flugzeuge" = planes
    ) %>%
    dm_get_table_description()
  expect_identical(
    dm_nycflights_small() %>%
      dm_set_table_description(!!!table_desc) %>%
      dm_get_table_description(),
    set_names(
      c("flights", "planes"),
      c("high in the sky\nflying from NY", "Flugzeuge")
    )
  )

  expect_identical(
    dm_nycflights_small() %>%
      dm_set_table_description(!!!table_desc) %>%
      dm_reset_table_description(flights) %>%
      dm_get_table_description(),
    set_names("planes", "Flugzeuge")
  )

  expect_identical(
    dm_nycflights_small() %>%
      dm_set_table_description(!!!table_desc) %>%
      dm_reset_table_description() %>%
      dm_get_table_description(),
    set_names(character())
  )

  expect_identical(
    dm_nycflights13(table_description = TRUE) %>%
      dm_get_table_description(),
    rlang::set_names(
      c("airlines", "airports", "flights", "planes", "weather"),
      c(
        "Airlines that operated the flights",
        "Airports of origin or destination of the flights",
        paste(
          "On-time data for all flights that",
          "departed NYC (i.e. JFK, LGA or EWR) in 2013",
          sep = "\n"
        ),
        "Planes used for the flights",
        "Hourly meteorological data for LGA, JFK and EWR in 2013"
      )
    )
  )
})

library(conflicted)
library(dm)
library(tidyverse)

dm <- dm_nycflights13(cycle = TRUE)
dm
dm_deconstruct(dm)

airlines <- pull_tbl(dm, "airlines", keyed = TRUE)
airports <- pull_tbl(dm, "airports", keyed = TRUE)
flights <- pull_tbl(dm, "flights", keyed = TRUE)
planes <- pull_tbl(dm, "planes", keyed = TRUE)
weather <- pull_tbl(dm, "weather", keyed = TRUE)

dm(airlines, airports, flights, planes, weather)

by_origin <-
  flights %>%
  group_by(origin) %>%
  summarize(n = n(), mean_arr_delay = mean(arr_delay)) %>%
  ungroup()

dm(airlines, airports, flights, planes, weather, by_origin) %>%
  dm_draw()

nest_join_dm_zoomed <- function(x, ...) {
  dm_zoomed <- x
  src_dm <- dm_get_src_impl(dm_zoomed)
  if (!is.null(src_dm)) {
    abort_only_for_local_src(src_dm)
  }

  vars <- src_tbls_impl(dm_zoomed)
  selected <- eval_select_table(quo(c(...)), vars)
  if (is_empty(selected)) {
    selected <- vars
  }

  orig_table <- orig_name_zoomed(dm_zoomed)
  if (!dm_has_pk_impl(dm_zoomed, orig_table)) {
    message(
      "The originally zoomed table didn't have a primary key, therefore `nest.dm_zoomed()` does nothing."
    )
    return(dm_zoomed)
  }

  orig_pk <- dm_get_pk_impl(dm_zoomed, orig_table)
  keys <- col_tracker_zoomed(dm_zoomed)
  if (!(orig_pk %in% keys)) {
    abort_pk_not_tracked(orig_table, orig_pk)
  }
  new_pk <- names(keys[keys == orig_pk])

  child_tables <-
    get_orig_in_fks(dm_zoomed, orig_table) %>%
    mutate(data = map(child_table, ~ dm_get_tables_impl(dm_zoomed)[[.x]])) %>%
    # FIXME: should we check and warn/message, if no child table is in selected?
    filter(child_table %in% selected) %>%
    # perform joins in the order given in the ellipsis
    arrange(match(child_table, selected))
  x <- tbl_zoomed(dm_zoomed)

  for (i in seq_len(nrow(child_tables))) {
    x <- nest_join(
      x,
      y = child_tables$data[[i]],
      by = set_names(child_tables$child_fk_cols[i], new_pk),
      name = child_tables$child_table[i]
    ) %>%
      # FIXME: why does `nest_join()` not produce a `list_of`?
      mutate(!!child_tables$child_table[i] := as_list_of(!!sym(child_tables$child_table[i])))
  }
  replace_zoomed_tbl(dm_zoomed, x)
}

# Run at installation time
boilerplate <- new_dm_def()

#' Validator
#'
#' `dm_validate()` checks the internal consistency of a `dm` object.
#'
#' In theory, with the exception of [new_dm()], all `dm` objects
#' created or modified by functions in this package should be valid,
#' and this function should not be needed.
#' Please file an issue if any dm operation creates an invalid object.
#'
#' @param x An object.
#'
#' @return Returns the `dm`, invisibly, after finishing all checks.
#'
#' @export
#' @examples
#' dm_validate(dm())
#'
#' bad_dm <- structure(list(bad = "dm"), class = "dm")
#' try(dm_validate(bad_dm))
#' @autoglobal
dm_validate <- function(x) {
  dm_local_error_call()
  check_dm(x)

  if (!identical(names(unclass(x)), "def")) {
    abort_dm_invalid("A `dm` needs to be a list of one item named `def`.")
  }

  def <- dm_get_def(x)

  table_names <- def$table
  if (any(table_names == "")) {
    abort_dm_invalid("Not all tables are named.")
  }

  check_df_structure(def, boilerplate, "dm definition")

  if (!all(map_lgl(def$data, ~ inherits(., "data.frame") || inherits(., "tbl_sql")))) {
    abort_dm_invalid(
      "Not all entries in `def$data` are of class `data.frame` or `tbl_dbi`. Check `dm_get_tables()`."
    )
  }
  if (!all_same_source(def$data)) {
    abort_dm_invalid("Not all tables in the object share the same `src`.")
  }

  check_df_structure(c_list_of(def$pks), c_list_of(boilerplate$pks), "`pks` column")
  check_df_structure(c_list_of(def$fks), c_list_of(boilerplate$fks), "`fks` column")
  check_df_structure(c_list_of(def$filters), c_list_of(boilerplate$filters), "`filters` column")

  dm_col_names <- set_names(map(def$data, colnames), table_names)

  fks <-
    def %>%
    select(ref_table = table, fks) %>%
    unnest_list_of_df("fks")

  check_fk_child_tables(fks$table, table_names)

  fks %>%
    unnest_col("column", character()) %>%
    check_colnames(dm_col_names, "FK")

  fks %>%
    unnest_col("ref_column", character()) %>%
    select(table = ref_table, column = ref_column) %>%
    check_colnames(dm_col_names, "Parent key")

  if (!all(map_int(def$pks, vctrs::vec_size) %in% 0:1)) {
    abort_dm_invalid(
      "Not all tables have maximally 1 primary key."
    )
  }

  pks <-
    def %>%
    select(table, pks) %>%
    unnest_list_of_df("pks")

  pks %>%
    unnest_col("column", character()) %>%
    check_colnames(dm_col_names, "PK")

  check_one_zoom(def, is_zoomed(x))
  if (!all(map_lgl(compact(def$zoom), ~ inherits(., "data.frame") || inherits(., "tbl_dbi")))) {
    abort_dm_invalid(
      "Not all entries in `def$zoom` are of class `data.frame`, `tbl_dbi` or `NULL`."
    )
  }

  uks <- def %>%
    select(table, uks) %>%
    unnest_list_of_df("uks")

  uks %>%
    unnest_col("column", character()) %>%
    check_colnames(dm_col_names, "UK")

  check_no_nulls(def)

  invisible(x)
}

#' Validator
#'
#' `validate_dm()` has been replaced by [dm_validate()] for consistency.
#'
#' @param x An object.
#'
#' @export
#' @rdname deprecated
#' @keywords internal
validate_dm <- function(x) {
  deprecate_warn("0.3.0", "dm::validate_dm()", "dm::dm_validate()")
  dm_validate(x)
}

debug_dm_validate <- function(dm) {
  # Uncomment to enable validation for troubleshooting
  # dm_validate(dm)
  dm
}

check_dm <- function(x) {
  if (!is_dm(x)) {
    abort_is_not_dm(class(x))
  }
}

check_df_structure <- function(check, boilerplate, where) {
  force(where)

  if (!identical(names(check), names(boilerplate))) {
    abort_dm_invalid(glue(
      "Inconsistent column names in {where}: {commas(names(check), Inf)} vs. {commas(names(boilerplate), Inf)}."
    ))
  }

  if (!identical(check[0, ], boilerplate[0, ])) {
    abort_dm_invalid(glue("Inconsistent column types in {where}."))
  }

  inner_names <- map(check, vec_names)
  if (!all(map_lgl(inner_names, is.null))) {
    abort_dm_invalid(glue("Inner names in {where}."))
  }
}

check_fk_child_tables <- function(child_tables, dm_tables) {
  if (!all(map_lgl(child_tables, ~ .x %in% dm_tables))) {
    abort_dm_invalid("FK child table names not in `dm` table names.")
  }
}

check_colnames <- function(key_tibble, dm_col_names, which) {
  good <- map2_lgl(key_tibble$table, key_tibble$column, ~ ..2 %in% dm_col_names[[..1]])
  if (!all(good)) {
    bad_key <- key_tibble[which(!good)[[1]], ]
    abort_dm_invalid(glue(
      "{which} column name not in `dm` tables' column names: `{bad_key$table}`$`{bad_key$column}`"
    ))
  }
}

check_one_zoom <- function(def, zoomed) {
  if (zoomed) {
    if (sum(!map_lgl(def$zoom, is_null)) > 1) {
      abort_dm_invalid("More than one table is zoomed.")
    }
    if (sum(!map_lgl(def$zoom, is_null)) < 1) {
      abort_dm_invalid("Class is `dm_zoomed` but no zoomed table available.")
    }
    if (sum(!map_lgl(def$col_tracker_zoom, is_null)) > 1) {
      abort_dm_invalid("Key tracking is active for more than one zoomed table.")
    }
    if (sum(!map_lgl(def$col_tracker_zoom, is_null)) < 1) {
      abort_dm_invalid("No key tracking is active despite `dm` a `dm_zoomed`.")
    }
  } else {
    if (sum(!map_lgl(def$zoom, is_null)) != 0) {
      abort_dm_invalid("Zoomed table(s) available despite `dm` not a `dm_zoomed`.")
    }
    if (sum(!map_lgl(def$col_tracker_zoom, is_null)) != 0) {
      abort_dm_invalid("Key tracker for zoomed table activated despite `dm` not a `dm_zoomed`.")
    }
  }
}

check_no_nulls <- function(def) {
  check_no_nulls_col(def$fks, "foreign keys")
  check_no_nulls_col(def$pks, "primary keys")
  check_no_nulls_col(def$filters, "filter conditions")
}

check_no_nulls_col <- function(x, where) {
  if (any(map_lgl(x, is.null))) {
    abort_dm_invalid(paste0("Found `NULL` entry in ", where, "."))
  }
}

# dm invalid --------------------------------------------------------------

abort_dm_invalid <- function(why) {
  cli::cli_abort(
    "This {.cls dm} is invalid, reason: {why}",
    class = dm_error_full("dm_invalid"),
    call = dm_error_call()
  )
}

#' Get connection
#'
#' @description
#' `dm_get_con()` returns the DBI connection for a `dm` object.
#' This works only if the tables are stored on a database, otherwise an error
#' is thrown.
#'
#' @details
#' All lazy tables in a dm object must be stored on the same database server
#' and accessed through the same connection, because a large part
#' of the package's functionality relies on efficient joins.
#'
#' @inheritParams dm_add_pk
#'
#' @return The [`DBI::DBIConnection-class`] object for a `dm` object.
#'
#' @export
#' @examplesIf dm:::dm_has_financial()
#' dm_financial() %>%
#'   dm_get_con()
dm_get_con <- function(dm) {
  dm_local_error_call()
  check_not_zoomed(dm)
  src <- dm_get_src_impl(dm)
  if (!inherits(src, "src_dbi")) {
    abort_con_only_for_dbi()
  }
  src$con
}

dm_get_src_impl <- function(x) {
  tables <- dm_get_tables_impl(x)
  tbl_src(tables[1][[1]])
}

#' Create code to deconstruct a dm object
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Emits code that assigns each table in the dm to a variable,
#' using [pull_tbl()] with `keyed = TRUE`.
#' These tables retain information about primary and foreign keys,
#' even after data transformations,
#' and can be converted back to a dm object with [dm()].
#'
#' @inheritParams dm_add_pk
#' @param dm_name The code to use to access the dm object,
#'   by default the expression passed to this function.
#'
#' @return This function is called for its side effect of printing
#'   generated code.
#'
#' @export
#' @examplesIf rlang::is_installed("DiagrammeR")
#' dm <- dm_nycflights13()
#' dm_deconstruct(dm)
#' airlines <- pull_tbl(dm, "airlines", keyed = TRUE)
#' airports <- pull_tbl(dm, "airports", keyed = TRUE)
#' flights <- pull_tbl(dm, "flights", keyed = TRUE)
#' planes <- pull_tbl(dm, "planes", keyed = TRUE)
#' weather <- pull_tbl(dm, "weather", keyed = TRUE)
#' by_origin <-
#'   flights %>%
#'   summarize(.by = origin, mean_arr_delay = mean(arr_delay, na.rm = TRUE))
#'
#' by_origin
#' dm(airlines, airports, flights, planes, weather, by_origin) %>%
#'   dm_draw()
dm_deconstruct <- function(dm, dm_name = NULL) {
  dm_name <- enquo(dm_name)
  if (quo_is_null(dm_name)) {
    dm_name <- enexpr(dm)
  } else {
    dm_name <- quo_get_expr(dm_name)
  }

  dm_name <- as.character(dm_name)

  names <- names(dm)
  code <- paste0(
    tick_if_needed(names),
    " <- pull_tbl(",
    dm_name,
    ", ",
    map_chr(names, deparse),
    ", keyed = TRUE)",
    collapse = "\n"
  )

  message(code)

  invisible()
}

#' JSON pack join
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' A wrapper around [pack_join()] which stores the joined data into a JSON column.
#' `json_pack_join()` returns all rows and columns in `x` with a new JSON columns that contains all packed matches from `y`.
#'
#' @inheritParams dplyr::nest_join
#' @param x,y A pair of data frames or data frame extensions (e.g. a tibble).
#' @seealso [pack_join()], [json_nest_join()]
#' @export
#' @examples
#' df1 <- tibble::tibble(x = 1:3)
#' df2 <- tibble::tibble(x = c(1, 1, 2), y = c("first", "second", "third"))
#' df3 <- json_pack_join(df1, df2)
#' df3
#' df3$df2
json_pack_join <- function(x, y, by = NULL, ..., copy = FALSE, keep = FALSE, name = NULL) {
  UseMethod("json_pack_join")
}

#' @export
json_pack_join.data.frame <- function(
  x,
  y,
  by = NULL,
  ...,
  copy = FALSE,
  keep = FALSE,
  name = NULL
) {
  check_suggested("jsonlite", "json_pack_join")
  check_dots_empty()

  name_var <- name %||% as_label(enexpr(y))
  pack_join(x, y, by, ..., copy = copy, keep = keep, name = name_var) %>%
    mutate(
      !!name_var := map(
        unname(split(!!sym(name_var), seq.int(n()))),
        jsonlite::toJSON,
        digits = NA
      )
    )
}

mw_cg_make_dm_select_tbl <- function(dm, ..., table_names = NULL, rm = FALSE) {
  check_dots_empty()

  # Checks
  check_is_character_vec(table_names)
  check_tbl_in_dm(dm, table_names)

  tbl_names_sel_rm <- syms(table_names)

  if (rm) {
    if (length(tbl_names_sel_rm) == 1) {
      tbl_names_sel_rm <- exprs(-!!tbl_names_sel_rm[[1]])
    } else {
      tbl_names_sel_rm <- exprs(-c(!!!tbl_names_sel_rm))
    }
  }

  # Return call object
  list(
    call = expr(dm_select_tbl(
      .,
      !!!tbl_names_sel_rm
    ))
  )
}

#' Create data model from database constraints
#'
#' @description If there are any permament tables on a DB, a new [`dm`] object can be created that contains those tables,
#' along with their primary and foreign key constraints.
#'
#' Currently this works for the following databases:
#'
#' - Postgres/Redshift
#' - MySQL/MariaDB
#' - SQLite
#' - MSSQL
#' - DuckDB
#'
#' The default database schema is used if `schema` is `NULL`:
#' `"dbo"` for MSSQL, `"public"` for Postgres/Redshift, and the current database for MariaDB/MySQL and for SQLite/DuckDB.
#'
#' @param dest A `src`-object on a DB or a connection to a DB.
#'
#' @family DB interaction functions
#'
#' @return A [`dm`] object with the tables from the DB and the respective key relations.
#'
#' @noRd
#' @examples
#' if (FALSE) {
#'   src_sqlite <- dplyr::src_sqlite(":memory:", create = TRUE)
#'   iris_key <- mutate(iris, key = row_number())
#'
#'   # setting key constraints currently doesn't work on
#'   # SQLite but this would be the code to set the PK
#'   # constraint on the DB
#'   iris_dm <- copy_dm_to(
#'     src_sqlite,
#'     dm(iris = iris_key),
#'     set_key_constraints = TRUE
#'   )
#'
#'   # and this would be the code to learn
#'   # the `dm` from the SQLite DB
#'   iris_dm_learned <- dm_learn_from_db(src_sqlite)
#' }
#' @autoglobal
dm_learn_from_db <- function(dest, dbname = NA, schema = NULL, names_pattern = "{.table}") {
  # assuming that we will not try to learn from (globally) temporary tables, which do not appear in sys.table
  con <- con_from_src_or_con(dest)
  src <- src_from_src_or_con(dest)

  if (is.null(con)) {
    return()
  }

  schema <- check_schema(con, schema)

  info <- dm_meta(con, catalog = dbname, schema = schema)

  df_info <-
    info %>%
    dm_select_tbl(-schemata) %>%
    collect()

  dm_name <-
    df_info$tables %>%
    select(catalog = table_catalog, .schema = table_schema, .table = table_name) %>%
    mutate(name = glue(!!names_pattern)) %>%
    pull() %>%
    unclass() %>%
    vec_as_names(repair = "unique")

  from <-
    df_info$tables %>%
    select(catalog = table_catalog, schema = table_schema, table = table_name) %>%
    # RSQLite doesn't accept three-component identifiers, so we need to quote only the schema and table name
    mutate(catalog = if (!all(is.na(catalog))) catalog) %>%
    pmap_chr(~ DBI::dbQuoteIdentifier(con, DBI::Id(...)))

  df_key_info <-
    df_info %>%
    dm_zoom_to(tables) %>%
    mutate(dm_name = !!dm_name, from = !!from) %>%
    dm_update_zoomed() %>%
    dm_zoom_to(columns) %>%
    arrange(ordinal_position) %>%
    select(-ordinal_position) %>%
    left_join(tables) %>%
    dm_update_zoomed() %>%
    dm_select_tbl(constraint_column_usage, key_column_usage, columns, table_constraints)

  table_info <-
    df_key_info %>%
    dm_zoom_to(columns) %>%
    group_by(dm_name, from) %>%
    summarize(vars = list(column_name)) %>%
    ungroup() %>%
    pull_tbl()

  table_info$from_id <- DBI::dbUnquoteIdentifier(con, DBI::SQL(table_info$from))
  tables <- map2(table_info$from_id, table_info$vars, ~ tbl(con, .x, vars = .y))
  names(tables) <- table_info$dm_name

  pks_df <-
    df_key_info %>%
    dm_zoom_to(table_constraints) %>%
    filter(constraint_type == "PRIMARY KEY") %>%
    dm_update_zoomed() %>%
    dm_zoom_to(key_column_usage) %>%
    semi_join(table_constraints) %>%
    anti_join(constraint_column_usage) %>%
    arrange(ordinal_position) %>%
    dm_update_zoomed() %>%
    dm_select_tbl(-table_constraints) %>%
    dm_flatten_to_tbl(key_column_usage, .recursive = TRUE) %>%
    select(
      constraint_catalog,
      constraint_schema,
      constraint_name,
      dm_name,
      column_name,
      is_autoincrement
    ) %>%
    group_by(constraint_catalog, constraint_schema, constraint_name, dm_name) %>%
    summarize(
      pks = list(tibble(
        column = list(column_name),
        autoincrement = any(as.logical(is_autoincrement))
      ))
    ) %>%
    ungroup() %>%
    select(table = dm_name, pks)

  fks_df <-
    df_key_info %>%
    dm_zoom_to(table_constraints) %>%
    filter(constraint_type == "FOREIGN KEY") %>%
    dm_update_zoomed() %>%
    dm_zoom_to(key_column_usage) %>%
    semi_join(table_constraints) %>%
    left_join(table_constraints, select = c(delete_rule)) %>%
    left_join(
      columns,
      select = c(column_name, dm_name, table_catalog, table_schema, table_name)
    ) %>%
    dm_update_zoomed() %>%
    dm_select_tbl(-table_constraints) %>%
    dm_zoom_to(constraint_column_usage) %>%
    #
    # inner_join(): Matching column sometimes not found on Postgres
    inner_join(
      columns,
      select = c(column_name, dm_name, table_catalog, table_schema, table_name)
    ) %>%
    #
    dm_update_zoomed() %>%
    dm_select_tbl(-columns) %>%
    dm_rename(constraint_column_usage, constraint_column_usage.table_catalog = table_catalog) %>%
    dm_rename(constraint_column_usage, constraint_column_usage.table_schema = table_schema) %>%
    dm_rename(constraint_column_usage, constraint_column_usage.table_name = table_name) %>%
    dm_rename(constraint_column_usage, constraint_column_usage.column_name = column_name) %>%
    dm_rename(constraint_column_usage, constraint_column_usage.dm_name = dm_name) %>%
    dm_rename(key_column_usage, key_column_usage.table_catalog = table_catalog) %>%
    dm_rename(key_column_usage, key_column_usage.table_schema = table_schema) %>%
    dm_rename(key_column_usage, key_column_usage.table_name = table_name) %>%
    dm_rename(key_column_usage, key_column_usage.column_name = column_name) %>%
    dm_rename(key_column_usage, key_column_usage.dm_name = dm_name) %>%
    # inner_join: Sometimes, constraint_schema is different, https://github.com/cynkra/dm/issues/2228
    dm_flatten_to_tbl(constraint_column_usage, .join = inner_join) %>%
    select(
      constraint_catalog,
      constraint_schema,
      constraint_name,
      ordinal_position,
      delete_rule,
      ref_table = constraint_column_usage.dm_name,
      ref_column = constraint_column_usage.column_name,
      table = key_column_usage.dm_name,
      column = key_column_usage.column_name,
    ) %>%
    arrange(
      constraint_catalog,
      constraint_schema,
      constraint_name,
      ordinal_position,
    ) %>%
    select(-ordinal_position) %>%
    # FIXME: Where to learn this in INFORMATION_SCHEMA?
    group_by(
      constraint_catalog,
      constraint_schema,
      constraint_name,
      ref_table,
    ) %>%
    summarize(
      fks = list(tibble(
        ref_column = list(ref_column),
        table = if (length(table) > 0) table[[1]] else NA_character_,
        column = list(column),
        on_delete = {
          x <- case_when(
            delete_rule == "CASCADE" ~ "cascade",
            .default = "no_action"
          ) %>%
            unique()
          if (!is_empty(x) & !is_scalar_character(x)) {
            cli::cli_abort("delete_rule for all fk_cols in one constraint_name should be the same.")
          }
          x
        }
      ))
    ) %>%
    ungroup() %>%
    select(-(1:3)) %>%
    group_by(table = ref_table) %>%
    summarize(fks = list(bind_rows(fks))) %>%
    ungroup()

  # FIXME: add uks_df
  def <- new_dm_def(tables, pks_df, fks_df = fks_df)
  dm_from_def(def)
}

schema_if <- function(schema, table, con, dbname = NULL) {
  if (is_null(dbname) || is.na(dbname) || dbname == "") {
    purrr::map2(
      schema,
      table,
      ~ {
        if (is.na(.x)) {
          DBI::Id(table = .y)
        } else {
          DBI::Id(schema = .x, table = .y)
        }
      }
    )
  } else {
    # 'schema_if()' only used internally (can e.g. be set to default schema beforehand)
    # so IMHO we don't need a formal 'dm_error' here
    if (anyNA(schema)) {
      cli::cli_abort("{.arg schema} must be given if {.arg dbname} is not {.code NULL}.")
    }
    purrr::map2(
      table,
      schema,
      ~ DBI::Id(catalog = dbname, schema = .y, table = .x)
    )
  }
}

mw_cg_make_dm_add_fk <- function(
  dm,
  ...,
  table_names = NULL,
  column_names = NULL,
  other_column_names = NULL
) {
  # Checks
  check_dots_empty()
  stopifnot(length(table_names) == 2)
  table_name <- table_names[[1]]
  other_table_name <- table_names[[2]]
  table <- dm[[table_name]]
  other_table <- dm[[other_table_name]]

  stopifnot(length(column_names) >= 1)
  stopifnot(length(other_column_names) %in% c(0, length(column_names)))
  stopifnot(column_names %in% colnames(table))
  stopifnot(other_column_names %in% colnames(other_table))

  # Call generation
  if (length(column_names) == 1) {
    column_names_sym <- sym(column_names)
  } else {
    column_names_sym <- expr(c(!!!syms(column_names)))
  }

  if (length(other_column_names) == 0) {
    list(
      call = expr(dm_add_fk(
        .,
        !!sym(table_name),
        !!column_names_sym,
        !!sym(other_table_name)
      ))
    )
  } else {
    if (length(other_column_names) == 1) {
      other_column_names_sym <- sym(other_column_names)
    } else {
      other_column_names_sym <- expr(c(!!!syms(other_column_names)))
    }

    list(
      call = expr(dm_add_fk(
        .,
        !!sym(table_name),
        !!column_names_sym,
        !!sym(other_table_name),
        !!other_column_names_sym
      ))
    )
  }
}

#' Creates a dm object for the \pkg{pixarfilms} data
#'
#' @description
#' Creates an example [`dm`] object from the tables in
#' \pkg{pixarfilms}, along with the references.
#'
#' @inheritParams rlang::args_dots_empty
#' @param color Boolean, if `TRUE` (default), the resulting `dm` object will
#'   have colors assigned to different tables for visualization with
#'   `dm_draw()`.
#' @param consistent Boolean, In the original `dm`  the  `film` column in
#' `pixar_films` contains missing values so cannot be made a proper primary key.
#' Set to `TRUE` to remove those records.
#' @param version The version of the data to use.
#'   `"v1"` (default) uses a vendored snapshot of \pkg{pixarfilms} 0.2.1.
#'   `"latest"` uses the data from the installed \pkg{pixarfilms} package.
#'
#' @return A `dm` object consisting of \pkg{pixarfilms} tables, complete with
#'   primary and foreign keys and optionally colored.
#'
#' @export
#' @autoglobal
#' @examplesIf rlang::is_installed("DiagrammeR")
#' dm_pixarfilms()
#' dm_pixarfilms() %>%
#'   dm_draw()
dm_pixarfilms <- function(..., color = TRUE, consistent = FALSE, version = "v1") {
  check_dots_empty()

  version <- arg_match(version, c("v1", "latest"))

  # Extract data objects
  if (version == "latest") {
    # Check for data package installed
    check_suggested("pixarfilms", "dm_pixarfilms")

    pixar_films <- pixarfilms::pixar_films
    pixar_people <- pixarfilms::pixar_people
    academy <- pixarfilms::academy
    box_office <- pixarfilms::box_office
    genres <- pixarfilms::genres
    public_response <- pixarfilms::public_response
  } else {
    data <- pixarfilms_v1()
    pixar_films <- data$pixar_films
    pixar_people <- data$pixar_people
    academy <- data$academy
    box_office <- data$box_office
    genres <- data$genres
    public_response <- data$public_response
  }

  if (consistent) {
    pixar_films <- filter(pixar_films, !is.na(film))
  }

  # Create dm object
  dm <- dm(
    pixar_films,
    pixar_people,
    academy,
    box_office,
    genres,
    public_response
  )

  # Add primary keys
  dm <-
    dm %>%
    dm_add_pk(pixar_films, film) %>%
    dm_add_pk(academy, c(film, award_type)) %>%
    dm_add_pk(box_office, film)

  if ("genre" %in% names(genres)) {
    dm <-
      dm %>%
      dm_add_pk(genres, c(film, genre))
  } else {
    dm <-
      dm %>%
      dm_add_pk(genres, c(film, category, value))
  }

  dm <-
    dm %>%
    dm_add_pk(public_response, film)

  # Add foreign keys between tables
  dm <-
    dm %>%
    dm_add_fk(pixar_people, film, pixar_films) %>%
    dm_add_fk(academy, film, pixar_films) %>%
    dm_add_fk(box_office, film, pixar_films) %>%
    dm_add_fk(genres, film, pixar_films) %>%
    dm_add_fk(public_response, film, pixar_films)

  # Set colors for relationship diagram
  if (color) {
    dm <-
      dm %>%
      dm_set_colors(
        "#5B9BD5" = pixar_films,
        "#ED7D31" = c(
          academy,
          box_office,
          genres,
          public_response
        ),
        "#70AD47" = pixar_people
      )
  }

  dm
}

pixarfilms_v1 <- function() {
  readRDS(system.file("extdata/pixarfilms-v1.rds", package = "dm"))
}

test_that("`dm_pixarfilms()` works", {
  expect_snapshot(
    dm_examine_constraints(dm_pixarfilms(consistent = FALSE))
  )
  expect_snapshot(
    dm_examine_constraints(dm_pixarfilms(consistent = TRUE))
  )
})

test_that("`dm_pixarfilms()` version argument works", {
  dm_v1 <- dm_pixarfilms(version = "v1")
  expect_s3_class(dm_v1, "dm")
  expect_identical(dm_get_tables(dm_v1)$pixar_films, pixarfilms_v1()$pixar_films)

  dm_latest <- dm_pixarfilms(version = "latest")
  expect_s3_class(dm_latest, "dm")
  expect_identical(dm_get_tables(dm_latest)$pixar_films, pixarfilms::pixar_films)
})

# This file is part of .Rbuildignore
local({
  if (file.exists("../../man")) {
    names <- c("cynkra-72.png", "energie-72.png")
    sources <- file.path("../../man/figures", names)
    target_path <- "../../pkgdown/assets/reference/figures"
    targets <- file.path(target_path, names)

    info <- file.info(c(sources, targets))
    need_copy <- info$mtime[1:2] > info$mtime[3:4]
    need_copy[is.na(need_copy)] <- TRUE
    # need_copy <- c(TRUE, TRUE)
    if (any(need_copy)) {
      message("Copying pkgdown assets")
      stopifnot(file.copy(sources[need_copy], target_path, overwrite = TRUE))
    }
  }

  sources <- c(
    "duckdb",
    "maria",
    "mssql",
    "postgres",
    "sqlite",
    NULL
  )

  # paste0("DM_TEST_FILTER='", paste(sources, collapse = "|"), "' make test") |> clipr::write_clip()

  test_files <- set_names(paste0("test-", sources, ".R"), sources)
  in_contents <- imap_chr(test_files, ~ {
    if (file.exists(.x)) {
      text <- brio::read_file(.x)
      gsub(paste0('"', .y, '"'), '"db-source"', text)
    } else {
      ""
    }
  })

  table <- sort(table(in_contents[in_contents != ""]))
  stopifnot(length(table) <= 2)
  winner <- names(table)[1]

  out_contents <- iwalk(test_files, ~ {
    text <- gsub('"db-source"', paste0('"', .y, '"'), winner)
    brio::write_file(text, .x)
  })
})

#' @export
dplyr::`%>%`

#' @export
dplyr::filter

#' @export
dplyr::mutate

#' @export
dplyr::transmute

#' @export
dplyr::summarise

#' @export
dplyr::summarize

#' @export
dplyr::select

#' @export
dplyr::rename

#' @export
dplyr::group_by

#' @export
dplyr::ungroup

#' @export
dplyr::left_join

#' @export
dplyr::inner_join

#' @export
dplyr::full_join

#' @export
dplyr::semi_join

#' @export
dplyr::anti_join

#' @export
dplyr::right_join

#' @export
dplyr::collect

#' @export
dplyr::compute

#' @export
dplyr::copy_to

#' @export
dplyr::src_tbls

#' @export
dplyr::tbl

#' @export
dplyr::arrange

#' @export
dplyr::rows_insert

#' @export
dplyr::rows_append

#' @export
dplyr::rows_update

#' @export
dplyr::rows_patch

#' @export
dplyr::rows_upsert

#' @export
dplyr::rows_delete

#' @export
tibble::glimpse

#' @export
tibble::tibble

#' @export
tidyr::unite

#' @export
tidyr::separate

library(tidyverse)
devtools::load_all()

# using an input object provided by user:
inp_obj <- quo(dm_nycflights13())
cg_block <- new_cg_block(cg_input_object = inp_obj)

# evaluation of object:
cg_block %>%
  cg_eval_block()

# visualisation of object:
cg_block %>%
  cg_eval_block() %>%
  dm_draw()

# Code Output in Shiny App:
cg_block

# user hits add table mtcars:
cg_block_2 <-
  cg_block %>%
  cg_add_call(dm_add_tbl(., mtcars))

# evaluation of object:
cg_block_2 %>%
  cg_eval_block()

# visualisation of object:
cg_block_2 %>%
  cg_eval_block() %>%
  dm_draw()

# Code Output in Shiny App:
cg_block_2

# user adds a primary key for table mtcars:
cg_block_3 <-
  cg_block_2 %>%
  cg_add_call(dm_add_pk(., mtcars, c(mpg, wt)))

# evaluation of object:
cg_block_3 %>%
  cg_eval_block()

# visualisation of object:
cg_block_3 %>%
  cg_eval_block() %>%
  dm_draw()

# Code Output in Shiny App:
cg_block_3

# user hits undo last action: FIXME: probably need a function for this
cg_block_4 <- cg_block_3
cg_block_4$cg_f_list <- cg_block_4$cg_f_list[-length(cg_block_4$cg_f_list)]

# visualisation of object:
cg_block_4 %>%
  cg_eval_block() %>%
  dm_draw()

# Code Output in Shiny App:
cg_block_4

#' Load a dm from a remote data source
#'
#' @description
#' `dm_from_con()` creates a [dm] from some or all tables in a [src]
#' (a database or an environment) or which are accessible via a DBI-Connection.
#' For Postgres/Redshift and SQL Server databases, primary and foreign keys
#' are imported from the database.
#'
#' @param con A [`DBI::DBIConnection-class`] or a `Pool` object.
#' @param table_names
#'   A character vector of the names of the tables to include.
#' @param learn_keys
#'   `r lifecycle::badge("experimental")`
#'
#'   Set to `TRUE` to query the definition of primary and
#'   foreign keys from the database.
#'   Currently works for Postgres/Redshift, MariaDB/MySQL, SQLite, SQL Server, and DuckDB databases.
#'   The default attempts to query and issues an informative message.
#' @param .names
#'   `r lifecycle::badge("experimental")`
#'
#'   A glue specification that describes how to name the tables
#'   within the output, currently only for MSSQL, Postgres/Redshift and MySQL/MariaDB.
#'   This can use `{.table}` to stand for the table name, and
#'   `{.schema}` to stand for the name of the schema which the table lives
#'   within. The default (`NULL`) is equivalent to `"{.table}"` when a single
#'   schema is specified in `schema`, and `"{.schema}.{.table}"` for the case
#'   where multiple schemas are given, and may change in future versions.
#' @param ... `r lifecycle::badge("experimental")`
#'
#'   Additional parameters for the schema learning query.
#'
#'   - `schema`: supported for MSSQL (default: `"dbo"`), Postgres/Redshift (default: `"public"`), MariaDB/MySQL
#'     (default: current database) and SQLite (default: main schema).
#'     Learn the tables in a specific schema (or database for MariaDB/MySQL).
#'   - `dbname`: supported for MSSQL. Access different databases on the connected MSSQL-server;
#'     default: active database.
#'   - `table_type`: supported for Postgres/Redshift (default: `"BASE TABLE"`). Specify the table type. Options are:
#'     1. `"BASE TABLE"` for a persistent table (normal table type)
#'     2. `"VIEW"` for a view
#'     3. `"FOREIGN TABLE"` for a foreign table
#'     4. `"LOCAL TEMPORARY"` for a temporary table
#'
#' @return A `dm` object.
#'
#' @export
#' @examplesIf dm:::dm_has_financial()
#' con <- dm_get_con(dm_financial())
#'
#' # Avoid DBI::dbDisconnect() here, because we don't own the connection
dm_from_con <- function(
  con = NULL,
  table_names = NULL,
  learn_keys = NULL,
  .names = NULL,
  ...
) {
  dm_local_error_call()
  stopifnot(is(con, "DBIConnection") || inherits(con, "Pool"))

  check_suggested("dbplyr", "dm_from_con")

  if (inherits(con, "Pool")) {
    con <- pool_con <- pool::poolCheckout(con)
    on.exit(pool::poolReturn(pool_con))
  }

  src <- src_from_src_or_con(con)

  # Use smart default for `.names`, if it wasn't provided
  dots <- list2(...)
  if (!is.null(.names)) {
    names_pattern <- .names
  } else if (is.null(dots$schema) || length(dots$schema) == 1) {
    names_pattern <- "{.table}"
  } else {
    names_pattern <- "{.schema}.{.table}"
    cli::cli_inform('Using {.code .names = "{names_pattern}"}')
  }

  if (is.null(learn_keys) || isTRUE(learn_keys)) {
    # FIXME: Try to make it work everywhere
    tryCatch(
      {
        dm_learned <- dm_learn_from_db(con, ..., names_pattern = names_pattern)
        if (is_null(learn_keys)) {
          cli::cli_inform(c(
            "Keys queried successfully.",
            i = "Use {.code learn_keys = TRUE} to enforce querying keys and to mute this message."
          ))
        }

        if (is_null(table_names)) {
          return(dm_learned)
        }

        tbls_in_dm <- src_tbls_impl(dm_learned)

        if (!all(table_names %in% tbls_in_dm)) {
          abort_tbl_access(setdiff(table_names, tbls_in_dm))
        }
        tbls_req <- intersect(tbls_in_dm, table_names)

        return(dm_learned %>% dm_select_tbl(!!!tbls_req))
      },
      error = function(e) {
        if (isTRUE(learn_keys)) {
          abort_learn_keys(e)
        }
        cli::cli_inform(
          "Keys could not be queried.",
          x = conditionMessage(e),
          i = "Use {.code learn_keys = FALSE} to avoid trying to query keys and to mute this message."
        )
        NULL
      }
    )
  }

  if (is_null(table_names)) {
    src_tbl_names <- get_src_tbl_names(src, ..., names_pattern = names_pattern)
  } else {
    src_tbl_names <- table_names
    if (is.null(names(src_tbl_names))) {
      names(src_tbl_names) <- src_tbl_names
    }
  }

  tbls <-
    src_tbl_names %>%
    quote_ids(con) %>%
    map(possibly(tbl, NULL), src = src)

  bad <- map_lgl(tbls, is_null)
  if (any(bad)) {
    if (is_null(table_names)) {
      warn_tbl_access(names(tbls)[bad])
      tbls <- tbls[!bad]
    } else {
      abort_tbl_access(names(tbls)[bad])
    }
  }

  new_dm(tbls)
}

#' Load a dm from a remote data source
#'
#' Deprecated  in dm 0.3.0 in favor of [dm_from_con()].
#'
#' @inheritParams dm_from_con
#' @param src A dbplyr source, DBI connection object or a Pool object.
#'
#' @export
#' @keywords internal
dm_from_src <- function(src = NULL, table_names = NULL, learn_keys = NULL, ...) {
  if (is_null(src)) {
    return(empty_dm())
  }

  deprecate_warn("0.3.0", "dm::dm_from_src()", "dm::dm_from_con()")
  dm_from_con(con = con_from_src_or_con(src), table_names, learn_keys, ...)
}

quote_ids <- function(x, con, schema = NULL) {
  if (is.null(con)) {
    return(x)
  }

  if (is_null(schema)) {
    map_if(x, ~ !inherits(.x, "Id"), ~ DBI::Id(table = .x))
  } else {
    map_if(x, ~ !inherits(.x, "Id"), ~ schema_if(rep(schema, length(.x)), .x, con)[[1]])
  }
}

# Errors ------------------------------------------------------------------

abort_learn_keys <- function(parent) {
  cli::cli_abort(
    c(
      "Failed to learn keys from database.",
      i = "Use {.code learn_keys = FALSE} to work around, or {.code dm:::dm_meta()} to debug."
    ),
    class = dm_error_full("learn_keys"),
    parent = parent,
    call = dm_error_call()
  )
}

abort_tbl_access <- function(bad) {
  cli::cli_abort(
    c(
      "{cli::qty(length(bad))}Table{?s} {.field {bad}} cannot be accessed.",
      i = "Use {.code tbl(src, ...)} to troubleshoot."
    ),
    class = dm_error_full("tbl_access"),
    call = dm_error_call()
  )
}

warn_tbl_access <- function(bad) {
  cli::cli_warn(
    c(
      "{cli::qty(length(bad))}Table{?s} {.field {bad}} cannot be accessed.",
      i = "Use {.code tbl(src, ...)} to troubleshoot.",
      i = "Set the {.arg table_name} argument to avoid this warning."
    ),
    class = dm_warning_full("tbl_access")
  )
}

test_that("dm_from_con() supports 'Pool'", {
  skip_if_not_installed("pool")
  skip_if_not_installed("RSQLite")

  # expect no error
  conn <- pool::dbPool(RSQLite::SQLite(), dbname = "", timeout = 10)
  on.exit(pool::poolClose(conn))
  DBI::dbWriteTable(conn, "mtcars", mtcars)
  dm <- dm::dm_from_con(conn, learn_keys = FALSE)
  expect_identical(names(dm), "mtcars")
})

dm_nyc_filtered <-
  dm_nycflights13() %>%
  dm_filter(airports, name == "John F Kennedy Intl")

dm_apply_filters_to_tbl(dm_nyc_filtered, flights)

dm_nycflights13() %>%
  dm_filter(airports, name == "John F Kennedy Intl") %>%
  dm_apply_filters()

# If you want to keep only those rows in the parent tables
# whose primary key values appear as foreign key values in
# `flights`, you can set a `TRUE` filter in `flights`:
dm_nycflights13() %>%
  dm_filter(flights, 1 == 1) %>%
  dm_applyp_filters() %>%
  dm_nrow()
# note that in this example, the only affected table is
# `airports` because the departure airports in `flights` are
# only the three New York airports.

dm_nycflights13() %>%
  dm_filter(flights, month == 3) %>%
  dm_apply_filters()

dm_nycflights13() %>%
  dm_filter(planes, engine %in% c("Reciprocating", "4 Cycle")) %>%
  compute()
dm_nycflights13() %>%
  dm_filter(flights, month == 3) %>%
  dm_apply_filters_to_tbl(planes)

shiny_input_dm_nodes_edges <- function(x, session, name) {
  # The client shouldn't have to deal with that (and I wonder why we need to
  # do this here even).
  x$nodes <- unlist(x$nodes)
  x$edges <- unlist(x$edges)
  x
}

gui_run <- function(dm = dm_nycflights13(), select_tables = TRUE, debug = FALSE) {
  stopifnot(is_dm(dm))
  local_options(dm.gui_init_dm = dm)

  app <- shiny::shinyApp(ui = gui_ui(select_tables = select_tables), server = gui_server)

  if (debug) {
    local_options(shiny.fullstacktrace = debug)
  }

  code <- shiny::runApp(app)
  rstudioapi::sendToConsole(code, execute = FALSE)
}

# Determine active versions of R to test against
tags <- xml2::read_html("https://svn.r-project.org/R/tags/")

bullets <-
  tags |>
  xml2::xml_find_all("//li") |>
  xml2::xml_text()

version_bullets <- grep("^R-([0-9]+-[0-9]+-[0-9]+)/$", bullets, value = TRUE)
versions <- unique(gsub("^R-([0-9]+)-([0-9]+)-[0-9]+/$", "\\1.\\2", version_bullets))

r_release <- head(sort(as.package_version(versions), decreasing = TRUE), 5)

deps <- desc::desc_get_deps()
r_crit <- deps$version[deps$package == "R"]
if (length(r_crit) == 1) {
  min_r <- as.package_version(gsub("^>= ([0-9]+[.][0-9]+)(?:.*)$", "\\1", r_crit))
  r_release <- r_release[r_release >= min_r]
}

r_versions <- c("devel", as.character(r_release))

macos <- data.frame(os = "macos-latest", r = r_versions[2:3])
windows <- data.frame(os = "windows-latest", r = r_versions[1:3])
linux_devel <- data.frame(os = "ubuntu-22.04", r = r_versions[1], `http-user-agent` = "release", check.names = FALSE)
linux <- data.frame(os = "ubuntu-22.04", r = r_versions[-1])
covr <- data.frame(os = "ubuntu-22.04", r = r_versions[2], covr = "true", desc = "with covr")

include_list <- list(macos, windows, linux_devel, linux, covr)

if (file.exists(".github/versions-matrix.R")) {
  custom <- source(".github/versions-matrix.R")$value
  include_list <- c(include_list, list(custom))
}

print(include_list)

filter <- read.dcf("DESCRIPTION")[1,]["Config/gha/filter"]
if (!is.na(filter)) {
  filter_expr <- parse(text = filter)[[1]]
  subset_fun_expr <- bquote(function(x) subset(x, .(filter_expr)))
  subset_fun <- eval(subset_fun_expr)
  include_list <- lapply(include_list, subset_fun)
  print(include_list)
}

to_json <- function(x) {
  if (nrow(x) == 0) return(character())
  parallel <- vector("list", length(x))
  for (i in seq_along(x)) {
    parallel[[i]] <- paste0('"', names(x)[[i]], '":"', x[[i]], '"')
  }
  paste0("{", do.call(paste, c(parallel, sep = ",")), "}")
}

configs <- unlist(lapply(include_list, to_json))
json <- paste0('{"include":[', paste(configs, collapse = ","), ']}')

writeLines(paste0("matrix=", json), Sys.getenv("GITHUB_OUTPUT"))
writeLines(json)

# column -----------------------------------------------------------------------

null_to_character0 <- function(x) {
  if (is.null(x)) {
    return(character())
  }
  x
}

get_child_fk_cols <- function(dm, table_name = "flights") {
  child_fk_cols <-
    dm_get_all_fks(dm) %>%
    filter(child_table == table_name) %>%
    dplyr::pull(child_fk_cols)
  null_to_character0(unlist(child_fk_cols))
}

get_parent_key_cols <- function(dm, table_name = "flights") {
  parent_key_cols <-
    dm_get_all_fks(dm) %>%
    filter(parent_table == table_name) %>%
    dplyr::pull(parent_key_cols)
  null_to_character0(unlist(parent_key_cols))
}

get_pk_cols <- function(dm, table_name = "flights") {
  pk_cols <-
    dm_get_all_pks(dm) %>%
    filter(table == table_name) %>%
    dplyr::pull(pk_col)
  null_to_character0(unlist(pk_cols))
}


# dm = dm_nycflights13()
# data_column(dm, "flights")
data_column <- function(dm, table_name = "airports") {
  stopifnot(length(table_name) == 1)
  table_colnames <- colnames(dm[[table_name]])
  table_types <- vapply(dm[[table_name]], vctrs::vec_ptype_abbr, "", USE.NAMES = FALSE)

  tibble(
    name = table_colnames,
    type = table_types,
    is_pk = table_colnames %in% get_pk_cols(dm, table_name),
    is_child_fk = table_colnames %in% get_child_fk_cols(dm, table_name),
    is_parent_key = table_colnames %in% get_parent_key_cols(dm, table_name)
  )
}


# data <- tibble(
#   name = c("col1", "col2"),
#   type = c("numeric", "character"),
#   is_pk = c(TRUE, FALSE),
#   is_fk = c(FALSE, TRUE)
# )

# table_name <- "bla"

reactable_column <- function(data, table_name) {
  # browser()
  data %>%
    reactable::reactable(
      columns = list(
        name = reactable::colDef(
          name = table_name,
          # FIXME: filterable hides the first row with scrollable = TRUE
          # filterable = TRUE,
          cell = function(value, index) {
            type <- shiny::div(
              style = list(float = "right"),
              dplyr::if_else(
                data$is_pk[index],
                list(htmltools::span(
                  style = "margin-right: 10px;",
                  title = "Primary key",
                  shiny::icon(verify_fa = FALSE, "key")
                )),
                list(NULL)
              ),
              dplyr::if_else(
                data$is_child_fk[index],
                list(htmltools::span(
                  style = "margin-right: 10px;",
                  title = "Child in foreign key",
                  shiny::icon(verify_fa = FALSE, "angle-double-right")
                )),
                list(NULL)
              ),
              dplyr::if_else(
                data$is_parent_key[index],
                list(htmltools::span(
                  style = "margin-right: 10px;",
                  title = "Parent key",
                  shiny::icon(verify_fa = FALSE, "angle-double-left")
                )),
                list(NULL)
              ),
              shiny::span(
                class = "tag",
                style = "color: #999; border-color: #999;",
                title = paste0("Data type: ", data$type[index]),
                data$type[index]
              )
            )
            shiny::tagList(value, type)
          }
        ),
        type = reactable::colDef(
          show = FALSE,
          filterable = TRUE,
          cell = function(value) {
            shiny::span(class = "tag", value)
          }
        ),
        is_pk = reactable::colDef(show = FALSE),
        is_child_fk = reactable::colDef(show = FALSE),
        is_parent_key = reactable::colDef(show = FALSE)
      ),
      # searchable = TRUE,
      height = "272px",
      pagination = FALSE,
      sortable = FALSE,
      highlight = TRUE,
      selection = "multiple",
      onClick = "select",
      # minRows = 5,
      # maxRows = 5,
      bordered = TRUE,
      theme = dm_theme()
    )
}
# reactable_column(data, table_name)

# theme ------------------------------------------------------------------------

dm_theme <- function() {
  reactable::reactableTheme(
    # Full-width search bar with search icon
    searchInputStyle = list(
      width = "100%"
    ),

    # cellPadding = "10px 8px",
    style = list(
      ".tag" = list(
        display = "inline-block",
        width = "28px",
        textAlign = "center",
        padding = "0.125rem 0.25rem",
        color = "#999",
        fontSize = "0.9rem",
        border = "1px solid #777",
        borderRadius = "2px"
      )
    )
  )
}

# No longer needed for dbplyr >= 2.4.0
# https://github.com/tidyverse/dbplyr/pull/1190
rlang::on_load({
  if (rlang::is_installed("dbplyr") && getNamespaceInfo("dbplyr", "spec")["version"] <= "2.3.4") {
    dbplyr <- asNamespace("dbplyr")
    rlang::env_binding_unlock(dbplyr, "sql_values_subquery.MariaDBConnection")
    try(
      dbplyr$sql_values_subquery.MariaDBConnection <- dbplyr$sql_values_subquery.DBIConnection,
      silent = TRUE
    )
    rlang::env_binding_lock(dbplyr, "sql_values_subquery.MariaDBConnection")
  }
})

default_eval <- function() {
  # Only build vignettes in pkgdown for now
  (Sys.getenv("IN_PKGDOWN") != "") || grepl("^cran-", Sys.getenv("GITHUB_HEAD_REF"))
}

knitr::opts_chunk$set(
  error = (Sys.getenv("IN_PKGDOWN") == "") || !dm:::dm_has_financial(),
  # Only build vignettes:
  # - on pkgdown
  # - if not running in GitHub Actions
  # - if running in a branch that starts with CRAN in GitHub Actions
  eval = default_eval(),
  message = TRUE,
  collapse = TRUE,
  comment = "#>"
)

if (rlang::is_installed("fansi")) {
  fansi::set_knit_hooks(knitr::knit_hooks, which = c("output", "message"))
} else if (default_eval() && Sys.getenv("GITHUB_JOB") != "rcc-suggests") {
  cli::cli_abort(
    "{.pkg fansi} required for rendering vignettes but not installed."
  )
}

options(crayon.enabled = knitr::is_html_output(excludes = "gfm"), width = 75, cli.width = 75)

knit_print.grViz <- function(x, ...) {
  x %>%
    DiagrammeRsvg::export_svg() %>%
    c("`````{=html}\n", ., "\n`````\n") %>%
    knitr::asis_output()
}

# If input loads dm or tidyverse, we load it here to omit warnings
input <- readLines(knitr::current_input())
if (rlang::has_length(grep('^library[(]"?dm"?[)]', input))) {
  library(dm)
}
if (rlang::has_length(grep('^library[(]"?dplyr"?[)]', input))) {
  library(dplyr)
}

## Link helper to enable links only on pkgdown
href <- function(title, url) {
  if (Sys.getenv("IN_PKGDOWN") == "") {
    paste0(title, " (", url, ")")
  } else {
    paste0("[", title, "](", url, ")")
  }
}

#' Creates a dm object for the Financial data
#'
#' @description
#' `dm_financial()` creates an example [`dm`] object from the tables at
#' https://relational.fel.cvut.cz/dataset/Financial.
#' The connection is established once per session,
#' subsequent calls return the same connection.
#'
#' @return A `dm` object.
#'
#' @export
#' @examplesIf dm:::dm_has_financial() && rlang::is_installed("DiagrammeR")
#' dm_financial() %>%
#'   dm_draw()
dm_financial <- function() {
  check_suggested("RMariaDB", "dm_financial")

  my_db <- financial_db_con()

  my_dm <-
    dm_from_con(my_db, learn_keys = FALSE) %>%
    dm_add_pk(districts, id) %>%
    dm_add_pk(accounts, id) %>%
    dm_add_pk(clients, id) %>%
    dm_add_pk(loans, id) %>%
    dm_add_pk(orders, id) %>%
    dm_add_pk(disps, id) %>%
    dm_add_pk(cards, id) %>%
    dm_add_pk(trans, id) %>%
    dm_add_fk(loans, account_id, accounts) %>%
    dm_add_fk(orders, account_id, accounts) %>%
    dm_add_fk(disps, account_id, accounts) %>%
    dm_add_fk(disps, client_id, clients) %>%
    dm_add_fk(accounts, district_id, districts) %>%
    dm_add_fk(cards, disp_id, disps) %>%
    dm_add_fk(trans, account_id, accounts) %>%
    dm_set_colors(darkgreen = loans)

  my_dm
}

dm_has_financial <- function() {
  # Not on CRAN:
  if (Sys.getenv("CI") != "true") {
    return(FALSE)
  }

  # Crashes observed with R < 3.5:
  if (getRversion() < "3.5") {
    return(FALSE)
  }

  # Connectivity:
  try_connect <- try(dm_financial(), silent = TRUE)
  if (inherits(try_connect, "try-error")) {
    return(FALSE)
  }

  # Accessing the connection:
  try_count <- try(collect(count(dm_financial()$districts)), silent = TRUE)
  if (inherits(try_connect, "try-error")) {
    return(FALSE)
  }

  TRUE
}

#' dm_financial_sqlite()
#'
#' `dm_financial_sqlite()` copies the data to a temporary SQLite database.
#' The data is downloaded once per session, subsequent calls return the same database.
#' The `trans` table is excluded due to its size.
#' @rdname dm_financial
#' @export
dm_financial_sqlite <- function() {
  check_suggested("RSQLite", "dm_financial_sqlite")

  my_dm <-
    dm_financial() %>%
    dm_select_tbl(-trans)

  sqlite_db <- DBI::dbConnect(RSQLite::SQLite())
  sqlite_dm <- copy_dm_to(sqlite_db, my_dm, temporary = FALSE)

  sqlite_dm
}

# FIXME: Adapt argument list as provided/required by the Shiny app
mw_cg_make_dm_disentangle <- function(dm, ..., table_names = NULL) {
  check_dots_empty()

  # Checks
  stopifnot(is.character(table_names))
  stopifnot(table_names %in% names(dm))
  stopifnot(length(table_names) == 1)

  # Return call object
  list(
    call = expr(dm_disentangle(
      .,
      # FIXME: Adapt signature
      !!sym(table_names)
    ))
  )
}

# FIXME: Implement choosing dm or connection object from .GlobalEnv
enum_ops <- function(dm, ..., table_names = NULL, column_names = NULL, op_name = NULL) {
  check_dots_empty()
  if (any_null(table_names, column_names, op_name)) {
    enum_ops_(op_name, dm, table_names, column_names)
  } else {
    list2(
      input = list(
        dm = dm,
        table_names = table_names,
        column_names = column_names,
        op_name = op_name
      ),
      !!!enum(op_name)(dm, table_names, column_names)
    )
  }
}

enum_ops_ <- function(op, dm, tbls, cols) {
  op %>%
    add_single(dm, tbls, cols) %>%
    add_multiple(dm, tbls, cols)
}

add_single <- function(op, dm, tbls, cols) {
  if (is.null(op)) {
    if (length(tbls) == 1) {
      op <- "dm_add_pk"
    } else if (length(tbls) == 2) {
      op <- "dm_add_fk"
    }
    if (length(dm) > 0 && is.null(cols)) {
      op <- c(op, "dm_rm_fk")
    }
    list(input = list(dm = dm), single = list(op_name = op))
  } else {
    nil <- set_names(list(), character())
    list(input = list(dm = dm, op_name = op), single = nil)
  }
}

add_multiple <- function(e, dm, tbls, cols) {
  e[["multiple"]] <- list(table_names = NULL, column_names = NULL)
  if (is.null(tbls)) {
    # FIXME: Restrict table_names based on selected operation
    e[["multiple"]][["table_names"]] <- names(eval_tidy(dm))
    e[["multiple"]][["column_names"]] <- cols
  } else {
    if (length(tbls) == 1) {
      if (is.null(cols)) {
        e[["multiple"]][["column_names"]] <- colnames(eval_tidy(dm)[[tbls]])
      } else {
        e[["input"]][["column_names"]] <- cols
      }
    }
    e[["input"]][["table_names"]] <- tbls
  }
  e[["multiple"]] <- compact(e[["multiple"]])
  e
}

# TODO: Add more operators
enum <- function(op) {
  switch(op, dm_add_pk = enum_ops_dm_add_pk)
}

enum_ops_dm_add_pk <- function(dm, tbls, cols) {
  stopifnot(length(tbls) == 1)
  columns <- if (length(cols) == 1) sym(cols) else expr(c(!!!syms(cols)))
  out <- list(call = expr(dm_add_pk(., !!sym(tbls), !!columns)))
  if (dm_has_pk(eval_tidy(dm), !!sym(tbls))) {
    out[["call"]] <- as.call(c(as.list(out[["call"]]), force = TRUE))
    out[["confirmation_message"]] <- paste(
      "This table already has a primary key.",
      "Please confirm overwriting the existing primary key."
    )
  }
  out
}

any_null <- function(...) {
  detect_index(list(...), is.null) != 0
}

library(tidyverse)
library(rlang)
devtools::load_all()

parent <- tibble(a = c(1L, 1:3), b = -1)
child <- tibble(a = 1:4, c = 3)

dm <- dm(parent, child)

# This is done by the Shiny app
ops <- enum_ops(dm)
ops

enum_ops(dm, op_name = "dm_add_pk")
enum_ops(dm, op_name = "dm_add_pk", table_names = "parent")

final_ops <- enum_ops(dm, op_name = "dm_add_pk", table_names = "parent", column_names = "a")
final_ops

cg_block <- new_cg_block(quo(dm))

cg_block %>%
  cg_add_call(!!final_ops$call)

cg_block %>%
  cg_add_call(!!final_ops$call) %>%
  cg_eval_block()

#' Select and rename tables
#'
#' @description
#' `dm_select_tbl()` keeps the selected tables and their relationships,
#' optionally renaming them.
#'
#' @return The input `dm` with tables renamed or removed.
#'
#' @param dm A [`dm`] object.
#' @param ... One or more table names of the tables of the [`dm`] object.
#' `tidyselect` is supported, see [dplyr::select()] for details on the semantics.
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_select_tbl(airports, fl = flights)
#' @export
dm_select_tbl <- function(dm, ...) {
  dm_local_error_call()
  check_not_zoomed(dm)
  check_no_filter(dm)

  selected <- eval_select_table(quo(c(...)), src_tbls_impl(dm))
  dm_select_tbl_impl(dm, selected)
}

#' Change the names of the tables in a `dm`
#'
#' @description
#' `dm_rename_tbl()` renames tables.
#'
#' @rdname dm_select_tbl
#'
#' @examplesIf rlang::is_installed("nycflights13")
#' dm_nycflights13() %>%
#'   dm_rename_tbl(ap = airports, fl = flights)
#' @export
dm_rename_tbl <- function(dm, ...) {
  dm_local_error_call()
  check_not_zoomed(dm)

  selected <- eval_rename_table_all(quo(c(...)), src_tbls_impl(dm))
  dm_select_tbl_impl(dm, selected)
}

dm_select_tbl_impl <- function(dm, selected) {
  if (anyDuplicated(names(selected))) {
    abort_need_unique_names(names(selected[duplicated(names(selected))]))
  }

  # Required to avoid an error further on
  if (is_empty(selected)) {
    return(empty_dm())
  }

  def <-
    dm_get_def(dm) %>%
    filter_recode_table_def(selected) %>%
    filter_recode_table_fks(selected)

  dm_from_def(def)
}

filter_recode_table_fks <- function(def, selected) {
  def$fks <-
    # as_list_of() is needed so that `fks` doesn't become a normal list
    as_list_of(map(
      def$fks,
      filter_recode_fks_of_table,
      selected = selected
    ))
  def
}

filter_recode_table_def <- function(data, selected) {
  # We want to keep the order mentioned in `selected` here.
  # data$table only contains unique values by definition.
  idx <- match(selected, data$table, nomatch = 0L)

  data[idx, ] %>%
    mutate(table = recode_compat(table, prep_recode(selected)))
}

filter_recode_fks_of_table <- function(data, selected) {
  # data$table can have multiple entries, we don't care about the order
  idx <- data$table %in% selected
  out <- data[idx, ]
  out$table <- recode_compat(out$table, prep_recode(selected))
  out
}

prep_recode <- function(x) {
  set_names(names2(x), x)
}

prep_compact_recode <- function(x) {
  x <- x[names(x) != x]
  prep_recode(x)
}

recode2 <- function(x, new) {
  recipe <- prep_compact_recode(new)
  if (is_empty(recipe)) {
    x
  } else {
    recode_compat(x, recipe)
  }
}

mw_cg_make_dm_add_pk <- function(dm, ..., table_names = NULL, column_names = NULL) {
  # Checks
  check_dots_empty()
  stopifnot(length(table_names) == 1)
  table <- dm[[table_names]]

  check_at_least_one_col(column_names)
  stopifnot(column_names %in% colnames(table))

  # Call generation
  if (length(column_names) == 1) {
    column_names_sym <- sym(column_names)
  } else {
    column_names_sym <- expr(c(!!!syms(column_names)))
  }

  if (dm_has_pk(dm, !!table_names)) {
    # FIXME: Include confirmation message
    list(
      call = expr(dm_add_pk(
        .,
        !!sym(table_names),
        !!column_names_sym,
        force = TRUE # we want to change the pks interactively, otherwise error
      ))
    )
  } else {
    list(
      call = expr(dm_add_pk(
        .,
        !!sym(table_names),
        !!column_names_sym
      ))
    )
  }
}

expect_pred_chain <- function(fc, chain) {
  # Check that node is unique
  expect_true(identical(fc$node, unique(fc$node)))

  filtered_fc <-
    fc %>%
    filter(node %in% !!chain)

  # Beware of https://github.com/r-lib/testthat/issues/929
  expect_equal(filtered_fc$node, !!chain)
}

expect_not_pred <- function(fc, node) {
  # Check that node is missing
  expect_false(any(node %in% fc$node))
}

expect_error_obj <- function(object, regexp = NULL, class = NULL, ..., info = NULL, label = NULL) {
  testthat::expect_error(
    obj <- rlang::eval_tidy({{ object }}),
    regexp = regexp,
    class = class,
    ...,
    all = all,
    info = info,
    label = label
  )
  invisible(obj)
}

expect_warning_obj <- function(
  object,
  regexp = NULL,
  class = NULL,
  ...,
  info = NULL,
  label = NULL
) {
  testthat::expect_warning(
    obj <- rlang::eval_tidy({{ object }}),
    regexp = regexp,
    class = class,
    ...,
    info = info,
    label = label
  )
  invisible(obj)
}

expect_message_obj <- function(
  object,
  regexp = NULL,
  class = NULL,
  ...,
  info = NULL,
  label = NULL
) {
  testthat::expect_message(
    obj <- rlang::eval_tidy({{ object }}),
    regexp = regexp,
    class = class,
    ...,
    info = info,
    label = label
  )
  invisible(obj)
}

expect_condition_obj <- function(
  object,
  regexp = NULL,
  class = NULL,
  ...,
  info = NULL,
  label = NULL
) {
  testthat::expect_condition(
    obj <- rlang::eval_tidy({{ object }}),
    regexp = regexp,
    class = class,
    ...,
    info = info,
    label = label
  )
  invisible(obj)
}

expect_deprecated_obj <- function(object, regexp = NULL, ..., info = NULL, label = NULL) {
  lifecycle::expect_deprecated(
    obj <- rlang::eval_tidy({{ object }}),
    regexp = regexp,
    ...,
    info = info,
    label = label
  )
  invisible(obj)
}

testthat::test_that("testthat wrappers return the object", {
  expect_equal(
    expect_message_obj(
      {
        message("abc")
        "foo"
      },
      "a"
    ),
    "foo"
  )
  expect_equal(
    expect_warning_obj(
      {
        warning("abc")
        "foo"
      },
      "a"
    ),
    "foo"
  )
  expect_equal(
    expect_condition_obj(
      {
        message("abc")
        "foo"
      },
      "a"
    ),
    "foo"
  )
})

# from https://mariadb.com/kb/en/about-mariadb-connector-odbc/#installing-mariadb-connectorodbc-on-debianubuntu
mkdir odbc_package
pushd odbc_package
wget https://downloads.mariadb.com/Connectors/odbc/connector-odbc-3.1.7/mariadb-connector-odbc-3.1.7-ga-debian-x86_64.tar.gz
tar -xvzf mariadb-connector-odbc-3.1.7-ga-debian-x86_64.tar.gz
sudo install lib/libmaodbc.so /usr/lib/
sudo install -d /usr/lib/mariadb/
sudo install -d /usr/lib/mariadb/plugin/
sudo install lib/mariadb/plugin/auth_gssapi_client.so /usr/lib/mariadb/plugin/
sudo install lib/mariadb/plugin/caching_sha2_password.so /usr/lib/mariadb/plugin/
sudo install lib/mariadb/plugin/client_ed25519.so /usr/lib/mariadb/plugin/
sudo install lib/mariadb/plugin/dialog.so /usr/lib/mariadb/plugin/
sudo install lib/mariadb/plugin/mysql_clear_password.so /usr/lib/mariadb/plugin/
sudo install lib/mariadb/plugin/sha256_password.so /usr/lib/mariadb/plugin/
popd
rm -r odbc_package

---
title: "Insert, update, or remove rows in a database"
date: "`r Sys.Date()`"
author: James Wondrasek
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{How to: Insert, update or remove rows in a database}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---


``````{r setup, include = FALSE}
source("setup/setup.R")
``````


This tutorial introduces the methods {dm} provides for modifying the data in the tables of a relational model.
There are 6 methods:

* [`dm_rows_insert()`](#insert) - adds new unique rows
* [`dm_rows_append()`](#insert) - adds new rows unconditionally
* [`dm_rows_update()`](#update) - changes values in rows
* [`dm_rows_patch()`](#patch) - fills in missing values
* [`dm_rows_upsert()`](#upsert) - adds new rows or changes values if pre-existing
* [`dm_rows_delete()`](#delete) - deletes rows

## The dm_rows_* process

All six methods take the same arguments and using them follows the same process:

1. Create a temporary *changeset dm* object that defines the intended changes on the RDBMS
1. If desired, simulate changes with `in_place = FALSE` to double-check
1. Apply changes with `in_place = TRUE`.


To start, a `dm` object is created containing the tables and rows that you want to change.
This changeset `dm` is then copied into the same source as the dm you want to modify.
With the dm in the same RDBMS as the destination dm, you call the appropriate method, such as `dm_rows_insert()`, to make your planned changes, along with an argument of `in_place = FALSE` so you can confirm you achieve the changes that you want.

This verification can be done visually, looking at row counts and the like, or using {dm}'s constraint checking method, `dm_examine_constraints()`.
The biggest danger is damaging key relations between data spread across multiple tables by deleting or duplicating rows and their keys.
`dm_examine_constraints()` will catch errors where primary keys are duplicated or foreign keys do not have a matching primary key (unless the foreign key value is `NA`).

With the changes confirmed, you execute the method again, this time with the argument `in_place = TRUE` to make the changes permanent.
Note that `in_place = FALSE` is the default: you must opt in to actually change data on the database.

Each method has its own requirements in order to maintain database consistency.
These involve constraints on primary key values that uniquely identify rows.

| Method | Requirements |
|--------|--------------|
| `dm_rows_insert()` | Records with existing primary keys are silently ignored (via `dplyr::rows_insert(conflict = "ignore")`). |
| `dm_rows_append()` | All records are inserted, the underlying database might check for uniqueness of primary keys (and fail the operation) if a constraint is set. |
| `dm_rows_update()` | Primary keys must match for all records to be updated.|
| `dm_rows_patch()` | Updates missing values in existing records. Primary keys must match for all records to be patched.|
| `dm_rows_upsert()` | Updates existing records and adds new records, based on the primary key.|
| `dm_rows_delete()` | Removes matching records based on the primary key. Primary keys must match for all records to be deleted.|

To ensure the integrity of all relations during the process, all methods automatically determine the correct processing order for the tables involved.
For operations that create records, parent tables (which hold primary keys) are processed before child tables (which hold foreign keys).
For `dm_rows_delete()`, child tables are processed before their parent tables.
Note that the user is still responsible for setting transactions to ensure integrity of operations across multiple tables.
For more details on this see `vignette("howto-dm-theory")` and `vignette("howto-dm-db")`.

## Usage {#usage}

To demonstrate the use of these table modifying methods, we will create a simple `dm` object with two tables linked by a foreign key.
Note that the `child` table has a foreign key missing (`NA`).

``````{r}
library(dm)
parent <- tibble(value = c("A", "B", "C"), pk = 1:3)
parent
child <- tibble(value = c("a", "b", "c"), pk = 1:3, fk = c(1, 1, NA))
child
demo_dm <-
  dm(parent = parent, child = child) %>%
  dm_add_pk(parent, pk) %>%
  dm_add_pk(child, pk) %>%
  dm_add_fk(child, fk, parent)

demo_dm %>%
  dm_draw(view_type = "all")
``````

{dm} doesn't check your key values when you create a dm, we add this check:[^null-fk]

[^null-fk]: Be aware that when using `dm_examine_constraints()`, missing (denoted by `NULL` in SQL, while `NA` in R) foreign keys are allowed and will be counted as a match.
In some cases this doesn't make sense and non-NULL columns should be enforced by the RDBMS.
Currently, {dm} does not specify or check non-NULL constraints for columns.

``````{r }
dm_examine_constraints(demo_dm)
``````

Then we copy `demo_dm` into an SQLite database.
Note: the default for the method used, `copy_dm_to()`, is to create temporary tables that will be automatically deleted when your session ends.
As `demo_sql` will be the destination dm for the examples, the argument `temporary = FALSE` is used to make this distinction apparent.


``````{r }
library(DBI)
sqlite_db <- DBI::dbConnect(RSQLite::SQLite())
demo_sql <- copy_dm_to(sqlite_db, demo_dm, temporary = FALSE)
demo_sql
``````

{dm}'s table modification methods can be piped together to create a repeatable sequence of operations that returns a dm incorporating all the changes required.
This is a common use case for {dm} -- manually building a sequence of operations using temporary results until it is complete and correct, and then committing the result.

## `dm_rows_insert()` {#insert}

To demonstrate `dm_rows_insert()`, we create a dm with tables containing the rows to insert and copy it to `sqlite_db`, the same source as `demo_sql`.
For all of the `dm_rows_...()` methods, the source and destination `dm` objects must be in the same RDBMS.
You will get an error message if this is not the case.

The code below adds `parent` and `child` table entries for the letter "D".
First, the changeset dm is created and temporarily copied to the database:

``````{r }
new_parent <- tibble(value = "D", pk = 4)
new_parent
new_child <- tibble(value = "d", pk = 4, fk = 4)
new_child
dm_insert_in <-
  dm(parent = new_parent, child = new_child) %>%
  copy_dm_to(sqlite_db, ., temporary = TRUE)
``````

The changeset dm is then used as an argument to `dm_rows_insert()`.

``````{r }
dm_insert_out <-
  demo_sql %>%
  dm_rows_insert(dm_insert_in)
``````

This gives us a warning that changes will not persist (i.e., they are temporary).
Inspecting the `child` table of the resulting `dm_insert_out` and `demo_sql`, we can see that's exactly what happened.
{dm} returned to us a dm object with our inserted rows in place, but the underlying database has not changed.

``````{r }
dm_insert_out$child
demo_sql$child
``````

We repeat the operation, this time with the argument `in_place = TRUE` and the changes now persist in `demo_sql`.

``````{r }
dm_insert_out <-
  demo_sql %>%
  dm_rows_insert(dm_insert_in, in_place = TRUE)

demo_sql$child
``````

## `dm_rows_update()` {#update}

`dm_rows_update()` works the same as `dm_rows_insert()`.
We create the dm object and copy it to the same source as the destination.
Here we will change the foreign key for the row in `child` containing "b" to point to the correct row in `parent`.
And we will persist the changes.

``````{r }
updated_child <- tibble(value = "b", pk = 2, fk = 2)
updated_child
dm_update_in <-
  dm(child = updated_child) %>%
  copy_dm_to(sqlite_db, ., temporary = TRUE)

dm_update_out <-
  demo_sql %>%
  dm_rows_update(dm_update_in, in_place = TRUE)

demo_sql$child
``````

## `dm_rows_delete()` {#delete}

`dm_rows_delete()` is not currently implemented to work with an RDBMS, so we will shift our demonstrations back to the local R environment.
We've made changes to `demo_sql`, so we use `collect()` to copy the current tables out of SQLite.
Note that persistence is not a concern for *local* `dm` objects.
Every operation returns a new dm object containing the changes made.

``````{r }
local_dm <- collect(demo_sql)

local_dm$parent
local_dm$child
dm_deleted <-
  dm(parent = new_parent, child = new_child) %>%
  dm_rows_delete(local_dm, .)

dm_deleted$child
``````

## `dm_rows_patch()` {#patch}

`dm_rows_patch()` updates missing values in existing records.
We use it here to fix the missing foreign key in the `child` table.

``````{r }
patched_child <- tibble(value = "c", pk = 3, fk = 3)
patched_child
dm_patched <-
  dm(child = patched_child) %>%
  dm_rows_patch(dm_deleted, .)

dm_patched$child
``````

## `dm_rows_upsert()` {#upsert}

`dm_rows_upsert()` updates rows with supplied values if they exist or inserts the supplied values as new rows if they don't.
In this example we add the letter "D" back to our dm, and update the foreign key for "b".

``````{r }
upserted_parent <- tibble(value = "D", pk = 4)
upserted_parent
upserted_child <- tibble(value = c("b", "d"), pk = c(2, 4), fk = c(3, 4))
upserted_child
dm_upserted <-
  dm(parent = upserted_parent, child = upserted_child) %>%
  dm_rows_upsert(dm_patched, .)

dm_upserted$parent
dm_upserted$child
``````

When done, do not forget to disconnect:

``````{r disconnect}
DBI::dbDisconnect(sqlite_db)
``````

## Conclusion {#conclusion}

The `dm_rows_...()` methods give you row-level granularity over the modifications you need to make to your relational model.
Using the common `in_place` argument, they all can construct and verify your modifications before committing them.
There are a few limitations, as mentioned in the tutorial, but these will be addressed in future updates to {dm}.

## Further Reading

If this tutorial answered some questions, but opened others, these resources might be of assistance.

Is your data in an RDBMS? `vignette("howto-dm-db")` offers a detailed look at working with an existing relational data model.

If your data is in data frames, then you may want to read `vignette("howto-dm-df")` next.

If you would like to know more about relational data models in order to get the most out of dm, check out `vignette("howto-dm-theory")`.

If you're familiar with relational data models but want to know how to work with them in dm, then any of `vignette("tech-dm-join")`, `vignette("tech-dm-filter")`, or `vignette("tech-dm-zoom")` is a good next step.

---
title: "Copy tables to and from a database"
date: "`r Sys.Date()`"
author: James Wondrasek, Kirill Müller
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{How to: Copy data to and from a database}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---


``````{r setup, include = FALSE}
source("setup/setup.R")
``````


In this tutorial, we introduce {dm} methods and techniques for copying individual tables and entire relational data models into a relational database management system (RDBMS).
This is an integral part of the {dm} workflow.
Copying tables to an RDBMS is often a step in the process of building a relational data model from locally hosted data.
If your data model is complete, copying it to an RDBMS in a single operation allows you to leverage the power of the database and make it accessible to others.
For modifying and persisting changes to your data at the row-level see `vignette("howto-dm-rows")`.

## Copy models or copy tables?

Using {dm} you can persist an entire relational data model with a single function call.
`copy_dm_to()` will move your entire model into a destination RDBMS.
This may be all you need to deploy a new model.
You may want to add new tables to an existing model on an RDBMS.
These requirements can be handled using the `compute()` and `copy_to()` methods.

Calling `compute()` or `copy_to()` requires write permission on the RDBMS; otherwise, an error is returned.
Therefore, for the following examples, we will instantiate a test `dm` object and move it into a local SQLite database with full permissions.
{dm} and {dbplyr} are designed to treat the code used to manipulate a **local** SQLite database and a **remote** RDBMS similarly.
The steps for this were already introduced in `vignette("howto-dm-db")` and will be discussed in more detail in the [Copying a relational model](#copy-model) section.

``````{r }
library(dm)
library(dbplyr)

fin_dm <-
  dm_financial() %>%
  dm_select_tbl(-trans) %>%
  collect()

local_db <- DBI::dbConnect(RSQLite::SQLite())
deployed_dm <- copy_dm_to(local_db, fin_dm, temporary = FALSE)
``````


## Copying and persisting individual tables {#copying-tables}

As part of your data analysis, you may combine tables from multiple sources and create links to existing tables via foreign keys, or create new tables holding data summaries.
The example below, already discussed in `vignette("howto-dm-db")`, computes the total amount of all loans for each account.

``````{r}
my_dm_total <-
  deployed_dm %>%
  dm_zoom_to(loans) %>%
  summarize(.by = account_id, total_amount = sum(amount, na.rm = TRUE)) %>%
  dm_insert_zoomed("total_loans")
``````

The derived table `total_loans` is a *lazy table* powered by the {[dbplyr](https://dbplyr.tidyverse.org/)} package: the results are not materialized, instead an SQL query is built and executed each time the data is requested.

``````{r}
my_dm_total$total_loans %>%
  sql_render()
``````

To avoid recomputing the query every time you use `total_loans`, call `compute()` right before inserting the derived table with `dm_insert_tbl()`.
`compute()` forces the computation of a query and stores the full results in a table on the RDBMS.

``````{r}
my_dm_total_computed <-
  deployed_dm %>%
  dm_zoom_to(loans) %>%
  summarize(.by = account_id, total_amount = sum(amount, na.rm = TRUE)) %>%
  compute() %>%
  dm_insert_zoomed("total_loans")

my_dm_total_computed$total_loans %>%
  sql_render()
``````

```{r echo = FALSE, eval = TRUE}
# https://github.com/tidyverse/dbplyr/issues/639, https://github.com/tidyverse/dbplyr/pull/649
remote_name_total_loans <- "dbplyr_"
stopifnot(grepl(remote_name_total_loans, sql_render(my_dm_total_computed$total_loans), fixed = TRUE))
```

Note the differences in queries returned by `sql_render()`.
`my_dm_total$total_loans` is still being lazily evaluated and the full query constructed from the chain of operations that generated it is still in place and needs to be run to access it.
Contrast that with `my_dm_total_computed$total_loans`, where the query has been realized and accessing its rows requires a simple `SELECT *` statement.
The table name, `` `r remote_name_total_loans` ``, was automatically generated as the `name` argument was not supplied to `compute()`.

The default is to create a **temporary** tables.
If you want results to persist across sessions in **permanent** tables, `compute()` must be called with the argument `temporary = FALSE` and a table name for the `name` argument.
See `?compute` for more details.

When called on a whole `dm` object (without zoom), `compute()` materializes all tables into new temporary tables by executing the associated SQL query and storing the full results.
Depending on the size of your data, this may take considerable time or may even be infeasible.
It may be useful occasionally to create snapshots of data that is subject to change.

``````{r }
my_dm_total_snapshot <-
  my_dm_total %>%
  compute()
``````


## Adding local data frames to an RDBMS {#data-frames}

If you need to add local data frames to an existing `dm` object, use the `copy_to()` method.
It takes the same arguments as `copy_dm_to()`, except the second argument takes a data frame rather than a dm.
The result is a derived `dm` object that contains the new table.

To demonstrate the use of `copy_to()`, the example below will use {dm} to pull consolidated data from several tables out of an RDBMS, estimate a linear model from the data, then insert the residuals back into the RDBMS and link it to the existing tables.
This is all done with a local SQLite database, but the process would work unchanged on any supported RDBMS.

``````{r}
loans_df <-
  deployed_dm %>%
  dm_flatten_to_tbl(loans, .recursive = TRUE) %>%
  select(id, amount, duration, A3) %>%
  collect()
``````

Please note the use of `recursive = TRUE` for `dm_flatten_to_tbl()`.
This method gathers all linked information into a single wide table.
It follows foreign key relations starting from the table supplied as its argument and gathers all the columns from related tables, disambiguating column names as it goes.

In the above code, the `select()` statement isolates the columns we need for our model.
`collect()` works similarly to `compute()` by forcing the execution of the underlying SQL query, but it returns the results as a local tibble.

Below, the local tibble, `loans_df`, is used to estimate the linear model and the residuals are stored along with the original associated `id` in a new tibble, `loans_residuals`.
The `id` column is necessary to link the new tibble to the tables in the dm it was collected from.

``````{r}
model <- lm(amount ~ duration + A3, data = loans_df)

loans_residuals <- tibble::tibble(
  id = loans_df$id,
  resid = unname(residuals(model))
)

loans_residuals
``````

Adding `loans_residuals` to the dm is done using `copy_to()`.
The call to the method includes the argument `temporary = FALSE` because we want this table to persist beyond our current session.
In the same pipeline we create the necessary primary and foreign keys to integrate the table with the rest of our relational model.
For more information on key creation, see `vignette("howto-dm-db")` and `vignette("howto-dm-theory")`.

``````{r}
my_dm_sqlite_resid <-
  copy_to(deployed_dm, loans_residuals, temporary = FALSE) %>%
  dm_add_pk(loans_residuals, id) %>%
  dm_add_fk(loans_residuals, id, loans)

my_dm_sqlite_resid %>%
  dm_set_colors(violet = loans_residuals) %>%
  dm_draw()
my_dm_sqlite_resid %>%
  dm_examine_constraints()
my_dm_sqlite_resid$loans_residuals
``````


## Persisting a relational model with `copy_dm_to()` {#copy-model}

Persistence, because it is intended to make permanent changes, requires write access to the source RDBMS.
The code below is a repeat of the code that opened the [Copying and persisting individual tables](#copying-tables) section at the beginning of the tutorial.
It uses the {dm} convenience function `dm_financial()` to create a dm object corresponding to a data model from a public dataset repository.
The dm object is downloaded locally first, before deploying it to a local SQLite database.

`dm_select_tbl()` is used to exclude the transaction table `trans` due to its size, then the `collect()` method retrieves the remaining tables and returns them as a local dm object.

``````{r }
dm_financial() %>%
  dm_nrow()
fin_dm <-
  dm_financial() %>%
  dm_select_tbl(-trans) %>%
  collect()

fin_dm
``````

It is just as simple to move a local relational model into an RDBMS.

``````{r }
destination_db <- DBI::dbConnect(RSQLite::SQLite())

deployed_dm <-
  copy_dm_to(destination_db, fin_dm, temporary = FALSE)

deployed_dm
``````

Note that in the call to `copy_dm_to()` the argument `temporary = FALSE` is supplied.
Without this argument, the model would still be copied into the database, but the argument would default to `temporary = TRUE` and the data would be deleted once your session ends.

In the output you can observe that the `src` for `deployed_dm` is SQLite, while for `fin_dm` the source is not indicated because it is a local data model.

Copying a relational model into an empty database is the simplest use case for `copy_dm_to()`.
If you want to copy a model into an RDBMS that is already populated, be aware that `copy_dm_to()` will not overwrite pre-existing tables.
In this case you will need to use the `table_names` argument to give the tables unique names.

`table_names` can be a named character vector, with the names matching the table names in the dm object and the values containing the desired names in the RDBMS, or a function or one-sided formula.
In the example below, `paste0()` is used to add a prefix to the table names to provide uniqueness.

``````{r }
dup_dm <-
  copy_dm_to(destination_db, fin_dm, temporary = FALSE, table_names = ~ paste0("dup_", .x))

dup_dm
remote_name(dup_dm$accounts)
remote_name(deployed_dm$accounts)
``````

Note the different table names for `dup_dm$accounts` and `deployed_dm$accounts`.
For both, the table name is `accounts` in the `dm` object, but they link to different tables on the database.
In `dup_dm`, the table is backed by the table `dup_accounts` in the RDBMS.
`dm_deployed$accounts` shows us that this table is still backed by the `accounts` table from the `copy_dm_to()` operation we performed in the preceding example.

Managing tables in the RDBMS is outside the scope of `dm`.
If you find you need to remove tables or perform operations directly on the RDBMS, see the {[DBI](https://dbi.r-dbi.org/)} package.

When done, do not forget to disconnect:

``````{r disconnect}
DBI::dbDisconnect(destination_db)
DBI::dbDisconnect(local_db)
``````

## Conclusion {#conclusion}

`dm` makes it straightforward to deploy your complete relational model to an RDBMS using the `copy_dm_to()` function.
For tables that are created from a relational model during analysis or development, `compute()` and `copy_to()` can be used to persist them (using argument `temporary = FALSE`) between sessions or to copy local tables to a database `dm`.
The `collect()` method downloads an entire `dm` object that fits into memory from the database.


## Further Reading

If you need finer-grained control over modifications to your relational model, see `vignette("howto-dm-rows")` for an introduction to row level operations, including updates, insertions, deletions and patching.

If you would like to know more about relational data models in order to get the most out of dm, check out `vignette("howto-dm-theory")`.

If you're familiar with relational data models but want to know how to work with them in dm, then any of `vignette("tech-dm-join")`, `vignette("tech-dm-filter")`, or `vignette("tech-dm-zoom")` is a good next step.

---
title: "Manipulating individual tables"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Technical: Manipulating individual tables}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

``````{r setup, include = FALSE}
source("setup/setup.R")
``````

This vignette deals with situations where you want to transform tables of your `dm` object and then update an existing table or add a new table to the `dm` object.
There are two approaches:

1. extract the tables relevant to the calculation, perform the necessary transformations, and (if needed) recombine the resulting table into a `dm`,
1. do all this within the `dm` object by zooming to a table and manipulating it.

Both approaches aim at maintaining the key relations whenever possible.
We will explore the first approach here. 
For the second approach, see `vignette("tech-dm-zoom")`.

## Enabling {dplyr}-workflow within a `dm`

The `dm_get_tables()` and `pull_tbl()` functions have a new experimental argument `keyed`, which defaults to `FALSE`.
If set to `TRUE`, a list of objects of class `dm_keyed_tbl` is returned instead.
Because `dm_keyed_tbl` inherits from `tbl` or `tbl_lazy`, many {dplyr} and {tidyr} verbs will work unchanged.
These objects will also attempt to track primary and foreign keys, so that they are available for joins and when recombining these tables later into a `dm` object.

When you are finished with transforming your data, you can use `dm()` or `new_dm()` to recombine the tables into a `dm` object.
The resulting tables in the `dm` will have all the primary and foreign keys available that could be tracked from the original table.
Reconstructing the `dm` object is not strictly necessary if you're primarily interested in deriving one or multiple separate tables for analysis.

If this workflow proves as useful as it seems, subsetting tables via `$`, `[[` will default to `keyed = TRUE` in a forthcoming major release of {dm}.


## Examples

So much for the theory, but how does it look and feel?
To explore this, we once more make use of our trusted {nycflights13} data.

### Use case 1: Add a new column to an existing table

Imagine you want to have a column in `flights`, specifying if a flight left before noon or after.
Just like with {dplyr}, we can tackle this with `mutate()`.
Let us do this step by step:

```{r zoom}
library(dm)
library(dplyr)

flights_dm <- dm_nycflights13(cycle = TRUE)
flights_dm
flights_keyed <-
  flights_dm %>%
  dm_get_tables(keyed = TRUE)

# The print output for a `dm_keyed_tbl` looks very much like that from a normal
# `tibble`, with additional details about keys.
flights_keyed$flights
flights_tbl_mutate <-
  flights_keyed$flights %>%
  mutate(am_pm_dep = if_else(dep_time < 1200, "am", "pm"), .after = dep_time)

flights_tbl_mutate
```

To update the original `dm` with a new `flights` table we use `dm()`.
The bang-bang-bang (`!!!`) is a technical necessity that will become superfluous in a forthcoming release.

```{r zoom2}
updated_flights_dm <- dm(
  flights = flights_tbl_mutate,
  !!!flights_keyed[c("airlines", "airports", "planes", "weather")]
)

# The only difference in the `dm` print output is the increased number of
# columns
updated_flights_dm
# The schematic view of the data model remains unchanged
dm_draw(updated_flights_dm)
```

### Use case 2: Creation of a surrogate key

The same course of action could, for example, be employed to create a surrogate key for a table, a synthetic simple key that replaces a compound key.
We can do this for the `weather` table.

```{r}
library(tidyr)

flights_keyed$weather

# Maybe there is some hidden candidate for a primary key that we overlooked?
enum_pk_candidates(flights_keyed$weather)
# Seems we have to construct a column with unique values
# This can be done by combining column `origin` with `time_hour`, if the latter
# is converted to a single time zone first; all within the `dm`:
weather_tbl_mutate <-
  flights_keyed$weather %>%
  # first convert all times to the same time zone:
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  # paste together as character the airport code and the time
  unite("origin_slot_id", origin, time_hour_fmt) %>%
  select(origin_slot_id, everything())

# check if we the result is as expected:
weather_tbl_mutate %>%
  enum_pk_candidates() %>%
  filter(candidate)
# We apply the same transformation to create
# the foreign key in the flights table:
flights_tbl_mutate <-
  flights_keyed$flights %>%
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  unite("origin_slot_id", origin, time_hour_fmt) %>%
  select(origin_slot_id, everything())

surrogate_flights_dm <-
  dm(
    weather = weather_tbl_mutate,
    flights = flights_tbl_mutate,
    !!!flights_keyed[c("airlines", "airports", "planes")]
  ) %>%
  dm_add_pk(weather, origin_slot_id) %>%
  dm_add_fk(flights, origin_slot_id, weather)

surrogate_flights_dm %>%
  dm_draw()
```

### Use case 3: Disentangle `dm`

If you look at the `dm` created by `dm_nycflights13(cycle = TRUE)`, you see that two columns of `flights` relate to the same table, `airports`.
One column stands for the departure airport and the other for the arrival airport.
This generates a cycle which leads to failures with many operations that only work on cycle-free data models, such as `dm_flatten_to_tbl()`, `dm_filter()` or `dm_wrap_tbl()`.
In such cases, it can be beneficial to "disentangle" the `dm` by duplicating the referred table.
One way to do this in the {dm}-framework is as follows:

```{r}
disentangled_flights_dm <-
  dm(
    destination = flights_keyed$airports,
    origin = flights_keyed$airports,
    !!!flights_keyed[c("flights", "airlines", "planes", "weather")]
  ) %>%
  # Key relations are also duplicated, so the wrong ones need to be removed
  dm_rm_fk(flights, dest, origin) %>%
  dm_rm_fk(flights, origin, destination)

disentangled_flights_dm %>%
  dm_draw()
```


### Use case 4: Add summary table to `dm`

Here is an example for adding a summary of a table as a new table to a `dm`. Foreign-key relations are taken care of automatically.
This example shows an alternative approach of deconstruction reconstruction using `pull_tbl()`.

```{r}
flights_derived <-
  flights_dm %>%
  pull_tbl(flights, keyed = TRUE) %>%
  dplyr::count(origin, carrier)

derived_flights_dm <- dm(flights_derived, !!!flights_keyed)

derived_flights_dm %>%
  dm_draw()
```

### Use case 5: Joining tables

If you would like to join some or all of the columns of one table to another, you can make use of one of the `..._join()` methods for a `dm_keyed_tbl`.
In many cases, using keyed tables derived from a `dm` object allows omitting the `by` argument without triggering a message, because they are safely inferred from the foreign keys stored in the `dm_keyed_tbl` objects.
For the syntax, please see the example below.

```{r}
planes_for_join <-
  flights_keyed$planes %>%
  select(tailnum, plane_type = type)

joined_flights_tbl <-
  flights_keyed$flights %>%
  # let's first reduce the number of columns of flights
  select(-dep_delay:-arr_delay, -air_time:-minute, -starts_with("sched_")) %>%
  # in the {dm}-method for the joins you can specify which columns you want to
  # add to the subsetted table
  left_join(planes_for_join)

joined_flights_dm <- dm(
  flights_plane_type = joined_flights_tbl,
  !!!flights_keyed[c("airlines", "airports", "weather")]
)

# this is how the table looks now
joined_flights_dm$flights_plane_type
# also here, the FK-relations are transferred to the new table
joined_flights_dm %>%
  dm_draw()
```

### Use case 6: Retrieve all tables

Retrieving all tables from a `dm` object requires a lot of boilerplate code.
The `dm_deconstruct()` function helps creating that boilerplate.
For a `dm` object, it prints the code necessary to create local variables for all tables.

```{r}
dm <- dm_nycflights13()
dm_deconstruct(dm)
```

This code can be copy-pasted into your script or function.

---
title: "Filtering in relational data models"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Technical: Filtering in relational data models}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
author: Katharina Brunner
---


``````{r setup, include = FALSE}
source("setup/setup.R")
``````

The {dm} package offers functions to work with relational data models in R.

This document introduces you to filtering functions, and shows how to apply them to the data that is separated into multiple tables.

Our example data is drawn from the [{nycflights13}](https://github.com/tidyverse/nycflights13) package that contains five inter-linked tables.

First, we will load the packages that we need:

```{r message=FALSE, warning=FALSE}
library(nycflights13)
library(dm)
```

## Data: nycflights13

To explore filtering with {dm}, we'll use the {nycflights13} data with its `flights`, `planes`, `airlines`, `airports` and `weather` tables.

This dataset contains information about the 336 776 flights that departed from New York City in 2013, with 3322 different planes and 1458 airports involved.
The data comes from the US Bureau of Transportation Statistics, and is documented in `?nycflights13::flights`.

To start with our exploration, we have to create a `dm` object from the {nycflights13} data.
The built-in `dm::dm_nycflights13()` function takes care of this.

By default it only uses a subset of the complete data though: only the flights on the 10th of each month are considered, reducing the number of rows in the `flights` table to 11 227.

A [data model object](https://dm.cynkra.com/articles/tech-dm-class.html) contains data from the source tables, and metadata about the tables.

If you would like to create a `dm` object from tables other than the example data, you can use the `new_dm()`, `dm()` or `as_dm()` functions.
See `vignette("howto-dm-df")` for details.

```{r}
dm <- dm_nycflights13()
```

The console output of the 'dm` object shows its data and metadata, and is colored for clarity:

```{r}
dm
```

Now we know that there are five tables in our `dm` object.
But how are they connected?
These relations are best displayed as a visualization of the entity-relationship model:

```{r}
dm_draw(dm)
```

You can look at a single table with `tbl`.
To print the `airports` table, call

```{r}
tbl(dm, "airports")
```

## Filtering a `dm` object {#filter}

`dm_filter()` allows you to select a subset of a `dm` object.

### How it works

Filtering a `dm` object is not that different from filtering a dataframe or tibble with `dplyr::filter()`.

The corresponding {dm} function is `dm::dm_filter()`.
With this function one or more filtering conditions can be set for one of the tables of the `dm` object.
These conditions are immediately evaluated for their respective tables and for all related tables.
For each resulting table, all related tables (directly or indirectly) with a filter condition them are taken into account in the following way:
    - filtering semi-joins are successively performed along the paths from each of the filtered tables to the requested table, each join reducing the left-hand side tables of the joins to only those of their rows with key values that have corresponding values in key columns of the right-hand side tables of the join.
    - eventually the requested table is returned, containing only the the remaining rows after the filtering joins

Currently, this only works if the graph induced by the foreign key relations is cycle free.
Fortunately, this is the default for `dm_nycflights13()`.


### Filtering Examples

Let's see filtering in action:

**We only want the data that is related to John F. Kennedy International Airport.**

```{r}
filtered_dm <-
  dm %>%
  dm_filter(airports = (name == "John F Kennedy Intl"))
filtered_dm
```

You can get the numbers of rows of each table with `dm_nrow()`.

```{r}
rows_per_table <-
  filtered_dm %>%
  dm_nrow()
rows_per_table
sum(rows_per_table)
```

```{r echo = FALSE, eval = TRUE}
sum_nrow <- NA
sum_nrow_filtered <- NA
```

```{r}
sum_nrow <- sum(dm_nrow(dm))
sum_nrow_filtered <- sum(dm_nrow(dm_apply_filters(filtered_dm)))
```

The total number of rows in the `dm` drops from `r format(sum_nrow, big.mark = " ")` to `r format(sum_nrow_filtered, big.mark = " ")` (the only unaffected table is the disconnected `weather` table).

Next example:

**Get a `dm` object containing data for flights from New York to the Dulles International Airport in Washington D.C., abbreviated with `IAD`.**

```{r}
dm %>%
  dm_filter(flights = (dest == "IAD")) %>%
  dm_nrow()
```

Applying multiple filters to different tables is also supported.

An example:

**Get all January 2013 flights from Delta Air Lines which didn't depart from John F. Kennedy International Airport.**

```{r}
dm_delta_may <-
  dm %>%
  dm_filter(
    airlines = (name == "Delta Air Lines Inc."),
    airports = (name != "John F Kennedy Intl"),
    flights = (month == 1)
  )
dm_delta_may
dm_delta_may %>%
  dm_nrow()
```

You can inspect the filtered tables by subsetting them.

In the `airlines` table, Delta is the only remaining carrier:

```{r}
dm_delta_may$airlines
```

Which planes were used to service these flights?

```{r}
dm_delta_may$planes
```

And indeed, all included flights departed in January (`month == 1`):

```{r}
dm_delta_may$flights %>%
  dplyr::count(month)
```

For comparison, let's review the equivalent manual query for `flights` in `dplyr` syntax:

```{r}
airlines_filtered <- filter(airlines, name == "Delta Air Lines Inc.")
airports_filtered <- filter(airports, name != "John F Kennedy Intl")
flights %>%
  semi_join(airlines_filtered, by = "carrier") %>%
  semi_join(airports_filtered, by = c("origin" = "faa")) %>%
  filter(month == 5)
```

The {dm} code is leaner because the foreign key information is encoded in the object.

## SQL statements behind filtering a `dm` object on a database

{dm} is meant to work with relational data models, locally as well as on databases.
In your project, the data is probably not stored locally but in a remote [relational database](https://dm.cynkra.com/articles/howto-dm-theory.html) that can be queried with SQL statements.

You can check the queries by using `sql_render()` from the [{dbplyr}](https://dbplyr.tidyverse.org/) package.

Example:

**Print the SQL statements for getting all January 2013 flights from Delta Air Lines, which did not depart from John F. Kennedy International Airport, with the data stored in a sqlite database.**

To show the SQL query behind a `dm_filter()`, we copy the `flights`, `airlines` and `airports` tables from the `nyflights13` dataset to a temporary in-memory database using the built-in function `copy_dm_to()` and `dbplyr::src_memdb`.

Then we filter the data, and print the corresponding SQL statement with `dbplyr::sql_render()`.

```{r, warning=FALSE}
dm %>%
  dm_select_tbl(flights, airlines, airports) %>%
  copy_dm_to(dbplyr::src_memdb(), .) %>%
  dm_filter(
    airlines = (name == "Delta Air Lines Inc."),
    airports = (name != "John F Kennedy Intl"),
    flights = (month == 1)
  ) %>%
  dm_get_tables() %>%
  purrr::map(dbplyr::sql_render)
```

Further reading: {dm}'s function for copying data [from and to databases](https://dm.cynkra.com/articles/howto-dm-copy.html).

---
title: "Function naming logic"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{Technical: Function naming logic}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---


``````{r setup, include = FALSE}
source("setup/setup.R")
``````

This vignette describes the logic behind the names of the {dm}-functions.

In general, the functions are implemented in such a way that no function can handle instances of both `dm` classes (`dm` and `dm_zoomed`) as an argument.
In most situations, a `dm_zoomed` behaves like a table and there are mostly no explicit functions available for this class, but rather methods (see vignette ["Zooming and manipulating tables"](https://dm.cynkra.com/articles/tech-dm-zoom.html).

## Prefixes

1. `dm_`: There are two cases for this prefix:
    - The function takes a `dm` object as a first argument
    - The function constructs a `dm` object from other objects
1. `check_`: Certain tests on the arguments are performed and an error is thrown in case they are not passed. If the tests are passed, the first argument is returned invisibly.
1. `examine_`: The function executes checks and returns an object containing an overview of the results of the checks.

Most important extended prefixes `dm_..._`:

1. `dm_examine_`: The function examines several aspects of a `dm` object
1. `dm_add_`: Adds either key constraints or new tables to a `dm`
1. `dm_get_`: Retrieves a specific piece of information about the `dm` (often a part of its constituents)
1. `dm_rm_`: Removes either key constraints or tables from a `dm`

## Suffixes

1. `_dm`: Validators, checkers, or coercers
1. `_to_tbl`: Data from a `dm` is processed in a way that a tibble is produced
1. `_candidates`: If key-candidates are to be calculated and presented in tibble-form
1. `_pk(s)`: Relates to primary key(s)
1. `_fk(s)`: Relates to foreign key(s)
1. `_tbl`: Add/remove/select/rename a tibble to/from/within a `dm` object
1. `_colors`: Relates to the colors for the visualization of a `dm` object
1. `_zoomed`: Insert, update, or discard the zoomed table. In combination with the prefix `dm_`, this indicates that the first argument is a `dm_zoomed` rather than a `dm`.

---
title: "Joining in relational data models"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{Technical: Joining in relational data models}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
author: Katharina Brunner
---


``````{r setup, include = FALSE}
source("setup/setup.R")
``````

The {dm} package offers functions to work with relational data models in R.
A common task for multiple, separated tables that have a shared attribute is merging the data.

This document introduces you to the joining functions of {dm} and shows how to apply them using data from the [{nycflights13}](https://github.com/tidyverse/nycflights13) package.

[Relational data models](https://dm.cynkra.com/articles/howto-dm-theory#model) consist of multiple tables that are linked with [foreign keys](https://dm.cynkra.com/articles/howto-dm-theory#fk). They are the building blocks for joining tables.
Read more about relational data models in the vignette ["Introduction to Relational Data Models"](https://dm.cynkra.com/articles/howto-dm-theory).

First, we load the packages that we need:

```{r}
library(dm)
```

## Data: nycflights13

To explore filtering with {dm}, we'll use the {nycflights13} data with its tables `flights`, `planes`, `airlines` and `airports`.

This dataset contains information about the 336,776 flights that departed from New York City in 2013, with 3,322 different planes and 1,458 airports involved.
The data comes from the US Bureau of Transportation Statistics, and is documented in `?nycflights13`.

First, we have to create a `dm` object from the {nycflights13} data.
This is implemented with `dm_nycflights13()`.

A [data model object](https://dm.cynkra.com/articles/tech-dm-class.html) contains the data as well as metadata.

If you would like to create a `dm` from other tables, please look at `?dm` and the function `new_dm()`.

```{r}
dm <- dm_nycflights13()
```

## Joining a `dm` object {#join}

{dm} allows you to join two tables of a `dm` object based on a shared column. You can use all join functions that you know from the {dplyr} package.
Currently {dplyr} supports four types of mutating joins, two types of filtering joins, and a nesting join.
See `?dplyr::join` for details.

### How it works

A join is the combination of two tables based on shared information.
In technical terms, we merge the tables that need to be directly connected by a [foreign key relation](https://dm.cynkra.com/articles/howto-dm-theory#fk).

The existing links can be inspected in two ways:

1. Visually, by drawing the data model with `dm_draw()`

```{r}
dm %>%
  dm_draw()
```

The directed arrows show explicitly the relation between different columns.

2. Printed to the console by calling `dm_get_all_fks()`

```{r}
dm %>%
  dm_get_all_fks()
```

### Joining Examples

Let's look at some examples:

**Add a column with airline names from the `airlines` table to the `flights` table.**

```{r}
dm_joined <-
  dm %>%
  dm_flatten_to_tbl(flights, airlines, .join = left_join)
dm_joined
```

As you can see below, the `dm_joined` data frame has one more column than the `flights` table.
The difference is the `name` column from the `airlines` table.

```{r}
dm$flights %>%
  names()

dm$airlines %>%
  names()

dm_joined %>%
  names()
```

The result is not a `dm` object anymore, but a (tibble) data frame:

```{r}
dm_joined %>%
  class()
```

Another example:

**Get all flights that can't be matched with airlines names.**

We expect the `flights` data from {nycflights13} package to be clean and well organized, so no flights should remain.
You can check this with an `anti_join`:

```{r}
dm %>%
  dm_flatten_to_tbl(flights, airlines, .join = anti_join)
```

An example with filtering on a `dm` and then merging:

**Get all May 2013 flights from Delta Air Lines which didn't depart from John F. Kennedy International Airport in - and join all the airports data into the `flights` table.**

```{r}
dm_nycflights13(subset = FALSE) %>%
  dm_filter(
    airlines = (name == "Delta Air Lines Inc."),
    airports = (name != "John F Kennedy Intl"),
    flights = (month == 5)
  ) %>% 
  dm_flatten_to_tbl(flights, airports, .join = left_join)
```

See `vignette("tech-dm-filter")` for more details on filtering.

A last example:

**Merge all tables into one big table.**

Sometimes you need everything in one place.
In this case, you can use the `dm_flatten_to_tbl()` function.
It joins all the tables in your `dm` object together into one wide table.
All you have to do is to specify the starting table.
The following joins are determined by the foreign key links.

```{r}
dm_nycflights13() %>%
  dm_select_tbl(-weather) %>%
  dm_flatten_to_tbl(.start = flights)
```

To be more precise, `dm_flatten_to_tbl()` will join all tables from one level of hierarchy (i.e., direct neighbors to table `.start`).
If you want to cover tables from all levels of hierarchy, use the argument `recursive = TRUE` for `dm_flatten_to_tbl()` instead.

Also, be aware that all column names need to be unique.
The `dm_flatten_to_tbl()` takes care of this by automatically renaming the relevant columns and informs the user if any names were changed, e.g. `dm_rename(airlines, airlines.name = name)`.

If you want to merge all tables, but get a nested table in return, use `dm_wrap_tbl()` with `pull_tbl()` instead:

```{r}
dm_nycflights13() %>%
  dm_wrap_tbl(root = flights) %>%
  pull_tbl(flights)

---
title: "Model verification - keys, constraints and normalization"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{Technical: Model verification - keys, constraints and normalization}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---


``````{r setup, include = FALSE}
source("setup/setup.R")
``````

In this document, we will present several specialized functions for conducting basic tests about key conditions and about relations between tables.
We will also describe functions that can be used for splitting and uniting tables.

```{r}
library(dm)
```


## Testing key constraints

This section contains information and examples about the following functions:

1. `check_key(.data, ...)`
1. `check_subset(t1, c1, t2, c2)`
1. `check_set_equality(t1, c1, t2, c2)`

When you have tables (data frames) that are connected by key relations, {dm} can help you to verify the assumed key relations and/or determine the existing key relations between the tables.
For example, if you have tables:

```{r}
data_1 <- tibble(a = c(1, 2, 1), b = c(1, 4, 1), c = c(5, 6, 7))
data_2 <- tibble(a = c(1, 2, 3), b = c(4, 5, 6), c = c(7, 8, 9))
```

and you want to know if `a` is a primary key for `data_1`, you can use the `check_key()` function:

```{r error = TRUE}
check_key(data_1, a)
```

Mind the error message when a test is not passed.

For `data_2`, column `a` is a key:

```{r}
check_key(data_2, a)
```

To see if a column of one table contains only those values that are also present in another column of another table, the `check_subset()` function can be used:

```{r}
check_subset(data_1, a, data_2, a)
```

This function is important for determining if a column is a foreign key to some other table.
What about the inverse relation?

```{r error = TRUE}
check_subset(data_2, a, data_1, a)
```

It should be kept in mind that `check_subset()` does not test if column `c2` is a unique key of table `t2`.
In order to find out if a (child) table `t1` contains a column `c1` that is a foreign key to a (parent) table `t2` with the corresponding column `c2`, the following method should be used:

```{r eval=FALSE}
check_key(t2, c2)
check_subset(t1, c1, t2, c2)
```

To check both directions at once, and to find out if the unique values of `c_1` in `t_1` are the same as those of `c_2` in `t_2`, {dm} provides the function `check_set_equality()`:

```{r error=TRUE}
check_set_equality(data_1, a, data_2, a)
```

Introducing one more table enables us to show how it looks when the test is passed:

```{r}
data_3 <- tibble(a = c(2, 1, 2), b = c(4, 5, 6), c = c(7, 8, 9))

check_set_equality(data_1, a, data_3, a)
```

If the test is passed, the return value of the function will be the first table parameter (invisibly).
This ensures that the functions can be conveniently used in a pipe configuration.

## Testing cardinalities between two tables

This section contains information and examples for the functions

1. `check_cardinality_0_n(parent_table, primary_key_column, child_table, foreign_key_column)`
1. `check_cardinality_1_n(parent_table, primary_key_column, child_table, foreign_key_column)`
1. `check_cardinality_0_1(parent_table, primary_key_column, child_table, foreign_key_column)`
1. `check_cardinality_1_1(parent_table, primary_key_column, child_table, foreign_key_column)`
1. `examine_cardinality(parent_table, primary_key_column, child_table, foreign_key_column)`

The four functions for testing for a specific kind of cardinality of the relation all require a parent table and a child table as inputs.
The functions first test if that requirement is fulfilled by checking if:

1. `primary_key_column` is a unique key for `parent_table`
1. The set of values of `foreign_key_column` is a subset of the set of values of `primary_key_column`

The cardinality specifications `0_n`, `1_n`, `0_1`, `1_1` refer to the expected relation that the child table has with the parent table.
The numbers '0', '1' and 'n' refer to the number of values in the child table's column (`foreign_key_column`) that correspond to each value of the parent table's column (`primary_key_column`).
'n' means more than one in this context, with no upper limit.

`0_n` means, that for each value of the `parent_key_column`, the number of corresponding records in the child table is unrestricted.
`1_n` means, that for each value of the `parent_key_column` there is at least one corresponding record in the child table.
This means that there is a "surjective" relation from the child table to the parent table w.r.t. the specified columns, i.e. for each parent table column value there exists at least one equal child table column value.

`0_1` means, that for each value of the `parent_key_column`, at least zero and at most one value has to correspond to it in the column of the child table.
This means that there is an "injective" relation from the child table to the parent table w.r.t. the specified columns, i.e. no parent table column value is addressed multiple times.
But not all of the parent table column values have to be referred to.

`1_1` means, that for each value of the `parent_key_column`, exactly one value has to correspond to it in the child table's column.
This means that there is a "bijective" ("injective" AND "surjective") relation between the child table and the parent table w.r.t. the specified columns, i.e. the set of values of the two columns is equal and there are no duplicates in either of them.

Also `examine_cardinality()` first performs the above mentioned tests to figure out, if the parent-child table relationship criteria are met.
Subsequently, two further checks are made to determine the nature of the relation (surjective, injective, bijective, or none of these) between the two columns.

### Examples

Given the following three data frames:

```{r}
d1 <- tibble(a = 1:5)
d2 <- tibble(c = c(1:5, 5))
d3 <- tibble(c = 1:4)
d4 <- tibble(a = c(2:5, 5))
```

Here are some examples of how the cardinality testing functions can be used:

```{r error=TRUE}
# This does not pass, `c` is not unique key of d2:
check_cardinality_0_n(d2, c, d1, a)

# This passes, multiple values in d2$c are allowed:
check_cardinality_0_n(d1, a, d2, c)

# This does not pass, injectivity is violated:
check_cardinality_1_1(d1, a, d2, c)

# This passes:
check_cardinality_0_1(d1, a, d3, c)
```

`examine_cardinality()` returns the type of relation, e.g.:

```{r}
examine_cardinality(d1, a, d3, c)
examine_cardinality(d1, a, d2, c)
examine_cardinality(d1, a, d1, a)
examine_cardinality(d1, a, d4, a)
```

Just like the underlying cardinality functions, it will also inform you if any restrictions on cardinality are violated:

```{r}
examine_cardinality(d2, c, d1, a)
```


## Table surgery

The relevant functions are:

1. `decompose_table(.data, new_id_column, ...)`
1. `reunite_parent_child(child_table, parent_table, id_column)`
1. `reunite_parent_child_from_list(list_of_parent_child_tables, id_column)`

The first function implements table normalization.
An existing table is split into a parent table (i.e. a lookup table) and a child table (containing the observations), linked by a key column (here: `new_id_column`).
Basically, a foreign key relation would be created, pointing from the  `new_id_column` of the child table to the parent table's corresponding column, which can be seen as the parent table's primary key column.
The function `decompose_table()` does that, as can be seen in the following example:

```{r}
mtcars_tibble <- tibble::as_tibble(mtcars)
mtcars_tibble
decomposed_table <- decompose_table(mtcars_tibble, am_gear_carb_id, am, gear, carb)
decomposed_table
```

A new column is created, with which the two tables can be joined again, essentially creating the original table.

The functions that do the inverse operation, i.e. join a parent and a child table and subsequently drop the `new_id_column`, are `reunite_parent_child()` and `reunite_parent_child_from_list()`.
The former takes as arguments two tables and the unquoted name of the ID column, and the latter takes as arguments a list of two tables plus the unquoted name of the ID column:

```{r}
parent_table <- decomposed_table$parent_table
child_table <- decomposed_table$child_table
reunite_parent_child(child_table, parent_table, id_column = am_gear_carb_id)
```

```{r eval = FALSE}
# Shortcut:
reunite_parent_child_from_list(decomposed_table, id_column = am_gear_carb_id)
```

Currently, these functions only exist as a low-level operation on tables.
We plan to extend this operation to `dm` objects in the future.

Storing all data related to a problem in a single table or data frame ("the dataset") can result in many repetitive values. Separation into multiple tables helps data quality but poses a different challenge: if information from different tables is required jointly, a "merge" or "join" operation must be carried out. {dm} is a new package that fills a gap in the R ecosystem: it makes working with multiple tables just as easy as working with a single table.

In the relational database model, a "data model" consists of:

- tables (both the definition and the data),

- primary and foreign key constraints.

Primary keys uniquely identify rows in a table. Foreign keys establish a link between rows in two tables. The {dm} package combines these concepts with data manipulation powered by the tidyverse: entire data models are handled in a single entity, a "dm" object.

Data models are defined by combining a set of tables and establishing relationships through primary and foreign key constraints. The {dm} package also provides tools for normalization (decomposition into multiple tables). Once a "dm" is defined, the following transformations can be applied:

- Basic data manipulation on individual tables: adding, removing or renaming columns, filtering, aggregation

- Joins: combining multiple tables into one

All transformations take relationships between tables into account. For example, the filter operation also affects related tables.

"dm" objects are storage agnostic, they work efficiently with in-memory data and with databases, and can be copied from and to databases.

Use cases for the {dm} package include a Shiny application that supports complex filters across multiple tables, and processing a data model with ~10⁹ rows on a MSSQL database.

In this presentation, Kirill introduces the basics of relational data models, demonstrates the most important features of the {dm} package, and presents a few use cases.

{dm} is an R package that provides tools for working with multiple related tables, stored as data frames or in a relational database.


## Background{#background}

Relational databases and flat tables, like data frames or spreadsheets, present data in fundamentally different ways.

In data frames and spreadsheets, all data is presented together in one large table with many rows and columns.
This means that the data is accessible in one location but has the disadvantage that the same values may be repeated multiple times, resulting in bloated tables with redundant data.
In the worst case scenario, a data frame may have many rows and columns but only a single value different in each row.

Relational databases, on the other hand, do not keep all data together but split it into multiple smaller tables.
That separation into sub-tables has several advantages:

- all information is stored only once, avoiding repetition and conserving memory
- all information is updated only once and in one place, improving consistency and avoiding errors that may result from updating the same value in multiple locations
- all information is organized by topic and segmented into smaller tables that are easier to handle

Separation of data, thus, helps with data quality, and explains the continuing popularity of relational databases in production-level data management.

The downside of this approach is that it is harder to merge together information from different data sources and to identify which entities refer to the same object, a common task when modeling or plotting data.
To be mapped uniquely, the entities would need to be designated as _keys_, and the separate tables collated together through a process called _joining_.

In R, there already exist packages that support handling inter-linked tables but the code is complex and requires multiple command sequences.
The goal of the {dm} package is to simplify the data management processes in R while keeping the advantages of relational data models and the core concept of splitting one table into multiple tables.
In this way, you can have the best of both worlds: manage your data as a collection of linked tables, then flatten multiple tables into one for an analysis with {dplyr} or other packages, on an as-needed basis.

Although {dm} is built upon relational data models, it is not a database itself.
It can work transparently with both relational database systems and in-memory data, and copy data [from and to databases](https://cynkra.github.io/dm/articles/dm.html#copy).

---
title: "Visualizing dm objects"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{Techincal: Visualizing dm objects}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---

``````{r setup, include = FALSE}
source("setup/setup.R")
``````

Once you have all your primary keys set and all foreign key relations defined, a graphical representation of your data model offers a condensed view of the tables and the relationships between the tables.
The following functions can be used to visualize the `dm` object:^[The code for the functions in this section is borrowed from the [{datamodelr}](https://github.com/bergant/datamodelr) package.]

1. `dm_draw()`
1. `dm_set_colors()`
1. `dm_get_colors()`
1. `dm_get_available_colors()`

We use the prepared example `dm` object `dm_nycflights13(cycle = TRUE)`:

```{r}
library(dm)
library(dplyr)
flights_dm_w_many_keys <- dm_nycflights13(cycle = TRUE, color = FALSE)
flights_dm_w_many_keys
```

The schema is drawn with `dm_draw()`.

```{r}
dm_draw(flights_dm_w_many_keys)
```

You can use colors to visually group your tables into families to reflect their logical grouping.
The available colors are either hexcoded colors or the standard R color names.
The function `dm_get_available_colors()` forwards to `grDevices::colors()`:

```{r}
dm_get_available_colors()
```

Colors are assigned with `dm_set_colors()` using syntax known in the {tidyverse} as {tidyselect}-syntax, here in the form: `color = table`.
[Select helper functions](https://tidyselect.r-lib.org/reference/language.html) are supported.
The result of `dm_set_colors()` is a `dm` object.
The information about the color is stored together with the rest of the metadata.

```{r}
flights_dm_w_many_keys_and_colors <-
  flights_dm_w_many_keys %>%
  dm_set_colors(
    maroon4 = flights,
    orange = starts_with("air"),
    "#5986C4" = planes
  )
```

Draw the schema with `dm_draw()`.

```{r}
dm_draw(flights_dm_w_many_keys_and_colors)
```

The colors can be queried with `dm_get_colors()`.

```{r}
dm_get_colors(flights_dm_w_many_keys_and_colors)
```

See the documentation for `dm_draw()` for further options.
One important argument is `view_type`.
Besides the default `"keys_only"`, it accepts `"all"` to display all columns, and `"title_only"` to show only the title of the table.

```{r}
flights_dm_w_many_keys_and_colors %>%
  dm_draw(view_type = "title_only")
```

If you would like to visualize only some of the tables, use `dm_select_tbl()` before drawing:

```{r}
flights_dm_w_many_keys_and_colors %>%
  dm_select_tbl(flights, airports, planes) %>%
  dm_draw()
```

Finally, for exporting a drawing to `svg` you could use `DiagrammeRsvg::export_svg()`:

```{r eval = default_eval() && rlang::is_installed("DiagrammeRsvg")}
flights_dm_w_many_keys_and_colors %>%
  dm_select_tbl(flights, airports, planes) %>%
  dm_draw() %>%
  DiagrammeRsvg::export_svg() %>%
  write("flights_dm_w_many_keys_and_color.svg")
``` 

---
title: "Create a dm object from data frames"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{How to: Create a dm object from data frames}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---


``````{r setup, include = FALSE}
source("setup/setup.R")
``````


dm allows you to create your own relational data models from local data frames.
Once your data model is complete, you can deploy it to a range of database management systems (DBMS) using {dm}.

## Creating a dm object from data frames

The example data set that we will be using is available through the [`nycflights13`](https://github.com/tidyverse/nycflights13) package.
The five tables that we are working with contain information about all flights that departed from the airports of New York to other destinations in the United States in 2013:

- `flights` represents the trips taken by planes
- `airlines` includes
  - the names of transport organizations (`name`)
  - their abbreviated codes (`carrier`)
- `airports` indicates the ports of departure (`origin`) and of destination (`dest`)
- `weather` contains meteorological information at each hour
- `planes` describes characteristics of the aircraft

Once we've loaded {nycflights13}, the aforementioned tables are all in our work environment, ready to be accessed.

```{r nycflights13, message=FALSE}
library(nycflights13)

airports
```

Your own data will probably not be available as an R package.
Whatever format it is in, you will need to be able to load it as data frames into your R session.
If the data is too large, consider using dm to connect to the database instead.
See `vignette("howto-dm-db")` for details on using dm with databases.

## Adding Tables

Our first step will be to tell `dm` which tables we want to work with and how they are connected.
For that we can use `dm()`, passing in the table names as arguments.

```{r}
library(dm)

flights_dm_no_keys <- dm(airlines, airports, flights, planes, weather)
flights_dm_no_keys
```

The `as_dm()` function is an alternative that works if you already have a list of tables.

## A dm is a list

`dm` objects behave like lists with a user- and console-friendly print format.
In fact, using a dm as a nicer list for organizing your data frames in your environment is an easy first step towards using dm and its data modeling functionality.

Subsetting syntax for a `dm` object (either by subscript or by name) is similar to syntax for lists, and so you don't need to learn any additional syntax to work with `dm` objects.

```{r listlike}
names(flights_dm_no_keys)
flights_dm_no_keys$airports
flights_dm_no_keys[c("airports", "flights")]
```

## Defining Keys

Even though we now have a `dm` object that contains all our data, we have not specified how our five tables are connected.
To do this, we need to define primary keys and foreign keys on the tables.

Primary keys and foreign keys are how relational database tables are linked with each other.
A primary key is a column or column tuple that has a unique value for each row within a table.
A foreign key is a column or column tuple containing the primary key for a row in another table.
Foreign keys act as cross references between tables.
They specify the relationships that gives us the *relational* database.
For more information on keys and a crash course on databases, see `vignette("howto-dm-theory")`.

## Primary Keys

`dm` offers `dm_enum_pk_candidates()` to identify viable primary keys for a table in the `dm` object, and `dm_add_pk()` to add them.

```{r}
dm_enum_pk_candidates(
  dm = flights_dm_no_keys,
  table = planes
)
```

Now, we can add the identified primary keys:

```{r}
flights_dm_only_pks <-
  flights_dm_no_keys %>%
  dm_add_pk(table = airlines, columns = carrier) %>%
  dm_add_pk(airports, faa) %>%
  dm_add_pk(planes, tailnum) %>%
  dm_add_pk(weather, c(origin, time_hour))
flights_dm_only_pks
```

Note that {dm} functions work with both named and positional argument specification, and compound keys can be specified using a vector argument.

## Foreign Keys

To define how our tables are related, we use `dm_add_fk()` to add foreign keys.
Naturally, this function will deal with two tables: a table *looking for* a reference, and a table that is *providing* the reference.
Accordingly, while calling `dm_add_fk()`, the `table` argument specifies the table that needs a foreign key to link it to a second table, and the `ref_table` argument specifies the table to be linked to, which needs a primary key already defined for it.

```{r}
dm_enum_fk_candidates(
  dm = flights_dm_only_pks,
  table = flights,
  ref_table = airlines
)
```


Having chosen a column from the successful candidates provided by `dm_enum_fk_candidates()`, we use the `dm_add_fk()` function to establish the foreign key linking the tables.
In the second call to `dm_add_fk()` we complete the process for the `flights` and `airlines` tables that we started above.
The `carrier` column in the `airlines` table will be added as the foreign key in `flights`.

```{r}
flights_dm_all_keys <-
  flights_dm_only_pks %>%
  dm_add_fk(table = flights, columns = tailnum, ref_table = planes) %>%
  dm_add_fk(flights, carrier, airlines) %>%
  dm_add_fk(flights, origin, airports) %>%
  dm_add_fk(flights, c(origin, time_hour), weather)
flights_dm_all_keys
```

Having created the required primary and foreign keys to link all the tables together, we now have a relational data model we can work with.

## Visualization

Visualizing a data model is a quick and easy way to verify that we have successfully created the model we were aiming for.
We can use `dm_draw()` at any stage of the process to generate a visual representation of the tables and the links between them:

```{r}
flights_dm_no_keys %>%
  dm_draw(rankdir = "TB", view_type = "all")
```


```{r}
flights_dm_no_keys %>%
  dm_add_pk(airlines, carrier) %>%
  dm_draw()
```


```{r}
flights_dm_only_pks %>%
  dm_add_fk(flights, tailnum, planes) %>%
  dm_draw()
```


```{r}
flights_dm_all_keys %>%
  dm_draw()
```


## Integrity Checks

As well as checking our data model visually, dm can examine the constraints we have created by the addition of keys and verify that they are sensible.

```{r}
flights_dm_no_keys %>%
  dm_examine_constraints()

flights_dm_only_pks %>%
  dm_examine_constraints()

flights_dm_all_keys %>%
  dm_examine_constraints()
```

The results are presented in a human-readable form, and available as a tibble for programmatic inspection.


## Programming

Helper functions are available to access details on keys and check results.

A data frame of primary keys is retrieved with `dm_get_all_pks()`:

```{r}
flights_dm_only_pks %>%
  dm_get_all_pks()
```

Similarly, a data frame of foreign keys is retrieved with `dm_get_all_fks()`:

```{r}
flights_dm_all_keys %>%
  dm_get_all_fks()
```

We can use `tibble::as_tibble()` on the result of `dm_examine_constraints()` to programmatically inspect which constraints are not satisfied:

```{r}
flights_dm_all_keys %>%
  dm_examine_constraints() %>%
  tibble::as_tibble()
```

## Conclusion {#conclusion}

In this tutorial, we have demonstrated how simple it is to create relational data models from local data frames using {dm}, including setting primary and foreign keys and visualizing the resulting relational model.

## Further reading

`vignette("howto-dm-db")` -- This article covers accessing and working with RDBMSs within your R session, including manipulating data, filling in missing relationships between tables, getting data out of the RDBMS and into your model, and deploying your data model to an RDBMS.

`vignette("howto-dm-theory")` -- Do you know all about data frames but very little about relational data models? This quick introduction will walk you through the key similarities and differences, and show you how to move from individual data frames to a relational data model.

---
title: "Zooming and manipulating tables"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Technical: Zooming and manipulating tables}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

``````{r setup, include = FALSE}
source("setup/setup.R")
``````

This vignette deals with situations where you want to transform tables of your `dm` and then update an existing table or add a new table to the `dm`.
There are two straightforward approaches:

1. extract the tables relevant to the calculation, perform the necessary transformations, and (if needed) recombine the resulting table into a `dm`,
1. do all this within the `dm` object by zooming to a table and manipulating it while maintaining the key relations whenever possible.

Both approaches aim at maintaining the key relations whenever possible.
We will explore the second approach here. 
For the first approach, see `vignette("tech-dm-zoom")`.

## Enabling {dplyr}-workflow within a `dm`

"Zooming" to a table of a `dm` means:

- all information stored in the original `dm` is kept, including the originally zoomed table
- an object of class `dm_zoomed` is produced, presenting a view of the table for transformations
- you do not need to specify the table when calling `select()`, `mutate()` and other table manipulation functions

{dm} provides methods for many of the {dplyr}-verbs for a `dm_zoomed` which behave the way you are used to, affecting only the zoomed table and leaving the rest of the `dm` untouched.
When you are finished with transforming the table, there are three options to proceed:

1. use `dm_update_zoomed()` if you want to replace the originally zoomed table with the new table
1. use `dm_insert_zoomed()` if you are creating a new table for your `dm`
1. use `dm_discard_zoomed()` if you do not need the result and want to discard it

When employing one of the first two options, the resulting table in the `dm` will have all the primary and foreign keys available that could be tracked from the originally zoomed table.

## Examples

So much for the theory, but how does it look and feel?
To explore this, we once more make use of our trusted {nycflights13} data.

### Use case 1: Add a new column to an existing table

Imagine you want to have a column in `flights`, specifying if a flight left before noon or after.
Just like with {dplyr}, we can tackle this with `mutate()`.
Let us do this step by step:

```{r zoom}
library(dm)
library(dplyr)
flights_dm <- dm_nycflights13()
flights_dm
flights_zoomed <-
  flights_dm %>%
  dm_zoom_to(flights)
# The print output for a `dm_zoomed` looks very much like that from a normal `tibble`.
flights_zoomed

flights_zoomed_mutate <-
  flights_zoomed %>%
  mutate(am_pm_dep = if_else(dep_time < 1200, "am", "pm")) %>%
  # in order to see our changes in the output we use `select()` for reordering the columns
  select(year:dep_time, am_pm_dep, everything())

flights_zoomed_mutate

# To update the original `dm` with a new `flights` table we use `dm_update_zoomed()`:
updated_flights_dm <-
  flights_zoomed_mutate %>%
  dm_update_zoomed()
# The only difference in the `dm` print output is the increased number of columns
updated_flights_dm
# The schematic view of the data model remains unchanged
dm_draw(updated_flights_dm)
```

### Use case 2: Creation of a surrogate key

The same course of action could, for example, be employed to create a surrogate key for a table, a synthetic simple key that replaces a compound key.
We can do this for the `weather` table.

```{r}
library(tidyr)

weather_zoomed <-
  flights_dm %>%
  dm_zoom_to(weather)
weather_zoomed
# Maybe there is some hidden candidate for a primary key that we overlooked
enum_pk_candidates(weather_zoomed)
# Seems we have to construct a column with unique values
# This can be done by combining column `origin` with `time_hour`, if the latter
# is converted to a single time zone first; all within the `dm`:
weather_zoomed_mutate <-
  weather_zoomed %>%
  # first convert all times to the same time zone:
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  # paste together as character the airport code and the time
  unite("origin_slot_id", origin, time_hour_fmt) %>%
  select(origin_slot_id, everything())
# check if we the result is as expected:
enum_pk_candidates(weather_zoomed_mutate) %>% filter(candidate)
flights_upd_weather_dm <-
  weather_zoomed_mutate %>%
  dm_update_zoomed() %>%
  dm_add_pk(weather, origin_slot_id)
flights_upd_weather_dm
# creating the coveted FK relation between `flights` and `weather`
extended_flights_dm <-
  flights_upd_weather_dm %>%
  dm_zoom_to(flights) %>%
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  # need to keep `origin` as FK to airports, so `remove = FALSE`
  unite("origin_slot_id", origin, time_hour_fmt, remove = FALSE) %>%
  dm_update_zoomed() %>%
  dm_add_fk(flights, origin_slot_id, weather)
extended_flights_dm %>% dm_draw()
```


### Use case 3: Disentangle `dm`

If you look at the `dm` created by `dm_nycflights13(cycle = TRUE)`, you see that two columns of `flights` relate to one and the same table, `airports`.
One column stands for the departure airport and the other for the arrival airport.

```{r}
dm_draw(dm_nycflights13(cycle = TRUE))
```

This generates a cycle which leads to failures with many operations that only work on cycle-free data models, such as `dm_flatten_to_tbl()`, `dm_filter()` or `dm_wrap_to_tbl()`.
In such cases, it can be beneficial to "disentangle" the `dm` by duplicating the referred table.
One way to do this in the {dm}-framework is as follows:

```{r}
disentangled_flights_dm <-
  dm_nycflights13(cycle = TRUE) %>%
  # zooming and immediately inserting essentially creates a copy of the original table
  dm_zoom_to(airports) %>%
  # reinserting the `airports` table under the name `destination`
  dm_insert_zoomed("destination") %>%
  # renaming the originally zoomed table
  dm_rename_tbl(origin = airports) %>%
  # Key relations are also duplicated, so the wrong ones need to be removed
  dm_rm_fk(flights, dest, origin) %>%
  dm_rm_fk(flights, origin, destination)
dm_draw(disentangled_flights_dm)
```

In a future update, we will provide a more convenient way to "disentangle" `dm` objects, so that the individual steps will be done automatically.

### Use case 4: Add summary table to `dm`

Here is an example for adding a summary of a table as a new table to a `dm` (FK-relations are taken care of automatically):

```{r}
dm_with_summary <-
  flights_dm %>%
  dm_zoom_to(flights) %>%
  dplyr::count(origin, carrier) %>%
  dm_insert_zoomed("dep_carrier_count")
dm_draw(dm_with_summary)
```

### Use case 5: Joining tables

If you would like to join some or all of the columns of one table to another, you can make use of one of the `join`-methods for a `dm_zoomed`.
In addition to the usual arguments for the {dplyr}-joins, by supplying the `select` argument you can specify which columns of the RHS-table you want to be included in the join.
For the syntax please see the example below.
The LHS-table of a join is always the zoomed table.

```{r}
joined_flights_dm <-
  flights_dm %>%
  dm_zoom_to(flights) %>%
  # let's first reduce the number of columns of flights
  select(-dep_delay:-arr_delay, -air_time:-time_hour) %>%
  # in the {dm}-method for the joins you can specify which columns you want to add to the zoomed table
  left_join(planes, select = c(tailnum, plane_type = type)) %>%
  dm_insert_zoomed("flights_plane_type")
# this is how the table looks now
joined_flights_dm$flights_plane_type
# also here, the FK-relations are transferred to the new table
dm_draw(joined_flights_dm)
```

### Tip: Accessing the zoomed table

At each point, you can retrieve the zoomed table by calling `pull_tbl()` on a `dm_zoomed`.
To use our last example once more:

```{r}
flights_dm %>%
  dm_zoom_to(flights) %>%
  select(-dep_delay:-arr_delay, -air_time:-time_hour) %>%
  left_join(planes, select = c(tailnum, plane_type = type)) %>%
  pull_tbl()
```

### Possible pitfalls and caveats

1. Currently, not all {dplyr}-verbs have their own method for a `dm_zoomed` object, so be aware that in some cases it will still be necessary to resort to extracting one or more tables from a `dm` and reinserting a transformed version back into the `dm` object.
The supported functions are:
`group_by()`, `ungroup()`, `summarise()`, `mutate()`, `transmute()`, `filter()`, `select()`, `relocate()`, `rename()`, `distinct()`, `arrange()`, `slice()`, `left_join()`, `inner_join()`, `full_join()`, `right_join()`, `semi_join()`, and `anti_join()`.

1. The same is true for {tidyr}-functions.
Methods are provided for: `unite()` and `separate()`.

1. There might be situations when you would like the key relations to remain intact, but they are dropped nevertheless.
This is because a rigid logic is implemented, that does drop a key when its associated column is acted upon with e.g. a `mutate()` call.
In these cases, the key relations will need to be re-established after finishing with the manipulations.

1. For each implemented {dplyr}-verb, there is a logic for tracking key relations between the tables.
Up to {dm} version 0.2.4 we tried to track the columns in a very detailed manner. This has become increasingly difficult, especially with `dplyr::across()`. As of {dm} 0.2.5, we give more responsibility to the {dm} user:
Now those columns are tracked whose **names** remain in the resulting table.
Affected by these changes are the methods for: `mutate()`, `transmute()`, `distinct()`.
When using one of these functions, be aware that if you want to replace a key column with a column with a different content but of the same name, this column will automatically become a key column.

---
title: "Class dm and basic operations"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Technical: Class dm and basic operations}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---


``````{r setup, include = FALSE}
source("setup/setup.R")
``````

The goal of the {dm} package and the `dm` class that comes with it, is to make your life easier when you are dealing with data from several different tables.

Let's take a look at the `dm` class.

## Class `dm`

The `dm` class consists of a collection of tables and metadata about the tables, such as

- the names of the tables
- the names of the columns of the tables
- the primary and foreign keys of the tables to link the tables together
- the data (either as data frames or as references to database tables)

All tables in a `dm` must be obtained from the same data source;
csv files and spreadsheets would need to be imported to data frames in R.


## Examples of `dm` objects {#ex_dm}

There are currently three options available for creating a `dm` object.
The relevant functions for creating `dm` objects are:

1. `dm()`
2. `as_dm()`
3. `new_dm()`
4. `dm_from_con()`

To illustrate these options, we will now create the same `dm` in several different ways.
We can use the tables from the well-known {nycflights13} package.

### Pass the tables directly

Create a `dm` object directly by providing data frames to `dm()`:

```{r}
library(nycflights13)
library(dm)
dm(airlines, airports, flights, planes, weather)
```

### Start with an empty `dm`

Start with an empty `dm` object that has been created with `dm()` or `new_dm()`, and add tables to that object:

```{r}
library(nycflights13)
library(dm)
empty_dm <- dm()
empty_dm
dm(empty_dm, airlines, airports, flights, planes, weather)
```

### Coerce a list of tables

Turn a named list of tables into a `dm` with `as_dm()`:
```{r}
as_dm(list(
  airlines = airlines,
  airports = airports,
  flights = flights,
  planes = planes,
  weather = weather
))
```

### Turn tables from a `src` into a `dm`

Squeeze all (or a subset of) tables belonging to a `src` object into a `dm` using `dm_from_con()`:

```{r message=FALSE}
sqlite_con <- dbplyr::nycflights13_sqlite()

flights_dm <- dm_from_con(sqlite_con)
flights_dm
```


The function `dm_from_con(con, table_names = NULL)` includes all available tables on a source in the `dm` object.
This means that you can use this, for example, on a postgres database that you access via `DBI::dbConnect(RPostgres::Postgres())` (with the appropriate arguments `dbname`, `host`, `port`, ...), to produce a `dm` object with all the tables on the database.

### Low-level construction

Another way of creating a `dm` object is calling `new_dm()` on a list of `tbl` objects:

```{r}
base_dm <- new_dm(list(
  airlines = airlines,
  airports = airports,
  flights = flights,
  planes = planes,
  weather = weather
))
base_dm
```

This constructor is optimized for speed and does not perform integrity checks.
Use with caution, validate using `dm_validate()` if necessary.

```{r}
dm_validate(base_dm)
```

## Access tables

We can get the list of tables with `dm_get_tables()` and the `src` object with `dm_get_con()`.

In order to pull a specific table from a `dm`, use:

```{r}
flights_dm[["airports"]]
```

But how can we use {dm}-functions to manage the primary keys of the tables in a `dm` object?

## Primary keys of `dm` objects {#pk}

Some useful functions for managing primary key settings are:

1. `dm_add_pk()`
1. `dm_get_all_pks()`
1. `dm_rm_pk()`
1. `dm_enum_pk_candidates()`

If you created a `dm` object according to the examples in ["Examples of `dm` objects"](#ex_dm), your object does not yet have any primary keys set.
So let's add one.

We use the `nycflights13` tables, i.e. `flights_dm` from above.

```{r}
dm_has_pk(flights_dm, airports)
flights_dm_with_key <- dm_add_pk(flights_dm, airports, faa)
flights_dm_with_key
```

The `dm` now has a primary key:

```{r}
dm_has_pk(flights_dm_with_key, airports)
```

To get an overview over all tables with primary keys, use `dm_get_all_pks()`:

```{r}
dm_get_all_pks(flights_dm_with_key)
```

Remove a primary key:

```{r}
dm_rm_pk(flights_dm_with_key, airports) %>%
  dm_has_pk(airports)
```

If you still need to get to know your data better, and it is already available in the form of a `dm` object, you can use the  `dm_enum_pk_candidates()` function in order to get information about which columns of the table are unique keys:

```{r}
dm_enum_pk_candidates(flights_dm_with_key, airports)
```

The `flights` table does not have any one-column primary key candidates:

```{r}
dm_enum_pk_candidates(flights_dm_with_key, flights) %>% dplyr::count(candidate)
```

`dm_add_pk()` has a `check` argument.
If set to `TRUE`, the function checks if the column of the table given by the user is unique.
For performance reasons, the default is `check = FALSE`.
See also [dm_examine_constraints()] for checking all constraints in a `dm`.

```{r error = TRUE}
try(
  dm_add_pk(flights_dm, airports, tzone, check = TRUE)
)
```


## Foreign keys

Useful functions for managing foreign key relations include:

1. `dm_add_fk()`
1. `dm_get_all_fks()`
1. `dm_rm_fk()`
1. `dm_enum_fk_candidates()`

Now it gets (even more) interesting: we want to define relations between different tables.
With the `dm_add_fk()` function you can define which column of which table points to another table's column.

This is done by choosing a foreign key from one table that will point to a primary key of another table.
The primary key of the referred table must be set with `dm_add_pk()`.
`dm_add_fk()` will find the primary key column of the referenced table by itself and make the indicated column of the child table point to it.

```{r}
flights_dm_with_key %>% dm_add_fk(flights, origin, airports)
```

This will throw an error:

```{r error=TRUE}
try(
  flights_dm %>% dm_add_fk(flights, origin, airports)
)
```

Let's create a `dm` object with a foreign key relation to work with later on:

```{r}
flights_dm_with_fk <- dm_add_fk(flights_dm_with_key, flights, origin, airports)
```

What if we tried to add another foreign key relation from `flights` to `airports` to the object? Column `dest` might work, since it also contains airport codes:

```{r error=TRUE}
try(
  flights_dm_with_fk %>% dm_add_fk(flights, dest, airports, check = TRUE)
)
```

Checks are opt-in and executed only if `check = TRUE`.
You can still add a foreign key with the default `check = FALSE`.
See also `dm_examine_constraints()` for checking all constraints in a `dm`.

Get an overview of all foreign key relations with`dm_get_all_fks()`:

```{r}
dm_get_all_fks(dm_nycflights13(cycle = TRUE))
```

Remove foreign key relations with `dm_rm_fk()` (parameter `columns = NULL` means that all relations will be removed, with a message):

```{r error=TRUE}
try(
  flights_dm_with_fk %>%
    dm_rm_fk(table = flights, column = dest, ref_table = airports) %>%
    dm_get_all_fks(c(flights, airports))
)

flights_dm_with_fk %>%
  dm_rm_fk(flights, origin, airports) %>%
  dm_get_all_fks(c(flights, airports))

flights_dm_with_fk %>%
  dm_rm_fk(flights, columns = NULL, airports) %>%
  dm_get_all_fks(c(flights, airports))
```

Since the primary keys are defined in the `dm` object, you do not usually need to provide the referenced column name of `ref_table`.

Another function for getting to know your data better (cf. `dm_enum_pk_candidates()` in ["Primary keys of `dm` objects"](#pk)) is `dm_enum_fk_candidates()`. Use it to get an overview over foreign key candidates that point from one table to another:

```{r}
dm_enum_fk_candidates(flights_dm_with_key, weather, airports)
```

---
title: "Relational data with dm: Cheat sheet"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{cheatsheet}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
resource_files:
  - cheatsheet
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  echo = FALSE
)
```

[For printing](cheatsheet/cheatsheet-printable.html)

<div class="page1">
> If you have more than one table, use dm.

<div class="page1col1">
## <i class="cheatsheet-icon fa-solid fa-circle-plus"></i> Create dm...

### <i class="cheatsheet-icon fa-solid fa-database"></i> from database: `dm_from_con()`

```{r}
get_src <- function(src) {
  if (!nzchar(Sys.getenv("CHEATSHEET"))) {
    file.path("cheatsheet-figures", src)
  } else {
    sprintf("%s/%s", "cheatsheet-figures", src)
  }
}
get_svg <- function(path, alt) {
  svg <- xml2::read_xml(path)
  xml2::xml_remove(xml2::xml_find_all(svg, "//comment()"))
  xml2::xml_attr(svg, "class") <- "cheatsheet-svg"
  xml2::xml_attr(svg, "role") <- "img"
  xml2::xml_attr(svg, "aria-describedby") <- sprintf("svgDescription-%s", basename(path))
  xml2::xml_add_child(
    svg,
    .value = "description",
    id = sprintf("svgDescription-%s", basename(path)),
    alt,
    .where = 0
  )
  # to drop the XML declaration that messes things up
  temp_file <- withr::local_tempfile()
  xml2::write_xml(svg, file = temp_file, options = "format_whitespace")
  lines <- readLines(temp_file)
  lines <- lines[!grepl("^<\\?", lines)]
  lines <- lines[!grepl("^<!", lines)]
  glue::glue_collapse(lines, sep = "")
}
create_figure <- function(class, src, alt, caption) {
  if (!nzchar(Sys.getenv("CHEATSHEET"))) {
    path <- purrr::map_chr(src, get_src)
    svgs <- purrr::map2_chr(path, alt, get_svg)
    svgs <- paste0(rev(svgs), collapse = "")

    cat("<div class='cheatsheet-div'>")
    html <- sprintf('<div class="cheatsheet-caption">%s</div>', commonmark::markdown_html(caption))
    cat(html)
    cat(svgs)
    cat("</div>")
  } else {
    create_img <- function(alt, src) {
      sprintf('<img alt="%s" src="%s"/>', alt, src)
    }
    src <- purrr::map_chr(src, get_src)
    imgs <- purrr::map2_chr(alt, src, create_img)
    imgs <- paste(imgs, collapse = "")
    html <- sprintf(
      '<figure class="%s">%s<figcaption>%s</figcaption></figure>',
      class, imgs, commonmark::markdown_html(caption)
    )
    cat(html)
  }
}
```


```{r, results='asis'}
create_figure(
  class = "right",
  alt = "diagram showing the transformation from left a database icon to right a dm object represented as three tables in a blue frame",
  src = "dm_from_con.svg",
  caption = "```r\ncon <- DBI::dbConnect(...)
dm_from_con(con)\n```\n"
)
```

### <i class="cheatsheet-icon fa-solid fa-table"></i> from data frames: `dm(df1, df2, ...)`

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "diagram showing the transformation from left three data frames to right a dm object represented as three tables in a blue frame",
  src = "dm.svg",
  caption = "```r\ndm(df1, df2, df3)\n```\n"
)
```
         
### <i class="cheatsheet-icon fa-solid fa-diagram-project"></i> from dm: `dm(dm1, df1...)`

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "diagram showing the transformation from left a dm with a data frame below to right a single dm object represented as three tables in a blue frame",
  src = "dm-df.svg",
  caption = "```r\ndm(dm1, df1)\n```\n"
)
```

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "diagram showing the transformation from left two dm objects to right a single dm object represented as three tables in a blue frame",
  src = "dm-dm.svg",
  caption = "```r\ndm(dm1, dm2)\n```\n"
)
```        

<hr class="d-none d-print-block"></hr>

### Add keys: `dm_add_pk()`, `dm_add_fk()`

**Automatic** for MariaDB, SQL Server, Postgres, and others.

#### <i class="cheatsheet-icon fa-solid fa-key"></i> Primary keys

**Identify** potential primary keys:

- `dm_enum_pk_candidates()`: columns, candidate, why.

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "left a single dm object represented as three tables in a blue frame, right the same but with markers for primary keys",
  src = "primary-keys.svg",
  caption = "**Add** primary keys:\n\n```r\ndm1 |>
  dm_add_pk(table, columns)\n```\n"
)
```        


**Identify** potential foreign keys:

- `dm_enum_fk_candidates()`: columns, candidate, why.

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "left a single dm object represented as three tables in a blue frame, right the same but with an arrow to indicate a foreign key relationship",
  src = "fk.svg",
  caption = "**Add** foreign keys:\n\n```r\ndm1 |>
  dm_add_fk(table, column)\n```\n"
)
``` 
</div>
<div class="page1col2">
::: {.important-div}
### <i class="cheatsheet-icon fa-solid fa-circle-info"></i> dm objects: relational data models

The dm package provides a **grammar of relational data models**. 
It helps maintain **referential integrity**.

A dm behaves like a list of tables (data frames or lazy tables) capturing
**relationships** between the tables.
:::

::: {.alert .alert-primary role="alert" style="display:grid;place-items:center;"}

<i class="cheatsheet-icon fa-solid fa-arrow-up-right-from-square"></i> Shiny app: `dm_gui(dm = dm1)`

```{r, results='asis'}
cat(
  '<img
    alt="Screenshot of the Shiny app provided by dm. It has three parts. On the left is a diagram view of the dm that one can click to, for instance, select tables or change their colors. On the right, there is the generated code as well as a menu to control primary and foreign keys."
    src="cheatsheet-figures/shiny.png"
    style="max-width: 75%; margin:auto;"
  />'
)
```
            
:::

<hr class="d-none d-print-block"></hr>

## <i class="cheatsheet-icon fa-solid fa-filter"></i> Resize dm

### Select tables: `dm_select_tbl(dm1, ...)`

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "diagram showing two dm objects, the one on the right has one table less",
  src = "select_tbl.svg",
  caption = "```r\ndm1 |>
  dm_select_tbl(-df3)\n```\n"
)
``` 

### Rename tables: `dm_rename_tbl(dm1, ...)`

### Select columns: `dm_select(dm1, table, ...)`

Automatic update of dm meta-information and table relations.


```{r, results='asis'}
create_figure(
  class = "right",
  alt = "diagram showing two dm objects, the third table of the one on the right has two colmns less",
  src = "dm_select.svg",
  caption = "```r\ndm1 |>
  dm_select(df3, -c3, -c4)\n```\n"
)
```

### Rename columns: `dm_rename(dm1, table, ...)`

### Filter rows: `dm_filter(dm1, table = (pred))`

Filter rows in the table where the condition is defined, 
but also all directly/indirectly connected tables.

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "diagram showing two dm objects, the one on the right has tables with less rows",
  src = "filter.svg",
  caption = '```r\ndm1 |>
  dm_filter(df3 = (x == "val"))\n```\n'
)
```
</div>
<div class="page1col3">

## <i class="cheatsheet-icon fa-solid fa-eye"></i> Visualize dm: `dm_draw()`

### Control diagram level of detail: display...

- Only keys (default): `dm_draw(view_type = "keys_only")`.
- All variables: `dm_draw(view_type = "all")`.
- Only table names: `dm_draw(view_type = "title_only")`.

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "Diagram with linked rectangles (tables). Only table names.",
  src = "draw-title_only.svg",
  caption = '```r\ndm |>
  dm_draw(
    view_type = "title_only",
    rankdir = "TB"
  )\n```\n'
)
``` 

### Control diagram scope

To visualize fewer tables first use `dm_select_tbl()`.

### Control diagram colors: `dm_set_colors()`

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "Diagram with linked rectangles (tables). The flight table is pink, the airlines and airpots tables whose name start with 'air', are orange.",
  src = "df-pk-fk-col.svg",
  caption = '```r\ndm |>
  dm_set_colors(
    pink = flights,
    orange = starts_with("air")
  ) |>
  dm_draw()\n```\n'
)
``` 

<hr class="d-none d-print-block"></hr>
          
## <i class="cheatsheet-icon fa-solid fa-check"></i> Data checks

### `dm_examine_constraints()`

tibble with information about which key constraints are met or violated.

### `dm_examine_cardinalities()`

tibble with information about the cardinality of the foreign keys constraints.

### `check_key(df, col1)`

returns an error if not an unique key.

### `check_subset(df1, df2)`

returns an error if `df1` is not a subset of `df2`.

### `check_set_equality(df1, df2)`

returns an error if `df1` and `df2` are not the same sets.

<hr class="d-none d-print-block"></hr>

### <i class="cheatsheet-icon fa-solid fa-tags"></i> Fix column names: `dm_disambiguate_cols()`

`dm_disambiguate_cols(dm1)` ensures that all columns in a dm have unique names.
</div>
</div>
<div class="page2">
<div class="page2col1">

## <i class="cheatsheet-icon fa-solid fa-table"></i> Transform dm into tibble

### Wide tibble: Cascade joins with `dm_flatten_to_tbl()`

Only direct neighbours: `dm_flatten_to_tbl()`

```{r, results='asis'}
create_figure(
  class = "big",
  alt = "On the left a dm object where table A is linked to table B and table C. Table D is linked to table B. On the right a table corresponding to a join of table A, B and C. Not D as it is not a direct neighboor",
  src = "flatten.svg",
  caption = "```r\ndm1 |>
  dm_flatten_to_tbl(
.start = df1
)\n```\n"
)
```

All neighbours: `dm_flatten_to_tbl(.recursive = TRUE)`

```{r, results='asis'}
create_figure(
  class = "big",
  alt = "On the left a dm object where table A is linked to table B and table C. Table D is linked to table B. On the right a table corresponding to a join of table A, B, C and D as they can all be reached from table A albeit indirectly.",
  src = "squash.svg",
  caption = "```r\ndm1 |>
  dm_flatten_to_tbl(
.start = df1,
.recursive = TRUE
)\n```\n"
)
```


### Single tibble dm: `dm_wrap_tbl()`

Parent tables are packed &#8212; `dm_pack_tbl()`.

Child tables are nested &#8212; `dm_nest_tbl()`.

```{r, results='asis'}
create_figure(
  class = "big",
  alt = "On the left a dm object (tables with arrows in a frame) where a green table is linked to a parent blue table and a child purple table. On the right another dm object with a single table in a frame. The blue parent table became a packed column in the green table (a column that contains a data frame) and the purple child table became a nested table (a table in each cell).",
  src = "dm_wrap_tbl.svg",
  caption = "```r\ndm1 |>
  dm_wrap_tbl(
root = green_df
)\n```\n"
)
```

### Retrieve one table of the dm: `pull_tbl()`

```{r, results='asis'}
create_figure(
  class = "big",
  alt = "On the left a dm object with three tables, on the right only a table. The primary key information has been kept which is represented by the first column of that table having kept its dark background.",
  src = "pull-keyed.svg",
  caption = "```r\ndm1 |>
  pull_tbl(
dm1,
green_df,
keyed = TRUE
)\n```\n"
)
```

```{r, results='asis'}
create_figure(
  class = "big",
  alt = "On the left a dm object with three tables. There is a magnifying glass on the green table. On the right only the green table.",
  src = "pull2.svg",
  caption = "If the dm is zoomed, retrieve zoomed table automatically.\n\n```r\ndm1 |>
  dm_zoom_to(green_df) |>
pull_tbl()\n```\n"
)
```
</div>
<div class="page2col2">
## <i class="cheatsheet-icon fa-solid fa-screwdriver-wrench"></i> Mutate, create, analyze tables

### <i class="cheatsheet-icon fa-solid fa-hammer"></i> Method 1: deconstruct and reconstruct

1. `dm_get_tables(keyed = TRUE)`: <br />
    to keep information on primary and foreign keys).

```{r, results='asis'}
create_figure(
  class = "",
  alt = "diagram showing the transformation from left a dm object to right a named list of the same table, instead of a solid border frame on the right the frame has a dotted border",
  src = "deconstruct.svg",
  caption = "```r\ndm_tbl <- dm1 |>
  dm_get_tables(keyed = TRUE)\n```\n"
)
```

2. tidyverse pipeline on the table of interest.

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "on the left a table, on the right the same table with one more column",
  src = "mutate.svg",
  caption = "```r\nnew_table1 <- dm_tbl$table1 |>
  mutate(...)\n```\n"
)
```

3. Optional: update the dm object:

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "on the left a dm of three tables, on the right the same dm but one of the tables has one column more",
  src = "reconstruct.svg",
  caption = "```r\ndm1 |>
  dm_select_tbl(-table1) |>
  dm(table1 = new_table1)\n```\n"
)
```

### <i class="cheatsheet-icon fa-solid fa-magnifying-glass"></i> Method 2: zoom

1. `dm_zoom_to()`: Zoom on a table.

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "On the left a dm object with three tables. On the right the same object but there is now a magnifying glass on the green table.",
  src = "zoom1.svg",
  caption = "```r\nzoomed_dm1 <- dm1 |>
  dm_zoom_to(green_df)\n```\n"
)
```

2. tidyverse pipeline (`mutate()`, etc.).

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "On the left a dm object with three tables including a green table with a magnifying glass on it. On the right the same object but there is now a second green table superimposed, with one more column compared to the green table in the dm.",
  src = "zoom2.svg",
  caption = "```r\nzoomed_dm2 <- zoomed_dm1 |>
  mutate(var = thing)\n```\n"
)
```

3. `dm_update_zoomed()` (replace) /
    `dm_insert_zoomed()`

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "On the left a dm object featuring a second green table superimposed, with one more column compared to the green table in the dm. On the right the magnifying glass and second green table disappeared. Instead the green table in the dm has the new column.",
  src = "zoom3.svg",
  caption = "```r\ndm3 <- zoomed_dm2 |>
  dm_update_zoomed()\n```\n"
)
```
    
</div>
<div class="page2col3">
## <i class="cheatsheet-icon fa-solid fa-database"></i> Modify database source of a dm

### <i class="cheatsheet-icon fa-solid fa-download"></i> Export dm object to database:  `copy_dm_to()`

Need a **database connection &#8212; `DBI::dbConnect()`**.

```{r, results='asis'}
create_figure(
  class = "right",
  alt = "On the left a dm object, with an arrow pointing to on the right a database icon",
  src = "copy_dm.svg",
  caption = "```r\ncon <- DBI::dbConnect(...)
  # Persistent tables:
persistent_dm <- copy_dm_to(
  con,
  dm1,
  temporary = FALSE
)

DBI::dbDisconnect(con)\n```\n"
)
```

## <i class="cheatsheet-icon fa-solid fa-plus-minus"></i> Insert, update or remove rows in a dm

Methods:

- `dm_rows_insert(dm1, dm2)`: adds new rows
- `dm_rows_update(dm1, dm2)`: changes values in rows
- `dm_rows_patch(dm1, dm2)`: fills in missing values
- `dm_rows_upsert(dm1, dm2)`: adds new or changes rows
- `dm_rows_delete(dm1, dm2)`: deletes rows

```r
dm1 |>
  dm_rows_insert(dm2, in_place = FALSE)
```

```{r, results='asis'}
create_figure(
  class = "left",
  alt = c(
    "Equal sign then a dm object where the blue table has now more rows.",
    "On the left a dm object with 3 tables. On the right, a similar dm where one of the tables, the blue one, has more rows. Between the two is a plus sign"
  ),
  src = c("insert2.svg", "insert1.svg"),
  caption = NULL
)
```

```r
dm1 |>
  dm_rows_insert(dm2, in_place = TRUE)
```

```{r, results='asis'}
create_figure(
  class = "left",
  alt = c(
    "Database icon.",
    "On the left a dm object with 3 tables, and a database icon superimposed. On the right, a similar dm where one of the tables, the blue one, has more rows. Between the two is a plus sign"
  ),
  src = c("insert4.svg", "insert3.svg"),
  caption = NULL
)
```

A dm is immutable, except with

- <i class="cheatsheet-icon fa-solid fa-check"></i> these functions AND
- <i class="cheatsheet-icon fa-solid fa-check"></i> a mutable backend (database) AND
- <i class="cheatsheet-icon fa-solid fa-check"></i> `in_place = TRUE`.
</div>
</div>

```{r, echo=FALSE, message=FALSE, warning=FALSE, results='hide'}
source(system.file("make-cheatsheet.R", package = "dm"))
```

Storing all data related to a problem in a single table or data frame ("the dataset") can result in many repetitive values. Separation into multiple tables helps data quality but requires "merge" or "join" operations. {dm} is a new package that fills a gap in the R ecosystem: it makes working with multiple tables just as easy as working with a single table.

A "data model" consists of tables (both the definition and the data), and primary and foreign keys. The {dm} package combines these concepts with data manipulation powered by the tidyverse: entire data models are handled in a single entity, a "dm" object.

Three principal use cases for {dm} can be identified:

1. When you consume a data model, {dm} helps access and manipulate a dataset consisting of multiple tables (database or local data frames) through a consistent interface.

2. When you use a third-party dataset, {dm} helps normalizing the data to remove redundancies as part of the cleaning process.

3. To create a relational data model, you can prepare the data using R and familiar tools and seamlessly export to a database.

The presentation revolves around these use cases and shows a few applications. The {dm} package is available on GitHub and will be submitted to CRAN in early February.

---
title: "Getting started with dm"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{First read: Getting started with dm}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---

``````{r setup, include = FALSE}
source("setup/setup.R")
``````


In this article, we're going show you how easy it is to move from connecting to the database holding your data to producing the results you need.
It's meant to be a quick and friendly introduction to {dm}, so it is low on details and caveats.
Links to detailed documentation are provided at the end.
(If your data is in data frames instead of a database and you're in a hurry, jump over to `vignette("howto-dm-df")`.)

## Creating a dm object

dm objects can be created from individual tables or loaded directly from a relational data model on an RDBMS (relational database management system).

For this demonstration, we're going to work with a model hosted on a public server.
The first thing we need is a connection to the RDBMS hosting the data.

``````{r connect, eval = FALSE}
library(RMariaDB)

fin_db <- dbConnect(
  MariaDB(),
  username = "guest",
  password = "ctu-relational",
  dbname = "Financial_ijs",
  host = "relational.fel.cvut.cz"
)
``````

``````{r connect-real, echo = FALSE}
library(RMariaDB)

fin_db <- dm:::financial_db_con()
``````

We create a dm object from an RDBMS using `dm_from_con()`, passing in the connection object we just created as the first argument.

``````{r load-full}
library(dm)

fin_dm <- dm_from_con(fin_db)
fin_dm
``````

The dm object interrogates the RDBMS for table and column information, and primary and foreign keys.
Currently, primary and foreign keys are only available from `r glue::glue_collapse(dm:::schema_supported_dbs()[["db_name"]], sep = ", ", last = " and ")`.


## Selecting tables

The dm object can be accessed like a named list of tables:

```{r names}
names(fin_dm)
fin_dm$loans
dplyr::count(fin_dm$trans)
```

Additionally, most `dm` functions are [pipe-friendly](https://r4ds.had.co.nz/pipes.html) and support [tidy evaluation](https://adv-r.hadley.nz/metaprogramming.html).
We can use `[` or the `dm_select_tbl()` verb to derive a smaller dm with the `loans`, `accounts`, `districts` and `trans` tables:

```{r select}
fin_dm_small <- fin_dm[c("loans", "accounts", "districts", "trans")]
fin_dm_small <-
  fin_dm %>%
  dm_select_tbl(loans, accounts, districts, trans)
```

## Linking tables by adding keys

In many cases, `dm_from_con()` already returns a dm with all keys set.
If not, dm allows us to define primary and foreign keys ourselves.
For this, we use `learn_keys = FALSE` to obtain a `dm` object with only the tables.

``````{r load}
library(dm)

fin_dm_small <-
  dm_from_con(fin_db, learn_keys = FALSE) %>%
  dm_select_tbl(loans, accounts, districts, trans)
``````

In our data model, `id` columns uniquely identify records in the `accounts` and `loans` tables, and was used as a primary key.
A primary key is defined with `dm_add_pk()`.
Each loan is linked to one account via the `account_id` column in the `loans` table, the relationship is established with `dm_add_fk()`.

```{r keys}
fin_dm_keys <-
  fin_dm_small %>%
  dm_add_pk(table = accounts, columns = id) %>%
  dm_add_pk(loans, id) %>%
  dm_add_fk(table = loans, columns = account_id, ref_table = accounts) %>%
  dm_add_pk(trans, id) %>%
  dm_add_fk(trans, account_id, accounts) %>%
  dm_add_pk(districts, id) %>%
  dm_add_fk(accounts, district_id, districts)
```

## Visualizing a data model

Having a diagram of the data model is the quickest way to verify we're on the right track.
We can display a visual summary of the dm at any time.
The default is to display the table name, any defined keys, and their links to other tables.

Visualizing the dm in its current state, we can see the keys we have created and how they link the tables together.
Color guides the eye.

```{r visualize_keys}
fin_dm_keys %>%
  dm_set_colors(darkgreen = c(loans, accounts), darkblue = trans, grey = districts) %>%
  dm_draw()
```

## Accessing a data model as a table

If we want to perform modeling or analysis on this relational model, we need to transform it into a tabular format that R functions can work with.
With the argument `recursive = TRUE`, `dm_flatten_to_tbl()` will automatically follow foreign keys across tables to gather all the available columns into a single table.

```{r squash}
fin_dm_keys %>%
  dm_flatten_to_tbl(loans, .recursive = TRUE)
```

Apart from the rows printed above, no data has been fetched from the database.
Use `select()` to reduce the number of columns fetched, and `collect()` to retrieve the entire result for local processing.

```{r model}
loans_df <-
  fin_dm_keys %>%
  dm_flatten_to_tbl(loans, .recursive = TRUE) %>%
  select(id, amount, duration, A3) %>%
  collect()

model <- lm(amount ~ duration + A3, data = loans_df)

model
```


## Operations on table data within a dm

We don't need to take the extra step of exporting the data to work with it.
Through the dm object, we have complete access to dplyr's data manipulation verbs.
These operate on the data within individual tables.

To work with a particular table we use `dm_zoom_to()` to set the context to our chosen table.
Then we can perform any of the dplyr operations we want.

```{r zoom}
fin_dm_total <-
  fin_dm_keys %>%
  dm_zoom_to(loans) %>%
  summarize(.by = account_id, total_amount = sum(amount, na.rm = TRUE)) %>%
  dm_insert_zoomed("total_loans")

fin_dm_total$total_loans
```

Note that, in the above example, we use `dm_insert_zoomed()` to add the results as a new table to our data model.
This table is temporary and will be deleted when our session ends.
If you want to make permanent changes to your data model on an RDBMS, please see the "Persisting results" section in `vignette("howto-dm-db")`.


## Checking constraints

It's always smart to check that your data model follows its specifications.
When building our own model or changing existing models by adding tables or keys, it is even more important that the new model is validated.

`dm_examine_constraints()` checks all primary and foreign keys and reports if they violate their expected constraints.

```{r constraints}
fin_dm_total %>%
  dm_examine_constraints()
```

For more on constraint checking, including cardinality, finding candidate columns for keys, and normalization, see `vignette("tech-dm-low-level")`.

``````{r disconnect, echo = FALSE, results = "hide"}
dbDisconnect(fin_db)
``````

## Next Steps

Now that you have been introduced to the basic operation of dm, the next step is to learn more about the dm methods that your particular use case requires.

Is your data in an RDBMS? Then move on to `vignette("howto-dm-db")` for a more detailed look at working with an existing relational data model.

If your data is in data frames, then you want to read `vignette("howto-dm-df")` next.

If you would like to know more about relational data models in order to get the most out of dm, check out `vignette("howto-dm-theory")`.

If you're familiar with relational data models, but want to know how to work with them in dm, then any of `vignette("tech-dm-join")`, `vignette("tech-dm-filter")`, or `vignette("tech-dm-zoom")` is a good next step.


## Standing on the shoulders of giants

The {dm} package follows the tidyverse principles:

- `dm` objects are immutable (your data will never be overwritten in place)
- most functions used on `dm` objects are pipeable (i.e., return new `dm` or table objects)
- tidy evaluation is used (unquoted function arguments are supported)

The {dm} package builds heavily upon the [{datamodelr} package](https://github.com/bergant/datamodelr), and upon the [tidyverse](https://tidyverse.org/).
We're looking forward to a good collaboration!

The [{polyply} package](https://github.com/russHyde/polyply) has a similar intent with a slightly different interface.

The [{data.cube} package](https://github.com/jangorecki/data.cube) has quite the same intent using `array`-like interface.

Articles in the [{rquery} package](https://github.com/WinVector/rquery) discuss [join controllers](https://github.com/WinVector/rquery/blob/master/extras/JoinController.md) and [join dependency sorting](https://github.com/WinVector/rquery/blob/master/extras/DependencySorting.md), with the intent to move the declaration of table relationships from code to data.

The [{tidygraph} package](https://github.com/thomasp85/tidygraph) stores a network as two related tables of `nodes` and `edges`, compatible with {dplyr} workflows.

In object-oriented programming languages, [object-relational mapping](https://en.wikipedia.org/wiki/Object-relational_mapping) is a similar concept that attempts to map a set of related tables to a class hierarchy.

dm 1.0.10

## R CMD check results

- [x] Checked locally, R 4.3.2
- [x] Checked on CI system, R 4.3.2
- [x] Checked on win-builder, R devel

## Current CRAN check results

- [x] Checked on 2024-01-21, no problems found.

<!-- Generated by galley: do not edit by hand -->

The goal of the {dm} package and the `dm` class that comes with it, is
to make your life easier when you are dealing with data from several
different tables.

Let’s take a look at the `dm` class.

## Class `dm`

The `dm` class consists of a collection of tables and metadata about the
tables, such as

-   the names of the tables
-   the names of the columns of the tables
-   the primary and foreign keys of the tables to link the tables
    together
-   the data (either as data frames or as references to database tables)

All tables in a `dm` must be obtained from the same data source; csv
files and spreadsheets would need to be imported to data frames in R.

## Examples of `dm` objects

There are currently three options available for creating a `dm` object.
The relevant functions for creating `dm` objects are:

1.  `dm()`
2.  `as_dm()`
3.  `new_dm()`
4.  `dm_from_con()`

To illustrate these options, we will now create the same `dm` in several
different ways. We can use the tables from the well-known {nycflights13}
package.

### Pass the tables directly

Create a `dm` object directly by providing data frames to `dm()`:

``` r
library(nycflights13)
library(dm)
dm(airlines, airports, flights, planes, weather)
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 0
#> Foreign keys: 0
```

### Start with an empty `dm`

Start with an empty `dm` object that has been created with `dm()` or
`new_dm()`, and add tables to that object:

``` r
library(nycflights13)
library(dm)
empty_dm <- dm()
empty_dm
#> dm()
dm(empty_dm, airlines, airports, flights, planes, weather)
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 0
#> Foreign keys: 0
```

### Coerce a list of tables

Turn a named list of tables into a `dm` with `as_dm()`:

``` r
as_dm(list(airlines = airlines,
           airports = airports,
           flights = flights,
           planes = planes,
           weather = weather))
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 0
#> Foreign keys: 0
```

### Turn tables from a `src` into a `dm`

Squeeze all (or a subset of) tables belonging to a `src` object into a
`dm` using `dm_from_con()`:

``` r
sqlite_con <- dbplyr::remote_con(dbplyr::nycflights13_sqlite())

flights_dm <- dm_from_con(sqlite_con)
flights_dm
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  sqlite 3.35.5 [/tmp/RtmpGalley/nycflights13.sqlite]
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `sqlite_stat1`, … (7 total)
#> Columns: 62
#> Primary keys: 0
#> Foreign keys: 0
```

The function `dm_from_con(con, table_names = NULL)` includes all
available tables on a source in the `dm` object. This means that you can
use this, for example, on a Postgres database that you access via
`DBI::dbConnect(RPostgres::Postgres())` (with the appropriate arguments `dbname`, `host`,
`port`, …), to produce a `dm` object with all the tables on the
database.

### Low-level construction

Another way of creating a `dm` object is calling `new_dm()` on a list of
`tbl` objects:

``` r
base_dm <- new_dm(list(trees = trees, mtcars = mtcars))
base_dm
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `trees`, `mtcars`
#> Columns: 14
#> Primary keys: 0
#> Foreign keys: 0
```

This constructor is optimized for speed and does not perform integrity
checks. Use with caution, validate using `dm_validate()` if necessary.

``` r
dm_validate(base_dm)
```

## Access tables

We can get the list of tables with `dm_get_tables()` and the `src`
object with `dm_get_src()`.

In order to pull a specific table from a `dm`, use:

``` r
tbl(flights_dm, "airports")
#> # Source:   table<`airports`> [?? x 8]
#> # Database: sqlite 3.35.5 [/tmp/RtmpGalley/nycflights13.sqlite]
#>    faa   name                    lat    lon   alt    tz dst   tzone        
#>    <chr> <chr>                 <dbl>  <dbl> <dbl> <dbl> <chr> <chr>        
#>  1 04G   Lansdowne Airport      41.1  -80.6  1044    -5 A     America/New_…
#>  2 06A   Moton Field Municipa…  32.5  -85.7   264    -6 A     America/Chic…
#>  3 06C   Schaumburg Regional    42.0  -88.1   801    -6 A     America/Chic…
#>  4 06N   Randall Airport        41.4  -74.4   523    -5 A     America/New_…
#>  5 09J   Jekyll Island Airport  31.1  -81.4    11    -5 A     America/New_…
#>  6 0A9   Elizabethton Municip…  36.4  -82.2  1593    -5 A     America/New_…
#>  7 0G6   Williams County Airp…  41.5  -84.5   730    -5 A     America/New_…
#>  8 0G7   Finger Lakes Regiona…  42.9  -76.8   492    -5 A     America/New_…
#>  9 0P2   Shoestring Aviation …  39.8  -76.6  1000    -5 U     America/New_…
#> 10 0S9   Jefferson County Intl  48.1 -123.    108    -8 A     America/Los_…
#> # … with more rows
```

But how can we use {dm}-functions to manage the primary keys of the
tables in a `dm` object?

## Primary keys of `dm` objects

Some useful functions for managing primary key settings are:

1.  `dm_add_pk()`
2.  `dm_has_pk()`
3.  `dm_get_pk()`
4.  `dm_rm_pk()`
5.  `dm_enum_pk_candidates()`
6.  `dm_get_all_pks()`

If you created a `dm` object according to the examples in [“Examples of
`dm` objects”](#ex_dm), your object does not yet have any primary keys
set. So let’s add one.

`dm_add_pk()` has an option to check if the column of the table given by
the user is a unique key; for performance reasons, the check will not be
executed unless requested. We use the `nycflights13` tables,
i.e. `flights_dm` from above.

``` r
dm_has_pk(flights_dm, airports)
#> [1] FALSE
flights_dm_with_key <- dm_add_pk(flights_dm, airports, faa)
flights_dm_with_key
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  sqlite 3.35.5 [/tmp/RtmpGalley/nycflights13.sqlite]
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `sqlite_stat1`, … (7 total)
#> Columns: 62
#> Primary keys: 1
#> Foreign keys: 0
```

The `dm` now has a primary key. Let’s check:

``` r
dm_has_pk(flights_dm_with_key, airports)
#> [1] TRUE
```

Get the name of the column that is marked as primary key of the table:

``` r
dm_get_pk(flights_dm_with_key, airports)
#> <list_of<character>[1]>
#> [[1]]
#> [1] "faa"
```

Remove a primary key:

``` r
dm_rm_pk(flights_dm_with_key, airports) %>%
  dm_has_pk(airports)
#> [1] FALSE
```

If you still need to get to know your data better, and it is already
available in the form of a `dm` object, you can use the
`dm_enum_pk_candidates()` function in order to get information about
which columns of the table are unique keys:

``` r
dm_enum_pk_candidates(flights_dm_with_key, airports)
#> # A tibble: 8 x 3
#>   columns candidate why                                                    
#>   <keys>  <lgl>     <chr>                                                  
#> 1 faa     TRUE      ""                                                     
#> 2 lon     TRUE      ""                                                     
#> 3 name    FALSE     "has duplicate values: Municipal Airport (5), All Airp…
#> 4 lat     FALSE     "has duplicate values: 38.88944 (2), 40.63975 (2)"     
#> 5 alt     FALSE     "has duplicate values: 0 (51), 13 (13), 14 (12), 15 (1…
#> 6 tz      FALSE     "has duplicate values: -5 (521), -6 (342), -9 (240), -…
#> 7 dst     FALSE     "has duplicate values: A (1388), U (47), N (23)"       
#> 8 tzone   FALSE     "has duplicate values: America/New_York (519), America…
```

The `flights` table does not have any one-column primary key candidates:

``` r
dm_enum_pk_candidates(flights_dm_with_key, flights) %>% dplyr::count(candidate)
#> # A tibble: 1 x 2
#>   candidate     n
#>   <lgl>     <int>
#> 1 FALSE        19
```

To get an overview over all tables with primary keys, use
`dm_get_all_pks()`:

``` r
dm_get_all_pks(dm_nycflights13(cycle = TRUE))
#> # A tibble: 4 x 2
#>   table    pk_col           
#>   <chr>    <keys>           
#> 1 airlines carrier          
#> 2 airports faa              
#> 3 planes   tailnum          
#> 4 weather  origin, time_hour
```

Here we used the prepared `dm` object `dm_nycflights13(cycle = TRUE)` as
an example. This object already has all keys pre-set.

## Foreign keys

Useful functions for managing foreign key relations include:

1.  `dm_add_fk()`
2.  `dm_has_fk()`
3.  `dm_get_fk()`
4.  `dm_rm_fk()`
5.  `dm_enum_fk_candidates()`
6.  `dm_get_all_fks()`

Now it gets (even more) interesting: we want to define relations between
different tables. With the `dm_add_fk()` function you can define which
column of which table points to another table’s column.

This is done by choosing a foreign key from one table that will point to
a primary key of another table. The primary key of the referred table
must be set with `dm_add_pk()`. `dm_add_fk()` will find the primary key
column of the referenced table by itself and make the indicated column
of the child table point to it.

``` r
flights_dm_with_key %>% dm_add_fk(flights, origin, airports)
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  sqlite 3.35.5 [/tmp/RtmpGalley/nycflights13.sqlite]
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `sqlite_stat1`, … (7 total)
#> Columns: 62
#> Primary keys: 1
#> Foreign keys: 1
```

This will throw an error:

``` r
flights_dm %>% dm_add_fk(flights, origin, airports)
#> Error: ref_table `airports` needs a primary key first. Use `dm_enum_pk_candidates()` to find appropriate columns and `dm_add_pk()` to define a primary key.
```

Let’s create a `dm` object with a foreign key relation to work with
later on:

``` r
flights_dm_with_fk <- dm_add_fk(flights_dm_with_key, flights, origin, airports)
```

What if we tried to add another foreign key relation from `flights` to
`airports` to the object? Column `dest` might work, since it also
contains airport codes:

``` r
flights_dm_with_fk %>% dm_add_fk(flights, dest, airports, check = TRUE)
#> Error: Column `dest` of table `flights` contains values (see examples above) that are not present in column `faa` of table `airports`.
```

As you can see, behind the scenes, checks are executed automatically
(unless `check = FALSE`) by the functions of `dm` to prevent steps that
would result in inconsistent representations.

Use `dm_has_fk()` for checking if a foreign key exists that is pointing
from one table to another:

``` r
flights_dm_with_fk %>% dm_has_fk(flights, planes)
#> [1] FALSE
flights_dm_with_fk %>% dm_has_fk(flights, airports)
#> [1] TRUE
```

If you want to access the name of the column which acts as a foreign key
of one table to another table’s column, use `dm_get_fk()`:

``` r
flights_dm_with_fk %>% dm_get_fk(flights, planes)
#> <list_of<character>[0]>
flights_dm_with_fk %>% dm_get_fk(flights, airports)
#> <list_of<character>[1]>
#> [[1]]
#> [1] "origin"
```

Remove foreign key relations with `dm_rm_fk()` (parameter
`column = NULL` means that all relations will be removed):

``` r
flights_dm_with_fk %>%
  dm_rm_fk(table = flights, column = dest, ref_table = airports) %>%
  dm_get_fk(flights, airports)
#> Error: (`dest`) is not a foreign key of table `flights` into table `airports`.
flights_dm_with_fk %>%
  dm_rm_fk(flights, origin, airports) %>%
  dm_get_fk(flights, airports)
#> <list_of<character>[0]>
flights_dm_with_fk %>%
  dm_rm_fk(flights, NULL, airports) %>%
  dm_get_fk(flights, airports)
#> <list_of<character>[0]>
```

Since the primary keys are defined in the `dm` object, you do not need
to provide the referenced column name of `ref_table`. This is always the
primary key column of the table.

Another function for getting to know your data better
(cf. `dm_enum_pk_candidates()` in [“Primary keys of `dm` objects”](#pk))
is `dm_enum_fk_candidates()`. Use it to get an overview over foreign key
candidates that point from one table to another:

``` r
dm_enum_fk_candidates(flights_dm_with_key, weather, airports)
#> # A tibble: 15 x 3
#>    columns    candidate why                                                
#>    <keys>     <lgl>     <chr>                                              
#>  1 origin     TRUE      ""                                                 
#>  2 year       FALSE     "values of `weather$year` not in `airports$faa`: 2…
#>  3 month      FALSE     "values of `weather$month` not in `airports$faa`: …
#>  4 day        FALSE     "values of `weather$day` not in `airports$faa`: 3 …
#>  5 hour       FALSE     "values of `weather$hour` not in `airports$faa`: 1…
#>  6 temp       FALSE     "values of `weather$temp` not in `airports$faa`: 3…
#>  7 dewp       FALSE     "values of `weather$dewp` not in `airports$faa`: 2…
#>  8 humid      FALSE     "values of `weather$humid` not in `airports$faa`: …
#>  9 wind_dir   FALSE     "values of `weather$wind_dir` not in `airports$faa…
#> 10 wind_speed FALSE     "values of `weather$wind_speed` not in `airports$f…
#> 11 wind_gust  FALSE     "values of `weather$wind_gust` not in `airports$fa…
#> 12 precip     FALSE     "values of `weather$precip` not in `airports$faa`:…
#> 13 pressure   FALSE     "values of `weather$pressure` not in `airports$faa…
#> 14 visib      FALSE     "values of `weather$visib` not in `airports$faa`: …
#> 15 time_hour  FALSE     "values of `weather$time_hour` not in `airports$fa…
```

Get an overview of all foreign key relations with`dm_get_all_fks()`:

``` r
dm_get_all_fks(dm_nycflights13(cycle = TRUE))
#> # A tibble: 5 x 4
#>   child_table child_fk_cols     parent_table parent_key_cols   
#>   <chr>       <keys>            <chr>        <keys>           
#> 1 flights     carrier           airlines     carrier          
#> 2 flights     origin            airports     faa              
#> 3 flights     dest              airports     faa              
#> 4 flights     tailnum           planes       tailnum          
#> 5 flights     origin, time_hour weather      origin, time_hour
```

<!-- Generated by galley: do not edit by hand -->

# Building data models from data frames with dm

dm allows you to create your own relational data models from local data
frames. Once your data model is complete, you can deploy it to a range
of DBMSs using dm.

## Creating a dm object from data frames

The example data set that we will be using is available through the
[`nycflights13`](https://github.com/tidyverse/nycflights13) package. The
five tables that we are working with contain information about all
flights that departed from the airports of New York to other
destinations in the United States in 2013:

-   `flights` represents the trips taken by planes
-   `airlines` includes
    -   the names of transport organizations (`name`)
    -   their abbreviated codes (`carrier`)
-   `airports` indicates the ports of departure (`origin`) and of
    destination (`dest`)
-   `weather` contains meteorological information at each hour
-   `planes` describes characteristics of the aircraft

Once we’ve loaded {nycflights13}, the aforementioned tables are all in
our work environment, ready to be accessed.

``` r
library(nycflights13)

airports
#> # A tibble: 1,458 x 8
#>    faa   name                    lat    lon   alt    tz dst   tzone        
#>    <chr> <chr>                 <dbl>  <dbl> <dbl> <dbl> <chr> <chr>        
#>  1 04G   Lansdowne Airport      41.1  -80.6  1044    -5 A     America/New_…
#>  2 06A   Moton Field Municipa…  32.5  -85.7   264    -6 A     America/Chic…
#>  3 06C   Schaumburg Regional    42.0  -88.1   801    -6 A     America/Chic…
#>  4 06N   Randall Airport        41.4  -74.4   523    -5 A     America/New_…
#>  5 09J   Jekyll Island Airport  31.1  -81.4    11    -5 A     America/New_…
#>  6 0A9   Elizabethton Municip…  36.4  -82.2  1593    -5 A     America/New_…
#>  7 0G6   Williams County Airp…  41.5  -84.5   730    -5 A     America/New_…
#>  8 0G7   Finger Lakes Regiona…  42.9  -76.8   492    -5 A     America/New_…
#>  9 0P2   Shoestring Aviation …  39.8  -76.6  1000    -5 U     America/New_…
#> 10 0S9   Jefferson County Intl  48.1 -123.    108    -8 A     America/Los_…
#> # … with 1,448 more rows
```

Your own data will probably not be available as an R package. Whatever
format it is in, you will need to be able to load it as data frames into
your R session. If the data is too large, consider using dm to connect
to the database instead. See `vignette("howto-dm-db")` for details on
using dm with databases.

## Adding Tables

Our first step will be to tell `dm` which tables we want to work with
and how they are connected. For that we can use `dm()`, passing in the
table names as arguments.

``` r
library(dm)

flights_dm_no_keys <- dm(airlines, airports, flights, planes, weather)
flights_dm_no_keys
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 0
#> Foreign keys: 0
```

The `as_dm()` function is an alternative that works if you already have
a list of tables.

## A dm is a list

dm objects behave like lists with a user- and console-friendly print
format. In fact, using a dm as a nicer list for organizing your data
frames in your environment is an easy first step towards using dm and
its data modeling functionality.

Member referencing, by subscript and by name, and list and slice
manipulation functions work just as you would expect on a dm object.

``` r
names(flights_dm_no_keys)
#> [1] "airlines" "airports" "flights"  "planes"   "weather"
flights_dm_no_keys$airports
#> # A tibble: 1,458 x 8
#>    faa   name                    lat    lon   alt    tz dst   tzone        
#>    <chr> <chr>                 <dbl>  <dbl> <dbl> <dbl> <chr> <chr>        
#>  1 04G   Lansdowne Airport      41.1  -80.6  1044    -5 A     America/New_…
#>  2 06A   Moton Field Municipa…  32.5  -85.7   264    -6 A     America/Chic…
#>  3 06C   Schaumburg Regional    42.0  -88.1   801    -6 A     America/Chic…
#>  4 06N   Randall Airport        41.4  -74.4   523    -5 A     America/New_…
#>  5 09J   Jekyll Island Airport  31.1  -81.4    11    -5 A     America/New_…
#>  6 0A9   Elizabethton Municip…  36.4  -82.2  1593    -5 A     America/New_…
#>  7 0G6   Williams County Airp…  41.5  -84.5   730    -5 A     America/New_…
#>  8 0G7   Finger Lakes Regiona…  42.9  -76.8   492    -5 A     America/New_…
#>  9 0P2   Shoestring Aviation …  39.8  -76.6  1000    -5 U     America/New_…
#> 10 0S9   Jefferson County Intl  48.1 -123.    108    -8 A     America/Los_…
#> # … with 1,448 more rows
flights_dm_no_keys[c("airports", "flights")]
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airports`, `flights`
#> Columns: 27
#> Primary keys: 0
#> Foreign keys: 0
```

## Defining Keys

Even though we now have a `dm` object that contains all our data, we
have not specified how our five tables are connected. To do this we need
to define primary keys and foreign keys on the tables.

Primary keys and foreign keys are how relational database tables are
linked with each other. A primary key is a column or column tuple that
has a unique value for each row within a table. A foreign key is a
column or column tuple containing the primary key for a row in another
table. Foreign keys act as cross references between tables. They specify
the relationships that gives us the *relational* database. For more
information on keys and a crash course on databases, see
`vignette("howto-dm-theory")`.

## Primary Keys

`dm` offers `dm_enum_pk_candidates()` to identify viable primary keys,
and `dm_add_pk()` to add them.

``` r
dm_enum_pk_candidates(
  dm = flights_dm_no_keys,
  table = planes
)
#> # A tibble: 9 x 3
#>   columns     candidate why                                                
#>   <keys>      <lgl>     <chr>                                              
#> 1 tailnum     TRUE      ""                                                 
#> 2 year        FALSE     "has duplicate values: 2001 (284), 2000 (244), 200…
#> 3 type        FALSE     "has duplicate values: Fixed wing multi engine (32…
#> 4 manufactur… FALSE     "has duplicate values: BOEING (1630), AIRBUS INDUS…
#> 5 model       FALSE     "has duplicate values: 737-7H4 (361), A320-232 (25…
#> 6 engines     FALSE     "has duplicate values: 2 (3288), 1 (27), 4 (4), 3 …
#> 7 seats       FALSE     "has duplicate values: 149 (452), 140 (411), 55 (3…
#> 8 speed       FALSE     "has 3299 missing values, and duplicate values: 43…
#> 9 engine      FALSE     "has duplicate values: Turbo-fan (2750), Turbo-jet…
```

Now, we add the primary keys that we have identified:

``` r
flights_dm_only_pks <-
  flights_dm_no_keys %>%
  dm_add_pk(table = airlines, columns = carrier) %>%
  dm_add_pk(airports, faa) %>%
  dm_add_pk(planes, tailnum)
flights_dm_only_pks
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 3
#> Foreign keys: 0
```

Note that we demonstrate both named and positional arguments above.

## Foreign Keys

To define how our tables are related, we use `dm_add_fk()` to add
foreign keys. In calling `dm_add_fk()` the `table` argument is the table
that needs a foreign key to link it to a second table. `ref_table` is
the table to be linked to and it needs a primary key already defined for
it.

``` r
dm_enum_fk_candidates(
  dm = flights_dm_only_pks,
  table = flights,
  ref_table = airlines
)
#> # A tibble: 19 x 3
#>    columns      candidate why                                              
#>    <keys>       <lgl>     <chr>                                            
#>  1 carrier      TRUE      ""                                               
#>  2 year         FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  3 month        FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  4 day          FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  5 dep_time     FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  6 sched_dep_t… FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  7 dep_delay    FALSE     "Can't combine `value1` <double> and `value1` <c…
#>  8 arr_time     FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  9 sched_arr_t… FALSE     "Can't combine `value1` <integer> and `value1` <…
#> 10 arr_delay    FALSE     "Can't combine `value1` <double> and `value1` <c…
#> 11 flight       FALSE     "Can't combine `value1` <integer> and `value1` <…
#> 12 tailnum      FALSE     "values of `flights$tailnum` not in `airlines$ca…
#> 13 origin       FALSE     "values of `flights$origin` not in `airlines$car…
#> 14 dest         FALSE     "values of `flights$dest` not in `airlines$carri…
#> 15 air_time     FALSE     "Can't combine `value1` <double> and `value1` <c…
#> 16 distance     FALSE     "Can't combine `value1` <double> and `value1` <c…
#> 17 hour         FALSE     "Can't combine `value1` <double> and `value1` <c…
#> 18 minute       FALSE     "Can't combine `value1` <double> and `value1` <c…
#> 19 time_hour    FALSE     "Can't combine `value1` <datetime<America/New_Yo…
```

Having chosen a column from the successful candidates provided by
`dm_enum_fk_candidates()`, we use the `dm_add_fk()` function to
establish the foreign key linking the tables. In the second call to
`dm_add_fk()` we complete the process for the `flights` and `airlines`
tables that we started above. The `carrier` column in the `airlines`
table will be added as the foreign key in `flights`.

``` r
flights_dm_all_keys <-
  flights_dm_only_pks %>%
  dm_add_fk(table = flights, columns = tailnum, ref_table = planes) %>%
  dm_add_fk(flights, carrier, airlines) %>%
  dm_add_fk(flights, origin, airports)
flights_dm_all_keys
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 3
#> Foreign keys: 3
```

Having created the required primary and foreign keys to link all the
tables together, we now have a relational data model we can work with.

## Visualization

Visualizing a data model is a quick and easy way to verify that we have
successfully created the model we were aiming for. We can use
`dm_draw()` at any stage of the process to generate a visual
representation of the tables and the links between them:

``` r
flights_dm_no_keys %>%
  dm_draw(rankdir = "TB", view_type = "all")
```

![](/home/kirill/git/R/dm/vignettes/out/howto-dm-df_files/figure-gfm/unnamed-chunk-6-1.png)<!-- -->

``` r
flights_dm_no_keys %>%
  dm_add_pk(airlines, carrier) %>%
  dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/howto-dm-df_files/figure-gfm/unnamed-chunk-6-2.png)<!-- -->

``` r
flights_dm_only_pks %>%
  dm_add_fk(flights, tailnum, planes) %>%
  dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/howto-dm-df_files/figure-gfm/unnamed-chunk-6-3.png)<!-- -->

``` r
flights_dm_all_keys %>%
  dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/howto-dm-df_files/figure-gfm/unnamed-chunk-6-4.png)<!-- -->

## Integrity Checks

As well as checking our data model visually, dm can examine the
constraints we have created by the addition of keys and verify that they
are sensible.

``` r
flights_dm_no_keys %>%
  dm_examine_constraints()
#> ℹ No constraints defined.

flights_dm_only_pks %>%
  dm_examine_constraints()
#> ℹ All constraints satisfied.

flights_dm_all_keys %>%
  dm_examine_constraints()
#> ! Unsatisfied constraints:
#> ● Table `flights`: foreign key tailnum into table `planes`: values of `flights$tailnum` not in `planes$tailnum`: N725MQ (575), N722MQ (513), N723MQ (507), N713MQ (483), N735MQ (396), …
```

The results are presented in a human-readable form, and available as a
tibble for programmatic inspection.

## Programming

Helper functions are available to access details on keys and check
results.

`dm_get_all_pks()` returns a data frame with our primary keys:

``` r
flights_dm_only_pks %>%
  dm_get_all_pks()
#> # A tibble: 3 x 2
#>   table    pk_col 
#>   <chr>    <keys> 
#> 1 airlines carrier
#> 2 airports faa    
#> 3 planes   tailnum
```

A data frame of foreign keys is retrieved with `dm_get_all_fks()`:

``` r
flights_dm_all_keys %>%
  dm_get_all_pks()
#> # A tibble: 3 x 2
#>   table    pk_col 
#>   <chr>    <keys> 
#> 1 airlines carrier
#> 2 airports faa    
#> 3 planes   tailnum
```

We can use `tibble::as_tibble()` on the result of
`dm_examine_constraints()` to programmatically inspect which constraints
are not satisfied:

``` r
flights_dm_all_keys %>%
  dm_examine_constraints() %>%
  tibble::as_tibble()
#> # A tibble: 6 x 6
#>   table   kind  columns ref_table is_key problem                           
#>   <chr>   <chr> <keys>  <chr>     <lgl>  <chr>                             
#> 1 flights FK    tailnum planes    FALSE  "values of `flights$tailnum` not …
#> 2 airlin… PK    carrier <NA>      TRUE   ""                                
#> 3 airpor… PK    faa     <NA>      TRUE   ""                                
#> 4 planes  PK    tailnum <NA>      TRUE   ""                                
#> 5 flights FK    carrier airlines  TRUE   ""                                
#> 6 flights FK    origin  airports  TRUE   ""
```

<!-- Generated by galley: do not edit by hand -->

{dm} was designed to make connecting to and working with an RDBMS as
straightforward as possible. To this end, a dm object can be created
from any database that has a {[DBI](https://dbi.r-dbi.org/)} backend
available ([see list](https://github.com/r-dbi/backends)).

When a dm object is created via a DBI connection to an RDBMS it can
import all the tables in the database, the active schema or a limited
set. For some RDBMS, such as Postgres and SQL Server, primary and
foreign keys are also imported and do not have to be manually added
afterwards.

To demonstrate, we will connect to a [relational dataset
repository](https://relational.fel.cvut.cz/) with a database server that
is publicly accessible without registration. It hosts a [financial
dataset](https://relational.fel.cvut.cz/dataset/Financial) that contains
loan data along with relevant account information and transactions. We
chose this dataset because the relationships between `loan`, `account`,
and `transactions` tables are a good representation of databases that
record real-world business transactions. The repository uses a MariaDB
server for which {dm} does not currently import primary or foreign keys,
so we will need to add them.

Below we open a connection to the publicly accessible database server
using their documented connection parameters. Connection details vary
from database to database. Before connecting to your own RDBMS you may
want to read `vignette("DBI", package = "DBI")` for further information.

``` r
library(RMariaDB)
my_db <- dbConnect(
  MariaDB(),
  username = "guest",
  password = "ctu-relational",
  dbname = "Financial_ijs",
  host = "relational.fel.cvut.cz"
)
```

Creating a dm object takes a single call to `dm_from_con()` with the DBI
connection object as its argument.

``` r
library(dm)

my_dm <- dm_from_con(my_db)
my_dm
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  mysql  [guest@relational.fel.cvut.cz:NA/Financial_ijs]
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `accounts`, `cards`, `clients`, `disps`, `districts`, … (9 total)
#> Columns: 57
#> Primary keys: 0
#> Foreign keys: 0
```

The components of the `my_dm` object are lazy tables powered by
{[dbplyr](https://dbplyr.tidyverse.org/)}. {dbplyr} translates the
{[dplyr](https://dplyr.tidyverse.org/)} grammar of data manipulation
into queries the database server understands. Lazy tables defer
downloading of table data until results are required for printing or
local processing.

## Building a dm from a subset of tables

A dm can also be constructed from individual tables or views. This is
useful for when you want to work with a subset of a database’s tables,
perhaps from different schemas.

Below we use the `$` notation to extract two tables from the financial
database. Then we create our dm by passing the tables in as arguments.
Note that the tables arguments have to all be from the same source, in
this case `my_db`.

``` r
dbListTables(my_db)
#> [1] "accounts"  "cards"     "clients"   "disps"     "districts" "loans"    
#> [7] "orders"    "tkeys"     "trans"

library(dbplyr)
#> 
#> Attaching package: 'dbplyr'
#> The following objects are masked from 'package:dm':
#> 
#>     ident, sql
loans <- tbl(my_db, "loans")
accounts <- tbl(my_db, "accounts")

my_manual_dm <- dm(loans, accounts)
my_manual_dm
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  mysql  [guest@relational.fel.cvut.cz:NA/Financial_ijs]
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `loans`, `accounts`
#> Columns: 11
#> Primary keys: 0
#> Foreign keys: 0
```

## Define Primary and Foreign Keys

Primary keys and foreign keys are how relational database tables are
linked with each other. A primary key is a column or column tuple that
has a unique value for each row within a table. A foreign key is a
column or column tuple containing the primary key for a row in another
table. Foreign keys act as cross references between tables. They specify
the relationships that gives us the *relational* database. For more
information on keys and a crash course on databases, see
`vignette("howto-dm-theory")`.

The [model
diagram](https://relational.fel.cvut.cz/assets/img/datasets-generated/financial.svg)
provided by our test database loosely illustrates the intended
relationships between tables. In the diagram we can see that the `loans`
table should be linked to the `accounts` table. Below we create those
links in 3 steps:

1.  Add a primary key `id` to the `accounts` table
2.  Add a primary key `id` to the `loans` table
3.  Add a foreign key `account_id` to the `loans` table referencing the
    `accounts` table

Then we assign colors to the tables and draw the structure of the dm.

Note that when the foreign key is created the primary key in the
referenced table does not need to be specified, but the primary key must
already be defined. And, as mentioned above, primary and foreign key
constraints on the database are currently only imported for Postgres and
SQL Server databases, and only when `dm_from_con()` is used. This
process of key definition needs to be done manually for other databases.

``` r
my_dm_keys <-
  my_manual_dm %>%
  dm_add_pk(accounts, id) %>%
  dm_add_pk(loans, id) %>%
  dm_add_fk(loans, account_id, accounts) %>%
  dm_set_colors(green = loans, orange = accounts)

my_dm_keys %>%
  dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/howto-dm-db_files/figure-gfm/unnamed-chunk-4-1.png)<!-- -->

Once you have instantiated a dm object you can continue to add tables to
it. For tables from the original source for the dm, use `dm()`

``` r
trans <- tbl(my_db, "trans")

my_dm_keys %>%
  dm(trans)
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  mysql  [guest@relational.fel.cvut.cz:NA/Financial_ijs]
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `loans`, `accounts`, `trans`
#> Columns: 21
#> Primary keys: 2
#> Foreign keys: 1
```

For tables from other sources or from the local environment
`dplyr::copy_to()` is used. `copy_to()` is discussed later in this
article.

## Transient nature of operations

Like other R objects, a dm is immutable and all operations performed on
it are transient unless stored in a new variable.

``` r
my_dm_keys
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  mysql  [guest@relational.fel.cvut.cz:NA/Financial_ijs]
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `loans`, `accounts`
#> Columns: 11
#> Primary keys: 2
#> Foreign keys: 1

my_dm_trans <-
  my_dm_keys %>%
  dm(trans)

my_dm_trans
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  mysql  [guest@relational.fel.cvut.cz:NA/Financial_ijs]
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `loans`, `accounts`, `trans`
#> Columns: 21
#> Primary keys: 2
#> Foreign keys: 1
```

And, like {dbplyr}, results are never written to a database unless
explicitly requested.

``` r
my_dm_keys %>%
  dm_flatten_to_tbl(loans)
#> Renamed columns:
#> * date -> loans.date, accounts.date
#> # Source:   lazy query [?? x 10]
#> # Database: mysql [guest@relational.fel.cvut.cz:NA/Financial_ijs]
#>       id account_id loans.date amount duration payments status district_id
#>    <int>      <int> <date>      <dbl>    <int>    <dbl> <chr>        <int>
#>  1  4959          2 1994-01-05  80952       24     3373 A                1
#>  2  4961         19 1996-04-29  30276       12     2523 B               21
#>  3  4962         25 1997-12-08  30276       12     2523 A               68
#>  4  4967         37 1998-10-14 318480       60     5308 D               20
#>  5  4968         38 1998-04-19 110736       48     2307 C               19
#>  6  4973         67 1996-05-02 165960       24     6915 A               16
#>  7  4986         97 1997-08-10 102876       12     8573 A               74
#>  8  4988        103 1997-12-06 265320       36     7370 D               44
#>  9  4989        105 1998-12-05 352704       48     7348 C               21
#> 10  4990        110 1997-09-08 162576       36     4516 C               36
#> # … with more rows, and 2 more variables: frequency <chr>,
#> #   accounts.date <date>

my_dm_keys %>%
  dm_flatten_to_tbl(loans) %>%
  sql_render()
#> Renamed columns:
#> * date -> loans.date, accounts.date
#> <SQL> SELECT `LHS`.`id` AS `id`, `account_id`, `loans.date`, `amount`, `duration`, `payments`, `status`, `district_id`, `frequency`, `accounts.date`
#> FROM (SELECT `id`, `account_id`, `date` AS `loans.date`, `amount`, `duration`, `payments`, `status`
#> FROM `loans`) `LHS`
#> LEFT JOIN (SELECT `id`, `district_id`, `frequency`, `date` AS `accounts.date`
#> FROM `accounts`) `RHS`
#> ON (`LHS`.`account_id` = `RHS`.`id`)
```

## Performing operations on tables by “zooming”

As the dm is a collection of tables, if we wish to perform operations on
an individual table we set it as the context for those operations using
`dm_zoom_to()`. See `vignette("tech-dm-zoom")` for more detail on
zooming.

dm operations are transient unless persistence is explicitly requested.
To make our chain of manipulations on the selected table permanent we
assign the result of `dm_insert_zoomed()` to a new object,
`my_dm_total`. This is a new dm object, derived from `my_dm_keys`, with
a new lazy table `total_loans` linked to the `accounts` table.

``` r
my_dm_total <-
  my_dm_keys %>%
  dm_zoom_to(loans) %>%
  group_by(account_id) %>%
  summarize(total_amount = sum(amount, na.rm = TRUE)) %>%
  ungroup() %>%
  dm_insert_zoomed("total_loans")
```

Context is set to the table “loans” using `dm_zoom_to(loans)`. You can
learn more about zooming in the tutorial `vignette("tech-dm-zoom")`. We
then use {[dplyr](https://dplyr.tidyverse.org/)} functions on the zoomed
table to generate a new summary table.

`summarize()` returns a temporary table with one row for each group
created by the preceding `group_by()` function. The columns in the
temporary table are constrained to the columns passed as arguments to
the `group_by()` function and the column(s) created by the `summarize()`
function.

`dm_insert_zoomed("total_loans")` adds the temporary table created by
`summarize()` to the data model under a new name, `total_loans`. Because
the grouping variable `account_id` is a primary key, the new derived
table is automatically linked to the `accounts` table.

``` r
my_dm_total %>%
  dm_set_colors(violet = total_loans) %>%
  dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/howto-dm-db_files/figure-gfm/unnamed-chunk-9-1.png)<!-- -->

The resulting table `total_loans` can be accessed like any other table
in the dm object.

``` r
my_dm_total$total_loans
#> # Source:   lazy query [?? x 2]
#> # Database: mysql [guest@relational.fel.cvut.cz:NA/Financial_ijs]
#>    account_id total_amount
#>         <int>        <dbl>
#>  1          2        80952
#>  2         19        30276
#>  3         25        30276
#>  4         37       318480
#>  5         38       110736
#>  6         67       165960
#>  7         97       102876
#>  8        103       265320
#>  9        105       352704
#> 10        110       162576
#> # … with more rows
```

It is a *lazy table* powered by the
{[dbplyr](https://dbplyr.tidyverse.org/)} package: the results are not
materialized, instead an SQL query is built and executed each time the
data is requested.

``` r
my_dm_total$total_loans %>%
  sql_render()
#> <SQL> SELECT `account_id`, SUM(`amount`) AS `total_amount`
#> FROM `loans`
#> GROUP BY `account_id`
```

Use `compute()` on a zoomed table to materialize it to a temporary table
and avoid recomputing. See `vignette("howto-dm-copy")` for more details.

## Downloading data

When it becomes necessary to move data locally for analysis or
reporting, the {dm} method `collect()` is used. Operations on dm objects
for databases are limited to report only the first ten results.
`collect()` forces the evaluation of all SQL queries and the generation
of the complete set of results. The resulting tables are transferred
from the RDBMS and stored as local tibbles.

``` r
my_dm_local <-
  my_dm_total %>%
  collect()

my_dm_local$total_loans
#> # A tibble: 682 x 2
#>    account_id total_amount
#>         <int>        <dbl>
#>  1          2        80952
#>  2         19        30276
#>  3         25        30276
#>  4         37       318480
#>  5         38       110736
#>  6         67       165960
#>  7         97       102876
#>  8        103       265320
#>  9        105       352704
#> 10        110       162576
#> # … with 672 more rows
```

Use this method with caution. If you are not sure of the size of the
dataset you will be downloading you can call `dm_nrow()` on your dm for
the row count of your data model’s tables.

``` r
my_dm_total %>%
  dm_nrow()
#>       loans    accounts total_loans 
#>         682        4500         682
```

## Persisting results

It is just as simple to move a local relational model into an RDBMS as
is using `collect()` to download it. The method used is `copy_dm_to()`
and it takes as arguments a database connection and a dm object. In the
example below a local SQLite database is used to demonstrate it, but
{dm} is designed to work with any RDBMS supported by {DBI}.

``` r
destination_db <- DBI::dbConnect(RSQLite::SQLite())

deployed_dm <- copy_dm_to(destination_db, my_dm_local)

deployed_dm
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  sqlite 3.35.5 []
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `loans`, `accounts`, `total_loans`
#> Columns: 13
#> Primary keys: 2
#> Foreign keys: 2
my_dm_local
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `loans`, `accounts`, `total_loans`
#> Columns: 13
#> Primary keys: 2
#> Foreign keys: 2
```

In the output you can observe that the `src` for `deployed_dm` is the
SQLite database, while for `my_dm_local` the source is the local R
environment.

Persisting tables are covered in more detail in
`vignette("howto-dm-copy")`.

## Conclusion

In this tutorial we have demonstrated how simple it is to load a
database into a dm object and begin working with it. While loading a dm
from most RDBMS currently requires you to manually set key relations, dm
provides methods to make this straightforward. It is planned that future
versions of dm will support automatic key creation for more RDBMS.

The next step is to read `vignette("howto-dm-copy")`, where copying your
tables to and from an RDBMS is covered. `vignette("howto-dm-rows")`
discusses manipulation of individual rows in a database.

## Further reading

`vignette("howto-dm-df")` -- Is your data in local data frames? This
article covers creating a data model from your local data frames,
including building the relationships in your data model, verifying your
model, and leveraging the power of dplyr to operate on your data model.

`vignette("howto-dm-theory")` -- Do you know all about data frames but
very little about relational data models? This quick introduction will
walk you through the key similarities and differences, and show you how
to move from individual data frames to a relational data model.

---
title: "Introduction to relational data models"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{How to: Understand relational data models}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
author: Katharina Brunner
---


``````{r setup, include = FALSE}
source("setup/setup.R")
``````

Computer scientists are familiar with multiple, linked tables.
But, because many R users tend to have backgrounds in other disciplines, we present **six important terms in relational data modeling** to help you to jump-start working with {dm}. These terms are:

1) [Data Frames and Tables](#tables)
2) [Data Model](#model)
3) [Primary Keys](#pk)
4) [Foreign Keys](#fk)
5) [Referential Integrity](#referential-integrity)
6) [Normalization](#normalization)
7) [Relational Databases](#relational-databases)

## 1. Data Frames and Tables

A data frame is a fundamental data structure in R.
Columns represent variables, rows represent observations.
In more technical terms, a data frame is a list of variables of identical length and unique row names.
If you imagine it visually, the result is a typical table structure.
That is why working with data from spreadsheets is so convenient and the users of the popular [{dplyr}](https://dplyr.tidyverse.org) package for data wrangling mainly rely on data frames.

The downside is that data frames and flat file systems like spreadsheets can result in bloated tables because they hold many repetitive values.
In the worst case, a data frame can contain multiple columns with only a single value different in each row.

This calls for better data organization by utilizing the resemblance between data frames and database tables, which also consist of columns and rows.
The elements are just named differently:

| Data Frame | Table                 |
|------------|-----------------------|
| Column     | Attribute (or Field)  |
| Row        | Tuple (or Record)     |

Additionally, number of rows and columns for a data frame are, respectively, analogous to the *cardinality* and *degree* of the table.

Relational databases, unlike data frames, do not keep all data in one large table but instead split it into multiple smaller tables.
That separation into sub-tables has several advantages:

- all information is stored only once, avoiding redundancy and conserving memory
- all information needs to be updated only once and in one place, improving consistency and avoiding errors that may result from updating (or forgetting to update) the same value in multiple locations
- all information is organized by topic and segmented into smaller tables that are easier to handle

It is for these reasons that separation of data helps with data quality, and they explain the popularity of relational databases in production-level data management.

The downside of this approach is that it is harder to merge together information from different data sources and to identify which entities refer to the same object, a common task when modeling or plotting data.

Thus, to take full advantage of the relational database approach, an associated **data model** is needed to overcome the challenges that arise when working with multiple tables.

Let's illustrate this challenge with the data from the [`nycflights13` dataset](https://github.com/tidyverse/nycflights13) that
contains detailed information about the 336,776 flights that departed from New York City in 2013.
The information is stored in five tables.

Details like the full name of an airport are not available immediately; these can only be obtained by joining or merging the constituent tables, which can result in long and inflated pipe chains full of `left_join()`, `anti_join()` and other forms of data merging.

In classical {dplyr} notation, you will need four `left_join()` calls to gradually merge the `flights` table to the `airlines`, `planes`, `airports`, and `weather` tables to create one wide data frame.

```{r}
library(dm)
library(nycflights13)

nycflights13_df <-
  flights %>%
  left_join(airlines, by = "carrier") %>%
  left_join(planes, by = "tailnum") %>%
  left_join(airports, by = c("origin" = "faa")) %>%
  left_join(weather, by = c("origin", "time_hour"))

nycflights13_df
```

{dm} offers a more elegant and shorter way to combine tables while augmenting {dplyr}/{dbplyr} workflows.

It is possible to have the best of both worlds: manage your data with {dm} as linked tables, and, when necessary, flatten multiple tables into a single data frame for analysis with {dplyr}.

The next step is to create a [data model](#model) based on multiple tables:

## 2. Data Model {#model}

A data model shows the structure between multiple tables that are linked together.

The `nycflights13` relations can be transferred into the following graphical representation:

```{r warning=FALSE, message=FALSE}
dm <- dm_nycflights13(cycle = TRUE)

dm %>%
  dm_draw()
```

The `flights` table is linked to four other tables: `airlines`, `planes`, `weather`, and `airports`. 
By using directed arrows, the visualization shows explicitly the connection between different columns (they are called attributes in the relational data sphere).

For example: The column `carrier` in `flights` can be joined with the column `carrier` from the `airlines` table.

The links between the tables are established through [primary keys](#pk) and [foreign keys](#fk).

As an aside, we can also now see how avoiding redundant data by building data models with multiple tables can save memory compared to storing data in single a data frame:

```{r}
object.size(dm)

object.size(nycflights13_df)
```

Further Reading: The {dm} methods for [visualizing data models](https://dm.cynkra.com/articles/tech-dm-draw.html).

## 3. Primary Keys {#pk}

In a relational data model, each table should have **one or several columns that uniquely identify a row**.
These columns define the *primary key* (abbreviated with "pk").
If the key consists of a single column, it is called a *simple key*.
A key consisting of more than one column is called a *compound key*.

Example: In the `airlines` table of `nycflights13` the column `carrier` is the primary key, a simple key.
The `weather` table has the combination of `origin` and `time_hour` as primary key, a compound key.

You can get all primary keys in a `dm` by calling `dm_get_all_pks()`:

```{r}
dm %>%
  dm_get_all_pks()
```

`dm_enum_pk_candidates()` checks suitability of each column to serve as a simple primary key:

```{r}
dm %>%
  dm_enum_pk_candidates(airports)
```

Further Reading: The {dm} package offers several functions for dealing with [primary keys](https://dm.cynkra.com/articles/tech-dm-class.html).

## 4. Foreign Keys {#fk}

The **counterpart of a primary key in one table is the foreign key in another table**.
In order to join two tables, the primary key of the first table needs to be referenced from the second table.
This column or these columns are called the *foreign key* (abbreviated with "fk").

For example, if you want to link the `airlines` table to the `flights` table, the primary key in `airlines` needs to match the foreign key in `flights`.
This condition is satisfied because the column `carrier` is present as a primary key in the `airlines` table as well as a foreign key in the `flights` table.
In the case of compound keys, the `origin` and `time_hour` columns (which form the primary key of the `weather` table) are also present in the `flights` table.

You can find foreign key candidates for simple keys with the function `dm_enum_fk_candidates()`; they are marked with `TRUE` in the `candidate` column.

```{r}
dm %>%
  dm_enum_fk_candidates(flights, airlines)
```

Additionally, you can also extract a summary of all foreign key relations present in a `dm` object using `dm_get_all_fks()`:

```{r}
dm %>%
  dm_get_all_fks()
```

Further Reading: All {dm} functions for working with [foreign keys](https://dm.cynkra.com/articles/tech-dm-class.html).

## 5. Referential Integrity {#referential-integrity}

A data set has referential integrity if all relations between tables are valid.
That is, every foreign key holds a primary key that is present in the parent table.
If a foreign key contains a reference where the corresponding row in the parent table is not available, that row is an orphan row and the database no longer has referential integrity.

{dm} allows checking referential integrity with the `dm_examine_constraints()` function.
The following conditions are checked:

- All primary key values must be unique and not missing (i.e., `NA`s are not allowed).
- Each foreign key value must have a corresponding primary key value.

In the example data model, for a substantial share of the flights, detailed information for the corresponding airplane is not available:

```{r}
dm %>%
  dm_examine_constraints()
```

Establishing referential integrity is important for providing clean data for analysis or downstream users.
See `vignette("howto-dm-rows")` for more information on adding, deleting, or updating individual rows, and `vignette("tech-dm-zoom")` for operations on the data in a data model.

## 6. Normalization {#normalization}

Normalization is a technical term that describes the **central design principle of a relational data model**:
splitting data into multiple tables.

A normalized data schema consists of several relations (tables) that are linked with attributes (columns).
The relations can be joined together by means of [primary](#pk) and [foreign keys](#fk).
The main goal of normalization is to keep data organization as clean and simple as possible by avoiding redundant data entries.

For example, if you want to change the name of one airport in the `nycflights13` dataset, you will only need to update a single data value.
This principle is sometimes called the *single point of truth*.

```{r}
#  Update in one single location...
airlines[airlines$carrier == "UA", "name"] <- "United broke my guitar"

airlines %>%
  filter(carrier == "UA")

# ...propagates to all related records
flights %>%
  left_join(airlines) %>%
  select(flight, name)
```

Another way to demonstrate normalization is splitting a table into two parts.

Let's look at the `planes` table, which consists of 3322 individual tail numbers and corresponding information for the specific airplane, like the year it was manufactured or the average cruising speed.

The function `decompose_table()` extracts two new tables and creates a new key `model_id`, that links both tables.

This results in a `parent_table` and a `child_table` that differ massively in the number of rows:

```{r}
planes %>%
  decompose_table(model_id, model, manufacturer, type, engines, seats, speed)
```

While `child_table` contains 3322 unique `tailnum` rows and therefore consists of 3322 rows, just like the original `planes` table, the `parent_table` shrunk to just 147 rows, enough to store all relevant combinations and avoid storing redundant information.

Further Reading: See the [Simple English Wikipedia article on database normalization](https://simple.wikipedia.org/wiki/Database_normalisation) for more details.

## 7. Relational Databases {#relational-databases}

{dm} is built upon relational data models but it is not a database itself.
Databases are systems for data management and many of them are constructed as relational databases (e.g., SQLite, MySQL, MSSQL, Postgres, etc.).
As you can guess from the names of the databases, SQL, short for Structured Querying Language, plays an important role: it was invented for the purpose of querying relational databases.

In production, the data is stored in a relational database and {dm} is used to work with the data.

Therefore, {dm} can copy data [from and to databases](https://dm.cynkra.com/articles/howto-dm-copy.html), and works transparently with both in-memory data and with relational database systems.

For example, let's create a local SQLite database and copy the `dm` object to it:

```{r}
con_sqlite <- DBI::dbConnect(RSQLite::SQLite())
con_sqlite
DBI::dbListTables(con_sqlite)

copy_dm_to(con_sqlite, dm)
DBI::dbListTables(con_sqlite)
```

In the opposite direction, `dm` can also be populated with data from a database. Unfortunately, keys currently can be learned only for Microsoft SQL Server and Postgres, but not for SQLite.
Therefore, the `dm` contains the tables but not the keys:

```{r}
dm_from_con(con_sqlite)
```

Remember to terminate the database connection:

```{r disconnect}
DBI::dbDisconnect(con_sqlite)
```

## Conclusion {#conclusion}

In this article, we have learned about some of the most fundamental concepts and data structures associated with the relational database management system (RDBMS).

## Further reading

`vignette("howto-dm-db")` -- This article covers accessing and working with RDBMSs within your R session, including manipulating data, filling in missing relationships between tables, getting data out of the RDBMS and into your model, and deploying your data model to an RDBMS.

`vignette("howto-dm-df")` -- Is your data in local data frames? This article covers creating a data model from your local data frames, including building the relationships in your data model, verifying your model, and leveraging the power of dplyr to operate on your data model.

<!-- Generated by galley: do not edit by hand -->

{dm} is an R package that provides tools for working with multiple
related tables, stored as data frames or in a relational database.

## Background

Relational databases and flat tables, like data frames or spreadsheets,
present data in fundamentally different ways.

In data frames and spreadsheets, all data is presented together in one
large table with many rows and columns. This means that the data is
accessible in one location but has the disadvantage that the same values
may be repeated multiple times, resulting in bloated tables with
redundant data. In the worst case scenario, a data frame may have many
rows and columns but only a single value different in each row.

Relational databases, on the other hand, do not keep all data together
but split it into multiple smaller tables. That separation into
sub-tables has several advantages:

-   all information is stored only once, avoiding repetition and
    conserving memory
-   all information is updated only once and in one place, improving
    consistency and avoiding errors that may result from updating the
    same value in multiple locations
-   all information is organized by topic and segmented into smaller
    tables that are easier to handle

Separation of data, thus, helps with data quality, and explains the
continuing popularity of relational databases in production-level data
management.

The downside of this approach is that it is harder to merge together
information from different data sources and to identify which entities
refer to the same object, a common task when modeling or plotting data.
To be mapped uniquely, the entities would need to be designated as
*keys*, and the separate tables collated together through a process
called *joining*.

In R, there already exist packages that support handling inter-linked
tables but the code is complex and requires multiple command sequences.
The goal of the {dm} package is to simplify the data management
processes in R while keeping the advantages of relational data models
and the core concept of splitting one table into multiple tables. In
this way, you can have the best of both worlds: manage your data as a
collection of linked tables, then flatten multiple tables into one for
an analysis with {dplyr} or other packages, on an as-needed basis.

Although {dm} is built upon relational data models, it is not a database
itself. It can work transparently with both relational database systems
and in-memory data, and copy data [from and to
databases](https://cynkra.github.io/dm/articles/dm.html#copy).

---
title: "Introduction to dm"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{dm overview}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fansi::set_knit_hooks(knitr::knit_hooks)
options(crayon.enabled = TRUE, width = 75, cli.width = 75)

knit_print.grViz <- function(x, ...) {
  x %>%
    DiagrammeRsvg::export_svg() %>%
    c("`````{=html}\n", ., "\n`````\n") %>%
    knitr::asis_output()
}
```

The goal of both the package dm and the class `dm` - which comes along with it - is to make your life easier when you have to deal with data models the data therein.

Let's first have a look at the new class:

## Class `dm`

The `dm` class contains a set of tables as well as information about their primary keys and the relations between the tables. You can have your `dm` object on different data sources, such as a local environment or a database. In the latter use case you need to have the package dbplyr installed.

`dm` is a wrapper around two classes, which are independent from each other:

1. part of the object is a [`src`-object](https://dplyr.tidyverse.org/reference/src.html) from dplyr
2. the other part is a class `data_model` object
 from the package [datamodelr](https://github.com/bergant/datamodelr)^[Use: `devtools::install_github("https://github.com/bergant/datamodelr")` to install it]

The `src`-object contains the information, where the tables of your data model live. And in the `data_model` object the meta-information is stored:

- what are the primary keys in your schema?
- how are the tables related to each other, i.e. what are the foreign key relations?

Last but not least, a third part of the object is a named list containing the tables on the `src`.

So much to the theory, let's see how to construct such an object and how it looks in `R`:

## Examples of `dm` objects {#ex_dm}

Relevant functions for creating `dm` objects are:

1. `dm()`
2. `as_dm()`

We can use the tables from the famous package {nycflights13}:
```{r message=FALSE}
library(dm)
library(dplyr)
flights_dm <- dm_from_src(src_df(pkg = "nycflights13"))
flights_dm
```

Here we make use of the fact, that the function `dm(src, data_model = NULL)` includes all available tables on a source in the `dm` object, should the `data_model`-argument be left `NULL`.
Thus you can use this for example on a postgres database that you access via `src_postgres()` (with the appropriate arguments `dbname`, `host`, `port`, ...), to produce a `dm` object with all the tables on the database.

Another way of creating a `dm` object is calling `as_dm()` on a list of `tbls`:
```{r}
trees_dm <- as_dm(list("trees1" = trees, "trees2" = trees))
trees_dm
```

## Methods for `dm` objects

tbl(dm, "table_name") etc. (copy_to...) FIXME: complete

But how can we use dm-functions to manage primary keys of the tables in a `dm` object?

## Primary keys of `dm` objects {#pk}

Useful functions for managing primary key settings are:

1. `dm_add_pk()`
2. `dm_has_pk()`
3. `dm_get_pk()`
4. `dm_rm_pk()`
5. `dm_enum_pk_candidates()`

If you created a `dm` object according to the examples in ["Examples of `dm` objects"](#ex_dm), your object does not yet have any primary keys set.
So let's add one.
By default `dm_add_pk()` checks if the column of the table given by the user is a unique key.
Since the `trees` dataset does not have any unique one-column keys, we will use the `nycflights13` tables, i.e. `flights_dm`.

```{r}
dm_has_pk(flights_dm, airports)
flights_dm_with_key <- dm_add_pk(flights_dm, airports, faa)
flights_dm_with_key
```
Still looks exactly the same as before... let's check:
```{r}
dm_has_pk(flights_dm_with_key, airports)
```
Get the name of the column that is marked as primary key of the table:
```{r}
dm_get_pk(flights_dm_with_key, airports)
```
Remove a primary key by:
```{r}
dm_rm_pk(flights_dm_with_key, airports) %>%
  dm_has_pk(airports)
```

If you still need to get to know your data better, and it is already available in the form of a `dm` object, you can use the function `dm_enum_pk_candidates()` in order to get information, which columns of a table are unique keys:
```{r}
dm_enum_pk_candidates(flights_dm_with_key, airports)
```
Table `flights` does not have any one-column primary key candidates:
```{r}
dm_enum_pk_candidates(flights_dm_with_key, flights) %>% count(candidate)
```

## Foreign keys

Useful functions for managing foreign key relations are:

1. `dm_add_fk()`
2. `dm_has_fk()`
3. `dm_get_fk()`
4. `dm_rm_fk()`
5. `dm_enum_fk_candidates()`

Now it gets (even more) interesting: we want to define relations between different tables.
With the function `dm_add_fk()` you can define, which column of which table points to another table's column.

When dealing with data models, a foreign key from one table has to always point to a primary key of another table.
The primary key of the referred table must be set with `dm_add_pk()`.
`dm_add_fk()` will find the primary key column of the referenced table by itself and make the indicated column of the child table point to it.

```{r}
flights_dm_with_key %>% dm_add_fk(flights, origin, airports)
```
This will throw an error:
```{r error=TRUE}
flights_dm %>% dm_add_fk(flights, origin, airports)
```
Let's create a `dm` object with a foreign key relation to work with later on:
```{r}
flights_dm_with_fk <- dm_add_fk(flights_dm_with_key, flights, origin, airports)
```
What if we try to add another foreign key relation from `flights` to `airports` to the object? Column `dest` might work, since it also contains airport codes:
```{r error=TRUE}
flights_dm_with_fk %>% dm_add_fk(flights, dest, airports)
```
As you can see, behind the scenes, checks are executed automatically by the functions of `dm` to prevent steps that do not comply with the logic of data models.

Use `dm_has_fk()` for checking if a foreign key exists, which is pointing from one table to another:
```{r}
flights_dm_with_fk %>% dm_has_fk(flights, planes)
flights_dm_with_fk %>% dm_has_fk(flights, airports)
```

If you want to access the name of the column, which acts as a foreign key of one table to another table's column, use `dm_get_fk()`:
```{r}
flights_dm_with_fk %>% dm_get_fk(flights, planes)
flights_dm_with_fk %>% dm_get_fk(flights, airports)
```

Remove foreign key relations with `dm_rm_fk()` (parameter `column = NULL` means removal of all relations):
```{r error=TRUE}
flights_dm_with_fk %>%
  dm_rm_fk(table = flights, column = dest, ref_table = airports) %>%
  dm_get_fk(flights, airports)
flights_dm_with_fk %>%
  dm_rm_fk(flights, origin, airports) %>%
  dm_get_fk(flights, airports)
flights_dm_with_fk %>%
  dm_rm_fk(flights, NULL, airports) %>%
  dm_get_fk(flights, airports)
```

Another function for getting to know your data better (cf. `dm_enum_pk_candidates()` in ["Primary keys of `dm` objects"](#pk)) is `dm_enum_fk_candidates()`. Use it to get an overview over foreign key candidates pointing from one table to another:
```{r}
dm_enum_fk_candidates(flights_dm_with_key, weather, airports)
```

## Visualising the data model

Relevant functions for visualising the data model:^[The functions in this section rely heavily on related functions of {datamodelr}.]

1. `dm_set_colors()`
1. `dm_get_colors()`
1. `dm_get_available_colors()`
1. `dm_draw()`

Once you have all primary keys set and all foreign key relations defined, it might be interesting to look at a graphical representation of your data model.

Let's first create a slightly more interconnected `dm` object from the {nycflights13} package tables:
```{r}
flights_dm_w_many_keys <-
  flights_dm %>%
  dm_add_pk(planes, tailnum) %>%
  dm_add_pk(airports, faa) %>%
  dm_add_pk(airlines, carrier) %>%
  dm_add_fk(flights, origin, airports) %>%
  dm_add_fk(flights, carrier, airlines) %>%
  dm_add_fk(flights, tailnum, planes, check = FALSE) %>%
  dm_add_fk(flights, dest, airports, check = FALSE)
flights_dm_w_many_keys
```
You probably realized the option `check = FALSE` in the last two function calls when adding foreign keys.
The check normally ensures, that the set of values of the column of the first table is a subset of the set of values of the primary key column of the referenced table.
Since in our case this isn't the case, but we want complexity, we turned the check off.
It is recommended not to do this, since you might run into errors or other unwanted effects, mainly when working on databases.

You can use colors to visually group your tables into families (which could reflect their logical grouping).
Currently, there are only eight different colors available (plus the option of having borders displayed around the tables or not).
To find out which colors exist, use `dm_get_available_colors()`:
```{r}
dm_get_available_colors()
```

The assignment is carried out with `dm_set_colors()`, using a syntax similar to `switch()`.
Any table you don't specifically assign will be displayed in the default color.

Use `dm_set_colors()` to assign the respective colors.
The result of `dm_set_colors()` is a `dm` object.
The information about the color is stored in its `data_model`-part.
```{r}
flights_dm_w_many_keys_and_colors <-
  flights_dm_w_many_keys %>%
  dm_set_colors(
    flights = "blue",
    airlines = ,
    airports = "orange",
    planes = "green_nb"
  )
```

Draw the schema with `dm_draw()`.
```{r}
dm_draw(flights_dm_w_many_keys_and_colors)
```

The colors can be queried with `dm_get_colors()`.
```{r}
dm_get_colors(flights_dm_w_many_keys_and_colors)
```

There are quite a few options for customization available.
Please read the documentation to find out more.
One argument of the function we'd like to draw your attention to, though, is `table_names`.
This argument accepts a character vector with the table names which are to be printed, in case you would like to only see a subset of the whole data model.
```{r}
dm_draw(flights_dm_w_many_keys_and_colors, table_names = c("flights", "airports", "planes"))
```

## Joining tables

## Filtering a `dm` object

<!-- Generated by galley: do not edit by hand -->

The {dm} package offers functions to work with relational data models in
R. A common task for multiple, separated tables that have a shared
attribute is merging the data.

This document introduces you to the joining functions of {dm} and shows
how to apply them using data from the
[{nycflights13}](https://github.com/tidyverse/nycflights13) package.

[Relational data
models](https://cynkra.github.io/dm/articles/howto-dm-theory#model)
consist of multiple tables that are linked with [foreign
keys](https://cynkra.github.io/dm/articles/howto-dm-theory#fk). They are
the building blocks for joining tables. Read more about relational data
models in the vignette [“Introduction to Relational Data
Models”](https://cynkra.github.io/dm/articles/howto-dm-theory).

First, we load the packages that we need:

``` r
library(dm)
library(tidyverse)
```

## Data: nycflights13

To explore filtering with {dm}, we’ll use the {nycflights13} data with
its tables `flights`, `planes`, `airlines` and `airports`.

This dataset contains information about the 336 776 flights that
departed from New York City in 2013, with 3322 different planes and 1458
airports involved. The data comes from the US Bureau of Transportation
Statistics, and is documented in `?nycflights13`.

First, we have to create a `dm` object from the {nycflights13} data.
This is implemented with `dm::dm_nycflights13()`.

A [data model
object](https://cynkra.github.io/dm/articles/tech-dm-class.html#class-dm)
contains the data as well as metadata.

If you would like to create a `dm` from other tables, please look at
`?dm` and the function `new_dm()`.

``` r
dm <- dm_nycflights13()
```

## Joining a `dm` object

{dm} allows you to join two tables of a `dm` object based on a shared
column. You can use all join functions that you know from the [{dplyr}
package](https://dplyr.tidyverse.org/reference/join.html). Currently
{dplyr} supports four types of mutating joins, two types of filtering
joins, and a nesting join. See `?dplyr::join` for details.

### How it works

A join is the combination of two tables based on shared information. In
technical terms, we merge the tables that need to be directly connected
by a [foreign key
relation](https://cynkra.github.io/dm/articles/howto-dm-theory#fk).

The existing links can be inspected in two ways:

1.  Visually, by drawing the data model with `dm_draw()`

``` r
dm %>%
  dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-join_files/figure-gfm/unnamed-chunk-3-1.png)<!-- -->

The directed arrows show explicitly the relation between different
columns.

1.  Printed to the console by calling `dm_get_all_fks()`

``` r
dm %>%
  dm_get_all_fks()
#> # A tibble: 4 x 4
#>   child_table child_fk_cols     parent_table parent_key_cols   
#>   <chr>       <keys>            <chr>        <keys>           
#> 1 flights     carrier           airlines     carrier          
#> 2 flights     origin            airports     faa              
#> 3 flights     tailnum           planes       tailnum          
#> 4 flights     origin, time_hour weather      origin, time_hour
```

### Joining Examples

Let’s look at some examples:

**Add a column with airline names from the `airlines` table to the
`flights` table.**

``` r
dm_joined <-
  dm %>%
  dm_join_to_tbl(flights, airlines, join = left_join)
dm_joined
#> # A tibble: 11,227 x 20
#>     year month   day dep_time sched_dep_time dep_delay arr_time
#>    <int> <int> <int>    <int>          <int>     <dbl>    <int>
#>  1  2013     1    10        3           2359         4      426
#>  2  2013     1    10       16           2359        17      447
#>  3  2013     1    10      450            500       -10      634
#>  4  2013     1    10      520            525        -5      813
#>  5  2013     1    10      530            530         0      824
#>  6  2013     1    10      531            540        -9      832
#>  7  2013     1    10      535            540        -5     1015
#>  8  2013     1    10      546            600       -14      645
#>  9  2013     1    10      549            600       -11      652
#> 10  2013     1    10      550            600       -10      649
#> # … with 11,217 more rows, and 13 more variables: sched_arr_time <int>,
#> #   arr_delay <dbl>, carrier <chr>, flight <int>, tailnum <chr>,
#> #   origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>, hour <dbl>,
#> #   minute <dbl>, time_hour <dttm>, name <chr>
```

As you can see below, the `dm_joined` dataframe has one more column than
the `flights` table. The difference is the `name` column from the
`airlines` table.

``` r
dm %>%
  tbl("flights") %>%
  names()
#>  [1] "year"           "month"          "day"            "dep_time"      
#>  [5] "sched_dep_time" "dep_delay"      "arr_time"       "sched_arr_time"
#>  [9] "arr_delay"      "carrier"        "flight"         "tailnum"       
#> [13] "origin"         "dest"           "air_time"       "distance"      
#> [17] "hour"           "minute"         "time_hour"

dm %>%
  tbl("airlines") %>%
  names()
#> [1] "carrier" "name"

dm_joined %>%
  names()
#>  [1] "year"           "month"          "day"            "dep_time"      
#>  [5] "sched_dep_time" "dep_delay"      "arr_time"       "sched_arr_time"
#>  [9] "arr_delay"      "carrier"        "flight"         "tailnum"       
#> [13] "origin"         "dest"           "air_time"       "distance"      
#> [17] "hour"           "minute"         "time_hour"      "name"
```

The result is not a `dm` object anymore, but a conventional dataframe:

``` r
dm_joined %>%
  class()
#> [1] "tbl_df"     "tbl"        "data.frame"
```

Another example:

**Get all flights that can’t be matched with airlines names.**

We expect the flights data from {nycflights13} package to be clean and
well organized, so no flights should remain. You can check this with an
`anti_join`:

``` r
dm %>%
  dm_join_to_tbl(flights, airlines, join = anti_join)
#> # A tibble: 0 x 19
#> # … with 19 variables: year <int>, month <int>, day <int>, dep_time <int>,
#> #   sched_dep_time <int>, dep_delay <dbl>, arr_time <int>,
#> #   sched_arr_time <int>, arr_delay <dbl>, carrier <chr>, flight <int>,
#> #   tailnum <chr>, origin <chr>, dest <chr>, air_time <dbl>,
#> #   distance <dbl>, hour <dbl>, minute <dbl>, time_hour <dttm>
```

An example with filtering on a `dm` and then merging:

**Get all flights from Delta Air Lines which didn’t depart from John F.
Kennedy International Airport in May 2013 - and join all the airports
data into the `flights` table.**

Currently, it is important to call `dm_apply_filters()` after piping
your conditions. Only then the underlying tables and key relations are
updated and you can perform a join on the filtered data. We are working
towards removing this inconvenience
[\#62](https://github.com/cynkra/dm/issues/62).

``` r
dm_nycflights13() %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(flights, month == 5) %>%
  dm_apply_filters() %>%
  dm_join_to_tbl(flights, airports, join = left_join)
#> # A tibble: 136 x 26
#>     year month   day dep_time sched_dep_time dep_delay arr_time
#>    <int> <int> <int>    <int>          <int>     <dbl>    <int>
#>  1  2013     5    10      554            600        -6      739
#>  2  2013     5    10      556            600        -4      825
#>  3  2013     5    10      606            610        -4      743
#>  4  2013     5    10      625            630        -5      843
#>  5  2013     5    10      632            635        -3      847
#>  6  2013     5    10      653            700        -7      923
#>  7  2013     5    10      654            700        -6     1001
#>  8  2013     5    10      656            700        -4     1008
#>  9  2013     5    10      656            700        -4      911
#> 10  2013     5    10      657            700        -3     1006
#> # … with 126 more rows, and 19 more variables: sched_arr_time <int>,
#> #   arr_delay <dbl>, carrier <chr>, flight <int>, tailnum <chr>,
#> #   origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>, hour <dbl>,
#> #   minute <dbl>, time_hour <dttm>, name <chr>, lat <dbl>, lon <dbl>,
#> #   alt <dbl>, tz <dbl>, dst <chr>, tzone <chr>
```

A last example:

**Merge all tables into one big table.**

Sometimes you need everything in one place. In this case you can use the
`dm_flatten_to_tbl()` function. It joins all the tables of your `dm`
object together into one wide table. All you have to do is to specify
the starting table. The following joins are determined by the foreign
key links.

``` r
dm_nycflights13() %>%
  dm_select_tbl(-weather) %>%
  dm_flatten_to_tbl(start = flights)
#> Renamed columns:
#> * year -> flights.year, planes.year
#> * name -> airlines.name, airports.name
#> # A tibble: 11,227 x 35
#>    flights.year month   day dep_time sched_dep_time dep_delay arr_time
#>           <int> <int> <int>    <int>          <int>     <dbl>    <int>
#>  1         2013     1    10        3           2359         4      426
#>  2         2013     1    10       16           2359        17      447
#>  3         2013     1    10      450            500       -10      634
#>  4         2013     1    10      520            525        -5      813
#>  5         2013     1    10      530            530         0      824
#>  6         2013     1    10      531            540        -9      832
#>  7         2013     1    10      535            540        -5     1015
#>  8         2013     1    10      546            600       -14      645
#>  9         2013     1    10      549            600       -11      652
#> 10         2013     1    10      550            600       -10      649
#> # … with 11,217 more rows, and 28 more variables: sched_arr_time <int>,
#> #   arr_delay <dbl>, carrier <chr>, flight <int>, tailnum <chr>,
#> #   origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>, hour <dbl>,
#> #   minute <dbl>, time_hour <dttm>, airlines.name <chr>,
#> #   airports.name <chr>, lat <dbl>, lon <dbl>, alt <dbl>, tz <dbl>,
#> #   dst <chr>, tzone <chr>, planes.year <int>, type <chr>,
#> #   manufacturer <chr>, model <chr>, engines <int>, seats <int>,
#> #   speed <int>, engine <chr>
```

Be aware that all column names need to be unique. The
`dm_flatten_to_tbl` cares about automatically renaming the relevant
columns and prints if something was changed,
e.g. `name -> airlines.name`.

---
title: "Updating tables and dm objects"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Updating tables dm objects}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fansi::set_knit_hooks(knitr::knit_hooks)
options(crayon.enabled = TRUE, width = 75, cli.width = 75)

knit_print.grViz <- function(x, ...) {
  x %>%
    DiagrammeRsvg::export_svg() %>%
    c("`````{=html}\n", ., "\n`````\n") %>%
    knitr::asis_output()
}
library(nycflights13)
library(dm)
```

This draft describes update operations on tables and dm objects.
The generics for the table operations can be private in dm for now and move to dplyr later.
Contains parts from <https://github.com/tidyverse/dplyr/issues/4654>.

## Tables

In dplyr, operations on tables in dplyr are generally transient or ephemeral.
Resulting table objects must be stored in a new object, otherwise they are lost.
All operations by default return a new table object that is disconnected from the original.


### Operations

The operations are modeled after existing database statements, with the exception of the new "patch".

* insert new rows — would error if keys already exist, similar to `tibble::add_row()`.

* update values — overrides existing values. (Similar to https://github.com/tidyverse/tidyr/issues/183)

* patch values — like update, but only replaces missing values (Also similar to https://github.com/tidyverse/tidyr/issues/183)

* upsert — update or insert depending on presence/absence of keys

All operations would either take multiple named inputs or a single unnamed data frame.
Additional restrictions and options may apply on other backends, these will be specified with arguments that start with a dot.
For extensibility, named inputs that start with a dot are silently discarded.

All operations require specification of a `.key` argument.
Unlike `by` with joins, the `.key` argument is mandatory, because RHS column names must be a subset of the LHS column names.
Keys must have the same name in the target and the input.
If the table container knows its keys (e.g., grouped data frames or data.table), `.key` may be omitted.

Both target table and source columns/table must be compatible:

- Source has no extra tables and columns.
- Key columns must be present in source.


### Mutable backends

Some *mutable* backends, most notably databases, Google sheets and data.table, permit in-place update of the source data.
Update operations on mutable backends should optionally allow updating the source data.
Because this is a potentially destructive exception from the dplyr guarantees, in-place updates must be "opt in".
The default result always will be a "lazy" table.
This allows previewing the results of an update operation before materializing.

On these backends, the following additional operations are useful:

* delete — remove rows that match keys, a variant of `anti_join()`

* truncate — remove all rows

Databases will require that the data is already on the database or ask the user to supply a `copy` argument.

On mutable backends, update operations return the input, invisibly, if the update was carried out in-place.

Challenges:

- How to make compatible with `sql_render()`?


## dm

Operations on a dm object are generally transient or ephemeral.
Resulting dm or table objects must be stored in a new object, otherwise they are lost.


### Design

- records from tables from a source dm are appended/updated/upserted/removed/replaced/... from the target dm
- both dm must be compatible
    - same source
    - source is a subset of target tables
- dry run supported: use transient operations instead of materialization
    - check integrity constraints are still valid after running
    - compare before-after state
- transactions out of scope, caller can use `DBI::withTransaction()`
- some operations need top-down, others need bottom-up
    - fixed set of operations, each op knows its "direction"

### Logic of operation

- check compatibility
- persist tables one by one, topologically sorted
- put back into dm if necessary


### API draft

```{r}
dm_insert <- function(target_dm, dm, ..., dry_run = FALSE) {
  check_dots_empty()

  dm_persist(target_dm, dm, operation = tbl_insert, top_down = TRUE, dry_run = dry_run)
}

dm_update <- function(target_dm, dm, ..., dry_run = FALSE) {
  check_dots_empty()

  dm_persist(target_dm, dm, operation = tbl_update, top_down = TRUE, dry_run = dry_run)
}

dm_upsert <- function(target_dm, dm, ..., dry_run = FALSE) {
  check_dots_empty()

  dm_persist(target_dm, dm, operation = tbl_upsert, top_down = TRUE, dry_run = dry_run)
}

dm_delete <- function(target_dm, dm, ..., dry_run = FALSE) {
  check_dots_empty()

  dm_persist(target_dm, dm, operation = tbl_delete, top_down = FALSE, dry_run = dry_run)
}

dm_truncate <- function(target_dm, dm, ..., dry_run = FALSE) {
  check_dots_empty()

  dm_persist(target_dm, dm, operation = tbl_truncate, top_down = FALSE, dry_run = dry_run)
}

dm_persist <- function(target_dm, dm, operation, top_down, dry_run = FALSE) {
  dm_check_persist(target_dm, dm)

  dm_run_persist(target_dm, dm, operation, top_down, dry_run)
}

dm_check_persist <- function(target_dm, dm) {
  check_not_zoomed(target_dm)
  check_not_zoomed(dm)

  check_same_src(target_dm, dm)
  walk2(dm_get_tables(target_dm), dm_get_tables(dm), check_columns_superset)
  check_keys_compatible(target_dm, dm)
}

dm_run_persist <- function(target_dm, dm, operation, top_down, dry_run) {
  # topologically sort tables
  # run operation(target_tbl, source_tbl, dry_run = dry_run) for each table
  # operation() always returns tbl, only need to patch if not the same tbl
  # new_tables is list of non-NULL operation() values

  target_dm %>%
    dm_patch_tbl(!!!new_tables)
}

dm_patch_tbl <- function(dm, ...) {
  check_not_zoomed(dm)

  new_tables <- list2(...)

  # FIXME: Better error message for unknown tables

  def <- dm_get_def(dm)
  idx <- match(names(new_tables), def$table)
  def[idx, "data"] <- unname(new_tables)
  dm_from_def(def)
}
```

# FIXME NEXT

- un-skip tests

# Later

- Store `keys` objects in dm, much easier for debugging
- Add "strict mode" to GitHub Actions: validation in `dm_from_def()`
- Named PK and unique constraints: https://github.com/r-dbi/DBI/pull/351#issuecomment-833438890
- `dm_paste()`: remove `select` argument from documentation, via `_impl()` function that takes dots and this argument, like `dm_rm_pk()`
- Persistent test dm objects
  - Use `copy_to(temporary = FALSE)`
  - sqlite and duckdb: use file that can be discarded
  - Requires schema support for all databases

<!-- Generated by galley: do not edit by hand -->

The {dm} package offers functions to work with relational data models in
R.

This document introduces you to filtering functions, and shows how to
apply them to the data that is separated into multiple tables.

Our example data is drawn from the
[{nycflights13}](https://github.com/tidyverse/nycflights13) package that
contains five inter-linked tables.

First, we will load the packages that we need:

``` r
library(tidyverse)
library(nycflights13)
library(dm)
```

## Data: nycflights13

To explore filtering with {dm}, we’ll use the {nycflights13} data with
its `flights`, `planes`, `airlines`, `airports` and `weather` tables.

This dataset contains information about the 336 776 flights that
departed from New York City in 2013, with 3322 different planes and 1458
airports involved. The data comes from the US Bureau of Transportation
Statistics, and is documented in `?nycflights13::flights`.

To start with our exploration, we have to create a `dm` object from the
{nycflights13} data. The built-in `dm::dm_nycflights13()` function takes
care of this.

By default it only uses a subset of the complete data though: only the
flights on the 10th of each month are considered, reducing the number of
rows in the `flights` table to 11 227.

A [data model
object](https://cynkra.github.io/dm/articles/tech-dm-class.html#class-dm)
contains data from the source tables, and metadata about the tables.

If you would like to create a `dm` object from tables other than the
example data, you can use the `new_dm()`, `dm()` or `as_dm()` functions.
See `vignette("howto-dm-df")` for details.

``` r
dm <- dm_nycflights13()
```

The console output of the ’dm\` object shows its data and metadata, and
is colored for clarity:

``` r
dm
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 4
#> Foreign keys: 4
```

Now we know that there are five tables in our `dm` object. But how are
they connected? These relations are best displayed as a visualization of
the entity-relationship model:

``` r
dm_draw(dm)
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-filter_files/figure-gfm/unnamed-chunk-4-1.png)<!-- -->

You can look at a single table with `tbl`. To print the `airports`
table, call

``` r
tbl(dm, "airports")
#> # A tibble: 1,458 x 8
#>    faa   name                    lat    lon   alt    tz dst   tzone        
#>    <chr> <chr>                 <dbl>  <dbl> <dbl> <dbl> <chr> <chr>        
#>  1 04G   Lansdowne Airport      41.1  -80.6  1044    -5 A     America/New_…
#>  2 06A   Moton Field Municipa…  32.5  -85.7   264    -6 A     America/Chic…
#>  3 06C   Schaumburg Regional    42.0  -88.1   801    -6 A     America/Chic…
#>  4 06N   Randall Airport        41.4  -74.4   523    -5 A     America/New_…
#>  5 09J   Jekyll Island Airport  31.1  -81.4    11    -5 A     America/New_…
#>  6 0A9   Elizabethton Municip…  36.4  -82.2  1593    -5 A     America/New_…
#>  7 0G6   Williams County Airp…  41.5  -84.5   730    -5 A     America/New_…
#>  8 0G7   Finger Lakes Regiona…  42.9  -76.8   492    -5 A     America/New_…
#>  9 0P2   Shoestring Aviation …  39.8  -76.6  1000    -5 U     America/New_…
#> 10 0S9   Jefferson County Intl  48.1 -123.    108    -8 A     America/Los_…
#> # … with 1,448 more rows
```

## Filtering a `dm` object

`dm_filter()` allows you to select a subset of a `dm` object.

### How it works

Filtering a `dm` object is not that different from filtering a dataframe
or tibble with `dplyr::filter()`.

The corresponding {dm} function is `dm::dm_filter()`. With this function
one or more filtering conditions can be set for one of the tables of the
`dm` object. These conditions are immediately evaluated for their
respective tables and in addition they are stored in the `dm`. There are
two ways in which a condition on one table could affect another table:

1.  Calling `dm_apply_filters_to_tbl()`, i.e. requesting a table from a
    `dm` after the filter conditions are applied. In this case, all
    tables that are connected to the requested table and have stored
    filter conditions associated with them are taken into account in the
    following way:
    -   filtering semi-joins are successively performed along the paths
        from each of the filtered tables to the requested table, each
        join reducing the left-hand side tables of the joins to only
        those of their rows with key values that have corresponding
        values in key columns of the right-hand side tables of the join.
    -   eventually the requested table is returned, containing only the
        the remaining rows after the filtering joins
2.  Calling `dm_apply_filters()` or `compute()` methods for `dm` objects
    on a `dm`: this results in a new `dm` that contains the same tables
    as before. Each table is the result of a `dm_apply_filters_to_tbl()`
    call, meaning that the effects of the filter conditions on each of
    the tables are taken into account.

Currently, this only works if the graph induced by the foreign key
relations is cycle free. Fortunately, this is the default for
`dm_nycflights13()`.

Keep in mind that several {dm} functions will refuse to work when
unevaluated filter conditions exist, such as,
e.g. `dm_enum_fk_candidates()`, `dm_enum_pk_candidates()`,
`dm_select_tbl()`, `dm_rename_tbl()`, `dm_select()`, `dm_rename()` and
`dm_nrow()`. In these cases consider applying the filters with
`dm_apply_filters()` first.

### Filtering Examples

Let’s see filtering in action:

**We only want the data that is related to John F. Kennedy International
Airport.**

``` r
filtered_dm <-
  dm %>%
  dm_filter(airports, name == "John F Kennedy Intl")
filtered_dm
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 4
#> Foreign keys: 4
#> ── Filters ────────────────────────────────────────────────────────────────
#> airports: name == "John F Kennedy Intl"
```

The filter expression is listed in the print output.

You can get the numbers of rows of each table with `dm_nrow()`. Before
doing that, you will need to apply the filters using
`dm_apply_filters()`:

``` r
rows_per_table <-
  filtered_dm %>%
  dm_apply_filters() %>%
  dm_nrow()
rows_per_table
#> airlines airports  flights   planes  weather 
#>       10        1     3661      783      228
sum(rows_per_table)
#> [1] 4683
```

``` r
sum_nrow <- sum(dm_nrow(dm))
sum_nrow_filtered <- sum(dm_nrow(dm_apply_filters(filtered_dm)))
```

The total number of rows in the `dm` drops from 16 884 to 4 683 (the
only unaffected table is the disconnected `weather` table).

Next example:

**Get a `dm` object containing data for flights from New York to the
Dulles International Airport in Washington D.C., abbreviated with
`IAD`.**

``` r
dm %>%
  dm_filter(flights, dest == "IAD") %>%
  dm_apply_filters() %>%
  dm_nrow()
#> airlines airports  flights   planes  weather 
#>        4        3      191       95      178
```

Chaining multiple filters on different tables is also supported.

An example:

**Get all flights from Delta Air Lines which didn’t depart from John F.
Kennedy International Airport in May 2013.**

``` r
dm_delta_may <-
  dm %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_filter(flights, month == 5)
dm_delta_may
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 4
#> Foreign keys: 4
#> ── Filters ────────────────────────────────────────────────────────────────
#> airlines: name == "Delta Air Lines Inc."
#> airports: name != "John F Kennedy Intl"
#> flights: month == 5
dm_delta_may %>%
  dm_apply_filters() %>%
  dm_nrow()
#> airlines airports  flights   planes  weather 
#>        1        2       79       61       27
```

You can inspect the filtered tables with `dm_apply_filters_to_tbl()`.

In the `airlines` table, Delta is the only remaining carrier:

``` r
dm_delta_may %>%
  dm_apply_filters_to_tbl("airlines")
#> # A tibble: 1 x 2
#>   carrier name                
#>   <chr>   <chr>               
#> 1 DL      Delta Air Lines Inc.
```

Which planes were used to service these flights?

``` r
dm_delta_may %>%
  dm_apply_filters_to_tbl("planes")
#> # A tibble: 61 x 9
#>    tailnum  year type       manufacturer  model  engines seats speed engine
#>    <chr>   <int> <chr>      <chr>         <chr>    <int> <int> <int> <chr> 
#>  1 N305DQ   2008 Fixed win… BOEING        737-7…       2   149    NA Turbo…
#>  2 N310DE   2009 Fixed win… BOEING        737-7…       2   149    NA Turbo…
#>  3 N312US   1990 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#>  4 N313US   1990 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#>  5 N314US   1991 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#>  6 N318NB   2000 Fixed win… AIRBUS INDUS… A319-…       2   145    NA Turbo…
#>  7 N322NB   2001 Fixed win… AIRBUS INDUS… A319-…       2   145    NA Turbo…
#>  8 N325NB   2001 Fixed win… AIRBUS INDUS… A319-…       2   145    NA Turbo…
#>  9 N329NW   1992 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#> 10 N332NW   1992 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#> # … with 51 more rows
```

And indeed, all included flights departed in May (`month == 5`):

``` r
dm_delta_may %>%
  dm_apply_filters_to_tbl("flights")
#> # A tibble: 79 x 19
#>     year month   day dep_time sched_dep_time dep_delay arr_time
#>    <int> <int> <int>    <int>          <int>     <dbl>    <int>
#>  1  2013     5    10      554            600        -6      739
#>  2  2013     5    10      556            600        -4      825
#>  3  2013     5    10      606            610        -4      743
#>  4  2013     5    10      625            630        -5      843
#>  5  2013     5    10      653            700        -7      923
#>  6  2013     5    10      656            700        -4      911
#>  7  2013     5    10      700            700         0      958
#>  8  2013     5    10      701            705        -4      952
#>  9  2013     5    10      714            715        -1      908
#> 10  2013     5    10      743            745        -2      956
#> # … with 69 more rows, and 12 more variables: sched_arr_time <int>,
#> #   arr_delay <dbl>, carrier <chr>, flight <int>, tailnum <chr>,
#> #   origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>, hour <dbl>,
#> #   minute <dbl>, time_hour <dttm>
```

For comparison, let’s review the equivalent manual query for `flights`
in `dplyr` syntax:

``` r
airlines_filtered <- filter(airlines, name == "Delta Air Lines Inc.")
airports_filtered <- filter(airports, name != "John F Kennedy Intl")
flights %>%
  semi_join(airlines_filtered, by = "carrier") %>%
  semi_join(airports_filtered, by = c("origin" = "faa")) %>%
  filter(month == 5)
#> # A tibble: 2,340 x 19
#>     year month   day dep_time sched_dep_time dep_delay arr_time
#>    <int> <int> <int>    <int>          <int>     <dbl>    <int>
#>  1  2013     5     1      554            600        -6      731
#>  2  2013     5     1      555            600        -5      819
#>  3  2013     5     1      603            610        -7      754
#>  4  2013     5     1      622            630        -8      848
#>  5  2013     5     1      654            700        -6      931
#>  6  2013     5     1      655            700        -5      944
#>  7  2013     5     1      656            705        -9     1005
#>  8  2013     5     1      658            700        -2      925
#>  9  2013     5     1      743            745        -2     1014
#> 10  2013     5     1      755            800        -5      929
#> # … with 2,330 more rows, and 12 more variables: sched_arr_time <int>,
#> #   arr_delay <dbl>, carrier <chr>, flight <int>, tailnum <chr>,
#> #   origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>, hour <dbl>,
#> #   minute <dbl>, time_hour <dttm>
```

The {dm} code is leaner because the foreign key information is encoded
in the object.

## SQL statements behind filtering a `dm` object on a database

{dm} is meant to work with relational data models, locally as well as on
databases. In your project, the data is probably not stored locally but
in a remote [relational
database](https://cynkra.github.io/dm/articles/howto-dm-theory.html#relational-databases)
that can be queried with SQL statements.

You can check the queries by using `sql_render()` from the
[{dbplyr}](https://dbplyr.tidyverse.org/) package.

Example:

**Print the SQL statements for getting all flights from Delta Air Lines,
which did not depart from John F. Kennedy International Airport in May
2013, with the data stored in a sqlite database.**

To show the SQL query behind a `dm_filter()`, we copy the `flights`,
`airlines` and `airports` tables from the `nyflights13` dataset to a
temporary in-memory database using the built-in function `copy_dm_to()`
and `dbplyr::src_memdb`.

Then we filter the data, and print the corresponding SQL statement with
`dbplyr::sql_render()`.

``` r
dm %>%
  dm_select_tbl(flights, airlines, airports) %>%
  dm_filter(flights, month == 5) %>%
  copy_dm_to(dbplyr::src_memdb(), .) %>%
  dm_filter(airlines, name == "Delta Air Lines Inc.") %>%
  dm_filter(airports, name != "John F Kennedy Intl") %>%
  dm_apply_filters() %>%
  dm_get_tables() %>%
  purrr::map(dbplyr::sql_render)
#> $flights
#> <SQL> SELECT * FROM (SELECT * FROM `flights_2020_08_28_07_13_03_12345_1` AS `LHS`
#> WHERE EXISTS (
#>   SELECT 1 FROM (SELECT *
#> FROM `airlines_2020_08_28_07_13_03_12345_1`
#> WHERE (`name` = 'Delta Air Lines Inc.')) AS `RHS`
#>   WHERE (`LHS`.`carrier` = `RHS`.`carrier`)
#> )) AS `LHS`
#> WHERE EXISTS (
#>   SELECT 1 FROM (SELECT *
#> FROM `airports_2020_08_28_07_13_03_12345_1`
#> WHERE (`name` != 'John F Kennedy Intl')) AS `RHS`
#>   WHERE (`LHS`.`origin` = `RHS`.`faa`)
#> )
#> 
#> $airlines
#> <SQL> SELECT * FROM (SELECT *
#> FROM `airlines_2020_08_28_07_13_03_12345_1`
#> WHERE (`name` = 'Delta Air Lines Inc.')) AS `LHS`
#> WHERE EXISTS (
#>   SELECT 1 FROM (SELECT * FROM `flights_2020_08_28_07_13_03_12345_1` AS `LHS`
#> WHERE EXISTS (
#>   SELECT 1 FROM (SELECT *
#> FROM `airports_2020_08_28_07_13_03_12345_1`
#> WHERE (`name` != 'John F Kennedy Intl')) AS `RHS`
#>   WHERE (`LHS`.`origin` = `RHS`.`faa`)
#> )) AS `RHS`
#>   WHERE (`LHS`.`carrier` = `RHS`.`carrier`)
#> )
#> 
#> $airports
#> <SQL> SELECT * FROM (SELECT *
#> FROM `airports_2020_08_28_07_13_03_12345_1`
#> WHERE (`name` != 'John F Kennedy Intl')) AS `LHS`
#> WHERE EXISTS (
#>   SELECT 1 FROM (SELECT * FROM `flights_2020_08_28_07_13_03_12345_1` AS `LHS`
#> WHERE EXISTS (
#>   SELECT 1 FROM (SELECT *
#> FROM `airlines_2020_08_28_07_13_03_12345_1`
#> WHERE (`name` = 'Delta Air Lines Inc.')) AS `RHS`
#>   WHERE (`LHS`.`carrier` = `RHS`.`carrier`)
#> )) AS `RHS`
#>   WHERE (`LHS`.`faa` = `RHS`.`origin`)
#> )
```

Further reading: {dm}’s function for copying data [from and to
databases](https://cynkra.github.io/dm/articles/dm.html#copy).

<!-- Generated by galley: do not edit by hand -->

This vignette describes which changes are necessary to adapt your code
when updating the {dm} package version from a version `0.0.5` or lower
to `0.0.6` or higher.

## Changes required when updating from version `0.0.5` to `0.0.6`

### Replace `cdm` with `dm`

During this update the prevalent prefix `cdm` was discarded in favor of
`dm`. The old prefix would still do its job, but a warning message would
be issued each time a function beginning with `cdm` was being used,
informing that the function is soft-deprecated and suggesting the use of
its newer version.

If you have a script which is based on an older {dm} version, it should
still work with the newer version, albeit complaining each time an
outdated function is being used. This can be repaired by:

1.  either going through the script step by step, testing the output of
    each line of code and use the new function names provided in the
    generated warnings to update the function calls.
2.  or just by replacing all occurrences of `cdm` by `dm` in this
    script. This can e.g. be done in RStudio using “Find” or in the
    terminal using `sed -e 's/cdm/dm/g' path-to-file` on Windows or
    `sed -i '' -e 's/cdm/dm/g' path-to-file` on a Mac. If the script
    errors after this step, you will need to check where exactly the
    error happens and manually repair the damage.

### Be careful with methods for `dm`: `tbl`, `[[`, `$`

Furthermore, you need to pay attention if you used one of `tbl.dm()`,
`[[.dm()`, `$.dm()`. During the same update the implementation for those
methods changed as well, and here you don’t get the convenient warning
messages. The change was, that before the update, the mentioned methods
would return the table after “filtering” it to just contain the rows
with values that relate via foreign key relations to other tables that
were filtered earlier. After the update just the table as is would be
returned. If you want to retain the former behavior, you need to replace
each of the methods with the function `dm_apply_filters_to_tbl()`, which
was made available with the update.

The methods are of course not to be avoided in general, if no filters
are set anyway the result will not change after the update.

Here a short example for the different cases:

Formerly you would access the “filtered” tables using the following
syntax:

``` r
library(dm)
flights_dm <- dm_nycflights13()
tbl(flights_dm, "airports")
#> # A tibble: 1,458 x 8
#>    faa   name                    lat    lon   alt    tz dst   tzone        
#>    <chr> <chr>                 <dbl>  <dbl> <dbl> <dbl> <chr> <chr>        
#>  1 04G   Lansdowne Airport      41.1  -80.6  1044    -5 A     America/New_…
#>  2 06A   Moton Field Municipa…  32.5  -85.7   264    -6 A     America/Chic…
#>  3 06C   Schaumburg Regional    42.0  -88.1   801    -6 A     America/Chic…
#>  4 06N   Randall Airport        41.4  -74.4   523    -5 A     America/New_…
#>  5 09J   Jekyll Island Airport  31.1  -81.4    11    -5 A     America/New_…
#>  6 0A9   Elizabethton Municip…  36.4  -82.2  1593    -5 A     America/New_…
#>  7 0G6   Williams County Airp…  41.5  -84.5   730    -5 A     America/New_…
#>  8 0G7   Finger Lakes Regiona…  42.9  -76.8   492    -5 A     America/New_…
#>  9 0P2   Shoestring Aviation …  39.8  -76.6  1000    -5 U     America/New_…
#> 10 0S9   Jefferson County Intl  48.1 -123.    108    -8 A     America/Los_…
#> # … with 1,448 more rows
flights_dm$planes
#> # A tibble: 3,322 x 9
#>    tailnum  year type       manufacturer  model  engines seats speed engine
#>    <chr>   <int> <chr>      <chr>         <chr>    <int> <int> <int> <chr> 
#>  1 N10156   2004 Fixed win… EMBRAER       EMB-1…       2    55    NA Turbo…
#>  2 N102UW   1998 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#>  3 N103US   1999 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#>  4 N104UW   1999 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#>  5 N10575   2002 Fixed win… EMBRAER       EMB-1…       2    55    NA Turbo…
#>  6 N105UW   1999 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#>  7 N107US   1999 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#>  8 N108UW   1999 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#>  9 N109UW   1999 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#> 10 N110UW   1999 Fixed win… AIRBUS INDUS… A320-…       2   182    NA Turbo…
#> # … with 3,312 more rows
flights_dm[["weather"]]
#> # A tibble: 861 x 15
#>    origin  year month   day  hour  temp  dewp humid wind_dir wind_speed
#>    <chr>  <int> <int> <int> <int> <dbl> <dbl> <dbl>    <dbl>      <dbl>
#>  1 EWR     2013     1    10     0  41    32    70.1      230       8.06
#>  2 EWR     2013     1    10     1  39.0  30.0  69.9      210       9.21
#>  3 EWR     2013     1    10     2  39.0  28.9  66.8      230       6.90
#>  4 EWR     2013     1    10     3  39.9  27.0  59.5      270       5.75
#>  5 EWR     2013     1    10     4  41    26.1  55.0      320       6.90
#>  6 EWR     2013     1    10     5  41    26.1  55.0      300      12.7 
#>  7 EWR     2013     1    10     6  39.9  25.0  54.8      280       6.90
#>  8 EWR     2013     1    10     7  41    25.0  52.6      330       6.90
#>  9 EWR     2013     1    10     8  43.0  25.0  48.7      330       8.06
#> 10 EWR     2013     1    10     9  45.0  23    41.6      320      17.3 
#> # … with 851 more rows, and 5 more variables: wind_gust <dbl>,
#> #   precip <dbl>, pressure <dbl>, visib <dbl>, time_hour <dttm>
```

After the update the same result is achieved by this type of function
call:

``` r
dm_apply_filters_to_tbl(flights_dm, airlines)
#> # A tibble: 16 x 2
#>    carrier name                       
#>    <chr>   <chr>                      
#>  1 9E      Endeavor Air Inc.          
#>  2 AA      American Airlines Inc.     
#>  3 AS      Alaska Airlines Inc.       
#>  4 B6      JetBlue Airways            
#>  5 DL      Delta Air Lines Inc.       
#>  6 EV      ExpressJet Airlines Inc.   
#>  7 F9      Frontier Airlines Inc.     
#>  8 FL      AirTran Airways Corporation
#>  9 HA      Hawaiian Airlines Inc.     
#> 10 MQ      Envoy Air                  
#> 11 OO      SkyWest Airlines Inc.      
#> 12 UA      United Air Lines Inc.      
#> 13 US      US Airways Inc.            
#> 14 VX      Virgin America             
#> 15 WN      Southwest Airlines Co.     
#> 16 YV      Mesa Airlines Inc.
```

# Contributing to dm

This outlines how to propose a change to dm and how to set up a local development environment with test databases.

## Fixing typos

You can fix typos, spelling mistakes, or grammatical errors in the documentation directly using the GitHub web interface, as long as the changes are made in the _source_ file. 
This generally means you'll need to edit [roxygen2 comments](https://roxygen2.r-lib.org/articles/roxygen2.html) in an `.R`, not a `.Rd` file. 
You can find the `.R` file that generates the `.Rd` by reading the comment in the first line.

## Bigger changes

If you want to make a bigger change, it's a good idea to first file an issue and make sure someone from the team agrees that it’s needed. 
If you’ve found a bug, please file an issue that illustrates the bug with a minimal 
[reprex](https://www.tidyverse.org/help/#reprex) (this will also help you write a unit test, if needed).

### Pull request process

*   Fork the package and clone onto your computer. If you haven't done this before, we recommend using `usethis::create_from_github("cynkra/dm", fork = TRUE)`.

*   Install all development dependencies with `devtools::install_dev_deps()`, and then make sure the package passes R CMD check by running `devtools::check()`. 
    If R CMD check doesn't pass cleanly, it's a good idea to ask for help before continuing. 
*   Create a Git branch for your pull request (PR). We recommend using `usethis::pr_init("brief-description-of-change")`.

*   Make your changes, commit to git, and then create a PR by running `usethis::pr_push()`, and following the prompts in your browser.
    The title of your PR should briefly describe the change.
    The body of your PR should contain `Fixes #issue-number`.

*  For user-facing changes, add a bullet to the top of `NEWS.md` (i.e. just below the first header). Follow the style described in <https://style.tidyverse.org/news.html>.

## Code style & design

### General remarks

*   New code should follow the tidyverse [style guide](https://style.tidyverse.org). 
    You can use the [styler](https://CRAN.R-project.org/package=styler) package to apply these styles, but please don't restyle code that has nothing to do with your PR.  

*  We use [roxygen2](https://cran.r-project.org/package=roxygen2), with [Markdown syntax](https://cran.r-project.org/web/packages/roxygen2/vignettes/rd-formatting.html), for documentation.  

* We use a specific [branch of downlit](https://github.com/r-lib/downlit/tree/f-readme-document) for knitting the README.

### Testing
  
* We use {testthat}.

* All exported functions should be tested.

* R scripts under "tests/testthat/"" whose name start with ["helper"](https://blog.r-hub.io/2020/11/18/testthat-utility-belt/#code-called-in-your-tests) are loaded with `devtools::load_all()` so they are available for both tests and interactive  debugging, but can't be found with `:::`.
  
* "helper-src.R" implements a complex mechanism so that some dm creating functions, such as `dm_for_filter()`, 
will create a remote dm in a different database management system depending on context. 
This allows tests that use those to be run on different setups through github actions. 
Additionally `my_db_test_src()` will return the relevant database. 
In order to test databases locally (typically to debug if CI tests fails and we can't debug from the online log)
  we can set the environ variable "DM_TEST_SRC" to "postgres", "mariadb", "mssql", "duckdb" or "sqlite".
  You might have to setup credentials in "helper-config-db.R" to do so.
* Some useful expectations can be found in "helper-expectations.R".
* In "helper-skip.R" are some helpers to skip tests in some contexts, these might be useful for instance if a feature is not supported on some databases.
* When using `expect_snapshot()` on a DBMS dependent call (i.e a call that uses `dm_for_filter()` or copies to `my_db_test_src()`), 
  the `variant` argument should be   set to `my_test_src_name` (a global  variable created when loading helpers) so that the snapshots end up in different directories.
    
### Function naming

See vignette [function naming logic](https://cynkra.github.io/dm/articles/tech-dm-naming.html).

### Error messages

We strive to standardise error messages in {dm}. A failure should be triggered by `abort()` through a function defined in "error-helpers.R" where the error class is defined with `dm_error_full()` and the error message is created by a separate function. Please follow the pattern used in `"error-helpers.R"`.  Exceptions might exist but they are mostly waiting to be harmonised.

## Test databases

This repository comes with a `docker-compose.yml` file that sets up the databases required for testing.
All shell commands are expected to be run from the top-level directory of a clone of this repository.

### macOS only: Install colima to run Docker containers

```sh
brew install colima docker-compose
colima start -c 4 -m 4 --vm-type vz --vz-rosetta --network-address
colima status
#  INFO[0000] colima is running using macOS Virtualization.Framework
#  INFO[0000] arch: aarch64
#  INFO[0000] runtime: docker
#  INFO[0000] mountType: virtiofs
#  INFO[0000] address: 192.168.64.2
#  INFO[0000] socket: unix:///Users/kirill/.colima/default/docker.sock
```

Take note of the `address` in the output, it will be used later.

See also <https://docs.google.com/document/d/1axInaYK6oK6riRio72uTAeQazuork1X0clY9UL9gYoE/edit?usp=sharing> for more details on colima.

### mssql: ODBC drivers

- Linux: <https://learn.microsoft.com/en-us/sql/connect/odbc/linux-mac/installing-the-microsoft-odbc-driver-for-sql-server?view=sql-server-ver16&tabs=ubuntu18-install%2Calpine17-install%2Cdebian8-install%2Credhat7-13-install%2Crhel7-offline>

- macOS: <https://learn.microsoft.com/en-us/sql/connect/odbc/linux-mac/install-microsoft-odbc-driver-sql-server-macos?view=sql-server-ver16>

### Start new database containers

```sh
make db-start
# May take several minutes to pull the images
```

### Start database containers without forcing recreation

```sh
make db-restart
# May take several minutes to pull the images
```

### Connectivity test

macOS:

```sh
DM_TEST_DOCKER_HOST=192.168.64.2 make connect
```

Linux:

```sh
DM_TEST_DOCKER_HOST=127.0.0.1 make connect
```

Controlled with environment variables:

- `DM_TEST_DOCKER_HOST`: `127.0.0.1` or `localhost` on Linux, see output of `colima status` on macOS

    (On Linux, using `localhost` instead of `127.0.0.1` may cause problems for MariaDB.)

See also the `Makefile`.

It is recommended to adapt your `.Renviron` once connectivity has been established.
The subsequent instructions omit setting the environment variables explicitly.


### Test against a specific database backend

```sh
make test-postgres
```


### Test against all backends

```sh
make -j1 test
```


### Test on Docker

```sh
# make docker-build # not necessary, image available on ghcr.io
make docker-test
```


## Code of Conduct

Please note that the dm project is released with a
[Contributor Code of Conduct](CODE_OF_CONDUCT.md). By contributing to this
project you agree to abide by its terms.

<!-- Generated by galley: do not edit by hand -->

In this document we will present several specialized functions for
conducting basic tests about key conditions and about relations between
tables. We will also describe functions that can be used for splitting
and uniting tables.

``` r
library(tidyverse)
#> ── Attaching packages ────────────────────────────────── tidyverse 1.3.1 ──
#> ✔ ggplot2 3.3.3     ✔ purrr   0.3.4
#> ✔ tibble  3.1.1     ✔ dplyr   1.0.5
#> ✔ tidyr   1.1.3     ✔ stringr 1.4.0
#> ✔ readr   1.4.0     ✔ forcats 0.5.1
#> ── Conflicts ───────────────────────────────────── tidyverse_conflicts() ──
#> ✖ purrr::%@%()         masks dm::%@%()
#> ✖ purrr::as_function() masks dm::as_function()
#> ✖ tidyr::extract()     masks dm::extract()
#> ✖ dplyr::filter()      masks dm::filter(), stats::filter()
#> ✖ purrr::flatten()     masks dm::flatten()
#> ✖ purrr::flatten_chr() masks dm::flatten_chr()
#> ✖ purrr::flatten_dbl() masks dm::flatten_dbl()
#> ✖ purrr::flatten_int() masks dm::flatten_int()
#> ✖ purrr::flatten_lgl() masks dm::flatten_lgl()
#> ✖ purrr::flatten_raw() masks dm::flatten_raw()
#> ✖ purrr::invoke()      masks dm::invoke()
#> ✖ purrr::is_null()     masks dm::is_null(), testthat::is_null()
#> ✖ dplyr::lag()         masks dm::lag(), stats::lag()
#> ✖ purrr::list_along()  masks dm::list_along()
#> ✖ dplyr::matches()     masks tidyr::matches(), dm::matches(), testthat::matches()
#> ✖ purrr::modify()      masks dm::modify()
#> ✖ purrr::prepend()     masks dm::prepend()
#> ✖ purrr::splice()      masks dm::splice()
library(dm)
```

## Testing key constraints

This section contains information and examples about the following
functions:

1.  `check_key(.data, ...)`
2.  `check_subset(t1, c1, t2, c2)`
3.  `check_set_equality(t1, c1, t2, c2)`

When you have tables (data frames) that are connected by key relations,
this package can help you to verify the assumed key relations and/or
determine the existing key relations between the tables. For example, if
you have tables:

``` r
data_1 <- tibble(a = c(1, 2, 1), b = c(1, 4, 1), c = c(5, 6, 7))
data_2 <- tibble(a = c(1, 2, 3), b = c(4, 5, 6), c = c(7, 8, 9))
```

and you want to know if `a` is a primary key for `data_1`, you can use
the `check_key()` function:

``` r
check_key(data_1, a)
#> Error: (`a`) not a unique key of `data_1`.
```

Mind the error message when a test is not passed.

For `data_2`, column `a` is a key:

``` r
check_key(data_2, a)
```

To see if a column of one table contains only those values that are also
present in another column of another table, the `check_subset()`
function can be used:

``` r
check_subset(data_1, a, data_2, a)
```

This function is important for determining if a column is a foreign key
to some other table. What about the inverse relation?

``` r
check_subset(data_2, a, data_1, a)
#> # A tibble: 1 x 3
#>       a     b     c
#>   <dbl> <dbl> <dbl>
#> 1     3     6     9
#> Error: Column `a` of table `data_2` contains values (see examples above) that are not present in column `a` of table `data_1`.
```

It should be kept in mind that `check_subset()` does not test if column
`c2` is a unique key of table `t2`. In order to find out if a (child)
table `t1` contains a column `c1` that is a foreign key to a (parent)
table `t2` with the corresponding column `c2`, the following method
should be used:

``` r
check_key(t2, c2)
check_subset(t1, c1, t2, c2)
```

To check both directions at once, and to find out if the unique values
of `c_1` in `t_1` are the same as those of `c_2` in `t_2`, {dm} provides
the function `check_set_equality()`:

``` r
check_set_equality(data_1, a, data_2, a)
#> # A tibble: 1 x 3
#>       a     b     c
#>   <dbl> <dbl> <dbl>
#> 1     3     6     9
#> Error: Column `a` of table `data_2` contains values (see examples above) that are not present in column `a` of table `data_1`..
```

Introducing one more table enables us to show how it looks when the test
is passed:

``` r
data_3 <- tibble(a = c(2, 1, 2), b = c(4, 5, 6), c = c(7, 8, 9))

check_set_equality(data_1, a, data_3, a)
```

If the test is passed, the return value of the function will be the
first table parameter (invisibly). This ensures that the functions can
be conveniently used in a pipe configuration.

## Testing cardinalities between two tables

This section contains information and examples for the functions

1.  `check_cardinality_0_n(parent_table, primary_key_column, child_table, foreign_key_column)`
2.  `check_cardinality_1_n(parent_table, primary_key_column, child_table, foreign_key_column)`
3.  `check_cardinality_0_1(parent_table, primary_key_column, child_table, foreign_key_column)`
4.  `check_cardinality_1_1(parent_table, primary_key_column, child_table, foreign_key_column)`
5.  `examine_cardinality(parent_table, primary_key_column, child_table, foreign_key_column)`

The four functions for testing for a specific kind of cardinality of the
relation all require a parent table and a child table as inputs. The
functions first test if that requirement is fulfilled by checking if:

1.  `primary_key_column` is a unique key for `parent_table`
2.  The set of values of `foreign_key_column` is a subset of the set of
    values of `primary_key_column`

The cardinality specifications `0_n`, `1_n`, `0_1`, `1_1` refer to the
expected relation that the child table has with the parent table. The
numbers ‘0’, ‘1’ and ‘n’ refer to the number of values in the child
table’s column (`foreign_key_column`) that correspond to each value of
the parent table’s column (`primary_key_column`). ‘n’ means more than
one in this context, with no upper limit.

`0_n` means, that for each value of the `parent_key_column`, the number
of corresponding records in the child table is unrestricted. `1_n`
means, that for each value of the `parent_key_column` there is at least
one corresponding record in the child table. This means that there is a
“surjective” relation from the child table to the parent table w.r.t.
the specified columns, i.e. for each parent table column value there
exists at least one equal child table column value.

`0_1` means, that for each value of the `parent_key_column`, at least
zero and at most one value has to correspond to it in the column of the
child table. This means that there is an “injective” relation from the
child table to the parent table w.r.t. the specified columns, i.e. no
parent table column value is addressed multiple times. But not all of
the parent table column values have to be referred to.

`1_1` means, that for each value of the `parent_key_column`, exactly one
value has to correspond to it in the child table’s column. This means
that there is a “bijective” (“injective” AND “surjective”) relation
between the child table and the parent table w.r.t. the specified
columns, i.e. the set of values of the two columns is equal and there
are no duplicates in either of them.

Also `examine_cardinality()` first performs the above mentioned tests to
figure out, if the parent-child table relationship criteria are met.
Subsequently, two further checks are made to determine the nature of the
relation (surjective, injective, bijective or none of these) between the
two columns.

### Examples

Given the following three data frames:

``` r
d1 <- tibble(a = 1:5)
d2 <- tibble(c = c(1:5,5))
d3 <- tibble(c = 1:4)
d4 <- tibble(a = c(2:5, 5))
```

Here are some examples of how the cardinality testing functions can be
used:

``` r
# This does not pass, `c` is not unique key of d2:
check_cardinality_0_n(d2, c, d1, a)
#> Error: (`c`) not a unique key of `d2`.

# This passes, multiple values in d2$c are allowed:
check_cardinality_0_n(d1, a, d2, c)

# This does not pass, injectivity is violated:
check_cardinality_1_1(d1, a, d2, c)
#> Error: 1..1 cardinality (bijectivity) is not given: Column `c` in table `d2` contains duplicate values.

# This passes:
check_cardinality_0_1(d1, a, d3, c)
```

`examine_cardinality()` returns the type of relation, e.g.:

``` r
examine_cardinality(d1, a, d3, c)
#> [1] "injective mapping (child: 0 or 1 -> parent: 1)"
examine_cardinality(d1, a, d2, c)
#> [1] "surjective mapping (child: 1 to n -> parent: 1)"
examine_cardinality(d1, a, d1, a)
#> [1] "bijective mapping (child: 1 -> parent: 1)"
examine_cardinality(d1, a, d4, a)
#> [1] "generic mapping (child: 0 to n -> parent: 1)"
```

## Table surgery

The relevant functions are:

1.  `decompose_table(.data, new_id_column, ...)`
2.  `reunite_parent_child(child_table, parent_table, id_column)`
3.  `reunite_parent_child_from_list(list_of_parent_child_tables, id_column)`

The first function implements table normalization. An existing table is
split into a parent table (i.e. a lookup table) and a child table
(containing the observations), linked by a key column (here:
`new_id_column`). Basically, a foreign key relation would be created,
pointing from the `new_id_column` of the child table to the parent
table’s corresponding column, which can be seen as the parent table’s
primary key column. The function `decompose_table()` does that, as can
be seen in the following example:

``` r
mtcars_tibble <- as_tibble(mtcars)
mtcars_tibble
#> # A tibble: 32 x 11
#>      mpg   cyl  disp    hp  drat    wt  qsec    vs    am  gear  carb
#>    <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl>
#>  1  21       6  160    110  3.9   2.62  16.5     0     1     4     4
#>  2  21       6  160    110  3.9   2.88  17.0     0     1     4     4
#>  3  22.8     4  108     93  3.85  2.32  18.6     1     1     4     1
#>  4  21.4     6  258    110  3.08  3.22  19.4     1     0     3     1
#>  5  18.7     8  360    175  3.15  3.44  17.0     0     0     3     2
#>  6  18.1     6  225    105  2.76  3.46  20.2     1     0     3     1
#>  7  14.3     8  360    245  3.21  3.57  15.8     0     0     3     4
#>  8  24.4     4  147.    62  3.69  3.19  20       1     0     4     2
#>  9  22.8     4  141.    95  3.92  3.15  22.9     1     0     4     2
#> 10  19.2     6  168.   123  3.92  3.44  18.3     1     0     4     4
#> # … with 22 more rows
decomposed_table <- decompose_table(mtcars_tibble, am_gear_carb_id, am, gear, carb)
decomposed_table
#> $child_table
#> # A tibble: 32 x 9
#>      mpg   cyl  disp    hp  drat    wt  qsec    vs am_gear_carb_id
#>    <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl>           <int>
#>  1  21       6  160    110  3.9   2.62  16.5     0               7
#>  2  21       6  160    110  3.9   2.88  17.0     0               7
#>  3  22.8     4  108     93  3.85  2.32  18.6     1               8
#>  4  21.4     6  258    110  3.08  3.22  19.4     1               1
#>  5  18.7     8  360    175  3.15  3.44  17.0     0               2
#>  6  18.1     6  225    105  2.76  3.46  20.2     1               1
#>  7  14.3     8  360    245  3.21  3.57  15.8     0               3
#>  8  24.4     4  147.    62  3.69  3.19  20       1               4
#>  9  22.8     4  141.    95  3.92  3.15  22.9     1               4
#> 10  19.2     6  168.   123  3.92  3.44  18.3     1               5
#> # … with 22 more rows
#> 
#> $parent_table
#> # A tibble: 13 x 4
#>    am_gear_carb_id    am  gear  carb
#>              <int> <dbl> <dbl> <dbl>
#>  1               7     1     4     4
#>  2               8     1     4     1
#>  3               1     0     3     1
#>  4               2     0     3     2
#>  5               3     0     3     4
#>  6               4     0     4     2
#>  7               5     0     4     4
#>  8               6     0     3     3
#>  9               9     1     4     2
#> 10              10     1     5     2
#> 11              11     1     5     4
#> 12              12     1     5     6
#> 13              13     1     5     8
```

A new column is created, with which the two tables can be joined again,
essentially creating the original table.

The functions that do the inverse operation, i.e. join a parent and a
child table and subsequently drop the `new_id_column`, are
`reunite_parent_child()` and `reunite_parent_child_from_list()`. The
former takes as arguments two tables and the unquoted name of the ID
column, and the latter takes as arguments a list of two tables plus the
unquoted name of the ID column:

``` r
parent_table <- decomposed_table$parent_table
child_table <- decomposed_table$child_table
reunite_parent_child(child_table, parent_table, id_column = am_gear_carb_id)
#> # A tibble: 32 x 11
#>      mpg   cyl  disp    hp  drat    wt  qsec    vs    am  gear  carb
#>    <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl>
#>  1  21       6  160    110  3.9   2.62  16.5     0     1     4     4
#>  2  21       6  160    110  3.9   2.88  17.0     0     1     4     4
#>  3  22.8     4  108     93  3.85  2.32  18.6     1     1     4     1
#>  4  21.4     6  258    110  3.08  3.22  19.4     1     0     3     1
#>  5  18.7     8  360    175  3.15  3.44  17.0     0     0     3     2
#>  6  18.1     6  225    105  2.76  3.46  20.2     1     0     3     1
#>  7  14.3     8  360    245  3.21  3.57  15.8     0     0     3     4
#>  8  24.4     4  147.    62  3.69  3.19  20       1     0     4     2
#>  9  22.8     4  141.    95  3.92  3.15  22.9     1     0     4     2
#> 10  19.2     6  168.   123  3.92  3.44  18.3     1     0     4     4
#> # … with 22 more rows
```

``` r
# Shortcut:
reunite_parent_child_from_list(decomposed_table, id_column = am_gear_carb_id)
```

Currently these functions only exist as a low-level operation on tables.
We plan to extend this operation to `dm` objects in the future.

---
title: "Tips and tricks"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tips and tricks}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fansi::set_knit_hooks(knitr::knit_hooks)
options(crayon.enabled = TRUE, width = 75, cli.width = 75)

knit_print.grViz <- function(x, ...) {
  x %>%
    DiagrammeRsvg::export_svg() %>%
    c("`````{=html}\n", ., "\n`````\n") %>%
    knitr::asis_output()
}
```


FIXME:
This vignette is not yet written, but will be available in a future update.

<!-- Generated by galley: do not edit by hand -->

## Introduction

This tutorial introduces the methods {dm} provides for modifying the
data in the tables of a relational model. There are 6 methods:

-   [`dm_rows_insert()`](#insert) - adds new rows
-   [`dm_rows_update()`](#update) - changes values in rows
-   [`dm_rows_patch()`](#patch) - fills in missing values
-   [`dm_rows_upsert()`](#upsert) - adds new rows or changes values if
    pre-existing
-   [`dm_rows_delete()`](#delete) - deletes rows
-   [`dm_rows_truncate()`](#truncate) - removes all rows, leaving table
    structure intact

## The dm\_rows\_\* process

All six methods take the same arguments and using them follows the same
process:

1.  Create a temporary *changeset dm* that defines the intended changes
    on the RDBMS
2.  If desired, simulate changes with `in_place = FALSE` to double-check
3.  Apply changes with `in_place = TRUE`.

To start, a dm object is created containing the tables, and rows, that
you want to change. This changeset dm is then copied into the same
source as the dm you want to modify. With the dm in the same RDBMS as
the destination dm, you call the appropriate method, such as
`dm_rows_insert()`, to make your planned changes, along with an argument
of `in_place = FALSE` so you can confirm you achieve the changes that
you want.

This verification can be done visually, looking at row counts and the
like, or using {dm}’s constraint checking method,
`dm_examine_constraints()`. The biggest danger is damaging key relations
between data spread across multiple tables by deleting or duplicating
rows and their keys. `dm_examine_constraints()` will catch errors where
primary keys are duplicated or foreign keys do not have a matching
primary key (unless the foreign key value is `NA`).

With the changes confirmed, you execute the method again, this time with
the argument `in_place = TRUE` to make the changes permanent. Note that
`in_place = FALSE` is the default: you must opt in to actually change
data on the database.

Each method has its own requirements in order to maintain database
consistency. These involve constraints on primary key values as they are
how rows are identified.

| Method               | Requirements                                                                                       |
|----------------------|----------------------------------------------------------------------------------------------------|
| `dm_rows_insert()`   | The primary keys must differ from existing records.                                                |
| `dm_rows_update()`   | Primary keys must match for all records to be updated.                                             |
| `dm_rows_patch()`    | Updates missing values in existing records. Primary keys must match for all records to be patched. |
| `dm_rows_upsert()`   | Updates existing records and adds new records, based on the primary key.                           |
| `dm_rows_delete()`   | Removes matching records based on the primary key.                                                 |
| `dm_rows_truncate()` | Removes all records, only for tables in the changeset dm.                                          |

To ensure the integrity of all relations during the process, all methods
automatically determine the correct processing order for the tables
involved. For operations that create records, parent tables are
processed before child tables. For `dm_rows_delete()` and
`dm_rows_truncate()`, child tables are processed before their parent
tables. For more details on this see `vignette("howto-dm-theory")` and
`vignette("howto-dm-db")`.

## Usage

To demonstrate the use of these table modifying methods we will create a
simple dm object with two tables linked by a foreign key. Note the
foreign key of `NA` in the `child` table.

``` r
library(tidyverse)
#> ── Attaching packages ────────────────────────────────── tidyverse 1.3.1 ──
#> ✔ ggplot2 3.3.3     ✔ purrr   0.3.4
#> ✔ tibble  3.1.1     ✔ dplyr   1.0.5
#> ✔ tidyr   1.1.3     ✔ stringr 1.4.0
#> ✔ readr   1.4.0     ✔ forcats 0.5.1
#> ── Conflicts ───────────────────────────────────── tidyverse_conflicts() ──
#> ✖ purrr::%@%()         masks dm::%@%()
#> ✖ purrr::as_function() masks dm::as_function()
#> ✖ tidyr::extract()     masks dm::extract()
#> ✖ dplyr::filter()      masks dm::filter(), stats::filter()
#> ✖ purrr::flatten()     masks dm::flatten()
#> ✖ purrr::flatten_chr() masks dm::flatten_chr()
#> ✖ purrr::flatten_dbl() masks dm::flatten_dbl()
#> ✖ purrr::flatten_int() masks dm::flatten_int()
#> ✖ purrr::flatten_lgl() masks dm::flatten_lgl()
#> ✖ purrr::flatten_raw() masks dm::flatten_raw()
#> ✖ purrr::invoke()      masks dm::invoke()
#> ✖ purrr::is_null()     masks dm::is_null(), testthat::is_null()
#> ✖ dplyr::lag()         masks dm::lag(), stats::lag()
#> ✖ purrr::list_along()  masks dm::list_along()
#> ✖ dplyr::matches()     masks tidyr::matches(), dm::matches(), testthat::matches()
#> ✖ purrr::modify()      masks dm::modify()
#> ✖ purrr::prepend()     masks dm::prepend()
#> ✖ purrr::splice()      masks dm::splice()
library(dm)
parent <- tibble(value = c("A", "B", "C"), pk = 1:3)
parent
#> # A tibble: 3 x 2
#>   value    pk
#>   <chr> <int>
#> 1 A         1
#> 2 B         2
#> 3 C         3
child <- tibble(value = c("a", "b", "c"), pk = 1:3, fk = c(1, 1, NA))
child
#> # A tibble: 3 x 3
#>   value    pk    fk
#>   <chr> <int> <dbl>
#> 1 a         1     1
#> 2 b         2     1
#> 3 c         3    NA
demo_dm <-
  dm(parent = parent, child = child) %>%
  dm_add_pk(parent, pk) %>%
  dm_add_pk(child, pk) %>%
  dm_add_fk(child, fk, parent)

demo_dm %>%
  dm_draw(view_type = "all")
```

![](/home/kirill/git/R/dm/vignettes/out/howto-dm-rows_files/figure-gfm/unnamed-chunk-1-1.png)<!-- -->

{dm} doesn’t check your key values when you create a dm, we add this
check:[1]

``` r
dm_examine_constraints(demo_dm)
#> ℹ All constraints satisfied.
```

Then we copy `demo_dm` into an SQLite database. Note: the default for
the method used, `copy_dm_to()`, is to create temporary tables that will
be automatically deleted when your session ends. As `demo_sql` will be
the destination dm for the examples, the argument `temporary = FALSE` is
used to make this distinction apparent.

``` r
library(DBI)
sqlite_db <- DBI::dbConnect(RSQLite::SQLite())
demo_sql <- copy_dm_to(sqlite_db, demo_dm, temporary = FALSE)
demo_sql
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  sqlite 3.35.5 []
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `parent`, `child`
#> Columns: 5
#> Primary keys: 2
#> Foreign keys: 1
```

{dm}’s table modification methods can be piped together to create a
repeatable sequence of operations that returns a dm incorporating all
the changes required. This is a common use case for {dm} – building by
hand a sequence of operations using temporary results until it is
complete and correct, then committing the result.

## `dm_rows_insert()`

To demonstrate `dm_rows_insert()` we create a dm with tables containing
the rows to insert and copy it to `sqlite_db`, the same source as
`demo_sql`. For all of the `dm_rows_*` methods the source and
destination dm objects must be in the same RDBMS. You will get an error
message if this is not the case.

The code below adds `parent` and `child` table entries for the letter
“D”. First, the changeset dm is created and temporarily copied to the
database:

``` r
new_parent <- tibble(value = "D", pk = 4)
new_parent
#> # A tibble: 1 x 2
#>   value    pk
#>   <chr> <dbl>
#> 1 D         4
new_child <- tibble(value = "d", pk = 4, fk = 4)
new_child
#> # A tibble: 1 x 3
#>   value    pk    fk
#>   <chr> <dbl> <dbl>
#> 1 d         4     4
dm_insert_in <-
  dm(parent = new_parent, child = new_child) %>%
  copy_dm_to(sqlite_db, ., temporary = TRUE)
```

The changeset dm is then used as an argument to `dm_rows_insert()`.

``` r
dm_insert_out <-
  demo_sql %>%
  dm_rows_insert(dm_insert_in)
#> Not persisting, use `in_place = FALSE` to turn off this message.
```

This gives us a warning that changes will not be persisted. Inspecting
the `child` table of the resulting `dm_insert_out` and `demo_sql`, we
can see that’s exactly what happened. {dm} returned to us a dm object
with our inserted rows in place, but the underlying database has not
changed.

``` r
dm_insert_out$child
#> # Source:   lazy query [?? x 3]
#> # Database: sqlite 3.35.5 []
#>   value    pk    fk
#>   <chr> <dbl> <dbl>
#> 1 a         1     1
#> 2 b         2     1
#> 3 c         3    NA
#> 4 d         4     4
demo_sql$child
#> # Source:   table<`child`> [?? x 3]
#> # Database: sqlite 3.35.5 []
#>   value    pk    fk
#>   <chr> <int> <dbl>
#> 1 a         1     1
#> 2 b         2     1
#> 3 c         3    NA
```

We repeat the operation, this time with the argument `in_place = TRUE`
and the changes now persist in `demo_sql`.

``` r
dm_insert_out <-
  demo_sql %>%
  dm_rows_insert(dm_insert_in, in_place = TRUE)

demo_sql$child
#> # Source:   table<`child`> [?? x 3]
#> # Database: sqlite 3.35.5 []
#>   value    pk    fk
#>   <chr> <int> <dbl>
#> 1 a         1     1
#> 2 b         2     1
#> 3 c         3    NA
#> 4 d         4     4
```

## `dm_rows_update()`

`dm_rows_update()` works the same as `dm_rows_insert()`. We create the
dm object and copy it to the same source as the destination. Here we
will change the foreign key for the row in `child` containing “b” to
point to the correct row in `parent`. And we will persist the changes.

``` r
updated_child <- tibble(value = "b", pk = 2, fk = 2)
updated_child
#> # A tibble: 1 x 3
#>   value    pk    fk
#>   <chr> <dbl> <dbl>
#> 1 b         2     2
dm_update_in <-
  dm(child = updated_child) %>%
  copy_dm_to(sqlite_db, ., temporary = TRUE)

dm_update_out <-
  demo_sql %>%
  dm_rows_update(dm_update_in, in_place = TRUE)

demo_sql$child
#> # Source:   table<`child`> [?? x 3]
#> # Database: sqlite 3.35.5 []
#>   value    pk    fk
#>   <chr> <int> <dbl>
#> 1 a         1     1
#> 2 b         2     2
#> 3 c         3    NA
#> 4 d         4     4
```

## `dm_rows_delete()`

`dm_rows_delete()` is not currently implemented to work with an RDBMS,
so we will shift our demonstrations back to the local R environment.
We’ve made changes to `demo_sql` so we use `collect()` to copy the
current tables out of SQLite. Note that persistence is not a concern
with local dm objects. Every operation returns a new dm object
containing the changes made.

``` r
local_dm <- collect(demo_sql)

local_dm$parent
#> # A tibble: 4 x 2
#>   value    pk
#>   <chr> <int>
#> 1 A         1
#> 2 B         2
#> 3 C         3
#> 4 D         4
local_dm$child
#> # A tibble: 4 x 3
#>   value    pk    fk
#>   <chr> <int> <dbl>
#> 1 a         1     1
#> 2 b         2     2
#> 3 c         3    NA
#> 4 d         4     4
dm_deleted <-
  dm(parent = new_parent, child = new_child) %>%
  dm_rows_delete(local_dm, .)
#> Not persisting, use `in_place = FALSE` to turn off this message.
#> Ignoring extra columns: value, fk
#> Ignoring extra columns: value

dm_deleted$child
#> # A tibble: 3 x 3
#>   value    pk    fk
#>   <chr> <int> <dbl>
#> 1 a         1     1
#> 2 b         2     2
#> 3 c         3    NA
```

## `dm_rows_patch()`

`dm_rows_patch()` updates missing values in existing records. We use it
here to fix the missing foreign key in the `child` table.

``` r
patched_child <- tibble(value = "c", pk = 3, fk = 3)
patched_child
#> # A tibble: 1 x 3
#>   value    pk    fk
#>   <chr> <dbl> <dbl>
#> 1 c         3     3
dm_patched <-
  dm(child = patched_child) %>%
  dm_rows_patch(dm_deleted, .)
#> Not persisting, use `in_place = FALSE` to turn off this message.

dm_patched$child
#> # A tibble: 3 x 3
#>   value    pk    fk
#>   <chr> <int> <dbl>
#> 1 a         1     1
#> 2 b         2     2
#> 3 c         3     3
```

## `dm_rows_upsert()`

`dm_rows_upsert()` updates rows with supplied values if they exist or
inserts the supplied values as new rows if they don’t. In this example
we add the letter “D” back to our dm, and update the foreign key for
“b”.

``` r
upserted_parent <- tibble(value = "D", pk = 4)
upserted_parent
#> # A tibble: 1 x 2
#>   value    pk
#>   <chr> <dbl>
#> 1 D         4
upserted_child <- tibble(value = c("b", "d"), pk = c(2, 4), fk = c(3, 4))
upserted_child
#> # A tibble: 2 x 3
#>   value    pk    fk
#>   <chr> <dbl> <dbl>
#> 1 b         2     3
#> 2 d         4     4
dm_upserted <-
  dm(parent = upserted_parent, child = upserted_child) %>%
  dm_rows_upsert(dm_patched, .)
#> Not persisting, use `in_place = FALSE` to turn off this message.

dm_upserted$parent
#> # A tibble: 4 x 2
#>   value    pk
#>   <chr> <dbl>
#> 1 A         1
#> 2 B         2
#> 3 C         3
#> 4 D         4
dm_upserted$child
#> # A tibble: 4 x 3
#>   value    pk    fk
#>   <chr> <dbl> <dbl>
#> 1 a         1     1
#> 2 b         2     3
#> 3 c         3     3
#> 4 d         4     4
```

## `dm_rows_truncate()`

`dm_rows_truncate()` deletes all the rows in a table while leaving all
other related information intact, including column names, column types,
and key relations. The function derives its name from the SQL
`TRUNCATE TABLE` statement, so we will return to our SQLite database to
demonstrate its use. The example below truncates only the `child` table.
Note how a modified version of the destination dm is used as “changeset
dm”: the rows in the changeset dm do not matter here.

``` r
dm_trunc_in <-
  demo_sql %>%
  dm_select_tbl(child)
dm_trunc_in
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  sqlite 3.35.5 []
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `child`
#> Columns: 3
#> Primary keys: 1
#> Foreign keys: 0
dm_trunc_out <-
  demo_sql %>%
  dm_rows_truncate(dm_trunc_in, in_place = TRUE)

demo_sql$child
#> # Source:   table<`child`> [?? x 3]
#> # Database: sqlite 3.35.5 []
#> # … with 3 variables: value <chr>, pk <int>, fk <dbl>
```

## Conclusion

The `dm_rows_*` methods give you row-level granularity over the
modifications you need to make to your relational model. By using the
`in_place` argument they all share you can construct and verify your
modifications before committing them. There are a few limitations, as
mentioned in the tutorial, but these will be addressed in future updates
to {dm}.

## Next Steps

If this tutorial answered some questions, but opened others, these
resources might be of assistance.

Is your data in an RDBMS? `vignette("howto-dm-db")` offers a detailed
look at working with an existing relational data model.

If your data is in data frames, then you may want to read
`vignette("howto-dm-df")` next.

If you would like to know more about relational data models in order
to get the most out of dm, check out `vignette("howto-dm-theory")`.

If you’re familiar with relational data models but want to know how to
work with them in dm, then any of `vignette("tech-dm-join")`,
`vignette("tech-dm-filter")`, or `vignette("tech-dm-zoom")` is a good
next step.

[1] Be aware that when using `dm_examine_constraints()` NULL (`NA`)
foreign keys are allowed and will be counted as a match. In some cases
this doesn’t make sense and non-NULL columns should be enforced by the
RDBMS. Currently {dm} does not specify or check non-NULL constraints for
columns.

<!-- Generated by galley: do not edit by hand -->

In this tutorial we introduce {dm} methods and techniques for copying
individual tables and entire relational data models into an RDBMS. This
is an integral part of the {dm} workflow. Copying tables to an RDBMS is
often a step in the process of building a relational data model from
locally hosted data. If your data model is complete, copying it to an
RDBMS in a single operation allows you to leverage the power of the
database and make it accessible to others. For modifying and persisting
changes to your data at the row-level see `vignette("howto-dm-rows")`.

## Copy models or copy tables?

Using {dm} you can persist an entire relational data model with a single
function call. `copy_dm_to()` will move your entire model into a
destination RDBMS. This may be all you need to deploy a new model. You
may want to add new tables to an existing model on an RDBMS. These
requirements can be handled using the `compute()` and `copy_to()`
methods.

Calling `compute()` or `copy_to()` requires write permission on the
RDBMS, otherwise an error is returned. Therefore for the following
examples we will instantiate a test dm and move it into a local SQLite
database with full permissions. {dm} and {dbplyr} are designed so there
is no difference between the code used to manipulate a local SQLite
database and a remote RDBMS. The steps for this were already introduced
in `vignette("howto-dm-db")` and will be discussed in more detail in the
[Copying a relational model](#copy-model) section.

``` r
library(dm)
library(tidyverse)
#> ── Attaching packages ────────────────────────────────── tidyverse 1.3.1 ──
#> ✔ ggplot2 3.3.3     ✔ purrr   0.3.4
#> ✔ tibble  3.1.1     ✔ dplyr   1.0.5
#> ✔ tidyr   1.1.3     ✔ stringr 1.4.0
#> ✔ readr   1.4.0     ✔ forcats 0.5.1
#> ── Conflicts ───────────────────────────────────── tidyverse_conflicts() ──
#> ✖ purrr::%@%()         masks dm::%@%()
#> ✖ purrr::as_function() masks dm::as_function()
#> ✖ tidyr::extract()     masks dm::extract()
#> ✖ dplyr::filter()      masks dm::filter(), stats::filter()
#> ✖ purrr::flatten()     masks dm::flatten()
#> ✖ purrr::flatten_chr() masks dm::flatten_chr()
#> ✖ purrr::flatten_dbl() masks dm::flatten_dbl()
#> ✖ purrr::flatten_int() masks dm::flatten_int()
#> ✖ purrr::flatten_lgl() masks dm::flatten_lgl()
#> ✖ purrr::flatten_raw() masks dm::flatten_raw()
#> ✖ purrr::invoke()      masks dm::invoke()
#> ✖ purrr::is_null()     masks dm::is_null(), testthat::is_null()
#> ✖ dplyr::lag()         masks dm::lag(), stats::lag()
#> ✖ purrr::list_along()  masks dm::list_along()
#> ✖ dplyr::matches()     masks tidyr::matches(), dm::matches(), testthat::matches()
#> ✖ purrr::modify()      masks dm::modify()
#> ✖ purrr::prepend()     masks dm::prepend()
#> ✖ purrr::splice()      masks dm::splice()
library(dbplyr)
#> 
#> Attaching package: 'dbplyr'
#> The following objects are masked from 'package:dplyr':
#> 
#>     ident, sql
#> The following objects are masked from 'package:dm':
#> 
#>     ident, sql

fin_dm <-
  dm_financial() %>%
  dm_select_tbl(-trans) %>%
  collect()

local_db <- DBI::dbConnect(RSQLite::SQLite())
deployed_dm <- copy_dm_to(local_db, fin_dm, temporary = FALSE)
```

## Copying and persisting individual tables

As part of your data analysis you may combine tables from multiple
sources and create links to existing tables via foreign keys, or create
new tables holding data summaries. The example below, already discussed
in `vignette("howto-dm-db")`, computes the total amount of all loans for
each account.

``` r
my_dm_total <-
  deployed_dm %>%
  dm_zoom_to(loans) %>%
  group_by(account_id) %>%
  summarize(total_amount = sum(amount, na.rm = TRUE)) %>%
  ungroup() %>%
  dm_insert_zoomed("total_loans")
```

The derived table `total_loans` is a *lazy table* powered by the
{[dbplyr](https://dbplyr.tidyverse.org/)} package: the results are not
materialized, instead an SQL query is built and executed each time the
data is requested.

``` r
my_dm_total$total_loans %>%
  sql_render()
#> <SQL> SELECT `account_id`, SUM(`amount`) AS `total_amount`
#> FROM `loans`
#> GROUP BY `account_id`
```

To avoid recomputing the query every time you use `total_loans`, call
`compute()` right before inserting the derived table with
`dm_insert_tbl()`. `compute()` forces the computation of a query and
stores the full results in a table on the RDBMS.

``` r
my_dm_total_computed <-
  deployed_dm %>%
  dm_zoom_to(loans) %>%
  group_by(account_id) %>%
  summarize(total_amount = sum(amount, na.rm = TRUE)) %>%
  ungroup() %>%
  compute() %>%
  dm_insert_zoomed("total_loans")

my_dm_total_computed$total_loans %>%
  sql_render()
#> <SQL> SELECT *
#> FROM `dbplyr_001`
```

Note the differences in queries returned by `sql_render()`.
`my_dm_total$total_loans` is still being lazily evaluated and the full
query constructed from the chain of operations that generated it, and is
required to run to access it, is still in place. Contrast that with
`my_dm_total_computed$total_loans` where the query has been realized and
accessing its rows requires a simple `SELECT *` statement. The table
name, ``, was automatically generated as the `name` argument was not
supplied.

The default is to create a temporary table. If you want results to
persist across sessions in permanent tables, `compute()` must be called
with the argument `temporary = FALSE` and a table name for the `name`
argument. See `?compute` for more details.

When called on a whole dm object (without zoom), `compute()`
materializes all tables into new (temporary or persistent) tables by
executing the associated SQL query and storing the full results.
Depending on the size of your data this may take considerable time or be
infeasible. It may be useful occasionally to create snapshots of data
that is subject to change.

``` r
my_dm_total_snapshot <-
  my_dm_total %>%
  compute()
```

## Adding local data frames to an RDBMS

If you need to add local data frames to an existing dm object, use the
`copy_to()` method. It takes the same arguments as `copy_dm_to()`,
except the second argument takes a data frame rather than a dm. The
result is a derived dm object that contains the new table.

To demonstrate the use of `copy_to()` the example below will use {dm} to
pull consolidated data from several tables out of an RDBMS, estimate a
linear model from the data, then insert the residuals back into the
RDBMS and link it to the existing tables. This is all done with a local
SQLite database, but the process would work unchanged on any supported
RDBMS.

``` r
loans_df <-
  deployed_dm %>%
  dm_squash_to_tbl(loans) %>%
  select(id, amount, duration, A3) %>%
  collect()
#> Renamed columns:
#> * date -> loans.date, accounts.date
```

Please note the use of `dm_squash_to_tbl()`. This method gathers all
linked information into a single wide table. It follows foreign key
relations starting from the table supplied as its argument and gathers
all the columns from related tables, disambiguating column names as it
goes.

In the above code, the `select()` statement isolates the columns we need
for our model. `collect()` works similarly to `compute()` by forcing the
execution of the underlying SQL query, but it returns the results as a
local tibble.

Below, the local tibble, `loans_df`, is used to estimate the linear
model and the residuals are stored along with the original associated
`id` in a new tibble, `loans_residuals`. The `id` column is necessary to
link the new tibble to the tables in the dm it was collected from.

``` r
model <- lm(amount ~ duration + A3, data = loans_df)

loans_residuals <- tibble::tibble(
  id = loans_df$id,
  resid = unname(residuals(model))
)

loans_residuals
#> # A tibble: 682 x 2
#>       id   resid
#>    <int>   <dbl>
#>  1  4959 -31912.
#>  2  4961 -27336.
#>  3  4962 -30699.
#>  4  4967  63621.
#>  5  4968 -94811.
#>  6  4973  59036.
#>  7  4986  41901.
#>  8  4988 123392.
#>  9  4989 147157.
#> 10  4990  33377.
#> # … with 672 more rows
```

Adding `loans_residuals` to the dm is done using `copy_to()`. The call
to the method includes the argument `temporary = FALSE` because we want
this table to persist beyond our current session. In the same pipeline
we create the necessary primary and foreign keys to integrate the table
with the rest of our relational model. For more information on key
creation see `vignette("howto-dm-db")` and
`vignette("howto-dm-theory")`.

``` r
my_dm_sqlite_resid <-
  copy_to(deployed_dm, loans_residuals, temporary = FALSE) %>%
  dm_add_pk(loans_residuals, id) %>%
  dm_add_fk(loans_residuals, id, loans)

my_dm_sqlite_resid %>%
  dm_set_colors(violet = loans_residuals) %>%
  dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/howto-dm-copy_files/figure-gfm/unnamed-chunk-9-1.png)<!-- -->

``` r
my_dm_sqlite_resid %>%
  dm_examine_constraints()
#> ℹ All constraints satisfied.
my_dm_sqlite_resid$loans_residuals
#> # Source:   table<loans_residuals_2020_08_28_07_13_03_12345_1> [?? x 2]
#> # Database: sqlite 3.35.5 []
#>       id   resid
#>    <int>   <dbl>
#>  1  4959 -31912.
#>  2  4961 -27336.
#>  3  4962 -30699.
#>  4  4967  63621.
#>  5  4968 -94811.
#>  6  4973  59036.
#>  7  4986  41901.
#>  8  4988 123392.
#>  9  4989 147157.
#> 10  4990  33377.
#> # … with more rows
```

## Persisting a relational model with `copy_dm_to()`

Persistence, because it is intended to make permanent changes, requires
write access to the source RDBMS. The code below is a repeat of the code
that opened the [Copying and persisting individual
tables](#copying-tables) section at the beginning of the tutorial. It
uses the {dm} convenience function `dm_financial()` to create a dm
object corresponding to a data model from a public dataset repository.
The dm object is downloaded locally first, before deploying it to a
local SQLite database.

`dm_select_tbl()` is used to exclude the transaction table `trans` due
to its size, then the `collect()` method retrieves the remaining tables
and returns them as a local dm object.

``` r
dm_financial() %>%
  dm_nrow()
#>  accounts     cards   clients     disps districts     loans    orders 
#>      4500       892      5369      5369        77       682      6471 
#>     tkeys     trans 
#>       234   1056320
fin_dm <-
  dm_financial() %>%
  dm_select_tbl(-trans) %>%
  collect()

fin_dm
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `accounts`, `cards`, `clients`, `disps`, `districts`, … (8 total)
#> Columns: 47
#> Primary keys: 7
#> Foreign keys: 6
```

It is just as simple to move a local relational model into an RDBMS.

``` r
destination_db <- DBI::dbConnect(RSQLite::SQLite())

deployed_dm <-
  copy_dm_to(destination_db, fin_dm, temporary = FALSE)

deployed_dm
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  sqlite 3.35.5 []
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `accounts`, `cards`, `clients`, `disps`, `districts`, … (8 total)
#> Columns: 47
#> Primary keys: 7
#> Foreign keys: 6
```

Note that in the call to `copy_dm_to()` the argument `temporary = FALSE`
is supplied. Without this argument the model would still be copied into
the database, but the argument would default to `temporary = TRUE` and
the data would be deleted once your session ends.

In the output you can observe that the `src` for `deployed_dm` is
SQLite, while for `fin_dm` the source is not indicated because it is a
local data model.

Copying a relational model into an empty database is the simplest use
case for `copy_dm_to()`. If you want to copy a model into an RDBMS that
is already populated, be aware that `copy_dm_to()` will not overwrite
pre-existing tables. In this case you will need to use the `table_names`
argument to give the tables unique names.

`table_names` can be a named character vector, with the names matching
the table names in the dm object and the values containing the desired
names in the RDBMS, or a function or one-sided formula. In the example
below, `paste0()` is used to add a prefix to the table names to provide
uniqueness.

``` r
dup_dm <-
  copy_dm_to(destination_db, fin_dm, temporary = FALSE, table_names = ~ paste0("dup_", .x))

dup_dm
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  sqlite 3.35.5 []
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `accounts`, `cards`, `clients`, `disps`, `districts`, … (8 total)
#> Columns: 47
#> Primary keys: 7
#> Foreign keys: 6
remote_name(dup_dm$accounts)
#> <IDENT> `dup_accounts`
remote_name(deployed_dm$accounts)
#> <IDENT> `accounts`
```

Note the different table names for `dup_dm$accounts` and
`deployed_dm$accounts`. For both, the table name is `accounts` in the
dm, but they link to different tables on the database. In `dup_dm` the
table is backed by the table `dup_accounts` in the RDBMS.
`dm_deployed$accounts` shows us that this table is still backed by the
`accounts` table from the `copy_dm_to()` operation we performed in the
preceding example.

Managing tables in the RDBMS is outside the scope of `dm`. If you find
you need to remove tables or perform operations directly on the RDBMS,
see the {[DBI](https://dbi.r-dbi.org/)} package.

## Conclusion

`dm` makes it straightforward to deploy your complete relational model
to an RDBMS using the `copy_dm_to()` function. For tables that are
created from a relational model during analysis or development,
`compute()` and `copy_to()` can be used to persist them between sessions
or to copy local tables to a database dm. The `collect()` method
downloads an entire dm object that fits into memory from the database.

## Next steps

If you need finer-grained control over modifications to your relational
model, see `vignette("howto-dm-rows")` for an introduction to row level
operations, including updates, insertions, deletions and patching.

If you would like to know more about relational data models in order
to get the most out of dm, check out `vignette("howto-dm-theory")`.

If you’re familiar with relational data models but want to know how to
work with them in dm, then any of `vignette("tech-dm-join")`,
`vignette("tech-dm-filter")`, or `vignette("tech-dm-zoom")` is a good
next step.

<!-- Generated by galley: do not edit by hand -->

This vignette deals with situations where you want to transform tables
of your `dm` and then update an existing table or add a new table to the
`dm`. There are two straightforward approaches:

1.  individually extract the tables relevant to the calculation, perform
    the necessary transformations, add the result to the `dm` (or
    replace an existing table) and establish the key relations.
2.  do all this within the `dm` object by zooming to a table and
    manipulating it while maintaining the key relations whenever
    possible.

The first approach is rather self-explanatory, so let us have a closer
look at the second way.

## Enabling {dplyr}-workflow within a `dm`

“Zooming” to a table of a `dm` means:

-   all information stored in the original `dm` is kept, including the
    originally zoomed table
-   an object of class `dm_zoomed` is produced, presenting a view of the
    table for transformations
-   you do not need to specify the table when calling `select()`,
    `mutate()` and other table manipulation functions

{dm} provides methods for many of the {dplyr}-verbs for a `dm_zoomed`
which behave the way you are used to, affecting only the zoomed table
and leaving the rest of the `dm` untouched. When you are finished with
transforming the table, there are three options to proceed:

1.  use `dm_update_zoomed()` if you want to replace the originally
    zoomed table with the new table
2.  use `dm_insert_zoomed()` if you are creating a new table for your
    `dm`
3.  use `dm_discard_zoomed()` if you do not need the result and want to
    discard it

When employing one of the first two options, the resulting table in the
`dm` will have all the primary and foreign keys available that could be
tracked from the originally zoomed table.

## Examples

So much for the theory, but how does it look and feel? To explore this,
we once more make use of our trusted {nycflights13} data.

### Use case 1: Add a new column to an existing table

Imagine you want to have a column in `flights`, specifying if a flight
left before noon or after. Just like with {dplyr}, we can tackle this
with `mutate()`. Let us do this step by step:

``` r
library(dm)
library(dplyr)
#> 
#> Attaching package: 'dplyr'
#> The following object is masked from 'package:testthat':
#> 
#>     matches
#> The following objects are masked from 'package:stats':
#> 
#>     filter, lag
#> The following objects are masked from 'package:base':
#> 
#>     intersect, setdiff, setequal, union
flights_dm <- dm_nycflights13()
flights_dm
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 4
#> Foreign keys: 4
flights_zoomed <-
  flights_dm %>%
  dm_zoom_to(flights)
# The print output for a `dm_zoomed` looks very much like that from a normal `tibble`.
flights_zoomed
#> # Zoomed table: flights
#> # A tibble:     11,227 x 19
#>     year month   day dep_time sched_dep_time dep_delay arr_time
#>    <int> <int> <int>    <int>          <int>     <dbl>    <int>
#>  1  2013     1    10        3           2359         4      426
#>  2  2013     1    10       16           2359        17      447
#>  3  2013     1    10      450            500       -10      634
#>  4  2013     1    10      520            525        -5      813
#>  5  2013     1    10      530            530         0      824
#>  6  2013     1    10      531            540        -9      832
#>  7  2013     1    10      535            540        -5     1015
#>  8  2013     1    10      546            600       -14      645
#>  9  2013     1    10      549            600       -11      652
#> 10  2013     1    10      550            600       -10      649
#> # … with 11,217 more rows, and 12 more variables: sched_arr_time <int>,
#> #   arr_delay <dbl>, carrier <chr>, flight <int>, tailnum <chr>,
#> #   origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>, hour <dbl>,
#> #   minute <dbl>, time_hour <dttm>

flights_zoomed_mutate <-
  flights_zoomed %>%
  mutate(am_pm_dep = if_else(dep_time < 1200, "am", "pm")) %>%
  # in order to see our changes in the output we use `select()` for reordering the columns
  select(year:dep_time, am_pm_dep, everything())
flights_zoomed_mutate
#> # Zoomed table: flights
#> # A tibble:     11,227 x 20
#>     year month   day dep_time am_pm_dep sched_dep_time dep_delay arr_time
#>    <int> <int> <int>    <int> <chr>              <int>     <dbl>    <int>
#>  1  2013     1    10        3 am                  2359         4      426
#>  2  2013     1    10       16 am                  2359        17      447
#>  3  2013     1    10      450 am                   500       -10      634
#>  4  2013     1    10      520 am                   525        -5      813
#>  5  2013     1    10      530 am                   530         0      824
#>  6  2013     1    10      531 am                   540        -9      832
#>  7  2013     1    10      535 am                   540        -5     1015
#>  8  2013     1    10      546 am                   600       -14      645
#>  9  2013     1    10      549 am                   600       -11      652
#> 10  2013     1    10      550 am                   600       -10      649
#> # … with 11,217 more rows, and 12 more variables: sched_arr_time <int>,
#> #   arr_delay <dbl>, carrier <chr>, flight <int>, tailnum <chr>,
#> #   origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>, hour <dbl>,
#> #   minute <dbl>, time_hour <dttm>

# To update the original `dm` with a new `flights` table we use `dm_update_zoomed()`:
updated_flights_dm <-
  flights_zoomed_mutate %>%
  dm_update_zoomed()
# The only difference in the `dm` print output is the increased number of columns
updated_flights_dm
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 54
#> Primary keys: 4
#> Foreign keys: 4
# The schematic view of the data model remains unchanged
dm_draw(updated_flights_dm)
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-zoom_files/figure-gfm/zoom-1.png)<!-- -->

### Use case 2: Creation of a surrogate key

The same course of action could, for example, be employed to create a
surrogate key for a table. We can do this for the `weather` table.

``` r
library(tidyr)
#> 
#> Attaching package: 'tidyr'
#> The following object is masked from 'package:dm':
#> 
#>     extract
#> The following object is masked from 'package:testthat':
#> 
#>     matches
weather_zoomed <-
  flights_dm %>%
  dm_zoom_to(weather)
weather_zoomed
#> # Zoomed table: weather
#> # A tibble:     861 x 15
#>    origin  year month   day  hour  temp  dewp humid wind_dir wind_speed
#>    <chr>  <int> <int> <int> <int> <dbl> <dbl> <dbl>    <dbl>      <dbl>
#>  1 EWR     2013     1    10     0  41    32    70.1      230       8.06
#>  2 EWR     2013     1    10     1  39.0  30.0  69.9      210       9.21
#>  3 EWR     2013     1    10     2  39.0  28.9  66.8      230       6.90
#>  4 EWR     2013     1    10     3  39.9  27.0  59.5      270       5.75
#>  5 EWR     2013     1    10     4  41    26.1  55.0      320       6.90
#>  6 EWR     2013     1    10     5  41    26.1  55.0      300      12.7 
#>  7 EWR     2013     1    10     6  39.9  25.0  54.8      280       6.90
#>  8 EWR     2013     1    10     7  41    25.0  52.6      330       6.90
#>  9 EWR     2013     1    10     8  43.0  25.0  48.7      330       8.06
#> 10 EWR     2013     1    10     9  45.0  23    41.6      320      17.3 
#> # … with 851 more rows, and 5 more variables: wind_gust <dbl>,
#> #   precip <dbl>, pressure <dbl>, visib <dbl>, time_hour <dttm>
# Maybe there is some hidden candidate for a primary key that we overlooked
enum_pk_candidates(weather_zoomed)
#> # A tibble: 15 x 3
#>    columns    candidate why                                                
#>    <keys>     <lgl>     <chr>                                              
#>  1 origin     FALSE     has duplicate values: EWR (287), JFK (287), LGA (2…
#>  2 year       FALSE     has duplicate values: 2013 (861)                   
#>  3 month      FALSE     has duplicate values: 1 (72), 2 (72), 4 (72), 5 (7…
#>  4 day        FALSE     has duplicate values: 10 (861)                     
#>  5 hour       FALSE     has duplicate values: 0 (36), 1 (36), 3 (36), 4 (3…
#>  6 temp       FALSE     has duplicate values: 62.06 (24), 75.92 (23), 33.0…
#>  7 dewp       FALSE     has duplicate values: 32.00 (53), 53.96 (47), 53.0…
#>  8 humid      FALSE     has duplicate values: 72.33 (8), 89.86 (8), 70.08 …
#>  9 wind_dir   FALSE     has duplicate values: 180 (39), 200 (39), 320 (39)…
#> 10 wind_speed FALSE     has duplicate values: 6.90468 (82), 8.05546 (79), …
#> 11 wind_gust  FALSE     has 718 missing values, and duplicate values: 23.0…
#> 12 precip     FALSE     has duplicate values: 0.00 (791), 0.01 (16), 0.02 …
#> 13 pressure   FALSE     has 92 missing values, and duplicate values: 1015.…
#> 14 visib      FALSE     has duplicate values: 10 (698), 9 (45), 8 (38), 7 …
#> 15 time_hour  FALSE     has duplicate values: 2013-01-10 00:00:00 (3), 201…
# Seems we have to construct a column with unique values
# This can be done by combining column `origin` with `time_hour`, if the latter
# is converted to a single time zone first; all within the `dm`:
weather_zoomed_mutate <-
  weather_zoomed %>%
  # first convert all times to the same time zone:
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  # paste together as character the airport code and the time
  unite("origin_slot_id", origin, time_hour_fmt) %>%
  select(origin_slot_id, everything())
# check if we the result is as expected:
enum_pk_candidates(weather_zoomed_mutate) %>% filter(candidate)
#> # A tibble: 1 x 3
#>   columns        candidate why  
#>   <keys>         <lgl>     <chr>
#> 1 origin_slot_id TRUE      ""
flights_upd_weather_dm <-
  weather_zoomed_mutate %>%
  dm_update_zoomed() %>%
  dm_add_pk(weather, origin_slot_id)
flights_upd_weather_dm
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 4
#> Foreign keys: 3
# creating the coveted FK relation between `flights` and `weather`
extended_flights_dm <-
  flights_upd_weather_dm %>%
  dm_zoom_to(flights) %>%
  mutate(time_hour_fmt = format(time_hour, tz = "UTC")) %>%
  # need to keep `origin` as FK to airports, so `remove = FALSE`
  unite("origin_slot_id", origin, time_hour_fmt, remove = FALSE) %>%
  dm_update_zoomed() %>%
  dm_add_fk(flights, origin_slot_id, weather)
extended_flights_dm %>% dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-zoom_files/figure-gfm/unnamed-chunk-1-1.png)<!-- -->

### Use case 3: Disentangle `dm`

If you look at the `dm` created by `dm_nycflights13(cycle = TRUE)`, you
see that two columns of `flights` relate to one and the same table,
`airports`. One column stands for the departure airport and the other
for the arrival airport.

``` r
dm_draw(dm_nycflights13(cycle = TRUE))
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-zoom_files/figure-gfm/unnamed-chunk-2-1.png)<!-- -->
In such cases it can be beneficial, to “disentangle” the `dm` by
duplicating the referred table. One way to do this in the {dm}-framework
is as follows:

``` r
disentangled_flights_dm <-
  dm_nycflights13(cycle = TRUE) %>%
  # zooming and immediately inserting essentially creates a copy of the original table
  dm_zoom_to(airports) %>%
  # reinserting the `airports` table under the name `destination`
  dm_insert_zoomed("destination") %>%
  # renaming the originally zoomed table
  dm_rename_tbl(origin = airports) %>%
  # Key relations are also duplicated, so the wrong ones need to be removed
  dm_rm_fk(flights, dest, origin) %>%
  dm_rm_fk(flights, origin, destination)
dm_draw(disentangled_flights_dm)
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-zoom_files/figure-gfm/unnamed-chunk-3-1.png)<!-- -->

In a future update we will provide a more convenient way to
“disentangle” `dm` objects, so that the individual steps will be done
automatically.

### Use case 4: Add summary table to `dm`

Here is an example for adding a summary of a table as a new table to a
`dm` (FK-relations are taken care of automatically):

``` r
dm_with_summary <-
  flights_dm %>%
  dm_zoom_to(flights) %>%
  count(origin, carrier) %>%
  dm_insert_zoomed("dep_carrier_count")
dm_draw(dm_with_summary)
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-zoom_files/figure-gfm/unnamed-chunk-4-1.png)<!-- -->

### Use case 5: Joining tables

If you would like to join some or all of the columns of one table to
another, you can make use of one of the `join`-methods for a
`dm_zoomed`. In addition to the usual arguments for the {dplyr}-joins,
by supplying the `select` argument you can specify which columns of the
RHS-table you want to be included in the join. For the syntax please see
the example below. The LHS-table of a join is always the zoomed table.

``` r
joined_flights_dm <-
  flights_dm %>%
  dm_zoom_to(flights) %>%
  # let's first reduce the number of columns of flights
  select(-dep_delay:-arr_delay, -air_time:-time_hour) %>%
  # in the {dm}-method for the joins you can specify which columns you want to add to the zoomed table
  left_join(planes, select = c(tailnum, plane_type = type)) %>%
  dm_insert_zoomed("flights_plane_type")
# this is how the table looks now
joined_flights_dm$flights_plane_type
#> # A tibble: 11,227 x 11
#>     year month   day dep_time sched_dep_time carrier flight tailnum origin
#>    <int> <int> <int>    <int>          <int> <chr>    <int> <chr>   <chr> 
#>  1  2013     1    10        3           2359 B6         727 N571JB  JFK   
#>  2  2013     1    10       16           2359 B6         739 N564JB  JFK   
#>  3  2013     1    10      450            500 US        1117 N171US  EWR   
#>  4  2013     1    10      520            525 UA        1018 N35204  EWR   
#>  5  2013     1    10      530            530 UA         404 N815UA  LGA   
#>  6  2013     1    10      531            540 AA        1141 N5EAAA  JFK   
#>  7  2013     1    10      535            540 B6         725 N784JB  JFK   
#>  8  2013     1    10      546            600 B6         380 N337JB  EWR   
#>  9  2013     1    10      549            600 EV        6055 N19554  LGA   
#> 10  2013     1    10      550            600 US        2114 N740UW  LGA   
#> # … with 11,217 more rows, and 2 more variables: dest <chr>,
#> #   plane_type <chr>
# also here, the FK-relations are transferred to the new table
dm_draw(joined_flights_dm)
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-zoom_files/figure-gfm/unnamed-chunk-5-1.png)<!-- -->

### Tip: Accessing the zoomed table

At each point you can retrieve the zoomed table by calling `pull_tbl()`
on a `dm_zoomed`. To use our last example once more:

``` r
flights_dm %>%
  dm_zoom_to(flights) %>%
  select(-dep_delay:-arr_delay, -air_time:-time_hour) %>%
  left_join(planes, select = c(tailnum, plane_type = type)) %>%
  pull_tbl()
#> # A tibble: 11,227 x 11
#>     year month   day dep_time sched_dep_time carrier flight tailnum origin
#>    <int> <int> <int>    <int>          <int> <chr>    <int> <chr>   <chr> 
#>  1  2013     1    10        3           2359 B6         727 N571JB  JFK   
#>  2  2013     1    10       16           2359 B6         739 N564JB  JFK   
#>  3  2013     1    10      450            500 US        1117 N171US  EWR   
#>  4  2013     1    10      520            525 UA        1018 N35204  EWR   
#>  5  2013     1    10      530            530 UA         404 N815UA  LGA   
#>  6  2013     1    10      531            540 AA        1141 N5EAAA  JFK   
#>  7  2013     1    10      535            540 B6         725 N784JB  JFK   
#>  8  2013     1    10      546            600 B6         380 N337JB  EWR   
#>  9  2013     1    10      549            600 EV        6055 N19554  LGA   
#> 10  2013     1    10      550            600 US        2114 N740UW  LGA   
#> # … with 11,217 more rows, and 2 more variables: dest <chr>,
#> #   plane_type <chr>
```

### Possible pitfalls and caveats

1.  Currently not all of the {dplyr}-verbs have their own method for a
    `dm_zoomed`, so be aware that in some cases it will still be
    necessary to resort to extracting one or more tables from a `dm` and
    reinserting a transformed version of theirs into the `dm`
    eventually. The supported functions are: `group_by()`, `ungroup()`,
    `summarise()`, `mutate()`, `transmute()`, `filter()`, `select()`,
    `rename()`, `distinct()`, `arrange()`, `slice()`, `left_join()`,
    `inner_join()`, `full_join()`, `right_join()`, `semi_join()` and
    `anti_join()`.

2.  The same is true for {tidyr}-functions. Methods are provided for:
    `unite()` and `separate()`.

3.  There might be situations when you would like the key relations to
    remain intact, but they are dropped nevertheless. This is because a
    rigid logic is implemented, that does drop a key when its associated
    column is acted upon with e.g. a `mutate()` call. In these cases the
    key relations will need to be established once more after finishing
    with the manipulations.

Complex datasets are often organized as several tables, interlinked through common values in columns.
Working with these datasets requires an understanding of the relationships
between tables. Occasionally, reorganization such as normalization, decomposition, or aggregation may be necessary, in particular if the data comes as a single denormalized table.
Defining a rigorous data model helps, but we found surprisingly few tools in the R ecosystem
that deal with this problem. One example is the datamodelr package on GitHub, which
focuses on describing data models and drawing diagrams.

The dm package offers tools to prepare, specify and validate a data model (e.g., decomposing
tables, defining and checking keys and cardinality). It focuses on both data and metadata,
and can work with any dplyr-compatible data source, in particular in-memory data frames
and SQL databases are supported. With a validated data model, join operations are well-
defined and can be generated automatically. An interface to datamodelr for drawing the
data model diagram is provided. A Shiny module for browsing and filtering the tables of the
data model or a submodel, respecting the relationships between the tables, makes it easy to
build custom apps for interactive work with data spread across multiple tables.

<!-- Generated by galley: do not edit by hand -->

Multiple, linked tables are common within computer science. Because many
R users have backgrounds in other disciplines, we present **six
important terms in relational data modeling** to help you to jump-start
working with {dm}. These terms are:

1.  [Data Frames and Tables](#tables)
2.  [Data Model](#model)
3.  [Primary Keys](#pk)
4.  [Foreign Keys](#fk)
5.  [Referential Integrity](#referential-integrity)
6.  [Normalization](#normalization)
7.  [Relational Databases](#relational-databases)

## 1. Data Frames and Tables

A data frame is a fundamental data structure in R. Columns represent
variables, rows represent observations. In more technical terms: a data
frame is a list of variables of identical length and unique row names.
If you imagine it visually, the result is a typical table structure.
That is why working with data from spreadsheets is so convenient and the
users of the popular [{dplyr}](https://dplyr.tidyverse.org) package for
data wrangling mainly rely on data frames.

The downside is that data frames and flat file systems like spreadsheets
can result in bloated tables because they hold many repetitive values.
In the worst case, a data frame can contain multiple columns with only a
single value different in each row.

This calls for better data organization by utilizing the resemblance
between data frames and database tables, which also consist of columns
and rows. The elements are just named differently:

| Data Frame | Table     |
|------------|-----------|
| Column     | Attribute |
| Row        | Tuple     |

Relational databases, unlike data frames, do not keep all data in one
large table but instead split it into multiple smaller tables. That
separation into sub-tables has several advantages:

-   all information is stored only once, avoiding redundancy and
    conserving memory
-   all information needs to be updated only once and in one place,
    improving consistency and avoiding errors that may result from
    updating the same value in multiple locations
-   all information is organized by topic and segmented into smaller
    tables that are easier to handle

It is for these reasons that separation of data helps with data quality,
and they explain the popularity of relational databases in
production-level data management.

The downside of this approach is that it is harder to merge together
information from different data sources and to identify which entities
refer to the same object, a common task when modeling or plotting data.

Thus, to take full advantage of the relational database approach, an
associated data model is needed to overcome the challenges that arise
when working with multiple tables.

Let’s illustrate this challenge with the data from the [`nycflights13`
dataset](https://github.com/tidyverse/nycflights13) that contains
detailed information about the 336776 flights that departed from New
York City in 2013. The information is stored in five tables.

Details like the full name of an airport are not available immediately;
these can only be obtained by joining or merging the constituent tables,
which can result in long and inflated pipe chains full of `left_join()`,
`anti_join()` and other forms of data merging.

In classical {dplyr} notation, you will need three `left_join()` calls
to merge the `flights` table gradually to `airlines`, `planes`,
`airports` and `weather` tables to create one wide data frame.

``` r
library(tidyverse)
#> ── Attaching packages ────────────────────────────────── tidyverse 1.3.1 ──
#> ✔ ggplot2 3.3.3     ✔ purrr   0.3.4
#> ✔ tibble  3.1.1     ✔ dplyr   1.0.5
#> ✔ tidyr   1.1.3     ✔ stringr 1.4.0
#> ✔ readr   1.4.0     ✔ forcats 0.5.1
#> ── Conflicts ───────────────────────────────────── tidyverse_conflicts() ──
#> ✖ purrr::%@%()         masks dm::%@%()
#> ✖ purrr::as_function() masks dm::as_function()
#> ✖ tidyr::extract()     masks dm::extract()
#> ✖ dplyr::filter()      masks dm::filter(), stats::filter()
#> ✖ purrr::flatten()     masks dm::flatten()
#> ✖ purrr::flatten_chr() masks dm::flatten_chr()
#> ✖ purrr::flatten_dbl() masks dm::flatten_dbl()
#> ✖ purrr::flatten_int() masks dm::flatten_int()
#> ✖ purrr::flatten_lgl() masks dm::flatten_lgl()
#> ✖ purrr::flatten_raw() masks dm::flatten_raw()
#> ✖ purrr::invoke()      masks dm::invoke()
#> ✖ purrr::is_null()     masks dm::is_null(), testthat::is_null()
#> ✖ dplyr::lag()         masks dm::lag(), stats::lag()
#> ✖ purrr::list_along()  masks dm::list_along()
#> ✖ dplyr::matches()     masks tidyr::matches(), dm::matches(), testthat::matches()
#> ✖ purrr::modify()      masks dm::modify()
#> ✖ purrr::prepend()     masks dm::prepend()
#> ✖ purrr::splice()      masks dm::splice()
library(dm)
library(nycflights13)

flights %>%
  left_join(airlines, by = "carrier") %>%
  left_join(planes, by = "tailnum") %>%
  left_join(airports, by = c("origin" = "faa")) %>%
  left_join(weather, by = c("origin", "time_hour"))
#> # A tibble: 336,776 x 48
#>    year.x month.x day.x dep_time sched_dep_time dep_delay arr_time
#>     <int>   <int> <int>    <int>          <int>     <dbl>    <int>
#>  1   2013       1     1      517            515         2      830
#>  2   2013       1     1      533            529         4      850
#>  3   2013       1     1      542            540         2      923
#>  4   2013       1     1      544            545        -1     1004
#>  5   2013       1     1      554            600        -6      812
#>  6   2013       1     1      554            558        -4      740
#>  7   2013       1     1      555            600        -5      913
#>  8   2013       1     1      557            600        -3      709
#>  9   2013       1     1      557            600        -3      838
#> 10   2013       1     1      558            600        -2      753
#> # … with 336,766 more rows, and 41 more variables: sched_arr_time <int>,
#> #   arr_delay <dbl>, carrier <chr>, flight <int>, tailnum <chr>,
#> #   origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>,
#> #   hour.x <dbl>, minute <dbl>, time_hour <dttm>, name.x <chr>,
#> #   year.y <int>, type <chr>, manufacturer <chr>, model <chr>,
#> #   engines <int>, seats <int>, speed <int>, engine <chr>, name.y <chr>,
#> #   lat <dbl>, lon <dbl>, alt <dbl>, tz <dbl>, dst <chr>, tzone <chr>,
#> #   year <int>, month.y <int>, day.y <int>, hour.y <int>, temp <dbl>,
#> #   dewp <dbl>, humid <dbl>, wind_dir <dbl>, wind_speed <dbl>,
#> #   wind_gust <dbl>, precip <dbl>, pressure <dbl>, visib <dbl>
```

{dm} offers a more elegant and shorter way to combine tables while
augmenting {dplyr}/{dbplyr} workflows.

It is possible to have the best of both worlds: manage your data with
{dm} as linked tables, then flatten multiple tables into one for your
analysis with {dplyr} on an as-needed basis.

The next step is to create a [data model](#model) based on multiple
tables:

## 2. Data Model

A data model shows the structure between multiple tables that can be
linked together.

The `nycflights13` relations can be transferred into the following
graphical representation:

``` r
dm <- dm_nycflights13(cycle = TRUE)

dm %>%
  dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/howto-dm-theory_files/figure-gfm/unnamed-chunk-2-1.png)<!-- -->

The `flights` table is linked to three other tables: `airlines`,
`planes` and `airports`. By using directed arrows, the visualization
shows explicitly the connection between different columns (they are
called attributes in the relational data sphere).

For example: The column `carrier` in `flights` can be joined with the
column `carrier` from the `airlines` table.

The links between the tables are established through [primary keys](#pk)
and [foreign keys](#fk).

Further Reading: The {dm} methods for [visualizing data
models](https://cynkra.github.io/dm/articles/tech-dm-draw.html).

## 3. Primary Keys

In a relational data model, each table should have **one or several
columns that uniquely identifies a row**. These columns define the
*primary key* (abbreviated with “pk”). If the key consists of a single
column, it is called *simple key*. A key consisting of more than one
column is called a *compound key*.

Example: In the `airlines` table of `nycflights13` the column `carrier`
is the primary key, a simple key. The `weather` table has the
combination of `origin` and `time_hour` as primary key, a compound key.

You can get all primary keys in a `dm` by calling `dm_get_all_pks()`:

``` r
dm %>%
  dm_get_all_pks()
#> # A tibble: 4 x 2
#>   table    pk_col           
#>   <chr>    <keys>           
#> 1 airlines carrier          
#> 2 airports faa              
#> 3 planes   tailnum          
#> 4 weather  origin, time_hour
```

`dm_enum_pk_candidates()` checks all columns if they are suitable as a
simple primary key:

``` r
dm %>%
  dm_enum_pk_candidates(airports)
#> # A tibble: 8 x 3
#>   columns candidate why                                                    
#>   <keys>  <lgl>     <chr>                                                  
#> 1 faa     TRUE      ""                                                     
#> 2 lon     TRUE      ""                                                     
#> 3 name    FALSE     "has duplicate values: Municipal Airport (5), All Airp…
#> 4 lat     FALSE     "has duplicate values: 38.88944 (2), 40.63975 (2)"     
#> 5 alt     FALSE     "has duplicate values: 0 (51), 13 (13), 14 (12), 15 (1…
#> 6 tz      FALSE     "has duplicate values: -5 (521), -6 (342), -9 (240), -…
#> 7 dst     FALSE     "has duplicate values: A (1388), U (47), N (23)"       
#> 8 tzone   FALSE     "has duplicate values: America/New_York (519), America…
```

Further Reading: The {dm} package offers several functions for dealing
with [primary
keys](https://cynkra.github.io/dm/articles/tech-dm-class.html#pk).

## 4. Foreign Keys

The **counterpart of a primary key in one table is the foreign key in
another table**. In order to join two tables, the primary key of the
first table needs to be referenced from the second table. This column or
these columns are called the *foreign key* (abbreviated with “fk”).

For example, if you want to link the `airlines` table to the `flights`
table, the primary key in `airlines` needs to match the foreign key in
`flights`. This condition is satisfied because the column `carrier` is
present as a primary key in the `airlines` table as well as a foreign
key in the `flights` table. In the case of compound keys, the `origin`
and `time_hour` columns (which form the primary key of the `weather`
table) are also present in the `flights` table.

You can find foreign key candidates for simple keys with the function
`dm_enum_fk_candidates()`, they are marked with `TRUE` in the
`candidate` column.

``` r
dm %>%
  dm_enum_fk_candidates(flights, airlines)
#> # A tibble: 19 x 3
#>    columns      candidate why                                              
#>    <keys>       <lgl>     <chr>                                            
#>  1 carrier      TRUE      ""                                               
#>  2 year         FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  3 month        FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  4 day          FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  5 dep_time     FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  6 sched_dep_t… FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  7 dep_delay    FALSE     "Can't combine `value1` <double> and `value1` <c…
#>  8 arr_time     FALSE     "Can't combine `value1` <integer> and `value1` <…
#>  9 sched_arr_t… FALSE     "Can't combine `value1` <integer> and `value1` <…
#> 10 arr_delay    FALSE     "Can't combine `value1` <double> and `value1` <c…
#> 11 flight       FALSE     "Can't combine `value1` <integer> and `value1` <…
#> 12 tailnum      FALSE     "values of `flights$tailnum` not in `airlines$ca…
#> 13 origin       FALSE     "values of `flights$origin` not in `airlines$car…
#> 14 dest         FALSE     "values of `flights$dest` not in `airlines$carri…
#> 15 air_time     FALSE     "Can't combine `value1` <double> and `value1` <c…
#> 16 distance     FALSE     "Can't combine `value1` <double> and `value1` <c…
#> 17 hour         FALSE     "Can't combine `value1` <double> and `value1` <c…
#> 18 minute       FALSE     "Can't combine `value1` <double> and `value1` <c…
#> 19 time_hour    FALSE     "Can't combine `value1` <datetime<America/New_Yo…
```

Further Reading: All {dm} functions for working with [foreign
keys](https://cynkra.github.io/dm/articles/tech-dm-class.html#foreign-keys).

## 5. Referential Integrity

A data set has referential integrity if all relations between tables are
valid. That is, every foreign key holds a primary key that is present in
the parent table. If a foreign key contains a reference where the
corresponding row in the parent table is not availabe, that row is an
orphan row and the database no longer has referential integrity.

{dm} allows checking referential integrity with the
`dm_examine_constraints()` function. The following conditions are
checked:

-   All primary key values must be unique.
-   All primary key values must be set, `NA` values are not allowed.
-   Each foreign key value must have a corresponding primary key value.

In the example data model, for a substantial share of the flights,
detailed information for the corresponding airplane is not available:

``` r
dm %>%
  dm_examine_constraints()
#> ! Unsatisfied constraints:
#> ● Table `flights`: foreign key dest into table `airports`: values of `flights$dest` not in `airports$faa`: SJU (187), BQN (28), STT (15), PSE (12)
#> ● Table `flights`: foreign key tailnum into table `planes`: values of `flights$tailnum` not in `planes$tailnum`: N722MQ (27), N725MQ (20), N520MQ (19), N723MQ (19), N508MQ (16), …
```

Establishing referential integrity is important for providing clean data
for analysis or downstream users. See `vignette("howto-dm-rows")` for
more information on adding, deleting or updating individual rows, and
`vignette("tech-dm-zoom")` for operations on the data in a data model.

## 6. Normalization

Normalization is a technical term that describes the **central design
principle of a relational data model**: splitting data into multiple
tables.

A normalized data schema consists of several relations (tables) that are
linked with attributes (columns). The relations can be joined together
by means of [primary](#pk) and [foreign keys](#fk). The main goal of
normalization is to keep data organization as clean and simple as
possible by avoiding redundant data entries.

For example, if you want to change the name of one airport in the
`nycflights13` dataset, you will only need to update a single data
value. This principle is sometimes called the *single point of truth*.

``` r
#  Update in one single location...
airlines[airlines$carrier == "UA", "name"] <- "United broke my guitar"

airlines %>%
  filter(carrier == "UA")
#> # A tibble: 1 x 2
#>   carrier name                  
#>   <chr>   <chr>                 
#> 1 UA      United broke my guitar

# ...propagates to all related records
flights %>%
  left_join(airlines) %>%
  select(flight, name)
#> Joining, by = "carrier"
#> # A tibble: 336,776 x 2
#>    flight name                    
#>     <int> <chr>                   
#>  1   1545 United broke my guitar  
#>  2   1714 United broke my guitar  
#>  3   1141 American Airlines Inc.  
#>  4    725 JetBlue Airways         
#>  5    461 Delta Air Lines Inc.    
#>  6   1696 United broke my guitar  
#>  7    507 JetBlue Airways         
#>  8   5708 ExpressJet Airlines Inc.
#>  9     79 JetBlue Airways         
#> 10    301 American Airlines Inc.  
#> # … with 336,766 more rows
```

Another way to demonstrate normalization is splitting a table into two
parts.

Let’s look at the `planes` table, which consists of 3322 individual tail
numbers and corresponding information for the specific airplane, like
the year it was manufactured or the average cruising speed.

The function `decompose_table()` extracts two new tables and creates a
new key `model_id`, that links both tables.

This results in a `parent_table` and a `child_table` that differ
massively in the number of rows:

``` r
planes %>%
  decompose_table(model_id, model, manufacturer, type, engines, seats, manufacturer, speed)
#> $child_table
#> # A tibble: 3,322 x 4
#>    tailnum  year engine    model_id
#>    <chr>   <int> <chr>        <int>
#>  1 N10156   2004 Turbo-fan      120
#>  2 N102UW   1998 Turbo-fan       93
#>  3 N103US   1999 Turbo-fan       93
#>  4 N104UW   1999 Turbo-fan       93
#>  5 N10575   2002 Turbo-fan      119
#>  6 N105UW   1999 Turbo-fan       93
#>  7 N107US   1999 Turbo-fan       93
#>  8 N108UW   1999 Turbo-fan       93
#>  9 N109UW   1999 Turbo-fan       93
#> 10 N110UW   1999 Turbo-fan       93
#> # … with 3,312 more rows
#> 
#> $parent_table
#> # A tibble: 147 x 7
#>    model_id model      manufacturer   type              engines seats speed
#>       <int> <chr>      <chr>          <chr>               <int> <int> <int>
#>  1      120 EMB-145XR  EMBRAER        Fixed wing multi…       2    55    NA
#>  2       93 A320-214   AIRBUS INDUST… Fixed wing multi…       2   182    NA
#>  3      119 EMB-145LR  EMBRAER        Fixed wing multi…       2    55    NA
#>  4       39 737-824    BOEING         Fixed wing multi…       2   149    NA
#>  5       68 767-332    BOEING         Fixed wing multi…       2   330    NA
#>  6       52 757-224    BOEING         Fixed wing multi…       2   178    NA
#>  7       94 A320-214   AIRBUS         Fixed wing multi…       2   182    NA
#>  8      112 CL-600-2D… BOMBARDIER INC Fixed wing multi…       2    95    NA
#>  9       30 737-724    BOEING         Fixed wing multi…       2   149    NA
#> 10       27 737-524    BOEING         Fixed wing multi…       2   149    NA
#> # … with 137 more rows
```

While `child_table` contains 3322 unique `tailnum` rows and therefore
consists of 3322 rows, just like the original `planes` table, the
`parent_table` shrunk to just 147 rows, enough to store all relevant
combinations and avoid storing redundant information.

Further Reading: See the [Simple English Wikipedia article on database
normalization](https://simple.wikipedia.org/wiki/Database_normalisation)
for more details.

## 7. Relational Databases

{dm} is built upon relational data models but it is not a database
itself. Databases are systems for data management and many of them are
constructed as relational databases, e.g. SQLite, MySQL, MSSQL,
Postgres. As you can guess from the names of the databases, SQL, short
for Structured Querying Language, plays an important role: it was
invented for the purpose of querying relational databases.

In production, the data is stored in a relational database and {dm} is
used to work with the data.

Therefore, {dm} can copy data [from and to
databases](https://cynkra.github.io/dm/articles/dm.html#copy), and works
transparently with both in-memory data and with relational database
systems.

For example, let’s create a local SQLite database and copy the `dm`
object to it:

``` r
con_sqlite <- DBI::dbConnect(RSQLite::SQLite())
con_sqlite
#> <SQLiteConnection>
#>   Path: 
#>   Extensions: TRUE
DBI::dbListTables(con_sqlite)
#> character(0)

copy_dm_to(con_sqlite, dm)
DBI::dbListTables(con_sqlite)
#> [1] "airlines_2020_08_28_07_13_03_12345_1"
#> [2] "airports_2020_08_28_07_13_03_12345_1"
#> [3] "flights_2020_08_28_07_13_03_12345_1" 
#> [4] "planes_2020_08_28_07_13_03_12345_1"  
#> [5] "sqlite_stat1"                        
#> [6] "sqlite_stat4"                        
#> [7] "weather_2020_08_28_07_13_03_12345_1"
```

In the opposite direction, `dm` can also be populated with data from a
DB. Unfortunately, keys currently can be learned only for MSSQL and
Postgres, but not for SQLite. Therefore, the dm contains the tables but
not the keys:

``` r
dm_from_con(con_sqlite)
#> Keys could not be queried, use `learn_keys = FALSE` to mute this message.
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  sqlite 3.35.5 []
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines_2020_08_28_07_13_03_12345_1`, `airports_2020_08_28_07_13_03_12345_1`, `flights_2020_08_28_07_13_03_12345_1`, `planes_2020_08_28_07_13_03_12345_1`, `sqlite_stat1`, … (7 total)
#> Columns: 62
#> Primary keys: 0
#> Foreign keys: 0
```

Remember to terminate the database connection:

``` r
DBI::dbDisconnect(con_sqlite)
```

<!-- Generated by galley: do not edit by hand -->

Once you have all your primary keys set and all foreign key relations
defined, a graphical representation of your data model offers a
condensed view of the tables and the relationships between the tables.
The following functions can be used to visualize the `dm` object:[1]

1.  `dm_draw()`
2.  `dm_set_colors()`
3.  `dm_get_colors()`
4.  `dm_get_available_colors()`

We use the prepared example `dm` object `dm_nycflights13(cycle = TRUE)`:

``` r
library(dm)
library(dplyr)
#> 
#> Attaching package: 'dplyr'
#> The following object is masked from 'package:testthat':
#> 
#>     matches
#> The following objects are masked from 'package:stats':
#> 
#>     filter, lag
#> The following objects are masked from 'package:base':
#> 
#>     intersect, setdiff, setequal, union
flights_dm_w_many_keys <- dm_nycflights13(color = FALSE)
flights_dm_w_many_keys
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
#> Columns: 53
#> Primary keys: 4
#> Foreign keys: 4
```

The schema is drawn with `dm_draw()`.

``` r
dm_draw(flights_dm_w_many_keys)
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-draw_files/figure-gfm/unnamed-chunk-2-1.png)<!-- -->

You can use colors to visually group your tables into families to
reflect their logical grouping. The available colors are either hexcoded
colors or the standard R color names. The function
`dm_get_available_colors()` forwards to `grDevices::colors()`:

``` r
dm_get_available_colors()
#>   [1] "default"              "white"                "aliceblue"           
#>   [4] "antiquewhite"         "antiquewhite1"        "antiquewhite2"       
#>   [7] "antiquewhite3"        "antiquewhite4"        "aquamarine"          
#>  [10] "aquamarine1"          "aquamarine2"          "aquamarine3"         
#>  [13] "aquamarine4"          "azure"                "azure1"              
#>  [16] "azure2"               "azure3"               "azure4"              
#>  [19] "beige"                "bisque"               "bisque1"             
#>  [22] "bisque2"              "bisque3"              "bisque4"             
#>  [25] "black"                "blanchedalmond"       "blue"                
#>  [28] "blue1"                "blue2"                "blue3"               
#>  [31] "blue4"                "blueviolet"           "brown"               
#>  [34] "brown1"               "brown2"               "brown3"              
#>  [37] "brown4"               "burlywood"            "burlywood1"          
#>  [40] "burlywood2"           "burlywood3"           "burlywood4"          
#>  [43] "cadetblue"            "cadetblue1"           "cadetblue2"          
#>  [46] "cadetblue3"           "cadetblue4"           "chartreuse"          
#>  [49] "chartreuse1"          "chartreuse2"          "chartreuse3"         
#>  [52] "chartreuse4"          "chocolate"            "chocolate1"          
#>  [55] "chocolate2"           "chocolate3"           "chocolate4"          
#>  [58] "coral"                "coral1"               "coral2"              
#>  [61] "coral3"               "coral4"               "cornflowerblue"      
#>  [64] "cornsilk"             "cornsilk1"            "cornsilk2"           
#>  [67] "cornsilk3"            "cornsilk4"            "cyan"                
#>  [70] "cyan1"                "cyan2"                "cyan3"               
#>  [73] "cyan4"                "darkblue"             "darkcyan"            
#>  [76] "darkgoldenrod"        "darkgoldenrod1"       "darkgoldenrod2"      
#>  [79] "darkgoldenrod3"       "darkgoldenrod4"       "darkgray"            
#>  [82] "darkgreen"            "darkgrey"             "darkkhaki"           
#>  [85] "darkmagenta"          "darkolivegreen"       "darkolivegreen1"     
#>  [88] "darkolivegreen2"      "darkolivegreen3"      "darkolivegreen4"     
#>  [91] "darkorange"           "darkorange1"          "darkorange2"         
#>  [94] "darkorange3"          "darkorange4"          "darkorchid"          
#>  [97] "darkorchid1"          "darkorchid2"          "darkorchid3"         
#> [100] "darkorchid4"          "darkred"              "darksalmon"          
#> [103] "darkseagreen"         "darkseagreen1"        "darkseagreen2"       
#> [106] "darkseagreen3"        "darkseagreen4"        "darkslateblue"       
#> [109] "darkslategray"        "darkslategray1"       "darkslategray2"      
#> [112] "darkslategray3"       "darkslategray4"       "darkslategrey"       
#> [115] "darkturquoise"        "darkviolet"           "deeppink"            
#> [118] "deeppink1"            "deeppink2"            "deeppink3"           
#> [121] "deeppink4"            "deepskyblue"          "deepskyblue1"        
#> [124] "deepskyblue2"         "deepskyblue3"         "deepskyblue4"        
#> [127] "dimgray"              "dimgrey"              "dodgerblue"          
#> [130] "dodgerblue1"          "dodgerblue2"          "dodgerblue3"         
#> [133] "dodgerblue4"          "firebrick"            "firebrick1"          
#> [136] "firebrick2"           "firebrick3"           "firebrick4"          
#> [139] "floralwhite"          "forestgreen"          "gainsboro"           
#> [142] "ghostwhite"           "gold"                 "gold1"               
#> [145] "gold2"                "gold3"                "gold4"               
#> [148] "goldenrod"            "goldenrod1"           "goldenrod2"          
#> [151] "goldenrod3"           "goldenrod4"           "gray"                
#> [154] "gray0"                "gray1"                "gray2"               
#> [157] "gray3"                "gray4"                "gray5"               
#> [160] "gray6"                "gray7"                "gray8"               
#> [163] "gray9"                "gray10"               "gray11"              
#> [166] "gray12"               "gray13"               "gray14"              
#> [169] "gray15"               "gray16"               "gray17"              
#> [172] "gray18"               "gray19"               "gray20"              
#> [175] "gray21"               "gray22"               "gray23"              
#> [178] "gray24"               "gray25"               "gray26"              
#> [181] "gray27"               "gray28"               "gray29"              
#> [184] "gray30"               "gray31"               "gray32"              
#> [187] "gray33"               "gray34"               "gray35"              
#> [190] "gray36"               "gray37"               "gray38"              
#> [193] "gray39"               "gray40"               "gray41"              
#> [196] "gray42"               "gray43"               "gray44"              
#> [199] "gray45"               "gray46"               "gray47"              
#> [202] "gray48"               "gray49"               "gray50"              
#> [205] "gray51"               "gray52"               "gray53"              
#> [208] "gray54"               "gray55"               "gray56"              
#> [211] "gray57"               "gray58"               "gray59"              
#> [214] "gray60"               "gray61"               "gray62"              
#> [217] "gray63"               "gray64"               "gray65"              
#> [220] "gray66"               "gray67"               "gray68"              
#> [223] "gray69"               "gray70"               "gray71"              
#> [226] "gray72"               "gray73"               "gray74"              
#> [229] "gray75"               "gray76"               "gray77"              
#> [232] "gray78"               "gray79"               "gray80"              
#> [235] "gray81"               "gray82"               "gray83"              
#> [238] "gray84"               "gray85"               "gray86"              
#> [241] "gray87"               "gray88"               "gray89"              
#> [244] "gray90"               "gray91"               "gray92"              
#> [247] "gray93"               "gray94"               "gray95"              
#> [250] "gray96"               "gray97"               "gray98"              
#> [253] "gray99"               "gray100"              "green"               
#> [256] "green1"               "green2"               "green3"              
#> [259] "green4"               "greenyellow"          "grey"                
#> [262] "grey0"                "grey1"                "grey2"               
#> [265] "grey3"                "grey4"                "grey5"               
#> [268] "grey6"                "grey7"                "grey8"               
#> [271] "grey9"                "grey10"               "grey11"              
#> [274] "grey12"               "grey13"               "grey14"              
#> [277] "grey15"               "grey16"               "grey17"              
#> [280] "grey18"               "grey19"               "grey20"              
#> [283] "grey21"               "grey22"               "grey23"              
#> [286] "grey24"               "grey25"               "grey26"              
#> [289] "grey27"               "grey28"               "grey29"              
#> [292] "grey30"               "grey31"               "grey32"              
#> [295] "grey33"               "grey34"               "grey35"              
#> [298] "grey36"               "grey37"               "grey38"              
#> [301] "grey39"               "grey40"               "grey41"              
#> [304] "grey42"               "grey43"               "grey44"              
#> [307] "grey45"               "grey46"               "grey47"              
#> [310] "grey48"               "grey49"               "grey50"              
#> [313] "grey51"               "grey52"               "grey53"              
#> [316] "grey54"               "grey55"               "grey56"              
#> [319] "grey57"               "grey58"               "grey59"              
#> [322] "grey60"               "grey61"               "grey62"              
#> [325] "grey63"               "grey64"               "grey65"              
#> [328] "grey66"               "grey67"               "grey68"              
#> [331] "grey69"               "grey70"               "grey71"              
#> [334] "grey72"               "grey73"               "grey74"              
#> [337] "grey75"               "grey76"               "grey77"              
#> [340] "grey78"               "grey79"               "grey80"              
#> [343] "grey81"               "grey82"               "grey83"              
#> [346] "grey84"               "grey85"               "grey86"              
#> [349] "grey87"               "grey88"               "grey89"              
#> [352] "grey90"               "grey91"               "grey92"              
#> [355] "grey93"               "grey94"               "grey95"              
#> [358] "grey96"               "grey97"               "grey98"              
#> [361] "grey99"               "grey100"              "honeydew"            
#> [364] "honeydew1"            "honeydew2"            "honeydew3"           
#> [367] "honeydew4"            "hotpink"              "hotpink1"            
#> [370] "hotpink2"             "hotpink3"             "hotpink4"            
#> [373] "indianred"            "indianred1"           "indianred2"          
#> [376] "indianred3"           "indianred4"           "ivory"               
#> [379] "ivory1"               "ivory2"               "ivory3"              
#> [382] "ivory4"               "khaki"                "khaki1"              
#> [385] "khaki2"               "khaki3"               "khaki4"              
#> [388] "lavender"             "lavenderblush"        "lavenderblush1"      
#> [391] "lavenderblush2"       "lavenderblush3"       "lavenderblush4"      
#> [394] "lawngreen"            "lemonchiffon"         "lemonchiffon1"       
#> [397] "lemonchiffon2"        "lemonchiffon3"        "lemonchiffon4"       
#> [400] "lightblue"            "lightblue1"           "lightblue2"          
#> [403] "lightblue3"           "lightblue4"           "lightcoral"          
#> [406] "lightcyan"            "lightcyan1"           "lightcyan2"          
#> [409] "lightcyan3"           "lightcyan4"           "lightgoldenrod"      
#> [412] "lightgoldenrod1"      "lightgoldenrod2"      "lightgoldenrod3"     
#> [415] "lightgoldenrod4"      "lightgoldenrodyellow" "lightgray"           
#> [418] "lightgreen"           "lightgrey"            "lightpink"           
#> [421] "lightpink1"           "lightpink2"           "lightpink3"          
#> [424] "lightpink4"           "lightsalmon"          "lightsalmon1"        
#> [427] "lightsalmon2"         "lightsalmon3"         "lightsalmon4"        
#> [430] "lightseagreen"        "lightskyblue"         "lightskyblue1"       
#> [433] "lightskyblue2"        "lightskyblue3"        "lightskyblue4"       
#> [436] "lightslateblue"       "lightslategray"       "lightslategrey"      
#> [439] "lightsteelblue"       "lightsteelblue1"      "lightsteelblue2"     
#> [442] "lightsteelblue3"      "lightsteelblue4"      "lightyellow"         
#> [445] "lightyellow1"         "lightyellow2"         "lightyellow3"        
#> [448] "lightyellow4"         "limegreen"            "linen"               
#> [451] "magenta"              "magenta1"             "magenta2"            
#> [454] "magenta3"             "magenta4"             "maroon"              
#> [457] "maroon1"              "maroon2"              "maroon3"             
#> [460] "maroon4"              "mediumaquamarine"     "mediumblue"          
#> [463] "mediumorchid"         "mediumorchid1"        "mediumorchid2"       
#> [466] "mediumorchid3"        "mediumorchid4"        "mediumpurple"        
#> [469] "mediumpurple1"        "mediumpurple2"        "mediumpurple3"       
#> [472] "mediumpurple4"        "mediumseagreen"       "mediumslateblue"     
#> [475] "mediumspringgreen"    "mediumturquoise"      "mediumvioletred"     
#> [478] "midnightblue"         "mintcream"            "mistyrose"           
#> [481] "mistyrose1"           "mistyrose2"           "mistyrose3"          
#> [484] "mistyrose4"           "moccasin"             "navajowhite"         
#> [487] "navajowhite1"         "navajowhite2"         "navajowhite3"        
#> [490] "navajowhite4"         "navy"                 "navyblue"            
#> [493] "oldlace"              "olivedrab"            "olivedrab1"          
#> [496] "olivedrab2"           "olivedrab3"           "olivedrab4"          
#> [499] "orange"               "orange1"              "orange2"             
#> [502] "orange3"              "orange4"              "orangered"           
#> [505] "orangered1"           "orangered2"           "orangered3"          
#> [508] "orangered4"           "orchid"               "orchid1"             
#> [511] "orchid2"              "orchid3"              "orchid4"             
#> [514] "palegoldenrod"        "palegreen"            "palegreen1"          
#> [517] "palegreen2"           "palegreen3"           "palegreen4"          
#> [520] "paleturquoise"        "paleturquoise1"       "paleturquoise2"      
#> [523] "paleturquoise3"       "paleturquoise4"       "palevioletred"       
#> [526] "palevioletred1"       "palevioletred2"       "palevioletred3"      
#> [529] "palevioletred4"       "papayawhip"           "peachpuff"           
#> [532] "peachpuff1"           "peachpuff2"           "peachpuff3"          
#> [535] "peachpuff4"           "peru"                 "pink"                
#> [538] "pink1"                "pink2"                "pink3"               
#> [541] "pink4"                "plum"                 "plum1"               
#> [544] "plum2"                "plum3"                "plum4"               
#> [547] "powderblue"           "purple"               "purple1"             
#> [550] "purple2"              "purple3"              "purple4"             
#> [553] "red"                  "red1"                 "red2"                
#> [556] "red3"                 "red4"                 "rosybrown"           
#> [559] "rosybrown1"           "rosybrown2"           "rosybrown3"          
#> [562] "rosybrown4"           "royalblue"            "royalblue1"          
#> [565] "royalblue2"           "royalblue3"           "royalblue4"          
#> [568] "saddlebrown"          "salmon"               "salmon1"             
#> [571] "salmon2"              "salmon3"              "salmon4"             
#> [574] "sandybrown"           "seagreen"             "seagreen1"           
#> [577] "seagreen2"            "seagreen3"            "seagreen4"           
#> [580] "seashell"             "seashell1"            "seashell2"           
#> [583] "seashell3"            "seashell4"            "sienna"              
#> [586] "sienna1"              "sienna2"              "sienna3"             
#> [589] "sienna4"              "skyblue"              "skyblue1"            
#> [592] "skyblue2"             "skyblue3"             "skyblue4"            
#> [595] "slateblue"            "slateblue1"           "slateblue2"          
#> [598] "slateblue3"           "slateblue4"           "slategray"           
#> [601] "slategray1"           "slategray2"           "slategray3"          
#> [604] "slategray4"           "slategrey"            "snow"                
#> [607] "snow1"                "snow2"                "snow3"               
#> [610] "snow4"                "springgreen"          "springgreen1"        
#> [613] "springgreen2"         "springgreen3"         "springgreen4"        
#> [616] "steelblue"            "steelblue1"           "steelblue2"          
#> [619] "steelblue3"           "steelblue4"           "tan"                 
#> [622] "tan1"                 "tan2"                 "tan3"                
#> [625] "tan4"                 "thistle"              "thistle1"            
#> [628] "thistle2"             "thistle3"             "thistle4"            
#> [631] "tomato"               "tomato1"              "tomato2"             
#> [634] "tomato3"              "tomato4"              "turquoise"           
#> [637] "turquoise1"           "turquoise2"           "turquoise3"          
#> [640] "turquoise4"           "violet"               "violetred"           
#> [643] "violetred1"           "violetred2"           "violetred3"          
#> [646] "violetred4"           "wheat"                "wheat1"              
#> [649] "wheat2"               "wheat3"               "wheat4"              
#> [652] "whitesmoke"           "yellow"               "yellow1"             
#> [655] "yellow2"              "yellow3"              "yellow4"             
#> [658] "yellowgreen"
```

Colors are assigned with `dm_set_colors()` using syntax known in the
{tidyverse} as {tidyselect}-syntax, here in the form: `color = table`.
`select_helper` functions are supported. The result of `dm_set_colors()`
is a `dm` object. The information about the color is stored together
with the rest of the metadata.

``` r
flights_dm_w_many_keys_and_colors <-
  flights_dm_w_many_keys %>%
  dm_set_colors(
    maroon4 = flights,
    orange = starts_with("air"),
    "#5986C4" = planes
  )
```

Draw the schema with `dm_draw()`.

``` r
dm_draw(flights_dm_w_many_keys_and_colors)
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-draw_files/figure-gfm/unnamed-chunk-5-1.png)<!-- -->

The colors can be queried with `dm_get_colors()`.

``` r
dm_get_colors(flights_dm_w_many_keys_and_colors)
#>  #FFA500FF  #FFA500FF  #8B1C62FF  #5986C4FF    default 
#> "airlines" "airports"  "flights"   "planes"  "weather"
```

See the documentation for `dm_draw()` for further options. One important
argument is `view_type`. Besides the default `"keys_only"`, it accepts
`"all"` to display all columns, and `"title_only"` to show only the
title of the table.

``` r
flights_dm_w_many_keys_and_colors %>%
  dm_draw(view_type = "title_only")
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-draw_files/figure-gfm/unnamed-chunk-7-1.png)<!-- -->

If you would like to visualize only a some of the tables, use
`dm_select_tbl()` before drawing:

``` r
flights_dm_w_many_keys_and_colors %>%
  dm_select_tbl(flights, airports, planes) %>%
  dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/tech-dm-draw_files/figure-gfm/unnamed-chunk-8-1.png)<!-- -->

Finally, for exporting a drawing to `svg` you could use
`DiagrammeRsvg::export_svg()`:

``` r
flights_dm_w_many_keys_and_colors %>% 
  dm_select_tbl(flights, airports, planes) %>% 
  dm_draw() %>% 
  DiagrammeRsvg::export_svg() %>% 
  write("flights_dm_w_many_keys_and_color.svg")
```

[1] The code for the functions in this section is borrowed from the
[{datamodelr}](https://github.com/bergant/datamodelr) package.

<!-- Generated by galley: do not edit by hand -->

This vignette describes the logic behind the names of the
{dm}-functions.

In general the functions are implemented in such a way, that no function
can handle both `dm` classes (`dm` and `dm_zoomed`) as an argument. A
`dm_zoomed` behaves in most situations like a table and there are mostly
no explicit functions available for this class, but rather methods (see
vignette [“Zooming and manipulating
tables”](https://cynkra.github.io/dm/articles/tech-dm-zoom.html).

## Prefixes

1.  `dm_`: There are two cases for this prefix:
    -   The function takes a `dm` object as a first argument
    -   The function constructs a `dm` object from other objects
2.  `check_`: Certain tests on the arguments are performed and an error
    is thrown in case they are not passed. If the tests are passed, the
    first argument is returned invisibly.
3.  `examine_`: The function executes checks and returns an object
    containing an overview of the results of the checks.

Most important extended prefixes `dm_..._`:

1.  `dm_examine_`: The function examines several aspects of a `dm`
    object
2.  `dm_add_`: Adds either key constraints or new tables to a `dm`
3.  `dm_get_`: Retrieves a specific piece of information about the `dm`
    (often a part of its constituents)
4.  `dm_rm_`: Removes either key constraints or tables from a `dm`

## Suffixes

1.  `_dm`: Validators, checkers or coercers
2.  `_to_tbl`: Data from a `dm` is processed in a way that a tibble is
    produced
3.  `_candidates`: If key-candidates are to be calculated and presented
    in tibble-form
4.  `_pk(s)`: Relates to primary key
5.  `_fk(s)`: Relates to foreign key
6.  `_tbl`: Add/remove/select/rename a tibble to/from/within a dm
7.  `_colors`: Relates to the colors for the visualization of the dm
8.  `_zoomed`: Insert, update or discard the zoomed table. In
    combination with the prefix `dm_` this indicates, that the first
    argument is a `dm_zoomed` rather than a `dm`.

<!-- Generated by galley: do not edit by hand -->

## Introduction

In this article we’re going show you how easy it is to move from
connecting to the database holding your data to producing the results
you need. It’s meant to be a quick and friendly introduction to {dm}, so
it is low on details and caveats. Links to detailed documentation are
provided at the end. (If your data is in data frames instead of a
database and you’re in a hurry, jump over to `vignette("howto-dm-df")`.)

## Creating a dm object

dm objects can be created from individual tables or loaded directly from
a relational data model on an RDBMS (relational database management
system).

For this demonstration we’re going work with a model hosted on a public
server. The first thing we need is a connection to the RDBMS hosting the
data.

``` r
library(RMariaDB)

fin_db <- dbConnect(
  MariaDB(),
  username = 'guest',
  password = 'relational',
  dbname = 'Financial_ijs',
  host = 'relational.fit.cvut.cz'
)
```

We create a dm object from an RDBMS using `dm_from_con()`, passing in
the connection object we just created as the first argument.

``` r
library(dm)

fin_dm <- dm_from_con(fin_db)
fin_dm
#> ── Table source ───────────────────────────────────────────────────────────
#> src:  mysql  [guest@relational.fit.cvut.cz:NA/Financial_ijs]
#> ── Metadata ───────────────────────────────────────────────────────────────
#> Tables: `accounts`, `cards`, `clients`, `disps`, `districts`, … (9 total)
#> Columns: 57
#> Primary keys: 0
#> Foreign keys: 0
```

The dm object interrogates the RDBMS for table and column information
and, where implemented, primary and foreign keys. Currently, primary and
foreign keys are only available from Postgres and SQL Server.

## Selecting tables

The dm object can be accessed like a named list of tables:

``` r
names(fin_dm)
#> [1] "accounts"  "cards"     "clients"   "disps"     "districts" "loans"    
#> [7] "orders"    "tkeys"     "trans"
fin_dm$loans
#> # Source:   table<`Financial_ijs`.`loans`> [?? x 7]
#> # Database: mysql [guest@relational.fit.cvut.cz:NA/Financial_ijs]
#>       id account_id date       amount duration payments status
#>    <int>      <int> <date>      <dbl>    <int>    <dbl> <chr> 
#>  1  4959          2 1994-01-05  80952       24     3373 A     
#>  2  4961         19 1996-04-29  30276       12     2523 B     
#>  3  4962         25 1997-12-08  30276       12     2523 A     
#>  4  4967         37 1998-10-14 318480       60     5308 D     
#>  5  4968         38 1998-04-19 110736       48     2307 C     
#>  6  4973         67 1996-05-02 165960       24     6915 A     
#>  7  4986         97 1997-08-10 102876       12     8573 A     
#>  8  4988        103 1997-12-06 265320       36     7370 D     
#>  9  4989        105 1998-12-05 352704       48     7348 C     
#> 10  4990        110 1997-09-08 162576       36     4516 C     
#> # … with more rows
dplyr::count(fin_dm$trans)
#> # Source:   lazy query [?? x 1]
#> # Database: mysql [guest@relational.fit.cvut.cz:NA/Financial_ijs]
#>         n
#>   <int64>
#> 1 1056320
```

At the same time, most `dm` functions are pipe-friendly and support tidy
evaluation. We can use `[` or the `dm_select_tbl()` verb to derive a
smaller dm with the `loans`, `accounts`, `districts` and `trans` tables:

``` r
fin_dm_small <- fin_dm[c("loans", "accounts", "districts", "trans")]
fin_dm_small <-
  fin_dm %>%
  dm_select_tbl(loans, accounts, districts, trans)
```

## Linking tables by adding keys

In many cases, `dm_from_con()` already returns a dm with all keys set.
If not, dm allows us to define primary and foreign keys ourselves.

In our data model, `id` columns uniquely identify records in the
`accounts` and `loans` table, a primary key is added with `dm_add_pk()`.
Each loan is linked to one account via the `account_id` column in the
`loans` table, the relationship is established with `dm_add_fk()`.

``` r
fin_dm_keys <-
  fin_dm_small %>%
  dm_add_pk(accounts, id) %>%
  dm_add_pk(loans, id) %>%
  dm_add_fk(loans, account_id, accounts) %>%
  dm_add_pk(trans, id) %>%
  dm_add_fk(trans, account_id, accounts) %>%
  dm_add_pk(districts, id) %>%
  dm_add_fk(accounts, district_id, districts)
```

## Visualizing a data model

Having a diagram of the data model is the quickest way to verify we’re
on the right track. We can display a visual summary of the dm at any
time. The default is to display the table name, any defined keys and
their links to other tables.

Visualizing the dm in its current state we see the keys we have created
and how they link the tables together. Color guides the eye.

``` r
fin_dm_keys %>%
  dm_set_colors(green = c(loans, accounts), darkblue = trans, grey = districts) %>%
  dm_draw()
```

![](/home/kirill/git/R/dm/vignettes/out/dm_files/figure-gfm/visualize_keys-1.png)<!-- -->

## Accessing a data model as a table

If we want to perform modeling or analysis on this relational model we
need to transform it into a tabular format that R functions can work
with. `dm_squash_to_tbl()` will automatically follow foreign keys across
tables to gather all the available columns into a single table.

``` r
fin_dm_keys %>%
  dm_squash_to_tbl(loans)
#> Renamed columns:
#> * date -> loans.date, accounts.date
#> # Source:   lazy query [?? x 25]
#> # Database: mysql [guest@relational.fit.cvut.cz:NA/Financial_ijs]
#>       id account_id loans.date amount duration payments status district_id
#>    <int>      <int> <date>      <dbl>    <int>    <dbl> <chr>        <int>
#>  1  4959          2 1994-01-05  80952       24     3373 A                1
#>  2  4961         19 1996-04-29  30276       12     2523 B               21
#>  3  4962         25 1997-12-08  30276       12     2523 A               68
#>  4  4967         37 1998-10-14 318480       60     5308 D               20
#>  5  4968         38 1998-04-19 110736       48     2307 C               19
#>  6  4973         67 1996-05-02 165960       24     6915 A               16
#>  7  4986         97 1997-08-10 102876       12     8573 A               74
#>  8  4988        103 1997-12-06 265320       36     7370 D               44
#>  9  4989        105 1998-12-05 352704       48     7348 C               21
#> 10  4990        110 1997-09-08 162576       36     4516 C               36
#> # … with more rows, and 17 more variables: frequency <chr>,
#> #   accounts.date <date>, A2 <chr>, A3 <chr>, A4 <int>, A5 <int>,
#> #   A6 <int>, A7 <int>, A8 <int>, A9 <int>, A10 <dbl>, A11 <int>,
#> #   A12 <dbl>, A13 <dbl>, A14 <int>, A15 <int>, A16 <int>
```

Apart from the rows printed above, no data has been fetched from the
database. Use `select()` to reduce the number of columns fetched, and
`collect()` to retrieve the entire result for local processing.

``` r
loans_df <-
  fin_dm_keys %>%
  dm_squash_to_tbl(loans) %>%
  select(id, amount, duration, A3) %>%
  collect()
#> Renamed columns:
#> * date -> loans.date, accounts.date

model <- lm(amount ~ duration + A3, data = loans_df)

model
#> 
#> Call:
#> lm(formula = amount ~ duration + A3, data = loans_df)
#> 
#> Coefficients:
#>     (Intercept)         duration   A3east Bohemia  A3north Bohemia  
#>           10196             4109           -16204           -28933  
#> A3north Moravia         A3Prague  A3south Bohemia  A3south Moravia  
#>            1467             4044            -1896           -12463  
#>  A3west Bohemia  
#>          -28572
```

## Operations on table data within a dm

We don’t need to take the extra step of exporting the data to work with
it. Through the dm object we have complete access to dplyr’s data
manipulation verbs. These operate on the data within individual tables.

To work with a particular table we use `dm_zoom_to()` to set the context
to our chosen table. Then we can perform any of the operations we want.

``` r
fin_dm_total <-
  fin_dm_keys %>%
  dm_zoom_to(loans) %>%
  group_by(account_id) %>%
  summarize(total_amount = sum(amount, na.rm = TRUE)) %>%
  ungroup() %>%
  dm_insert_zoomed("total_loans")

fin_dm_total$total_loans
#> # Source:   lazy query [?? x 2]
#> # Database: mysql [guest@relational.fit.cvut.cz:NA/Financial_ijs]
#>    account_id total_amount
#>         <int>        <dbl>
#>  1          2        80952
#>  2         19        30276
#>  3         25        30276
#>  4         37       318480
#>  5         38       110736
#>  6         67       165960
#>  7         97       102876
#>  8        103       265320
#>  9        105       352704
#> 10        110       162576
#> # … with more rows
```

Note that in the above example we use `dm_insert_zoomed()` to add the
results as a new table to our data model. This table is temporary and
will be deleted when our session ends. If you want to make permanent
changes to your data model on an RDBMS please see the “Persisting
results” section in `vignette("howto-dm-db")`.

## Checking constraints

It’s always smart to check that your data model follows its
specifications. When building our own model or changing existing models
by adding tables or keys, it is even more important that the new model
is validated.

`dm_examine_constrains()` checks all primary and foreign keys and
reports if they violate their expected constraints.

``` r
fin_dm_total %>%
  dm_examine_constraints()
#> ℹ All constraints satisfied.
```

For more on constraint checking, including cardinality, finding
candidate columns for keys, and normalization, see
`vignette("tech-dm-low-level")`.

## Next Steps

Now that you have been introduced to the basic operation of dm, the next
step is to learn more about the dm methods that your particular use case
requires.

Is your data in an RDBMS? Then move on to `vignette("howto-dm-db")` for
a more detailed looking at working with an existing relational data
model.

If your data is in data frames, then you want to read
`vignette("howto-dm-df")` next.

If you would like to know more about relational data models in order
to get the most out of dm, check out `vignette("howto-dm-theory")`.

If you’re familiar with relational data models but want to know how to
work with them in dm, then any of `vignette("tech-dm-join")`,
`vignette("tech-dm-filter")`, or `vignette("tech-dm-zoom")` is a good
next step.

## Further reading

`vignette("howto-dm-db")` -- This article covers accessing and working
with RDBMSs within your R session, including manipulating data, filling
in missing relationships between tables, getting data out of the RDBMS
and into your model, and deploying your data model to an RDBMS.

`vignette("howto-dm-df")` -- Is your data in local data frames? This
article covers creating a data model from your local data frames,
including building the relationships in your data model, verifying your
model, and leveraging the power of dplyr to operate on your data model.

`vignette("howto-dm-theory")` -- Do you know all about data frames but
very little about relational data models? This quick introduction will
walk you through the key similarities and differences, and show you how
to move from individual data frames to a relational data model.

`vignette("tech-dm-join")` -- Joining is how separate, linked tables are
virtually combined in order to perform operations, generate summaries,
or prepare data for extraction. This article covers dm’s methods for
creating the different kind of joins along with usage examples.

`vignette("tech-dm-zoom")` -- Learn how to manipulate the data in the
tables in your dm using dplyr verbs.

`vignette("tech-dm-filter")` -- Need a subset of your data across tables?
This introduction to dm’s filtering methods will show you how to apply
them to data that is separated into multiple tables.

`vignette("tech-dm-draw")` -- A quick guide to dm’s methods for drawing
your data model for exploration, verification or documentation purposes.
Learn how to customize the diagrams.

`vignette("tech-dm-low-level")` -- If your data model needs to be rock
solid, this article covers the dm methods for checking every aspect of
the model. It covers key constraints, cardinalities between tables, and
methods for normalization as you construct your data model.

`vignette("tech-dm-class")` -- A detailed overview of the `"dm"` class.

`vignette("tech-dm-naming")` -- Making sense of dm’s API.

`vignette("tech-dm-cdm")` -- If you’re an early adopter, used dm version
0.0.5 or lower and want to migrate your code, this is for you.

## Standing on the shoulders of giants

The {dm} package follows the tidyverse principles:

-   `dm` objects are immutable (your data will never be overwritten in
    place)
-   most functions used on `dm` objects are pipeable (i.e., return new
    `dm` or table objects)
-   tidy evaluation is used (unquoted function parameters are supported)

The {dm} package builds heavily upon the [{datamodelr}
package](https://github.com/bergant/datamodelr), and upon the
[tidyverse](https://www.tidyverse.org/). We’re looking forward to a good
collaboration!

The [{polyply} package](https://github.com/russHyde/polyply) has a
similar intent with a slightly different interface.

The [{data.cube} package](https://github.com/jangorecki/data.cube) has
quite the same intent using `array`-like interface.

Articles in the [{rquery} package](https://github.com/WinVector/rquery)
discuss [join
controllers](https://github.com/WinVector/rquery/blob/master/extras/JoinController.md)
and [join dependency
sorting](https://github.com/WinVector/rquery/blob/master/extras/DependencySorting.md),
with the intent to move the declaration of table relationships from code
to data.

The [{tidygraph} package](https://github.com/thomasp85/tidygraph) stores
a network as two related tables of `nodes` and `edges`, compatible with
{dplyr} workflows.

In object-oriented programming languages, [object-relational
mapping](https://en.wikipedia.org/wiki/Object-relational_mapping) is a
similar concept that attempts to map a set of related tables to a class
hierarchy.

# output

    Code
      dm() %>% dm_examine_constraints()
    Message
      i No constraints defined.
    Code
      dm_nycflights_small() %>% dm_examine_constraints()
    Message
      ! Unsatisfied constraints:
    Output
      * Table `flights`: foreign key `dest` into table `airports`: values of `flights$dest` not in `airports$faa`: SJU (30), BQN (6), STT (4), PSE (2)
      * Table `flights`: foreign key `tailnum` into table `planes`: values of `flights$tailnum` not in `planes$tailnum`: N725MQ (6), N537MQ (5), N722MQ (5), N730MQ (5), N736MQ (5), ...
    Code
      dm_nycflights_small_cycle() %>% dm_examine_constraints()
    Message
      ! Unsatisfied constraints:
    Output
      * Table `flights`: foreign key `dest` into table `airports`: values of `flights$dest` not in `airports$faa`: SJU (30), BQN (6), STT (4), PSE (2)
      * Table `flights`: foreign key `tailnum` into table `planes`: values of `flights$tailnum` not in `planes$tailnum`: N725MQ (6), N537MQ (5), N722MQ (5), N730MQ (5), N736MQ (5), ...
    Code
      dm_nycflights_small_cycle() %>% dm_select_tbl(-flights) %>%
        dm_examine_constraints()
    Message
      i All constraints satisfied.
    Code
      # n column
      dm_for_filter_w_cycle() %>% dm_examine_constraints()
    Message
      i All constraints satisfied.

# output as tibble

    Code
      dm_nycflights_small_cycle() %>% dm_examine_constraints() %>% as_tibble()
    Output
      # A tibble: 7 x 6
        table    kind  columns ref_table is_key problem                               
        <chr>    <chr> <keys>  <chr>     <lgl>  <chr>                                 
      1 flights  FK    dest    airports  FALSE  "values of `flights$dest` not in `air~
      2 flights  FK    tailnum planes    FALSE  "values of `flights$tailnum` not in `~
      3 airlines PK    carrier <NA>      TRUE   ""                                    
      4 airports PK    faa     <NA>      TRUE   ""                                    
      5 planes   PK    tailnum <NA>      TRUE   ""                                    
      6 flights  FK    carrier airlines  TRUE   ""                                    
      7 flights  FK    origin  airports  TRUE   ""                                    

# .max_value parameter works

    Code
      dm_nycflights_small_cycle() %>% dm_examine_constraints(.max_value = Inf)
    Message
      ! Unsatisfied constraints:
    Output
      * Table `flights`: foreign key `dest` into table `airports`: values of `flights$dest` not in `airports$faa`: SJU (30), BQN (6), STT (4), PSE (2)
      * Table `flights`: foreign key `tailnum` into table `planes`: values of `flights$tailnum` not in `planes$tailnum`: N725MQ (6), N537MQ (5), N722MQ (5), N730MQ (5), N736MQ (5), N3GBAA (4), N509MQ (4), N719MQ (4), N723MQ (4), N734MQ (4), N739MQ (4), N856MQ (4), N3CJAA (3), N3FRAA (3), N500MQ (3), N513MQ (3), N515MQ (3), N526MQ (3), N532MQ (3), N534MQ (3), N5FAAA (3), N713MQ (3), N846MQ (3), N0EGMQ (2), N3AEMQ (2), N3BCAA (2), N3BTAA (2), N3CCAA (2), N3CDAA (2), N3CHAA (2), N3CSAA (2), N3CUAA (2), N3DXAA (2), N3ERAA (2), N3ETAA (2), N3EYAA (2), N3FAAA (2), N3FWAA (2), N3GFAA (2), N3GRAA (2), N3HYAA (2), N3JWAA (2), N4YFAA (2), N4YGAA (2), N501MQ (2), N503MQ (2), N504MQ (2), N507JB (2), N507MQ (2), N516MQ (2), N517MQ (2), N518MQ (2), N522MQ (2), N525UA (2), N527MQ (2), N539AA (2), N542MQ (2), N586UA (2), N5DEAA (2), N5DWAA (2), N5FLAA (2), N832MQ (2), N835MQ (2), N850MQ (2), N854MQ (2), N8EGMQ (2), N318AT (1), N395AA (1), N3ACAA (1), N3AHAA (1), N3APAA (1), N3BFAA (1), N3BGAA (1), N3BKAA (1), N3BSAA (1), N3BUAA (1), N3CEAA (1), N3CFAA (1), N3CGAA (1), N3CWAA (1), N3CXAA (1), N3DFAA (1), N3DGAA (1), N3DHAA (1), N3DPAA (1), N3DRAA (1), N3DSAA (1), N3DUAA (1), N3DWAA (1), N3DYAA (1), N3EAAA (1), N3ECAA (1), N3EHAA (1), N3EUAA (1), N3EXAA (1), N3GAAA (1), N3GTAA (1), N3GUAA (1), N3HAAA (1), N3HCAA (1), N3HXAA (1), N3JAAA (1), N3JBAA (1), N3JJAA (1), N3JMAA (1), N3JNAA (1), N3JPAA (1), N4WKAA (1), N4WMAA (1), N4WNAA (1), N4WWAA (1), N4XBAA (1), N4XKAA (1), N4XMAA (1), N4XNAA (1), N4XSAA (1), N4YBAA (1), N4YHAA (1), N4YJAA (1), N4YKAA (1), N4YSAA (1), N502MQ (1), N505MQ (1), N506MQ (1), N510MQ (1), N511MQ (1), N512MQ (1), N514MQ (1), N526JB (1), N527AA (1), N527JB (1), N527UA (1), N528MQ (1), N531MQ (1), N546MQ (1), N592UA (1), N5BSAA (1), N5CBAA (1), N5CPAA (1), N5DAAA (1), N5DMAA (1), N5EAAA (1), N5EGAA (1), N5EKAA (1), N5EUAA (1), N5FEAA (1), N5FFAA (1), N5FPAA (1), N5FSAA (1), N620MQ (1), N623MQ (1), N630MQ (1), N631MQ (1), N635MQ (1), N636MQ (1), N638MQ (1), N659MQ (1), N660MQ (1), N694MQ (1), N695MQ (1), N829MQ (1), N902MQ (1), N909MQ (1), N922MQ (1), N923MQ (1), N932MQ (1)

# output for compound keys

    Code
      bad_dm() %>% dm_examine_constraints()
    Message
      ! Unsatisfied constraints:
    Output
      * Table `tbl_2`: primary key `id`, `x`: has duplicate values: 3, E (2)
      * Table `tbl_3`: primary key `id`: has duplicate values: 4 (2)
      * Table `tbl_1`: foreign key `a`, `x` into table `tbl_2`: values of `tbl_1$a`, `tbl_1$x` not in `tbl_2$id`, `tbl_2$x`: 4, E (1), 5, F (1)
      * Table `tbl_1`: foreign key `b` into table `tbl_3`: values of `tbl_1$b` not in `tbl_3$id`: 1 (1), 5 (1)

# Non-explicit PKs should be tested too

    Code
      dm_for_card() %>% dm_examine_constraints()
    Message
      ! Unsatisfied constraints:
    Output
      * Table `dc_3`: unique key `b`, `a`: has duplicate values: e, 5 (2)
      * Table `dc_4`: foreign key `b`, `a` into table `dc_3`: values of `dc_4$b`, `dc_4$a` not in `dc_3$b`, `dc_3$a`: f, 6 (1)
      * Table `dc_6`: foreign key `c` into table `dc_1`: values of `dc_6$c` not in `dc_1$a`: 6 (1)

# `dm_examine_constraints()` API

    Code
      dm_examine_constraints(dm_test_obj(), progress = FALSE)
    Condition
      Warning:
      The `progress` argument of `dm_examine_constraints()` is deprecated as of dm 1.0.0.
      i Please use the `.progress` argument instead.
    Message
      i No constraints defined.
    Code
      dm_examine_constraints(dm = dm_test_obj())
    Condition
      Warning:
      The `dm` argument of `dm_examine_constraints()` is deprecated as of dm 1.0.0.
      i Please use the `.dm` argument instead.
    Message
      i No constraints defined.

# `dm_examine_constraints()` API (2)

    Code
      dm_examine_constraints(dm_test_obj(), foo = "bar")
    Condition
      Error in `dm_examine_constraints()`:
      ! `...` must be empty.
      x Problematic argument:
      * foo = "bar"

# output

    Code
      abort_key_set_force_false("pk_table")
    Condition
      Error in `abort_key_set_force_false()`:
      ! Table pk_table already has a primary key. Use `force = TRUE` to change the existing primary key.
    Code
      abort_not_unique_key("Christmas", "Rudolph")
    Condition
      Error in `abort_not_unique_key()`:
      ! (`Rudolph`) not a unique key of `Christmas`.
    Code
      abort_not_unique_key("Christmas", c("elves", "Rudolph", "mulled_wine"))
    Condition
      Error in `abort_not_unique_key()`:
      ! (`elves`, `Rudolph`, `mulled_wine`) not a unique key of `Christmas`.
    Code
      abort_table_not_in_dm("laziness", "hard_work")
    Condition
      Error in `abort_table_not_in_dm()`:
      ! Table laziness not in <dm> object. Available table names: hard_work.
    Code
      abort_not_subset_of("playing", "game", "hunting", "game")
    Condition
      Error in `abort_not_subset_of()`:
      ! Column (`game`) of table playing contains values (see examples above) that are not present in column (`game`) of table hunting.
    Code
      abort_sets_not_equal(c("A problem occurred.",
        "And another, even worse problem, occurred shortly after."))
    Condition
      Error in `abort_sets_not_equal()`:
      ! A problem occurred.
      And another, even worse problem, occurred shortly after.
    Code
      abort_not_bijective("child_table_name", "fk_col_name")
    Condition
      Error in `abort_not_bijective()`:
      ! 1..1 cardinality (bijectivity) is not given: Column (`fk_col_name`) in table child_table_name contains duplicate values.
    Code
      abort_not_injective("child_table_name", "fk_col_name")
    Condition
      Error in `abort_not_injective()`:
      ! 0..1 cardinality (injectivity from child table to parent table) is not given: Column (`fk_col_name`) in table child_table_name contains duplicate values.
    Code
      abort_ref_tbl_has_no_pk("parent_table")
    Condition
      Error in `abort_ref_tbl_has_no_pk()`:
      ! ref_table parent_table needs a primary key first. Use `dm_enum_pk_candidates()` to find appropriate columns and `dm_add_pk()` to define a primary key.
    Code
      abort_is_not_fkc()
    Condition
      Error in `abort_is_not_fkc()`:
      ! No foreign keys to remove.
    Code
      abort_rm_fk_col_missing()
    Condition
      Error in `abort_rm_fk_col_missing()`:
      ! Parameter `columns` has to be set. Pass NULL for removing all references.
    Code
      abort_last_col_missing()
    Condition
      Error in `abort_last_col_missing()`:
      ! The last color can't be missing.
    Code
      abort_no_cycles(create_graph_from_dm(dm_for_filter_w_cycle()))
    Condition
      Error in `abort_no_cycles()`:
      ! Cycles in the relationship graph not yet supported.
      i Shortest cycle: tf_5 -> tf_6 -> tf_7 -> tf_2 -> tf_3 -> tf_4 -> tf_5
    Code
      abort_tables_not_reachable_from_start()
    Condition
      Error in `abort_tables_not_reachable_from_start()`:
      ! All selected tables must be reachable from `.start`.
    Code
      abort_dupl_new_id_col_name("tibbletable")
    Condition
      Error in `abort_dupl_new_id_col_name()`:
      ! `new_id_column` can't have an identical name as one of the columns of tibbletable.
    Code
      abort_no_overwrite()
    Condition
      Error in `abort_no_overwrite()`:
      ! The `overwrite` argument is not supported.
    Code
      abort_pk_not_defined()
    Condition
      Error in `abort_pk_not_defined()`:
      ! No primary keys to remove.
    Code
      abort_fk_exists("child", c("child_1", "child_2"), "parent")
    Condition
      Error in `abort_fk_exists()`:
      ! (`child_1`, `child_2`) is already a foreign key of table child into table parent.
    Code
      abort_first_rm_fks("parent", c("child_1", "child_2"))
    Condition
      Error in `abort_first_rm_fks()`:
      ! There are foreign keys pointing from tables child_1 and child_2 to table parent. First remove those, or set `fail_fk = FALSE`.
    Code
      abort_update_not_supported()
    Condition
      Error in `abort_update_not_supported()`:
      ! Updating <dm> objects not supported.
    Code
      abort_only_possible_wo_filters("find_wisdom")
    Condition
      Error in `abort_only_possible_wo_filters()`:
      ! Not supported on a <dm> with filter conditions. Consider using `dm_apply_filters()` first.
    Code
      abort_tables_not_neighbors("subjects", "king")
    Condition
      Error in `abort_tables_not_neighbors()`:
      ! Tables subjects and king are not directly linked by a foreign key relation.
    Code
      abort_only_parents("dm_flatten", "table", "recursive")
    Condition
      Error in `abort_only_parents()`:
      ! All join partners of table `table` must be its direct neighbors. Use `recursive = TRUE` for recursive flattening.
    Code
      abort_not_same_src()
    Condition
      Error in `abort_not_same_src()`:
      ! Not all tables in the object share the same `src`.
    Code
      abort_what_a_weird_object("monster")
    Condition
      Error in `abort_what_a_weird_object()`:
      ! Don't know how to determine table source for object of class <monster>.
    Code
      abort_not_same_src()
    Condition
      Error in `abort_not_same_src()`:
      ! Not all tables in the object share the same `src`.
    Code
      abort_squash_limited()
    Condition
      Error in `abort_squash_limited()`:
      ! Recursive flattening only supports `left_join()`, `inner_join()`, or `full_join()`.
    Code
      abort_apply_filters_first("join_tightly")
    Condition
      Error in `abort_apply_filters_first()`:
      ! `dm_..._to_tbl()` with join using `join_tightly()` generally wouldn't produce the correct result when filters are set. Please consider calling `dm_apply_filters()` first.
    Code
      abort_no_flatten_with_nest_join()
    Condition
      Error in `abort_no_flatten_with_nest_join()`:
      ! `join = nest_join` is not supported. Consider `join = left_join`.
    Code
      abort_is_not_dm("blob")
    Condition
      Error in `abort_is_not_dm()`:
      ! Required class <dm> but instead is <blob>.
    Code
      abort_con_only_for_dbi()
    Condition
      Error in `abort_con_only_for_dbi()`:
      ! A local <dm> doesn't have a DB connection.
    Code
      abort_only_possible_wo_zoom("dm_zoom_to")
    Condition
      Error in `abort_only_possible_wo_zoom()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.
    Code
      abort_only_possible_w_zoom("dm_update_zoomed")
    Condition
      Error in `abort_only_possible_w_zoom()`:
      ! Not supported on an unzoomed <dm>. Consider using `dm_zoom_to()` first.
    Code
      abort_learn_keys(error_cnd(message = "some reason"))
    Condition
      Error in `abort_learn_keys()`:
      ! Failed to learn keys from database.
      i Use `learn_keys = FALSE` to work around, or `dm:::dm_meta()` to debug.
      Caused by error:
      ! some reason
    Code
      abort_tbl_access("accessdenied")
    Condition
      Error in `abort_tbl_access()`:
      ! Table accessdenied cannot be accessed.
      i Use `tbl(src, ...)` to troubleshoot.
    Code
      abort_need_unique_names("clone")
    Condition
      Error in `abort_need_unique_names()`:
      ! Each new table needs to have a unique name. Duplicate new name: clone.
    Code
      abort_fk_not_tracked("hook", "eye")
    Condition
      Error in `abort_fk_not_tracked()`:
      ! The foreign key that existed between the originally zoomed table hook and eye got lost in transformations. Please explicitly provide the `by` argument.
    Code
      abort_dm_invalid("it's ugly.")
    Condition
      Error in `abort_dm_invalid()`:
      ! This <dm> is invalid, reason: it's ugly.
    Code
      abort_no_table_provided()
    Condition
      Error in `abort_no_table_provided()`:
      ! Argument `table` is missing.
    Code
      abort_table_not_zoomed("blur", c("focus_1", "focus_2"))
    Condition
      Error in `abort_table_not_zoomed()`:
      ! Table `blur` not zoomed, zoomed tables: `focus_1` and `focus_2`.
    Code
      abort_not_pulling_multiple_zoomed()
    Condition
      Error in `abort_not_pulling_multiple_zoomed()`:
      ! If more than one zoomed table is available, you need to specify argument `table`.
    Code
      abort_cols_not_avail(c("pink5", "elephant"))
    Condition
      Error in `abort_cols_not_avail()`:
      ! The colors "pink5" and "elephant" are not available. Call `dm_get_available_colors()` for possible color names or use hex color codes.
    Code
      abort_only_named_args("give_names", "frobnicability")
    Condition
      Error in `abort_only_named_args()`:
      ! All `...` arguments must be named. The names represent frobnicability.
    Code
      abort_wrong_syntax_set_cols()
    Condition
      Error in `abort_wrong_syntax_set_cols()`:
      ! You seem to be using outdated syntax for setting colors, type `?dm_set_colors()` for examples.
    Code
      abort_pk_not_tracked("house", "house_number")
    Condition
      Error in `abort_pk_not_tracked()`:
      ! The primary key column(s) `house_number` of the originally zoomed table house got lost in transformations.
    Code
      abort_only_for_local_src(mtcars)
    Condition
      Error in `abort_only_for_local_src()`:
      ! Only supported for a local `src`, not on a database with `src`-class: <data.frame>.
    Code
      abort_parameter_not_correct_class("number", correct_class = "numeric", class = "logical")
    Condition
      Error in `abort_parameter_not_correct_class()`:
      ! Parameter `number` needs to be of class <numeric> but is of class <logical>.
    Code
      abort_parameter_not_correct_length("length_1_parameter", 1, letters[1:26])
    Condition
      Error in `abort_parameter_not_correct_length()`:
      ! Argument `length_1_parameter` needs to be of length 1 but is of length 26 ("a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", ..., "y", and "z").
    Code
      warn_if_arg_not("NULL", c("MSSQL", "Postgres"), arg_name = "dbms_dependent_arg")
    Condition
      Warning:
      Argument `dbms_dependent_arg` ignored: currently only supported for MSSQL and Postgres.
    Output
      NULL
    Code
      abort_no_schemas_supported("FantasticDatabaseManagementSystem",
        "hyperconnection")
    Condition
      Error in `abort_no_schemas_supported()`:
      ! The concept of schemas is not supported for DBMS `FantasticDatabaseManagementSystem`.
    Code
      abort_no_schemas_supported(con = 1)
    Condition
      Error in `abort_no_schemas_supported()`:
      ! Currently schemas are not supported for a connection of class <numeric>.
    Code
      abort_no_schemas_supported()
    Condition
      Error in `abort_no_schemas_supported()`:
      ! Schemas are not available locally.
    Code
      abort_temporary_not_in_schema()
    Condition
      Error in `abort_temporary_not_in_schema()`:
      ! If argument `temporary` is TRUE, argument `schema` has to be NULL.
    Code
      abort_one_of_schema_table_names()
    Condition
      Error in `abort_one_of_schema_table_names()`:
      ! Only one of the arguments `schema` and `table_names` can be different from NULL.

# singular and plural

    Code
      abort_not_subset_of("t1", c("a", "b"), "t2", c("c", "d"))
    Condition
      Error in `abort_not_subset_of()`:
      ! Columns (`a`, `b`) of table t1 contain values (see examples above) that are not present in columns (`c`, `d`) of table t2.
    Code
      abort_not_bijective("child_tbl", c("col_a", "col_b"))
    Condition
      Error in `abort_not_bijective()`:
      ! 1..1 cardinality (bijectivity) is not given: Columns (`col_a`, `col_b`) in table child_tbl contain duplicate values.
    Code
      abort_not_injective("child_tbl", c("col_a", "col_b"))
    Condition
      Error in `abort_not_injective()`:
      ! 0..1 cardinality (injectivity from child table to parent table) is not given: Columns (`col_a`, `col_b`) in table child_tbl contain duplicate values.
    Code
      abort_need_unique_names(c("clone_a", "clone_b"))
    Condition
      Error in `abort_need_unique_names()`:
      ! Each new table needs to have a unique name. Duplicate new names: clone_a and clone_b.
    Code
      abort_first_rm_fks("parent", "child_1")
    Condition
      Error in `abort_first_rm_fks()`:
      ! There is a foreign key pointing from table child_1 to table parent. First remove those, or set `fail_fk = FALSE`.
    Code
      abort_tbl_access(c("table_a", "table_b"))
    Condition
      Error in `abort_tbl_access()`:
      ! Tables table_a and table_b cannot be accessed.
      i Use `tbl(src, ...)` to troubleshoot.
    Code
      abort_cols_not_avail("pink5")
    Condition
      Error in `abort_cols_not_avail()`:
      ! The color "pink5" is not available. Call `dm_get_available_colors()` for possible color names or use hex color codes.

---

    Code
      warn_tbl_access("table_a")
    Condition
      Warning:
      Table table_a cannot be accessed.
      i Use `tbl(src, ...)` to troubleshoot.
      i Set the `table_name` argument to avoid this warning.
    Code
      warn_tbl_access(c("table_a", "table_b"))
    Condition
      Warning:
      Tables table_a and table_b cannot be accessed.
      i Use `tbl(src, ...)` to troubleshoot.
      i Set the `table_name` argument to avoid this warning.

# basic test: 'join()'-methods for `zoomed.dm` work (2)

    Code
      # keys are correctly tracked if selected columns from 'y' have same name as key columns from 'x'
      dm_zoomed() %>% left_join(tf_3, select = c(d = g, f, f1)) %>% dm_update_zoomed() %>%
        get_all_keys()
    Message
      Renaming ambiguous columns: %>%
        dm_rename(tf_2, d.tf_2 = d) %>%
        dm_rename(tf_3, d.tf_3 = d)
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d.tf_2        tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      # keys are correctly tracked if selected columns from 'y' have same name as key columns from 'x'
      dm_zoomed() %>% semi_join(tf_3, select = c(d = g, f, f1)) %>% dm_update_zoomed() %>%
        get_all_keys()
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      

# key tracking works

    Code
      # rename()
      zoomed_grouped_out_dm %>% rename(c_new = c) %>% dm_update_zoomed() %>%
        get_all_keys()
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c_new  FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      zoomed_grouped_out_dm %>% rename(e_new = e) %>% dm_update_zoomed() %>%
        get_all_keys()
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e_new, e1     tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      zoomed_grouped_in_dm %>% rename(f_new = f) %>% dm_update_zoomed() %>%
        get_all_keys()
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col    autoincrement
        <chr> <keys>    <lgl>        
      1 tf_1  a         TRUE         
      2 tf_2  c         FALSE        
      3 tf_3  f_new, f1 FALSE        
      4 tf_4  h         FALSE        
      5 tf_5  k         FALSE        
      6 tf_6  o         FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f_new, f1       no_action
      3 tf_4        j, j1         tf_3         f_new, f1       no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      # summarize()
      zoomed_grouped_out_dm %>% summarize(d_mean = mean(d)) %>% dm_insert_zoomed(
        "new_tbl") %>% get_all_keys()
    Output
      $pks
      # A tibble: 7 x 3
        table   pk_col autoincrement
        <chr>   <keys> <lgl>        
      1 tf_1    a      TRUE         
      2 tf_2    c      FALSE        
      3 tf_3    f, f1  FALSE        
      4 tf_4    h      FALSE        
      5 tf_5    k      FALSE        
      6 tf_6    o      FALSE        
      7 new_tbl c      FALSE        
      
      $fks
      # A tibble: 6 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 new_tbl     e, e1         tf_3         f, f1           no_action
      5 tf_5        l             tf_4         h               cascade  
      6 tf_5        m             tf_6         n               no_action
      
    Code
      zoomed_grouped_in_dm %>% summarize(g_list = list(g)) %>% dm_insert_zoomed(
        "new_tbl") %>% get_all_keys()
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      

# key tracking works (2)

    Code
      # transmute()
      zoomed_grouped_out_dm %>% transmute(d_mean = mean(d)) %>% dm_insert_zoomed(
        "new_tbl") %>% get_all_keys()
    Output
      $pks
      # A tibble: 7 x 3
        table   pk_col autoincrement
        <chr>   <keys> <lgl>        
      1 tf_1    a      TRUE         
      2 tf_2    c      FALSE        
      3 tf_3    f, f1  FALSE        
      4 tf_4    h      FALSE        
      5 tf_5    k      FALSE        
      6 tf_6    o      FALSE        
      7 new_tbl c      FALSE        
      
      $fks
      # A tibble: 6 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 new_tbl     e, e1         tf_3         f, f1           no_action
      5 tf_5        l             tf_4         h               cascade  
      6 tf_5        m             tf_6         n               no_action
      

# key tracking works (3)

    Code
      zoomed_grouped_in_dm %>% transmute(g_list = list(g)) %>% dm_insert_zoomed(
        "new_tbl") %>% get_all_keys()
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      

# key tracking works (4)

    Code
      # mutate()
      zoomed_grouped_out_dm %>% mutate(d_mean = mean(d)) %>% select(-d) %>%
        dm_insert_zoomed("new_tbl") %>% get_all_keys()
    Output
      $pks
      # A tibble: 7 x 3
        table   pk_col autoincrement
        <chr>   <keys> <lgl>        
      1 tf_1    a      TRUE         
      2 tf_2    c      FALSE        
      3 tf_3    f, f1  FALSE        
      4 tf_4    h      FALSE        
      5 tf_5    k      FALSE        
      6 tf_6    o      FALSE        
      7 new_tbl c      FALSE        
      
      $fks
      # A tibble: 6 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 new_tbl     e, e1         tf_3         f, f1           no_action
      5 tf_5        l             tf_4         h               cascade  
      6 tf_5        m             tf_6         n               no_action
      
    Code
      zoomed_grouped_in_dm %>% mutate(f = paste0(g, g)) %>% dm_insert_zoomed(
        "new_tbl") %>% get_all_keys()
    Output
      $pks
      # A tibble: 7 x 3
        table   pk_col autoincrement
        <chr>   <keys> <lgl>        
      1 tf_1    a      TRUE         
      2 tf_2    c      FALSE        
      3 tf_3    f, f1  FALSE        
      4 tf_4    h      FALSE        
      5 tf_5    k      FALSE        
      6 tf_6    o      FALSE        
      7 new_tbl f, f1  FALSE        
      
      $fks
      # A tibble: 7 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      6 tf_2        e, e1         new_tbl      f, f1           no_action
      7 tf_4        j, j1         new_tbl      f, f1           no_action
      
    Code
      zoomed_grouped_in_dm %>% mutate(g_new = list(g)) %>% dm_insert_zoomed("new_tbl") %>%
        get_all_keys()
    Output
      $pks
      # A tibble: 7 x 3
        table   pk_col autoincrement
        <chr>   <keys> <lgl>        
      1 tf_1    a      TRUE         
      2 tf_2    c      FALSE        
      3 tf_3    f, f1  FALSE        
      4 tf_4    h      FALSE        
      5 tf_5    k      FALSE        
      6 tf_6    o      FALSE        
      7 new_tbl f, f1  FALSE        
      
      $fks
      # A tibble: 7 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      6 tf_2        e, e1         new_tbl      f, f1           no_action
      7 tf_4        j, j1         new_tbl      f, f1           no_action
      

# key tracking works (5)

    Code
      # chain of renames & other transformations
      zoomed_grouped_out_dm %>% summarize(d_mean = mean(d)) %>% ungroup() %>% rename(
        e_new = e) %>% group_by(e_new, e1) %>% transmute(c = paste0(c, "_animal")) %>%
        dm_insert_zoomed("new_tbl") %>% get_all_keys()
    Output
      $pks
      # A tibble: 7 x 3
        table   pk_col autoincrement
        <chr>   <keys> <lgl>        
      1 tf_1    a      TRUE         
      2 tf_2    c      FALSE        
      3 tf_3    f, f1  FALSE        
      4 tf_4    h      FALSE        
      5 tf_5    k      FALSE        
      6 tf_6    o      FALSE        
      7 new_tbl c      FALSE        
      
      $fks
      # A tibble: 6 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 new_tbl     e_new, e1     tf_3         f, f1           no_action
      5 tf_5        l             tf_4         h               cascade  
      6 tf_5        m             tf_6         n               no_action
      

# key tracking works (6)

    Code
      zoomed_grouped_in_dm %>% select(g_new = g) %>% get_all_keys("tf_3")
    Output
      $pks
      # A tibble: 1 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_3  f, f1  FALSE        
      
      $fks
      # A tibble: 2 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        e, e1         tf_3         f, f1           no_action
      2 tf_4        j, j1         tf_3         f, f1           no_action
      

# output for compound keys

    Code
      grouped_zoomed_comp_dm_1 %>% mutate(count = n()) %>% col_tracker_zoomed()
    Output
            origin         year        month          day         hour         temp 
          "origin"       "year"      "month"        "day"       "hour"       "temp" 
              dewp        humid     wind_dir   wind_speed    wind_gust       precip 
            "dewp"      "humid"   "wind_dir" "wind_speed"  "wind_gust"     "precip" 
          pressure        visib    time_hour 
        "pressure"      "visib"  "time_hour" 
    Code
      grouped_zoomed_comp_dm_2 %>% mutate(count = n()) %>% col_tracker_zoomed()
    Output
            origin         year        month          day         hour         temp 
          "origin"       "year"      "month"        "day"       "hour"       "temp" 
              dewp        humid     wind_dir   wind_speed    wind_gust       precip 
            "dewp"      "humid"   "wind_dir" "wind_speed"  "wind_gust"     "precip" 
          pressure        visib    time_hour 
        "pressure"      "visib"  "time_hour" 
    Code
      grouped_zoomed_comp_dm_1 %>% transmute(count = n()) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 41
      Primary keys: 3
      Foreign keys: 3
    Code
      grouped_zoomed_comp_dm_2 %>% transmute(count = n()) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 41
      Primary keys: 4
      Foreign keys: 4
    Code
      grouped_zoomed_comp_dm_1 %>% summarize(count = n()) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 41
      Primary keys: 3
      Foreign keys: 3
    Code
      grouped_zoomed_comp_dm_2 %>% summarize(count = n()) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 41
      Primary keys: 4
      Foreign keys: 4
    Code
      zoomed_comp_dm %>% select(time_hour, wind_dir) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 40
      Primary keys: 3
      Foreign keys: 3
    Code
      zoomed_comp_dm %>% select(time_hour, origin, wind_dir) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 41
      Primary keys: 4
      Foreign keys: 4
    Code
      zoomed_comp_dm %>% rename(th = time_hour, wd = wind_dir) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 53
      Primary keys: 4
      Foreign keys: 4
    Code
      zoomed_comp_dm %>% distinct(origin, wind_dir) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 40
      Primary keys: 3
      Foreign keys: 3
    Code
      zoomed_comp_dm %>% distinct(origin, wind_dir, time_hour) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 41
      Primary keys: 4
      Foreign keys: 4
    Code
      zoomed_comp_dm %>% filter(pressure < 1020) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 53
      Primary keys: 4
      Foreign keys: 4
    Code
      zoomed_comp_dm %>% pull(origin) %>% unique()
    Output
      [1] "EWR" "JFK" "LGA"
    Code
      zoomed_comp_dm %>% slice(c(1:3, 5:3))
    Message
      Keeping PK column, but `slice.dm_zoomed()` can potentially damage the uniqueness of PK columns (duplicated indices). Set argument `.keep_pk` to `TRUE` or `FALSE` to ensure the behavior you intended.
    Output
      # Zoomed table: weather
      # A tibble:     6 x 15
        origin  year month   day  hour  temp  dewp humid wind_dir wind_speed wind_gust
        <chr>  <int> <int> <int> <int> <dbl> <dbl> <dbl>    <dbl>      <dbl>     <dbl>
      1 EWR     2013     1    10     0  41    32    70.1      230       8.06        NA
      2 EWR     2013     1    10     1  39.0  30.0  69.9      210       9.21        NA
      3 EWR     2013     1    10     2  39.0  28.9  66.8      230       6.90        NA
      4 EWR     2013     1    10     4  41    26.1  55.0      320       6.90        NA
      5 EWR     2013     1    10     3  39.9  27.0  59.5      270       5.75        NA
      6 EWR     2013     1    10     2  39.0  28.9  66.8      230       6.90        NA
      # i 4 more variables: precip <dbl>, pressure <dbl>, visib <dbl>,
      #   time_hour <chr>
    Code
      zoomed_comp_dm %>% slice(c(1:3, 5:3), .keep_pk = TRUE) %>% col_tracker_zoomed()
    Output
            origin         year        month          day         hour         temp 
          "origin"       "year"      "month"        "day"       "hour"       "temp" 
              dewp        humid     wind_dir   wind_speed    wind_gust       precip 
            "dewp"      "humid"   "wind_dir" "wind_speed"  "wind_gust"     "precip" 
          pressure        visib    time_hour 
        "pressure"      "visib"  "time_hour" 
    Code
      zoomed_comp_dm %>% left_join(flights, multiple = "all") %>% nrow()
    Message
      Renaming ambiguous columns: %>%
        dm_rename(weather, year.weather = year) %>%
        dm_rename(weather, month.weather = month) %>%
        dm_rename(weather, day.weather = day) %>%
        dm_rename(weather, hour.weather = hour) %>%
        dm_rename(flights, year.flights = year) %>%
        dm_rename(flights, month.flights = month) %>%
        dm_rename(flights, day.flights = day) %>%
        dm_rename(flights, hour.flights = hour)
    Output
      [1] 1800
    Code
      zoomed_comp_dm %>% right_join(flights, multiple = "all") %>% nrow()
    Message
      Renaming ambiguous columns: %>%
        dm_rename(weather, year.weather = year) %>%
        dm_rename(weather, month.weather = month) %>%
        dm_rename(weather, day.weather = day) %>%
        dm_rename(weather, hour.weather = hour) %>%
        dm_rename(flights, year.flights = year) %>%
        dm_rename(flights, month.flights = month) %>%
        dm_rename(flights, day.flights = day) %>%
        dm_rename(flights, hour.flights = hour)
    Output
      [1] 1761
    Code
      zoomed_comp_dm %>% inner_join(flights, multiple = "all") %>% nrow()
    Message
      Renaming ambiguous columns: %>%
        dm_rename(weather, year.weather = year) %>%
        dm_rename(weather, month.weather = month) %>%
        dm_rename(weather, day.weather = day) %>%
        dm_rename(weather, hour.weather = hour) %>%
        dm_rename(flights, year.flights = year) %>%
        dm_rename(flights, month.flights = month) %>%
        dm_rename(flights, day.flights = day) %>%
        dm_rename(flights, hour.flights = hour)
    Output
      [1] 1761
    Code
      zoomed_comp_dm %>% full_join(flights, multiple = "all") %>% nrow()
    Message
      Renaming ambiguous columns: %>%
        dm_rename(weather, year.weather = year) %>%
        dm_rename(weather, month.weather = month) %>%
        dm_rename(weather, day.weather = day) %>%
        dm_rename(weather, hour.weather = hour) %>%
        dm_rename(flights, year.flights = year) %>%
        dm_rename(flights, month.flights = month) %>%
        dm_rename(flights, day.flights = day) %>%
        dm_rename(flights, hour.flights = hour)
    Output
      [1] 1800
    Code
      zoomed_comp_dm %>% semi_join(flights) %>% nrow()
    Output
      [1] 105
    Code
      zoomed_comp_dm %>% anti_join(flights) %>% nrow()
    Output
      [1] 39
    Code
      zoomed_comp_dm %>% nest_join(flights) %>% nrow()
    Output
      [1] 144

# .by key tracking works with zoomed summarise()

    Code
      dm_zoom_to(dm_for_filter(), tf_2) %>% summarize(d_mean = mean(d), .by = c(c, e,
        e1)) %>% dm_insert_zoomed("new_tbl") %>% get_all_keys()
    Output
      $pks
      # A tibble: 7 x 3
        table   pk_col autoincrement
        <chr>   <keys> <lgl>        
      1 tf_1    a      TRUE         
      2 tf_2    c      FALSE        
      3 tf_3    f, f1  FALSE        
      4 tf_4    h      FALSE        
      5 tf_5    k      FALSE        
      6 tf_6    o      FALSE        
      7 new_tbl c      FALSE        
      
      $fks
      # A tibble: 6 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 new_tbl     e, e1         tf_3         f, f1           no_action
      5 tf_5        l             tf_4         h               cascade  
      6 tf_5        m             tf_6         n               no_action
      
    Code
      dm_zoom_to(dm_for_filter(), tf_3) %>% summarize(g_list = list(g), .by = g) %>%
        dm_insert_zoomed("new_tbl") %>% get_all_keys()
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      

# snapshot test

    Code
      dm <- dm_nycflights13()
      dm_deconstruct(dm)
    Message
      airlines <- pull_tbl(dm, "airlines", keyed = TRUE)
      airports <- pull_tbl(dm, "airports", keyed = TRUE)
      flights <- pull_tbl(dm, "flights", keyed = TRUE)
      planes <- pull_tbl(dm, "planes", keyed = TRUE)
      weather <- pull_tbl(dm, "weather", keyed = TRUE)

# non-syntactic names

    Code
      dm <- dm(`if` = tibble(a = 1), `a b` = tibble(b = 1))
      dm_deconstruct(dm)
    Message
      `if` <- pull_tbl(dm, "if", keyed = TRUE)
      `a b` <- pull_tbl(dm, "a b", keyed = TRUE)

# dm_select_tbl() remembers all FKs

    Code
      dm_nycflights_small() %>% dm_add_fk(flights, origin, airports) %>%
        dm_select_tbl(airports, flights) %>% dm_paste()
    Message
      dm::dm(
        airports,
        flights,
      ) %>%
        dm::dm_add_pk(airports, faa) %>%
        dm::dm_add_fk(flights, dest, airports) %>%
        dm::dm_add_fk(flights, origin, airports)

# check_key() API

    Code
      check_key(tibble(a = 1), a)
      check_key(.data = tibble(a = 1), a)
    Condition
      Warning:
      The `.data` argument of `check_key()` is deprecated as of dm 1.0.0.
      i Please use the `x` argument instead.
    Code
      check_key(a, .data = tibble(a = 1))
    Condition
      Warning:
      The `.data` argument of `check_key()` is deprecated as of dm 1.0.0.
      i Please use the `x` argument instead.

# output for legacy API

    Code
      check_subset(data_mcard_1(), a, data_mcard_2(), a)
    Condition
      Warning:
      The `c1` argument of `check_subset()` is deprecated as of dm 1.0.0.
      i Please use the `x_select` argument instead.
      i Use `y_select` instead of `c2`, and `x` and `y` instead of `t1` and `t2`.
      i Using `by_position = TRUE` for compatibility.

# output

    Code
      check_subset(data_mcard_1(), data_mcard_2(), x_select = c(x = a))
    Condition
      Error in `check_subset()`:
      ! `by_position = FALSE` or `by_position = NULL` require column names in `x` to match those in `y`.

---

    Code
      check_subset(data_mcard_2(), data_mcard_1(), x_select = a)
    Output
      # A tibble: 1 x 1
            a
        <dbl>
      1     3
    Condition
      Error in `check_subset()`:
      ! Column (`a`) of table data_mcard_2() contains values (see examples above) that are not present in column (`a`) of table data_mcard_1().

# output for compound keys

    Code
      check_subset(data_mcard_2(), data_mcard_1(), x_select = c(a, b))
    Output
      # A tibble: 3 x 2
            a     b
        <dbl> <dbl>
      1     1     4
      2     2     5
      3     3     6
    Condition
      Error in `check_subset()`:
      ! Columns (`a`, `b`) of table data_mcard_2() contain values (see examples above) that are not present in columns (`a`, `b`) of table data_mcard_1().

# check_set_equality() checks properly if 2 sets of values are equal?

    Code
      check_set_equality(data_mcard_1(), data_mcard_2(), x_select = c(a, c))
    Output
      # A tibble: 2 x 2
            a     c
        <dbl> <dbl>
      1     1     5
      2     2     6
      # A tibble: 2 x 2
            a     c
        <dbl> <dbl>
      1     2     8
      2     3     9
    Condition
      Error in `check_set_equality()`:
      ! Columns (`a`, `c`) of table data_mcard_1() contain values (see examples above) that are not present in columns (`a`, `c`) of table data_mcard_2().
      Columns (`a`, `c`) of table data_mcard_2() contain values (see examples above) that are not present in columns (`a`, `c`) of table data_mcard_1().

# dm_examine_cardinalities() works

    Code
      dm_examine_cardinalities(dm_for_card())
    Output
      * FK: dc_2$(`a`, `b`) -> dc_1$(`a`, `b`): injective mapping (child: 0 or 1 -> parent: 1)
      * FK: dc_3$(`a`, `b`) -> dc_1$(`a`, `b`): surjective mapping (child: 1 to n -> parent: 1)
      * FK: dc_5$(`b`, `a`) -> dc_1$(`b`, `a`): bijective mapping (child: 1 -> parent: 1)
      * FK: dc_3$(`b`, `a`) -> dc_4$(`b`, `a`): generic mapping (child: 0 to n -> parent: 1)
      * FK: dc_6$(`c`) -> dc_1$(`a`): Column (`c`) of table `dc_6` not a subset of column (`a`) of table `dc_1`.
      * FK: dc_4$(`b`, `a`) -> dc_3$(`b`, `a`): Columns (`b`, `a`) not a unique key of `dc_3`.
    Message
      ! Not all FK constraints satisfied, call `dm_examine_constraints()` for details.
    Code
      dm_examine_cardinalities(dm_for_card()) %>% as_tibble()
    Output
      # A tibble: 6 x 5
        child_table child_fk_cols parent_table parent_key_cols cardinality            
        <chr>       <keys>        <chr>        <keys>          <chr>                  
      1 dc_2        a, b          dc_1         a, b            injective mapping (chi~
      2 dc_3        a, b          dc_1         a, b            surjective mapping (ch~
      3 dc_5        b, a          dc_1         b, a            bijective mapping (chi~
      4 dc_6        c             dc_1         a               Column (`c`) of table ~
      5 dc_4        b, a          dc_3         b, a            Columns (`b`, `a`) not~
      6 dc_3        b, a          dc_4         b, a            generic mapping (child~
    Code
      dm_for_card() %>% dm_rm_fk(dc_6, c, dc_1, a) %>% dm_rm_fk(dc_4, c(b, a), dc_3,
      c(b, a)) %>% dm_examine_cardinalities()
    Output
      * FK: dc_2$(`a`, `b`) -> dc_1$(`a`, `b`): injective mapping (child: 0 or 1 -> parent: 1)
      * FK: dc_3$(`a`, `b`) -> dc_1$(`a`, `b`): surjective mapping (child: 1 to n -> parent: 1)
      * FK: dc_5$(`b`, `a`) -> dc_1$(`b`, `a`): bijective mapping (child: 1 -> parent: 1)
      * FK: dc_3$(`b`, `a`) -> dc_4$(`b`, `a`): generic mapping (child: 0 to n -> parent: 1)
    Code
      dm_examine_cardinalities(dm())
    Message
      ! No FKs available in `dm`.

# `dm_examine_cardinalities()` API

    Code
      dm_examine_cardinalities(dm_test_obj(), progress = FALSE)
    Condition
      Warning:
      The `progress` argument of `dm_examine_cardinalities()` is deprecated as of dm 1.0.0.
      i Please use the `.progress` argument instead.
    Message
      ! No FKs available in `dm`.
    Code
      dm_examine_cardinalities(dm = dm_test_obj())
    Condition
      Warning:
      The `dm` argument of `dm_examine_cardinalities()` is deprecated as of dm 1.0.0.
      i Please use the `.dm` argument instead.
    Message
      ! No FKs available in `dm`.

# `dm_examine_cardinalities()` API (2)

    Code
      dm_examine_cardinalities(dm_test_obj(), foo = "bar")
    Condition
      Error in `dm_examine_cardinalities()`:
      ! `...` must be empty.
      x Problematic argument:
      * foo = "bar"

# `json_pack()` works

    Code
      df <- tibble::tibble(x1 = 1:3, x2 = 4:6, x3 = 7:9, y = 1:3)
      packed <- json_pack(df, x = c(x1, x2, x3), y = y)
      packed
    Output
      # A tibble: 3 x 2
        x                              y          
        <chr>                          <chr>      
      1 "{\"x1\":1,\"x2\":4,\"x3\":7}" "{\"y\":1}"
      2 "{\"x1\":2,\"x2\":5,\"x3\":8}" "{\"y\":2}"
      3 "{\"x1\":3,\"x2\":6,\"x3\":9}" "{\"y\":3}"

# `json_unpack()` fails without cols argument

    Code
      json_unpack(df)
    Condition
      Error in `json_unpack()`:
      ! The `cols` argument must be provided.

# autoincrement produces valid R code

    Code
      dm
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `x`, `y`
      Columns: 5
      Primary keys: 2
      Foreign keys: 1

# autoincrement produces valid SQL queries and R code - RSQLite

    Code
      df$sql_table
    Output
      <SQL> CREATE TEMPORARY TABLE `x` (
        `x_id` INTEGER,
        `x_data` TEXT,
        PRIMARY KEY (`x_id`)
      )
      <SQL> CREATE TEMPORARY TABLE `y` (
        `y_id` INTEGER,
        `x_id` INTEGER,
        `y_data` TEXT,
        PRIMARY KEY (`y_id`),
        FOREIGN KEY (`x_id`) REFERENCES `x` (`x_id`)
      )

---

    Code
      dm_paste(dm)
    Message
      dm::dm(
        x,
        y,
      ) %>%
        dm::dm_add_pk(x, x_id, autoincrement = TRUE) %>%
        dm::dm_add_pk(y, y_id) %>%
        dm::dm_add_fk(y, x_id, x)

# decompose_table() decomposes tables nicely on chosen source

    Code
      out$parent_table
    Output
      # A tibble: 2 x 4
        aef_id     a e     f    
         <int> <int> <chr> <lgl>
      1      1     1 c     TRUE 
      2      2     2 b     FALSE
    Code
      list_of_data_ts_parent_and_child()$parent_table
    Output
      # A tibble: 2 x 4
        aef_id     a e     f    
         <int> <int> <chr> <lgl>
      1      1     1 c     TRUE 
      2      2     2 b     FALSE
    Code
      out$child_table
    Output
      # A tibble: 3 x 4
            b     c d     aef_id
        <dbl> <int> <chr>  <int>
      1   1.1     5 a          1
      2   4.2     6 b          2
      3   1.1     7 c          1
    Code
      list_of_data_ts_parent_and_child()$child_table
    Output
      # A tibble: 3 x 4
            b     c d     aef_id
        <dbl> <int> <chr>  <int>
      1   1.1     5 a          1
      2   4.2     6 b          2
      3   1.1     7 c          1

# dumma

    Code
      # dummy

# dm_rows_insert()

    Code
      print(dm_nrow(flights_sqlite))
    Output
      airlines airports  flights   planes  weather 
            15       86        0      945        0 
    Code
      flights_hour10 <- dm_nycflights13() %>% dm_select_tbl(flights, weather) %>%
        dm_zoom_to(flights) %>% filter(month == 1, day == 10, hour == 10) %>%
        dm_update_zoomed() %>% dm_zoom_to(weather) %>% filter(month == 1, day == 10,
      hour == 10) %>% dm_update_zoomed()
      print(dm_nrow(flights_hour10))
    Output
      flights weather 
           43       3 
    Code
      flights_hour10_sqlite <- copy_dm_to(sqlite, flights_hour10)
      out <- dm_rows_append(flights_sqlite, flights_hour10_sqlite)
    Message
      Result is returned as a dm object with lazy tables. Use `in_place = FALSE` to mute this message, or `in_place = TRUE` to write to the underlying tables.
    Code
      print(dm_nrow(flights_sqlite))
    Output
      airlines airports  flights   planes  weather 
            15       86        0      945        0 
    Code
      dm_rows_append(flights_sqlite, flights_hour10_sqlite, in_place = TRUE)
      print(dm_nrow(flights_sqlite))
    Output
      airlines airports  flights   planes  weather 
            15       86       43      945        3 
    Code
      flights_hour11 <- dm_nycflights13() %>% dm_select_tbl(flights, weather) %>%
        dm_zoom_to(flights) %>% filter(month == 1, day == 10, hour == 11) %>%
        dm_update_zoomed() %>% dm_zoom_to(weather) %>% filter(month == 1, day == 10,
      hour == 11) %>% dm_update_zoomed()
      flights_hour11_sqlite <- copy_dm_to(sqlite, flights_hour11)
      flights_new <- dm_rows_append(flights_sqlite, flights_hour11_sqlite, in_place = FALSE)
      print(dm_nrow(flights_new))
    Output
      airlines airports  flights   planes  weather 
            15       86       88      945        6 
    Code
      print(dm_nrow(flights_sqlite))
    Output
      airlines airports  flights   planes  weather 
            15       86       43      945        3 
    Code
      flights_new %>% dm_examine_constraints()
    Message
      ! Unsatisfied constraints:
    Output
      * Table `flights`: foreign key `tailnum` into table `planes`: values of `flights$tailnum` not in `planes$tailnum`: N0EGMQ (1), N3BCAA (1), N3CCAA (1), N3CFAA (1), N3EHAA (1), ...
    Code
      dm_rows_append(flights_sqlite, flights_hour11_sqlite, in_place = TRUE)
      print(dm_nrow(flights_sqlite))
    Output
      airlines airports  flights   planes  weather 
            15       86       88      945        6 
    Code
      DBI::dbDisconnect(sqlite)

# dm_rows_update()

    Code
      dm_copy %>% pull_tbl(tf_2) %>% arrange_all()
    Output
            d c        e        e1
        <int> <chr>    <chr> <int>
      1     2 elephant D         4
      2     3 lion     E         5
      3     4 seal     F         6
      4     5 worm     G         7
      5     6 dog      E         5
      6     7 cat      F         6
    Code
      dm_copy %>% dm_rows_update(dm_update_copy) %>% pull_tbl(tf_2) %>% arrange_all()
    Message
      Result is returned as a dm object with lazy tables. Use `in_place = FALSE` to mute this message, or `in_place = TRUE` to write to the underlying tables.
    Output
            d c        e        e1
        <int> <chr>    <chr> <int>
      1     2 elephant D         4
      2     3 lion     E         5
      3     4 seal     F         6
      4     5 worm     G         7
      5     6 dog      E         5
      6     7 cat      F         6
    Code
      dm_copy %>% pull_tbl(tf_2) %>% arrange_all()
    Output
            d c        e        e1
        <int> <chr>    <chr> <int>
      1     2 elephant D         4
      2     3 lion     E         5
      3     4 seal     F         6
      4     5 worm     G         7
      5     6 dog      E         5
      6     7 cat      F         6
    Code
      dm_copy %>% dm_rows_update(dm_update_copy, in_place = FALSE) %>% pull_tbl(tf_2) %>%
        arrange_all()
    Output
            d c        e        e1
        <int> <chr>    <chr> <int>
      1     2 elephant D         4
      2     3 lion     E         5
      3     4 seal     F         6
      4     5 worm     G         7
      5     6 dog      E         5
      6     7 cat      F         6
    Code
      dm_copy %>% dm_get_tables() %>% map(arrange_all)
    Output
      $tf_1
             a b    
         <int> <chr>
       1     1 A    
       2     2 B    
       3     3 C    
       4     4 D    
       5     5 E    
       6     6 F    
       7     7 G    
       8     8 H    
       9     9 I    
      10    10 J    
      
      $tf_2
            d c        e        e1
        <int> <chr>    <chr> <int>
      1     2 elephant D         4
      2     3 lion     E         5
      3     4 seal     F         6
      4     5 worm     G         7
      5     6 dog      E         5
      6     7 cat      F         6
      
      $tf_3
         f        f1 g    
         <chr> <int> <chr>
       1 C         2 one  
       2 C         3 two  
       3 D         4 three
       4 E         5 four 
       5 F         6 five 
       6 G         7 six  
       7 H         7 seven
       8 I         7 eight
       9 J        10 nine 
      10 K        11 ten  
      
      $tf_4
        i     h     j        j1
        <chr> <chr> <chr> <int>
      1 five  c     E         5
      2 four  b     D         4
      3 seven e     F         6
      4 six   d     F         6
      5 three a     C         3
      
      $tf_5
        l     m             ww     k
        <chr> <chr>      <int> <int>
      1 b     house          2     1
      2 c     tree           2     2
      3 d     streetlamp     2     3
      4 e     streetlamp     2     4
      
      $tf_6
           zz n          o    
        <int> <chr>      <chr>
      1     1 garden     i    
      2     1 hill       g    
      3     1 house      e    
      4     1 streetlamp h    
      5     1 tree       f    
      
    Code
      dm_copy %>% dm_rows_update(dm_update_copy, in_place = TRUE)
      dm_copy %>% dm_get_tables() %>% map(arrange_all)
    Output
      $tf_1
             a b    
         <int> <chr>
       1     1 A    
       2     2 q    
       3     3 C    
       4     4 D    
       5     5 E    
       6     6 F    
       7     7 G    
       8     8 H    
       9     9 I    
      10    10 J    
      
      $tf_2
            d c        e        e1
        <int> <chr>    <chr> <int>
      1     2 elephant D         4
      2     3 lion     E         5
      3     4 seal     F         6
      4     5 worm     G         7
      5     6 dog      E         5
      6     7 cat      F         6
      
      $tf_3
         f        f1 g    
         <chr> <int> <chr>
       1 C         2 one  
       2 C         3 two  
       3 D         4 three
       4 E         5 four 
       5 F         6 five 
       6 G         7 six  
       7 H         7 seven
       8 I         7 eight
       9 J        10 nine 
      10 K        11 ten  
      
      $tf_4
        i      h     j        j1
        <chr>  <chr> <chr> <int>
      1 five   c     E         5
      2 four   b     D         4
      3 sieben e     F         6
      4 six    d     F         6
      5 three  a     C         3
      
      $tf_5
        l     m             ww     k
        <chr> <chr>      <int> <int>
      1 b     house          2     1
      2 c     tree           2     2
      3 d     streetlamp     3     3
      4 e     streetlamp     2     4
      
      $tf_6
           zz n          o    
        <int> <chr>      <chr>
      1     1 garden     i    
      2     1 hill       g    
      3     1 house      e    
      4     1 streetlamp h    
      5     1 tree       f    
      

# dm_rows_truncate()

    Code
      dm_copy %>% pull_tbl(tf_2) %>% arrange_all()
    Output
        c            d e        e1
        <chr>    <int> <chr> <int>
      1 cat          7 F         6
      2 dog          6 E         5
      3 elephant     2 D         4
      4 lion         3 E         5
      5 seal         4 F         6
      6 worm         5 G         7
    Code
      dm_copy %>% dm_rows_truncate(dm_truncate_copy) %>% pull_tbl(tf_2) %>%
        arrange_all()
    Condition
      Warning:
      `dm_rows_truncate()` was deprecated in dm 1.0.0.
    Message
      Result is returned as a dm object with lazy tables. Use `in_place = FALSE` to mute this message, or `in_place = TRUE` to write to the underlying tables.
    Output
      # i 4 variables: c <chr>, d <int>, e <chr>, e1 <int>
    Code
      dm_copy %>% pull_tbl(tf_2) %>% arrange_all()
    Output
        c            d e        e1
        <chr>    <int> <chr> <int>
      1 cat          7 F         6
      2 dog          6 E         5
      3 elephant     2 D         4
      4 lion         3 E         5
      5 seal         4 F         6
      6 worm         5 G         7
    Code
      dm_copy %>% dm_rows_truncate(dm_truncate_copy, in_place = FALSE) %>% pull_tbl(
        tf_2) %>% arrange_all()
    Condition
      Warning:
      `dm_rows_truncate()` was deprecated in dm 1.0.0.
    Output
      # i 4 variables: c <chr>, d <int>, e <chr>, e1 <int>
    Code
      dm_copy %>% dm_get_tables() %>% map(arrange_all)
    Output
      $tf_1
             a b    
         <int> <chr>
       1     1 A    
       2     2 B    
       3     3 C    
       4     4 D    
       5     5 E    
       6     6 F    
       7     7 G    
       8     8 H    
       9     9 I    
      10    10 J    
      
      $tf_2
        c            d e        e1
        <chr>    <int> <chr> <int>
      1 cat          7 F         6
      2 dog          6 E         5
      3 elephant     2 D         4
      4 lion         3 E         5
      5 seal         4 F         6
      6 worm         5 G         7
      
      $tf_3
         f        f1 g    
         <chr> <int> <chr>
       1 C         2 one  
       2 C         3 two  
       3 D         4 three
       4 E         5 four 
       5 F         6 five 
       6 G         7 six  
       7 H         7 seven
       8 I         7 eight
       9 J        10 nine 
      10 K        11 ten  
      
      $tf_4
        h     i     j        j1
        <chr> <chr> <chr> <int>
      1 a     three C         3
      2 b     four  D         4
      3 c     five  E         5
      4 d     six   F         6
      5 e     seven F         6
      
      $tf_5
           ww     k l     m         
        <int> <int> <chr> <chr>     
      1     2     1 b     house     
      2     2     2 c     tree      
      3     2     3 d     streetlamp
      4     2     4 e     streetlamp
      
      $tf_6
           zz n          o    
        <int> <chr>      <chr>
      1     1 garden     i    
      2     1 hill       g    
      3     1 house      e    
      4     1 streetlamp h    
      5     1 tree       f    
      
    Code
      dm_copy %>% dm_rows_truncate(dm_truncate_copy, in_place = TRUE)
    Condition
      Warning:
      `dm_rows_truncate()` was deprecated in dm 1.0.0.
      Warning:
      `sql_rows_truncate()` was deprecated in dm 1.0.0.
      Warning:
      `sql_rows_truncate()` was deprecated in dm 1.0.0.
    Code
      dm_copy %>% dm_get_tables() %>% map(arrange_all)
    Output
      $tf_1
             a b    
         <int> <chr>
       1     1 A    
       2     2 B    
       3     3 C    
       4     4 D    
       5     5 E    
       6     6 F    
       7     7 G    
       8     8 H    
       9     9 I    
      10    10 J    
      
      $tf_2
      # i 4 variables: c <chr>, d <int>, e <chr>, e1 <int>
      
      $tf_3
         f        f1 g    
         <chr> <int> <chr>
       1 C         2 one  
       2 C         3 two  
       3 D         4 three
       4 E         5 four 
       5 F         6 five 
       6 G         7 six  
       7 H         7 seven
       8 I         7 eight
       9 J        10 nine 
      10 K        11 ten  
      
      $tf_4
        h     i     j        j1
        <chr> <chr> <chr> <int>
      1 a     three C         3
      2 b     four  D         4
      3 c     five  E         5
      4 d     six   F         6
      5 e     seven F         6
      
      $tf_5
      # i 4 variables: ww <int>, k <int>, l <chr>, m <chr>
      
      $tf_6
           zz n          o    
        <int> <chr>      <chr>
      1     1 garden     i    
      2     1 hill       g    
      3     1 house      e    
      4     1 streetlamp h    
      5     1 tree       f    
      

# dm_rows_update() requires primary keys

    Code
      dm_rows_update(x, y, in_place = FALSE)
    Condition
      Error in `dm_rows_update()`:
      ! `dm_rows_update()` requires the <dm> object to have primary keys for all target tables.

# cdm_filter() behaves correctly

    Code
      dm_for_filter_simple() %>% dm_filter(tf_1, a > 3, a < 8) %>% cdm_apply_filters() %>%
        dm_get_tables()
    Output
      $tf_1
      # A tibble: 4 x 2
            a b    
        <int> <chr>
      1     4 D    
      2     5 E    
      3     6 F    
      4     7 G    
      
      $tf_2
      # A tibble: 4 x 3
        c         d e    
        <chr> <int> <chr>
      1 seal      4 F    
      2 worm      5 G    
      3 dog       6 E    
      4 cat       7 F    
      
      $tf_3
      # A tibble: 3 x 2
        f     g    
        <chr> <chr>
      1 E     four 
      2 F     five 
      3 G     six  
      
      $tf_4
      # A tibble: 3 x 4
        h     i     j        j1
        <chr> <chr> <chr> <int>
      1 c     five  E         5
      2 d     six   F         6
      3 e     seven F         6
      
      $tf_5
      # A tibble: 3 x 4
           ww     k l     m         
        <int> <int> <chr> <chr>     
      1     2     2 c     tree      
      2     2     3 d     streetlamp
      3     2     4 e     streetlamp
      
      $tf_6
      # A tibble: 2 x 3
           zz n          o    
        <int> <chr>      <chr>
      1     1 tree       f    
      2     1 streetlamp h    
      

# dm_squash_to_tbl() deprecation warning is correct

    Code
      dm_squash_to_tbl(dm_for_flatten(), fact)
    Condition
      Warning:
      `dm_squash_to_tbl()` was deprecated in dm 1.0.0.
      i Please use `.recursive = TRUE` in `dm_flatten_to_tbl()` instead.
    Message
      Renaming ambiguous columns: %>%
        dm_rename(fact, fact.something = something) %>%
        dm_rename(dim_1, dim_1.something = something) %>%
        dm_rename(dim_2, dim_2.something = something) %>%
        dm_rename(dim_3, dim_3.something = something) %>%
        dm_rename(dim_4, dim_4.something = something)
    Output
      # A tibble: 10 x 11
         fact     dim_1_key_1 dim_1_key_2 dim_2_key dim_3_key dim_4_key fact.something
         <chr>          <int> <chr>       <chr>     <chr>         <int>          <int>
       1 acorn             14 N           c         X                 7              1
       2 blubber           13 M           d         W                 8              2
       3 cindere~          12 L           e         V                 9              3
       4 depth             11 K           f         U                10              4
       5 elysium           10 J           g         T                11              5
       6 fantasy            9 I           h         S                12              6
       7 gorgeous           8 H           i         R                13              7
       8 halo               7 G           j         Q                14              8
       9 ill-adv~           6 F           k         P                15              9
      10 jitter             5 E           l         O                16             10
      # i 4 more variables: dim_1.something <chr>, dim_2.something <chr>,
      #   dim_3.something <int>, dim_4.something <int>

# dummy

    Code
      # dummy

# dm_sql()

    Code
      dm_for_filter_df() %>% dm_sql(my_test_con())
    Output
      $pre
      $pre$tf_1
      <SQL> CREATE TEMPORARY TABLE `tf_1` (
        `a` INT AUTO_INCREMENT,
        `b` VARCHAR(255),
        PRIMARY KEY (`a`)
      )
      
      $pre$tf_2
      <SQL> CREATE TEMPORARY TABLE `tf_2` (
        `c` VARCHAR(255),
        `d` INT,
        `e` VARCHAR(255),
        `e1` INT,
        PRIMARY KEY (`c`)
      )
      
      $pre$tf_3
      <SQL> CREATE TEMPORARY TABLE `tf_3` (
        `f` VARCHAR(255),
        `f1` INT,
        `g` VARCHAR(255),
        PRIMARY KEY (`f`, `f1`)
      )
      
      $pre$tf_4
      <SQL> CREATE TEMPORARY TABLE `tf_4` (
        `h` VARCHAR(255),
        `i` VARCHAR(255),
        `j` VARCHAR(255),
        `j1` INT,
        PRIMARY KEY (`h`)
      )
      
      $pre$tf_5
      <SQL> CREATE TEMPORARY TABLE `tf_5` (
        `ww` INT,
        `k` INT,
        `l` VARCHAR(255),
        `m` VARCHAR(255),
        PRIMARY KEY (`k`)
      )
      
      $pre$tf_6
      <SQL> CREATE TEMPORARY TABLE `tf_6` (
        `zz` INT,
        `n` VARCHAR(255),
        `o` VARCHAR(255),
        PRIMARY KEY (`o`)
      )
      
      
      $load
      $load$tf_1
      <SQL> INSERT INTO `tf_1` (`a`, `b`)
      SELECT `a`, CAST(`b` AS CHAR) AS `b`
      FROM (
        (
          SELECT NULL AS `a`, NULL AS `b`
          WHERE (0 = 1)
        )
        UNION ALL
        (
        VALUES
          (1, 'A'),
          (2, 'B'),
          (3, 'C'),
          (4, 'D'),
          (5, 'E'),
          (6, 'F'),
          (7, 'G'),
          (8, 'H'),
          (9, 'I'),
          (10, 'J')
        )
      ) `values_table`
      
      $load$tf_2
      <SQL> INSERT INTO `tf_2` (`c`, `d`, `e`, `e1`)
      SELECT
        CAST(`c` AS CHAR) AS `c`,
        `d`,
        CAST(`e` AS CHAR) AS `e`,
        `e1`
      FROM (
        (
          SELECT NULL AS `c`, NULL AS `d`, NULL AS `e`, NULL AS `e1`
          WHERE (0 = 1)
        )
        UNION ALL
        (
        VALUES
          ('cat', 7, 'F', 6),
          ('dog', 6, 'E', 5),
          ('elephant', 2, 'D', 4),
          ('lion', 3, 'E', 5),
          ('seal', 4, 'F', 6),
          ('worm', 5, 'G', 7)
        )
      ) `values_table`
      
      $load$tf_3
      <SQL> INSERT INTO `tf_3` (`f`, `f1`, `g`)
      SELECT
        CAST(`f` AS CHAR) AS `f`,
        `f1`,
        CAST(`g` AS CHAR) AS `g`
      FROM (
        (
          SELECT NULL AS `f`, NULL AS `f1`, NULL AS `g`
          WHERE (0 = 1)
        )
        UNION ALL
        (
        VALUES
          ('C', 2, 'one'),
          ('C', 3, 'two'),
          ('D', 4, 'three'),
          ('E', 5, 'four'),
          ('F', 6, 'five'),
          ('G', 7, 'six'),
          ('H', 7, 'seven'),
          ('I', 7, 'eight'),
          ('J', 10, 'nine'),
          ('K', 11, 'ten')
        )
      ) `values_table`
      
      $load$tf_4
      <SQL> INSERT INTO `tf_4` (`h`, `i`, `j`, `j1`)
      SELECT
        CAST(`h` AS CHAR) AS `h`,
        CAST(`i` AS CHAR) AS `i`,
        CAST(`j` AS CHAR) AS `j`,
        `j1`
      FROM (
        (
          SELECT NULL AS `h`, NULL AS `i`, NULL AS `j`, NULL AS `j1`
          WHERE (0 = 1)
        )
        UNION ALL
        (
        VALUES
          ('a', 'three', 'C', 3),
          ('b', 'four', 'D', 4),
          ('c', 'five', 'E', 5),
          ('d', 'six', 'F', 6),
          ('e', 'seven', 'F', 6)
        )
      ) `values_table`
      
      $load$tf_5
      <SQL> INSERT INTO `tf_5` (`ww`, `k`, `l`, `m`)
      SELECT
        `ww`,
        `k`,
        CAST(`l` AS CHAR) AS `l`,
        CAST(`m` AS CHAR) AS `m`
      FROM (
        (
          SELECT NULL AS `ww`, NULL AS `k`, NULL AS `l`, NULL AS `m`
          WHERE (0 = 1)
        )
        UNION ALL
        (
        VALUES
          (2, 1, 'b', 'house'),
          (2, 2, 'c', 'tree'),
          (2, 3, 'd', 'streetlamp'),
          (2, 4, 'e', 'streetlamp')
        )
      ) `values_table`
      
      $load$tf_6
      <SQL> INSERT INTO `tf_6` (`zz`, `n`, `o`)
      SELECT
        `zz`,
        CAST(`n` AS CHAR) AS `n`,
        CAST(`o` AS CHAR) AS `o`
      FROM (
        (
          SELECT NULL AS `zz`, NULL AS `n`, NULL AS `o`
          WHERE (0 = 1)
        )
        UNION ALL
        (
        VALUES
          (1, 'garden', 'i'),
          (1, 'hill', 'g'),
          (1, 'house', 'e'),
          (1, 'streetlamp', 'h'),
          (1, 'tree', 'f')
        )
      ) `values_table`
      
      
      $post
      $post$uk
      $post$uk$tf_1
      <SQL> ALTER TABLE `tf_1` ADD UNIQUE (`a`)
      
      $post$uk$tf_2
      <SQL> ALTER TABLE `tf_2` ADD UNIQUE (`c`)
      
      $post$uk$tf_3
      <SQL> ALTER TABLE `tf_3` ADD UNIQUE (`f`, `f1`)
      <SQL> ALTER TABLE `tf_3` ADD UNIQUE (`g`)
      
      $post$uk$tf_4
      <SQL> ALTER TABLE `tf_4` ADD UNIQUE (`h`)
      
      $post$uk$tf_5
      <SQL> ALTER TABLE `tf_5` ADD UNIQUE (`k`)
      
      $post$uk$tf_6
      <SQL> ALTER TABLE `tf_6` ADD UNIQUE (`o`)
      <SQL> ALTER TABLE `tf_6` ADD UNIQUE (`n`)
      
      
      $post$fk
      named list()
      
      $post$indexes
      $post$indexes$tf_2
      <SQL> CREATE INDEX tf_2__d ON `tf_2` (`d`)
      <SQL> CREATE INDEX tf_2__e_e1 ON `tf_2` (`e`, `e1`)
      
      $post$indexes$tf_4
      <SQL> CREATE INDEX tf_4__j_j1 ON `tf_4` (`j`, `j1`)
      
      $post$indexes$tf_5
      <SQL> CREATE INDEX tf_5__l ON `tf_5` (`l`)
      <SQL> CREATE INDEX tf_5__m ON `tf_5` (`m`)
      
      
      

# dm_add_fk() works as intended?

    Code
      dm_test_obj() %>% dm_add_pk(dm_table_4, c) %>% dm_add_fk(dm_table_1, a,
        dm_table_4) %>% get_all_keys()
    Output
      $pks
      # A tibble: 1 x 3
        table      pk_col autoincrement
        <chr>      <keys> <lgl>        
      1 dm_table_4 c      FALSE        
      
      $fks
      # A tibble: 1 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 dm_table_1  a             dm_table_4   c               no_action
      

# dm_rm_fk() works with partial matching

    Code
      dm_for_filter() %>% dm_rm_fk(tf_5) %>% get_all_keys()
    Message
      Removing foreign keys: %>%
        dm_rm_fk(tf_5, l, tf_4) %>%
        dm_rm_fk(tf_5, m, tf_6, n))
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 3 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      
    Code
      dm_for_filter() %>% dm_rm_fk(columns = l) %>% get_all_keys()
    Message
      Removing foreign keys: %>%
        dm_rm_fk(tf_5, l, tf_4)
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 4 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        m             tf_6         n               no_action
      
    Code
      dm_for_filter() %>% dm_rm_fk(columns = c(e, e1)) %>% get_all_keys()
    Message
      Removing foreign keys: %>%
        dm_rm_fk(tf_2, c(e, e1), tf_3)
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 4 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_4        j, j1         tf_3         f, f1           no_action
      3 tf_5        l             tf_4         h               cascade  
      4 tf_5        m             tf_6         n               no_action
      
    Code
      dm_for_filter() %>% dm_rm_fk(ref_table = tf_3) %>% get_all_keys()
    Message
      Removing foreign keys: %>%
        dm_rm_fk(tf_2, c(e, e1), tf_3) %>%
        dm_rm_fk(tf_4, c(j, j1), tf_3)
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 3 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_5        l             tf_4         h               cascade  
      3 tf_5        m             tf_6         n               no_action
      
    Code
      dm_for_filter() %>% dm_rm_fk(ref_columns = c(f, f1)) %>% get_all_keys()
    Message
      Removing foreign keys: %>%
        dm_rm_fk(tf_2, c(e, e1), tf_3) %>%
        dm_rm_fk(tf_4, c(j, j1), tf_3)
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 3 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_5        l             tf_4         h               cascade  
      3 tf_5        m             tf_6         n               no_action
      
    Code
      dm_for_filter() %>% dm_rm_fk() %>% get_all_keys()
    Message
      Removing foreign keys: %>%
        dm_rm_fk(tf_2, d, tf_1) %>%
        dm_rm_fk(tf_2, c(e, e1), tf_3) %>%
        dm_rm_fk(tf_4, c(j, j1), tf_3) %>%
        dm_rm_fk(tf_5, l, tf_4) %>%
        dm_rm_fk(tf_5, m, tf_6, n))
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 0 x 5
      # i 5 variables: child_table <chr>, child_fk_cols <keys>, parent_table <chr>,
      #   parent_key_cols <keys>, on_delete <chr>
      

# dm_enum_fk_candidates() works as intended?

    Code
      dm_nycflights13() %>% dm_enum_fk_candidates(flights, airports) %>% mutate(why = if_else(
        why != "", "<reason>", ""))
    Output
      # A tibble: 19 x 3
         columns        candidate why       
         <keys>         <lgl>     <chr>     
       1 origin         TRUE      ""        
       2 year           FALSE     "<reason>"
       3 month          FALSE     "<reason>"
       4 day            FALSE     "<reason>"
       5 dep_time       FALSE     "<reason>"
       6 sched_dep_time FALSE     "<reason>"
       7 dep_delay      FALSE     "<reason>"
       8 arr_time       FALSE     "<reason>"
       9 sched_arr_time FALSE     "<reason>"
      10 arr_delay      FALSE     "<reason>"
      11 carrier        FALSE     "<reason>"
      12 flight         FALSE     "<reason>"
      13 tailnum        FALSE     "<reason>"
      14 dest           FALSE     "<reason>"
      15 air_time       FALSE     "<reason>"
      16 distance       FALSE     "<reason>"
      17 hour           FALSE     "<reason>"
      18 minute         FALSE     "<reason>"
      19 time_hour      FALSE     "<reason>"

# can add foreign key with cascade

    Code
      dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_add_fk(a, x, b, x, on_delete = "cascade") %>%
        dm_get_all_fks()
    Output
      # A tibble: 1 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 a           x             b            x               cascade  

# bogus arguments are rejected

    Code
      dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_add_fk(a, x, b, x, on_delete = "bogus")
    Condition
      Error in `dm_add_fk()`:
      ! `on_delete` must be one of "no_action" or "cascade", not "bogus".
    Code
      dm(a = tibble(x = 1), b = tibble(x = 1)) %>% dm_add_fk(a, x, b, x, on_delete = letters)
    Condition
      Error in `dm_add_fk()`:
      ! `on_delete` must be one of "no_action" or "cascade", not "a".

# dm_get_all_fks() with parent_table arg

    Code
      nyc_comp() %>% dm_get_all_fks(weather)
    Output
      # A tibble: 1 x 5
        child_table child_fk_cols     parent_table parent_key_cols   on_delete
        <chr>       <keys>            <chr>        <keys>            <chr>    
      1 flights     origin, time_hour weather      origin, time_hour no_action
    Code
      nyc_comp() %>% dm_get_all_fks(c("airlines", "weather"))
    Output
      # A tibble: 2 x 5
        child_table child_fk_cols     parent_table parent_key_cols   on_delete
        <chr>       <keys>            <chr>        <keys>            <chr>    
      1 flights     carrier           airlines     carrier           no_action
      2 flights     origin, time_hour weather      origin, time_hour no_action
    Code
      nyc_comp() %>% dm_get_all_fks(ends_with("ports"))
    Output
      # A tibble: 1 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 flights     dest          airports     faa             no_action
    Code
      nyc_comp() %>% dm_get_all_fks(everything())
    Output
      # A tibble: 4 x 5
        child_table child_fk_cols     parent_table parent_key_cols   on_delete
        <chr>       <keys>            <chr>        <keys>            <chr>    
      1 flights     carrier           airlines     carrier           no_action
      2 flights     dest              airports     faa               no_action
      3 flights     tailnum           planes       tailnum           no_action
      4 flights     origin, time_hour weather      origin, time_hour no_action

# dm_get_all_fks() with parent_table arg fails nicely

    Can't subset tables that don't exist.
    x Table `timetable` doesn't exist.

# error message for non-dm object

    Code
      dm_get_def(structure(list(table = "a"), class = "bogus"))
    Condition
      Error in `abort_is_not_dm()`:
      ! Required class <dm> but instead is <bogus>.

# can upgrade from v1

    Code
      def <- dm_get_def(dm_v1, quiet = TRUE)
      def <- dm_get_def(dm_v1)
    Message
      Upgrading dm object created with dm <= 0.2.1.
      Upgrading dm object created with dm <= 0.2.4.
      Upgrading dm object created with dm <= 0.3.0.
      Upgrading dm object created with dm <= 1.0.3.
    Code
      dm <- dm_from_def(def)
      dm_validate(dm)
      is_zoomed(dm)
    Output
      [1] FALSE

# can upgrade zoomed from v1

    Code
      def <- dm_get_def(dm_v1_zoomed, quiet = TRUE)
      def <- dm_get_def(dm_v1_zoomed)
    Message
      Upgrading dm object created with dm <= 0.2.1.
      Upgrading dm object created with dm <= 0.2.4.
      Upgrading dm object created with dm <= 0.3.0.
      Upgrading dm object created with dm <= 1.0.3.
    Code
      dm <- dm_from_def(def, zoomed = TRUE)
      dm_validate(dm)
      is_zoomed(dm)
    Output
      [1] TRUE

# can upgrade from v2

    Code
      def <- dm_get_def(dm_v2, quiet = TRUE)
      def <- dm_get_def(dm_v2)
    Message
      Upgrading dm object created with dm <= 0.2.4.
      Upgrading dm object created with dm <= 0.3.0.
      Upgrading dm object created with dm <= 1.0.3.
    Code
      dm <- dm_from_def(def)
      dm_validate(dm)
      is_zoomed(dm)
    Output
      [1] FALSE

# can upgrade zoomed from v2

    Code
      def <- dm_get_def(dm_v2_zoomed, quiet = TRUE)
      def <- dm_get_def(dm_v2_zoomed)
    Message
      Upgrading dm object created with dm <= 0.2.4.
      Upgrading dm object created with dm <= 0.3.0.
      Upgrading dm object created with dm <= 1.0.3.
    Code
      dm <- dm_from_def(def, zoomed = TRUE)
      dm_validate(dm)
      is_zoomed(dm)
    Output
      [1] TRUE

# can upgrade from v3

    Code
      def <- dm_get_def(dm_v3, quiet = TRUE)
      def <- dm_get_def(dm_v3)
    Message
      Upgrading dm object created with dm <= 0.3.0.
      Upgrading dm object created with dm <= 1.0.3.
    Code
      dm <- dm_from_def(def)
      dm_validate(dm)
      is_zoomed(dm)
    Output
      [1] FALSE
    Code
      dm_get_all_uks(dm_v3)
    Message
      Upgrading dm object created with dm <= 0.3.0.
      Upgrading dm object created with dm <= 1.0.3.
    Output
      # A tibble: 7 x 3
        table uk_col kind       
        <chr> <keys> <chr>      
      1 tf_1  a      PK         
      2 tf_2  c      PK         
      3 tf_3  f, f1  PK         
      4 tf_4  h      PK         
      5 tf_5  k      PK         
      6 tf_6  o      PK         
      7 tf_6  n      implicit UK
    Code
      dm_get_all_pks(dm_v3)
    Message
      Upgrading dm object created with dm <= 0.3.0.
      Upgrading dm object created with dm <= 1.0.3.
    Output
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      FALSE        
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
    Code
      dm_get_all_fks(dm_v3)
    Message
      Upgrading dm object created with dm <= 0.3.0.
      Upgrading dm object created with dm <= 1.0.3.
    Output
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action

# can upgrade zoomed from v3

    Code
      def <- dm_get_def(dm_v3_zoomed, quiet = TRUE)
      def <- dm_get_def(dm_v3_zoomed)
    Message
      Upgrading dm object created with dm <= 0.3.0.
      Upgrading dm object created with dm <= 1.0.3.
    Code
      dm <- dm_from_def(def, zoomed = TRUE)
      dm_validate(dm)
      is_zoomed(dm)
    Output
      [1] TRUE

# can upgrade to v4

    Code
      def <- dm_get_def(dm_v4, quiet = TRUE)
      def <- dm_get_def(dm_v4)
    Message
      Upgrading dm object created with dm <= 1.0.3.
    Code
      dm <- dm_from_def(def)
      dm_validate(dm)
      is_zoomed(dm)
    Output
      [1] FALSE
    Code
      dm_get_all_uks(dm_v4)
    Message
      Upgrading dm object created with dm <= 1.0.3.
    Output
      # A tibble: 7 x 3
        table uk_col kind       
        <chr> <keys> <chr>      
      1 tf_1  a      PK         
      2 tf_2  c      PK         
      3 tf_3  f, f1  PK         
      4 tf_4  h      PK         
      5 tf_5  k      PK         
      6 tf_6  o      PK         
      7 tf_6  n      implicit UK
    Code
      dm_get_all_pks(dm_v4)
    Message
      Upgrading dm object created with dm <= 1.0.3.
    Output
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      FALSE        
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
    Code
      dm_get_all_fks(dm_v4)
    Message
      Upgrading dm object created with dm <= 1.0.3.
    Output
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action

# can upgrade zoomed to v4

    Code
      def <- dm_get_def(dm_v4_zoomed, quiet = TRUE)
      def <- dm_get_def(dm_v4_zoomed)
    Message
      Upgrading dm object created with dm <= 1.0.3.
    Code
      dm <- dm_from_def(def, zoomed = TRUE)
      dm_validate(dm)
      is_zoomed(dm)
    Output
      [1] TRUE

# dummy

    Code
      TRUE
    Output
      [1] TRUE

# dm_meta() data model

    Code
      dm_meta(my_test_src()) %>% dm_paste(options = c("select", "keys", "color"))
    Message
      dm::dm(
        schemata,
        tables,
        columns,
        table_constraints,
        key_column_usage,
        constraint_column_usage,
      ) %>%
        dm::dm_select(schemata, catalog_name, schema_name) %>%
        dm::dm_select(tables, table_catalog, table_schema, table_name, table_type) %>%
        dm::dm_select(columns, table_catalog, table_schema, table_name, column_name, ordinal_position, column_default, is_nullable, is_autoincrement) %>%
        dm::dm_select(table_constraints, constraint_catalog, constraint_schema, constraint_name, table_catalog, table_schema, table_name, constraint_type, delete_rule) %>%
        dm::dm_select(key_column_usage, constraint_catalog, constraint_schema, constraint_name, table_catalog, table_schema, table_name, column_name, ordinal_position) %>%
        dm::dm_select(constraint_column_usage, table_catalog, table_schema, table_name, column_name, constraint_catalog, constraint_schema, constraint_name, ordinal_position) %>%
        dm::dm_add_pk(schemata, c(catalog_name, schema_name)) %>%
        dm::dm_add_pk(tables, c(table_catalog, table_schema, table_name)) %>%
        dm::dm_add_pk(columns, c(table_catalog, table_schema, table_name, column_name)) %>%
        dm::dm_add_pk(table_constraints, c(constraint_catalog, constraint_schema, constraint_name)) %>%
        dm::dm_add_pk(key_column_usage, c(constraint_catalog, constraint_schema, constraint_name, ordinal_position)) %>%
        dm::dm_add_pk(constraint_column_usage, c(constraint_catalog, constraint_schema, constraint_name, ordinal_position)) %>%
        dm::dm_add_fk(tables, c(table_catalog, table_schema), schemata) %>%
        dm::dm_add_fk(columns, c(table_catalog, table_schema, table_name), tables) %>%
        dm::dm_add_fk(table_constraints, c(table_catalog, table_schema, table_name), tables) %>%
        dm::dm_add_fk(key_column_usage, c(table_catalog, table_schema, table_name, column_name), columns) %>%
        dm::dm_add_fk(constraint_column_usage, c(table_catalog, table_schema, table_name, column_name), columns) %>%
        dm::dm_add_fk(key_column_usage, c(constraint_catalog, constraint_schema, constraint_name), table_constraints) %>%
        dm::dm_add_fk(constraint_column_usage, c(constraint_catalog, constraint_schema, constraint_name), table_constraints) %>%
        dm::dm_add_fk(constraint_column_usage, c(constraint_catalog, constraint_schema, constraint_name, ordinal_position), key_column_usage) %>%
        dm::dm_set_colors(`#0000FFFF` = schemata) %>%
        dm::dm_set_colors(`#A52A2AFF` = tables) %>%
        dm::dm_set_colors(`#A52A2AFF` = columns) %>%
        dm::dm_set_colors(`#008B00FF` = table_constraints) %>%
        dm::dm_set_colors(`#FFA500FF` = key_column_usage) %>%
        dm::dm_set_colors(`#FFA500FF` = constraint_column_usage)

# dm_disentangle() works

    Code
      dm_disentangle(dm_for_filter_w_cycle(), tf_1) %>% dm_get_all_fks()
    Message
      Replaced table `tf_3` with `tf_3-1`, `tf_3-2`.
      Replaced table `tf_4` with `tf_4-1`, `tf_4-2`.
      Replaced table `tf_5` with `tf_5-1`, `tf_5-2`.
      Replaced table `tf_6` with `tf_6-1`, `tf_6-2`.
      Replaced table `tf_7` with `tf_7-1`, `tf_7-2`.
    Output
      # A tibble: 11 x 5
         child_table child_fk_cols parent_table parent_key_cols on_delete
         <chr>       <keys>        <chr>        <keys>          <chr>    
       1 tf_2        d             tf_1         a               no_action
       2 tf_7-1      q             tf_2         c               no_action
       3 tf_4-1      j, j1         tf_3-1       f, f1           no_action
       4 tf_2        e, e1         tf_3-2       f, f1           no_action
       5 tf_4-2      j, j1         tf_3-2       f, f1           no_action
       6 tf_5-1      l             tf_4-1       h               cascade  
       7 tf_5-2      l             tf_4-2       h               cascade  
       8 tf_5-1      m             tf_6-1       n               no_action
       9 tf_5-2      m             tf_6-2       n               no_action
      10 tf_6-1      o             tf_7-1       p               no_action
      11 tf_6-2      o             tf_7-2       p               no_action
    Code
      dm_disentangle(dm_for_filter_w_cycle(), tf_5) %>% dm_get_all_fks()
    Message
      Replaced table `tf_1` with `tf_1-1`, `tf_1-2`.
      Replaced table `tf_2` with `tf_2-1`, `tf_2-2`.
      Replaced table `tf_3` with `tf_3-1`, `tf_3-2`.
      Replaced table `tf_4` with `tf_4-1`, `tf_4-2`.
      Replaced table `tf_6` with `tf_6-1`, `tf_6-2`.
      Replaced table `tf_7` with `tf_7-1`, `tf_7-2`.
    Output
      # A tibble: 12 x 5
         child_table child_fk_cols parent_table parent_key_cols on_delete
         <chr>       <keys>        <chr>        <keys>          <chr>    
       1 tf_2-1      d             tf_1-1       a               no_action
       2 tf_2-2      d             tf_1-2       a               no_action
       3 tf_7-1      q             tf_2-1       c               no_action
       4 tf_7-2      q             tf_2-2       c               no_action
       5 tf_4-1      j, j1         tf_3-1       f, f1           no_action
       6 tf_2-1      e, e1         tf_3-1       f, f1           no_action
       7 tf_2-2      e, e1         tf_3-2       f, f1           no_action
       8 tf_4-2      j, j1         tf_3-2       f, f1           no_action
       9 tf_5        l             tf_4-1       h               cascade  
      10 tf_5        m             tf_6-2       n               no_action
      11 tf_6-1      o             tf_7-1       p               no_action
      12 tf_6-2      o             tf_7-2       p               no_action
    Code
      dm_disentangle(entangled_dm(), a, quiet = TRUE) %>% dm_get_all_fks()
    Output
      # A tibble: 22 x 5
         child_table child_fk_cols parent_table parent_key_cols on_delete
         <chr>       <keys>        <chr>        <keys>          <chr>    
       1 a           a             b-1          b               no_action
       2 a           a             c-2          c               no_action
       3 b-1         b             d-1          d               no_action
       4 c-1         c             d-1          d               no_action
       5 c-2         c             d-2          d               no_action
       6 b-2         b             d-2          d               no_action
       7 d-1         d             e-1          e               no_action
       8 d-2         d             e-3          e               no_action
       9 d-1         d             f-2          f               no_action
      10 d-2         d             f-4          f               no_action
      # i 12 more rows
    Code
      dm_disentangle(entangled_dm(), c) %>% dm_get_all_fks()
    Message
      Replaced table `a` with `a-1`, `a-2`.
      Replaced table `b` with `b-1`, `b-2`.
      Replaced table `d` with `d-1`, `d-2`.
      Replaced table `e` with `e-1`, `e-2`, `e-3`, `e-4`.
      Replaced table `f` with `f-1`, `f-2`, `f-3`, `f-4`.
      Replaced table `g` with `g-1`, `g-2`, `g-3`, `g-4`.
      Replaced table `h` with `h-1`, `h-2`, `h-3`, `h-4`.
    Output
      # A tibble: 22 x 5
         child_table child_fk_cols parent_table parent_key_cols on_delete
         <chr>       <keys>        <chr>        <keys>          <chr>    
       1 a-1         a             b-1          b               no_action
       2 a-2         a             b-2          b               no_action
       3 a-1         a             c            c               no_action
       4 b-1         b             d-1          d               no_action
       5 c           c             d-2          d               no_action
       6 b-2         b             d-2          d               no_action
       7 d-1         d             e-1          e               no_action
       8 d-2         d             e-3          e               no_action
       9 d-1         d             f-2          f               no_action
      10 d-2         d             f-4          f               no_action
      # i 12 more rows
    Code
      dm_disentangle(entangled_dm_2(), a) %>% dm_get_all_fks()
    Message
      Replaced table `b` with `b-1`, `b-2`.
      Replaced table `c` with `c-1`, `c-2`.
      Replaced table `d` with `d-1`, `d-2`.
      Replaced table `e` with `e-1`, `e-2`.
    Output
      # A tibble: 9 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 a           a             d-1          d               no_action
      2 b-1         b             d-1          d               no_action
      3 c-1         c             d-1          d               no_action
      4 b-2         b             d-2          d               no_action
      5 c-2         c             d-2          d               no_action
      6 d-1         d             e-1          e               no_action
      7 a           a             e-2          e               no_action
      8 d-2         d             e-2          e               no_action
      9 f           f             g            g               no_action
    Code
      dm_disentangle(entangled_dm_2(), d, quiet = TRUE) %>% dm_get_all_fks()
    Output
      # A tibble: 7 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 a-1         a             d            d               no_action
      2 b           b             d            d               no_action
      3 c           c             d            d               no_action
      4 a-1         a             e-1          e               no_action
      5 d           d             e-2          e               no_action
      6 a-2         a             e-2          e               no_action
      7 f           f             g            g               no_action

---

    Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

---

    Must pass `start` argument.

# `dm_nest_tbl()` failure modes

    Code
      dm_nest_tbl(dm_for_filter(), tf_1)
    Condition
      Error in `dm_nest_tbl()`:
      ! "tf_1" can't be nested because it is not a terminal child table.
      children: "tf_2"

---

    Code
      dm_nest_tbl(dm_for_filter(), tf_2)
    Condition
      Error in `dm_nest_tbl()`:
      ! "tf_2" can't be nested because it is not a terminal child table.
      parents: "tf_1" and "tf_3"

---

    Code
      dm_nest_tbl(dm_for_filter(), tf_3)
    Condition
      Error in `dm_nest_tbl()`:
      ! "tf_3" can't be nested because it is not a terminal child table.
      children: "tf_2" and "tf_4"

---

    Code
      dm_nest_tbl(dm_for_filter(), tf_4)
    Condition
      Error in `dm_nest_tbl()`:
      ! "tf_4" can't be nested because it is not a terminal child table.
      parents: "tf_3"
      children: "tf_5"

# `dm_pack_tbl()`, `dm_unpack_tbl()`, `dm_nest_tbl()`, `dm_unnest_tbl()` work

    Code
      dm_packed <- dm_pack_tbl(dm_for_filter(), tf_1)
      dm_packed
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_2`, `tf_3`, `tf_4`, `tf_5`, `tf_6`
      Columns: 19
      Primary keys: 5
      Foreign keys: 4
    Code
      dm_packed_nested <- dm_nest_tbl(dm_packed, tf_2)
      dm_packed_nested
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_3`, `tf_4`, `tf_5`, `tf_6`
      Columns: 15
      Primary keys: 4
      Foreign keys: 3
    Code
      dm_packed_nested_unnested <- dm_unnest_tbl(dm_packed_nested, tf_3, tf_2, ptype = dm_for_filter())
      dm_packed_nested_unnested
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_3`, `tf_4`, `tf_5`, `tf_6`, `tf_2`
      Columns: 19
      Primary keys: 5
      Foreign keys: 4
    Code
      dm_packed_nested_unnested_unpacked <- dm_unpack_tbl(dm_packed_nested_unnested,
        tf_2, tf_1, ptype = dm_for_filter())
      dm_packed_nested_unnested_unpacked
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_3`, `tf_4`, `tf_5`, `tf_6`, `tf_2`, `tf_1`
      Columns: 20
      Primary keys: 6
      Foreign keys: 5

# `dm_pack_tbl()` failure modes

    Code
      dm_pack_tbl(dm_for_filter(), tf_2)
    Condition
      Error in `dm_pack_tbl()`:
      ! "tf_2" can't be packed because it is not a terminal parent table.
      parents : "tf_1" and "tf_3"

---

    Code
      dm_pack_tbl(dm_for_filter(), tf_3)
    Condition
      Error in `dm_pack_tbl()`:
      ! "tf_3" can't be packed because it is not a terminal parent table.
      children: "tf_2" and "tf_4"

---

    Code
      dm_pack_tbl(dm_for_filter(), tf_4)
    Condition
      Error in `dm_pack_tbl()`:
      ! "tf_4" can't be packed because it is not a terminal parent table.
      parents : "tf_3"
      children: "tf_5"

# json joins work

    Code
      packed <- json_pack_join(df1, df2, by = "key")
      packed
    Output
      # A tibble: 3 x 3
         col1 key   df2       
        <int> <chr> <list>    
      1     1 a     <json [1]>
      2     1 a     <json [1]>
      3     2 b     <json [1]>
    Code
      packed$df2
    Output
      [[1]]
      [{"col2":3,"col3":"X"}] 
      
      [[2]]
      [{"col2":4,"col3":"Y"}] 
      
      [[3]]
      [{"col2":3.14159265358979,"col3":"Z"}] 
      

---

    Code
      nested <- json_nest_join(df1, df2, by = "key")
      nested
    Output
      # A tibble: 2 x 3
         col1 key   df2       
        <int> <chr> <list>    
      1     1 a     <json [1]>
      2     2 b     <json [1]>
    Code
      nested$df2
    Output
      [[1]]
      [{"col2":3,"col3":"X"},{"col2":4,"col3":"Y"}] 
      
      [[2]]
      [{"col2":3.14159265358979,"col3":"Z"}] 
      

# dm_add_tbl() snapshots

    Code
      dm_add_tbl(dm_for_filter(), tf_1 = data_card_1(), repair = "check_unique")
    Condition
      Warning:
      `dm_add_tbl()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
      i Use `.name_repair = "unique"` if necessary.
      Error in `abort_need_unique_names()`:
      ! Each new table needs to have a unique name. Duplicate new name: tf_1.

---

    Code
      dm_add_tbl(dm_for_flatten(), res_flat = result_from_flatten()) %>% dm_paste(
        options = c("select", "keys"))
    Condition
      Warning:
      `dm_add_tbl()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
      i Use `.name_repair = "unique"` if necessary.
    Message
      dm::dm(
        fact,
        dim_1,
        dim_2,
        dim_3,
        dim_4,
        res_flat,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something) %>%
        dm::dm_select(dim_1, dim_1_pk_1, dim_1_pk_2, something) %>%
        dm::dm_select(dim_2, dim_2_pk, something) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_select(res_flat, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, fact.something, dim_1.something, dim_2.something, dim_3.something, dim_4.something) %>%
        dm::dm_add_pk(dim_1, c(dim_1_pk_1, dim_1_pk_2)) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, c(dim_1_key_1, dim_1_key_2), dim_1) %>%
        dm::dm_add_fk(fact, dim_2_key, dim_2) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)

# dm_rm_tbl() snapshot

    Code
      dm_rm_tbl(dm_for_flatten(), dim_1) %>% dm_paste(options = c("select", "keys"))
    Condition
      Warning:
      `dm_rm_tbl()` was deprecated in dm 1.0.0.
      i Please use `dm_select_tbl()` instead.
    Message
      dm::dm(
        fact,
        dim_2,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something) %>%
        dm::dm_select(dim_2, dim_2_pk, something) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, dim_2_key, dim_2) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)
    Code
      dm_rm_tbl(dm_for_flatten(), fact) %>% dm_paste(options = c("select", "keys"))
    Condition
      Warning:
      `dm_rm_tbl()` was deprecated in dm 1.0.0.
      i Please use `dm_select_tbl()` instead.
    Message
      dm::dm(
        dim_1,
        dim_2,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(dim_1, dim_1_pk_1, dim_1_pk_2, something) %>%
        dm::dm_select(dim_2, dim_2_pk, something) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_1, c(dim_1_pk_1, dim_1_pk_2)) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk)

# dm_meta() contents

    Code
      meta %>% dm_get_tables() %>% map(select, -any_of("constraint_name")) %>% map(
        arrange_all) %>% map(collect) %>% jsonlite::toJSON(pretty = TRUE) %>% gsub(
        schema_name, "schema_name", .) %>% gsub("(_catalog\": \")[^\"]*(\")",
        "\\1catalog\\2", .) %>% writeLines()
    Output
      {
        "schemata": [
          {
            "catalog_name": "master",
            "schema_name": "schema_name"
          }
        ],
        "tables": [
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_1",
            "table_type": "BASE TABLE"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "table_type": "BASE TABLE"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "table_type": "BASE TABLE"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "table_type": "BASE TABLE"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "table_type": "BASE TABLE"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_6",
            "table_type": "BASE TABLE"
          }
        ],
        "columns": [
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_1",
            "column_name": "a",
            "ordinal_position": 1,
            "is_nullable": "NO"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_1",
            "column_name": "b",
            "ordinal_position": 2,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "column_name": "c",
            "ordinal_position": 1,
            "is_nullable": "NO"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "column_name": "d",
            "ordinal_position": 2,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "column_name": "e",
            "ordinal_position": 3,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "column_name": "e1",
            "ordinal_position": 4,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "column_name": "f",
            "ordinal_position": 1,
            "is_nullable": "NO"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "column_name": "f1",
            "ordinal_position": 2,
            "is_nullable": "NO"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "column_name": "g",
            "ordinal_position": 3,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "column_name": "h",
            "ordinal_position": 1,
            "is_nullable": "NO"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "column_name": "i",
            "ordinal_position": 2,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "column_name": "j",
            "ordinal_position": 3,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "column_name": "j1",
            "ordinal_position": 4,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "column_name": "k",
            "ordinal_position": 2,
            "is_nullable": "NO"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "column_name": "l",
            "ordinal_position": 3,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "column_name": "m",
            "ordinal_position": 4,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "column_name": "ww",
            "ordinal_position": 1,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_6",
            "column_name": "n",
            "ordinal_position": 2,
            "is_nullable": "YES"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_6",
            "column_name": "o",
            "ordinal_position": 3,
            "is_nullable": "NO"
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_6",
            "column_name": "zz",
            "ordinal_position": 1,
            "is_nullable": "YES"
          }
        ],
        "table_constraints": [
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_1",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "constraint_type": "FOREIGN KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "constraint_type": "FOREIGN KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "constraint_type": "FOREIGN KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "constraint_type": "FOREIGN KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "constraint_type": "FOREIGN KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_6",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_6",
            "constraint_type": "UNIQUE"
          }
        ],
        "key_column_usage": [
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_1",
            "column_name": "a",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "column_name": "c",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "column_name": "d",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "column_name": "e",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_2",
            "column_name": "e1",
            "ordinal_position": 2
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "column_name": "f",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "column_name": "f1",
            "ordinal_position": 2
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "column_name": "h",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "column_name": "j",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "column_name": "j1",
            "ordinal_position": 2
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "column_name": "k",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "column_name": "l",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_5",
            "column_name": "m",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_6",
            "column_name": "n",
            "ordinal_position": 1
          },
          {
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_6",
            "column_name": "o",
            "ordinal_position": 1
          }
        ],
        "constraint_column_usage": [
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_1",
            "column_name": "a",
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "ordinal_position": 1
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "column_name": "f",
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "ordinal_position": 1
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "column_name": "f",
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "ordinal_position": 1
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "column_name": "f1",
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "ordinal_position": 2
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_3",
            "column_name": "f1",
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "ordinal_position": 2
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_4",
            "column_name": "h",
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "ordinal_position": 1
          },
          {
            "table_catalog": "catalog",
            "table_schema": "schema_name",
            "table_name": "tf_6",
            "column_name": "n",
            "constraint_catalog": "catalog",
            "constraint_schema": "schema_name",
            "ordinal_position": 1
          }
        ]
      }

# `check_suggested()` works

    Code
      check_suggested("rlang", TRUE, top_level_fun = "foo")
    Output
      [1] TRUE
    Code
      check_suggested("dm", NA, top_level_fun = "foo")
    Output
      [1] FALSE
    Code
      check_suggested("iurtnkjvmomweicopbt", NA, top_level_fun = "foo")
    Output
      [1] FALSE
    Code
      check_suggested("iurtnkjvmomweicopbt", NA, message = "not installed!")
    Output
      [1] FALSE
    Code
      check_suggested("dm", FALSE, top_level_fun = "foo")
    Output
      [1] FALSE
    Code
      check_suggested("iurtnkjvmomweicopbt", FALSE, top_level_fun = "foo")
    Output
      [1] FALSE
    Code
      check_suggested("iurtnkjvmomweicopbt", FALSE, message = "not installed!")
    Output
      [1] FALSE

# `check_suggested()` works for error messages

    Code
      check_suggested("iurtnkjvmomweicopbt", TRUE, top_level_fun = "foo")
    Condition
      Error:
      ! `foo()` needs the 'iurtnkjvmomweicopbt' package. Do you need `install.packages("iurtnkjvmomweicopbt")` ?
    Code
      check_suggested("iurtnkjvmomweicopbt", TRUE, message = "not installed!")
    Condition
      Error:
      ! not installed!

# Learning keys from SQLite works

    Code
      dm_paste(learned_dm, options = "keys")
    Message
      dm::dm(
        first,
        second,
        third,
      ) %>%
        dm::dm_add_pk(first, id) %>%
        dm::dm_add_pk(second, id) %>%
        dm::dm_add_pk(third, id) %>%
        dm::dm_add_fk(second, first_id, first) %>%
        dm::dm_add_fk(third, first_id, first) %>%
        dm::dm_add_fk(third, second_id, second)

# Learning keys from an attached SQLite database works

    Code
      dm_paste(learned_dm, options = "keys")
    Message
      dm::dm(
        other.child_tbl,
        other.parent_tbl,
      ) %>%
        dm::dm_add_pk(other.child_tbl, id) %>%
        dm::dm_add_pk(other.parent_tbl, id) %>%
        dm::dm_add_fk(other.child_tbl, parent_id, other.parent_tbl)

# dm_meta() contents

    Code
      meta %>% dm_select_tbl(-schemata) %>% dm_zoom_to(table_constraints) %>% filter(
        constraint_type %in% c("PRIMARY KEY", "FOREIGN KEY")) %>% dm_update_zoomed() %>%
        dm_get_tables() %>% map(select, -any_of("column_default"), -contains(
        "catalog"), -contains("schema")) %>% map(collect) %>% map(
        arrange_all_but_constraint_name) %>% map(~ if ("constraint_name" %in%
        colnames(.x)) {
        .x %>% mutate(constraint_name = as.integer(forcats::fct_inorder(
          constraint_name)))
      } else {
        .x
      }) %>% imap(~ if (is_mariadb(con_db) && .y == "columns") {
        mutate(.x, is_autoincrement = as.logical(is_autoincrement))
      } else {
        .x
      }) %>% imap(~ if (is_mariadb(con_db) && .y == "table_constraints") {
        mutate(.x, delete_rule = if_else(delete_rule == "RESTRICT", "NO ACTION",
        delete_rule))
      } else {
        .x
      }) %>% map(arrange_all) %>% jsonlite::toJSON(pretty = TRUE) %>% gsub(
        schema_name, "schema_name", .) %>% gsub("(_catalog\": \")[^\"]*(\")",
        "\\1catalog\\2", .) %>% writeLines()
    Output
      {
        "tables": [
          {
            "table_name": "tf_1",
            "table_type": "BASE TABLE"
          },
          {
            "table_name": "tf_2",
            "table_type": "BASE TABLE"
          },
          {
            "table_name": "tf_3",
            "table_type": "BASE TABLE"
          },
          {
            "table_name": "tf_4",
            "table_type": "BASE TABLE"
          },
          {
            "table_name": "tf_5",
            "table_type": "BASE TABLE"
          },
          {
            "table_name": "tf_6",
            "table_type": "BASE TABLE"
          }
        ],
        "columns": [
          {
            "table_name": "tf_1",
            "column_name": "a",
            "ordinal_position": 1,
            "is_nullable": "NO",
            "is_autoincrement": true
          },
          {
            "table_name": "tf_1",
            "column_name": "b",
            "ordinal_position": 2,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_2",
            "column_name": "c",
            "ordinal_position": 1,
            "is_nullable": "NO",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_2",
            "column_name": "d",
            "ordinal_position": 2,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_2",
            "column_name": "e",
            "ordinal_position": 3,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_2",
            "column_name": "e1",
            "ordinal_position": 4,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_3",
            "column_name": "f",
            "ordinal_position": 1,
            "is_nullable": "NO",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_3",
            "column_name": "f1",
            "ordinal_position": 2,
            "is_nullable": "NO",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_3",
            "column_name": "g",
            "ordinal_position": 3,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_4",
            "column_name": "h",
            "ordinal_position": 1,
            "is_nullable": "NO",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_4",
            "column_name": "i",
            "ordinal_position": 2,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_4",
            "column_name": "j",
            "ordinal_position": 3,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_4",
            "column_name": "j1",
            "ordinal_position": 4,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_5",
            "column_name": "k",
            "ordinal_position": 2,
            "is_nullable": "NO",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_5",
            "column_name": "l",
            "ordinal_position": 3,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_5",
            "column_name": "m",
            "ordinal_position": 4,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_5",
            "column_name": "ww",
            "ordinal_position": 1,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_6",
            "column_name": "n",
            "ordinal_position": 2,
            "is_nullable": "YES",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_6",
            "column_name": "o",
            "ordinal_position": 3,
            "is_nullable": "NO",
            "is_autoincrement": false
          },
          {
            "table_name": "tf_6",
            "column_name": "zz",
            "ordinal_position": 1,
            "is_nullable": "YES",
            "is_autoincrement": false
          }
        ],
        "table_constraints": [
          {
            "constraint_name": 1,
            "table_name": "tf_1",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_name": 2,
            "table_name": "tf_2",
            "constraint_type": "FOREIGN KEY",
            "delete_rule": "NO ACTION"
          },
          {
            "constraint_name": 3,
            "table_name": "tf_2",
            "constraint_type": "FOREIGN KEY",
            "delete_rule": "NO ACTION"
          },
          {
            "constraint_name": 4,
            "table_name": "tf_2",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_name": 5,
            "table_name": "tf_3",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_name": 6,
            "table_name": "tf_4",
            "constraint_type": "FOREIGN KEY",
            "delete_rule": "NO ACTION"
          },
          {
            "constraint_name": 7,
            "table_name": "tf_4",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_name": 8,
            "table_name": "tf_5",
            "constraint_type": "FOREIGN KEY",
            "delete_rule": "CASCADE"
          },
          {
            "constraint_name": 9,
            "table_name": "tf_5",
            "constraint_type": "FOREIGN KEY",
            "delete_rule": "NO ACTION"
          },
          {
            "constraint_name": 10,
            "table_name": "tf_5",
            "constraint_type": "PRIMARY KEY"
          },
          {
            "constraint_name": 11,
            "table_name": "tf_6",
            "constraint_type": "PRIMARY KEY"
          }
        ],
        "key_column_usage": [
          {
            "constraint_name": 1,
            "table_name": "tf_1",
            "column_name": "a",
            "ordinal_position": 1
          },
          {
            "constraint_name": 2,
            "table_name": "tf_1",
            "column_name": "a",
            "ordinal_position": 1
          },
          {
            "constraint_name": 3,
            "table_name": "tf_2",
            "column_name": "c",
            "ordinal_position": 1
          },
          {
            "constraint_name": 4,
            "table_name": "tf_2",
            "column_name": "c",
            "ordinal_position": 1
          },
          {
            "constraint_name": 5,
            "table_name": "tf_2",
            "column_name": "d",
            "ordinal_position": 1
          },
          {
            "constraint_name": 6,
            "table_name": "tf_2",
            "column_name": "e",
            "ordinal_position": 1
          },
          {
            "constraint_name": 6,
            "table_name": "tf_2",
            "column_name": "e1",
            "ordinal_position": 2
          },
          {
            "constraint_name": 7,
            "table_name": "tf_3",
            "column_name": "f",
            "ordinal_position": 1
          },
          {
            "constraint_name": 7,
            "table_name": "tf_3",
            "column_name": "f1",
            "ordinal_position": 2
          },
          {
            "constraint_name": 8,
            "table_name": "tf_3",
            "column_name": "f",
            "ordinal_position": 1
          },
          {
            "constraint_name": 8,
            "table_name": "tf_3",
            "column_name": "f1",
            "ordinal_position": 2
          },
          {
            "constraint_name": 9,
            "table_name": "tf_3",
            "column_name": "g",
            "ordinal_position": 1
          },
          {
            "constraint_name": 10,
            "table_name": "tf_4",
            "column_name": "h",
            "ordinal_position": 1
          },
          {
            "constraint_name": 11,
            "table_name": "tf_4",
            "column_name": "h",
            "ordinal_position": 1
          },
          {
            "constraint_name": 12,
            "table_name": "tf_4",
            "column_name": "j",
            "ordinal_position": 1
          },
          {
            "constraint_name": 12,
            "table_name": "tf_4",
            "column_name": "j1",
            "ordinal_position": 2
          },
          {
            "constraint_name": 13,
            "table_name": "tf_5",
            "column_name": "k",
            "ordinal_position": 1
          },
          {
            "constraint_name": 14,
            "table_name": "tf_5",
            "column_name": "k",
            "ordinal_position": 1
          },
          {
            "constraint_name": 15,
            "table_name": "tf_5",
            "column_name": "l",
            "ordinal_position": 1
          },
          {
            "constraint_name": 16,
            "table_name": "tf_5",
            "column_name": "m",
            "ordinal_position": 1
          },
          {
            "constraint_name": 17,
            "table_name": "tf_6",
            "column_name": "n",
            "ordinal_position": 1
          },
          {
            "constraint_name": 18,
            "table_name": "tf_6",
            "column_name": "o",
            "ordinal_position": 1
          },
          {
            "constraint_name": 19,
            "table_name": "tf_6",
            "column_name": "o",
            "ordinal_position": 1
          }
        ],
        "constraint_column_usage": [
          {
            "table_name": "tf_1",
            "column_name": "a",
            "constraint_name": 1,
            "ordinal_position": 1
          },
          {
            "table_name": "tf_3",
            "column_name": "f",
            "constraint_name": 2,
            "ordinal_position": 1
          },
          {
            "table_name": "tf_3",
            "column_name": "f",
            "constraint_name": 3,
            "ordinal_position": 1
          },
          {
            "table_name": "tf_3",
            "column_name": "f1",
            "constraint_name": 2,
            "ordinal_position": 2
          },
          {
            "table_name": "tf_3",
            "column_name": "f1",
            "constraint_name": 3,
            "ordinal_position": 2
          },
          {
            "table_name": "tf_4",
            "column_name": "h",
            "constraint_name": 4,
            "ordinal_position": 1
          },
          {
            "table_name": "tf_6",
            "column_name": "n",
            "constraint_name": 5,
            "ordinal_position": 1
          }
        ]
      }

# dm_from_con() with mariaDB

    Code
      dm::dm_get_all_fks(my_dm)
    Output
      # A tibble: 8 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 disps       account_id    accounts     id              cascade  
      2 loans       account_id    accounts     id              cascade  
      3 orders      account_id    accounts     id              cascade  
      4 trans       account_id    accounts     id              cascade  
      5 disps       client_id     clients      id              cascade  
      6 cards       disp_id       disps        id              cascade  
      7 accounts    district_id   districts    id              cascade  
      8 clients     district_id   districts    id              cascade  

---

    Code
      dm::dm_get_all_pks(my_dm)
    Output
      # A tibble: 9 x 3
        table     pk_col autoincrement
        <chr>     <keys> <lgl>        
      1 accounts  id     FALSE        
      2 cards     id     FALSE        
      3 clients   id     FALSE        
      4 disps     id     FALSE        
      5 districts id     FALSE        
      6 loans     id     FALSE        
      7 orders    id     FALSE        
      8 tkeys     id     FALSE        
      9 trans     id     FALSE        

---

    Code
      dm::dm_get_all_fks(my_dm)
    Output
      # A tibble: 9 x 5
        child_table              child_fk_cols  parent_table parent_key_cols on_delete
        <chr>                    <keys>         <chr>        <keys>          <chr>    
      1 Accidents.oseba          id_nesreca     Accidents.n~ id_nesreca      cascade  
      2 Accidents.nesreca        upravna_enota  Accidents.u~ id_upravna_eno~ cascade  
      3 Accidents.oseba          upravna_enota  Accidents.u~ id_upravna_eno~ cascade  
      4 Financial_std.Loan_Acc   account_id     Financial_s~ account_id      cascade  
      5 Financial_std.Loan_Acc   loan_id        Financial_s~ loan_id         cascade  
      6 Financial_std.Loan_Order loan_id        Financial_s~ loan_id         cascade  
      7 Financial_std.Loan_Trans loan_id        Financial_s~ loan_id         cascade  
      8 Financial_std.Loan_Order order_id       Financial_s~ order_id        cascade  
      9 Financial_std.Loan_Trans transaction_id Financial_s~ transaction_id  cascade  

---

    Code
      dm::dm_get_all_pks(my_dm)
    Output
      # A tibble: 10 x 3
         table                    pk_col                  autoincrement
         <chr>                    <keys>                  <lgl>        
       1 Accidents.nesreca        id_nesreca              FALSE        
       2 Accidents.upravna_enota  id_upravna_enota        FALSE        
       3 Ad.ad                    ts, ad_id, user_id      FALSE        
       4 Financial_std.Loan_Acc   loan_id, account_id     FALSE        
       5 Financial_std.Loan_Order order_id, loan_id       FALSE        
       6 Financial_std.Loan_Trans transaction_id, loan_id FALSE        
       7 Financial_std.acc        account_id              FALSE        
       8 Financial_std.loan       loan_id                 FALSE        
       9 Financial_std.order      order_id                FALSE        
      10 Financial_std.trans      transaction_id          TRUE         

# datamodel-code for drawing

    Code
      dm_get_data_model(dm_for_filter())
    Output
      $tables
        table segment display
      1  tf_1    <NA>    <NA>
      2  tf_2    <NA>    <NA>
      3  tf_3    <NA>    <NA>
      4  tf_4    <NA>    <NA>
      5  tf_5    <NA>    <NA>
      6  tf_6    <NA>    <NA>
      
      $columns
         table column id        kind key  ref ref_col  keyId           uk_col
      1   tf_1      a  1          PK   1 <NA>    <NA>   <NA>             <NA>
      2   tf_1      b  2        <NA>   0 <NA>    <NA>   <NA>             <NA>
      3   tf_2      c  1          PK   1 <NA>    <NA>   <NA>             <NA>
      4   tf_2      d  2        <NA>   0 tf_1       a tf_2_1                 
      5   tf_2      e  3        <NA>   0 <NA>    <NA>   <NA>             <NA>
      6   tf_2     e1  4        <NA>   0 <NA>    <NA>   <NA>             <NA>
      7   tf_3      f  1        <NA>   0 <NA>    <NA>   <NA>             <NA>
      8   tf_3     f1  2        <NA>   0 <NA>    <NA>   <NA>             <NA>
      9   tf_3      g  3 explicit UK   1 <NA>    <NA>   <NA>             <NA>
      10  tf_4      h  1          PK   1 <NA>    <NA>   <NA>             <NA>
      11  tf_4      i  2        <NA>   0 <NA>    <NA>   <NA>             <NA>
      12  tf_4      j  3        <NA>   0 <NA>    <NA>   <NA>             <NA>
      13  tf_4     j1  4        <NA>   0 <NA>    <NA>   <NA>             <NA>
      14  tf_5     ww  1        <NA>   0 <NA>    <NA>   <NA>             <NA>
      15  tf_5      k  2          PK   1 <NA>    <NA>   <NA>             <NA>
      16  tf_5      l  3        <NA>   0 tf_4       h tf_5_1                 
      17  tf_5      m  4        <NA>   0 tf_6       n tf_5_2 , style="dashed"
      18  tf_6     zz  1        <NA>   0 <NA>    <NA>   <NA>             <NA>
      19  tf_6      n  2 implicit UK   1 <NA>    <NA>   <NA>             <NA>
      20  tf_6      o  3          PK   1 <NA>    <NA>   <NA>             <NA>
      21  tf_3  f, f1 NA          PK   1 <NA>    <NA>   <NA>             <NA>
      22  tf_2  e, e1 NA        <NA>   0 tf_3   f, f1 tf_2_2                 
      23  tf_4  j, j1 NA        <NA>   0 tf_3   f, f1 tf_4_1                 
      
      $references
      # A tibble: 5 x 8
        table column ref   ref_col keyId  uk_col               ref_id ref_col_num
        <chr> <chr>  <chr> <chr>   <chr>  <chr>                 <int>       <int>
      1 tf_2  d      tf_1  a       tf_2_1 ""                        1           1
      2 tf_2  e, e1  tf_3  f, f1   tf_2_2 ""                        2           1
      3 tf_4  j, j1  tf_3  f, f1   tf_4_1 ""                        3           1
      4 tf_5  l      tf_4  h       tf_5_1 ""                        4           1
      5 tf_5  m      tf_6  n       tf_5_2 ", style=\"dashed\""      5           1
      
      attr(,"class")
      [1] "data_model"
    Code
      dm_get_data_model(dm_for_filter(), column_types = TRUE)
    Output
      $tables
        table segment display
      1  tf_1    <NA>    <NA>
      2  tf_2    <NA>    <NA>
      3  tf_3    <NA>    <NA>
      4  tf_4    <NA>    <NA>
      5  tf_5    <NA>    <NA>
      6  tf_6    <NA>    <NA>
      
      $columns
         table column id type        kind key  ref ref_col  keyId           uk_col
      1   tf_1      a  1  int          PK   1 <NA>    <NA>   <NA>             <NA>
      2   tf_1      b  2  chr        <NA>   0 <NA>    <NA>   <NA>             <NA>
      3   tf_2      c  1  chr          PK   1 <NA>    <NA>   <NA>             <NA>
      4   tf_2      d  2  int        <NA>   0 tf_1       a tf_2_1                 
      5   tf_2      e  3  chr        <NA>   0 <NA>    <NA>   <NA>             <NA>
      6   tf_2     e1  4  int        <NA>   0 <NA>    <NA>   <NA>             <NA>
      7   tf_3      f  1  chr        <NA>   0 <NA>    <NA>   <NA>             <NA>
      8   tf_3     f1  2  int        <NA>   0 <NA>    <NA>   <NA>             <NA>
      9   tf_3      g  3  chr explicit UK   1 <NA>    <NA>   <NA>             <NA>
      10  tf_4      h  1  chr          PK   1 <NA>    <NA>   <NA>             <NA>
      11  tf_4      i  2  chr        <NA>   0 <NA>    <NA>   <NA>             <NA>
      12  tf_4      j  3  chr        <NA>   0 <NA>    <NA>   <NA>             <NA>
      13  tf_4     j1  4  int        <NA>   0 <NA>    <NA>   <NA>             <NA>
      14  tf_5     ww  1  int        <NA>   0 <NA>    <NA>   <NA>             <NA>
      15  tf_5      k  2  int          PK   1 <NA>    <NA>   <NA>             <NA>
      16  tf_5      l  3  chr        <NA>   0 tf_4       h tf_5_1                 
      17  tf_5      m  4  chr        <NA>   0 tf_6       n tf_5_2 , style="dashed"
      18  tf_6     zz  1  int        <NA>   0 <NA>    <NA>   <NA>             <NA>
      19  tf_6      n  2  chr implicit UK   1 <NA>    <NA>   <NA>             <NA>
      20  tf_6      o  3  chr          PK   1 <NA>    <NA>   <NA>             <NA>
      21  tf_3  f, f1 NA <NA>          PK   1 <NA>    <NA>   <NA>             <NA>
      22  tf_2  e, e1 NA <NA>        <NA>   0 tf_3   f, f1 tf_2_2                 
      23  tf_4  j, j1 NA <NA>        <NA>   0 tf_3   f, f1 tf_4_1                 
      
      $references
      # A tibble: 5 x 8
        table column ref   ref_col keyId  uk_col               ref_id ref_col_num
        <chr> <chr>  <chr> <chr>   <chr>  <chr>                 <int>       <int>
      1 tf_2  d      tf_1  a       tf_2_1 ""                        1           1
      2 tf_2  e, e1  tf_3  f, f1   tf_2_2 ""                        2           1
      3 tf_4  j, j1  tf_3  f, f1   tf_4_1 ""                        3           1
      4 tf_5  l      tf_4  h       tf_5_1 ""                        4           1
      5 tf_5  m      tf_6  n       tf_5_2 ", style=\"dashed\""      5           1
      
      attr(,"class")
      [1] "data_model"

# copy_dm_to() copies data frames from any source

    Code
      copy_dm_to(default_local_src(), dm_for_filter())
    Condition
      Error:
      ! The `dest` argument of `copy_dm_to()` must refer to a DBI connection as of dm 0.1.6.
      i Please use `collect.dm()` instead.

# copy_dm_to() rejects overwrite and types arguments

    Code
      copy_dm_to(my_test_src(), dm_for_filter(), overwrite = TRUE)
    Condition
      Error in `copy_dm_to()`:
      ! `...` must be empty.
      x Problematic argument:
      * overwrite = TRUE

---

    Code
      copy_dm_to(my_test_src(), dm_for_filter(), types = character())
    Condition
      Error in `copy_dm_to()`:
      ! `...` must be empty.
      x Problematic argument:
      * types = character()

# `json_nest()` works remotely

    Code
      query <- remote %>% json_nest(a = starts_with("a")) %>% arrange(grp) %>%
        dbplyr::sql_render()
      gsub("test_frame_[_0-9]+", "test_frame_...", query)
    Output
      <SQL> SELECT "q01".*
      FROM (
      SELECT "grp", (SELECT "a_i", "a_j" FROM (SELECT *
      FROM "#test_frame_...") "*tmp_a*" WHERE ("grp" = "*tmp*"."grp") FOR JSON PATH) AS "a" FROM (SELECT *
      FROM "#test_frame_...") "*tmp*" GROUP BY "grp"
      ) "q01"
      ORDER BY "grp"
    Code
      remote %>% json_nest(a = starts_with("a")) %>% arrange(grp) %>% collect()
    Output
      # A tibble: 2 x 2
          grp a                                                              
        <dbl> <chr>                                                          
      1     1 "[{\"a_i\":\"a\",\"a_j\":\"A\"},{\"a_i\":\"b\",\"a_j\":\"B\"}]"
      2     2 "[{\"a_i\":\"c\",\"a_j\":\"C\"},{\"a_i\":\"d\",\"a_j\":\"D\"}]"
    Code
      query <- remote %>% json_nest(a = starts_with("a"), .names_sep = "_") %>%
        arrange(grp) %>% dbplyr::sql_render()
      gsub("test_frame_[_0-9]+", "test_frame_...", query)
    Output
      <SQL> SELECT "q01".*
      FROM (
      SELECT "grp", (SELECT "a_i" "i", "a_j" "j" FROM (SELECT *
      FROM "#test_frame_...") "*tmp_a*" WHERE ("grp" = "*tmp*"."grp") FOR JSON PATH) AS "a" FROM (SELECT *
      FROM "#test_frame_...") "*tmp*" GROUP BY "grp"
      ) "q01"
      ORDER BY "grp"
    Code
      remote %>% json_nest(a = starts_with("a"), .names_sep = "_") %>% arrange(grp) %>%
        collect()
    Output
      # A tibble: 2 x 2
          grp a                                                      
        <dbl> <chr>                                                  
      1     1 "[{\"i\":\"a\",\"j\":\"A\"},{\"i\":\"b\",\"j\":\"B\"}]"
      2     2 "[{\"i\":\"c\",\"j\":\"C\"},{\"i\":\"d\",\"j\":\"D\"}]"

# `dm_set_colors()` works

    Code
      dm_nycflights_small() %>% dm_set_colors(blue = starts_with("air"), green = contains(
        "h")) %>% dm_get_colors()
    Output
       #0000FFFF  #0000FFFF  #00FF00FF    default  #00FF00FF 
      "airlines" "airports"  "flights"   "planes"  "weather" 

# helpers

    Code
      dm_get_all_columns(dm_for_filter())
    Output
      # A tibble: 20 x 3
         table column    id
         <chr> <chr>  <int>
       1 tf_1  a          1
       2 tf_1  b          2
       3 tf_2  c          1
       4 tf_2  d          2
       5 tf_2  e          3
       6 tf_2  e1         4
       7 tf_3  f          1
       8 tf_3  f1         2
       9 tf_3  g          3
      10 tf_4  h          1
      11 tf_4  i          2
      12 tf_4  j          3
      13 tf_4  j1         4
      14 tf_5  ww         1
      15 tf_5  k          2
      16 tf_5  l          3
      17 tf_5  m          4
      18 tf_6  zz         1
      19 tf_6  n          2
      20 tf_6  o          3

---

    Code
      dm_get_all_column_types(dm_for_filter())
    Output
      # A tibble: 20 x 4
         table column    id type 
         <chr> <chr>  <int> <chr>
       1 tf_1  a          1 int  
       2 tf_1  b          2 chr  
       3 tf_2  c          1 chr  
       4 tf_2  d          2 int  
       5 tf_2  e          3 chr  
       6 tf_2  e1         4 int  
       7 tf_3  f          1 chr  
       8 tf_3  f1         2 int  
       9 tf_3  g          3 chr  
      10 tf_4  h          1 chr  
      11 tf_4  i          2 chr  
      12 tf_4  j          3 chr  
      13 tf_4  j1         4 int  
      14 tf_5  ww         1 int  
      15 tf_5  k          2 int  
      16 tf_5  l          3 chr  
      17 tf_5  m          4 chr  
      18 tf_6  zz         1 int  
      19 tf_6  n          2 chr  
      20 tf_6  o          3 chr  

# DiagrammeR-specific options are soft-deprecated

    Code
      invisible(dm_draw(dm_nycflights13(), graph_attrs = "rankdir=LR"))
    Condition
      Warning:
      The `graph_attrs` argument of `dm_draw()` is deprecated as of dm 1.1.0.
      i Use `backend_opts = list(graph_attrs = ...)` instead.

---

    Code
      invisible(dm_draw(dm_nycflights13(), font_size = 14L))
    Condition
      Warning:
      The `font_size` argument of `dm_draw()` is deprecated as of dm 1.1.0.
      i Use `backend_opts = list(font_size = ...)` instead.

# unsupported backend_opts fail with clear error

    Code
      dm_draw(dm_nycflights13(), backend_opts = list(columnArrows = TRUE))
    Condition
      Error in `dm_draw()`:
      ! Unsupported `backend_opts` for backend "DiagrammeR": columnArrows.
      i Supported options are: graph_attrs, node_attrs, edge_attrs, focus, graph_name, column_arrow, and font_size.

---

    Code
      dm_draw(dm_nycflights13(), backend_opts = list(foo = 1, bar = 2))
    Condition
      Error in `dm_draw()`:
      ! Unsupported `backend_opts` for backend "DiagrammeR": foo and bar.
      i Supported options are: graph_attrs, node_attrs, edge_attrs, focus, graph_name, column_arrow, and font_size.

# column_arrow backend option works

    Code
      invisible(dm_draw(dm_nycflights13(), columnArrows = FALSE))
    Condition
      Warning:
      The `columnArrows` argument of `dm_draw()` is deprecated as of dm 1.1.0.
      i Use `backend_opts = list(column_arrow = ...)` instead.

# `new_fks_in()` generates expected tibble

    Code
      new_fks_in(child_uuid = "flights-uuid", child_fk_cols = new_keys(list(list(
        "origin", "dest"))), parent_key_cols = new_keys(list(list("faa"))))
    Output
      # A tibble: 1 x 3
        child_uuid   child_fk_cols parent_key_cols
        <chr>        <keys>        <keys>         
      1 flights-uuid origin, dest  faa            

# `new_fks_out()` generates expected tibble

    Code
      new_fks_out(child_fk_cols = new_keys(list(list("origin", "dest"))),
      parent_uuid = "airports-uuid", parent_key_cols = new_keys(list(list("faa"))))
    Output
      # A tibble: 1 x 3
        child_fk_cols parent_uuid   parent_key_cols
        <keys>        <chr>         <keys>         
      1 origin, dest  airports-uuid faa            

# `new_keyed_tbl()` generates expected output

    Code
      dm <- dm_nycflights13(cycle = TRUE)
      keyed_tbl <- new_keyed_tbl(x = dm$airports, pk = "faa", fks_in = new_fks_in(
        child_uuid = "flights-uuid", child_fk_cols = new_keys(list("origin", "dest")),
        parent_key_cols = new_keys(list("faa"))), fks_out = new_fks_out(
        child_fk_cols = new_keys(list("origin", "dest")), parent_uuid = "airports-uuid",
        parent_key_cols = new_keys(list("faa"))), uuid = "0a0c060f-0d01-0b03-0402-05090800070e")
      keyed_get_info(keyed_tbl)
    Output
      $pk
      [1] "faa"
      
      $uks
      # A tibble: 0 x 1
      # i 1 variable: column <list>
      
      $fks_in
      # A tibble: 2 x 3
        child_uuid   child_fk_cols parent_key_cols
        <chr>        <keys>        <keys>         
      1 flights-uuid origin        faa            
      2 flights-uuid dest          faa            
      
      $fks_out
      # A tibble: 2 x 3
        child_fk_cols parent_uuid   parent_key_cols
        <keys>        <chr>         <keys>         
      1 origin        airports-uuid faa            
      2 dest          airports-uuid faa            
      
      $uuid
      [1] "0a0c060f-0d01-0b03-0402-05090800070e"
      

# dm_get_keyed_tables_impl()

    Code
      dm_nycflights13(cycle = TRUE) %>% dm_get_keyed_tables_impl() %>% map(
        keyed_get_info)
    Output
      $airlines
      $airlines$pk
      [1] "carrier"
      
      $airlines$uks
      # A tibble: 0 x 1
      # i 1 variable: column <list>
      
      $airlines$fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0800020b-0c07-030f-0a0e-0105060d0904 carrier       carrier        
      
      $airlines$fks_out
      # A tibble: 0 x 3
      # i 3 variables: child_fk_cols <keys>, parent_uuid <chr>,
      #   parent_key_cols <keys>
      
      $airlines$uuid
      [1] "0109020c-0b0a-030e-0d04-05060f070008"
      
      
      $airports
      $airports$pk
      [1] "faa"
      
      $airports$uks
      # A tibble: 0 x 1
      # i 1 variable: column <list>
      
      $airports$fks_in
      # A tibble: 2 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0800020b-0c07-030f-0a0e-0105060d0904 origin        faa            
      2 0800020b-0c07-030f-0a0e-0105060d0904 dest          faa            
      
      $airports$fks_out
      # A tibble: 0 x 3
      # i 3 variables: child_fk_cols <keys>, parent_uuid <chr>,
      #   parent_key_cols <keys>
      
      $airports$uuid
      [1] "04080601-0b0a-0c02-0503-0e070f0d0009"
      
      
      $flights
      $flights$pk
      NULL
      
      $flights$uks
      # A tibble: 0 x 1
      # i 1 variable: column <list>
      
      $flights$fks_in
      # A tibble: 0 x 3
      # i 3 variables: child_uuid <chr>, child_fk_cols <keys>, parent_key_cols <keys>
      
      $flights$fks_out
      # A tibble: 5 x 3
        child_fk_cols     parent_uuid                          parent_key_cols  
        <keys>            <chr>                                <keys>           
      1 carrier           0109020c-0b0a-030e-0d04-05060f070008 carrier          
      2 origin            04080601-0b0a-0c02-0503-0e070f0d0009 faa              
      3 dest              04080601-0b0a-0c02-0503-0e070f0d0009 faa              
      4 tailnum           0c0e080a-0307-0904-0b06-0205010f000d tailnum          
      5 origin, time_hour 0a090204-0108-0b00-0c0d-0705060e0f03 origin, time_hour
      
      $flights$uuid
      [1] "0800020b-0c07-030f-0a0e-0105060d0904"
      
      
      $planes
      $planes$pk
      [1] "tailnum"
      
      $planes$uks
      # A tibble: 0 x 1
      # i 1 variable: column <list>
      
      $planes$fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0800020b-0c07-030f-0a0e-0105060d0904 tailnum       tailnum        
      
      $planes$fks_out
      # A tibble: 0 x 3
      # i 3 variables: child_fk_cols <keys>, parent_uuid <chr>,
      #   parent_key_cols <keys>
      
      $planes$uuid
      [1] "0c0e080a-0307-0904-0b06-0205010f000d"
      
      
      $weather
      $weather$pk
      [1] "origin"    "time_hour"
      
      $weather$uks
      # A tibble: 0 x 1
      # i 1 variable: column <list>
      
      $weather$fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols     parent_key_cols  
        <chr>                                <keys>            <keys>           
      1 0800020b-0c07-030f-0a0e-0105060d0904 origin, time_hour origin, time_hour
      
      $weather$fks_out
      # A tibble: 0 x 3
      # i 3 variables: child_fk_cols <keys>, parent_uuid <chr>,
      #   parent_key_cols <keys>
      
      $weather$uuid
      [1] "0a090204-0108-0b00-0c0d-0705060e0f03"
      
      

# `new_keyed_tbl()` formatting

    Code
      keyed_tbl_impl(dm_nycflights13(cycle = TRUE), "flights")
    Output
      # A tibble: 1,761 x 19
      # Keys:     --- | 0 | 5
          year month   day dep_time sched_dep_time dep_delay arr_time sched_arr_time
         <int> <int> <int>    <int>          <int>     <dbl>    <int>          <int>
       1  2013     1    10        3           2359         4      426            437
       2  2013     1    10       16           2359        17      447            444
       3  2013     1    10      450            500       -10      634            648
       4  2013     1    10      520            525        -5      813            820
       5  2013     1    10      530            530         0      824            829
       6  2013     1    10      531            540        -9      832            850
       7  2013     1    10      535            540        -5     1015           1017
       8  2013     1    10      546            600       -14      645            709
       9  2013     1    10      549            600       -11      652            724
      10  2013     1    10      550            600       -10      649            703
      # i 1,751 more rows
      # i 11 more variables: arr_delay <dbl>, carrier <chr>, flight <int>,
      #   tailnum <chr>, origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>,
      #   hour <dbl>, minute <dbl>, time_hour <dttm>
    Code
      keyed_tbl_impl(dm_nycflights13(cycle = TRUE), "airports")
    Output
      # A tibble: 86 x 8
      # Keys:     `faa` | 2 | 0
         faa   name                                 lat    lon   alt    tz dst   tzone
         <chr> <chr>                              <dbl>  <dbl> <dbl> <dbl> <chr> <chr>
       1 ALB   Albany Intl                         42.7  -73.8   285    -5 A     Amer~
       2 ATL   Hartsfield Jackson Atlanta Intl     33.6  -84.4  1026    -5 A     Amer~
       3 AUS   Austin Bergstrom Intl               30.2  -97.7   542    -6 A     Amer~
       4 BDL   Bradley Intl                        41.9  -72.7   173    -5 A     Amer~
       5 BHM   Birmingham Intl                     33.6  -86.8   644    -6 A     Amer~
       6 BNA   Nashville Intl                      36.1  -86.7   599    -6 A     Amer~
       7 BOS   General Edward Lawrence Logan Intl  42.4  -71.0    19    -5 A     Amer~
       8 BTV   Burlington Intl                     44.5  -73.2   335    -5 A     Amer~
       9 BUF   Buffalo Niagara Intl                42.9  -78.7   724    -5 A     Amer~
      10 BUR   Bob Hope                            34.2 -118.    778    -8 A     Amer~
      # i 76 more rows
    Code
      keyed_tbl_impl(dm_nycflights13(cycle = TRUE), "airlines")
    Output
      # A tibble: 15 x 2
      # Keys:     `carrier` | 1 | 0
         carrier name                       
         <chr>   <chr>                      
       1 9E      Endeavor Air Inc.          
       2 AA      American Airlines Inc.     
       3 AS      Alaska Airlines Inc.       
       4 B6      JetBlue Airways            
       5 DL      Delta Air Lines Inc.       
       6 EV      ExpressJet Airlines Inc.   
       7 F9      Frontier Airlines Inc.     
       8 FL      AirTran Airways Corporation
       9 HA      Hawaiian Airlines Inc.     
      10 MQ      Envoy Air                  
      11 UA      United Air Lines Inc.      
      12 US      US Airways Inc.            
      13 VX      Virgin America             
      14 WN      Southwest Airlines Co.     
      15 YV      Mesa Airlines Inc.         

# `dm()` and `new_dm()` can handle a list of `dm_keyed_tbl` objects

    Code
      tbl_sum(keyed_tbl_impl(dm_output, "d1"))
    Output
                             A tibble                            Keys 
                           "144 x 15" "`origin`, `time_hour` | 0 | 0" 

---

    Code
      tbl_sum(keyed_tbl_impl(dm_output, "d2"))
    Output
             A tibble            Keys 
             "86 x 8" "`faa` | 0 | 0" 

---

    Code
      tbl_sum(keyed_tbl_impl(new_dm_output, "d1"))
    Output
                             A tibble                            Keys 
                           "144 x 15" "`origin`, `time_hour` | 0 | 0" 

---

    Code
      tbl_sum(keyed_tbl_impl(new_dm_output, "d2"))
    Output
             A tibble            Keys 
             "86 x 8" "`faa` | 0 | 0" 

# `dm()` and `new_dm()` can handle a mix of tables and `dm_keyed_tbl` objects

    Code
      tbl_sum(keyed_tbl_impl(dm_output, "d1"))
    Output
                             A tibble                            Keys 
                           "144 x 15" "`origin`, `time_hour` | 0 | 0" 

---

    Code
      tbl_sum(keyed_tbl_impl(dm_output, "d2"))
    Output
           A tibble          Keys 
           "86 x 8" "--- | 0 | 0" 

---

    Code
      tbl_sum(keyed_tbl_impl(new_dm_output, "d1"))
    Output
                             A tibble                            Keys 
                           "144 x 15" "`origin`, `time_hour` | 0 | 0" 

---

    Code
      tbl_sum(keyed_tbl_impl(new_dm_output, "d2"))
    Output
           A tibble          Keys 
           "86 x 8" "--- | 0 | 0" 

# `dm()` handles missing key column names gracefully

    Code
      dm(x = keyed$x["b"], y = keyed$y) %>% dm_paste()
    Message
      dm::dm(
        x,
        y,
      ) %>%
        dm::dm_add_pk(y, c(a, b))
    Code
      dm(x = keyed$x, y = keyed$y["b"]) %>% dm_paste()
    Message
      dm::dm(
        x,
        y,
      )

# keyed_by()

    Code
      keyed_by(x, y)
    Output
        a 
      "b" 
    Code
      keyed_by(y, x)
    Output
        b 
      "a" 

# keyed_by() failure modes

    Code
      keyed_by(x_no_fk, y_no_fk)
    Condition
      Error in `keyed_by()`:
      ! Can't infer `by`: foreign key information lost?

---

    Code
      keyed_by(x_both_fk, y_both_fk)
    Condition
      Error in `keyed_by()`:
      ! Can't infer `by`: foreign key available in both directions.

---

    Code
      keyed_by(x_multi_fk, y_multi_fk)
    Condition
      Error in `keyed_by()`:
      ! Can't infer `by`: multiple foreign keys available.

# joins without child PK

    Code
      keyed_build_join_spec(x, y) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            a
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 a     b    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      NULL
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             a              
      
      $new_fks_out
      # A tibble: 1 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 a             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "0800020b-0c07-030f-0a0e-0105060d0904"
      
    Code
      dm(x, y, r = left_join(x, y)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, a) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, a, y) %>%
        dm::dm_add_fk(x, a, r, a)
    Code
      keyed_build_join_spec(y, x) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 1
            a
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 b     a    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      NULL
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             b              
      
      $new_fks_out
      # A tibble: 1 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 b             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "03000c09-0a07-050d-020e-01040b08060f"
      
    Code
      dm(x, y, r = left_join(y, x)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, a) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, b) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, b, y) %>%
        dm::dm_add_fk(x, a, r, b)

# joins with other child PK

    Code
      keyed_build_join_spec(x, y) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 2
            a     c
        <dbl> <dbl>
      1     1     1
      
      $y_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 a     b    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "c"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             a              
      
      $new_fks_out
      # A tibble: 1 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 a             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "0800020b-0c07-030f-0a0e-0105060d0904"
      
    Code
      dm(x, y, r = left_join(x, y)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, a, c) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, a, c) %>%
        dm::dm_add_pk(x, c) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, c) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, a, y) %>%
        dm::dm_add_fk(x, a, r, a)
    Code
      keyed_build_join_spec(y, x) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 2
            a     c
        <dbl> <dbl>
      1     1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 b     a    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "c"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             b              
      
      $new_fks_out
      # A tibble: 1 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 b             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "03000c09-0a07-050d-020e-01040b08060f"
      
    Code
      dm(x, y, r = left_join(y, x)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, a, c) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, b, c) %>%
        dm::dm_add_pk(x, c) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, c) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, b, y) %>%
        dm::dm_add_fk(x, a, r, b)

# joins with other child PK and name conflict

    Code
      keyed_build_join_spec(x, y) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 2
            a     b
        <dbl> <dbl>
      1     1     1
      
      $y_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 a     b    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "b"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             a              
      
      $new_fks_out
      # A tibble: 1 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 a             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "0800020b-0c07-030f-0a0e-0105060d0904"
      
    Code
      dm(x, y, r = left_join(x, y)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, a, b) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, a, b) %>%
        dm::dm_add_pk(x, b) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, a, y) %>%
        dm::dm_add_fk(x, a, r, a)
    Code
      keyed_build_join_spec(y, x) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 2
            a     b
        <dbl> <dbl>
      1     1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 b     a    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "b"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             b              
      
      $new_fks_out
      # A tibble: 1 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 b             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "03000c09-0a07-050d-020e-01040b08060f"
      
    Code
      dm(x, y, r = left_join(y, x)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, a, b) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, b, b.y) %>%
        dm::dm_add_pk(x, b) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, b, y) %>%
        dm::dm_add_fk(x, a, r)

# joins with same child PK

    Code
      keyed_build_join_spec(x, y) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            a
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 a     b    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "a"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             a              
      
      $new_fks_out
      # A tibble: 1 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 a             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "0800020b-0c07-030f-0a0e-0105060d0904"
      
    Code
      dm(x, y, r = left_join(x, y)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, a) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, a) %>%
        dm::dm_add_pk(x, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, a) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, a, y) %>%
        dm::dm_add_fk(x, a, r)
    Code
      keyed_build_join_spec(y, x) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 1
            a
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 b     a    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "b"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             b              
      
      $new_fks_out
      # A tibble: 1 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 b             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "03000c09-0a07-050d-020e-01040b08060f"
      
    Code
      dm(x, y, r = left_join(y, x)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, a) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, b) %>%
        dm::dm_add_pk(x, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, b, y) %>%
        dm::dm_add_fk(x, a, r)

# joins with same child PK and same name

    Code
      keyed_build_join_spec(x, y) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 b     b    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "b"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 b             b              
      
      $new_fks_out
      # A tibble: 1 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 b             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "0800020b-0c07-030f-0a0e-0105060d0904"
      
    Code
      dm(x, y, r = left_join(x, y)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, b) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, b) %>%
        dm::dm_add_pk(x, b) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, b) %>%
        dm::dm_add_fk(x, b, y) %>%
        dm::dm_add_fk(r, b, y) %>%
        dm::dm_add_fk(x, b, r)
    Code
      keyed_build_join_spec(y, x) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 b     b    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "b"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 b             b              
      
      $new_fks_out
      # A tibble: 1 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 b             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "03000c09-0a07-050d-020e-01040b08060f"
      
    Code
      dm(x, y, r = left_join(y, x)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, b) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, b) %>%
        dm::dm_add_pk(x, b) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, b) %>%
        dm::dm_add_fk(x, b, y) %>%
        dm::dm_add_fk(r, b, y) %>%
        dm::dm_add_fk(x, b, r)

# joins with other FK from parent

    Code
      keyed_build_join_spec(x, y) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            a
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 2
            b     c
        <dbl> <dbl>
      1     1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 a     b    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "a"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             a              
      
      $new_fks_out
      # A tibble: 2 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 a             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      2 c             0800020b-0c07-030f-0a0e-0105060d0904 c              
      
      $new_uuid
      [1] "0c0e080a-0307-0904-0b06-0205010f000d"
      
    Code
      dm(x, y, z, r = left_join(x, y)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        z,
        r,
      ) %>%
        dm::dm_select(x, a) %>%
        dm::dm_select(y, b, c) %>%
        dm::dm_select(z, c) %>%
        dm::dm_select(r, a, c) %>%
        dm::dm_add_pk(x, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, a) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, a, y) %>%
        dm::dm_add_fk(y, c, z, c) %>%
        dm::dm_add_fk(r, c, z, c) %>%
        dm::dm_add_fk(x, a, r)
    Code
      keyed_build_join_spec(y, x) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 2
            b     c
        <dbl> <dbl>
      1     1     1
      
      $y_tbl
      # A tibble: 1 x 1
            a
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 b     a    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "b"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             b              
      
      $new_fks_out
      # A tibble: 2 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 c             0800020b-0c07-030f-0a0e-0105060d0904 c              
      2 b             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "0108090a-0403-0f05-0d0c-0b0006020e07"
      
    Code
      dm(x, y, z, r = left_join(y, x)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        z,
        r,
      ) %>%
        dm::dm_select(x, a) %>%
        dm::dm_select(y, b, c) %>%
        dm::dm_select(z, c) %>%
        dm::dm_select(r, b, c) %>%
        dm::dm_add_pk(x, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, b, y) %>%
        dm::dm_add_fk(y, c, z, c) %>%
        dm::dm_add_fk(r, c, z, c) %>%
        dm::dm_add_fk(x, a, r)

# joins with other FK from parent and name conflict

    Code
      keyed_build_join_spec(x, y) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            a
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 2
            b     a
        <dbl> <dbl>
      1     1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 a     b    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "a"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             a              
      
      $new_fks_out
      # A tibble: 2 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 a             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      2 a             0800020b-0c07-030f-0a0e-0105060d0904 a              
      
      $new_uuid
      [1] "0c0e080a-0307-0904-0b06-0205010f000d"
      
    Code
      dm(x, y, z, r = left_join(x, y)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        z,
        r,
      ) %>%
        dm::dm_select(x, a) %>%
        dm::dm_select(y, b, a) %>%
        dm::dm_select(z, a) %>%
        dm::dm_select(r, a, a.y) %>%
        dm::dm_add_pk(x, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, a) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, a, y) %>%
        dm::dm_add_fk(y, a, z, a) %>%
        dm::dm_add_fk(r, a, z, a) %>%
        dm::dm_add_fk(x, a, r)
    Code
      keyed_build_join_spec(y, x) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 2
            b     a
        <dbl> <dbl>
      1     1     1
      
      $y_tbl
      # A tibble: 1 x 1
            a
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 b     a    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "b"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             b              
      
      $new_fks_out
      # A tibble: 2 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 a             0800020b-0c07-030f-0a0e-0105060d0904 a              
      2 b             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      
      $new_uuid
      [1] "0108090a-0403-0f05-0d0c-0b0006020e07"
      
    Code
      dm(x, y, z, r = left_join(y, x)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        z,
        r,
      ) %>%
        dm::dm_select(x, a) %>%
        dm::dm_select(y, b, a) %>%
        dm::dm_select(z, a) %>%
        dm::dm_select(r, b, a) %>%
        dm::dm_add_pk(x, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, b, y) %>%
        dm::dm_add_fk(y, a, z, a) %>%
        dm::dm_add_fk(r, a, z, a) %>%
        dm::dm_add_fk(x, a, r)

# joins with other FK from child

    Code
      keyed_build_join_spec(x, y) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 2
            a     c
        <dbl> <dbl>
      1     1     1
      
      $y_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 a     b    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "a"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             a              
      
      $new_fks_out
      # A tibble: 2 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 a             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      2 c             0800020b-0c07-030f-0a0e-0105060d0904 c              
      
      $new_uuid
      [1] "0c0e080a-0307-0904-0b06-0205010f000d"
      
    Code
      dm(x, y, z, r = left_join(x, y)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        z,
        r,
      ) %>%
        dm::dm_select(x, a, c) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(z, c) %>%
        dm::dm_select(r, a, c) %>%
        dm::dm_add_pk(x, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, a) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, a, y) %>%
        dm::dm_add_fk(x, c, z, c) %>%
        dm::dm_add_fk(r, c, z, c) %>%
        dm::dm_add_fk(x, a, r)
    Code
      keyed_build_join_spec(y, x) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 2
            a     c
        <dbl> <dbl>
      1     1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 b     a    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "b"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             b              
      
      $new_fks_out
      # A tibble: 2 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 b             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      2 c             0800020b-0c07-030f-0a0e-0105060d0904 c              
      
      $new_uuid
      [1] "0108090a-0403-0f05-0d0c-0b0006020e07"
      
    Code
      dm(x, y, z, r = left_join(y, x)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        z,
        r,
      ) %>%
        dm::dm_select(x, a, c) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(z, c) %>%
        dm::dm_select(r, b, c) %>%
        dm::dm_add_pk(x, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, b, y) %>%
        dm::dm_add_fk(x, c, z, c) %>%
        dm::dm_add_fk(r, c, z, c) %>%
        dm::dm_add_fk(x, a, r)

# joins with other FK from child and name conflict

    Code
      keyed_build_join_spec(x, y) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 2
            a     b
        <dbl> <dbl>
      1     1     1
      
      $y_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 a     b    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "a"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             a              
      
      $new_fks_out
      # A tibble: 2 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 a             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      2 b             0800020b-0c07-030f-0a0e-0105060d0904 b              
      
      $new_uuid
      [1] "0c0e080a-0307-0904-0b06-0205010f000d"
      
    Code
      dm(x, y, z, r = left_join(x, y)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        z,
        r,
      ) %>%
        dm::dm_select(x, a, b) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(z, b) %>%
        dm::dm_select(r, a, b) %>%
        dm::dm_add_pk(x, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, a) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, a, y) %>%
        dm::dm_add_fk(x, b, z, b) %>%
        dm::dm_add_fk(r, b, z, b) %>%
        dm::dm_add_fk(x, a, r)
    Code
      keyed_build_join_spec(y, x) %>% to_snapshot_json()
    Output
      $x_tbl
      # A tibble: 1 x 1
            b
        <dbl>
      1     1
      
      $y_tbl
      # A tibble: 1 x 2
            a     b
        <dbl> <dbl>
      1     1     1
      
      $by
      # A tibble: 1 x 2
        x     y    
        <chr> <chr>
      1 b     a    
      
      $suffix
      [1] ".x" ".y"
      
      $new_pk
      [1] "b"
      
      $new_fks_in
      # A tibble: 1 x 3
        child_uuid                           child_fk_cols parent_key_cols
        <chr>                                <keys>        <keys>         
      1 0109020c-0b0a-030e-0d04-05060f070008 a             b              
      
      $new_fks_out
      # A tibble: 2 x 3
        child_fk_cols parent_uuid                          parent_key_cols
        <keys>        <chr>                                <keys>         
      1 b             04080601-0b0a-0c02-0503-0e070f0d0009 b              
      2 b             0800020b-0c07-030f-0a0e-0105060d0904 b              
      
      $new_uuid
      [1] "0108090a-0403-0f05-0d0c-0b0006020e07"
      
    Code
      dm(x, y, z, r = left_join(y, x)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        z,
        r,
      ) %>%
        dm::dm_select(x, a, b) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(z, b) %>%
        dm::dm_select(r, b, b.y) %>%
        dm::dm_add_pk(x, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, b, y) %>%
        dm::dm_add_fk(x, b, z, b) %>%
        dm::dm_add_fk(r, b, z, b) %>%
        dm::dm_add_fk(x, a, r)

# left join works as expected with keyed tables

    Code
      dm <- dm_nycflights13()
      keyed_tbl_impl(dm, "weather") %>% left_join(keyed_tbl_impl(dm, "flights"),
      multiple = "all")
    Output
      # A tibble: 1,800 x 32
      # Keys:     --- | 1 | 4
         origin year.x month.x day.x hour.x  temp  dewp humid wind_dir wind_speed
         <chr>   <int>   <int> <int>  <int> <dbl> <dbl> <dbl>    <dbl>      <dbl>
       1 EWR      2013       1    10      0  41    32    70.1      230       8.06
       2 EWR      2013       1    10      1  39.0  30.0  69.9      210       9.21
       3 EWR      2013       1    10      2  39.0  28.9  66.8      230       6.90
       4 EWR      2013       1    10      3  39.9  27.0  59.5      270       5.75
       5 EWR      2013       1    10      4  41    26.1  55.0      320       6.90
       6 EWR      2013       1    10      5  41    26.1  55.0      300      12.7 
       7 EWR      2013       1    10      5  41    26.1  55.0      300      12.7 
       8 EWR      2013       1    10      6  39.9  25.0  54.8      280       6.90
       9 EWR      2013       1    10      6  39.9  25.0  54.8      280       6.90
      10 EWR      2013       1    10      6  39.9  25.0  54.8      280       6.90
      # i 1,790 more rows
      # i 22 more variables: wind_gust <dbl>, precip <dbl>, pressure <dbl>,
      #   visib <dbl>, time_hour <dttm>, year.y <int>, month.y <int>, day.y <int>,
      #   dep_time <int>, sched_dep_time <int>, dep_delay <dbl>, arr_time <int>,
      #   sched_arr_time <int>, arr_delay <dbl>, carrier <chr>, flight <int>,
      #   tailnum <chr>, dest <chr>, air_time <dbl>, distance <dbl>, hour.y <dbl>,
      #   minute <dbl>

# semi_join()

    Code
      dm(x, y, r = semi_join(x, y)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, a) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, a) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(r, a, y)
    Code
      dm(x, y, r = semi_join(y, x)) %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        x,
        y,
        r,
      ) %>%
        dm::dm_select(x, a) %>%
        dm::dm_select(y, b) %>%
        dm::dm_select(r, b) %>%
        dm::dm_add_pk(y, b) %>%
        dm::dm_add_pk(r, b) %>%
        dm::dm_add_fk(x, a, y) %>%
        dm::dm_add_fk(x, a, r)

# arrange for keyed tables produces expected output

    Code
      keyed_tbl_impl(dm, "airlines") %>% arrange(desc(name))
    Output
      # A tibble: 15 x 2
      # Keys:     `carrier` | 1 | 0
         carrier name                       
         <chr>   <chr>                      
       1 VX      Virgin America             
       2 UA      United Air Lines Inc.      
       3 US      US Airways Inc.            
       4 WN      Southwest Airlines Co.     
       5 YV      Mesa Airlines Inc.         
       6 B6      JetBlue Airways            
       7 HA      Hawaiian Airlines Inc.     
       8 F9      Frontier Airlines Inc.     
       9 EV      ExpressJet Airlines Inc.   
      10 MQ      Envoy Air                  
      11 9E      Endeavor Air Inc.          
      12 DL      Delta Air Lines Inc.       
      13 AA      American Airlines Inc.     
      14 AS      Alaska Airlines Inc.       
      15 FL      AirTran Airways Corporation

# group_by for keyed tables produces expected output

    Code
      dm <- dm_nycflights13(cycle = TRUE)
      keyed_tbl_impl(dm, "flights") %>% group_by(month)
    Output
      # A tibble: 1,761 x 19
      # Groups:   month [2]
      # Keys:     --- | 0 | 5
          year month   day dep_time sched_dep_time dep_delay arr_time sched_arr_time
         <int> <int> <int>    <int>          <int>     <dbl>    <int>          <int>
       1  2013     1    10        3           2359         4      426            437
       2  2013     1    10       16           2359        17      447            444
       3  2013     1    10      450            500       -10      634            648
       4  2013     1    10      520            525        -5      813            820
       5  2013     1    10      530            530         0      824            829
       6  2013     1    10      531            540        -9      832            850
       7  2013     1    10      535            540        -5     1015           1017
       8  2013     1    10      546            600       -14      645            709
       9  2013     1    10      549            600       -11      652            724
      10  2013     1    10      550            600       -10      649            703
      # i 1,751 more rows
      # i 11 more variables: arr_delay <dbl>, carrier <chr>, flight <int>,
      #   tailnum <chr>, origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>,
      #   hour <dbl>, minute <dbl>, time_hour <dttm>
    Code
      keyed_tbl_impl(dm, "airports") %>% group_by(tzone)
    Output
      # A tibble: 86 x 8
      # Groups:   tzone [6]
      # Keys:     `faa` | 2 | 0
         faa   name                                 lat    lon   alt    tz dst   tzone
         <chr> <chr>                              <dbl>  <dbl> <dbl> <dbl> <chr> <chr>
       1 ALB   Albany Intl                         42.7  -73.8   285    -5 A     Amer~
       2 ATL   Hartsfield Jackson Atlanta Intl     33.6  -84.4  1026    -5 A     Amer~
       3 AUS   Austin Bergstrom Intl               30.2  -97.7   542    -6 A     Amer~
       4 BDL   Bradley Intl                        41.9  -72.7   173    -5 A     Amer~
       5 BHM   Birmingham Intl                     33.6  -86.8   644    -6 A     Amer~
       6 BNA   Nashville Intl                      36.1  -86.7   599    -6 A     Amer~
       7 BOS   General Edward Lawrence Logan Intl  42.4  -71.0    19    -5 A     Amer~
       8 BTV   Burlington Intl                     44.5  -73.2   335    -5 A     Amer~
       9 BUF   Buffalo Niagara Intl                42.9  -78.7   724    -5 A     Amer~
      10 BUR   Bob Hope                            34.2 -118.    778    -8 A     Amer~
      # i 76 more rows
    Code
      keyed_tbl_impl(dm, "airports") %>% group_by(faa)
    Output
      # A tibble: 86 x 8
      # Groups:   faa [86]
      # Keys:     `faa` | 2 | 0
         faa   name                                 lat    lon   alt    tz dst   tzone
         <chr> <chr>                              <dbl>  <dbl> <dbl> <dbl> <chr> <chr>
       1 ALB   Albany Intl                         42.7  -73.8   285    -5 A     Amer~
       2 ATL   Hartsfield Jackson Atlanta Intl     33.6  -84.4  1026    -5 A     Amer~
       3 AUS   Austin Bergstrom Intl               30.2  -97.7   542    -6 A     Amer~
       4 BDL   Bradley Intl                        41.9  -72.7   173    -5 A     Amer~
       5 BHM   Birmingham Intl                     33.6  -86.8   644    -6 A     Amer~
       6 BNA   Nashville Intl                      36.1  -86.7   599    -6 A     Amer~
       7 BOS   General Edward Lawrence Logan Intl  42.4  -71.0    19    -5 A     Amer~
       8 BTV   Burlington Intl                     44.5  -73.2   335    -5 A     Amer~
       9 BUF   Buffalo Niagara Intl                42.9  -78.7   724    -5 A     Amer~
      10 BUR   Bob Hope                            34.2 -118.    778    -8 A     Amer~
      # i 76 more rows

# summarize for keyed tables produces expected output

    Code
      dm <- dm_nycflights13(cycle = TRUE)
      keyed_tbl_impl(dm, "airports") %>% summarise(mean_alt = mean(alt))
    Output
      # A tibble: 1 x 1
      # Keys:     --- | 0 | 0
        mean_alt
           <dbl>
      1     632.
    Code
      keyed_tbl_impl(dm, "airports") %>% group_by(tzone, dst) %>% summarise(mean_alt = mean(
        alt))
    Output
      # A tibble: 6 x 3
      # Groups:   tzone [6]
      # Keys:     `tzone`, `dst` | 0 | 0
        tzone               dst   mean_alt
        <chr>               <chr>    <dbl>
      1 America/Chicago     A         680.
      2 America/Denver      A        5399.
      3 America/Los_Angeles A         313.
      4 America/New_York    A         396.
      5 America/Phoenix     N        1135 
      6 Pacific/Honolulu    N          13 

# pks_df_from_keys_info()

    Code
      dm %>% dm_get_keyed_tables_impl() %>% pks_df_from_keys_info() %>%
        to_snapshot_json()
    Output
      # A tibble: 5 x 2
        table    pks             
        <chr>    <list>          
      1 airlines <tibble [1 x 2]>
      2 airports <tibble [1 x 2]>
      3 flights  <NULL>          
      4 planes   <tibble [1 x 2]>
      5 weather  <tibble [1 x 2]>

# uks_df_from_keys_info()

    Code
      dm %>% dm_get_keyed_tables_impl() %>% uks_df_from_keys_info() %>%
        to_snapshot_json()
    Output
      # A tibble: 6 x 2
        table uks             
        <chr> <list>          
      1 tf_1  <tibble [0 x 1]>
      2 tf_2  <tibble [0 x 1]>
      3 tf_3  <tibble [1 x 1]>
      4 tf_4  <tibble [0 x 1]>
      5 tf_5  <tibble [1 x 1]>
      6 tf_6  <tibble [0 x 1]>

# fks_df_from_keys_info()

    Code
      dm %>% dm_get_keyed_tables_impl() %>% fks_df_from_keys_info() %>%
        to_snapshot_json()
    Output
      # A tibble: 4 x 2
        table    fks             
        <chr>    <list>          
      1 airlines <tibble [1 x 4]>
      2 airports <tibble [2 x 4]>
      3 planes   <tibble [1 x 4]>
      4 weather  <tibble [1 x 4]>

# dummy

    Code
      # dummy

# dm_sql()

    Code
      dm_for_filter_df() %>% dm_sql(my_test_con())
    Output
      $pre
      $pre$tf_1
      <SQL> CREATE TEMPORARY TABLE "tf_1" (
        "a" INT IDENTITY,
        "b" varchar(255),
        PRIMARY KEY ("a")
      )
      
      $pre$tf_2
      <SQL> CREATE TEMPORARY TABLE "tf_2" (
        "c" varchar(255),
        "d" INT,
        "e" varchar(255),
        "e1" INT,
        PRIMARY KEY ("c")
      )
      
      $pre$tf_3
      <SQL> CREATE TEMPORARY TABLE "tf_3" (
        "f" varchar(255),
        "f1" INT,
        "g" varchar(255),
        PRIMARY KEY ("f", "f1")
      )
      
      $pre$tf_4
      <SQL> CREATE TEMPORARY TABLE "tf_4" (
        "h" varchar(255),
        "i" varchar(255),
        "j" varchar(255),
        "j1" INT,
        PRIMARY KEY ("h")
      )
      
      $pre$tf_5
      <SQL> CREATE TEMPORARY TABLE "tf_5" (
        "ww" INT,
        "k" INT,
        "l" varchar(255),
        "m" varchar(255),
        PRIMARY KEY ("k")
      )
      
      $pre$tf_6
      <SQL> CREATE TEMPORARY TABLE "tf_6" (
        "zz" INT,
        "n" varchar(255),
        "o" varchar(255),
        PRIMARY KEY ("o")
      )
      
      
      $load
      $load$tf_1
      <SQL> SET IDENTITY_INSERT "tf_1" ON
      INSERT INTO "tf_1" ("a", "b")
      SELECT
        TRY_CAST(TRY_CAST("a" AS NUMERIC) AS INT) AS "a",
        TRY_CAST("b" AS VARCHAR(MAX)) AS "b"
      FROM (
        VALUES
          (1, 'A'),
          (2, 'B'),
          (3, 'C'),
          (4, 'D'),
          (5, 'E'),
          (6, 'F'),
          (7, 'G'),
          (8, 'H'),
          (9, 'I'),
          (10, 'J')
      ) AS drvd("a", "b")
      SET IDENTITY_INSERT "tf_1" OFF
      
      $load$tf_2
      <SQL> INSERT INTO "tf_2" ("c", "d", "e", "e1")
      SELECT
        TRY_CAST("c" AS VARCHAR(MAX)) AS "c",
        TRY_CAST(TRY_CAST("d" AS NUMERIC) AS INT) AS "d",
        TRY_CAST("e" AS VARCHAR(MAX)) AS "e",
        TRY_CAST(TRY_CAST("e1" AS NUMERIC) AS INT) AS "e1"
      FROM (
        VALUES
          ('cat', 7, 'F', 6),
          ('dog', 6, 'E', 5),
          ('elephant', 2, 'D', 4),
          ('lion', 3, 'E', 5),
          ('seal', 4, 'F', 6),
          ('worm', 5, 'G', 7)
      ) AS drvd("c", "d", "e", "e1")
      
      $load$tf_3
      <SQL> INSERT INTO "tf_3" ("f", "f1", "g")
      SELECT
        TRY_CAST("f" AS VARCHAR(MAX)) AS "f",
        TRY_CAST(TRY_CAST("f1" AS NUMERIC) AS INT) AS "f1",
        TRY_CAST("g" AS VARCHAR(MAX)) AS "g"
      FROM (
        VALUES
          ('C', 2, 'one'),
          ('C', 3, 'two'),
          ('D', 4, 'three'),
          ('E', 5, 'four'),
          ('F', 6, 'five'),
          ('G', 7, 'six'),
          ('H', 7, 'seven'),
          ('I', 7, 'eight'),
          ('J', 10, 'nine'),
          ('K', 11, 'ten')
      ) AS drvd("f", "f1", "g")
      
      $load$tf_4
      <SQL> INSERT INTO "tf_4" ("h", "i", "j", "j1")
      SELECT
        TRY_CAST("h" AS VARCHAR(MAX)) AS "h",
        TRY_CAST("i" AS VARCHAR(MAX)) AS "i",
        TRY_CAST("j" AS VARCHAR(MAX)) AS "j",
        TRY_CAST(TRY_CAST("j1" AS NUMERIC) AS INT) AS "j1"
      FROM (
        VALUES
          ('a', 'three', 'C', 3),
          ('b', 'four', 'D', 4),
          ('c', 'five', 'E', 5),
          ('d', 'six', 'F', 6),
          ('e', 'seven', 'F', 6)
      ) AS drvd("h", "i", "j", "j1")
      
      $load$tf_5
      <SQL> INSERT INTO "tf_5" ("ww", "k", "l", "m")
      SELECT
        TRY_CAST(TRY_CAST("ww" AS NUMERIC) AS INT) AS "ww",
        TRY_CAST(TRY_CAST("k" AS NUMERIC) AS INT) AS "k",
        TRY_CAST("l" AS VARCHAR(MAX)) AS "l",
        TRY_CAST("m" AS VARCHAR(MAX)) AS "m"
      FROM (
        VALUES
          (2, 1, 'b', 'house'),
          (2, 2, 'c', 'tree'),
          (2, 3, 'd', 'streetlamp'),
          (2, 4, 'e', 'streetlamp')
      ) AS drvd("ww", "k", "l", "m")
      
      $load$tf_6
      <SQL> INSERT INTO "tf_6" ("zz", "n", "o")
      SELECT
        TRY_CAST(TRY_CAST("zz" AS NUMERIC) AS INT) AS "zz",
        TRY_CAST("n" AS VARCHAR(MAX)) AS "n",
        TRY_CAST("o" AS VARCHAR(MAX)) AS "o"
      FROM (
        VALUES
          (1, 'garden', 'i'),
          (1, 'hill', 'g'),
          (1, 'house', 'e'),
          (1, 'streetlamp', 'h'),
          (1, 'tree', 'f')
      ) AS drvd("zz", "n", "o")
      
      
      $post
      $post$uk
      $post$uk$tf_1
      <SQL> ALTER TABLE "tf_1" ADD UNIQUE ("a")
      
      $post$uk$tf_2
      <SQL> ALTER TABLE "tf_2" ADD UNIQUE ("c")
      
      $post$uk$tf_3
      <SQL> ALTER TABLE "tf_3" ADD UNIQUE ("f", "f1")
      <SQL> ALTER TABLE "tf_3" ADD UNIQUE ("g")
      
      $post$uk$tf_4
      <SQL> ALTER TABLE "tf_4" ADD UNIQUE ("h")
      
      $post$uk$tf_5
      <SQL> ALTER TABLE "tf_5" ADD UNIQUE ("k")
      
      $post$uk$tf_6
      <SQL> ALTER TABLE "tf_6" ADD UNIQUE ("o")
      <SQL> ALTER TABLE "tf_6" ADD UNIQUE ("n")
      
      
      $post$fk
      $post$fk$tf_2
      <SQL> ALTER TABLE "tf_2" ADD FOREIGN KEY ("d") REFERENCES "tf_1" ("a")
      <SQL> ALTER TABLE "tf_2" ADD FOREIGN KEY ("e", "e1") REFERENCES "tf_3" ("f", "f1")
      
      $post$fk$tf_4
      <SQL> ALTER TABLE "tf_4" ADD FOREIGN KEY ("j", "j1") REFERENCES "tf_3" ("f", "f1")
      
      $post$fk$tf_5
      <SQL> ALTER TABLE "tf_5" ADD FOREIGN KEY ("l") REFERENCES "tf_4" ("h") ON DELETE CASCADE
      <SQL> ALTER TABLE "tf_5" ADD FOREIGN KEY ("m") REFERENCES "tf_6" ("n")
      
      
      $post$indexes
      $post$indexes$tf_2
      <SQL> CREATE INDEX tf_2__d ON "tf_2" ("d")
      <SQL> CREATE INDEX tf_2__e_e1 ON "tf_2" ("e", "e1")
      
      $post$indexes$tf_4
      <SQL> CREATE INDEX tf_4__j_j1 ON "tf_4" ("j", "j1")
      
      $post$indexes$tf_5
      <SQL> CREATE INDEX tf_5__l ON "tf_5" ("l")
      <SQL> CREATE INDEX tf_5__m ON "tf_5" ("m")
      
      
      

---

    Code
      dm(x = data.frame(a = strrep("x", 300))) %>% dm_sql(my_test_con())
    Output
      $pre
      $pre$x
      <SQL> CREATE TEMPORARY TABLE "x" (
        "a" varchar(300)
      )
      
      
      $load
      $load$x
      <SQL> INSERT INTO "x" ("a")
      SELECT TRY_CAST("a" AS VARCHAR(MAX)) AS "a"
      FROM (  VALUES ('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx')) AS drvd("a")
      
      
      $post
      $post$uk
      named list()
      
      $post$fk
      named list()
      
      $post$indexes
      named list()
      
      

---

    Code
      dm(x = data.frame(a = strrep("x", 10000))) %>% dm_sql(my_test_con())
    Output
      $pre
      $pre$x
      <SQL> CREATE TEMPORARY TABLE "x" (
        "a" varchar(max)
      )
      
      
      $load
      $load$x
      <SQL> INSERT INTO "x" ("a")
      SELECT TRY_CAST("a" AS VARCHAR(MAX)) AS "a"
      FROM (  VALUES ('[encoded data, 10000 characters]')) AS drvd("a")
      
      
      $post
      $post$uk
      named list()
      
      $post$fk
      named list()
      
      $post$indexes
      named list()
      
      

# `dm_set_colors()` works

    Code
      dm_nycflights_small() %>% dm_set_colors(blue = "flights", green = "airports") %>%
        dm_get_colors()
    Output
         default  #00FF00FF  #0000FFFF    default    default 
      "airlines" "airports"  "flights"   "planes"  "weather" 

# output

    Code
      # empty
      empty_dm() %>% dm_paste()
    Message
      dm::dm(
      )
    Code
      # empty table
      dm(a = tibble()) %>% dm_paste(options = "tables")
    Message
      a <- tibble::tibble(
      )
      dm::dm(
        a,
      )
    Code
      # baseline
      dm_for_filter() %>% dm_paste()
    Message
      dm::dm(
        tf_1,
        tf_2,
        tf_3,
        tf_4,
        tf_5,
        tf_6,
      ) %>%
        dm::dm_add_pk(tf_1, a, autoincrement = TRUE) %>%
        dm::dm_add_pk(tf_2, c) %>%
        dm::dm_add_pk(tf_3, c(f, f1)) %>%
        dm::dm_add_pk(tf_4, h) %>%
        dm::dm_add_pk(tf_5, k) %>%
        dm::dm_add_pk(tf_6, o) %>%
        dm::dm_add_uk(tf_3, g) %>%
        dm::dm_add_fk(tf_2, d, tf_1) %>%
        dm::dm_add_fk(tf_2, c(e, e1), tf_3) %>%
        dm::dm_add_fk(tf_4, c(j, j1), tf_3) %>%
        dm::dm_add_fk(tf_5, l, tf_4, on_delete = "cascade") %>%
        dm::dm_add_fk(tf_5, m, tf_6, n)
    Code
      # changing the tab width
      dm_for_filter() %>% dm_paste(tab_width = 4)
    Message
      dm::dm(
          tf_1,
          tf_2,
          tf_3,
          tf_4,
          tf_5,
          tf_6,
      ) %>%
          dm::dm_add_pk(tf_1, a, autoincrement = TRUE) %>%
          dm::dm_add_pk(tf_2, c) %>%
          dm::dm_add_pk(tf_3, c(f, f1)) %>%
          dm::dm_add_pk(tf_4, h) %>%
          dm::dm_add_pk(tf_5, k) %>%
          dm::dm_add_pk(tf_6, o) %>%
          dm::dm_add_uk(tf_3, g) %>%
          dm::dm_add_fk(tf_2, d, tf_1) %>%
          dm::dm_add_fk(tf_2, c(e, e1), tf_3) %>%
          dm::dm_add_fk(tf_4, c(j, j1), tf_3) %>%
          dm::dm_add_fk(tf_5, l, tf_4, on_delete = "cascade") %>%
          dm::dm_add_fk(tf_5, m, tf_6, n)
    Code
      # we don't care if the tables really exist
      dm_for_filter() %>% dm_rename_tbl(tf_1_new = tf_1) %>% dm_paste()
    Message
      dm::dm(
        tf_1_new,
        tf_2,
        tf_3,
        tf_4,
        tf_5,
        tf_6,
      ) %>%
        dm::dm_add_pk(tf_1_new, a, autoincrement = TRUE) %>%
        dm::dm_add_pk(tf_2, c) %>%
        dm::dm_add_pk(tf_3, c(f, f1)) %>%
        dm::dm_add_pk(tf_4, h) %>%
        dm::dm_add_pk(tf_5, k) %>%
        dm::dm_add_pk(tf_6, o) %>%
        dm::dm_add_uk(tf_3, g) %>%
        dm::dm_add_fk(tf_2, d, tf_1_new) %>%
        dm::dm_add_fk(tf_2, c(e, e1), tf_3) %>%
        dm::dm_add_fk(tf_4, c(j, j1), tf_3) %>%
        dm::dm_add_fk(tf_5, l, tf_4, on_delete = "cascade") %>%
        dm::dm_add_fk(tf_5, m, tf_6, n)
    Code
      # produce `dm_select()` statements in addition to the rest
      dm_for_filter() %>% dm_select(tf_5, k = k, m) %>% dm_select(tf_1, a) %>% dm(x = copy_to_my_test_src(
        tibble(q = 1L), qq)) %>% dm_paste(options = "select")
    Message
      dm::dm(
        tf_1,
        tf_2,
        tf_3,
        tf_4,
        tf_5,
        tf_6,
        x,
      ) %>%
        dm::dm_select(tf_1, a) %>%
        dm::dm_select(tf_2, c, d, e, e1) %>%
        dm::dm_select(tf_3, f, f1, g) %>%
        dm::dm_select(tf_4, h, i, j, j1) %>%
        dm::dm_select(tf_5, k, m) %>%
        dm::dm_select(tf_6, zz, n, o) %>%
        dm::dm_select(x, q)
    Code
      # produce code with colors
      dm_for_filter() %>% dm_set_colors(orange = tf_1:tf_3, darkgreen = tf_5:tf_6) %>%
        dm_paste()
    Message
      dm::dm(
        tf_1,
        tf_2,
        tf_3,
        tf_4,
        tf_5,
        tf_6,
      ) %>%
        dm::dm_add_pk(tf_1, a, autoincrement = TRUE) %>%
        dm::dm_add_pk(tf_2, c) %>%
        dm::dm_add_pk(tf_3, c(f, f1)) %>%
        dm::dm_add_pk(tf_4, h) %>%
        dm::dm_add_pk(tf_5, k) %>%
        dm::dm_add_pk(tf_6, o) %>%
        dm::dm_add_uk(tf_3, g) %>%
        dm::dm_add_fk(tf_2, d, tf_1) %>%
        dm::dm_add_fk(tf_2, c(e, e1), tf_3) %>%
        dm::dm_add_fk(tf_4, c(j, j1), tf_3) %>%
        dm::dm_add_fk(tf_5, l, tf_4, on_delete = "cascade") %>%
        dm::dm_add_fk(tf_5, m, tf_6, n) %>%
        dm::dm_set_colors(`#FFA500FF` = tf_1) %>%
        dm::dm_set_colors(`#FFA500FF` = tf_2) %>%
        dm::dm_set_colors(`#FFA500FF` = tf_3) %>%
        dm::dm_set_colors(`#006400FF` = tf_5) %>%
        dm::dm_set_colors(`#006400FF` = tf_6)
    Code
      # tick if needed
      a <- tibble(x = 1)
      names(a) <- "a b"
      dm(a) %>% dm_zoom_to(a) %>% dm_insert_zoomed("a b") %>% dm_add_pk(a, "a b") %>%
        dm_add_fk("a b", "a b", a) %>% dm_set_colors(green = "a b") %>% dm_paste(
        options = "all")
    Message
      a <- tibble::tibble(
        `a b` = numeric(0),
      )
      `a b` <- tibble::tibble(
        `a b` = numeric(0),
      )
      dm::dm(
        a,
        `a b`,
      ) %>%
        dm::dm_add_pk(a, `a b`) %>%
        dm::dm_add_fk(`a b`, `a b`, a) %>%
        dm::dm_set_colors(`#00FF00FF` = `a b`)
    Code
      # FK referencing non default PK
      b <- tibble(x = 1, y = "A", z = "A")
      c <- tibble(x = "A", y = "A")
      dm(b, c) %>% dm_add_pk(c, x) %>% dm_add_fk(b, y, c) %>% dm_add_fk(b, z, c, y) %>%
        dm_paste()
    Message
      dm::dm(
        b,
        c,
      ) %>%
        dm::dm_add_pk(c, x) %>%
        dm::dm_add_fk(b, y, c) %>%
        dm::dm_add_fk(b, z, c, y)
    Code
      dm_for_filter() %>% dm_add_uk(tf_5, l) %>% dm_add_uk(tf_6, n) %>% dm_paste()
    Message
      dm::dm(
        tf_1,
        tf_2,
        tf_3,
        tf_4,
        tf_5,
        tf_6,
      ) %>%
        dm::dm_add_pk(tf_1, a, autoincrement = TRUE) %>%
        dm::dm_add_pk(tf_2, c) %>%
        dm::dm_add_pk(tf_3, c(f, f1)) %>%
        dm::dm_add_pk(tf_4, h) %>%
        dm::dm_add_pk(tf_5, k) %>%
        dm::dm_add_pk(tf_6, o) %>%
        dm::dm_add_uk(tf_3, g) %>%
        dm::dm_add_uk(tf_5, l) %>%
        dm::dm_add_uk(tf_6, n) %>%
        dm::dm_add_fk(tf_2, d, tf_1) %>%
        dm::dm_add_fk(tf_2, c(e, e1), tf_3) %>%
        dm::dm_add_fk(tf_4, c(j, j1), tf_3) %>%
        dm::dm_add_fk(tf_5, l, tf_4, on_delete = "cascade") %>%
        dm::dm_add_fk(tf_5, m, tf_6, n)
    Code
      # on_delete if needed
      dm(b, c) %>% dm_add_pk(c, x) %>% dm_add_fk(b, y, c, on_delete = "cascade") %>%
        dm_add_fk(b, z, c, y, on_delete = "no_action") %>% dm_paste()
    Message
      dm::dm(
        b,
        c,
      ) %>%
        dm::dm_add_pk(c, x) %>%
        dm::dm_add_fk(b, y, c, on_delete = "cascade") %>%
        dm::dm_add_fk(b, z, c, y)
    Code
      # all of nycflights13
      dm_nycflights13() %>% dm_paste(options = "all")
    Message
      airlines <- tibble::tibble(
        carrier = character(0),
        name = character(0),
      )
      airports <- tibble::tibble(
        faa = character(0),
        name = character(0),
        lat = numeric(0),
        lon = numeric(0),
        alt = numeric(0),
        tz = numeric(0),
        dst = character(0),
        tzone = character(0),
      )
      flights <- tibble::tibble(
        year = integer(0),
        month = integer(0),
        day = integer(0),
        dep_time = integer(0),
        sched_dep_time = integer(0),
        dep_delay = numeric(0),
        arr_time = integer(0),
        sched_arr_time = integer(0),
        arr_delay = numeric(0),
        carrier = character(0),
        flight = integer(0),
        tailnum = character(0),
        origin = character(0),
        dest = character(0),
        air_time = numeric(0),
        distance = numeric(0),
        hour = numeric(0),
        minute = numeric(0),
        time_hour = structure(numeric(0), class = c("POSIXct", "POSIXt"), tzone = "America/New_York"),
      )
      planes <- tibble::tibble(
        tailnum = character(0),
        year = integer(0),
        type = character(0),
        manufacturer = character(0),
        model = character(0),
        engines = integer(0),
        seats = integer(0),
        speed = integer(0),
        engine = character(0),
      )
      weather <- tibble::tibble(
        origin = character(0),
        year = integer(0),
        month = integer(0),
        day = integer(0),
        hour = integer(0),
        temp = numeric(0),
        dewp = numeric(0),
        humid = numeric(0),
        wind_dir = numeric(0),
        wind_speed = numeric(0),
        wind_gust = numeric(0),
        precip = numeric(0),
        pressure = numeric(0),
        visib = numeric(0),
        time_hour = structure(numeric(0), class = c("POSIXct", "POSIXt"), tzone = "America/New_York"),
      )
      dm::dm(
        airlines,
        airports,
        flights,
        planes,
        weather,
      ) %>%
        dm::dm_add_pk(airlines, carrier) %>%
        dm::dm_add_pk(airports, faa) %>%
        dm::dm_add_pk(planes, tailnum) %>%
        dm::dm_add_pk(weather, c(origin, time_hour)) %>%
        dm::dm_add_fk(flights, carrier, airlines) %>%
        dm::dm_add_fk(flights, origin, airports) %>%
        dm::dm_add_fk(flights, tailnum, planes) %>%
        dm::dm_add_fk(flights, c(origin, time_hour), weather) %>%
        dm::dm_set_colors(`#ED7D31FF` = airlines) %>%
        dm::dm_set_colors(`#ED7D31FF` = airports) %>%
        dm::dm_set_colors(`#5B9BD5FF` = flights) %>%
        dm::dm_set_colors(`#ED7D31FF` = planes) %>%
        dm::dm_set_colors(`#70AD47FF` = weather)
    Code
      # deprecation warning for select argument
      dm() %>% dm_paste(select = TRUE)
    Condition
      Warning:
      The `select` argument of `dm_paste()` is deprecated as of dm 0.1.2.
      i Please use the `options` argument instead.
    Message
      dm::dm(
      )
    Code
      # error for bad option
      writeLines(conditionMessage(expect_error(dm_paste(dm(), options = c("bogus",
        "all", "mad")))))
    Output
      Options unknown: "bogus" and "mad". Must be one of "all", "tables", "keys", "select", and "color".

# output 2

    Code
      # no error for factor column that leads to code with width > 500
      dm(tibble(a = factor(levels = expand.grid(letters, as.character(1:5)) %>%
        transmute(x = paste0(Var1, Var2)) %>% pull()))) %>% dm_paste(options = "tables")
    Message
      `tibble(...)` <- tibble::tibble(
        a = structure(integer(0), class = "factor", levels = c("a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1", "i1", "j1", "k1", "l1", "m1", "n1", "o1", "p1", "q1", "r1", "s1", "t1", "u1", "v1", "w1", "x1", "y1", "z1", "a2", "b2", "c2", "d2", "e2", "f2", "g2", "h2", "i2", "j2", "k2", "l2", "m2", "n2", "o2", "p2", "q2", "r2", "s2", "t2", "u2", "v2", "w2", "x2", "y2", "z2", "a3", "b3", "c3", "d3", "e3", "f3", "g3", "h3", "i3", "j3", "k3", "l3", "m3", "n3", "o3", "p3", "q3", "r3", "s3", "t3", "u3", "v3", "w3", "x3", "y3", "z3", "a4", "b4", "c4", "d4", "e4", "f4", "g4", "h4", "i4", "j4", "k4", "l4", "m4", "n4", "o4", "p4", "q4", "r4", "s4", "t4", "u4", "v4", "w4", "x4", "y4", "z4", "a5", "b5", "c5", "d5", "e5", "f5", "g5", "h5", "i5", "j5", "k5", "l5", "m5", "n5", "o5", "p5", "q5", "r5", "s5", "t5", "u5", "v5", "w5", "x5", "y5", "z5")),
      )
      dm::dm(
        `tibble(...)`,
      )

# chunking behavior for large dm

    Code
      writeLines(dm:::dm_paste_impl(create_large_dm(3), c("keys"), 2, chunk_size = 3))
    Output
      dm_step_1 <- dm::dm(
        main,
        table_1,
        table_2,
        table_3,
      )
      
      dm_step_2 <- dm_step_1 %>%
        dm::dm_add_pk(main, id) %>%
        dm::dm_add_pk(table_1, id) %>%
        dm::dm_add_pk(table_2, id)
      
      dm_step_3 <- dm_step_2 %>%
        dm::dm_add_pk(table_3, id)
      
      dm_step_3 %>%
        dm::dm_add_fk(table_1, main_id, main) %>%
        dm::dm_add_fk(table_2, main_id, main) %>%
        dm::dm_add_fk(table_3, main_id, main)

# build_copy_queries snapshot test for pixarfilms

    Code
      pixar_dm %>% build_copy_queries(src_db, .) %>% as.list()
    Output
      $name
      [1] "pixar_films"     "academy"         "box_office"      "genres"         
      [5] "public_response"
      
      $remote_name
            pixar_films           academy        box_office            genres 
          "pixar_films"         "academy"      "box_office"          "genres" 
        public_response 
      "public_response" 
      
      $columns
      $columns[[1]]
      [1] "number"       "film"         "release_date" "run_time"     "film_rating" 
      
      $columns[[2]]
      [1] "film"       "award_type" "status"    
      
      $columns[[3]]
      [1] "film"                 "budget"               "box_office_us_canada"
      [4] "box_office_other"     "box_office_worldwide"
      
      $columns[[4]]
      [1] "film"  "genre"
      
      $columns[[5]]
      [1] "film"            "rotten_tomatoes" "metacritic"      "cinema_score"   
      [5] "critics_choice" 
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE pixar_films (
        number STRING,
        film STRING,
        release_date DATE,
        run_time DOUBLE,
        film_rating STRING,
        PRIMARY KEY (film)
      )
      <SQL> CREATE TEMPORARY TABLE academy (
        film STRING,
        award_type STRING,
        status STRING,
        PRIMARY KEY (film, award_type),
        FOREIGN KEY (film) REFERENCES pixar_films (film)
      )
      <SQL> CREATE TEMPORARY TABLE box_office (
        film STRING,
        budget DOUBLE,
        box_office_us_canada DOUBLE,
        box_office_other DOUBLE,
        box_office_worldwide DOUBLE,
        PRIMARY KEY (film),
        FOREIGN KEY (film) REFERENCES pixar_films (film)
      )
      <SQL> CREATE TEMPORARY TABLE genres (
        film STRING,
        genre STRING,
        PRIMARY KEY (film, genre),
        FOREIGN KEY (film) REFERENCES pixar_films (film)
      )
      <SQL> CREATE TEMPORARY TABLE public_response (
        film STRING,
        rotten_tomatoes DOUBLE,
        metacritic DOUBLE,
        cinema_score STRING,
        critics_choice DOUBLE,
        PRIMARY KEY (film),
        FOREIGN KEY (film) REFERENCES pixar_films (film)
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      <SQL> CREATE INDEX academy__film ON academy (film)
      
      $sql_index[[3]]
      <SQL> CREATE INDEX box_office__film ON box_office (film)
      
      $sql_index[[4]]
      <SQL> CREATE INDEX genres__film ON genres (film)
      
      $sql_index[[5]]
      <SQL> CREATE INDEX public_response__film ON public_response (film)
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      [1] "academy__film"
      
      $index_name[[3]]
      [1] "box_office__film"
      
      $index_name[[4]]
      [1] "genres__film"
      
      $index_name[[5]]
      [1] "public_response__film"
      
      

# build_copy_queries snapshot test for dm_for_filter()

    Code
      dm_for_filter() %>% collect() %>% build_copy_queries(src_db, .) %>% as.list()
    Message
      `on_delete = "cascade"` not supported for duckdb
    Output
      $name
      [1] "tf_1" "tf_3" "tf_6" "tf_2" "tf_4" "tf_5"
      
      $remote_name
        tf_1   tf_3   tf_6   tf_2   tf_4   tf_5 
      "tf_1" "tf_3" "tf_6" "tf_2" "tf_4" "tf_5" 
      
      $columns
      $columns[[1]]
      [1] "a" "b"
      
      $columns[[2]]
      [1] "f"  "f1" "g" 
      
      $columns[[3]]
      [1] "zz" "n"  "o" 
      
      $columns[[4]]
      [1] "c"  "d"  "e"  "e1"
      
      $columns[[5]]
      [1] "h"  "i"  "j"  "j1"
      
      $columns[[6]]
      [1] "ww" "k"  "l"  "m" 
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE tf_1 (
        a INTEGER,
        b STRING,
        PRIMARY KEY (a)
      )
      <SQL> CREATE TEMPORARY TABLE tf_3 (
        f STRING,
        f1 INTEGER,
        g STRING,
        PRIMARY KEY (f, f1)
      )
      <SQL> CREATE TEMPORARY TABLE tf_6 (
        zz INTEGER,
        n STRING,
        o STRING,
        PRIMARY KEY (o),
        UNIQUE (n)
      )
      <SQL> CREATE TEMPORARY TABLE tf_2 (
        c STRING,
        d INTEGER,
        e STRING,
        e1 INTEGER,
        PRIMARY KEY (c),
        FOREIGN KEY (d) REFERENCES tf_1 (a),
        FOREIGN KEY (e, e1) REFERENCES tf_3 (f, f1)
      )
      <SQL> CREATE TEMPORARY TABLE tf_4 (
        h STRING,
        i STRING,
        j STRING,
        j1 INTEGER,
        PRIMARY KEY (h),
        FOREIGN KEY (j, j1) REFERENCES tf_3 (f, f1)
      )
      <SQL> CREATE TEMPORARY TABLE tf_5 (
        ww INTEGER,
        k INTEGER,
        l STRING,
        m STRING,
        PRIMARY KEY (k),
        FOREIGN KEY (m) REFERENCES tf_6 (n),
        FOREIGN KEY (l) REFERENCES tf_4 (h)
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      NULL
      
      $sql_index[[3]]
      NULL
      
      $sql_index[[4]]
      <SQL> CREATE INDEX tf_2__d ON tf_2 (d)
      <SQL> CREATE INDEX tf_2__e_e1 ON tf_2 (e, e1)
      
      $sql_index[[5]]
      <SQL> CREATE INDEX tf_4__j_j1 ON tf_4 (j, j1)
      
      $sql_index[[6]]
      <SQL> CREATE INDEX tf_5__m ON tf_5 (m)
      <SQL> CREATE INDEX tf_5__l ON tf_5 (l)
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      NULL
      
      $index_name[[3]]
      NULL
      
      $index_name[[4]]
      [1] "tf_2__d"    "tf_2__e_e1"
      
      $index_name[[5]]
      [1] "tf_4__j_j1"
      
      $index_name[[6]]
      [1] "tf_5__m" "tf_5__l"
      
      

# build_copy_queries avoids duplicate indexes

    Code
      as.list(queries)
    Output
      $name
      [1] "parent1"  "parent2"  "child"    "child__a"
      
      $remote_name
      $remote_name$parent1
      <Id> "parent1"
      
      $remote_name$parent2
      <Id> "parent2"
      
      $remote_name$child
      <Id> "child"
      
      $remote_name$child__a
      <Id> "child__a"
      
      
      $columns
      $columns[[1]]
      [1] "key"
      
      $columns[[2]]
      [1] "a__key"
      
      $columns[[3]]
      [1] "a__key"
      
      $columns[[4]]
      [1] "key"
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE parent1 (
        "key" DOUBLE,
        PRIMARY KEY ("key")
      )
      <SQL> CREATE TEMPORARY TABLE parent2 (
        a__key DOUBLE,
        PRIMARY KEY (a__key)
      )
      <SQL> CREATE TEMPORARY TABLE child (
        a__key DOUBLE,
        FOREIGN KEY (a__key) REFERENCES parent2 (a__key)
      )
      <SQL> CREATE TEMPORARY TABLE child__a (
        "key" DOUBLE,
        FOREIGN KEY ("key") REFERENCES parent2 (a__key)
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      NULL
      
      $sql_index[[3]]
      <SQL> CREATE INDEX child__a__key ON child (a__key)
      
      $sql_index[[4]]
      <SQL> CREATE INDEX child__a__key__1 ON child__a ("key")
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      NULL
      
      $index_name[[3]]
      [1] "child__a__key"
      
      $index_name[[4]]
      [1] "child__a__key__1"
      
      

# build_copy_queries snapshot test for pixarfilms

    Code
      pixar_dm %>% build_copy_queries(src_db, .) %>% as.list()
    Output
      $name
      [1] "pixar_films"     "academy"         "box_office"      "genres"         
      [5] "public_response"
      
      $remote_name
            pixar_films           academy        box_office            genres 
          "pixar_films"         "academy"      "box_office"          "genres" 
        public_response 
      "public_response" 
      
      $columns
      $columns[[1]]
      [1] "number"       "film"         "release_date" "run_time"     "film_rating" 
      
      $columns[[2]]
      [1] "film"       "award_type" "status"    
      
      $columns[[3]]
      [1] "film"                 "budget"               "box_office_us_canada"
      [4] "box_office_other"     "box_office_worldwide"
      
      $columns[[4]]
      [1] "film"  "genre"
      
      $columns[[5]]
      [1] "film"            "rotten_tomatoes" "metacritic"      "cinema_score"   
      [5] "critics_choice" 
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE "pixar_films" (
        "number" TEXT,
        "film" TEXT,
        "release_date" DATE,
        "run_time" DOUBLE PRECISION,
        "film_rating" TEXT,
        PRIMARY KEY ("film")
      )
      <SQL> CREATE TEMPORARY TABLE "academy" (
        "film" TEXT,
        "award_type" TEXT,
        "status" TEXT,
        PRIMARY KEY ("film", "award_type"),
        FOREIGN KEY ("film") REFERENCES "pixar_films" ("film")
      )
      <SQL> CREATE TEMPORARY TABLE "box_office" (
        "film" TEXT,
        "budget" DOUBLE PRECISION,
        "box_office_us_canada" DOUBLE PRECISION,
        "box_office_other" DOUBLE PRECISION,
        "box_office_worldwide" DOUBLE PRECISION,
        PRIMARY KEY ("film"),
        FOREIGN KEY ("film") REFERENCES "pixar_films" ("film")
      )
      <SQL> CREATE TEMPORARY TABLE "genres" (
        "film" TEXT,
        "genre" TEXT,
        PRIMARY KEY ("film", "genre"),
        FOREIGN KEY ("film") REFERENCES "pixar_films" ("film")
      )
      <SQL> CREATE TEMPORARY TABLE "public_response" (
        "film" TEXT,
        "rotten_tomatoes" DOUBLE PRECISION,
        "metacritic" DOUBLE PRECISION,
        "cinema_score" TEXT,
        "critics_choice" DOUBLE PRECISION,
        PRIMARY KEY ("film"),
        FOREIGN KEY ("film") REFERENCES "pixar_films" ("film")
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      <SQL> CREATE INDEX academy__film ON "academy" ("film")
      
      $sql_index[[3]]
      <SQL> CREATE INDEX box_office__film ON "box_office" ("film")
      
      $sql_index[[4]]
      <SQL> CREATE INDEX genres__film ON "genres" ("film")
      
      $sql_index[[5]]
      <SQL> CREATE INDEX public_response__film ON "public_response" ("film")
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      [1] "academy__film"
      
      $index_name[[3]]
      [1] "box_office__film"
      
      $index_name[[4]]
      [1] "genres__film"
      
      $index_name[[5]]
      [1] "public_response__film"
      
      

# build_copy_queries snapshot test for dm_for_filter()

    Code
      dm_for_filter() %>% collect() %>% build_copy_queries(src_db, .) %>% as.list()
    Output
      $name
      [1] "tf_1" "tf_3" "tf_6" "tf_2" "tf_4" "tf_5"
      
      $remote_name
        tf_1   tf_3   tf_6   tf_2   tf_4   tf_5 
      "tf_1" "tf_3" "tf_6" "tf_2" "tf_4" "tf_5" 
      
      $columns
      $columns[[1]]
      [1] "a" "b"
      
      $columns[[2]]
      [1] "f"  "f1" "g" 
      
      $columns[[3]]
      [1] "zz" "n"  "o" 
      
      $columns[[4]]
      [1] "c"  "d"  "e"  "e1"
      
      $columns[[5]]
      [1] "h"  "i"  "j"  "j1"
      
      $columns[[6]]
      [1] "ww" "k"  "l"  "m" 
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE "tf_1" (
        "a" SERIAL,
        "b" TEXT,
        PRIMARY KEY ("a")
      )
      <SQL> CREATE TEMPORARY TABLE "tf_3" (
        "f" TEXT,
        "f1" INTEGER,
        "g" TEXT,
        PRIMARY KEY ("f", "f1")
      )
      <SQL> CREATE TEMPORARY TABLE "tf_6" (
        "zz" INTEGER,
        "n" TEXT,
        "o" TEXT,
        PRIMARY KEY ("o"),
        UNIQUE ("n")
      )
      <SQL> CREATE TEMPORARY TABLE "tf_2" (
        "c" TEXT,
        "d" INTEGER,
        "e" TEXT,
        "e1" INTEGER,
        PRIMARY KEY ("c"),
        FOREIGN KEY ("d") REFERENCES "tf_1" ("a"),
        FOREIGN KEY ("e", "e1") REFERENCES "tf_3" ("f", "f1")
      )
      <SQL> CREATE TEMPORARY TABLE "tf_4" (
        "h" TEXT,
        "i" TEXT,
        "j" TEXT,
        "j1" INTEGER,
        PRIMARY KEY ("h"),
        FOREIGN KEY ("j", "j1") REFERENCES "tf_3" ("f", "f1")
      )
      <SQL> CREATE TEMPORARY TABLE "tf_5" (
        "ww" INTEGER,
        "k" INTEGER,
        "l" TEXT,
        "m" TEXT,
        PRIMARY KEY ("k"),
        FOREIGN KEY ("m") REFERENCES "tf_6" ("n"),
        FOREIGN KEY ("l") REFERENCES "tf_4" ("h") ON DELETE CASCADE
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      NULL
      
      $sql_index[[3]]
      NULL
      
      $sql_index[[4]]
      <SQL> CREATE INDEX tf_2__d ON "tf_2" ("d")
      <SQL> CREATE INDEX tf_2__e_e1 ON "tf_2" ("e", "e1")
      
      $sql_index[[5]]
      <SQL> CREATE INDEX tf_4__j_j1 ON "tf_4" ("j", "j1")
      
      $sql_index[[6]]
      <SQL> CREATE INDEX tf_5__m ON "tf_5" ("m")
      <SQL> CREATE INDEX tf_5__l ON "tf_5" ("l")
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      NULL
      
      $index_name[[3]]
      NULL
      
      $index_name[[4]]
      [1] "tf_2__d"    "tf_2__e_e1"
      
      $index_name[[5]]
      [1] "tf_4__j_j1"
      
      $index_name[[6]]
      [1] "tf_5__m" "tf_5__l"
      
      

# build_copy_queries avoids duplicate indexes

    Code
      as.list(queries)
    Output
      $name
      [1] "parent1"  "parent2"  "child"    "child__a"
      
      $remote_name
      $remote_name$parent1
      <Id> "parent1"
      
      $remote_name$parent2
      <Id> "parent2"
      
      $remote_name$child
      <Id> "child"
      
      $remote_name$child__a
      <Id> "child__a"
      
      
      $columns
      $columns[[1]]
      [1] "key"
      
      $columns[[2]]
      [1] "a__key"
      
      $columns[[3]]
      [1] "a__key"
      
      $columns[[4]]
      [1] "key"
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE "parent1" (
        "key" DOUBLE PRECISION,
        PRIMARY KEY ("key")
      )
      <SQL> CREATE TEMPORARY TABLE "parent2" (
        "a__key" DOUBLE PRECISION,
        PRIMARY KEY ("a__key")
      )
      <SQL> CREATE TEMPORARY TABLE "child" (
        "a__key" DOUBLE PRECISION,
        FOREIGN KEY ("a__key") REFERENCES "parent2" ("a__key")
      )
      <SQL> CREATE TEMPORARY TABLE "child__a" (
        "key" DOUBLE PRECISION,
        FOREIGN KEY ("key") REFERENCES "parent2" ("a__key")
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      NULL
      
      $sql_index[[3]]
      <SQL> CREATE INDEX child__a__key ON "child" ("a__key")
      
      $sql_index[[4]]
      <SQL> CREATE INDEX child__a__key__1 ON "child__a" ("key")
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      NULL
      
      $index_name[[3]]
      [1] "child__a__key"
      
      $index_name[[4]]
      [1] "child__a__key__1"
      
      

# zoom2 output for compound keys

    Code
      nyc_comp() %>% dm_zoom2_to(weather)
    Output
      # A tibble: 144 x 15
      # Keys:     `origin`, `time_hour` | 1 | 0
         origin  year month   day  hour  temp  dewp humid wind_dir wind_speed
         <chr>  <int> <int> <int> <int> <dbl> <dbl> <dbl>    <dbl>      <dbl>
       1 EWR     2013     1    10     0  41    32    70.1      230       8.06
       2 EWR     2013     1    10     1  39.0  30.0  69.9      210       9.21
       3 EWR     2013     1    10     2  39.0  28.9  66.8      230       6.90
       4 EWR     2013     1    10     3  39.9  27.0  59.5      270       5.75
       5 EWR     2013     1    10     4  41    26.1  55.0      320       6.90
       6 EWR     2013     1    10     5  41    26.1  55.0      300      12.7 
       7 EWR     2013     1    10     6  39.9  25.0  54.8      280       6.90
       8 EWR     2013     1    10     7  41    25.0  52.6      330       6.90
       9 EWR     2013     1    10     8  43.0  25.0  48.7      330       8.06
      10 EWR     2013     1    10     9  45.0  23    41.6      320      17.3 
      # i 134 more rows
      # i 5 more variables: wind_gust <dbl>, precip <dbl>, pressure <dbl>,
      #   visib <dbl>, time_hour <chr>
    Code
      nyc_comp() %>% dm_zoom2_to(weather) %>% dm_update_zoom2ed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 53
      Primary keys: 4
      Foreign keys: 4
    Code
      nyc_comp_2 <- nyc_comp() %>% dm_zoom2_to(weather) %>% dm_insert_zoom2ed(
        "weather_2")
      nyc_comp_2 %>% get_all_keys()
    Output
      $pks
      # A tibble: 5 x 3
        table     pk_col            autoincrement
        <chr>     <keys>            <lgl>        
      1 airlines  carrier           FALSE        
      2 airports  faa               FALSE        
      3 planes    tailnum           FALSE        
      4 weather   origin, time_hour FALSE        
      5 weather_2 origin, time_hour FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols     parent_table parent_key_cols   on_delete
        <chr>       <keys>            <chr>        <keys>            <chr>    
      1 flights     carrier           airlines     carrier           no_action
      2 flights     dest              airports     faa               no_action
      3 flights     tailnum           planes       tailnum           no_action
      4 flights     origin, time_hour weather      origin, time_hour no_action
      5 flights     origin, time_hour weather_2    origin, time_hour no_action
      
    Code
      nyc_comp_3 <- nyc_comp() %>% dm_zoom2_to(flights) %>% dm_insert_zoom2ed(
        "flights_2")
      nyc_comp_3 %>% get_all_keys()
    Output
      $pks
      # A tibble: 4 x 3
        table    pk_col            autoincrement
        <chr>    <keys>            <lgl>        
      1 airlines carrier           FALSE        
      2 airports faa               FALSE        
      3 planes   tailnum           FALSE        
      4 weather  origin, time_hour FALSE        
      
      $fks
      # A tibble: 8 x 5
        child_table child_fk_cols     parent_table parent_key_cols   on_delete
        <chr>       <keys>            <chr>        <keys>            <chr>    
      1 flights     carrier           airlines     carrier           no_action
      2 flights_2   carrier           airlines     carrier           no_action
      3 flights     dest              airports     faa               no_action
      4 flights_2   dest              airports     faa               no_action
      5 flights     tailnum           planes       tailnum           no_action
      6 flights_2   tailnum           planes       tailnum           no_action
      7 flights     origin, time_hour weather      origin, time_hour no_action
      8 flights_2   origin, time_hour weather      origin, time_hour no_action
      

# zoom2 select() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% select(id) %>% dm_update_zoom2ed() %>% dm_paste(
        options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 select() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% select(child_id, val) %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id)

# zoom2 rename() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% rename(parent_id = id) %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        parent_id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(child, child_id)

# zoom2 rename() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% rename(cid = child_id, pid = parent_id) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        cid = integer(0),
        pid = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id)

# zoom2 summarise() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% summarise(n = n()) %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        n = integer(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(child, child_id)

# zoom2 summarise() on child table with group_by

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% group_by(parent_id) %>% summarise(n = n()) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        parent_id = integer(0),
        n = integer(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, parent_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 summarise() insert

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% group_by(parent_id) %>% summarise(n = n()) %>%
        dm_insert_zoom2ed("child_summary") %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      child_summary <- tibble::tibble(
        parent_id = integer(0),
        n = integer(0),
      )
      dm::dm(
        parent,
        child,
        child_summary,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_pk(child_summary, parent_id) %>%
        dm::dm_add_fk(child, parent_id, parent) %>%
        dm::dm_add_fk(child_summary, parent_id, parent)

# zoom2 summarise() on child table with .by

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% summarise(n = n(), .by = parent_id) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        parent_id = integer(0),
        n = integer(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, parent_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 summarise() insert with .by

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% summarise(n = n(), .by = parent_id) %>%
        dm_insert_zoom2ed("child_summary") %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      child_summary <- tibble::tibble(
        parent_id = integer(0),
        n = integer(0),
      )
      dm::dm(
        parent,
        child,
        child_summary,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_pk(child_summary, parent_id) %>%
        dm::dm_add_fk(child, parent_id, parent) %>%
        dm::dm_add_fk(child_summary, parent_id, parent)

# zoom2 reframe() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% reframe(n = n()) %>% dm_update_zoom2ed() %>% dm_paste(
        options = "all")
    Message
      parent <- tibble::tibble(
        n = integer(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(child, child_id)

# zoom2 reframe() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% reframe(n = n(), .by = parent_id) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        parent_id = integer(0),
        n = integer(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 tally() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% tally() %>% dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        n = integer(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(child, child_id)

# zoom2 tally() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% tally() %>% dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        n = integer(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id)

# zoom2 left_join()

    Code
      d <- dm(grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")), parent = tibble(
        id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")), child = tibble(
        child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>%
        dm_add_pk(grandparent, gp_id) %>% dm_add_pk(parent, id) %>% dm_add_pk(child,
        child_id) %>% dm_add_fk(parent, gp_id, grandparent) %>% dm_add_fk(child,
        parent_id, parent)
      d %>% dm_zoom2_to(child) %>% left_join(dm_zoom2_to(d, parent)) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      grandparent <- tibble::tibble(
        gp_id = integer(0),
        gp_name = character(0),
      )
      parent <- tibble::tibble(
        id = integer(0),
        gp_id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
        gp_id = integer(0),
        name = character(0),
      )
      dm::dm(
        grandparent,
        parent,
        child,
      ) %>%
        dm::dm_add_pk(grandparent, gp_id) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(parent, gp_id, grandparent) %>%
        dm::dm_add_fk(child, gp_id, grandparent) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 inner_join()

    Code
      d <- dm(grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")), parent = tibble(
        id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")), child = tibble(
        child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>%
        dm_add_pk(grandparent, gp_id) %>% dm_add_pk(parent, id) %>% dm_add_pk(child,
        child_id) %>% dm_add_fk(parent, gp_id, grandparent) %>% dm_add_fk(child,
        parent_id, parent)
      d %>% dm_zoom2_to(child) %>% inner_join(dm_zoom2_to(d, parent)) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      grandparent <- tibble::tibble(
        gp_id = integer(0),
        gp_name = character(0),
      )
      parent <- tibble::tibble(
        id = integer(0),
        gp_id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
        gp_id = integer(0),
        name = character(0),
      )
      dm::dm(
        grandparent,
        parent,
        child,
      ) %>%
        dm::dm_add_pk(grandparent, gp_id) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(parent, gp_id, grandparent) %>%
        dm::dm_add_fk(child, gp_id, grandparent) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 semi_join()

    Code
      d <- dm(grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")), parent = tibble(
        id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")), child = tibble(
        child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>%
        dm_add_pk(grandparent, gp_id) %>% dm_add_pk(parent, id) %>% dm_add_pk(child,
        child_id) %>% dm_add_fk(parent, gp_id, grandparent) %>% dm_add_fk(child,
        parent_id, parent)
      d %>% dm_zoom2_to(child) %>% semi_join(dm_zoom2_to(d, parent)) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      grandparent <- tibble::tibble(
        gp_id = integer(0),
        gp_name = character(0),
      )
      parent <- tibble::tibble(
        id = integer(0),
        gp_id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        grandparent,
        parent,
        child,
      ) %>%
        dm::dm_add_pk(grandparent, gp_id) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(parent, gp_id, grandparent) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 anti_join()

    Code
      d <- dm(grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")), parent = tibble(
        id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")), child = tibble(
        child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>%
        dm_add_pk(grandparent, gp_id) %>% dm_add_pk(parent, id) %>% dm_add_pk(child,
        child_id) %>% dm_add_fk(parent, gp_id, grandparent) %>% dm_add_fk(child,
        parent_id, parent)
      d %>% dm_zoom2_to(child) %>% anti_join(dm_zoom2_to(d, parent)) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      grandparent <- tibble::tibble(
        gp_id = integer(0),
        gp_name = character(0),
      )
      parent <- tibble::tibble(
        id = integer(0),
        gp_id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        grandparent,
        parent,
        child,
      ) %>%
        dm::dm_add_pk(grandparent, gp_id) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(parent, gp_id, grandparent) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 right_join()

    Code
      d <- dm(grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")), parent = tibble(
        id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")), child = tibble(
        child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>%
        dm_add_pk(grandparent, gp_id) %>% dm_add_pk(parent, id) %>% dm_add_pk(child,
        child_id) %>% dm_add_fk(parent, gp_id, grandparent) %>% dm_add_fk(child,
        parent_id, parent)
      d %>% dm_zoom2_to(child) %>% right_join(dm_zoom2_to(d, parent)) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      grandparent <- tibble::tibble(
        gp_id = integer(0),
        gp_name = character(0),
      )
      parent <- tibble::tibble(
        id = integer(0),
        gp_id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
        gp_id = integer(0),
        name = character(0),
      )
      dm::dm(
        grandparent,
        parent,
        child,
      ) %>%
        dm::dm_add_pk(grandparent, gp_id) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(parent, gp_id, grandparent) %>%
        dm::dm_add_fk(child, gp_id, grandparent) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 full_join()

    Code
      d <- dm(grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")), parent = tibble(
        id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")), child = tibble(
        child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>%
        dm_add_pk(grandparent, gp_id) %>% dm_add_pk(parent, id) %>% dm_add_pk(child,
        child_id) %>% dm_add_fk(parent, gp_id, grandparent) %>% dm_add_fk(child,
        parent_id, parent)
      d %>% dm_zoom2_to(child) %>% full_join(dm_zoom2_to(d, parent)) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      grandparent <- tibble::tibble(
        gp_id = integer(0),
        gp_name = character(0),
      )
      parent <- tibble::tibble(
        id = integer(0),
        gp_id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
        gp_id = integer(0),
        name = character(0),
      )
      dm::dm(
        grandparent,
        parent,
        child,
      ) %>%
        dm::dm_add_pk(grandparent, gp_id) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(parent, gp_id, grandparent) %>%
        dm::dm_add_fk(child, gp_id, grandparent) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 cross_join()

    Code
      d <- dm(grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")), parent = tibble(
        id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")), child = tibble(
        child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>%
        dm_add_pk(grandparent, gp_id) %>% dm_add_pk(parent, id) %>% dm_add_pk(child,
        child_id) %>% dm_add_fk(parent, gp_id, grandparent) %>% dm_add_fk(child,
        parent_id, parent)
      d %>% dm_zoom2_to(child) %>% cross_join(dm_zoom2_to(d, parent)) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      grandparent <- tibble::tibble(
        gp_id = integer(0),
        gp_name = character(0),
      )
      parent <- tibble::tibble(
        id = integer(0),
        gp_id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
        id = integer(0),
        gp_id = integer(0),
        name = character(0),
      )
      dm::dm(
        grandparent,
        parent,
        child,
      ) %>%
        dm::dm_add_pk(grandparent, gp_id) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_fk(parent, gp_id, grandparent) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 left_join() insert

    Code
      d <- dm(grandparent = tibble(gp_id = 1:2, gp_name = c("x", "y")), parent = tibble(
        id = 1:3, gp_id = c(1L, 1L, 2L), name = c("a", "b", "c")), child = tibble(
        child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>%
        dm_add_pk(grandparent, gp_id) %>% dm_add_pk(parent, id) %>% dm_add_pk(child,
        child_id) %>% dm_add_fk(parent, gp_id, grandparent) %>% dm_add_fk(child,
        parent_id, parent)
      d %>% dm_zoom2_to(child) %>% left_join(dm_zoom2_to(d, parent)) %>%
        dm_insert_zoom2ed("child_parent") %>% dm_paste(options = "all")
    Message
      grandparent <- tibble::tibble(
        gp_id = integer(0),
        gp_name = character(0),
      )
      parent <- tibble::tibble(
        id = integer(0),
        gp_id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      child_parent <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
        gp_id = integer(0),
        name = character(0),
      )
      dm::dm(
        grandparent,
        parent,
        child,
        child_parent,
      ) %>%
        dm::dm_add_pk(grandparent, gp_id) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_pk(child_parent, child_id) %>%
        dm::dm_add_fk(parent, gp_id, grandparent) %>%
        dm::dm_add_fk(child_parent, gp_id, grandparent) %>%
        dm::dm_add_fk(child, parent_id, parent) %>%
        dm::dm_add_fk(child_parent, parent_id, parent) %>%
        dm::dm_add_fk(child, parent_id, child_parent, parent_id)

# zoom2 filter() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% filter(id > 1) %>% dm_update_zoom2ed() %>% dm_paste(
        options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 filter() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% filter(val != "a") %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 mutate() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% mutate(name2 = toupper(name)) %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
        name2 = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 mutate() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% mutate(val2 = toupper(val)) %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
        val2 = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 arrange() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% arrange(desc(id)) %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 arrange() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% arrange(desc(child_id)) %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 distinct() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% distinct() %>% dm_update_zoom2ed() %>% dm_paste(
        options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 distinct() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% distinct() %>% dm_update_zoom2ed() %>% dm_paste(
        options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 slice() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% slice(1:2) %>% dm_update_zoom2ed() %>% dm_paste(
        options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 slice() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% slice(1:3) %>% dm_update_zoom2ed() %>% dm_paste(
        options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 transmute() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% transmute(id, name_upper = toupper(name)) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name_upper = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 transmute() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% transmute(child_id, parent_id) %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 relocate() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% relocate(name, .before = id) %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        name = character(0),
        id = integer(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 relocate() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% relocate(val, .before = child_id) %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        val = character(0),
        child_id = integer(0),
        parent_id = integer(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 group_by() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% group_by(name) %>% dm_update_zoom2ed() %>% dm_paste(
        options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 group_by() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% group_by(parent_id) %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 ungroup() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% group_by(name) %>% ungroup() %>% dm_update_zoom2ed() %>%
        dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 ungroup() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% group_by(parent_id) %>% ungroup() %>%
        dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 count() on parent table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(parent) %>% count(name) %>% dm_update_zoom2ed() %>% dm_paste(
        options = "all")
    Message
      parent <- tibble::tibble(
        name = character(0),
        n = integer(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(child, child_id)

# zoom2 count() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>% dm_add_pk(parent,
        id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child, parent_id, parent) %>%
        dm_zoom2_to(child) %>% count(parent_id) %>% dm_update_zoom2ed() %>% dm_paste(
        options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        parent_id = integer(0),
        n = integer(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 unite() on parent table

    Code
      dm(parent = tibble(id = 1:3, first = c("a", "b", "c"), last = c("x", "y", "z")),
      child = tibble(child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>%
        dm_add_pk(parent, id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child,
        parent_id, parent) %>% dm_zoom2_to(parent) %>% tidyr::unite(full_name, first,
        last) %>% dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        full_name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 unite() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), first = c("a", "b", "c", "d"), last = c("w",
        "x", "y", "z"))) %>% dm_add_pk(parent, id) %>% dm_add_pk(child, child_id) %>%
        dm_add_fk(child, parent_id, parent) %>% dm_zoom2_to(child) %>% tidyr::unite(
        full_name, first, last) %>% dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        full_name = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 separate() on parent table

    Code
      dm(parent = tibble(id = 1:3, full_name = c("a_x", "b_y", "c_z")), child = tibble(
        child_id = 1:4, parent_id = c(1L, 1L, 2L, 3L), val = letters[1:4])) %>%
        dm_add_pk(parent, id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child,
        parent_id, parent) %>% dm_zoom2_to(parent) %>% tidyr::separate(full_name,
        into = c("first", "last")) %>% dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        first = character(0),
        last = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        val = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# zoom2 separate() on child table

    Code
      dm(parent = tibble(id = 1:3, name = c("a", "b", "c")), child = tibble(child_id = 1:
        4, parent_id = c(1L, 1L, 2L, 3L), full_val = c("a_1", "b_2", "c_3", "d_4"))) %>%
        dm_add_pk(parent, id) %>% dm_add_pk(child, child_id) %>% dm_add_fk(child,
        parent_id, parent) %>% dm_zoom2_to(child) %>% tidyr::separate(full_val, into = c(
        "letter", "number")) %>% dm_update_zoom2ed() %>% dm_paste(options = "all")
    Message
      parent <- tibble::tibble(
        id = integer(0),
        name = character(0),
      )
      child <- tibble::tibble(
        child_id = integer(0),
        parent_id = integer(0),
        letter = character(0),
        number = character(0),
      )
      dm::dm(
        parent,
        child,
      ) %>%
        dm::dm_add_pk(parent, id) %>%
        dm::dm_add_pk(child, child_id) %>%
        dm::dm_add_fk(child, parent_id, parent)

# dummy

    Code
      # dummy

# dm_sql()

    Code
      dm_for_filter_df() %>% dm_sql(my_test_con())
    Output
      $pre
      $pre$tf_1
      <SQL> CREATE TEMPORARY TABLE "tf_1" (
        "a" SERIAL,
        "b" TEXT,
        PRIMARY KEY ("a")
      )
      
      $pre$tf_2
      <SQL> CREATE TEMPORARY TABLE "tf_2" (
        "c" TEXT,
        "d" INTEGER,
        "e" TEXT,
        "e1" INTEGER,
        PRIMARY KEY ("c")
      )
      
      $pre$tf_3
      <SQL> CREATE TEMPORARY TABLE "tf_3" (
        "f" TEXT,
        "f1" INTEGER,
        "g" TEXT,
        PRIMARY KEY ("f", "f1")
      )
      
      $pre$tf_4
      <SQL> CREATE TEMPORARY TABLE "tf_4" (
        "h" TEXT,
        "i" TEXT,
        "j" TEXT,
        "j1" INTEGER,
        PRIMARY KEY ("h")
      )
      
      $pre$tf_5
      <SQL> CREATE TEMPORARY TABLE "tf_5" (
        "ww" INTEGER,
        "k" INTEGER,
        "l" TEXT,
        "m" TEXT,
        PRIMARY KEY ("k")
      )
      
      $pre$tf_6
      <SQL> CREATE TEMPORARY TABLE "tf_6" (
        "zz" INTEGER,
        "n" TEXT,
        "o" TEXT,
        PRIMARY KEY ("o")
      )
      
      
      $load
      $load$tf_1
      <SQL> INSERT INTO "tf_1" ("a", "b")
      SELECT CAST("a" AS INTEGER) AS "a", CAST("b" AS TEXT) AS "b"
      FROM (
        VALUES
          (1, 'A'),
          (2, 'B'),
          (3, 'C'),
          (4, 'D'),
          (5, 'E'),
          (6, 'F'),
          (7, 'G'),
          (8, 'H'),
          (9, 'I'),
          (10, 'J')
      ) AS drvd("a", "b")
      
      $load$tf_2
      <SQL> INSERT INTO "tf_2" ("c", "d", "e", "e1")
      SELECT
        CAST("c" AS TEXT) AS "c",
        CAST("d" AS INTEGER) AS "d",
        CAST("e" AS TEXT) AS "e",
        CAST("e1" AS INTEGER) AS "e1"
      FROM (
        VALUES
          ('cat', 7, 'F', 6),
          ('dog', 6, 'E', 5),
          ('elephant', 2, 'D', 4),
          ('lion', 3, 'E', 5),
          ('seal', 4, 'F', 6),
          ('worm', 5, 'G', 7)
      ) AS drvd("c", "d", "e", "e1")
      
      $load$tf_3
      <SQL> INSERT INTO "tf_3" ("f", "f1", "g")
      SELECT
        CAST("f" AS TEXT) AS "f",
        CAST("f1" AS INTEGER) AS "f1",
        CAST("g" AS TEXT) AS "g"
      FROM (
        VALUES
          ('C', 2, 'one'),
          ('C', 3, 'two'),
          ('D', 4, 'three'),
          ('E', 5, 'four'),
          ('F', 6, 'five'),
          ('G', 7, 'six'),
          ('H', 7, 'seven'),
          ('I', 7, 'eight'),
          ('J', 10, 'nine'),
          ('K', 11, 'ten')
      ) AS drvd("f", "f1", "g")
      
      $load$tf_4
      <SQL> INSERT INTO "tf_4" ("h", "i", "j", "j1")
      SELECT
        CAST("h" AS TEXT) AS "h",
        CAST("i" AS TEXT) AS "i",
        CAST("j" AS TEXT) AS "j",
        CAST("j1" AS INTEGER) AS "j1"
      FROM (
        VALUES
          ('a', 'three', 'C', 3),
          ('b', 'four', 'D', 4),
          ('c', 'five', 'E', 5),
          ('d', 'six', 'F', 6),
          ('e', 'seven', 'F', 6)
      ) AS drvd("h", "i", "j", "j1")
      
      $load$tf_5
      <SQL> INSERT INTO "tf_5" ("ww", "k", "l", "m")
      SELECT
        CAST("ww" AS INTEGER) AS "ww",
        CAST("k" AS INTEGER) AS "k",
        CAST("l" AS TEXT) AS "l",
        CAST("m" AS TEXT) AS "m"
      FROM (
        VALUES
          (2, 1, 'b', 'house'),
          (2, 2, 'c', 'tree'),
          (2, 3, 'd', 'streetlamp'),
          (2, 4, 'e', 'streetlamp')
      ) AS drvd("ww", "k", "l", "m")
      
      $load$tf_6
      <SQL> INSERT INTO "tf_6" ("zz", "n", "o")
      SELECT
        CAST("zz" AS INTEGER) AS "zz",
        CAST("n" AS TEXT) AS "n",
        CAST("o" AS TEXT) AS "o"
      FROM (
        VALUES
          (1, 'garden', 'i'),
          (1, 'hill', 'g'),
          (1, 'house', 'e'),
          (1, 'streetlamp', 'h'),
          (1, 'tree', 'f')
      ) AS drvd("zz", "n", "o")
      
      
      $post
      $post$uk
      $post$uk$tf_1
      <SQL> ALTER TABLE "tf_1" ADD UNIQUE ("a")
      
      $post$uk$tf_2
      <SQL> ALTER TABLE "tf_2" ADD UNIQUE ("c")
      
      $post$uk$tf_3
      <SQL> ALTER TABLE "tf_3" ADD UNIQUE ("f", "f1")
      <SQL> ALTER TABLE "tf_3" ADD UNIQUE ("g")
      
      $post$uk$tf_4
      <SQL> ALTER TABLE "tf_4" ADD UNIQUE ("h")
      
      $post$uk$tf_5
      <SQL> ALTER TABLE "tf_5" ADD UNIQUE ("k")
      
      $post$uk$tf_6
      <SQL> ALTER TABLE "tf_6" ADD UNIQUE ("o")
      <SQL> ALTER TABLE "tf_6" ADD UNIQUE ("n")
      
      
      $post$fk
      $post$fk$tf_2
      <SQL> ALTER TABLE "tf_2" ADD FOREIGN KEY ("d") REFERENCES "tf_1" ("a")
      <SQL> ALTER TABLE "tf_2" ADD FOREIGN KEY ("e", "e1") REFERENCES "tf_3" ("f", "f1")
      
      $post$fk$tf_4
      <SQL> ALTER TABLE "tf_4" ADD FOREIGN KEY ("j", "j1") REFERENCES "tf_3" ("f", "f1")
      
      $post$fk$tf_5
      <SQL> ALTER TABLE "tf_5" ADD FOREIGN KEY ("l") REFERENCES "tf_4" ("h") ON DELETE CASCADE
      <SQL> ALTER TABLE "tf_5" ADD FOREIGN KEY ("m") REFERENCES "tf_6" ("n")
      
      
      $post$indexes
      $post$indexes$tf_2
      <SQL> CREATE INDEX tf_2__d ON "tf_2" ("d")
      <SQL> CREATE INDEX tf_2__e_e1 ON "tf_2" ("e", "e1")
      
      $post$indexes$tf_4
      <SQL> CREATE INDEX tf_4__j_j1 ON "tf_4" ("j", "j1")
      
      $post$indexes$tf_5
      <SQL> CREATE INDEX tf_5__l ON "tf_5" ("l")
      <SQL> CREATE INDEX tf_5__m ON "tf_5" ("m")
      
      
      

# enumerate_all_paths() works

    Code
      enumerate_all_paths(dm_for_filter_w_cycle(), "tf_1")
    Output
      $table_mapping
      # A tibble: 10 x 2
         new_table table
         <chr>     <chr>
       1 tf_7-1    tf_7 
       2 tf_6-1    tf_6 
       3 tf_5-1    tf_5 
       4 tf_4-1    tf_4 
       5 tf_4-2    tf_4 
       6 tf_5-2    tf_5 
       7 tf_6-2    tf_6 
       8 tf_3-1    tf_3 
       9 tf_3-2    tf_3 
      10 tf_7-2    tf_7 
      
      $new_fks
      # A tibble: 11 x 5
         new_child_table child_cols new_parent_table parent_cols on_delete
         <chr>           <keys>     <chr>            <keys>      <chr>    
       1 tf_2            d          tf_1             a           no_action
       2 tf_7-1          q          tf_2             c           no_action
       3 tf_6-1          o          tf_7-1           p           no_action
       4 tf_5-1          m          tf_6-1           n           no_action
       5 tf_5-1          l          tf_4-1           h           cascade  
       6 tf_4-1          j, j1      tf_3-1           f, f1       no_action
       7 tf_2            e, e1      tf_3-2           f, f1       no_action
       8 tf_4-2          j, j1      tf_3-2           f, f1       no_action
       9 tf_5-2          l          tf_4-2           h           cascade  
      10 tf_5-2          m          tf_6-2           n           no_action
      11 tf_6-2          o          tf_7-2           p           no_action
      
    Code
      enumerate_all_paths(dm_for_filter_w_cycle(), "tf_5")
    Output
      $table_mapping
      # A tibble: 12 x 2
         new_table table
         <chr>     <chr>
       1 tf_4-1    tf_4 
       2 tf_2-1    tf_2 
       3 tf_7-1    tf_7 
       4 tf_6-1    tf_6 
       5 tf_6-2    tf_6 
       6 tf_7-2    tf_7 
       7 tf_2-2    tf_2 
       8 tf_4-2    tf_4 
       9 tf_3-1    tf_3 
      10 tf_1-1    tf_1 
      11 tf_1-2    tf_1 
      12 tf_3-2    tf_3 
      
      $new_fks
      # A tibble: 12 x 5
         new_child_table child_cols new_parent_table parent_cols on_delete
         <chr>           <keys>     <chr>            <keys>      <chr>    
       1 tf_5            l          tf_4-1           h           cascade  
       2 tf_4-1          j, j1      tf_3-1           f, f1       no_action
       3 tf_2-1          e, e1      tf_3-1           f, f1       no_action
       4 tf_7-1          q          tf_2-1           c           no_action
       5 tf_6-1          o          tf_7-1           p           no_action
       6 tf_2-1          d          tf_1-1           a           no_action
       7 tf_5            m          tf_6-2           n           no_action
       8 tf_6-2          o          tf_7-2           p           no_action
       9 tf_7-2          q          tf_2-2           c           no_action
      10 tf_2-2          d          tf_1-2           a           no_action
      11 tf_2-2          e, e1      tf_3-2           f, f1       no_action
      12 tf_4-2          j, j1      tf_3-2           f, f1       no_action
      
    Code
      enumerate_all_paths(entangled_dm(), "a")
    Output
      $table_mapping
      # A tibble: 22 x 2
         new_table table
         <chr>     <chr>
       1 b-1       b    
       2 c-1       c    
       3 d-1       d    
       4 e-1       e    
       5 f-1       f    
       6 g-1       g    
       7 f-2       f    
       8 e-2       e    
       9 g-2       g    
      10 c-2       c    
      # i 12 more rows
      
      $new_fks
      # A tibble: 22 x 5
         new_child_table child_cols new_parent_table parent_cols on_delete
         <chr>           <keys>     <chr>            <keys>      <chr>    
       1 a               a          b-1              b           no_action
       2 b-1             b          d-1              d           no_action
       3 c-1             c          d-1              d           no_action
       4 d-1             d          e-1              e           no_action
       5 e-1             e          g-1              g           no_action
       6 f-1             f          g-1              g           no_action
       7 g-1             g          h-1              h           no_action
       8 d-1             d          f-2              f           no_action
       9 f-2             f          g-2              g           no_action
      10 e-2             e          g-2              g           no_action
      # i 12 more rows
      
    Code
      enumerate_all_paths(entangled_dm(), "c")
    Output
      $table_mapping
      # A tibble: 22 x 2
         new_table table
         <chr>     <chr>
       1 a-1       a    
       2 b-1       b    
       3 d-1       d    
       4 e-1       e    
       5 f-1       f    
       6 g-1       g    
       7 f-2       f    
       8 e-2       e    
       9 g-2       g    
      10 b-2       b    
      # i 12 more rows
      
      $new_fks
      # A tibble: 22 x 5
         new_child_table child_cols new_parent_table parent_cols on_delete
         <chr>           <keys>     <chr>            <keys>      <chr>    
       1 a-1             a          c                c           no_action
       2 a-1             a          b-1              b           no_action
       3 b-1             b          d-1              d           no_action
       4 d-1             d          e-1              e           no_action
       5 e-1             e          g-1              g           no_action
       6 f-1             f          g-1              g           no_action
       7 g-1             g          h-1              h           no_action
       8 d-1             d          f-2              f           no_action
       9 f-2             f          g-2              g           no_action
      10 e-2             e          g-2              g           no_action
      # i 12 more rows
      
    Code
      enumerate_all_paths(entangled_dm_2(), "a")
    Output
      $table_mapping
      # A tibble: 8 x 2
        new_table table
        <chr>     <chr>
      1 b-1       b    
      2 c-1       c    
      3 d-1       d    
      4 d-2       d    
      5 b-2       b    
      6 c-2       c    
      7 e-1       e    
      8 e-2       e    
      
      $new_fks
      # A tibble: 9 x 5
        new_child_table child_cols new_parent_table parent_cols on_delete
        <chr>           <keys>     <chr>            <keys>      <chr>    
      1 a               a          d-1              d           no_action
      2 b-1             b          d-1              d           no_action
      3 c-1             c          d-1              d           no_action
      4 d-1             d          e-1              e           no_action
      5 a               a          e-2              e           no_action
      6 d-2             d          e-2              e           no_action
      7 b-2             b          d-2              d           no_action
      8 c-2             c          d-2              d           no_action
      9 f               f          g                g           no_action
      
    Code
      enumerate_all_paths(entangled_dm_2(), "d")
    Output
      $table_mapping
      # A tibble: 4 x 2
        new_table table
        <chr>     <chr>
      1 a-1       a    
      2 a-2       a    
      3 e-1       e    
      4 e-2       e    
      
      $new_fks
      # A tibble: 7 x 5
        new_child_table child_cols new_parent_table parent_cols on_delete
        <chr>           <keys>     <chr>            <keys>      <chr>    
      1 a-1             a          d                d           no_action
      2 a-1             a          e-1              e           no_action
      3 b               b          d                d           no_action
      4 c               c          d                d           no_action
      5 d               d          e-2              e           no_action
      6 a-2             a          e-2              e           no_action
      7 f               f          g                g           no_action
      

# `json_pack()` works remotely

    Code
      query <- json_pack(remote, a = starts_with("a")) %>% dbplyr::sql_render()
      gsub("test_frame_[_0-9]+", "test_frame_...", query)
    Output
      <SQL> SELECT "grp", JSON_BUILD_OBJECT('a_i', "a_i", 'a_j', "a_j") AS "a"
      FROM "test_frame_..."
    Code
      json_pack(remote, a = starts_with("a"))
    Output
          grp a                         
        <dbl> <pq_json>                 
      1     1 {"a_i" : "a", "a_j" : "A"}
      2     1 {"a_i" : "b", "a_j" : "B"}
      3     2 {"a_i" : "c", "a_j" : "C"}
      4     2 {"a_i" : "d", "a_j" : "D"}
    Code
      query <- json_pack(remote, a = starts_with("a"), .names_sep = "_") %>% dbplyr::sql_render()
      gsub("test_frame_[_0-9]+", "test_frame_...", query)
    Output
      <SQL> SELECT "grp", JSON_BUILD_OBJECT('i', "a_i", 'j', "a_j") AS "a"
      FROM "test_frame_..."
    Code
      json_pack(remote, a = starts_with("a"), .names_sep = "_")
    Output
          grp a                     
        <dbl> <pq_json>             
      1     1 {"i" : "a", "j" : "A"}
      2     1 {"i" : "b", "j" : "B"}
      3     2 {"i" : "c", "j" : "C"}
      4     2 {"i" : "d", "j" : "D"}

# print() and format() methods for subclass `dm_zoomed` work

    Code
      dm_for_filter() %>% dm_zoom_to(tf_5) %>% as_dm_zoomed_df() %>% tbl_sum()
    Output
      Zoomed table     A tibble 
            "tf_5"      "4 x 4" 

---

    Code
      dm_for_filter() %>% dm_zoom_to(tf_2) %>% as_dm_zoomed_df() %>% tbl_sum()
    Output
      Zoomed table     A tibble 
            "tf_2"      "6 x 4" 

# zoom output for compound keys

    Code
      nyc_comp() %>% dm_zoom_to(weather)
    Output
      # Zoomed table: weather
      # A tibble:     144 x 15
         origin  year month   day  hour  temp  dewp humid wind_dir wind_speed
         <chr>  <int> <int> <int> <int> <dbl> <dbl> <dbl>    <dbl>      <dbl>
       1 EWR     2013     1    10     0  41    32    70.1      230       8.06
       2 EWR     2013     1    10     1  39.0  30.0  69.9      210       9.21
       3 EWR     2013     1    10     2  39.0  28.9  66.8      230       6.90
       4 EWR     2013     1    10     3  39.9  27.0  59.5      270       5.75
       5 EWR     2013     1    10     4  41    26.1  55.0      320       6.90
       6 EWR     2013     1    10     5  41    26.1  55.0      300      12.7 
       7 EWR     2013     1    10     6  39.9  25.0  54.8      280       6.90
       8 EWR     2013     1    10     7  41    25.0  52.6      330       6.90
       9 EWR     2013     1    10     8  43.0  25.0  48.7      330       8.06
      10 EWR     2013     1    10     9  45.0  23    41.6      320      17.3 
      # i 134 more rows
      # i 5 more variables: wind_gust <dbl>, precip <dbl>, pressure <dbl>,
      #   visib <dbl>, time_hour <chr>
    Code
      nyc_comp() %>% dm_zoom_to(weather) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 53
      Primary keys: 4
      Foreign keys: 4
    Code
      nyc_comp_2 <- nyc_comp() %>% dm_zoom_to(weather) %>% dm_insert_zoomed(
        "weather_2")
      nyc_comp_2 %>% get_all_keys()
    Output
      $pks
      # A tibble: 5 x 3
        table     pk_col            autoincrement
        <chr>     <keys>            <lgl>        
      1 airlines  carrier           FALSE        
      2 airports  faa               FALSE        
      3 planes    tailnum           FALSE        
      4 weather   origin, time_hour FALSE        
      5 weather_2 origin, time_hour FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols     parent_table parent_key_cols   on_delete
        <chr>       <keys>            <chr>        <keys>            <chr>    
      1 flights     carrier           airlines     carrier           no_action
      2 flights     dest              airports     faa               no_action
      3 flights     tailnum           planes       tailnum           no_action
      4 flights     origin, time_hour weather      origin, time_hour no_action
      5 flights     origin, time_hour weather_2    origin, time_hour no_action
      
    Code
      attr(graph_edges(create_graph_from_dm(nyc_comp_2)), "vnames")
    Output
      [1] "airlines|flights"  "airports|flights"  "flights|planes"   
      [4] "flights|weather"   "flights|weather_2"
    Code
      nyc_comp_3 <- nyc_comp() %>% dm_zoom_to(flights) %>% dm_insert_zoomed(
        "flights_2")
      nyc_comp_3 %>% get_all_keys()
    Output
      $pks
      # A tibble: 4 x 3
        table    pk_col            autoincrement
        <chr>    <keys>            <lgl>        
      1 airlines carrier           FALSE        
      2 airports faa               FALSE        
      3 planes   tailnum           FALSE        
      4 weather  origin, time_hour FALSE        
      
      $fks
      # A tibble: 8 x 5
        child_table child_fk_cols     parent_table parent_key_cols   on_delete
        <chr>       <keys>            <chr>        <keys>            <chr>    
      1 flights     carrier           airlines     carrier           no_action
      2 flights_2   carrier           airlines     carrier           no_action
      3 flights     dest              airports     faa               no_action
      4 flights_2   dest              airports     faa               no_action
      5 flights     tailnum           planes       tailnum           no_action
      6 flights_2   tailnum           planes       tailnum           no_action
      7 flights     origin, time_hour weather      origin, time_hour no_action
      8 flights_2   origin, time_hour weather      origin, time_hour no_action
      
    Code
      attr(graph_edges(create_graph_from_dm(nyc_comp_3)), "vnames")
    Output
      [1] "airlines|flights"   "airlines|flights_2" "airports|flights"  
      [4] "airports|flights_2" "flights|planes"     "planes|flights_2"  
      [7] "flights|weather"    "weather|flights_2" 

# dm_add_pk() - abort_key_set_force_false

    Code
      dm_add_pk(d, a, x)
    Condition
      Error in `dm_add_pk()`:
      ! Table a already has a primary key. Use `force = TRUE` to change the existing primary key.

# dm_add_pk() - abort_not_unique_key (via check)

    Code
      dm_add_pk(d, a, x, check = TRUE)
    Condition
      Error in `check_key()`:
      ! (`x`) not a unique key of `a`.

# dm_rm_pk() - abort_pk_not_defined

    Code
      dm_rm_pk(d, a)
    Condition
      Error in `dm_rm_pk()`:
      ! No primary keys to remove.

# dm_add_fk() - abort_ref_tbl_has_no_pk

    Code
      dm_add_fk(d, a, x, ref_table = b)
    Condition
      Error in `dm_add_fk()`:
      ! ref_table b needs a primary key first. Use `dm_enum_pk_candidates()` to find appropriate columns and `dm_add_pk()` to define a primary key.

# dm_add_fk() - abort_fk_exists

    Code
      dm_add_fk(d, a, x, ref_table = b)
    Condition
      Error in `dm_add_fk()`:
      ! (`x`) is already a foreign key of table a into table b.

# dm_add_fk() - abort_not_subset_of (via check)

    Code
      dm_add_fk(d, a, x, ref_table = b, check = TRUE)
    Condition
      Error in `dm_add_fk()`:
      ! Column (`x`) of table a contains values (see examples above) that are not present in column (`x`) of table b.

# dm_rm_fk() - abort_is_not_fkc

    Code
      dm_rm_fk(d, a, x, ref_table = b)
    Condition
      Error in `dm_rm_fk()`:
      ! No foreign keys to remove.

# dm_add_uk() - abort_no_uk_if_pk (PK exists)

    Code
      dm_add_uk(d, a, x)
    Condition
      Error in `dm_add_uk()`:
      ! A PK (`x`) for table a already exists, not adding UK.

# dm_add_uk() - abort_no_uk_if_pk (UK exists)

    Code
      dm_add_uk(d, a, x)
    Condition
      Error in `dm_add_uk()`:
      ! A UK (`x`) for table a already exists, not adding UK.

# dm_rm_uk() - abort_uk_not_defined

    Code
      dm_rm_uk(d, a)
    Condition
      Error in `dm_rm_uk()`:
      ! No unique keys to remove.

# check_key() - abort_not_unique_key

    Code
      check_key(t, x)
    Condition
      Error in `check_key()`:
      ! (`x`) not a unique key of `t`.

# check_cardinality_1_1() - abort_not_bijective

    Code
      check_cardinality_1_1(parent, child, by_position = TRUE)
    Condition
      Error in `check_cardinality_1_1()`:
      ! 1..1 cardinality (bijectivity) is not given: Column (`x`) in table child contains duplicate values.

# check_cardinality_0_1() - abort_not_injective

    Code
      check_cardinality_0_1(parent, child, by_position = TRUE)
    Condition
      Error in `check_cardinality_0_1()`:
      ! 0..1 cardinality (injectivity from child table to parent table) is not given: Column (`x`) in table child contains duplicate values.

# dm_set_colors() - abort_is_not_dm

    Code
      dm_set_colors("not_a_dm", blue = a)
    Condition
      Error in `dm_set_colors()`:
      ! Required class <dm> but instead is <character>.

# dm_get_colors() - abort_is_not_dm

    Code
      dm_get_colors("not_a_dm")
    Condition
      Error in `dm_get_colors()`:
      ! Required class <dm> but instead is <character>.

# dm_discard_zoomed() - abort_is_not_dm

    Code
      dm_discard_zoomed("not_a_dm")
    Condition
      Error in `dm_discard_zoomed()`:
      ! Required class <dm> but instead is <character>.

# dm_wrap_tbl() - abort_is_not_dm

    Code
      dm_wrap_tbl("not_a_dm", root = a)
    Condition
      Error in `dm_wrap_tbl()`:
      ! Required class <dm> but instead is <character>.

# dm_set_colors() - abort_only_named_args

    Code
      dm_set_colors(d, a)
    Condition
      Error in `dm_set_colors()`:
      ! All `...` arguments must be named. The names represent the colors.

# dm_set_colors() - abort_wrong_syntax_set_cols

    Code
      dm_set_colors(d, a = "blue")
    Condition
      Error in `dm_set_colors()`:
      ! You seem to be using outdated syntax for setting colors, type `?dm_set_colors()` for examples.

# dm_set_colors() - abort_cols_not_avail

    Code
      dm_set_colors(d, nonexistent_color_xyz = a)
    Condition
      Error in `dm_set_colors()`:
      ! The color "nonexistent_color_xyz" is not available. Call `dm_get_available_colors()` for possible color names or use hex color codes.

# dm_get_con() - abort_con_only_for_dbi

    Code
      dm_get_con(d)
    Condition
      Error in `dm_get_con()`:
      ! A local <dm> doesn't have a DB connection.

# $<-.dm - abort_update_not_supported

    Code
      d$a <- tibble(x = 2)
    Condition
      Error in `$<-`:
      ! Updating <dm> objects not supported.

# [<-.dm - abort_update_not_supported

    Code
      d[["a"]] <- tibble(x = 2)
    Condition
      Error in `[[<-`:
      ! Updating <dm> objects not supported.

# dm_validate() - abort_is_not_dm

    Code
      dm_validate("not_a_dm")
    Condition
      Error in `dm_validate()`:
      ! Required class <dm> but instead is <character>.

# dm_validate() - abort_dm_invalid

    Code
      dm_validate(bad)
    Condition
      Error in `dm_validate()`:
      ! This <dm> is invalid, reason: A `dm` needs to be a list of one item named `def`.

# dm_zoom_to() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_zoom_to(d, b)
    Condition
      Error in `dm_zoom_to()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_update_zoomed() on unzoomed dm - abort_only_possible_w_zoom

    Code
      dm_update_zoomed(d)
    Condition
      Error in `dm_update_zoomed()`:
      ! Not supported on an unzoomed <dm>. Consider using `dm_zoom_to()` first.

# dm_flatten_to_tbl() with recursive and unsupported join - abort_squash_limited

    Code
      dm_flatten_to_tbl(d, a, .recursive = TRUE, .join = right_join)
    Condition
      Error in `dm_flatten_to_tbl()`:
      ! Recursive flattening only supports `left_join()`, `inner_join()`, or `full_join()`.

# dm_flatten_to_tbl() with nest_join - abort_no_flatten_with_nest_join

    Code
      dm_flatten_to_tbl(d, a, .join = nest_join)
    Condition
      Error in `dm_flatten_to_tbl()`:
      ! `join = nest_join` is not supported. Consider `join = left_join`.

# dm_rename_tbl() - abort_need_unique_names

    Code
      dm_rename_tbl(d, a = b)
    Condition
      Error in `dm_rename_tbl()`:
      ! Each new table needs to have a unique name. Duplicate new name: a.

# decompose_table() - abort_dupl_new_id_col_name

    Code
      decompose_table(t, x, y)
    Condition
      Error in `decompose_table()`:
      ! `new_id_column` can't have an identical name as one of the columns of t.

# dm_paste() - abort_unknown_option

    Code
      dm_paste(d, options = "nonexistent_option")
    Condition
      Error in `dm_paste()`:
      ! Option unknown: "nonexistent_option". Must be one of "all", "tables", "keys", "select", and "color".

# check_set_equality() - abort_sets_not_equal

    Code
      check_set_equality(parent, child, by_position = TRUE)
    Output
      # A tibble: 1 x 1
            x
        <int>
      1     3
    Condition
      Error in `check_set_equality()`:
      ! Column (`x`) of table parent contains values (see examples above) that are not present in column (`x`) of table child.

# dm_draw() - unsupported backend_opts

    Code
      dm_draw(d, backend_opts = list(unsupported_option = TRUE))
    Condition
      Error in `dm_draw()`:
      ! Unsupported `backend_opts` for backend "DiagrammeR": unsupported_option.
      i Supported options are: graph_attrs, node_attrs, edge_attrs, focus, graph_name, column_arrow, and font_size.

# copy_to.dm() - abort_only_data_frames_supported

    Code
      copy_to(d, 42, name = "b")
    Condition
      Error in `copy_to()`:
      ! Only class <data.frame> is supported for argument `df`.

# copy_to.dm() - abort_no_overwrite

    Code
      copy_to(d, tibble(y = 1), name = "b", overwrite = TRUE)
    Condition
      Error in `copy_to()`:
      ! The `overwrite` argument is not supported.

# copy_to.dm() - abort_one_name_for_copy_to

    Code
      copy_to(d, tibble(y = 1), name = c("b", "c"))
    Condition
      Error in `copy_to()`:
      ! Argument `name` must have length 1, not length 2.

# pull_tbl() on dm with table not in dm - abort_table_not_in_dm

    Code
      pull_tbl(d, nonexistent)
    Condition
      Error in `pull_tbl()`:
      ! Table nonexistent not in <dm> object. Available table names: a.

# pull_tbl() on zoomed dm with wrong table - abort_table_not_zoomed

    Code
      pull_tbl(d, b)
    Condition
      Error in `pull_tbl()`:
      ! Table `b` not zoomed, zoomed tables: `a`.

# dm_flatten_to_tbl() with unrelated tables - abort_tables_not_reachable_from_start

    Code
      dm_flatten_to_tbl(d, a, c)
    Condition
      Error in `dm_flatten_to_tbl()`:
      ! All selected tables must be reachable from `.start`.

# dm_flatten_to_tbl() with grandparent - abort_only_parents

    Code
      dm_flatten_to_tbl(d, a, b, c)
    Condition
      Error in `dm_flatten_to_tbl()`:
      ! All join partners of table `.start` must be its direct neighbors. Use `.recursive = TRUE` for recursive flattening.

# dm_flatten_to_tbl() with cycle - abort_no_cycles

    Code
      dm_flatten_to_tbl(dm_for_filter_w_cycle(), tf_5, .recursive = TRUE)
    Condition
      Error in `dm_flatten_to_tbl()`:
      ! Cycles in the relationship graph not yet supported.
      i Shortest cycle: tf_5 -> tf_6 -> tf_7 -> tf_2 -> tf_3 -> tf_4 -> tf_5

# dm_select_tbl() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_select_tbl(d, a)
    Condition
      Error in `dm_select_tbl()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_add_pk() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_add_pk(d, b, x)
    Condition
      Error in `dm_add_pk()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_add_fk() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_add_fk(d, a, x, ref_table = b)
    Condition
      Error in `dm_add_fk()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_get_con() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_get_con(d)
    Condition
      Error in `dm_get_con()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_insert_zoomed() on unzoomed dm - abort_only_possible_w_zoom

    Code
      dm_insert_zoomed(d)
    Condition
      Error in `dm_insert_zoomed()`:
      ! Not supported on an unzoomed <dm>. Consider using `dm_zoom_to()` first.

# dm_draw() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_draw(d)
    Condition
      Error in `dm_draw()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_paste() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_paste(d)
    Condition
      Error in `dm_paste()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_flatten() - abort_tables_not_neighbors

    Code
      dm_flatten(d, a, parent_tables = c)
    Condition
      Error in `dm_flatten()`:
      ! All selected tables must be reachable from `table`.

# pull_tbl() - abort_no_table_provided

    Code
      pull_tbl(d, )
    Condition
      Error in `pull_tbl()`:
      ! Argument `table` is missing.

# dm_flatten_to_tbl() - abort_no_flatten_with_nest_join (via dm_join_to_tbl)

    Code
      dm_join_to_tbl(d, a, b, join = nest_join)
    Condition
      Error in `dm_join_to_tbl()`:
      ! `join = nest_join` is not supported. Consider `join = left_join`.

# dm_examine_constraints() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_examine_constraints(d)
    Condition
      Error in `dm_examine_constraints()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_disambiguate_cols() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_disambiguate_cols(d)
    Condition
      Error in `dm_disambiguate_cols()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_rm_pk() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_rm_pk(d, a)
    Condition
      Error in `dm_rm_pk()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_rm_fk() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_rm_fk(d, a, x, ref_table = b)
    Condition
      Error in `dm_rm_fk()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_add_uk() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_add_uk(d, a, x)
    Condition
      Error in `dm_add_uk()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_rm_uk() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_rm_uk(d, a)
    Condition
      Error in `dm_rm_uk()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_has_pk() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_has_pk(d, a)
    Condition
      Error in `dm_has_pk()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_get_pk() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_get_pk(d, a)
    Condition
      Error in `dm_get_pk()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_get_all_pks() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_get_all_pks(d)
    Condition
      Error in `dm_get_all_pks()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_get_all_fks() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_get_all_fks(d)
    Condition
      Error in `dm_get_all_fks()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_get_all_uks() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_get_all_uks(d)
    Condition
      Error in `dm_get_all_uks()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_enum_pk_candidates() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_enum_pk_candidates(d, a)
    Condition
      Error in `dm_enum_pk_candidates()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_enum_fk_candidates() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_enum_fk_candidates(d, a, b)
    Condition
      Error in `dm_enum_fk_candidates()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_rename() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_rename(d, a, y = x)
    Condition
      Error in `dm_rename()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_select() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_select(d, a, x)
    Condition
      Error in `dm_select()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_set_table_description() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_set_table_description(d, a = "test")
    Condition
      Error in `dm_set_table_description()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# dm_get_tables() on zoomed dm - abort_only_possible_wo_zoom

    Code
      dm_get_tables(d)
    Condition
      Error in `dm_get_tables()`:
      ! Not supported on a <dm_zoomed>. Consider using one of `dm_update_zoomed()`, `dm_insert_zoomed()` or `dm_discard_zoomed()` first.

# check_cardinality_0_n() - abort_not_subset_of

    Code
      check_cardinality_0_n(parent, child, by_position = TRUE)
    Output
      # A tibble: 1 x 1
            x
        <dbl>
      1     3
    Condition
      Error in `check_cardinality_0_n()`:
      ! Column (`x`) of table child contains values (see examples above) that are not present in column (`x`) of table parent.

# check_cardinality_1_n() - abort_not_unique_key and not_subset_of

    Code
      check_cardinality_1_n(parent, child, by_position = TRUE)
    Condition
      Error in `check_cardinality_1_n()`:
      ! (`x`) not a unique key of `parent`.

# check_subset() - abort_not_subset_of

    Code
      check_subset(child, parent, by_position = TRUE)
    Output
      # A tibble: 1 x 1
            x
        <dbl>
      1     3
    Condition
      Error in `check_subset()`:
      ! Column (`x`) of table child contains values (see examples above) that are not present in column (`x`) of table parent.

# build_copy_queries snapshot test for pixarfilms

    Code
      pixar_dm %>% build_copy_queries(src_db, .) %>% as.list()
    Output
      $name
      [1] "pixar_films"     "academy"         "box_office"      "genres"         
      [5] "public_response"
      
      $remote_name
            pixar_films           academy        box_office            genres 
          "pixar_films"         "academy"      "box_office"          "genres" 
        public_response 
      "public_response" 
      
      $columns
      $columns[[1]]
      [1] "number"       "film"         "release_date" "run_time"     "film_rating" 
      
      $columns[[2]]
      [1] "film"       "award_type" "status"    
      
      $columns[[3]]
      [1] "film"                 "budget"               "box_office_us_canada"
      [4] "box_office_other"     "box_office_worldwide"
      
      $columns[[4]]
      [1] "film"  "genre"
      
      $columns[[5]]
      [1] "film"            "rotten_tomatoes" "metacritic"      "cinema_score"   
      [5] "critics_choice" 
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE `pixar_films` (
        `number` VARCHAR(255),
        `film` VARCHAR(255),
        `release_date` DATE,
        `run_time` DOUBLE,
        `film_rating` VARCHAR(255),
        PRIMARY KEY (`film`)
      )
      <SQL> CREATE TEMPORARY TABLE `academy` (
        `film` VARCHAR(255),
        `award_type` VARCHAR(255),
        `status` VARCHAR(255),
        PRIMARY KEY (`film`, `award_type`)
      )
      <SQL> CREATE TEMPORARY TABLE `box_office` (
        `film` VARCHAR(255),
        `budget` DOUBLE,
        `box_office_us_canada` DOUBLE,
        `box_office_other` DOUBLE,
        `box_office_worldwide` DOUBLE,
        PRIMARY KEY (`film`)
      )
      <SQL> CREATE TEMPORARY TABLE `genres` (
        `film` VARCHAR(255),
        `genre` VARCHAR(255),
        PRIMARY KEY (`film`, `genre`)
      )
      <SQL> CREATE TEMPORARY TABLE `public_response` (
        `film` VARCHAR(255),
        `rotten_tomatoes` DOUBLE,
        `metacritic` DOUBLE,
        `cinema_score` VARCHAR(255),
        `critics_choice` DOUBLE,
        PRIMARY KEY (`film`)
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      NULL
      
      $sql_index[[3]]
      NULL
      
      $sql_index[[4]]
      NULL
      
      $sql_index[[5]]
      NULL
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      NULL
      
      $index_name[[3]]
      NULL
      
      $index_name[[4]]
      NULL
      
      $index_name[[5]]
      NULL
      
      

# build_copy_queries snapshot test for dm_for_filter()

    Code
      dm_for_filter() %>% collect() %>% build_copy_queries(src_db, .) %>% as.list()
    Output
      $name
      [1] "tf_1" "tf_3" "tf_6" "tf_2" "tf_4" "tf_5"
      
      $remote_name
        tf_1   tf_3   tf_6   tf_2   tf_4   tf_5 
      "tf_1" "tf_3" "tf_6" "tf_2" "tf_4" "tf_5" 
      
      $columns
      $columns[[1]]
      [1] "a" "b"
      
      $columns[[2]]
      [1] "f"  "f1" "g" 
      
      $columns[[3]]
      [1] "zz" "n"  "o" 
      
      $columns[[4]]
      [1] "c"  "d"  "e"  "e1"
      
      $columns[[5]]
      [1] "h"  "i"  "j"  "j1"
      
      $columns[[6]]
      [1] "ww" "k"  "l"  "m" 
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE `tf_1` (
        `a` INT AUTO_INCREMENT,
        `b` VARCHAR(255),
        PRIMARY KEY (`a`)
      )
      <SQL> CREATE TEMPORARY TABLE `tf_3` (
        `f` VARCHAR(255),
        `f1` INT,
        `g` VARCHAR(255),
        PRIMARY KEY (`f`, `f1`)
      )
      <SQL> CREATE TEMPORARY TABLE `tf_6` (
        `zz` INT,
        `n` VARCHAR(255),
        `o` VARCHAR(255),
        PRIMARY KEY (`o`),
        UNIQUE (`n`)
      )
      <SQL> CREATE TEMPORARY TABLE `tf_2` (
        `c` VARCHAR(255),
        `d` INT,
        `e` VARCHAR(255),
        `e1` INT,
        PRIMARY KEY (`c`)
      )
      <SQL> CREATE TEMPORARY TABLE `tf_4` (
        `h` VARCHAR(255),
        `i` VARCHAR(255),
        `j` VARCHAR(255),
        `j1` INT,
        PRIMARY KEY (`h`)
      )
      <SQL> CREATE TEMPORARY TABLE `tf_5` (
        `ww` INT,
        `k` INT,
        `l` VARCHAR(255),
        `m` VARCHAR(255),
        PRIMARY KEY (`k`)
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      NULL
      
      $sql_index[[3]]
      NULL
      
      $sql_index[[4]]
      NULL
      
      $sql_index[[5]]
      NULL
      
      $sql_index[[6]]
      NULL
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      NULL
      
      $index_name[[3]]
      NULL
      
      $index_name[[4]]
      NULL
      
      $index_name[[5]]
      NULL
      
      $index_name[[6]]
      NULL
      
      

# build_copy_queries avoids duplicate indexes

    Code
      as.list(queries)
    Output
      $name
      [1] "parent1"  "parent2"  "child"    "child__a"
      
      $remote_name
      $remote_name$parent1
      <Id> "parent1"
      
      $remote_name$parent2
      <Id> "parent2"
      
      $remote_name$child
      <Id> "child"
      
      $remote_name$child__a
      <Id> "child__a"
      
      
      $columns
      $columns[[1]]
      [1] "key"
      
      $columns[[2]]
      [1] "a__key"
      
      $columns[[3]]
      [1] "a__key"
      
      $columns[[4]]
      [1] "key"
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE `parent1` (
        `key` DOUBLE,
        PRIMARY KEY (`key`)
      )
      <SQL> CREATE TEMPORARY TABLE `parent2` (
        `a__key` DOUBLE,
        PRIMARY KEY (`a__key`)
      )
      <SQL> CREATE TEMPORARY TABLE `child` (
        `a__key` DOUBLE
      )
      <SQL> CREATE TEMPORARY TABLE `child__a` (
        `key` DOUBLE
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      NULL
      
      $sql_index[[3]]
      NULL
      
      $sql_index[[4]]
      NULL
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      NULL
      
      $index_name[[3]]
      NULL
      
      $index_name[[4]]
      NULL
      
      

# `dm_wrap_tbl()` and `dm_unwrap_tbl()` work

    Code
      wrapped <- dm_wrap_tbl(dm_for_filter(), tf_4)
      wrapped
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_4`
      Columns: 6
      Primary keys: 1
      Foreign keys: 0
    Code
      wrapped$tf_4
    Output
      # A tibble: 5 x 6
        h     i     j        j1 tf_3$g $tf_2            tf_5            
        <chr> <chr> <chr> <int> <chr>  <nested>         <nested>        
      1 a     three C         3 two    <tibble [0 x 3]> <tibble [0 x 4]>
      2 b     four  D         4 three  <tibble [1 x 3]> <tibble [1 x 4]>
      3 c     five  E         5 four   <tibble [2 x 3]> <tibble [1 x 4]>
      4 d     six   F         6 five   <tibble [2 x 3]> <tibble [1 x 4]>
      5 e     seven F         6 five   <tibble [2 x 3]> <tibble [1 x 4]>
    Code
      wrapped$tf_4$tf_3$tf_2[[3]]
    Output
      # A tibble: 2 x 3
        c         d tf_1$b
        <chr> <int> <chr> 
      1 lion      3 C     
      2 dog       6 F     
    Code
      wrapped$tf_4$tf_5[[2]]
    Output
      # A tibble: 1 x 4
           ww     k m     tf_6$zz $o   
        <int> <int> <chr>   <int> <chr>
      1     2     1 house       1 e    

---

    Code
      unwrapped <- dm_unwrap_tbl(dm_wrap_tbl(dm_for_filter(), tf_4), dm_for_filter())
      unwrapped
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_4`, `tf_5`, `tf_3`, `tf_6`, `tf_2`, `tf_1`
      Columns: 20
      Primary keys: 6
      Foreign keys: 5
    Code
      unwrapped$tf_4
    Output
      # A tibble: 5 x 4
        h     i     j        j1
        <chr> <chr> <chr> <int>
      1 a     three C         3
      2 b     four  D         4
      3 c     five  E         5
      4 d     six   F         6
      5 e     seven F         6
    Code
      unwrapped$tf_1
    Output
      # A tibble: 5 x 2
            a b    
        <int> <chr>
      1     2 B    
      2     3 C    
      3     6 F    
      4     4 D    
      5     7 G    
    Code
      unwrapped$tf_6
    Output
      # A tibble: 3 x 3
        n             zz o    
        <chr>      <int> <chr>
      1 house          1 e    
      2 tree           1 f    
      3 streetlamp     1 h    

# `node_type_from_graph()` works

    Code
      node_type_from_graph(graph)
    Output
                   tf_1              tf_2              tf_3              tf_4 
      "terminal parent"    "intermediate"    "intermediate"    "intermediate" 
                   tf_5              tf_6 
         "intermediate" "terminal parent" 

---

    Code
      node_type_from_graph(graph, drop = "tf_4")
    Output
                   tf_1              tf_2              tf_3              tf_5 
      "terminal parent"    "intermediate"    "intermediate"    "intermediate" 
                   tf_6 
      "terminal parent" 

# `dm_wrap_tbl()` fails for cyclic dm

    Code
      dm_wrap_tbl(cyclic_dm, a)
    Condition
      Error in `dm_wrap_tbl()`:
      ! The <dm> is not cycle free and can't be wrapped into a single tibble.

# snapshot of code_generation_middleware.R is unchanged

    Code
      parent <- tibble(a = c(1L, 1:3), b = -1)
      child <- tibble(a = 1:4, c = 3)
      dm <- dm(parent, child)
      ops <- enum_ops(dm)
      ops
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      
      $single
      $single$op_name
      [1] "dm_rm_fk"
      
      
      $multiple
      $multiple$table_names
      [1] "parent" "child" 
      
      
    Code
      enum_ops(dm, op_name = "dm_add_pk")
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      $input$op_name
      [1] "dm_add_pk"
      
      
      $single
      named list()
      
      $multiple
      $multiple$table_names
      [1] "parent" "child" 
      
      
    Code
      enum_ops(dm, op_name = "dm_add_pk", table_names = "parent")
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      $input$op_name
      [1] "dm_add_pk"
      
      $input$table_names
      [1] "parent"
      
      
      $single
      named list()
      
      $multiple
      $multiple$column_names
      [1] "a" "b"
      
      
    Code
      enum_ops(dm, op_name = "dm_add_pk", table_names = "parent", column_names = "a")$
        call
    Output
      dm_add_pk(., parent, a)
    Code
      enum_ops(dm, op_name = "dm_add_pk", table_names = "parent", column_names = c(
        "a", "b"))$call
    Output
      dm_add_pk(., parent, c(a, b))

# snapshot of code_generation_middleware_2.R is unchanged

    Code
      parent <- tibble(a = c(1L, 1:3), b = -1)
      child <- tibble(a = 1:4, c = 3)
      dm <- dm(parent, child)
      ops <- enum_ops(dm)
      ops
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      
      $single
      $single$op_name
      [1] "dm_rm_fk"
      
      
      $multiple
      $multiple$table_names
      [1] "parent" "child" 
      
      
    Code
      ops$multiple
    Output
      $table_names
      [1] "parent" "child" 
      
    Code
      user_choice <- list(table_names = "parent")
      ops_2 <- exec(enum_ops, !!!ops$input, !!!user_choice)
      ops_2
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      $input$table_names
      [1] "parent"
      
      
      $single
      $single$op_name
      [1] "dm_add_pk" "dm_rm_fk" 
      
      
      $multiple
      $multiple$column_names
      [1] "a" "b"
      
      
    Code
      ops_2$single
    Output
      $op_name
      [1] "dm_add_pk" "dm_rm_fk" 
      
    Code
      user_choice_2 <- list(op_name = "dm_add_pk")
      ops_3 <- exec(enum_ops, !!!ops_2$input, !!!user_choice_2)
      ops_3
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      $input$op_name
      [1] "dm_add_pk"
      
      $input$table_names
      [1] "parent"
      
      
      $single
      named list()
      
      $multiple
      $multiple$column_names
      [1] "a" "b"
      
      
    Code
      ops_3$multiple
    Output
      $column_names
      [1] "a" "b"
      
    Code
      user_choice_3 <- list(column_names = c("a"))
      ops_4 <- exec(enum_ops, !!!ops_3$input, !!!user_choice_3)
      ops_4
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      $input$table_names
      [1] "parent"
      
      $input$column_names
      [1] "a"
      
      $input$op_name
      [1] "dm_add_pk"
      
      
      $call
      dm_add_pk(., parent, a)
      
    Code
      ops_4$call
    Output
      dm_add_pk(., parent, a)

# snapshot of code_generation_middleware_3.R is unchanged

    Code
      parent <- tibble(a = c(1L, 1:3), b = -1)
      child <- tibble(a = 1:4, c = 3)
      dm <- dm(parent, child)
      ops <- enum_ops(dm)
      ops
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      
      $single
      $single$op_name
      [1] "dm_rm_fk"
      
      
      $multiple
      $multiple$table_names
      [1] "parent" "child" 
      
      
    Code
      ops$multiple
    Output
      $table_names
      [1] "parent" "child" 
      
    Code
      user_choice <- list(table_names = "parent")
      ops_2 <- exec(enum_ops, !!!ops$input, !!!user_choice)
      ops_2
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      $input$table_names
      [1] "parent"
      
      
      $single
      $single$op_name
      [1] "dm_add_pk" "dm_rm_fk" 
      
      
      $multiple
      $multiple$column_names
      [1] "a" "b"
      
      
    Code
      ops_2$multiple
    Output
      $column_names
      [1] "a" "b"
      
    Code
      user_choice_2 <- list(column_names = c("a"))
      ops_3 <- exec(enum_ops, !!!ops_2$input, !!!user_choice_2)
      ops_3
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      $input$column_names
      [1] "a"
      
      $input$table_names
      [1] "parent"
      
      
      $single
      $single$op_name
      [1] "dm_add_pk"
      
      
      $multiple
      named list()
      
    Code
      ops_3$single
    Output
      $op_name
      [1] "dm_add_pk"
      
    Code
      user_choice_3 <- list(op_name = "dm_add_pk")
      ops_4 <- exec(enum_ops, !!!ops_3$input, !!!user_choice_3)
      ops_4
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 0
      Foreign keys: 0
      
      $input$table_names
      [1] "parent"
      
      $input$column_names
      [1] "a"
      
      $input$op_name
      [1] "dm_add_pk"
      
      
      $call
      dm_add_pk(., parent, a)
      
    Code
      ops_4$call
    Output
      dm_add_pk(., parent, a)

# snapshot of code_generation_middleware_4.R is unchanged

    Code
      parent <- tibble(a = c(1L, 1:3), b = -1)
      child <- tibble(a = 1:4, c = 3)
      dm <- dm(parent, child) %>% dm_add_pk(parent, b)
      ops <- enum_ops(dm)
      ops
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 1
      Foreign keys: 0
      
      
      $single
      $single$op_name
      [1] "dm_rm_fk"
      
      
      $multiple
      $multiple$table_names
      [1] "parent" "child" 
      
      
    Code
      ops$multiple
    Output
      $table_names
      [1] "parent" "child" 
      
    Code
      user_choice <- list(table_names = "parent")
      ops_2 <- exec(enum_ops, !!!ops$input, !!!user_choice)
      ops_2
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 1
      Foreign keys: 0
      
      $input$table_names
      [1] "parent"
      
      
      $single
      $single$op_name
      [1] "dm_add_pk" "dm_rm_fk" 
      
      
      $multiple
      $multiple$column_names
      [1] "a" "b"
      
      
    Code
      ops_2$multiple
    Output
      $column_names
      [1] "a" "b"
      
    Code
      user_choice_2 <- list(column_names = c("a"))
      ops_3 <- exec(enum_ops, !!!ops_2$input, !!!user_choice_2)
      ops_3
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 1
      Foreign keys: 0
      
      $input$column_names
      [1] "a"
      
      $input$table_names
      [1] "parent"
      
      
      $single
      $single$op_name
      [1] "dm_add_pk"
      
      
      $multiple
      named list()
      
    Code
      ops_3$single
    Output
      $op_name
      [1] "dm_add_pk"
      
    Code
      user_choice_3 <- list(op_name = "dm_add_pk")
      ops_4 <- exec(enum_ops, !!!ops_3$input, !!!user_choice_3)
      ops_4
    Output
      $input
      $input$dm
      -- Metadata --------------------------------------------------------------------
      Tables: `parent`, `child`
      Columns: 4
      Primary keys: 1
      Foreign keys: 0
      
      $input$table_names
      [1] "parent"
      
      $input$column_names
      [1] "a"
      
      $input$op_name
      [1] "dm_add_pk"
      
      
      $call
      dm_add_pk(., parent, a, force = TRUE)
      
      $confirmation_message
      [1] "This table already has a primary key. Please confirm overwriting the existing primary key."
      
    Code
      ops_4$call
    Output
      dm_add_pk(., parent, a, force = TRUE)

# `dm_flatten()` snapshot with `dm_paste()`

    Code
      dm_flatten(dm_for_flatten(), fact) %>% dm_paste(options = c("select", "keys"))
    Message
      Renaming ambiguous columns: %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_2, something.dim_2 = something) %>%
        dm_rename(dim_3, something.dim_3 = something) %>%
        dm_rename(dim_4, something.dim_4 = something)
      dm::dm(
        fact,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something, something.dim_1, something.dim_2, something.dim_3, something.dim_4)

---

    Code
      dm_flatten(dm_for_flatten(), fact, parent_tables = c(dim_1, dim_2)) %>%
        dm_paste(options = c("select", "keys"))
    Message
      Renaming ambiguous columns: %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_2, something.dim_2 = something)
      dm::dm(
        fact,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something, something.dim_1, something.dim_2) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)

# `dm_flatten()` recursive snapshot with `dm_paste()`

    Code
      dm_flatten(dm_more_complex(), tf_5, parent_tables = c(tf_4, tf_3), recursive = TRUE) %>%
        dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        tf_1,
        tf_2,
        tf_5,
        tf_6,
        tf_7,
        tf_6_2,
        tf_4_2,
        a,
        b,
        c,
        d,
        e,
      ) %>%
        dm::dm_select(tf_1, a, b) %>%
        dm::dm_select(tf_2, c, d, e, e1) %>%
        dm::dm_select(tf_5, ww, k, l, m, i, j, j1, g) %>%
        dm::dm_select(tf_6, zz, n, o) %>%
        dm::dm_select(tf_7, p, q) %>%
        dm::dm_select(tf_6_2, p, f, f1) %>%
        dm::dm_select(tf_4_2, r, s, t) %>%
        dm::dm_select(a, a_1, a_2) %>%
        dm::dm_select(b, b_1, b_2, b_3) %>%
        dm::dm_select(c, c_1) %>%
        dm::dm_select(d, d_1, b_1) %>%
        dm::dm_select(e, e_1, b_1) %>%
        dm::dm_add_pk(tf_1, a) %>%
        dm::dm_add_pk(tf_2, c) %>%
        dm::dm_add_pk(tf_5, k) %>%
        dm::dm_add_pk(tf_6, n) %>%
        dm::dm_add_pk(tf_6_2, p) %>%
        dm::dm_add_pk(tf_4_2, r) %>%
        dm::dm_add_pk(a, a_1) %>%
        dm::dm_add_pk(b, b_1) %>%
        dm::dm_add_pk(c, c_1) %>%
        dm::dm_add_pk(d, d_1) %>%
        dm::dm_add_pk(e, e_1) %>%
        dm::dm_add_fk(tf_2, d, tf_1) %>%
        dm::dm_add_fk(tf_5, m, tf_6) %>%
        dm::dm_add_fk(tf_5, l, tf_4_2) %>%
        dm::dm_add_fk(b, b_2, a) %>%
        dm::dm_add_fk(d, b_1, b) %>%
        dm::dm_add_fk(e, b_1, b) %>%
        dm::dm_add_fk(b, b_3, c)

# `dm_flatten()` allow_deep snapshot with `dm_paste()`

    Code
      dm_flatten(dm_deep, A, allow_deep = TRUE) %>% dm_paste(options = c("select",
        "keys"))
    Message
      dm::dm(
        A,
        C,
      ) %>%
        dm::dm_select(A, id, b_id, val_a, c_id, val_b) %>%
        dm::dm_select(C, id, val_c) %>%
        dm::dm_add_pk(C, id) %>%
        dm::dm_add_fk(A, c_id, C)

# `dm_flatten()` with tidyselect via parent_tables

    Code
      dm_flatten(dm_for_flatten(), fact, parent_tables = c(ends_with("3"), ends_with(
        "1"))) %>% dm_paste(options = c("select", "keys"))
    Message
      Renaming ambiguous columns: %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_3, something.dim_3 = something)
      dm::dm(
        fact,
        dim_2,
        dim_4,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something, something.dim_1, something.dim_3) %>%
        dm::dm_select(dim_2, dim_2_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, dim_2_key, dim_2) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)

# `dm_flatten()` handles column disambiguation correctly

    Code
      dm_flatten(dm_for_flatten(), fact, parent_tables = c(dim_1, dim_2)) %>%
        dm_paste(options = c("select", "keys"))
    Message
      Renaming ambiguous columns: %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_2, something.dim_2 = something)
      dm::dm(
        fact,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something, something.dim_1, something.dim_2) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)

# `dm_flatten()` reports renames

    Code
      dm_flatten(dm_for_flatten(), fact, parent_tables = c(dim_1, dim_2)) %>%
        dm_paste(options = c("select", "keys"))
    Message
      Renaming ambiguous columns: %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_2, something.dim_2 = something)
      dm::dm(
        fact,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something, something.dim_1, something.dim_2) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)

# `dm_flatten()` join argument snapshot with `dm_paste()`

    Code
      dm_flatten(dm_for_flatten(), fact, parent_tables = c(dim_1, dim_2), join = inner_join) %>%
        dm_paste(options = c("select", "keys"))
    Message
      Renaming ambiguous columns: %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_2, something.dim_2 = something)
      dm::dm(
        fact,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something, something.dim_1, something.dim_2) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)

# dm_rename() works for replacing pk

    Code
      dm_for_filter() %>% dm_rename(tf_3, new_f = f) %>% dm_get_all_pks_impl()
    Output
      # A tibble: 6 x 3
        table pk_col    autoincrement
        <chr> <keys>    <lgl>        
      1 tf_1  a         TRUE         
      2 tf_2  c         FALSE        
      3 tf_3  new_f, f1 FALSE        
      4 tf_4  h         FALSE        
      5 tf_5  k         FALSE        
      6 tf_6  o         FALSE        

# dm_rename() works for replacing fks

    Code
      dm_for_filter() %>% dm_rename(tf_2, new_d = d, new_e = e) %>%
        dm_get_all_fks_impl()
    Output
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        new_d         tf_1         a               no_action
      2 tf_2        new_e, e1     tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action

# dm_select() works for replacing pk

    Code
      dm_for_filter() %>% dm_select(tf_3, new_f = f) %>% dm_get_all_pks_impl()
    Output
      # A tibble: 5 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_4  h      FALSE        
      4 tf_5  k      FALSE        
      5 tf_6  o      FALSE        

# dm_select() works for replacing fks, and removes missing ones

    Code
      dm_for_filter() %>% dm_select(tf_2, new_d = d) %>% dm_get_all_fks_impl()
    Output
      # A tibble: 4 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        new_d         tf_1         a               no_action
      2 tf_4        j, j1         tf_3         f, f1           no_action
      3 tf_5        l             tf_4         h               cascade  
      4 tf_5        m             tf_6         n               no_action

# output for compound keys

    Code
      dm_select(dm_for_flatten(), fact, dim_1_key_1, dim_1_key_2) %>% dm_paste(
        options = c("select", "keys"))
    Message
      dm::dm(
        fact,
        dim_1,
        dim_2,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(fact, dim_1_key_1, dim_1_key_2) %>%
        dm::dm_select(dim_1, dim_1_pk_1, dim_1_pk_2, something) %>%
        dm::dm_select(dim_2, dim_2_pk, something) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_1, c(dim_1_pk_1, dim_1_pk_2)) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, c(dim_1_key_1, dim_1_key_2), dim_1)
    Code
      dm_select(dm_for_flatten(), dim_1, dim_1_pk_1, dim_1_pk_2) %>% dm_paste(
        options = c("select", "keys"))
    Message
      dm::dm(
        fact,
        dim_1,
        dim_2,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something) %>%
        dm::dm_select(dim_1, dim_1_pk_1, dim_1_pk_2) %>%
        dm::dm_select(dim_2, dim_2_pk, something) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_1, c(dim_1_pk_1, dim_1_pk_2)) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, c(dim_1_key_1, dim_1_key_2), dim_1) %>%
        dm::dm_add_fk(fact, dim_2_key, dim_2) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)
    Code
      dm_select(dm_for_flatten(), fact, -dim_1_key_1) %>% dm_paste(options = c(
        "select", "keys"))
    Message
      dm::dm(
        fact,
        dim_1,
        dim_2,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something) %>%
        dm::dm_select(dim_1, dim_1_pk_1, dim_1_pk_2, something) %>%
        dm::dm_select(dim_2, dim_2_pk, something) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_1, c(dim_1_pk_1, dim_1_pk_2)) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, dim_2_key, dim_2) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)
    Code
      dm_select(dm_for_flatten(), dim_1, -dim_1_pk_1) %>% dm_paste(options = c(
        "select", "keys"))
    Message
      dm::dm(
        fact,
        dim_1,
        dim_2,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something) %>%
        dm::dm_select(dim_1, dim_1_pk_2, something) %>%
        dm::dm_select(dim_2, dim_2_pk, something) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, dim_2_key, dim_2) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)

# key tracking works

    Code
      dm_zoomed() %>% unite("new_col", c, e) %>% dm_update_zoomed() %>% get_all_keys()
    Output
      $pks
      # A tibble: 5 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_3  f, f1  FALSE        
      3 tf_4  h      FALSE        
      4 tf_5  k      FALSE        
      5 tf_6  o      FALSE        
      
      $fks
      # A tibble: 4 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_4        j, j1         tf_3         f, f1           no_action
      3 tf_5        l             tf_4         h               cascade  
      4 tf_5        m             tf_6         n               no_action
      
    Code
      dm_zoomed() %>% unite("new_col", c, e, remove = FALSE) %>% dm_update_zoomed() %>%
        get_all_keys()
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      dm_zoomed() %>% unite("new_col", c, e, remove = FALSE) %>% dm_update_zoomed() %>%
        dm_add_fk(tf_2, new_col, tf_6) %>% dm_zoom_to(tf_2) %>% separate(new_col, c(
        "c", "e"), remove = TRUE) %>% dm_update_zoomed() %>% get_all_keys()
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      dm_zoomed() %>% unite("new_col", c, e, remove = FALSE) %>% dm_update_zoomed() %>%
        dm_add_fk(tf_2, new_col, tf_6) %>% dm_zoom_to(tf_2) %>% separate(new_col, c(
        "c", "e"), remove = FALSE) %>% dm_update_zoomed() %>% get_all_keys()
    Output
      $pks
      # A tibble: 6 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      6 tf_6  o      FALSE        
      
      $fks
      # A tibble: 6 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      6 tf_2        new_col       tf_6         o               no_action
      

# output for compound keys

    Code
      unite_weather_dm <- nyc_comp() %>% dm_zoom_to(weather) %>% mutate(chr_col = "airport") %>%
        unite("new_col", origin, chr_col) %>% dm_update_zoomed()
      unite_weather_dm %>% get_all_keys()
    Output
      $pks
      # A tibble: 3 x 3
        table    pk_col  autoincrement
        <chr>    <keys>  <lgl>        
      1 airlines carrier FALSE        
      2 airports faa     FALSE        
      3 planes   tailnum FALSE        
      
      $fks
      # A tibble: 3 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 flights     carrier       airlines     carrier         no_action
      2 flights     dest          airports     faa             no_action
      3 flights     tailnum       planes       tailnum         no_action
      
    Code
      unite_weather_dm %>% get_all_keys()
    Output
      $pks
      # A tibble: 3 x 3
        table    pk_col  autoincrement
        <chr>    <keys>  <lgl>        
      1 airlines carrier FALSE        
      2 airports faa     FALSE        
      3 planes   tailnum FALSE        
      
      $fks
      # A tibble: 3 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 flights     carrier       airlines     carrier         no_action
      2 flights     dest          airports     faa             no_action
      3 flights     tailnum       planes       tailnum         no_action
      
    Code
      unite_flights_dm <- nyc_comp() %>% dm_zoom_to(flights) %>% mutate(chr_col = "airport") %>%
        unite("new_col", origin, chr_col) %>% dm_update_zoomed()
      unite_flights_dm %>% get_all_keys()
    Output
      $pks
      # A tibble: 4 x 3
        table    pk_col            autoincrement
        <chr>    <keys>            <lgl>        
      1 airlines carrier           FALSE        
      2 airports faa               FALSE        
      3 planes   tailnum           FALSE        
      4 weather  origin, time_hour FALSE        
      
      $fks
      # A tibble: 3 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 flights     carrier       airlines     carrier         no_action
      2 flights     dest          airports     faa             no_action
      3 flights     tailnum       planes       tailnum         no_action
      
    Code
      unite_flights_dm %>% get_all_keys()
    Output
      $pks
      # A tibble: 4 x 3
        table    pk_col            autoincrement
        <chr>    <keys>            <lgl>        
      1 airlines carrier           FALSE        
      2 airports faa               FALSE        
      3 planes   tailnum           FALSE        
      4 weather  origin, time_hour FALSE        
      
      $fks
      # A tibble: 3 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 flights     carrier       airlines     carrier         no_action
      2 flights     dest          airports     faa             no_action
      3 flights     tailnum       planes       tailnum         no_action
      
    Code
      nyc_comp() %>% dm_zoom_to(weather) %>% separate(origin, c("o1", "o2"), sep = "^..",
      remove = TRUE) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 54
      Primary keys: 3
      Foreign keys: 3
    Code
      nyc_comp() %>% dm_zoom_to(weather) %>% separate(origin, c("o1", "o2"), sep = "^..",
      remove = FALSE) %>% dm_update_zoomed()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 55
      Primary keys: 4
      Foreign keys: 4

# `json_pack()` works remotely

    Code
      query <- json_pack(remote, a = starts_with("a")) %>% dbplyr::sql_render()
      gsub("test_frame_[_0-9]+", "test_frame_...", query)
    Output
      <SQL> SELECT "grp", (SELECT value FROM OPENJSON((SELECT "a_i", "a_j" FOR JSON PATH))) AS "a" FROM (SELECT *
      FROM "#test_frame_...") "*tmp*"
    Code
      json_pack(remote, a = starts_with("a"))
    Output
          grp a                              
        <dbl> <chr>                          
      1     1 "{\"a_i\":\"a\",\"a_j\":\"A\"}"
      2     1 "{\"a_i\":\"b\",\"a_j\":\"B\"}"
      3     2 "{\"a_i\":\"c\",\"a_j\":\"C\"}"
      4     2 "{\"a_i\":\"d\",\"a_j\":\"D\"}"
    Code
      query <- json_pack(remote, a = starts_with("a"), .names_sep = "_") %>% dbplyr::sql_render()
      gsub("test_frame_[_0-9]+", "test_frame_...", query)
    Output
      <SQL> SELECT "grp", (SELECT value FROM OPENJSON((SELECT "a_i" "i", "a_j" "j" FOR JSON PATH))) AS "a" FROM (SELECT *
      FROM "#test_frame_...") "*tmp*"
    Code
      json_pack(remote, a = starts_with("a"), .names_sep = "_")
    Output
          grp a                          
        <dbl> <chr>                      
      1     1 "{\"i\":\"a\",\"j\":\"A\"}"
      2     1 "{\"i\":\"b\",\"j\":\"B\"}"
      3     2 "{\"i\":\"c\",\"j\":\"C\"}"
      4     2 "{\"i\":\"d\",\"j\":\"D\"}"

# errors: duplicate table names, src mismatches

    Code
      dm_bind(dm_for_filter(), dm_for_flatten(), dm_for_filter())
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
      Error in `dm_bind()`:
      ! Names must be unique.
      x These names are duplicated:
        * "tf_1" at locations 1 and 12.
        * "tf_2" at locations 2 and 13.
        * "tf_3" at locations 3 and 14.
        * "tf_4" at locations 4 and 15.
        * "tf_5" at locations 5 and 16.
        * ...

# test error output for src mismatches

    Code
      writeLines(conditionMessage(expect_error(dm_bind(dm_for_flatten(),
      dm_for_filter_duckdb()))))
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
    Output
      All <dm> objects need to share the same `src`.

# output

    Code
      dm_bind()
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
    Output
      dm()
    Code
      dm_bind(empty_dm())
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
    Output
      dm()
    Code
      dm_bind(dm_for_filter()) %>% collect()
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_1`, `tf_2`, `tf_3`, `tf_4`, `tf_5`, `tf_6`
      Columns: 20
      Primary keys: 6
      Foreign keys: 5
    Code
      dm_bind(dm_for_filter(), dm_for_flatten(), dm_for_filter(), repair = "unique",
      quiet = TRUE) %>% collect()
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_1...1`, `tf_2...2`, `tf_3...3`, `tf_4...4`, `tf_5...5`, ... (17 total)
      Columns: 56
      Primary keys: 16
      Foreign keys: 14
    Code
      writeLines(conditionMessage(expect_error(dm_bind(dm_for_filter(),
      dm_for_flatten(), dm_for_filter()))))
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
    Output
      Names must be unique.
      x These names are duplicated:
        * "tf_1" at locations 1 and 12.
        * "tf_2" at locations 2 and 13.
        * "tf_3" at locations 3 and 14.
        * "tf_4" at locations 4 and 15.
        * "tf_5" at locations 5 and 16.
        * ...

---

    Code
      dm_bind(dm_for_filter(), dm_for_flatten(), dm_for_filter(), repair = "unique") %>%
        collect()
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
    Message
      New names:
      * `tf_1` -> `tf_1...1`
      * `tf_2` -> `tf_2...2`
      * `tf_3` -> `tf_3...3`
      * `tf_4` -> `tf_4...4`
      * `tf_5` -> `tf_5...5`
      * `tf_6` -> `tf_6...6`
      * `tf_1` -> `tf_1...12`
      * `tf_2` -> `tf_2...13`
      * `tf_3` -> `tf_3...14`
      * `tf_4` -> `tf_4...15`
      * `tf_5` -> `tf_5...16`
      * `tf_6` -> `tf_6...17`
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_1...1`, `tf_2...2`, `tf_3...3`, `tf_4...4`, `tf_5...5`, ... (17 total)
      Columns: 56
      Primary keys: 16
      Foreign keys: 14

# output for compound keys

    Code
      dm_bind(dm_for_filter(), dm_for_flatten()) %>% dm_paste(options = c("select",
        "keys"))
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
    Message
      dm::dm(
        tf_1,
        tf_2,
        tf_3,
        tf_4,
        tf_5,
        tf_6,
        fact,
        dim_1,
        dim_2,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(tf_1, a, b) %>%
        dm::dm_select(tf_2, c, d, e, e1) %>%
        dm::dm_select(tf_3, f, f1, g) %>%
        dm::dm_select(tf_4, h, i, j, j1) %>%
        dm::dm_select(tf_5, ww, k, l, m) %>%
        dm::dm_select(tf_6, zz, n, o) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something) %>%
        dm::dm_select(dim_1, dim_1_pk_1, dim_1_pk_2, something) %>%
        dm::dm_select(dim_2, dim_2_pk, something) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_add_pk(tf_1, a, autoincrement = TRUE) %>%
        dm::dm_add_pk(tf_2, c) %>%
        dm::dm_add_pk(tf_3, c(f, f1)) %>%
        dm::dm_add_pk(tf_4, h) %>%
        dm::dm_add_pk(tf_5, k) %>%
        dm::dm_add_pk(tf_6, o) %>%
        dm::dm_add_pk(dim_1, c(dim_1_pk_1, dim_1_pk_2)) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_uk(tf_3, g) %>%
        dm::dm_add_fk(tf_2, d, tf_1) %>%
        dm::dm_add_fk(tf_2, c(e, e1), tf_3) %>%
        dm::dm_add_fk(tf_4, c(j, j1), tf_3) %>%
        dm::dm_add_fk(tf_5, l, tf_4, on_delete = "cascade") %>%
        dm::dm_add_fk(tf_5, m, tf_6, n) %>%
        dm::dm_add_fk(fact, c(dim_1_key_1, dim_1_key_2), dim_1) %>%
        dm::dm_add_fk(fact, dim_2_key, dim_2) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)
    Code
      dm_bind(dm_for_flatten(), dm_for_filter()) %>% dm_paste(options = c("select",
        "keys"))
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
    Message
      dm::dm(
        fact,
        dim_1,
        dim_2,
        dim_3,
        dim_4,
        tf_1,
        tf_2,
        tf_3,
        tf_4,
        tf_5,
        tf_6,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something) %>%
        dm::dm_select(dim_1, dim_1_pk_1, dim_1_pk_2, something) %>%
        dm::dm_select(dim_2, dim_2_pk, something) %>%
        dm::dm_select(dim_3, dim_3_pk, something) %>%
        dm::dm_select(dim_4, dim_4_pk, something) %>%
        dm::dm_select(tf_1, a, b) %>%
        dm::dm_select(tf_2, c, d, e, e1) %>%
        dm::dm_select(tf_3, f, f1, g) %>%
        dm::dm_select(tf_4, h, i, j, j1) %>%
        dm::dm_select(tf_5, ww, k, l, m) %>%
        dm::dm_select(tf_6, zz, n, o) %>%
        dm::dm_add_pk(dim_1, c(dim_1_pk_1, dim_1_pk_2)) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_pk(tf_1, a, autoincrement = TRUE) %>%
        dm::dm_add_pk(tf_2, c) %>%
        dm::dm_add_pk(tf_3, c(f, f1)) %>%
        dm::dm_add_pk(tf_4, h) %>%
        dm::dm_add_pk(tf_5, k) %>%
        dm::dm_add_pk(tf_6, o) %>%
        dm::dm_add_uk(tf_3, g) %>%
        dm::dm_add_fk(fact, c(dim_1_key_1, dim_1_key_2), dim_1) %>%
        dm::dm_add_fk(fact, dim_2_key, dim_2) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4) %>%
        dm::dm_add_fk(tf_2, d, tf_1) %>%
        dm::dm_add_fk(tf_2, c(e, e1), tf_3) %>%
        dm::dm_add_fk(tf_4, c(j, j1), tf_3) %>%
        dm::dm_add_fk(tf_5, l, tf_4, on_delete = "cascade") %>%
        dm::dm_add_fk(tf_5, m, tf_6, n)

---

    Code
      dm_bind(dm_for_flatten(), dm_for_flatten(), repair = "unique") %>% dm_paste(
        options = c("select", "keys"))
    Condition
      Warning:
      `dm_bind()` was deprecated in dm 1.0.0.
      i Please use `dm()` instead.
    Message
      New names:
      * `fact` -> `fact...1`
      * `dim_1` -> `dim_1...2`
      * `dim_2` -> `dim_2...3`
      * `dim_3` -> `dim_3...4`
      * `dim_4` -> `dim_4...5`
      * `fact` -> `fact...6`
      * `dim_1` -> `dim_1...7`
      * `dim_2` -> `dim_2...8`
      * `dim_3` -> `dim_3...9`
      * `dim_4` -> `dim_4...10`
      dm::dm(
        fact...1,
        dim_1...2,
        dim_2...3,
        dim_3...4,
        dim_4...5,
        fact...6,
        dim_1...7,
        dim_2...8,
        dim_3...9,
        dim_4...10,
      ) %>%
        dm::dm_select(fact...1, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something) %>%
        dm::dm_select(dim_1...2, dim_1_pk_1, dim_1_pk_2, something) %>%
        dm::dm_select(dim_2...3, dim_2_pk, something) %>%
        dm::dm_select(dim_3...4, dim_3_pk, something) %>%
        dm::dm_select(dim_4...5, dim_4_pk, something) %>%
        dm::dm_select(fact...6, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something) %>%
        dm::dm_select(dim_1...7, dim_1_pk_1, dim_1_pk_2, something) %>%
        dm::dm_select(dim_2...8, dim_2_pk, something) %>%
        dm::dm_select(dim_3...9, dim_3_pk, something) %>%
        dm::dm_select(dim_4...10, dim_4_pk, something) %>%
        dm::dm_add_pk(dim_1...2, c(dim_1_pk_1, dim_1_pk_2)) %>%
        dm::dm_add_pk(dim_2...3, dim_2_pk) %>%
        dm::dm_add_pk(dim_3...4, dim_3_pk) %>%
        dm::dm_add_pk(dim_4...5, dim_4_pk) %>%
        dm::dm_add_pk(dim_1...7, c(dim_1_pk_1, dim_1_pk_2)) %>%
        dm::dm_add_pk(dim_2...8, dim_2_pk) %>%
        dm::dm_add_pk(dim_3...9, dim_3_pk) %>%
        dm::dm_add_pk(dim_4...10, dim_4_pk) %>%
        dm::dm_add_fk(fact...1, c(dim_1_key_1, dim_1_key_2), dim_1...2) %>%
        dm::dm_add_fk(fact...1, dim_2_key, dim_2...3) %>%
        dm::dm_add_fk(fact...1, dim_3_key, dim_3...4) %>%
        dm::dm_add_fk(fact...1, dim_4_key, dim_4...5) %>%
        dm::dm_add_fk(fact...6, c(dim_1_key_1, dim_1_key_2), dim_1...7) %>%
        dm::dm_add_fk(fact...6, dim_2_key, dim_2...8) %>%
        dm::dm_add_fk(fact...6, dim_3_key, dim_3...9) %>%
        dm::dm_add_fk(fact...6, dim_4_key, dim_4...10)

# `pack_join()` works

    Code
      pack_join(df1, df2)
    Message
      Joining with `by = join_by(key)`
    Output
      # A tibble: 2 x 3
         col1 key   df2$col2
        <int> <chr>    <int>
      1     1 a            3
      2     2 b            4

---

    Code
      pack_join(df1, df2, name = "packed_col")
    Message
      Joining with `by = join_by(key)`
    Output
      # A tibble: 2 x 3
         col1 key   packed_col$col2
        <int> <chr>           <int>
      1     1 a                   3
      2     2 b                   4

---

    Code
      pack_join(df1, df3, by = c(key = "key3"))
    Output
      # A tibble: 2 x 3
         col1 key   df3$col3
        <int> <chr>    <int>
      1     1 a            3
      2     2 b            4

---

    Code
      pack_join(df1, df3, by = c(key = "key3"), keep = TRUE)
    Output
      # A tibble: 2 x 3
         col1 key   df3$col3 $key3
        <int> <chr>    <int> <chr>
      1     1 a            3 a    
      2     2 b            4 b    

# `pack_join()` works with remote table

    `x` and `y` must share the same src, set `copy = TRUE` (may be slow).

---

    Code
      pack_join(df1, dm_remote$tf_1, by = c(col1 = "a"), copy = TRUE)
    Output
      # A tibble: 2 x 3
         col1 key   `dm_remote$tf_1`$b
        <int> <chr> <chr>             
      1     1 a     A                 
      2     2 b     B                 

---

    Code
      pack_join(df4, df5, by = c(df5 = "col"))
    Output
      # A tibble: 0 x 1
      # i 1 variable: df5 <tibble[,0]>

---

    Code
      pack_join(df5, df6, by = c(col = "df6"))
    Output
      # A tibble: 0 x 2
      # i 2 variables: col <int>, df6 <tibble[,0]>

# build_copy_queries snapshot test for pixarfilms

    Code
      pixar_dm %>% build_copy_queries(src_db, .) %>% as.list()
    Output
      $name
      [1] "pixar_films"     "academy"         "box_office"      "genres"         
      [5] "public_response"
      
      $remote_name
            pixar_films           academy        box_office            genres 
          "pixar_films"         "academy"      "box_office"          "genres" 
        public_response 
      "public_response" 
      
      $columns
      $columns[[1]]
      [1] "number"       "film"         "release_date" "run_time"     "film_rating" 
      
      $columns[[2]]
      [1] "film"       "award_type" "status"    
      
      $columns[[3]]
      [1] "film"                 "budget"               "box_office_us_canada"
      [4] "box_office_other"     "box_office_worldwide"
      
      $columns[[4]]
      [1] "film"  "genre"
      
      $columns[[5]]
      [1] "film"            "rotten_tomatoes" "metacritic"      "cinema_score"   
      [5] "critics_choice" 
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE "pixar_films" (
        "number" varchar(255),
        "film" varchar(255),
        "release_date" DATE,
        "run_time" FLOAT,
        "film_rating" varchar(255),
        PRIMARY KEY ("film")
      )
      <SQL> CREATE TEMPORARY TABLE "academy" (
        "film" varchar(255),
        "award_type" varchar(255),
        "status" varchar(255),
        PRIMARY KEY ("film", "award_type"),
        FOREIGN KEY ("film") REFERENCES "pixar_films" ("film")
      )
      <SQL> CREATE TEMPORARY TABLE "box_office" (
        "film" varchar(255),
        "budget" FLOAT,
        "box_office_us_canada" FLOAT,
        "box_office_other" FLOAT,
        "box_office_worldwide" FLOAT,
        PRIMARY KEY ("film"),
        FOREIGN KEY ("film") REFERENCES "pixar_films" ("film")
      )
      <SQL> CREATE TEMPORARY TABLE "genres" (
        "film" varchar(255),
        "genre" varchar(255),
        PRIMARY KEY ("film", "genre"),
        FOREIGN KEY ("film") REFERENCES "pixar_films" ("film")
      )
      <SQL> CREATE TEMPORARY TABLE "public_response" (
        "film" varchar(255),
        "rotten_tomatoes" FLOAT,
        "metacritic" FLOAT,
        "cinema_score" varchar(255),
        "critics_choice" FLOAT,
        PRIMARY KEY ("film"),
        FOREIGN KEY ("film") REFERENCES "pixar_films" ("film")
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      <SQL> CREATE INDEX academy__film ON "academy" ("film")
      
      $sql_index[[3]]
      <SQL> CREATE INDEX box_office__film ON "box_office" ("film")
      
      $sql_index[[4]]
      <SQL> CREATE INDEX genres__film ON "genres" ("film")
      
      $sql_index[[5]]
      <SQL> CREATE INDEX public_response__film ON "public_response" ("film")
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      [1] "academy__film"
      
      $index_name[[3]]
      [1] "box_office__film"
      
      $index_name[[4]]
      [1] "genres__film"
      
      $index_name[[5]]
      [1] "public_response__film"
      
      

# build_copy_queries snapshot test for dm_for_filter()

    Code
      dm_for_filter() %>% collect() %>% build_copy_queries(src_db, .) %>% as.list()
    Output
      $name
      [1] "tf_1" "tf_3" "tf_6" "tf_2" "tf_4" "tf_5"
      
      $remote_name
        tf_1   tf_3   tf_6   tf_2   tf_4   tf_5 
      "tf_1" "tf_3" "tf_6" "tf_2" "tf_4" "tf_5" 
      
      $columns
      $columns[[1]]
      [1] "a" "b"
      
      $columns[[2]]
      [1] "f"  "f1" "g" 
      
      $columns[[3]]
      [1] "zz" "n"  "o" 
      
      $columns[[4]]
      [1] "c"  "d"  "e"  "e1"
      
      $columns[[5]]
      [1] "h"  "i"  "j"  "j1"
      
      $columns[[6]]
      [1] "ww" "k"  "l"  "m" 
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE "tf_1" (
        "a" INT IDENTITY,
        "b" varchar(255),
        PRIMARY KEY ("a")
      )
      <SQL> CREATE TEMPORARY TABLE "tf_3" (
        "f" varchar(255),
        "f1" INT,
        "g" varchar(255),
        PRIMARY KEY ("f", "f1")
      )
      <SQL> CREATE TEMPORARY TABLE "tf_6" (
        "zz" INT,
        "n" varchar(255),
        "o" varchar(255),
        PRIMARY KEY ("o"),
        UNIQUE ("n")
      )
      <SQL> CREATE TEMPORARY TABLE "tf_2" (
        "c" varchar(255),
        "d" INT,
        "e" varchar(255),
        "e1" INT,
        PRIMARY KEY ("c"),
        FOREIGN KEY ("d") REFERENCES "tf_1" ("a"),
        FOREIGN KEY ("e", "e1") REFERENCES "tf_3" ("f", "f1")
      )
      <SQL> CREATE TEMPORARY TABLE "tf_4" (
        "h" varchar(255),
        "i" varchar(255),
        "j" varchar(255),
        "j1" INT,
        PRIMARY KEY ("h"),
        FOREIGN KEY ("j", "j1") REFERENCES "tf_3" ("f", "f1")
      )
      <SQL> CREATE TEMPORARY TABLE "tf_5" (
        "ww" INT,
        "k" INT,
        "l" varchar(255),
        "m" varchar(255),
        PRIMARY KEY ("k"),
        FOREIGN KEY ("m") REFERENCES "tf_6" ("n"),
        FOREIGN KEY ("l") REFERENCES "tf_4" ("h") ON DELETE CASCADE
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      NULL
      
      $sql_index[[3]]
      NULL
      
      $sql_index[[4]]
      <SQL> CREATE INDEX tf_2__d ON "tf_2" ("d")
      <SQL> CREATE INDEX tf_2__e_e1 ON "tf_2" ("e", "e1")
      
      $sql_index[[5]]
      <SQL> CREATE INDEX tf_4__j_j1 ON "tf_4" ("j", "j1")
      
      $sql_index[[6]]
      <SQL> CREATE INDEX tf_5__m ON "tf_5" ("m")
      <SQL> CREATE INDEX tf_5__l ON "tf_5" ("l")
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      NULL
      
      $index_name[[3]]
      NULL
      
      $index_name[[4]]
      [1] "tf_2__d"    "tf_2__e_e1"
      
      $index_name[[5]]
      [1] "tf_4__j_j1"
      
      $index_name[[6]]
      [1] "tf_5__m" "tf_5__l"
      
      

# build_copy_queries avoids duplicate indexes

    Code
      as.list(queries)
    Output
      $name
      [1] "parent1"  "parent2"  "child"    "child__a"
      
      $remote_name
      $remote_name$parent1
      <Id> "parent1"
      
      $remote_name$parent2
      <Id> "parent2"
      
      $remote_name$child
      <Id> "child"
      
      $remote_name$child__a
      <Id> "child__a"
      
      
      $columns
      $columns[[1]]
      [1] "key"
      
      $columns[[2]]
      [1] "a__key"
      
      $columns[[3]]
      [1] "a__key"
      
      $columns[[4]]
      [1] "key"
      
      
      $sql_table
      <SQL> CREATE TEMPORARY TABLE "parent1" (
        "key" FLOAT,
        PRIMARY KEY ("key")
      )
      <SQL> CREATE TEMPORARY TABLE "parent2" (
        "a__key" FLOAT,
        PRIMARY KEY ("a__key")
      )
      <SQL> CREATE TEMPORARY TABLE "child" (
        "a__key" FLOAT,
        FOREIGN KEY ("a__key") REFERENCES "parent2" ("a__key")
      )
      <SQL> CREATE TEMPORARY TABLE "child__a" (
        "key" FLOAT,
        FOREIGN KEY ("key") REFERENCES "parent2" ("a__key")
      )
      
      $sql_index
      $sql_index[[1]]
      NULL
      
      $sql_index[[2]]
      NULL
      
      $sql_index[[3]]
      <SQL> CREATE INDEX child__a__key ON "child" ("a__key")
      
      $sql_index[[4]]
      <SQL> CREATE INDEX child__a__key__1 ON "child__a" ("key")
      
      
      $index_name
      $index_name[[1]]
      NULL
      
      $index_name[[2]]
      NULL
      
      $index_name[[3]]
      [1] "child__a__key"
      
      $index_name[[4]]
      [1] "child__a__key__1"
      
      

# dm_add_pk() works as intended?

    Code
      dm(x = tibble(x = integer())) %>% dm_add_pk(x)
    Condition
      Error in `dm_add_pk()`:
      ! `columns` is absent but must be supplied.

# dm_rm_pk() supports partial filters

    Code
      dm_for_filter() %>% dm_rm_pk(tf_4) %>% get_all_keys()
    Output
      $pks
      # A tibble: 5 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_5  k      FALSE        
      5 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      dm_for_filter() %>% dm_rm_pk(tf_3) %>% get_all_keys()
    Output
      $pks
      # A tibble: 5 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_4  h      FALSE        
      4 tf_5  k      FALSE        
      5 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      dm_for_filter() %>% dm_rm_pk(tf_6) %>% get_all_keys()
    Output
      $pks
      # A tibble: 5 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_4  h      FALSE        
      5 tf_5  k      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      dm_for_filter() %>% dm_rm_pk(tf_4, rm_referencing_fks = TRUE) %>% get_all_keys()
    Condition
      Warning:
      The `rm_referencing_fks` argument of `dm_rm_pk()` is deprecated as of dm 0.2.1.
      i When removing a primary key, potential associated foreign keys will be pointing at an implicit unique key.
    Output
      $pks
      # A tibble: 5 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_3  f, f1  FALSE        
      4 tf_5  k      FALSE        
      5 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      dm_for_filter() %>% dm_rm_pk(columns = c) %>% get_all_keys()
    Message
      Removing primary keys: %>%
        dm_rm_pk(tf_2)
    Output
      $pks
      # A tibble: 5 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_3  f, f1  FALSE        
      3 tf_4  h      FALSE        
      4 tf_5  k      FALSE        
      5 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      dm_for_filter() %>% dm_rm_pk(columns = c(f, f1)) %>% get_all_keys()
    Message
      Removing primary keys: %>%
        dm_rm_pk(tf_3)
    Output
      $pks
      # A tibble: 5 x 3
        table pk_col autoincrement
        <chr> <keys> <lgl>        
      1 tf_1  a      TRUE         
      2 tf_2  c      FALSE        
      3 tf_4  h      FALSE        
      4 tf_5  k      FALSE        
      5 tf_6  o      FALSE        
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      
    Code
      dm_for_filter() %>% dm_rm_pk(fail_fk = FALSE) %>% get_all_keys()
    Condition
      Warning:
      The `fail_fk` argument of `dm_rm_pk()` is deprecated as of dm 1.0.4.
      i When removing a primary key, potential associated foreign keys will be pointing at an implicit unique key.
    Message
      Removing primary keys: %>%
        dm_rm_pk(tf_1) %>%
        dm_rm_pk(tf_2) %>%
        dm_rm_pk(tf_3) %>%
        dm_rm_pk(tf_4) %>%
        dm_rm_pk(tf_5) %>%
        dm_rm_pk(tf_6)
    Output
      $pks
      # A tibble: 0 x 3
      # i 3 variables: table <chr>, pk_col <keys>, autoincrement <lgl>
      
      $fks
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         n               no_action
      

# dm_enum_pk_candidates() works properly?

    Code
      dm_enum_pk_candidates(dm_test_obj(), dm_table_1)
    Output
      # A tibble: 2 x 3
        columns candidate why  
        <keys>  <lgl>     <chr>
      1 a       TRUE      ""   
      2 b       TRUE      ""   
    Code
      dm_enum_pk_candidates(dm_test_obj(), dm_table_2)
    Output
      # A tibble: 1 x 3
        columns candidate why                        
        <keys>  <lgl>     <chr>                      
      1 c       FALSE     has duplicate values: 5 (2)
    Code
      dm_enum_pk_candidates(dm_test_obj(), dm_table_5)
    Output
      # A tibble: 1 x 3
        columns candidate why                 
        <keys>  <lgl>     <chr>               
      1 c       FALSE     has 1 missing values
    Code
      dm_enum_pk_candidates(dm_test_obj(), dm_table_6)
    Output
      # A tibble: 1 x 3
        columns candidate why                                              
        <keys>  <lgl>     <chr>                                            
      1 c       FALSE     has 1 missing values, and duplicate values: 3 (2)

# output

    Code
      dm(x = tibble(a = c(1, 1))) %>% dm_add_pk(x, a, check = TRUE)
    Condition
      Error in `check_key()`:
      ! (`a`) not a unique key of `x`.

# dm_get_all_pks() with table arg

    Code
      nyc_comp() %>% dm_get_all_pks("weather")
    Output
      # A tibble: 1 x 3
        table   pk_col            autoincrement
        <chr>   <keys>            <lgl>        
      1 weather origin, time_hour FALSE        
    Code
      nyc_comp() %>% dm_get_all_pks(c(airlines, weather))
    Output
      # A tibble: 2 x 3
        table    pk_col            autoincrement
        <chr>    <keys>            <lgl>        
      1 airlines carrier           FALSE        
      2 weather  origin, time_hour FALSE        
    Code
      nyc_comp() %>% dm_get_all_pks(starts_with("weat"))
    Output
      # A tibble: 1 x 3
        table   pk_col            autoincrement
        <chr>   <keys>            <lgl>        
      1 weather origin, time_hour FALSE        
    Code
      nyc_comp() %>% dm_get_all_pks(matches("^a.*s$"))
    Output
      # A tibble: 2 x 3
        table    pk_col  autoincrement
        <chr>    <keys>  <lgl>        
      1 airlines carrier FALSE        
      2 airports faa     FALSE        

# dm_get_all_pks() with compound keys

    Code
      nyc_comp()
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 53
      Primary keys: 4
      Foreign keys: 4
    Code
      nyc_comp() %>% dm_get_all_pks()
    Output
      # A tibble: 4 x 3
        table    pk_col            autoincrement
        <chr>    <keys>            <lgl>        
      1 airlines carrier           FALSE        
      2 airports faa               FALSE        
      3 planes   tailnum           FALSE        
      4 weather  origin, time_hour FALSE        

# autoincrement fails with compound keys

    Code
      dm(x) %>% dm_add_pk(x, columns = c(x_id, z), autoincrement = TRUE)
    Condition
      Error in `dm_add_pk()`:
      ! Composite primary keys cannot be autoincremented.
      i Provide only a single column name to `columns`.

# set autoincrement PK

    Code
      dm(x, y = x) %>% dm_add_pk(x, columns = c(x_id), autoincrement = TRUE) %>%
        dm_add_pk(y, columns = c(x_id, z)) %>% dm_get_all_pks()
    Output
      # A tibble: 2 x 3
        table pk_col  autoincrement
        <chr> <keys>  <lgl>        
      1 x     x_id    TRUE         
      2 y     x_id, z FALSE        

# dm_rows_append() works with autoincrement PKs and FKS locally

    Code
      local_dm$t1
    Output
      # A tibble: 3 x 2
            a o    
        <int> <chr>
      1     5 a    
      2     6 b    
      3     7 c    
    Code
      local_dm$t2
    Output
      # A tibble: 3 x 3
            c     d o    
        <int> <int> <chr>
      1    10     7 c    
      2     9     6 b    
      3     8     5 a    
    Code
      local_dm$t3
    Output
      # A tibble: 3 x 2
            e o    
        <int> <chr>
      1     6 b    
      2     5 a    
      3     7 c    
    Code
      local_dm$t4
    Output
      # A tibble: 3 x 3
            g     h o    
        <int> <int> <chr>
      1     1     8 a    
      2     2     9 b    
      3     3    10 c    
    Code
      filled_dm$t1
    Output
      # A tibble: 3 x 2
            a o    
        <int> <chr>
      1     1 a    
      2     2 b    
      3     3 c    
    Code
      filled_dm$t2
    Output
      # A tibble: 3 x 3
            c     d o    
        <int> <int> <chr>
      1     1     3 c    
      2     2     2 b    
      3     3     1 a    
    Code
      filled_dm$t3
    Output
      # A tibble: 0 x 2
      # i 2 variables: e <int>, o <chr>
    Code
      filled_dm$t4
    Output
      # A tibble: 3 x 3
            g     h o    
        <int> <int> <chr>
      1     1     3 a    
      2     2     2 b    
      3     3     1 c    
    Code
      filled_twice_dm$t1
    Output
      # A tibble: 6 x 2
            a o    
        <int> <chr>
      1     1 a    
      2     2 b    
      3     3 c    
      4     4 a    
      5     5 b    
      6     6 c    
    Code
      filled_twice_dm$t2
    Output
      # A tibble: 6 x 3
            c     d o    
        <int> <int> <chr>
      1     1     3 c    
      2     2     2 b    
      3     3     1 a    
      4     4     6 c    
      5     5     5 b    
      6     6     4 a    
    Code
      filled_twice_dm$t3
    Output
      # A tibble: 0 x 2
      # i 2 variables: e <int>, o <chr>
    Code
      filled_twice_dm$t4
    Output
      # A tibble: 6 x 3
            g     h o    
        <int> <int> <chr>
      1     1     3 a    
      2     2     2 b    
      3     3     1 c    
      4     4     6 a    
      5     5     5 b    
      6     6     4 c    

# `dm_flatten_to_tbl()` does the right things for 'left_join()'

    Code
      prepare_dm_for_flatten(dm_for_flatten(), tables = c("fact", "dim_1", "dim_2",
        "dim_3", "dim_4"), gotta_rename = TRUE) %>% dm_get_tables()
    Message
      Renaming ambiguous columns: %>%
        dm_rename(fact, something.fact = something) %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_2, something.dim_2 = something) %>%
        dm_rename(dim_3, something.dim_3 = something) %>%
        dm_rename(dim_4, something.dim_4 = something)
    Output
      $fact
         fact     dim_1_key_1 dim_1_key_2 dim_2_key dim_3_key dim_4_key something.fact
         <chr>          <int> <chr>       <chr>     <chr>         <int>          <int>
       1 acorn             14 N           c         X                 7              1
       2 blubber           13 M           d         W                 8              2
       3 cindere~          12 L           e         V                 9              3
       4 depth             11 K           f         U                10              4
       5 elysium           10 J           g         T                11              5
       6 fantasy            9 I           h         S                12              6
       7 gorgeous           8 H           i         R                13              7
       8 halo               7 G           j         Q                14              8
       9 ill-adv~           6 F           k         P                15              9
      10 jitter             5 E           l         O                16             10
      
      $dim_1
         dim_1_pk_1 dim_1_pk_2 something.dim_1
              <int> <chr>      <chr>          
       1          1 A          c              
       2          2 B          d              
       3          3 C          e              
       4          4 D          f              
       5          5 E          g              
       6          6 F          h              
       7          7 G          i              
       8          8 H          j              
       9          9 I          k              
      10         10 J          l              
      # i more rows
      
      $dim_2
         dim_2_pk something.dim_2
         <chr>    <chr>          
       1 a        E              
       2 b        F              
       3 c        G              
       4 d        H              
       5 e        I              
       6 f        J              
       7 g        K              
       8 h        L              
       9 i        M              
      10 j        N              
      # i more rows
      
      $dim_3
         dim_3_pk something.dim_3
         <chr>              <int>
       1 E                      3
       2 F                      4
       3 G                      5
       4 H                      6
       5 I                      7
       6 J                      8
       7 K                      9
       8 L                     10
       9 M                     11
      10 N                     12
      # i more rows
      
      $dim_4
         dim_4_pk something.dim_4
            <int>           <int>
       1       19              19
       2       18              20
       3       17              21
       4       16              22
       5       15              23
       6       14              24
       7       13              25
       8       12              26
       9       11              27
      10       10              28
      # i more rows
      
    Code
      dm_flatten_to_tbl(dm_for_flatten(), fact)
    Message
      Renaming ambiguous columns: %>%
        dm_rename(fact, something.fact = something) %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_2, something.dim_2 = something) %>%
        dm_rename(dim_3, something.dim_3 = something) %>%
        dm_rename(dim_4, something.dim_4 = something)
    Output
         fact     dim_1_key_1 dim_1_key_2 dim_2_key dim_3_key dim_4_key something.fact
         <chr>          <int> <chr>       <chr>     <chr>         <int>          <int>
       1 acorn             14 N           c         X                 7              1
       2 blubber           13 M           d         W                 8              2
       3 cindere~          12 L           e         V                 9              3
       4 depth             11 K           f         U                10              4
       5 elysium           10 J           g         T                11              5
       6 fantasy            9 I           h         S                12              6
       7 gorgeous           8 H           i         R                13              7
       8 halo               7 G           j         Q                14              8
       9 ill-adv~           6 F           k         P                15              9
      10 jitter             5 E           l         O                16             10
      # i 4 more variables: something.dim_1 <chr>, something.dim_2 <chr>,
      #   something.dim_3 <int>, something.dim_4 <int>
    Code
      result_from_flatten_new()
    Output
         fact     dim_1_key_1 dim_1_key_2 dim_2_key dim_3_key dim_4_key something.fact
         <chr>          <int> <chr>       <chr>     <chr>         <int>          <int>
       1 acorn             14 N           c         X                 7              1
       2 blubber           13 M           d         W                 8              2
       3 cindere~          12 L           e         V                 9              3
       4 depth             11 K           f         U                10              4
       5 elysium           10 J           g         T                11              5
       6 fantasy            9 I           h         S                12              6
       7 gorgeous           8 H           i         R                13              7
       8 halo               7 G           j         Q                14              8
       9 ill-adv~           6 F           k         P                15              9
      10 jitter             5 E           l         O                16             10
      # i 4 more variables: something.dim_1 <chr>, something.dim_2 <chr>,
      #   something.dim_3 <int>, something.dim_4 <int>

# `dm_flatten_to_tbl()` does the right things for 'inner_join()'

    Code
      out
    Output
         fact     dim_1_key_1 dim_1_key_2 dim_2_key dim_3_key dim_4_key something.fact
         <chr>          <int> <chr>       <chr>     <chr>         <int>          <int>
       1 acorn             14 N           c         X                 7              1
       2 blubber           13 M           d         W                 8              2
       3 cindere~          12 L           e         V                 9              3
       4 depth             11 K           f         U                10              4
       5 elysium           10 J           g         T                11              5
       6 fantasy            9 I           h         S                12              6
       7 gorgeous           8 H           i         R                13              7
       8 halo               7 G           j         Q                14              8
       9 ill-adv~           6 F           k         P                15              9
      10 jitter             5 E           l         O                16             10
      # i 4 more variables: something.dim_1 <chr>, something.dim_2 <chr>,
      #   something.dim_3 <int>, something.dim_4 <int>

# dummy

    Code
      # dummy

# dm_sql()

    Code
      dm_for_filter_df() %>% dm_sql(my_test_con())
    Output
      $pre
      $pre$tf_1
      <SQL> CREATE TEMPORARY TABLE `tf_1` (
        `a` INTEGER,
        `b` TEXT,
        PRIMARY KEY (`a`),
        UNIQUE (`a`)
      )
      
      $pre$tf_3
      <SQL> CREATE TEMPORARY TABLE `tf_3` (
        `f` TEXT,
        `f1` INTEGER,
        `g` TEXT,
        PRIMARY KEY (`f`, `f1`),
        UNIQUE (`f`, `f1`),
        UNIQUE (`g`)
      )
      
      $pre$tf_6
      <SQL> CREATE TEMPORARY TABLE `tf_6` (
        `zz` INTEGER,
        `n` TEXT,
        `o` TEXT,
        PRIMARY KEY (`o`),
        UNIQUE (`o`),
        UNIQUE (`n`)
      )
      
      $pre$tf_2
      <SQL> CREATE TEMPORARY TABLE `tf_2` (
        `c` TEXT,
        `d` INTEGER,
        `e` TEXT,
        `e1` INTEGER,
        PRIMARY KEY (`c`),
        UNIQUE (`c`),
        FOREIGN KEY (`d`) REFERENCES `tf_1` (`a`),
        FOREIGN KEY (`e`, `e1`) REFERENCES `tf_3` (`f`, `f1`)
      )
      
      $pre$tf_4
      <SQL> CREATE TEMPORARY TABLE `tf_4` (
        `h` TEXT,
        `i` TEXT,
        `j` TEXT,
        `j1` INTEGER,
        PRIMARY KEY (`h`),
        UNIQUE (`h`),
        FOREIGN KEY (`j`, `j1`) REFERENCES `tf_3` (`f`, `f1`)
      )
      
      $pre$tf_5
      <SQL> CREATE TEMPORARY TABLE `tf_5` (
        `ww` INTEGER,
        `k` INTEGER,
        `l` TEXT,
        `m` TEXT,
        PRIMARY KEY (`k`),
        UNIQUE (`k`),
        FOREIGN KEY (`m`) REFERENCES `tf_6` (`n`),
        FOREIGN KEY (`l`) REFERENCES `tf_4` (`h`) ON DELETE CASCADE
      )
      
      
      $load
      $load$tf_1
      <SQL> INSERT INTO `tf_1` (`a`, `b`)
      SELECT CAST(`a` AS INTEGER) AS `a`, CAST(`b` AS TEXT) AS `b`
      FROM (
        SELECT NULL AS `a`, NULL AS `b`
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          (1, 'A'),
          (2, 'B'),
          (3, 'C'),
          (4, 'D'),
          (5, 'E'),
          (6, 'F'),
          (7, 'G'),
          (8, 'H'),
          (9, 'I'),
          (10, 'J')
      ) AS `values_table`
      
      $load$tf_3
      <SQL> INSERT INTO `tf_3` (`f`, `f1`, `g`)
      SELECT
        CAST(`f` AS TEXT) AS `f`,
        CAST(`f1` AS INTEGER) AS `f1`,
        CAST(`g` AS TEXT) AS `g`
      FROM (
        SELECT NULL AS `f`, NULL AS `f1`, NULL AS `g`
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          ('C', 2, 'one'),
          ('C', 3, 'two'),
          ('D', 4, 'three'),
          ('E', 5, 'four'),
          ('F', 6, 'five'),
          ('G', 7, 'six'),
          ('H', 7, 'seven'),
          ('I', 7, 'eight'),
          ('J', 10, 'nine'),
          ('K', 11, 'ten')
      ) AS `values_table`
      
      $load$tf_6
      <SQL> INSERT INTO `tf_6` (`zz`, `n`, `o`)
      SELECT
        CAST(`zz` AS INTEGER) AS `zz`,
        CAST(`n` AS TEXT) AS `n`,
        CAST(`o` AS TEXT) AS `o`
      FROM (
        SELECT NULL AS `zz`, NULL AS `n`, NULL AS `o`
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          (1, 'garden', 'i'),
          (1, 'hill', 'g'),
          (1, 'house', 'e'),
          (1, 'streetlamp', 'h'),
          (1, 'tree', 'f')
      ) AS `values_table`
      
      $load$tf_2
      <SQL> INSERT INTO `tf_2` (`c`, `d`, `e`, `e1`)
      SELECT
        CAST(`c` AS TEXT) AS `c`,
        CAST(`d` AS INTEGER) AS `d`,
        CAST(`e` AS TEXT) AS `e`,
        CAST(`e1` AS INTEGER) AS `e1`
      FROM (
        SELECT NULL AS `c`, NULL AS `d`, NULL AS `e`, NULL AS `e1`
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          ('cat', 7, 'F', 6),
          ('dog', 6, 'E', 5),
          ('elephant', 2, 'D', 4),
          ('lion', 3, 'E', 5),
          ('seal', 4, 'F', 6),
          ('worm', 5, 'G', 7)
      ) AS `values_table`
      
      $load$tf_4
      <SQL> INSERT INTO `tf_4` (`h`, `i`, `j`, `j1`)
      SELECT
        CAST(`h` AS TEXT) AS `h`,
        CAST(`i` AS TEXT) AS `i`,
        CAST(`j` AS TEXT) AS `j`,
        CAST(`j1` AS INTEGER) AS `j1`
      FROM (
        SELECT NULL AS `h`, NULL AS `i`, NULL AS `j`, NULL AS `j1`
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          ('a', 'three', 'C', 3),
          ('b', 'four', 'D', 4),
          ('c', 'five', 'E', 5),
          ('d', 'six', 'F', 6),
          ('e', 'seven', 'F', 6)
      ) AS `values_table`
      
      $load$tf_5
      <SQL> INSERT INTO `tf_5` (`ww`, `k`, `l`, `m`)
      SELECT
        CAST(`ww` AS INTEGER) AS `ww`,
        CAST(`k` AS INTEGER) AS `k`,
        CAST(`l` AS TEXT) AS `l`,
        CAST(`m` AS TEXT) AS `m`
      FROM (
        SELECT NULL AS `ww`, NULL AS `k`, NULL AS `l`, NULL AS `m`
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          (2, 1, 'b', 'house'),
          (2, 2, 'c', 'tree'),
          (2, 3, 'd', 'streetlamp'),
          (2, 4, 'e', 'streetlamp')
      ) AS `values_table`
      
      
      $post
      $post$uk
      named list()
      
      $post$fk
      named list()
      
      $post$indexes
      $post$indexes$tf_2
      <SQL> CREATE INDEX tf_2__d ON `tf_2` (`d`)
      <SQL> CREATE INDEX tf_2__e_e1 ON `tf_2` (`e`, `e1`)
      
      $post$indexes$tf_4
      <SQL> CREATE INDEX tf_4__j_j1 ON `tf_4` (`j`, `j1`)
      
      $post$indexes$tf_5
      <SQL> CREATE INDEX tf_5__m ON `tf_5` (`m`)
      <SQL> CREATE INDEX tf_5__l ON `tf_5` (`l`)
      
      
      

---

    Code
      dm(x = data.frame(a = strrep("x", 300))) %>% dm_sql(my_test_con())
    Output
      $pre
      $pre$x
      <SQL> CREATE TEMPORARY TABLE `x` (
        `a` TEXT
      )
      
      
      $load
      $load$x
      <SQL> INSERT INTO `x` (`a`)
      SELECT CAST(`a` AS TEXT) AS `a`
      FROM (
        SELECT NULL AS `a`
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES ('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx')
      ) AS `values_table`
      
      
      $post
      $post$uk
      named list()
      
      $post$fk
      named list()
      
      $post$indexes
      named list()
      
      

---

    Code
      dm(x = data.frame(a = strrep("x", 10000))) %>% dm_sql(my_test_con())
    Output
      $pre
      $pre$x
      <SQL> CREATE TEMPORARY TABLE `x` (
        `a` TEXT
      )
      
      
      $load
      $load$x
      <SQL> INSERT INTO `x` (`a`)
      SELECT CAST(`a` AS TEXT) AS `a`
      FROM (
        SELECT NULL AS `a`
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES ('[encoded data, 10000 characters]')
      ) AS `values_table`
      
      
      $post
      $post$uk
      named list()
      
      $post$fk
      named list()
      
      $post$indexes
      named list()
      
      

# waldo

    Code
      dm %>% waldo::compare(dm, max_diffs = 10)
    Output
      v No differences

---

    Code
      dm %>% dm_select_tbl(-airlines) %>% waldo::compare(dm, max_diffs = 10)
    Output
      `old` is length 4
      `new` is length 5
      
      `names(old)[1:3]`:            "airports" "flights" "planes"
      `names(new)[1:4]`: "airlines" "airports" "flights" "planes"
      
      `old$airlines` is absent
      `new$airlines` is a list

---

    Code
      dm %>% dm_select(airlines, -name) %>% waldo::compare(dm, max_diffs = 10)
    Output
      `old$airlines$data` is length 1
      `new$airlines$data` is length 2
      
      `names(old$airlines$data)`: "carrier"       
      `names(new$airlines$data)`: "carrier" "name"
      
      `old$airlines$data$name` is absent
      `new$airlines$data$name` is a character vector ('Endeavor Air Inc.', 'American Airlines Inc.', 'Alaska Airlines Inc.', 'JetBlue Airways', 'Delta Air Lines Inc.', ...)

---

    Code
      dm %>% dm_rm_fk() %>% waldo::compare(dm, max_diffs = 10)
    Message
      Removing foreign keys: %>%
        dm_rm_fk(flights, carrier, airlines) %>%
        dm_rm_fk(flights, origin, airports) %>%
        dm_rm_fk(flights, tailnum, planes) %>%
        dm_rm_fk(flights, c(origin, time_hour), weather)
    Output
      `old$airlines$fks` is length 0
      `new$airlines$fks` is length 1
      
      `names(old$airlines$fks)`:          
      `names(new$airlines$fks)`: "flights"
      
      `old$airlines$fks$flights` is absent
      `new$airlines$fks$flights` is a list
      
      `old$airports$fks` is length 0
      `new$airports$fks` is length 1
      
      `names(old$airports$fks)`:          
      `names(new$airports$fks)`: "flights"
      
      `old$airports$fks$flights` is absent
      `new$airports$fks$flights` is a list
      
      `old$planes$fks` is length 0
      `new$planes$fks` is length 1
      
      `names(old$planes$fks)`:          
      `names(new$planes$fks)`: "flights"
      
      `old$planes$fks$flights` is absent
      `new$planes$fks$flights` is a list
      
      `old$weather$fks` is length 0
      `new$weather$fks` is length 1
      
      And 2 more differences ...

---

    Code
      dm %>% dm_rm_pk() %>% waldo::compare(dm, max_diffs = 10)
    Message
      Removing primary keys: %>%
        dm_rm_pk(airlines) %>%
        dm_rm_pk(airports) %>%
        dm_rm_pk(planes) %>%
        dm_rm_pk(weather)
    Output
      `old$airlines$pks$column` is length 0
      `new$airlines$pks$column` is length 1
      
      `old$airlines$pks$column[[1]]` is absent
      `new$airlines$pks$column[[1]]` is a character vector ('carrier')
      
      `old$airlines$pks$autoincrement`:      
      `new$airlines$pks$autoincrement`: FALSE
      
      `old$airports$pks$column` is length 0
      `new$airports$pks$column` is length 1
      
      `old$airports$pks$column[[1]]` is absent
      `new$airports$pks$column[[1]]` is a character vector ('faa')
      
      `old$airports$pks$autoincrement`:      
      `new$airports$pks$autoincrement`: FALSE
      
      `old$planes$pks$column` is length 0
      `new$planes$pks$column` is length 1
      
      `old$planes$pks$column[[1]]` is absent
      `new$planes$pks$column[[1]]` is a character vector ('tailnum')
      
      `old$planes$pks$autoincrement`:      
      `new$planes$pks$autoincrement`: FALSE
      
      `old$weather$pks$column` is length 0
      `new$weather$pks$column` is length 1
      
      And 2 more differences ...

---

    Code
      dm %>% dm_set_colors(yellow = flights) %>% waldo::compare(dm, max_diffs = 10)
    Output
      `old$flights$display`: "#FFFF00FF"
      `new$flights$display`: "#5B9BD5FF"

---

    Code
      dm %>% dm_zoom_to(flights) %>% waldo::compare(dm, max_diffs = 10)
    Output
      `old$flights$zoom` is an S3 object of class <tbl_df/tbl/data.frame>, a list
      `new$flights$zoom` is NULL
      
      `old$flights$col_tracker_zoom` is a character vector ('year', 'month', 'day', 'dep_time', 'sched_dep_time', ...)
      `new$flights$col_tracker_zoom` is NULL

# Simple duckplyr test

    Code
      summarize(dm_zoom_to(dm(a = duckplyr::duckdb_tibble(x = 1:3, .prudence = "stingy")),
      a), mean_x = mean(x))
    Output
      # Zoomed table:          a
      # A duckplyr data frame: 1 variable
        mean_x
         <dbl>
      1      2

# unique keys

    Code
      nyc_1_uk <- dm_add_uk(dm_nycflights_small(), flights, everything())
      nyc_1_uk %>% dm_get_all_uks()
    Output
      # A tibble: 4 x 3
        table    uk_col                      kind       
        <chr>    <keys>                      <chr>      
      1 airlines carrier                     PK         
      2 airports faa                         PK         
      3 planes   tailnum                     PK         
      4 flights  year, month, ... (19 total) explicit UK
    Code
      nyc_1_uk %>% dm_add_uk(flights, c(origin, dest, time_hour)) %>% dm_add_uk(
        planes, c(year, manufacturer, model)) %>% dm_get_all_uks()
    Output
      # A tibble: 6 x 3
        table    uk_col                      kind       
        <chr>    <keys>                      <chr>      
      1 airlines carrier                     PK         
      2 airports faa                         PK         
      3 planes   tailnum                     PK         
      4 flights  year, month, ... (19 total) explicit UK
      5 flights  origin, dest, time_hour     explicit UK
      6 planes   year, manufacturer, model   explicit UK
    Code
      nyc_1_uk %>% dm_add_uk(flights, c(origin, dest, time_hour)) %>% dm_add_uk(
        planes, c(year, manufacturer, model)) %>% dm_rm_uk(flights, c(origin, dest,
        time_hour)) %>% dm_get_all_uks()
    Output
      # A tibble: 5 x 3
        table    uk_col                      kind       
        <chr>    <keys>                      <chr>      
      1 airlines carrier                     PK         
      2 airports faa                         PK         
      3 planes   tailnum                     PK         
      4 flights  year, month, ... (19 total) explicit UK
      5 planes   year, manufacturer, model   explicit UK
    Code
      nyc_1_uk %>% dm_add_uk(flights, c(origin, dest, time_hour)) %>% dm_add_uk(
        planes, c(year, manufacturer, model)) %>% dm_rm_uk(flights, everything()) %>%
        dm_get_all_uks()
    Output
      # A tibble: 5 x 3
        table    uk_col                    kind       
        <chr>    <keys>                    <chr>      
      1 airlines carrier                   PK         
      2 airports faa                       PK         
      3 planes   tailnum                   PK         
      4 flights  origin, dest, time_hour   explicit UK
      5 planes   year, manufacturer, model explicit UK
    Code
      nyc_1_uk %>% dm_rm_uk() %>% dm_get_all_uks()
    Message
      Removing unique keys: %>%
        dm_rm_uk(flights, c(year, month, day, dep_time, sched_dep_time, dep_delay, arr_time, sched_arr_time, arr_delay, carrier, flight, tailnum, origin, dest, air_time, distance, hour, minute, time_hour))
    Output
      # A tibble: 3 x 3
        table    uk_col  kind 
        <chr>    <keys>  <chr>
      1 airlines carrier PK   
      2 airports faa     PK   
      3 planes   tailnum PK   
    Code
      nyc_1_uk %>% dm_rm_uk(flights) %>% dm_get_all_uks()
    Message
      Removing unique keys: %>%
        dm_rm_uk(flights, c(year, month, day, dep_time, sched_dep_time, dep_delay, arr_time, sched_arr_time, arr_delay, carrier, flight, tailnum, origin, dest, air_time, distance, hour, minute, time_hour))
    Output
      # A tibble: 3 x 3
        table    uk_col  kind 
        <chr>    <keys>  <chr>
      1 airlines carrier PK   
      2 airports faa     PK   
      3 planes   tailnum PK   
    Code
      nyc_1_uk %>% dm_rm_uk(flights, everything()) %>% dm_get_all_uks()
    Output
      # A tibble: 3 x 3
        table    uk_col  kind 
        <chr>    <keys>  <chr>
      1 airlines carrier PK   
      2 airports faa     PK   
      3 planes   tailnum PK   
    Code
      nyc_1_uk %>% dm_add_uk(flights, c(origin, dest, time_hour)) %>% dm_add_uk(
        planes, c(year, manufacturer, model)) %>% dm_rm_uk() %>% dm_get_all_uks()
    Message
      Removing unique keys: %>%
        dm_rm_uk(flights, c(year, month, day, dep_time, sched_dep_time, dep_delay, arr_time, sched_arr_time, arr_delay, carrier, flight, tailnum, origin, dest, air_time, distance, hour, minute, time_hour)) %>%
        dm_rm_uk(flights, c(origin, dest, time_hour)) %>%
        dm_rm_uk(planes, c(year, manufacturer, model))
    Output
      # A tibble: 3 x 3
        table    uk_col  kind 
        <chr>    <keys>  <chr>
      1 airlines carrier PK   
      2 airports faa     PK   
      3 planes   tailnum PK   
    Code
      nyc_1_uk %>% dm_add_uk(flights, c(origin, dest, time_hour)) %>% dm_add_uk(
        planes, manufacturer) %>% dm_rm_uk() %>% dm_get_all_uks()
    Message
      Removing unique keys: %>%
        dm_rm_uk(flights, c(year, month, day, dep_time, sched_dep_time, dep_delay, arr_time, sched_arr_time, arr_delay, carrier, flight, tailnum, origin, dest, air_time, distance, hour, minute, time_hour)) %>%
        dm_rm_uk(flights, c(origin, dest, time_hour)) %>%
        dm_rm_uk(planes, manufacturer)
    Output
      # A tibble: 3 x 3
        table    uk_col  kind 
        <chr>    <keys>  <chr>
      1 airlines carrier PK   
      2 airports faa     PK   
      3 planes   tailnum PK   
    Code
      dm_examine_constraints(dm_nycflights_small() %>% dm_add_uk(planes, c(year,
        manufacturer, model)))
    Message
      ! Unsatisfied constraints:
    Output
      * Table `planes`: unique key `year`, `manufacturer`, `model`: has duplicate values: 2002, EMBRAER, EMB-145LR (19), 2001, EMBRAER, EMB-145LR (18), 2008, BOMBARDIER INC, CL-600-2D24 (18), 2007, BOMBARDIER INC, CL-600-2D24 (17), 1999, EMBRAER, EMB-145LR (16), ...
      * Table `flights`: foreign key `dest` into table `airports`: values of `flights$dest` not in `airports$faa`: SJU (30), BQN (6), STT (4), PSE (2)
      * Table `flights`: foreign key `tailnum` into table `planes`: values of `flights$tailnum` not in `planes$tailnum`: N725MQ (6), N537MQ (5), N722MQ (5), N730MQ (5), N736MQ (5), ...
    Code
      dm_add_fk(dm_nycflights_small(), flights, time_hour, weather, time_hour) %>%
        dm_get_all_uks()
    Output
      # A tibble: 4 x 3
        table    uk_col    kind       
        <chr>    <keys>    <chr>      
      1 airlines carrier   PK         
      2 airports faa       PK         
      3 planes   tailnum   PK         
      4 weather  time_hour implicit UK
    Code
      dm_add_fk(dm_nycflights_small(), flights, time_hour, weather, time_hour) %>%
        dm_add_uk(weather, time_hour) %>% dm_get_all_uks()
    Output
      # A tibble: 4 x 3
        table    uk_col    kind       
        <chr>    <keys>    <chr>      
      1 airlines carrier   PK         
      2 airports faa       PK         
      3 planes   tailnum   PK         
      4 weather  time_hour explicit UK
    Code
      dm_rename(dm_for_filter(), tf_6, p = n) %>% dm_get_all_uks()
    Output
      # A tibble: 8 x 3
        table uk_col kind       
        <chr> <keys> <chr>      
      1 tf_1  a      PK         
      2 tf_2  c      PK         
      3 tf_3  f, f1  PK         
      4 tf_4  h      PK         
      5 tf_5  k      PK         
      6 tf_6  o      PK         
      7 tf_3  g      explicit UK
      8 tf_6  p      implicit UK
    Code
      dm_rename(dm_for_filter(), tf_6, p = n) %>% dm_get_all_fks()
    Output
      # A tibble: 5 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
      5 tf_5        m             tf_6         p               no_action
    Code
      dm_select(dm_for_filter(), tf_6, -n) %>% dm_get_all_uks()
    Output
      # A tibble: 7 x 3
        table uk_col kind       
        <chr> <keys> <chr>      
      1 tf_1  a      PK         
      2 tf_2  c      PK         
      3 tf_3  f, f1  PK         
      4 tf_4  h      PK         
      5 tf_5  k      PK         
      6 tf_6  o      PK         
      7 tf_3  g      explicit UK
    Code
      dm_select(dm_for_filter(), tf_6, -n) %>% dm_get_all_fks()
    Output
      # A tibble: 4 x 5
        child_table child_fk_cols parent_table parent_key_cols on_delete
        <chr>       <keys>        <chr>        <keys>          <chr>    
      1 tf_2        d             tf_1         a               no_action
      2 tf_2        e, e1         tf_3         f, f1           no_action
      3 tf_4        j, j1         tf_3         f, f1           no_action
      4 tf_5        l             tf_4         h               cascade  
    Code
      nyc_1_uk %>% dm_get_all_uks(flights)
    Output
      # A tibble: 1 x 3
        table   uk_col                      kind       
        <chr>   <keys>                      <chr>      
      1 flights year, month, ... (19 total) explicit UK
    Code
      nyc_1_uk %>% dm_get_all_uks("airports")
    Output
      # A tibble: 1 x 3
        table    uk_col kind 
        <chr>    <keys> <chr>
      1 airports faa    PK   
    Code
      nyc_1_uk %>% dm_get_all_uks(c(airports, weather, flights, airlines))
    Output
      # A tibble: 3 x 3
        table    uk_col                      kind       
        <chr>    <keys>                      <chr>      
      1 airports faa                         PK         
      2 airlines carrier                     PK         
      3 flights  year, month, ... (19 total) explicit UK
    Code
      nyc_1_uk %>% dm_get_all_uks(starts_with("a"))
    Output
      # A tibble: 2 x 3
        table    uk_col  kind 
        <chr>    <keys>  <chr>
      1 airlines carrier PK   
      2 airports faa     PK   
    Code
      nyc_1_uk %>% dm_get_all_uks(everything())
    Output
      # A tibble: 4 x 3
        table    uk_col                      kind       
        <chr>    <keys>                      <chr>      
      1 airlines carrier                     PK         
      2 airports faa                         PK         
      3 planes   tailnum                     PK         
      4 flights  year, month, ... (19 total) explicit UK

---

    (`year`, `manufacturer`, `model`) not a unique key of `planes`.

---

    A PK (`carrier`) for table airlines already exists, not adding UK.

---

    A UK (`year`, `month`, `day`, `dep_time`, `sched_dep_time`, ... (19 total)) for table flights already exists, not adding UK.

# dm_get_all_uks() with table arg fails nicely

    object 'nyc_1_uk' not found

---

    object 'nyc_1_uk' not found

# `dm_flatten_to_tbl()` does the right things for 'inner_join()'

    Code
      out
    Output
         fact     dim_1_key_1 dim_1_key_2 dim_2_key dim_3_key dim_4_key something.fact
         <chr>          <int> <chr>       <chr>     <chr>         <int>          <int>
       1 acorn             14 N           c         X                 7              1
       2 blubber           13 M           d         W                 8              2
       3 cindere~          12 L           e         V                 9              3
       4 depth             11 K           f         U                10              4
       5 elysium           10 J           g         T                11              5
       6 fantasy            9 I           h         S                12              6
       7 gorgeous           8 H           i         R                13              7
       8 halo               7 G           j         Q                14              8
       9 ill-adv~           6 F           k         P                15              9
      10 jitter             5 E           l         O                16             10
      # i 4 more variables: something.dim_1 <chr>, something.dim_2 <chr>,
      #   something.dim_3 <int>, something.dim_4 <int>

# output

    Code
      dm_for_filter() %>% dm_select_tbl(tf_7)
    Condition
      Error in `eval_select_indices()`:
      ! Can't subset tables that don't exist.
      x Table `tf_7` doesn't exist.
    Code
      dm_for_filter() %>% dm_rename_tbl(tf_0 = tf_7)
    Condition
      Error in `eval_rename_indices()`:
      ! Can't rename tables that don't exist.
      x Table `tf_7` doesn't exist.

# `dm_flatten_to_tbl()` does the right things for 'left_join()'

    Code
      prepare_dm_for_flatten(dm_for_flatten(), tables = c("fact", "dim_1", "dim_2",
        "dim_3", "dim_4"), gotta_rename = TRUE) %>% dm_get_tables()
    Message
      Renaming ambiguous columns: %>%
        dm_rename(fact, something.fact = something) %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_2, something.dim_2 = something) %>%
        dm_rename(dim_3, something.dim_3 = something) %>%
        dm_rename(dim_4, something.dim_4 = something)
    Output
      $fact
      # A tibble: 10 x 7
         fact     dim_1_key_1 dim_1_key_2 dim_2_key dim_3_key dim_4_key something.fact
         <chr>          <int> <chr>       <chr>     <chr>         <int>          <int>
       1 acorn             14 N           c         X                 7              1
       2 blubber           13 M           d         W                 8              2
       3 cindere~          12 L           e         V                 9              3
       4 depth             11 K           f         U                10              4
       5 elysium           10 J           g         T                11              5
       6 fantasy            9 I           h         S                12              6
       7 gorgeous           8 H           i         R                13              7
       8 halo               7 G           j         Q                14              8
       9 ill-adv~           6 F           k         P                15              9
      10 jitter             5 E           l         O                16             10
      
      $dim_1
      # A tibble: 20 x 3
         dim_1_pk_1 dim_1_pk_2 something.dim_1
              <int> <chr>      <chr>          
       1          1 A          c              
       2          2 B          d              
       3          3 C          e              
       4          4 D          f              
       5          5 E          g              
       6          6 F          h              
       7          7 G          i              
       8          8 H          j              
       9          9 I          k              
      10         10 J          l              
      11         11 K          m              
      12         12 L          n              
      13         13 M          o              
      14         14 N          p              
      15         15 O          q              
      16         16 P          r              
      17         17 Q          s              
      18         18 R          t              
      19         19 S          u              
      20         20 T          v              
      
      $dim_2
      # A tibble: 20 x 2
         dim_2_pk something.dim_2
         <chr>    <chr>          
       1 a        E              
       2 b        F              
       3 c        G              
       4 d        H              
       5 e        I              
       6 f        J              
       7 g        K              
       8 h        L              
       9 i        M              
      10 j        N              
      11 k        O              
      12 l        P              
      13 m        Q              
      14 n        R              
      15 o        S              
      16 p        T              
      17 q        U              
      18 r        V              
      19 s        W              
      20 t        X              
      
      $dim_3
      # A tibble: 20 x 2
         dim_3_pk something.dim_3
         <chr>              <int>
       1 E                      3
       2 F                      4
       3 G                      5
       4 H                      6
       5 I                      7
       6 J                      8
       7 K                      9
       8 L                     10
       9 M                     11
      10 N                     12
      11 O                     13
      12 P                     14
      13 Q                     15
      14 R                     16
      15 S                     17
      16 T                     18
      17 U                     19
      18 V                     20
      19 W                     21
      20 X                     22
      
      $dim_4
      # A tibble: 13 x 2
         dim_4_pk something.dim_4
            <int>           <int>
       1       19              19
       2       18              20
       3       17              21
       4       16              22
       5       15              23
       6       14              24
       7       13              25
       8       12              26
       9       11              27
      10       10              28
      11        9              29
      12        8              30
      13        7              31
      
    Code
      dm_flatten_to_tbl(dm_for_flatten(), fact)
    Message
      Renaming ambiguous columns: %>%
        dm_rename(fact, something.fact = something) %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_2, something.dim_2 = something) %>%
        dm_rename(dim_3, something.dim_3 = something) %>%
        dm_rename(dim_4, something.dim_4 = something)
    Output
      # A tibble: 10 x 11
         fact     dim_1_key_1 dim_1_key_2 dim_2_key dim_3_key dim_4_key something.fact
         <chr>          <int> <chr>       <chr>     <chr>         <int>          <int>
       1 acorn             14 N           c         X                 7              1
       2 blubber           13 M           d         W                 8              2
       3 cindere~          12 L           e         V                 9              3
       4 depth             11 K           f         U                10              4
       5 elysium           10 J           g         T                11              5
       6 fantasy            9 I           h         S                12              6
       7 gorgeous           8 H           i         R                13              7
       8 halo               7 G           j         Q                14              8
       9 ill-adv~           6 F           k         P                15              9
      10 jitter             5 E           l         O                16             10
      # i 4 more variables: something.dim_1 <chr>, something.dim_2 <chr>,
      #   something.dim_3 <int>, something.dim_4 <int>
    Code
      result_from_flatten_new()
    Output
      # A tibble: 10 x 11
         fact     dim_1_key_1 dim_1_key_2 dim_2_key dim_3_key dim_4_key something.fact
         <chr>          <int> <chr>       <chr>     <chr>         <int>          <int>
       1 acorn             14 N           c         X                 7              1
       2 blubber           13 M           d         W                 8              2
       3 cindere~          12 L           e         V                 9              3
       4 depth             11 K           f         U                10              4
       5 elysium           10 J           g         T                11              5
       6 fantasy            9 I           h         S                12              6
       7 gorgeous           8 H           i         R                13              7
       8 halo               7 G           j         Q                14              8
       9 ill-adv~           6 F           k         P                15              9
      10 jitter             5 E           l         O                16             10
      # i 4 more variables: something.dim_1 <chr>, something.dim_2 <chr>,
      #   something.dim_3 <int>, something.dim_4 <int>

# `dm_flatten_to_tbl()` does the right things for 'inner_join()'

    Code
      out
    Output
      # A tibble: 10 x 11
         fact     dim_1_key_1 dim_1_key_2 dim_2_key dim_3_key dim_4_key something.fact
         <chr>          <int> <chr>       <chr>     <chr>         <int>          <int>
       1 acorn             14 N           c         X                 7              1
       2 blubber           13 M           d         W                 8              2
       3 cindere~          12 L           e         V                 9              3
       4 depth             11 K           f         U                10              4
       5 elysium           10 J           g         T                11              5
       6 fantasy            9 I           h         S                12              6
       7 gorgeous           8 H           i         R                13              7
       8 halo               7 G           j         Q                14              8
       9 ill-adv~           6 F           k         P                15              9
      10 jitter             5 E           l         O                16             10
      # i 4 more variables: something.dim_1 <chr>, something.dim_2 <chr>,
      #   something.dim_3 <int>, something.dim_4 <int>

# dummy

    Code
      # dummy

# dm_sql()

    Code
      dm_for_filter_df() %>% dm_sql(my_test_con())
    Condition
      Warning:
      Autoincrementing columns not yet supported for DuckDB, these won't be set in the remote database but are preserved in the `dm`
    Message
      `on_delete = "cascade"` not supported for duckdb
    Output
      $pre
      $pre$tf_1
      <SQL> CREATE TEMPORARY TABLE tf_1 (
        a INTEGER,
        b STRING,
        PRIMARY KEY (a),
        UNIQUE (a)
      )
      
      $pre$tf_3
      <SQL> CREATE TEMPORARY TABLE tf_3 (
        f STRING,
        f1 INTEGER,
        g STRING,
        PRIMARY KEY (f, f1),
        UNIQUE (f, f1),
        UNIQUE (g)
      )
      
      $pre$tf_6
      <SQL> CREATE TEMPORARY TABLE tf_6 (
        zz INTEGER,
        n STRING,
        o STRING,
        PRIMARY KEY (o),
        UNIQUE (o),
        UNIQUE (n)
      )
      
      $pre$tf_2
      <SQL> CREATE TEMPORARY TABLE tf_2 (
        c STRING,
        d INTEGER,
        e STRING,
        e1 INTEGER,
        PRIMARY KEY (c),
        UNIQUE (c),
        FOREIGN KEY (d) REFERENCES tf_1 (a),
        FOREIGN KEY (e, e1) REFERENCES tf_3 (f, f1)
      )
      
      $pre$tf_4
      <SQL> CREATE TEMPORARY TABLE tf_4 (
        h STRING,
        i STRING,
        j STRING,
        j1 INTEGER,
        PRIMARY KEY (h),
        UNIQUE (h),
        FOREIGN KEY (j, j1) REFERENCES tf_3 (f, f1)
      )
      
      $pre$tf_5
      <SQL> CREATE TEMPORARY TABLE tf_5 (
        ww INTEGER,
        k INTEGER,
        l STRING,
        m STRING,
        PRIMARY KEY (k),
        UNIQUE (k),
        FOREIGN KEY (m) REFERENCES tf_6 (n),
        FOREIGN KEY (l) REFERENCES tf_4 (h)
      )
      
      
      $load
      $load$tf_1
      <SQL> INSERT INTO tf_1 (a, b)
      SELECT CAST(a AS INTEGER) AS a, CAST(b AS TEXT) AS b
      FROM (
        SELECT NULL AS a, NULL AS b
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          (1, 'A'),
          (2, 'B'),
          (3, 'C'),
          (4, 'D'),
          (5, 'E'),
          (6, 'F'),
          (7, 'G'),
          (8, 'H'),
          (9, 'I'),
          (10, 'J')
      ) values_table
      
      $load$tf_3
      <SQL> INSERT INTO tf_3 (f, f1, g)
      SELECT CAST(f AS TEXT) AS f, CAST(f1 AS INTEGER) AS f1, CAST(g AS TEXT) AS g
      FROM (
        SELECT NULL AS f, NULL AS f1, NULL AS g
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          ('C', 2, 'one'),
          ('C', 3, 'two'),
          ('D', 4, 'three'),
          ('E', 5, 'four'),
          ('F', 6, 'five'),
          ('G', 7, 'six'),
          ('H', 7, 'seven'),
          ('I', 7, 'eight'),
          ('J', 10, 'nine'),
          ('K', 11, 'ten')
      ) values_table
      
      $load$tf_6
      <SQL> INSERT INTO tf_6 (zz, n, o)
      SELECT CAST(zz AS INTEGER) AS zz, CAST(n AS TEXT) AS n, CAST(o AS TEXT) AS o
      FROM (
        SELECT NULL AS zz, NULL AS n, NULL AS o
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          (1, 'garden', 'i'),
          (1, 'hill', 'g'),
          (1, 'house', 'e'),
          (1, 'streetlamp', 'h'),
          (1, 'tree', 'f')
      ) values_table
      
      $load$tf_2
      <SQL> INSERT INTO tf_2 (c, d, e, e1)
      SELECT
        CAST(c AS TEXT) AS c,
        CAST(d AS INTEGER) AS d,
        CAST(e AS TEXT) AS e,
        CAST(e1 AS INTEGER) AS e1
      FROM (
        SELECT NULL AS c, NULL AS d, NULL AS e, NULL AS e1
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          ('cat', 7, 'F', 6),
          ('dog', 6, 'E', 5),
          ('elephant', 2, 'D', 4),
          ('lion', 3, 'E', 5),
          ('seal', 4, 'F', 6),
          ('worm', 5, 'G', 7)
      ) values_table
      
      $load$tf_4
      <SQL> INSERT INTO tf_4 (h, i, j, j1)
      SELECT
        CAST(h AS TEXT) AS h,
        CAST(i AS TEXT) AS i,
        CAST(j AS TEXT) AS j,
        CAST(j1 AS INTEGER) AS j1
      FROM (
        SELECT NULL AS h, NULL AS i, NULL AS j, NULL AS j1
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          ('a', 'three', 'C', 3),
          ('b', 'four', 'D', 4),
          ('c', 'five', 'E', 5),
          ('d', 'six', 'F', 6),
          ('e', 'seven', 'F', 6)
      ) values_table
      
      $load$tf_5
      <SQL> INSERT INTO tf_5 (ww, k, l, m)
      SELECT
        CAST(ww AS INTEGER) AS ww,
        CAST(k AS INTEGER) AS k,
        CAST(l AS TEXT) AS l,
        CAST(m AS TEXT) AS m
      FROM (
        SELECT NULL AS ww, NULL AS k, NULL AS l, NULL AS m
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES
          (2, 1, 'b', 'house'),
          (2, 2, 'c', 'tree'),
          (2, 3, 'd', 'streetlamp'),
          (2, 4, 'e', 'streetlamp')
      ) values_table
      
      
      $post
      $post$uk
      named list()
      
      $post$fk
      named list()
      
      $post$indexes
      $post$indexes$tf_2
      <SQL> CREATE INDEX tf_2__d ON tf_2 (d)
      <SQL> CREATE INDEX tf_2__e_e1 ON tf_2 (e, e1)
      
      $post$indexes$tf_4
      <SQL> CREATE INDEX tf_4__j_j1 ON tf_4 (j, j1)
      
      $post$indexes$tf_5
      <SQL> CREATE INDEX tf_5__m ON tf_5 (m)
      <SQL> CREATE INDEX tf_5__l ON tf_5 (l)
      
      
      

---

    Code
      dm(x = data.frame(a = strrep("x", 300))) %>% dm_sql(my_test_con())
    Output
      $pre
      $pre$x
      <SQL> CREATE TEMPORARY TABLE x (
        a STRING
      )
      
      
      $load
      $load$x
      <SQL> INSERT INTO x (a)
      SELECT CAST(a AS TEXT) AS a
      FROM (
        SELECT NULL AS a
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES ('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx')
      ) values_table
      
      
      $post
      $post$uk
      named list()
      
      $post$fk
      named list()
      
      $post$indexes
      named list()
      
      

---

    Code
      dm(x = data.frame(a = strrep("x", 10000))) %>% dm_sql(my_test_con())
    Output
      $pre
      $pre$x
      <SQL> CREATE TEMPORARY TABLE x (
        a STRING
      )
      
      
      $load
      $load$x
      <SQL> INSERT INTO x (a)
      SELECT CAST(a AS TEXT) AS a
      FROM (
        SELECT NULL AS a
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES ('[encoded data, 10000 characters]')
      ) values_table
      
      
      $post
      $post$uk
      named list()
      
      $post$fk
      named list()
      
      $post$indexes
      named list()
      
      

# `json_nest()` works remotely

    Code
      query <- remote %>% json_nest(a = starts_with("a")) %>% arrange(grp) %>%
        dbplyr::sql_render()
      gsub("test_frame_[_0-9]+", "test_frame_...", query)
    Output
      <SQL> SELECT "grp", JSON_AGG(JSON_BUILD_OBJECT('a_i', "a_i", 'a_j', "a_j")) AS "a"
      FROM "test_frame_..."
      GROUP BY "grp"
      ORDER BY "grp"
    Code
      remote %>% json_nest(a = starts_with("a")) %>% arrange(grp) %>% collect()
    Output
      # A tibble: 2 x 2
          grp a                                                       
        <dbl> <pq_json>                                               
      1     1 [{"a_i" : "a", "a_j" : "A"}, {"a_i" : "b", "a_j" : "B"}]
      2     2 [{"a_i" : "c", "a_j" : "C"}, {"a_i" : "d", "a_j" : "D"}]
    Code
      query <- remote %>% json_nest(a = starts_with("a"), .names_sep = "_") %>%
        arrange(grp) %>% dbplyr::sql_render()
      gsub("test_frame_[_0-9]+", "test_frame_...", query)
    Output
      <SQL> SELECT "grp", JSON_AGG(JSON_BUILD_OBJECT('i', "a_i", 'j', "a_j")) AS "a"
      FROM "test_frame_..."
      GROUP BY "grp"
      ORDER BY "grp"
    Code
      remote %>% json_nest(a = starts_with("a"), .names_sep = "_") %>% arrange(grp) %>%
        collect()
    Output
      # A tibble: 2 x 2
          grp a                                               
        <dbl> <pq_json>                                       
      1     1 [{"i" : "a", "j" : "A"}, {"i" : "b", "j" : "B"}]
      2     2 [{"i" : "c", "j" : "C"}, {"i" : "d", "j" : "D"}]

# dm_disambiguate_cols() works as intended

    Code
      dm_for_flatten() %>% dm_disambiguate_cols() %>% dm_paste(options = c("select",
        "keys"))
    Message
      Renaming ambiguous columns: %>%
        dm_rename(fact, something.fact = something) %>%
        dm_rename(dim_1, something.dim_1 = something) %>%
        dm_rename(dim_2, something.dim_2 = something) %>%
        dm_rename(dim_3, something.dim_3 = something) %>%
        dm_rename(dim_4, something.dim_4 = something)
      dm::dm(
        fact,
        dim_1,
        dim_2,
        dim_3,
        dim_4,
      ) %>%
        dm::dm_select(fact, fact, dim_1_key_1, dim_1_key_2, dim_2_key, dim_3_key, dim_4_key, something.fact) %>%
        dm::dm_select(dim_1, dim_1_pk_1, dim_1_pk_2, something.dim_1) %>%
        dm::dm_select(dim_2, dim_2_pk, something.dim_2) %>%
        dm::dm_select(dim_3, dim_3_pk, something.dim_3) %>%
        dm::dm_select(dim_4, dim_4_pk, something.dim_4) %>%
        dm::dm_add_pk(dim_1, c(dim_1_pk_1, dim_1_pk_2)) %>%
        dm::dm_add_pk(dim_2, dim_2_pk) %>%
        dm::dm_add_pk(dim_3, dim_3_pk) %>%
        dm::dm_add_pk(dim_4, dim_4_pk) %>%
        dm::dm_add_fk(fact, c(dim_1_key_1, dim_1_key_2), dim_1) %>%
        dm::dm_add_fk(fact, dim_2_key, dim_2) %>%
        dm::dm_add_fk(fact, dim_3_key, dim_3) %>%
        dm::dm_add_fk(fact, dim_4_key, dim_4)

# `check_suggested()` works

    Code
      check_suggested("rlang", "foo")
    Output
      [1] TRUE
    Code
      check_suggested("dm", "foo", use = NA)
    Output
      [1] FALSE
    Code
      check_suggested("iurtnkjvmomweicopbt", "foo", use = NA)
    Output
      [1] FALSE
    Code
      check_suggested("iurtnkjvmomweicopbt", use = NA)
    Output
      [1] FALSE
    Code
      check_suggested("dm", "foo", use = FALSE)
    Output
      [1] FALSE
    Code
      check_suggested("iurtnkjvmomweicopbt", "foo", use = FALSE)
    Output
      [1] FALSE
    Code
      check_suggested("iurtnkjvmomweicopbt", use = FALSE)
    Output
      [1] FALSE

# `check_suggested()` works for error messages

    Code
      check_suggested("iurtnkjvmomweicopbt", "foo")
    Condition
      Error:
      ! The package "iurtnkjvmomweicopbt" is required to use `foo()`.

---

    Code
      check_suggested(c("iurtnkjvmomweicopbt (>= 0.5)", "xxx", "cli"), "foo", use = NA)
    Message
      `foo()` is improved by the "iurtnkjvmomweicopbt (>= 0.5)" and "xxx" packages. Consider installing them.
    Output
      [1] FALSE

# output

    Code
      commas(character())
    Output
      
    Code
      commas(letters[1])
    Output
      a
    Code
      commas(letters[1:2])
    Output
      a, b
    Code
      commas(letters[1:3])
    Output
      a, b, c
    Code
      commas(letters[seq_len(MAX_COMMAS - 1)])
    Output
      a, b, c, d, e
    Code
      commas(letters[seq_len(MAX_COMMAS)])
    Output
      a, b, c, d, e, f
    Code
      commas(letters[seq_len(MAX_COMMAS + 1)])
    Output
      a, b, c, d, e, ... (7 total)
    Code
      commas(letters[1:4], max_commas = 3)
    Output
      a, b, ... (4 total)
    Code
      commas(letters, capped = TRUE)
    Output
      a, b, c, d, e, ...
    Code
      commas(letters, fun = tick)
    Output
      `a`, `b`, `c`, `d`, `e`, ... (26 total)

# functions working with graphs do the right thing?

    Code
      attr(graph_edges(create_graph_from_dm(nyc_comp())), "vnames")
    Output
      [1] "airlines|flights" "airports|flights" "flights|planes"   "flights|weather" 

# empty graph

    Code
      print(g0)
    Output
      <dm_graph> undirected, 0 vertices, 0 edges
    Code
      print(g1)
    Output
      <dm_graph> undirected, 1 vertex, 0 edges
      Vertices: x 

# graph_from_data_frame: print output is consistent

    Code
      print(g_directed)
    Output
      <dm_graph> directed, 3 vertices, 2 edges
      Vertices: a, b, c 
      Edges: a|b, b|c 
    Code
      print(g_undirected)
    Output
      <dm_graph> undirected, 3 vertices, 2 edges
      Vertices: a, b, c 
      Edges: a|b, b|c 

# graph_from_data_frame: zero-row data frame builds graph with no edges

    Code
      print(g)
    Output
      <dm_graph> undirected, 2 vertices, 0 edges
      Vertices: x, y 

# dm_filter() deprecations

    Code
      dm_filter(dm_for_filter(), tf_1, a > 4)
    Condition
      Warning:
      The `table` argument of `dm_filter()` is deprecated as of dm 1.0.0.
      i `dm_filter()` now takes named filter expressions, the names correspond to the tables to be filtered. You no longer need to call `dm_apply_filters()` to materialize the filters.
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_1`, `tf_2`, `tf_3`, `tf_4`, `tf_5`, `tf_6`
      Columns: 20
      Primary keys: 6
      Foreign keys: 5
      -- Filters ---------------------------------------------------------------------
      tf_1: a > 4
    Code
      dm_filter(dm = dm_for_filter(), tf_1, a > 4)
    Condition
      Warning:
      The `dm` argument of `dm_filter()` is deprecated as of dm 1.0.0.
      i Please use the `.dm` argument instead.
      Warning:
      The `table` argument of `dm_filter()` is deprecated as of dm 1.0.0.
      i `dm_filter()` now takes named filter expressions, the names correspond to the tables to be filtered. You no longer need to call `dm_apply_filters()` to materialize the filters.
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_1`, `tf_2`, `tf_3`, `tf_4`, `tf_5`, `tf_6`
      Columns: 20
      Primary keys: 6
      Foreign keys: 5
      -- Filters ---------------------------------------------------------------------
      tf_1: a > 4
    Code
      dm_filter(dm_for_filter(), tf_1, a > 4) %>% dm_apply_filters()
    Condition
      Warning:
      The `table` argument of `dm_filter()` is deprecated as of dm 1.0.0.
      i `dm_filter()` now takes named filter expressions, the names correspond to the tables to be filtered. You no longer need to call `dm_apply_filters()` to materialize the filters.
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_1`, `tf_2`, `tf_3`, `tf_4`, `tf_5`, `tf_6`
      Columns: 20
      Primary keys: 6
      Foreign keys: 5
    Code
      dm_filter(dm_for_filter(), tf_1 = a > 4) %>% dm_apply_filters()
    Condition
      Warning:
      `dm_apply_filters()` was deprecated in dm 1.0.0.
      i Calling `dm_apply_filters()` after `dm_filter()` is no longer necessary.
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `tf_1`, `tf_2`, `tf_3`, `tf_4`, `tf_5`, `tf_6`
      Columns: 20
      Primary keys: 6
      Foreign keys: 5
    Code
      dm_filter(dm_for_filter(), tf_1, a > 4) %>% dm_apply_filters_to_tbl(tf_2)
    Condition
      Warning:
      The `table` argument of `dm_filter()` is deprecated as of dm 1.0.0.
      i `dm_filter()` now takes named filter expressions, the names correspond to the tables to be filtered. You no longer need to call `dm_apply_filters()` to materialize the filters.
    Output
      # A tibble: 3 x 4
        c         d e        e1
        <chr> <int> <chr> <int>
      1 worm      5 G         7
      2 dog       6 E         5
      3 cat       7 F         6
    Code
      dm_filter(dm_for_filter(), tf_1 = a > 4) %>% dm_apply_filters_to_tbl(tf_2)
    Condition
      Warning:
      `dm_apply_filters_to_tbl()` was deprecated in dm 1.0.0.
      i Access tables directly after `dm_filter()`.
    Output
      # A tibble: 3 x 4
        c         d e        e1
        <chr> <int> <chr> <int>
      1 worm      5 G         7
      2 dog       6 E         5
      3 cat       7 F         6
    Code
      dm_filter(dm_for_filter(), tf_1, a > 4) %>% dm_get_filters()
    Condition
      Warning:
      The `table` argument of `dm_filter()` is deprecated as of dm 1.0.0.
      i `dm_filter()` now takes named filter expressions, the names correspond to the tables to be filtered. You no longer need to call `dm_apply_filters()` to materialize the filters.
    Output
      # A tibble: 1 x 3
        table filter     zoomed
        <chr> <list>     <lgl> 
      1 tf_1  <language> FALSE 
    Code
      dm_filter(dm_for_filter(), tf_1 = a > 4) %>% dm_get_filters()
    Condition
      Warning:
      `dm_get_filters()` was deprecated in dm 1.0.0.
      i Filter conditions are no longer stored with the dm object.
    Output
      # A tibble: 0 x 3
      # i 3 variables: table <chr>, filter <list>, zoomed <lgl>

# data structure

    Code
      dm_more_complex() %>% dm_paste(options = c("select", "keys"))
    Message
      dm::dm(
        tf_1,
        tf_2,
        tf_3,
        tf_4,
        tf_5,
        tf_6,
        tf_7,
        tf_6_2,
        tf_4_2,
        a,
        b,
        c,
        d,
        e,
      ) %>%
        dm::dm_select(tf_1, a, b) %>%
        dm::dm_select(tf_2, c, d, e, e1) %>%
        dm::dm_select(tf_3, f, f1, g) %>%
        dm::dm_select(tf_4, h, i, j, j1) %>%
        dm::dm_select(tf_5, ww, k, l, m) %>%
        dm::dm_select(tf_6, zz, n, o) %>%
        dm::dm_select(tf_7, p, q) %>%
        dm::dm_select(tf_6_2, p, f, f1) %>%
        dm::dm_select(tf_4_2, r, s, t) %>%
        dm::dm_select(a, a_1, a_2) %>%
        dm::dm_select(b, b_1, b_2, b_3) %>%
        dm::dm_select(c, c_1) %>%
        dm::dm_select(d, d_1, b_1) %>%
        dm::dm_select(e, e_1, b_1) %>%
        dm::dm_add_pk(tf_1, a) %>%
        dm::dm_add_pk(tf_2, c) %>%
        dm::dm_add_pk(tf_3, c(f, f1)) %>%
        dm::dm_add_pk(tf_4, h) %>%
        dm::dm_add_pk(tf_5, k) %>%
        dm::dm_add_pk(tf_6, n) %>%
        dm::dm_add_pk(tf_6_2, p) %>%
        dm::dm_add_pk(tf_4_2, r) %>%
        dm::dm_add_pk(a, a_1) %>%
        dm::dm_add_pk(b, b_1) %>%
        dm::dm_add_pk(c, c_1) %>%
        dm::dm_add_pk(d, d_1) %>%
        dm::dm_add_pk(e, e_1) %>%
        dm::dm_add_fk(tf_2, d, tf_1) %>%
        dm::dm_add_fk(tf_2, c(e, e1), tf_3) %>%
        dm::dm_add_fk(tf_4, c(j, j1), tf_3) %>%
        dm::dm_add_fk(tf_6_2, c(f, f1), tf_3) %>%
        dm::dm_add_fk(tf_5, l, tf_4) %>%
        dm::dm_add_fk(tf_5, m, tf_6) %>%
        dm::dm_add_fk(tf_5, l, tf_4_2) %>%
        dm::dm_add_fk(b, b_2, a) %>%
        dm::dm_add_fk(d, b_1, b) %>%
        dm::dm_add_fk(e, b_1, b) %>%
        dm::dm_add_fk(b, b_3, c)

# we get filtered/unfiltered tables with respective funs

    Code
      dm_for_filter() %>% dm_filter(tf_1 = a > 3 & a < 8) %>% dm_get_tables() %>% map(
        harmonize_tbl)
    Output
      $tf_1
      # A tibble: 4 x 2
            a b    
        <int> <chr>
      1     4 D    
      2     5 E    
      3     6 F    
      4     7 G    
      
      $tf_2
      # A tibble: 4 x 4
        c         d e        e1
        <chr> <int> <chr> <int>
      1 cat       7 F         6
      2 dog       6 E         5
      3 seal      4 F         6
      4 worm      5 G         7
      
      $tf_3
      # A tibble: 3 x 3
        f        f1 g    
        <chr> <int> <chr>
      1 E         5 four 
      2 F         6 five 
      3 G         7 six  
      
      $tf_4
      # A tibble: 3 x 4
        h     i     j        j1
        <chr> <chr> <chr> <int>
      1 c     five  E         5
      2 d     six   F         6
      3 e     seven F         6
      
      $tf_5
      # A tibble: 3 x 4
           ww     k l     m         
        <int> <int> <chr> <chr>     
      1     2     2 c     tree      
      2     2     3 d     streetlamp
      3     2     4 e     streetlamp
      
      $tf_6
      # A tibble: 2 x 3
           zz n          o    
        <int> <chr>      <chr>
      1     1 streetlamp h    
      2     1 tree       f    
      

# dm_filter() works as intended for reversed dm

    Code
      dm_for_filter_rev() %>% dm_filter(tf_1 = a < 8 & a > 3) %>% dm_get_tables() %>%
        map(harmonize_tbl)
    Output
      $tf_6
      # A tibble: 2 x 3
           zz n          o    
        <int> <chr>      <chr>
      1     1 streetlamp h    
      2     1 tree       f    
      
      $tf_5
      # A tibble: 3 x 4
           ww     k l     m         
        <int> <int> <chr> <chr>     
      1     2     2 c     tree      
      2     2     3 d     streetlamp
      3     2     4 e     streetlamp
      
      $tf_4
      # A tibble: 3 x 4
        h     i     j        j1
        <chr> <chr> <chr> <int>
      1 c     five  E         5
      2 d     six   F         6
      3 e     seven F         6
      
      $tf_3
      # A tibble: 3 x 3
        f        f1 g    
        <chr> <int> <chr>
      1 E         5 four 
      2 F         6 five 
      3 G         7 six  
      
      $tf_2
      # A tibble: 4 x 4
        c         d e        e1
        <chr> <int> <chr> <int>
      1 cat       7 F         6
      2 dog       6 E         5
      3 seal      4 F         6
      4 worm      5 G         7
      
      $tf_1
      # A tibble: 4 x 2
            a b    
        <int> <chr>
      1     4 D    
      2     5 E    
      3     6 F    
      4     7 G    
      

# dm_filter() works as intended for inbetween table

    Code
      dm_for_filter() %>% dm_filter(tf_3 = g == "five") %>% dm_get_tables() %>% map(
        harmonize_tbl)
    Output
      $tf_1
      # A tibble: 2 x 2
            a b    
        <int> <chr>
      1     4 D    
      2     7 G    
      
      $tf_2
      # A tibble: 2 x 4
        c         d e        e1
        <chr> <int> <chr> <int>
      1 cat       7 F         6
      2 seal      4 F         6
      
      $tf_3
      # A tibble: 1 x 3
        f        f1 g    
        <chr> <int> <chr>
      1 F         6 five 
      
      $tf_4
      # A tibble: 2 x 4
        h     i     j        j1
        <chr> <chr> <chr> <int>
      1 d     six   F         6
      2 e     seven F         6
      
      $tf_5
      # A tibble: 2 x 4
           ww     k l     m         
        <int> <int> <chr> <chr>     
      1     2     3 d     streetlamp
      2     2     4 e     streetlamp
      
      $tf_6
      # A tibble: 1 x 3
           zz n          o    
        <int> <chr>      <chr>
      1     1 streetlamp h    
      

# dm_filter() output for compound keys

    Code
      nyc_comp() %>% dm_filter(flights = sched_dep_time <= 1200) %>% dm_nrow()
    Output
      airlines airports  flights   planes  weather 
            14       63      672      502       47 
    Code
      nyc_comp() %>% dm_filter(weather = pressure < 1020) %>% dm_nrow()
    Output
      airlines airports  flights   planes  weather 
             0        0        0        0        0 

# code generation works

    Code
      call_to_char(body(function(.) dm_add_tbl(., weather)))
    Output
      [1] "dm_add_tbl(weather)"
    Code
      call_to_char(expr(dm_add_tbl(., weather, airports, flights, airlines, planes,
        mtcars, penguins)))
    Output
      [1] "dm_add_tbl(weather, airports, flights, airlines, planes, mtcars, penguins)"

---

    Code
      new_cg_block()
      new_cg_block(quo(dm_nycflights13()), list(function(.) dm_add_pk(., flights,
        flight_id)))
    Output
      dm_nycflights13() %>%
        dm_add_pk(flights, flight_id)

---

    Code
      table <- "flights"
      columns <- "carrier"
      cg_block <- new_cg_block(quo(dm_nycflights13())) %>% cg_add_call(dm_rm_fk(.,
        table = !!sym(table), columns = !!sym(columns), ref_table = airlines)) %>%
        cg_add_call(dm_rm_fk(., table = flights, columns = c(origin, time_hour),
        ref_table = weather)) %>% cg_add_call(dm_add_fk(., table = !!sym(table),
      columns = !!sym(columns), ref_table = airlines))
      cg_block
    Output
      dm_nycflights13() %>%
        dm_rm_fk(table = flights, columns = carrier, ref_table = airlines) %>%
        dm_rm_fk(table = flights, columns = c(origin, time_hour), ref_table = weather) %>%
        dm_add_fk(table = flights, columns = carrier, ref_table = airlines)
    Code
      cg_eval_block(cg_block)
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `planes`, `weather`
      Columns: 53
      Primary keys: 4
      Foreign keys: 3

---

    Code
      cg_block_2 <- new_cg_block(cg_block$cg_input_object, list(function(.) dm(.,
        mtcars), function(.) dm_select_tbl(., -planes)))
      cg_block_2
    Output
      dm_nycflights13() %>%
        dm(mtcars) %>%
        dm_select_tbl(-planes)
    Code
      cg_eval_block(cg_block_2)
    Output
      -- Metadata --------------------------------------------------------------------
      Tables: `airlines`, `airports`, `flights`, `weather`, `mtcars`
      Columns: 55
      Primary keys: 3
      Foreign keys: 3

---

    Code
      format(new_cg_block(quo(dm_nycflights13()), list(function(.) dm_add_pk(.,
        flights, flight_id))))
    Output
      [1] "dm_nycflights13() %>%"           "  dm_add_pk(flights, flight_id)"

# `dm_pixarfilms()` works

    Code
      dm_examine_constraints(dm_pixarfilms(consistent = FALSE))
    Message
      ! Unsatisfied constraints:
    Output
      * Table `pixar_films`: primary key `film`: has 1 missing values

---

    Code
      dm_examine_constraints(dm_pixarfilms(consistent = TRUE))
    Message
      i All constraints satisfied.

# `json_nest()` and `json_unnest()` work

    Code
      df <- tibble::tibble(x = c(1, 1, 1, 2, 2, 3), y = 1:6, z = 6:1)
      nested <- json_nest(df, data = c(y, z))
      nested
    Output
      # A tibble: 3 x 2
            x data                                                     
        <dbl> <chr>                                                    
      1     1 "[{\"y\":1,\"z\":6},{\"y\":2,\"z\":5},{\"y\":3,\"z\":4}]"
      2     2 "[{\"y\":4,\"z\":3},{\"y\":5,\"z\":2}]"                  
      3     3 "[{\"y\":6,\"z\":1}]"                                    

# `json_nest()` fails with unnamed elements

    Code
      json_nest(df, c(y, z))
    Condition
      Error in `json_nest()`:
      ! All elements of `...` must be named.

# `json_unnest()` fails without cols argument

    Code
      json_unnest(df)
    Condition
      Error in `json_unnest()`:
      ! The `cols` argument must be provided.

# check_cardinality_...() functions work without `x_select` and `y_select`

    Code
      examine_cardinality(data_card_1(), data_card_11())
    Output
      [1] "injective mapping (child: 0 or 1 -> parent: 1)"
    Code
      examine_cardinality(data_card_1(), data_card_12())
    Output
      [1] "surjective mapping (child: 1 to n -> parent: 1)"
    Code
      examine_cardinality(data_card_1(), suppress_mssql_message(compute(data_card_1())))
    Output
      [1] "bijective mapping (child: 1 -> parent: 1)"
    Code
      examine_cardinality(data_card_1(), data_card_11())
    Output
      [1] "injective mapping (child: 0 or 1 -> parent: 1)"

# check_cardinality_...() API errors

    Code
      check_card_api(data_mcard_1(), a, data_mcard_2(), b)
    Condition
      Warning:
      The `pk_column` argument of `eval()` is deprecated as of dm 1.0.0.
      i Please use the `x_select` argument instead.
      * Use `y_select` instead of `fk_column`, and `x` and `y` instead of `parent_table` and `child_table`.
      * Using `by_position = TRUE` for compatibility.
    Output
      [[1]]
      # A tibble: 3 x 1
            a
        <dbl>
      1     1
      2     2
      3     1
      
      [[2]]
      # A tibble: 3 x 1
            b
        <dbl>
      1     4
      2     5
      3     6
      
      [[3]]
      [1] "data_mcard_1()"
      
      [[4]]
      [1] "data_mcard_2()"
      

---

    Code
      check_card_api(data_mcard_1(), data_mcard_2(), x_select = a, y_select = c)
    Condition
      Error:
      ! `by_position = FALSE` or `by_position = NULL` require column names in `x` to match those in `y`.

# check_cardinality_...() functions are checking the cardinality correctly?

    Code
      expect_dm_error(check_cardinality_0_n(data_card_1(), data_card_2(), x_select = a,
      y_select = a), class = "not_subset_of")
    Output
      # A tibble: 1 x 1
            a
        <dbl>
      1     6
    Code
      expect_dm_error(check_cardinality_1_1(data_card_5(), data_card_4(), x_select = a,
      y_select = c(a = c)), class = "not_bijective")
      expect_dm_error(check_cardinality_1_1(data_card_4(), data_card_5(), x_select = c,
      y_select = c(c = a)), class = "not_unique_key")
      expect_dm_error(check_cardinality_1_1(data_card_4(), data_card_1(), x_select = c,
      y_select = c(c = a)), class = "not_unique_key")
      expect_dm_error(check_cardinality_0_1(data_card_1(), data_card_4(), x_select = a,
      y_select = c(a = c)), class = "not_injective")
      expect_dm_error(check_cardinality_0_n(data_card_4(), data_card_1(), x_select = c,
      y_select = c(c = a)), class = "not_unique_key")
      expect_dm_error(check_cardinality_1_1(data_card_4(), data_card_1(), x_select = c,
      y_select = c(c = a)), class = "not_unique_key")
      expect_dm_error(check_cardinality_1_1(data_card_1(), data_card_4(), x_select = a,
      y_select = c(a = c)), class = "not_bijective")

# check_cardinality_...() functions are supporting compound keys

    Code
      expect_dm_error(check_cardinality_0_n(data_card_1(), data_card_2(), x_select = c(
        a, b), y_select = c(a, b)), class = "not_subset_of")
    Output
      # A tibble: 4 x 2
            a b    
        <dbl> <chr>
      1     3 b    
      2     4 c    
      3     5 d    
      4     6 e    
    Code
      expect_dm_error(check_cardinality_1_1(data_card_1(), data_card_12(), x_select = c(
        a, b), y_select = c(a, b)), class = "not_bijective")
      expect_dm_error(check_cardinality_1_1(data_card_12(), data_card_1(), x_select = c(
        a, b), y_select = c(a, b)), class = "not_unique_key")
      expect_dm_error(check_cardinality_0_1(data_card_1(), data_card_12(), x_select = c(
        b, a), y_select = c(b, a)), class = "not_injective")

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm.R
\name{glimpse.dm}
\alias{glimpse.dm}
\alias{glimpse.dm_zoomed}
\title{Get a glimpse of your \code{dm} object}
\usage{
\method{glimpse}{dm}(x, width = NULL, ...)

\method{glimpse}{dm_zoomed}(x, width = NULL, ...)
}
\arguments{
\item{x}{A \code{dm} object.}

\item{width}{Controls the maximum number of columns on a line used in
printing. If \code{NULL}, \code{getOption("width")} will be consulted.}

\item{...}{Passed to \code{\link[pillar:glimpse]{pillar::glimpse()}}.}
}
\description{
\code{glimpse()} provides an overview (dimensions, column data types, primary
keys, etc.) of all tables included in the \code{dm} object. It will additionally
print details about outgoing foreign keys for the child table.

\code{glimpse()} is provided by the pillar package, and re-exported by \pkg{dm}.
See \code{\link[pillar:glimpse]{pillar::glimpse()}} for more details.
}
\examples{

dm_nycflights13() \%>\% glimpse()

dm_nycflights13() \%>\%
  dm_zoom_to(flights) \%>\%
  glimpse()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema.R
\name{db_schema_drop}
\alias{db_schema_drop}
\title{Remove a schema from a database}
\usage{
db_schema_drop(con, schema, force = FALSE, ...)
}
\arguments{
\item{con}{An object of class \code{"src"} or \code{"DBIConnection"}.}

\item{schema}{Class \code{character} or \code{SQL} (cf. Details), name of the schema}

\item{force}{Boolean, default \code{FALSE}. Set to \code{TRUE} to drop a schema and
all objects it contains at once. Currently only supported for Postgres/Redshift.}

\item{...}{Passed on to the individual methods.}
}
\value{
\code{NULL} invisibly.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{db_schema_drop()} deletes a schema from the database.
For certain DBMS it is possible to force the removal of a non-empty schema, see below.
}
\details{
Methods are not available for all DBMS.

An error is thrown if no schema of that name exists.

The argument \code{schema} (and \code{dbname} for MSSQL) can be provided as \code{SQL} objects.
Keep in mind, that in this case it is assumed that they are already correctly quoted as identifiers.

Additional arguments are:
\itemize{
\item \code{dbname}: supported for MSSQL. Remove a schema from a different
database on the connected MSSQL-server; default: database addressed by \code{con}.
}
}
\seealso{
Other schema handling functions:
\code{\link[=db_schema_create]{db_schema_create()}},
\code{\link[=db_schema_exists]{db_schema_exists()}},
\code{\link[=db_schema_list]{db_schema_list()}}
}
\concept{schema handling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table-surgery.R
\name{decompose_table}
\alias{decompose_table}
\title{Decompose a table into two linked tables}
\usage{
decompose_table(.data, new_id_column, ...)
}
\arguments{
\item{.data}{Data frame from which columns \code{...} are to be extracted.}

\item{new_id_column}{Name of the identifier column (primary key column) for the parent table.
A column of this name is also added in 'child table'.}

\item{...}{The columns to be extracted from the \code{.data}.

One or more unquoted expressions separated by commas.
You can treat variable names as if they were positions, so you
can use expressions like x:y to select ranges of variables.

The arguments in ... are automatically quoted and evaluated in a context where column names represent column positions.
They also support
unquoting and splicing.
See vignette("programming") for an introduction to those concepts.

See select helpers for more details, and the examples about tidyselect helpers, such as starts_with(), everything(), ...}
}
\value{
A named list of length two:
\itemize{
\item entry "child_table": the child table with column \code{new_id_column} referring to the same column in \code{parent_table},
\item entry "parent_table": the "lookup table" for \code{child_table}.
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Perform table surgery by extracting a 'parent table' from a table, linking the original table and the new table by a key, and returning both tables.

\code{decompose_table()} accepts a data frame, a name for the 'ID column' that will be newly created, and the names
of the columns that will be extracted into the new data frame.

It creates a 'parent table', which consists of the columns specified in the ellipsis, and a new 'ID column'.
Then it removes those
columns from the original table, which is now called the 'child table, and adds the 'ID column'.
}
\section{Life cycle}{

This function is marked "experimental" because it seems more useful
when applied to a table in a dm object.
Changing the interface later seems harmless because these functions are
most likely used interactively.
}

\examples{
decomposed_table <- decompose_table(mtcars, new_id, am, gear, carb)
decomposed_table$child_table
decomposed_table$parent_table
}
\seealso{
Other table surgery functions:
\code{\link[=reunite_parent_child]{reunite_parent_child()}}
}
\concept{table surgery functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primary-keys.R
\name{dm_get_pk}
\alias{dm_get_pk}
\title{Primary key column names}
\usage{
dm_get_pk(dm, table, ...)
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

This function is deprecated because of its limited use
and its unintuitive return value.
Use \code{\link[=dm_get_all_pks]{dm_get_all_pks()}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_from_con.R
\name{dm_from_con}
\alias{dm_from_con}
\title{Load a dm from a remote data source}
\usage{
dm_from_con(
  con = NULL,
  table_names = NULL,
  learn_keys = NULL,
  .names = NULL,
  ...
)
}
\arguments{
\item{con}{A \code{\link[DBI:DBIConnection-class]{DBI::DBIConnection}} or a \code{Pool} object.}

\item{table_names}{A character vector of the names of the tables to include.}

\item{learn_keys}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Set to \code{TRUE} to query the definition of primary and
foreign keys from the database.
Currently works for Postgres/Redshift, MariaDB/MySQL, SQLite, SQL Server, and DuckDB databases.
The default attempts to query and issues an informative message.}

\item{.names}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

A glue specification that describes how to name the tables
within the output, currently only for MSSQL, Postgres/Redshift and MySQL/MariaDB.
This can use \code{{.table}} to stand for the table name, and
\code{{.schema}} to stand for the name of the schema which the table lives
within. The default (\code{NULL}) is equivalent to \code{"{.table}"} when a single
schema is specified in \code{schema}, and \code{"{.schema}.{.table}"} for the case
where multiple schemas are given, and may change in future versions.}

\item{...}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Additional parameters for the schema learning query.
\itemize{
\item \code{schema}: supported for MSSQL (default: \code{"dbo"}), Postgres/Redshift (default: \code{"public"}), MariaDB/MySQL
(default: current database) and SQLite (default: main schema).
Learn the tables in a specific schema (or database for MariaDB/MySQL).
\item \code{dbname}: supported for MSSQL. Access different databases on the connected MSSQL-server;
default: active database.
\item \code{table_type}: supported for Postgres/Redshift (default: \code{"BASE TABLE"}). Specify the table type. Options are:
\enumerate{
\item \code{"BASE TABLE"} for a persistent table (normal table type)
\item \code{"VIEW"} for a view
\item \code{"FOREIGN TABLE"} for a foreign table
\item \code{"LOCAL TEMPORARY"} for a temporary table
}
}}
}
\value{
A \code{dm} object.
}
\description{
\code{dm_from_con()} creates a \link{dm} from some or all tables in a \link[dplyr:src]{src}
(a database or an environment) or which are accessible via a DBI-Connection.
For Postgres/Redshift and SQL Server databases, primary and foreign keys
are imported from the database.
}
\examples{
\dontshow{if (dm:::dm_has_financial()) withAutoprint(\{ # examplesIf}
con <- dm_get_con(dm_financial())

# Avoid DBI::dbDisconnect() here, because we don't own the connection
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_pack.R
\name{json_pack}
\alias{json_pack}
\title{JSON pack}
\usage{
json_pack(.data, ..., .names_sep = NULL)
}
\arguments{
\item{.data}{A data frame, a data frame extension (e.g. a tibble), or  a lazy data frame (e.g. from dbplyr or dtplyr).}

\item{...}{<\code{\link[tidyr:tidyr_tidy_select]{tidy-select}}> Columns to pack, specified
using name-variable pairs of the form \code{new_col = c(col1, col2, col3)}.
The right hand side can be any valid tidy select expression.}

\item{.names_sep}{If \code{NULL}, the default, the names will be left as is.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

A wrapper around \code{\link[tidyr:pack]{tidyr::pack()}} which stores the packed data into JSON columns.
}
\examples{
df <- tibble::tibble(x1 = 1:3, x2 = 4:6, x3 = 7:9, y = 1:3)
packed <- json_pack(df, x = c(x1, x2, x3), y = y)
packed
}
\seealso{
\code{\link[tidyr:pack]{tidyr::pack()}}, \code{\link[=json_pack_join]{json_pack_join()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_unpack.R
\name{json_unpack}
\alias{json_unpack}
\title{Unpack a JSON column}
\usage{
json_unpack(data, cols, ..., names_sep = NULL, names_repair = "check_unique")
}
\arguments{
\item{data}{A data frame, a data frame extension (e.g. a tibble), or  a lazy
data frame (e.g. from dbplyr or dtplyr).}

\item{cols}{<\code{\link[tidyr:tidyr_tidy_select]{tidy-select}}> Columns to unpack.}

\item{...}{Arguments passed to methods.}

\item{names_sep}{If \code{NULL}, the default, the names will be left
as is. In \code{pack()}, inner names will come from the former outer names;
in \code{unpack()}, the new outer names will come from the inner names.

If a string, the inner and outer names will be used together. In
\code{unpack()}, the names of the new outer columns will be formed by pasting
together the outer and the inner column names, separated by \code{names_sep}. In
\code{pack()}, the new inner names will have the outer names + \code{names_sep}
automatically stripped. This makes \code{names_sep} roughly symmetric between
packing and unpacking.}

\item{names_repair}{Used to check that output data frame has valid
names. Must be one of the following options:
\itemize{
\item \verb{"minimal}": no name repair or checks, beyond basic existence,
\item \verb{"unique}": make sure names are unique and not empty,
\item \verb{"check_unique}": (the default), no name repair, but check they are unique,
\item \verb{"universal}": make the names unique and syntactic
\item a function: apply custom name repair.
\item \link[tidyr]{tidyr_legacy}: use the name repair from tidyr 0.8.
\item a formula: a purrr-style anonymous function (see \code{\link[rlang:as_function]{rlang::as_function()}})
}

See \code{\link[vctrs:vec_as_names]{vctrs::vec_as_names()}} for more details on these terms and the
strategies used to enforce them.}
}
\value{
An object of the same type as \code{data}
}
\description{
A wrapper around \code{\link[tidyr:unpack]{tidyr::unpack()}} that extracts its data from a JSON column.
The inverse of \code{\link[=json_pack]{json_pack()}}.
}
\examples{
tibble(a = 1, b = '{ "c": 2, "d": 3 }') \%>\%
  json_unpack(b)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/examine-cardinalities.R
\name{dm_examine_cardinalities}
\alias{dm_examine_cardinalities}
\title{Learn about your data model}
\usage{
dm_examine_cardinalities(
  .dm,
  ...,
  .progress = NA,
  dm = deprecated(),
  progress = deprecated()
)
}
\arguments{
\item{.dm}{A \code{dm} object.}

\item{...}{These dots are for future extensions and must be empty.}

\item{.progress}{Whether to display a progress bar, if \code{NA} (the default)
hide in non-interactive mode, show in interactive mode. Requires the
'progress' package.}

\item{dm, progress}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}}
}
\value{
A tibble with the following columns:
\describe{
\item{\code{child_table}}{child table,}
\item{\code{child_fk_cols}}{foreign key column(s) in child table as list of character vectors,}
\item{\code{parent_table}}{parent table,}
\item{\code{parent_key_cols}}{key column(s) in parent table as list of character vectors,}
\item{\code{cardinality}}{the nature of cardinality along the foreign key.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

This function returns a tibble with information about
the cardinality of the FK constraints.
The printing for this object is special, use \code{\link[tibble:as_tibble]{as_tibble()}}
to print as a regular tibble.
}
\details{
Uses \code{\link[=examine_cardinality]{examine_cardinality()}} on each foreign key that is defined in the \code{\link{dm}}.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_examine_cardinalities()
\dontshow{\}) # examplesIf}
}
\seealso{
Other cardinality functions:
\code{\link[=examine_cardinality]{examine_cardinality()}}
}
\concept{cardinality functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unique-keys.R
\name{dm_get_all_uks}
\alias{dm_get_all_uks}
\title{Get all unique keys of a \code{\link{dm}} object}
\usage{
dm_get_all_uks(dm, table = NULL, ...)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{One or more table names, unquoted,
to return unique key information for.
The default \code{NULL} returns information for all tables.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A tibble with the following columns:
\describe{
\item{\code{table}}{table name,}
\item{\code{uk_col}}{column name(s) of primary key, as list of character vectors,}
\item{\code{kind}}{kind of unique key, see details.}
}
}
\description{
\code{dm_get_all_uks()} checks the \code{dm} object for unique keys
(primary keys, explicit and implicit unique keys) and returns the tables and
the respective unique key columns.
}
\details{
There are 3 kinds of unique keys:
\itemize{
\item \code{PK}: Primary key, set by \code{\link[=dm_add_pk]{dm_add_pk()}}
\item \verb{explicit UK}: Unique key, set by \code{\link[=dm_add_uk]{dm_add_uk()}}
\item \verb{implicit UK}: Unique key, not explicitly set, but referenced by a foreign key.
}
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_get_all_uks()
\dontshow{\}) # examplesIf}
}
\seealso{
Other primary key functions:
\code{\link[=dm_add_pk]{dm_add_pk()}},
\code{\link[=dm_add_uk]{dm_add_uk()}},
\code{\link[=dm_get_all_pks]{dm_get_all_pks()}},
\code{\link[=dm_has_pk]{dm_has_pk()}},
\code{\link[=dm_rm_pk]{dm_rm_pk()}},
\code{\link[=dm_rm_uk]{dm_rm_uk()}},
\code{\link[=enum_pk_candidates]{enum_pk_candidates()}}
}
\concept{primary key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_deconstruct.R
\name{dm_deconstruct}
\alias{dm_deconstruct}
\title{Create code to deconstruct a dm object}
\usage{
dm_deconstruct(dm, dm_name = NULL)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{dm_name}{The code to use to access the dm object,
by default the expression passed to this function.}
}
\value{
This function is called for its side effect of printing
generated code.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Emits code that assigns each table in the dm to a variable,
using \code{\link[=pull_tbl]{pull_tbl()}} with \code{keyed = TRUE}.
These tables retain information about primary and foreign keys,
even after data transformations,
and can be converted back to a dm object with \code{\link[=dm]{dm()}}.
}
\examples{
\dontshow{if (rlang::is_installed("DiagrammeR")) withAutoprint(\{ # examplesIf}
dm <- dm_nycflights13()
dm_deconstruct(dm)
airlines <- pull_tbl(dm, "airlines", keyed = TRUE)
airports <- pull_tbl(dm, "airports", keyed = TRUE)
flights <- pull_tbl(dm, "flights", keyed = TRUE)
planes <- pull_tbl(dm, "planes", keyed = TRUE)
weather <- pull_tbl(dm, "weather", keyed = TRUE)
by_origin <-
  flights \%>\%
  summarize(.by = origin, mean_arr_delay = mean(arr_delay, na.rm = TRUE))

by_origin
dm(airlines, airports, flights, planes, weather, by_origin) \%>\%
  dm_draw()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr-src.R
\name{dplyr_src}
\alias{dplyr_src}
\alias{dm_get_src}
\alias{tbl.dm}
\alias{src_tbls.dm}
\alias{copy_to.dm}
\title{dm as data source}
\usage{
dm_get_src(x)

\method{tbl}{dm}(src, from, ...)

\method{src_tbls}{dm}(x, ...)

\method{copy_to}{dm}(
  dest,
  df,
  name = deparse(substitute(df)),
  overwrite = FALSE,
  temporary = TRUE,
  repair = "unique",
  quiet = FALSE,
  ...
)
}
\arguments{
\item{src}{A \code{dm} object.}

\item{from}{A length one character variable containing the name of the requested table}

\item{...}{See original function documentation}

\item{dest}{For \code{copy_to.dm()}: The \code{dm} object to which a table should be copied.}

\item{df}{For \code{copy_to.dm()}: A table (can be on a different \code{src})}

\item{name}{For \code{copy_to.dm()}: See \code{\link[dplyr:copy_to]{dplyr::copy_to()}}}

\item{overwrite}{For \code{copy_to.dm()}: See \code{\link[dplyr:copy_to]{dplyr::copy_to()}}; \code{TRUE} leads to an error}

\item{temporary}{For \code{copy_to.dm()}: If the \code{dm} is on a DB, the copied version of \code{df} will only be written temporarily to the DB.
After the connection is reset it will no longer be available.}

\item{repair, quiet}{Name repair options; cf. \code{\link[vctrs:vec_as_names]{vctrs::vec_as_names()}}}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

These methods are deprecated because of their limited use,
and because the notion of a "source" seems to be getting phased out from dplyr.
Use other ways to access the tables in a \code{dm}.
}
\details{
Use \code{\link[=dm_get_con]{dm_get_con()}} instead of \code{dm_get_src()} to get the DBI connetion for a
\code{dm} object

Use \code{\link[base:Extract]{[[}} instead of \code{tbl()} to access individual tables in a \code{dm} object.

Get the names from \code{\link[=dm_get_tables]{dm_get_tables()}} instead of calling \code{dm_get_src()}
to list the table names in a \code{dm} object.

Use \code{\link[=copy_to]{copy_to()}} on a table and then \code{\link[=dm]{dm()}} instead of \code{copy_to()}
on a \code{dm} object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disambiguate.R
\name{dm_disambiguate_cols}
\alias{dm_disambiguate_cols}
\title{Resolve column name ambiguities}
\usage{
dm_disambiguate_cols(
  dm,
  .sep = ".",
  ...,
  .quiet = FALSE,
  .position = c("suffix", "prefix")
)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{.sep}{The character variable that separates the names of the table and the names of the ambiguous columns.}

\item{...}{These dots are for future extensions and must be empty.}

\item{.quiet}{Boolean.
By default, this function lists the renamed columns in a message, pass \code{TRUE} to suppress this message.}

\item{.position}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
By default, table names are appended to the column names to resolve conflicts.
Prepending table names was the default for versions before 1.0.0,
use \code{"prefix"} to achieve this behavior.}
}
\value{
A \code{dm} whose column names are unambiguous.
}
\description{
This function ensures that all columns in a \code{dm} have unique names.
}
\details{
The function first checks if there are any column names that are not unique.
If there are, those columns will be assigned new, unique, names by prefixing their existing name
with the name of their table and a separator.
Columns that act as primary or foreign keys will not be renamed
because only the foreign key column will remain when two tables are joined,
making that column name "unique" as well.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_disambiguate_cols()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primary-keys.R
\name{dm_add_pk}
\alias{dm_add_pk}
\title{Add a primary key}
\usage{
dm_add_pk(
  dm,
  table,
  columns,
  ...,
  autoincrement = FALSE,
  check = FALSE,
  force = FALSE
)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.}

\item{columns}{Table columns, unquoted.
To define a compound key, use \code{c(col1, col2)}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{autoincrement}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
If \code{TRUE}, the  column specified in \code{columns} will be populated
automatically with a sequence of integers.}

\item{check}{Boolean, if \code{TRUE}, a check is made if the combination of columns is a unique key of the table.}

\item{force}{Boolean, if \code{FALSE} (default), an error will be thrown if there is already a primary key
set for this table.
If \code{TRUE}, a potential old \code{pk} is deleted before setting a new one.}
}
\value{
An updated \code{dm} with an additional primary key.
}
\description{
\code{dm_add_pk()} marks the specified columns as the primary key of the specified table.
If \code{check == TRUE}, then it will first check if
the given combination of columns is a unique key of the table.
If \code{force == TRUE}, the function will replace an already
set key, without altering foreign keys previously pointing to that primary key.
}
\details{
There can be only one primary key per table in a \code{\link{dm}}.
It's possible though to set an unlimited number of unique keys using \code{\link[=dm_add_uk]{dm_add_uk()}}
or adding foreign keys pointing to columns other than the primary key columns with \code{\link[=dm_add_fk]{dm_add_fk()}}.
}
\examples{
\dontshow{if (rlang::is_installed(c("nycflights13", "DiagrammeR"))) withAutoprint(\{ # examplesIf}
nycflights_dm <- dm(
  planes = nycflights13::planes,
  airports = nycflights13::airports,
  weather = nycflights13::weather
)

nycflights_dm \%>\%
  dm_draw()

# Create primary keys:
nycflights_dm \%>\%
  dm_add_pk(planes, tailnum) \%>\%
  dm_add_pk(airports, faa, check = TRUE) \%>\%
  dm_add_pk(weather, c(origin, time_hour)) \%>\%
  dm_draw()

# Keys can be checked during creation:
try(
  nycflights_dm \%>\%
    dm_add_pk(planes, manufacturer, check = TRUE)
)
\dontshow{\}) # examplesIf}
}
\seealso{
Other primary key functions:
\code{\link[=dm_add_uk]{dm_add_uk()}},
\code{\link[=dm_get_all_pks]{dm_get_all_pks()}},
\code{\link[=dm_get_all_uks]{dm_get_all_uks()}},
\code{\link[=dm_has_pk]{dm_has_pk()}},
\code{\link[=dm_rm_pk]{dm_rm_pk()}},
\code{\link[=dm_rm_uk]{dm_rm_uk()}},
\code{\link[=enum_pk_candidates]{enum_pk_candidates()}}
}
\concept{primary key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_sql.R
\name{dm_sql}
\alias{dm_sql}
\alias{dm_ddl_pre}
\alias{dm_dml_load}
\alias{dm_ddl_post}
\title{Create \emph{DDL} and \emph{DML} scripts for a \code{dm} a and database connection}
\usage{
dm_sql(dm, dest, table_names = NULL, temporary = TRUE)

dm_ddl_pre(dm, dest, table_names = NULL, temporary = TRUE)

dm_dml_load(dm, dest, table_names = NULL, temporary = TRUE)

dm_ddl_post(dm, dest, table_names = NULL, temporary = TRUE)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{dest}{Connection to database.}

\item{table_names}{A named character vector or named vector of \link[DBI:Id]{DBI::Id},
\link[DBI:SQL]{DBI::SQL} or dbplyr objects created with \code{\link[dbplyr:ident]{dbplyr::ident()}}, \code{\link[dbplyr:in_schema]{dbplyr::in_schema()}}
or \code{\link[dbplyr:in_catalog]{dbplyr::in_catalog()}},
with one unique element for each table in \code{dm}.
The default, \code{NULL}, means to use the original table names.}

\item{temporary}{Should the tables be marked as \emph{temporary}? Defaults to \code{TRUE}.}
}
\value{
Nested list of SQL statements.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Generate SQL scripts to create tables, load data and set constraints, keys and indices.
This function powers \code{\link[=copy_dm_to]{copy_dm_to()}} and is useful if you need more control
over the process of copying a \code{dm} to a database.
}
\details{
\itemize{
\item{ \code{dm_ddl_pre()} generates \verb{CREATE TABLE} statements (including \verb{PRIMARY KEY} definition). }
\item{ \code{dm_dml_load()} generates \verb{INSERT INTO} statements. }
\item{ \code{dm_ddl_post()} generates scripts for \verb{FOREIGN KEY}, \verb{UNIQUE KEY} and \code{INDEX}. }
\item{ \code{dm_sql()} calls all three above and returns a complete set of scripts. }
}
}
\examples{
\dontshow{if (rlang::is_installed(c("RSQLite", "dbplyr"))) withAutoprint(\{ # examplesIf}
con <- DBI::dbConnect(RSQLite::SQLite())
dm <- dm_nycflights13()
s <- dm_sql(dm, con)
s
DBI::dbDisconnect(con)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr.R
\name{dplyr_join}
\alias{dplyr_join}
\alias{left_join.dm_zoomed}
\alias{left_join.dm_keyed_tbl}
\alias{inner_join.dm_zoomed}
\alias{inner_join.dm_keyed_tbl}
\alias{full_join.dm_zoomed}
\alias{full_join.dm_keyed_tbl}
\alias{right_join.dm_zoomed}
\alias{right_join.dm_keyed_tbl}
\alias{semi_join.dm_zoomed}
\alias{semi_join.dm_keyed_tbl}
\alias{anti_join.dm_zoomed}
\alias{anti_join.dm_keyed_tbl}
\alias{nest_join.dm_zoomed}
\alias{cross_join.dm_zoomed}
\alias{cross_join.dm_keyed_tbl}
\title{\pkg{dplyr} join methods for zoomed dm objects}
\usage{
\method{left_join}{dm_zoomed}(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL,
  select = NULL
)

\method{left_join}{dm_keyed_tbl}(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = FALSE,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL
)

\method{inner_join}{dm_zoomed}(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL,
  select = NULL
)

\method{inner_join}{dm_keyed_tbl}(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = FALSE,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL
)

\method{full_join}{dm_zoomed}(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  relationship = NULL,
  select = NULL
)

\method{full_join}{dm_keyed_tbl}(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = FALSE,
  na_matches = c("na", "never"),
  multiple = "all",
  relationship = NULL
)

\method{right_join}{dm_zoomed}(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL,
  select = NULL
)

\method{right_join}{dm_keyed_tbl}(
  x,
  y,
  by = NULL,
  copy = NULL,
  suffix = NULL,
  ...,
  keep = FALSE,
  na_matches = c("na", "never"),
  multiple = "all",
  unmatched = "drop",
  relationship = NULL
)

\method{semi_join}{dm_zoomed}(
  x,
  y,
  by = NULL,
  copy = NULL,
  ...,
  na_matches = c("na", "never"),
  suffix = NULL,
  select = NULL
)

\method{semi_join}{dm_keyed_tbl}(x, y, by = NULL, copy = NULL, ..., na_matches = c("na", "never"))

\method{anti_join}{dm_zoomed}(
  x,
  y,
  by = NULL,
  copy = NULL,
  ...,
  na_matches = c("na", "never"),
  suffix = NULL,
  select = NULL
)

\method{anti_join}{dm_keyed_tbl}(x, y, by = NULL, copy = NULL, ..., na_matches = c("na", "never"))

\method{nest_join}{dm_zoomed}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  keep = NULL,
  name = NULL,
  ...,
  na_matches = c("na", "never"),
  unmatched = "drop"
)

\method{cross_join}{dm_zoomed}(x, y, ..., copy = NULL, suffix = c(".x", ".y"))

\method{cross_join}{dm_keyed_tbl}(x, y, ..., copy = NULL, suffix = c(".x", ".y"))
}
\arguments{
\item{x, y}{tbls to join. \code{x} is the \code{dm_zoomed} and \code{y} is another table in the \code{dm}.}

\item{by}{If left \code{NULL} (default), the join will be performed by via the foreign key relation that exists between the originally zoomed table (now \code{x})
and the other table (\code{y}).
If you provide a value (for the syntax see \code{\link[dplyr:join]{dplyr::join}}), you can also join tables that are not connected in the \code{dm}.}

\item{copy}{Disabled, since all tables in a \code{dm} are by definition on the same \code{src}.}

\item{suffix}{Disabled, since columns are disambiguated automatically if necessary, changing the column names to \code{table_name.column_name}.}

\item{...}{see \code{\link[dplyr:join]{dplyr::join}}}

\item{keep}{Should the join keys from both \code{x} and \code{y} be preserved in the
output?
\itemize{
\item If \code{NULL}, the default, joins on equality retain only the keys from \code{x},
while joins on inequality retain the keys from both inputs.
\item If \code{TRUE}, all keys from both inputs are retained.
\item If \code{FALSE}, only keys from \code{x} are retained. For right and full joins,
the data in key columns corresponding to rows that only exist in \code{y} are
merged into the key columns from \code{x}. Can't be used when joining on
inequality conditions.
}}

\item{na_matches}{Should two \code{NA} or two \code{NaN} values match?
\itemize{
\item \code{"na"}, the default, treats two \code{NA} or two \code{NaN} values as equal, like
\code{\%in\%}, \code{\link[=match]{match()}}, and \code{\link[=merge]{merge()}}.
\item \code{"never"} treats two \code{NA} or two \code{NaN} values as different, and will
never match them together or to any other values. This is similar to joins
for database sources and to \code{base::merge(incomparables = NA)}.
}}

\item{multiple}{Handling of rows in \code{x} with multiple matches in \code{y}.
For each row of \code{x}:
\itemize{
\item \code{"all"}, the default, returns every match detected in \code{y}. This is the
same behavior as SQL.
\item \code{"any"} returns one match detected in \code{y}, with no guarantees on which
match will be returned. It is often faster than \code{"first"} and \code{"last"}
if you just need to detect if there is at least one match.
\item \code{"first"} returns the first match detected in \code{y}.
\item \code{"last"} returns the last match detected in \code{y}.
}}

\item{unmatched}{How should unmatched keys that would result in dropped rows
be handled?
\itemize{
\item \code{"drop"} drops unmatched keys from the result.
\item \code{"error"} throws an error if unmatched keys are detected.
}

\code{unmatched} is intended to protect you from accidentally dropping rows
during a join. It only checks for unmatched keys in the input that could
potentially drop rows.
\itemize{
\item For left joins, it checks \code{y}.
\item For right joins, it checks \code{x}.
\item For inner joins, it checks both \code{x} and \code{y}. In this case, \code{unmatched} is
also allowed to be a character vector of length 2 to specify the behavior
for \code{x} and \code{y} independently.
}}

\item{relationship}{Handling of the expected relationship between the keys of
\code{x} and \code{y}. If the expectations chosen from the list below are
invalidated, an error is thrown.
\itemize{
\item \code{NULL}, the default, doesn't expect there to be any relationship between
\code{x} and \code{y}. However, for equality joins it will check for a many-to-many
relationship (which is typically unexpected) and will warn if one occurs,
encouraging you to either take a closer look at your inputs or make this
relationship explicit by specifying \code{"many-to-many"}.

See the \emph{Many-to-many relationships} section for more details.
\item \code{"one-to-one"} expects:
\itemize{
\item Each row in \code{x} matches at most 1 row in \code{y}.
\item Each row in \code{y} matches at most 1 row in \code{x}.
}
\item \code{"one-to-many"} expects:
\itemize{
\item Each row in \code{y} matches at most 1 row in \code{x}.
}
\item \code{"many-to-one"} expects:
\itemize{
\item Each row in \code{x} matches at most 1 row in \code{y}.
}
\item \code{"many-to-many"} doesn't perform any relationship checks, but is provided
to allow you to be explicit about this relationship if you know it
exists.
}

\code{relationship} doesn't handle cases where there are zero matches. For that,
see \code{unmatched}.}

\item{select}{Select a subset of the \strong{RHS-table}'s columns, the syntax being \code{select = c(col_1, col_2, col_3)} (unquoted or quoted).
This argument is specific for the \code{join}-methods for \code{dm_zoomed}.
The table's \code{by} column(s) are automatically added if missing in the selection.}

\item{name}{The name of the list-column created by the join. If \code{NULL},
the default, the name of \code{y} is used.}
}
\description{
Use these methods without the '.dm_zoomed' suffix (see examples).
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
flights_dm <- dm_nycflights13()
dm_zoom_to(flights_dm, flights) \%>\%
  left_join(airports, select = c(faa, name))

# this should illustrate that tables don't necessarily need to be connected
dm_zoom_to(flights_dm, airports) \%>\%
  semi_join(airlines, by = "name")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_gui.R
\name{dm_gui}
\alias{dm_gui}
\title{Shiny app for defining dm objects}
\usage{
dm_gui(..., dm = NULL, select_tables = TRUE, debug = FALSE)
}
\arguments{
\item{...}{These dots are for future extensions and must be empty.}

\item{dm}{An initial dm object, currently required.}

\item{select_tables}{Show selectize input to select tables?}

\item{debug}{Set to \code{TRUE} to simplify debugging of the app.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

This function starts a Shiny application that allows to define \code{dm} objects
from a database or from local data frames.
The application generates R code that can be inserted or copy-pasted
into an R script or function.
}
\details{
In a future release, the app will also allow composing \code{dm} objects directly
from database connections or data frames.

The signature of this function is subject to change without notice.
This should not pose too many problems, because it will usually be run
interactively.
}
\examples{
\dontrun{
dm <- dm_nycflights13(cycle = TRUE)
dm_gui(dm = dm)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_pack_join.R
\name{json_pack_join}
\alias{json_pack_join}
\title{JSON pack join}
\usage{
json_pack_join(x, y, by = NULL, ..., copy = FALSE, keep = FALSE, name = NULL)
}
\arguments{
\item{x, y}{A pair of data frames or data frame extensions (e.g. a tibble).}

\item{by}{A join specification created with \code{\link[dplyr:join_by]{join_by()}}, or a character
vector of variables to join by.

If \code{NULL}, the default, \verb{*_join()} will perform a natural join, using all
variables in common across \code{x} and \code{y}. A message lists the variables so
that you can check they're correct; suppress the message by supplying \code{by}
explicitly.

To join on different variables between \code{x} and \code{y}, use a \code{\link[dplyr:join_by]{join_by()}}
specification. For example, \code{join_by(a == b)} will match \code{x$a} to \code{y$b}.

To join by multiple variables, use a \code{\link[dplyr:join_by]{join_by()}} specification with
multiple expressions. For example, \code{join_by(a == b, c == d)} will match
\code{x$a} to \code{y$b} and \code{x$c} to \code{y$d}. If the column names are the same between
\code{x} and \code{y}, you can shorten this by listing only the variable names, like
\code{join_by(a, c)}.

\code{\link[dplyr:join_by]{join_by()}} can also be used to perform inequality, rolling, and overlap
joins. See the documentation at \link[dplyr:join_by]{?join_by} for details on
these types of joins.

For simple equality joins, you can alternatively specify a character vector
of variable names to join by. For example, \code{by = c("a", "b")} joins \code{x$a}
to \code{y$a} and \code{x$b} to \code{y$b}. If variable names differ between \code{x} and \code{y},
use a named character vector like \code{by = c("x_a" = "y_a", "x_b" = "y_b")}.

To perform a cross-join, generating all combinations of \code{x} and \code{y}, see
\code{\link[dplyr:cross_join]{cross_join()}}.}

\item{...}{Other parameters passed onto methods.}

\item{copy}{If \code{x} and \code{y} are not from the same data source,
and \code{copy} is \code{TRUE}, then \code{y} will be copied into the
same src as \code{x}.  This allows you to join tables across srcs, but
it is a potentially expensive operation so you must opt into it.}

\item{keep}{Should the new list-column contain join keys? The default
will preserve the join keys for inequality joins.}

\item{name}{The name of the list-column created by the join. If \code{NULL},
the default, the name of \code{y} is used.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

A wrapper around \code{\link[=pack_join]{pack_join()}} which stores the joined data into a JSON column.
\code{json_pack_join()} returns all rows and columns in \code{x} with a new JSON columns that contains all packed matches from \code{y}.
}
\examples{
df1 <- tibble::tibble(x = 1:3)
df2 <- tibble::tibble(x = c(1, 1, 2), y = c("first", "second", "third"))
df3 <- json_pack_join(df1, df2)
df3
df3$df2
}
\seealso{
\code{\link[=pack_join]{pack_join()}}, \code{\link[=json_nest_join]{json_nest_join()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unique-keys.R
\name{dm_rm_uk}
\alias{dm_rm_uk}
\title{Remove a unique key}
\usage{
dm_rm_uk(dm, table = NULL, columns = NULL, ...)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.
Pass \code{NULL} to remove all matching keys.}

\item{columns}{Table columns, unquoted.
To refer to a compound key, use \code{c(col1, col2)}.
Pass \code{NULL} (the default) to remove all matching keys.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
An updated \code{dm} without the indicated unique key(s).
}
\description{
\code{dm_rm_uk()} removes one or more unique keys from a table and leaves the \code{\link{dm}} object otherwise unaltered.
An error is thrown if no unique key matches the selection criteria.
If the selection criteria are ambiguous, a message with unambiguous replacement code is shown.
Foreign keys are never removed.
}
\seealso{
Other primary key functions:
\code{\link[=dm_add_pk]{dm_add_pk()}},
\code{\link[=dm_add_uk]{dm_add_uk()}},
\code{\link[=dm_get_all_pks]{dm_get_all_pks()}},
\code{\link[=dm_get_all_uks]{dm_get_all_uks()}},
\code{\link[=dm_has_pk]{dm_has_pk()}},
\code{\link[=dm_rm_pk]{dm_rm_pk()}},
\code{\link[=enum_pk_candidates]{enum_pk_candidates()}}
}
\concept{primary key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/examine-constraints.R
\name{dm_examine_constraints}
\alias{dm_examine_constraints}
\title{Validate your data model}
\usage{
dm_examine_constraints(
  .dm,
  ...,
  .progress = NA,
  .max_value = 6L,
  dm = deprecated(),
  progress = deprecated()
)
}
\arguments{
\item{.dm}{A \code{dm} object.}

\item{...}{These dots are for future extensions and must be empty.}

\item{.progress}{Whether to display a progress bar, if \code{NA} (the default)
hide in non-interactive mode, show in interactive mode. Requires the
'progress' package.}

\item{.max_value}{Maximum number of distinct problematic values to report
in the \code{problem} column, defaults to \code{6}.
Set to \code{Inf} to report all values.}

\item{dm, progress}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}}
}
\value{
A tibble with the following columns:
\describe{
\item{\code{table}}{the table in the \code{dm},}
\item{\code{kind}}{"PK" or "FK",}
\item{\code{columns}}{the table columns that define the key,}
\item{\code{ref_table}}{for foreign keys, the referenced table,}
\item{\code{is_key}}{logical,}
\item{\code{problem}}{if \code{is_key = FALSE}, the reason for that.}
}
}
\description{
This function returns a tibble with information about
which key constraints are met (\code{is_key = TRUE}) or violated (\code{FALSE}).
The printing for this object is special, use \code{\link[tibble:as_tibble]{as_tibble()}}
to print as a regular tibble.
}
\details{
For the primary key constraints, it is tested if the values in the respective columns are all unique.
For the foreign key constraints, the tests check if for each foreign key constraint, the values of the foreign key column
form a subset of the values of the referenced column.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_examine_constraints()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primary-keys.R
\name{dm_has_pk}
\alias{dm_has_pk}
\title{Check for primary key}
\usage{
dm_has_pk(dm, table, ...)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A logical value: \code{TRUE} if the given table has a primary key, \code{FALSE} otherwise.
}
\description{
\code{dm_has_pk()} checks if a given table has columns marked as its primary key.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_has_pk(flights)
dm_nycflights13() \%>\%
  dm_has_pk(planes)
\dontshow{\}) # examplesIf}
}
\seealso{
Other primary key functions:
\code{\link[=dm_add_pk]{dm_add_pk()}},
\code{\link[=dm_add_uk]{dm_add_uk()}},
\code{\link[=dm_get_all_pks]{dm_get_all_pks()}},
\code{\link[=dm_get_all_uks]{dm_get_all_uks()}},
\code{\link[=dm_rm_pk]{dm_rm_pk()}},
\code{\link[=dm_rm_uk]{dm_rm_uk()}},
\code{\link[=enum_pk_candidates]{enum_pk_candidates()}}
}
\concept{primary key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/key-helpers.R
\name{check_subset}
\alias{check_subset}
\title{Check column values for subset}
\usage{
check_subset(x, y, ..., x_select = NULL, y_select = NULL, by_position = NULL)
}
\arguments{
\item{x, y}{A data frame or lazy table.}

\item{...}{These dots are for future extensions and must be empty.}

\item{x_select, y_select}{Key columns to restrict the check, processed with
\code{\link[dplyr:select]{dplyr::select()}}.}

\item{by_position}{Set to \code{TRUE} to ignore column names and match
by position instead.
The default means matching by name, use \code{x_select} and/or \code{y_select}
to align the names.}
}
\value{
Returns \code{x}, invisibly, if the check is passed.
Otherwise an error is thrown and the reason for it is explained.
}
\description{
\code{check_subset()} tests if \code{x} is a subset of \code{y}.
For convenience, the \code{x_select} and \code{y_select} arguments allow restricting the check
to a set of key columns without affecting the return value.
}
\examples{
data_1 <- tibble::tibble(a = c(1, 2, 1), b = c(1, 4, 1), c = c(5, 6, 7))
data_2 <- tibble::tibble(a = c(1, 2, 3), b = c(4, 5, 6), c = c(7, 8, 9))
# this is passing:
check_subset(data_1, data_2, x_select = a, y_select = a)

# this is failing:
try(check_subset(data_2, data_1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pixarfilms.R
\name{dm_pixarfilms}
\alias{dm_pixarfilms}
\title{Creates a dm object for the \pkg{pixarfilms} data}
\usage{
dm_pixarfilms(..., color = TRUE, consistent = FALSE, version = "v1")
}
\arguments{
\item{...}{These dots are for future extensions and must be empty.}

\item{color}{Boolean, if \code{TRUE} (default), the resulting \code{dm} object will
have colors assigned to different tables for visualization with
\code{dm_draw()}.}

\item{consistent}{Boolean, In the original \code{dm}  the  \code{film} column in
\code{pixar_films} contains missing values so cannot be made a proper primary key.
Set to \code{TRUE} to remove those records.}

\item{version}{The version of the data to use.
\code{"v1"} (default) uses a vendored snapshot of \pkg{pixarfilms} 0.2.1.
\code{"latest"} uses the data from the installed \pkg{pixarfilms} package.}
}
\value{
A \code{dm} object consisting of \pkg{pixarfilms} tables, complete with
primary and foreign keys and optionally colored.
}
\description{
Creates an example \code{\link{dm}} object from the tables in
\pkg{pixarfilms}, along with the references.
}
\examples{
\dontshow{if (rlang::is_installed("DiagrammeR")) withAutoprint(\{ # examplesIf}
dm_pixarfilms()
dm_pixarfilms() \%>\%
  dm_draw()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm.R
\name{materialize}
\alias{materialize}
\alias{compute.dm}
\alias{collect.dm}
\title{Materialize}
\usage{
\method{compute}{dm}(x, ..., temporary = TRUE)

\method{collect}{dm}(x, ..., progress = NA)
}
\arguments{
\item{x}{A \code{dm} object.}

\item{...}{Passed on to \code{\link[=compute]{compute()}}.}

\item{temporary}{Must remain \code{TRUE}.}

\item{progress}{Whether to display a progress bar, if \code{NA} (the default)
hide in non-interactive mode, show in interactive mode. Requires the
'progress' package.}
}
\value{
A \code{dm} object of the same structure as the input.
}
\description{
\code{compute()} materializes all tables in a \code{dm} to new temporary
tables on the database.

\code{collect()} downloads the tables in a \code{dm} object as local \link{tibble}s.
}
\details{
Called on a \code{dm} object, these methods create a copy of all tables in the \code{dm}.
Depending on the size of your data this may take a long time.

To create permament tables, first create the database schema using \code{\link[=copy_dm_to]{copy_dm_to()}}
or \code{\link[=dm_sql]{dm_sql()}}, and then use \code{\link[=dm_rows_append]{dm_rows_append()}}.
}
\examples{
\dontshow{if (dm:::dm_has_financial() && rlang::is_installed("RSQLite")) withAutoprint(\{ # examplesIf}
financial <- dm_financial_sqlite()

financial \%>\%
  pull_tbl(districts) \%>\%
  dbplyr::remote_name()

# compute() copies the data to new tables:
financial \%>\%
  compute() \%>\%
  pull_tbl(districts) \%>\%
  dbplyr::remote_name()

# collect() returns a local dm:
financial \%>\%
  collect() \%>\%
  pull_tbl(districts) \%>\%
  class()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rows-dm.R
\name{rows-dm}
\alias{rows-dm}
\alias{dm_rows_insert}
\alias{dm_rows_...}
\alias{dm_rows_append}
\alias{dm_rows_update}
\alias{dm_rows_patch}
\alias{dm_rows_upsert}
\alias{dm_rows_delete}
\title{Modifying rows for multiple tables}
\usage{
dm_rows_insert(x, y, ..., in_place = NULL, progress = NA)

dm_rows_append(x, y, ..., in_place = NULL, progress = NA)

dm_rows_update(x, y, ..., in_place = NULL, progress = NA)

dm_rows_patch(x, y, ..., in_place = NULL, progress = NA)

dm_rows_upsert(x, y, ..., in_place = NULL, progress = NA)

dm_rows_delete(x, y, ..., in_place = NULL, progress = NA)
}
\arguments{
\item{x}{Target \code{dm} object.}

\item{y}{\code{dm} object with new data.}

\item{...}{These dots are for future extensions and must be empty.}

\item{in_place}{Should \code{x} be modified in place? This argument is only
relevant for mutable backends (e.g. databases, data.tables).

When \code{TRUE}, a modified version of \code{x} is returned invisibly;
when \code{FALSE}, a new object representing the resulting changes is returned.}

\item{progress}{Whether to display a progress bar, if \code{NA} (the default)
hide in non-interactive mode, show in interactive mode. Requires the
'progress' package.}
}
\value{
A dm object of the same \code{\link[=dm_ptype]{dm_ptype()}} as \code{x}.
If \code{in_place = TRUE}, the underlying data is updated as a side effect,
and \code{x} is returned, invisibly.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

These functions provide a framework for updating data in existing tables.
Unlike \code{\link[=compute]{compute()}}, \code{\link[=copy_to]{copy_to()}} or \code{\link[=copy_dm_to]{copy_dm_to()}}, no new tables are created
on the database.
All operations expect that both existing and new data are presented
in two compatible \link{dm} objects on the same data source.

The functions make sure that the tables in the target dm
are processed in topological order so that parent (dimension)
tables receive insertions before child (fact) tables.

These operations, in contrast to all other operations,
may lead to irreversible changes to the underlying database.
Therefore, in-place operation must be requested explicitly with \code{in_place = TRUE}.
By default, an informative message is given.

\code{dm_rows_insert()} adds new records via \code{\link[=rows_insert]{rows_insert()}} with \code{conflict = "ignore"}.
Duplicate records will be silently discarded.
This operation requires primary keys on all tables, use \code{dm_rows_append()}
to insert unconditionally.

\code{dm_rows_append()} adds new records via \code{\link[=rows_append]{rows_append()}}.
The primary keys must differ from existing records.
This must be ensured by the caller and might be checked by the underlying database.
Use \code{in_place = FALSE} and apply \code{\link[=dm_examine_constraints]{dm_examine_constraints()}} to check beforehand.

\code{dm_rows_update()} updates existing records via \code{\link[=rows_update]{rows_update()}}.
Primary keys must match for all records to be updated.

\code{dm_rows_patch()} updates missing values in existing records
via \code{\link[=rows_patch]{rows_patch()}}.
Primary keys must match for all records to be patched.

\code{dm_rows_upsert()} updates existing records and adds new records,
based on the primary key, via \code{\link[=rows_upsert]{rows_upsert()}}.

\code{dm_rows_delete()} removes matching records via \code{\link[=rows_delete]{rows_delete()}},
based on the primary key.
The order in which the tables are processed is reversed.
}
\examples{
\dontshow{if (rlang::is_installed(c("RSQLite", "nycflights13", "dbplyr"))) withAutoprint(\{ # examplesIf}
# Establish database connection:
sqlite <- DBI::dbConnect(RSQLite::SQLite())

# Entire dataset with all dimension tables populated
# with flights and weather data truncated:
flights_init <-
  dm_nycflights13() \%>\%
  dm_zoom_to(flights) \%>\%
  filter(FALSE) \%>\%
  dm_update_zoomed() \%>\%
  dm_zoom_to(weather) \%>\%
  filter(FALSE) \%>\%
  dm_update_zoomed()

# Target database:
flights_sqlite <- copy_dm_to(sqlite, flights_init, temporary = FALSE)
print(dm_nrow(flights_sqlite))

# First update:
flights_jan <-
  dm_nycflights13() \%>\%
  dm_select_tbl(flights, weather) \%>\%
  dm_zoom_to(flights) \%>\%
  filter(month == 1) \%>\%
  dm_update_zoomed() \%>\%
  dm_zoom_to(weather) \%>\%
  filter(month == 1) \%>\%
  dm_update_zoomed()
print(dm_nrow(flights_jan))

# Copy to temporary tables on the target database:
flights_jan_sqlite <- copy_dm_to(sqlite, flights_jan)

# Dry run by default:
dm_rows_append(flights_sqlite, flights_jan_sqlite)
print(dm_nrow(flights_sqlite))

# Explicitly request persistence:
dm_rows_append(flights_sqlite, flights_jan_sqlite, in_place = TRUE)
print(dm_nrow(flights_sqlite))

# Second update:
flights_feb <-
  dm_nycflights13() \%>\%
  dm_select_tbl(flights, weather) \%>\%
  dm_zoom_to(flights) \%>\%
  filter(month == 2) \%>\%
  dm_update_zoomed() \%>\%
  dm_zoom_to(weather) \%>\%
  filter(month == 2) \%>\%
  dm_update_zoomed()

# Copy to temporary tables on the target database:
flights_feb_sqlite <- copy_dm_to(sqlite, flights_feb)

# Explicit dry run:
flights_new <- dm_rows_append(
  flights_sqlite,
  flights_feb_sqlite,
  in_place = FALSE
)
print(dm_nrow(flights_new))
print(dm_nrow(flights_sqlite))

# Check for consistency before applying:
flights_new \%>\%
  dm_examine_constraints()

# Apply:
dm_rows_append(flights_sqlite, flights_feb_sqlite, in_place = TRUE)
print(dm_nrow(flights_sqlite))

DBI::dbDisconnect(sqlite)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm-flatten.R
\name{dm_flatten}
\alias{dm_flatten}
\title{Flatten a table in a \code{dm} by joining its parent tables}
\usage{
dm_flatten(
  dm,
  table,
  ...,
  parent_tables = NULL,
  recursive = FALSE,
  allow_deep = FALSE,
  join = left_join
)
}
\arguments{
\item{dm}{A \code{\link{dm}} object.}

\item{table}{The table to flatten by joining its parent tables.
An interesting choice could be
for example a fact table in a star schema.}

\item{...}{These dots are for future extensions and must be empty.}

\item{parent_tables}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Unquoted names of the parent tables to be joined into \code{table}.
The order of the tables here determines the order of the joins.
If \code{NULL} (the default), all direct parent tables
are joined in non-recursive mode,
or all reachable ancestor tables in recursive mode.
\code{tidyselect} is supported, see \code{\link[dplyr:select]{dplyr::select()}} for details on the semantics.}

\item{recursive}{Logical, defaults to \code{FALSE}.
If \code{TRUE}, recursively flatten parent tables before joining them
into \code{table}.
Uses simple recursion: recursively flattening the parents
and then doing a join in order.
If \code{FALSE}, fails if a parent table has further parents
(unless \code{allow_deep} is \code{TRUE}).
Cannot be \code{TRUE} when \code{allow_deep} is \code{TRUE}.}

\item{allow_deep}{Logical, defaults to \code{FALSE}.
Only relevant if \code{recursive = FALSE}.
If \code{TRUE}, parent tables with further parents are allowed
and will remain in the result with a
foreign-key relationship to the flattened table.
Cannot be \code{TRUE} when \code{recursive} is \code{TRUE}.}

\item{join}{The type of join to use when combining parent tables,
see \code{\link[dplyr:join]{dplyr::join()}}.
Defaults to \code{\link[dplyr:left_join]{dplyr::left_join()}}.
\code{nest_join} is not supported.
When \code{recursive = TRUE}, only \code{\link[dplyr:left_join]{dplyr::left_join()}},
\code{\link[dplyr:inner_join]{dplyr::inner_join()}}, and \code{\link[dplyr:full_join]{dplyr::full_join()}} are supported.}
}
\value{
A \code{\link{dm}} object with the flattened table and removed parent tables.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{dm_flatten()} updates a table in-place by joining its parent tables into it,
and removes the now integrated parent tables from the dm.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_select_tbl(-weather) \%>\%
  dm_flatten(flights, recursive = TRUE)
\dontshow{\}) # examplesIf}
}
\seealso{
Other flattening functions:
\code{\link[=dm_flatten_to_tbl]{dm_flatten_to_tbl()}}
}
\concept{flattening functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-cardinalities.R
\name{examine_cardinality}
\alias{examine_cardinality}
\alias{check_cardinality_0_n}
\alias{check_cardinality_...}
\alias{check_cardinality_1_n}
\alias{check_cardinality_1_1}
\alias{check_cardinality_0_1}
\title{Check table relations}
\usage{
check_cardinality_0_n(
  x,
  y,
  ...,
  x_select = NULL,
  y_select = NULL,
  by_position = NULL
)

check_cardinality_1_n(
  x,
  y,
  ...,
  x_select = NULL,
  y_select = NULL,
  by_position = NULL
)

check_cardinality_1_1(
  x,
  y,
  ...,
  x_select = NULL,
  y_select = NULL,
  by_position = NULL
)

check_cardinality_0_1(
  x,
  y,
  ...,
  x_select = NULL,
  y_select = NULL,
  by_position = NULL
)

examine_cardinality(
  x,
  y,
  ...,
  x_select = NULL,
  y_select = NULL,
  by_position = NULL
)
}
\arguments{
\item{x}{Parent table, data frame or lazy table.}

\item{y}{Child table, data frame or lazy table.}

\item{...}{These dots are for future extensions and must be empty.}

\item{x_select, y_select}{Key columns to restrict the check, processed with
\code{\link[dplyr:select]{dplyr::select()}}.}

\item{by_position}{Set to \code{TRUE} to ignore column names and match
by position instead.
The default means matching by name, use \code{x_select} and/or \code{y_select}
to align the names.}
}
\value{
\code{check_cardinality_...()} return \code{x}, invisibly,
if the check is passed, to support pipes.
Otherwise an error is thrown and the reason for it is explained.

\code{examine_cardinality()} returns a character variable specifying the type of relationship between the two columns.
}
\description{
All \code{check_cardinality_...()} functions test the following conditions:
\enumerate{
\item Are all rows in \code{x} unique?
\item Are the rows in \code{y} a subset of the rows in \code{x}?
\item Does the relation between \code{x} and \code{y} meet the cardinality requirements?
One row from \code{x} must correspond to the requested number of rows in \code{y},
e.g. \verb{_0_1} means that there must be zero or one rows in \code{y} for each
row in \code{x}.
}

\code{examine_cardinality()} also checks the first two points and subsequently determines the type of cardinality.

For convenience, the \code{x_select} and \code{y_select} arguments allow restricting the check
to a set of key columns without affecting the return value.
}
\details{
All cardinality functions accept a parent and a child table (\code{x} and \code{y}).
All rows in \code{x} must be unique, and all rows in \code{y} must be a subset of the
rows in \code{x}.
The \code{x_select} and \code{y_select} arguments allow restricting the check
to a set of key columns without affecting the return value.
If given, both arguments must refer to the same number of key columns.

The cardinality specifications "0_n", "1_n", "0_1", "1_1" refer to the expected relation that the child table has with the parent table.
"0", "1" and "n" refer to the occurrences of value combinations
in \code{y} that correspond to each combination in the
columns of the parent table.
"n" means "more than one" in this context, with no upper limit.

\strong{"0_n"}: no restrictions, each row in \code{x} has at least 0 and at most
n corresponding occurrences in \code{y}.

\strong{"1_n"}: each row in \code{x} has at least 1 and at most
n corresponding occurrences in \code{y}.
This means that there is a "surjective" mapping from the child table
to the parent table, i.e. each parent table row exists at least once in the
child table.

\strong{"0_1"}: each row in \code{x} has at least 0 and at most
1 corresponding occurrence in \code{y}.
This means that there is a "injective" mapping from the child table
to the parent table, i.e. no combination of values in the
parent table columns is addressed multiple times.
But not all parent table rows have to be referred to.

\strong{"1_1"}: each row in \code{x} occurs exactly once in \code{y}.
This means that there is a "bijective" ("injective" AND "surjective") mapping
between the child table and the parent table, i.e. the
sets of rows are identical.

Finally, \code{examine_cardinality()} tests for and returns the nature of the relationship
(injective, surjective, bijective, or none of these)
between the two given sets of columns.
If either \code{x} is not unique or there are rows in \code{y} that are missing from \code{x},
the requirements for a cardinality test is not fulfilled.
No error will be thrown, but
the result will contain the information which prerequisite was violated.
}
\examples{
d1 <- tibble::tibble(a = 1:5)
d2 <- tibble::tibble(a = c(1:4, 4L))
d3 <- tibble::tibble(c = c(1:5, 5L), d = 0)
# This does not pass, `a` is not unique key of d2:
try(check_cardinality_0_n(d2, d1))

# Columns are matched by name by default:
try(check_cardinality_0_n(d1, d3))

# This passes, multiple values in d3$c are allowed:
check_cardinality_0_n(d1, d2)

# This does not pass, injectivity is violated:
try(check_cardinality_1_1(d1, d3, y_select = c(a = c)))
try(check_cardinality_0_1(d1, d3, x_select = c(c = a)))

# What kind of cardinality is it?
examine_cardinality(d1, d3, x_select = c(c = a))
examine_cardinality(d1, d2)
}
\seealso{
Other cardinality functions:
\code{\link[=dm_examine_cardinalities]{dm_examine_cardinalities()}}
}
\concept{cardinality functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table-surgery.R
\name{reunite_parent_child}
\alias{reunite_parent_child}
\alias{reunite_parent_child_from_list}
\title{Merge two tables that are linked by a foreign key relation}
\usage{
reunite_parent_child(child_table, parent_table, id_column)

reunite_parent_child_from_list(list_of_parent_child_tables, id_column)
}
\arguments{
\item{child_table}{Table (possibly created by \code{decompose_table()}) that references \code{parent_table}}

\item{parent_table}{Table (possibly created by \code{decompose_table()}).}

\item{id_column}{Identical name of referencing / referenced column in \code{child_table}/\code{parent_table}.}

\item{list_of_parent_child_tables}{Cf arguments \code{child_table} and \code{parent_table} from
\code{reunite_parent_child()}, but both in a named list (as created by \code{decompose_table()}).}
}
\value{
A wide table produced by joining the two given tables.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Perform table fusion by combining two tables by a common (key) column, and then removing this column.

\code{reunite_parent_child()}: After joining the two tables by the column \code{id_column}, this column will be removed.
The transformation is roughly the
inverse of what \code{decompose_table()} does.

\code{reunite_parent_child_from_list()}: After joining the two tables
by the column \code{id_column}, \code{id_column} is removed.

This function is almost exactly the inverse of \code{decompose_table()} (the order
of the columns is not retained, and the original row names are lost).
}
\section{Life cycle}{

These functions are marked "experimental" because they seem more useful
when applied to a table in a dm object.
Changing the interface later seems harmless because these functions are
most likely used interactively.
}

\examples{
decomposed_table <- decompose_table(mtcars, new_id, am, gear, carb)
ct <- decomposed_table$child_table
pt <- decomposed_table$parent_table

reunite_parent_child(ct, pt, new_id)
reunite_parent_child_from_list(decomposed_table, new_id)
}
\seealso{
Other table surgery functions:
\code{\link[=decompose_table]{decompose_table()}}
}
\concept{table surgery functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paste.R
\name{dm_paste}
\alias{dm_paste}
\title{Create R code for a dm object}
\usage{
dm_paste(dm, select = NULL, ..., tab_width = 2, options = NULL, path = NULL)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{select}{Deprecated, see \code{"select"} in the \code{options} argument.}

\item{...}{Must be empty.}

\item{tab_width}{Indentation width for code from the second line onwards}

\item{options}{Formatting options. A character vector containing some of:
\itemize{
\item \code{"tables"}: \code{\link[=tibble]{tibble()}} calls for empty table definitions
derived from \code{\link[=dm_ptype]{dm_ptype()}}, overrides \code{"select"}.
\item \code{"select"}: \code{\link[=dm_select]{dm_select()}} statements for columns that are part
of the dm.
\item \code{"keys"}: \code{\link[=dm_add_pk]{dm_add_pk()}}, \code{\link[=dm_add_fk]{dm_add_fk()}} and \code{\link[=dm_add_uk]{dm_add_uk()}} statements for adding keys.
\item \code{"color"}: \code{\link[=dm_set_colors]{dm_set_colors()}} statements to set color.
\item \code{"all"}: All options above except \code{"select"}
}

Default \code{NULL} is equivalent to \code{c("keys", "color")}}

\item{path}{Output file, if \code{NULL} the code is printed to the console.}
}
\value{
Code for producing the prototype of the given \code{dm}.
}
\description{
\code{dm_paste()} takes an existing \code{dm} and emits the code necessary for its creation.
}
\details{
The code emitted by the function reproduces the structure of the \code{dm} object.
The \code{options} argument controls the level of detail: keys, colors,
table definitions.
Data in the tables is never included, see \code{\link[=dm_ptype]{dm_ptype()}} for the underlying logic.
}
\examples{
dm() \%>\%
  dm_paste()
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}

dm_nycflights13() \%>\%
  dm_paste()

dm_nycflights13() \%>\%
  dm_paste(options = "select")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm.R
\name{pull_tbl}
\alias{pull_tbl}
\title{Retrieve a table}
\usage{
pull_tbl(dm, table, ..., keyed = FALSE)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{One unquoted table name for \code{pull_tbl.dm()}, ignored for \code{pull_tbl.dm_zoomed()}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{keyed}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Set to \code{TRUE} to return objects of the internal class \code{"dm_keyed_tbl"}
that will contain information on primary and foreign keys
in the individual table objects.
This allows using dplyr workflows on those tables and later reconstruct them
into a \code{dm} object.
See \code{\link[=dm_deconstruct]{dm_deconstruct()}} for a function that generates corresponding code
for an existing dm object, and \code{vignette("tech-dm-keyed")} for details.}
}
\value{
The requested table.
}
\description{
This generic has methods for both \code{dm} classes:
\enumerate{
\item With \code{pull_tbl.dm()} you can chose which table of the \code{dm} you want to retrieve.
\item With \code{pull_tbl.dm_zoomed()} you will retrieve the zoomed table in the current state.
}
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
# For an unzoomed dm you need to specify the table to pull:
dm_nycflights13() \%>\%
  pull_tbl(airports)

# If zoomed, pulling detaches the zoomed table from the dm:
dm_nycflights13() \%>\%
  dm_zoom_to(airports) \%>\%
  pull_tbl()
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=dm_deconstruct]{dm_deconstruct()}} to generate code of the form
\code{pull_tbl(..., keyed = TRUE)} from an existing \code{dm} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select.R
\name{dm_rename}
\alias{dm_rename}
\title{Rename columns}
\usage{
dm_rename(dm, table, ...)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.}

\item{...}{One or more unquoted expressions separated by commas.
You can treat
variable names as if they were positions, and use expressions like \code{x:y}
to select the ranges of variables.

Use named arguments, e.g. \code{new_name = old_name}, to rename the selected variables.

The arguments in ... are automatically quoted and evaluated in a context where
column names represent column positions.
They also support unquoting and splicing.
See \code{vignette("programming", package = "dplyr")} for an introduction to those concepts.

See select helpers for more details, and the examples about \link[tidyselect:language]{tidyselect helpers}, such as \code{starts_with()}, \code{everything()}, etc.}
}
\value{
An updated \code{dm} with the columns of \code{table} renamed.
}
\description{
Rename the columns of your \code{\link{dm}} using syntax that is similar to \code{\link[dplyr:rename]{dplyr::rename()}}.
}
\details{
If key columns are renamed, then the meta-information of the \code{dm} is updated accordingly.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_rename(airports, code = faa, altitude = alt)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unique-keys.R
\name{dm_add_uk}
\alias{dm_add_uk}
\title{Add a unique key}
\usage{
dm_add_uk(dm, table, columns, ..., check = FALSE)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.}

\item{columns}{Table columns, unquoted.
To define a compound key, use \code{c(col1, col2)}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{check}{Boolean, if \code{TRUE}, a check is made if the combination of columns is a unique key of the table.}
}
\value{
An updated \code{dm} with an additional unqiue key.
}
\description{
\code{dm_add_uk()} marks the specified columns as a unique key of the specified table.
If \code{check == TRUE}, then it will first check if
the given combination of columns is a unique key of the table.
}
\details{
The difference between a primary key (PK) and a unique key (UK) consists in the following:
\itemize{
\item When a local \code{dm} is copied to a database (DB) with \code{copy_dm_to()}, a PK will be set on the DB by default, whereas a UK is being ignored.
\item A PK can be set as an \code{autoincrement} key (also implemented on certain DBMS when the \code{dm} is transferred to the DB)
\item There can be only one PK for each table, whereas there can be unlimited UKs
\item A UK will be used, if the same table has an autoincrement PK in addition, to ensure that during delta load processes
on the DB (cf. \code{\link[=dm_rows_append]{dm_rows_append()}}) the foreign keys are updated accordingly.
If no UK is available, the insertion is done row-wise, which also ensures a correct matching, but can be much slower.
\item A UK can generally enhance the data model by adding additional information
\item There can also be implicit UKs, when the columns addressed by a foreign key are neither a PK nor a UK.
These implicit UKs are also listed by \code{\link[=dm_get_all_uks]{dm_get_all_uks()}}
}
}
\examples{
\dontshow{if (rlang::is_installed(c("nycflights13", "DiagrammeR"))) withAutoprint(\{ # examplesIf}

nycflights_dm <- dm(
  planes = nycflights13::planes,
  airports = nycflights13::airports,
  weather = nycflights13::weather
)

# Create unique keys:
nycflights_dm \%>\%
  dm_add_uk(planes, tailnum) \%>\%
  dm_add_uk(airports, faa, check = TRUE) \%>\%
  dm_add_uk(weather, c(origin, time_hour)) \%>\%
  dm_get_all_uks()

# Keys can be checked during creation:
try(
  nycflights_dm \%>\%
    dm_add_uk(planes, manufacturer, check = TRUE)
)
\dontshow{\}) # examplesIf}
}
\seealso{
Other primary key functions:
\code{\link[=dm_add_pk]{dm_add_pk()}},
\code{\link[=dm_get_all_pks]{dm_get_all_pks()}},
\code{\link[=dm_get_all_uks]{dm_get_all_uks()}},
\code{\link[=dm_has_pk]{dm_has_pk()}},
\code{\link[=dm_rm_pk]{dm_rm_pk()}},
\code{\link[=dm_rm_uk]{dm_rm_uk()}},
\code{\link[=enum_pk_candidates]{enum_pk_candidates()}}
}
\concept{primary key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_unnest_tbl.R
\name{dm_unpack_tbl}
\alias{dm_unpack_tbl}
\title{Unpack columns from a wrapped table}
\usage{
dm_unpack_tbl(dm, child_table, col, ptype)
}
\arguments{
\item{dm}{A dm.}

\item{child_table}{A table in the dm with packed columns.}

\item{col}{The column to unpack (unquoted).}

\item{ptype}{A dm, only used to query names of primary and foreign keys.}
}
\description{
#' @description
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\details{
\code{dm_unpack_tbl()} targets a specific column to unpack
from the given table in a given dm.
A ptype or a set of keys should be given,
not both.

\code{\link[=dm_pack_tbl]{dm_pack_tbl()}} is an inverse operation to \code{dm_unpack_tbl()}
if differences in row and column order are ignored.
The opposite is true if referential constraints between both tables
are satisfied
and if all rows in the parent table have at least one child row,
i.e. if the relationship is of cardinality 1:n or 1:1.
}
\examples{
flights_wrapped <-
  dm_nycflights13() \%>\%
  dm_wrap_tbl(flights)

# The ptype is required for reconstruction.
# It can be an empty dm, only primary and foreign keys are considered.
ptype <- dm_ptype(dm_nycflights13())

flights_wrapped \%>\%
  dm_unpack_tbl(flights, airlines, ptype)
}
\seealso{
\code{\link[=dm_unwrap_tbl]{dm_unwrap_tbl()}}, \code{\link[=dm_unnest_tbl]{dm_unnest_tbl()}},
\code{\link[=dm_nest_tbl]{dm_nest_tbl()}}, \code{\link[=dm_pack_tbl]{dm_pack_tbl()}}, \code{\link[=dm_wrap_tbl]{dm_wrap_tbl()}},
\code{\link[=dm_examine_constraints]{dm_examine_constraints()}}, \code{\link[=dm_examine_cardinalities]{dm_examine_cardinalities()}},
\code{\link[=dm_ptype]{dm_ptype()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primary-keys.R
\name{enum_pk_candidates}
\alias{enum_pk_candidates}
\alias{dm_enum_pk_candidates}
\title{Primary key candidate}
\usage{
enum_pk_candidates(table, ...)

dm_enum_pk_candidates(dm, table, ...)
}
\arguments{
\item{table}{A table in the \code{dm}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{dm}{A \code{dm} object.}
}
\value{
A tibble with the following columns:
\describe{
\item{\code{columns}}{columns of \code{table},}
\item{\code{candidate}}{boolean: are these columns a candidate for a primary key,}
\item{\code{why}}{if not a candidate for a primary key column, explanation for this.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{enum_pk_candidates()} checks for each column of a
table if the column contains only unique values, and is thus
a suitable candidate for a primary key of the table.

\code{dm_enum_pk_candidates()} performs these checks
for a table in a \link{dm} object.
}
\section{Life cycle}{

These functions are marked "experimental" because we are not yet sure about
the interface, in particular if we need both \code{dm_enum...()} and \code{enum...()}
variants.
Changing the interface later seems harmless because these functions are
most likely used interactively.
}

\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
nycflights13::flights \%>\%
  enum_pk_candidates()
\dontshow{\}) # examplesIf}
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}

dm_nycflights13() \%>\%
  dm_enum_pk_candidates(airports)
\dontshow{\}) # examplesIf}
}
\seealso{
Other primary key functions:
\code{\link[=dm_add_pk]{dm_add_pk()}},
\code{\link[=dm_add_uk]{dm_add_uk()}},
\code{\link[=dm_get_all_pks]{dm_get_all_pks()}},
\code{\link[=dm_get_all_uks]{dm_get_all_uks()}},
\code{\link[=dm_has_pk]{dm_has_pk()}},
\code{\link[=dm_rm_pk]{dm_rm_pk()}},
\code{\link[=dm_rm_uk]{dm_rm_uk()}}
}
\concept{primary key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primary-keys.R
\name{dm_get_all_pks}
\alias{dm_get_all_pks}
\title{Get all primary keys of a \code{\link{dm}} object}
\usage{
dm_get_all_pks(dm, table = NULL, ...)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{One or more table names, unquoted,
to return primary key information for.
If given, primary keys are returned in that order.
The default \code{NULL} returns information for all tables.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A tibble with the following columns:
\describe{
\item{\code{table}}{table name,}
\item{\code{pk_col}}{column name(s) of primary key, as list of character vectors.}
}
}
\description{
\code{dm_get_all_pks()} checks the \code{dm} object for primary keys and
returns the tables and the respective primary key columns.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_get_all_pks()
\dontshow{\}) # examplesIf}
}
\seealso{
Other primary key functions:
\code{\link[=dm_add_pk]{dm_add_pk()}},
\code{\link[=dm_add_uk]{dm_add_uk()}},
\code{\link[=dm_get_all_uks]{dm_get_all_uks()}},
\code{\link[=dm_has_pk]{dm_has_pk()}},
\code{\link[=dm_rm_pk]{dm_rm_pk()}},
\code{\link[=dm_rm_uk]{dm_rm_uk()}},
\code{\link[=enum_pk_candidates]{enum_pk_candidates()}}
}
\concept{primary key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_from_con.R
\name{dm_from_src}
\alias{dm_from_src}
\title{Load a dm from a remote data source}
\usage{
dm_from_src(src = NULL, table_names = NULL, learn_keys = NULL, ...)
}
\arguments{
\item{src}{A dbplyr source, DBI connection object or a Pool object.}

\item{table_names}{A character vector of the names of the tables to include.}

\item{learn_keys}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Set to \code{TRUE} to query the definition of primary and
foreign keys from the database.
Currently works for Postgres/Redshift, MariaDB/MySQL, SQLite, SQL Server, and DuckDB databases.
The default attempts to query and issues an informative message.}

\item{...}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Additional parameters for the schema learning query.
\itemize{
\item \code{schema}: supported for MSSQL (default: \code{"dbo"}), Postgres/Redshift (default: \code{"public"}), MariaDB/MySQL
(default: current database) and SQLite (default: main schema).
Learn the tables in a specific schema (or database for MariaDB/MySQL).
\item \code{dbname}: supported for MSSQL. Access different databases on the connected MSSQL-server;
default: active database.
\item \code{table_type}: supported for Postgres/Redshift (default: \code{"BASE TABLE"}). Specify the table type. Options are:
\enumerate{
\item \code{"BASE TABLE"} for a persistent table (normal table type)
\item \code{"VIEW"} for a view
\item \code{"FOREIGN TABLE"} for a foreign table
\item \code{"LOCAL TEMPORARY"} for a temporary table
}
}}
}
\description{
Deprecated  in dm 0.3.0 in favor of \code{\link[=dm_from_con]{dm_from_con()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ptype.R
\name{dm_ptype}
\alias{dm_ptype}
\title{Prototype for a dm object}
\usage{
dm_ptype(dm)
}
\arguments{
\item{dm}{A \code{dm} object.}
}
\description{
The prototype contains all tables, all primary and foreign keys,
but no data.
All tables are truncated and converted to zero-row tibbles,
also for remote data models.
Columns retain their type.
This is useful for performing creation and population of a database
in separate steps.
}
\examples{
\dontshow{if (dm:::dm_has_financial()) withAutoprint(\{ # examplesIf}
dm_financial() \%>\%
  dm_ptype()

dm_financial() \%>\%
  dm_ptype() \%>\%
  dm_nrow()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreign-keys.R
\name{dm_enum_fk_candidates}
\alias{dm_enum_fk_candidates}
\alias{enum_fk_candidates}
\title{Foreign key candidates}
\usage{
dm_enum_fk_candidates(dm, table, ref_table, ...)

enum_fk_candidates(dm_zoomed, ref_table, ...)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{The table whose columns should be tested for suitability as foreign keys.}

\item{ref_table}{A table with a primary key.}

\item{...}{These dots are for future extensions and must be empty.}

\item{dm_zoomed}{A \code{dm} with a zoomed table.}
}
\value{
A tibble with the following columns:
\describe{
\item{\code{columns}}{columns of \code{table},}
\item{\code{candidate}}{boolean: are these columns a candidate for a foreign key,}
\item{\code{why}}{if not a candidate for a foreign key, explanation for for this.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Determine which columns would be good candidates to be used as foreign keys of a table,
to reference the primary key column of another table of the \code{\link{dm}} object.
}
\details{
\code{dm_enum_fk_candidates()} first checks if \code{ref_table} has a primary key set,
if not, an error is thrown.

If \code{ref_table} does have a primary key, then a join operation will be tried using
that key as the \code{by} argument of join() to match it to each column of \code{table}.
Attempting to join incompatible columns triggers an error.

The outcome of the join operation determines the value of the \code{why} column in the result:
\itemize{
\item an empty value for a column of \code{table} that is a suitable foreign key candidate
\item the count and percentage of missing matches for a column that is not suitable
\item the error message triggered for unsuitable candidates that may include the types of mismatched columns
}

\code{enum_fk_candidates()} works like \code{dm_enum_fk_candidates()} with the zoomed table as \code{table}.
}
\section{Life cycle}{

These functions are marked "experimental" because we are not yet sure about
the interface, in particular if we need both \code{dm_enum...()} and \code{enum...()}
variants.
Changing the interface later seems harmless because these functions are
most likely used interactively.
}

\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_enum_fk_candidates(flights, airports)

dm_nycflights13() \%>\%
  dm_zoom_to(flights) \%>\%
  enum_fk_candidates(airports)
\dontshow{\}) # examplesIf}
}
\seealso{
Other foreign key functions:
\code{\link[=dm_add_fk]{dm_add_fk()}},
\code{\link[=dm_get_all_fks]{dm_get_all_fks()}},
\code{\link[=dm_rm_fk]{dm_rm_fk()}}
}
\concept{foreign key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreign-keys.R
\name{dm_rm_fk}
\alias{dm_rm_fk}
\title{Remove foreign keys}
\usage{
dm_rm_fk(
  dm,
  table = NULL,
  columns = NULL,
  ref_table = NULL,
  ref_columns = NULL,
  ...
)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.
Pass \code{NULL} to remove all matching keys.}

\item{columns}{Table columns, unquoted.
To refer to a compound key, use \code{c(col1, col2)}.
Pass \code{NULL} (the default) to remove all matching keys.}

\item{ref_table}{The table referenced by the \code{table} argument.
Pass \code{NULL} to remove all matching keys.}

\item{ref_columns}{The columns of \code{table} that should no longer be referencing the primary key of \code{ref_table}.
To refer to a compound key, use \code{c(col1, col2)}.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
An updated \code{dm} without the matching foreign key relation(s).
}
\description{
\code{dm_rm_fk()} can remove either one reference between two tables, or multiple references at once (with a message).
An error is thrown if no matching foreign key is found.
}
\examples{
\dontshow{if (rlang::is_installed(c("nycflights13", "DiagrammeR"))) withAutoprint(\{ # examplesIf}
dm_nycflights13(cycle = TRUE) \%>\%
  dm_rm_fk(flights, dest, airports) \%>\%
  dm_draw()
\dontshow{\}) # examplesIf}
}
\seealso{
Other foreign key functions:
\code{\link[=dm_add_fk]{dm_add_fk()}},
\code{\link[=dm_enum_fk_candidates]{dm_enum_fk_candidates()}},
\code{\link[=dm_get_all_fks]{dm_get_all_fks()}}
}
\concept{foreign key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_nest_tbl.R
\name{dm_nest_tbl}
\alias{dm_nest_tbl}
\title{Nest a table inside its dm}
\usage{
dm_nest_tbl(dm, child_table, into = NULL)
}
\arguments{
\item{dm}{A dm.}

\item{child_table}{A terminal table with one parent table.}

\item{into}{The table to nest \code{child_tables} into, optional as it can be guessed
from the foreign keys unambiguously but useful to be explicit.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{dm_nest_tbl()} converts a child table to a nested column in its parent
table.
The child table should not have children itself (i.e. it needs to be a
\emph{terminal child table}).
}
\examples{
nested_dm <-
  dm_nycflights13() \%>\%
  dm_select_tbl(airlines, flights) \%>\%
  dm_nest_tbl(flights)

nested_dm

nested_dm$airlines
}
\seealso{
\code{\link[=dm_wrap_tbl]{dm_wrap_tbl()}}, \code{\link[=dm_unwrap_tbl]{dm_unwrap_tbl()}}, \code{\link[=dm_pack_tbl]{dm_pack_tbl()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/key-helpers.R
\name{check_set_equality}
\alias{check_set_equality}
\title{Check column values for set equality}
\usage{
check_set_equality(
  x,
  y,
  ...,
  x_select = NULL,
  y_select = NULL,
  by_position = NULL
)
}
\arguments{
\item{x, y}{A data frame or lazy table.}

\item{...}{These dots are for future extensions and must be empty.}

\item{x_select, y_select}{Key columns to restrict the check, processed with
\code{\link[dplyr:select]{dplyr::select()}}.}

\item{by_position}{Set to \code{TRUE} to ignore column names and match
by position instead.
The default means matching by name, use \code{x_select} and/or \code{y_select}
to align the names.}
}
\value{
Returns \code{x}, invisibly, if the check is passed.
Otherwise an error is thrown and the reason for it is explained.
}
\description{
\code{check_set_equality()} is a wrapper of \code{\link[=check_subset]{check_subset()}}.

It tests if one table is a subset of another and vice versa, i.e., if both sets are the same.
If not, it throws an error.
}
\examples{
data_1 <- tibble::tibble(a = c(1, 2, 1), b = c(1, 4, 1), c = c(5, 6, 7))
data_2 <- tibble::tibble(a = c(1, 2, 3), b = c(4, 5, 6), c = c(7, 8, 9))
# this is failing:
try(check_set_equality(data_1, data_2, x_select = a, y_select = a))

data_3 <- tibble::tibble(a = c(2, 1, 2), b = c(4, 5, 6), c = c(7, 8, 9))
# this is passing:
check_set_equality(data_1, data_3, x_select = a, y_select = a)
# this is still failing:
try(check_set_equality(data_2, data_3))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pack_join.R
\name{pack_join}
\alias{pack_join}
\alias{pack_join.dm_zoomed}
\title{Pack Join}
\usage{
pack_join(x, y, by = NULL, ..., copy = FALSE, keep = FALSE, name = NULL)

\method{pack_join}{dm_zoomed}(x, y, by = NULL, ..., copy = FALSE, keep = FALSE, name = NULL)
}
\arguments{
\item{x, y}{A pair of data frames or data frame extensions (e.g. a tibble).}

\item{by}{A join specification created with \code{\link[dplyr:join_by]{join_by()}}, or a character
vector of variables to join by.

If \code{NULL}, the default, \verb{*_join()} will perform a natural join, using all
variables in common across \code{x} and \code{y}. A message lists the variables so
that you can check they're correct; suppress the message by supplying \code{by}
explicitly.

To join on different variables between \code{x} and \code{y}, use a \code{\link[dplyr:join_by]{join_by()}}
specification. For example, \code{join_by(a == b)} will match \code{x$a} to \code{y$b}.

To join by multiple variables, use a \code{\link[dplyr:join_by]{join_by()}} specification with
multiple expressions. For example, \code{join_by(a == b, c == d)} will match
\code{x$a} to \code{y$b} and \code{x$c} to \code{y$d}. If the column names are the same between
\code{x} and \code{y}, you can shorten this by listing only the variable names, like
\code{join_by(a, c)}.

\code{\link[dplyr:join_by]{join_by()}} can also be used to perform inequality, rolling, and overlap
joins. See the documentation at \link[dplyr:join_by]{?join_by} for details on
these types of joins.

For simple equality joins, you can alternatively specify a character vector
of variable names to join by. For example, \code{by = c("a", "b")} joins \code{x$a}
to \code{y$a} and \code{x$b} to \code{y$b}. If variable names differ between \code{x} and \code{y},
use a named character vector like \code{by = c("x_a" = "y_a", "x_b" = "y_b")}.

To perform a cross-join, generating all combinations of \code{x} and \code{y}, see
\code{\link[dplyr:cross_join]{cross_join()}}.}

\item{...}{Other parameters passed onto methods.}

\item{copy}{If \code{x} and \code{y} are not from the same data source,
and \code{copy} is \code{TRUE}, then \code{y} will be copied into the
same src as \code{x}.  This allows you to join tables across srcs, but
it is a potentially expensive operation so you must opt into it.}

\item{keep}{Should the new list-column contain join keys? The default
will preserve the join keys for inequality joins.}

\item{name}{The name of the list-column created by the join. If \code{NULL},
the default, the name of \code{y} is used.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{pack_join()} returns all rows and columns in \code{x} with a new packed column
that contains all matches from \code{y}.
}
\examples{
df1 <- tibble::tibble(x = 1:3)
df2 <- tibble::tibble(x = c(1, 1, 2), y = c("first", "second", "third"))
pack_join(df1, df2)
}
\seealso{
\code{\link[dplyr:nest_join]{dplyr::nest_join()}}, \code{\link[tidyr:pack]{tidyr::pack()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nycflights13.R
\name{dm_nycflights13}
\alias{dm_nycflights13}
\title{Creates a dm object for the \pkg{nycflights13} data}
\usage{
dm_nycflights13(
  ...,
  cycle = FALSE,
  color = TRUE,
  subset = TRUE,
  compound = TRUE,
  table_description = FALSE
)
}
\arguments{
\item{...}{These dots are for future extensions and must be empty.}

\item{cycle}{Boolean.
If \code{FALSE} (default), only one foreign key relation
(from \code{flights$origin} to \code{airports$faa}) between the \code{flights} table and the \code{airports} table is
established.
If \code{TRUE}, a \code{dm} object with a double reference
between those tables will be produced.}

\item{color}{Boolean, if \code{TRUE} (default), the resulting \code{dm} object will have
colors assigned to different tables for visualization with \code{dm_draw()}.}

\item{subset}{Boolean, if \code{TRUE} (default), the \code{flights} table is reduced to flights with column \code{day} equal to 10.}

\item{compound}{Boolean, if \code{FALSE}, no link will be established between tables \code{flights} and \code{weather},
because this requires compound keys.}

\item{table_description}{Boolean, if \code{TRUE}, a description will be added for each table that will be displayed
when drawing the table with \code{\link[=dm_draw]{dm_draw()}}.}
}
\value{
A \code{dm} object consisting of \pkg{nycflights13} tables, complete with primary and foreign keys and optionally colored.
}
\description{
Creates an example \code{\link{dm}} object from the tables in \pkg{nycflights13},
along with the references.
See \code{\link[nycflights13:flights]{nycflights13::flights}} for a description of the data.
As described in \code{\link[nycflights13:planes]{nycflights13::planes}}, the relationship
between the \code{flights} table and the \code{planes} tables is "weak", it does not satisfy
data integrity constraints.
}
\examples{
\dontshow{if (rlang::is_installed("DiagrammeR")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_draw()
\dontshow{\}) # examplesIf}
}
\seealso{
\code{vignette("howto-dm-df")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw-dm.R
\name{dm_draw}
\alias{dm_draw}
\title{Draw a diagram of the data model}
\usage{
dm_draw(
  dm,
  rankdir = "LR",
  ...,
  col_attr = NULL,
  view_type = c("keys_only", "all", "title_only"),
  column_types = NULL,
  backend = c("DiagrammeR"),
  backend_opts = list(),
  columnArrows = lifecycle::deprecated(),
  graph_attrs = lifecycle::deprecated(),
  node_attrs = lifecycle::deprecated(),
  edge_attrs = lifecycle::deprecated(),
  focus = lifecycle::deprecated(),
  graph_name = lifecycle::deprecated(),
  font_size = lifecycle::deprecated()
)
}
\arguments{
\item{dm}{A \code{\link{dm}} object.}

\item{rankdir}{Graph attribute for direction (e.g., 'BT' = bottom --> top).}

\item{...}{These dots are for future extensions and must be empty.}

\item{col_attr}{Deprecated, use \code{colummn_types} instead.}

\item{view_type}{Can be "keys_only" (default), "all" or "title_only".
It defines the level of details for rendering tables
(only primary and foreign keys, all columns, or no columns).}

\item{column_types}{Set to \code{TRUE} to show column types.}

\item{backend}{Currently, only the default \code{"DiagrammeR"} is accepted.
Pass this value explicitly if your code relies on the type of the return value.}

\item{backend_opts}{A named list of backend-specific options.
For the \code{"DiagrammeR"} backend, supported options are:
\itemize{
\item \code{graph_attrs}: Additional graph attributes (default \code{""}).
\item \code{node_attrs}: Additional node attributes (default \code{""}).
\item \code{edge_attrs}: Additional edge attributes (default \code{""}).
\item \code{focus}: A list of parameters for rendering (table filter).
\item \code{graph_name}: The name of the graph (default \code{"Data Model"}).
\item \code{column_arrow}: Edges from columns to columns (default \code{TRUE}).
\item \code{font_size}: \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Font size for \code{header} (default \code{16}), \code{column} (default \code{16}),
and \code{table_description} (default \code{8}).
Can be set as a named integer vector, e.g.
\code{c(table_headers = 18L, table_description = 6L)}.
}}

\item{columnArrows}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use
\code{backend_opts = list(column_arrow = ...)} instead.}

\item{graph_attrs}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use
\code{backend_opts = list(graph_attrs = ...)} instead.}

\item{node_attrs}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use
\code{backend_opts = list(node_attrs = ...)} instead.}

\item{edge_attrs}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use
\code{backend_opts = list(edge_attrs = ...)} instead.}

\item{focus}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use
\code{backend_opts = list(focus = ...)} instead.}

\item{graph_name}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use
\code{backend_opts = list(graph_name = ...)} instead.}

\item{font_size}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use
\code{backend_opts = list(font_size = ...)} instead.}
}
\value{
An object with a \code{\link[=print]{print()}} method, which,
when printed, produces the output seen in the viewer as a side effect.
Currently, this is an object of class \code{grViz} (see also
\code{\link[DiagrammeR:grViz]{DiagrammeR::grViz()}}), but this is subject to change.
}
\description{
\code{dm_draw()} draws a diagram, a visual representation of the data model.
}
\details{
Currently, \pkg{dm} uses \pkg{DiagrammeR} to draw diagrams.
Use \code{\link[DiagrammeRsvg:export_svg]{DiagrammeRsvg::export_svg()}} to convert the diagram to an SVG file.

The backend for drawing the diagrams might change in the future.
If you rely on DiagrammeR, pass an explicit value for the \code{backend} argument.
}
\examples{
\dontshow{if (rlang::is_installed(c("nycflights13", "DiagrammeR"))) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_draw()

dm_nycflights13(cycle = TRUE) \%>\%
  dm_draw(view_type = "title_only")

head(dm_get_available_colors())
length(dm_get_available_colors())

dm_nycflights13() \%>\%
  dm_get_colors()
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=dm_set_colors]{dm_set_colors()}} for defining the table colors.

\code{\link[=dm_set_table_description]{dm_set_table_description()}} for adding details to one or more tables in the diagram
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/descriptions.R
\name{dm_set_table_description}
\alias{dm_set_table_description}
\alias{dm_get_table_description}
\alias{dm_reset_table_description}
\title{Add info about a dm's tables}
\usage{
dm_set_table_description(dm, ...)

dm_get_table_description(dm, table = NULL, ...)

dm_reset_table_description(dm, table = NULL, ...)
}
\arguments{
\item{dm}{A \code{\link{dm}} object.}

\item{...}{For \code{dm_set_table_description()}: Descriptions for tables to set in the form \code{description = table}.
\code{tidyselect} is supported, see \code{\link[dplyr:select]{dplyr::select()}} for details on the semantics.

For \code{dm_get_table_description()} and \code{dm_reset_table_description()}: These dots are for future extensions and must be empty.}

\item{table}{One or more table names, unquoted, for which to
\enumerate{
\item get information about the current description(s) with \code{\link[=dm_get_table_description]{dm_get_table_description()}}.
\item remove descriptions with \code{\link[=dm_reset_table_description]{dm_reset_table_description()}}.
}

In both cases the default applies to all tables in the \code{dm}.}
}
\value{
For \code{dm_set_table_description()}: A \code{dm} object containing descriptions for specified tables.

For \code{dm_get_table_description}: A named vector of tables, with the descriptions in the names.

For \code{dm_reset_table_description()}: A \code{dm} object without descriptions for specified tables.
}
\description{
When creating a diagram from a \code{dm} using \code{\link[=dm_draw]{dm_draw()}} the table descriptions set with \code{dm_set_table_description()} will be displayed.
}
\details{
Multi-line descriptions can be achieved using the newline symbol \verb{\\n}.
Descriptions are set with \code{dm_set_table_description()}.
The currently set descriptions can be checked using \code{dm_get_table_description()}.
Descriptions can be removed using \code{dm_reset_table_description()}.
}
\examples{
\dontshow{if (rlang::is_installed(c("nycflights13", "labelled", "DiagrammeR"))) withAutoprint(\{ # examplesIf}
desc_flights <- rlang::set_names(
  "flights",
  paste(
    "On-time data for all flights",
    "that departed NYC (i.e. JFK, LGA or EWR) in 2013.",
    sep = "\n"
  )
)
nyc_desc <- dm_nycflights13() \%>\%
  dm_set_table_description(
    !!desc_flights,
    "Weather at the airport of\norigin at time of departure" = weather
  )
nyc_desc \%>\%
  dm_draw()

dm_get_table_description(nyc_desc)
dm_reset_table_description(nyc_desc, flights) \%>\%
  dm_draw(font_size = c(header = 18L, table_description = 9L, column = 15L))

pull_tbl(nyc_desc, flights) \%>\%
  labelled::label_attribute()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate.R
\name{dm_validate}
\alias{dm_validate}
\title{Validator}
\usage{
dm_validate(x)
}
\arguments{
\item{x}{An object.}
}
\value{
Returns the \code{dm}, invisibly, after finishing all checks.
}
\description{
\code{dm_validate()} checks the internal consistency of a \code{dm} object.
}
\details{
In theory, with the exception of \code{\link[=new_dm]{new_dm()}}, all \code{dm} objects
created or modified by functions in this package should be valid,
and this function should not be needed.
Please file an issue if any dm operation creates an invalid object.
}
\examples{
dm_validate(dm())

bad_dm <- structure(list(bad = "dm"), class = "dm")
try(dm_validate(bad_dm))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreign-keys.R
\name{dm_get_all_fks}
\alias{dm_get_all_fks}
\title{Get foreign key constraints}
\usage{
dm_get_all_fks(dm, parent_table = NULL, ...)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{parent_table}{One or more table names, unquoted,
to return foreign key information for.
If given, foreign keys are returned in that order.
The default \code{NULL} returns information for all tables.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A tibble with the following columns:
\describe{
\item{\code{child_table}}{child table,}
\item{\code{child_fk_cols}}{foreign key column(s) in child table as list of character vectors,}
\item{\code{parent_table}}{parent table,}
\item{\code{parent_key_cols}}{key column(s) in parent table as list of character vectors.}
\item{\code{on_delete}}{behavior on deletion of rows in the parent table.}
}
}
\description{
Get a summary of all foreign key relations in a \code{\link{dm}}.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_get_all_fks()
\dontshow{\}) # examplesIf}
}
\seealso{
Other foreign key functions:
\code{\link[=dm_add_fk]{dm_add_fk()}},
\code{\link[=dm_enum_fk_candidates]{dm_enum_fk_candidates()}},
\code{\link[=dm_rm_fk]{dm_rm_fk()}}
}
\concept{foreign key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw-dm.R
\name{dm_set_colors}
\alias{dm_set_colors}
\alias{dm_get_colors}
\alias{dm_get_available_colors}
\title{Color in database diagrams}
\usage{
dm_set_colors(dm, ...)

dm_get_colors(dm)

dm_get_available_colors()
}
\arguments{
\item{dm}{A \code{\link{dm}} object.}

\item{...}{Colors to set in the form \code{color = table}.
Allowed colors are all hex coded colors (quoted) and the color names from \code{dm_get_available_colors()}.
\code{tidyselect} is supported, see \code{\link[dplyr:select]{dplyr::select()}} for details on the semantics.}
}
\value{
For \code{dm_set_colors()}: the updated data model.

For \code{dm_get_colors()}, a named character vector of table names
with the colors in the names.
This allows calling \code{dm_set_colors(!!!dm_get_colors(...))}.
Use \code{\link[tibble:enframe]{tibble::enframe()}} to convert this to a tibble.

For \code{dm_get_available_colors()}, a vector with the available colors.
}
\description{
\code{dm_set_colors()} allows to define the colors that will be used to display the tables of the data model with \code{\link[=dm_draw]{dm_draw()}}.
The colors can either be specified with hex color codes or using the names of the built-in R colors.
An overview of the colors corresponding to the standard color names can be found at
the bottom of
\url{https://rpubs.com/krlmlr/colors}.

\code{dm_get_colors()} returns the colors defined for a data model.

\code{dm_get_available_colors()} returns an overview of the names of the available colors
These are the standard colors also returned by \code{\link[grDevices:colors]{grDevices::colors()}} plus a default
table color with the name "default".
}
\examples{
\dontshow{if (rlang::is_installed(c("nycflights13", "DiagrammeR"))) withAutoprint(\{ # examplesIf}
dm_nycflights13(color = FALSE) \%>\%
  dm_set_colors(
    darkblue = starts_with("air"),
    "#5986C4" = flights
  ) \%>\%
  dm_draw()

# Splicing is supported:
nyc_cols <-
  dm_nycflights13() \%>\%
  dm_get_colors()
nyc_cols

dm_nycflights13(color = FALSE) \%>\%
  dm_set_colors(!!!nyc_cols) \%>\%
  dm_draw()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select-tbl.R
\name{dm_select_tbl}
\alias{dm_select_tbl}
\alias{dm_rename_tbl}
\title{Select and rename tables}
\usage{
dm_select_tbl(dm, ...)

dm_rename_tbl(dm, ...)
}
\arguments{
\item{dm}{A \code{\link{dm}} object.}

\item{...}{One or more table names of the tables of the \code{\link{dm}} object.
\code{tidyselect} is supported, see \code{\link[dplyr:select]{dplyr::select()}} for details on the semantics.}
}
\value{
The input \code{dm} with tables renamed or removed.
}
\description{
\code{dm_select_tbl()} keeps the selected tables and their relationships,
optionally renaming them.

\code{dm_rename_tbl()} renames tables.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_select_tbl(airports, fl = flights)
\dontshow{\}) # examplesIf}
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_rename_tbl(ap = airports, fl = flights)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_unnest_tbl.R
\name{dm_unnest_tbl}
\alias{dm_unnest_tbl}
\title{Unnest columns from a wrapped table}
\usage{
dm_unnest_tbl(dm, parent_table, col, ptype)
}
\arguments{
\item{dm}{A dm.}

\item{parent_table}{A table in the dm with nested columns.}

\item{col}{The column to unnest (unquoted).}

\item{ptype}{A dm, only used to query names of primary and foreign keys.}
}
\value{
A dm.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{dm_unnest_tbl()} target a specific column to unnest
from the given table in a given dm.
A ptype or a set of keys should be given, not both.
}
\details{
\code{\link[=dm_nest_tbl]{dm_nest_tbl()}} is an inverse operation to \code{dm_unnest_tbl()}
if differences in row and column order are ignored.
The opposite is true if referential constraints between both tables
are satisfied.
}
\examples{
airlines_wrapped <-
  dm_nycflights13() \%>\%
  dm_wrap_tbl(airlines)

# The ptype is required for reconstruction.
# It can be an empty dm, only primary and foreign keys are considered.
ptype <- dm_ptype(dm_nycflights13())

airlines_wrapped \%>\%
  dm_unnest_tbl(airlines, flights, ptype)
}
\seealso{
\code{\link[=dm_unwrap_tbl]{dm_unwrap_tbl()}}, \code{\link[=dm_unpack_tbl]{dm_unpack_tbl()}},
\code{\link[=dm_nest_tbl]{dm_nest_tbl()}}, \code{\link[=dm_pack_tbl]{dm_pack_tbl()}}, \code{\link[=dm_wrap_tbl]{dm_wrap_tbl()}},
\code{\link[=dm_examine_constraints]{dm_examine_constraints()}}, \code{\link[=dm_examine_cardinalities]{dm_examine_cardinalities()}},
\code{\link[=dm_ptype]{dm_ptype()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreign-keys.R
\name{dm_has_fk}
\alias{dm_has_fk}
\alias{dm_get_fk}
\title{Check if foreign keys exists}
\usage{
dm_has_fk(dm, table, ref_table, ...)

dm_get_fk(dm, table, ref_table, ...)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.}

\item{ref_table}{The table which \code{table} will be referencing.}

\item{...}{These dots are for future extensions and must be empty.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

These functions are deprecated because of their limited use
since the introduction of foreign keys to arbitrary columns in dm 0.2.1.
Use \code{\link[=dm_get_all_fks]{dm_get_all_fks()}} with table manipulation functions instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{head.dm_zoomed}
\alias{head.dm_zoomed}
\alias{tail.dm_zoomed}
\title{\pkg{utils} table manipulation methods for \code{dm_zoomed} objects}
\usage{
\method{head}{dm_zoomed}(x, n = 6L, ...)

\method{tail}{dm_zoomed}(x, n = 6L, ...)
}
\arguments{
\item{x}{object of class \code{dm_zoomed}}

\item{n}{an integer vector of length up to \code{dim(x)} (or 1,
    for non-dimensioned objects).  A \code{logical} is silently coerced to
    integer.  Values specify the indices to be
    selected in the corresponding dimension (or along the length) of the
    object. A positive value of \code{n[i]} includes the first/last
    \code{n[i]} indices in that dimension, while a negative value
    excludes the last/first \code{abs(n[i])}, including all remaining
    indices. \code{NA} or non-specified values (when \code{length(n) <
      length(dim(x))}) select all indices in that dimension. Must
    contain at least one non-missing value.}

\item{...}{arguments to be passed to or from other methods.}
}
\value{
A \code{dm_zoomed} object.
}
\description{
Extract the first or last rows from a table.
Use these methods without the '.dm_zoomed' suffix (see examples).
The methods for regular \code{dm} objects extract the first or last tables.
}
\details{
see manual for the corresponding functions in \pkg{utils}.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
zoomed <- dm_nycflights13() \%>\%
  dm_zoom_to(flights) \%>\%
  head(4)
zoomed
dm_insert_zoomed(zoomed, new_tbl_name = "head_flights")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreign-keys.R
\name{dm_add_fk}
\alias{dm_add_fk}
\title{Add foreign keys}
\usage{
dm_add_fk(
  dm,
  table,
  columns,
  ref_table,
  ref_columns = NULL,
  ...,
  check = FALSE,
  on_delete = c("no_action", "cascade")
)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.}

\item{columns}{The columns of \code{table} which are to become the foreign key columns that
reference \code{ref_table}.
To define a compound key, use \code{c(col1, col2)}.}

\item{ref_table}{The table which \code{table} will be referencing.}

\item{ref_columns}{The column(s) of \code{table} which are to become the referenced column(s) in \code{ref_table}.
By default, the primary key is used.
To define a compound key, use \code{c(col1, col2)}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{check}{Boolean, if \code{TRUE}, a check will be performed to determine if the values of
\code{columns} are a subset of the values of the key column(s) of \code{ref_table}.}

\item{on_delete}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Defines behavior if a row in the parent table is deleted.
- \code{"no_action"}, the default, means that no action is taken
and the operation is aborted if child rows exist
- \code{"cascade"} means that the child row is also deleted
This setting is picked up by \code{\link[=copy_dm_to]{copy_dm_to()}} with \code{set_key_constraints = TRUE},
and by \code{\link[=dm_sql]{dm_sql()}},
and might be considered by \code{\link[=dm_rows_delete]{dm_rows_delete()}} in a future version.}
}
\value{
An updated \code{dm} with an additional foreign key relation.
}
\description{
\code{dm_add_fk()} marks the specified \code{columns} as the foreign key of table \code{table} with
respect to a key of table \code{ref_table}.
Usually the referenced columns are a primary key in \code{ref_table}.
However, it is also possible to specify other columns via the \code{ref_columns} argument.
If \code{check == TRUE}, then it will first check if the values in \code{columns} are a subset
of the values of the key in table \code{ref_table}.
}
\details{
It is possible that a foreign key (FK) is pointing to columns that are neither primary (PK) nor
explicit unique keys (UK).
This can happen
\enumerate{
\item when a FK is added without a corresponding PK or UK being present in the parent table
\item when the PK or UK is removed (\code{\link[=dm_rm_pk]{dm_rm_pk()}}/\code{\link[=dm_rm_uk]{dm_rm_uk()}}) without first removing the associated FKs.
}

These columns are then a so-called "implicit unique key" of the referenced table and can be listed via \code{\link[=dm_get_all_uks]{dm_get_all_uks()}}.
}
\examples{
\dontshow{if (rlang::is_installed(c("nycflights13", "DiagrammeR"))) withAutoprint(\{ # examplesIf}
nycflights_dm <- dm(
  planes = nycflights13::planes,
  flights = nycflights13::flights,
  weather = nycflights13::weather
)

nycflights_dm \%>\%
  dm_draw()

# Create foreign keys:
nycflights_dm \%>\%
  dm_add_pk(planes, tailnum) \%>\%
  dm_add_fk(flights, tailnum, planes) \%>\%
  dm_add_pk(weather, c(origin, time_hour)) \%>\%
  dm_add_fk(flights, c(origin, time_hour), weather) \%>\%
  dm_draw()

# Keys can be checked during creation:
try(
  nycflights_dm \%>\%
    dm_add_pk(planes, tailnum) \%>\%
    dm_add_fk(flights, tailnum, planes, check = TRUE)
)
\dontshow{\}) # examplesIf}
}
\seealso{
Other foreign key functions:
\code{\link[=dm_enum_fk_candidates]{dm_enum_fk_candidates()}},
\code{\link[=dm_get_all_fks]{dm_get_all_fks()}},
\code{\link[=dm_rm_fk]{dm_rm_fk()}}
}
\concept{foreign key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_wrap.R
\name{dm_wrap_tbl}
\alias{dm_wrap_tbl}
\title{Wrap dm into a single tibble dm}
\usage{
dm_wrap_tbl(dm, root, strict = TRUE, progress = NA)
}
\arguments{
\item{dm}{A cycle free dm object.}

\item{root}{Table to wrap the dm into (unquoted).}

\item{strict}{Whether to fail for cyclic dms that cannot be wrapped into a
single table, if \code{FALSE} a partially wrapped dm will be returned.}

\item{progress}{Whether to display a progress bar, if \code{NA} (the default)
hide in non-interactive mode, show in interactive mode. Requires the
'progress' package.}
}
\value{
A \code{dm} object.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{dm_wrap_tbl()} creates a single tibble dm containing the \code{root} table
enhanced with all the data related to it
through the relationships stored in the dm.
It runs a sequence of \code{\link[=dm_nest_tbl]{dm_nest_tbl()}} and \code{\link[=dm_pack_tbl]{dm_pack_tbl()}} operations
on the dm.
}
\details{
\code{dm_wrap_tbl()} is an inverse to \code{dm_unwrap_tbl()},
i.e., wrapping after unwrapping returns the same information
(disregarding row and column order).
The opposite is not generally true:
since \code{dm_wrap_tbl()} keeps only rows related directly or indirectly to
rows in the \code{root} table.
Even if all referential constraints are satisfied,
unwrapping after wrapping loses rows in parent tables
that don't have a corresponding row in the child table.

This function differs from \code{dm_flatten_to_tbl()} and \code{dm_squash_to_tbl()} ,
which always return a single table, and not a \code{dm} object.
}
\examples{
dm_nycflights13() \%>\%
  dm_wrap_tbl(root = airlines)
}
\seealso{
\code{\link[=dm_unwrap_tbl]{dm_unwrap_tbl()}}, \code{\link[=dm_nest_tbl]{dm_nest_tbl()}},
\code{\link[=dm_examine_constraints]{dm_examine_constraints()}},
\code{\link[=dm_examine_cardinalities]{dm_examine_cardinalities()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm.R
\name{dm}
\alias{dm}
\alias{new_dm}
\alias{is_dm}
\alias{as_dm}
\title{Data model class}
\usage{
dm(
  ...,
  .name_repair = c("check_unique", "unique", "universal", "minimal"),
  .quiet = FALSE
)

new_dm(tables = list())

is_dm(x)

as_dm(x, ...)
}
\arguments{
\item{...}{Tables or existing \code{dm} objects to add to the \code{dm} object.
Unnamed tables are auto-named, \code{dm} objects must not be named.}

\item{.name_repair, .quiet}{Options for name repair.
Forwarded as \code{repair} and \code{quiet} to \code{\link[vctrs:vec_as_names]{vctrs::vec_as_names()}}.}

\item{tables}{A named list of the tables (tibble-objects, not names),
to be included in the \code{dm} object.}

\item{x}{An object.}
}
\value{
For \code{dm()}, \code{new_dm()}, \code{as_dm()}: A \code{dm} object.

For \code{is_dm()}: A scalar logical, \code{TRUE} if is this object is a \code{dm}.
}
\description{
The \code{dm} class holds a list of tables and their relationships.
It is inspired by \href{https://github.com/bergant/datamodelr}{datamodelr},
and extends the idea by offering operations to access the data in the tables.

\code{dm()} creates a \code{dm} object from \link{tbl} objects
(tibbles or lazy data objects).

\code{new_dm()} is a low-level constructor that creates a new \code{dm} object.
\itemize{
\item If called without arguments, it will create an empty \code{dm}.
\item If called with arguments, no validation checks will be made to ascertain that
the inputs are of the expected class and internally consistent;
use \code{\link[=dm_validate]{dm_validate()}} to double-check the returned object.
}

\code{is_dm()} returns \code{TRUE} if the input is of class \code{dm}.

\code{as_dm()} coerces objects to the \code{dm} class
}
\examples{
dm(trees, mtcars)

new_dm(list(trees = trees, mtcars = mtcars))

as_dm(list(trees = trees, mtcars = mtcars))
\dontshow{if (rlang::is_installed(c("nycflights13", "dbplyr"))) withAutoprint(\{ # examplesIf}

is_dm(dm_nycflights13())

dm_nycflights13()$airports

dm_nycflights13()["airports"]

dm_nycflights13()[["airports"]]

dm_nycflights13() \%>\% names()
\dontshow{\}) # examplesIf}
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
library(dm)
library(nycflights13)

# using `data.frame` objects
new_dm(tibble::lst(weather, airports))

# using `dm_keyed_tbl` objects
dm <- dm_nycflights13()
y1 <- dm$planes \%>\%
  mutate() \%>\%
  select(everything())
y2 <- dm$flights \%>\%
  left_join(dm$airlines, by = "carrier")

new_dm(list("tbl1" = y1, "tbl2" = y2))
\dontshow{\}) # examplesIf}
}
\seealso{
\itemize{
\item \code{\link[=dm_from_con]{dm_from_con()}} for connecting to all tables in a database
and importing the primary and foreign keys
\item \code{\link[=dm_get_tables]{dm_get_tables()}} for returning a list of tables
\item \code{\link[=dm_add_pk]{dm_add_pk()}} and \code{\link[=dm_add_fk]{dm_add_fk()}} for adding primary and foreign keys
\item \code{\link[=copy_dm_to]{copy_dm_to()}} for DB interaction
\item \code{\link[=dm_draw]{dm_draw()}} for visualization
\item \code{\link[=dm_flatten_to_tbl]{dm_flatten_to_tbl()}} for flattening
\item \code{\link[=dm_filter]{dm_filter()}} for filtering
\item \code{\link[=dm_select_tbl]{dm_select_tbl()}} for creating a \code{dm} with only a subset of the tables
\item \code{\link[=dm_nycflights13]{dm_nycflights13()}}  for creating an example \code{dm} object
\item \code{\link[=decompose_table]{decompose_table()}} for table surgery
\item \code{\link[=check_key]{check_key()}} and \code{\link[=check_subset]{check_subset()}} for checking for key properties
\item \code{\link[=examine_cardinality]{examine_cardinality()}} for checking the cardinality of the relation between two tables
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_nest.R
\name{json_nest}
\alias{json_nest}
\title{JSON nest}
\usage{
json_nest(.data, ..., .names_sep = NULL)
}
\arguments{
\item{.data}{A data frame, a data frame extension (e.g. a tibble), or  a lazy data frame (e.g. from dbplyr or dtplyr).}

\item{...}{<\code{\link[tidyr:tidyr_tidy_select]{tidy-select}}> Columns to pack, specified
using name-variable pairs of the form \code{new_col = c(col1, col2, col3)}.
The right hand side can be any valid tidy select expression.}

\item{.names_sep}{If \code{NULL}, the default, the names will be left as is.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

A wrapper around \code{\link[tidyr:nest]{tidyr::nest()}} which stores the nested data into JSON columns.
}
\examples{
df <- tibble::tibble(x = c(1, 1, 1, 2, 2, 3), y = 1:6, z = 6:1)
nested <- json_nest(df, data = c(y, z))
nested
}
\seealso{
\code{\link[tidyr:nest]{tidyr::nest()}}, \code{\link[=json_nest_join]{json_nest_join()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select.R
\name{dm_select}
\alias{dm_select}
\title{Select columns}
\usage{
dm_select(dm, table, ...)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.}

\item{...}{One or more unquoted expressions separated by commas.
You can treat
variable names as if they were positions, and use expressions like \code{x:y}
to select the ranges of variables.

Use named arguments, e.g. \code{new_name = old_name}, to rename the selected variables.

The arguments in ... are automatically quoted and evaluated in a context where
column names represent column positions.
They also support unquoting and splicing.
See \code{vignette("programming", package = "dplyr")} for an introduction to those concepts.

See select helpers for more details, and the examples about \link[tidyselect:select_helpers]{tidyselect helpers}, such as \code{starts_with()}, \code{everything()}, etc.}
}
\value{
An updated \code{dm} with the columns of \code{table} reduced and/or renamed.
}
\description{
Select columns of your \code{\link{dm}} using syntax that is similar to \code{\link[dplyr:select]{dplyr::select()}}.
}
\details{
If key columns are renamed, then the meta-information of the \code{dm} is updated accordingly.
If key columns are removed, then all related relations are dropped as well.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_select(airports, code = faa, altitude = alt)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db-interface.R
\name{copy_dm_to}
\alias{copy_dm_to}
\title{Copy data model to data source}
\usage{
copy_dm_to(
  dest,
  dm,
  ...,
  set_key_constraints = TRUE,
  table_names = NULL,
  temporary = TRUE,
  schema = NULL,
  progress = NA,
  unique_table_names = NULL,
  copy_to = NULL
)
}
\arguments{
\item{dest}{An object of class \code{"src"} or \code{"DBIConnection"}.}

\item{dm}{A \code{dm} object.}

\item{...}{These dots are for future extensions and must be empty.}

\item{set_key_constraints}{If \code{TRUE} will mirror
the primary, foreign, and unique key constraints
and create indexes for foreign key constraints
for the primary and foreign keys in the \code{dm} object.
Set to \code{FALSE} if your data model currently does not satisfy
primary or foreign key constraints.}

\item{table_names}{Desired names for the tables on \code{dest}; the names within the \code{dm} remain unchanged.
Can be \code{NULL}, a named character vector, or a vector of \code{\link[DBI:Id]{DBI::Id}} objects.

If left \code{NULL} (default), the names will be determined automatically depending on the \code{temporary} argument:
\enumerate{
\item \code{temporary = TRUE} (default): unique table names based on the names of the tables in the \code{dm} are created.
\item \code{temporary = FALSE}: the table names in the \code{dm} are used as names for the tables on \code{dest}.
}

If a function or one-sided formula, \code{table_names} is converted to a function
using \code{\link[rlang:as_function]{rlang::as_function()}}.
This function is called with the unquoted table names of the \code{dm} object
as the only argument.
The output of this function is processed by \code{\link[DBI:dbQuoteIdentifier]{DBI::dbQuoteIdentifier()}},
that result should be a vector of identifiers of the same length
as the original table names.

Use a variant of
\code{table_names = ~ DBI::SQL(paste0("schema_name", ".", .x))}
to specify the same schema for all tables.
Use \code{table_names = identity} with \code{temporary = TRUE}
to avoid giving temporary tables unique names.

If a named character vector,
the names of this vector need to correspond to the table names in the \code{dm},
and its values are the desired names on \code{dest}.
The value is processed by \code{\link[DBI:dbQuoteIdentifier]{DBI::dbQuoteIdentifier()}},
that result should be a vector of identifiers of the same length
as the original table names.

Use qualified names corresponding to your database's syntax
to specify e.g. database and schema for your tables.}

\item{temporary}{If \code{TRUE}, only temporary tables will be created.
These tables will vanish when disconnecting from the database.}

\item{schema}{Name of schema to copy the \code{dm} to.
If \code{schema} is provided, an error will be thrown if \code{temporary = FALSE} or
\code{table_names} is not \code{NULL}.

Not all DBMS are supported.}

\item{progress}{Whether to display a progress bar, if \code{NA} (the default)
hide in non-interactive mode, show in interactive mode. Requires the
'progress' package.}

\item{unique_table_names, copy_to}{Must be \code{NULL}.}
}
\value{
A \code{dm} object on the given \code{src} with the same table names
as the input \code{dm}.
}
\description{
\code{copy_dm_to()} takes a \code{\link[dbplyr:src_dbi]{dbplyr::src_dbi}} object or a \code{\link[DBI:DBIConnection-class]{DBI::DBIConnection}} object as its first argument
and a \code{\link{dm}} object as its second argument.
The latter is copied to the former.
The default is to create temporary tables, set \code{temporary = FALSE} to create permanent tables.
Unless \code{set_key_constraints} is \code{FALSE}, primary key, foreign key, and unique constraints
are set, and indexes for foreign keys are created, on all databases.
}
\examples{
\dontshow{if (rlang::is_installed(c("RSQLite", "nycflights13", "dbplyr"))) withAutoprint(\{ # examplesIf}
con <- DBI::dbConnect(RSQLite::SQLite())

# Copy to temporary tables, unique table names by default:
temp_dm <- copy_dm_to(
  con,
  dm_nycflights13(),
  set_key_constraints = FALSE
)

# Persist, explicitly specify table names:
persistent_dm <- copy_dm_to(
  con,
  dm_nycflights13(),
  temporary = FALSE,
  table_names = ~ paste0("flights_", .x)
)
dbplyr::remote_name(persistent_dm$planes)

DBI::dbDisconnect(con)
\dontshow{\}) # examplesIf}
}
\concept{DB interaction functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flatten.R
\name{dm_flatten_to_tbl}
\alias{dm_flatten_to_tbl}
\title{Flatten a part of a \code{dm} into a wide table}
\usage{
dm_flatten_to_tbl(dm, .start, ..., .recursive = FALSE, .join = left_join)
}
\arguments{
\item{dm}{A \code{\link{dm}} object.}

\item{.start}{The table from which all outgoing foreign key relations are considered
when establishing a processing order for the joins.
An interesting choice could be
for example a fact table in a star schema.}

\item{...}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Unquoted names of the tables to be included in addition to the \code{.start} table.
The order of the tables here determines the order of the joins.
If the argument is empty, all tables that can be reached will be included.
\code{tidyselect} is supported, see \code{\link[dplyr:select]{dplyr::select()}} for details on the semantics.}

\item{.recursive}{Logical, defaults to \code{FALSE}. Should not only parent tables be joined to \code{.start}, but also their ancestors?}

\item{.join}{The type of join to be performed, see \code{\link[dplyr:join]{dplyr::join()}}.}
}
\value{
A single table that results from consecutively joining all affected tables to the \code{.start} table.
}
\description{
\code{dm_flatten_to_tbl()} gathers all information of interest in one place in a wide table.
It performs a disambiguation of column names and a cascade of joins.
}
\details{
With \code{...} left empty, this function will join together all the tables of your \code{\link{dm}}
object that can be reached from the \code{.start} table, in the direction of the foreign key relations
(pointing from the child tables to the parent tables), using the foreign key relations to
determine the argument \code{by} for the necessary joins.
The result is one table with unique column names.
Use the \code{...} argument if you would like to control which tables should be joined to the \code{.start} table.

Mind that calling \code{dm_flatten_to_tbl()} with \code{.join = right_join} and no table order determined in the \code{...} argument
will not lead to a well-defined result if two or more foreign tables are to be joined to \code{.start}.
The resulting
table would depend on the order the tables that are listed in the \code{dm}.
Therefore, trying this will result in a warning.

Since \code{.join = nest_join} does not make sense in this direction (LHS = child table, RHS = parent table: for valid key constraints
each nested column entry would be a tibble of one row), an error will be thrown if this method is chosen.

The difference between \code{.recursive = FALSE} and \code{.recursive = TRUE} is
the following (see the examples):
\itemize{
\item \code{.recursive = FALSE} allows only one level of hierarchy
(i.e., direct neighbors to table \code{.start}), while
\item \code{.recursive = TRUE} will go through all levels of hierarchy while joining.
}

Additionally, these functions differ from \code{dm_wrap_tbl()}, which always
returns a \code{dm} object.
}
\examples{
\dontshow{if (dm:::dm_has_financial()) withAutoprint(\{ # examplesIf}

dm_financial() \%>\%
  dm_select_tbl(-loans) \%>\%
  dm_flatten_to_tbl(.start = cards)

dm_financial() \%>\%
  dm_select_tbl(-loans) \%>\%
  dm_flatten_to_tbl(.start = cards, .recursive = TRUE)
\dontshow{\}) # examplesIf}
}
\seealso{
Other flattening functions:
\code{\link[=dm_flatten]{dm_flatten()}}
}
\concept{flattening functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyr.R
\name{tidyr_table_manipulation}
\alias{tidyr_table_manipulation}
\alias{unite.dm_zoomed}
\alias{unite.dm_keyed_tbl}
\alias{separate.dm_zoomed}
\alias{separate.dm_keyed_tbl}
\title{\pkg{tidyr} table manipulation methods for zoomed dm objects}
\usage{
\method{unite}{dm_zoomed}(data, col, ..., sep = "_", remove = TRUE, na.rm = FALSE)

\method{unite}{dm_keyed_tbl}(data, col, ..., sep = "_", remove = TRUE, na.rm = FALSE)

\method{separate}{dm_zoomed}(
  data,
  col,
  into,
  sep = "[^[:alnum:]]+",
  remove = TRUE,
  convert = FALSE,
  extra = "warn",
  fill = "warn",
  ...
)

\method{separate}{dm_keyed_tbl}(
  data,
  col,
  into,
  sep = "[^[:alnum:]]+",
  remove = TRUE,
  convert = FALSE,
  extra = "warn",
  fill = "warn",
  ...
)
}
\arguments{
\item{data}{object of class \code{dm_zoomed}}

\item{col}{The name of the new column, as a string or symbol.

This argument is passed by expression and supports
\link[rlang:topic-inject]{quasiquotation} (you can unquote strings
and symbols). The name is captured from the expression with
\code{\link[rlang:defusing-advanced]{rlang::ensym()}} (note that this kind of interface where
symbols do not represent actual objects is now discouraged in the
tidyverse; we support it here for backward compatibility).}

\item{...}{For \code{unite.dm_zoomed}: see \code{\link[tidyr:unite]{tidyr::unite()}}

For \code{separate.dm_zoomed}: see \code{\link[tidyr:separate]{tidyr::separate()}}}

\item{sep}{Separator to use between values.}

\item{remove}{If \code{TRUE}, remove input columns from output data frame.}

\item{na.rm}{If \code{TRUE}, missing values will be removed prior to uniting
each value.}

\item{into}{Names of new variables to create as character vector.
Use \code{NA} to omit the variable in the output.}

\item{convert}{If \code{TRUE}, will run \code{\link[=type.convert]{type.convert()}} with
\code{as.is = TRUE} on new columns. This is useful if the component
columns are integer, numeric or logical.

NB: this will cause string \code{"NA"}s to be converted to \code{NA}s.}

\item{extra}{If \code{sep} is a character vector, this controls what
happens when there are too many pieces. There are three valid options:
\itemize{
\item \code{"warn"} (the default): emit a warning and drop extra values.
\item \code{"drop"}: drop any extra values without a warning.
\item \code{"merge"}: only splits at most \code{length(into)} times
}}

\item{fill}{If \code{sep} is a character vector, this controls what
happens when there are not enough pieces. There are three valid options:
\itemize{
\item \code{"warn"} (the default): emit a warning and fill from the right
\item \code{"right"}: fill with missing values on the right
\item \code{"left"}: fill with missing values on the left
}}
}
\description{
Use these methods without the '.dm_zoomed' suffix (see examples).
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
zoom_united <- dm_nycflights13() \%>\%
  dm_zoom_to(flights) \%>\%
  select(year, month, day) \%>\%
  unite("month_day", month, day)
zoom_united
zoom_united \%>\%
  separate(month_day, c("month", "day"))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm-package.R
\docType{package}
\name{dm-package}
\alias{dm-package}
\title{dm: Relational Data Models}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides tools for working with multiple related tables, stored as data frames or in a relational database. Multiple tables (data and metadata) are stored in a compound object, which can then be manipulated with a pipe-friendly syntax.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://dm.cynkra.com/}
  \item \url{https://github.com/cynkra/dm}
  \item Report bugs at \url{https://github.com/cynkra/dm/issues}
}

}
\author{
\strong{Maintainer}: Kirill Müller \email{kirill@cynkra.com} (\href{https://orcid.org/0000-0002-1416-3412}{ORCID})

Authors:
\itemize{
  \item Kirill Müller \email{kirill@cynkra.com} (\href{https://orcid.org/0000-0002-1416-3412}{ORCID})
  \item Tobias Schieferdecker \email{tobias@cynkra.com}
  \item Darko Bergant \email{darko.bergant@gmail.com}
}

Other contributors:
\itemize{
  \item Antoine Fabri [contributor]
  \item Katharina Brunner [contributor]
  \item James Wondrasek [contributor]
  \item Indrajeet Patil (\href{https://orcid.org/0000-0003-1995-6531}{ORCID}) [contributor]
  \item Maëlle Salmon (\href{https://orcid.org/0000-0002-2815-0399}{ORCID}) [contributor]
  \item energie360° AG \email{info@energie360.ch} [funder]
  \item cynkra GmbH \email{mail@cynkra.com} (\href{https://ror.org/0335t7e62}{ROR}) [funder, copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_get_con.R
\name{dm_get_con}
\alias{dm_get_con}
\title{Get connection}
\usage{
dm_get_con(dm)
}
\arguments{
\item{dm}{A \code{dm} object.}
}
\value{
The \code{\link[DBI:DBIConnection-class]{DBI::DBIConnection}} object for a \code{dm} object.
}
\description{
\code{dm_get_con()} returns the DBI connection for a \code{dm} object.
This works only if the tables are stored on a database, otherwise an error
is thrown.
}
\details{
All lazy tables in a dm object must be stored on the same database server
and accessed through the same connection, because a large part
of the package's functionality relies on efficient joins.
}
\examples{
\dontshow{if (dm:::dm_has_financial()) withAutoprint(\{ # examplesIf}
dm_financial() \%>\%
  dm_get_con()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add-tbl.R
\name{dm_mutate_tbl}
\alias{dm_mutate_tbl}
\title{Update tables in a \code{\link{dm}}}
\usage{
dm_mutate_tbl(dm, ...)
}
\arguments{
\item{dm}{A \code{\link{dm}} object.}

\item{...}{One or more tables to update in the \code{dm}.
Must be named.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Updates one or more existing tables in a \code{\link{dm}}.
For now, the column names must be identical.
This restriction may be levied optionally in the future.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_mutate_tbl(flights = nycflights13::flights[1:3, ])
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=dm]{dm()}}, \code{\link[=dm_select_tbl]{dm_select_tbl()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/key-helpers.R
\name{check_key}
\alias{check_key}
\title{Check if column(s) can be used as keys}
\usage{
check_key(x, ..., .data = deprecated())
}
\arguments{
\item{x}{The data frame whose columns should be tested for key properties.}

\item{...}{The names of the columns to be checked, processed with
\code{\link[dplyr:select]{dplyr::select()}}. If omitted, all columns will be checked.}

\item{.data}{Deprecated.}
}
\value{
Returns \code{x}, invisibly, if the check is passed.
Otherwise an error is thrown and the reason for it is explained.
}
\description{
\code{check_key()} accepts a data frame and, optionally, columns.
It throws an error
if the specified columns are NOT a unique key of the data frame.
If the columns given in the ellipsis ARE a key, the data frame itself is returned silently, so that it can be used for piping.
}
\examples{
data <- tibble::tibble(a = c(1, 2, 1), b = c(1, 4, 1), c = c(5, 6, 7))
# this is failing:
try(check_key(data, a, b))

# this is passing:
check_key(data, a, c)
check_key(data)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-dm.R
\name{dm_filter}
\alias{dm_filter}
\title{Filtering}
\usage{
dm_filter(.dm, ...)
}
\arguments{
\item{.dm}{A \code{dm} object.}

\item{...}{Named logical predicates.
The names correspond to tables in the \code{dm} object.
The predicates are defined in terms of the variables in the corresponding table,
they are passed on to \code{\link[dplyr:filter]{dplyr::filter()}}.

Multiple conditions are combined with \code{&}.
Only the rows where the condition evaluates
to \code{TRUE} are kept.}
}
\value{
An updated \code{dm} object with filters executed across all tables.
}
\description{
Filtering a table of a \code{\link{dm}} object may affect other tables that are connected to it
directly or indirectly via foreign key relations.

\code{dm_filter()} can be used to define filter conditions for tables using syntax that is similar to \code{\link[dplyr:filter]{dplyr::filter()}}.
The filters work across related tables:
The resulting \code{dm} object only contains rows that are related
(directly or indirectly) to rows that remain after applying the filters
on all tables.
}
\details{
As of dm 1.0.0, these conditions are no longer stored in the \code{dm} object,
instead they are applied to all tables during the call to \code{dm_filter()}.
Calling \code{dm_apply_filters()} or \code{dm_apply_filters_to_tbl()} is no longer necessary.

Use \code{\link[=dm_zoom_to]{dm_zoom_to()}} and \code{\link[dplyr:filter]{dplyr::filter()}} to filter rows without affecting related tables.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nyc <- dm_nycflights13()
dm_nyc \%>\%
  dm_nrow()

dm_nyc_filtered <-
  dm_nycflights13() \%>\%
  dm_filter(airports = (name == "John F Kennedy Intl"))

dm_nyc_filtered \%>\%
  dm_nrow()

# If you want to keep only those rows in the parent tables
# whose primary key values appear as foreign key values in
# `flights`, you can set a `TRUE` filter in `flights`:
dm_nyc \%>\%
  dm_filter(flights = (1 == 1)) \%>\%
  dm_nrow()
# note that in this example, the only affected table is
# `airports` because the departure airports in `flights` are
# only the three New York airports.
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoom.R
\name{dm_zoom_to}
\alias{dm_zoom_to}
\alias{zoomed_df}
\alias{dm_zoomed_df}
\alias{dm_insert_zoomed}
\alias{dm_update_zoomed}
\alias{dm_discard_zoomed}
\title{Mark table for manipulation}
\usage{
dm_zoom_to(dm, table)

dm_insert_zoomed(dm, new_tbl_name = NULL, repair = "unique", quiet = FALSE)

dm_update_zoomed(dm)

dm_discard_zoomed(dm)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.}

\item{new_tbl_name}{Name of the new table.}

\item{repair}{Either a string or a function. If a string, it must be one of
\code{"check_unique"}, \code{"minimal"}, \code{"unique"}, \code{"universal"}, \code{"unique_quiet"},
or \code{"universal_quiet"}. If a function, it is invoked with a vector of
minimal names and must return minimal names, otherwise an error is thrown.
\itemize{
\item Minimal names are never \code{NULL} or \code{NA}. When an element doesn't
have a name, its minimal name is an empty string.
\item Unique names are unique. A suffix is appended to duplicate
names to make them unique.
\item Universal names are unique and syntactic, meaning that you can
safely use the names as variables without causing a syntax
error.
}

The \code{"check_unique"} option doesn't perform any name repair.
Instead, an error is raised if the names don't suit the
\code{"unique"} criteria.

The options \code{"unique_quiet"} and \code{"universal_quiet"} are here to help the
user who calls this function indirectly, via another function which exposes
\code{repair} but not \code{quiet}. Specifying \code{repair = "unique_quiet"} is like
specifying \verb{repair = "unique", quiet = TRUE}. When the \code{"*_quiet"} options
are used, any setting of \code{quiet} is silently overridden.}

\item{quiet}{By default, the user is informed of any renaming
caused by repairing the names. This only concerns unique and
universal repairing. Set \code{quiet} to \code{TRUE} to silence the
messages.

Users can silence the name repair messages by setting the
\code{"rlib_name_repair_verbosity"} global option to \code{"quiet"}.}
}
\value{
For \code{dm_zoom_to()}: A \code{dm_zoomed} object.

For \code{dm_insert_zoomed()}, \code{dm_update_zoomed()} and \code{dm_discard_zoomed()}: A \code{dm} object.
}
\description{
Zooming to a table of a \code{\link{dm}} allows for the use of many \code{dplyr}-verbs directly on this table, while retaining the
context of the \code{dm} object.

\code{dm_zoom_to()} zooms to the given table.

\code{dm_update_zoomed()} overwrites the originally zoomed table with the manipulated table.
The filter conditions for the zoomed table are added to the original filter conditions.

\code{dm_insert_zoomed()} adds a new table to the \code{dm}.

\code{dm_discard_zoomed()} discards the zoomed table and returns the \code{dm} as it was before zooming.

Please refer to \code{vignette("tech-db-zoom", package = "dm")}
for a more detailed introduction.
}
\details{
Whenever possible, the key relations of the original table are transferred to the resulting table
when using \code{dm_insert_zoomed()} or \code{dm_update_zoomed()}.

Functions from \code{dplyr} that are supported for a \code{dm_zoomed}: \code{\link[=group_by]{group_by()}}, \code{\link[=summarise]{summarise()}}, \code{\link[=mutate]{mutate()}},
\code{\link[=transmute]{transmute()}}, \code{\link[=filter]{filter()}}, \code{\link[=select]{select()}}, \code{\link[=rename]{rename()}} and \code{\link[=ungroup]{ungroup()}}.
You can use these functions just like you would
with a normal table.

Calling \code{\link[=filter]{filter()}} on a zoomed \code{dm} is different from calling \code{\link[=dm_filter]{dm_filter()}}:
only with the latter, the filter expression is added to the list of table filters stored in the dm.

Furthermore, different \code{join()}-variants from \pkg{dplyr} are also supported,
e.g. \code{\link[=left_join]{left_join()}} and \code{\link[=semi_join]{semi_join()}}.
(Support for \code{\link[dplyr:nest_join]{nest_join()}} is planned.)
The join-methods for \code{dm_zoomed} infer the columns to join by from the primary and foreign keys,
and have an extra argument \code{select} that allows choosing the columns of the RHS table.

And -- last but not least -- also the \pkg{tidyr}-functions \code{\link[=unite]{unite()}} and \code{\link[=separate]{separate()}} are supported for \code{dm_zoomed}.
}
\examples{
\dontshow{if (rlang::is_installed(c("nycflights13", "DiagrammeR"))) withAutoprint(\{ # examplesIf}
flights_zoomed <- dm_zoom_to(dm_nycflights13(), flights)

flights_zoomed

flights_zoomed_transformed <-
  flights_zoomed \%>\%
  mutate(am_pm_dep = ifelse(dep_time < 1200, "am", "pm")) \%>\%
  # `by`-argument of `left_join()` can be explicitly given
  # otherwise the key-relation is used
  left_join(airports) \%>\%
  select(year:dep_time, am_pm_dep, everything())

flights_zoomed_transformed

# replace table `flights` with the zoomed table
flights_zoomed_transformed \%>\%
  dm_update_zoomed()

# insert the zoomed table as a new table
flights_zoomed_transformed \%>\%
  dm_insert_zoomed("extended_flights") \%>\%
  dm_draw()

# discard the zoomed table
flights_zoomed_transformed \%>\%
  dm_discard_zoomed()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr.R
\name{dplyr_table_manipulation}
\alias{dplyr_table_manipulation}
\alias{filter.dm_zoomed}
\alias{filter.dm_keyed_tbl}
\alias{filter_out.dm_zoomed}
\alias{filter_out.dm_keyed_tbl}
\alias{mutate.dm_zoomed}
\alias{mutate.dm_keyed_tbl}
\alias{transmute.dm_zoomed}
\alias{transmute.dm_keyed_tbl}
\alias{select.dm_zoomed}
\alias{select.dm_keyed_tbl}
\alias{relocate.dm_zoomed}
\alias{relocate.dm_keyed_tbl}
\alias{rename.dm_zoomed}
\alias{rename.dm_keyed_tbl}
\alias{distinct.dm_zoomed}
\alias{distinct.dm_keyed_tbl}
\alias{arrange.dm_zoomed}
\alias{arrange.dm_keyed_tbl}
\alias{slice.dm_zoomed}
\alias{slice.dm_keyed_tbl}
\alias{group_by.dm_zoomed}
\alias{group_by.dm_keyed_tbl}
\alias{ungroup.dm_zoomed}
\alias{ungroup.dm_keyed_tbl}
\alias{summarise.dm_zoomed}
\alias{summarise.dm_keyed_tbl}
\alias{reframe.dm_zoomed}
\alias{reframe.dm_keyed_tbl}
\alias{count.dm_zoomed}
\alias{count.dm_keyed_tbl}
\alias{tally.dm_zoomed}
\alias{tally.dm_keyed_tbl}
\alias{pull.dm_zoomed}
\alias{compute.dm_zoomed}
\title{\pkg{dplyr} table manipulation methods for zoomed dm objects}
\usage{
\method{filter}{dm_zoomed}(.data, ..., .by = NULL, .preserve = FALSE)

\method{filter}{dm_keyed_tbl}(.data, ..., .by = NULL, .preserve = FALSE)

\method{filter_out}{dm_zoomed}(.data, ..., .by = NULL, .preserve = FALSE)

\method{filter_out}{dm_keyed_tbl}(.data, ..., .by = NULL, .preserve = FALSE)

\method{mutate}{dm_zoomed}(
  .data,
  ...,
  .by = NULL,
  .keep = c("all", "used", "unused", "none"),
  .before = NULL,
  .after = NULL
)

\method{mutate}{dm_keyed_tbl}(
  .data,
  ...,
  .by = NULL,
  .keep = c("all", "used", "unused", "none"),
  .before = NULL,
  .after = NULL
)

\method{transmute}{dm_zoomed}(.data, ...)

\method{transmute}{dm_keyed_tbl}(.data, ...)

\method{select}{dm_zoomed}(.data, ...)

\method{select}{dm_keyed_tbl}(.data, ...)

\method{relocate}{dm_zoomed}(.data, ..., .before = NULL, .after = NULL)

\method{relocate}{dm_keyed_tbl}(.data, ..., .before = NULL, .after = NULL)

\method{rename}{dm_zoomed}(.data, ...)

\method{rename}{dm_keyed_tbl}(.data, ...)

\method{distinct}{dm_zoomed}(.data, ..., .keep_all = FALSE)

\method{distinct}{dm_keyed_tbl}(.data, ..., .keep_all = FALSE)

\method{arrange}{dm_zoomed}(.data, ..., .by_group = FALSE, .locale = NULL)

\method{arrange}{dm_keyed_tbl}(.data, ..., .by_group = FALSE, .locale = NULL)

\method{slice}{dm_zoomed}(.data, ..., .by = NULL, .preserve = FALSE, .keep_pk = NULL)

\method{slice}{dm_keyed_tbl}(.data, ..., .by = NULL, .preserve = FALSE)

\method{group_by}{dm_zoomed}(.data, ..., .add = FALSE, .drop = group_by_drop_default(.data))

\method{group_by}{dm_keyed_tbl}(.data, ..., .add = FALSE, .drop = group_by_drop_default(.data))

\method{ungroup}{dm_zoomed}(x, ...)

\method{ungroup}{dm_keyed_tbl}(x, ...)

\method{summarise}{dm_zoomed}(.data, ..., .by = NULL, .groups = NULL)

\method{summarise}{dm_keyed_tbl}(.data, ..., .by = NULL, .groups = NULL)

\method{reframe}{dm_zoomed}(.data, ..., .by = NULL)

\method{reframe}{dm_keyed_tbl}(.data, ..., .by = NULL)

\method{count}{dm_zoomed}(
  x,
  ...,
  wt = NULL,
  sort = FALSE,
  name = NULL,
  .drop = group_by_drop_default(x)
)

\method{count}{dm_keyed_tbl}(
  x,
  ...,
  wt = NULL,
  sort = FALSE,
  name = NULL,
  .drop = group_by_drop_default(x)
)

\method{tally}{dm_zoomed}(x, wt = NULL, sort = FALSE, name = NULL)

\method{tally}{dm_keyed_tbl}(x, wt = NULL, sort = FALSE, name = NULL)

\method{pull}{dm_zoomed}(.data, var = -1, name = NULL, ...)

\method{compute}{dm_zoomed}(x, ...)
}
\arguments{
\item{.data}{object of class \code{dm_zoomed}}

\item{...}{see corresponding function in package \pkg{dplyr} or \pkg{tidyr}}

\item{.by}{<\code{\link[dplyr:dplyr_tidy_select]{tidy-select}}> Optionally, a selection of columns to
group by for just this operation, functioning as an alternative to \code{\link[dplyr:group_by]{group_by()}}. For
details and examples, see \link[dplyr:dplyr_by]{?dplyr_by}.}

\item{.preserve}{Relevant when the \code{.data} input is grouped. If \code{.preserve = FALSE} (the default), the grouping structure is recalculated based on the
resulting data, otherwise the grouping is kept as is.}

\item{.keep}{Control which columns from \code{.data} are retained in the output. Grouping
columns and columns created by \code{...} are always kept.
\itemize{
\item \code{"all"} retains all columns from \code{.data}. This is the default.
\item \code{"used"} retains only the columns used in \code{...} to create new
columns. This is useful for checking your work, as it displays inputs
and outputs side-by-side.
\item \code{"unused"} retains only the columns \emph{not} used in \code{...} to create new
columns. This is useful if you generate new columns, but no longer need
the columns used to generate them.
\item \code{"none"} doesn't retain any extra columns from \code{.data}. Only the grouping
variables and columns created by \code{...} are kept.
}}

\item{.before, .after}{<\code{\link[dplyr:dplyr_tidy_select]{tidy-select}}> Optionally, control where new columns
should appear (the default is to add to the right hand side). See
\code{\link[dplyr:relocate]{relocate()}} for more details.}

\item{.keep_all}{For \code{distinct.dm_zoomed()}: see \code{\link[dplyr:distinct]{dplyr::distinct()}}}

\item{.by_group}{If \code{TRUE}, will sort first by grouping variable. Applies to
grouped data frames only.}

\item{.locale}{The locale to sort character vectors in.
\itemize{
\item If \code{NULL}, the default, uses the \code{"C"} locale unless the deprecated
\code{dplyr.legacy_locale} global option escape hatch is active. See the
\link[dplyr]{dplyr-locale} help page for more details.
\item If a single string from \code{\link[stringi:stri_locale_list]{stringi::stri_locale_list()}} is supplied, then
this will be used as the locale to sort with. For example, \code{"en"} will
sort with the American English locale. This requires the stringi package.
\item If \code{"C"} is supplied, then character vectors will always be sorted in the
C locale. This does not require stringi and is often much faster than
supplying a locale identifier.
}

The C locale is not the same as English locales, such as \code{"en"},
particularly when it comes to data containing a mix of upper and lower case
letters. This is explained in more detail on the \link[dplyr:dplyr-locale]{locale}
help page under the \verb{Default locale} section.}

\item{.keep_pk}{For \code{slice.dm_zoomed}: Logical, if \code{TRUE}, the primary key will be retained during this transformation. If \code{FALSE}, it will be dropped.
By default, the value is \code{NULL}, which causes the function to issue a message in case a primary key is available for the zoomed table.
This argument is specific for the \code{slice.dm_zoomed()} method.}

\item{.add}{When \code{FALSE}, the default, \code{group_by()} will
override existing groups. To add to the existing groups, use
\code{.add = TRUE}.}

\item{.drop}{Drop groups formed by factor levels that don't appear in the
data? The default is \code{TRUE} except when \code{.data} has been previously
grouped with \code{.drop = FALSE}. See \code{\link[dplyr:group_by_drop_default]{group_by_drop_default()}} for details.}

\item{x}{For \code{ungroup.dm_zoomed}: object of class \code{dm_zoomed}}

\item{.groups}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Grouping structure of the
result.
\itemize{
\item \code{"drop_last"}: drops the last level of grouping. This was the
only supported option before version 1.0.0.
\item \code{"drop"}: All levels of grouping are dropped.
\item \code{"keep"}: Same grouping structure as \code{.data}.
\item \code{"rowwise"}: Each row is its own group.
}

When \code{.groups} is not specified, it is set to \code{"drop_last"} for a grouped
data frame, and \code{"keep"} for a rowwise data frame. In addition, a message
informs you of how the result will be grouped unless the result is
ungrouped, the option \code{"dplyr.summarise.inform"} is set to \code{FALSE}, or when
\code{summarise()} is called from a function in a package.}

\item{wt}{<\code{\link[rlang:args_data_masking]{data-masking}}> Frequency weights.
Can be \code{NULL} or a variable:
\itemize{
\item If \code{NULL} (the default), counts the number of rows in each group.
\item If a variable, computes \code{sum(wt)} for each group.
}}

\item{sort}{If \code{TRUE}, will show the largest groups at the top.}

\item{name}{The name of the new column in the output.

If omitted, it will default to \code{n}. If there's already a column called \code{n},
it will use \code{nn}. If there's a column called \code{n} and \code{nn}, it'll use
\code{nnn}, and so on, adding \code{n}s until it gets a new name.}

\item{var}{A variable specified as:
\itemize{
\item a literal variable name
\item a positive integer, giving the position counting from the left
\item a negative integer, giving the position counting from the right.
}

The default returns the last column (on the assumption that's the
column you've created most recently).

This argument is taken by expression and supports
\link[rlang:topic-inject]{quasiquotation} (you can unquote column
names and column locations).}
}
\description{
Use these methods without the '.dm_zoomed' suffix (see examples).
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
zoomed <- dm_nycflights13() \%>\%
  dm_zoom_to(flights) \%>\%
  arrange(desc(day)) \%>\%
  summarize(.by = month, avg_air_time = mean(air_time, na.rm = TRUE))
zoomed
dm_insert_zoomed(zoomed, new_tbl_name = "avg_air_time_per_month")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_wrap.R
\name{dm_unwrap_tbl}
\alias{dm_unwrap_tbl}
\title{Unwrap a single table dm}
\usage{
dm_unwrap_tbl(dm, ptype, progress = NA)
}
\arguments{
\item{dm}{A dm.}

\item{ptype}{A dm, only used to query names of primary and foreign keys.}

\item{progress}{Whether to display a progress bar, if \code{NA} (the default)
hide in non-interactive mode, show in interactive mode. Requires the
'progress' package.}
}
\value{
A dm.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{dm_unwrap_tbl()} unwraps all tables in a dm object so that the resulting dm
matches a given ptype dm.
It runs a sequence of \code{\link[=dm_unnest_tbl]{dm_unnest_tbl()}} and \code{\link[=dm_unpack_tbl]{dm_unpack_tbl()}} operations
on the dm.
}
\examples{

roundtrip <-
  dm_nycflights13() \%>\%
  dm_wrap_tbl(root = flights) \%>\%
  dm_unwrap_tbl(ptype = dm_ptype(dm_nycflights13()))
roundtrip

# The roundtrip has the same structure but fewer rows:
dm_nrow(dm_nycflights13())
dm_nrow(roundtrip)
}
\seealso{
\code{\link[=dm_wrap_tbl]{dm_wrap_tbl()}}, \code{\link[=dm_unnest_tbl]{dm_unnest_tbl()}},
\code{\link[=dm_examine_constraints]{dm_examine_constraints()}},
\code{\link[=dm_examine_cardinalities]{dm_examine_cardinalities()}},
\code{\link[=dm_ptype]{dm_ptype()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_get_tables.R
\name{dm_get_tables}
\alias{dm_get_tables}
\title{Get tables}
\usage{
dm_get_tables(x, ..., keyed = FALSE)
}
\arguments{
\item{x}{A \code{dm} object.}

\item{...}{These dots are for future extensions and must be empty.}

\item{keyed}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Set to \code{TRUE} to return objects of the internal class \code{"dm_keyed_tbl"}
that will contain information on primary and foreign keys
in the individual table objects.
This allows using dplyr workflows on those tables and later reconstruct them
into a \code{dm} object.
See \code{\link[=dm_deconstruct]{dm_deconstruct()}} for a function that generates corresponding code
for an existing dm object, and \code{vignette("tech-dm-keyed")} for details.}
}
\value{
A named list with the tables (data frames or lazy tables)
constituting the \code{dm}.
}
\description{
\code{dm_get_tables()} returns a named list of \pkg{dplyr} \link{tbl} objects
of a \code{dm} object.
}
\examples{
dm_nycflights13() \%>\%
  dm_get_tables()

dm_nycflights13() \%>\%
  dm_get_tables(keyed = TRUE)

dm_nycflights13() \%>\%
  dm_get_tables(keyed = TRUE) \%>\%
  new_dm()
}
\seealso{
\code{\link[=dm]{dm()}} and \code{\link[=new_dm]{new_dm()}} for constructing a \code{dm} object from tables.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema.R
\name{db_schema_list}
\alias{db_schema_list}
\title{List schemas on a database}
\usage{
db_schema_list(con, include_default = TRUE, ...)
}
\arguments{
\item{con}{An object of class \code{"src"} or \code{"DBIConnection"}.}

\item{include_default}{Boolean, if \code{TRUE} (default), also the default schema
on the database is included in the result}

\item{...}{Passed on to the individual methods.}
}
\value{
A tibble with the following columns:
\describe{
\item{\code{schema_name}}{the names of the schemas,}
\item{\code{schema_owner}}{the schema owner names.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{db_schema_list()} lists the available schemas on the database.
}
\details{
Methods are not available for all DBMS.

Additional arguments are:
\itemize{
\item \code{dbname}: supported for MSSQL. List schemas on a different database on the connected MSSQL-server;
default: database addressed by \code{con}.
}
}
\seealso{
Other schema handling functions:
\code{\link[=db_schema_create]{db_schema_create()}},
\code{\link[=db_schema_drop]{db_schema_drop()}},
\code{\link[=db_schema_exists]{db_schema_exists()}}
}
\concept{schema handling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dm_nest_tbl.R
\name{dm_pack_tbl}
\alias{dm_pack_tbl}
\title{dm_pack_tbl()}
\usage{
dm_pack_tbl(dm, parent_table, into = NULL)
}
\arguments{
\item{dm}{A dm.}

\item{parent_table}{A terminal table with one child table.}

\item{into}{The table to pack \code{parent_tables} into, optional as it can be guessed
from the foreign keys unambiguously but useful to be explicit.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{dm_pack_tbl()} converts a parent table to a packed column in its child
table.
The parent table should not have parent tables itself (i.e. it needs to be a
\emph{terminal parent table}).
}
\examples{
dm_packed <-
  dm_nycflights13() \%>\%
  dm_pack_tbl(planes)

dm_packed

dm_packed$flights

dm_packed$flights$planes
}
\seealso{
\code{\link[=dm_wrap_tbl]{dm_wrap_tbl()}}, \code{\link[=dm_unwrap_tbl]{dm_unwrap_tbl()}}, \code{\link[=dm_nest_tbl]{dm_nest_tbl()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R, R/filter-dm.R, R/flatten.R,
%   R/graph.R, R/validate.R, R/zzx-deprecated.R
\name{deprecated}
\alias{deprecated}
\alias{sql_schema_create}
\alias{sql_schema_drop}
\alias{sql_schema_exists}
\alias{sql_schema_list}
\alias{dm_apply_filters}
\alias{dm_apply_filters_to_tbl}
\alias{dm_get_filters}
\alias{dm_join_to_tbl}
\alias{dm_is_referenced}
\alias{dm_get_referencing_tables}
\alias{validate_dm}
\alias{check_if_subset}
\alias{check_cardinality}
\alias{cdm_get_src}
\alias{cdm_get_con}
\alias{cdm_get_tables}
\alias{cdm_get_filter}
\alias{cdm_add_tbl}
\alias{cdm_rm_tbl}
\alias{cdm_copy_to}
\alias{cdm_disambiguate_cols}
\alias{cdm_draw}
\alias{cdm_set_colors}
\alias{cdm_get_colors}
\alias{cdm_get_available_colors}
\alias{cdm_filter}
\alias{cdm_nrow}
\alias{cdm_flatten_to_tbl}
\alias{cdm_squash_to_tbl}
\alias{cdm_join_to_tbl}
\alias{cdm_apply_filters}
\alias{cdm_apply_filters_to_tbl}
\alias{cdm_add_pk}
\alias{cdm_add_fk}
\alias{cdm_has_fk}
\alias{cdm_get_fk}
\alias{cdm_get_all_fks}
\alias{cdm_rm_fk}
\alias{cdm_enum_fk_candidates}
\alias{cdm_is_referenced}
\alias{cdm_get_referencing_tables}
\alias{cdm_learn_from_db}
\alias{cdm_check_constraints}
\alias{cdm_nycflights13}
\alias{cdm_paste}
\alias{cdm_has_pk}
\alias{cdm_get_pk}
\alias{cdm_get_all_pks}
\alias{cdm_rm_pk}
\alias{cdm_enum_pk_candidates}
\alias{cdm_select_tbl}
\alias{cdm_rename_tbl}
\alias{cdm_select}
\alias{cdm_rename}
\alias{cdm_zoom_to_tbl}
\alias{cdm_insert_zoomed_tbl}
\alias{cdm_update_zoomed_tbl}
\alias{cdm_zoom_out}
\alias{dm_rm_tbl}
\alias{dm_add_tbl}
\alias{dm_bind}
\alias{dm_squash_to_tbl}
\alias{rows_truncate}
\alias{sql_rows_truncate}
\alias{dm_rows_truncate}
\title{Deprecated functions}
\usage{
sql_schema_create(dest, schema, ...)

sql_schema_drop(dest, schema, force = FALSE, ...)

sql_schema_exists(dest, schema, ...)

sql_schema_list(dest, include_default = TRUE, ...)

dm_apply_filters(dm)

dm_apply_filters_to_tbl(dm, table)

dm_get_filters(dm)

dm_join_to_tbl(dm, table_1, table_2, join = left_join)

dm_is_referenced(dm, table)

dm_get_referencing_tables(dm, table)

validate_dm(x)

check_if_subset(t1, c1, t2, c2)

check_cardinality(parent_table, pk_column, child_table, fk_column)

cdm_get_src(x)

cdm_get_con(x)

cdm_get_tables(x)

cdm_get_filter(x)

cdm_add_tbl(dm, ..., repair = "unique", quiet = FALSE)

cdm_rm_tbl(dm, ...)

cdm_copy_to(
  dest,
  dm,
  ...,
  types = NULL,
  overwrite = NULL,
  indexes = NULL,
  unique_indexes = NULL,
  set_key_constraints = TRUE,
  unique_table_names = FALSE,
  table_names = NULL,
  temporary = TRUE
)

cdm_disambiguate_cols(dm, sep = ".", quiet = FALSE)

cdm_draw(
  dm,
  rankdir = "LR",
  col_attr = "column",
  view_type = "keys_only",
  columnArrows = TRUE,
  graph_attrs = "",
  node_attrs = "",
  edge_attrs = "",
  focus = NULL,
  graph_name = "Data Model"
)

cdm_set_colors(dm, ...)

cdm_get_colors(dm)

cdm_get_available_colors()

cdm_filter(dm, table, ...)

cdm_nrow(dm)

cdm_flatten_to_tbl(dm, start, ..., join = left_join)

cdm_squash_to_tbl(dm, start, ..., join = left_join)

cdm_join_to_tbl(dm, table_1, table_2, join = left_join)

cdm_apply_filters(dm)

cdm_apply_filters_to_tbl(dm, table)

cdm_add_pk(dm, table, column, check = FALSE, force = FALSE)

cdm_add_fk(dm, table, column, ref_table, check = FALSE)

cdm_has_fk(dm, table, ref_table)

cdm_get_fk(dm, table, ref_table)

cdm_get_all_fks(dm)

cdm_rm_fk(dm, table, columns, ref_table)

cdm_enum_fk_candidates(dm, table, ref_table)

cdm_is_referenced(dm, table)

cdm_get_referencing_tables(dm, table)

cdm_learn_from_db(dest)

cdm_check_constraints(dm)

cdm_nycflights13(cycle = FALSE, color = TRUE, subset = TRUE)

cdm_paste(dm, select = FALSE, tab_width = 2)

cdm_has_pk(dm, table)

cdm_get_pk(dm, table)

cdm_get_all_pks(dm)

cdm_rm_pk(dm, table, rm_referencing_fks = FALSE)

cdm_enum_pk_candidates(dm, table)

cdm_select_tbl(dm, ...)

cdm_rename_tbl(dm, ...)

cdm_select(dm, table, ...)

cdm_rename(dm, table, ...)

cdm_zoom_to_tbl(dm, table)

cdm_insert_zoomed_tbl(
  dm,
  new_tbl_name = NULL,
  repair = "unique",
  quiet = FALSE
)

cdm_update_zoomed_tbl(dm)

cdm_zoom_out(dm)

dm_rm_tbl(dm, ...)

dm_add_tbl(dm, ..., repair = "unique", quiet = FALSE)

dm_bind(..., repair = "check_unique", quiet = FALSE)

dm_squash_to_tbl(dm, start, ..., join = left_join)

rows_truncate(x, ..., in_place = FALSE)

sql_rows_truncate(x, ...)

dm_rows_truncate(x, y, ..., in_place = NULL, progress = NA)
}
\arguments{
\item{...}{These dots are for future extensions and must be empty.}

\item{force}{Boolean, if \code{FALSE} (default), an error will be thrown if there is already a primary key
set for this table.
If \code{TRUE}, a potential old \code{pk} is deleted before setting a new one.}

\item{dm}{A \code{\link{dm}} object.}

\item{table}{A table in the \code{dm}.}

\item{table_1}{One of the tables involved in the join.}

\item{table_2}{The second table of the join.}

\item{join}{The type of join to be performed, see \code{\link[dplyr:join]{dplyr::join()}}.}

\item{x}{An object.}

\item{check}{Boolean, if \code{TRUE}, a check is made if the combination of columns is a unique key of the table.}

\item{columns}{Table columns, unquoted.
To define a compound key, use \code{c(col1, col2)}.}
}
\description{
These functions are deprecated in favor of better alternatives.
Most functions with the \code{cdm_} prefix have an identical alternative
with a \code{dm_} prefix.

\verb{sql_schema_*()} functions have been replaced with the corresponding
\verb{db_schema_*()} functions.

\code{dm_join_to_tbl()} is deprecated in favor of \code{\link[=dm_flatten_to_tbl]{dm_flatten_to_tbl()}}.

\code{dm_is_referenced()} is soft-deprecated, use the information returned from
\code{\link[=dm_get_all_fks]{dm_get_all_fks()}} instead.

\code{dm_get_referencing_tables()} is soft-deprecated, use the information
returned from \code{\link[=dm_get_all_fks]{dm_get_all_fks()}} instead.

\code{validate_dm()} has been replaced by \code{\link[=dm_validate]{dm_validate()}} for consistency.

\code{dm_add_tbl} is deprecated as of dm 1.0.0, because the same functionality
is offered by \code{\link[=dm]{dm()}} with \code{.name_repair = "unique"}.

\code{dm_bind()}  is deprecated as of dm 1.0.0, because the same functionality
is offered by \code{\link[=dm]{dm()}}.

\code{dm_squash_to_tbl()}  is deprecated as of dm 1.0.0, because the same functionality
is offered by \code{\link[=dm_flatten_to_tbl]{dm_flatten_to_tbl()}} with \code{.recursive = TRUE}.

\code{rows_truncate()} is deprecated as of dm 1.0.0, because it's a DDL operation
and requires different permissions than the
\code{\link[dplyr:rows_insert]{dplyr::rows_*()}} functions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema.R
\name{db_schema_exists}
\alias{db_schema_exists}
\title{Check for existence of a schema on a database}
\usage{
db_schema_exists(con, schema, ...)
}
\arguments{
\item{con}{An object of class \code{"src"} or \code{"DBIConnection"}.}

\item{schema}{Class \code{character} or \code{SQL}, name of the schema}

\item{...}{Passed on to the individual methods.}
}
\value{
A boolean: \code{TRUE} if schema exists, \code{FALSE} otherwise.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{db_schema_exists()} checks, if a schema exists on the database.
}
\details{
Methods are not available for all DBMS.

Additional arguments are:
\itemize{
\item \code{dbname}: supported for MSSQL. Check if a schema exists on a different
database on the connected MSSQL-server; default: database addressed by \code{con}.
}
}
\seealso{
Other schema handling functions:
\code{\link[=db_schema_create]{db_schema_create()}},
\code{\link[=db_schema_drop]{db_schema_drop()}},
\code{\link[=db_schema_list]{db_schema_list()}}
}
\concept{schema handling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reexport.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{\%>\%}
\alias{filter}
\alias{mutate}
\alias{transmute}
\alias{summarise}
\alias{summarize}
\alias{select}
\alias{rename}
\alias{group_by}
\alias{ungroup}
\alias{left_join}
\alias{inner_join}
\alias{full_join}
\alias{semi_join}
\alias{anti_join}
\alias{right_join}
\alias{collect}
\alias{compute}
\alias{copy_to}
\alias{src_tbls}
\alias{tbl}
\alias{arrange}
\alias{rows_insert}
\alias{rows_append}
\alias{rows_update}
\alias{rows_patch}
\alias{rows_upsert}
\alias{rows_delete}
\alias{glimpse}
\alias{tibble}
\alias{unite}
\alias{separate}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{dplyr}{\code{\link[dplyr:\%>\%]{\%>\%}}, \code{\link[dplyr:anti_join]{anti_join()}}, \code{\link[dplyr:arrange]{arrange()}}, \code{\link[dplyr:collect]{collect()}}, \code{\link[dplyr:compute]{compute()}}, \code{\link[dplyr:copy_to]{copy_to()}}, \code{\link[dplyr:filter]{filter()}}, \code{\link[dplyr:full_join]{full_join()}}, \code{\link[dplyr:group_by]{group_by()}}, \code{\link[dplyr:inner_join]{inner_join()}}, \code{\link[dplyr:left_join]{left_join()}}, \code{\link[dplyr:mutate]{mutate()}}, \code{\link[dplyr:rename]{rename()}}, \code{\link[dplyr:right_join]{right_join()}}, \code{\link[dplyr:rows_append]{rows_append()}}, \code{\link[dplyr:rows_delete]{rows_delete()}}, \code{\link[dplyr:rows_insert]{rows_insert()}}, \code{\link[dplyr:rows_patch]{rows_patch()}}, \code{\link[dplyr:rows_update]{rows_update()}}, \code{\link[dplyr:rows_upsert]{rows_upsert()}}, \code{\link[dplyr:select]{select()}}, \code{\link[dplyr:semi_join]{semi_join()}}, \code{\link[dplyr:src_tbls]{src_tbls()}}, \code{\link[dplyr:summarise]{summarise()}}, \code{\link[dplyr:summarize]{summarize()}}, \code{\link[dplyr:tbl]{tbl()}}, \code{\link[dplyr:transmute]{transmute()}}, \code{\link[dplyr:ungroup]{ungroup()}}}

  \item{tibble}{\code{\link[tibble:glimpse]{glimpse}}, \code{\link[tibble:tibble]{tibble()}}}

  \item{tidyr}{\code{\link[tidyr:separate]{separate()}}, \code{\link[tidyr:unite]{unite()}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/financial.R
\name{dm_financial}
\alias{dm_financial}
\alias{dm_financial_sqlite}
\title{Creates a dm object for the Financial data}
\usage{
dm_financial()

dm_financial_sqlite()
}
\value{
A \code{dm} object.
}
\description{
\code{dm_financial()} creates an example \code{\link{dm}} object from the tables at
https://relational.fel.cvut.cz/dataset/Financial.
The connection is established once per session,
subsequent calls return the same connection.

\code{dm_financial_sqlite()} copies the data to a temporary SQLite database.
The data is downloaded once per session, subsequent calls return the same database.
The \code{trans} table is excluded due to its size.
}
\examples{
\dontshow{if (dm:::dm_has_financial() && rlang::is_installed("DiagrammeR")) withAutoprint(\{ # examplesIf}
dm_financial() \%>\%
  dm_draw()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_unnest.R
\name{json_unnest}
\alias{json_unnest}
\title{Unnest a JSON column}
\usage{
json_unnest(data, cols, ..., names_sep = NULL, names_repair = "check_unique")
}
\arguments{
\item{data}{A data frame, a data frame extension (e.g. a tibble), or  a lazy
data frame (e.g. from dbplyr or dtplyr).}

\item{cols}{<\code{\link[tidyr:tidyr_tidy_select]{tidy-select}}> List-columns to unnest.

When selecting multiple columns, values from the same row will be recycled
to their common size.}

\item{...}{Arguments passed to methods.}

\item{names_sep}{If \code{NULL}, the default, the outer names will come from the
inner names. If a string, the outer names will be formed by pasting
together the outer and the inner column names, separated by \code{names_sep}.}

\item{names_repair}{Used to check that output data frame has valid
names. Must be one of the following options:
\itemize{
\item \verb{"minimal}": no name repair or checks, beyond basic existence,
\item \verb{"unique}": make sure names are unique and not empty,
\item \verb{"check_unique}": (the default), no name repair, but check they are unique,
\item \verb{"universal}": make the names unique and syntactic
\item a function: apply custom name repair.
\item \link[tidyr]{tidyr_legacy}: use the name repair from tidyr 0.8.
\item a formula: a purrr-style anonymous function (see \code{\link[rlang:as_function]{rlang::as_function()}})
}

See \code{\link[vctrs:vec_as_names]{vctrs::vec_as_names()}} for more details on these terms and the
strategies used to enforce them.}
}
\value{
An object of the same type as \code{data}
}
\description{
A wrapper around \code{\link[tidyr:unnest]{tidyr::unnest()}} that extracts its data from a JSON column.
The inverse of \code{\link[=json_nest]{json_nest()}}.
}
\examples{
tibble(a = 1, b = '[{ "c": 2 }, { "c": 3 }]') \%>\%
  json_unnest(b)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primary-keys.R
\name{dm_rm_pk}
\alias{dm_rm_pk}
\title{Remove a primary key}
\usage{
dm_rm_pk(dm, table = NULL, columns = NULL, ..., fail_fk = NULL)
}
\arguments{
\item{dm}{A \code{dm} object.}

\item{table}{A table in the \code{dm}.
Pass \code{NULL} to remove all matching keys.}

\item{columns}{Table columns, unquoted.
To refer to a compound key, use \code{c(col1, col2)}.
Pass \code{NULL} (the default) to remove all matching keys.}

\item{...}{These dots are for future extensions and must be empty.}

\item{fail_fk}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}}
}
\value{
An updated \code{dm} without the indicated primary key(s).
}
\description{
If a table name is provided, \code{dm_rm_pk()} removes the primary key from this table and leaves the \code{\link{dm}} object otherwise unaltered.
If no table is given, the \code{dm} is stripped of all primary keys at once.
An error is thrown if no primary key matches the selection criteria.
If the selection criteria are ambiguous, a message with unambiguous replacement code is shown.
Foreign keys are never removed.
}
\examples{
\dontshow{if (rlang::is_installed(c("nycflights13", "DiagrammeR"))) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_rm_pk(airports) \%>\%
  dm_draw()
\dontshow{\}) # examplesIf}
}
\seealso{
Other primary key functions:
\code{\link[=dm_add_pk]{dm_add_pk()}},
\code{\link[=dm_add_uk]{dm_add_uk()}},
\code{\link[=dm_get_all_pks]{dm_get_all_pks()}},
\code{\link[=dm_get_all_uks]{dm_get_all_uks()}},
\code{\link[=dm_has_pk]{dm_has_pk()}},
\code{\link[=dm_rm_uk]{dm_rm_uk()}},
\code{\link[=enum_pk_candidates]{enum_pk_candidates()}}
}
\concept{primary key functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-helpers.R
\name{dm_nrow}
\alias{dm_nrow}
\title{Number of rows}
\usage{
dm_nrow(dm)
}
\arguments{
\item{dm}{A \code{\link{dm}} object.}
}
\value{
A named vector with the number of rows for each table.
}
\description{
Returns a named vector with the number of rows for each table.
}
\examples{
\dontshow{if (rlang::is_installed("nycflights13")) withAutoprint(\{ # examplesIf}
dm_nycflights13() \%>\%
  dm_filter(airports = (faa \%in\% c("EWR", "LGA"))) \%>\%
  dm_nrow()
\dontshow{\}) # examplesIf}
}
